Package: tmap
Title: Thematic Maps
Version: 4.3
Authors@R: c(
    person("Martijn", "Tennekes", , "mtennekes@gmail.com", role = c("aut", "cre")),
    person("Jakub", "Nowosad", , "nowosad.jakub@gmail.com", role = "ctb"),
    person("Joel", "Gombin", , "joel.gombin@gmail.com", role = "ctb"),
    person("Sebastian", "Jeworutzki", , "Sebastian.Jeworutzki@rub.de", role = "ctb"),
    person("Kent", "Russell", , "kent.russell@timelyportfolio.com", role = "ctb"),
    person("Richard", "Zijdeman", , "richard.zijdeman@iisg.nl", role = "ctb"),
    person("John", "Clouse", , "johnaclouse@gmail.com", role = "ctb"),
    person("Robin", "Lovelace", , "rob00x@gmail.com", role = "ctb"),
    person("Jannes", "Muenchow", , "jannes.muenchow@uni-jena.de", role = "ctb"),
    person("Olivier", "Roy", role = "ctb"),
    person("Edzer", "Pebesma", , "edzer.pebesma@uni-muenster.de", role = "ctb"),
    person("Hugh", "Graham", , "hugh.a.graham@gmail.com", role = "ctb"),
    person("Michael D.", "Sumner", , "mdsumner@gmail.com", role = "ctb"),
    person("Tim", "Appelhans", , "tim.appelhans@gmail.com", role = "ctb"),
    person("Nick", "Bearman", , "nick@nickbearman.com", role = "ctb"),
    person("Pukar", "Bhandari", , "pukar.bhandari@outlook.com", role = "ctb"),
    person("Stéphane", "Guillou", , "stephane.guillou@member.fsf.org", role = "ctb"),
    person("Monika Anna", "Tomaszewska", , "monika.anna.tomaszewska@gmail.com", role = "ctb"),
    person("Ajoke", "Onojeghuo", , "a.onojeghuo@gmail.com", role = "ctb"),
    person("Jerome", "Guelat", , "jerome.guelat@vogelwarte.ch", role = "ctb"))
Description: Thematic maps are geographical maps in which spatial data
    distributions are visualized. This package offers a flexible,
    layer-based, and easy to use approach to create thematic maps, such as
    choropleths and bubble maps.
License: GPL-3
URL: https://github.com/r-tmap/tmap, https://r-tmap.github.io/tmap/
BugReports: https://github.com/r-tmap/tmap/issues
Depends: R (>= 4.1)
Imports: classInt (>= 0.4-3), cli, cols4all (>= 0.8-1), data.table,
        grid, htmltools, htmlwidgets, base64enc, leafem (>= 0.2.4),
        leafgl, leaflegend, leaflet (>= 2.0.2), leafsync, maptiles,
        methods, rlang, sf (>= 0.9-3), stars (>= 0.4-2), stats, s2,
        tmaptools (>= 3.1), units (>= 0.6-1), servr
Suggests: av, transformr, colorspace, ggplot2, gifski, knitr, shiny,
        terra, testthat (>= 3.2.0), widgetframe, lobstr, rsvg,
        rstudioapi
Config/Needs/check: Nowosad/spDataLarge, lwgeom, r-tmap/tmaptools,
        r-tmap/tmap
Config/Needs/coverage: Nowosad/spDataLarge, lwgeom, r-tmap/tmaptools,
        r-tmap/tmap
Config/Needs/website: bookdown, rmarkdown, sfnetworks,
        r-tmap/tmaptools, r-tmap/tmap.glyphs, r-tmap/tmap.networks,
        walkerke/mapgl, r-tmap/tmap.mapgl, geofacet, sits,
        r-tmap/tmap.cartogram, mapsf, mapview
Config/testthat/edition: 3
Encoding: UTF-8
LazyData: true
RoxygenNote: 7.3.3
NeedsCompilation: no
Packaged: 2026-04-14 07:55:02 UTC; mtes
Author: Martijn Tennekes [aut, cre],
  Jakub Nowosad [ctb],
  Joel Gombin [ctb],
  Sebastian Jeworutzki [ctb],
  Kent Russell [ctb],
  Richard Zijdeman [ctb],
  John Clouse [ctb],
  Robin Lovelace [ctb],
  Jannes Muenchow [ctb],
  Olivier Roy [ctb],
  Edzer Pebesma [ctb],
  Hugh Graham [ctb],
  Michael D. Sumner [ctb],
  Tim Appelhans [ctb],
  Nick Bearman [ctb],
  Pukar Bhandari [ctb],
  Stéphane Guillou [ctb],
  Monika Anna Tomaszewska [ctb],
  Ajoke Onojeghuo [ctb],
  Jerome Guelat [ctb]
Maintainer: Martijn Tennekes <mtennekes@gmail.com>
Repository: CRAN
Date/Publication: 2026-04-20 08:20:02 UTC
Built: R 4.5.2; ; 2026-04-20 11:06:52 UTC; unix

Package: tmap
Title: Thematic Maps
Version: 4.4
Authors@R: c(
    person("Martijn", "Tennekes", , "mtennekes@gmail.com", role = c("aut", "cre")),
    person("Jakub", "Nowosad", , "nowosad.jakub@gmail.com", role = "ctb"),
    person("Joel", "Gombin", , "joel.gombin@gmail.com", role = "ctb"),
    person("Sebastian", "Jeworutzki", , "Sebastian.Jeworutzki@rub.de", role = "ctb"),
    person("Kent", "Russell", , "kent.russell@timelyportfolio.com", role = "ctb"),
    person("Richard", "Zijdeman", , "richard.zijdeman@iisg.nl", role = "ctb"),
    person("John", "Clouse", , "johnaclouse@gmail.com", role = "ctb"),
    person("Robin", "Lovelace", , "rob00x@gmail.com", role = "ctb"),
    person("Jannes", "Muenchow", , "jannes.muenchow@uni-jena.de", role = "ctb"),
    person("Olivier", "Roy", role = "ctb"),
    person("Edzer", "Pebesma", , "edzer.pebesma@uni-muenster.de", role = "ctb"),
    person("Hugh", "Graham", , "hugh.a.graham@gmail.com", role = "ctb"),
    person("Michael D.", "Sumner", , "mdsumner@gmail.com", role = "ctb"),
    person("Tim", "Appelhans", , "tim.appelhans@gmail.com", role = "ctb"),
    person("Nick", "Bearman", , "nick@nickbearman.com", role = "ctb"),
    person("Pukar", "Bhandari", , "pukar.bhandari@outlook.com", role = "ctb"),
    person("Stéphane", "Guillou", , "stephane.guillou@member.fsf.org", role = "ctb"),
    person("Monika Anna", "Tomaszewska", , "monika.anna.tomaszewska@gmail.com", role = "ctb"),
    person("Ajoke", "Onojeghuo", , "a.onojeghuo@gmail.com", role = "ctb"),
    person("Jerome", "Guelat", , "jerome.guelat@vogelwarte.ch", role = "ctb"),
    person("Artur", "Wroblewski", , "wrobell@riseup.net", role = "ctb"))
Description: Thematic maps are geographical maps in which spatial data
    distributions are visualized. This package offers a flexible,
    layer-based, and easy to use approach to create thematic maps, such as
    choropleths and bubble maps.
License: GPL-3
URL: https://github.com/r-tmap/tmap, https://r-tmap.github.io/tmap/
BugReports: https://github.com/r-tmap/tmap/issues
Depends: R (>= 4.1)
Imports: classInt (>= 0.4-3), cli, cols4all (>= 0.10), data.table,
        grid, htmltools, htmlwidgets, base64enc, leafem (>= 0.2.4),
        leafgl, leaflegend, leaflet (>= 2.0.2), leafsync, maptiles,
        methods, rlang, sf (>= 0.9-3), stars (>= 0.4-2), stats, s2,
        tmaptools (>= 3.1), units (>= 0.6-1), servr
Suggests: av, transformr, colorspace, ggplot2, gifski, knitr, shiny,
        terra, testthat (>= 3.2.0), widgetframe, lobstr, rsvg,
        rstudioapi
Config/Needs/check: Nowosad/spDataLarge, lwgeom, r-tmap/tmaptools,
        r-tmap/tmap
Config/Needs/coverage: Nowosad/spDataLarge, lwgeom, r-tmap/tmaptools,
        r-tmap/tmap
Config/Needs/website: bookdown, rmarkdown, sfnetworks,
        r-tmap/tmaptools, r-tmap/tmap.glyphs, r-tmap/tmap.networks,
        walkerke/mapgl, r-tmap/tmap.mapgl, geofacet, sits,
        r-tmap/tmap.cartogram, mapsf, mapview
Config/testthat/edition: 3
Encoding: UTF-8
LazyData: true
Config/roxygen2/version: 8.0.0
NeedsCompilation: no
Packaged: 2026-06-13 10:18:46 UTC; mtes
Author: Martijn Tennekes [aut, cre],
  Jakub Nowosad [ctb],
  Joel Gombin [ctb],
  Sebastian Jeworutzki [ctb],
  Kent Russell [ctb],
  Richard Zijdeman [ctb],
  John Clouse [ctb],
  Robin Lovelace [ctb],
  Jannes Muenchow [ctb],
  Olivier Roy [ctb],
  Edzer Pebesma [ctb],
  Hugh Graham [ctb],
  Michael D. Sumner [ctb],
  Tim Appelhans [ctb],
  Nick Bearman [ctb],
  Pukar Bhandari [ctb],
  Stéphane Guillou [ctb],
  Monika Anna Tomaszewska [ctb],
  Ajoke Onojeghuo [ctb],
  Jerome Guelat [ctb],
  Artur Wroblewski [ctb]
Maintainer: Martijn Tennekes <mtennekes@gmail.com>
Repository: CRAN
Date/Publication: 2026-06-14 06:20:02 UTC

# Generated by roxygen2: do not edit by hand

S3method("+",tmap)
S3method(knitr::knit_print,tmap)
S3method(knitr::knit_print,tmap_arrange)
S3method(print,tm_element)
S3method(print,tm_shape)
S3method(print,tmap)
S3method(print,tmap_arrange)
S3method(tmapGetShapeMeta1,Raster)
S3method(tmapGetShapeMeta1,SpatRaster)
S3method(tmapGetShapeMeta1,SpatVector)
S3method(tmapGetShapeMeta1,Spatial)
S3method(tmapGetShapeMeta1,default)
S3method(tmapGetShapeMeta1,sf)
S3method(tmapGetShapeMeta1,sfc)
S3method(tmapGetShapeMeta1,stars)
S3method(tmapGetShapeMeta2,SpatRaster)
S3method(tmapGetShapeMeta2,SpatVector)
S3method(tmapGetShapeMeta2,default)
S3method(tmapGetShapeMeta2,sf)
S3method(tmapGetShapeMeta2,sfc)
S3method(tmapGetShapeMeta2,stars)
S3method(tmapGridAuxPlot,default)
S3method(tmapGridAuxPlot,tm_aux_basemap)
S3method(tmapGridAuxPlot,tm_aux_graticules)
S3method(tmapGridAuxPlot,tm_aux_grid)
S3method(tmapGridAuxPlot,tm_aux_tiles)
S3method(tmapGridAuxPrepare,default)
S3method(tmapGridAuxPrepare,tm_aux_basemap)
S3method(tmapGridAuxPrepare,tm_aux_graticules)
S3method(tmapGridAuxPrepare,tm_aux_grid)
S3method(tmapGridAuxPrepare,tm_aux_tiles)
S3method(tmapGridCompHeight,tm_chart)
S3method(tmapGridCompHeight,tm_compass)
S3method(tmapGridCompHeight,tm_credits)
S3method(tmapGridCompHeight,tm_inset_grob)
S3method(tmapGridCompHeight,tm_inset_map)
S3method(tmapGridCompHeight,tm_legend_landscape)
S3method(tmapGridCompHeight,tm_legend_portrait)
S3method(tmapGridCompHeight,tm_logo)
S3method(tmapGridCompHeight,tm_minimap)
S3method(tmapGridCompHeight,tm_scalebar)
S3method(tmapGridCompHeight,tm_title)
S3method(tmapGridCompPlot,tm_chart_bar)
S3method(tmapGridCompPlot,tm_chart_box)
S3method(tmapGridCompPlot,tm_chart_donut)
S3method(tmapGridCompPlot,tm_chart_heatmap)
S3method(tmapGridCompPlot,tm_chart_histogram)
S3method(tmapGridCompPlot,tm_chart_violin)
S3method(tmapGridCompPlot,tm_compass)
S3method(tmapGridCompPlot,tm_credits)
S3method(tmapGridCompPlot,tm_inset_grob)
S3method(tmapGridCompPlot,tm_inset_map)
S3method(tmapGridCompPlot,tm_legend_landscape)
S3method(tmapGridCompPlot,tm_legend_portrait)
S3method(tmapGridCompPlot,tm_logo)
S3method(tmapGridCompPlot,tm_minimap)
S3method(tmapGridCompPlot,tm_scalebar)
S3method(tmapGridCompPlot,tm_title)
S3method(tmapGridCompPrepare,default)
S3method(tmapGridCompPrepare,tm_chart)
S3method(tmapGridCompPrepare,tm_chart_bar)
S3method(tmapGridCompPrepare,tm_chart_histogram)
S3method(tmapGridCompPrepare,tm_compass)
S3method(tmapGridCompPrepare,tm_credits)
S3method(tmapGridCompPrepare,tm_inset_gg)
S3method(tmapGridCompPrepare,tm_inset_grob)
S3method(tmapGridCompPrepare,tm_inset_image)
S3method(tmapGridCompPrepare,tm_inset_map)
S3method(tmapGridCompPrepare,tm_inset_tmap)
S3method(tmapGridCompPrepare,tm_legend_landscape)
S3method(tmapGridCompPrepare,tm_legend_portrait)
S3method(tmapGridCompPrepare,tm_logo)
S3method(tmapGridCompPrepare,tm_minimap)
S3method(tmapGridCompPrepare,tm_scalebar)
S3method(tmapGridCompPrepare,tm_title)
S3method(tmapGridCompWidth,tm_chart)
S3method(tmapGridCompWidth,tm_compass)
S3method(tmapGridCompWidth,tm_credits)
S3method(tmapGridCompWidth,tm_inset_grob)
S3method(tmapGridCompWidth,tm_inset_map)
S3method(tmapGridCompWidth,tm_legend_landscape)
S3method(tmapGridCompWidth,tm_legend_portrait)
S3method(tmapGridCompWidth,tm_logo)
S3method(tmapGridCompWidth,tm_minimap)
S3method(tmapGridCompWidth,tm_scalebar)
S3method(tmapGridCompWidth,tm_title)
S3method(tmapGridDataPlot,default)
S3method(tmapGridDataPlot,tm_data_borders)
S3method(tmapGridDataPlot,tm_data_bubbles)
S3method(tmapGridDataPlot,tm_data_dots)
S3method(tmapGridDataPlot,tm_data_fill)
S3method(tmapGridDataPlot,tm_data_iso)
S3method(tmapGridDataPlot,tm_data_labels)
S3method(tmapGridDataPlot,tm_data_labels_highlighted)
S3method(tmapGridDataPlot,tm_data_lines)
S3method(tmapGridDataPlot,tm_data_markers)
S3method(tmapGridDataPlot,tm_data_polygons)
S3method(tmapGridDataPlot,tm_data_raster)
S3method(tmapGridDataPlot,tm_data_squares)
S3method(tmapGridDataPlot,tm_data_symbols)
S3method(tmapGridDataPlot,tm_data_text)
S3method(tmapLeafletAuxPlot,tm_aux_basemap)
S3method(tmapLeafletAuxPlot,tm_aux_graticules)
S3method(tmapLeafletAuxPlot,tm_aux_grid)
S3method(tmapLeafletAuxPlot,tm_aux_tiles)
S3method(tmapLeafletAuxPrepare,tm_aux_basemap)
S3method(tmapLeafletAuxPrepare,tm_aux_graticules)
S3method(tmapLeafletAuxPrepare,tm_aux_grid)
S3method(tmapLeafletAuxPrepare,tm_aux_tiles)
S3method(tmapLeafletCompHeight,tm_compass)
S3method(tmapLeafletCompHeight,tm_credits)
S3method(tmapLeafletCompHeight,tm_legend_landscape)
S3method(tmapLeafletCompHeight,tm_legend_portrait)
S3method(tmapLeafletCompHeight,tm_logo)
S3method(tmapLeafletCompHeight,tm_minimap)
S3method(tmapLeafletCompHeight,tm_mouse_coordinates)
S3method(tmapLeafletCompHeight,tm_scalebar)
S3method(tmapLeafletCompHeight,tm_title)
S3method(tmapLeafletCompPlot,tm_compass)
S3method(tmapLeafletCompPlot,tm_credits)
S3method(tmapLeafletCompPlot,tm_legend_landscape)
S3method(tmapLeafletCompPlot,tm_legend_portrait)
S3method(tmapLeafletCompPlot,tm_logo)
S3method(tmapLeafletCompPlot,tm_minimap)
S3method(tmapLeafletCompPlot,tm_mouse_coordinates)
S3method(tmapLeafletCompPlot,tm_scalebar)
S3method(tmapLeafletCompPlot,tm_title)
S3method(tmapLeafletCompPrepare,default)
S3method(tmapLeafletCompPrepare,tm_chart_none)
S3method(tmapLeafletCompPrepare,tm_compass)
S3method(tmapLeafletCompPrepare,tm_credits)
S3method(tmapLeafletCompPrepare,tm_inset_gg)
S3method(tmapLeafletCompPrepare,tm_inset_grob)
S3method(tmapLeafletCompPrepare,tm_inset_image)
S3method(tmapLeafletCompPrepare,tm_legend_landscape)
S3method(tmapLeafletCompPrepare,tm_legend_portrait)
S3method(tmapLeafletCompPrepare,tm_logo)
S3method(tmapLeafletCompPrepare,tm_minimap)
S3method(tmapLeafletCompPrepare,tm_mouse_coordinates)
S3method(tmapLeafletCompPrepare,tm_scalebar)
S3method(tmapLeafletCompPrepare,tm_title)
S3method(tmapLeafletCompWidth,tm_compass)
S3method(tmapLeafletCompWidth,tm_credits)
S3method(tmapLeafletCompWidth,tm_legend_landscape)
S3method(tmapLeafletCompWidth,tm_legend_portrait)
S3method(tmapLeafletCompWidth,tm_logo)
S3method(tmapLeafletCompWidth,tm_minimap)
S3method(tmapLeafletCompWidth,tm_mouse_coordinates)
S3method(tmapLeafletCompWidth,tm_scalebar)
S3method(tmapLeafletCompWidth,tm_title)
S3method(tmapLeafletDataPlot,default)
S3method(tmapLeafletDataPlot,tm_data_borders)
S3method(tmapLeafletDataPlot,tm_data_bubbles)
S3method(tmapLeafletDataPlot,tm_data_dots)
S3method(tmapLeafletDataPlot,tm_data_fill)
S3method(tmapLeafletDataPlot,tm_data_iso)
S3method(tmapLeafletDataPlot,tm_data_labels)
S3method(tmapLeafletDataPlot,tm_data_labels_highlighted)
S3method(tmapLeafletDataPlot,tm_data_lines)
S3method(tmapLeafletDataPlot,tm_data_markers)
S3method(tmapLeafletDataPlot,tm_data_polygons)
S3method(tmapLeafletDataPlot,tm_data_raster)
S3method(tmapLeafletDataPlot,tm_data_squares)
S3method(tmapLeafletDataPlot,tm_data_symbols)
S3method(tmapLeafletDataPlot,tm_data_text)
S3method(tmapReproject,dimensions)
S3method(tmapReproject,sfc)
S3method(tmapReproject,stars)
S3method(tmapShape,Raster)
S3method(tmapShape,SpatRaster)
S3method(tmapShape,SpatVector)
S3method(tmapShape,Spatial)
S3method(tmapShape,default)
S3method(tmapShape,sf)
S3method(tmapShape,stars)
S3method(tmapSplitShp,default)
S3method(tmapSplitShp,stars)
S3method(tmapSubsetShp,Raster)
S3method(tmapSubsetShp,SpatRaster)
S3method(tmapSubsetShp,SpatVector)
S3method(tmapSubsetShp,Spatial)
S3method(tmapSubsetShp,default)
S3method(tmapSubsetShp,sf)
S3method(tmapSubsetShp,sfc)
S3method(tmapSubsetShp,stars)
export(.TMAP)
export(.TMAP_GRID)
export(.TMAP_LEAFLET)
export(.tmap_providers)
export(chart_save)
export(data_class)
export(data_type)
export(format_aes_results)
export(get_fact_levels_na)
export(get_scale_defaults)
export(lwd_to_mm)
export(make_by_vars)
export(marker_icon)
export(opt_tm_bubbles)
export(opt_tm_dots)
export(opt_tm_labels)
export(opt_tm_lines)
export(opt_tm_markers)
export(opt_tm_polygons)
export(opt_tm_raster)
export(opt_tm_rgb)
export(opt_tm_sf)
export(opt_tm_squares)
export(opt_tm_symbols)
export(opt_tm_text)
export(qtm)
export(renderTmap)
export(rtm)
export(rtmp)
export(shapeTM)
export(theme_ps)
export(tm_add_legend)
export(tm_animate)
export(tm_animate_fast)
export(tm_basemap)
export(tm_borders)
export(tm_bubbles)
export(tm_chart_bar)
export(tm_chart_box)
export(tm_chart_donut)
export(tm_chart_heatmap)
export(tm_chart_histogram)
export(tm_chart_none)
export(tm_chart_violin)
export(tm_check_fix)
export(tm_compass)
export(tm_components)
export(tm_const)
export(tm_credits)
export(tm_crs)
export(tm_dots)
export(tm_element)
export(tm_element_list)
export(tm_extra_inner_margin)
export(tm_facets)
export(tm_facets_flip)
export(tm_facets_grid)
export(tm_facets_hstack)
export(tm_facets_pagewise)
export(tm_facets_stack)
export(tm_facets_vstack)
export(tm_facets_wrap)
export(tm_fill)
export(tm_format)
export(tm_graticules)
export(tm_grid)
export(tm_group)
export(tm_inset)
export(tm_iso)
export(tm_label_format)
export(tm_labels)
export(tm_labels_highlighted)
export(tm_layout)
export(tm_legend)
export(tm_legend_bivariate)
export(tm_legend_combine)
export(tm_legend_hide)
export(tm_lines)
export(tm_logo)
export(tm_markers)
export(tm_minimap)
export(tm_mouse_coordinates)
export(tm_options)
export(tm_place_legends_bottom)
export(tm_place_legends_inside)
export(tm_place_legends_left)
export(tm_place_legends_right)
export(tm_place_legends_top)
export(tm_plot)
export(tm_plot_order)
export(tm_polygons)
export(tm_pos)
export(tm_pos_auto_in)
export(tm_pos_auto_out)
export(tm_pos_in)
export(tm_pos_on_top)
export(tm_pos_out)
export(tm_raster)
export(tm_remove_layer)
export(tm_rgb)
export(tm_rgba)
export(tm_scale)
export(tm_scale_asis)
export(tm_scale_bivariate)
export(tm_scale_categorical)
export(tm_scale_continuous)
export(tm_scale_continuous_log)
export(tm_scale_continuous_log10)
export(tm_scale_continuous_log1p)
export(tm_scale_continuous_log2)
export(tm_scale_continuous_pseudo_log)
export(tm_scale_continuous_sqrt)
export(tm_scale_discrete)
export(tm_scale_intervals)
export(tm_scale_ordinal)
export(tm_scale_rank)
export(tm_scale_rgb)
export(tm_scale_rgba)
export(tm_scalebar)
export(tm_seq)
export(tm_sf)
export(tm_shape)
export(tm_squares)
export(tm_style)
export(tm_symbols)
export(tm_text)
export(tm_tiles)
export(tm_title)
export(tm_title_in)
export(tm_title_out)
export(tm_vars)
export(tm_view)
export(tm_xlab)
export(tm_ylab)
export(tmapAddLayerOptions)
export(tmapChartBinned)
export(tmapChartBinned2d)
export(tmapChartBinned2d_catcat)
export(tmapChartBinned2d_numcat)
export(tmapChartBinned2d_numnum)
export(tmapChartBinned_categorical)
export(tmapChartBinned_numeric)
export(tmapChartNone)
export(tmapChartPass)
export(tmapChartRaw)
export(tmapChartRaw_nna)
export(tmapGetCompGroupArgs)
export(tmapGetShapeMeta1)
export(tmapGetShapeMeta2)
export(tmapGpar)
export(tmapGridAuxPlot)
export(tmapGridAuxPrepare)
export(tmapGridCompHeight)
export(tmapGridCompPlot)
export(tmapGridCompPrepare)
export(tmapGridCompWidth)
export(tmapGridDataPlot)
export(tmapLeafletAuxPlot)
export(tmapLeafletAuxPrepare)
export(tmapLeafletCompPrepare)
export(tmapLeafletDataPlot)
export(tmapMode)
export(tmapOutput)
export(tmapProxy)
export(tmapReproject)
export(tmapScale)
export(tmapScaleAsIs)
export(tmapScaleAuto)
export(tmapScaleBivariate)
export(tmapScaleCategorical)
export(tmapScaleIntervals)
export(tmapScaleRank)
export(tmapSeq)
export(tmapShape)
export(tmapSplitShp)
export(tmapSubmitOptions)
export(tmapSubsetShp)
export(tmapTpar)
export(tmapTransCentroid)
export(tmapTransLines)
export(tmapTransPolygons)
export(tmapTransRaster)
export(tmapUsrCls)
export(tmapValuesBVV_bgcol)
export(tmapValuesBVV_col)
export(tmapValuesBVV_fill)
export(tmapValuesCVV_angle)
export(tmapValuesCVV_area)
export(tmapValuesCVV_bgcol)
export(tmapValuesCVV_bgcol_alpha)
export(tmapValuesCVV_col)
export(tmapValuesCVV_col_alpha)
export(tmapValuesCVV_fill)
export(tmapValuesCVV_fill_alpha)
export(tmapValuesCVV_fontface)
export(tmapValuesCVV_lty)
export(tmapValuesCVV_lwd)
export(tmapValuesCVV_num)
export(tmapValuesCVV_shape)
export(tmapValuesCVV_size)
export(tmapValuesCVV_skip)
export(tmapValuesCVV_text)
export(tmapValuesCVV_xmod)
export(tmapValuesCVV_ymod)
export(tmapValuesCheck_angle)
export(tmapValuesCheck_area)
export(tmapValuesCheck_bgcol)
export(tmapValuesCheck_bgcol_alpha)
export(tmapValuesCheck_col)
export(tmapValuesCheck_col_alpha)
export(tmapValuesCheck_fill)
export(tmapValuesCheck_fill_alpha)
export(tmapValuesCheck_fontface)
export(tmapValuesCheck_lty)
export(tmapValuesCheck_lwd)
export(tmapValuesCheck_num)
export(tmapValuesCheck_shape)
export(tmapValuesCheck_size)
export(tmapValuesCheck_skip)
export(tmapValuesCheck_text)
export(tmapValuesCheck_xmod)
export(tmapValuesColorize_angle)
export(tmapValuesColorize_area)
export(tmapValuesColorize_bgcol)
export(tmapValuesColorize_bgcol_alpha)
export(tmapValuesColorize_col)
export(tmapValuesColorize_col_alpha)
export(tmapValuesColorize_fill)
export(tmapValuesColorize_fill_alpha)
export(tmapValuesColorize_fontface)
export(tmapValuesColorize_lty)
export(tmapValuesColorize_lwd)
export(tmapValuesColorize_num)
export(tmapValuesColorize_shape)
export(tmapValuesColorize_size)
export(tmapValuesColorize_skip)
export(tmapValuesColorize_text)
export(tmapValuesColorize_xmod)
export(tmapValuesIsDiv_angle)
export(tmapValuesIsDiv_area)
export(tmapValuesIsDiv_bgcol)
export(tmapValuesIsDiv_bgcol_alpha)
export(tmapValuesIsDiv_col)
export(tmapValuesIsDiv_col_alpha)
export(tmapValuesIsDiv_fill)
export(tmapValuesIsDiv_fill_alpha)
export(tmapValuesIsDiv_fontface)
export(tmapValuesIsDiv_lty)
export(tmapValuesIsDiv_lwd)
export(tmapValuesIsDiv_num)
export(tmapValuesIsDiv_shape)
export(tmapValuesIsDiv_size)
export(tmapValuesIsDiv_skip)
export(tmapValuesIsDiv_text)
export(tmapValuesIsDiv_xmod)
export(tmapValuesRange_angle)
export(tmapValuesRange_area)
export(tmapValuesRange_bgcol)
export(tmapValuesRange_bgcol_alpha)
export(tmapValuesRange_col)
export(tmapValuesRange_col_alpha)
export(tmapValuesRange_fill)
export(tmapValuesRange_fill_alpha)
export(tmapValuesRange_fontface)
export(tmapValuesRange_lty)
export(tmapValuesRange_lwd)
export(tmapValuesRange_num)
export(tmapValuesRange_shape)
export(tmapValuesRange_size)
export(tmapValuesRange_skip)
export(tmapValuesRange_text)
export(tmapValuesRange_xmod)
export(tmapValuesScale_angle)
export(tmapValuesScale_area)
export(tmapValuesScale_bgcol)
export(tmapValuesScale_bgcol_alpha)
export(tmapValuesScale_col)
export(tmapValuesScale_col_alpha)
export(tmapValuesScale_fill)
export(tmapValuesScale_fill_alpha)
export(tmapValuesScale_fontface)
export(tmapValuesScale_lty)
export(tmapValuesScale_lwd)
export(tmapValuesScale_num)
export(tmapValuesScale_shape)
export(tmapValuesScale_size)
export(tmapValuesScale_skip)
export(tmapValuesScale_text)
export(tmapValuesScale_xmod)
export(tmapValuesSubmit_angle)
export(tmapValuesSubmit_area)
export(tmapValuesSubmit_bgcol)
export(tmapValuesSubmit_bgcol_alpha)
export(tmapValuesSubmit_col)
export(tmapValuesSubmit_col_alpha)
export(tmapValuesSubmit_fill)
export(tmapValuesSubmit_fill_alpha)
export(tmapValuesSubmit_fontface)
export(tmapValuesSubmit_lty)
export(tmapValuesSubmit_lwd)
export(tmapValuesSubmit_num)
export(tmapValuesSubmit_shape)
export(tmapValuesSubmit_size)
export(tmapValuesSubmit_skip)
export(tmapValuesSubmit_text)
export(tmapValuesSubmit_xmod)
export(tmapValuesSubmit_ymod)
export(tmapValuesVV_angle)
export(tmapValuesVV_area)
export(tmapValuesVV_bgcol)
export(tmapValuesVV_bgcol_alpha)
export(tmapValuesVV_col)
export(tmapValuesVV_col_alpha)
export(tmapValuesVV_fill)
export(tmapValuesVV_fill_alpha)
export(tmapValuesVV_fontface)
export(tmapValuesVV_lty)
export(tmapValuesVV_lwd)
export(tmapValuesVV_num)
export(tmapValuesVV_shape)
export(tmapValuesVV_size)
export(tmapValuesVV_skip)
export(tmapValuesVV_text)
export(tmapValuesVV_xmod)
export(tmap_animation)
export(tmap_arrange)
export(tmap_design_mode)
export(tmap_devel_mode)
export(tmap_format)
export(tmap_format_add)
export(tmap_grob)
export(tmap_icons)
export(tmap_last)
export(tmap_leaflet)
export(tmap_mode)
export(tmap_options)
export(tmap_options_diff)
export(tmap_options_mode)
export(tmap_options_reset)
export(tmap_options_save)
export(tmap_overview)
export(tmap_provider_credits)
export(tmap_providers)
export(tmap_save)
export(tmap_style)
export(tmap_style_catalog)
export(tmap_style_catalogue)
export(tmap_tip)
export(toTitleCase)
export(transform_values)
export(ttm)
export(ttmp)
import(classInt)
import(cols4all)
import(data.table)
import(grid)
import(htmltools)
import(htmlwidgets)
import(leafem)
import(leaflegend)
import(leaflet)
import(leafsync)
import(stats)
import(tmaptools)
import(units)
importFrom(grDevices,col2rgb)
importFrom(grDevices,colorRampPalette)
importFrom(grDevices,colors)
importFrom(grDevices,dev.off)
importFrom(grDevices,dev.size)
importFrom(grDevices,png)
importFrom(grDevices,rgb)
importFrom(htmlwidgets,saveWidget)
importFrom(htmlwidgets,shinyWidgetOutput)
importFrom(leaflet,icons)
importFrom(methods,as)
importFrom(rlang,expr)
importFrom(rlang,missing_arg)
importFrom(s2,as_s2_geography)
importFrom(s2,s2_buffer_cells)
importFrom(s2,s2_intersects)
importFrom(sf,st_as_sf)
importFrom(sf,st_crs)
importFrom(sf,st_geometry)
importFrom(sf,st_sf)
importFrom(sf,st_transform)
importFrom(stars,st_as_stars)
importFrom(stars,st_dimensions)
importFrom(stars,st_get_dimension_values)
importFrom(stars,st_set_dimensions)
importFrom(stars,st_warp)
importFrom(utils,capture.output)
importFrom(utils,data)
importFrom(utils,download.file)
importFrom(utils,getS3method)
importFrom(utils,head)
importFrom(utils,methods)
importFrom(utils,setTxtProgressBar)
importFrom(utils,str)
importFrom(utils,tail)
importFrom(utils,txtProgressBar)

crs_is_ortho = function(crs) {
	grepl("Orthographic", crs$wkt, fixed = TRUE)
}

crs_is_local = function(crs) {
	if (!inherits(crs, "crs")) {
		FALSE
	} else if (is.na(crs)) {
		TRUE
	} else {
		grepl("ENGCRS", crs$wkt, fixed = TRUE)
	}
}

# could be improved #1152
full_bbox = function(crs) {
	if (sf::st_is_longlat(crs)) {
		sf::st_bbox(sf::st_transform(sf::st_bbox(), crs = crs))
	} else {
		# to make sure the projected full bounding box is returned
		ply = tmaptools::bb_poly(sf::st_bbox())
		spl = suppressMessages(ply |> sf::st_sample(1e4, type = "regular") |> sf::st_sfc(crs = 4326))
		spl2 = sf::st_cast(ply, "POINT")
		spl12 = c(spl, spl2)
		sf::st_bbox(sf::st_transform(spl12, crs = crs))
	}
}


ortho_lonlat = function(lon, lat) {
	paste0("+proj=ortho +lat_0=", lat, " +lon_0=", lon)
}

transform_ortho = function(shp, crs, tmapID = NULL) {
	is_sfc = !is.null(tmapID) # also: inherits(shp, "sfc")
	if (is_sfc) shp = st_sf(shp, tmapID = tmapID)
	if (crs_is_ortho(crs)) {
		tryCatch({
			suppressWarnings({
				shp4326 = sf::st_transform(shp, 4326)
				visible = crs_ortho_visible(crs, projected = FALSE)
				if (!sf::st_is_valid(visible)) visible = sf::st_make_valid(visible)
				shp = suppressMessages(sf::st_intersection(shp4326, visible))
			})
		}, error = function(e) {
			shp
		})
	}
	shp2 = sf::st_transform(shp, crs = crs)

	if (is_sfc) {
		list(shp = st_geometry(shp2), tmapID = shp2$tmapID)
	} else {
		shp2
	}
}

crs_ortho_visible = function(crs, projected = TRUE, max_cells = 1e5) {
	wkt = sf::st_crs(crs)$wkt
	lst = strsplit(wkt, ",")[[1]]
	lat0 = as.numeric(lst[[which(grepl("Latitude of natural origin", lst))+1]])
	lon0 = as.numeric(lst[[which(grepl("Longitude of natural origin", lst))+1]])

	b = s2::s2_buffer_cells(s2::as_s2_geography(paste0("POINT(", lon0, " ", lat0, ")")), 9800000, max_cells = max_cells) # visible half

	pole_n = s2::s2_intersects("POINT(0 90)", b)
	pole_s = s2::s2_intersects("POINT(0 -90)", b)

	sfc = sf::st_as_sfc(b)
	co = sf::st_coordinates(sfc)[,1:2]

	if (pole_n || pole_s) {
		# STEP 1: create pattern from -360 to 360 (to fix potential 180/-180 meridian problems)

		# step 1a: arrange coordinates by lon
		lonmin_id = which.min(co[,1])
		lonmax_id = which.max(co[,1])

		if (pole_n) {
			co2 = co[c(lonmin_id:nrow(co), 1:lonmax_id), ]
		} else {
			co2 = co[c(lonmax_id:nrow(co), 1:lonmin_id), ]
		}

		# step 1b: replicate one before and one after
		co_left = co2
		co_left[,1] = co_left[,1] - 360
		co_right = co2
		co_right[,1] = co_right[,1] + 360

		if (pole_n) {
			co3 = rbind(co_left, co2, co_right)
		} else {
			co3 = rbind(co_right, co2, co_left)
		}

		co4 = co3[co3[,1] >= -360 & co3[,1] <= 360, ]

		# step 1c; make sure to close the polygon
		if (pole_n) {
			co4[1,1] = -360
			co4[nrow(co4),1] = 360
		} else {
			co4[1,1] = 360
			co4[nrow(co4),1] = -360
		}
		lat = co4[1,2]
		co4[nrow(co4),2] = lat

		# STEP 2: add block for the pole
		if (pole_n) {
			to_add = t(sapply(seq(360, -360, by = -45), c, 90))
		} else {
			to_add = t(sapply(seq(-360, 360, by = 45), c, -90))
		}
		co5 = rbind(co4,
					to_add,
					co4[1,,drop=FALSE])

		sfc = sf::st_sfc(sf::st_polygon(list(co5)), crs = sf::st_crs(sfc))

	} else {
		if (lon0 >= 0) {
			lon_shift = lon0 - 180

			co2 = co
			co2[,1][co2[,1] < lon_shift] = co2[,1][co2[,1] < lon_shift] + 360

			co_left = co2
			co_left[,1] = co_left[,1] - 360
			co_right = co2
			co_right[,1] = co_right[,1] + 360

			sfc = sf::st_sfc(sf::st_multipolygon(list(list(co_left), list(co2), list(co_right))), crs = sf::st_crs(sfc))
		} else {
			lon_shift = lon0 + 180

			co2 = co
			co2[,1][co2[,1] > lon_shift] = co2[,1][co2[,1] > lon_shift] - 360

			co_left = co2
			co_left[,1] = co_left[,1] - 360
			co_right = co2
			co_right[,1] = co_right[,1] + 360

			co3 = rbind(co_left, co2, co_right)

			sfc = sf::st_sfc(sf::st_multipolygon(list(list(co_left), list(co2), list(co_right))), crs = sf::st_crs(sfc))
		}


	}

	if (projected) sf::st_transform(sfc, crs) else sfc
}

end_of_the_world = function(crs, earth_datum) {
	wkt = sf::st_crs(crs)$wkt

	# orthographic
	if (grepl("Orthographic", wkt, fixed = TRUE)) {
		lst = strsplit(wkt, ",")[[1]]
		lat0 = as.numeric(lst[[which(grepl("Latitude of natural origin", lst))+1]])
		lon0 = as.numeric(lst[[which(grepl("Longitude of natural origin", lst))+1]])

		b = s2::s2_buffer_cells(s2::as_s2_geography(paste0("POINT(", lon0, " ", lat0, ")")), 9800000) # visible half

		sf::st_as_sfc(b) |>
			sf::st_transform(crs)
	} else {
		sf::st_bbox(stars::st_as_stars()) |>
			sf::st_as_sfc() |>
			sf::st_set_crs(NA) |>
			sf::st_segmentize(1) |>
			sf::st_set_crs(earth_datum) |>
			sf::st_transform(crs)
	}


}

crop_lat = function(bb, crs, limit_latitude_3857 = NULL) {
	if ((crs == 3857 || crs == st_crs(3857)) && (!is.null(limit_latitude_3857) && (!identical(limit_latitude_3857, FALSE)))) {
		crp = sf::st_bbox(c(xmin = -180, xmax = 180, ymin = limit_latitude_3857[1], ymax = limit_latitude_3857[2]), crs = 4326)
		crp2 = sf::st_transform(crp, crs = crs)
		bb['xmin'] = max(bb['xmin'], crp2['xmin'])
		bb['xmax'] = min(bb['xmax'], crp2['xmax'])
		bb['ymin'] = max(bb['ymin'], crp2['ymin'])
		bb['ymax'] = min(bb['ymax'], crp2['ymax'])
		bb
	} else {
		bb
	}
}


to_longest_linestring = function(shp) {
	crs = sf::st_crs(shp)
	if (sf::st_geometry_type(shp, by_geometry = FALSE) == "LINESTRING") {
		shp
	} else {
		shp_splitted = sf_expand(shp)
		do.call(sf::st_sfc, c(lapply(split(shp_splitted, shp_splitted$split__id), function(s) {
			s$geometry[[which.max(sf::st_length(s))]]
		}), list(crs = crs)))
	}
}

distances_bbox_sides = function(bbox, steps = 4) {
	crs_bbx = sf::st_crs(bbox)
	if (crs_is_local(crs_bbx)) return(c(top = NA, mid = NA, bottom = NA))
	xs = seq(bbox$xmin, bbox$xmax, length.out = steps + 1)
	ys = seq(bbox$ymin, bbox$ymax, length.out = 3)

	get_dst = function(y) {
		suppressWarnings({
			df = data.frame(x=xs, y = y) |>
				st_as_sf(coords = c("x", "y"), crs = crs_bbx)
			dfll = sf::st_transform(df, crs = 4326)
			dst = sf::st_distance(dfll)
			u = units(dst)
			units::set_units(sum(sapply(1:steps, function(i) dst[i, i + 1])), u, mode = "standard")
		})
	}

	d_top = get_dst(ys[3])
	d_mid = get_dst(ys[2])
	d_bottom = get_dst(ys[1])

	c(top = d_top, mid = d_mid, bottom = d_bottom)
}

#' @export
#' @rdname tmapGridLeaflet
tmapGridDataPlot.tm_data_raster = function(a, shpTM, dt, gp, bbx, facet_row, facet_col, facet_page, id, pane, group, glid, o, ...) {
	gts = get("gts", .TMAP_GRID)
	#bbx = get("bbx", .TMAP_GRID)

	g = get("g", .TMAP_GRID)

	g$fasp


	rc_text = frc(facet_row, facet_col)


	bb_target <- bb_asp(bbx, g$fasp)
	bb_real <-  stm_bbox_all(shpTM)

	shp = shpTM$shp
	tmapID = shpTM$tmapID

	if (is_regular_grid(shp)) {

		if (nrow(dt) == length(tmapID)) {
			# no matching needed

			if (all(dt$col_alpha == 1)) {
				color = dt$col
			} else {
				dt = merge_alpha(dt, name = "col")
				color = dt$ca
			}
		} else {
			# matching: to be improved
			tid = intersect(tmapID, dt$tmapID__)

			color = rep(NA, length(tmapID)) #"#FFFFFF"

			sel = which(tmapID %in% tid)
			tid2 = tmapID[sel]

			if (all(dt$col_alpha == 1)) {
				color[sel] = dt$col[match(tid2, dt$tmapID__)]
			} else {
				dt = merge_alpha(dt, name = "col")
				color[sel] = dt$ca[match(tid2, dt$tmapID__)]
			}
		}


		if (all(abs(bb_real-bb_target)< 1e-3)) {
			width <- 1
			height <- 1
			cent <- c(mean.default(c(bb_target[1], bb_target[3])), mean.default(c(bb_target[2], bb_target[4])))
		} else {
			width <- (bb_real[3] - bb_real[1]) / (bb_target[3] - bb_target[1])
			height <- (bb_real[4] - bb_real[2]) / (bb_target[4] - bb_target[2])
			cent <- c(mean.default(c(bb_real[1], bb_real[3])), mean.default(c(bb_real[2], bb_real[4])))
		}

		cx <- (cent[1] - bb_target[1]) / (bb_target[3] - bb_target[1])
		cy <- (cent[2] - bb_target[2]) / (bb_target[4] - bb_target[2])

		m <- matrix(color, ncol=nrow(shp), nrow=ncol(shp), byrow = TRUE)

		y_is_pos <- local({
			vals = stars::st_get_dimension_values(shp, "y")
			if (!is.null(vals)) {
				!all(diff(vals) < 0)
			} else {
				rst = attr(shp, "raster")
				name_y = names(get_xy_dim(shp))[2]
				delta = shp[[name_y]]$delta
				delta > 0
			}
		})

		if (y_is_pos) {
			m <- m[nrow(m):1L, ]
		}
		m[is.na(m)] = NA #"#0000FF"

		grb = grid::rasterGrob(m, x=cx, y=cy, width=width, height=height, interpolate = a$interpolate, name = paste0("raster_", id)) #gpl$raster.misc$interpolate
		gt = grid::addGrob(gts[[facet_page]], grb, gPath = grid::gPath(paste0("gt_facet_", rc_text)))
		gts[[facet_page]] = gt
		assign("gts", gts, envir = .TMAP_GRID)
	} else {
		m = matrix(tmapID, nrow = nrow(shp), ncol = ncol(shp))
		shp2 = structure(list(tmapID = m), class = "stars", dimensions = shp)
		shpTM = shapeTM(sf::st_geometry(sf::st_as_sf(shp2)), tmapID)

		#dt[, ":="(ord__ = 1, fill = col, fill_alpha = col_alpha, lty = "solid")]
		#dt[, ":="(col_alpha = 0)]

		dt[, ":="(ord__ = 1, lty = "solid")]

		a2 = structure(list(), class = "tm_data_polygons")
		tmapGridDataPlot(a2, shpTM, dt, gp, bbx, facet_row, facet_col, facet_page, id, pane, group, glid, o)
		#grid.shape(s, gp=grid::gpar(fill=color, col=NA), bg.col=NA, i, k)
	}
	NULL
}

crs_is_ortho = function(crs) {
	grepl("Orthographic", crs$wkt, fixed = TRUE)
}

crs_is_local = function(crs) {
	if (!inherits(crs, "crs")) {
		FALSE
	} else if (is.na(crs)) {
		TRUE
	} else {
		grepl("ENGCRS", crs$wkt, fixed = TRUE)
	}
}

# could be improved #1152
full_bbox = function(crs) {
	if (sf::st_is_longlat(crs)) {
		sf::st_bbox(sf::st_transform(sf::st_bbox(), crs = crs))
	} else {
		# to make sure the projected full bounding box is returned
		ply = tmaptools::bb_poly(sf::st_bbox())
		spl = suppressMessages(ply |> sf::st_sample(1e4, type = "regular") |> sf::st_sfc(crs = 4326))
		spl2 = sf::st_cast(ply, "POINT")
		spl12 = c(spl, spl2)
		sf::st_bbox(sf::st_transform(spl12, crs = crs))
	}
}


ortho_lonlat = function(lon, lat) {
	paste0("+proj=ortho +lat_0=", lat, " +lon_0=", lon)
}

transform_ortho = function(shp, crs, tmapID = NULL) {
	is_sfc = !is.null(tmapID) # also: inherits(shp, "sfc")
	if (is_sfc) shp = st_sf(shp, tmapID = tmapID)
	if (crs_is_ortho(crs)) {
		tryCatch({
			suppressWarnings({
				shp4326 = sf::st_transform(shp, 4326)
				visible = crs_ortho_visible(crs, projected = FALSE)
				if (!sf::st_is_valid(visible)) visible = sf::st_make_valid(visible)
				shp = suppressMessages(sf::st_intersection(shp4326, visible))
			})
		}, error = function(e) {
			shp
		})
	}
	shp2 = sf::st_transform(shp, crs = crs)

	if (is_sfc) {
		list(shp = st_geometry(shp2), tmapID = shp2$tmapID)
	} else {
		shp2
	}
}

crs_ortho_visible = function(crs, projected = TRUE, max_cells = 1e5) {
	wkt = sf::st_crs(crs)$wkt
	lst = strsplit(wkt, ",")[[1]]
	lat0 = as.numeric(lst[[which(grepl("Latitude of natural origin", lst))+1]])
	lon0 = as.numeric(lst[[which(grepl("Longitude of natural origin", lst))+1]])

	b = s2::s2_buffer_cells(s2::as_s2_geography(paste0("POINT(", lon0, " ", lat0, ")")), 9800000, max_cells = max_cells) # visible half

	pole_n = s2::s2_intersects("POINT(0 90)", b)
	pole_s = s2::s2_intersects("POINT(0 -90)", b)

	sfc = sf::st_as_sfc(b)
	co = sf::st_coordinates(sfc)[,1:2]

	if (pole_n || pole_s) {
		# STEP 1: create pattern from -360 to 360 (to fix potential 180/-180 meridian problems)

		# step 1a: arrange coordinates by lon
		lonmin_id = which.min(co[,1])
		lonmax_id = which.max(co[,1])

		if (pole_n) {
			co2 = co[c(lonmin_id:nrow(co), 1:lonmax_id), ]
		} else {
			co2 = co[c(lonmax_id:nrow(co), 1:lonmin_id), ]
		}

		# step 1b: replicate one before and one after
		co_left = co2
		co_left[,1] = co_left[,1] - 360
		co_right = co2
		co_right[,1] = co_right[,1] + 360

		if (pole_n) {
			co3 = rbind(co_left, co2, co_right)
		} else {
			co3 = rbind(co_right, co2, co_left)
		}

		co4 = co3[co3[,1] >= -360 & co3[,1] <= 360, ]

		# step 1c; make sure to close the polygon
		if (pole_n) {
			co4[1,1] = -360
			co4[nrow(co4),1] = 360
		} else {
			co4[1,1] = 360
			co4[nrow(co4),1] = -360
		}
		lat = co4[1,2]
		co4[nrow(co4),2] = lat

		# STEP 2: add block for the pole
		if (pole_n) {
			to_add = t(sapply(seq(360, -360, by = -45), c, 90))
		} else {
			to_add = t(sapply(seq(-360, 360, by = 45), c, -90))
		}
		co5 = rbind(co4,
					to_add,
					co4[1,,drop=FALSE])

		sfc = sf::st_sfc(sf::st_polygon(list(co5)), crs = sf::st_crs(sfc))

	} else {
		if (lon0 >= 0) {
			lon_shift = lon0 - 180

			co2 = co
			co2[,1][co2[,1] < lon_shift] = co2[,1][co2[,1] < lon_shift] + 360

			co_left = co2
			co_left[,1] = co_left[,1] - 360
			co_right = co2
			co_right[,1] = co_right[,1] + 360

			sfc = sf::st_sfc(sf::st_multipolygon(list(list(co_left), list(co2), list(co_right))), crs = sf::st_crs(sfc))
		} else {
			lon_shift = lon0 + 180

			co2 = co
			co2[,1][co2[,1] > lon_shift] = co2[,1][co2[,1] > lon_shift] - 360

			co_left = co2
			co_left[,1] = co_left[,1] - 360
			co_right = co2
			co_right[,1] = co_right[,1] + 360

			co3 = rbind(co_left, co2, co_right)

			sfc = sf::st_sfc(sf::st_multipolygon(list(list(co_left), list(co2), list(co_right))), crs = sf::st_crs(sfc))
		}


	}

	if (projected) sf::st_transform(sfc, crs) else sfc
}

end_of_the_world = function(crs, earth_datum) {
	wkt = sf::st_crs(crs)$wkt

	# orthographic
	if (grepl("Orthographic", wkt, fixed = TRUE)) {
		lst = strsplit(wkt, ",")[[1]]
		lat0 = as.numeric(lst[[which(grepl("Latitude of natural origin", lst))+1]])
		lon0 = as.numeric(lst[[which(grepl("Longitude of natural origin", lst))+1]])

		b = s2::s2_buffer_cells(s2::as_s2_geography(paste0("POINT(", lon0, " ", lat0, ")")), 9800000) # visible half

		sf::st_as_sfc(b) |>
			sf::st_transform(crs)
	} else {
		sf::st_bbox(stars::st_as_stars()) |>
			sf::st_as_sfc() |>
			sf::st_set_crs(NA) |>
			sf::st_segmentize(1) |>
			sf::st_set_crs(earth_datum) |>
			sf::st_transform(crs)
	}


}

crop_lat = function(bb, crs, limit_latitude_3857 = NULL) {
	if ((crs == 3857 || crs == st_crs(3857)) && (!is.null(limit_latitude_3857) && (!identical(limit_latitude_3857, FALSE)))) {
		crp = sf::st_bbox(c(xmin = -180, xmax = 180, ymin = limit_latitude_3857[1], ymax = limit_latitude_3857[2]), crs = 4326)
		crp2 = sf::st_transform(crp, crs = crs)
		bb['xmin'] = max(bb['xmin'], crp2['xmin'])
		bb['xmax'] = min(bb['xmax'], crp2['xmax'])
		bb['ymin'] = max(bb['ymin'], crp2['ymin'])
		bb['ymax'] = min(bb['ymax'], crp2['ymax'])
		bb
	} else {
		bb
	}
}


to_longest_linestring = function(shp) {
	crs = sf::st_crs(shp)
	if (sf::st_geometry_type(shp, by_geometry = FALSE) == "LINESTRING") {
		shp
	} else {
		shp_splitted = sf_expand(shp)
		do.call(sf::st_sfc, c(lapply(split(shp_splitted, shp_splitted$split__id), function(s) {
			s$geometry[[which.max(sf::st_length(s))]]
		}), list(crs = crs)))
	}
}

distances_bbox_sides = function(bbox, steps = 4) {
	crs_bbx = sf::st_crs(bbox)
	if (crs_is_local(crs_bbx)) return(c(top = NA, mid = NA, bottom = NA))
	xs = seq(bbox$xmin, bbox$xmax, length.out = steps + 1)
	ys = seq(bbox$ymin, bbox$ymax, length.out = 3)

	get_dst = function(y) {
		suppressWarnings({
			df = data.frame(x=unname(xs), y = unname(y)) |>
				st_as_sf(coords = c("x", "y"), crs = crs_bbx)
			dfll = sf::st_transform(df, crs = 4326)
			dst = sf::st_distance(dfll)
			u = units(dst)
			units::set_units(sum(sapply(1:steps, function(i) dst[i, i + 1])), u, mode = "standard")
		})
	}

	d_top = get_dst(ys[3])
	d_mid = get_dst(ys[2])
	d_bottom = get_dst(ys[1])

	c(top = d_top, mid = d_mid, bottom = d_bottom)
}

#' @export
#' @rdname tmapGridLeaflet
tmapGridDataPlot.tm_data_raster = function(a, shpTM, dt, gp, bbx, facet_row, facet_col, facet_page, id, pane, group, glid, o, ...) {
	gts = get("gts", .TMAP_GRID)
	#bbx = get("bbx", .TMAP_GRID)

	g = get("g", .TMAP_GRID)

	g$fasp


	rc_text = frc(facet_row, facet_col)


	bb_target <- bb_asp(bbx, g$fasp)
	bb_real <-  stm_bbox_all(shpTM)

	shp = shpTM$shp
	tmapID = shpTM$tmapID

	if (is_regular_grid(shp)) {

		if (nrow(dt) == length(tmapID)) {
			# no matching needed

			if (all(dt$col_alpha == 1)) {
				color = dt$col
			} else {
				dt = merge_alpha(dt, name = "col")
				color = dt$ca
			}
		} else {
			# matching: to be improved
			tid = intersect(tmapID, dt$tmapID__)

			color = rep(NA, length(tmapID)) #"#FFFFFF"

			sel = which(tmapID %in% tid)
			tid2 = tmapID[sel]

			if (all(dt$col_alpha == 1)) {
				color[sel] = dt$col[match(tid2, dt$tmapID__)]
			} else {
				dt = merge_alpha(dt, name = "col")
				color[sel] = dt$ca[match(tid2, dt$tmapID__)]
			}
		}


		if (all(abs(bb_real-bb_target)< 1e-3)) {
			width <- 1
			height <- 1
			cent <- c(mean.default(c(bb_target[1], bb_target[3])), mean.default(c(bb_target[2], bb_target[4])))
		} else {
			width <- (bb_real[3] - bb_real[1]) / (bb_target[3] - bb_target[1])
			height <- (bb_real[4] - bb_real[2]) / (bb_target[4] - bb_target[2])
			cent <- c(mean.default(c(bb_real[1], bb_real[3])), mean.default(c(bb_real[2], bb_real[4])))
		}

		cx <- (cent[1] - bb_target[1]) / (bb_target[3] - bb_target[1])
		cy <- (cent[2] - bb_target[2]) / (bb_target[4] - bb_target[2])

		m <- matrix(color, ncol=nrow(shp), nrow=ncol(shp), byrow = TRUE)

		y_is_pos <- local({
			vals = stars::st_get_dimension_values(shp, "y")
			if (!is.null(vals)) {
				!all(diff(vals) < 0)
			} else {
				rst = attr(shp, "raster")
				name_y = names(get_xy_dim(shp))[2]
				delta = shp[[name_y]]$delta
				delta > 0
			}
		})

		if (y_is_pos) {
			m <- m[nrow(m):1L, ]
		}
		m[is.na(m)] = NA #"#0000FF"

		grb = grid::rasterGrob(m, x=cx, y=cy, width=width, height=height, interpolate = a$interpolate, name = paste0("raster_", id)) #gpl$raster.misc$interpolate

		#gt = grid::addGrob(gts[[facet_page]], grb, gPath = grid::gPath(paste0("gt_facet_", rc_text)))

		if (!is.null(a$blend) && a$blend != "over") {
			existing = gts[[facet_page]]$children$gt_main$children[[paste0("gt_facet_", rc_text)]]$children
			dst = do.call(grid::grobTree, existing)
			grb = blend_grobs(grb, a$blend, dst = dst)
		}

		gt = grid::addGrob(gts[[facet_page]], grb, gPath = grid::gPath(paste0("gt_facet_", rc_text)))
		gts[[facet_page]] = gt
		assign("gts", gts, envir = .TMAP_GRID)

	} else {
		m = matrix(tmapID, nrow = nrow(shp), ncol = ncol(shp))
		shp2 = structure(list(tmapID = m), class = "stars", dimensions = shp)
		shpTM = shapeTM(sf::st_geometry(sf::st_as_sf(shp2)), tmapID)

		#dt[, ":="(ord__ = 1, fill = col, fill_alpha = col_alpha, lty = "solid")]
		#dt[, ":="(col_alpha = 0)]

		dt[, ":="(ord__ = 1, lty = "solid")]

		a2 = structure(list(), class = "tm_data_polygons")
		tmapGridDataPlot(a2, shpTM, dt, gp, bbx, facet_row, facet_col, facet_page, id, pane, group, glid, o)
		#grid.shape(s, gp=grid::gpar(fill=color, col=NA), bg.col=NA, i, k)
	}
	NULL
}

#' Map layer: iso (contour)
#'
#' Map layer that draws iso (contour) lines. Stack of [tm_lines()] and [tm_labels_highlighted].
#'
#' @param col `r .doc_vv("col")`
#' @param text `r .doc_vv("text")`
#' @param ... passed on to [tm_lines()] and [tm_labels_highlighted()]. For the text color and alpha transparency of the text labels, please use `text_col` and `text_alpha` instead of `col` and `col_alpha`.
#' @param options_lines The options for [tm_lines()]
#' @param options_labels The options for [tm_labels_highlighted()]
#' @export
tm_iso = function(col = tm_const(),
				  text = tm_vars(x = 1),
				  ...,
				  options_lines = opt_tm_lines(),
				  options_labels = opt_tm_labels()) {
	args = list(...)
	args$options = NULL
	argsL = args[intersect(names(formals("tm_lines")), names(args))]

	nms_col = c("col", "col.scale", "col.legend", "col.chart", "col.free", "col_alpha", "col_alpha.scale", "col_alpha.legend", "col_alpha.chart", "col_alpha.free")

	# update args for labels
	args[nms_col] = NULL
	rename_id = names(args) %in% paste0("text_", nms_col)

	names(args)[rename_id] = substr(names(args)[rename_id], 6, nchar(names(args)[rename_id]))
	argsT = args[intersect(names(formals("tm_labels_highlighted")), names(args))]

	do.call("tm_lines", c(list(col=col, options = options_lines), argsL)) +
	do.call("tm_labels_highlighted", c(list(text=text, options = options_labels), argsT))
}

# from LLM, to be improved
pretty_datetime_seq <- function(x, n, rounding_method_start = "floor", rounding_method_end = "ceiling") {
	# --- Input Validation ---
	if (!inherits(x, c("POSIXct", "POSIXlt", "Date"))) {
		stop("Input 'x' must be a POSIXct, POSIXlt, or Date vector.")
	}
	if (!is.numeric(n) || length(n) != 1 || n <= 0) {
		stop("Input 'n' must be a positive numeric value.")
	}
	rounding_methods <- c("none", "floor", "ceiling", "round")
	if (!(rounding_method_start %in% rounding_methods) || !(rounding_method_end %in% rounding_methods)) {
		stop("Invalid rounding_method. Choose from 'none', 'floor', 'ceiling', or 'round'.")
	}

	# --- Determine Start and End Times (handling different classes, timezones, and NAs) ---
	if (inherits(x, "Date")) {
		# Use na.rm = TRUE to handle potential NA values
		start_time <- as.POSIXct(min(x, na.rm = TRUE), tz = "UTC")
		end_time <- as.POSIXct(max(x, na.rm = TRUE), tz = "UTC")
		original_class <- "Date"
	} else {
		# Use na.rm = TRUE and explicitly set the timezone
		start_time <- as.POSIXct(min(x, na.rm = TRUE), tz = attr(x, "tzone"))
		end_time <- as.POSIXct(max(x, na.rm = TRUE), tz = attr(x, "tzone"))
		original_class <- class(x)[1]
	}

	# Check if start_time or end_time are NA (meaning ALL values in x were NA)
	if (is.na(start_time) || is.na(end_time)) {
		return(NA)  # Or throw an error: stop("Input vector 'x' contains only NA values.")
	}


	# --- Rounding Function (internal) ---
	round_datetime <- function(dt, interval_text, method) {

		interval_unit <- gsub("[0-9. ]", "", interval_text)

		if (interval_unit %in% c("months", "years")){
			dt_lt <- as.POSIXlt(dt)

			if (interval_unit == "months") {
				if(method == "floor"){
					dt_lt$mday <- 1
					dt_lt$hour <- 0
					dt_lt$min <- 0
					dt_lt$sec <- 0
				} else if (method == "ceiling"){
					dt_lt$mon <- dt_lt$mon + as.numeric(gsub("[^0-9.]", "", interval_text))
					dt_lt$mday <- 1
					dt_lt$hour <- 0
					dt_lt$min <- 0
					dt_lt$sec <- 0
				} else if (method == "round"){
					#Get next month
					next_mon <- as.POSIXlt(dt)
					next_mon$mon <- next_mon$mon + as.numeric(gsub("[^0-9.]", "", interval_text))
					next_mon$mday <- 1
					next_mon$hour <- 0
					next_mon$min <- 0
					next_mon$sec <- 0
					#Calculate distances
					dist_up <- as.numeric(difftime(next_mon, dt, units="secs"))
					dist_down <- as.numeric(difftime(dt, round_datetime(dt, interval_text, "floor"), units="secs"))
					if(dist_up < dist_down) {
						dt_lt <- next_mon
					} else{
						dt_lt$mday <- 1
						dt_lt$hour <- 0
						dt_lt$min <- 0
						dt_lt$sec <- 0
					}

				}
			} else if (interval_unit == "years"){
				if (method == "floor") {
					dt_lt$mon <- 0  # January
					dt_lt$mday <- 1
					dt_lt$hour <- 0
					dt_lt$min <- 0
					dt_lt$sec <- 0
				} else if (method == "ceiling") {
					dt_lt$year <- dt_lt$year + as.numeric(gsub("[^0-9.]", "", interval_text))
					dt_lt$mon <- 0  # January
					dt_lt$mday <- 1
					dt_lt$hour <- 0
					dt_lt$min <- 0
					dt_lt$sec <- 0
				} else if (method == "round") {
					#Get next year
					next_year <- as.POSIXlt(dt)
					next_year$year <- next_year$year + as.numeric(gsub("[^0-9.]", "", interval_text))
					next_year$mon <- 0 # January
					next_year$mday <- 1
					next_year$hour <- 0
					next_year$min <- 0
					next_year$sec <- 0
					#Calculate distances
					dist_up <- as.numeric(difftime(next_year, dt, units = "secs"))
					dist_down <- as.numeric(difftime(dt, round_datetime(dt, interval_text, "floor"), units = "secs"))
					if(dist_up < dist_down){
						dt_lt <- next_year
					} else {
						dt_lt$mon <- 0  # January
						dt_lt$mday <- 1
						dt_lt$hour <- 0
						dt_lt$min <- 0
						dt_lt$sec <- 0
					}
				}
			}
			return(as.POSIXct(dt_lt)) #Convert back
		} else {
			#For secs, mins, hours, days and weeks, we use numeric operations
			interval_num <- as.numeric(gsub("[^0-9.]", "", interval_text))
			interval_seconds <- switch(interval_unit,
									   "secs" = interval_num,
									   "mins" = interval_num * 60,
									   "hours" = interval_num * 3600,
									   "days" = interval_num * 86400,
									   "weeks" = interval_num * 86400 * 7,
									   stop("Unsupported interval unit")
			)
			dt_numeric <- as.numeric(dt)
			rounded_numeric <- switch(method,
									  "floor" = floor(dt_numeric / interval_seconds) * interval_seconds,
									  "ceiling" = ceiling(dt_numeric / interval_seconds) * interval_seconds,
									  "round" = round(dt_numeric / interval_seconds) * interval_seconds,
									  dt_numeric # none
			)
			return(as.POSIXct(rounded_numeric, origin = "1970-01-01", tz = attr(dt, "tzone")))
		}
	}

	# --- Determine Ideal Interval ---
	# Calculate total_seconds ACCURATELY using numeric representation
	total_seconds <- as.numeric(end_time) - as.numeric(start_time)  # KEY CHANGE
	target_interval_seconds <- total_seconds / n

	# --- Find "Pretty" Interval ---
	pretty_intervals <- c(
		1, 2, 5, 10, 15, 30,
		60, 120, 300, 600, 900, 1800, 3600,
		3600 * 2, 3600 * 3, 3600 * 4, 3600 * 6, 3600 * 8, 3600 * 12,
		86400, 86400 * 2, 86400 * 7,
		86400 * 30, 86400 * 30 * 2, 86400 * 30 * 3, 86400 * 30 * 4, 86400 * 30 * 6,
		86400 * 365, 86400 * 365 * 2, 86400 * 365 * 5, 86400 * 365 * 10
	)
	best_interval_seconds <- pretty_intervals[which.min(abs(pretty_intervals - target_interval_seconds))]

	interval_text <- if (best_interval_seconds < 60) {
		paste(best_interval_seconds, "secs")
	} else if (best_interval_seconds < 3600) {
		paste(best_interval_seconds / 60, "mins")
	} else if (best_interval_seconds < 86400) {
		paste(best_interval_seconds / 3600, "hours")
	} else if (best_interval_seconds < (86400 * 7)) {
		paste(best_interval_seconds / 86400, "days")
	} else if (best_interval_seconds < (86400 * 30)) {
		paste(best_interval_seconds / (86400 * 7), "weeks")
	} else if (best_interval_seconds < (86400 * 365)) {
		#Months is approximate, so we use the interval_text to calculate
		interval_text <- paste(round(best_interval_seconds / (86400*30)), "months")
		best_interval_seconds <- as.numeric(gsub("[^0-9.]", "", interval_text)) * 86400*30
		interval_text #return
	} else {
		#Years is approximate, so we use the interval_text to calculate
		interval_text <- paste(round(best_interval_seconds / (86400 * 365)), "years")
		best_interval_seconds <- as.numeric(gsub("[^0-9.]", "", interval_text)) * 86400 * 365
		interval_text #return
	}

	# --- Round Start and End Times ---
	rounded_start_time <- round_datetime(start_time, interval_text, rounding_method_start)
	rounded_end_time <- round_datetime(end_time, interval_text, rounding_method_end)

	# --- Generate Sequence ---
	date_time_sequence <- seq.POSIXt(from = rounded_start_time, to = rounded_end_time, by = interval_text)

	# --- Convert back to original class ---
	if (original_class == "Date") {
		date_time_sequence <- as.Date(date_time_sequence)
	} else if (original_class == "POSIXlt") {
		date_time_sequence <- as.POSIXlt(date_time_sequence)
	}

	return(date_time_sequence)
}


pretty_datetime_seq3 = function(x, by) {
	s = seq(min(x), max(x), by = by)
	pretty_datetime_seq(x, n = length(s))
}

#' @export
#' @keywords internal
#' @rdname tmapGridLeaflet
tmapGridAuxPrepare.tm_aux_basemap = function(a, bs, id, o) {
	tmapGridAuxPrepare.tm_aux_tiles(a, bs, id, o)
}

#' @export
#' @keywords internal
#' @rdname tmapGridLeaflet
tmapGridAuxPrepare.tm_aux_tiles = function(a, bs, id, o) {
	g = get("g", envir = .TMAP_GRID)
	rlang::check_installed("maptiles")

	crs = sf::st_crs(bs[[1]])

	crs3857 = (crs == 3857 || crs == st_crs(3857))

	isproj = !sf::st_is_longlat(crs)

	bs_orig = bs
	if (isproj) {
		# plain lat-lon to find zoom levels
		bs = lapply(bs, function(b) {
			sf::st_bbox(sf::st_transform(sf::st_as_sfc(b), crs = "EPSG:4326"))
		})
	}

	# tiles are in mercator
	bs3857 = lapply(bs_orig, function(b) {
		sf::st_bbox(sf::st_transform(tmaptools::bb_poly(b), crs = "EPSG:3857"))
	})

	bs = lapply(bs, function(b) {
		# not sure why needed
		bb_ll_valid(bb_asp(b, g$fasp))
	})

	if (is.na(a$zoom)) {
		zs = vapply(bs, findZoom, FUN.VALUE = integer(1))
	} else {
		zs = rep(a$zoom, length(bs))
	}

	serv = a$server[1]

	if (substr(serv, 1, 4) == "http") {
		if (serv %in% .TMAP_GRID$maptiles_urls) {
			tile_id = which(serv == .TMAP_GRID$maptiles_urls)[1]
		} else {
			tile_id = length(.TMAP_GRID$maptiles_urls) + 1L
			.TMAP_GRID$maptiles_urls = c(.TMAP_GRID$maptiles_urls, serv)
		}
		sub = if (is.na(a$sub)) NA else strsplit(a$sub, split = "")[[1]]
		serv = maptiles::create_provider(paste0("id_", tile_id), url = serv, citation = "", sub = sub)
		api = if (!is.null(a$api)) {
			a$api
		} else NULL
	} else if (substr(serv, 1, 6) %in% c("Thunde", "Stadia")) {
		is_stadia = substr(serv, 1, 6) == "Stadia"
		api = if (!is.null(a$api)) {
			a$api
		} else if (is_stadia) {
			Sys.getenv("STADIA_MAPS")
		} else {
			Sys.getenv("THUNDERFOREST_MAPS")
		}
		if (api == "") message_basemaps(is_stadia)
	} else {
		api = NULL
	}

	xs = mapply(function(b, z) {
		m = tryCatch({
			if (is.null(api)) {
				maptiles::get_tiles(x = b, provider = serv, zoom = z, crop = TRUE)
			} else {
				maptiles::get_tiles(x = b, provider = serv, apikey = api, zoom = z, crop = TRUE)
			}
		}, error = function(e) {
			tryCatch({
				if (is.null(api)) {
					maptiles::get_tiles(x = b, provider = serv, zoom = z - 1, crop = TRUE)
				} else {
					maptiles::get_tiles(x = b, provider = serv, apikey = api, zoom = z - 1, crop = TRUE)
				}
			}, error = function(e) {
				NULL
			})
		})
		if (!is.null(m)) {
			names(m)[1:3] = c("red", "green", "blue")
			if (terra::nlyr(m) == 4) names(m)[4] = "alpha"

		} else {
			message_basemaps_none(serv, z)
		}
		m
	}, bs3857, zs, SIMPLIFY = FALSE)

	if (isproj && !crs3857) {
		if (!all(vapply(xs, is.null, FUN.VALUE = logical(1)))) {
			message_basemaps_blurry(serv)
			xs = mapply(function(x,b) {
				if (is.null(x)) return(NULL)

				ex = terra::ext(as.vector(b[c(1,3,2,4)]))
				asp = (ex[2] - ex[1]) / (ex[4] - ex[3])

				tot = terra::ncell(x) * 2

				nc = round(sqrt(tot * asp))
				nr = round(tot / nc)

				r = terra::rast(ex, nrows = nr, ncols = nc, crs = crs$wkt)
				terra::project(x, r, method = "near")
			}, xs, bs_orig, SIMPLIFY = FALSE)
		}
	}

	ss = lapply(xs, function(x) {
		if (is.null(x)) NULL else do.call(tmapShape, list(shp = x, is.main = FALSE, crs = crs, bbox = NULL, unit=NULL, filter=NULL, layer = NULL, shp_name = "x", smeta = list(), o = o, tmf = NULL))
	})

	srgb = tm_scale_rgb(max_color_value = 255, value.na = "#FFFFFF")


	ds = lapply(ss, function(s) {
		if (is.null(s)) return(NULL)
		d = s$dt
		d[, c("col", "legnr", "crtnr") := do.call(srgb$FUN, list(x1 = red, x2 = green, x3 = blue, scale = srgb, legend = list(), o = o, aes = "col", layer = "raster", layer_args = opt_tm_rgb(interpolate = TRUE)$mapping.args, sortRev = NA, bypass_ord = TRUE))]
		if ("alpha" %in% names(d)) {
			d[, col_alpha:=alpha/255 * a$alpha]
			d[is.na(col_alpha), col_alpha:=0]
		} else {
			d[, col_alpha:=a$alpha]
		}
		d
	})

	shpTMs = lapply(ss, function(s) {
		if (is.null(s)) NULL else s$shpTM
	})


	bmaps_shpTHs = structure(list(shpTMs), names = id)
	bmaps_dts = structure(list(ds), names = id)

	g$bmaps_shpTHs = c(g$bmaps_shpTHs, bmaps_shpTHs)
	g$bmaps_dts = c(g$bmaps_dts, bmaps_dts)

	assign("g", g, envir = .TMAP_GRID)
	paste0(a$server, collapse = "__")
}

#' @export
#' @rdname tmapGridLeaflet
tmapLeafletDataPlot.tm_data_raster = function(a, shpTM, dt, pdt, popup.format, hdt, idt, gp, bbx, facet_row, facet_col, facet_page, id, pane, group, glid, o, ...) {

	rc_text = frc(facet_row, facet_col)


	shp = shpTM$shp
	tmapID = shpTM$tmapID

	if (is_regular_grid(shp)) {

		tid = intersect(tmapID, dt$tmapID__)

		color = rep(NA, length(tmapID)) # NA

		sel = which(tmapID %in% tid)
		tid2 = tmapID[sel]

		color[sel] = dt$col[match(tid2, dt$tmapID__)]


		#color = rep("#FFFFFF", length(tmapID))
		#color[match(dt$tmapID__, tmapID)] = dt$color

		pal <- na.omit(unique(color))
		pal <- pal[substr(pal, 8,10)!="00"] ## remove transparant colors

		if (!length(pal)) return(NULL)

		res <- split_alpha_channel(pal, alpha = 1)
		pal_col <- res$col
		pal_opacity <- if (length(res$opacity) == 0L) 0 else max(res$opacity)

		if ("col_alpha" %in% names(dt)) pal_opacity = max(dt$col_alpha)


		col_ids <- match(color, pal)

		m <- matrix(col_ids, ncol = ncol(shp))

		#matrix(color, ncol = ncol(shp))

		#m <- matrix(tmapID, ncol = ncol(shp))


		#m = tmapID

		#m[1,5] = 4

		shp2 = stars::st_as_stars(m, dimensions = shp)

		lf = get_lf(facet_row, facet_col, facet_page)

		opts = leaflet::gridOptions(pane = pane)

		lf %>%
			leafem::addStarsImage(shp2, band = 1, colors = pal_col, opacity = pal_opacity, group = group, options = opts) %>%
			assign_lf(facet_row, facet_col, facet_page)
	} else {
		#shp2 = stars::st_as_stars(list(values = tmapID), dimensions = shp)
		#shpTM = shapeTM(sf::st_geometry(sf::st_as_sf(shp2)), as.vector(tmapID))

		m = matrix(tmapID, nrow = nrow(shp), ncol = ncol(shp))
		shp2 = structure(list(tmapID = m), class = "stars", dimensions = shp)

		shp3 = sf::st_geometry(sf::st_as_sf(shp2))

		crs = get_option_class(o$crs_step4, "sf")

		shpTM = shapeTM(sf::st_transform(shp3, crs), tmapID)


		gp$lty = "solid"
		a2 = structure(list(), class = "tm_data_polygons")
		tmapLeafletDataPlot(a2, shpTM, dt, pdt, popup.format = NULL, hdt = NULL, idt = NULL, gp, bbx, facet_row, facet_col, facet_page, id, pane, group, glid, o)
		#grid.shape(s, gp=gpar(fill=color, col=NA), bg.col=NA, i, k)
	}
	NULL
}

#' @method tmapGridCompPrepare tm_chart
#' @export
tmapGridCompPrepare.tm_chart = function(comp, o) {
	comp
}

#' @method tmapGridCompPrepare tm_chart_bar
#' @export
tmapGridCompPrepare.tm_chart_bar = function(comp, o) {
	if (nlevels(comp$tab$color) == 1) comp$plot.axis.x = TRUE
	comp
}

#' @method tmapGridCompPrepare tm_chart_histogram
#' @export
tmapGridCompPrepare.tm_chart_histogram = function(comp, o) {
	if (nlevels(comp$tab$color) == 1) comp$plot.axis.x = TRUE
	comp
}



#' @method tmapGridCompWidth tm_chart
#' @export
tmapGridCompWidth.tm_chart = function(comp, o) {

	textS = comp$text.size #* o$scale
	#textP = comp$padding[c(3,1)] * textS * o$lin

	comp$margins = c(0.02, 0.02, 0.02, 0.02)
	marW = comp$margins[c(2,4)] * textS * o$lin

	comp$nlines = get_vector_id(comp$width, paste(comp$type, {if (comp$predefined) "min" else "max"}, sep = "."))

	body = comp$nlines * textS * o$lin
	ws = c(marW[1], body, marW[2])

	sides = switch(comp$position$align.h, left = "second", right = "first", "both")
	wsu = set_unit_with_stretch(ws, sides = sides)
	comp$flexCol = NA
	comp$Win = sum(ws)
	comp$wsu = wsu
	comp

}


#' @method tmapGridCompHeight tm_chart
#' @export
tmapGridCompHeight.tm_chart = function(comp, o) {

	textS = comp$text.size #* o$scale
	#textP = comp$padding[c(3,1)] * textS * o$lin

	comp$margins = c(0.02, 0.02, 0.02, 0.02)
	comp$nlines = get_vector_id(comp$height, paste(comp$type, {if (comp$predefined) "min" else "max"}, sep = "."))


	marH = comp$margins[c(3,1)] * textS * o$lin
	hs = c(marH[1], comp$nlines * textS * o$lin, marH[2])


	sides = switch(comp$position$align.v, top = "second", bottom = "first", "both")
	hsu = set_unit_with_stretch(hs, sides = sides)

	Hin = sum(hs)
	comp$flexRow = NA
	comp$Hin = Hin #  sum(textP[1], textH, textP[2])
	comp$hsu = hsu

	comp

}

#' @method tmapGridCompPlot tm_chart_histogram
#' @export
tmapGridCompPlot.tm_chart_histogram = function(comp, o, fH, fW) {
	tmapGridCompPlot.tm_chart_bar(comp, o, fH, fW)
}

#' @method tmapGridCompPlot tm_chart_bar
#' @export
tmapGridCompPlot.tm_chart_bar = function(comp, o, fH, fW) {
	u = 1/(comp$nlines)
	#vpComp = viewport(x=u, y=u, height=1-2*u, width=1-2*u, just=c("left", "bottom"))
	scale = o$scale * comp$scale

	textsize = o$chart.text.size * scale

	wsu = comp$wsu
	hsu = comp$hsu

	vp = grid::viewport(layout = grid::grid.layout(ncol = length(wsu),
												   nrow = length(hsu),
												   widths = wsu,
												   heights = hsu))

	grobBG = if (getOption("tmap.design.mode")) rectGrob(gp=gpar(fill="#CAB2D6")) else NULL

	g = ggplot2::ggplot(comp$tab, ggplot2::aes(x = bin, y = freq, fill = color)) +
		ggplot2::geom_bar(width = 1, lwd = lwd_to_mm(scale),color = "#000000", stat = "identity", na.rm = TRUE) +
		ggplot2::scale_fill_manual(values = comp$pal) +
		theme_chart(plot.axis.x = comp$plot.axis.x, plot.axis.y = comp$plot.axis.y, scale = scale, text.color = o$chart.text.color, text.size = textsize)

	g = g + comp$extra.ggplot2

	g2 = ggplot2::ggplotGrob(g)

	# other grid cells are aligns (1 and 5) and margins (2 and 4)
	histogram = gridCell(3,3, {
		gTree(children=gList(grobBG,
							 g2),
			  name="compass")
	})

	grid::grobTree(histogram, vp = vp)
}

## borrowed from https://github.com/CMAP-REPOS/cmapplot
# .lwd <- ggplot2::.pt / ggplot2::.stroke

#' @export
#' @rdname tmap_internal
lwd_to_mm = function(value, unit = "bigpts") {
	grid::convertUnit(grid::unit(value, unit), "mm", valueOnly = TRUE)
}


# if (FALSE) {
# 	library(ggplot2)
# 	ggplot(mpg, aes(class)) +
# 		geom_bar(fill = NA, color = "black", linewidth = gg_lwd_convert(12))
# 	grid.lines(y = c(0.5, 0.5), gp=gpar(lwd = 12))
#
# }




#' @method tmapGridCompPlot tm_chart_donut
#' @export
tmapGridCompPlot.tm_chart_donut = function(comp, o, fH, fW) {
	u = 1/(comp$nlines)
	#vpComp = viewport(x=u, y=u, height=1-2*u, width=1-2*u, just=c("left", "bottom"))
	scale = o$scale * comp$scale
	textsize = o$chart.text.size * scale

	wsu = comp$wsu
	hsu = comp$hsu

	vp = grid::viewport(layout = grid::grid.layout(ncol = length(wsu),
												   nrow = length(hsu),
												   widths = wsu,
												   heights = hsu))

	grobBG = if (getOption("tmap.design.mode")) rectGrob(gp=gpar(fill="#CAB2D6")) else NULL

	hsize = 2

	g = ggplot2::ggplot(comp$tab, ggplot2::aes(x = hsize, y = freq, fill = color)) +
		ggplot2::geom_bar(stat = "identity", width = 1, color = "#000000", linewidth = lwd_to_mm(scale)) +
		ggplot2::coord_polar(theta = "y", start = 0) +
		#ggplot2::theme_void() +
		#ggplot2::geom_text(ggplot2::aes(label = scales::percent(Percent), y = Count/2), position = ggplot2::position_stack(vjust = 0.5)) +
		ggplot2::scale_fill_manual(values = comp$pal) +
		theme_chart(plot.grid.y = FALSE, scale = scale, text.color = o$chart.text.color, text.size = textsize) +
		ggplot2::xlim(c(0, hsize + 0.5))
		#ggplot2::theme(
	#		legend.position = "right",
#			legend.title = ggplot2::element_blank()
#		)


	g2 = ggplot2::ggplotGrob(g)

	# other grid cells are aligns (1 and 5) and margins (2 and 4)
	chart = gridCell(3,3, {
		gTree(children=gList(grobBG,
							 g2),
			  name="compass")
	})

	grid::grobTree(chart, vp = vp)
}




#' @method tmapGridCompPlot tm_chart_violin
#' @export
tmapGridCompPlot.tm_chart_violin = function(comp, o, fH, fW) {
	scale = o$scale * comp$scale
	textsize = o$chart.text.size * scale

	wsu = comp$wsu
	hsu = comp$hsu

	vp = grid::viewport(layout = grid::grid.layout(ncol = length(wsu),
												   nrow = length(hsu),
												   widths = wsu,
												   heights = hsu))

	grobBG = if (getOption("tmap.design.mode")) rectGrob(gp=gpar(fill="#CAB2D6")) else NULL

	hsize = 2

	g = ggplot2::ggplot(comp$df, ggplot2::aes(x = x, y = y)) +
		ggplot2::geom_violin(fill = comp$object.color) +
		theme_chart(plot.axis.x = comp$plot.axis.x, plot.axis.y = comp$plot.axis.y, scale = scale, text.color = o$chart.text.color, text.size = textsize)

	g2 = ggplot2::ggplotGrob(g)

	# other grid cells are aligns (1 and 5) and margins (2 and 4)
	chart = gridCell(3,3, {
		gTree(children=gList(grobBG,
							 g2),
			  name="compass")
	})

	grid::grobTree(chart, vp = vp)
}


#' @method tmapGridCompPlot tm_chart_box
#' @export
tmapGridCompPlot.tm_chart_box = function(comp, o, fH, fW) {
	scale = o$scale * comp$scale
	textsize = o$chart.text.size * scale

	wsu = comp$wsu
	hsu = comp$hsu

	vp = grid::viewport(layout = grid::grid.layout(ncol = length(wsu),
												   nrow = length(hsu),
												   widths = wsu,
												   heights = hsu))

	grobBG = if (getOption("tmap.design.mode")) rectGrob(gp=gpar(fill="#CAB2D6")) else NULL

	hsize = 2

	g = ggplot2::ggplot(comp$df, ggplot2::aes(x = x, y = y)) +
		ggplot2::geom_boxplot(fill = comp$object.color) +
		theme_chart(plot.axis.x = comp$plot.axis.x, plot.axis.y = comp$plot.axis.y, scale = scale, text.color = o$chart.text.color, text.size = textsize)

	g2 = ggplot2::ggplotGrob(g)

	# other grid cells are aligns (1 and 5) and margins (2 and 4)
	chart = gridCell(3,3, {
		gTree(children=gList(grobBG,
							 g2),
			  name="compass")
	})

	grid::grobTree(chart, vp = vp)
}


#' @method tmapGridCompPlot tm_chart_heatmap
#' @export
tmapGridCompPlot.tm_chart_heatmap = function(comp, o, fH, fW) {
	u = 1/(comp$nlines)
	#vpComp = viewport(x=u, y=u, height=1-2*u, width=1-2*u, just=c("left", "bottom"))
	scale = o$scale * comp$scale

	textsize = o$chart.text.size * scale

	wsu = comp$wsu
	hsu = comp$hsu

	vp = grid::viewport(layout = grid::grid.layout(ncol = length(wsu),
												   nrow = length(hsu),
												   widths = wsu,
												   heights = hsu))

	grobBG = if (getOption("tmap.design.mode")) rectGrob(gp=gpar(fill="#CAB2D6")) else NULL

	g = ggplot2::ggplot(comp$tab, ggplot2::aes(x = bin2, y = bin1, fill = freq)) +
		ggplot2::geom_tile() +
		ggplot2::scale_fill_distiller(type = "seq",
							  direction = 1,
							  palette = "Greys") +
		theme_chart(plot.axis.x = TRUE, plot.axis.y = TRUE, scale = scale, text.color = o$chart.text.color, text.size = textsize, plot.grid.x = FALSE, plot.grid.y = FALSE) #comp$plot.axis.x

	g = g + comp$extra.ggplot2

	g2 = ggplot2::ggplotGrob(g)

	# other grid cells are aligns (1 and 5) and margins (2 and 4)
	histogram = gridCell(3,3, {
		gTree(children=gList(grobBG,
							 g2),
			  name="compass")
	})

	grid::grobTree(histogram, vp = vp)
}

# To generate all options:
# args = intersect(names(tmap_options_mode("view")),names(tmap_options_mode("plot")))
# cat(paste(args, collapse = ", "))
# x = sapply(args, function(a) paste0("#' @param ", a, " `r .doc_opt(\"", a, "\")`"))
# cat(paste(x, collapse = "\n"))
#
# To generate param records:
# cat(paste(sapply(names, function(x) {
#	paste0("#' @param ", x, " `r .doc_opt(\"", x ,"\")`")
# }), collapse = "\n"))
# To generate option topics (the 'what', last part)
# nms = names(tmap_options())
# nms_s = strsplit(nms, split = ".", fixed = TRUE)
# nms_lst = sapply(nms_s, tail, 1)
# whats = sort(unique(nms_lst))
# cat(paste0("c(", paste(sapply(whats, function(w) paste0(w, " = \"", w, "\"")), collapse = ",\n"), ")"))
#' tmap options
#'
#' Get or set the tmap options globally. For map specific options, we recommend to use [tm_options()] or [tm_layout()] via which the layout-related options can be set. [tmap_options()] functions similar to [base::options()].
#'
#' @param ... List of tmap options to be set, or option names (characters) to be returned (see details)
#' @param crs Map crs (see [tm_shape()]). `NA` means the crs is specified in [tm_shape()]. The crs that is used by the transformation functions is defined in [tm_shape()].
#' @param facet.max Maximum number of facets
#' @param free.scales For backward compatibility: if this value is set, it will be used to impute the free arguments in the layer functions
#' @param raster.max_cells Maximum number of raster grid cells. Can be mode specific `c(plot = 3000, view = 1000, 1000)` (the last value is the fall back default)
#' @param raster.warp Should rasters be warped or transformed in case a different projection (crs) is used? Warping creates a new regular raster in the target crs, whereas transforming creates a (usually non-regular) raster in the target crs. The former is lossy, but much faster and is therefore the default.
#' When a different projection (crs) is used, a (usually) regular raster will be
#' @param show.messages Show messages?
#' @param show.warnings Show warnings?
#' @param output.format Output format
#' @param output.size Output size
#' @param output.dpi Output dpi
#' @param animation.dpi Output dpi for animations
#' @param value.const Default visual value constants e.g. the default fill color for `tm_shape(World) + tm_polygons()`. A list is required with per visual variable a value.
#' @param value.na Default visual values that are used to visualize NA data values. A list is required with per visual variable a value.
#' @param value.null Default visual values that are used to visualize null (out-of-scope) data values. A list is required with per visual variable a value.
#' @param value.blank Default visual values that correspond to blank. For color these are `"#00000000"` meaning transparent. A list is required with per visual variable a value.
#' @param values.var Default values when a data variable to mapped to a visual variable, e.g. a color palette. A list is required with per visual variable a value.
#' @param values.range Default range for values. See `values.range` of [tm_scale_categorical()]. A list is required with per visual variable a value.
#' @param value.neutral Default values for when a data variable to mapped to a visual variable, e.g. a color palette. A list is required with per visual variable a value.
#' @param values.scale Default scales (as in object sizes) for values. See `values.range` of [tm_scale_categorical()]. A list is required with per visual variable a value.
#' @param scales.var Default scale functions per visual variable and type of data variable. A list is required with per visual variable per data type.
#' @param scale.misc.args Default values of scale function-specific arguments. A list is required with per scale function and optional per visual variable.
#' @param continuous.nclass_per_legend_break The number of continuous legend breaks within one 'unit' (label).  The default value is 50.
#' @param continuous.nclasses the number of classes of a continuous scale. Should be odd.  The default value is 101.
#' @param label.format Format for the labels. These are the default values for [tm_label_format()]
#' @param label.na Default label for missing values.
#' @param scale Overall scale of the map
#' @param asp Aspect ratio of each map. When `asp` is set to `NA` (default) the aspect ratio will be adjusted to the used shapes. When set to 0 the aspect ratio is
#'   adjusted to the size of the device divided by the number of columns and rows.
#' @param bg Draw map background?
#' @param bg.color Background color of the map.
#' @param outer.bg Draw map background (outside the frame)?
#' @param outer.bg.color Background color of map outside the frame.
#' @param frame Draw map frame?
#' @param frame.color `r .doc_opt("frame.color")`
#' @param frame.alpha `r .doc_opt("frame.alpha")`
#' @param frame.lwd `r .doc_opt("frame.lwd")`
#' @param frame.r `r .doc_opt("frame.r")`
#' @param frame.double_line `r .doc_opt("frame.double_line")`
#' @param outer.margins `r .doc_opt("outer.margins")`
#' @param inner.margins `r .doc_opt("inner.margins")`
#' @param inner.margins.extra `r .doc_opt("inner.margins.extra")`
#' @param meta.margins `r .doc_opt("meta.margins")`
#' @param meta.auto_margins `r .doc_opt("meta.auto_margins")`
#' @param between_margin Margin between the map.
#' @param panel.margin `r .doc_opt("panel.margin")`
#' @param component.offset `r .doc_opt("component.offset")`
#' @param component.stack_margin `r .doc_opt("component.stack_margin")`
#' @param xlab.show `r .doc_opt("xlab.show")`
#' @param xlab.text `r .doc_opt("xlab.text")`
#' @param xlab.size `r .doc_opt("xlab.size")`
#' @param xlab.color `r .doc_opt("xlab.color")`
#' @param xlab.rotation `r .doc_opt("xlab.rotation")`
#' @param xlab.space `r .doc_opt("xlab.space")`
#' @param xlab.fontface `r .doc_opt("xlab.fontface")`
#' @param xlab.fontfamily `r .doc_opt("xlab.fontfamily")`
#' @param xlab.alpha `r .doc_opt("xlab.alpha")`
#' @param xlab.side `r .doc_opt("xlab.side")`
#' @param ylab.show `r .doc_opt("ylab.show")`
#' @param ylab.text `r .doc_opt("ylab.text")`
#' @param ylab.size `r .doc_opt("ylab.size")`
#' @param ylab.color `r .doc_opt("ylab.color")`
#' @param ylab.rotation `r .doc_opt("ylab.rotation")`
#' @param ylab.space `r .doc_opt("ylab.space")`
#' @param ylab.fontface `r .doc_opt("ylab.fontface")`
#' @param ylab.fontfamily `r .doc_opt("ylab.fontfamily")`
#' @param ylab.alpha `r .doc_opt("ylab.alpha")`
#' @param ylab.side `r .doc_opt("ylab.side")`
#' @param panel.type `r .doc_opt("panel.type")`
#' @param panel.wrap.pos The panel positions for wrapped facets created with [tm_facets_grid()]. One of `"left"`, `"right"`, `"top"` (default) or `"bottom"`.
#' @param panel.xtab.pos The panel positions for grid facets created with [tm_facets_grid()]. Vector of two, where the first determines the locations of row panels (`"left"` or `"right"`) and the second the location of column panels ( `"top"` or `"bottom")
#' @param unit Unit of the coordinate
#' @param color.sepia_intensity `r .doc_opt("color.sepia_intensity")`
#' @param color.saturation `r .doc_opt("color.saturation")`
#' @param color_vision_deficiency_sim Color vision deficiency simulation. Either `"protan"`, `"deutan"`, or `"tritan"`.
#' @param text.fontface `r .doc_opt("text.fontface")`
#' @param text.fontfamily `r .doc_opt("text.fontfamily")`
#' @param r `r .doc_opt("r")`
#' @param component.position `r .doc_opt("component.position")`
#' @param component.autoscale `r .doc_opt("component.autoscale")`
#' @param legend.show `r .doc_opt("legend.show")`
#' @param legend.orientation `r .doc_opt("legend.orientation")`
#' @param legend.position `r .doc_opt("legend.position")`
#' @param legend.width `r .doc_opt("legend.width")`
#' @param legend.height `r .doc_opt("legend.height")`
#' @param legend.stack `r .doc_opt("legend.stack")`
#' @param legend.reverse `r .doc_opt("legend.reverse")`
#' @param legend.na.show `r .doc_opt("legend.na.show")`
#' @param legend.title.color `r .doc_opt("legend.title.color")`
#' @param legend.title.size `r .doc_opt("legend.title.size")`
#' @param legend.title.fontface `r .doc_opt("legend.title.fontface")`
#' @param legend.title.fontfamily `r .doc_opt("legend.title.fontfamily")`
#' @param legend.title.alpha `r .doc_opt("legend.title.alpha")`
#' @param legend.xlab.color `r .doc_opt("legend.xlab.color")`
#' @param legend.xlab.size `r .doc_opt("legend.xlab.size")`
#' @param legend.xlab.fontface `r .doc_opt("legend.xlab.fontface")`
#' @param legend.xlab.fontfamily `r .doc_opt("legend.xlab.fontfamily")`
#' @param legend.xlab.rot `r .doc_opt("legend.xlab.rot")`
#' @param legend.xlab.alpha `r .doc_opt("legend.xlab.alpha")`
#' @param legend.ylab.color `r .doc_opt("legend.ylab.color")`
#' @param legend.ylab.size `r .doc_opt("legend.ylab.size")`
#' @param legend.ylab.fontface `r .doc_opt("legend.ylab.fontface")`
#' @param legend.ylab.fontfamily `r .doc_opt("legend.ylab.fontfamily")`
#' @param legend.ylab.rot `r .doc_opt("legend.ylab.rot")`
#' @param legend.ylab.alpha `r .doc_opt("legend.ylab.alpha")`
#' @param legend.text.color `r .doc_opt("legend.text.color")`
#' @param legend.text.size `r .doc_opt("legend.text.size")`
#' @param legend.text.fontface `r .doc_opt("legend.text.fontface")`
#' @param legend.text.fontfamily `r .doc_opt("legend.text.fontfamily")`
#' @param legend.text.alpha `r .doc_opt("legend.text.alpha")`
#' @param legend.frame `r .doc_opt("legend.frame")`
#' @param legend.frame.lwd `r .doc_opt("legend.frame.lwd")`
#' @param legend.frame.r `r .doc_opt("legend.frame.r")`
#' @param legend.bg.color `r .doc_opt("legend.bg.color")`
#' @param legend.bg.alpha `r .doc_opt("legend.bg.alpha")`
#' @param legend.only Should only legends be printed (so without map)?
#' @param legend.absolute_fontsize `r .doc_opt("legend.absolute_fontsize")`
#' @param legend.settings.portrait `r .doc_opt("legend.settings.portrait")`
#' @param legend.settings.landscape `r .doc_opt("legend.settings.landscape")`
#' @param chart.show `r .doc_opt("chart.show")`
#' @param chart.plot.axis.x `r .doc_opt("chart.plot.axis.x")`
#' @param chart.plot.axis.y `r .doc_opt("chart.plot.axis.y")`
#' @param chart.position `r .doc_opt("chart.position")`
#' @param chart.width `r .doc_opt("chart.width")`
#' @param chart.height `r .doc_opt("chart.height")`
#' @param chart.stack `r .doc_opt("chart.stack")`
#' @param chart.reverse `r .doc_opt("chart.reverse")`
#' @param chart.na.show `r .doc_opt("chart.na.show")`
#' @param chart.title.color `r .doc_opt("chart.title.color")`
#' @param chart.title.size `r .doc_opt("chart.title.size")`
#' @param chart.title.fontface `r .doc_opt("chart.title.fontface")`
#' @param chart.title.fontfamily `r .doc_opt("chart.title.fontfamily")`
#' @param chart.title.alpha `r .doc_opt("chart.title.alpha")`
#' @param chart.xlab.color `r .doc_opt("chart.xlab.color")`
#' @param chart.xlab.size `r .doc_opt("chart.xlab.size")`
#' @param chart.xlab.fontface `r .doc_opt("chart.xlab.fontface")`
#' @param chart.xlab.fontfamily `r .doc_opt("chart.xlab.fontfamily")`
#' @param chart.xlab.alpha `r .doc_opt("chart.xlab.alpha")`
#' @param chart.ylab.color `r .doc_opt("chart.ylab.color")`
#' @param chart.ylab.size `r .doc_opt("chart.ylab.size")`
#' @param chart.ylab.fontface `r .doc_opt("chart.ylab.fontface")`
#' @param chart.ylab.fontfamily `r .doc_opt("chart.ylab.fontfamily")`
#' @param chart.ylab.alpha `r .doc_opt("chart.ylab.alpha")`
#' @param chart.text.color `r .doc_opt("chart.text.color")`
#' @param chart.text.size `r .doc_opt("chart.text.size")`
#' @param chart.text.fontface `r .doc_opt("chart.text.fontface")`
#' @param chart.text.fontfamily `r .doc_opt("chart.text.fontfamily")`
#' @param chart.text.alpha `r .doc_opt("chart.text.alpha")`
#' @param chart.frame `r .doc_opt("chart.frame")`
#' @param chart.frame.lwd `r .doc_opt("chart.frame.lwd")`
#' @param chart.frame.r `r .doc_opt("chart.frame.r")`
#' @param chart.bg.color `r .doc_opt("chart.bg.color")`
#' @param chart.bg.alpha `r .doc_opt("chart.bg.alpha")`
#' @param chart.object.color `r .doc_opt("chart.object.color")`
#' @param title.size `r .doc_opt("title.size")`
#' @param title.color `r .doc_opt("title.color")`
#' @param title.fontface `r .doc_opt("title.fontface")`
#' @param title.fontfamily `r .doc_opt("title.fontfamily")`
#' @param title.alpha `r .doc_opt("title.alpha")`
#' @param title.padding `r .doc_opt("title.padding")`
#' @param title.frame `r .doc_opt("title.frame")`
#' @param title.frame.lwd `r .doc_opt("title.frame.lwd")`
#' @param title.frame.r `r .doc_opt("title.frame.r")`
#' @param title.position `r .doc_opt("title.position")`
#' @param title.width `r .doc_opt("title.width")`
#' @param credits.size `r .doc_opt("credits.size")`
#' @param credits.color `r .doc_opt("credits.color")`
#' @param credits.fontface `r .doc_opt("credits.fontface")`
#' @param credits.fontfamily `r .doc_opt("credits.fontfamily")`
#' @param credits.alpha `r .doc_opt("credits.alpha")`
#' @param credits.padding `r .doc_opt("credits.padding")`
#' @param credits.position `r .doc_opt("credits.position")`
#' @param credits.width `r .doc_opt("credits.width")`
#' @param credits.height `r .doc_opt("credits.height")`
#' @param compass.north `r .doc_opt("compass.north")`
#' @param compass.type `r .doc_opt("compass.type")`
#' @param compass.text.size `r .doc_opt("compass.text.size")`
#' @param compass.size `r .doc_opt("compass.size")`
#' @param compass.show.labels `r .doc_opt("compass.show.labels")`
#' @param compass.cardinal.directions `r .doc_opt("compass.cardinal.directions")`
#' @param compass.text.color `r .doc_opt("compass.text.color")`
#' @param compass.color.dark `r .doc_opt("compass.color.dark")`
#' @param compass.color.light `r .doc_opt("compass.color.light")`
#' @param compass.lwd `r .doc_opt("compass.lwd")`
#' @param compass.margins `r .doc_opt("compass.margins")`
#' @param compass.position `r .doc_opt("compass.position")`
#' @param logo.height `r .doc_opt("logo.height")`
#' @param logo.margins `r .doc_opt("logo.margins")`
#' @param logo.between_margin `r .doc_opt("logo.between_margin")`
#' @param logo.position `r .doc_opt("logo.position")`
#' @param scalebar.breaks See [tm_scalebar()]
#' @param scalebar.width See [tm_scalebar()]
#' @param scalebar.allow_clipping See [tm_scalebar()]
#' @param scalebar.text.size `r .doc_opt("scalebar.text.size")`
#' @param scalebar.text.color `r .doc_opt("scalebar.text.color")`
#' @param scalebar.color.dark `r .doc_opt("scalebar.color.dark")`
#' @param scalebar.color.light `r .doc_opt("scalebar.color.light")`
#' @param scalebar.lwd `r .doc_opt("scalebar.lwd")`
#' @param scalebar.size `r .doc_opt("scalebar.size")`
#' @param scalebar.margins `r .doc_opt("scalebar.margins")`
#' @param scalebar.position `r .doc_opt("scalebar.position")`
#' @param grid.show `r .doc_opt("grid.show")`
#' @param grid.labels.pos `r .doc_opt("grid.labels.pos")`
#' @param grid.x `r .doc_opt("grid.x")`
#' @param grid.y `r .doc_opt("grid.y")`
#' @param grid.n.x `r .doc_opt("grid.n.x")`
#' @param grid.n.y `r .doc_opt("grid.n.y")`
#' @param grid.crs `r .doc_opt("grid.crs")`
#' @param grid.col `r .doc_opt("grid.col")`
#' @param grid.lwd `r .doc_opt("grid.lwd")`
#' @param grid.alpha `r .doc_opt("grid.alpha")`
#' @param grid.labels.show `r .doc_opt("grid.labels.show")`
#' @param grid.labels.size `r .doc_opt("grid.labels.size")`
#' @param grid.labels.col `r .doc_opt("grid.labels.col")`
#' @param grid.labels.rot `r .doc_opt("grid.labels.rot")`
#' @param grid.labels.format `r .doc_opt("grid.labels.format")`
#' @param grid.labels.cardinal `r .doc_opt("grid.labels.cardinal")`
#' @param grid.labels.margin.x `r .doc_opt("grid.labels.margin.x")`
#' @param grid.labels.margin.y `r .doc_opt("grid.labels.margin.y")`
#' @param grid.labels.space.x `r .doc_opt("grid.labels.space.x")`
#' @param grid.labels.space.y `r .doc_opt("grid.labels.space.y")`
#' @param grid.labels.inside_frame `r .doc_opt("grid.labels.inside_frame")`
#' @param grid.ticks `r .doc_opt("grid.ticks")`
#' @param grid.lines `r .doc_opt("grid.lines")`
#' @param grid.ndiscr `r .doc_opt("grid.ndiscr")`
#' @param mouse_coordinates.position `r .doc_opt("mouse_coordinates.position")`
#' @param minimap.server `r .doc_opt("minimap.server")`
#' @param minimap.toggle `r .doc_opt("minimap.toggle")`
#' @param minimap.position `r .doc_opt("minimap.position")`
#' @param panel.show `r .doc_opt("panel.show")`
#' @param panel.labels `r .doc_opt("panel.labels")`
#' @param panel.label.size `r .doc_opt("panel.label.size")`
#' @param panel.label.color `r .doc_opt("panel.label.color")`
#' @param panel.label.fontface `r .doc_opt("panel.label.fontface")`
#' @param panel.label.fontfamily `r .doc_opt("panel.label.fontfamily")`
#' @param panel.label.alpha `r .doc_opt("panel.label.alpha")`
#' @param panel.label.bg.color `r .doc_opt("panel.label.bg.color")`
#' @param panel.label.frame `r .doc_opt("panel.label.frame")`
#' @param panel.label.frame.lwd `r .doc_opt("panel.label.frame.lwd")`
#' @param panel.label.frame.r `r .doc_opt("panel.label.frame.r")`
#' @param panel.label.height `r .doc_opt("panel.label.height")`
#' @param panel.label.rot Rotation angles of the panel labels. Vector of four values that determine the panel label rotation when they are placed left, top, right, and bottom. The default angles are 90, 0, 270 and 0 respectively. Note that the second value is the most common, since labels are by default shown on top (see `panel.wrap.pos`). In cross-table facets created with `tm_facets_grid()`, the first two values are used by default (see `panel.xtab.pos`).
#' @inheritParams tm_view
#' @param qtm.scalebar `r .doc_opt("qtm.scalebar")`
#' @param qtm.minimap `r .doc_opt("qtm.minimap")`
#' @param qtm.mouse_coordinates `r .doc_opt("qtm.mouse_coordinates")`
#' @param earth_boundary The earth boundary
#' @param earth_boundary.color `r .doc_opt("earth_boundary.color")`
#' @param earth_boundary.lwd `r .doc_opt("earth_boundary.lwd")`
#' @param earth_datum  Earth datum
#' @param space Should the space be drawn? Only applicable is earth_boundary is enabled.
#' @param space.color `r .doc_opt("space.color")`
#' @param space_overlay Should the space be drawn as overlay (to make sure spatial features or rasters do not exceed the earth boundary), or as background? By default `TRUE` when a raster is warped.
#' @param check_and_fix Should attempt to fix an invalid shapefile
#' @param basemap.show `r .doc_opt("basemap.show")`
#' @param basemap.server `r .doc_opt("basemap.server")`
#' @param basemap.alpha `r .doc_opt("basemap.alpha")`
#' @param basemap.zoom `r .doc_opt("basemap.zoom")`
#' @param tiles.show `r .doc_opt("tiles.show")`
#' @param tiles.server `r .doc_opt("tiles.server")`
#' @param tiles.alpha `r .doc_opt("tiles.alpha")`
#' @param tiles.zoom `r .doc_opt("tiles.zoom")`
#' @param attr.color `r .doc_opt("attr.color")`
#' @param crs_extra Only used internally (work in progress)
#' @param crs_global The used crs for world maps
#' @param title deprecated See [tm_title()]
#' @param main.title deprecated See [tm_title()]
#' @param main.title.size,main.title.color,main.title.fontface,main.title.fontfamily,main.title.position deprecated. Use the `title.` options instead.
#' @param fontface,fontfamily renamed to `text.fontface` and `text.fontfamily`
#' @param component.resize_as_group `r .doc_opt("component.resize_as_group")`
#' @param component.frame_combine `r .doc_opt("component.frame_combine")`
#' @param component.stack `r .doc_opt("component.stack")`
#' @param component.equalize `r .doc_opt("component.equalize")`
#' @param component.frame `r .doc_opt("component.frame")`
#' @param component.frame.color `r .doc_opt("component.frame.color")`
#' @param component.frame.alpha `r .doc_opt("component.frame.alpha")`
#' @param component.frame.lwd `r .doc_opt("component.frame.lwd")`
#' @param component.frame.r `r .doc_opt("component.frame.r")`
#' @param component.bg `r .doc_opt("component.bg")`
#' @param component.bg.color `r .doc_opt("component.bg.color")`
#' @param component.bg.alpha `r .doc_opt("component.bg.alpha")`
#' @param legend.frame.color `r .doc_opt("legend.frame.color")`
#' @param legend.frame.alpha `r .doc_opt("legend.frame.alpha")`
#' @param legend.bg `r .doc_opt("legend.bg")`
#' @param add_legend.position `r .doc_opt("add_legend.position")`
#' @param chart.frame.color `r .doc_opt("chart.frame.color")`
#' @param chart.frame.alpha `r .doc_opt("chart.frame.alpha")`
#' @param chart.bg `r .doc_opt("chart.bg")`
#' @param title.frame.color `r .doc_opt("title.frame.color")`
#' @param title.frame.alpha `r .doc_opt("title.frame.alpha")`
#' @param inset.position `r .doc_opt("inset.position")`
#' @param inset_map.height `r .doc_opt("inset_map.height")`
#' @param inset_map.width `r .doc_opt("inset_map.width")`
#' @param inset_map.margins `r .doc_opt("inset_map.margins")`
#' @param inset_map.between_margin `r .doc_opt("inset_map.between_margin")`
#' @param inset_map.position `r .doc_opt("inset_map.position")`
#' @param inset_map.frame `r .doc_opt("inset_map.frame")`
#' @param inset.height `r .doc_opt("inset.height")`
#' @param inset.width `r .doc_opt("inset.width")`
#' @param inset.margins `r .doc_opt("inset.margins")`
#' @param inset.between_margin `r .doc_opt("inset.between_margin")`
#' @param inset.frame `r .doc_opt("inset.frame")`
#' @param inset.bg `r .doc_opt("inset.bg")`
#' @param inset.bg.color `r .doc_opt("inset.bg.color")`
#' @param inset.bg.alpha `r .doc_opt("inset.bg.alpha")`
#' @param inset_grob.height `r .doc_opt("inset_grob.height")`
#' @param inset_grob.width `r .doc_opt("inset_grob.width")`
#' @param inset_gg.height `r .doc_opt("inset_gg.height")`
#' @param inset_gg.width `r .doc_opt("inset_gg.width")`
#' @param scalebar.text.fontface `r .doc_opt("scalebar.text.fontface")`
#' @param scalebar.text.fontfamily `r .doc_opt("scalebar.text.fontfamily")`
#' @param grid.labels.fontface `r .doc_opt("grid.labels.fontface")`
#' @param grid.labels.fontfamily `r .doc_opt("grid.labels.fontfamily")`
#' @param panel.label.bg `r .doc_opt("panel.label.bg")`
#' @param panel.label.bg.alpha `r .doc_opt("panel.label.bg.alpha")`
#' @param panel.label.frame.color `r .doc_opt("panel.label.frame.color")`
#' @param panel.label.frame.alpha `r .doc_opt("panel.label.frame.alpha")`
#' @param crs_basemap `r .doc_opt("crs_basemap")`
#' @inheritParams tm_plot
# For bbox
#' @inheritParams tm_shape
#' @export
#' @example ./examples/tmap_options.R
#' @name tmap_options
tmap_options = function(..., crs, facet.max, free.scales, raster.max_cells, raster.warp, show.messages, show.warnings, output.format, output.size, output.dpi, animation.dpi, value.const, value.na, value.null, value.blank, values.var, values.range, value.neutral, values.scale, scales.var, scale.misc.args, continuous.nclass_per_legend_break, continuous.nclasses, label.format, label.na, scale, asp, bg, bg.color, outer.bg, outer.bg.color, frame, frame.color, frame.alpha, frame.lwd, frame.r, frame.double_line, outer.margins, inner.margins, inner.margins.extra, meta.margins, meta.auto_margins, between_margin, panel.margin, xlab.show, xlab.text, xlab.size, xlab.color, xlab.rotation, xlab.space, xlab.fontface, xlab.fontfamily, xlab.alpha, xlab.side, ylab.show, ylab.text, ylab.size, ylab.color, ylab.rotation, ylab.space, ylab.fontface, ylab.fontfamily, ylab.alpha, ylab.side, panel.type, panel.wrap.pos, panel.xtab.pos, unit, color.sepia_intensity, color.saturation, color_vision_deficiency_sim, text.fontface, text.fontfamily, r, component.position, component.offset, component.stack_margin, component.autoscale, component.resize_as_group, component.frame_combine, component.stack, legend.stack, chart.stack, component.equalize, component.frame, component.frame.color, component.frame.alpha, component.frame.lwd, component.frame.r, component.bg, component.bg.color, component.bg.alpha, legend.show, legend.orientation, legend.position, legend.width, legend.height, legend.reverse, legend.na.show, legend.title.color, legend.title.size, legend.title.fontface, legend.title.fontfamily, legend.title.alpha, legend.xlab.color, legend.xlab.size, legend.xlab.rot, legend.xlab.fontface, legend.xlab.fontfamily, legend.xlab.alpha, legend.ylab.color, legend.ylab.size, legend.ylab.rot, legend.ylab.fontface, legend.ylab.fontfamily, legend.ylab.alpha, legend.text.color, legend.text.size, legend.text.fontface, legend.text.fontfamily, legend.text.alpha, legend.frame, legend.frame.color, legend.frame.alpha, legend.frame.lwd, legend.frame.r, legend.bg, legend.bg.color, legend.bg.alpha, legend.only, legend.absolute_fontsize, legend.settings.portrait, legend.settings.landscape, add_legend.position, chart.show, chart.plot.axis.x, chart.plot.axis.y, chart.position, chart.width, chart.height, chart.reverse, chart.na.show, chart.title.color, chart.title.size, chart.title.fontface, chart.title.fontfamily, chart.title.alpha, chart.xlab.color, chart.xlab.size, chart.xlab.fontface, chart.xlab.fontfamily, chart.xlab.alpha, chart.ylab.color, chart.ylab.size, chart.ylab.fontface, chart.ylab.fontfamily, chart.ylab.alpha, chart.text.color, chart.text.size, chart.text.fontface, chart.text.fontfamily, chart.text.alpha, chart.frame, chart.frame.color, chart.frame.alpha, chart.frame.lwd, chart.frame.r, chart.bg, chart.bg.color, chart.bg.alpha, chart.object.color, title.size, title.color, title.fontface, title.fontfamily, title.alpha, title.padding, title.frame, title.frame.color, title.frame.alpha, title.frame.lwd, title.frame.r, title.position, title.width, credits.size, credits.color, credits.fontface, credits.fontfamily, credits.alpha, credits.padding, credits.position, credits.width, credits.height, compass.north, compass.type, compass.text.size, compass.size, compass.show.labels, compass.cardinal.directions, compass.text.color, compass.color.dark, compass.color.light, compass.lwd, compass.margins, compass.position, inset.position, logo.height, logo.margins, logo.between_margin, logo.position, inset_map.height, inset_map.width, inset_map.margins, inset_map.between_margin, inset_map.position, inset_map.frame, inset.height, inset.width, inset.margins, inset.between_margin, inset.frame, inset.bg, inset.bg.color, inset.bg.alpha, inset_grob.height, inset_grob.width, inset_gg.height, inset_gg.width, scalebar.breaks, scalebar.width, scalebar.allow_clipping, scalebar.text.size, scalebar.text.color, scalebar.text.fontface, scalebar.text.fontfamily, scalebar.color.dark, scalebar.color.light, scalebar.lwd, scalebar.size, scalebar.margins, scalebar.position, grid.show, grid.labels.pos, grid.x, grid.y, grid.n.x, grid.n.y, grid.crs, grid.col, grid.lwd, grid.alpha, grid.labels.show, grid.labels.size, grid.labels.col, grid.labels.fontface, grid.labels.fontfamily, grid.labels.rot, grid.labels.format, grid.labels.cardinal, grid.labels.margin.x, grid.labels.margin.y, grid.labels.space.x, grid.labels.space.y, grid.labels.inside_frame, grid.ticks, grid.lines, grid.ndiscr, mouse_coordinates.position, minimap.server, minimap.toggle, minimap.position, panel.show, panel.labels, panel.label.size, panel.label.color, panel.label.fontface, panel.label.fontfamily, panel.label.alpha, panel.label.bg, panel.label.bg.color, panel.label.bg.alpha, panel.label.frame, panel.label.frame.color, panel.label.frame.alpha, panel.label.frame.lwd, panel.label.frame.r, panel.label.height, panel.label.rot, qtm.scalebar, qtm.minimap, qtm.mouse_coordinates, earth_boundary, earth_boundary.color, earth_boundary.lwd, earth_datum, space, space.color, space_overlay, check_and_fix, basemap.show, basemap.server, basemap.alpha, basemap.zoom, tiles.show, tiles.server, tiles.alpha, tiles.zoom, attr.color, crs_extra, crs_global, crs_basemap,
						use_gradient, # plot mode
						use_browser, use_WebGL, control.position, control.bases, control.overlays, control.collapse, set_bounds, # view mode
						set_view, set_zoom_limits, use_circle_markers, leaflet.options, # view mode
						title = NULL,
						main.title = NULL,
						main.title.size = NULL,
						main.title.color = NULL,
						main.title.fontface = NULL,
						main.title.fontfamily = NULL,
						main.title.position = NULL,
						fontface = NULL,
						fontfamily = NULL
						) {
	o = get("tmapOptions", envir = .TMAP)
	nms = names(o)
	show.warnings = o$show.warnings

	## case 1: option list is given. E.g. opts = list(crs = 3035, panel.show = TRUE); tmap_options(opts)
	## case 2: option name is given. E.g. tmap_options("crs", "panel.show")
	## case 3: named options are set. E.g. tmap_options(crs = 3035, panel.show = TRUE)
	## case 4: tmap_options is called without arguments


	# get current style name (default: white), and set new style name (with "(modified)")
	sty_cur = getOption("tmap.style")
	sty_new = if (substr(sty_cur, nchar(sty_cur) - 9, nchar(sty_cur)) == "(modified)") sty_cur else paste(sty_cur, "(modified)")

	args = lapply(as.list(rlang::call_match(dots_expand = TRUE)[-1]), eval, envir = parent.frame())
	set_new_style = FALSE

	#lst = list(...)
	if (length(args) >= 1 && is.null(names(args[1]))) {
		arg = args[[1]]
		if (is.list(arg)) {
			if (length(args) > 1 && show.warnings) warning("Only the first argument is used; the other arguments are ignored.")
			## case 1: option list is given
			args = arg

			style_attr = attr(args, "style")
			if (!is.null(style_attr)) {
				sty_new = style_attr
				set_new_style = TRUE
			}

		} else {
			## case 2: option name is given
			args = sapply(args, "[", 1)
			if (!all(args %in% nms) && show.warnings) warning("The following options do not exist: ", paste(setdiff(args, nms), collapse = ", "))
			args = intersect(args, nms)
			return(o[args])
		}
	}

	mode_opts = setdiff(unique(unlist(lapply(o$modes, names))), "name")

	all_opts = union(mode_opts, names(.TMAP$defaultTmapOptions))

	unknown_args = setdiff(names(args), all_opts)

	fun = "tmap_options"
	if ("view.legend.position" %in% unknown_args) {
		args$legend.position = args$view.legend.position
		args$view.legend.position = NULL
		v3_opt(fun, "view.legend.position", "legend.position")
	}
	if ("set.bounds" %in% unknown_args) {
		args$set_bounds = args$set.bounds
		args$set.bounds = NULL
		v3_opt(fun, "set.bounds", "set_bounds")
	}
	if ("set.view" %in% unknown_args) {
		args$set_view = args$set.view
		args$set.view = NULL
		v3_opt(fun, "set.view", "set_view")
	}
	if ("set.zoom.limits" %in% unknown_args) {
		args$set_zoom_limits = args$set.zoom.limits
		args$set.zoom.limits = NULL
		v3_opt(fun, "set.zoom.limits", "set_zoom_limits")
	}
	if ("max.raster" %in% unknown_args) {
		args$raster.max_cells = args$max.raster
		args$max.raster = NULL
		v3_opt(fun, "max.raster", "raster.max_cells")
	}
	if ("fontfamily" %in% unknown_args) {
		args$text.fontfamily = args$fontfamily
		args$fontfamily = NULL
		v3_opt(fun, "fontfamily", "text.fontfamily")
	}
	if ("fontface" %in% unknown_args) {
		args$text.fontface = args$fontface
		args$fontface = NULL
		v3_opt(fun, "fontface", "text.fontface")
	}
	if ("overlays" %in% unknown_args) {
		args$tiles.server = args$overlays
		args$overlays = NULL
		v3_opt(fun, "overlays", "tiles.server")
	}
	if ("basemaps" %in% unknown_args) {
		args$basemap.server = args$basemaps
		args$basemaps = NULL
		v3_opt(fun, "basemaps", "basemap.server")
	}


	unknown_args = setdiff(names(args), all_opts)
	if (length(unknown_args) == 1) {
		stop("the following option does not exist: ", unknown_args)
	} else if (length(unknown_args) > 1) {
		stop("the following options do not exist: ", paste(unknown_args, collapse = ", "))
	}

	if (!length(args)) {
		# case 4
		return(o)
	} else {
		# case 1 and 3
		if (is.null(names(args))) {
			stop("Option list is unnamed")
		}
		backup = o[names(args)]
		o[names(args)] = complete_options(args, backup) # needed to impute position and value(s) args


		options(tmap.style=sty_new)
		attr(o, "style") = sty_new
		attr(o, "specified") = names(args)
		assign("tmapOptions", o, envir = .TMAP)

		if (set_new_style) {
			if (o$show.messages) message("tmap options successfully loaded as style \"", sty_new, "\"")
			styles = get("tmapStyles", envir = .TMAP)
			styles[[sty_new]] = suppressMessages(tmap_options_diff())
			assign("tmapStyles", styles, envir = .TMAP)
		}
		invisible(backup)
	}
}

#' @param mode mode, e.g. `"plot"` or `"view"`
#' @param style style. If specified, the style specific options are returned
#' @param mode.specific Should only mode-specific options be returned? `TRUE` by default.
#' @param default.options return the default options or the current options?
#' @rdname tmap_options
#' @export
tmap_options_mode = function(mode = NA, style = NULL, mode.specific = TRUE, default.options = FALSE) {
	if (default.options) {
		o = .TMAP$defaultTmapOptions
	} else {
		o = get("tmapOptions", envir = .TMAP)

		if (!is.null(style)) {
			check_style(style)
			s = get("tmapStyles", envir = .TMAP)[[style]]
			o = complete_options(s, o)
		}
	}

	if (is.na(mode)) mode = getOption("tmap.mode")
	opt2 = o$modes[[mode]]

	specified = attr(o, "specified")

	# all general options with mode-specific defaults (except the specified ones tmap_options(bg.color = "red")
	int_opt = setdiff(intersect(names(o), names(opt2)), specified)

	# all mode-specific options
	diff_opt = setdiff(names(opt2), names(o))

	if (length(int_opt)) o[int_opt] = opt2[int_opt]
	if (length(diff_opt)) o = c(o, opt2[diff_opt])

	o$modes = NULL

	if (mode.specific) {
		o[setdiff(names(opt2), "name")]
	} else {
		o
	}
}


tmap_option = function(name, type = NULL) {
	get_option_class(tmap_options()[[name]], class = type, spatial_class = FALSE)
}



tmap_graphics_name = function(mode) {
	if (missing(mode)) mode = getOption("tmap.mode")
	get("tmapOptions", envir = .TMAP)$modes[[mode]]$name
}



tmapOption = function(...) {
	structure(list(...), class = "tmapOption")
}




#' Internal tmap function to add a default value for the layer functions. Deprecated
#'
#' Internal tmap function to add a default value for the layer functions. Deprecated. Use [tmapSubmitOptions()] instead.
#'
#' @param option, one of: `"value.const"`, `"value.na"`, `"value.blank"`, `"values.var"`, `'values.range'`, `"value.neutral"`, `"scales.var"`
#' @param id name of the visual variable with layer, in the format `"x.y"`,
#'   where `x` is the visual variable and `y` is the layer.
#'   It is also possible to set `x` only; then it applies to all layer functions.
#' @param value value
#' @keywords internal
#' @export
tmapAddLayerOptions = function(option, id, value) {
	if (!(option %in% c("value.const", "value.na", "value.blank", "values.var", "values.range", "value.neutral", "scales.var"))) {
		stop("Unknown option")
	}

	d = .TMAP$defaultTmapOptions
	d[[option]][[id]] = value

	o = .TMAP$tmapOptions
	o[[option]][[id]] = value

	.TMAP$defaultTmapOptions = d
	.TMAP$tmapOptions = o
}

#' @rdname tmap_options
#' @export
tmap_options_diff <- function() {
	.tmapOptions <- get("tmapOptions", envir = .TMAP)
	iden <- mapply(identical, .tmapOptions, .TMAP$defaultTmapOptions)

	if (all(iden)) {
		message("current tmap options are similar to the default tmap options (style \"white\")")
	} else {
		message("current tmap options (style \"", attr(.tmapOptions, "style"), "\") that are different from default tmap options (style \"white\"):")
		.tmapOptions[!iden]
	}
}

#' @rdname tmap_options
#' @export
tmap_options_reset <- function() {
	assign("tmapOptions", .TMAP$defaultTmapOptions, envir = .TMAP)
	options(tmap.style="white")
	message("tmap options successfully reset")
	invisible(NULL)
}

#' @export
#' @param style style, see [tmap_style()] for available styles
#' @rdname tmap_options
tmap_options_save <- function(style) {
	show.messages <- get("tmapOptions", envir = .TMAP)$show.messages

	stylediff <- suppressMessages(tmap_options_diff())

	.tmapOptions <- get("tmapOptions", envir = .TMAP)

	if (is.null(stylediff)) {
		if (show.messages) message("current style is the same as the default style, so nothing to save")
		return(invisible(.tmapOptions))
	}

	options(tmap.style=style)
	attr(.tmapOptions, "style") <- style
	assign("tmapOptions", .tmapOptions, envir = .TMAP)

	styles <- get("tmapStyles", envir = .TMAP)
	styles[[style]] <- suppressMessages(tmap_options_diff())
	assign("tmapStyles", styles, envir = .TMAP)

	if (show.messages) message("current tmap options saved as style \"", style, "\"")
	invisible(.tmapOptions)
}

#' Internal method for submitting options
#'
#' Internal method for submitting options
#'
#' @param options list of options
#' @param styleOptions list of lists of options, each with a style name
#' @export
#' @keywords internal
tmapSubmitOptions = function(options = NULL, styleOptions = NULL) {
	o = .TMAP$defaultTmapOptions
	s = .TMAP$tmapStyles

	if (!is.null(options)) {
		backup = o[names(options)]
		o[names(options)] = complete_options(options, backup, erase_style = FALSE)
	}

	if (!is.null(styleOptions)) {
		for (st in names(styleOptions)) {
			backup = s[[st]]
			s[[st]] = complete_options(styleOptions[[st]], backup, erase_style = FALSE)
		}
	}
	.TMAP$defaultTmapOptions = o
	.TMAP$tmapStyles = s
	suppressMessages(tmap_options_reset())
}

#' Internal method for submitting a new mode (deprecated)
#'
#' Internal method for submitting a new mode. Deprecated. Use [tmapSubmitOptions()] instead.
#'
#' @param id id of the mode: please use lowercase and one-word. This will be used with [tmap_mode()].
#' @param name name of the mode: please use title case. This will be used to recognize internal functions, e.g. `tmapLeafletInit`
#' @param ... mode specific options
#' @export
#' @keywords internal
tmapMode = function(id, name, ..., styleOptions = NULL) {
	d = .TMAP$defaultTmapOptions
	o = .TMAP$tmapOptions

	mlist = c(list(name = name), list(...))

	o$modes[[id]] = mlist
	d$modes[[id]] = mlist

	if (!is.null(styleOptions)) {
		os = get("tmapStyles", envir = .TMAP)
		for (s in names(styleOptions)) {
			os[[s]]$modes[[id]] = styleOptions[[s]]
			#lst = structure(list(modes = structure(styleOptions[s], names = id)), style = s)
			#suppressMessages(tmap_options(lst))
		}
		assign("tmapStyles", os, envir = .TMAP)
	}
	.TMAP$defaultTmapOptions = d
	.TMAP$tmapOptions = o
	suppressMessages(tmap_style(getOption("tmap.style")))
}

#' Map layer: iso (contour)
#'
#' Map layer that draws iso (contour) lines. Stack of [tm_lines()] and [tm_labels_highlighted].
#'
#' @param col `r .doc_vv("col")` `r .doc_unit("col")`
#' @param text `r .doc_vv("text")` `r .doc_unit("text")`
#' @param ... passed on to [tm_lines()] and [tm_labels_highlighted()]. For the text color and alpha transparency of the text labels, please use `text_col` and `text_alpha` instead of `col` and `col_alpha`.
#' @param options_lines The options for [tm_lines()]
#' @param options_labels The options for [tm_labels_highlighted()]
#' @export
tm_iso = function(col = tm_const(),
				  text = tm_vars(x = 1),
				  ...,
				  options_lines = opt_tm_lines(),
				  options_labels = opt_tm_labels()) {
	args = list(...)
	args$options = NULL
	argsL = args[intersect(names(formals("tm_lines")), names(args))]

	nms_col = c("col", "col.scale", "col.legend", "col.chart", "col.free", "col_alpha", "col_alpha.scale", "col_alpha.legend", "col_alpha.chart", "col_alpha.free")

	# update args for labels
	args[nms_col] = NULL
	rename_id = names(args) %in% paste0("text_", nms_col)

	names(args)[rename_id] = substr(names(args)[rename_id], 6, nchar(names(args)[rename_id]))
	argsT = args[intersect(names(formals("tm_labels_highlighted")), names(args))]

	do.call("tm_lines", c(list(col=col, options = options_lines), argsL)) +
	do.call("tm_labels_highlighted", c(list(text=text, options = options_labels), argsT))
}

#' @export
#' @keywords internal
#' @rdname tmapGridLeaflet
tmapGridAuxPrepare.tm_aux_basemap = function(a, bs, id, o) {
	tmapGridAuxPrepare.tm_aux_tiles(a, bs, id, o)
}

#' @export
#' @keywords internal
#' @rdname tmapGridLeaflet
tmapGridAuxPrepare.tm_aux_tiles = function(a, bs, id, o) {
	g = get("g", envir = .TMAP_GRID)
	rlang::check_installed("maptiles")

	crs = sf::st_crs(bs[[1]])

	crs3857 = (crs == 3857 || crs == st_crs(3857))

	isproj = !sf::st_is_longlat(crs)

	bs_orig = bs
	if (isproj) {
		# plain lat-lon to find zoom levels
		bs = lapply(bs, function(b) {
			sf::st_bbox(sf::st_transform(sf::st_as_sfc(b), crs = "EPSG:4326"))
		})
	}

	# tiles are in mercator
	bs3857 = lapply(bs_orig, function(b) {
		sf::st_bbox(sf::st_transform(tmaptools::bb_poly(b), crs = "EPSG:3857"))
	})

	bs = lapply(bs, function(b) {
		# not sure why needed
		bb_ll_valid(bb_asp(b, g$fasp))
	})

	if (is.na(a$zoom)) {
		zs = vapply(bs, findZoom, FUN.VALUE = integer(1))
	} else {
		zs = rep(a$zoom, length(bs))
	}

	serv = a$server[1]

	if (substr(serv, 1, 4) == "http") {
		if (serv %in% .TMAP_GRID$maptiles_urls) {
			tile_id = which(serv == .TMAP_GRID$maptiles_urls)[1]
		} else {
			tile_id = length(.TMAP_GRID$maptiles_urls) + 1L
			.TMAP_GRID$maptiles_urls = c(.TMAP_GRID$maptiles_urls, serv)
		}
		sub = if (is.na(a$sub)) NA else strsplit(a$sub, split = "")[[1]]
		serv = maptiles::create_provider(paste0("id_", tile_id), url = serv, citation = "", sub = sub)
		api = if (!is.null(a$api)) {
			a$api
		} else NULL
	} else if (substr(serv, 1, 6) %in% c("Thunde", "Stadia")) {
		is_stadia = substr(serv, 1, 6) == "Stadia"
		api = if (!is.null(a$api)) {
			a$api
		} else if (is_stadia) {
			Sys.getenv("STADIA_MAPS")
		} else {
			Sys.getenv("THUNDERFOREST_MAPS")
		}
		if (api == "") message_basemaps(is_stadia)
	} else {
		if (!(serv %in% tmap_providers("plot"))) {
			eq = basemap_equivalent(serv, "plot")
			if (!is.na(eq)) {
				message_basemaps_equivalent(serv, "plot", eq)
				serv = eq
			} else {
				message_basemaps_invalid_provider(serv, "plot", o$basemap.server[1])
				serv = o$basemap.server[1]
			}
		}
		api = NULL
	}

	xs = mapply(function(b, z) {
		m = tryCatch({
			if (is.null(api)) {
				maptiles::get_tiles(x = b, provider = serv, zoom = z, crop = TRUE)
			} else {
				maptiles::get_tiles(x = b, provider = serv, apikey = api, zoom = z, crop = TRUE)
			}
		}, error = function(e) {
			tryCatch({
				if (is.null(api)) {
					maptiles::get_tiles(x = b, provider = serv, zoom = z - 1, crop = TRUE)
				} else {
					maptiles::get_tiles(x = b, provider = serv, apikey = api, zoom = z - 1, crop = TRUE)
				}
			}, error = function(e) {
				NULL
			})
		})
		if (!is.null(m)) {
			names(m)[1:3] = c("red", "green", "blue")
			if (terra::nlyr(m) == 4) names(m)[4] = "alpha"

		} else {
			message_basemaps_none(serv, z)
		}
		m
	}, bs3857, zs, SIMPLIFY = FALSE)

	if (isproj && !crs3857) {
		if (!all(vapply(xs, is.null, FUN.VALUE = logical(1)))) {
			message_basemaps_blurry(serv)
			xs = mapply(function(x,b) {
				if (is.null(x)) return(NULL)

				ex = terra::ext(as.vector(b[c(1,3,2,4)]))
				asp = (ex[2] - ex[1]) / (ex[4] - ex[3])

				tot = terra::ncell(x) * 2

				nc = round(sqrt(tot * asp))
				nr = round(tot / nc)

				r = terra::rast(ex, nrows = nr, ncols = nc, crs = crs$wkt)
				terra::project(x, r, method = "near")
			}, xs, bs_orig, SIMPLIFY = FALSE)
		}
	}

	ss = lapply(xs, function(x) {
		if (is.null(x)) NULL else do.call(tmapShape, list(shp = x, is.main = FALSE, crs = crs, bbox = NULL, unit=NULL, filter=NULL, layer = NULL, shp_name = "x", smeta = list(), o = o, tmf = NULL))
	})

	srgb = tm_scale_rgb(max_color_value = 255, value.na = "#FFFFFF")


	ds = lapply(ss, function(s) {
		if (is.null(s)) return(NULL)
		d = s$dt
		d[, c("col", "legnr", "crtnr") := do.call(srgb$FUN, list(x1 = red, x2 = green, x3 = blue, scale = srgb, legend = list(), o = o, aes = "col", layer = "raster", layer_args = opt_tm_rgb(interpolate = TRUE)$mapping.args, sortRev = NA, bypass_ord = TRUE))]
		if ("alpha" %in% names(d)) {
			d[, col_alpha:=alpha/255 * a$alpha]
			d[is.na(col_alpha), col_alpha:=0]
		} else {
			d[, col_alpha:=a$alpha]
		}
		d
	})

	shpTMs = lapply(ss, function(s) {
		if (is.null(s)) NULL else s$shpTM
	})


	bmaps_shpTHs = structure(list(shpTMs), names = id)
	bmaps_dts = structure(list(ds), names = id)

	g$bmaps_shpTHs = c(g$bmaps_shpTHs, bmaps_shpTHs)
	g$bmaps_dts = c(g$bmaps_dts, bmaps_dts)

	assign("g", g, envir = .TMAP_GRID)
	paste0(a$server, collapse = "__")
}

#' @export
#' @rdname tmapGridLeaflet
tmapLeafletDataPlot.tm_data_raster = function(a, shpTM, dt, pdt, popup.format, hdt, idt, gp, bbx, facet_row, facet_col, facet_page, id, pane, group, glid, o, ...) {

	rc_text = frc(facet_row, facet_col)


	shp = shpTM$shp
	tmapID = shpTM$tmapID

	if (is_regular_grid(shp)) {

		tid = intersect(tmapID, dt$tmapID__)

		color = rep(NA, length(tmapID)) # NA

		sel = which(tmapID %in% tid)
		tid2 = tmapID[sel]

		color[sel] = dt$col[match(tid2, dt$tmapID__)]


		#color = rep("#FFFFFF", length(tmapID))
		#color[match(dt$tmapID__, tmapID)] = dt$color

		pal <- na.omit(unique(color))
		pal <- pal[substr(pal, 8,10)!="00"] ## remove transparant colors

		if (!length(pal)) return(NULL)

		res <- split_alpha_channel(pal, alpha = 1)
		pal_col <- res$col
		pal_opacity <- if (length(res$opacity) == 0L) 0 else max(res$opacity)

		if ("col_alpha" %in% names(dt)) pal_opacity = max(dt$col_alpha)


		col_ids <- match(color, pal)

		m <- matrix(col_ids, ncol = ncol(shp))

		#matrix(color, ncol = ncol(shp))

		#m <- matrix(tmapID, ncol = ncol(shp))


		#m = tmapID

		#m[1,5] = 4

		shp2 = stars::st_as_stars(m, dimensions = shp)

		lf = get_lf(facet_row, facet_col, facet_page)

		opts = leaflet::gridOptions(pane = pane)

		lf %>%
			leafem::addStarsImage(shp2, band = 1, colors = pal_col, opacity = pal_opacity, group = group, options = opts) %>%
			blend_lf(a$blend, pane) %>%
			assign_lf(facet_row, facet_col, facet_page)
	} else {
		#shp2 = stars::st_as_stars(list(values = tmapID), dimensions = shp)
		#shpTM = shapeTM(sf::st_geometry(sf::st_as_sf(shp2)), as.vector(tmapID))

		m = matrix(tmapID, nrow = nrow(shp), ncol = ncol(shp))
		shp2 = structure(list(tmapID = m), class = "stars", dimensions = shp)

		shp3 = sf::st_geometry(sf::st_as_sf(shp2))

		crs = get_option_class(o$crs_step4, "sf")

		shpTM = shapeTM(sf::st_transform(shp3, crs), tmapID)


		gp$lty = "solid"
		a2 = structure(list(), class = "tm_data_polygons")
		tmapLeafletDataPlot(a2, shpTM, dt, pdt, popup.format = NULL, hdt = NULL, idt = NULL, gp, bbx, facet_row, facet_col, facet_page, id, pane, group, glid, o)
		#grid.shape(s, gp=gpar(fill=color, col=NA), bg.col=NA, i, k)
	}
	NULL
}

# To generate all options:
# args = intersect(names(tmap_options_mode("view")),names(tmap_options_mode("plot")))
# cat(paste(args, collapse = ", "))
# x = sapply(args, function(a) paste0("#' @param ", a, " `r .doc_opt(\"", a, "\")`"))
# cat(paste(x, collapse = "\n"))
#
# To generate param records:
# cat(paste(sapply(names, function(x) {
#	paste0("#' @param ", x, " `r .doc_opt(\"", x ,"\")`")
# }), collapse = "\n"))
# To generate option topics (the 'what', last part)
# nms = names(tmap_options())
# nms_s = strsplit(nms, split = ".", fixed = TRUE)
# nms_lst = sapply(nms_s, tail, 1)
# whats = sort(unique(nms_lst))
# cat(paste0("c(", paste(sapply(whats, function(w) paste0(w, " = \"", w, "\"")), collapse = ",\n"), ")"))
#' tmap options
#'
#' Get or set the tmap options globally. For map specific options, we recommend to use [tm_options()] or [tm_layout()] via which the layout-related options can be set. [tmap_options()] functions similar to [base::options()].
#'
#' @param ... List of tmap options to be set, or option names (characters) to be returned (see details)
#' @param crs Map crs (see [tm_shape()]). `NA` means the crs is specified in [tm_shape()]. The crs that is used by the transformation functions is defined in [tm_shape()].
#' @param facet.max Maximum number of facets in a static plot (rows x columns); animation frames are not counted.
#' @param facet_levels.max Maximum number of facet levels processed across all faceting dimensions, animation frames included (rows x columns x frames). Default `1e5`.
#' @param free.scales For backward compatibility: if this value is set, it will be used to impute the free arguments in the layer functions
#' @param raster.max_cells Maximum number of raster grid cells. Can be mode specific `c(plot = 3000, view = 1000, 1000)` (the last value is the fall back default)
#' @param raster.warp Should rasters be warped or transformed in case a different projection (crs) is used? Warping creates a new regular raster in the target crs, whereas transforming creates a (usually non-regular) raster in the target crs. The former is lossy, but much faster and is therefore the default.
#' When a different projection (crs) is used, a (usually) regular raster will be
#' @param show.messages Show messages?
#' @param show.warnings Show warnings?
#' @param output.format Output format
#' @param output.size Output size
#' @param output.dpi Output dpi
#' @param animation.dpi Output dpi for animations
#' @param value.const Default visual value constants e.g. the default fill color for `tm_shape(World) + tm_polygons()`. A list is required with per visual variable a value.
#' @param value.na Default visual values that are used to visualize NA data values. A list is required with per visual variable a value.
#' @param value.null Default visual values that are used to visualize null (out-of-scope) data values. A list is required with per visual variable a value.
#' @param value.blank Default visual values that correspond to blank. For color these are `"#00000000"` meaning transparent. A list is required with per visual variable a value.
#' @param values.var Default values when a data variable to mapped to a visual variable, e.g. a color palette. A list is required with per visual variable a value.
#' @param values.range Default range for values. See `values.range` of [tm_scale_categorical()]. A list is required with per visual variable a value.
#' @param value.neutral Default values for when a data variable to mapped to a visual variable, e.g. a color palette. A list is required with per visual variable a value.
#' @param values.scale Default scales (as in object sizes) for values. See `values.range` of [tm_scale_categorical()]. A list is required with per visual variable a value.
#' @param scales.var Default scale functions per visual variable and type of data variable. A list is required with per visual variable per data type.
#' @param scale.misc.args Default values of scale function-specific arguments. A list is required with per scale function and optional per visual variable.
#' @param continuous.nclass_per_legend_break The number of continuous legend breaks within one 'unit' (label).  The default value is 50.
#' @param continuous.nclasses the number of classes of a continuous scale. Should be odd.  The default value is 101.
#' @param label.format Format for the labels. These are the default values for [tm_label_format()]
#' @param label.na Default label for missing values.
#' @param scale Overall scale of the map
#' @param asp Aspect ratio of each map. When `asp` is set to `NA` (default) the aspect ratio will be adjusted to the used shapes. When set to 0 the aspect ratio is
#'   adjusted to the size of the device divided by the number of columns and rows.
#' @param bg Draw map background?
#' @param bg.color Background color of the map.
#' @param outer.bg Draw map background (outside the frame)?
#' @param outer.bg.color Background color of map outside the frame.
#' @param frame Draw map frame?
#' @param frame.color `r .doc_opt("frame.color")`
#' @param frame.alpha `r .doc_opt("frame.alpha")`
#' @param frame.lwd `r .doc_opt("frame.lwd")`
#' @param frame.r `r .doc_opt("frame.r")`
#' @param frame.double_line `r .doc_opt("frame.double_line")`
#' @param outer.margins `r .doc_opt("outer.margins")`
#' @param inner.margins `r .doc_opt("inner.margins")`
#' @param inner.margins.extra `r .doc_opt("inner.margins.extra")`
#' @param meta.margins `r .doc_opt("meta.margins")`
#' @param meta.auto_margins `r .doc_opt("meta.auto_margins")`
#' @param between_margin Margin between the map.
#' @param panel.margin `r .doc_opt("panel.margin")`
#' @param component.offset `r .doc_opt("component.offset")`
#' @param component.stack_margin `r .doc_opt("component.stack_margin")`
#' @param xlab.show `r .doc_opt("xlab.show")`
#' @param xlab.text `r .doc_opt("xlab.text")`
#' @param xlab.size `r .doc_opt("xlab.size")`
#' @param xlab.color `r .doc_opt("xlab.color")`
#' @param xlab.rotation `r .doc_opt("xlab.rotation")`
#' @param xlab.space `r .doc_opt("xlab.space")`
#' @param xlab.fontface `r .doc_opt("xlab.fontface")`
#' @param xlab.fontfamily `r .doc_opt("xlab.fontfamily")`
#' @param xlab.alpha `r .doc_opt("xlab.alpha")`
#' @param xlab.side `r .doc_opt("xlab.side")`
#' @param ylab.show `r .doc_opt("ylab.show")`
#' @param ylab.text `r .doc_opt("ylab.text")`
#' @param ylab.size `r .doc_opt("ylab.size")`
#' @param ylab.color `r .doc_opt("ylab.color")`
#' @param ylab.rotation `r .doc_opt("ylab.rotation")`
#' @param ylab.space `r .doc_opt("ylab.space")`
#' @param ylab.fontface `r .doc_opt("ylab.fontface")`
#' @param ylab.fontfamily `r .doc_opt("ylab.fontfamily")`
#' @param ylab.alpha `r .doc_opt("ylab.alpha")`
#' @param ylab.side `r .doc_opt("ylab.side")`
#' @param panel.type `r .doc_opt("panel.type")`
#' @param panel.wrap.pos The panel positions for wrapped facets created with [tm_facets_grid()]. One of `"left"`, `"right"`, `"top"` (default) or `"bottom"`.
#' @param panel.xtab.pos The panel positions for grid facets created with [tm_facets_grid()]. Vector of two, where the first determines the locations of row panels (`"left"` or `"right"`) and the second the location of column panels ( `"top"` or `"bottom")
#' @param unit Unit of the coordinate
#' @param color.sepia_intensity `r .doc_opt("color.sepia_intensity")`
#' @param color.saturation `r .doc_opt("color.saturation")`
#' @param color_vision_deficiency_sim Color vision deficiency simulation. Either `"protan"`, `"deutan"`, or `"tritan"`.
#' @param text.fontface `r .doc_opt("text.fontface")`
#' @param text.fontfamily `r .doc_opt("text.fontfamily")`
#' @param r `r .doc_opt("r")`
#' @param component.position `r .doc_opt("component.position")`
#' @param component.autoscale `r .doc_opt("component.autoscale")`
#' @param legend.show `r .doc_opt("legend.show")`
#' @param legend.orientation `r .doc_opt("legend.orientation")`
#' @param legend.position `r .doc_opt("legend.position")`
#' @param legend.width `r .doc_opt("legend.width")`
#' @param legend.height `r .doc_opt("legend.height")`
#' @param legend.stack `r .doc_opt("legend.stack")`
#' @param legend.reverse `r .doc_opt("legend.reverse")`
#' @param legend.na.show `r .doc_opt("legend.na.show")`
#' @param legend.title.color `r .doc_opt("legend.title.color")`
#' @param legend.title.size `r .doc_opt("legend.title.size")`
#' @param legend.title.fontface `r .doc_opt("legend.title.fontface")`
#' @param legend.title.fontfamily `r .doc_opt("legend.title.fontfamily")`
#' @param legend.title.alpha `r .doc_opt("legend.title.alpha")`
#' @param legend.xlab.color `r .doc_opt("legend.xlab.color")`
#' @param legend.xlab.size `r .doc_opt("legend.xlab.size")`
#' @param legend.xlab.fontface `r .doc_opt("legend.xlab.fontface")`
#' @param legend.xlab.fontfamily `r .doc_opt("legend.xlab.fontfamily")`
#' @param legend.xlab.rot `r .doc_opt("legend.xlab.rot")`
#' @param legend.xlab.alpha `r .doc_opt("legend.xlab.alpha")`
#' @param legend.ylab.color `r .doc_opt("legend.ylab.color")`
#' @param legend.ylab.size `r .doc_opt("legend.ylab.size")`
#' @param legend.ylab.fontface `r .doc_opt("legend.ylab.fontface")`
#' @param legend.ylab.fontfamily `r .doc_opt("legend.ylab.fontfamily")`
#' @param legend.ylab.rot `r .doc_opt("legend.ylab.rot")`
#' @param legend.ylab.alpha `r .doc_opt("legend.ylab.alpha")`
#' @param legend.text.color `r .doc_opt("legend.text.color")`
#' @param legend.text.size `r .doc_opt("legend.text.size")`
#' @param legend.text.fontface `r .doc_opt("legend.text.fontface")`
#' @param legend.text.fontfamily `r .doc_opt("legend.text.fontfamily")`
#' @param legend.text.alpha `r .doc_opt("legend.text.alpha")`
#' @param legend.frame `r .doc_opt("legend.frame")`
#' @param legend.frame.lwd `r .doc_opt("legend.frame.lwd")`
#' @param legend.frame.r `r .doc_opt("legend.frame.r")`
#' @param legend.bg.color `r .doc_opt("legend.bg.color")`
#' @param legend.bg.alpha `r .doc_opt("legend.bg.alpha")`
#' @param legend.only Should only legends be printed (so without map)?
#' @param legend.absolute_fontsize `r .doc_opt("legend.absolute_fontsize")`
#' @param legend.settings.portrait `r .doc_opt("legend.settings.portrait")`
#' @param legend.settings.landscape `r .doc_opt("legend.settings.landscape")`
#' @param chart.show `r .doc_opt("chart.show")`
#' @param chart.plot.axis.x `r .doc_opt("chart.plot.axis.x")`
#' @param chart.plot.axis.y `r .doc_opt("chart.plot.axis.y")`
#' @param chart.position `r .doc_opt("chart.position")`
#' @param chart.width `r .doc_opt("chart.width")`
#' @param chart.height `r .doc_opt("chart.height")`
#' @param chart.stack `r .doc_opt("chart.stack")`
#' @param chart.reverse `r .doc_opt("chart.reverse")`
#' @param chart.na.show `r .doc_opt("chart.na.show")`
#' @param chart.title.color `r .doc_opt("chart.title.color")`
#' @param chart.title.size `r .doc_opt("chart.title.size")`
#' @param chart.title.fontface `r .doc_opt("chart.title.fontface")`
#' @param chart.title.fontfamily `r .doc_opt("chart.title.fontfamily")`
#' @param chart.title.alpha `r .doc_opt("chart.title.alpha")`
#' @param chart.xlab.color `r .doc_opt("chart.xlab.color")`
#' @param chart.xlab.size `r .doc_opt("chart.xlab.size")`
#' @param chart.xlab.fontface `r .doc_opt("chart.xlab.fontface")`
#' @param chart.xlab.fontfamily `r .doc_opt("chart.xlab.fontfamily")`
#' @param chart.xlab.alpha `r .doc_opt("chart.xlab.alpha")`
#' @param chart.ylab.color `r .doc_opt("chart.ylab.color")`
#' @param chart.ylab.size `r .doc_opt("chart.ylab.size")`
#' @param chart.ylab.fontface `r .doc_opt("chart.ylab.fontface")`
#' @param chart.ylab.fontfamily `r .doc_opt("chart.ylab.fontfamily")`
#' @param chart.ylab.alpha `r .doc_opt("chart.ylab.alpha")`
#' @param chart.text.color `r .doc_opt("chart.text.color")`
#' @param chart.text.size `r .doc_opt("chart.text.size")`
#' @param chart.text.fontface `r .doc_opt("chart.text.fontface")`
#' @param chart.text.fontfamily `r .doc_opt("chart.text.fontfamily")`
#' @param chart.text.alpha `r .doc_opt("chart.text.alpha")`
#' @param chart.frame `r .doc_opt("chart.frame")`
#' @param chart.frame.lwd `r .doc_opt("chart.frame.lwd")`
#' @param chart.frame.r `r .doc_opt("chart.frame.r")`
#' @param chart.bg.color `r .doc_opt("chart.bg.color")`
#' @param chart.bg.alpha `r .doc_opt("chart.bg.alpha")`
#' @param chart.object.color `r .doc_opt("chart.object.color")`
#' @param title.size `r .doc_opt("title.size")`
#' @param title.color `r .doc_opt("title.color")`
#' @param title.fontface `r .doc_opt("title.fontface")`
#' @param title.fontfamily `r .doc_opt("title.fontfamily")`
#' @param title.alpha `r .doc_opt("title.alpha")`
#' @param title.padding `r .doc_opt("title.padding")`
#' @param title.frame `r .doc_opt("title.frame")`
#' @param title.frame.lwd `r .doc_opt("title.frame.lwd")`
#' @param title.frame.r `r .doc_opt("title.frame.r")`
#' @param title.position `r .doc_opt("title.position")`
#' @param title.width `r .doc_opt("title.width")`
#' @param credits.size `r .doc_opt("credits.size")`
#' @param credits.color `r .doc_opt("credits.color")`
#' @param credits.fontface `r .doc_opt("credits.fontface")`
#' @param credits.fontfamily `r .doc_opt("credits.fontfamily")`
#' @param credits.alpha `r .doc_opt("credits.alpha")`
#' @param credits.padding `r .doc_opt("credits.padding")`
#' @param credits.position `r .doc_opt("credits.position")`
#' @param credits.width `r .doc_opt("credits.width")`
#' @param credits.height `r .doc_opt("credits.height")`
#' @param compass.north `r .doc_opt("compass.north")`
#' @param compass.type `r .doc_opt("compass.type")`
#' @param compass.text.size `r .doc_opt("compass.text.size")`
#' @param compass.size `r .doc_opt("compass.size")`
#' @param compass.show.labels `r .doc_opt("compass.show.labels")`
#' @param compass.cardinal.directions `r .doc_opt("compass.cardinal.directions")`
#' @param compass.text.color `r .doc_opt("compass.text.color")`
#' @param compass.color.dark `r .doc_opt("compass.color.dark")`
#' @param compass.color.light `r .doc_opt("compass.color.light")`
#' @param compass.lwd `r .doc_opt("compass.lwd")`
#' @param compass.margins `r .doc_opt("compass.margins")`
#' @param compass.position `r .doc_opt("compass.position")`
#' @param logo.height `r .doc_opt("logo.height")`
#' @param logo.margins `r .doc_opt("logo.margins")`
#' @param logo.between_margin `r .doc_opt("logo.between_margin")`
#' @param logo.position `r .doc_opt("logo.position")`
#' @param scalebar.breaks See [tm_scalebar()]
#' @param scalebar.width See [tm_scalebar()]
#' @param scalebar.allow_clipping See [tm_scalebar()]
#' @param scalebar.text.size `r .doc_opt("scalebar.text.size")`
#' @param scalebar.text.color `r .doc_opt("scalebar.text.color")`
#' @param scalebar.color.dark `r .doc_opt("scalebar.color.dark")`
#' @param scalebar.color.light `r .doc_opt("scalebar.color.light")`
#' @param scalebar.lwd `r .doc_opt("scalebar.lwd")`
#' @param scalebar.size `r .doc_opt("scalebar.size")`
#' @param scalebar.margins `r .doc_opt("scalebar.margins")`
#' @param scalebar.position `r .doc_opt("scalebar.position")`
#' @param grid.show `r .doc_opt("grid.show")`
#' @param grid.labels.pos `r .doc_opt("grid.labels.pos")`
#' @param grid.x `r .doc_opt("grid.x")`
#' @param grid.y `r .doc_opt("grid.y")`
#' @param grid.n.x `r .doc_opt("grid.n.x")`
#' @param grid.n.y `r .doc_opt("grid.n.y")`
#' @param grid.crs `r .doc_opt("grid.crs")`
#' @param grid.col `r .doc_opt("grid.col")`
#' @param grid.lwd `r .doc_opt("grid.lwd")`
#' @param grid.alpha `r .doc_opt("grid.alpha")`
#' @param grid.labels.show `r .doc_opt("grid.labels.show")`
#' @param grid.labels.size `r .doc_opt("grid.labels.size")`
#' @param grid.labels.col `r .doc_opt("grid.labels.col")`
#' @param grid.labels.rot `r .doc_opt("grid.labels.rot")`
#' @param grid.labels.format `r .doc_opt("grid.labels.format")`
#' @param grid.labels.cardinal `r .doc_opt("grid.labels.cardinal")`
#' @param grid.labels.margin.x `r .doc_opt("grid.labels.margin.x")`
#' @param grid.labels.margin.y `r .doc_opt("grid.labels.margin.y")`
#' @param grid.labels.space.x `r .doc_opt("grid.labels.space.x")`
#' @param grid.labels.space.y `r .doc_opt("grid.labels.space.y")`
#' @param grid.labels.inside_frame `r .doc_opt("grid.labels.inside_frame")`
#' @param grid.ticks `r .doc_opt("grid.ticks")`
#' @param grid.lines `r .doc_opt("grid.lines")`
#' @param grid.ndiscr `r .doc_opt("grid.ndiscr")`
#' @param mouse_coordinates.position `r .doc_opt("mouse_coordinates.position")`
#' @param minimap.server `r .doc_opt("minimap.server")`
#' @param minimap.toggle `r .doc_opt("minimap.toggle")`
#' @param minimap.position `r .doc_opt("minimap.position")`
#' @param panel.show `r .doc_opt("panel.show")`
#' @param panel.labels `r .doc_opt("panel.labels")`
#' @param panel.label.size `r .doc_opt("panel.label.size")`
#' @param panel.label.color `r .doc_opt("panel.label.color")`
#' @param panel.label.fontface `r .doc_opt("panel.label.fontface")`
#' @param panel.label.fontfamily `r .doc_opt("panel.label.fontfamily")`
#' @param panel.label.alpha `r .doc_opt("panel.label.alpha")`
#' @param panel.label.bg.color `r .doc_opt("panel.label.bg.color")`
#' @param panel.label.frame `r .doc_opt("panel.label.frame")`
#' @param panel.label.frame.lwd `r .doc_opt("panel.label.frame.lwd")`
#' @param panel.label.frame.r `r .doc_opt("panel.label.frame.r")`
#' @param panel.label.height `r .doc_opt("panel.label.height")`
#' @param panel.label.rot Rotation angles of the panel labels. Vector of four values that determine the panel label rotation when they are placed left, top, right, and bottom. The default angles are 90, 0, 270 and 0 respectively. Note that the second value is the most common, since labels are by default shown on top (see `panel.wrap.pos`). In cross-table facets created with `tm_facets_grid()`, the first two values are used by default (see `panel.xtab.pos`).
#' @inheritParams tm_view
#' @param qtm.scalebar `r .doc_opt("qtm.scalebar")`
#' @param qtm.minimap `r .doc_opt("qtm.minimap")`
#' @param qtm.mouse_coordinates `r .doc_opt("qtm.mouse_coordinates")`
#' @param earth_boundary The earth boundary
#' @param earth_boundary.color `r .doc_opt("earth_boundary.color")`
#' @param earth_boundary.lwd `r .doc_opt("earth_boundary.lwd")`
#' @param earth_datum  Earth datum
#' @param space Should the space be drawn? Only applicable is earth_boundary is enabled.
#' @param space.color `r .doc_opt("space.color")`
#' @param space_overlay Should the space be drawn as overlay (to make sure spatial features or rasters do not exceed the earth boundary), or as background? By default `TRUE` when a raster is warped.
#' @param check_and_fix Should attempt to fix an invalid shapefile
#' @param basemap.show `r .doc_opt("basemap.show")`
#' @param basemap.server `r .doc_opt("basemap.server")`
#' @param basemap.alpha `r .doc_opt("basemap.alpha")`
#' @param basemap.zoom `r .doc_opt("basemap.zoom")`
#' @param tiles.show `r .doc_opt("tiles.show")`
#' @param tiles.server `r .doc_opt("tiles.server")`
#' @param tiles.alpha `r .doc_opt("tiles.alpha")`
#' @param tiles.zoom `r .doc_opt("tiles.zoom")`
#' @param attr.color `r .doc_opt("attr.color")`
#' @param crs_extra Only used internally (work in progress)
#' @param crs_global The used crs for world maps
#' @param title deprecated See [tm_title()]
#' @param main.title deprecated See [tm_title()]
#' @param main.title.size,main.title.color,main.title.fontface,main.title.fontfamily,main.title.position deprecated. Use the `title.` options instead.
#' @param fontface,fontfamily renamed to `text.fontface` and `text.fontfamily`
#' @param component.resize_as_group `r .doc_opt("component.resize_as_group")`
#' @param component.frame_combine `r .doc_opt("component.frame_combine")`
#' @param component.stack `r .doc_opt("component.stack")`
#' @param component.equalize `r .doc_opt("component.equalize")`
#' @param component.frame `r .doc_opt("component.frame")`
#' @param component.frame.color `r .doc_opt("component.frame.color")`
#' @param component.frame.alpha `r .doc_opt("component.frame.alpha")`
#' @param component.frame.lwd `r .doc_opt("component.frame.lwd")`
#' @param component.frame.r `r .doc_opt("component.frame.r")`
#' @param component.bg `r .doc_opt("component.bg")`
#' @param component.bg.color `r .doc_opt("component.bg.color")`
#' @param component.bg.alpha `r .doc_opt("component.bg.alpha")`
#' @param legend.frame.color `r .doc_opt("legend.frame.color")`
#' @param legend.frame.alpha `r .doc_opt("legend.frame.alpha")`
#' @param legend.bg `r .doc_opt("legend.bg")`
#' @param add_legend.position `r .doc_opt("add_legend.position")`
#' @param chart.frame.color `r .doc_opt("chart.frame.color")`
#' @param chart.frame.alpha `r .doc_opt("chart.frame.alpha")`
#' @param chart.bg `r .doc_opt("chart.bg")`
#' @param title.frame.color `r .doc_opt("title.frame.color")`
#' @param title.frame.alpha `r .doc_opt("title.frame.alpha")`
#' @param inset.position `r .doc_opt("inset.position")`
#' @param inset_map.height `r .doc_opt("inset_map.height")`
#' @param inset_map.width `r .doc_opt("inset_map.width")`
#' @param inset_map.margins `r .doc_opt("inset_map.margins")`
#' @param inset_map.between_margin `r .doc_opt("inset_map.between_margin")`
#' @param inset_map.position `r .doc_opt("inset_map.position")`
#' @param inset_map.frame `r .doc_opt("inset_map.frame")`
#' @param inset.height `r .doc_opt("inset.height")`
#' @param inset.width `r .doc_opt("inset.width")`
#' @param inset.margins `r .doc_opt("inset.margins")`
#' @param inset.between_margin `r .doc_opt("inset.between_margin")`
#' @param inset.frame `r .doc_opt("inset.frame")`
#' @param inset.bg `r .doc_opt("inset.bg")`
#' @param inset.bg.color `r .doc_opt("inset.bg.color")`
#' @param inset.bg.alpha `r .doc_opt("inset.bg.alpha")`
#' @param inset_grob.height `r .doc_opt("inset_grob.height")`
#' @param inset_grob.width `r .doc_opt("inset_grob.width")`
#' @param inset_gg.height `r .doc_opt("inset_gg.height")`
#' @param inset_gg.width `r .doc_opt("inset_gg.width")`
#' @param scalebar.text.fontface `r .doc_opt("scalebar.text.fontface")`
#' @param scalebar.text.fontfamily `r .doc_opt("scalebar.text.fontfamily")`
#' @param grid.labels.fontface `r .doc_opt("grid.labels.fontface")`
#' @param grid.labels.fontfamily `r .doc_opt("grid.labels.fontfamily")`
#' @param panel.label.bg `r .doc_opt("panel.label.bg")`
#' @param panel.label.bg.alpha `r .doc_opt("panel.label.bg.alpha")`
#' @param panel.label.frame.color `r .doc_opt("panel.label.frame.color")`
#' @param panel.label.frame.alpha `r .doc_opt("panel.label.frame.alpha")`
#' @param crs_basemap `r .doc_opt("crs_basemap")`
#' @inheritParams tm_plot
# For bbox
#' @inheritParams tm_shape
#' @export
#' @example ./examples/tmap_options.R
#' @name tmap_options
tmap_options = function(..., crs, facet.max, facet_levels.max, free.scales, raster.max_cells, raster.warp, show.messages, show.warnings, output.format, output.size, output.dpi, animation.dpi, value.const, value.na, value.null, value.blank, values.var, values.range, value.neutral, values.scale, scales.var, scale.misc.args, continuous.nclass_per_legend_break, continuous.nclasses, label.format, label.na, scale, asp, bg, bg.color, outer.bg, outer.bg.color, frame, frame.color, frame.alpha, frame.lwd, frame.r, frame.double_line, outer.margins, inner.margins, inner.margins.extra, meta.margins, meta.auto_margins, between_margin, panel.margin, xlab.show, xlab.text, xlab.size, xlab.color, xlab.rotation, xlab.space, xlab.fontface, xlab.fontfamily, xlab.alpha, xlab.side, ylab.show, ylab.text, ylab.size, ylab.color, ylab.rotation, ylab.space, ylab.fontface, ylab.fontfamily, ylab.alpha, ylab.side, panel.type, panel.wrap.pos, panel.xtab.pos, unit, color.sepia_intensity, color.saturation, color_vision_deficiency_sim, text.fontface, text.fontfamily, r, component.position, component.offset, component.stack_margin, component.autoscale, component.resize_as_group, component.frame_combine, component.stack, legend.stack, chart.stack, component.equalize, component.frame, component.frame.color, component.frame.alpha, component.frame.lwd, component.frame.r, component.bg, component.bg.color, component.bg.alpha, legend.show, legend.orientation, legend.position, legend.width, legend.height, legend.reverse, legend.na.show, legend.title.color, legend.title.size, legend.title.fontface, legend.title.fontfamily, legend.title.alpha, legend.xlab.color, legend.xlab.size, legend.xlab.rot, legend.xlab.fontface, legend.xlab.fontfamily, legend.xlab.alpha, legend.ylab.color, legend.ylab.size, legend.ylab.rot, legend.ylab.fontface, legend.ylab.fontfamily, legend.ylab.alpha, legend.text.color, legend.text.size, legend.text.fontface, legend.text.fontfamily, legend.text.alpha, legend.frame, legend.frame.color, legend.frame.alpha, legend.frame.lwd, legend.frame.r, legend.bg, legend.bg.color, legend.bg.alpha, legend.only, legend.absolute_fontsize, legend.settings.portrait, legend.settings.landscape, add_legend.position, chart.show, chart.plot.axis.x, chart.plot.axis.y, chart.position, chart.width, chart.height, chart.reverse, chart.na.show, chart.title.color, chart.title.size, chart.title.fontface, chart.title.fontfamily, chart.title.alpha, chart.xlab.color, chart.xlab.size, chart.xlab.fontface, chart.xlab.fontfamily, chart.xlab.alpha, chart.ylab.color, chart.ylab.size, chart.ylab.fontface, chart.ylab.fontfamily, chart.ylab.alpha, chart.text.color, chart.text.size, chart.text.fontface, chart.text.fontfamily, chart.text.alpha, chart.frame, chart.frame.color, chart.frame.alpha, chart.frame.lwd, chart.frame.r, chart.bg, chart.bg.color, chart.bg.alpha, chart.object.color, title.size, title.color, title.fontface, title.fontfamily, title.alpha, title.padding, title.frame, title.frame.color, title.frame.alpha, title.frame.lwd, title.frame.r, title.position, title.width, credits.size, credits.color, credits.fontface, credits.fontfamily, credits.alpha, credits.padding, credits.position, credits.width, credits.height, compass.north, compass.type, compass.text.size, compass.size, compass.show.labels, compass.cardinal.directions, compass.text.color, compass.color.dark, compass.color.light, compass.lwd, compass.margins, compass.position, inset.position, logo.height, logo.margins, logo.between_margin, logo.position, inset_map.height, inset_map.width, inset_map.margins, inset_map.between_margin, inset_map.position, inset_map.frame, inset.height, inset.width, inset.margins, inset.between_margin, inset.frame, inset.bg, inset.bg.color, inset.bg.alpha, inset_grob.height, inset_grob.width, inset_gg.height, inset_gg.width, scalebar.breaks, scalebar.width, scalebar.allow_clipping, scalebar.text.size, scalebar.text.color, scalebar.text.fontface, scalebar.text.fontfamily, scalebar.color.dark, scalebar.color.light, scalebar.lwd, scalebar.size, scalebar.margins, scalebar.position, grid.show, grid.labels.pos, grid.x, grid.y, grid.n.x, grid.n.y, grid.crs, grid.col, grid.lwd, grid.alpha, grid.labels.show, grid.labels.size, grid.labels.col, grid.labels.fontface, grid.labels.fontfamily, grid.labels.rot, grid.labels.format, grid.labels.cardinal, grid.labels.margin.x, grid.labels.margin.y, grid.labels.space.x, grid.labels.space.y, grid.labels.inside_frame, grid.ticks, grid.lines, grid.ndiscr, mouse_coordinates.position, minimap.server, minimap.toggle, minimap.position, panel.show, panel.labels, panel.label.size, panel.label.color, panel.label.fontface, panel.label.fontfamily, panel.label.alpha, panel.label.bg, panel.label.bg.color, panel.label.bg.alpha, panel.label.frame, panel.label.frame.color, panel.label.frame.alpha, panel.label.frame.lwd, panel.label.frame.r, panel.label.height, panel.label.rot, qtm.scalebar, qtm.minimap, qtm.mouse_coordinates, earth_boundary, earth_boundary.color, earth_boundary.lwd, earth_datum, space, space.color, space_overlay, check_and_fix, basemap.show, basemap.server, basemap.alpha, basemap.zoom, tiles.show, tiles.server, tiles.alpha, tiles.zoom, attr.color, crs_extra, crs_global, crs_basemap,
						use_gradient, # plot mode
						use_browser, use_WebGL, control.position, control.bases, control.overlays, control.collapse, set_bounds, # view mode
						set_view, set_zoom_limits, use_circle_markers, leaflet.options, # view mode
						title = NULL,
						main.title = NULL,
						main.title.size = NULL,
						main.title.color = NULL,
						main.title.fontface = NULL,
						main.title.fontfamily = NULL,
						main.title.position = NULL,
						fontface = NULL,
						fontfamily = NULL
						) {
	o = get("tmapOptions", envir = .TMAP)
	nms = names(o)
	show.warnings = o$show.warnings

	## case 1: option list is given. E.g. opts = list(crs = 3035, panel.show = TRUE); tmap_options(opts)
	## case 2: option name is given. E.g. tmap_options("crs", "panel.show")
	## case 3: named options are set. E.g. tmap_options(crs = 3035, panel.show = TRUE)
	## case 4: tmap_options is called without arguments


	# get current style name (default: white), and set new style name (with "(modified)")
	sty_cur = getOption("tmap.style")
	sty_new = if (substr(sty_cur, nchar(sty_cur) - 9, nchar(sty_cur)) == "(modified)") sty_cur else paste(sty_cur, "(modified)")

	args = lapply(as.list(rlang::call_match(dots_expand = TRUE)[-1]), eval, envir = parent.frame())
	set_new_style = FALSE

	#lst = list(...)
	if (length(args) >= 1 && is.null(names(args[1]))) {
		arg = args[[1]]
		if (is.list(arg)) {
			if (length(args) > 1 && show.warnings) warning("Only the first argument is used; the other arguments are ignored.")
			## case 1: option list is given
			args = arg

			style_attr = attr(args, "style")
			if (!is.null(style_attr)) {
				sty_new = style_attr
				set_new_style = TRUE
			}

		} else {
			## case 2: option name is given
			args = sapply(args, "[", 1)
			if (!all(args %in% nms) && show.warnings) warning("The following options do not exist: ", paste(setdiff(args, nms), collapse = ", "))
			args = intersect(args, nms)
			return(o[args])
		}
	}

	mode_opts = setdiff(unique(unlist(lapply(o$modes, names))), "name")

	all_opts = union(mode_opts, names(.TMAP$defaultTmapOptions))

	unknown_args = setdiff(names(args), all_opts)

	fun = "tmap_options"
	if ("view.legend.position" %in% unknown_args) {
		args$legend.position = args$view.legend.position
		args$view.legend.position = NULL
		v3_opt(fun, "view.legend.position", "legend.position")
	}
	if ("set.bounds" %in% unknown_args) {
		args$set_bounds = args$set.bounds
		args$set.bounds = NULL
		v3_opt(fun, "set.bounds", "set_bounds")
	}
	if ("set.view" %in% unknown_args) {
		args$set_view = args$set.view
		args$set.view = NULL
		v3_opt(fun, "set.view", "set_view")
	}
	if ("set.zoom.limits" %in% unknown_args) {
		args$set_zoom_limits = args$set.zoom.limits
		args$set.zoom.limits = NULL
		v3_opt(fun, "set.zoom.limits", "set_zoom_limits")
	}
	if ("max.raster" %in% unknown_args) {
		args$raster.max_cells = args$max.raster
		args$max.raster = NULL
		v3_opt(fun, "max.raster", "raster.max_cells")
	}
	if ("fontfamily" %in% unknown_args) {
		args$text.fontfamily = args$fontfamily
		args$fontfamily = NULL
		v3_opt(fun, "fontfamily", "text.fontfamily")
	}
	if ("fontface" %in% unknown_args) {
		args$text.fontface = args$fontface
		args$fontface = NULL
		v3_opt(fun, "fontface", "text.fontface")
	}
	if ("overlays" %in% unknown_args) {
		args$tiles.server = args$overlays
		args$overlays = NULL
		v3_opt(fun, "overlays", "tiles.server")
	}
	if ("basemaps" %in% unknown_args) {
		args$basemap.server = args$basemaps
		args$basemaps = NULL
		v3_opt(fun, "basemaps", "basemap.server")
	}


	unknown_args = setdiff(names(args), all_opts)
	if (length(unknown_args) == 1) {
		stop("the following option does not exist: ", unknown_args)
	} else if (length(unknown_args) > 1) {
		stop("the following options do not exist: ", paste(unknown_args, collapse = ", "))
	}

	if (!length(args)) {
		# case 4
		return(o)
	} else {
		# case 1 and 3
		if (is.null(names(args))) {
			stop("Option list is unnamed")
		}
		backup = o[names(args)]
		o[names(args)] = complete_options(args, backup) # needed to impute position and value(s) args


		options(tmap.style=sty_new)
		attr(o, "style") = sty_new
		attr(o, "specified") = names(args)
		assign("tmapOptions", o, envir = .TMAP)

		if (set_new_style) {
			if (o$show.messages) message("tmap options successfully loaded as style \"", sty_new, "\"")
			styles = get("tmapStyles", envir = .TMAP)
			styles[[sty_new]] = suppressMessages(tmap_options_diff())
			assign("tmapStyles", styles, envir = .TMAP)
		}
		invisible(backup)
	}
}

#' @param mode mode, e.g. `"plot"` or `"view"`
#' @param style style. If specified, the style specific options are returned
#' @param mode.specific Should only mode-specific options be returned? `TRUE` by default.
#' @param default.options return the default options or the current options?
#' @rdname tmap_options
#' @export
tmap_options_mode = function(mode = NA, style = NULL, mode.specific = TRUE, default.options = FALSE) {
	if (default.options) {
		o = .TMAP$defaultTmapOptions
	} else {
		o = get("tmapOptions", envir = .TMAP)

		if (!is.null(style)) {
			check_style(style)
			s = get("tmapStyles", envir = .TMAP)[[style]]
			o = complete_options(s, o)
		}
	}

	if (is.na(mode)) mode = getOption("tmap.mode")
	opt2 = o$modes[[mode]]

	specified = attr(o, "specified")

	# all general options with mode-specific defaults (except the specified ones tmap_options(bg.color = "red")
	int_opt = setdiff(intersect(names(o), names(opt2)), specified)

	# all mode-specific options
	diff_opt = setdiff(names(opt2), names(o))

	if (length(int_opt)) o[int_opt] = opt2[int_opt]
	if (length(diff_opt)) o = c(o, opt2[diff_opt])

	o$modes = NULL

	if (mode.specific) {
		o[setdiff(names(opt2), "name")]
	} else {
		o
	}
}


tmap_option = function(name, type = NULL) {
	get_option_class(tmap_options()[[name]], class = type, spatial_class = FALSE)
}



tmap_graphics_name = function(mode) {
	if (missing(mode)) mode = getOption("tmap.mode")
	get("tmapOptions", envir = .TMAP)$modes[[mode]]$name
}



tmapOption = function(...) {
	structure(list(...), class = "tmapOption")
}




#' Internal tmap function to add a default value for the layer functions. Deprecated
#'
#' Internal tmap function to add a default value for the layer functions. Deprecated. Use [tmapSubmitOptions()] instead.
#'
#' @param option, one of: `"value.const"`, `"value.na"`, `"value.blank"`, `"values.var"`, `'values.range'`, `"value.neutral"`, `"scales.var"`
#' @param id name of the visual variable with layer, in the format `"x.y"`,
#'   where `x` is the visual variable and `y` is the layer.
#'   It is also possible to set `x` only; then it applies to all layer functions.
#' @param value value
#' @keywords internal
#' @export
tmapAddLayerOptions = function(option, id, value) {
	if (!(option %in% c("value.const", "value.na", "value.blank", "values.var", "values.range", "value.neutral", "scales.var"))) {
		stop("Unknown option")
	}

	d = .TMAP$defaultTmapOptions
	d[[option]][[id]] = value

	o = .TMAP$tmapOptions
	o[[option]][[id]] = value

	.TMAP$defaultTmapOptions = d
	.TMAP$tmapOptions = o
}

#' @rdname tmap_options
#' @export
tmap_options_diff <- function() {
	.tmapOptions <- get("tmapOptions", envir = .TMAP)
	iden <- mapply(identical, .tmapOptions, .TMAP$defaultTmapOptions)

	if (all(iden)) {
		message("current tmap options are similar to the default tmap options (style \"white\")")
	} else {
		message("current tmap options (style \"", attr(.tmapOptions, "style"), "\") that are different from default tmap options (style \"white\"):")
		.tmapOptions[!iden]
	}
}

#' @rdname tmap_options
#' @export
tmap_options_reset <- function() {
	assign("tmapOptions", .TMAP$defaultTmapOptions, envir = .TMAP)
	options(tmap.style="white")
	message("tmap options successfully reset")
	invisible(NULL)
}

#' @export
#' @param style style, see [tmap_style()] for available styles
#' @rdname tmap_options
tmap_options_save <- function(style) {
	show.messages <- get("tmapOptions", envir = .TMAP)$show.messages

	stylediff <- suppressMessages(tmap_options_diff())

	.tmapOptions <- get("tmapOptions", envir = .TMAP)

	if (is.null(stylediff)) {
		if (show.messages) message("current style is the same as the default style, so nothing to save")
		return(invisible(.tmapOptions))
	}

	options(tmap.style=style)
	attr(.tmapOptions, "style") <- style
	assign("tmapOptions", .tmapOptions, envir = .TMAP)

	styles <- get("tmapStyles", envir = .TMAP)
	styles[[style]] <- suppressMessages(tmap_options_diff())
	assign("tmapStyles", styles, envir = .TMAP)

	if (show.messages) message("current tmap options saved as style \"", style, "\"")
	invisible(.tmapOptions)
}

#' Internal method for submitting options
#'
#' Internal method for submitting options
#'
#' @param options list of options
#' @param styleOptions list of lists of options, each with a style name
#' @export
#' @keywords internal
tmapSubmitOptions = function(options = NULL, styleOptions = NULL) {
	o = .TMAP$defaultTmapOptions
	s = .TMAP$tmapStyles

	if (!is.null(options)) {
		backup = o[names(options)]
		o[names(options)] = complete_options(options, backup, erase_style = FALSE)
	}
	if (!is.null(styleOptions)) {
		for (st in names(styleOptions)) {
			backup = s[[st]]
			s[[st]] = complete_options(styleOptions[[st]], backup, erase_style = FALSE)
		}
	}

	# detect newly registered modes; reset pool if needed
	added = setdiff(names(o$modes), names(.TMAP$defaultTmapOptions$modes))
	if (length(added) > 0L && !is.null(.TMAP$mode_pool)) {
		.TMAP$mode_pool = NULL
		cli::cli_inform(c(
			"i" = "New mode{?s} registered: {.val {added}}",
			"i" = "Mode pool reset. Use {.run tmap_mode_pool()} to set a new pool."
		))
	}

	.TMAP$defaultTmapOptions = o
	.TMAP$tmapStyles = s
	suppressMessages(tmap_options_reset())
}

#' Internal method for submitting a new mode (deprecated)
#'
#' Internal method for submitting a new mode. Deprecated. Use [tmapSubmitOptions()] instead.
#'
#' @param id id of the mode: please use lowercase and one-word. This will be used with [tmap_mode()].
#' @param name name of the mode: please use title case. This will be used to recognize internal functions, e.g. `tmapLeafletInit`
#' @param ... mode specific options
#' @export
#' @keywords internal
tmapMode = function(id, name, ..., styleOptions = NULL) {
	d = .TMAP$defaultTmapOptions
	o = .TMAP$tmapOptions

	mlist = c(list(name = name), list(...))

	o$modes[[id]] = mlist
	d$modes[[id]] = mlist

	if (!is.null(styleOptions)) {
		os = get("tmapStyles", envir = .TMAP)
		for (s in names(styleOptions)) {
			os[[s]]$modes[[id]] = styleOptions[[s]]
			#lst = structure(list(modes = structure(styleOptions[s], names = id)), style = s)
			#suppressMessages(tmap_options(lst))
		}
		assign("tmapStyles", os, envir = .TMAP)
	}
	.TMAP$defaultTmapOptions = d
	.TMAP$tmapOptions = o
	suppressMessages(tmap_style(getOption("tmap.style")))
}

#' @rdname tmap_providers
#' @param provider provider name
#' @export
tmap_provider_credits = function(provider) {
	x = tmap_providers(credits = TRUE)
	if (provider %in% names(x)) {
		x[[provider]]
	} else
		""
}

#' Get basemap tiles providers
#'
#' Get basemap tiles providers and the credits (attribution text). [tmap_providers()] returns a list (or vector) with provider nams (or credits). [tmap_provider_credits()]
#'
#' @param mode mode. If not specified the default mode is used
#' @param credits If `TRUE` the credit (attribution) text is returned. If `FALSE` (default) the provider name.
#' @param as.list Should the output be returned as list where names are provider names? By default `TRUE` when `credits` is also `TRUE`.
#' @return list or vector (see `as.list`) with providers (or credits). [tmap_provider_credits()] returns the credits text for the provided provider.
#' @export
#' @name tmap_providers
tmap_providers = function(mode, credits = FALSE, as.list = credits) {
	if (missing(mode)) mode = getOption("tmap.mode")
	gs = tmap_graphics_name(mode)

	fun = paste0("tmap", gs, "Providers")
	x = do.call(fun, list(credits = credits))
	if (as.list) {
		x
	} else {
		unlist(x, use.names = FALSE)
	}
}

#' Create animation
#'
#' Create a gif animation or video from an animated tmap plot. First use [tm_animate()] or [tm_animate_fast()] to animate the plot, and then apply [tmap_animation()] to save it as a gif or video file (e.g. mp4).
#'
#' @param tm tmap or a list of tmap objects. If `tm` is a tmap object,
#'   animation frames should be created using either [tm_animate()] or
#'   [tm_animate_fast()].
#' @param filename filename. If omitted (default), the animation will be shown
#'   in the viewer or browser. If specified, it should be a gif file or a video
#'   file (i.e. mp4). The package `gifski` is required to create a gif animation.
#'   The package `av` (which uses the `FFmpeg` library) is required for video formats.
#'   The mp4 format is recommended but many other video formats are supported,
#'   such as wmv, avi, and mkv.
#' @param width,height Dimensions of the animation file (in pixels).
#'   Required when `tm` is a list, and recommended to specify in advance when
#'   `tm` is a `tmap` object. If not specified in the latter case, it will be
#'   determined by the aspect ratio of the map.
#' @param dpi dots per inch. By default 100, but this can be set with the option
#'   `animation.dpi` in [tmap_options()].
#' @param outer.margins (passed on to [tmap_save()]) overrides the outer.margins
#'   argument of `tm_layout()` (unless set to `NA`)
#' @param asp (passed on to [tmap_save()]) if specified, it overrides the `asp`
#'   argument of `tm_layout()`. Tip: set to `0` if map frame should be placed on
#'   the edges of the image.
#' @param scale (passed on to [tmap_save()]) overrides the scale argument of
#'   [tm_layout()] (unless set to `NA`)
#' @param ... arguments passed on to [av::av_encode_video()]
#' @note Not only tmap plots are supported, but any series of R plots.
#' @concept animation
#' @example ./examples/tmap_animation.R
#' @import tmaptools
#' @export
tmap_animation <- function(tm, filename = NULL, width = NA, height = NA, dpi = NA, outer.margins = NA, asp = NULL, scale = NA, ...) {
	args = list(...)
	args_called = names(rlang::call_match()[-1])

	if (any(c("fps", "delay") %in% args_called)) {
		cli::cli_inform("{.filed [{.fun tmap_animation}]} please specify the frames per second {.arg fps} in {.fun tm_animate}")
		args$fps = NULL
		args$delay = NULL
	}

	if ("loop" %in% names(args)) {

		if (args$loop) {
			cli::cli_inform("{.filed [{.fun tmap_animation}]} please use {.code play = {.str loop}} in {.fun tm_animate} instead of {.code loop = TRUE}")
		} else {
			cli::cli_inform("{.filed [{.fun tmap_animation}]} please use {.code play = {.str once}} in {.fun tm_animate} instead of {.code loop = FALSE}")
		}
		args$loop = NULL
	}

	.tmapOptions <- get("tmapOptions", envir = .TMAP)

	showAni = missing(filename)
	if (showAni) filename = tempfile(fileext = ".gif")

	progress = .tmapOptions$show.messages

	if (is.na(dpi)) dpi <- .tmapOptions$animation.dpi

	gif = substr(filename, nchar(filename) - 2, nchar(filename)) == "gif"

	#syscall <- if (.Platform$OS.type == "unix") system else shell ## macOS == unix

	# check system requirements
	if (gif) {
		rlang::check_installed("gifski", reason = "for creating gif animations.")
	} else {
		rlang::check_installed("av", reason = "for ffmpeg animations")
	}


	# create plots
	d <- paste(tempdir(), "/tmap_plots", sep="/")
	dir.create(d, showWarnings = FALSE)

	if (progress) cat("Creating frames\n")

	even = if (gif) round else {
		function(x) round(x / 2) *2
	} # av requires dimensions to be divisible by 2


	if (inherits(tm, "tmap_arrange") || (is.list(tm) && !inherits(tm, "tmap"))) {

		if (progress) pb = txtProgressBar()

		if (is.na(width) || is.na(height)) stop("The arguments width and height need to be specified both.")

		width = even(width)
		height = even(height)

		for (i in 1:length(tm)) {
			if (progress) setTxtProgressBar(pb, i/length(tm))
			tmi = tm[[i]]
			if (!inherits(tmi, "tmap")) stop("List item ", i, " of tm is not a tmap object.")
			suppressMessages(tmap_save(tm[[i]], filename = paste0(d, "/plot", sprintf("%03d", i), ".png"), width=width, height=height, dpi=dpi, outer.margins=outer.margins, asp=asp, scale=scale))
		}
	} else if (inherits(tm, "tmap")) {
		if (is.na(width) || is.na(height)) {
			sasp <- get_asp_ratio(tm, width = 700, height = 700, res = dpi)

			if (is.na(width) && is.na(height)) {
				height <- sqrt(.tmapOptions$output.size / sasp) * dpi
				width <- height * sasp
			} else if (is.na(width)) {
				width = height * sasp
			} else {
				height = width / sasp
			}
		}

		# round and in case of av to even numbers
		width = even(width)
		height = even(height)

		suppressMessages(tmap_save(tm, filename = paste(d, "plot%03d.png", sep="/"), width=width, height=height, dpi=dpi, outer.margins=outer.margins, asp=asp, scale=scale))
	} else {
		stop("tm should be a tmap object or a list of tmap objects")
	}

	fps = .TMAP$animation$fps
	play = .TMAP$animation$play

	files <- list.files(path = d, pattern = "^plot[0-9]{3}\\.png$", full.names = TRUE)

	if (play == "pingpong") files = c(files, rev(files))

	if (play != "once" && !gif) {
		cli::cli_inform("{.field [tmap_animation]} looped animations are only supported for gif videos")
	}

	do.call(create_animation, c(list(filename = filename, files, width = width, height = height, fps = fps, loop = (play != "once"), progress = progress, gif = gif, showAni = showAni), args))
}

create_animation = function(filename, files, width = NA, height = NA, fps = 2, loop = TRUE, progress = FALSE, gif = TRUE, showAni = FALSE, dpr = 2, knit = FALSE, knit_opts = NULL, ...) {

	k <- length(files)

	if (is.na(width) || is.na(height)) {
		x = stars::read_stars(files[1])
		width = dim(x)[1]
		height = dim(x)[2]
	}
	if (!is.na(fps)) delay = 100 / fps

	if (progress) cat("\nCreating animation\n")

	if (gif) {
		gifski::gifski(files,
					   delay = delay / 100,
					   width = width, height = height,
					   gif_file  = filename,
					   progress = progress,
					   loop = loop)
	} else {
		args = list(...)
		args = args[intersect(names(args), names(formals(av::av_encode_video)))]
		if (!("verbose" %in% names(args))) args$verbose = FALSE
		do.call(av::av_encode_video,
				c(list(files,
					   output = filename,
					   framerate = 100/delay),
				  args))
	}


	# cleaning up plots
	unlink(dirname(files[1]), recursive = TRUE)

	if (showAni) {
		file_base64 <- base64enc::dataURI(file = filename, mime = "image/gif")
		print(htmltools::browsable(
			htmltools::tagList(list(
				htmltools::tags$div(
					style = paste0(
						"width: auto",
						"height: auto"
					),
					htmltools::tags$img(
						src = file_base64,
						alt = "Animation",
						style = paste0(
							"max-width: 99%; ",
							"min-width: 99%; ",
							"width: auto; ",
							"height: auto; ",
							"margin: 0 auto; "
						)
					)
				)
			))
		)
		)
	} else {
		if (progress) cat("Animation saved to", suppressWarnings(normalizePath(filename)), "\n")
		invisible()
	}
}

#' @export
tmapLeafletCompPrepare.tm_title = function(comp, o) {
	comp$show = TRUE
	comp
}


#' @export
tmapLeafletCompHeight.tm_title = function(comp, o) {
	comp
}

#' @export
tmapLeafletCompWidth.tm_title = function(comp, o) {
	comp
}


#' @export
tmapLeafletCompPlot.tm_title = function(comp, lf, o) {
	if (inherits(lf, "shiny.tag.list")) {
		ncld <- length(lf[[1]])
		lf[[1]] <- mapply(function(l, i) {
			title <- l$children[[1]]$title %||% ""
			if (title!="") {
				l$children[[1]] <- l$children[[1]] %>% htmlwidgets::onRender(paste("
					function(el, x) {
						var tldiv = el.getElementsByClassName(\"leaflet-top leaflet-left\")[0];
						var titlediv = document.createElement('div');
						titlediv.className = \"info legend leaflet-control\";
						titlediv.innerHTML = \"<b>", title, "</b>\";
						tldiv.insertBefore(titlediv, tldiv.childNodes[0]);
					}", sep="")
				)
			}
			l
		}, lf[[1]], 0:(ncld-1), SIMPLIFY = FALSE)
	} else {
		title <- comp$text %||% ""
		if (title!="") {
			lf <- lf %>% htmlwidgets::onRender(paste("
						function(el, x) {
							var tldiv = el.getElementsByClassName(\"leaflet-top leaflet-left\")[0];
							var titlediv = document.createElement('div');
							titlediv.className = \"info legend leaflet-control\";
							titlediv.innerHTML = \"<b>", title, "</b>\";
							tldiv.insertBefore(titlediv, tldiv.childNodes[0]);
						}", sep="")
			)
		}
	}
	lf

}



#' @export
tmapLeafletCompPrepare.tm_compass = function(comp, o) {
	comp$show = TRUE
	comp
}


#' @export
tmapLeafletCompHeight.tm_compass = function(comp, o) {
	comp
}

#' @export
tmapLeafletCompWidth.tm_compass = function(comp, o) {
	comp
}


#' @export
tmapLeafletCompPlot.tm_compass = function(comp, lf, o) {
	lf %>% leafem::addLogo(system.file(paste0("img/compass_", comp$type, ".png"), package = "tmap"), position = leaflet_pos(comp$position), width = 120, height = 120)
}



#' @export
tmapLeafletCompPrepare.tm_scalebar = function(comp, o) {
	comp$show = TRUE
	comp
}


#' @export
tmapLeafletCompHeight.tm_scalebar = function(comp, o) {
	comp
}

#' @export
tmapLeafletCompWidth.tm_scalebar = function(comp, o) {
	comp
}


#' @export
tmapLeafletCompPlot.tm_scalebar = function(comp, lf, o) {
	lf %>% leaflet::addScaleBar(position = leaflet_pos(comp$position))
}




#' @export
tmapLeafletCompPrepare.tm_credits = function(comp, o) {
	comp$show = TRUE
	comp
}


#' @export
tmapLeafletCompHeight.tm_credits = function(comp, o) {
	comp
}

#' @export
tmapLeafletCompWidth.tm_credits = function(comp, o) {
	comp
}


#' @export
tmapLeafletCompPlot.tm_credits = function(comp, lf, o) {
	#lf %>% leafem::addLogo(system.file(paste0("img/credits_", comp$type, ".png"), package = "tmsap"), src = "local", position = paste(unlist(comp$position[c("pos.v", "pos.h")]), collapse = ""), width = 120, height = 120)
	#message("tm_credits not implemented yet for view mode")
	lf %>% leaflet::addTiles(urlTemplate = "", attribution = comp$text)
}


#' @export
tmapLeafletCompPrepare.tm_mouse_coordinates = function(comp, o) {
	comp$show = TRUE
	comp
}


#' @export
tmapLeafletCompHeight.tm_mouse_coordinates = function(comp, o) {
	comp
}

#' @export
tmapLeafletCompWidth.tm_mouse_coordinates = function(comp, o) {
	comp
}


#' @export
tmapLeafletCompPlot.tm_mouse_coordinates = function(comp, lf, o) {
	lf %>% leafem::addMouseCoordinates()
}



#' @export
tmapLeafletCompPrepare.tm_minimap = function(comp, o) {
	comp$show = TRUE
	comp$height = 20 * comp$height
	comp$width = 20 * comp$width

	extra = comp[setdiff(intersect(names(comp), names(formals(leaflet::addMiniMap))), c("position", "map"))]

	comp$specified_tiles = !is.na(comp$server)

	comp$args = c(list(toggleDisplay = comp$toggle), extra)
	comp
}


#' @export
tmapLeafletCompHeight.tm_minimap = function(comp, o) {


	comp
}

#' @export
tmapLeafletCompWidth.tm_minimap = function(comp, o) {
	comp
}


#' @export
tmapLeafletCompPlot.tm_minimap = function(comp, lf, o) {
	comp$args$tiles = if (comp$specified_tiles) {
		comp$server
	} else if (length(.TMAP_LEAFLET$tiles)) {
		.TMAP_LEAFLET$tiles[[1]][[1]]$server[1]
	} else {
		o$basemap.server[1]
	}
	lf2 = do.call(addMiniMap, c(list(map = lf, position = leaflet_pos(comp$position)), comp$args))
	if (!comp$specified_tiles && (length(comp$args$tiles) > 0)) {
		lf2 <- lf2 %>%
			htmlwidgets::onRender("
			    function(el, x) {
			      var myMap = this;
			      myMap.on('baselayerchange',
			        function (e) {
			          myMap.minimap.changeLayer(L.tileLayer.provider(e.name));
			        })
			    }")
	}
	lf2
}

#' @export
tmapLeafletCompPrepare.tm_logo = function(comp, o) {
	height = 20 * comp$height
	comp$logo = lapply(comp$file, function(lf){
		x = tmap_icons(lf)
		scale = height / x$iconHeight
		x$iconWidth = x$iconWidth * scale
		x$iconHeight = x$iconHeight * scale
		x$iconAnchorX = x$iconAnchorX * scale
		x$iconAnchorY = x$iconAnchorY * scale
		x
	})

	comp$asp = vapply(comp$logo, function(lg) {
		lg$iconWidth / lg$iconHeight
	}, FUN.VALUE = numeric(1))
	comp$show = TRUE
	comp
}


#' @export
tmapLeafletCompHeight.tm_logo = function(comp, o) {
	comp
}

#' @export
tmapLeafletCompWidth.tm_logo = function(comp, o) {
	comp
}

#' @export
tmapLeafletCompPlot.tm_logo = function(comp, lf, o) {
	ws = c(rev(cumsum(vapply(comp$logo, function(l)l$iconWidth,FUN.VALUE = numeric(1)) + 10))[-1], 0)

	for (i in 1:length(comp$logo)) {
		lf = lf %>% leafem::addLogo(comp$logo[[i]]$iconUrl, position = leaflet_pos(comp$position), width = comp$logo[[i]]$iconWidth, height = comp$logo[[i]]$iconHeight, offset.x = 10 + ws[i], offset.y = 20)
	}
	lf
}



#' @export
tmapLeafletCompPrepare.tm_inset_image = function(comp, o) {
	comp$file = comp$x
	comp = tmapLeafletCompPrepare.tm_logo(comp, o)
	class(comp) = c("tm_logo", class(comp))
	comp
}

#' @export
tmapLeafletCompPrepare.tm_inset_grob = function(comp, o) {
	height = 20 * comp$height
	width = 20 * comp$width

	grob.dim = c(width=width, height=height, render.width=width, render.height=height)

	comp$logo = list(grob2icon(comp$x, grob.dim = grob.dim, just = c(0.5, 0.5)))
	comp$logo = lapply(comp$logo, function(x){
		scale = height / x$iconHeight
		x$iconWidth = unname(x$iconWidth * scale)
		x$iconHeight = unname(x$iconHeight * scale)
		x$iconAnchorX = unname(x$iconAnchorX * scale)
		x$iconAnchorY = unname(x$iconAnchorY * scale)
		x
	})

	comp$asp = vapply(comp$logo, function(lg) {
		lg$iconWidth / lg$iconHeight
	}, FUN.VALUE = numeric(1))
	comp$show = TRUE
	class(comp) = c("tm_logo", class(comp))
	comp

}

#' @export
tmapLeafletCompPrepare.tm_inset_gg = function(comp, o) {
	rlang::check_installed("ggplot2", reason = "for plotting ggplot2 charts")
	comp$x = ggplot2::ggplotGrob(comp$x)
	tmapLeafletCompPrepare.tm_inset_grob(comp, o)
}

#' Specify icons
#'
#' Specifies icons from a png images, which can be used as markers in thematic maps.
#' The function `marker_icon()` is the specification of the default marker.
#'
#' @param file character value/vector containing the file path(s) or url(s).
#' @param names names to be given to the icons. Useful when icons are assigned to factor levels.
#' @param width width of the icon. If `keep.asp`, this is interpreted as the maximum width.
#' @param height height of the icon. If `keep.asp`, this is interpreted as the maximum height.
#' @param keep.asp keep the aspect ratio of the png image. If `TRUE` and the aspect
#'   ratio differs from `width/height`, either `width` or `height` is adjusted accordingly.
#' @param just justification of the icons relative to the point coordinates.
#'   The first value specifies horizontal and the second value vertical justification.
#'   Possible values are: `"left"` , `"right"`, `"center"`, `"bottom"`, and `"top"`.
#'   Numeric values of 0 specify left alignment and 1 right alignment.
#'   The default value of `just` is `c("center", "center")`.
#' @param merge merge icons to one icon list (see return value)? If `FALSE`, a list is created per file. By default `TRUE`, unless `names` are specified.
#' @param as.local if the `file` is a url, should it be saved to local temporary file?
#' @param ... arguments passed on to [leaflet::icons()].
#'   When `iconWidth`, `iconHeight`, `iconAnchorX`, and `iconAnchorY` are specified,
#'   they override `width` and `height`, and `just`.
#' @return icon data (see [leaflet::icons()])
#' @importFrom leaflet icons
#' @export
#' @seealso [tm_symbols()]
tmap_icons <- function(file, names = NULL, width=48, height=48, keep.asp=TRUE, just=c("center", "center"), merge = NA, as.local=TRUE, ...) {
	if (is.na(merge)) merge = is.null(names)
	icon_names <- names(file)
	icons <- lapply(file, tmap_one_icon, width=width, height=height, keep.asp=keep.asp, just=just, as.local=as.local, ...)
	if (merge) {
		merge_icons(icons, icon_names)
	} else {
		if (!is.null(names)) {
			if (length(names) != length(icons)) stop("Icons and names have different lengths")
			names(icons) = names
		}
		icons
	}

}

file_extension <- function(filenames) {
	sub(pattern = "^(.*\\.|[^.]+)(?=[^.]*)", replacement = "", filenames, perl = TRUE)
}

tmap_one_icon <- function(file, width, height, keep.asp, just, as.local, ...) {
	args <- list(...)
	args$iconUrl <- NULL # already specified with file
	pu <- is_path_or_url(file)
	if (is.na(pu)) {
		stop(file, " is neither a valid path nor url", call.=FALSE)
	}
	if (!pu) {
		tmpfile <- file.path(tempdir(), basename(file))
		if (!file.exists(tmpfile)) download.file(file, destfile=tmpfile, mode="wb")
		localfile <- tmpfile
	} else {
		localfile <- file
	}

	if (!pu && as.local) file <- localfile

	# use exact dimensions when provided
	if (any(c("iconWidth", "iconHeight") %in% names(args))) keep.asp <- FALSE

	# adjust to png dimensions
	if (keep.asp) {
		ext = file_extension(localfile)

		if (ext %in% c("png", "jpg", "bmp", "jpeg", "tiff")) {
			x = stars::read_stars(localfile)
			dms = unname(dim(x))
			dms[1:2] = dms[2:1]
		} else {
			rlang::check_installed("rsvg")
			bitmap <- rsvg::rsvg(localfile, width = width)
			dim(bitmap) # h*w*c
			dms = dim(bitmap)
		}

		xasp <- dms[2]/dms[1]
		iasp <- width/height
		if (xasp > iasp) {
			height <- floor(width/xasp)
		} else {
			width <- floor(height*xasp)
		}
	}

	# override dimensions
	if (!("iconWidth" %in% names(args))) args$iconWidth <- width
	if (!("iconHeight" %in% names(args))) args$iconHeight <- height

	just <- c(ifelse(is_num_string(just[1]), as.numeric(just[1]), ifelse(just[1]=="left", 1, ifelse(just[1]=="right", 0, .5))),
			  ifelse(is_num_string(just[2]), as.numeric(just[2]), ifelse(just[2]=="bottom", 1, ifelse(just[2]=="top", 0, .5))))

	# override anchor data
	if (!("iconAnchorX" %in% names(args))) args$iconAnchorX <- round(args$iconWidth * (1-just[1]))
	if (!("iconAnchorY" %in% names(args))) args$iconAnchorY <- round(args$iconHeight * just[2])

	do.call(leaflet::icons, c(list(iconUrl=file), args))
}



#' @rdname tmap_icons
#' @export
marker_icon <- function() {
	file <- system.file("htmlwidgets/lib/leaflet/images/marker-icon.png", package="leaflet")
	if (!file.exists(file)) stop("leaflet marker icon not found")
	structure(leaflet::icons(iconUrl = file, iconWidth=25, iconHeight=41, iconAnchorX = 12, iconAnchorY = 41), class = "tmap_icons")
}


# fix.borders is needed to prevent that right-hand side and bottom edges disappear
pngGrob <- function(file, fix.borders=FALSE, n=NULL, height.inch=NULL, target.dpi=NULL) {
	rlang::check_installed("png", reason = "for this function to work.")
	pu <- is_path_or_url(file)
	if (is.na(pu)) {
		stop(file, " is neither a valid path nor url", call.=FALSE)
	}

	if (!pu) {
		tmpfile <- file.path(tempdir(), basename(file))
		if (!file.exists(tmpfile)) download.file(file, destfile=tmpfile, mode="wb")
		file <- tmpfile
	}

	ext = file_extension(file)
	if (ext %in% c("png", "jpg", "bmp", "jpeg", "tiff")){
		x <- aperm(stars::read_stars(file)[[1]], c(2, 1, 3))
		if (max(x) > 1) {
			x[] = x[] / max(x)
		}
	} else {
		rlang::check_installed("rsvg")
		x <- rsvg::rsvg(file, height = height.inch * target.dpi)
	}

	if (fix.borders) {
		if (dim(x)[3]==3) {
			x <- array(c(x, rep(1, dim(x)[1]*dim(x)[2])), dim = c(dim(x)[1], dim(x)[2], dim(x)[3]+1))
		}
		x2 <- add_zero_borders_to_3d_array(x, n=n, height.inch=height.inch,target.dpi=target.dpi)
		rasterGrob(x2, interpolate=TRUE)
	} else {
		rasterGrob(x, interpolate=TRUE)
	}
}

add_zero_borders_to_3d_array <- function(x, perc=NA, n=NULL, height.inch=NULL, target.dpi=NULL) {
	dims <- dim(x)

	if (is.na(perc)) {
		dpi <- dims[2] / height.inch
		compress <- dpi/target.dpi
		borders <- rep(compress * n, 2)
	} else {
		borders <- round(dims / 100 * perc)
	}

	res <- lapply(1:dims[3], function(i) {
		rbind(cbind(x[,,i], matrix(0, nrow=nrow(x), ncol=borders[2])), matrix(0, nrow=borders[1], ncol=ncol(x)+borders[2]))
	})
	array(unlist(res, use.names = FALSE), dim = c(dim(x)[1]+borders[1], dim(x)[2]+borders[2], dim(x)[3]))
}



icon2grob <- function(icon) {
	if (!is.list(icon)) stop("icon is not a list")
	if (!"iconUrl" %in% names(icon)) stop("iconUrl not defined")
	if (length(icon$iconUrl)==1) {
		pngGrob(icon$iconUrl)
	} else {
		lapply(icon$iconUrl, pngGrob)
	}
}

grob2icon <- function(grob, grob.dim, just) {
	tmp <- tempfile(fileext=".png")
	png(filename=tmp, width=grob.dim[3], height=grob.dim[4], bg = "transparent")
	grid.draw(grob)
	dev.off()
	w <- grob.dim[1]
	h <- grob.dim[2]
	icons(iconUrl = tmp, iconWidth = w, iconHeight = h, iconAnchorX = w * (1-just[1]), iconAnchorY = h * just[2])
}

split_icon <- function(icon) {
	ni <- max(vapply(icon, length, integer(1)))
	icon_max <- lapply(icon, function(ic) {
		rep(ic, length.out=ni)
	})

	if ("iconNames" %in% names(icon_max)) {
		icon_names <- icon_max$iconNames
		icon_max$iconNames <- NULL
	} else {
		icon_names <- NULL
	}

	res <- lapply(1:ni, function(i) {
		lapply(icon_max, function(ic) {
			ic[i]
		})
	})

	if (!is.null(icon_names)) names(res) <- icon_names

	res
}

merge_icons <- function(icons, icon_names = NULL) {
	list_names <- unique(unlist(lapply(icons, names), use.names = FALSE))
	names(list_names) <- list_names

	res <- lapply(list_names, function(ln) {
		unname(sapply(icons, function(ic) {
			if (ln %in% names(ic)) {
				ic[[ln]][1]
			} else NA
		}))
	})
	if (!is.null(icon_names)) res$iconNames <- icon_names
	structure(res, class = "tmap_icons")
}


is_path_or_url <- function(file) {
	if (file.exists(file)) {
		TRUE
	} else {
		con.url <- suppressWarnings(try({
			u <- url(file, open='rb')
			close(u)
		}, silent=TRUE))
		try.error <- inherits(con.url, "try-error")
		if (try.error) NA else FALSE
	}
}

prepreprocess_meta = function(o, vp) {

	within(o, {
		vp = vp
		if (is.null(vp)) {
			devsize = graphics::par("fin") #dev.size() MAY NOT BE EQUAL IN RSTUDIO: https://github.com/rstudio/rstudio/issues/10723
		} else {
			if (is.character(vp)) seekViewport(vp) else pushViewport(vp)
			devsize = c(grid::convertWidth(grid::unit(1, "npc"), unitTo = "inch", valueOnly = TRUE),
						grid::convertHeight(grid::unit(1, "npc"), unitTo = "inch", valueOnly = TRUE))
		}

		# dasp device aspect ratio
		dasp = devsize[1] / devsize[2]

		# needed for spnc viewport (to retain aspect ratio)
		if (dasp > 1) {
			cw = dasp
			ch = 1
		} else {
			ch = 1/dasp
			cw = 1
		}

		lin = graphics::par("cin")[2]# * scale

		lineH = lin / devsize[2] * scale
		lineW = lin / devsize[1] * scale

		# not needed?
		#nlinesH = 1/lineH
		#nlinesW = 1/lineW

	})
}
preprocess_meta = function(o, cdt) {
	npp = NULL
	within(o, {
		nby = fn #get_nby(fl)
		isdef = !sapply(fl, is.null)

		if (is.na(panel.type)) panel.type = if (identical(panel.show, FALSE)) {
			"none"
		} else if (identical(panel.show, TRUE)) {
			# force panel labels
			if (type %in% c("wrap", "stack") || (npp == 1)) {
				"wrap"
			} else {
				"xtab"
			}
		} else if ((npp == 1) && is.na(panel.labels[[1]])) {
			if (n > 1 || any(isdef)) {
				"wrap"
			} else {
				"none"
			}
		} else if (!(type %in% c("wrap", "stack")) && !isdef[1] && !isdef[2]) {
			"none"
		} else if ((type %in% c("wrap", "stack")) || (npp == 1)) {
			"wrap"
		} else {
			"xtab"
		}

		inner.margins = get_option_class(inner.margins, class = main_class)

		# legend.present.auto:
		#   find out whether there are legends for all facets, per row, per col
		#   use them to automatically determine meta.margins (in preprocess_meta)
		# # legend.present.fix
		#	find legend boxes that are assigned to outer margins

		if (nrow(cdt) == 0) {
			legend.present.auto = c(all = FALSE, per_row = FALSE, per_col = FALSE, per_facet = FALSE)
			legend.present.fix = rep(FALSE, 4)
		} else {
			if (type %in% c("wrap", "stack")) {
				#o$legend.present.auto = c(all = any(is.na(cdt$by1__) & cdt$class == "autoout"), per_row = any(!is.na(cdt$by1__) & cdt$class == "autoout"), per_col = FALSE)
				legend.present.auto = c(all = any(cdt$class == "autoout" & is.na(cdt$by1__)),
										per_row = FALSE, per_col = FALSE,
										per_facet = any(cdt$class == "autoout" & !is.na(cdt$by1__)))
		} else {
				legend.present.auto = c(all = any(is.na(cdt$by1__) & is.na(cdt$by2__) & cdt$class == "autoout"),
										per_row = any(!is.na(cdt$by1__) & is.na(cdt$by2__) & cdt$class == "autoout"),
										per_col = any(is.na(cdt$by1__) & !is.na(cdt$by2__) & cdt$class == "autoout"),
										per_facet = any(!is.na(cdt$by1__) & !is.na(cdt$by2__) & cdt$class == "autoout"))
			}
			legend.present.fix = c(any(cdt$class == "out" & cdt$cell.v == "bottom"),
								   any(cdt$class == "out" & cdt$cell.h == "left"),
								   any(cdt$class == "out" & cdt$cell.v == "top"),
								   any(cdt$class == "out" & cdt$cell.h == "right"))
		}


		# in case there are per-facet legends but no no marginal legends, and nrows or ncols equals 1, place them outside (to do this, set them to all-facet here, change legend.position.all below accordingly, and finally determine legend position in step4_plot)
		legend_move_all = FALSE

		# legend_move_all is needed to move the 'global' legend not in a stacked-per-facets cell, but in the other orientation (#1071)

		if (legend.present.auto[4] && (!any(legend.present.auto[2:3]))) {

			if (type == "stack") {
				if (all(legend.present.auto[c(1, 4)])) legend_move_all = TRUE
				legend.present.auto[1] = TRUE
				legend.present.auto[4] = FALSE
				set_to_stack_message = FALSE
			} else {
				set_to_stack_message = TRUE
			}
		} else {
			set_to_stack_message = FALSE
		}

	})
}

process_meta = function(o, d, cdt, aux) {

	if (o$legend.only) {
		return(within(o, {
			meta.buffers = c(0, 0, 0, 0)
			meta.margins = c(0, 0, 0, 0)
			xylab.margins = rep(0, 4)
			panel.xtab.size = c(0, 0, 0, 0)
			panel.xtab.margin = rep(0, 4)
			panel.wrap.margin = rep(0, 4)
			grid.buffers = c(0, 0, 0, 0)
			grid.margins = c(0, 0, 0, 0)
			panel.wrap.size = c(0, 0, 0, 0)
			fixedMargins  =  outer.margins + meta.buffers * 2 + meta.margins
			nrows = 1L
			ncols = 1L

			between_marginH = between_margin * lineH
			between_marginW = between_margin * lineW


			#overall scale down factor for facets
			width_forn = max(1e-9, ((1 - sum(fixedMargins[c(2, 4)])) - (ncols * sum(panel.wrap.size[c(2,4)])) - (ncols - 1) * between_marginW) / ncols)
			width_for1 = max(1e-9, ((1 - sum(fixedMargins[c(2, 4)])) - (sum(panel.wrap.size[c(2,4)]))))

			height_forn = max(1e-9, ((1 - sum(fixedMargins[c(1, 3)])) - (nrows * sum(panel.wrap.size[c(1,3)])) - (nrows - 1) * between_marginH) / nrows)
			height_for1 = max(1e-9, ((1 - sum(fixedMargins[c(1, 3)])) - (sum(panel.wrap.size[c(1,3)]))))

			scale_down = (1 / sqrt((width_for1 * height_for1) / (width_forn * height_forn))) ^ (1 / scale.factor)

		}))
	}


	gs = tmap_graphics_name()

	# add tm_grid values to o
	gid = which(vapply(aux, FUN = inherits, "tm_grid", FUN.VALUE = logical(1)))[1]
	if (!is.na(gid)) {
		a = aux[[gid]]$args
		a$zindex = NULL
		a$group = NULL
		o[paste0("grid.", names(a))] = a
	}

	# add credits into to o (for view mode in order to reset default attribution text)
	cid = which(vapply(cdt$comp, FUN = inherits, "tm_credits", FUN.VALUE = logical(1)))[1]
	o$credits.defined = (!is.na(cid))

	bbx = d$bbox[[1]]
	within(o, {
		# sasp shape aspect ratio (NA if free coordinates)
		diff_asp = any(d$asp != d$asp[1])
		sasp = ifelse(diff_asp, NA, d$asp[1])

		# preferred aspect ratio (just for this function): if asp is defined (not 0 or NA), use that, otherwise use sasp (shape asp) if available (if not; 1)
		pasp = if (is.na(sasp)) {
			if (!is.na(asp) && asp > 0) {
				asp
			} else {
				1
			}
		} else {
			if (!is.na(asp) && asp > 0) {
				asp
			} else {
				sasp
			}
		}


		if (gs == "Grid") {

			bufferH = lineH / 2
			bufferW = lineW / 2

			# calculate space for margins, panels, etc

			meta.automatic = is.na(meta.margins[1])

			#one.row = (!is.na(o$nrows) && o$nrows == 1)
			#one.col = (!is.na(o$ncols) && o$ncols == 1)

			if (meta.automatic) meta.margins = c(0, 0, 0, 0) else meta.margins = rep(meta.margins, length.out = 4)

			meta.buffers = sign(meta.margins) * c(bufferH, bufferW, bufferH, bufferW) # outside and inside

			panel.xtab.size = if (panel.type == "xtab") {
				c(ifelse("bottom" %in% panel.xtab.pos, panel.label.height * lineH, 0),
				  ifelse("left" %in% panel.xtab.pos, panel.label.height * lineW, 0),
				  ifelse("top" %in% panel.xtab.pos, panel.label.height * lineH, 0),
				  ifelse("right" %in% panel.xtab.pos, panel.label.height * lineW, 0))
			} else c(0, 0, 0, 0)

			panel.margin = get_option_class(panel.margin, panel.type, spatial_class = FALSE)

			panel.xtab.margin = if (panel.type == "xtab") {
				c(ifelse("bottom" %in% panel.xtab.pos, panel.margin * lineH, 0),
				  ifelse("left" %in% panel.xtab.pos, panel.margin * lineW, 0),
				  ifelse("top" %in% panel.xtab.pos, panel.margin * lineH, 0),
				  ifelse("right" %in% panel.xtab.pos, panel.margin * lineW, 0))
			} else c(0, 0, 0, 0)

			panel.wrap.margin =	if (panel.type == "wrap") {
				c(ifelse(panel.wrap.pos == "bottom", panel.margin * lineH, 0),
				  ifelse(panel.wrap.pos == "left", panel.margin * lineW, 0),
				  ifelse(panel.wrap.pos == "top", panel.margin * lineH, 0),
				  ifelse(panel.wrap.pos == "right", panel.margin * lineW, 0))
			} else c(0, 0, 0, 0)


			panel.wrap.size = if (panel.type == "wrap") {
				c(ifelse(panel.wrap.pos == "bottom", panel.label.height * lineH, 0),
				  ifelse(panel.wrap.pos == "left", panel.label.height * lineW, 0),
				  ifelse(panel.wrap.pos == "top", panel.label.height * lineH, 0),
				  ifelse(panel.wrap.pos == "right", panel.label.height * lineW, 0))
			} else c(0, 0, 0, 0)

			xylab.margins = rep(0, 4)
			if (xlab.show) {
				#if (is.na(xlab.fontface)) xlab.fontface = text.fontface
				xylab.margins[ifelse(xlab.side == "bottom", 1, 3)] = if (xlab.rotation %in% c(0, 180)) {
					(number_text_lines(xlab.text) + xlab.space)  * lineH
				} else {
					(text_width_inch(xlab.text, space = FALSE) / lin + xlab.space) * lineH
				}
			}
			if (ylab.show) {
				#if (is.na(ylab.fontface)) ylab.fontface = text.fontface
				xylab.margins[ifelse(ylab.side == "left", 2, 4)] = if (ylab.rotation %in% c(90, 270)) {
					(number_text_lines(ylab.text) + ylab.space)  * lineW * ylab.size
				} else {
					(text_width_inch(ylab.text, space = FALSE) / lin + ylab.space) * lineW * ylab.size
				}
			}


			grid.buffers = if (grid.show) {
				as.integer(c("bottom", "left", "top", "right") %in% grid.labels.pos) * c(bufferH, bufferW, bufferH, bufferW)
			} else {
				rep(0, 4)
			}



			grid.labels.show = rep(grid.labels.show, length.out = 2) # also happens in tmapGridGridPrep
			if (grid.show && any(grid.labels.show) && !grid.labels.inside_frame) {
				proj = sf::st_crs(bbx)
				if (!is.na(o$grid.crs)) {
					bbx_orig <- bbx
					bbx <- suppressWarnings(bb(bbx, current.projection = proj, projection = o$grid.crs))
				}

				lineHin <- convertHeight(unit(grid.labels.size, "lines"), "inch", valueOnly=TRUE)

				if (grid.labels.show[1]) {
					gridx = pretty30(bbx[c(1,3)], n = 5, longlat = !is.na(o$grid.crs) && sf::st_is_longlat(proj))
					xbbstringWin <- max(convertWidth(stringWidth(do.call("fancy_breaks", c(list(vec=gridx, intervals = FALSE), grid.labels.format))), "inch", valueOnly = TRUE)) * grid.labels.size
					xgridHin <- ifelse(!is.na(grid.labels.space.x), grid.labels.space.x * lineHin, ifelse(grid.labels.rot[1] %in% c(0, 180), 1.375 * lineHin, xbbstringWin + lineHin * .75) + grid.labels.margin.x * lineHin)

				} else {
					xgridHin = 0
				}

				if (grid.labels.show[2]) {
					gridy = pretty30(bbx[c(2,4)], n = 5, longlat = !is.na(o$grid.crs) && sf::st_is_longlat(proj))
					ybbstringWin = max(
						convertWidth(
							stringWidth(do.call("fancy_breaks", c(
								list(vec=gridy, intervals=FALSE), grid.labels.format))), "inch", valueOnly = TRUE)
						)

					ybbstringWin = ybbstringWin * grid.labels.size
					ygridWin = ifelse(!is.na(grid.labels.space.y), grid.labels.space.y * lineHin, ifelse(grid.labels.rot[2] %in% c(0, 180), ybbstringWin + lineHin * .75, 1.375 * lineHin) + grid.labels.margin.y * lineHin)
				} else {
					ygridWin = 0
				}

				marks_new = c(xgridHin, ygridWin, xgridHin, ygridWin) / lin
				grid.margins = as.integer(c("bottom", "left", "top", "right") %in% grid.labels.pos) * marks_new * c(lineH, lineW, lineH, lineW)
			} else {
				grid.margins = rep(0, 4)
			}

			between_marginH = between_margin * lineH
			between_marginW = between_margin * lineW

			fixedMargins = outer.margins + meta.buffers * 2 + meta.margins + xylab.margins + panel.xtab.size + grid.buffers + grid.margins
		} else {
			grid.buffers = rep(0, 4)
			grid.labels.show = c(FALSE, FALSE)
			grid.margins = rep(0, 4)
			fixedMargins = rep(0, 4)
			panel.wrap.size = rep(0, 4)
			between_marginH = 0
			between_marginW = 0
		}


		masp = ((1 - sum(fixedMargins[c(2, 4)])) / (1 - sum(fixedMargins[c(1, 3)]))) * dasp

		# Aspect ratios:
		# sasp: shape
		# asp: user specified
		# pasp: prefered (asp or if not specified, sasp)
		# masp: multiples (facets) area

		# determine where to place automatic legends (i.e. legends with local legend.position = NA and with legend.position = tm_pos_auto_out() enabled)
		# this is also neede to find out which margins are taken from meta.auto_margins

		legend.position.sides = legend.position
		legend.position.all = legend.position

	#	legsG = cdt[, leg]

		# determine orientation of stacked maps
		# it also implies where legends will be drawn: horizontal orientation=legends bottom or top, vertical orientation=legends left or right
		# !!! this also applies for single maps
		mx_width = (1 - sum(fixedMargins[c(1, 3)])) * devsize[1]
		mx_height = (1 - sum(fixedMargins[c(2, 4)])) * devsize[2]

		# cdt[, scale := pmax(legW/mx_width, legH/mx_height, 1)]
		# cdt[, ":="(legW_sc = legW/scale, legH_sc = legH/scale)]


		if (gs == "Grid") {
			if (type =="stack") {
				if (is.na(orientation)) {
					if (nrow(cdt)) {
						legs_auto = cdt[class=="autoout"]
					} else {
						legs_auto = cdt
					}


					if (nrow(legs_auto) && npp == 1) {


						legWmax = min(max(legs_auto$legW) / devsize[1], max(meta.auto_margins[c(2,4)]))
						legHmax = min(max(legs_auto$legH) / devsize[2], max(meta.auto_margins[c(1,3)]))


						av_width = mx_width - legWmax * devsize[1]
						av_height = mx_height - legHmax * devsize[2]

						shp_height_hor = if ((av_width / mx_height) < pasp) av_width / pasp else mx_height
						shp_height_ver = if ((mx_width / av_height) < pasp) mx_width / pasp else av_height

						orientation = if (shp_height_hor >= shp_height_ver) "vertical" else "horizontal"
					} else {
						orientation = if ((!is.na(nrows) && nrows == 1) || (!is.na(ncols) && ncols == npp)) {
							"horizontal"
						} else if ((!is.na(nrows) && nrows == npp) || (!is.na(ncols) && ncols == 1)) {
							"vertical"
						} else if ((npp == 1 && (pasp > masp)) || (npp > 1 && (pasp < masp))) "horizontal" else "vertical"
					}
				}
			}
		} else {
			if (is.na(nrows) && is.na(ncols)) {
				orientation = if ((npp == 1 && (pasp > masp)) || (npp > 1 && (pasp < masp))) "horizontal" else "vertical"
			} else {
				if (is.na(nrows)) nrows = floor(npp / ncols)
				if (is.na(ncols)) ncols = floor(npp / nrows)
				orientation = if (ncols >= nrows) "horizontal" else "vertical"
			}
		}

		if (gs == "Grid") {
			## find position for all-facet legend
			if (legend.present.auto[1]) {
				if (!legend.present.auto[2] & !legend.present.auto[3]) {
					# only 'all facets' outside legends (either bottom or right)
					# was: n > 1 && masp > pasp
					legend.position.all = local({
						moving = function(x) if (legend_move_all) !x else x
						if (moving((type != "stack" && npp == 1 && pasp > masp) || (type != "stack" && npp > 1 && masp < 1) || (type == "stack" && orientation == "horizontal"))) {
							list(cell.h = "center", cell.v = legend.position$cell.v)
						} else {
							list(cell.h = legend.position$cell.h, cell.v = "center")
						}
					})

				} else if (legend.present.auto[2] & !legend.present.auto[3]) {
					# central goes center bottom
					legend.position.all = list(cell.h = "center", cell.v = legend.position$cell.v)
				} else if (!legend.present.auto[2] & legend.present.auto[3]) {
					# central goes right center
					legend.position.all = list(cell.h = legend.position$cell.h, cell.v = "center")
				}
			}

			if (legend_move_all) {
				# 'all' legend is set via legend.position.all
				# the per stack-per-facet legend is 'the other' position
				# so both (left or right) and (top or bottom) should be enabled
				margins.used.all = c(legend.position$cell.v == "bottom",
									 legend.position$cell.h == "left",
									 legend.position$cell.v == "top",
									 legend.position$cell.h == "right") * legend.present.auto[1]

			} else {
				margins.used.all = c(legend.position.all$cell.v == "bottom",
									 legend.position.all$cell.h == "left",
									 legend.position.all$cell.v == "top",
									 legend.position.all$cell.h == "right") * legend.present.auto[1]
			}

			margins.used.sides = c(bottom = legend.position.sides$cell.v == "bottom",
								   left = legend.position.sides$cell.h == "left",
								   top = legend.position.sides$cell.v == "top",
								   right = legend.position.sides$cell.h == "right") * legend.present.auto[c(3,2,3,2)]


			margins.used =  margins.used.all | margins.used.sides | legend.present.fix

			# tm_shape(World) + tm_polygons(fill = "HPI", lwd = "life_exp")

			if (nrow(cdt)) {
				cdt2 = data.table::copy(cdt[cdt$class %in% c("autoout", "out"),])

				# CODE COPIED FROM STEP4_plot L157
				# TO DO: fix this
				if (o$type != "grid" && o$npp > 1) {
					#if (o$nrows == 1 && o$ncols == 1)
					if (identical(orientation, "horizontal")) {
						# -use by2 and not by1 when they form a row
						cdt2[, by2__ := by1__]
						cdt2[, by1__ := NA]
					}
				}



				stacks = o$component.stack

				cdt2[is.na(by1__) & is.na(by2__) & class == "autoout", ':='(cell.h = legend.position.all$cell.h, cell.v = legend.position.all$cell.v)]

				cdt2[!is.na(by1__) & is.na(by2__) & class == "autoout", ':='(cell.h = legend.position.sides$cell.h, cell.v = "by")]
				cdt2[is.na(by1__) & !is.na(by2__) & class == "autoout", ':='(cell.h = "by", cell.v = legend.position.sides$cell.v)]

				# in case tm_pos is called to update the 'non-by' cell
				cdt2$calledCV = vapply(cdt2$comp, function(comp) "cell.v" %in% comp$position$called, FUN.VALUE = logical(1))
				cdt2$calledCH = vapply(cdt2$comp, function(comp) "cell.h" %in% comp$position$called, FUN.VALUE = logical(1))
				cdt2[!is.na(by1__) & is.na(by2__) & !calledCV, ':='(cell.v = "by")]
				cdt2[is.na(by1__) & !is.na(by2__) & !calledCH, ':='(cell.h = "by")]

				cdt2[!is.na(by1__) & is.na(by2__) & !calledCV, ':='(stack = ifelse(stack_auto, stacks["per_row"], stack))]
				cdt2[is.na(by1__) & !is.na(by2__) & !calledCH, ':='(stack = ifelse(stack_auto, stacks["per_col"], stack))]

				cdt2$calledCV = NULL
				cdt2$calledCH = NULL


				cdt2[is.na(by1__) & is.na(by2__) & class == "autoout", ':='(stack = ifelse(stack_auto, ifelse(cell.h == "center", stacks["all_col"], ifelse(cell.v == "center", stacks["all_row"], stacks["all"])), stack))]
				cdt2[!is.na(by1__) & is.na(by2__) & class == "autoout", ':='(stack = ifelse(stack_auto, stacks["per_row"], stack))]
				cdt2[is.na(by1__) & !is.na(by2__) & class == "autoout", ':='(stack = ifelse(stack_auto, stacks["per_col"], stack))]


				cdt2[class == "autoout", class := "out"]

				if (nrow(cdt2) == 0) {
					meta.auto_margins = c(0, 0, 0, 0)
				} else {
					if (type %in% c("stack", "anistack")) {
						# workaround: stacking mode is determined later (step4 L156), because it requires ncols and nrows
						# for stack, this is already known, so therefore we can better estimate the meta width and height

						cdt2[is.na(by1__), by1__:=1]

						meta.auto_margins = pmin(meta.auto_margins, do.call(pmax, lapply(unique(cdt2$by1__), function(b1) {
							cdt2b = cdt2[by1__==b1, ]

							cdt2b[stack_auto == TRUE, stack:= ifelse(npp==1, ifelse(cell.h %in% c("left", "right"), o$legend.stack["all_row"], o$legend.stack["all_col"]), ifelse(orientation == "vertical", o$legend.stack["per_row"], o$legend.stack["per_col"]))]

							c(max(max(c(0,cdt2b[cell.v == "bottom" & stack == "vertical", .(V=sum(legH)),by = c("cell.h", "by3__")]$V)),
								  max(c(0,cdt2b[cell.v == "bottom" & stack == "horizontal", .(V=max(c(0, legH))),by = c("cell.h", "by3__")]$V))) / o$devsize[2],
							  max(max(c(0,cdt2b[cell.h == "left" & stack == "horizontal", .(V=sum(legW)),by = c("cell.v", "by3__")]$V)),
							  	max(c(0,cdt2b[cell.h == "left" & stack == "vertical", .(V=max(c(0, legW))),by = c("cell.v", "by3__")]$V))) / o$devsize[1],
							  max(max(c(0,cdt2b[cell.v == "top" & stack == "vertical", .(V=sum(legH)),by = c("cell.h", "by3__")]$V)),
							  	max(c(0,cdt2b[cell.v == "top" & stack == "horizontal", .(V=max(c(0, legH))),by = c("cell.h", "by3__")]$V))) / o$devsize[2],
							  max(max(c(0,cdt2b[cell.h == "right" & stack == "horizontal", .(V=sum(legW)),by = c("cell.v", "by3__")]$V)),
							  	max(c(0,cdt2b[cell.h == "right" & stack == "vertical", .(V=max(c(0, legW))),by = c("cell.v", "by3__")]$V))) / o$devsize[1])
						})))
					} else {
						meta.auto_margins = pmin(meta.auto_margins,
												 c(max(cdt$legH[cdt$cell.v == "bottom" & cdt$class %in% c("autoout", "out")], 0) / o$devsize[2],
												   max(cdt$legW[cdt$cell.h == "left" & cdt$class %in% c("autoout", "out")], 0) / o$devsize[1],
												   max(cdt$legH[cdt$cell.v == "top" & cdt$class %in% c("autoout", "out")], 0) / o$devsize[2],
												   max(cdt$legW[cdt$cell.h == "right" & cdt$class %in% c("autoout", "out")], 0) / o$devsize[1]))
					}

					# add margins (compensate for legend frames)
					# the final calculations of these margins are computed in tmapGridComp (this is just to compute the meta.auto_margins)
					# those calculations are take the component.offset into account

					sel_tb = c(3,1)[meta.auto_margins[c(3,1)]!=0]
					sel_lr = c(2,4)[meta.auto_margins[c(2,4)]!=0]
					if (length(sel_tb)) meta.auto_margins[sel_tb] = meta.auto_margins[sel_tb] + 2 * (o$frame.lwd * o$scale / 144) / o$devsize[2]
					if (length(sel_lr)) meta.auto_margins[sel_lr] = meta.auto_margins[sel_lr] + 2 * (o$frame.lwd * o$scale / 144) / o$devsize[1]
				}
			}

			if (meta.automatic && any(margins.used)) {
				meta.auto_margins = rep(meta.auto_margins, length.out = 4)
				meta.margins[margins.used] = meta.auto_margins[margins.used]

				# redo calculations
				meta.buffers = sign(meta.margins) * c(bufferH, bufferW, bufferH, bufferW) # outside and inside
				fixedMargins  =  outer.margins + meta.buffers * 2 + meta.margins + xylab.margins + panel.xtab.size + grid.buffers + grid.margins
			}
		} else {
			meta.buffers = c(0, 0, 0, 0)
			meta.margins = c(0, 0, 0, 0)
		}

		# determine number of rows and cols
		npp = prod(nby[1:2]) # number per page

		if (!is.na(o$ncols) && !is.na(o$nrows)) {
			npp = min(npp, o$ncols * o$nrows)
		}

		if (type %in% c("grid", "anigrid")) {
			nrows = nby[1]
			ncols = nby[2]
		} else if (type %in% c("stack", "anistack")) {
			if (orientation == "horizontal") {
				nrows = 1
				ncols = npp
			} else {
				nrows = npp
				ncols = 1
			}
		} else {
			if (is.na(nrows) && !is.na(ncols)) {
				nrows = ceiling((nby[1] / ncols))
			} else if (!is.na(nrows) && is.na(ncols)) {
				ncols = ceiling((nby[1] / nrows))
			} else if (is.na(nrows) && is.na(ncols)) {

				# loop through col row combinations to find best nrow/ncol
				# b needed to compare landscape vs portrait. E.g if prefered asp is 2, 1 is equally good as 4
				ncols = which.min(vapply(1L:npp, function(nc) {
					nr = ceiling(npp / nc)

					# calculate available width and height. They can be negative, at this stage this is avoided my taking at least a small number
					width = max(1e-9, ((1 - sum(fixedMargins[c(2, 4)])) - (nc * sum(panel.wrap.size[c(2,4)])) - (nc - 1) * between_marginW) / nc)
					height = max(1e-9, ((1 - sum(fixedMargins[c(1, 3)])) - (nr * sum(panel.wrap.size[c(1,3)])) - (nr - 1) * between_marginH) / nr)

					a = (width / height) * dasp
					b = ifelse(a<pasp, pasp/a, a/pasp)
					b
				}, FUN.VALUE = numeric(1)))


				nrows = ceiling(npp / ncols)
			}
			if ((nrows == 1 || ncols == 1) && set_to_stack_message) message_wrapstack(nrows == 1)
		}

		#overall scale down factor for facets
		width_forn = max(1e-9, ((1 - sum(fixedMargins[c(2, 4)])) - (ncols * sum(panel.wrap.size[c(2,4)])) - (ncols - 1) * between_marginW) / ncols)
		width_for1 = max(1e-9, ((1 - sum(fixedMargins[c(2, 4)])) - (sum(panel.wrap.size[c(2,4)]))))

		height_forn = max(1e-9, ((1 - sum(fixedMargins[c(1, 3)])) - (nrows * sum(panel.wrap.size[c(1,3)])) - (nrows - 1) * between_marginH) / nrows)
		height_for1 = max(1e-9, ((1 - sum(fixedMargins[c(1, 3)])) - (sum(panel.wrap.size[c(1,3)]))))

		scale_down = (1 / sqrt((width_for1 * height_for1) / (width_forn * height_forn))) ^ (1 / scale.factor)

		#
		# title.size = title.size * scale
		# legend.title.size = legend.title.size * scale
		# legend.text.size = legend.text.size * scale
		#
		# panel.label.size = panel.label.size * scale


		if (is.na(panel.labels[1])) {
			# quick fix
			# to do: checks
			panel.labels = fl[1:3]
		} else {
			if (!is.list(panel.labels)) {
				panel.labels = list(panel.labels, NULL)
			}
		}
		if (type %in% c("stack", "wrap")) {
			if (!is.null(panel.labels[[1]])) {
				panel.labels.dim = 1
			} else if (!is.null(panel.labels[[3]])) {
				panel.labels.dim = 3
			} else {
				panel.type = "none"
			}
			if (nby[3] != 1L) {
				npages = nby[3]
			} else {
				npages = ceiling(n / (nrows * ncols))
			}
		} else {
			npages = ceiling(n / (nrows * ncols))
		}




		legend.position = NA

		if (is.na(space_overlay)) {
			space_overlay = .TMAP$raster_wrap
		}

		if (gs == "Leaflet") {
			if (!is.logical(set_bounds)) if (length(set_bounds) !=4 || !is.numeric(set_bounds)) stop("Incorrect set_bounds argument", call.=FALSE)


			if (!is.na(set_view[1])) {
				if (!is.numeric(set_view)) stop("set_view is not numeric")
				if (!length(set_view) %in% c(1, 3)) stop("set_view does not have length 1 or 3")
			}
			if (!is.na(set_zoom_limits[1])) {
				if (!is.numeric(set_zoom_limits)) stop("set_zoom_limits is not numeric")
				if (!length(set_zoom_limits)==2) stop("set_zoom_limits does not have length 2")
				if (set_zoom_limits[1] >= set_zoom_limits[2]) stop("incorrect set_zoom_limits")
			} else {
				set_zoom_limits <- c(NA, NA)
			}
			if (!is.na(set_view[1]) && !is.na(set_zoom_limits[1])) {
				if (set_view[length(set_view)] < set_zoom_limits[1]) {
					if (show.warnings) warning("default zoom smaller than minimum zoom, now it is set to the minimum zoom")
					set_view[length(set_view)] <- set_zoom_limits[1]
				}
				if (set_view[length(set_view)] > set_zoom_limits[2]) {
					if (show.warnings) warning("default zoom larger than maximum zoom, now it is set to the maximum zoom")
					set_view[length(set_view)] <- set_zoom_limits[2]
				}
			}
		}

	})

}

#' @rdname tmap_providers
#' @param provider provider name
#' @export
tmap_provider_credits = function(provider) {
	x = tmap_providers(credits = TRUE)
	if (provider %in% names(x)) {
		x[[provider]]
	} else
		""
}

#' Get basemap tile providers
#'
#' Get basemap tile providers and their credits (attribution text).
#' \code{tmap_providers()} returns a list or vector of provider names or
#' credits. \code{tmap_provider_credits()} returns the attribution text for a
#' specific provider.
#'
#' \code{.tmap_providers} is an environment populated with all available
#' provider names as named entries. Its primary purpose is to enable
#' autocomplete in IDEs such as RStudio: typing \code{.tmap_providers$} in the
#' console or a script triggers a dropdown list of all available providers,
#' making it easy to discover and select provider names without consulting the
#' documentation. It is not intended to be called as a function.
#'
#' @param mode mode. If not specified the current active mode is used.
#' @param credits If \code{TRUE} the credit (attribution) text is returned.
#'   If \code{FALSE} (default) the provider name.
#' @param as.list Should the output be returned as a list where names are
#'   provider names? By default \code{TRUE} when \code{credits} is also
#'   \code{TRUE}.
#' @return \code{tmap_providers()} returns a list or vector (see
#'   \code{as.list}) of provider names or credits.
#'   \code{tmap_provider_credits()} returns the attribution text for the
#'   specified provider. \code{.tmap_providers} is an environment; see
#'   Details.
#' @export
#' @name tmap_providers
#' @examples
#' # List all providers for the current mode
#' tmap_providers()
#'
#' # Use IDE autocomplete to discover providers interactively:
#' # type .tmap_providers$ in the RStudio console
tmap_providers = function(mode, credits = FALSE, as.list = credits) {
	if (missing(mode)) mode = getOption("tmap.mode")
	gs = tmap_graphics_name(mode)
	fun = paste0("tmap", gs, "Providers")
	x = do.call(fun, list(credits = credits))
	if (as.list) {
		x
	} else {
		unlist(x, use.names = FALSE)
	}
}

update_scale_args = function(scaleName, args, aes, o) {
	if (scaleName %in% names(o$scale.misc.args)) {
		def = o$scale.misc.args[[scaleName]]
		for (nam in names(def)) {
			if (is.null(args[[nam]])) args[[nam]] = getAesValue(def[[nam]], aes)
		}
	}
	args
}


# extention of substr: when -1, then cound from tail
substr2 = function(x, start, stop) {
	if (start < 1 || stop < 1) {
		n = nchar(x)
		if (start < 1) {
			start = n + start
		}
		if (stop < 1) {
			stop = n + stop
		}
	}
	substr(x, start, stop)
}


tmapScaleContinuous = function(x1, scale, legend, chart, o, aes, layer, layer_args, sortRev, bypass_ord, submit_legend = TRUE) {
	# update misc argument from tmap option scale.misc.args
	scale = update_scale_args(substr2(class(scale)[1], 10, 0), scale, aes, o)

	cls = data_class(x1, midpoint_enabled = !is.null(scale$midpoint))
	maincls = class(scale)[1]

	if (attr(cls, "unique") && any(is.na(scale$limits)) && is.null(scale$ticks)) {
		scale$ticks = NA
		cli::cli_inform("The visual variable {.arg {aes}} of the layer {.str {layer}} contains a unique value. Therefore a discrete scale is applied (tm_scale_discrete).")
		return(tmapScaleDiscrete(x1, scale, legend, chart, o, aes, layer, layer_args, sortRev, bypass_ord, submit_legend = submit_legend))
	}

	#if (cls[1] == "na") stop("data contain only NAs, so ", maincls, " cannot be applied", call. = FALSE)

	if (cls[1] != "num") {
		if (!is.factor(x1)) x1 = as.factor(x1)
		x1 = as.integer(x1)
		if (length(x1) > 1) warning(maincls, " is supposed to be applied to numerical data", call. = FALSE)
	}

	x1 = without_units(x1)

	if (aes %in% c("lty", "shape", "pattern")) stop("tm_scale_continuous cannot be used for layer ", layer, ", aesthetic ", aes, call. = FALSE)

	scale = get_scale_defaults(scale, o, aes, layer, cls)

	vnum = (is.numeric(scale$values) || inherits(scale$values, "tmapSeq"))
	#vnum=FALSE

	show.messages <- o$show.messages
	show.warnings <- o$show.warnings

	with(scale, {
		allna = all(is.na(x1))


		# if (all(is.na(x1))) {
		# }


		tr = get(paste0("trans_", trans))

		trargs = formals(tr$fun)
		trargs$x = NULL
		trargs_match = intersect(names(trargs), names(trans.args))
		if (!is.null(trargs_match) && length(trargs_match)) {
			trargs[trargs_match] = trans.args[trargs_match]
		}


		if (!allna) {
			xrange = range(x1, na.rm = TRUE)

			if (xrange[1] < tr$domain[1]) stop("Values found that are lower than the valid domain", call. = FALSE)
			if (xrange[2] > tr$domain[2]) stop("Values found that are higher than the valid domain", call. = FALSE)
		}

		udiv = identical(use_div(brks = NULL, midpoint), TRUE)

		ticks.specified = !is.null(ticks)
		limits.specified = (length(limits) == 2L)

		if (!limits.specified) {
			if (allna) {
				if (show.messages) message("Variable(s) \"", paste(aes, collapse = "\", \""), "\" only contains NAs. Legend disabled for tm_scale_continuous, unless limits are specified")
				chart = within(chart, {
					labels = label.na
					vvalues = c(value.na, value.na)
					breaks = c(0, 1)
					na.show = TRUE
					x1 = x1[1]
				})
				return(tmapScale_returnNA(n = length(x1), legend = legend, chart = chart, value.na = value.na, label.na = label.na, label.show = label.show, na.show = legend$na.show, sortRev = sortRev, bypass_ord = bypass_ord))
			}
			if (is.na(limits)) {
				limits = range(x1, na.rm = TRUE)
			} else {
				if (length(limits) == 1L) limits = range(c(limits, x1), na.rm = TRUE)
				if (limits[1] < tr$domain[1]) limits[1] = tr$domain[1]
				if (limits[2] > tr$domain[2]) limits[2] = tr$domain[2]
			}
			if (ticks.specified) limits = range(c(limits, ticks))
		}

		if (limits[1] < tr$domain[1]) stop("Lower limit too low", call. = FALSE)
		if (limits[2] > tr$domain[2]) stop("Upper limit too high", call. = FALSE)

		if (ticks.specified) {
			if (limits.specified) {
				if (any(ticks < limits[1])) stop("(Some) ticks are lower than the lowest limit. Please remove these ticks or adjust the lower limit.", call. = FALSE)
				if (any(ticks > limits[2])) stop("(Some) ticks are higher than the upper limit. Please remove these ticks or adjust the upper limit.", call. = FALSE)
			}
			n = length(ticks) - 1
			ticks_t = do.call(tr$fun, c(list(x = ticks), trargs))
		}

		if (limits.specified) {
			outliers.trunc = rep(outliers.trunc, length.out = 2)
			x1_low = x1 < limits[1]
			x1_high = x1 > limits[2]
			if (any(x1_low, na.rm = TRUE)) {
				if (!outliers.trunc[1]) message("Values have been found that are lower than the lowest limit. These 'outliers' have been set to NA. Use outliers.trunc to truncate them to the lower limit")
				x1[which(x1_low)] = if (outliers.trunc[1]) limits[1] else NA
			}
			if (any(x1_high, na.rm = TRUE)) {
				if (!outliers.trunc[2]) message("Values have been found that are higher than the upper limit. These 'outliers' have been set to NA. Use outliers.trunc to truncate them to the upper limit")
				x1[which(x1_high)] = if (outliers.trunc[2]) limits[2] else NA
			}
		}

		x_t = do.call(tr$fun, c(list(x = x1), trargs))
		limits_t = do.call(tr$fun, c(list(x = limits), trargs))
		domain_t = do.call(tr$fun, c(list(x = tr$domain), trargs))

		if (!vnum) {
			breaks = cont_breaks(limits_t, n=o$continuous.nclasses)
			if (is.null(labels)) {
				ncont = n
			} else {
				if (ticks.specified && length(labels) != n+1) {
					if (show.warnings) warning("The length of legend labels is ", length(labels), ", which differs from the length of the breaks (", (n+1), "). Therefore, legend labels will be ignored", call.=FALSE)
					labels = NULL
				} else {
					ncont = length(labels)
				}
			}
			q = num2breaks(x = x_t, n = o$continuous.nclasses, style = "fixed", breaks=breaks, approx=TRUE, interval.closure = "left", var=paste(layer, aes, sep = "-"), args = list())

			breaks = q$brks
			nbrks = length(breaks)
			n2 = nbrks - 1

		}




		# update range if NA (automatic)
		if (is.na(values.range[1])) {
			fun_range = paste0("tmapValuesRange_", aes)
			values.range = do.call(fun_range, args = list(x = values, n = n, isdiv = udiv))
		}
		if (length(values.range) == 1 && !is.na(values.range[1])) values.range = c(0, values.range)


		check_values(layer, aes, values)

		fun_isdiv = paste0("tmapValuesIsDiv_", aes)

		isdiv = !is.null(midpoint) || do.call(fun_isdiv, args = list(x = values))

		# determine midpoint
		if ((is.null(midpoint) || is.na(midpoint)) && isdiv) {
			rng <- range(x_t, na.rm = TRUE)
			if (rng[1] < 0 && rng[2] > 0 && is.null(midpoint)) {

				if (show.messages) message("Variable(s) \"", paste(aes, collapse = "\", \""), "\" contains positive and negative values, so midpoint is set to 0. Set midpoint = NA to show the full range of visual values.")
				midpoint <- 0
			} else {
				if (vnum) {
					midpoint = mean(limits_t)
				} else {
					if ((n2 %% 2) == 1) {
						# number of classes is odd, so take middle class (average of those breaks)
						midpoint <- mean.default(breaks[c((n2+1) / 2, (n2+3) / 2)])
					} else {
						midpoint <- breaks[(n2+2) / 2]
					}
				}
			}
		}

		fun_getVV = paste0("tmapValuesVV_", aes)

		if (!vnum) {
			#### discretisize

			# number of visual values in legend item (belonging to one label)
			nvv = o$continuous.nclass_per_legend_break


			VV = do.call(fun_getVV, list(x = values, value.na = value.na, isdiv = isdiv, n = n2, dvalues = breaks, midpoint = midpoint, range = values.range, scale = values.scale * o$scale, are_breaks = TRUE, rep = values.repeat, o = o))

			vv = VV$vvalues
			value.na = VV$value.na

			sfun = paste0("tmapValuesScale_", aes)
			cfun = paste0("tmapValuesColorize_", aes)
			if (is.na(value.neutral)) value.neutral = VV$value.neutral else value.neutral = do.call(sfun, list(x = do.call(cfun, list(x = value.neutral, pc = o$pc)), scale = values.scale))

			ids = classInt::findCols(q)
			vals = vv[ids]
		} else {
			if (is.numeric(values)) {
				values = tm_seq(values[1], values[length(values)], power = "lin")
			}
			VV = transform_values(x_t, limits_t, values.range, values$power, values.scale * o$scale)

			vals = VV$x
			if (is.na(value.neutral)) value.neutral = VV$neutral
		}

		isna = is.na(vals)
		anyNA = any(isna)

		na.show = update_na.show(label.show, legend$na.show, anyNA)

		if (is.null(sortRev)) {
			ids = NULL
		} else if (is.na(sortRev)) {
			ids = rep(1L, length(x1))
		} else if (sortRev) {
			if (vnum) {
				ids = rank(-vals)
			} else {
				ids = (as.integer(n2) + 1L) - ids
			}
		} else if (vnum) {
			ids = rank(vals)
		}

		if (anyNA) {
			vals[isna] = value.na
			if (!is.null(sortRev)) ids[isna] = 0L
		}

		labels_exp = !ticks.specified && trans == "log" && trargs$base != exp(1)


		if (ticks.specified) {
			b_t = ticks_t
			b = do.call(tr$rev, c(list(x = b_t), trargs))
		} else {
			if (labels_exp) {
				b = do.call(tr$rev, c(list(x = pretty(limits_t, n = n)), trargs))
			} else {
				b  = prettyTicks(do.call(tr$rev, c(list(x = seq(limits_t[1], limits_t[2], length.out = n)), trargs)))
			}

			if (!(aes %in% c("col", "fill"))) b = b[b!=0]
			b_t = do.call(tr$fun, c(list(x = b), trargs))
		}
		if (length(b_t) == 2) {
			sel = TRUE
			labels_select = TRUE
		} else {
			# #1039
			# to include min/max if they are closer to the next 'not included' tick than the included tick
			# e.g. tm_shape(World) + tm_polygons("gender", fill.scale = tm_scale_continuous())
			# lowest value is 0.009 whereas ticks are 0.2, 0.4, 0.6, 0.8.
			stp1 = (b_t[2] - b_t[1]) / 2
			stp2 = (b_t[length(b_t)] - b_t[length(b_t) - 1L]) / 2
			limits_t_tmp = c(limits_t[1] - stp1, limits_t[2] + stp2)
			sel = (b_t>=limits_t_tmp[1] & b_t<=limits_t_tmp[2])

			labels_select = ((b_t>=limits_t[1] & b_t<=limits_t[2])[sel])
		}
		b_t = b_t[sel]
		b = b[sel]



		nbrks_cont <- length(b_t)

		if (vnum) {
			vvalues = transform_values(b_t, limits_t, values.range, values$power, values.scale * o$scale, include.neutral = FALSE)
		} else {
			id = as.integer(cut(b_t, breaks=breaks, include.lowest = TRUE))

			if (sum(!is.na(id)) < 2) {
				if (is.null(scale$ticks)) {
					cli::cli_abort("{.field [tm_scale_continuous]} too few tick labels specified. Either increase {.arg n} or specify tick values via {.arg ticks}")
				} else {
					cli::cli_abort("{.field [tm_scale_continuous]} too few tick labels specified via {.arg ticks}")
				}
			}

			# impute missing head and tail (#1039)
			missing_head = is.na(id[1])
			if (missing_head) id[1] = 2 * id[2] - id[3]

			nid = length(id)

			missing_tail = is.na(id[nid])
			if (missing_tail) id[nid] = 2 * id[nid-1] - id[nid-2]

			id_step = id[-1] - head(id, -1)
			id_step = c(id_step[1], id_step, id_step[length(id_step)])
			id_lst = mapply(function(i, s1, s2){
				#res = round(seq(i-floor(id_step/2), i+ceiling(id_step/2), length.out=11))[1:10]
				res1 = round(seq(i-floor(s1/2), i, length.out=(nvv/2)+1L))
				res2 = round(seq(i, i+ceiling(s2/2), length.out=(nvv/2)+1L))[2:(nvv/2)]
				res = c(res1, res2)
				res[res<1 | res>o$continuous.nclasses] = NA
				res
			}, id, head(id_step, -1), id_step[-1], SIMPLIFY = FALSE)
			vvalues = lapply(id_lst, function(i) {
				if (legend$reverse) rev(vv[i]) else vv[i]
			})
		}






		if (legend$reverse) vvalues = rev(vvalues)

		if (na.show) vvalues = c(vvalues, value.na)

		# temporarily stack gradient values
		if (!vnum) vvalues = cont_collapse(vvalues)

		# create legend values
		values = b

		# create legend labels for continuous cases
		if (is.null(labels)) {
			if (labels_exp) {
				labels = paste(trargs$base, b_t, sep = "^")
			} else {
				labels = do.call("fancy_breaks", c(list(vec=b, as.count = FALSE, intervals=FALSE, interval.closure="left"), label.format))
			}
		} else {
			labels = rep(labels, length.out=nbrks_cont)
			attr(labels, "align") <- label.format$text.align
		}

		if (legend$reverse) {
			labels.align = attr(labels, "align")
			labels = rev(labels)
			labels_select = rev(labels_select)
			attr(labels, "align") = labels.align
		}
		if (na.show) {
			labels.align = attr(labels, "align")
			labels = c(labels, label.na)
			labels_select = c(labels_select, TRUE)
			attr(labels, "align") = labels.align
		}

		layer_args = c(layer_args, list(ticks = scale$ticks, limits = scale$limits))

		legend = within(legend, {
			nitems = length(labels)
			labels = labels
			labels_select = labels_select
			dvalues = values
			vvalues = vvalues
			vneutral = value.neutral
			na.show = get("na.show", envir = parent.env(environment()))
			scale = "continuous"
			tr = tr
			limits = limits
			layer_args = layer_args
		})


		# NOTE: tr and limits are included in the output to facilitate the transformation of the leaflet continuous legend ticks (https://github.com/rstudio/leaflet/issues/665)
		#vvalues_mids = sapply(cont_split(vvalues), "[", nvv/2)
		#vvalues_mids[vvalues_mids == "NA"] = NA


		chartFun = paste0("tmapChart", toTitleCase(chart$summary))

		chart = do.call(chartFun, list(chart,
									   bin_colors = NULL,
									   breaks_def = chart$breaks,
									   na.show = na.show,
									   x1 = x1))



		# chart = within(chart, {
		# 	labels = labels
		# 	vvalues = sapply(cont_split(vvalues), "[", 5)
		# 	breaks_def = legend$dvalues
		# 	na.show = get("na.show", envir = parent.env(environment()))
		# 	x1 = x1
		# })



		if (submit_legend) {
			if (bypass_ord) {
				format_aes_results(vals, legend = legend, chart = chart)
			} else {
				format_aes_results(vals, ids, legend, chart = chart)
			}
		} else {
			list(vals = vals, ids = ids, legend = legend, chart = chart, bypass_ord = bypass_ord)
		}

	})
}

#' @export
tmapLeafletCompPrepare.tm_title = function(comp, o) {
	comp$show = TRUE
	comp
}


#' @export
tmapLeafletCompHeight.tm_title = function(comp, o) {
	comp
}

#' @export
tmapLeafletCompWidth.tm_title = function(comp, o) {
	comp
}


#' @export
tmapLeafletCompPlot.tm_title = function(comp, lf, o) {
	pos = (comp$position |>
		leaflet_pos(collapse = "-") |>
		strsplit(split = "-", fixed = TRUE))[[1]]
	if (inherits(lf, "shiny.tag.list")) {
		ncld <- length(lf[[1]])
		lf[[1]] <- mapply(function(l, i) {
			title <- l$children[[1]]$title %||% ""
			if (title!="") {
				l$children[[1]] <- l$children[[1]] %>% htmlwidgets::onRender(paste("
					function(el, x) {
						var tldiv = el.getElementsByClassName(\"leaflet-", pos[1], " leaflet-", pos[2],"\")[0];
						var titlediv = document.createElement('div');
						titlediv.className = \"info legend leaflet-control\";
						titlediv.innerHTML = \"<b>", title, "</b>\";
						tldiv.insertBefore(titlediv, tldiv.childNodes[0]);
					}", sep="")
				)
			}
			l
		}, lf[[1]], 0:(ncld-1), SIMPLIFY = FALSE)
	} else {
		title <- comp$text %||% ""
		if (title!="") {
			lf <- lf %>% htmlwidgets::onRender(paste("
						function(el, x) {
						    var tldiv = el.getElementsByClassName(\"leaflet-", pos[1], " leaflet-", pos[2],"\")[0];
							var titlediv = document.createElement('div');
							titlediv.className = \"info legend leaflet-control\";
							titlediv.innerHTML = \"<b>", title, "</b>\";
							tldiv.insertBefore(titlediv, tldiv.childNodes[0]);
						}", sep="")
			)
		}
	}
	lf

}



#' @export
tmapLeafletCompPrepare.tm_compass = function(comp, o) {
	comp$show = TRUE
	comp
}


#' @export
tmapLeafletCompHeight.tm_compass = function(comp, o) {
	comp
}

#' @export
tmapLeafletCompWidth.tm_compass = function(comp, o) {
	comp
}


#' @export
tmapLeafletCompPlot.tm_compass = function(comp, lf, o) {
	lf %>% leafem::addLogo(system.file(paste0("img/compass_", comp$type, ".png"), package = "tmap"), position = leaflet_pos(comp$position), width = 120, height = 120)
}



#' @export
tmapLeafletCompPrepare.tm_scalebar = function(comp, o) {
	comp$show = TRUE
	comp
}


#' @export
tmapLeafletCompHeight.tm_scalebar = function(comp, o) {
	comp
}

#' @export
tmapLeafletCompWidth.tm_scalebar = function(comp, o) {
	comp
}


#' @export
tmapLeafletCompPlot.tm_scalebar = function(comp, lf, o) {
	lf %>% leaflet::addScaleBar(position = leaflet_pos(comp$position))
}




#' @export
tmapLeafletCompPrepare.tm_credits = function(comp, o) {
	comp$show = TRUE
	comp
}


#' @export
tmapLeafletCompHeight.tm_credits = function(comp, o) {
	comp
}

#' @export
tmapLeafletCompWidth.tm_credits = function(comp, o) {
	comp
}


#' @export
tmapLeafletCompPlot.tm_credits = function(comp, lf, o) {
	#lf %>% leafem::addLogo(system.file(paste0("img/credits_", comp$type, ".png"), package = "tmsap"), src = "local", position = paste(unlist(comp$position[c("pos.v", "pos.h")]), collapse = ""), width = 120, height = 120)
	#message("tm_credits not implemented yet for view mode")
	lf %>% leaflet::addTiles(urlTemplate = "", attribution = comp$text)
}


#' @export
tmapLeafletCompPrepare.tm_mouse_coordinates = function(comp, o) {
	comp$show = TRUE
	comp
}


#' @export
tmapLeafletCompHeight.tm_mouse_coordinates = function(comp, o) {
	comp
}

#' @export
tmapLeafletCompWidth.tm_mouse_coordinates = function(comp, o) {
	comp
}


#' @export
tmapLeafletCompPlot.tm_mouse_coordinates = function(comp, lf, o) {
	lf %>% leafem::addMouseCoordinates()
}



#' @export
tmapLeafletCompPrepare.tm_minimap = function(comp, o) {
	comp$show = TRUE
	comp$height = 20 * comp$height
	comp$width = 20 * comp$width

	extra = comp[setdiff(intersect(names(comp), names(formals(leaflet::addMiniMap))), c("position", "map"))]

	comp$specified_tiles = !is.na(comp$server)

	comp$args = c(list(toggleDisplay = comp$toggle), extra)
	comp
}


#' @export
tmapLeafletCompHeight.tm_minimap = function(comp, o) {


	comp
}

#' @export
tmapLeafletCompWidth.tm_minimap = function(comp, o) {
	comp
}


#' @export
tmapLeafletCompPlot.tm_minimap = function(comp, lf, o) {
	comp$args$tiles = if (comp$specified_tiles) {
		comp$server
	} else if (length(.TMAP_LEAFLET$tiles)) {
		.TMAP_LEAFLET$tiles[[1]][[1]]$server[1]
	} else {
		o$basemap.server[1]
	}
	lf2 = do.call(addMiniMap, c(list(map = lf, position = leaflet_pos(comp$position)), comp$args))
	if (!comp$specified_tiles && (length(comp$args$tiles) > 0)) {
		lf2 <- lf2 %>%
			htmlwidgets::onRender("
			    function(el, x) {
			      var myMap = this;
			      myMap.on('baselayerchange',
			        function (e) {
			          myMap.minimap.changeLayer(L.tileLayer.provider(e.name));
			        })
			    }")
	}
	lf2
}

#' @export
tmapLeafletCompPrepare.tm_logo = function(comp, o) {
	height = 20 * comp$height
	comp$logo = lapply(comp$file, function(lf){
		x = tmap_icons(lf)
		scale = height / x$iconHeight
		x$iconWidth = x$iconWidth * scale
		x$iconHeight = x$iconHeight * scale
		x$iconAnchorX = x$iconAnchorX * scale
		x$iconAnchorY = x$iconAnchorY * scale
		x
	})

	comp$asp = vapply(comp$logo, function(lg) {
		lg$iconWidth / lg$iconHeight
	}, FUN.VALUE = numeric(1))
	comp$show = TRUE
	comp
}


#' @export
tmapLeafletCompHeight.tm_logo = function(comp, o) {
	comp
}

#' @export
tmapLeafletCompWidth.tm_logo = function(comp, o) {
	comp
}

#' @export
tmapLeafletCompPlot.tm_logo = function(comp, lf, o) {
	ws = c(rev(cumsum(vapply(comp$logo, function(l)l$iconWidth,FUN.VALUE = numeric(1)) + 10))[-1], 0)

	for (i in 1:length(comp$logo)) {
		lf = lf %>% leafem::addLogo(comp$logo[[i]]$iconUrl, position = leaflet_pos(comp$position), width = comp$logo[[i]]$iconWidth, height = comp$logo[[i]]$iconHeight, offset.x = 10 + ws[i], offset.y = 20)
	}
	lf
}



#' @export
tmapLeafletCompPrepare.tm_inset_image = function(comp, o) {
	comp$file = comp$x
	comp = tmapLeafletCompPrepare.tm_logo(comp, o)
	class(comp) = c("tm_logo", class(comp))
	comp
}

#' @export
tmapLeafletCompPrepare.tm_inset_grob = function(comp, o) {
	height = 20 * comp$height
	width = 20 * comp$width

	grob.dim = c(width=width, height=height, render.width=width, render.height=height)

	comp$logo = list(grob2icon(comp$x, grob.dim = grob.dim, just = c(0.5, 0.5)))
	comp$logo = lapply(comp$logo, function(x){
		scale = height / x$iconHeight
		x$iconWidth = unname(x$iconWidth * scale)
		x$iconHeight = unname(x$iconHeight * scale)
		x$iconAnchorX = unname(x$iconAnchorX * scale)
		x$iconAnchorY = unname(x$iconAnchorY * scale)
		x
	})

	comp$asp = vapply(comp$logo, function(lg) {
		lg$iconWidth / lg$iconHeight
	}, FUN.VALUE = numeric(1))
	comp$show = TRUE
	class(comp) = c("tm_logo", class(comp))
	comp

}

#' @export
tmapLeafletCompPrepare.tm_inset_gg = function(comp, o) {
	rlang::check_installed("ggplot2", reason = "for plotting ggplot2 charts")
	comp$x = ggplot2::ggplotGrob(comp$x)
	tmapLeafletCompPrepare.tm_inset_grob(comp, o)
}

prepreprocess_meta = function(o, vp) {

	within(o, {
		vp = vp
		if (is.null(vp)) {
			devsize = graphics::par("fin") #dev.size() MAY NOT BE EQUAL IN RSTUDIO: https://github.com/rstudio/rstudio/issues/10723
		} else {
			if (is.character(vp)) seekViewport(vp) else pushViewport(vp)
			devsize = c(grid::convertWidth(grid::unit(1, "npc"), unitTo = "inch", valueOnly = TRUE),
						grid::convertHeight(grid::unit(1, "npc"), unitTo = "inch", valueOnly = TRUE))
		}

		# dasp device aspect ratio
		dasp = devsize[1] / devsize[2]

		# needed for spnc viewport (to retain aspect ratio)
		if (dasp > 1) {
			cw = dasp
			ch = 1
		} else {
			ch = 1/dasp
			cw = 1
		}

		lin = graphics::par("cin")[2]# * scale

		lineH = lin / devsize[2] * scale
		lineW = lin / devsize[1] * scale

		# not needed?
		#nlinesH = 1/lineH
		#nlinesW = 1/lineW

	})
}
preprocess_meta = function(o, cdt) {
	npp = NULL
	within(o, {
		nby = fn #get_nby(fl)
		isdef = !sapply(fl, is.null)

		if (is.na(panel.type)) panel.type = if (identical(panel.show, FALSE)) {
			"none"
		} else if (identical(panel.show, TRUE)) {
			# force panel labels
			if (type %in% c("wrap", "stack") || (npp == 1)) {
				"wrap"
			} else {
				"xtab"
			}
		} else if ((npp == 1) && is.na(panel.labels[[1]])) {
			if (n > 1 || any(isdef)) {
				"wrap"
			} else {
				"none"
			}
		} else if (!(type %in% c("wrap", "stack")) && !isdef[1] && !isdef[2]) {
			"none"
		} else if ((type %in% c("wrap", "stack")) || (npp == 1)) {
			"wrap"
		} else {
			"xtab"
		}

		inner.margins = get_option_class(inner.margins, class = main_class)

		# legend.present.auto:
		#   find out whether there are legends for all facets, per row, per col
		#   use them to automatically determine meta.margins (in preprocess_meta)
		# # legend.present.fix
		#	find legend boxes that are assigned to outer margins

		if (nrow(cdt) == 0) {
			legend.present.auto = c(all = FALSE, per_row = FALSE, per_col = FALSE, per_facet = FALSE)
			legend.present.fix = rep(FALSE, 4)
		} else {
			if (type %in% c("wrap", "stack")) {
				#o$legend.present.auto = c(all = any(is.na(cdt$by1__) & cdt$class == "autoout"), per_row = any(!is.na(cdt$by1__) & cdt$class == "autoout"), per_col = FALSE)
				legend.present.auto = c(all = any(cdt$class == "autoout" & is.na(cdt$by1__)),
										per_row = FALSE, per_col = FALSE,
										per_facet = any(cdt$class == "autoout" & !is.na(cdt$by1__)))
		} else {
				legend.present.auto = c(all = any(is.na(cdt$by1__) & is.na(cdt$by2__) & cdt$class == "autoout"),
										per_row = any(!is.na(cdt$by1__) & is.na(cdt$by2__) & cdt$class == "autoout"),
										per_col = any(is.na(cdt$by1__) & !is.na(cdt$by2__) & cdt$class == "autoout"),
										per_facet = any(!is.na(cdt$by1__) & !is.na(cdt$by2__) & cdt$class == "autoout"))
			}
			legend.present.fix = c(any(cdt$class == "out" & cdt$cell.v == "bottom"),
								   any(cdt$class == "out" & cdt$cell.h == "left"),
								   any(cdt$class == "out" & cdt$cell.v == "top"),
								   any(cdt$class == "out" & cdt$cell.h == "right"))
		}


		# in case there are per-facet legends but no no marginal legends, and nrows or ncols equals 1, place them outside (to do this, set them to all-facet here, change legend.position.all below accordingly, and finally determine legend position in step4_plot)
		legend_move_all = FALSE

		# legend_move_all is needed to move the 'global' legend not in a stacked-per-facets cell, but in the other orientation (#1071)

		if (legend.present.auto[4] && (!any(legend.present.auto[2:3]))) {

			if (type == "stack") {
				if (all(legend.present.auto[c(1, 4)])) legend_move_all = TRUE
				legend.present.auto[1] = TRUE
				legend.present.auto[4] = FALSE
				set_to_stack_message = FALSE
			} else {
				set_to_stack_message = TRUE
			}
		} else {
			set_to_stack_message = FALSE
		}

	})
}

process_meta = function(o, d, cdt, aux) {

	if (o$legend.only) {
		return(within(o, {
			meta.buffers = c(0, 0, 0, 0)
			meta.margins = c(0, 0, 0, 0)
			xylab.margins = rep(0, 4)
			panel.xtab.size = c(0, 0, 0, 0)
			panel.xtab.margin = rep(0, 4)
			panel.wrap.margin = rep(0, 4)
			grid.buffers = c(0, 0, 0, 0)
			grid.margins = c(0, 0, 0, 0)
			panel.wrap.size = c(0, 0, 0, 0)
			fixedMargins  =  outer.margins + meta.buffers * 2 + meta.margins
			nrows = 1L
			ncols = 1L

			between_marginH = between_margin * lineH
			between_marginW = between_margin * lineW


			#overall scale down factor for facets
			width_forn = max(1e-9, ((1 - sum(fixedMargins[c(2, 4)])) - (ncols * sum(panel.wrap.size[c(2,4)])) - (ncols - 1) * between_marginW) / ncols)
			width_for1 = max(1e-9, ((1 - sum(fixedMargins[c(2, 4)])) - (sum(panel.wrap.size[c(2,4)]))))

			height_forn = max(1e-9, ((1 - sum(fixedMargins[c(1, 3)])) - (nrows * sum(panel.wrap.size[c(1,3)])) - (nrows - 1) * between_marginH) / nrows)
			height_for1 = max(1e-9, ((1 - sum(fixedMargins[c(1, 3)])) - (sum(panel.wrap.size[c(1,3)]))))

			scale_down = (1 / sqrt((width_for1 * height_for1) / (width_forn * height_forn))) ^ (1 / scale.factor)

		}))
	}


	gs = tmap_graphics_name()

	# add tm_grid values to o
	gid = which(vapply(aux, FUN = inherits, "tm_grid", FUN.VALUE = logical(1)))[1]
	if (!is.na(gid)) {
		a = aux[[gid]]$args
		a$zindex = NULL
		a$group = NULL
		o[paste0("grid.", names(a))] = a
	}

	# add credits into to o (for view mode in order to reset default attribution text)
	cid = which(vapply(cdt$comp, FUN = inherits, "tm_credits", FUN.VALUE = logical(1)))[1]
	o$credits.defined = (!is.na(cid))

	bbx = d$bbox[[1]]
	within(o, {
		# sasp shape aspect ratio (NA if free coordinates)
		diff_asp = any(d$asp != d$asp[1])
		sasp = ifelse(diff_asp, NA, d$asp[1])

		# preferred aspect ratio (just for this function): if asp is defined (not 0 or NA), use that, otherwise use sasp (shape asp) if available (if not; 1)
		pasp = if (is.na(sasp)) {
			if (!is.na(asp) && asp > 0) {
				asp
			} else {
				1
			}
		} else {
			if (!is.na(asp) && asp > 0) {
				asp
			} else {
				sasp
			}
		}


		if (gs == "Grid") {

			bufferH = lineH / 2
			bufferW = lineW / 2

			# calculate space for margins, panels, etc

			meta.automatic = is.na(meta.margins[1])

			#one.row = (!is.na(o$nrows) && o$nrows == 1)
			#one.col = (!is.na(o$ncols) && o$ncols == 1)

			if (meta.automatic) meta.margins = c(0, 0, 0, 0) else meta.margins = rep(meta.margins, length.out = 4)

			meta.buffers = sign(meta.margins) * c(bufferH, bufferW, bufferH, bufferW) # outside and inside

			panel.xtab.size = if (panel.type == "xtab") {
				c(ifelse("bottom" %in% panel.xtab.pos, panel.label.height * lineH, 0),
				  ifelse("left" %in% panel.xtab.pos, panel.label.height * lineW, 0),
				  ifelse("top" %in% panel.xtab.pos, panel.label.height * lineH, 0),
				  ifelse("right" %in% panel.xtab.pos, panel.label.height * lineW, 0))
			} else c(0, 0, 0, 0)

			panel.margin = get_option_class(panel.margin, panel.type, spatial_class = FALSE)

			panel.xtab.margin = if (panel.type == "xtab") {
				c(ifelse("bottom" %in% panel.xtab.pos, panel.margin * lineH, 0),
				  ifelse("left" %in% panel.xtab.pos, panel.margin * lineW, 0),
				  ifelse("top" %in% panel.xtab.pos, panel.margin * lineH, 0),
				  ifelse("right" %in% panel.xtab.pos, panel.margin * lineW, 0))
			} else c(0, 0, 0, 0)

			panel.wrap.margin =	if (panel.type == "wrap") {
				c(ifelse(panel.wrap.pos == "bottom", panel.margin * lineH, 0),
				  ifelse(panel.wrap.pos == "left", panel.margin * lineW, 0),
				  ifelse(panel.wrap.pos == "top", panel.margin * lineH, 0),
				  ifelse(panel.wrap.pos == "right", panel.margin * lineW, 0))
			} else c(0, 0, 0, 0)


			panel.wrap.size = if (panel.type == "wrap") {
				c(ifelse(panel.wrap.pos == "bottom", panel.label.height * lineH, 0),
				  ifelse(panel.wrap.pos == "left", panel.label.height * lineW, 0),
				  ifelse(panel.wrap.pos == "top", panel.label.height * lineH, 0),
				  ifelse(panel.wrap.pos == "right", panel.label.height * lineW, 0))
			} else c(0, 0, 0, 0)

			xylab.margins = rep(0, 4)
			if (xlab.show) {
				#if (is.na(xlab.fontface)) xlab.fontface = text.fontface
				xylab.margins[ifelse(xlab.side == "bottom", 1, 3)] = if (xlab.rotation %in% c(0, 180)) {
					(number_text_lines(xlab.text) + xlab.space)  * lineH
				} else {
					(text_width_inch(xlab.text, space = FALSE) / lin + xlab.space) * lineH
				}
			}
			if (ylab.show) {
				#if (is.na(ylab.fontface)) ylab.fontface = text.fontface
				xylab.margins[ifelse(ylab.side == "left", 2, 4)] = if (ylab.rotation %in% c(90, 270)) {
					(number_text_lines(ylab.text) + ylab.space)  * lineW * ylab.size
				} else {
					(text_width_inch(ylab.text, space = FALSE) / lin + ylab.space) * lineW * ylab.size
				}
			}


			grid.buffers = if (grid.show) {
				as.integer(c("bottom", "left", "top", "right") %in% grid.labels.pos) * c(bufferH, bufferW, bufferH, bufferW)
			} else {
				rep(0, 4)
			}



			grid.labels.show = rep(grid.labels.show, length.out = 2) # also happens in tmapGridGridPrep

			if (grid.show && any(grid.labels.show) && !grid.labels.inside_frame) {
				proj = sf::st_crs(bbx)
				if (!is.na(o$grid.crs)) {
					bbx_orig = bbx
					bbx = suppressWarnings(bb(bbx, current.projection = proj, projection = o$grid.crs))
				}

				# convert*() below forces grid to open a page to resolve units -> blank
				# frame on the user's device / extra page in tmap_save. Measure on a
				# null device instead.
				cur_dev = grDevices::dev.cur()
				grDevices::pdf(NULL)
				on.exit({
					grDevices::dev.off()
					if (cur_dev > 1L) grDevices::dev.set(cur_dev)
				}, add = TRUE)

				lineHin = convertHeight(unit(grid.labels.size, "lines"), "inch", valueOnly = TRUE)

				if (grid.labels.show[1]) {
					gridx = pretty30(bbx[c(1,3)], n = 5, longlat = !is.na(o$grid.crs) && sf::st_is_longlat(proj))
					xbbstringWin = max(convertWidth(stringWidth(do.call("fancy_breaks", c(list(vec=gridx, intervals = FALSE), grid.labels.format))), "inch", valueOnly = TRUE)) * grid.labels.size
					xgridHin = ifelse(!is.na(grid.labels.space.x), grid.labels.space.x * lineHin, ifelse(grid.labels.rot[1] %in% c(0, 180), 1.375 * lineHin, xbbstringWin + lineHin * .75) + grid.labels.margin.x * lineHin)
				} else {
					xgridHin = 0
				}

				if (grid.labels.show[2]) {
					gridy = pretty30(bbx[c(2,4)], n = 5, longlat = !is.na(o$grid.crs) && sf::st_is_longlat(proj))
					ybbstringWin = max(convertWidth(stringWidth(do.call("fancy_breaks", c(list(vec=gridy, intervals=FALSE), grid.labels.format))), "inch", valueOnly = TRUE))
					ybbstringWin = ybbstringWin * grid.labels.size
					ygridWin = ifelse(!is.na(grid.labels.space.y), grid.labels.space.y * lineHin, ifelse(grid.labels.rot[2] %in% c(0, 180), ybbstringWin + lineHin * .75, 1.375 * lineHin) + grid.labels.margin.y * lineHin)
				} else {
					ygridWin = 0
				}

				grDevices::dev.off()
				if (cur_dev > 1L) grDevices::dev.set(cur_dev)
				on.exit()  # drop the guard now that we've closed cleanly

				marks_new = c(xgridHin, ygridWin, xgridHin, ygridWin) / lin
				grid.margins = as.integer(c("bottom", "left", "top", "right") %in% grid.labels.pos) * marks_new * c(lineH, lineW, lineH, lineW)
			} else {
				grid.margins = rep(0, 4)
			}

			between_marginH = between_margin * lineH
			between_marginW = between_margin * lineW

			fixedMargins = outer.margins + meta.buffers * 2 + meta.margins + xylab.margins + panel.xtab.size + grid.buffers + grid.margins
		} else {
			grid.buffers = rep(0, 4)
			grid.labels.show = c(FALSE, FALSE)
			grid.margins = rep(0, 4)
			fixedMargins = rep(0, 4)
			panel.wrap.size = rep(0, 4)
			between_marginH = 0
			between_marginW = 0
		}


		masp = ((1 - sum(fixedMargins[c(2, 4)])) / (1 - sum(fixedMargins[c(1, 3)]))) * dasp

		# Aspect ratios:
		# sasp: shape
		# asp: user specified
		# pasp: prefered (asp or if not specified, sasp)
		# masp: multiples (facets) area

		# determine where to place automatic legends (i.e. legends with local legend.position = NA and with legend.position = tm_pos_auto_out() enabled)
		# this is also neede to find out which margins are taken from meta.auto_margins

		legend.position.sides = legend.position
		legend.position.all = legend.position

	#	legsG = cdt[, leg]

		# determine orientation of stacked maps
		# it also implies where legends will be drawn: horizontal orientation=legends bottom or top, vertical orientation=legends left or right
		# !!! this also applies for single maps
		mx_width = (1 - sum(fixedMargins[c(1, 3)])) * devsize[1]
		mx_height = (1 - sum(fixedMargins[c(2, 4)])) * devsize[2]

		# cdt[, scale := pmax(legW/mx_width, legH/mx_height, 1)]
		# cdt[, ":="(legW_sc = legW/scale, legH_sc = legH/scale)]


		if (gs == "Grid") {
			if (type =="stack") {
				if (is.na(orientation)) {
					if (nrow(cdt)) {
						legs_auto = cdt[class=="autoout"]
					} else {
						legs_auto = cdt
					}


					if (nrow(legs_auto) && npp == 1) {


						legWmax = min(max(legs_auto$legW) / devsize[1], max(meta.auto_margins[c(2,4)]))
						legHmax = min(max(legs_auto$legH) / devsize[2], max(meta.auto_margins[c(1,3)]))


						av_width = mx_width - legWmax * devsize[1]
						av_height = mx_height - legHmax * devsize[2]

						shp_height_hor = if ((av_width / mx_height) < pasp) av_width / pasp else mx_height
						shp_height_ver = if ((mx_width / av_height) < pasp) mx_width / pasp else av_height

						orientation = if (shp_height_hor >= shp_height_ver) "vertical" else "horizontal"
					} else {
						orientation = if ((!is.na(nrows) && nrows == 1) || (!is.na(ncols) && ncols == npp)) {
							"horizontal"
						} else if ((!is.na(nrows) && nrows == npp) || (!is.na(ncols) && ncols == 1)) {
							"vertical"
						} else if ((npp == 1 && (pasp > masp)) || (npp > 1 && (pasp < masp))) "horizontal" else "vertical"
					}
				}
			}
		} else {
			if (is.na(nrows) && is.na(ncols)) {
				orientation = if ((npp == 1 && (pasp > masp)) || (npp > 1 && (pasp < masp))) "horizontal" else "vertical"
			} else {
				if (is.na(nrows)) nrows = floor(npp / ncols)
				if (is.na(ncols)) ncols = floor(npp / nrows)
				orientation = if (ncols >= nrows) "horizontal" else "vertical"
			}
		}

		if (gs == "Grid") {
			## find position for all-facet legend
			if (legend.present.auto[1]) {
				if (!legend.present.auto[2] & !legend.present.auto[3]) {
					# only 'all facets' outside legends (either bottom or right)
					# was: n > 1 && masp > pasp
					legend.position.all = local({
						moving = function(x) if (legend_move_all) !x else x
						if (moving((type != "stack" && npp == 1 && pasp > masp) || (type != "stack" && npp > 1 && masp < 1) || (type == "stack" && orientation == "horizontal"))) {
							list(cell.h = "center", cell.v = legend.position$cell.v)
						} else {
							list(cell.h = legend.position$cell.h, cell.v = "center")
						}
					})

				} else if (legend.present.auto[2] & !legend.present.auto[3]) {
					# central goes center bottom
					legend.position.all = list(cell.h = "center", cell.v = legend.position$cell.v)
				} else if (!legend.present.auto[2] & legend.present.auto[3]) {
					# central goes right center
					legend.position.all = list(cell.h = legend.position$cell.h, cell.v = "center")
				}
			}

			if (legend_move_all) {
				# 'all' legend is set via legend.position.all
				# the per stack-per-facet legend is 'the other' position
				# so both (left or right) and (top or bottom) should be enabled
				margins.used.all = c(legend.position$cell.v == "bottom",
									 legend.position$cell.h == "left",
									 legend.position$cell.v == "top",
									 legend.position$cell.h == "right") * legend.present.auto[1]

			} else {
				margins.used.all = c(legend.position.all$cell.v == "bottom",
									 legend.position.all$cell.h == "left",
									 legend.position.all$cell.v == "top",
									 legend.position.all$cell.h == "right") * legend.present.auto[1]
			}

			margins.used.sides = c(bottom = legend.position.sides$cell.v == "bottom",
								   left = legend.position.sides$cell.h == "left",
								   top = legend.position.sides$cell.v == "top",
								   right = legend.position.sides$cell.h == "right") * legend.present.auto[c(3,2,3,2)]


			margins.used =  margins.used.all | margins.used.sides | legend.present.fix

			# tm_shape(World) + tm_polygons(fill = "HPI", lwd = "life_exp")

			if (nrow(cdt)) {
				cdt2 = data.table::copy(cdt[cdt$class %in% c("autoout", "out"),])

				# CODE COPIED FROM STEP4_plot L157
				# TO DO: fix this
				if (o$type != "grid" && o$npp > 1) {
					#if (o$nrows == 1 && o$ncols == 1)
					if (identical(orientation, "horizontal")) {
						# -use by2 and not by1 when they form a row
						cdt2[, by2__ := by1__]
						cdt2[, by1__ := NA]
					}
				}



				stacks = o$component.stack

				cdt2[is.na(by1__) & is.na(by2__) & class == "autoout", ':='(cell.h = legend.position.all$cell.h, cell.v = legend.position.all$cell.v)]

				cdt2[!is.na(by1__) & is.na(by2__) & class == "autoout", ':='(cell.h = legend.position.sides$cell.h, cell.v = "by")]
				cdt2[is.na(by1__) & !is.na(by2__) & class == "autoout", ':='(cell.h = "by", cell.v = legend.position.sides$cell.v)]

				# in case tm_pos is called to update the 'non-by' cell
				cdt2$calledCV = vapply(cdt2$comp, function(comp) "cell.v" %in% comp$position$called, FUN.VALUE = logical(1))
				cdt2$calledCH = vapply(cdt2$comp, function(comp) "cell.h" %in% comp$position$called, FUN.VALUE = logical(1))
				cdt2[!is.na(by1__) & is.na(by2__) & !calledCV, ':='(cell.v = "by")]
				cdt2[is.na(by1__) & !is.na(by2__) & !calledCH, ':='(cell.h = "by")]

				cdt2[!is.na(by1__) & is.na(by2__) & !calledCV, ':='(stack = ifelse(stack_auto, stacks["per_row"], stack))]
				cdt2[is.na(by1__) & !is.na(by2__) & !calledCH, ':='(stack = ifelse(stack_auto, stacks["per_col"], stack))]

				cdt2$calledCV = NULL
				cdt2$calledCH = NULL


				cdt2[is.na(by1__) & is.na(by2__) & class == "autoout", ':='(stack = ifelse(stack_auto, ifelse(cell.h == "center", stacks["all_col"], ifelse(cell.v == "center", stacks["all_row"], stacks["all"])), stack))]
				cdt2[!is.na(by1__) & is.na(by2__) & class == "autoout", ':='(stack = ifelse(stack_auto, stacks["per_row"], stack))]
				cdt2[is.na(by1__) & !is.na(by2__) & class == "autoout", ':='(stack = ifelse(stack_auto, stacks["per_col"], stack))]


				cdt2[class == "autoout", class := "out"]

				if (nrow(cdt2) == 0) {
					meta.auto_margins = c(0, 0, 0, 0)
				} else {
					if (type %in% c("stack", "anistack")) {
						# workaround: stacking mode is determined later (step4 L156), because it requires ncols and nrows
						# for stack, this is already known, so therefore we can better estimate the meta width and height

						cdt2[is.na(by1__), by1__:=1]

						meta.auto_margins = pmin(meta.auto_margins, do.call(pmax, lapply(unique(cdt2$by1__), function(b1) {
							cdt2b = cdt2[by1__==b1, ]

							cdt2b[stack_auto == TRUE, stack:= ifelse(npp==1, ifelse(cell.h %in% c("left", "right"), o$legend.stack["all_row"], o$legend.stack["all_col"]), ifelse(orientation == "vertical", o$legend.stack["per_row"], o$legend.stack["per_col"]))]

							c(max(max(c(0,cdt2b[cell.v == "bottom" & stack == "vertical", .(V=sum(legH)),by = c("cell.h", "by3__")]$V)),
								  max(c(0,cdt2b[cell.v == "bottom" & stack == "horizontal", .(V=max(c(0, legH))),by = c("cell.h", "by3__")]$V))) / o$devsize[2],
							  max(max(c(0,cdt2b[cell.h == "left" & stack == "horizontal", .(V=sum(legW)),by = c("cell.v", "by3__")]$V)),
							  	max(c(0,cdt2b[cell.h == "left" & stack == "vertical", .(V=max(c(0, legW))),by = c("cell.v", "by3__")]$V))) / o$devsize[1],
							  max(max(c(0,cdt2b[cell.v == "top" & stack == "vertical", .(V=sum(legH)),by = c("cell.h", "by3__")]$V)),
							  	max(c(0,cdt2b[cell.v == "top" & stack == "horizontal", .(V=max(c(0, legH))),by = c("cell.h", "by3__")]$V))) / o$devsize[2],
							  max(max(c(0,cdt2b[cell.h == "right" & stack == "horizontal", .(V=sum(legW)),by = c("cell.v", "by3__")]$V)),
							  	max(c(0,cdt2b[cell.h == "right" & stack == "vertical", .(V=max(c(0, legW))),by = c("cell.v", "by3__")]$V))) / o$devsize[1])
						})))
					} else {
						meta.auto_margins = pmin(meta.auto_margins,
												 c(max(cdt$legH[cdt$cell.v == "bottom" & cdt$class %in% c("autoout", "out")], 0) / o$devsize[2],
												   max(cdt$legW[cdt$cell.h == "left" & cdt$class %in% c("autoout", "out")], 0) / o$devsize[1],
												   max(cdt$legH[cdt$cell.v == "top" & cdt$class %in% c("autoout", "out")], 0) / o$devsize[2],
												   max(cdt$legW[cdt$cell.h == "right" & cdt$class %in% c("autoout", "out")], 0) / o$devsize[1]))
					}

					# add margins (compensate for legend frames)
					# the final calculations of these margins are computed in tmapGridComp (this is just to compute the meta.auto_margins)
					# those calculations are take the component.offset into account

					sel_tb = c(3,1)[meta.auto_margins[c(3,1)]!=0]
					sel_lr = c(2,4)[meta.auto_margins[c(2,4)]!=0]
					if (length(sel_tb)) meta.auto_margins[sel_tb] = meta.auto_margins[sel_tb] + 2 * (o$frame.lwd * o$scale / 144) / o$devsize[2]
					if (length(sel_lr)) meta.auto_margins[sel_lr] = meta.auto_margins[sel_lr] + 2 * (o$frame.lwd * o$scale / 144) / o$devsize[1]
				}
			}

			if (meta.automatic && any(margins.used)) {
				meta.auto_margins = rep(meta.auto_margins, length.out = 4)
				meta.margins[margins.used] = meta.auto_margins[margins.used]

				# redo calculations
				meta.buffers = sign(meta.margins) * c(bufferH, bufferW, bufferH, bufferW) # outside and inside
				fixedMargins  =  outer.margins + meta.buffers * 2 + meta.margins + xylab.margins + panel.xtab.size + grid.buffers + grid.margins
			}
		} else {
			meta.buffers = c(0, 0, 0, 0)
			meta.margins = c(0, 0, 0, 0)
		}

		# determine number of rows and cols
		npp = prod(nby[1:2]) # number per page

		if (!is.na(o$ncols) && !is.na(o$nrows)) {
			npp = min(npp, o$ncols * o$nrows)
		}

		if (type %in% c("grid", "anigrid")) {
			nrows = nby[1]
			ncols = nby[2]
		} else if (type %in% c("stack", "anistack")) {
			if (orientation == "horizontal") {
				nrows = 1
				ncols = npp
			} else {
				nrows = npp
				ncols = 1
			}
		} else {
			if (is.na(nrows) && !is.na(ncols)) {
				nrows = ceiling((nby[1] / ncols))
			} else if (!is.na(nrows) && is.na(ncols)) {
				ncols = ceiling((nby[1] / nrows))
			} else if (is.na(nrows) && is.na(ncols)) {

				# loop through col row combinations to find best nrow/ncol
				# b needed to compare landscape vs portrait. E.g if prefered asp is 2, 1 is equally good as 4
				ncols = which.min(vapply(1L:npp, function(nc) {
					nr = ceiling(npp / nc)

					# calculate available width and height. They can be negative, at this stage this is avoided my taking at least a small number
					width = max(1e-9, ((1 - sum(fixedMargins[c(2, 4)])) - (nc * sum(panel.wrap.size[c(2,4)])) - (nc - 1) * between_marginW) / nc)
					height = max(1e-9, ((1 - sum(fixedMargins[c(1, 3)])) - (nr * sum(panel.wrap.size[c(1,3)])) - (nr - 1) * between_marginH) / nr)

					a = (width / height) * dasp
					b = ifelse(a<pasp, pasp/a, a/pasp)
					b
				}, FUN.VALUE = numeric(1)))


				nrows = ceiling(npp / ncols)
			}
			if ((nrows == 1 || ncols == 1) && set_to_stack_message) message_wrapstack(nrows == 1)
		}

		#overall scale down factor for facets
		width_forn = max(1e-9, ((1 - sum(fixedMargins[c(2, 4)])) - (ncols * sum(panel.wrap.size[c(2,4)])) - (ncols - 1) * between_marginW) / ncols)
		width_for1 = max(1e-9, ((1 - sum(fixedMargins[c(2, 4)])) - (sum(panel.wrap.size[c(2,4)]))))

		height_forn = max(1e-9, ((1 - sum(fixedMargins[c(1, 3)])) - (nrows * sum(panel.wrap.size[c(1,3)])) - (nrows - 1) * between_marginH) / nrows)
		height_for1 = max(1e-9, ((1 - sum(fixedMargins[c(1, 3)])) - (sum(panel.wrap.size[c(1,3)]))))

		scale_down = (1 / sqrt((width_for1 * height_for1) / (width_forn * height_forn))) ^ (1 / scale.factor)

		#
		# title.size = title.size * scale
		# legend.title.size = legend.title.size * scale
		# legend.text.size = legend.text.size * scale
		#
		# panel.label.size = panel.label.size * scale


		if (is.na(panel.labels[1])) {
			# quick fix
			# to do: checks
			panel.labels = fl[1:3]
		} else {
			if (!is.list(panel.labels)) {
				panel.labels = list(panel.labels, NULL)
			}
		}
		if (type %in% c("stack", "wrap")) {
			if (!is.null(panel.labels[[1]])) {
				panel.labels.dim = 1
			} else if (!is.null(panel.labels[[3]])) {
				panel.labels.dim = 3
			} else {
				panel.labels.dim = NA
				panel.type = "none"
			}


			if (nby[3] != 1L) {
				npages = nby[3]
				panel.labels.dim_data = 3
			} else {
				npages = ceiling(n / (nrows * ncols))
				panel.labels.dim_data = panel.labels.dim
			}
		} else {
			npages = ceiling(n / (nrows * ncols))
		}




		legend.position = NA

		if (is.na(space_overlay)) {
			space_overlay = .TMAP$raster_wrap
		}

		if (gs == "Leaflet") {
			if (!is.logical(set_bounds)) if (length(set_bounds) !=4 || !is.numeric(set_bounds)) stop("Incorrect set_bounds argument", call.=FALSE)


			if (!is.na(set_view[1])) {
				if (!is.numeric(set_view)) stop("set_view is not numeric")
				if (!length(set_view) %in% c(1, 3)) stop("set_view does not have length 1 or 3")
			}
			if (!is.na(set_zoom_limits[1])) {
				if (!is.numeric(set_zoom_limits)) stop("set_zoom_limits is not numeric")
				if (!length(set_zoom_limits)==2) stop("set_zoom_limits does not have length 2")
				if (set_zoom_limits[1] >= set_zoom_limits[2]) stop("incorrect set_zoom_limits")
			} else {
				set_zoom_limits <- c(NA, NA)
			}
			if (!is.na(set_view[1]) && !is.na(set_zoom_limits[1])) {
				if (set_view[length(set_view)] < set_zoom_limits[1]) {
					if (show.warnings) warning("default zoom smaller than minimum zoom, now it is set to the minimum zoom")
					set_view[length(set_view)] <- set_zoom_limits[1]
				}
				if (set_view[length(set_view)] > set_zoom_limits[2]) {
					if (show.warnings) warning("default zoom larger than maximum zoom, now it is set to the maximum zoom")
					set_view[length(set_view)] <- set_zoom_limits[2]
				}
			}
		}

	})

}

update_scale_args = function(scaleName, args, aes, o) {
	if (scaleName %in% names(o$scale.misc.args)) {
		def = o$scale.misc.args[[scaleName]]
		for (nam in names(def)) {
			if (is.null(args[[nam]])) args[[nam]] = getAesValue(def[[nam]], aes)
		}
	}
	args
}


# extention of substr: when -1, then cound from tail
substr2 = function(x, start, stop) {
	if (start < 1 || stop < 1) {
		n = nchar(x)
		if (start < 1) {
			start = n + start
		}
		if (stop < 1) {
			stop = n + stop
		}
	}
	substr(x, start, stop)
}


tmapScaleContinuous = function(x1, scale, legend, chart, o, aes, layer, layer_args, sortRev, bypass_ord, submit_legend = TRUE) {
	# update misc argument from tmap option scale.misc.args
	scale = update_scale_args(substr2(class(scale)[1], 10, 0), scale, aes, o)

	cls = data_class(x1, midpoint_enabled = !is.null(scale$midpoint))
	maincls = class(scale)[1]

	if (attr(cls, "unique") && any(is.na(scale$limits)) && is.null(scale$ticks)) {
		scale$ticks = NA
		cli::cli_inform("The visual variable {.arg {aes}} of the layer {.str {layer}} contains a unique value. Therefore a discrete scale is applied (tm_scale_discrete).")
		return(tmapScaleDiscrete(x1, scale, legend, chart, o, aes, layer, layer_args, sortRev, bypass_ord, submit_legend = submit_legend))
	}

	#if (cls[1] == "na") stop("data contain only NAs, so ", maincls, " cannot be applied", call. = FALSE)

	if (cls[1] != "num") {
		if (!is.factor(x1)) x1 = as.factor(x1)
		x1 = as.integer(x1)
		if (length(x1) > 1) warning(maincls, " is supposed to be applied to numerical data", call. = FALSE)
	}

	x1 = without_units(x1)

	if (aes %in% c("lty", "shape", "pattern")) stop("tm_scale_continuous cannot be used for layer ", layer, ", aesthetic ", aes, call. = FALSE)

	scale = get_scale_defaults(scale, o, aes, layer, cls)

	vnum = (is.numeric(scale$values) || inherits(scale$values, "tmapSeq"))
	#vnum=FALSE

	show.messages <- o$show.messages
	show.warnings <- o$show.warnings

	with(scale, {
		allna = all(is.na(x1))


		# if (all(is.na(x1))) {
		# }


		tr = get(paste0("trans_", trans))

		trargs = formals(tr$fun)
		trargs$x = NULL
		trargs_match = intersect(names(trargs), names(trans.args))
		if (!is.null(trargs_match) && length(trargs_match)) {
			trargs[trargs_match] = trans.args[trargs_match]
		}


		if (!allna) {
			xrange = range(x1, na.rm = TRUE)

			if (xrange[1] < tr$domain[1]) stop("Values found that are lower than the valid domain", call. = FALSE)
			if (xrange[2] > tr$domain[2]) stop("Values found that are higher than the valid domain", call. = FALSE)
		}

		udiv = identical(use_div(brks = NULL, midpoint), TRUE)

		ticks.specified = !is.null(ticks)
		limits.specified = (length(limits) == 2L)

		if (!limits.specified) {
			if (allna) {
				if (show.messages) message("Variable(s) \"", paste(aes, collapse = "\", \""), "\" only contains NAs. Legend disabled for tm_scale_continuous, unless limits are specified")
				chart = within(chart, {
					labels = label.na
					vvalues = c(value.na, value.na)
					breaks = c(0, 1)
					na.show = TRUE
					x1 = x1[1]
				})
				return(tmapScale_returnNA(n = length(x1), legend = legend, chart = chart, value.na = value.na, label.na = label.na, label.show = label.show, na.show = legend$na.show, sortRev = sortRev, bypass_ord = bypass_ord))
			}
			if (is.na(limits)) {
				limits = range(x1, na.rm = TRUE)
			} else {
				if (length(limits) == 1L) limits = range(c(limits, x1), na.rm = TRUE)
				if (limits[1] < tr$domain[1]) limits[1] = tr$domain[1]
				if (limits[2] > tr$domain[2]) limits[2] = tr$domain[2]
			}
			if (ticks.specified) limits = range(c(limits, ticks))
		}

		if (limits[1] < tr$domain[1]) stop("Lower limit too low", call. = FALSE)
		if (limits[2] > tr$domain[2]) stop("Upper limit too high", call. = FALSE)

		if (ticks.specified) {
			if (limits.specified) {
				if (any(ticks < limits[1])) stop("(Some) ticks are lower than the lowest limit. Please remove these ticks or adjust the lower limit.", call. = FALSE)
				if (any(ticks > limits[2])) stop("(Some) ticks are higher than the upper limit. Please remove these ticks or adjust the upper limit.", call. = FALSE)
			}
			n = length(ticks) - 1
			ticks_t = do.call(tr$fun, c(list(x = ticks), trargs))
		}

		if (limits.specified) {
			outliers.trunc = rep(outliers.trunc, length.out = 2)
			x1_low = x1 < limits[1]
			x1_high = x1 > limits[2]
			if (any(x1_low, na.rm = TRUE)) {
				if (!outliers.trunc[1]) message("Values have been found that are lower than the lowest limit. These 'outliers' have been set to NA. Use outliers.trunc to truncate them to the lower limit")
				x1[which(x1_low)] = if (outliers.trunc[1]) limits[1] else NA
			}
			if (any(x1_high, na.rm = TRUE)) {
				if (!outliers.trunc[2]) message("Values have been found that are higher than the upper limit. These 'outliers' have been set to NA. Use outliers.trunc to truncate them to the upper limit")
				x1[which(x1_high)] = if (outliers.trunc[2]) limits[2] else NA
			}
		}

		x_t = do.call(tr$fun, c(list(x = x1), trargs))
		limits_t = do.call(tr$fun, c(list(x = limits), trargs))
		domain_t = do.call(tr$fun, c(list(x = tr$domain), trargs))

		if (!vnum) {
			breaks = cont_breaks(limits_t, n=o$continuous.nclasses)
			if (is.null(labels)) {
				ncont = n
			} else {
				if (ticks.specified && length(labels) != n+1) {
					if (show.warnings) warning("The length of legend labels is ", length(labels), ", which differs from the length of the breaks (", (n+1), "). Therefore, legend labels will be ignored", call.=FALSE)
					labels = NULL
				} else {
					ncont = length(labels)
				}
			}
			q = num2breaks(x = x_t, n = o$continuous.nclasses, style = "fixed", breaks=breaks, approx=TRUE, interval.closure = "left", var=paste(layer, aes, sep = "-"), args = list())

			breaks = q$brks
			nbrks = length(breaks)
			n2 = nbrks - 1

		}




		# update range if NA (automatic)
		if (is.na(values.range[1])) {
			fun_range = paste0("tmapValuesRange_", aes)
			values.range = do.call(fun_range, args = list(x = values, n = n, isdiv = udiv))
		}
		if (length(values.range) == 1 && !is.na(values.range[1])) values.range = c(0, values.range)


		check_values(layer, aes, values)

		fun_isdiv = paste0("tmapValuesIsDiv_", aes)

		isdiv = !is.null(midpoint) || do.call(fun_isdiv, args = list(x = values))

		# determine midpoint
		if ((is.null(midpoint) || is.na(midpoint)) && isdiv) {
			rng <- range(x_t, na.rm = TRUE)
			if (rng[1] < 0 && rng[2] > 0 && is.null(midpoint)) {

				if (show.messages) message_midpoint(aes, "tm_scale_continous")
				midpoint <- 0
			} else {
				if (vnum) {
					midpoint = mean(limits_t)
				} else {
					if ((n2 %% 2) == 1) {
						# number of classes is odd, so take middle class (average of those breaks)
						midpoint <- mean.default(breaks[c((n2+1) / 2, (n2+3) / 2)])
					} else {
						midpoint <- breaks[(n2+2) / 2]
					}
				}
			}
		}

		fun_getVV = paste0("tmapValuesVV_", aes)

		if (!vnum) {
			#### discretisize

			# number of visual values in legend item (belonging to one label)
			nvv = o$continuous.nclass_per_legend_break


			VV = do.call(fun_getVV, list(x = values, value.na = value.na, isdiv = isdiv, n = n2, dvalues = breaks, midpoint = midpoint, range = values.range, scale = values.scale * o$scale, are_breaks = TRUE, rep = values.repeat, o = o))

			vv = VV$vvalues
			value.na = VV$value.na

			sfun = paste0("tmapValuesScale_", aes)
			cfun = paste0("tmapValuesColorize_", aes)
			if (is.na(value.neutral)) value.neutral = VV$value.neutral else value.neutral = do.call(sfun, list(x = do.call(cfun, list(x = value.neutral, pc = o$pc)), scale = values.scale))

			ids = classInt::findCols(q)
			vals = vv[ids]
		} else {
			if (is.numeric(values)) {
				values = tm_seq(values[1], values[length(values)], power = "lin")
			}
			VV = transform_values(x_t, limits_t, values.range, values$power, values.scale * o$scale)

			vals = VV$x
			if (is.na(value.neutral)) value.neutral = VV$neutral
		}

		isna = is.na(vals)
		anyNA = any(isna)

		na.show = update_na.show(label.show, legend$na.show, anyNA)

		if (is.null(sortRev)) {
			ids = NULL
		} else if (is.na(sortRev)) {
			ids = rep(1L, length(x1))
		} else if (sortRev) {
			if (vnum) {
				ids = rank(-vals)
			} else {
				ids = (as.integer(n2) + 1L) - ids
			}
		} else if (vnum) {
			ids = rank(vals)
		}

		if (anyNA) {
			vals[isna] = value.na
			if (!is.null(sortRev)) ids[isna] = 0L
		}

		labels_exp = !ticks.specified && trans == "log" && trargs$base != exp(1)


		if (ticks.specified) {
			b_t = ticks_t
			b = do.call(tr$rev, c(list(x = b_t), trargs))
		} else {
			if (labels_exp) {
				b = do.call(tr$rev, c(list(x = pretty(limits_t, n = n)), trargs))
			} else {
				b  = prettyTicks(do.call(tr$rev, c(list(x = seq(limits_t[1], limits_t[2], length.out = n)), trargs)))
			}

			if (!(aes %in% c("col", "fill"))) b = b[b!=0]
			b_t = do.call(tr$fun, c(list(x = b), trargs))
		}
		if (length(b_t) == 2) {
			sel = TRUE
			labels_select = c(TRUE, TRUE)
		} else {
			# #1039
			# to include min/max if they are closer to the next 'not included' tick than the included tick
			# e.g. tm_shape(World) + tm_polygons("gender", fill.scale = tm_scale_continuous())
			# lowest value is 0.009 whereas ticks are 0.2, 0.4, 0.6, 0.8.
			stp1 = (b_t[2] - b_t[1]) / 2
			stp2 = (b_t[length(b_t)] - b_t[length(b_t) - 1L]) / 2
			limits_t_tmp = c(limits_t[1] - stp1, limits_t[2] + stp2)
			sel = (b_t>=limits_t_tmp[1] & b_t<=limits_t_tmp[2])

			labels_select = ((b_t>=limits_t[1] & b_t<=limits_t[2])[sel])
		}
		b_t = b_t[sel]
		b = b[sel]



		nbrks_cont <- length(b_t)

		if (vnum) {
			vvalues = transform_values(b_t, limits_t, values.range, values$power, values.scale * o$scale, include.neutral = FALSE)
		} else {
			id = as.integer(cut(b_t, breaks=breaks, include.lowest = TRUE))

			if (sum(!is.na(id)) < 2) {
				if (is.null(scale$ticks)) {
					cli::cli_abort("{.field [tm_scale_continuous]} too few tick labels specified. Either increase {.arg n} or specify tick values via {.arg ticks}")
				} else {
					cli::cli_abort("{.field [tm_scale_continuous]} too few tick labels specified via {.arg ticks}")
				}
			}

			# impute missing head and tail (#1039)
			missing_head = is.na(id[1])
			if (missing_head) id[1] = 2 * id[2] - id[3]

			nid = length(id)

			missing_tail = is.na(id[nid])
			if (missing_tail) id[nid] = 2 * id[nid-1] - id[nid-2]

			id_step = id[-1] - head(id, -1)
			id_step = c(id_step[1], id_step, id_step[length(id_step)])
			id_lst = mapply(function(i, s1, s2){
				#res = round(seq(i-floor(id_step/2), i+ceiling(id_step/2), length.out=11))[1:10]
				res1 = round(seq(i-floor(s1/2), i, length.out=(nvv/2)+1L))
				res2 = round(seq(i, i+ceiling(s2/2), length.out=(nvv/2)+1L))[2:(nvv/2)]
				res = c(res1, res2)
				res[res<1 | res>o$continuous.nclasses] = NA
				res
			}, id, head(id_step, -1), id_step[-1], SIMPLIFY = FALSE)
			vvalues = lapply(id_lst, function(i) {
				if (legend$reverse) rev(vv[i]) else vv[i]
			})
		}






		if (legend$reverse) vvalues = rev(vvalues)

		if (na.show) vvalues = c(vvalues, value.na)

		# temporarily stack gradient values
		if (!vnum) vvalues = cont_collapse(vvalues)

		# create legend values
		values = b

		# create legend labels for continuous cases
		if (is.null(labels)) {
			if (labels_exp) {
				labels = paste(trargs$base, b_t, sep = "^")
			} else {
				labels = do.call("fancy_breaks", c(list(vec=b, as.count = FALSE, intervals=FALSE, interval.closure="left"), label.format))
			}
		} else {
			labels = rep(labels, length.out=nbrks_cont)
			attr(labels, "align") <- label.format$text.align
		}

		if (legend$reverse) {
			labels.align = attr(labels, "align")
			labels = rev(labels)
			labels_select = rev(labels_select)
			attr(labels, "align") = labels.align
		}
		if (na.show) {
			labels.align = attr(labels, "align")
			labels = c(labels, label.na)
			labels_select = c(labels_select, TRUE)
			attr(labels, "align") = labels.align
		}

		layer_args = c(layer_args, list(ticks = scale$ticks, limits = scale$limits))

		legend = within(legend, {
			nitems = length(labels)
			labels = labels
			labels_select = labels_select
			dvalues = values
			vvalues = vvalues
			vneutral = value.neutral
			na.show = get("na.show", envir = parent.env(environment()))
			scale = "continuous"
			tr = tr
			limits = limits
			layer_args = layer_args
		})


		# NOTE: tr and limits are included in the output to facilitate the transformation of the leaflet continuous legend ticks (https://github.com/rstudio/leaflet/issues/665)
		#vvalues_mids = sapply(cont_split(vvalues), "[", nvv/2)
		#vvalues_mids[vvalues_mids == "NA"] = NA


		chartFun = paste0("tmapChart", toTitleCase(chart$summary))

		chart = do.call(chartFun, list(chart,
									   bin_colors = NULL,
									   breaks_def = chart$breaks,
									   na.show = na.show,
									   x1 = x1))



		# chart = within(chart, {
		# 	labels = labels
		# 	vvalues = sapply(cont_split(vvalues), "[", 5)
		# 	breaks_def = legend$dvalues
		# 	na.show = get("na.show", envir = parent.env(environment()))
		# 	x1 = x1
		# })



		if (submit_legend) {
			if (bypass_ord) {
				format_aes_results(vals, legend = legend, chart = chart)
			} else {
				format_aes_results(vals, ids, legend, chart = chart)
			}
		} else {
			list(vals = vals, ids = ids, legend = legend, chart = chart, bypass_ord = bypass_ord)
		}

	})
}

tmapScaleRGB_RGBA = function(xlist, scale, legend, chart, o, aes, layer, layer_args, sortRev, bypass_ord, submit_legend = TRUE, called) {

	k = length(xlist)
	n = length(xlist[[1]])

	cls = vapply(xlist, function(xi) {
		data_class(xi)[1]
	}, FUN.VALUE = character(1))


	# cls1 = data_class(x1)
	# cls2 = data_class(x2)
	# cls3 = data_class(x3)

	if (!(aes %in% c("col", "fill"))) stop("tm_scale_rgb cannot be used for layer ", layer, ", aesthetic ", aes, call. = FALSE)
	if (any(cls != "num")) {
		stop(called, " requires ", {if (called == "tm_scale_rgb") 3 else 4}, " variables", call. = FALSE)
	}

	#scale = get_scale_defaults(scale, opt, aes, layer, cls)
	scale$value.na = if (is.na(scale$value.na) || identical(scale$value.na, TRUE)) getAesOption("value.na", o, aes, layer, cls = cls[1]) else scale$value.na

	isna = Reduce("|", lapply(xlist, is.na))

	mx = max(unlist(xlist, recursive = FALSE, use.names = FALSE), na.rm = TRUE)

	scale = within(scale, {
		if (is.logical(stretch)) {
			if (stretch)
				stretch.method = "percent"
			else max_color_value = max(max_color_value, mx)
		}
		if (is.character(stretch)) {
			if (!stretch %in% c("percent", "histogram"))
				stretch.method = "percent"
			else stretch.method = stretch
			stretch = TRUE
		}
	})
	cutoff = function(x, probs, stretch.method = "percent", max_color_value = 255L) {
		if (stretch.method == "percent") {
			qs = if (all(probs == c(0, 1)))
				range(x, na.rm = TRUE)
			else quantile(x, probs, na.rm = TRUE)
			x = (x - qs[1])/(qs[2] - qs[1])
			x[x > 1] = 1
			x[x < 0] = 0
			x * max_color_value
		}
		else if (stretch.method == "histogram") {
			x = (stats::ecdf(x))(x)
			x * max_color_value
		}
		else {
			qs = range(x)
			(x - qs[1])/(qs[2] - qs[1]) * max_color_value
		}
	}

	if (scale$stretch) {
		xlist2 = lapply(xlist, cutoff, probs = scale$probs, stretch.method = scale$stretch.method, max_color_value = scale$max_color_value)
	} else {
		xlist2 = xlist
	}

	if (any(isna)) {
		values = rep(scale$value.na, n)
		values[!isna] = do.call(grDevices::rgb, c(lapply(xlist2, function(xl) xl[!isna]), list(maxColorValue = scale$max_color_value)))
	} else {
		values = do.call(grDevices::rgb, c(xlist2, list(maxColorValue = scale$max_color_value)))
	}

	if (layer_args$saturation != 1) {
		pc = o$pc
		pc$saturation = pc$saturation * layer_args$saturation
		values = do.call(process_color, c(list(col = values), pc))
	}

	legend = list(title = NA,
				  nitems = 0,
				  labels = NA,
				  dvalues = NA,
				  vvalues = NA,
				  vneutral = "grey50",
				  na.show = NA,
				  scale = "RGB",
				  show = FALSE,
				  active = FALSE)

	chart = list(show = FALSE)

	if (submit_legend) {
		if (bypass_ord) {
			format_aes_results(values, legend = legend, chart = chart)
		} else {
			format_aes_results(values, ids, legend, chart = chart)
		}
	} else {
		list(vals = values, ids = ids, legend = legend, chart = chart, bypass_ord = bypass_ord)
	}
}


tmapScaleRGB = function(x1, x2, x3, scale, legend, chart, o, aes, layer, layer_args, sortRev, bypass_ord, submit_legend = TRUE) {
	tmapScaleRGB_RGBA(xlist = list(x1, x2, x3), scale = scale, legend = legend, chart = chart, o = o, aes = aes, layer = layer, layer_args = layer_args, sortRev = sortRev, bypass_ord = bypass_ord, submit_legend = submit_legend)
}

tmapScaleRGBA = function(x1, x2, x3, x4, scale, legend, chart, o, aes, layer, layer_args, sortRev, bypass_ord, submit_legend = TRUE) {
	tmapScaleRGB_RGBA(xlist = list(x1, x2, x3, x4), scale = scale, legend = legend, chart = chart, o = o, aes = aes, layer = layer, layer_args = layer_args, sortRev = sortRev, bypass_ord = bypass_ord, submit_legend = submit_legend)
}

#' @export
tmapReproject.sfc = function(shp, tmapID, bbox = NULL, ..., crs) {
	if (is.na(sf::st_crs(shp))) {
		shp2 = shp
	} else {
		res = transform_ortho(shp, crs, tmapID)
		shp2 = res$shp
		tmapID = res$tmapID
	}

	if (!is.null(bbox$x)) {
		if (identical(bbox$x, "FULL")) {
			bbox = list(x = full_bbox(crs))
		} else {
			bbox = list(x = do.call(tmaptools::bb, c(bbox, list(projection = crs))))
		}
	}
	shapeTM(shp2, tmapID, bbox, ...)
}

#' @export
tmapShape.sf = function(shp, is.main, crs, bbox, unit, filter, layer, shp_name, smeta, o, tmf) {
	if (identical(crs, "auto")) crs = auto_crs(shp, crs_extra = o$crs_extra, crs_global = o$crs_global) else crs = sf::st_crs(crs)
	reproj = (!is.null(crs) && !is.na(crs) && sf::st_crs(shp) != crs)

	if (reproj) {
		shp = transform_ortho(shp, crs)
	}

	sfc = sf::st_geometry(shp)

	if (inherits(sfc[[1]], c("XYZ", "XYM", "XYZM"))) {
		sfc = sf::st_zm(sfc)
	}

	if (o$check_and_fix) sfc = check_fix(sfc, shp_name, reproj, o$show.messages)

	# if check_fix fails, is_valid contains the valid ids
	isv = attr(sfc, "is_valid")
	if (!is.null(isv)) {
		shp = shp[isv, ]
	}
	dt = data.table::as.data.table(sf::st_drop_geometry(shp))

	dtcols = copy(names(dt))


	if (is.null(filter)) filter = rep_len(TRUE, nrow(dt))
	dt[, ':='(tmapID__ = 1L:.N, sel__ = filter)]

	make_by_vars(dt, tmf, smeta)

	shpTM = shapeTM(shp = sfc, tmapID = 1L:(length(sfc)), bbox = bbox)

	structure(list(shpTM = shpTM, dt = dt, is.main = is.main, dtcols = dtcols, shpclass = "sfc", bbox = bbox, unit = unit, shp_name = shp_name, smeta = smeta, type_ids = NULL, type_vars = NULL), class = "tmapShape")
}

#' @export
tmapSubsetShp.sf = function(shp, vars) {

	if ("AREA_" %in% vars && !("AREA_" %in% names(shp))) {
		shp$AREA_ = sf::st_area(shp)
	}
	if ("LENGTH_" %in% vars && !("LENGTH_" %in% names(shp))) {
		shp$LENGTH_ = sf::st_length(shp)

	}
	if ("MAP_COLORS" %in% vars) {
		shp$MAP_COLORS = as.factor(tmaptools::map_coloring(shp, ncols = 7))
	}

	if (!length(vars)) {
		vars = "dummy__"
		shp$dummy__ = logical(nrow(shp))
	}
	shp[, vars]
}


#' @export
tmapSubsetShp.sfc = function(shp, vars) {
	s = sf::st_sf(dummy__ = TRUE, geometry = shp)
	if ("AREA" %in% vars) {
		s$AREA_ = sf::st_area(shp)
	}
	if ("LENGTH" %in% vars) {
		s$LENGTH_ = sf::st_length(shp)
	}
	if ("MAP_COLORS" %in% vars) {
		s$MAP_COLORS = as.factor(tmaptools::map_coloring(shp, ncols = 7))
	}
	s
}



#' @export
tmapGetShapeMeta2.sf = function(shp, smeta, o) {
	vars = setdiff(names(shp), attr(shp, "sf_column"))
	smeta$vars_levs = lapply(seq_along(vars), function(i) {
		get_fact_levels_na(shp[[i]], o)
	})
	names(smeta$vars_levs) = vars
	smeta
}


#' @export
tmapGetShapeMeta1.sf = function(shp, layer, o) {
	vars = setdiff(names(shp), attr(shp, "sf_column"))
	names(vars) = vars
	#vars_levs = lapply(vars, function(v) {get_fact_levels_na(shp[[v]], o)})
	dims = character(0)
	dims_vals = list()

	list(vars = vars,
		 dims = dims,
		 dims_vals = dims_vals)
}


#' @export
tmapGetShapeMeta2.sfc = function(shp, smeta, o) {
	vars = character(0)
	smeta$vars_levs = list()
	smeta
}


#' @export
tmapGetShapeMeta1.sfc = function(shp, layer, o) {
	vars = character(0)
	dims = character(0)
	dims_vals = list()

	list(vars = vars,
		 dims = dims,
		 dims_vals = dims_vals)
}

#' @export
#' @rdname tmap_internal
tmapScaleBivariate = function(x1, x2, scale, legend, chart, o, aes, layer, layer_args, sortRev, bypass_ord, submit_legend = TRUE) {

	if (!(aes %in% c("fill", "col", "shape"))) stop("tm_scale_bivariate cannot be used for layer ", layer, ", aesthetic ", aes, call. = FALSE)

	scale = get_scale_defaults(scale, o, aes, layer, cls = "biv")


	dummyLegend = legend
	dummyLegend$show = FALSE

	dummyChart = list(show = FALSE, summary = "pass")

	res = mapply(function(x, s) {
		f = s$FUN
		s$FUN = NULL
		# update label.format
		s$label.format = process_label_format(scale$label.format, o$label.format)
		r = do.call(f, list(x1 = x, scale = s, legend = dummyLegend, chart = dummyChart, o = o, aes = aes, layer = layer, layer_args = layer_args, sortRev = sortRev, bypass_ord = bypass_ord, submit_legend = FALSE))

		leg = r$legend
		vv = leg$vvalues
		lab = leg$labels
		na = leg$na.show
		if (na) {
			vv = head(vv, -1)
			lab = head(lab, -1)
		}
		vid = match(r$vals, vv)
		n = length(lab)
		list(vv = vv, lab = lab, na = na, vid = vid, n = n, ids = r$ids, scale = leg$scale, chart = r$chart)

	}, list(x1, x2), scale[1:2], SIMPLIFY = FALSE)

	if (res[[1]]$scale == "continuous" || res[[2]]$scale == "continuous") stop("tm_scale_bivariate not implemented for continuous scales", call. = FALSE)

	n1 = res[[1]]$n # rows m
	n2 = res[[2]]$n # cols n

	vals = res[[1]]$vid + (res[[2]]$vid - 1L) * n1

	with(scale, {
		check_values(layer, aes, values)

		fun_getBVV = paste0("tmapValuesBVV_", aes)
		VV = do.call(fun_getBVV, list(x = values, value.na = value.na, m = n1, n = n2, scale = values.scale * o$scale, rep = values.repeat, o = o))

		vvalues = VV$vvalues
		value.na = VV$value.na


		#pal = c4a(cols4all::.P$c4a$bivs$pu_gn_bivs, n = n2, m = n1)[n1:1,]
		#colorNA = c4a_na(cols4all::.P$c4a$bivs$pu_gn_bivs)

		#pal = as.vector(pal)


		pal = as.vector(vvalues)#[n1:1,])

		attr(pal, "biv") = TRUE
		attr(pal, "m") = n1
		attr(pal, "n") = n2

		vals = pal[vals]

		vals[is.na(vals)] = value.na

		ids = res[[1]]$ids + res[[2]]$ids

		na = res[[1]]$na || res[[2]]$na


		legend$labels = c(res[[1]]$lab, res[[2]]$lab)
		legend$nitems = n1 + na
		legend$dvalues = vals
		legend$vvalues = pal
		legend$vneutral = pal[4]
		legend$na.show = na
		legend$layer_args = layer_args



		chartFun = paste0("tmapChart", toTitleCase(chart$summary))

		chart = do.call(chartFun, list(chart,
									   chart1 = res[[1]]$chart,
									   chart2 = res[[2]]$chart))


		# for consistency with the other tmapScaleX
		# not needed because submit_legend is only needed to supress legends of the two variables by which a bivariate legend is created.
		if (submit_legend) {
			if (bypass_ord) {
				format_aes_results(vals, legend = legend, chart = chart)
			} else {
				format_aes_results(vals, ids, legend, chart)
			}
		} else {
			list(vals = vals, ids = ids, legend = legend, bypass_ord = bypass_ord)
		}



	})

}

#' Internal method that reprojects a shape object
#'
#' Internal method that reprojects a shape object
#'
#' @param shp shape
#' @param tmapID tmap id values
#' @param bbox bounding box
#' @param ... other args
#' @param crs crs
#' @export
#' @keywords internal
tmapReproject = function(shp, tmapID, bbox, ..., crs) {
	UseMethod("tmapReproject")
}


reproject_bbox = function(bbox, crs) {
	s = tmaptools::bb_poly(bbox) #st_as_sfc does not add intermediate points
	s2 = sf::st_transform(s, crs)
	sf::st_bbox(s2)
}




#' Internal method that processed shape objects
#'
#' Internal method that processed shape objects
#'
#' @param shp shp
#' @param is.main is.main
#' @param crs crs
#' @param bbox Bounding box
#' @param unit unit
#' @param filter filter
#' @param shp_name shp_name
#' @param smeta smeta
#' @param o o
#' @param tmf tmf
#' @export
#' @import data.table
#' @keywords internal
tmapShape = function(shp, is.main, crs, bbox, unit, filter, layer, shp_name, smeta, o, tmf) {
	UseMethod("tmapShape")
}

#' Internal method that subsets data from shape objects
#'
#' Internal method that subsets data from shape objects
#'
#' @param shp shape
#' @param vars a vector of variable names
#' @export
#' @keywords internal
tmapSubsetShp = function(shp, vars) {
	UseMethod("tmapSubsetShp")
}

#' Internal method that extracts meta data from shape objects
#'
#' Internal method that extracts meta data from shape objects
#'
#' @param shp the shape object
#' @param layer (optional) the layer of the shape object
#' @param o the list of options
#' @export
#' @keywords internal
tmapGetShapeMeta1 = function(shp, layer, o) {
	UseMethod("tmapGetShapeMeta1")
}

#' Internal method that extracts more meta data from shape objects
#'
#' Internal method that extracts meta data from shape objects
#'
#' @param shp the shape
#' @param smeta meta (from `tmapGetShapeMeta1()`)
#' @param o the list of options
#' @export
#' @keywords internal
tmapGetShapeMeta2 = function(shp, smeta, o) {
	UseMethod("tmapGetShapeMeta2")
}

trans_log = list(
	fun = function(x, base = exp(1)) log(x, base),
	rev = function(x, base = exp(1)) base^x,
	domain = c(1e-100, Inf)
)

# trans_logistic = list(
# 	fun = function(x) {
# 		1/ (1 + exp(-x))
# 	},
# 	rev = function(x) {
# 		log(x/(1 - x))
# 	},
# 	domain = c(-Inf, Inf)
# )


trans_sqrt = list(
	fun = sqrt,
	rev = function(x) x^2,
	domain = c(0, Inf)
)


trans_log1p = list(
	fun = log1p,
	rev = expm1,
	domain = c(0, Inf)
)

trans_pseudo_log = list(
	fun = function(x, base = exp(1), sigma = 1) asinh(x/(2 * sigma))/log(base), 
	rev = function(x, base = exp(1), sigma = 1) 2 * sigma * sinh(x * log(base)), 
	#d_transform = function(x) 1/(sqrt(4 + x^2/pseudo_log_sigma^2) * pseudo_log_sigma * log(pseudo_log_base)) 
	#d_inverse = function(x) 2 * pseudo_log_sigma * cosh(x * log(pseudo_log_base)) * log(pseudo_log_base))
	domain = c(-Inf, Inf)
)

trans_identity = list(
	fun = function(x) x,
	rev = function(x) x,
	domain = c(-Inf, Inf)
)

#' @export
#' @rdname tmapGridLeaflet
tmapLeafletDataPlot.tm_data_text = function(a, shpTM, dt, pdt, popup.format, hdt, idt, gp, bbx, facet_row, facet_col, facet_page, id, pane, group, glid, o, ...) {
	args = list(...)

	lf = get_lf(facet_row, facet_col, facet_page)

	rc_text = frc(facet_row, facet_col)

	res = select_sf(shpTM, dt)
	shp = res$shp
	dt = res$dt

	#shadow = gp$shadow

	gp = impute_gp(gp, dt)
	gp = rescale_gp(gp, o$scale_down)

	coords = sf::st_coordinates(shp)

	# in case shp is a multipoint (point_per == "segment"), expand gp:
	cp = expand_coords_gp(coords, gp, ndt = nrow(dt))
	coords = cp$coords
	gp = cp$gp

	if (cp$expanded) {
		shpTM_match = match(shpTM$tmapID_expanded, shpTM$tmapID)
	} else {
		shpTM_match = TRUE
	}

	text = as.character(dt$text[shpTM_match])
	if (all(text == "")) return(NULL)

	idt = dt$tmapID__[shpTM_match] |>
		submit_labels("text", pane, group)

	opt = leaflet::pathOptions(interactive = FALSE, pane = pane)


	clusterOpts = getClusterOpts(a$clustering)

	cidt = if (!is.null(clusterOpts)) submit_labels(idt[1], "cluster", pane, group) else NULL


	cex_set = unique(gp$cex)
	alpha_set = unique(gp$col_alpha)
	face_set = unique(gp$fontface)
	col_set = unique(gp$col)

	bgcol_set = unique(gp$bgcol)
	bgcol_alpha = unique(gp$bgcol_alpha)

	if (all(bgcol_set == "#00000000")) {
		bg_enabled = FALSE
	} else {
		bgres = split_alpha_channel(gp$bgcol, gp$bgcol_alpha)
		gp$bgcol = bgres$col
		gp$bgcol_alpha = bgres$opacity
		bg_enabled = TRUE
	}



	if (length(face_set) != 1) message("Variable fontfaces not supported by view mode")

	vary = (length(cex_set) != 1) || (length(alpha_set) != 1) || (length(face_set) != 1) || (length(col_set) != 1) || (length(bgcol_set) != 1) || (length(bgcol_alpha) != 1)

	sizeChar <- paste(round(gp$cex * 12), "px", sep="")

	# direction <- ifelse(gpl$text.just == "left", "right",
	# 			 ifelse(gpl$text.just == "right", "left",
	# 			 ifelse(gpl$text.just == "top", "bottom",
	# 			 ifelse(gpl$text.just == "bottom", "top", "center"))))
	direction = "right"


	# apply xmod and ymod
	delta = delta_per_lineheight(bbx)

	coords[,1] = coords[,1] + delta * gp$cex * gp$xmod
	coords[,2] = coords[,2] + delta * gp$cex * gp$ymod

	scale_px = function(textlines, cex) {
		ceiling(textlines * cex * 12)
	}


	make_shadow = function(cex, textcol) {
		if (a$halo) {
			if (is.na(a$halo.col)) a$halo.col = ifelse(is_light(textcol), "#000000DD", "#FFFFFFDD")
			make_halo_css(col = a$halo.col, mode = "halo", width = scale_px(a$halo.width, cex), blur = scale_px(a$halo.blur, cex), diag_scale = sqrt(2)/2, alpha = a$halo.alpha)
		} else if (a$shadow) {
			if (is.na(a$shadow.col)) a$shadow.col = ifelse(is_light(textcol), "#000000", "#FFFFFF")
			make_halo_css(col = a$shadow.col, mode = "shadow", offset.x = scale_px(a$shadow.offset.x, cex), offset.y = scale_px(a$shadow.offset.y, cex), alpha = 1)
		} else {
			NULL
		}
	}

	if (!vary) {
		lf = lf %>% addLabelOnlyMarkers(lng = coords[, 1], lat = coords[,2],
										label=text,
										group=group,
										layerId = idt,
										labelOptions = labelOptions(noHide = TRUE,
																	textOnly = TRUE,
																	pane = pane,
																	direction = direction,
																	opacity=gp$col_alpha[1],
																	textsize=sizeChar[1],
																	style=list("color"=gp$col[1],
																			   "background-color" = if (bg_enabled) paste0("rgba(", paste(col2rgb(gp$bgcol[1]), collapse = ", "), ",", gp$bgcol_alpha[1], ")") else NULL,
																			  "border" = if (a$bg.border) paste0(a$bg.border.lwd, "px solid rgba(", paste(col2rgb(a$bg.border.col[1]), collapse = ", "), ",", gp$bgcol_alpha[1], ")") else NULL,
																			  "line-height" = sizeChar[1],
																			  "text-shadow" = make_shadow(gp$cex[1], gp$col[1]),
																			  "padding" = paste0(scale_px(a$bg.padding / 2, gp$cex[1]), "px"))),
										options = markerOptions(pane = pane),
										clusterOptions = clusterOpts,
										clusterId = cidt)
	} else {
		for (i in 1:length(text)) {
			lf = lf %>% addLabelOnlyMarkers(lng = coords[i,1], lat = coords[i,2],
											label=text[i],
											group=group,
											layerId = idt[i],
											labelOptions = labelOptions(noHide = TRUE,
																		textOnly = TRUE,
																		pane = pane,
																		direction = direction,
																		opacity=gp$col_alpha[i],
																		textsize=sizeChar[i],
																		style=list("color"=gp$col[i],
																				   "background-color" = if (bg_enabled) paste0("rgba(", paste(col2rgb(gp$bgcol[i]), collapse = ", "), ",", gp$bgcol_alpha[i], ")") else NULL,
																				   "border" = if (a$bg.border) paste0(a$bg.border.lwd, "px solid rgba(", paste(col2rgb(a$bg.border.col[i]), collapse = ", "), ",", gp$bgcol_alpha[i], ")") else NULL,
																				   "line-height" = sizeChar[i],
																				   "text-shadow" = make_shadow(gp$cex[i], gp$col[i]),
																				   "padding" = paste0(scale_px(a$bg.padding / 2, gp$cex[i]), "px"))),
											options = markerOptions(pane = pane),
											clusterOptions = clusterOpts,
											clusterId = cidt)
		}
	}
	assign_lf(lf, facet_row, facet_col, facet_page)



	NULL
}


#' @export
#' @rdname tmapGridLeaflet
tmapLeafletDataPlot.tm_data_labels = function(a, shpTM, dt, pdt, popup.format, hdt, idt, gp, bbx, facet_row, facet_col, facet_page, id, pane, group, glid, o, ...) {
	NextMethod()
}

#' @export
#' @rdname tmapGridLeaflet
tmapLeafletDataPlot.tm_data_labels_highlighted = function(a, shpTM, dt, pdt, popup.format, hdt, idt, gp, bbx, facet_row, facet_col, facet_page, id, pane, group, glid, o, ...) {
	NextMethod()
}




make_halo_css <- function(
		col = "white",
		alpha = 0.9,
		width = 1,
		blur = 0,
		diagonals = TRUE,
		diag_scale = 1,
		mode = c("halo", "shadow"),
		offset.x = 2,
		offset.y = 2
) {

	mode <- match.arg(mode)

	# convert color to rgba
	rgb <- grDevices::col2rgb(col)
	col_css <- sprintf(
		"rgba(%d,%d,%d,%s)",
		rgb[1], rgb[2], rgb[3], alpha
	)

	# --- Simple drop shadow mode ---
	if (mode == "shadow") {
		return(sprintf(
			"%spx %spx %spx %s",
			offset.x, offset.y, blur, col_css
		))
	}

	# --- Halo mode ---
	# Cardinal directions
	offsets <- rbind(
		c(-1, 0),
		c( 1, 0),
		c( 0,-1),
		c( 0, 1)
	)

	# Optional diagonals
	if (diagonals) {
		diag_offsets <- rbind(
			c(-1,-1),
			c( 1,-1),
			c(-1, 1),
			c( 1, 1)
		)

		# allow smaller diagonal distance for rounder halo
		diag_offsets <- diag_offsets * diag_scale

		offsets <- rbind(offsets, diag_offsets)
	}

	# Scale by width
	offsets <- round(offsets * width, 2)

	parts <- apply(offsets, 1, function(x) {
		sprintf("%spx %spx 0 %s", x[1], x[2], col_css)
	})

	# Optional glow
	if (blur > 0) {
		parts <- c(parts, sprintf("0 0 %spx %s", blur, col_css))
	}

	paste(parts, collapse = ", ")
}

#' Popup specification for interactive maps
#'
#' `tm_popup()` specifies the popups that are shown in interactive (`"view"`)
#' mode when a feature is clicked. It is passed to the `popup` argument of the
#' layer functions ([tm_polygons()], [tm_symbols()], [tm_lines()], etc.). It
#' replaces the (now deprecated) layer arguments `popup.vars` and `popup.format`.
#'
#' @param vars Names of the data variables that are shown in the popup table.
#'   A (possibly named) character vector; when named, the names are used as
#'   labels in the popup table. Besides a character vector, the following
#'   special values are supported (identical to the former `popup.vars`
#'   argument):
#'   \describe{
#'     \item{`TRUE`}{show all variables of the shape object;}
#'     \item{`FALSE`}{disable popups;}
#'     \item{`NA` (default)}{automatic: if visual variables (e.g. `fill`) are
#'       used, only those are shown, otherwise all variables of the shape
#'       object are shown.}
#'   }
#' @param title Name of the data variable used as the popup title (the bold
#'   header shown above the popup table). This overrules the layer argument
#'   `id`, analogous to how `hover` overrules `id` for hover labels. The
#'   default (`NA`) means that the popup title is derived from `id` (the former,
#'   and still default, behaviour). A length-one character vector is expected;
#'   a named value is allowed and reserved for future use.
#' @param format A list of formatting options for the popup values, the output
#'   of [tm_label_format()]. Only applicable to numeric data variables. If one
#'   list of formatting options is provided, it is applied to all numeric
#'   variables of `vars`. A (named) list of lists can also be provided; in that
#'   case, each list of formatting options is applied to the named variable.
#' @param width Width of the popup content (view mode). A bare number is
#'   interpreted as pixels (e.g. `300` means `"300px"`); a character string is
#'   used as-is, so any CSS length is accepted (`"300px"`, `"20em"`, `"50%"`).
#'   The default `"auto"` lets the popup size to its content.
#' @param max.height Maximum height of the popup table before it becomes
#'   vertically scrollable (view mode). A bare number is interpreted as `em`
#'   (e.g. `5` means `"5em"`, roughly "show 5 lines"); a character string is
#'   used as-is. Default `"25em"`. Use `max.height = "none"` (or `NA`/`Inf`) to
#'   remove the cap, so the popup grows to fit its content and never scrolls.
#' @param title.align,label.align,value.align Horizontal alignment of the
#'   popup title (bold header), the variable-name (label) column, and the value
#'   column respectively. Each one of `"left"`, `"center"`, or `"right"`.
#'   Defaults: title `"left"`, label `"left"`, value `"right"`.
#' @param title.color,label.color,value.color Text color of the popup title,
#'   the label column, and the value column respectively. `NULL` (the default
#'   for title and value) inherits the browser/popup default; `label.color`
#'   defaults to `"#888888"` (grey).
#' @param css Optional free-form CSS, injected verbatim as a `<style>` block in
#'   each popup, for full restyling beyond the arguments above. Target the
#'   semantic classes `.tmap-popup` (container), `.tmap-popup-table`,
#'   `.tmap-popup-title`, `.tmap-popup-label`, and `.tmap-popup-value`. To size
#'   the popup box itself, target the backend's own popup element
#'   (`.leaflet-popup-content` in view mode, `.maplibregl-popup-content` in
#'   maplibre mode). Default `NULL`.
#' @return A `tm_popup` object.
#' @seealso [tm_polygons()], [tm_symbols()], [tm_lines()]
#' @export
tm_popup = function(vars = NA,
					title = NA,
					format = tm_label_format(),
					width = "auto",
					max.height = "25em",
					title.align = "left",
					title.color = NULL,
					label.align = "left",
					label.color = "#888888",
					value.align = "right",
					value.color = NULL,
					css = NULL) {
	structure(
		list(vars = vars,
			 title = title,
			 format = format,
			 layout = list(width = width,
			 			   max.height = max.height,
			 			   title.align = title.align,
			 			   title.color = title.color,
			 			   label.align = label.align,
			 			   label.color = label.color,
			 			   value.align = value.align,
			 			   value.color = value.color,
			 			   css = css)),
		class = c("tm_popup", "list")
	)
}

# Internal: default popup layout, and completion of a (possibly partial or
# NULL) layout list with these defaults. Used by view_format_popups() so that
# callers passing layout = NULL (e.g. extension layers, or the inset-frame and
# raster code paths) fall back to the documented defaults.
popup_layout_default = function() {
	list(width = "auto",
		 max.height = "25em",
		 title.align = "left",
		 title.color = NULL,
		 label.align = "left",
		 label.color = "#888888",
		 value.align = "right",
		 value.color = NULL,
		 css = NULL)
}

complete_popup_layout = function(x) {
	d = popup_layout_default()
	if (is.null(x) || !length(x)) {
		out = d
	} else {
		x = x[!vapply(x, is.null, logical(1))]
		out = utils::modifyList(d, x)
	}
	# Allow bare numbers as CSS lengths. A unit-less number is interpreted with
	# a sensible default unit per property: pixels for width (e.g. width = 300
	# -> "300px") and em for max.height (e.g. max.height = 5 -> "5em", i.e.
	# roughly "show 5 lines"). Character values are passed through unchanged, so
	# any CSS length ("300px", "20em", "50%", "auto") still works.
	out$width = .as_css_length(out$width, "px")
	out$max.height = .as_css_length(out$max.height, "em")
	out
}

.as_css_length = function(x, unit = "px") {
	# NA / non-finite means "no cap" (used by max.height to disable scrolling).
	if (length(x) == 1L && is.na(x)) return("none")
	if (is.numeric(x)) {
		if (!is.finite(x)) return("none")
		return(paste0(x, unit))
	}
	x
}

# Internal: build an inline style="" attribute from an optional alignment and
# color, omitting any component that is NULL/NA. Returns "" when neither is set.
# Shared by the popup formatters so the title/label/value cells are styled
# consistently.
.popup_style_attr = function(align = NULL, color = NULL) {
	parts = c(if (!is.null(align) && !is.na(align)) paste0("text-align:", align),
			  if (!is.null(color) && !is.na(color)) paste0("color:", color))
	if (!length(parts)) "" else paste0(" style=\"", paste(parts, collapse = "; "), ";\"")
}

#' @export
print.tm_popup = function(x, ...) {
	cli::cli_text("{.cls tm_popup} specification")
	invisible(x)
}

# Internal: resolve the popup specification of a layer function.
#
# It combines the new `popup` argument (a `tm_popup` object, or a bare value
# that is interpreted as `vars`) with the deprecated `popup.vars` /
# `popup.format` arguments, and returns a list with the three element fields
# that are stored on the tm_element: `vars`, `format`, and `title`.
#
# The `*.called` flags indicate whether the user explicitly supplied the
# corresponding argument (typically `"<arg>" %in% args_called`). They are used
# to decide (a) which source wins and (b) whether to emit a deprecation
# message. Because the convenience wrappers (e.g. `tm_dots()`, `tm_fill()`)
# forward these arguments via `...` rather than as their own formals, the
# `args_called` of the primitive layer function reliably reflects user intent.
process_popup = function(popup = tm_popup(),
						 popup.vars = NA,
						 popup.format = tm_label_format(),
						 popup.called = FALSE,
						 popup.vars.called = FALSE,
						 popup.format.called = FALSE,
						 layer_fun = NULL) {

	depr_called = popup.vars.called || popup.format.called

	if (popup.called) {
		# New API used.
		if (depr_called) message_popup_both(layer_fun)
		if (!inherits(popup, "tm_popup")) {
			# Permissive shorthand, e.g. popup = c("a", "b") or popup = FALSE.
			popup = tm_popup(vars = popup)
		}
	} else if (depr_called) {
		# Only the deprecated arguments were used.
		message_popup_deprecated(layer_fun)
		popup = tm_popup(vars = popup.vars, format = popup.format)
	}
	# else: keep the (default) `popup`, i.e. tm_popup().

	# Guard against a NULL format (e.g. tm_popup(format = NULL)).
	format = popup$format
	if (is.null(format)) format = tm_label_format()

	# Layout (scalar styling). It is present on tm_popup objects; for a bare
	# shorthand value it falls back to the defaults.
	layout = popup$layout
	if (is.null(layout)) layout = popup_layout_default()

	list(vars = popup$vars,
		 format = format,
		 title = popup$title,
		 layout = layout)
}

message_popup_deprecated = function(layer_fun = NULL) {
	fn = if (is.null(layer_fun)) "the layer function" else paste0(layer_fun, "()")
	cli::cli_inform(
		c("{.field [deprecated]} The arguments {.arg popup.vars} and {.arg popup.format} of {.code {fn}} are deprecated.",
		  "i" = "Use {.arg popup} with {.fn tm_popup} instead, e.g. {.code popup = tm_popup(vars = ..., format = ...)}."),
		.frequency = "regularly",
		.frequency_id = "popup-vars-deprecated"
	)
}

message_popup_both = function(layer_fun = NULL) {
	fn = if (is.null(layer_fun)) "the layer function" else paste0(layer_fun, "()")
	cli::cli_inform(
		c("{.field [popup]} Both {.arg popup} and the deprecated {.arg popup.vars}/{.arg popup.format} were supplied to {.code {fn}}.",
		  "i" = "The deprecated arguments are ignored in favour of {.arg popup}."),
		.frequency = "regularly",
		.frequency_id = "popup-both"
	)
}

#' Map layer: simple features
#'
#' Map layer that draws simple features as they are. Supported visual variables
#' are: `fill` (the fill color), `col` (the border color), `size` the point size,
#' `shape` the symbol shape, `lwd` (line width), `lty` (line type), `fill_alpha` (fill color alpha transparency)
#' and `col_alpha` (border color alpha transparency).
#'
#' @param fill,fill.scale,fill.legend,fill.free  `r .doc_vv("fill")`
#' @param col,col.scale,col.legend,col.free  `r .doc_vv("col")`
#' @param size,size.scale,size.legend,size.free  `r .doc_vv("size")`
#' @param shape,shape.scale,shape.legend,shape.free  `r .doc_vv("shape")`
#' @param lwd,lwd.scale,lwd.legend,lwd.free  `r .doc_vv("lwd")`
#' @param lty,lty.scale,lty.legend,lty.free  `r .doc_vv("lty")`
#' @param fill_alpha,fill_alpha.scale,fill_alpha.legend,fill_alpha.free  `r .doc_vv("fill_alpha")`
#' @param col_alpha,col_alpha.scale,col_alpha.legend,col_alpha.free  `r .doc_vv("col_alpha")`
#' @param linejoin,lineend line join and line end. See [gpar()][grid::gpar()] for details.
#' @param plot.order.list Specification in which order the spatial features are drawn.
#'   This consists of a list of three elementary geometry types: for polygons, lines and, points.
#'   For each of these types, which are drawn in that order, a [tm_plot_order()] is required.
#' @param options options passed on to the corresponding `opt_<layer_function>` function
#' @inherit tm_polygons details
#' @inheritParams tm_polygons
#' @param ... passed on to [tm_polygons()], [tm_lines()], and [tm_dots()]
#' @example ./examples/tm_sf.R
#' @export
tm_sf = function(fill = tm_const(),
                 fill.scale = tm_scale(),
                 fill.legend = tm_legend(),
                 fill.free = NA,
                 col = tm_const(),
                 col.scale = tm_scale(),
                 col.legend = tm_legend(),
                 col.free = NA,
                 size = tm_const(),
                 size.scale = tm_scale(),
                 size.legend = tm_legend(),
                 size.free = NA,
                 shape = tm_const(),
                 shape.scale = tm_scale(),
                 shape.legend = tm_legend(),
                 shape.free = NA,
                 lwd = tm_const(),
                 lwd.scale = tm_scale(),
                 lwd.legend = tm_legend(),
                 lwd.free = NA,
                 lty = tm_const(),
                 lty.scale = tm_scale(),
                 lty.legend = tm_legend(),
                 lty.free = NA,
                 fill_alpha = tm_const(),
                 fill_alpha.scale = tm_scale(),
                 fill_alpha.legend = tm_legend(),
                 fill_alpha.free = NA,
                 col_alpha = tm_const(),
                 col_alpha.scale = tm_scale(),
                 col_alpha.legend = tm_legend(),
                 col_alpha.free = NA,
                 linejoin = "round",
                 lineend = "round",
                 plot.order.list = list(
                 	polygons = tm_plot_order("AREA"), lines = tm_plot_order("LENGTH"), points = tm_plot_order("size")
                 	),
				 options = opt_tm_sf(),
                  zindex = NA,
                  group = NA,
                  group.control = "check",
                  ...) {
	args = list(...)
	args_called = names(rlang::call_match()[-1])

	#paste(paste(names(formals(tm_polygons)), names(formals(tm_polygons)), sep = " = "), collapse = ", ")
	tm = tm_polygons(
		fill = fill, fill.scale = fill.scale, fill.legend = fill.legend,
		fill.free = fill.free, col = col, col.scale = col.scale, col.legend = col.legend,
		col.free = col.free, lwd = lwd, lwd.scale = lwd.scale, lwd.legend = lwd.legend,
		lwd.free = lwd.free, lty = lty, lty.scale = lty.scale, lty.legend = lty.legend,
		lty.free = lty.free, fill_alpha = fill_alpha, fill_alpha.scale = fill_alpha.scale,
		fill_alpha.legend = fill_alpha.legend, fill_alpha.free = fill_alpha.free,
		col_alpha = col_alpha, col_alpha.scale = col_alpha.scale,
		col_alpha.legend = col_alpha.legend, col_alpha.free = col_alpha.free,
		linejoin = linejoin, lineend = lineend, plot.order = plot.order.list$polygons,
		options = options$polygons) +
	tm_lines(
		col = col, col.scale = col.scale, col.legend = col.legend,
		col.free = col.free, lwd = lwd, lwd.scale = lwd.scale,
		lwd.legend = lwd.legend, lwd.free = lwd.free, lty = lty, lty.scale = lty.scale,
		lty.legend = lty.legend, lty.free = lty.free, col_alpha = col_alpha,
		col_alpha.scale = col_alpha.scale, col_alpha.legend = col_alpha.legend,
		col_alpha.free = col_alpha.free, linejoin = linejoin, lineend = lineend,
		plot.order = plot.order.list$lines, options = options$lines)


	if ("shape" %in% args_called || "col" %in% args_called) {
		tm + tm_symbols(size = size, size.scale = size.scale, size.legend = size.legend,
						size.free = size.free, fill = fill, fill.scale = fill.scale,
						fill.legend = fill.legend, fill.free = fill.free, col = col,
						col.scale = col.scale, col.legend = col.legend, col.free = col.free,
						shape = shape, shape.scale = shape.scale, shape.legend = shape.legend,
						shape.free = shape.free, lwd = lwd, lwd.scale = lwd.scale, lwd.legend = lwd.legend,
						lwd.free = lwd.free, lty = lty, lty.scale = lty.scale, lty.legend = lty.legend,
						lty.free = lty.free, fill_alpha = fill_alpha, fill_alpha.scale = fill_alpha.scale,
						fill_alpha.legend = fill_alpha.legend, fill_alpha.free = fill_alpha.free,
						col_alpha = col_alpha, col_alpha.scale = col_alpha.scale,
						col_alpha.legend = col_alpha.legend, col_alpha.free = col_alpha.free,
						plot.order = plot.order.list$points, options = options$points)
	} else {
		tm + tm_dots(fill = fill, fill.scale = fill.scale, fill.legend = fill.legend,
					 fill.free = fill.free, size = size, size.scale = size.scale,
					 size.legend = size.legend, size.free = size.free, lwd = lwd,
					 lwd.scale = lwd.scale, lwd.legend = lwd.legend, lwd.free = lwd.free,
					 lty = lty, lty.scale = lty.scale, lty.legend = lty.legend, lty.free = lty.free,
					 fill_alpha = fill_alpha, fill_alpha.scale = fill_alpha.scale,
					 fill_alpha.legend = fill_alpha.legend, fill_alpha.free = fill_alpha.free,
					 plot.order = plot.order.list$points, options = options$points)
	}

}

#' @param points.icon.scale scaling number that determines how large the icons (or grobs) are in plot mode in comparison to proportional symbols (such as bubbles). For view mode, use the argument `grob.dim`
#' @param points.grob.dim vector of four values that determine how grob objects (see details) are shown in view mode. The first and second value are the width and height of the displayed icon. The third and fourth value are the width and height of the rendered png image that is used for the icon. Generally, the third and fourth value should be large enough to render a ggplot2 graphic successfully. Only needed for the view mode.
#' @param points.just justification of the points relative to the point coordinates. Either one of the following values: \code{"left"} , \code{"right"}, \code{"center"}, \code{"bottom"}, and \code{"top"}, or a vector of two values where first value specifies horizontal and the second value vertical justification. Besides the mentioned values, also numeric values between 0 and 1 can be used. 0 means left justification for the first value and bottom justification for the second value. Note that in view mode, only one value is used.
#' @rdname tm_sf
#' @inheritParams tm_symbols
#' @inheritParams tm_polygons
#' @inheritParams tm_lines
#' @export
opt_tm_sf = function(polygons.only = "yes",
					 lines.only = "yes",
					 points_only = "yes",
					 point_per = "feature",
					 points.icon.scale = 3,
					 points.just = NA,
					 points.grob.dim = c(width = 48, height = 48, render.width = 256, render.height = 256)) {

	list(points = list(trans.args = list(points_only = points_only, point_per = point_per, on_surface = FALSE, along_lines = FALSE),
					   mapping.args = list(icon.scale = points.icon.scale,
					   					just = points.just,
					   					grob.dim = points.grob.dim)),
		 lines = list(trans.args = list(lines.only = lines.only),
		 			 mapping.args = list()),
		 polygons = list(trans.args = list(polygons.only = polygons.only),
		 				mapping.args = list()))

}

#' @export
tmapReproject.sfc = function(shp, tmapID, bbox = NULL, ..., crs) {
	if (is.na(sf::st_crs(shp))) {
		shp2 = shp
	} else {
		res = transform_ortho(shp, crs, tmapID)
		shp2 = res$shp
		tmapID = res$tmapID
	}

	if (!is.null(bbox$x)) {
		if (identical(bbox$x, "FULL")) {
			bbox = list(x = full_bbox(crs))
		} else {
			bbox = list(x = do.call(tmaptools::bb, c(bbox, list(projection = crs))))
		}
	}
	shapeTM(shp = shp2, tmapID = tmapID, bbox = bbox, ...)
}

#' @export
tmapShape.sf = function(shp, is.main, crs, bbox, unit, filter, layer, shp_name, smeta, o, tmf) {
	if (identical(crs, "auto")) crs = auto_crs(shp, crs_extra = o$crs_extra, crs_global = o$crs_global) else crs = sf::st_crs(crs)
	reproj = (!is.null(crs) && !is.na(crs) && sf::st_crs(shp) != crs)

	if (reproj) {
		shp = transform_ortho(shp, crs)
	}

	sfc = sf::st_geometry(shp)

	if (inherits(sfc[[1]], c("XYZ", "XYM", "XYZM"))) {
		sfc = sf::st_zm(sfc)
	}

	if (o$check_and_fix) sfc = check_fix(sfc, shp_name, reproj, o$show.messages)

	# if check_fix fails, is_valid contains the valid ids
	isv = attr(sfc, "is_valid")
	if (!is.null(isv)) {
		shp = shp[isv, ]
	}
	df = sf::st_drop_geometry(shp)

	# Encode expression columns (e.g. created with parse()) as character strings
	# so they survive the data.table-based data pipeline. They are decoded again
	# at plot time (tm_text / tm_labels) to render as plotmath expressions.
	is_expr = vapply(df, is.expression, FUN.VALUE = logical(1))
	if (any(is_expr)) {
		df[is_expr] = lapply(df[is_expr], encode_expr)
	}

	dt = data.table::as.data.table(df)

	dtcols = copy(names(dt))


	if (is.null(filter)) filter = rep_len(TRUE, nrow(dt))
	dt[, ':='(tmapID__ = 1L:.N, sel__ = filter)]

	make_by_vars(dt, tmf, smeta)

	shpTM = shapeTM(shp = sfc, tmapID = 1L:(length(sfc)), bbox = bbox)

	structure(list(shpTM = shpTM, dt = dt, is.main = is.main, dtcols = dtcols, shpclass = "sfc", bbox = bbox, unit = unit, shp_name = shp_name, smeta = smeta, type_ids = NULL, type_vars = NULL), class = "tmapShape")
}

#' @export
tmapSubsetShp.sf = function(shp, vars) {

	if ("AREA_" %in% vars && !("AREA_" %in% names(shp))) {
		shp$AREA_ = sf::st_area(shp)
	}
	if ("LENGTH_" %in% vars && !("LENGTH_" %in% names(shp))) {
		shp$LENGTH_ = sf::st_length(shp)

	}
	if ("MAP_COLORS" %in% vars) {
		shp$MAP_COLORS = as.factor(tmaptools::map_coloring(shp, ncols = 7))
	}

	if (!length(vars)) {
		vars = "dummy__"
		shp$dummy__ = logical(nrow(shp))
	}
	shp[, vars]
}


#' @export
tmapSubsetShp.sfc = function(shp, vars) {
	s = sf::st_sf(dummy__ = TRUE, geometry = shp)
	if ("AREA" %in% vars) {
		s$AREA_ = sf::st_area(shp)
	}
	if ("LENGTH" %in% vars) {
		s$LENGTH_ = sf::st_length(shp)
	}
	if ("MAP_COLORS" %in% vars) {
		s$MAP_COLORS = as.factor(tmaptools::map_coloring(shp, ncols = 7))
	}
	s
}



#' @export
tmapGetShapeMeta2.sf = function(shp, smeta, o) {
	vars = setdiff(names(shp), attr(shp, "sf_column"))
	smeta$vars_levs = lapply(seq_along(vars), function(i) {
		get_fact_levels_na(shp[[i]], o)
	})
	names(smeta$vars_levs) = vars
	smeta
}


#' @export
tmapGetShapeMeta1.sf = function(shp, layer, o) {
	vars = setdiff(names(shp), attr(shp, "sf_column"))
	names(vars) = vars
	#vars_levs = lapply(vars, function(v) {get_fact_levels_na(shp[[v]], o)})
	dims = character(0)
	dims_vals = list()

	list(vars = vars,
		 dims = dims,
		 dims_vals = dims_vals)
}


#' @export
tmapGetShapeMeta2.sfc = function(shp, smeta, o) {
	vars = character(0)
	smeta$vars_levs = list()
	smeta
}


#' @export
tmapGetShapeMeta1.sfc = function(shp, layer, o) {
	vars = character(0)
	dims = character(0)
	dims_vals = list()

	list(vars = vars,
		 dims = dims,
		 dims_vals = dims_vals)
}

#' Thematic Map Visualization
#'
#' Thematic maps are geographical maps in which spatial data distributions are visualized.
#' This package offers a flexible, layer-based, and easy to use approach to create
#' thematic maps, such as choropleths and bubble maps. It is based on the grammar
#' of graphics, and resembles the syntax of ggplot2.
#'
#' @author Martijn Tennekes \email{mtennekes@@gmail.com}
#' @seealso \href{https://r-tmap.github.io/tmap/index.html}{Main documentation}
#' @concept GIS
#' @concept thematic maps
#' @concept statistical maps
#' @concept choropleth
#' @concept bubble map
#' @references Tennekes, M., 2018, {tmap}: Thematic Maps in {R}, Journal of Statistical Software, 84(6), 1-39, \doi{10.18637/jss.v084.i06}
"_PACKAGE"

gridCell = function(rows, cols, e) {
	vp = grid::viewport(layout.pos.row = rows, layout.pos.col = cols)
	grid::grobTree(e, vp = vp)
}




process_comp_box = function(comp, sc, o) {
	comp = within(comp, {
		if ("frame.color" %in% comp$called) {
			frame = TRUE
			frame.lwd = frame.lwd * sc
		} else {
			if (!is.logical(frame)) {
				cls = class(comp)[1]
				if (valid_colors(frame)) {
					cli::cli_warn("{.field {.fun {cls}}} use {.code frame.color = {.str {frame}}} instead of {.code frame = {.str {frame}}}")
					frame.color = frame
					frame = TRUE
				} else {
					cli::cli_abort("{.field {.fun {cls}}} {.arg frame} should be a logical")
				}

			}
			frame.color = if (!frame) NA else if (is.null(frame.color)) o$attr.color else frame.color
			frame.lwd = if (!frame) 0 else frame.lwd * sc
		}
		frame.r = frame.r * sc
	})
	comp
}

tmapGridComp2 = function(grp, comp, o, stack, pos.h, pos.v, maxH, maxW, offsetIn.h, offsetIn.v, marginInH, marginInV, are_nums, fH, fW) {
	n = length(comp)
	legWin = vapply(comp, "[[", FUN.VALUE = numeric(1), "Win")   #  vapply(comp, leg_standard$fun_width, FUN.VALUE = numeric(1), o = o)
	legHin = vapply(comp, "[[", FUN.VALUE = numeric(1), "Hin")#vapply(comp, leg_standard$fun_height, FUN.VALUE = numeric(1), o = o)

	group.just = c(pos.h, pos.v)
	frame_combine = grp$frame_combine

	legWin[is.infinite(legWin)] = maxW
	legHin[is.infinite(legHin)] = maxH

	scaleW = legWin / maxW
	scaleH = legHin / maxH

	# because of legend frames (for which margins are added in tmapGridComp), the scale may be a bit above 1, even though automatic layout is applied and there is enough space

	if (o$component.autoscale) {
		if (any(scaleW > 1.1) || any(scaleH > 1.1)) message_comp_scale()
	} else {
		scaleW[scaleW > 1.1] = 1
		scaleH[scaleH > 1.1] = 1
	}


	clipW = pmax(1, scaleW)
	clipH = pmax(1, scaleH)
	if (grp$resize_as_group) {
		clipT = rep(max(clipW, clipH), n)
	} else {
		clipT = pmax(clipW, clipH)
	}

	legWin = legWin / clipT
	legHin = legHin / clipT

	if (frame_combine) {
		if (stack == "vertical") {
			legWin = rep(max(legWin), n)
		} else {
			legHin = rep(max(legHin), n)
		}
	}

	# rescale due to stacking
	if (stack == "vertical") {
		scaleS = sum(legHin) / maxH
	} else {
		scaleS = sum(legWin) / maxW
	}

	if (o$component.autoscale) {
		if (scaleS > 1.01) {
			if (scaleS > 1.05) {
				message_comp_high_wide(stack)

			}
			legWin = legWin / scaleS
			legHin = legHin / scaleS
			clipT = clipT * scaleS
		}
	}

	legW = grid::unit(legWin, "inches")
	legH = grid::unit(legHin, "inches")

	if (stack == "vertical") {
		W = max(legWin)
		H = sum(legHin)

		Hs = unit_add_between(legHin, marginInV)
		Ws = W
	} else {
		W = sum(legWin)
		H = max(legHin)

		Ws = unit_add_between(legWin, marginInH)
		Hs = H
	}





	if (are_nums) {
		group.just = as.numeric(group.just)
		just.h = grp$position$just.h
		just.v = grp$position$just.v

		# add dummy offsets
		Hs = unit_add_sides(Hs, 0)
		Ws = unit_add_sides(Ws, 0)

		ancher_x = maxW * group.just[1]
		if (just.h == "left") {
			Ws = unit_add_sides(Ws, c(ancher_x, maxW - ancher_x- sum(Ws)))
		} else if (just.h == "right") {
			Ws = unit_add_sides(Ws, c(ancher_x - sum(Ws), maxW - ancher_x))
		} else {
			Ws = unit_add_sides(Ws, c(ancher_x - (sum(Ws) / 2), maxW - ancher_x- (sum(Ws) / 2)))
		}

		ancher_y = maxH * (1 - group.just[2])
		if (just.v == "top") {
			Hs = unit_add_sides(Hs, c(ancher_y, maxH - ancher_y - sum(Hs)))
		} else if (just.v == "bottom") {
			Hs = unit_add_sides(Hs, c(ancher_y - sum(Hs), maxH - ancher_y))
		} else {
			Hs = unit_add_sides(Hs, c(ancher_y - (sum(Hs) / 2), maxH - ancher_y - (sum(Hs) / 2)))
		}

	} else {
		nullsH = switch(group.just[1], left = c(0, 1), center = c(0.5, 0.5), right = c(1, 0))
		nullsV = switch(group.just[2], top = c(0, 1), center = c(0.5, 0.5), bottom = c(1, 0))

		Hs = unit_add_sides(Hs, offsetIn.v)
		Hs = unit_add_sides(Hs, nullsV)

		Ws = unit_add_sides(Ws, offsetIn.h)
		Ws = unit_add_sides(Ws, nullsH)

	}

	Hn = length(Hs)
	Wn = length(Ws)

	Hid = seq(3, Hn - 2, by = 2)
	Wid = seq(3, Wn - 2, by = 2)

	if (are_nums) {
		Wu = rep("inch", Wn)
		Hu = rep("inch", Hn)
	} else {
		Wu = c("null", rep("inch", Wn-2), "null")
		Hu = c("null", rep("inch", Hn-2), "null")
	}

	# po(Ws, Hs)
	# Ws = distr_space_over_nulls(Ws, maxW)
	# Hs = distr_space_over_nulls(Hs, maxH)
	# po(Ws, Hs)

	vp = grid::viewport(layout = grid::grid.layout(ncol = Wn,
												   nrow = Hn,
												   widths = grid::unit(Ws, Wu),
												   heights = grid::unit(Hs, Hu)))


	comp = mapply(function(leg, scale, W, H) {
		leg$scale = scale

		if ("flexCol" %in% names(leg)) leg$wsu = distr_space_over_nulls(u = leg$wsu * scale, tot = W, stretchID = leg$flexCol)
		if ("flexRow" %in% names(leg)) leg$hsu = distr_space_over_nulls(u = leg$hsu * scale, tot = H, stretchID = leg$flexRow)
		leg
	}, comp, 1/clipT, legW, legH, SIMPLIFY = FALSE, USE.NAMES = FALSE)


	legGrobs = lapply(comp, tmapGridCompPlot, o = o, fH = fH, fW = fW)


	#sq = function(x) do.call(seq, as.list(unname(range(x))))
	sc = min(1/clipT) * o$scale

	comp = lapply(comp, process_comp_box, sc = sc, o = o)
	grp = process_comp_box(grp, sc, o)

	design = getOption("tmap.design.mode")

	draw_rect = grp$frame || grp$bg
	if (draw_rect && frame_combine) {
		if (grp$frame) {
			groupframe = gridCell(range(Hid), range(Wid), rndrectGrob(gp=grid::gpar(fill = NA, col = grp$frame.color, lwd = grp$frame.lwd), r = grp$frame.r))
		} else {
			groupframe = NULL
		}

		if (grp$bg && !design) {
			groupbg = gridCell(range(Hid), range(Wid), rndrectGrob(gp=grid::gpar(fill = grp$bg.color, alpha = grp$bg.alpha, col = NA, lwd = 0), r = grp$frame.r))
		} else {
			groupbg = NULL
		}
	} else {
		groupbg = NULL
		groupframe = NULL
	}

	equalize = grp$equalize


	if (equalize) {
		if (stack == "horizontal") {
			legFH = grid::unit(1, "npc")
			legFW = legW
		} else {
			legFH = legH
			legFW = grid::unit(1, "npc")
		}
	} else {
		legFH = legH
		legFW = legW
	}


	grbs = do.call(grid::gList, mapply(function(leg, lG, lH, lW, fH, fW, iW, iH) {
		frame = if (draw_rect && !frame_combine) {
			bg.color = if (leg$bg && !design) leg$bg.color else NA
			frame.color = if (leg$frame) leg$frame.color else NA
			rndrectGrob(gp=grid::gpar(fill = bg.color, alpha = leg$bg.alpha, col = frame.color, lwd = leg$frame.lwd), r = leg$frame.r)
		} else NULL
		if (stack == "vertical") {
			x = switch(leg$position$align.h, "left" = lW/2, "right" = grid::unit(W, "inch") -lW/2, grid::unit(0.5, "npc"))
			y = grid::unit(0.5, "npc")
		} else {
			x = grid::unit(0.5, "npc")
			y = switch(leg$position$align.v, "top" = grid::unit(H, "inch")-lH/2, "bottom" = lH/2, grid::unit(0.5, "npc"))
		}
		gridCell(iH, iW, {
			grid::gList(
				if (equalize) {
					grid::grobTree(frame, vp = grid::viewport(width = fW, height = fH))
				} else {
					grid::grobTree(frame, vp = grid::viewport(x = x, y = y, width = fW, height = fH))
				},
				grid::grobTree(lG, vp = grid::viewport(x = x, width = lW, y = y, height = lH)))
		})
	}, comp, legGrobs, legH, legW, legFH, legFW, Wid, Hid, SIMPLIFY = FALSE))


	if (design) {
		df = expand.grid(col = 1:ncol(vp$layout),
						 row = 1:nrow(vp$layout))

		grDesign = lapply(1:nrow(df), function(i) gridCell(df$row[i], df$col[i], grid::rectGrob(gp=gpar(fill=NA,col="#000000", lwd=2))))
	} else {
		grDesign = NULL
	}

	do.call(grid::grobTree, c(list(groupbg), grbs, list(groupframe), grDesign, list(vp=vp)))
}

#' @export
#' @rdname tmap_internal
tmapGetCompGroupArgs = function(comp) {
	# get component group settings
	nms = setdiff(names(formals(tm_components)), "group_id")
	grp = comp[[1]][nms]

	any_legend_chart_inset = any(vapply(comp, inherits, FUN.VALUE = logical(1), c("tm_legend", "tm_chart")))
	grp_called = setdiff(unique(do.call(c, lapply(comp, FUN = "[[", "called_via_comp_group"))), "group_id")

	if (!("frame" %in% grp_called)) {
		# update because of #1214
		frms = vapply(comp, function(ci) ci$frame, FUN.VALUE = logical(1))
		if (any(frms) && (any(!frms))) grp$frame = any_legend_chart_inset
	}
	if (!("bg" %in%grp_called)) grp$bg = any_legend_chart_inset
	grp$called = comp[[1]]$called
	grp
}


tmapGridComp = function(comp, o, facet_row = NULL, facet_col = NULL, facet_page, class, stack, stack_auto, pos.h, pos.v, bbox) {

	grp = tmapGetCompGroupArgs(comp)


	gts = get("gts", envir = .TMAP_GRID)
	g = get("g", envir = .TMAP_GRID)

	gt = gts[[facet_page]]
	n = length(comp)
	# rows and cols in main grid
	rows = if (facet_row[1] == -2) g$meta_rows[1] else if (facet_row[1] == -1) g$meta_rows[2] else {
		panel_rows = if (class == "in") {
			NULL
		} else if (o$panel.type == "xtab") {
		    g$rows_panel_row_ids[facet_row]
		} else if (o$panel.type == "wrap") {
			g$rows_panel_ids[facet_row]
		} else {
			NULL
		}
		c(g$rows_facet_ids[facet_row], panel_rows)
	}
	cols = if (facet_col[1] == -2) g$meta_cols[1] else if (facet_col[1] == -1) g$meta_cols[2] else {
		panel_cols = if (class == "in") {
			NULL
		} else if (o$panel.type == "xtab") {
			g$cols_panel_col_ids[facet_col]
		} else if (o$panel.type == "wrap") {
			g$cols_panel_ids[facet_col]
		} else {
			NULL
		}
		c(g$cols_facet_ids[facet_col], panel_cols)
	}


	# make rows and cols a range
	rows = seq(min(rows), max(rows))
	cols = seq(min(cols), max(cols))


	# multiple not needed anymore
	stack = stack[1] # check what happens when stack_auto is T, e.g. facet_stack
	pos.h = pos.h[1]
	pos.v = pos.v[1]



	rowsIn = g$rowsIn[rows]
	colsIn = g$colsIn[cols]
	if (pos.h %in% c("LEFT", "RIGHT")) {
		pos.h = tolower(pos.h)
		CASE.h = toupper
	} else {
		if (pos.h %in% c("CENTER", "CENTRE")) pos.h = "center"
		CASE.h = function(x)x
	}
	if (pos.v %in% c("TOP", "BOTTOM")) {
		pos.v = tolower(pos.v)
		CASE.v = toupper
	} else {
		if (pos.v %in% c("CENTER", "CENTRE")) pos.v = "center"
		CASE.v = function(x)x
	}

	are_nums = !(any(is.na(suppressWarnings(as.numeric(c(pos.h, pos.v))))))

	if (are_nums) {
		component.offset.h = 0
		component.offset.v = 0
	} else {
		offset = grp$offset
		if (!is.null(names(offset)) && all(c("inside", "INSIDE", "outside", "OUTSIDE") %in% names(offset))) {
			component.offset.h = get_option_class(offset, class = CASE.h(paste0(class, "side")), spatial_class = FALSE)
			component.offset.v = get_option_class(offset, class = CASE.v(paste0(class, "side")), spatial_class = FALSE)
		} else {
			offset = rep_len(offset, 2L)
			component.offset.h = offset[1]
			component.offset.v = offset[2]

		}
	}

	offsetIn.h = component.offset.h * o$lin# + (o$frame.lwd * o$scale / 144) # 1 line = 1/72 inch, frame lines are centered (so /2)
	offsetIn.v = component.offset.v * o$lin#+ (o$frame.lwd * o$scale / 144)

	stack_margin = grp$stack_margin
	if (!is.null(names(stack_margin)) && all(c("combined", "apart") %in% names(stack_margin))) {
		stack_margin = rep_len(get_option_class(stack_margin, class = ifelse(comp[[1]]$frame_combine, "combined", "apart"), spatial_class = FALSE), 2L)
	} else {
		stack_margin = rep_len(stack_margin, 2L)
	}

	marginInH = stack_margin[1] * o$lin
	marginInV = stack_margin[2] * o$lin

	marginInTotH = (n - 1L) * marginInH
	marginInTotV = (n - 1L) * marginInV
	offsetInTot.h  = 2 * offsetIn.h
	offsetInTot.v  = 2 * offsetIn.v

	totH = sum(rowsIn) - offsetInTot.v - marginInTotH
	totW = sum(colsIn) - offsetInTot.h - marginInTotV

	## update scale_bar width: identified by the WnativeID = 3 item (null for other components)
	comp = mapply(function(cmp, bb) {
		bbw = bb[3] - bb[1]

		# TRUE if component is scalebar
		if (!is.null(cmp$WnativeID)) {
			# find out whether the bounding of borrowed from the map (yes if scale bar is drawn in another facet)
			bbox_nb = attr(bbox, "borrow")
			if (is.null(bbox_nb)) {
				bb_facet = sum(colsIn)
			} else {
				bb_facet = sum(g$colsIn[g$cols_facet_ids[bbox_nb$col]])
			}

			oldIn = as.numeric(cmp$wsu[cmp$WnativeID])

			if (is.null(cmp$WnativeRange) || ("width" %in% cmp$call)) {
				# in case breaks not defined: allow the width to be maximal fraction of facet width
				#newIn = min(totW, bb_facet * cmp$width)
				newIn = min(totW, cmp$Win)
			} else {
				# in case breaks are defined: compute width
				Wextra = (text_width_inch(paste0("   ", cmp$units$unit)) + text_width_inch(paste0(tail(cmp$breaks, 1), cmp$breaks[1])) / 2 ) * cmp$text.size

				# bbw are number of (CRS) units of map width, totalCoords translated to scalebar units
				totalCoords =  bbw * cmp$units$to

				Wcomp = bb_facet * (cmp$WnativeRange / totalCoords) + Wextra

				newIn = min(totW, Wcomp)
			}

			margins =  sum(as.numeric(cmp$wsu[c(2,4)]))

			newIn_without_margins = newIn - margins

			cmp$wsu[cmp$WnativeID] = unit(newIn_without_margins, "inch")
			cmp$Win = newIn_without_margins + margins

			# get cpi: coordinates per inch

			cmp$dst = distances_bbox_sides(bb)

			cmp$which_side_of_map = if (class == "in" && pos.v == "bottom") {
				"bottom"
			} else if (class == "in" && pos.v == "top") {
				"top"
			} else if (class == "in" && pos.v == "center") {
				"mid"
			} else {
				"bottom"
			}

			#cmp$cpi = unname(bbw / bb_facet)
			cmp$bb_facet = bb_facet
			#po(bb, bbw, bbox_nb, oldIn, bb_facet, newIn, totW, cmp$Win, margins, newIn_without_margins, cmp$Win)
			#po(cmp$cpi)
		}
		cmp
	}, comp, bbox, SIMPLIFY = FALSE)

	legWin = vapply(comp, "[[", FUN.VALUE = numeric(1), "Win")   #  vapply(comp, leg_standard$fun_width, FUN.VALUE = numeric(1), o = o)
	legHin = vapply(comp, "[[", FUN.VALUE = numeric(1), "Hin")#vapply(comp, leg_standard$fun_height, FUN.VALUE = numeric(1), o = o)


	# get total value (width or height)
	getH = function(s, lH) {
		if (!length(s)) return(NULL)
		if (s[1] == "vertical") {
			sum(unit_add_between(lH, marginInV))
		} else {
			max(lH)
		}
	}
	getW = function(s, lW) {
		if (!length(s)) return(NULL)
		if (s[1] == "vertical") {
			max(lW)
		} else {
			sum(unit_add_between(lW, marginInH))
		}
	}

	grbs = tmapGridComp2(grp = grp, comp = comp, o = o, stack = stack, pos.h = pos.h, pos.v = pos.v, maxH = totH, maxW = totW, offsetIn.h = offsetIn.h, offsetIn.v = offsetIn.v, marginInH = marginInH, marginInV = marginInV, are_nums = are_nums, fH = totH, fW = totW)#sum(rowsIn), fW = sum(colsIn))

	gt = add_to_gt(gt, grbs, row = rows, col = cols)

	gts[[facet_page]] = gt

	assign("gts", gts, envir = .TMAP_GRID)

}

getdts = function(aes, unm, p, q, o, dt, shpvars, layer, group, glid, mfun, args, plot.order) {



	dev = getOption("tmap.devel.mode")

	nm = aes$aes
	nm__ord = paste0(nm, "__ord")


	#nm and unm are both visual variables
	#nm is the 'prototype', for which methods are written (tmapScale_defaults.R)
	#unm is the name known by the user, so used in messaging and also as identifier in data

	# glid = group-layer id (a unique name across all layers)
	unm__ord = paste0(unm, "__ord")

	# should the results of the data (needed for the plotting function)?
	# sorting order will be plotting order
	# -1 for NULL features (filtered out, or dropped units)
	# 0 for NA features
	# 1-n for features based on scale values (n for latest=plotted on top)
	# 1 for features with non-NA value for not-selected aes
	sortRev = if (plot.order$aes == "DATA") NULL else if (plot.order$aes == unm) plot.order$reverse else NA


	bypass_ord = plot.order$aes == "DATA"

	val = aes$value
	if (q$limitvars) val = val[1]

	with(q, {
		get_num_facets = function(bys) {
			k = as.integer(substr(bys, 3, 3))
			o$fn[k]
		}

		nvars = length(val) #m
		nvari = vapply(val, length, integer(1))

		vars = unlist(val, recursive = FALSE)

		# active grouping variables (to keep)
		grp_bv = by123__[sort(c({if (nvars > 1) v else integer(0)}, b))]

		sfun = paste0("tmapValuesScale_", nm)
		cfun = paste0("tmapValuesColorize_", nm)

		if (inherits(val, "tmapUsrCls")) {
			temp = local({
				# not data driven variable: visual variable to support tm_vars(multivariate = TRUE) of other visual variable (e.g. fill of donut maps)
				k = length(aes$value[[1]]$x)
				vls = unname(aes$value[[1]]$x)
				dtl = data.table::data.table(tmapID__ = 1L:k, sel = TRUE, value = factor(vls, levels = vls))


				varname = paste(unm, 1L, sep = "_")
				legname = paste("legnr", 1L, sep = "_")
				crtname = paste("crtnr", 1L, sep = "_")

				l = update_l(o = o, l = aes$legend, v = "value", mfun = mfun, unm = unm, active = TRUE)
				crt = update_crt(o = o, crt = aes$chart, v = "value", mfun = mfun, unm = unm, active = TRUE)

				s = aes$scale
				._f = s$FUN
				s$FUN = NULL
				# update label.format
				s$label.format = process_label_format(s$label.format, o$label.format)

				if (all(is.ena(l$title))) l$title = ""
				l$glid = glid
				l$layer = layer


				arglist = list(scale = s, legend = l, chart = crt,
							   o = o, aes = nm,
							   layer = layer,
							   layer_args = args,
							   sortRev = NA,
							   bypass_ord = TRUE,
							   submit_legend = TRUE)
				dtl[, c(varname, legname, crtname) := do.call(._f, c(unname(.SD), arglist)), .SDcols = "value"]

				list(val = paste(dtl[[varname]], collapse = "__"),
					 legnr = dtl$legnr_1[1],
					 ctrnr = dtl$crtnr_1[1])
			})
			nvars = 1
			nvarsi = 1
			vars = temp$val
			grp_bv = by123__[sort(b)]

			nm = "skip"

			sfun = paste0("tmapValuesScale_", nm)
			cfun = paste0("tmapValuesColorize_", nm)
			aes$data_vars = FALSE
		} else {
			temp = NULL
		}

		if (!aes$data_vars && !aes$geo_vars && !aes$bypass_vars) {
			#cat("step2_grp_lyr_aes_const", unm," \n")
			# constant values (take first value (of possible multivariate per facet)
			if (any(nvari) > 1) warning("Mapping values considered as direct visual variables, which cannot be used with multivariate variables", call. = FALSE)
			val1 = sapply(vars, "[[", 1, USE.NAMES = FALSE)

			if (!aes$bypass_vars) {
				check_fun = paste0("tmapValuesCheck_", nm)
				check = do.call(check_fun, list(x = val1, is_var = TRUE))
				if (!check) {
					# to do: add "layer" name e.g. tm_fill is still "polygons" and not "fill"
					info = attr(check, "info")

					fn_call <-  call(paste0("tm_", layer[1]))
					cli::cli_abort(c(
						"Visual values used for the variable {.val {unm}} are incorrect.",
						i = info),
						call = fn_call
					)
				}

				val1 = do.call(sfun, list(x = val1, scale = o$scale))
				val1 = do.call(cfun, list(x = val1, pc = o$pc))
			} else {
# 				sc = aes$scale
#
#
# 				if (sc$FUN != "tmapScaleCategorical") {
# 					cli::cli_inform("A variable has been specified for {.val {unm}}. Please use {.fun tm_scale_continuous} and specify at least the argument {.arg levels}. Other scales are not supported for remote spatial data yet.")
# 				} else {
#
# 					asfd = tmapScaleCategorical(NA, scale = sc, legend = aes$legend, chart = aes$chart, o = o, aes = aes$aes, layer = layer, layer_args = args, sortRev = sortRev, bypass_ord = bypass_ord)
#
# browser()
# 					sc = get_scale_defaults(sc, o, aes$aes, layer, cls = "fact", ct = NULL)
# 				}
#
# 				scale_num = scale_save(sc)
# 				val1 = paste0("scale", sprintf("%04d", scale_num), ":", val1)
			}


			dtl = copy(dt[, c("tmapID__", "sel__", by123__[b]), with = FALSE])

			if (nvars > 1 && limitvars) {
				# not allowed: take first one
				warning(limitvars_warn, call. = FALSE)
				val1 = val1[1]
				nvars = 1L
			}
			vnames = paste0("value", sprintf("%02d", 1L:nvars))
			for (i in 1L:nvars) {
				vname = vnames[i]
				dtl[, (vname) := val1[i]]
			}

			if (length(v)) update_fl(k = v, m = nvars)

			if (nvars > 1) {
				dtl = melt(dtl, id.vars = c("tmapID__", "sel__", by123__[b]), measure.vars = vnames, variable.name = var__, value.name = unm)
				dtl[, (var__) := as.integer(get(..var__))]
			} else {
				setnames(dtl, vnames[1], unm)
			}

			# impute null (filter argument of tm_shape) with value.null
			if (!bypass_ord) dtl[, (nm__ord) := 1L]

			if (any(!dtl$sel__) || !q$drop.units) {
				# also needed for drop.units later on
				cls = data_class(dtl[[unm]])
				value.null = getAesOption("value.null", o, nm, layer, cls = cls)
				value.null = do.call(sfun, list(x = value.null, scale = o$scale))
				value.null = do.call(cfun, list(x = value.null, pc = o$pc))

				# todo: combine these:
				dtl[sel__==FALSE, (unm) := value.null]
				if (!bypass_ord) dtl[sel__==FALSE, (nm__ord) := -1L]

				if (!q$drop.units) {

					if (bypass_ord) {
						imp = structure(list(value.null, FALSE), names = c(unm, "sel__"))
					} else {
						imp = structure(list(value.null, -1L, FALSE), names = c(unm, nm__ord, "sel__"))
					}
					levs = lapply(get_num_facets(grp_bv), seq.int, from = 1)
					names(levs) = grp_bv
					dtl = completeDT2(dtl, cols = c(list("tmapID__" = unique(dtl$tmapID__)), levs), defs = imp)
				}

			}

			dtl[, legnr := vector("integer", length = nrow(dtl))]
			dtl[, crtnr := vector("integer", length = nrow(dtl))]

			if (!is.null(temp)) {
				dtl_leg = data.table::data.table(sel = TRUE, legnr = temp$legnr, crtnr = temp$ctrnr)
			} else {
				dtl_leg = dtl[, .SD[1], by = c(grp_bv)][, tmapID__ := NULL][, legnr := (vapply(get(..unm), function(s) legend_save(list(mfun = mfun, unm = unm, active = FALSE, vneutral = s)), FUN.VALUE = integer(1)))][, crtnr := (vapply(get(..unm), function(s) chart_save(list()), FUN.VALUE = integer(1)))][, (unm) := NULL]
			}

		} else {
			#cat("step2_grp_lyr_aes_var", nm," \n")

			if (aes$bypass_vars) {
				if (any(nvari) > 1) warning("Multiple data variables defined for one map variable, but for remote spatial data sources, only one (first) will be used", call. = FALSE)
				val1 = sapply(vars, "[[", 1, USE.NAMES = FALSE)
				#
				# sc = aes$scale
				# scale_num = scale_save(sc)
				# val1 = paste0("scale", sprintf("%04d", scale_num), ":", val1)
				dt[[val1]] = NA
				vars = val1
				val = val1
			}

			relevant_vars = c("tmapID__", "sel__" , vars, by123__[b])
			dtl = copy(dt[, relevant_vars, with = FALSE])

			# impute free argument.
			fr = rep(aes$free, length.out = 3)
			if (anyNA(fr)) {
				if (is.null(o$free.scales)) {
					# set to FALSE by default, and for the vars dimension to TRUE.
					fr = rep(FALSE, 3)
					if (length(v)) fr[v] = TRUE
				} else {
					# v3 backward compatibility: free.scales is set via tm_facets
					fr = rep(o$free.scales, length.out = 3)
				}
			}

			# group by variables with free scales
			grp_b_fr = by123__[intersect(which(fr), b)]
			grp_bv_fr = by123__[sort(c({if (nvars > 1 && fr[v]) v else integer(0)}, intersect(which(fr), b)))]
			#grp_bv_fr = by123__[intersect(which(fr), b)]

			if (length(v) && fr[v] && !all(nvari == nvari[1])) stop("number of variables per map variable should be consistent when free = FALSE", call. = FALSE)

			# multiple variables
			if (nvars > 1) {
				if (limitvars) {
					#cat("step2_grp_lyr_aes_var_limiter\n")
					# not allowed: take first one
					warning(limitvars_warn, call. = FALSE)
					val = val[[1]]
					val_orig = val_orig[[1]]
				} else {
					if (!fr[v]) {
						#cat("step2_grp_lyr_aes_var_multi_vars_!free_scale\n")

						# multiple variables, !free scale => stack all variable columns in long format
						nms = val[[1]]

						dtlks = lapply(1L:nvari[1], function(k) {
							vk = vapply(val, "[", character(1), k)
							melt(dtl, id.vars = c("tmapID__", by123__[b]), measure.vars = vk, variable.name = by123__[v], value.name = nms[k])
						})
						dtl = dtlks[[1]]
						if (nvari[1] > 1) {
							for (i in 2L:nvari[1]) {
								dtl[, (nms[i]) := dtlks[[i]][[nms[i]]]]
							}
						}

						for (col in by123__[v]) {
							dtl[, (col) := as.integer(get(..col))]
						}

						vars = vapply(val, "[[", character(1), 1)
						val = val[[1]]
						names(val) = val[1]
					} else {
						#cat("step2_grp_lyr_aes_var_multi_vars_free_scale\n")
						vars = vapply(val, "[[", character(1), 1)
						val = lapply(val, function(vl) {names(vl) = vl[1]; vl})
					}
				}
			} else {
				#cat("step2_grp_lyr_aes_var_one_var\n")
				val_name = names(val)[1]
				val = val[[1]]
				names(val) = val_name
				vars = vars[1] # only needed for update_fl?
			}
			if (length(v)) update_fl(k = v, lev = vars)

			apply_scale = function(s, l, crt, v, varname, ordname, legname, crtname, sortRev, bypass_ord, bypass_vars = FALSE) {
				l = update_l(o = o, l = l, v = v, mfun = mfun, unm = unm, active = TRUE)
				l$glid = glid
				l$layer = layer
				l$group = group
				crt = update_crt(o = o, crt = crt, v = v, mfun = mfun, unm = unm, active = TRUE)

				if (length(s) == 0) cli::cli_abort("mapping not implemented for visual variable {unm}")
				._f = s$FUN
				s$FUN = NULL
				# update label.format
				format_called = (length(s$label.format) > 0L)
				s$label.format = process_label_format(s$label.format, o$label.format)
				if (format_called) {
					.TMAP$popup.format[[v]] = s$label.format
				}

				cls = data_class(dtl[[v[1]]])
				#if (is.na(s$legend$title)) s$legend$title = v

				# for bivariate legends
				if (length(v) > 1L) {
					cls2 = data_class(dtl[[v[2]]])
					if (is.ena(l$xlab)) l$xlab = paste0(v[2], attr(cls, "units"))
					if (is.ena(l$ylab)) l$ylab = paste0(v[1], attr(cls2, "units"))
					if (is.ena(l$title)) l$title = ""
				} else {
					if (length(l$title) > 1) {
						warning("This probably shouldn't happen. Please check code.")
					}
					# Error in if (is.ena(l$title)) l$title = paste0(names(v), attr(cls, "units")) :
					# the condition has length > 1
					# Calls: <Anonymous> ... with -> with.default -> eval -> eval -> apply_scale
					# Detected in types of titles. # Example to illustrate the type of titles
					unit = if (!is.null(s$unit)) s$unit else ""
					if (unit != "") unit = paste0(" (", unit, ")")

					if (all(is.ena(l$title))) l$title = paste0(names(v), attr(cls, "units"), unit)
				}

				if (._f != "tmapScaleAuto") {
					# number of variables needed
					fnames = names(formals(._f))
					fnvar =  which(fnames == "scale") - 1L
					if (fnames[1] != "...") {
						if (fnvar > length(v)) {
							stop("Too few variables defined")
						} else if (fnvar < length(v)) {
							warning("Too many variables defined")
							v = v[1L:fnvar]
						}
					}
				}


				#aesname = aes$aes

				value.null = if ("value.null" %in% names(s) && !is.na(s$value.null)) s$value.null else {
					vn = getAesOption("value.null", o, nm, layer, cls = cls)
					vn = do.call(sfun, list(x = vn, scale = o$scale))
					do.call(cfun, list(x = vn, pc = o$pc))
				}

				arglist = list(scale = s, legend = l, chart = crt, o = o, aes = nm,
							   layer = layer,
							   layer_args = args,
							   sortRev = sortRev,
							   bypass_ord = bypass_ord,
							   submit_legend = TRUE)

				if (!all(dtl$sel__)) {
					if (bypass_ord) {
						dtl[, c(varname, legname, crtname) := list(value.null, 0L, 0L)]
					} else {
						dtl[, c(varname, ordname, legname, crtname) := list(value.null, -1L, 0L, 0L)]
					}

					if (is.na(value.null)) cli::cli_abort("{.code value.null} not implemented for visual variable {unm}")
					if (bypass_ord) {
						dtl[sel__ == TRUE, c(varname, legname, crtname) := do.call(._f, c(unname(.SD), arglist)), grp_b_fr, .SDcols = v]
					} else {
						dtl[sel__ == TRUE, c(varname, ordname, legname, crtname) := do.call(._f, c(unname(.SD), arglist)), grp_b_fr, .SDcols = v]
					}
				} else {
					if (bypass_ord) {
						dtl[, c(varname, legname, crtname) := do.call(._f, c(unname(.SD), arglist)), grp_b_fr, .SDcols = v]
					} else {
						dtl[, c(varname, ordname, legname, crtname) := do.call(._f, c(unname(.SD), arglist)), grp_b_fr, .SDcols = v]
					}
				}
				if (!q$drop.units) {
					imp = structure(list(value.null, 0L, 0L, TRUE), names = c(unm, legname, crtname, "sel__"))
					levs = lapply(get_num_facets(grp_bv), seq.int, from = 1)
					names(levs) = grp_bv

					# leg- and crt numbers are assigned to the first data record. If that one is removed because of drop.NA.facets, this number should be assigned to another data record (#932)
					if (q$drop.NA.facets) {
						legnr_non0 = (dtl[[legname]] != 0L)
						if (sum(legnr_non0) == 1L) legnr = dtl[[legname]][which(legnr_non0)] else legnr = NA
						crtnr_non0 = (dtl[[crtname]] != 0)
						if (sum(crtnr_non0) == 1L) crtnr = dtl[[crtname]][which(crtnr_non0)] else crtnr = NA
					} else {
						legnr = NA
						crtnr = NA
					}
					dtl = completeDT2(dtl, cols = c(list("tmapID__" = unique(dtl$tmapID__)), levs), defs = imp)
					if (!is.na(legnr) && all(dtl[[legname]] == 0)) dtl[[legname]][1] = legnr
					if (!is.na(crtnr) && all(dtl[[crtname]] == 0)) dtl[[crtname]][1] = crtnr
				}
				if (bypass_vars) {
					dtl[[varname]] = paste0("scale", sprintf("%03d", dtl$legnr), ":", v)
				}

				dtl
			}

			if (length(v) && fr[v] && nvars > 1) {
				#cat("step2_grp_lyr_aes_var_multi_aes_columns\n")
				# apply aes function for each var column
				if (inherits(aes$scale, "tm_scale")) {
					scale = rep(list(aes$scale), length.out = nvars)
				} else if (islistof(aes$scale, "tm_scale")) {
					scale = rep(aes$scale, length.out = nvars)
				} else {
					cli::cli_abort("incorrect scale specification")
				}

				if (inherits(aes$legend, "tm_legend")) {
					legend = rep(list(aes$legend), length.out = nvars)
				} else if (islistof(aes$legend, "tm_legend")) {
					legend = rep(aes$legend, length.out = nvars)
				} else {
					cli::cli_abort("incorrect legend specification")
				}

				if (inherits(aes$chart, "tm_chart")) {
					crt = rep(list(aes$chart), length.out = nvars)
				} else if (islistof(aes$chart, "tm_chart")) {
					crt = rep(aes$chart, length.out = nvars)
				} else {
					cli::cli_abort("incorrect chart specification")
				}


				varnames = paste(unm, 1L:nvars, sep = "_")
				ordnames = paste(unm__ord, 1L:nvars, sep = "_")
				legnames = paste("legnr", 1L:nvars, sep = "_")
				crtnames = paste("crtnr", 1L:nvars, sep = "_")
				for (i in 1L:nvars) {
					dtl = apply_scale(scale[[i]], legend[[i]], crt[[i]], val[[i]], varnames[[i]], ordnames[[i]], legnames[[i]], crtnames[[i]], sortRev = sortRev, bypass_ord = bypass_ord)
				}

				dtl_leg = melt(dtl, id.vars = c("tmapID__", by__), measure.vars = legnames, variable.name = var__, value.name = "legnr")
				dtl_crt = melt(dtl, id.vars = c("tmapID__", by__), measure.vars = crtnames, variable.name = var__, value.name = "crtnr")
				if (!bypass_ord) dtl_ord = melt(dtl, id.vars = c("tmapID__", by__), measure.vars = ordnames, variable.name = var__, value.name = nm__ord)
				dtl = melt(dtl, id.vars = c("tmapID__", by__), measure.vars = varnames, variable.name = var__, value.name = unm)
				if (!bypass_ord) dtl[, (nm__ord) := dtl_ord[[nm__ord]]]

				dtl[, (var__) := as.integer(get(..var__))]
				dtl_leg[, (var__) := as.integer(get(..var__))]
				dtl_crt[, (var__) := as.integer(get(..var__))]

				#sel = !vapply(dtl_leg$legend, is.null, logical(1))
				dtl_leg = dtl_leg[legnr != 0, c(grp_bv_fr, "legnr"), with = FALSE]
				dtl_crt = dtl_crt[crtnr != 0, c(grp_bv_fr, "crtnr"), with = FALSE]

			} else {
				#cat("step2_grp_lyr_aes_var_one_aes_column\n")

				# apply aes function to the (only) var column
				if (inherits(aes$scale, "tm_scale")) {
					s = aes$scale
				} else if (islistof(aes$scale, "tm_scale")) {
					s = aes$scale[[1]]
				} else {
					cli::cli_abort("incorrect scale specification")
				}

				if (length(s) == 0) cli::cli_abort("mapping not implemented for visual variable {unm}")

				if (inherits(aes$legend, "tm_legend")) {
					l = aes$legend
				} else if (islistof(aes$legend, "tm_legend")) {
					cli::cli_warn("multiple legends are specified, while only one is required; the first will be used")
					l = aes$legend[[1]]
				} else {
					cli::cli_abort("incorrect legend specification")
				}

				if (inherits(aes$chart, "tm_chart")) {
					crt = aes$chart
				} else if (islistof(aes$chart, "tm_chart")) {
					cli::cli_warn("multiple charts are specified, while only one is required; the first will be used")
					crt = aes$chart[[1]]
				} else {
					cli::cli_abort("incorrect chart specification")
				}

				dtl = apply_scale(s, l, crt, val, unm, nm__ord, "legnr", "crtnr", sortRev, bypass_ord, aes$bypass_vars)

				dtl_leg = dtl[legnr != 0L, c(grp_bv_fr, "legnr", "crtnr"), with = FALSE]
			}
		}
		if (dev) timing_add(s4 = paste0("aes ", aes$aes))

		if (bypass_ord) {
			list(dt = dtl[, c("tmapID__", grp_bv, unm), with = FALSE],
				 leg = dtl_leg)
		} else {
			list(dt = dtl[, c("tmapID__", grp_bv, unm, nm__ord), with = FALSE],
				 leg = dtl_leg)
		}
	})
}

tmapGridShape = function(bbx, facet_row, facet_col, facet_page, o) {
	gts = get("gts", envir = .TMAP_GRID)
	g = get("g", envir = .TMAP_GRID)

	design = getOption("tmap.design.mode")


	fbbx = bb_asp(bbx, g$fasp)

	rc_text = frc(facet_row, facet_col)

	rowid = g$rows_facet_ids[facet_row]
	colid = g$cols_facet_ids[facet_col]

	bg = if (o$earth_boundary) {
		if (!o$space || design) o$space.color = NA
		eotw = end_of_the_world(o$crs_step4, earth_datum = o$earth_datum)
		grid::gList(rndrectGrob(gp=gpar(fill=o$space.color, lwd=NA, lineend="square"), r = o$frame.r * o$scale, name = "inner_rect"),
					sf::st_as_grob(eotw, gp = grid::gpar(fill = o$bg.color, lwd = NA), name = "inner_world"))

	} else if (design) {
		NULL
	} else {
		rndrectGrob(gp=gpar(fill=o$bg.color, lwd=NA, lineend="square"), r = o$frame.r * o$scale, name = "inner_rect")
	}

	clip = if (is.na(o$bg.color) && o$earth_boundary && !o$space_overlay) {
		grid::as.path(bg)
	} else TRUE

	gtmap = grid::grobTree(bg,
						   vp = grid::vpStack(grid::viewport(layout.pos.col = colid, layout.pos.row = rowid, name = paste0("vp_facet_", rc_text)),
						   				   grid::viewport(xscale = fbbx[c(1,3)], yscale = fbbx[c(2,4)], name = paste0("vp_map_", rc_text), clip = clip)), name = paste0("gt_facet_", rc_text))

	gts[[facet_page]] = grid::addGrob(gts[[facet_page]], gtmap, gPath = grid::gPath("gt_main"))

	assign("gts", gts, envir = .TMAP_GRID)
	NULL
}


tmapGridOverlay = function(bbx, facet_row, facet_col, facet_page, o) {
	gts = get("gts", envir = .TMAP_GRID)
	g = get("g", envir = .TMAP_GRID)

	gt = gts[[facet_page]]

	rc_text = frc(facet_row, facet_col)
	frame.lwd = if (!o$frame) 0 else o$frame.lwd * o$scale
	r = o$frame.r * o$scale

	boundary = if (o$earth_boundary) {
		fbbx = bb_asp(bbx, g$fasp)

		eotw = end_of_the_world(o$crs_step4, earth_datum = o$earth_datum)

		sf::st_as_grob(eotw, gp = grid::gpar(fill = NA, col = o$earth_boundary.color, lwd = o$earth_boundary.lwd), name = "outer_world", vp = grid::viewport(xscale = fbbx[c(1,3)], yscale = fbbx[c(2,4)]))
	} else NULL

	gframe = if (o$frame) {
		pH = convertHeight(unit(1, "points"), unitTo = "npc", valueOnly = TRUE)*frame.lwd
		pW = convertWidth(unit(1, "points"), unitTo = "npc", valueOnly = TRUE)*frame.lwd

		if (o$frame.double_line) {
			bgcol = substr(o$bg.color, 1, 7) # remove transparency
			grid::grobTree(
				rndrectGrob(width = 1-4*pW, height=1-4*pH, gp=gpar(col=bgcol, fill=NA, lwd=4*frame.lwd, lineend="square"), r = r, name = "outer_frame"),
				rndrectGrob(gp=gpar(col=o$frame.color, fill=NA, lwd=1.5*frame.lwd, lineend="square", alpha = o$frame.alpha), r = r, name = "between_frame"),
				rndrectGrob(width = 1-8*pW, height=1-8*pH, gp=gpar(col=o$frame.color, fill=NA, lwd=frame.lwd, lineend="square", alpha = o$frame.alpha), r = r, name = "inner_frame"))
		} else {
			rndrectGrob(width = 1, height = 1, gp=gpar(col=o$frame.color, fill=NA, lwd=frame.lwd, lineend="square", alpha = o$frame.alpha), r = r, name = "outer_frame")
			#rectGrob(width = 1-0.5*pW, height = 1-0.5*pH, gp=gpar(col=frame.col, fill=NA, lwd=frame.lwd, lineend="square"), name = "outer_frame")
		}

	} else NULL



	rowid = g$rows_facet_ids[facet_row]
	colid = g$cols_facet_ids[facet_col]

	if (o$earth_boundary && o$space_overlay) {
		space_fill = local({
			if (o$frame) {
				gframe_path = roundrect_to_pathGrob(gframe)
			} else {
				gframe_path = viewport_frame_path()
			}
			boundary_npc = local({
				x = as.vector(boundary$x)
				x2 = (x - boundary$vp$xscale[1]) / diff(boundary$vp$xscale)
				y = as.vector(boundary$y)
				y2 = (y - boundary$vp$yscale[1]) / diff(boundary$vp$yscale)
				boundary$vp = NULL
				boundary$x = unit(x2, "npc")
				boundary$y = unit(y2, "npc")
				boundary
			})
			combine_paths(gframe_path, boundary_npc, fill = o$space.color)
		})
		gt = gt |>
			grid::addGrob(space_fill, paste0("gt_facet_", rc_text))
	}

	if (o$earth_boundary) {
		gt = gt |>
			grid::addGrob(boundary, paste0("gt_facet_", rc_text))
	}

	gt = gt |>
		add_to_gt(gframe, row = rowid, col = colid)
	gts[[facet_page]] = gt

	assign("gts", gts, envir = .TMAP_GRID)
	NULL

}



## functions to create a grob that draws space 'space'
## needed because clipping by earth boundary does not always work (e.g. not for rasterGrobs #1170)

combine_paths = function(pg1, pg2, rule = "winding", fill = "black") {
	# Combine coordinates and assign IDs
	combined_x = unit.c(pg1$x, pg2$x)
	combined_y = unit.c(pg1$y, pg2$y)
	combined_id = c(rep(1, length(pg1$x)), rep(2, length(pg2$x)))

	# Return compound path
	pathGrob(
		x = combined_x,
		y = combined_y,
		id = combined_id,
		rule = rule,
		gp = gpar(fill = fill, col = NA)
	)
}

roundrect_to_pathGrob = function(roundrect, n_corner = 10, fill = "black", col = NA, rule = "winding") {
	stopifnot(inherits(roundrect, "roundrect"))

	# Extract and convert parameters to NPC coordinates
	x0 = convertX(roundrect$x - roundrect$width / 2, "npc", valueOnly = TRUE)
	x1 = convertX(roundrect$x + roundrect$width / 2, "npc", valueOnly = TRUE)
	y0 = convertY(roundrect$y - roundrect$height / 2, "npc", valueOnly = TRUE)
	y1 = convertY(roundrect$y + roundrect$height / 2, "npc", valueOnly = TRUE)
	rx = convertWidth(roundrect$r, "npc", valueOnly = TRUE)
	ry = convertHeight(roundrect$r, "npc", valueOnly = TRUE)

	# Clamp corner radius to avoid overlaps
	rx = min(rx, (x1 - x0) / 2)
	ry = min(ry, (y1 - y0) / 2)

	# Helper: ellipse arc between angles
	arc = function(cx, cy, rx, ry, start, end, n = 10) {
		theta = seq(start, end, length.out = n)
		list(
			x = cx + rx * cos(theta),
			y = cy + ry * sin(theta)
		)
	}

	# Define corner arcs (clockwise from upper-left)
	ul = arc(x0 + rx, y1 - ry, rx, ry, pi / 2, pi, n_corner)
	ur = arc(x1 - rx, y1 - ry, rx, ry, 0, pi / 2, n_corner)
	lr = arc(x1 - rx, y0 + ry, rx, ry, -pi / 2, 0, n_corner)
	ll = arc(x0 + rx, y0 + ry, rx, ry, -pi, -pi / 2, n_corner)

	# Assemble path: arcs and straight edges between them
	x_coords = c(
		rev(ul$x),
		ur$x[n_corner],
		rev(ur$x),
		lr$x[n_corner],
		rev(lr$x),
		ll$x[n_corner],
		rev(ll$x),
		ul$x[n_corner]
	)

	y_coords = c(
		rev(ul$y),
		ur$y[n_corner],
		rev(ur$y),
		lr$y[n_corner],
		rev(lr$y),
		ll$y[n_corner],
		rev(ll$y),
		ul$y[n_corner]
	)

	# Return as filled path grob
	pathGrob(
		x = unit(x_coords, "npc"),
		y = unit(y_coords, "npc"),
		id = rep(1, length(x_coords)),
		rule = rule,
		gp = gpar(fill = fill, col = col)
	)
}


viewport_frame_path = function(fill = "black", col = NA, rule = "winding") {
	# Clockwise path around the viewport (0,0) to (1,1)
	x = unit(rev(c(0, 1, 1, 0, 0)), "npc")
	y = unit(rev(c(0, 0, 1, 1, 0)), "npc")

	pathGrob(
		x = x,
		y = y,
		id = rep(1, length(x)),
		rule = rule,
		gp = gpar(fill = fill, col = col)
	)
}

#' @export
#' @rdname tmapGridLeaflet
tmapLeafletDataPlot.tm_data_text = function(a, shpTM, dt, pdt, popup.format, hdt, idt, gp, bbx, facet_row, facet_col, facet_page, id, pane, group, glid, o, ...) {
	args = list(...)

	lf = get_lf(facet_row, facet_col, facet_page)

	rc_text = frc(facet_row, facet_col)

	res = select_sf(shpTM, dt)
	shp = res$shp
	dt = res$dt

	#shadow = gp$shadow

	gp = impute_gp(gp, dt)
	gp = rescale_gp(gp, o$scale_down)

	coords = sf::st_coordinates(shp)

	# in case shp is a multipoint (point_per == "segment"), expand gp:
	cp = expand_coords_gp(coords, gp, ndt = nrow(dt))
	coords = cp$coords
	gp = cp$gp

	if (cp$expanded) {
		shpTM_match = match(shpTM$tmapID_expanded, shpTM$tmapID)
	} else {
		shpTM_match = TRUE
	}

	text = decode_expr_chr(as.character(dt$text[shpTM_match]))
	if (all(text == "")) return(NULL)

	idt = dt$tmapID__[shpTM_match] |>
		submit_labels("text", pane, group)

	opt = leaflet::pathOptions(interactive = FALSE, pane = pane)


	clusterOpts = getClusterOpts(a$clustering)

	cidt = if (!is.null(clusterOpts)) submit_labels(idt[1], "cluster", pane, group) else NULL


	cex_set = unique(gp$cex)
	alpha_set = unique(gp$col_alpha)
	face_set = unique(gp$fontface)
	col_set = unique(gp$col)

	bgcol_set = unique(gp$bgcol)
	bgcol_alpha = unique(gp$bgcol_alpha)

	if (all(bgcol_set == "#00000000")) {
		bg_enabled = FALSE
	} else {
		bgres = split_alpha_channel(gp$bgcol, gp$bgcol_alpha)
		gp$bgcol = bgres$col
		gp$bgcol_alpha = bgres$opacity
		bg_enabled = TRUE
	}



	if (length(face_set) != 1) message("Variable fontfaces not supported by view mode")

	vary = (length(cex_set) != 1) || (length(alpha_set) != 1) || (length(face_set) != 1) || (length(col_set) != 1) || (length(bgcol_set) != 1) || (length(bgcol_alpha) != 1)

	sizeChar <- paste(round(gp$cex * 12), "px", sep="")

	# direction <- ifelse(gpl$text.just == "left", "right",
	# 			 ifelse(gpl$text.just == "right", "left",
	# 			 ifelse(gpl$text.just == "top", "bottom",
	# 			 ifelse(gpl$text.just == "bottom", "top", "center"))))
	direction = "right"


	# apply xmod and ymod
	delta = delta_per_lineheight(bbx)

	coords[,1] = coords[,1] + delta * gp$cex * gp$xmod
	coords[,2] = coords[,2] + delta * gp$cex * gp$ymod

	scale_px = function(textlines, cex) {
		ceiling(textlines * cex * 12)
	}


	make_shadow = function(cex, textcol) {
		if (a$halo) {
			if (is.na(a$halo.col)) a$halo.col = ifelse(is_light(textcol), "#000000DD", "#FFFFFFDD")
			make_halo_css(col = a$halo.col, mode = "halo", width = scale_px(a$halo.width, cex), blur = scale_px(a$halo.blur, cex), diag_scale = sqrt(2)/2, alpha = a$halo.alpha)
		} else if (a$shadow) {
			if (is.na(a$shadow.col)) a$shadow.col = ifelse(is_light(textcol), "#000000", "#FFFFFF")
			make_halo_css(col = a$shadow.col, mode = "shadow", offset.x = scale_px(a$shadow.offset.x, cex), offset.y = scale_px(a$shadow.offset.y, cex), alpha = 1)
		} else {
			NULL
		}
	}

	if (!vary) {
		lf = lf %>% addLabelOnlyMarkers(lng = coords[, 1], lat = coords[,2],
										label=text,
										group=group,
										layerId = idt,
										labelOptions = labelOptions(noHide = TRUE,
																	textOnly = TRUE,
																	pane = pane,
																	direction = direction,
																	opacity=gp$col_alpha[1],
																	textsize=sizeChar[1],
																	style=list("color"=gp$col[1],
																			   "background-color" = if (bg_enabled) paste0("rgba(", paste(col2rgb(gp$bgcol[1]), collapse = ", "), ",", gp$bgcol_alpha[1], ")") else NULL,
																			  "border" = if (a$bg.border) paste0(a$bg.border.lwd, "px solid rgba(", paste(col2rgb(a$bg.border.col[1]), collapse = ", "), ",", gp$bgcol_alpha[1], ")") else NULL,
																			  "line-height" = sizeChar[1],
																			  "text-shadow" = make_shadow(gp$cex[1], gp$col[1]),
																			  "padding" = paste0(scale_px(a$bg.padding / 2, gp$cex[1]), "px"))),
										options = markerOptions(pane = pane),
										clusterOptions = clusterOpts,
										clusterId = cidt)
	} else {
		for (i in 1:length(text)) {
			lf = lf %>% addLabelOnlyMarkers(lng = coords[i,1], lat = coords[i,2],
											label=text[i],
											group=group,
											layerId = idt[i],
											labelOptions = labelOptions(noHide = TRUE,
																		textOnly = TRUE,
																		pane = pane,
																		direction = direction,
																		opacity=gp$col_alpha[i],
																		textsize=sizeChar[i],
																		style=list("color"=gp$col[i],
																				   "background-color" = if (bg_enabled) paste0("rgba(", paste(col2rgb(gp$bgcol[i]), collapse = ", "), ",", gp$bgcol_alpha[i], ")") else NULL,
																				   "border" = if (a$bg.border) paste0(a$bg.border.lwd, "px solid rgba(", paste(col2rgb(a$bg.border.col[i]), collapse = ", "), ",", gp$bgcol_alpha[i], ")") else NULL,
																				   "line-height" = sizeChar[i],
																				   "text-shadow" = make_shadow(gp$cex[i], gp$col[i]),
																				   "padding" = paste0(scale_px(a$bg.padding / 2, gp$cex[i]), "px"))),
											options = markerOptions(pane = pane),
											clusterOptions = clusterOpts,
											clusterId = cidt)
		}
	}
	assign_lf(lf, facet_row, facet_col, facet_page)



	NULL
}


#' @export
#' @rdname tmapGridLeaflet
tmapLeafletDataPlot.tm_data_labels = function(a, shpTM, dt, pdt, popup.format, hdt, idt, gp, bbx, facet_row, facet_col, facet_page, id, pane, group, glid, o, ...) {
	NextMethod()
}

#' @export
#' @rdname tmapGridLeaflet
tmapLeafletDataPlot.tm_data_labels_highlighted = function(a, shpTM, dt, pdt, popup.format, hdt, idt, gp, bbx, facet_row, facet_col, facet_page, id, pane, group, glid, o, ...) {
	NextMethod()
}




make_halo_css <- function(
		col = "white",
		alpha = 0.9,
		width = 1,
		blur = 0,
		diagonals = TRUE,
		diag_scale = 1,
		mode = c("halo", "shadow"),
		offset.x = 2,
		offset.y = 2
) {

	mode <- match.arg(mode)

	# convert color to rgba
	rgb <- grDevices::col2rgb(col)
	col_css <- sprintf(
		"rgba(%d,%d,%d,%s)",
		rgb[1], rgb[2], rgb[3], alpha
	)

	# --- Simple drop shadow mode ---
	if (mode == "shadow") {
		return(sprintf(
			"%spx %spx %spx %s",
			offset.x, offset.y, blur, col_css
		))
	}

	# --- Halo mode ---
	# Cardinal directions
	offsets <- rbind(
		c(-1, 0),
		c( 1, 0),
		c( 0,-1),
		c( 0, 1)
	)

	# Optional diagonals
	if (diagonals) {
		diag_offsets <- rbind(
			c(-1,-1),
			c( 1,-1),
			c(-1, 1),
			c( 1, 1)
		)

		# allow smaller diagonal distance for rounder halo
		diag_offsets <- diag_offsets * diag_scale

		offsets <- rbind(offsets, diag_offsets)
	}

	# Scale by width
	offsets <- round(offsets * width, 2)

	parts <- apply(offsets, 1, function(x) {
		sprintf("%spx %spx 0 %s", x[1], x[2], col_css)
	})

	# Optional glow
	if (blur > 0) {
		parts <- c(parts, sprintf("0 0 %spx %s", blur, col_css))
	}

	paste(parts, collapse = ", ")
}

#' Map layer: simple features
#'
#' Map layer that draws simple features as they are. Supported visual variables
#' are: `fill` (the fill color), `col` (the border color), `size` the point size,
#' `shape` the symbol shape, `lwd` (line width), `lty` (line type), `fill_alpha` (fill color alpha transparency)
#' and `col_alpha` (border color alpha transparency).
#'
#' @param fill,fill.scale,fill.legend,fill.free  `r .doc_vv("fill")` `r .doc_unit("fill")`
#' @param col,col.scale,col.legend,col.free  `r .doc_vv("col")` `r .doc_unit("col")`
#' @param size,size.scale,size.legend,size.free  `r .doc_vv("size")` `r .doc_unit("size")`
#' @param shape,shape.scale,shape.legend,shape.free  `r .doc_vv("shape")` `r .doc_unit("shape")`
#' @param lwd,lwd.scale,lwd.legend,lwd.free  `r .doc_vv("lwd")` `r .doc_unit("lwd")`
#' @param lty,lty.scale,lty.legend,lty.free  `r .doc_vv("lty")` `r .doc_unit("lty")`
#' @param fill_alpha,fill_alpha.scale,fill_alpha.legend,fill_alpha.free  `r .doc_vv("fill_alpha")` `r .doc_unit("fill_alpha")`
#' @param col_alpha,col_alpha.scale,col_alpha.legend,col_alpha.free  `r .doc_vv("col_alpha")` `r .doc_unit("col_alpha")`
#' @param linejoin,lineend line join and line end. See [gpar()][grid::gpar()] for details.
#' @param plot.order.list Specification in which order the spatial features are drawn.
#'   This consists of a list of three elementary geometry types: for polygons, lines and, points.
#'   For each of these types, which are drawn in that order, a [tm_plot_order()] is required.
#' @param options options passed on to the corresponding `opt_<layer_function>` function
#' @param blend Compositing operator for layer blending, applied to each
#'   sublayer (polygons, lines, points). Default `"over"` applies no blending.
#'   See the "Layer blending" section of [tm_polygons()] for supported values.
#' @inherit tm_polygons details
#' @inheritParams tm_polygons
#' @param ... passed on to [tm_polygons()], [tm_lines()], and [tm_dots()]
#' @example ./examples/tm_sf.R
#' @export
tm_sf = function(fill = tm_const(),
                 fill.scale = tm_scale(),
                 fill.legend = tm_legend(),
                 fill.free = NA,
                 col = tm_const(),
                 col.scale = tm_scale(),
                 col.legend = tm_legend(),
                 col.free = NA,
                 size = tm_const(),
                 size.scale = tm_scale(),
                 size.legend = tm_legend(),
                 size.free = NA,
                 shape = tm_const(),
                 shape.scale = tm_scale(),
                 shape.legend = tm_legend(),
                 shape.free = NA,
                 lwd = tm_const(),
                 lwd.scale = tm_scale(),
                 lwd.legend = tm_legend(),
                 lwd.free = NA,
                 lty = tm_const(),
                 lty.scale = tm_scale(),
                 lty.legend = tm_legend(),
                 lty.free = NA,
                 fill_alpha = tm_const(),
                 fill_alpha.scale = tm_scale(),
                 fill_alpha.legend = tm_legend(),
                 fill_alpha.free = NA,
                 col_alpha = tm_const(),
                 col_alpha.scale = tm_scale(),
                 col_alpha.legend = tm_legend(),
                 col_alpha.free = NA,
                 linejoin = "round",
                 lineend = "round",
                 plot.order.list = list(
                 	polygons = tm_plot_order("AREA"), lines = tm_plot_order("LENGTH"), points = tm_plot_order("size")
                 	),
				 options = opt_tm_sf(),
                  zindex = NA,
                  group = NA,
                  group.control = "check",
                  blend = "over",
                  ...) {
	args = list(...)
	args_called = names(rlang::call_match()[-1])

	# Arguments forwarded to every sublayer: the interactive arguments
	# (popup, popup.vars, popup.format, hover, id) supplied via `...`, plus
	# blend. Other `...` arguments retain their previous behaviour (they are
	# not forwarded to the sublayers).
	interactive_nms = c("popup", "popup.vars", "popup.format", "hover", "id")
	common = c(args[intersect(names(args), interactive_nms)], list(blend = blend))

	#paste(paste(names(formals(tm_polygons)), names(formals(tm_polygons)), sep = " = "), collapse = ", ")
	tm = do.call(tm_polygons, c(list(
		fill = fill, fill.scale = fill.scale, fill.legend = fill.legend,
		fill.free = fill.free, col = col, col.scale = col.scale, col.legend = col.legend,
		col.free = col.free, lwd = lwd, lwd.scale = lwd.scale, lwd.legend = lwd.legend,
		lwd.free = lwd.free, lty = lty, lty.scale = lty.scale, lty.legend = lty.legend,
		lty.free = lty.free, fill_alpha = fill_alpha, fill_alpha.scale = fill_alpha.scale,
		fill_alpha.legend = fill_alpha.legend, fill_alpha.free = fill_alpha.free,
		col_alpha = col_alpha, col_alpha.scale = col_alpha.scale,
		col_alpha.legend = col_alpha.legend, col_alpha.free = col_alpha.free,
		linejoin = linejoin, lineend = lineend, plot.order = plot.order.list$polygons,
		options = options$polygons), common)) +
	do.call(tm_lines, c(list(
		col = col, col.scale = col.scale, col.legend = col.legend,
		col.free = col.free, lwd = lwd, lwd.scale = lwd.scale,
		lwd.legend = lwd.legend, lwd.free = lwd.free, lty = lty, lty.scale = lty.scale,
		lty.legend = lty.legend, lty.free = lty.free, col_alpha = col_alpha,
		col_alpha.scale = col_alpha.scale, col_alpha.legend = col_alpha.legend,
		col_alpha.free = col_alpha.free, linejoin = linejoin, lineend = lineend,
		plot.order = plot.order.list$lines, options = options$lines), common))


	if ("shape" %in% args_called || "col" %in% args_called) {
		tm + do.call(tm_symbols, c(list(size = size, size.scale = size.scale, size.legend = size.legend,
						size.free = size.free, fill = fill, fill.scale = fill.scale,
						fill.legend = fill.legend, fill.free = fill.free, col = col,
						col.scale = col.scale, col.legend = col.legend, col.free = col.free,
						shape = shape, shape.scale = shape.scale, shape.legend = shape.legend,
						shape.free = shape.free, lwd = lwd, lwd.scale = lwd.scale, lwd.legend = lwd.legend,
						lwd.free = lwd.free, lty = lty, lty.scale = lty.scale, lty.legend = lty.legend,
						lty.free = lty.free, fill_alpha = fill_alpha, fill_alpha.scale = fill_alpha.scale,
						fill_alpha.legend = fill_alpha.legend, fill_alpha.free = fill_alpha.free,
						col_alpha = col_alpha, col_alpha.scale = col_alpha.scale,
						col_alpha.legend = col_alpha.legend, col_alpha.free = col_alpha.free,
						plot.order = plot.order.list$points, options = options$points), common))
	} else {
		tm + do.call(tm_dots, c(list(fill = fill, fill.scale = fill.scale, fill.legend = fill.legend,
					 fill.free = fill.free, size = size, size.scale = size.scale,
					 size.legend = size.legend, size.free = size.free, lwd = lwd,
					 lwd.scale = lwd.scale, lwd.legend = lwd.legend, lwd.free = lwd.free,
					 lty = lty, lty.scale = lty.scale, lty.legend = lty.legend, lty.free = lty.free,
					 fill_alpha = fill_alpha, fill_alpha.scale = fill_alpha.scale,
					 fill_alpha.legend = fill_alpha.legend, fill_alpha.free = fill_alpha.free,
					 plot.order = plot.order.list$points, options = options$points), common))
	}

}

#' @param points.icon.scale scaling number that determines how large the icons (or grobs) are in plot mode in comparison to proportional symbols (such as bubbles). For view mode, use the argument `grob.dim`
#' @param points.grob.dim vector of four values that determine how grob objects (see details) are shown in view mode. The first and second value are the width and height of the displayed icon. The third and fourth value are the width and height of the rendered png image that is used for the icon. Generally, the third and fourth value should be large enough to render a ggplot2 graphic successfully. Only needed for the view mode.
#' @param points.just justification of the points relative to the point coordinates. Either one of the following values: \code{"left"} , \code{"right"}, \code{"center"}, \code{"bottom"}, and \code{"top"}, or a vector of two values where first value specifies horizontal and the second value vertical justification. Besides the mentioned values, also numeric values between 0 and 1 can be used. 0 means left justification for the first value and bottom justification for the second value. Note that in view mode, only one value is used.
#' @rdname tm_sf
#' @inheritParams tm_symbols
#' @inheritParams tm_polygons
#' @inheritParams tm_lines
#' @export
opt_tm_sf = function(polygons.only = "yes",
					 lines.only = "yes",
					 points_only = "yes",
					 point_per = "feature",
					 points.icon.scale = 3,
					 points.just = NA,
					 points.grob.dim = c(width = 48, height = 48, render.width = 256, render.height = 256)) {

	list(points = list(trans.args = list(points_only = points_only, point_per = point_per, on_surface = FALSE, along_lines = FALSE),
					   mapping.args = list(icon.scale = points.icon.scale,
					   					just = points.just,
					   					grob.dim = points.grob.dim)),
		 lines = list(trans.args = list(lines.only = lines.only),
		 			 mapping.args = list()),
		 polygons = list(trans.args = list(polygons.only = polygons.only),
		 				mapping.args = list()))

}

#' @export
#' @keywords internal
#' @rdname tmapGridLeaflet
tmapLeafletCompPrepare = function(comp, o) {
	UseMethod("tmapLeafletCompPrepare")
}

tmapLeafletCompHeight = function(comp, o) {
	UseMethod("tmapLeafletCompHeight")
}

tmapLeafletCompWidth = function(comp, o) {
	UseMethod("tmapLeafletCompWidth")
}

tmapLeafletCompPlot = function(comp, lf, o) {
	UseMethod("tmapLeafletCompPlot")
}

#' @export
#' @keywords internal
#' @rdname tmapGridLeaflet
tmapLeafletCompPrepare.tm_chart_none = function(comp, o) {
	comp$show = FALSE
	comp
}


#' @export
#' @keywords internal
#' @rdname tmapGridLeaflet
tmapLeafletCompPrepare.default = function(comp, o) {
	cls = class(comp)[1]
	id = paste("view_mode", cls, sep = "_")
	cli::cli_inform("{.field [view mode]} Map component {.fun {cls}} not supported in {.str view} mode.",
					.frequency_id = id,
					.frequency = "once")
	comp$show = FALSE
	comp
}

#' @export
tmapLeafletCompPrepare.tm_legend_portrait = function(comp, o) {
	tmapLeaflet_legend_comp(comp, o)
}

tmapLeaflet_legend_comp = function(comp, o) {
	icons = all(comp$gp$shape >= 1000)

	# replace gp visual values with user-specified used (e.g. tm_shape(World) + tm_polygons("HPI", fill.legend = tm_legend(col = "red")))
	comp$gp = add_user_specified_values(comp$gp, comp[intersect(names(comp), names(comp$gp))])

	# icons replaced by 'normal' shapes: apply icon.scale
	if (icons && all(comp$gp$shape < 1000) && !is.null(comp$layer_args$icon.scale)) {
		comp$gp$size = comp$gp$size * comp$layer_args$icon.scale
	}

	within(comp, {
		if ("biv" %in% names(attributes(gp$fill))) {
			warning("Bivariate legend not implemented for view mode", call. = FALSE)
			show = FALSE
		}

		nuq = vapply(comp$gp, length, FUN.VALUE = integer(1))
		varying = names(nuq)[which(nuq>1)]

		if (all(c("col", "fill") %in% varying)) {
			message("Legend in view mode doesn't support both fill and col varying. Setting col to the first value")
			nuq["col"] = 1
			varying = setdiff(varying, "col")
			gp$col = gp$col[1]
		}
		type = if ((!is.na(gp$fill[1]) && any(nchar(gp$fill) > 50)) || (!is.na(gp$fill_alpha[1]) && any(nchar(gp$fill_alpha) > 50)) ||
				   (!is.na(gp$col[1]) && any(nchar(gp$col) > 50)) || (!is.na(gp$col_alpha[1]) && any(nchar(gp$col_alpha) > 50))) {
			"gradient"

		} else if (any(c("bgcol", "bgcol_alpha") %in% varying)) {
			"none"
		} else if ("tm_data_lines" %in% mfun) {
			"lines"
		} else {
			"symbols"
		}

		bg.color = do.call("process_color", c(list(col=bg.color), o$pc))
		title.color = do.call("process_color", c(list(col=title.color), o$pc))
		text.color = do.call("process_color", c(list(col=text.color), o$pc))

		gp2 = gp_to_lpar(gp, mfun = comp$mfun, shape = comp$item.shape, size_factor = 14)
	})
}


#' @export
tmapLeafletCompHeight.tm_legend_portrait = function(comp, o) {
	within(comp, {
		if (type == "gradient") {
			height = if (is.na(height)) {
				get_legend_option(item.height, "gradient") * nitems * absolute_fontsize * 0.7
			} else {
				if (height < 0) {
					-height
				} else {
					height * absolute_fontsize
				}
			}
		}
	})
}



#' @export
tmapLeafletCompWidth.tm_legend_portrait = function(comp, o) {
	within(comp, {
		if (type == "gradient") {
			width = if (is.na(width)) {
				get_legend_option(item.width, "gradient") * absolute_fontsize
			} else {
				if (width < 0) {
					-width
				} else {
					width * absolute_fontsize
				}
			}
		}
	})
}





#' @export
tmapLeafletCompPrepare.tm_legend_landscape = function(comp, o) {
	tmapLeaflet_legend_comp(comp, o)
}

#' @export
tmapLeafletCompHeight.tm_legend_landscape = function(comp, o) {
	within(comp, {
		if (type == "gradient") {
			height = if (is.na(height)) {
				get_legend_option(item.height, "gradient") * absolute_fontsize
			} else {
				if (height < 0) {
					-height
				} else {
					height * absolute_fontsize
				}
			}
		}
	})
}



#' @export
tmapLeafletCompWidth.tm_legend_landscape = function(comp, o) {
	within(comp, {
		if (type == "gradient") {
			width = if (is.na(width)) {
				# nitems set to 2 because labels are not supported yet
				get_legend_option(item.width, "gradient") * 2 * absolute_fontsize
			} else {
				if (width < 0) {
					-width
				} else {
					width * absolute_fontsize
				}
			}
		}
	})
}

# envir = environment()
.onLoad = function(...) {
	options(tmap.style = "white", tmap.mode = "plot", tmap.design.mode = FALSE,
			tmap.devel.mode = FALSE)
	assign("defaultTmapOptions", .defaultTmapOptions, envir = .TMAP)
	assign("tmapOptions", .defaultTmapOptions, envir = .TMAP)

	# makeActiveBinding("tmap_pals", function() {
	# 	remove_non_letters = function(x) gsub("[-, _, \\,, (, ), \\ , \\.]",  "", x)
	# 	hcl_pals = grDevices::hcl.pals()
	# 	base_pals = grDevices::palette.pals()
	# 	pals = ls(asNamespace("pals"))[substr(ls(asNamespace("pals")), 1, 4) != "pal."]
	# 	list(base_hcl = structure(as.list(hcl_pals), names = remove_non_letters(hcl_pals)),
	# 		 base_pal = structure(as.list(base_pals), names = remove_non_letters(base_pals)),
	# 		 pals = structure(as.list(pals), names = pals))
	# }, env = envir)

	assign("last_map", NULL, envir = .TMAP)
	assign("last_map_new", NULL, envir = .TMAP)


	# flag for old v3 code
	assign("v3", FALSE, envir = .TMAP)

	assign("tips", sample(.tips), envir = .TMAP)
	assign("tip_nr", 1L, envir = .TMAP)

	assign("tmapStyles", .defaultTmapStyles, envir = .TMAP)

	.TMAP$round_to = as.vector(sapply((-9):9, function(i) {
			sapply(c(1, 2.5, 5), function(j) {
				j*10^i
			})
	})) # needed for pretty ticks for continuous scale with trans enabled (like log scale)

	.TMAP$mode_last = "view"

	.TMAP_GRID$maptiles_urls = character(0) # needed for caching

	fill_providers()
}

.tips <- c(
	"tmap works with {.href [visual variables](https://r-tmap.github.io/tmap/articles/basics_vv)}.",
	"tmap can be {.href [extended](https://r-tmap.github.io/tmap/articles/adv_extensions)} in several ways.",
	"{.href [Charts](https://r-tmap.github.io/tmap/articles/basics_charts)} can be added next to legends.",
	"Layout {.href [styles](https://r-tmap.github.io/tmap/articles/adv_options#setting-options-and-styles)} can be created.",
	"Automatic {.href [map projection](https://r-tmap.github.io/tmap/articles/foundations_crs)} recommendations can be set via {.code tm_crs(crs = {.str auto})}.",
	"Layout can be adjusted to look identical as {.href [ggplot2](https://r-tmap.github.io/tmap/articles/versus_ggplot2#mimicking-ggplot2-layout)}.",
	"In {.str plot} mode, a special design-mode can be enabled via {.code tmap_design_mode()}."
)


#' @rdname tmap_providers
#' @export
.tmap_providers <- new.env(FALSE, parent=globalenv())


#' @export
#' @rdname tmap_internal
.TMAP = new.env(FALSE, parent = globalenv())

#' @export
#' @rdname tmap_internal
.TMAP_LEAFLET = new.env(FALSE, parent = globalenv())

#' @export
#' @rdname tmap_internal
.TMAP_GRID = new.env(FALSE, parent = globalenv())


fill_providers = function() {
	rm(list = ls(envir = .tmap_providers), envir = .tmap_providers)
	y = tmap_providers(as.list = TRUE)
	list2env(y, envir = .tmap_providers)
}

gridCell = function(rows, cols, e) {
	vp = grid::viewport(layout.pos.row = rows, layout.pos.col = cols)
	grid::grobTree(e, vp = vp)
}




process_comp_box = function(comp, sc, o) {
	comp = within(comp, {
		if ("frame.color" %in% comp$called) {
			frame = TRUE
			frame.lwd = frame.lwd * sc
		} else {
			if (!is.logical(frame)) {
				cls = class(comp)[1]
				if (valid_colors(frame)) {
					cli::cli_warn("{.field {.fun {cls}}} use {.code frame.color = {.str {frame}}} instead of {.code frame = {.str {frame}}}")
					frame.color = frame
					frame = TRUE
				} else {
					cli::cli_abort("{.field {.fun {cls}}} {.arg frame} should be a logical")
				}

			}
			frame.color = if (!frame) NA else if (is.null(frame.color)) o$attr.color else frame.color
			frame.lwd = if (!frame) 0 else frame.lwd * sc
		}
		frame.r = frame.r * sc
	})
	comp
}

tmapGridComp2 = function(grp, comp, o, stack, pos.h, pos.v, maxH, maxW, offsetIn.h, offsetIn.v, marginInH, marginInV, are_nums, fH, fW) {
	n = length(comp)
	legWin = vapply(comp, "[[", FUN.VALUE = numeric(1), "Win")   #  vapply(comp, leg_standard$fun_width, FUN.VALUE = numeric(1), o = o)
	legHin = vapply(comp, "[[", FUN.VALUE = numeric(1), "Hin")#vapply(comp, leg_standard$fun_height, FUN.VALUE = numeric(1), o = o)

	group.just = c(pos.h, pos.v)
	frame_combine = grp$frame_combine

	legWin[is.infinite(legWin)] = maxW
	legHin[is.infinite(legHin)] = maxH

	scaleW = legWin / maxW
	scaleH = legHin / maxH

	# because of legend frames (for which margins are added in tmapGridComp), the scale may be a bit above 1, even though automatic layout is applied and there is enough space

	if (o$component.autoscale) {
		if (any(scaleW > 1.1) || any(scaleH > 1.1)) message_comp_scale()
	} else {
		scaleW[scaleW > 1.1] = 1
		scaleH[scaleH > 1.1] = 1
	}


	clipW = pmax(1, scaleW)
	clipH = pmax(1, scaleH)
	if (grp$resize_as_group) {
		clipT = rep(max(clipW, clipH), n)
	} else {
		clipT = pmax(clipW, clipH)
	}

	legWin = legWin / clipT
	legHin = legHin / clipT

	if (frame_combine) {
		if (stack == "vertical") {
			legWin = rep(max(legWin), n)
		} else {
			legHin = rep(max(legHin), n)
		}
	}

	# rescale due to stacking
	if (stack == "vertical") {
		scaleS = sum(legHin) / maxH
	} else {
		scaleS = sum(legWin) / maxW
	}

	if (o$component.autoscale) {
		if (scaleS > 1.01) {
			if (scaleS > 1.05) {
				message_comp_high_wide(stack)

			}
			legWin = legWin / scaleS
			legHin = legHin / scaleS
			clipT = clipT * scaleS
		}
	}

	legW = grid::unit(legWin, "inches")
	legH = grid::unit(legHin, "inches")

	if (stack == "vertical") {
		W = max(legWin)
		H = sum(legHin)

		Hs = unit_add_between(legHin, marginInV)
		Ws = W
	} else {
		W = sum(legWin)
		H = max(legHin)

		Ws = unit_add_between(legWin, marginInH)
		Hs = H
	}





	if (are_nums) {
		group.just = as.numeric(group.just)
		just.h = grp$position$just.h
		just.v = grp$position$just.v

		# add dummy offsets
		Hs = unit_add_sides(Hs, 0)
		Ws = unit_add_sides(Ws, 0)

		ancher_x = maxW * group.just[1]
		if (just.h == "left") {
			Ws = unit_add_sides(Ws, c(ancher_x, maxW - ancher_x- sum(Ws)))
		} else if (just.h == "right") {
			Ws = unit_add_sides(Ws, c(ancher_x - sum(Ws), maxW - ancher_x))
		} else {
			Ws = unit_add_sides(Ws, c(ancher_x - (sum(Ws) / 2), maxW - ancher_x- (sum(Ws) / 2)))
		}

		ancher_y = maxH * (1 - group.just[2])
		if (just.v == "top") {
			Hs = unit_add_sides(Hs, c(ancher_y, maxH - ancher_y - sum(Hs)))
		} else if (just.v == "bottom") {
			Hs = unit_add_sides(Hs, c(ancher_y - sum(Hs), maxH - ancher_y))
		} else {
			Hs = unit_add_sides(Hs, c(ancher_y - (sum(Hs) / 2), maxH - ancher_y - (sum(Hs) / 2)))
		}

	} else {
		nullsH = switch(group.just[1], left = c(0, 1), center = c(0.5, 0.5), right = c(1, 0))
		nullsV = switch(group.just[2], top = c(0, 1), center = c(0.5, 0.5), bottom = c(1, 0))

		Hs = unit_add_sides(Hs, offsetIn.v)
		Hs = unit_add_sides(Hs, nullsV)

		Ws = unit_add_sides(Ws, offsetIn.h)
		Ws = unit_add_sides(Ws, nullsH)

	}

	Hn = length(Hs)
	Wn = length(Ws)

	Hid = seq(3, Hn - 2, by = 2)
	Wid = seq(3, Wn - 2, by = 2)

	if (are_nums) {
		Wu = rep("inch", Wn)
		Hu = rep("inch", Hn)
	} else {
		Wu = c("null", rep("inch", Wn-2), "null")
		Hu = c("null", rep("inch", Hn-2), "null")
	}

	# po(Ws, Hs)
	# Ws = distr_space_over_nulls(Ws, maxW)
	# Hs = distr_space_over_nulls(Hs, maxH)
	# po(Ws, Hs)

	vp = grid::viewport(layout = grid::grid.layout(ncol = Wn,
												   nrow = Hn,
												   widths = grid::unit(Ws, Wu),
												   heights = grid::unit(Hs, Hu)))


	comp = mapply(function(leg, scale, W, H) {
		leg$scale = scale

		if ("flexCol" %in% names(leg)) leg$wsu = distr_space_over_nulls(u = leg$wsu * scale, tot = W, stretchID = leg$flexCol)
		if ("flexRow" %in% names(leg)) leg$hsu = distr_space_over_nulls(u = leg$hsu * scale, tot = H, stretchID = leg$flexRow)
		leg
	}, comp, 1/clipT, legW, legH, SIMPLIFY = FALSE, USE.NAMES = FALSE)


	legGrobs = lapply(comp, tmapGridCompPlot, o = o, fH = fH, fW = fW)


	#sq = function(x) do.call(seq, as.list(unname(range(x))))
	sc = min(1/clipT) * o$scale

	comp = lapply(comp, process_comp_box, sc = sc, o = o)
	grp = process_comp_box(grp, sc, o)

	design = getOption("tmap.design.mode")

	draw_rect = grp$frame || grp$bg
	if (draw_rect && frame_combine) {
		if (grp$frame) {
			groupframe = gridCell(range(Hid), range(Wid), rndrectGrob(gp=grid::gpar(fill = NA, col = grp$frame.color, lwd = grp$frame.lwd), r = grp$frame.r))
		} else {
			groupframe = NULL
		}

		if (grp$bg && !design) {
			groupbg = gridCell(range(Hid), range(Wid), rndrectGrob(gp=grid::gpar(fill = grp$bg.color, alpha = grp$bg.alpha, col = NA, lwd = 0), r = grp$frame.r))
		} else {
			groupbg = NULL
		}
	} else {
		groupbg = NULL
		groupframe = NULL
	}

	equalize = grp$equalize


	if (equalize) {
		if (stack == "horizontal") {
			legFH = grid::unit(1, "npc")
			legFW = legW
		} else {
			legFH = legH
			legFW = grid::unit(1, "npc")
		}
	} else {
		legFH = legH
		legFW = legW
	}


	grbs = do.call(grid::gList, mapply(function(leg, lG, lH, lW, fH, fW, iW, iH) {
		frame = if (draw_rect && !frame_combine) {
			bg.color = if (leg$bg && !design) leg$bg.color else NA
			frame.color = if (leg$frame) leg$frame.color else NA
			rndrectGrob(gp=grid::gpar(fill = bg.color, alpha = leg$bg.alpha, col = frame.color, lwd = leg$frame.lwd), r = leg$frame.r)
		} else NULL
		if (stack == "vertical") {
			x = switch(leg$position$align.h, "left" = lW/2, "right" = grid::unit(W, "inch") -lW/2, grid::unit(0.5, "npc"))
			y = grid::unit(0.5, "npc")
		} else {
			x = grid::unit(0.5, "npc")
			y = switch(leg$position$align.v, "top" = grid::unit(H, "inch")-lH/2, "bottom" = lH/2, grid::unit(0.5, "npc"))
		}
		gridCell(iH, iW, {
			grid::gList(
				if (equalize) {
					grid::grobTree(frame, vp = grid::viewport(width = fW, height = fH))
				} else {
					grid::grobTree(frame, vp = grid::viewport(x = x, y = y, width = fW, height = fH))
				},
				grid::grobTree(lG, vp = grid::viewport(x = x, width = lW, y = y, height = lH)))
		})
	}, comp, legGrobs, legH, legW, legFH, legFW, Wid, Hid, SIMPLIFY = FALSE))


	if (design) {
		df = expand.grid(col = 1:ncol(vp$layout),
						 row = 1:nrow(vp$layout))

		grDesign = lapply(1:nrow(df), function(i) gridCell(df$row[i], df$col[i], grid::rectGrob(gp=gpar(fill=NA,col="#000000", lwd=2))))
	} else {
		grDesign = NULL
	}

	do.call(grid::grobTree, c(list(groupbg), grbs, list(groupframe), grDesign, list(vp=vp)))
}

#' @export
#' @rdname tmap_internal
tmapGetCompGroupArgs = function(comp) {
	# get component group settings
	nms = setdiff(names(formals(tm_components)), "group_id")
	grp = comp[[1]][nms]

	any_legend_chart_inset = any(vapply(comp, inherits, FUN.VALUE = logical(1), c("tm_legend", "tm_chart")))
	grp_called = setdiff(unique(do.call(c, lapply(comp, FUN = "[[", "called_via_comp_group"))), "group_id")

	if (!("frame" %in% grp_called)) {
		# update because of #1214
		frms = vapply(comp, function(ci) ci$frame, FUN.VALUE = logical(1))
		if (any(frms) && (any(!frms))) grp$frame = any_legend_chart_inset
	}
	if (!("bg" %in%grp_called)) grp$bg = any_legend_chart_inset
	grp$called = comp[[1]]$called
	grp
}


tmapGridComp = function(comp, o, facet_row = NULL, facet_col = NULL, facet_page, class, stack, stack_auto, pos.h, pos.v, bbox) {

	grp = tmapGetCompGroupArgs(comp)


	gts = get("gts", envir = .TMAP_GRID)
	g = get("g", envir = .TMAP_GRID)

	gt = gts[[facet_page]]
	n = length(comp)
	# rows and cols in main grid
	rows = if (facet_row[1] == -2) g$meta_rows[1] else if (facet_row[1] == -1) g$meta_rows[2] else {
		panel_rows = if (class == "in") {
			NULL
		} else if (o$panel.type == "xtab") {
		    g$rows_panel_row_ids[facet_row]
		} else if (o$panel.type == "wrap") {
			g$rows_panel_ids[facet_row]
		} else {
			NULL
		}
		c(g$rows_facet_ids[facet_row], panel_rows)
	}
	cols = if (facet_col[1] == -2) g$meta_cols[1] else if (facet_col[1] == -1) g$meta_cols[2] else {
		panel_cols = if (class == "in") {
			NULL
		} else if (o$panel.type == "xtab") {
			g$cols_panel_col_ids[facet_col]
		} else if (o$panel.type == "wrap") {
			g$cols_panel_ids[facet_col]
		} else {
			NULL
		}
		c(g$cols_facet_ids[facet_col], panel_cols)
	}


	# make rows and cols a range
	rows = seq(min(rows), max(rows))
	cols = seq(min(cols), max(cols))


	# multiple not needed anymore
	stack = stack[1] # check what happens when stack_auto is T, e.g. facet_stack
	pos.h = pos.h[1]
	pos.v = pos.v[1]



	rowsIn = g$rowsIn[rows]
	colsIn = g$colsIn[cols]
	if (pos.h %in% c("LEFT", "RIGHT")) {
		pos.h = tolower(pos.h)
		CASE.h = toupper
	} else {
		if (pos.h %in% c("CENTER", "CENTRE")) pos.h = "center"
		CASE.h = function(x)x
	}
	if (pos.v %in% c("TOP", "BOTTOM")) {
		pos.v = tolower(pos.v)
		CASE.v = toupper
	} else {
		if (pos.v %in% c("CENTER", "CENTRE")) pos.v = "center"
		CASE.v = function(x)x
	}

	are_nums = !(any(is.na(suppressWarnings(as.numeric(c(pos.h, pos.v))))))

	if (are_nums) {
		component.offset.h = 0
		component.offset.v = 0
	} else {
		offset = grp$offset
		if (!is.null(names(offset)) && all(c("inside", "INSIDE", "outside", "OUTSIDE") %in% names(offset))) {
			component.offset.h = get_option_class(offset, class = CASE.h(paste0(class, "side")), spatial_class = FALSE)
			component.offset.v = get_option_class(offset, class = CASE.v(paste0(class, "side")), spatial_class = FALSE)
		} else {
			offset = rep_len(offset, 2L)
			component.offset.h = offset[1]
			component.offset.v = offset[2]

		}
	}

	offsetIn.h = component.offset.h * o$lin# + (o$frame.lwd * o$scale / 144) # 1 line = 1/72 inch, frame lines are centered (so /2)
	offsetIn.v = component.offset.v * o$lin#+ (o$frame.lwd * o$scale / 144)

	stack_margin = grp$stack_margin
	if (!is.null(names(stack_margin)) && all(c("combined", "apart") %in% names(stack_margin))) {
		stack_margin = rep_len(get_option_class(stack_margin, class = ifelse(comp[[1]]$frame_combine, "combined", "apart"), spatial_class = FALSE), 2L)
	} else {
		stack_margin = rep_len(stack_margin, 2L)
	}

	if (stack == "horizontal") {
		stack_margin[2] = 0
	} else {
		stack_margin[1] = 0
	}

	marginInH = stack_margin[1] * o$lin
	marginInV = stack_margin[2] * o$lin

	marginInTotH = (n - 1L) * marginInH
	marginInTotV = (n - 1L) * marginInV
	offsetInTot.h  = 2 * offsetIn.h
	offsetInTot.v  = 2 * offsetIn.v

	totH = sum(rowsIn) - offsetInTot.v - marginInTotV
	totW = sum(colsIn) - offsetInTot.h - marginInTotH

	## update scale_bar width: identified by the WnativeID = 3 item (null for other components)
	comp = mapply(function(cmp, bb) {
		bbw = bb[3] - bb[1]

		# TRUE if component is scalebar
		if (!is.null(cmp$WnativeID)) {
			# find out whether the bounding of borrowed from the map (yes if scale bar is drawn in another facet)
			bbox_nb = attr(bbox, "borrow")
			if (is.null(bbox_nb)) {
				bb_facet = sum(colsIn)
			} else {
				bb_facet = sum(g$colsIn[g$cols_facet_ids[bbox_nb$col]])
			}

			oldIn = as.numeric(cmp$wsu[cmp$WnativeID])

			if (is.null(cmp$WnativeRange) || ("width" %in% cmp$call)) {
				# in case breaks not defined: allow the width to be maximal fraction of facet width
				#newIn = min(totW, bb_facet * cmp$width)
				newIn = min(totW, cmp$Win)
			} else {
				# in case breaks are defined: compute width
				txt = paste0(cmp$breaks[1], " ", tail(cmp$breaks, 1), " ", cmp$units$unit, cmp$units$unit)

				Wextra = text_width_inch(txt) / 1.2 * cmp$text.size

				totalCoords =  as.numeric(units::set_units(units::as_units(bbw, st_crs(bb)$units), cmp$units$unit, mode = "standard"))

				Wcomp = bb_facet * (cmp$WnativeRange / totalCoords) + Wextra

				newIn = min(totW, Wcomp)
			}

			margins =  sum(as.numeric(cmp$wsu[c(2,4)]))

			newIn_without_margins = newIn - margins

			cmp$wsu[cmp$WnativeID] = unit(newIn_without_margins, "inch")
			cmp$Win = newIn_without_margins + margins

			# get cpi: coordinates per inch

			cmp$dst = distances_bbox_sides(bb)

			cmp$which_side_of_map = if (class == "in" && pos.v == "bottom") {
				"bottom"
			} else if (class == "in" && pos.v == "top") {
				"top"
			} else if (class == "in" && pos.v == "center") {
				"mid"
			} else {
				"bottom"
			}

			#cmp$cpi = unname(bbw / bb_facet)
			cmp$bb_facet = bb_facet
			#po(bb, bbw, bbox_nb, oldIn, bb_facet, newIn, totW, cmp$Win, margins, newIn_without_margins, cmp$Win)
			#po(cmp$cpi)
		}
		cmp
	}, comp, bbox, SIMPLIFY = FALSE)

	legWin = vapply(comp, "[[", FUN.VALUE = numeric(1), "Win")   #  vapply(comp, leg_standard$fun_width, FUN.VALUE = numeric(1), o = o)
	legHin = vapply(comp, "[[", FUN.VALUE = numeric(1), "Hin")#vapply(comp, leg_standard$fun_height, FUN.VALUE = numeric(1), o = o)


	# get total value (width or height)
	getH = function(s, lH) {
		if (!length(s)) return(NULL)
		if (s[1] == "vertical") {
			sum(unit_add_between(lH, marginInV))
		} else {
			max(lH)
		}
	}
	getW = function(s, lW) {
		if (!length(s)) return(NULL)
		if (s[1] == "vertical") {
			max(lW)
		} else {
			sum(unit_add_between(lW, marginInH))
		}
	}

	grbs = tmapGridComp2(grp = grp, comp = comp, o = o, stack = stack, pos.h = pos.h, pos.v = pos.v, maxH = totH, maxW = totW, offsetIn.h = offsetIn.h, offsetIn.v = offsetIn.v, marginInH = marginInH, marginInV = marginInV, are_nums = are_nums, fH = totH, fW = totW)#sum(rowsIn), fW = sum(colsIn))

	gt = add_to_gt(gt, grbs, row = rows, col = cols)

	gts[[facet_page]] = gt

	assign("gts", gts, envir = .TMAP_GRID)

}

#' Save tmap
#'
#' Save tmap to a file. This can be either a static plot (e.g. png) or an interactive map (html).
#'
#' @param tm tmap object
# The docs for filename could be put in tmap_mode.
#' @param filename filename including extension, and optionally the path.
#'   The extensions pdf, eps, svg, wmf (Windows only), png, jpg, bmp, tiff, and html are supported.
#'   If the extension is missing, the file will be saved as a static plot in `"plot"`
#'   mode and as an interactive map (html) in another mode.
#'   The default format for static plots is png, but this can be changed using
#'   the option `"output.format"` in [tmap_options()]. If `NA` (the default),
#'   the file is saved as "tmap01" in the default format, and the number incremented
#'   if the file already exists.
#' @param device graphic device to use. Either a device function
#'   (e.g., [`png`][grDevices::png()] or [`cairo_pdf`][grDevices::cairo_pdf()])
#'   or a text indicating selected graphic device: "pdf", "eps", "svg", "wmf" (Windows only), "png", "jpg", "bmp", "tiff".
#'   If `NULL`, the graphic device is guessed based on the `filename` argument.
#' @param height,width The dimensions of the plot (not applicable for html files).
#'   Units are set with the argument `units`. If one of them is not specified,
#'   this is calculated using the formula asp = width / height, where asp is the
#'   estimated aspect ratio of the map. If both are missing, they are set such
#'   that `width * height` is equal to the option `"output.size"` in [tmap_options()].
#'   This is by default 49, meaning that is the map is a square (so aspect ratio of 1)
#'   both width and height are set to 7.
#' @param units units for width and height (`"in"`, `"cm"`, or `"mm"`).
#'   By default, pixels (`"px"`) are used if either width or height is set to a
#'   value greater than 50. Else, the units are inches (`"in"`).
#' @param dpi dots per inch. Only applicable for raster graphics. By default it
#'   is set to 300, but this can be changed using the option `"output.dpi"` in [tmap_options()].
#' @param outer.margins overrides the outer.margins argument of [tm_options()] (unless set to `NA`)
#' @param asp if specified, it overrides the asp argument of [tm_options()].
#'   **Tip**: set to `0` if map frame should be placed on the edges of the image.
#' @param scale overrides the scale argument of [tm_options()] (unless set to `NA`)
#' @param insets_tm tmap object of an inset map, or a list of tmap objects of
#'   multiple inset maps. The number of tmap objects should be equal to the number
#'   of viewports specified with `insets_vp`.
#' @param insets_vp [`viewport`][grid::viewport()] of an inset map, or a list
#'   of [`viewport`][grid::viewport()]s of multiple inset maps. The number of
#'   viewports should be equal to the number of tmap objects specified with `insets_tm`.
#' @param add.titles add titles to leaflet object.
#' @param in.iframe should an interactive map be saved as an iframe?
#'   If so, two HTML files will be saved; one small parent HTML file with the
#'   iframe container, and one large child HTML file with the actual widget.
#'   See [widgetframe::saveWidgetframe()] for details. By default `FALSE`,
#'   which means that one large HTML file is saved (see [saveWidget()][htmlwidgets::saveWidget()]).
#' @param selfcontained when an interactive map is saved, should the resources
#'   (e.g. JavaScript libraries) be contained in the HTML file? If `FALSE`, they
#'   are placed in an adjacent directory (see also [htmlwidgets::saveWidget()]).
#'   Note that the HTML file will often still be large when `selfcontained = FALSE`,
#'   since the map data (polygons and popups), which are also contained in the HTML file,
#'   usually take more space then the map resources.
#' @param verbose Deprecated. It is now controlled by the tmap option `show.messages`
#'    (see [tmap_options()])
#' @inheritDotParams htmlwidgets::saveWidget
#' @inheritDotParams widgetframe::saveWidgetframe
#' @returns the filename, invisibly, if export is successful.
#' @importFrom htmlwidgets saveWidget
#' @import tmaptools
#' @example ./examples/tmap_save.R
#' @export
tmap_save = function(tm=NULL, filename=NA, device=NULL, width=NA, height=NA, units = NA,
					  dpi=NA, outer.margins=NA, asp=NULL, scale=NA, insets_tm=NULL,
					 insets_vp=NULL, add.titles = TRUE, in.iframe = FALSE, selfcontained = !in.iframe, verbose = NULL, ...) {
	.tmapOptions = get("tmapOptions", envir = .TMAP)
	show.warnings =.tmapOptions$show.warnings
	if (!missing(verbose) && show.warnings) warning("The argument verbose is deprecated. Please use the option show.messages of tmap_options instead.")



	verbose = .tmapOptions$show.messages


	lastcall = x = get("last_map", envir = .TMAP)
	if (missing(tm)) {
		tm = suppressWarnings(tmap_last())
		if (is.null(tm)) stop("A map has not been created yet")
		is.arrange = FALSE
	} else if (inherits(tm, "tmap")) {
		is.arrange = FALSE
	} else if (inherits(tm, "tmap_arrange")) {
		is.arrange = TRUE
	} else {
		stop("Unknown format. tm should be either a tmap output, or a list of tmap outputs")
	}

	tmap.mode = getOption("tmap.mode")
	default_ext = ifelse(tmap.mode == "plot", .tmapOptions$output.format, "html")


	if (is.na(filename)) {
		filename_default = paste("tmap01", default_ext, sep = ".")
		if (!file.exists(filename_default)) {
			filename = filename_default
		} else {
			files = list.files(pattern = paste0("^tmap[0-9]{2}\\.", default_ext, "$"))
			fid = max(as.integer(substr(files, 5, 6))) + 1L
			filename = paste0("tmap", sprintf("%02d", fid), ".",  default_ext)
		}
	}

	if (is.na(dpi)) dpi = .tmapOptions$output.dpi

	on.exit({
		assign("last_map", lastcall, envir = .TMAP)
	})



	get_ext = function(filename, default_ext) {
		pieces = strsplit(filename, "\\.")[[1]]
		if (length(pieces)==1) return(default_ext)
		tolower(pieces[length(pieces)])
	}

	convert_to_inches = function(x, units) {
		x = switch(units, px = x/dpi, `in` = x, cm = x/2.54, mm = x/2.54/10)
	}
	convert_to_pixels = function(x, units) {
		x = switch(units, px = x, `in` = dpi*x, cm = dpi*x/2.54, mm = dpi*x/2.54/10)
	}

	ext = get_ext(filename, default_ext)

	interactive = (ext=="html")

	interactive_mode = if (tmap.mode == "plot") "view" else tmap.mode

	options(tmap.mode=ifelse(interactive, interactive_mode, "plot"))

	if (interactive) {
		if (is.arrange) {
			lf = print_tmap_arrange(tm, show = FALSE, add.titles=add.titles)
		} else {
			lf = print.tmap(tm, show = FALSE)
		}

		tryCatch({
			wd = getwd()
			on.exit(setwd(wd), add = TRUE)
			wd_new = dirname(filename)
			base_filename = basename(filename)
			setwd(wd_new)
			if (in.iframe) {
				rlang::check_installed("widgetframe", reason = "for option in.frame.")
				widgetframe::saveWidgetframe(lf, file=base_filename, selfcontained = selfcontained, ...)
			} else {
				htmlwidgets::saveWidget(lf, file=base_filename, selfcontained = selfcontained, ...)
			}
		}, error = function(e) {
			stop("Unable to save the interactive map. Note that saving interactive small multiples is not supported yet. The error message from htmlwidgets::saveWidget is ", call. = FALSE)
		})

		options(tmap.mode=tmap.mode)
		if (verbose) {
			message("Interactive map saved to ", suppressWarnings(normalizePath(filename)))
		}
		return(invisible(filename))
	}

	if (is.na(width) || is.na(height)) {
		if (!is.na(width)) {
			if (is.na(units)) units = choose_unit(width)
			temp_size = convert_to_pixels(width, units)
		} else if (!is.na(height)) {
			if (is.na(units)) units = choose_unit(height)
			temp_size = convert_to_pixels(height, units)
		} else {
			units = "px"
			temp_size = 700
		}

		if (is.arrange) {
			sasp = 1
		} else {
			show.messages = tmap_options(show.messages = FALSE)
			on.exit(tmap_options(show.messages))
			sasp = get_asp_ratio(tm, width = temp_size, height = temp_size, res = dpi)
			tmap_options(show.messages)
		}

		if (is.na(width) && !is.na(height)) {
			width = height * sasp
		} else if (is.na(height)  && !is.na(width)) {
			height = width / sasp
		} else {
			units = "in"
			height = sqrt(.tmapOptions$output.size / sasp)
			width = height * sasp
		}
	} else {
		if (is.na(units)) units = choose_unit(max(width, height))
	}
	units_target = ifelse(units=="px" && ext %in% c("png", "jpg", "jpeg", "bmp", "tiff"), "px", "in")

	if (units_target=="in") {
	  width = convert_to_inches(width, units)
	  height = convert_to_inches(height, units)

	  if (ext=="pdf") {
	    round_to_1_72 = function(x) x %/% (1/72) / 72
	    width = round_to_1_72(width)
	    height = round_to_1_72(height)
	  }
	} else {
		width = convert_to_pixels(width, units)
		height = convert_to_pixels(height, units)
	}
	old_dev = grDevices::dev.cur()
	dev = plot_device(device = device, ext = ext, filename = filename,
					   dpi = dpi, units_target = units_target)

	folder = dirname(filename)
	if (!dir.exists(folder)) dir.create(folder, recursive = TRUE)
	dev(filename = filename, width = width, height = height, ...)

	on.exit(capture.output({
			dev.off()
			if (old_dev > 1) grDevices::dev.set(old_dev) # restore old device unless null device
			}), add = TRUE)

	if (is.arrange) {
		opts = attr(tm, "opts")
		if (!is.na(outer.margins[1])) opts$outer.margins = outer.margins
		if (!missing(asp)) opts$asp = asp

		attr(tm, "opts") = opts
		print(tm)
	} else {
		args = list()
		if (!is.na(outer.margins[1])) args$outer.margins = outer.margins
		if (!missing(asp)) args$asp = asp
		if (!is.na(scale)) args$scale = scale
		args$animate_disable = TRUE
		print(tm + do.call("tm_options", args))
	}

	if (!is.arrange && !missing(insets_tm) && !missing(insets_vp)) {
	  args_inset = if (!is.na(scale)) list(scale = scale) else list()
	  if (inherits(insets_tm, "list") && inherits(insets_vp, "list")) {
	    if (length(insets_tm) != length(insets_vp)) stop("Number of insets unequal to number of viewports")
	    mapply(function(tm_i, vp_i) {
	      print(tm_i + do.call("tm_options", args_inset), vp=vp_i)
	    }, insets_tm, insets_vp)
	  } else if (inherits(insets_tm, "tmap") && inherits(insets_vp, "viewport")) {
	    print(insets_tm + do.call("tm_options", args_inset), vp = insets_vp)
	  } else {
	    stop("Insets and/or its viewports not in the correct format")
	  }
	}

	if (verbose) message("Map saved to ", suppressWarnings(normalizePath(filename)))
	if (ext %in% c("png", "jpg", "jpeg", "bmp", "tiff")) {
		if (units_target == "px") {
			wp = width
			hp = height
			wi = convert_to_inches(width, "px")
			hi = convert_to_inches(height, "px")
		} else {
			wi = width
			hi = height
			wp = convert_to_pixels(width, "in")
			hp = convert_to_pixels(height, "in")
		}
		if (verbose) message("Resolution: ", format(wp), " by ", format(hp), " pixels")
		if (verbose) message("Size: ", format(wi), " by ", format(hi), " inches (", format(dpi), " dpi)")
	} else {
		wi = width
		hi = height
		if (verbose) message("Size: ", format(wi), " by ", format(hi), " inches")
	}

	options(tmap.mode=tmap.mode)

	output = list(filename = filename)
	if (ext %in% c("png", "jpg", "jpeg", "bmp", "tiff")) {
		output$size_pixels = c(width = wp, height = hp)
		output$dpi = dpi
	}
	output$size_inch = c(width = wi, height = hi)

	output = c(output, .TMAP$geo_ref)

	invisible(output)
}

plot_device = function(device, ext, filename, dpi, units_target){

	force(filename)
	force(dpi)
	force(units_target)

	if (is.function(device)) {
		args = formals(device)
		call_args = list()
		if ("file" %in% names(args)) {
			call_args$file = filename
		}
		if ("res" %in% names(args)) {
			call_args$res = dpi
		}
		if ("units" %in% names(args)) {
			call_args$units = units_target
		}
		dev = function(...) do.call(device, utils::modifyList(list(...), call_args))
		return(dev)
	}
	ps = function(..., filename, width, height)
		grDevices::postscript(
			...,
			file = filename,
			width = width,
			height = height,
			onefile = FALSE,
			horizontal = FALSE,
			paper = "special"
		)
	jpeg = function(..., width, height)
		grDevices::jpeg(
			...,
			width = width,
			height = height,
			res = dpi,
			units = units_target
		)
	devices = list(
		ps = ps,
		eps = ps,
		tex = function(..., filename, width, height)
			grDevices::pictex(..., file = filename, width = width, height = height),
		pdf = function(..., filename, version = "1.4")
			grDevices::pdf(..., file = filename, version = version),
		svg = function(..., filename)
			grDevices::svg(..., file = filename),
		wmf = function(..., width, height)
			grDevices::win.metafile(..., width = width, height = height),
		emf = function(..., width, height)
			grDevices::win.metafile(..., width = width, height = height),
		png = function(..., width, height)
				grDevices::png(
					...,
					type = "cairo-png", # several advantages, main: default "cairo" doesn't respect width and height (breaks av animations)
					width = width,
					height = height,
					res = dpi,
					units = units_target
				),
		jpeg = jpeg,
		jpg = jpeg,
		bmp =
			function(..., width, height)
				grDevices::bmp(
					...,
					width = width,
					height = height,
					res = dpi,
					units = units_target
				),
		tiff = function(..., width, height)
			grDevices::tiff(
				...,
				width = width,
				height = height,
				res = dpi,
				units = units_target
			)
	)
	if (is.null(device)) {
		dev = devices[[ext]]
		if (is.null(dev)) {
			stop("'", dev, "'", " graphic device does not exist", call. = FALSE)
		}
		return(dev)
	}
	if (!device %in% names(devices)) {
		stop('"', device, '"', " graphic device does not exist", call. = FALSE)
	}
}

choose_unit = function(x) {
	units = ifelse(x > 50, "px", "in")
	if (x > 15 && x < 100) {
		message("The argument 'units' has been set to \"",
				units, "\" since the specified width or height is ",
				ifelse(units == "px", "greater than ", "less than or equal to "), 50, ". Specify units = \"", ifelse(units == "px", "in", "px"),"\" to change this.")
	}
	units
}

tmapGridShape = function(bbx, facet_row, facet_col, facet_page, o) {
	gts = get("gts", envir = .TMAP_GRID)
	g = get("g", envir = .TMAP_GRID)

	design = getOption("tmap.design.mode")


	fbbx = bb_asp(bbx, g$fasp)

	rc_text = frc(facet_row, facet_col)

	rowid = g$rows_facet_ids[facet_row]
	colid = g$cols_facet_ids[facet_col]

	bg = if (o$earth_boundary) {
		if (!o$space || design) o$space.color = NA
		eotw = end_of_the_world(o$crs_step4, earth_datum = o$earth_datum)
		grid::gList(rndrectGrob(gp=gpar(fill=o$space.color, lwd=NA, lineend="square"), r = o$frame.r * o$scale, name = "inner_rect"),
					sf::st_as_grob(eotw, gp = grid::gpar(fill = o$bg.color, lwd = NA), name = "inner_world"))

	} else if (design || !o$bg) {
		NULL
	} else {
		rndrectGrob(gp=gpar(fill=o$bg.color, lwd=NA, lineend="square"), r = o$frame.r * o$scale, name = "inner_rect")
	}

	clip = if (is.na(o$bg.color) && o$earth_boundary && !o$space_overlay) {
		grid::as.path(bg)
	} else TRUE

	gtmap = grid::grobTree(bg,
						   vp = grid::vpStack(grid::viewport(layout.pos.col = colid, layout.pos.row = rowid, name = paste0("vp_facet_", rc_text)),
						   				   grid::viewport(xscale = fbbx[c(1,3)], yscale = fbbx[c(2,4)], name = paste0("vp_map_", rc_text), clip = clip)), name = paste0("gt_facet_", rc_text))

	gts[[facet_page]] = grid::addGrob(gts[[facet_page]], gtmap, gPath = grid::gPath("gt_main"))

	assign("gts", gts, envir = .TMAP_GRID)
	NULL
}


tmapGridOverlay = function(bbx, facet_row, facet_col, facet_page, o) {
	gts = get("gts", envir = .TMAP_GRID)
	g = get("g", envir = .TMAP_GRID)

	gt = gts[[facet_page]]

	rc_text = frc(facet_row, facet_col)
	frame.lwd = if (!o$frame) 0 else o$frame.lwd * o$scale
	r = o$frame.r * o$scale

	boundary = if (o$earth_boundary) {
		fbbx = bb_asp(bbx, g$fasp)

		eotw = end_of_the_world(o$crs_step4, earth_datum = o$earth_datum)

		sf::st_as_grob(eotw, gp = grid::gpar(fill = NA, col = o$earth_boundary.color, lwd = o$earth_boundary.lwd), name = "outer_world", vp = grid::viewport(xscale = fbbx[c(1,3)], yscale = fbbx[c(2,4)]))
	} else NULL

	gframe = if (o$frame) {
		pH = convertHeight(unit(1, "points"), unitTo = "npc", valueOnly = TRUE)*frame.lwd
		pW = convertWidth(unit(1, "points"), unitTo = "npc", valueOnly = TRUE)*frame.lwd

		if (o$frame.double_line) {
			bgcol = substr(o$bg.color, 1, 7) # remove transparency
			grid::grobTree(
				rndrectGrob(width = 1-4*pW, height=1-4*pH, gp=gpar(col=bgcol, fill=NA, lwd=4*frame.lwd, lineend="square"), r = r, name = "outer_frame"),
				rndrectGrob(gp=gpar(col=o$frame.color, fill=NA, lwd=1.5*frame.lwd, lineend="square", alpha = o$frame.alpha), r = r, name = "between_frame"),
				rndrectGrob(width = 1-8*pW, height=1-8*pH, gp=gpar(col=o$frame.color, fill=NA, lwd=frame.lwd, lineend="square", alpha = o$frame.alpha), r = r, name = "inner_frame"))
		} else {
			rndrectGrob(width = 1, height = 1, gp=gpar(col=o$frame.color, fill=NA, lwd=frame.lwd, lineend="square", alpha = o$frame.alpha), r = r, name = "outer_frame")
			#rectGrob(width = 1-0.5*pW, height = 1-0.5*pH, gp=gpar(col=frame.col, fill=NA, lwd=frame.lwd, lineend="square"), name = "outer_frame")
		}

	} else NULL



	rowid = g$rows_facet_ids[facet_row]
	colid = g$cols_facet_ids[facet_col]

	if (o$earth_boundary && o$space_overlay) {
		space_fill = local({
			if (o$frame) {
				gframe_path = roundrect_to_pathGrob(gframe)
			} else {
				gframe_path = viewport_frame_path()
			}
			boundary_npc = local({
				x = as.vector(boundary$x)
				x2 = (x - boundary$vp$xscale[1]) / diff(boundary$vp$xscale)
				y = as.vector(boundary$y)
				y2 = (y - boundary$vp$yscale[1]) / diff(boundary$vp$yscale)
				boundary$vp = NULL
				boundary$x = unit(x2, "npc")
				boundary$y = unit(y2, "npc")
				boundary
			})
			combine_paths(gframe_path, boundary_npc, fill = o$space.color)
		})
		gt = gt |>
			grid::addGrob(space_fill, paste0("gt_facet_", rc_text))
	}

	if (o$earth_boundary) {
		gt = gt |>
			grid::addGrob(boundary, paste0("gt_facet_", rc_text))
	}

	gt = gt |>
		add_to_gt(gframe, row = rowid, col = colid)
	gts[[facet_page]] = gt

	assign("gts", gts, envir = .TMAP_GRID)
	NULL

}



## functions to create a grob that draws space 'space'
## needed because clipping by earth boundary does not always work (e.g. not for rasterGrobs #1170)

combine_paths = function(pg1, pg2, rule = "winding", fill = "black") {
	# Combine coordinates and assign IDs
	combined_x = unit.c(pg1$x, pg2$x)
	combined_y = unit.c(pg1$y, pg2$y)
	combined_id = c(rep(1, length(pg1$x)), rep(2, length(pg2$x)))

	# Return compound path
	pathGrob(
		x = combined_x,
		y = combined_y,
		id = combined_id,
		rule = rule,
		gp = gpar(fill = fill, col = NA)
	)
}

roundrect_to_pathGrob = function(roundrect, n_corner = 10, fill = "black", col = NA, rule = "winding") {
	stopifnot(inherits(roundrect, "roundrect"))

	# Extract and convert parameters to NPC coordinates
	x0 = convertX(roundrect$x - roundrect$width / 2, "npc", valueOnly = TRUE)
	x1 = convertX(roundrect$x + roundrect$width / 2, "npc", valueOnly = TRUE)
	y0 = convertY(roundrect$y - roundrect$height / 2, "npc", valueOnly = TRUE)
	y1 = convertY(roundrect$y + roundrect$height / 2, "npc", valueOnly = TRUE)
	rx = convertWidth(roundrect$r, "npc", valueOnly = TRUE)
	ry = convertHeight(roundrect$r, "npc", valueOnly = TRUE)

	# Clamp corner radius to avoid overlaps
	rx = min(rx, (x1 - x0) / 2)
	ry = min(ry, (y1 - y0) / 2)

	# Helper: ellipse arc between angles
	arc = function(cx, cy, rx, ry, start, end, n = 10) {
		theta = seq(start, end, length.out = n)
		list(
			x = cx + rx * cos(theta),
			y = cy + ry * sin(theta)
		)
	}

	# Define corner arcs (clockwise from upper-left)
	ul = arc(x0 + rx, y1 - ry, rx, ry, pi / 2, pi, n_corner)
	ur = arc(x1 - rx, y1 - ry, rx, ry, 0, pi / 2, n_corner)
	lr = arc(x1 - rx, y0 + ry, rx, ry, -pi / 2, 0, n_corner)
	ll = arc(x0 + rx, y0 + ry, rx, ry, -pi, -pi / 2, n_corner)

	# Assemble path: arcs and straight edges between them
	x_coords = c(
		rev(ul$x),
		ur$x[n_corner],
		rev(ur$x),
		lr$x[n_corner],
		rev(lr$x),
		ll$x[n_corner],
		rev(ll$x),
		ul$x[n_corner]
	)

	y_coords = c(
		rev(ul$y),
		ur$y[n_corner],
		rev(ur$y),
		lr$y[n_corner],
		rev(lr$y),
		ll$y[n_corner],
		rev(ll$y),
		ul$y[n_corner]
	)

	# Return as filled path grob
	pathGrob(
		x = unit(x_coords, "npc"),
		y = unit(y_coords, "npc"),
		id = rep(1, length(x_coords)),
		rule = rule,
		gp = gpar(fill = fill, col = col)
	)
}


viewport_frame_path = function(fill = "black", col = NA, rule = "winding") {
	# Clockwise path around the viewport (0,0) to (1,1)
	x = unit(rev(c(0, 1, 1, 0, 0)), "npc")
	y = unit(rev(c(0, 0, 1, 1, 0)), "npc")

	pathGrob(
		x = x,
		y = y,
		id = rep(1, length(x)),
		rule = rule,
		gp = gpar(fill = fill, col = col)
	)
}

#' Internal methods for tmap extensions
#'
#' Internal methods for tmap extensions
#'
#' @param ... arguments
#' @param subclass subclass
#' @export
#' @name tmap_internal
#' @keywords internal
tm_element = function(..., subclass = NULL) {
	structure(list(...), class = c(subclass, "tm_element", "list"))
}

#' @export
#' @rdname tmap_internal
tm_element_list = function(...) {
	structure(list(...), class = "tmap")
}

tm_element_list_subclass = function(tml) {
	vapply(tml, function(x) {
		setdiff(class(x), "tm_element")[1]
	}, FUN.VALUE = character(1))
}

tm_element_list_sel = function(tml, subclass) {
	ids = which(tm_element_list_subclass(tml) %in% subclass)
	x = tml[ids]
	attr(x, "ids") = ids
	x
}

#' Stacking of tmap elements
#'
#' The plus operator allows you to stack tmap elements (functions with a prefix `tm_`)
#'
#' @param e1 first tmap element
#' @param e2 second tmap element
#' @name tmap-element
#' @export
"+.tmap" = function(e1, e2) {
	assign("last_map_new", rlang::call_match(), envir = .TMAP)
	if (inherits(e2, "tm_legend")) {
		names(e2) = paste0("legend.", names(e2))

		e2 = tm_element_list(do.call(tm_element, c(e2, list(calls = "v3_tm_legend", subclass = c("tm_legend_v3", "tm_options")))))
	}
	structure(c(e1, e2), class = "tmap")
}

#' Retrieve the last map to be modified or created
#'
#' Retrieve the last map to be modified or created. Works in the same way
#' as `ggplot2::last_plot()`, although there is a difference:
#' `tmap_last()` returns the last call instead of the stacked [`tmap-element`]s.
#'
#' @return call
#' @export
#' @seealso [tmap_save()]
tmap_last = function() {
	.x = get("last_map", envir = .TMAP)
	if (is.null(.x) && get("tmapOptions", envir = .TMAP)$show.warnings) {
		warning("A map has not been created yet.")
	}
	eval(.x)
}

save_last_map = function() {
	lt = get("last_map", envir = .TMAP)
	ltnew = get("last_map_new", envir = .TMAP)
	if (!is.null(ltnew)) lt = replace_last_tmap_by_correct_call(ltnew, lt)
	assign("last_map", lt, envir = .TMAP)
	assign("last_map_new", NULL, envir = .TMAP)
}

replace_last_tmap_by_correct_call = function(mc, lt) {
	if (is.symbol(mc)) {
		mc
	} else if (as.character(mc[1])=="last_map") {
		lt
	} else {
		if (as.character(mc[1]) %in% c("+.tmap", "+")) {
			if (!is.null(mc[[2]])) mc[2] = list(replace_last_tmap_by_correct_call(mc[[2]], lt))
			if (!is.null(mc[[3]])) mc[3] = list(replace_last_tmap_by_correct_call(mc[[3]], lt))
		}
		mc
	}
}

#' Print tm_element
#'
#' @param x x
#' @param ... passed on
#' @export
#' @name print.tm_element
#' @keywords internal
print.tm_element = function(x, ...) {
	cat("tm_element object\n")
	str(x, 2)
}

#' @rdname print.tm_element
#' @export
print.tm_shape = function(x, ...) {
	NextMethod()
}

view_format_popups <- function(id=NULL, titles, format, values) {

	pop_vars = intersect(names(format), names(values))

	format = format[pop_vars]


	# isnull <- vapply(values, is.null, logical(1))
	#
	# titles <- titles[!isnull]
	# titles[names(titles)!=""] <- names(titles)[names(titles)!=""]
	#
	# values <- values[!isnull]

	# islist <- is.list(format) && length(format)>0 && is.list(format[[1]])
	# if (!islist) {
	# 	format <- lapply(1:length(titles), function(i) format)
	# }
	h = lapply(format, function(f) {
		if (f$html.escape) {
			htmltools::htmlEscape
		} else {
			function(x) x
		}
	})
	if (!is.null(id)) {
		labels <- (paste("<b>", id, "</b>", sep=""))
	} else {
		labels <- ""
	}

	titles_format <- mapply(function(ti, hi) {
		hi(ti)
	}, titles, h, SIMPLIFY = FALSE)
	values_format <- mapply(function(v, f, hi) {
		if (inherits(v, "units")) {
			popup_append <- paste0(" ", as.character(attr(v, "units")))
		} else {
			popup_append <- ""
		}
		numbers <- hi(if (is.numeric(v)) do.call("fancy_breaks", c(list(vec=as.numeric(v), intervals=FALSE), f)) else v)
		paste0(numbers, popup_append)
	}, values, format, h, SIMPLIFY = FALSE)


	labels2 <- mapply(function(l, v) {
		paste0("<tr><td style=\"color: #888888;\"><nobr>", l, "</nobr></td><td align=\"right\"><nobr>", v, "</nobr></td>")
	}, titles_format, values_format, SIMPLIFY=FALSE)

	labels3 <- paste0(do.call("paste", c(labels2, list(sep="</tr>"))), "</tr>")

	padding_right <- ifelse(length(titles_format) > 13, 15, 0) # add padding for horizontal scroll bar. These will appear on most browsers when there are over 13 normal lines (tested: RStudio, FF, Chrome)

	x <- paste0("<style> div.leaflet-popup-content {width:auto !important;overflow-y:auto; overflow-x:hidden;}</style><div style=\"max-height:25em;padding-right:", padding_right, "px;\"><table>
			   <thead><tr><th colspan=\"2\">", labels, "</th></thead></tr>", labels3, "</table></div>")

	x
}

process_label_format = function(lf, mlf) {
	if (is.function(lf)) lf = list(fun = lf)

	to_be_assigned = setdiff(names(mlf), names(lf))
	big.num.abbr.set = "big.num.abbr" %in% names(lf)
	lf[to_be_assigned] = mlf[to_be_assigned]
	attr(lf, "big.num.abbr.set") = big.num.abbr.set
	lf
}

# process_popup_format = function(gpf, gtlf, vars, show.warnings) {
# 	# check if g$legend.format is list of lists or functions
# 	islist = is.list(gpf) && length(gpf)>0 && is.list(gpf[[1]])
#
# 	if (!islist) {
# 		process_legend_format(gpf, gtlf, nx=1)
# 	} else {
# 		nms = names(gpf)
# 		if (is.na(vars[1])) {
# 			if (show.warnings) warning("popup.vars not specified whereas popup.format is a list", call. = FALSE)
# 			return(process_legend_format(gpf[[1]], gtlf, nx=1))
# 		}
# 		if (!all(nms %in% vars)) stop("popup.format names do not correspond to popup.vars", call. = FALSE)
# 		lapply(vars, function(v) {
# 			if (v %in% nms) {
# 				process_legend_format(gpf[[v]], gtlf, nx=1)
# 			} else {
# 				process_legend_format(list(), gtlf, nx=1)
# 			}
# 		})
# 	}
# }

#' Export tmap to the format of the used graphics mode
#'
#' * `tmap_grob()` returns a [`grob`][grid::grob()] object (`"plot" mode`)
#' * `tmap_leaflet()` a [`leaflet`][leaflet::leaflet()] object (`"view"` mode).
#'
#' @param x a tmap object.
#' @param asp,scale the desired aspect ratio and scale of the map. Only applicable for `"plot"` mode.
#' @param show show the map?
#' @inheritDotParams print.tmap
#' @return
#' * `tmap_grob()` returns a [`grob`][grid::grob()] object (`"plot"` mode)
#' * `tmap_leaflet()` a [`leaflet`][leaflet::leaflet()] object (`"view"` mode).
#'   In case small multiples are shown, a list is returned.
#' @export
#' @examples
#' map = tm_shape(World) + tm_polygons()
#' tmap_leaflet(map, show = TRUE)
tmap_leaflet = function(x,
						#mode = "view",
						show = FALSE,
						#add.titles = TRUE,
						#in.shiny = FALSE,
						...) {
	current_mode = getOption("tmap.mode")
	on.exit({
		options(tmap.mode = current_mode)
	})
	options(tmap.mode = "view")
	print.tmap(x, show = show, ...)
}

#' @rdname tmap_leaflet
#' @export
tmap_grob = function(x,
					 asp = NA,
					 scale = 1,
					 show = FALSE,
					 ...) {
	current_mode = getOption("tmap.mode")
	current_dev = getOption("tmap.devel.mode")
	on.exit({
		options(tmap.mode = current_mode, tmap.devel.mode = current_dev)
	})
	options(tmap.mode = "plot")
	print.tmap(x + tm_options(asp = asp, scale = scale), show = show, ...)
}

.onLoad = function(...) {
	options(tmap.style = "white", tmap.mode = "plot", mode_pool = NULL, tmap.design.mode = FALSE,
			tmap.devel.mode = FALSE)

	# Resolve the leaflet marker icon on the user's machine rather than at
	# package build time. Calling marker_icon() at top level freezes the
	# build server's leaflet path into the binary (e.g. D:/RCompile/...).
	if (requireNamespace("leaflet", quietly = TRUE)) {
		opts = .defaultTmapOptions
		opts$value.const$shape.markers = marker_icon()
		utils::assignInMyNamespace(".defaultTmapOptions", opts)
	}

	assign("defaultTmapOptions", .defaultTmapOptions, envir = .TMAP)
	assign("tmapOptions", .defaultTmapOptions, envir = .TMAP)

	assign("last_map", NULL, envir = .TMAP)
	assign("last_map_new", NULL, envir = .TMAP)
	# flag for old v3 code
	assign("v3", FALSE, envir = .TMAP)
	assign("tips", sample(.tips), envir = .TMAP)
	assign("tip_nr", 1L, envir = .TMAP)
	assign("tmapStyles", .defaultTmapStyles, envir = .TMAP)
	.TMAP$round_to = as.vector(sapply((-9):9, function(i) {
		sapply(c(1, 2.5, 5), function(j) {
			j*10^i
		})
	})) # needed for pretty ticks for continuous scale with trans enabled (like log scale)
	.TMAP$mode_last = "view"
	.TMAP_GRID$maptiles_urls = character(0) # needed for caching
	fill_providers()
}

.tips <- c(
	"tmap works with {.href [visual variables](https://r-tmap.github.io/tmap/articles/basics_vv)}.",
	"tmap can be {.href [extended](https://r-tmap.github.io/tmap/articles/adv_extensions)} in several ways.",
	"{.href [Charts](https://r-tmap.github.io/tmap/articles/basics_charts)} can be added next to legends.",
	"Layout {.href [styles](https://r-tmap.github.io/tmap/articles/adv_options#setting-options-and-styles)} can be created.",
	"Automatic {.href [map projection](https://r-tmap.github.io/tmap/articles/foundations_crs)} recommendations can be set via {.code tm_crs(crs = {.str auto})}.",
	"Layout can be adjusted to look identical as {.href [ggplot2](https://r-tmap.github.io/tmap/articles/versus_ggplot2#mimicking-ggplot2-layout)}.",
	"In {.str plot} mode, a special design-mode can be enabled via {.code tmap_design_mode()}."
)


#' @rdname tmap_providers
#' @export
.tmap_providers <- new.env(FALSE, parent=globalenv())


#' @export
#' @rdname tmap_internal
.TMAP = new.env(FALSE, parent = globalenv())

#' @export
#' @rdname tmap_internal
.TMAP_LEAFLET = new.env(FALSE, parent = globalenv())

#' @export
#' @rdname tmap_internal
.TMAP_GRID = new.env(FALSE, parent = globalenv())


fill_providers = function() {
	rm(list = ls(envir = .tmap_providers), envir = .tmap_providers)
	y = tmap_providers(as.list = TRUE)
	list2env(y, envir = .tmap_providers)
}

#' Save tmap
#'
#' Save tmap to a file. This can be either a static plot (e.g. png) or an interactive map (html).
#'
#' @details
#'   Extra arguments (`...`) are passed on to the function that writes the
#'   output. For interactive (html) maps this is [htmlwidgets::saveWidget()], or
#'   [widgetframe::saveWidgetframe()] when `in.iframe = TRUE`. For static maps it
#'   is the graphic device: the function supplied to `device`, or, when
#'   `device = NULL`, the default device for the file extension:
#'   [grDevices::pdf()] (`"pdf"`), [grDevices::postscript()] (`"eps"`),
#'   [grDevices::svg()] (`"svg"`), [grDevices::png()] (`"png"`),
#'   [grDevices::jpeg()] (`"jpg"`/`"jpeg"`), [grDevices::bmp()] (`"bmp"`), and
#'   [grDevices::tiff()] (`"tiff"`). The most useful device arguments are listed
#'   under `...`.
#'
#'   Note that `filename`, `width`, and `height` are always set by `tmap_save`,
#'   as are `res` and `units` for the raster formats (png, jpg, bmp, tiff); these
#'   cannot be overridden through `...`. Use the `dpi`, `width`, `height`, and
#'   `units` arguments of `tmap_save` instead.
#'
#' @param tm tmap object
# The docs for filename could be put in tmap_mode.
#' @param filename filename including extension, and optionally the path.
#'   The extensions pdf, eps, svg, wmf (Windows only), png, jpg, bmp, tiff, and html are supported.
#'   If the extension is missing, the file will be saved as a static plot in `"plot"`
#'   mode and as an interactive map (html) in another mode.
#'   The default format for static plots is png, but this can be changed using
#'   the option `"output.format"` in [tmap_options()]. If `NA` (the default),
#'   the file is saved as "tmap01" in the default format, and the number incremented
#'   if the file already exists.
#' @param device graphic device to use. Either a device function
#'   (e.g., [`png`][grDevices::png()] or [`cairo_pdf`][grDevices::cairo_pdf()])
#'   or a text indicating selected graphic device: "pdf", "eps", "svg", "wmf" (Windows only), "png", "jpg", "bmp", "tiff".
#'   If `NULL`, the graphic device is guessed based on the `filename` argument.
#' @param height,width The dimensions of the plot (not applicable for html files).
#'   Units are set with the argument `units`. If one of them is not specified,
#'   this is calculated using the formula asp = width / height, where asp is the
#'   estimated aspect ratio of the map. If both are missing, they are set such
#'   that `width * height` is equal to the option `"output.size"` in [tmap_options()].
#'   This is by default 49, meaning that is the map is a square (so aspect ratio of 1)
#'   both width and height are set to 7.
#' @param units units for width and height (`"in"`, `"cm"`, or `"mm"`).
#'   By default, pixels (`"px"`) are used if either width or height is set to a
#'   value greater than 50. Else, the units are inches (`"in"`).
#' @param dpi dots per inch. Only applicable for raster graphics. By default it
#'   is set to 300, but this can be changed using the option `"output.dpi"` in [tmap_options()].
#' @param outer.margins overrides the outer.margins argument of [tm_options()] (unless set to `NA`)
#' @param asp if specified, it overrides the asp argument of [tm_options()].
#'   **Tip**: set to `0` if map frame should be placed on the edges of the image.
#' @param scale overrides the scale argument of [tm_options()] (unless set to `NA`)
#' @param insets_tm tmap object of an inset map, or a list of tmap objects of
#'   multiple inset maps. The number of tmap objects should be equal to the number
#'   of viewports specified with `insets_vp`.
#' @param insets_vp [`viewport`][grid::viewport()] of an inset map, or a list
#'   of [`viewport`][grid::viewport()]s of multiple inset maps. The number of
#'   viewports should be equal to the number of tmap objects specified with `insets_tm`.
#' @param add.titles add titles to leaflet object.
#' @param in.iframe should an interactive map be saved as an iframe?
#'   If so, two HTML files will be saved; one small parent HTML file with the
#'   iframe container, and one large child HTML file with the actual widget.
#'   See [widgetframe::saveWidgetframe()] for details. By default `FALSE`,
#'   which means that one large HTML file is saved (see [saveWidget()][htmlwidgets::saveWidget()]).
#' @param selfcontained when an interactive map is saved, should the resources
#'   (e.g. JavaScript libraries) be contained in the HTML file? If `FALSE`, they
#'   are placed in an adjacent directory (see also [htmlwidgets::saveWidget()]).
#'   Note that the HTML file will often still be large when `selfcontained = FALSE`,
#'   since the map data (polygons and popups), which are also contained in the HTML file,
#'   usually take more space then the map resources.
#' @param verbose Deprecated. It is now controlled by the tmap option `show.messages`
#'    (see [tmap_options()])
#' @param ... additional arguments passed on to the underlying output function
#'   (see Details). For interactive (html) maps these are passed to
#'   [htmlwidgets::saveWidget()] or [widgetframe::saveWidgetframe()]. For static
#'   maps they are passed to the graphic device; the most useful are:
#'   \describe{
#'     \item{`colormodel` (pdf, eps)}{Color model of the output. The default is
#'       `"srgb"`; use `colormodel = "cmyk"` for print workflows that require
#'       CMYK. See [grDevices::pdf()].}
#'     \item{`compression` (tiff)}{Compression method, e.g. `"lzw"`, `"zip"`, or
#'       `"jpeg"`. The default `"none"` can produce very large files; the
#'       lossless `"lzw"` or `"zip"` is usually a good choice for publication.}
#'     \item{`quality` (jpg)}{JPEG quality, from `0` to `100` (default `75`).}
#'     \item{`bg` (all)}{Background color. Use `bg = "transparent"` for a
#'       transparent background (png, tiff, and svg support transparency; jpg and
#'       bmp do not).}
#'     \item{`pointsize` (all)}{Default point size of text, which scales the text
#'       drawn by the device.}
#'     \item{`family` (pdf, eps, svg), `fonts` (pdf, eps)}{Font family, and fonts
#'       to embed in the file.}
#'     \item{`antialias` (png, jpg, bmp, tiff)}{Type of anti-aliasing, e.g.
#'       `"default"`, `"none"`, `"gray"`, or `"subpixel"`. This only applies to
#'       cairo-type devices. `png` always uses cairo (`type = "cairo-png"`), but
#'       `jpg`, `bmp`, and `tiff` use the platform default device type, so on a
#'       build without cairo support `antialias` is ignored for those formats.}
#'   }
#' @returns the filename, invisibly, if export is successful.
#' @importFrom htmlwidgets saveWidget
#' @import tmaptools
#' @example ./examples/tmap_save.R
#' @export
tmap_save = function(tm=NULL, filename=NA, device=NULL, width=NA, height=NA, units = NA,
					  dpi=NA, outer.margins=NA, asp=NULL, scale=NA, insets_tm=NULL,
					 insets_vp=NULL, add.titles = TRUE, in.iframe = FALSE, selfcontained = !in.iframe, verbose = NULL, ...) {
	.tmapOptions = get("tmapOptions", envir = .TMAP)
	show.warnings =.tmapOptions$show.warnings
	if (!missing(verbose) && show.warnings) warning("The argument verbose is deprecated. Please use the option show.messages of tmap_options instead.")



	verbose = .tmapOptions$show.messages


	lastcall = x = get("last_map", envir = .TMAP)
	if (missing(tm)) {
		tm = suppressWarnings(tmap_last())
		if (is.null(tm)) stop("A map has not been created yet")
		is.arrange = FALSE
	} else if (inherits(tm, "tmap")) {
		is.arrange = FALSE
	} else if (inherits(tm, "tmap_arrange")) {
		is.arrange = TRUE
	} else {
		stop("Unknown format. tm should be either a tmap output, or a list of tmap outputs")
	}

	tmap.mode = getOption("tmap.mode")
	default_ext = ifelse(tmap.mode == "plot", .tmapOptions$output.format, "html")


	if (is.na(filename)) {
		filename_default = paste("tmap01", default_ext, sep = ".")
		if (!file.exists(filename_default)) {
			filename = filename_default
		} else {
			files = list.files(pattern = paste0("^tmap[0-9]{2}\\.", default_ext, "$"))
			fid = max(as.integer(substr(files, 5, 6))) + 1L
			filename = paste0("tmap", sprintf("%02d", fid), ".",  default_ext)
		}
	}

	if (is.na(dpi)) dpi = .tmapOptions$output.dpi

	on.exit({
		assign("last_map", lastcall, envir = .TMAP)
	})



	get_ext = function(filename, default_ext) {
		pieces = strsplit(filename, "\\.")[[1]]
		if (length(pieces)==1) return(default_ext)
		tolower(pieces[length(pieces)])
	}

	convert_to_inches = function(x, units) {
		x = switch(units, px = x/dpi, `in` = x, cm = x/2.54, mm = x/2.54/10)
	}
	convert_to_pixels = function(x, units) {
		x = switch(units, px = x, `in` = dpi*x, cm = dpi*x/2.54, mm = dpi*x/2.54/10)
	}

	ext = get_ext(filename, default_ext)

	interactive = (ext=="html")

	interactive_mode = if (tmap.mode == "plot") "view" else tmap.mode

	options(tmap.mode=ifelse(interactive, interactive_mode, "plot"))

	if (interactive) {
		if (is.arrange) {
			lf = print_tmap_arrange(tm, show = FALSE, add.titles=add.titles)
		} else {
			lf = print.tmap(tm, show = FALSE)
		}

		tryCatch({
			wd = getwd()
			on.exit(setwd(wd), add = TRUE)
			wd_new = dirname(filename)
			base_filename = basename(filename)
			setwd(wd_new)
			if (in.iframe) {
				rlang::check_installed("widgetframe", reason = "for option in.frame.")
				widgetframe::saveWidgetframe(lf, file=base_filename, selfcontained = selfcontained, ...)
			} else {
				htmlwidgets::saveWidget(lf, file=base_filename, selfcontained = selfcontained, ...)
			}
		}, error = function(e) {
			stop("Unable to save the interactive map. Note that saving interactive small multiples is not supported yet. The error message from htmlwidgets::saveWidget is ", call. = FALSE)
		})

		options(tmap.mode=tmap.mode)
		if (verbose) {
			message("Interactive map saved to ", suppressWarnings(normalizePath(filename)))
		}
		return(invisible(filename))
	}

	if (is.na(width) || is.na(height)) {
		if (!is.na(width)) {
			if (is.na(units)) units = choose_unit(width)
			temp_size = convert_to_pixels(width, units)
		} else if (!is.na(height)) {
			if (is.na(units)) units = choose_unit(height)
			temp_size = convert_to_pixels(height, units)
		} else {
			units = "px"
			temp_size = 700
		}

		if (is.arrange) {
			sasp = 1
		} else {
			show.messages = tmap_options(show.messages = FALSE)
			on.exit(tmap_options(show.messages))
			sasp = get_asp_ratio(tm, width = temp_size, height = temp_size, res = dpi)
			tmap_options(show.messages)
		}

		if (is.na(width) && !is.na(height)) {
			width = height * sasp
		} else if (is.na(height)  && !is.na(width)) {
			height = width / sasp
		} else {
			units = "in"
			height = sqrt(.tmapOptions$output.size / sasp)
			width = height * sasp
		}
	} else {
		if (is.na(units)) units = choose_unit(max(width, height))
	}
	units_target = ifelse(units=="px" && ext %in% c("png", "jpg", "jpeg", "bmp", "tiff"), "px", "in")

	if (units_target=="in") {
	  width = convert_to_inches(width, units)
	  height = convert_to_inches(height, units)

	  if (ext=="pdf") {
	    round_to_1_72 = function(x) x %/% (1/72) / 72
	    width = round_to_1_72(width)
	    height = round_to_1_72(height)
	  }
	} else {
		width = convert_to_pixels(width, units)
		height = convert_to_pixels(height, units)
	}
	old_dev = grDevices::dev.cur()
	dev = plot_device(device = device, ext = ext, filename = filename,
					   dpi = dpi, units_target = units_target)

	folder = dirname(filename)
	if (!dir.exists(folder)) dir.create(folder, recursive = TRUE)
	dev(filename = filename, width = width, height = height, ...)

	on.exit(capture.output({
			dev.off()
			if (old_dev > 1) grDevices::dev.set(old_dev) # restore old device unless null device
			}), add = TRUE)

	if (is.arrange) {
		opts = attr(tm, "opts")
		if (!is.na(outer.margins[1])) opts$outer.margins = outer.margins
		if (!missing(asp)) opts$asp = asp

		attr(tm, "opts") = opts
		print(tm)
	} else {
		args = list()
		if (!is.na(outer.margins[1])) args$outer.margins = outer.margins
		if (!missing(asp)) args$asp = asp
		if (!is.na(scale)) args$scale = scale
		args$animate_disable = TRUE
		print(tm + do.call("tm_options", args))
	}

	if (!is.arrange && !missing(insets_tm) && !missing(insets_vp)) {
	  args_inset = if (!is.na(scale)) list(scale = scale) else list()
	  if (inherits(insets_tm, "list") && inherits(insets_vp, "list")) {
	    if (length(insets_tm) != length(insets_vp)) stop("Number of insets unequal to number of viewports")
	    mapply(function(tm_i, vp_i) {
	      print(tm_i + do.call("tm_options", args_inset), vp=vp_i)
	    }, insets_tm, insets_vp)
	  } else if (inherits(insets_tm, "tmap") && inherits(insets_vp, "viewport")) {
	    print(insets_tm + do.call("tm_options", args_inset), vp = insets_vp)
	  } else {
	    stop("Insets and/or its viewports not in the correct format")
	  }
	}

	if (verbose) message("Map saved to ", suppressWarnings(normalizePath(filename)))
	if (ext %in% c("png", "jpg", "jpeg", "bmp", "tiff")) {
		if (units_target == "px") {
			wp = width
			hp = height
			wi = convert_to_inches(width, "px")
			hi = convert_to_inches(height, "px")
		} else {
			wi = width
			hi = height
			wp = convert_to_pixels(width, "in")
			hp = convert_to_pixels(height, "in")
		}
		if (verbose) message("Resolution: ", format(wp), " by ", format(hp), " pixels")
		if (verbose) message("Size: ", format(wi), " by ", format(hi), " inches (", format(dpi), " dpi)")
	} else {
		wi = width
		hi = height
		if (verbose) message("Size: ", format(wi), " by ", format(hi), " inches")
	}

	options(tmap.mode=tmap.mode)

	output = list(filename = filename)
	if (ext %in% c("png", "jpg", "jpeg", "bmp", "tiff")) {
		output$size_pixels = c(width = wp, height = hp)
		output$dpi = dpi
	}
	output$size_inch = c(width = wi, height = hi)

	output = c(output, .TMAP$geo_ref)

	invisible(output)
}

plot_device = function(device, ext, filename, dpi, units_target){

	force(filename)
	force(dpi)
	force(units_target)

	if (is.function(device)) {
		args = formals(device)
		call_args = list()
		if ("file" %in% names(args)) {
			call_args$file = filename
		}
		if ("res" %in% names(args)) {
			call_args$res = dpi
		}
		if ("units" %in% names(args)) {
			call_args$units = units_target
		}
		dev = function(...) do.call(device, utils::modifyList(list(...), call_args))
		return(dev)
	}
	ps = function(..., filename, width, height)
		grDevices::postscript(
			...,
			file = filename,
			width = width,
			height = height,
			onefile = FALSE,
			horizontal = FALSE,
			paper = "special"
		)
	jpeg = function(..., width, height)
		grDevices::jpeg(
			...,
			width = width,
			height = height,
			res = dpi,
			units = units_target
		)
	devices = list(
		ps = ps,
		eps = ps,
		tex = function(..., filename, width, height)
			grDevices::pictex(..., file = filename, width = width, height = height),
		pdf = function(..., filename, version = "1.4")
			grDevices::pdf(..., file = filename, version = version),
		svg = function(..., filename)
			grDevices::svg(..., file = filename),
		wmf = function(..., width, height)
			grDevices::win.metafile(..., width = width, height = height),
		emf = function(..., width, height)
			grDevices::win.metafile(..., width = width, height = height),
		png = function(..., width, height)
				grDevices::png(
					...,
					type = "cairo-png", # several advantages, main: default "cairo" doesn't respect width and height (breaks av animations)
					width = width,
					height = height,
					res = dpi,
					units = units_target
				),
		jpeg = jpeg,
		jpg = jpeg,
		bmp =
			function(..., width, height)
				grDevices::bmp(
					...,
					width = width,
					height = height,
					res = dpi,
					units = units_target
				),
		tiff = function(..., width, height)
			grDevices::tiff(
				...,
				width = width,
				height = height,
				res = dpi,
				units = units_target
			)
	)
	if (is.null(device)) {
		dev = devices[[ext]]
		if (is.null(dev)) {
			stop("'", dev, "'", " graphic device does not exist", call. = FALSE)
		}
		return(dev)
	}
	if (!device %in% names(devices)) {
		stop('"', device, '"', " graphic device does not exist", call. = FALSE)
	}
}

choose_unit = function(x) {
	units = ifelse(x > 50, "px", "in")
	if (x > 15 && x < 100) {
		message("The argument 'units' has been set to \"",
				units, "\" since the specified width or height is ",
				ifelse(units == "px", "greater than ", "less than or equal to "), 50, ". Specify units = \"", ifelse(units == "px", "in", "px"),"\" to change this.")
	}
	units
}

view_format_popups = function(id = NULL, titles, format, values, layout = NULL) {

	layout = complete_popup_layout(layout)

	pop_vars = intersect(names(format), names(values))
	format = format[pop_vars]

	h = lapply(format, function(f) {
		if (f$html.escape) {
			htmltools::htmlEscape
		} else {
			function(x) x
		}
	})

	if (!is.null(id)) {
		labels = paste0("<b>", id, "</b>")
	} else {
		labels = ""
	}

	titles_format = mapply(function(ti, hi) {
		hi(ti)
	}, titles, h, SIMPLIFY = FALSE)

	values_format = mapply(function(v, f, hi) {
		if (inherits(v, "units")) {
			popup_append = paste0(" ", as.character(attr(v, "units")))
		} else {
			popup_append = ""
		}
		numbers = hi(if (is.numeric(v)) do.call("fancy_breaks", c(list(vec = as.numeric(v), intervals = FALSE), f)) else v)
		paste0(numbers, popup_append)
	}, values, format, h, SIMPLIFY = FALSE)

	# Per-region styling (alignment + color), shared with the maplibre backend.
	td_label = paste0("<td class=\"tmap-popup-label\"", .popup_style_attr(layout$label.align, layout$label.color), "><nobr>")
	td_value = paste0("<td class=\"tmap-popup-value\"", .popup_style_attr(layout$value.align, layout$value.color), "><nobr>")
	th_attr  = .popup_style_attr(layout$title.align, layout$title.color)

	labels2 = mapply(function(l, v) {
		paste0("<tr>", td_label, l, "</nobr></td>", td_value, v, "</nobr></td>")
	}, titles_format, values_format, SIMPLIFY = FALSE)

	labels3 = paste0(do.call("paste", c(labels2, list(sep = "</tr>"))), "</tr>")

	padding_right = ifelse(length(titles_format) > 13, 15, 0) # add padding for horizontal scroll bar. These will appear on most browsers when there are over 13 normal lines (tested: RStudio, FF, Chrome)

	# Scrolling is controlled solely by max.height: a length caps the height and
	# the table scrolls past it; "none" means no cap, so the popup grows to fit
	# and (with overflow-y:auto) never shows a scrollbar. max-height and overflow
	# are kept on the SAME (inner) element so they act together.
	inner_style = paste0("max-height:", layout$max.height, ";overflow-y:auto;overflow-x:hidden;padding-right:", padding_right, "px;")

	# When a fixed width is set, let the table fill it so the extra space falls
	# between the label and value columns (value column right-aligned to the
	# popup's right edge) rather than as blank space to the right of the table.
	table_style = if (identical(layout$width, "auto")) "" else " style=\"width:100%;\""

	# Free-form user CSS (tm_popup(css=)) injected after the width rule so it can
	# override anything above. Target classes: .tmap-popup, .tmap-popup-table,
	# .tmap-popup-title, .tmap-popup-label, .tmap-popup-value.
	css_block = if (!is.null(layout$css) && length(layout$css) == 1L && !is.na(layout$css) && nzchar(layout$css)) {
		paste0("<style>", layout$css, "</style>")
	} else {
		""
	}

	paste0(
		"<style> div.leaflet-popup-content {width:", layout$width, " !important;}</style>",
		css_block,
		"<div class=\"tmap-popup\" style=\"", inner_style, "\">",
		"<table class=\"tmap-popup-table\"", table_style, ">",
		"<thead><tr><th class=\"tmap-popup-title\" colspan=\"2\"", th_attr, ">", labels, "</th></tr></thead>",
		labels3,
		"</table></div>"
	)
}

.defaultTmapOptions = structure(
	list(
		# mode specific options or default values
		modes = list(plot = list(name = "Grid",
								 use_gradient = FALSE,
								 crs_basemap = 3857,
								 limit_latitude_3857 = c(-84, 84)),
								 #crs = "auto"),
					 view = list(name = "Leaflet",
					 			use_browser = FALSE,
					 			use_WebGL = NA,
					 			legend.position = tm_pos_in(pos.h = "right", pos.v = "bottom", align.h = "left", align.v = "top", just.h = "left", just.v = "top"),
					 			crs_basemap = list(dimensions = 3857, 4326),
					 			facet.max = 16,
					 			#legend.bg.alpha = 0.8,
					 			#view.legend.position = c("right", "top"),
					 			control.position = c("left", "top"),
					 			control.collapse = TRUE,
					 			basemap.show = TRUE,
					 			set_bounds = FALSE,
					 			set_view = NA,
					 			set_zoom_limits = NA,
					 			qtm.scalebar = TRUE,
					 			qtm.minimap = FALSE,
					 			qtm.mouse_coordinates = TRUE,
					 			use_circle_markers = TRUE,
					 			leaflet.options = list())),

		crs = NA,
		bbox = NULL,

		# facets
		facet.max = 64, # was max.facets
		free.scales = NULL, # for backward compatibility: if this value is set, it will be used to impute the free arguments in the layer functions

		# spatial object class specific options
		raster.max_cells = 1e7, # was max.raster
		raster.warp = TRUE,

		# general
		show.messages = TRUE,
		show.warnings = TRUE,

		# output
		output.format = "png",
		output.size = 49,
		output.dpi = 300,
		animation.dpi = 100,


		# default visual variable values
		value.const = list(fill.symbols = "grey60",
						   fill.dots = "black",
						   fill = "grey85",
						   col.polygons = "grey25",
						   col.symbols = "grey25",
						   col.raster = "grey40",
						   col.text = "black",
						   col = "black",
						   bgcol.labels_highlighted = "white",
						   bgcol = "#00000000",
						   lwd = 1,
						   lty = "solid",
						   text = "Abc",
						   text.labels = "",
						   fontface = NULL,
						   shape.squares = 22,
						   shape.dots = 19,
						   shape.markers = marker_icon(),
						   shape = 21,
						   size.bubbles = 1.3333,
						   size.squares = 1.3333,
						   size.dots = .3,
						   size = 1,
						   fill_alpha = 1,
						   col_alpha = 1,
						   bgcol_alpha = 1,
						   angle = 0,
						   num = 1),
		value.na = list(
			fill = "grey75",
			col = "grey75",
			col.raster = "#00000000",
			bgcol = "grey75",
			lty = "solid",
			text = "Unknown",
			fontface = NULL,
			fill_alpha = 1,
			col_alpha = 1,
			bgcol_alpha = 1,
			col_alpha.raster = 0,
			angle = 0
		),
		value.null = list(
			fill = "grey95",
			col = "grey95",
			col.polygons = "grey40",
			bgcol = "grey95",
			lty = "solid",
			lwd = 1,
			text = "",
			fontface = NULL,
			fill_alpha = 1,
			col_alpha = 1,
			bgcol_alpha = 1,
			shape.squares = 22,
			shape.dots = 19,
			shape = 21,
			size.bubbles = 1.3333,
			size.squares = 1.3333,
			size.dots = .3,
			size = 1,
			angle = 0,
			num = 0
		),
		value.blank = list(
			fill = "#00000000",
			col = "#00000000",
			bgcol = "#00000000",
			lty = "blank",
			lwd = NA,
			text = "",
			fontface = NULL,
			fill_alpha = 0,
			col_alpha = 0,
			bgcol_alpha = 0,
			angle = 0,
			num = 0
		),
		values.var = list(fill = list(seq = "-hcl.blues3", div = "pu_gn_div",
									  unord = "cols4all.area7", ord = "-hcl.blues3", cyc = "tol.rainbow_pu_rd", biv = "pu_gn_bivs"),
						  col = list(seq = "-hcl.blues3", div = "pu_gn_div",
						  		   unord = "cols4all.line7", ord = "-hcl.blues3", cyc = "tol.rainbow_pu_rd", biv = "pu_gn_bivs"),
						  fill.dots = list(seq = "-hcl.blues3", div = "pu_gn_div",
						  				 unord = "cols4all.line7", ord = "-hcl.blues3", cyc = "tol.rainbow_pu_rd", biv = "pu_gn_bivs"),
						  bgcol = list(seq = "-hcl.blues3", div = "pu_gn_div",
						  			 unord = "cols4all.area7", ord = "-hcl.blues3", cyc = "tol.rainbow_pu_rd", biv = "pu_gn_bivs"),
						  size = tm_seq(0, 1, power = "sqrt"),
						  size.bubbles = tm_seq(0, 1, power = "sqrt"),
						  lwd = c(0, 3),
						  lty = c("dashed", "dotted", "dotdash", "longdash", "twodash"),
						  text = LETTERS,
						  fontface = c("plain", "italic", "bold"),
						  fill_alpha = c(0.25, 1),
						  col_alpha = c(0.25, 1),
						  bgcol_alpha = c(0.25, 1),
						  shape = 21:25,
						  area = c(0, 1),
						  xmod = c(0, 1),
						  ymod = c(0, 1),
						  angle = c(0, 360),
						  num = c(0, 1)),
		values.range = list(fill = NA, col = NA, size = c(0, 1), lwd = c(0, 1),
							lty = NA, text = NA, fontface = NA, fill_alpha = NA,
							col_alpha = NA, shape = NA, angle = NA, num = c(0,1)), # NA = automatic, NULL is not applicable
		value.neutral = list(size = 0.75,
							 lwd = 2,
							 lty = "solid",
							 fill_alpha = 1,
							 col_alpha = 1,
							 bgcol_alpha = 1,
							 text = "Abc",
							 fontface = "plain",
							 angle = 0,
							 num = 1),
		values.scale = list(
			1,
			lwd.lines = 1,
			size.symbols = 1,
			size.bubbles = 1.3333,
			size.squares = 1.3333
		),

		# scales
		scales.var = list(fill = list(fact = "categorical", num = "intervals", int = "discrete", datetime = "intervals", date = "intervals"),
						  col = list(fact = "categorical", num = "intervals", int = "discrete", datetime = "intervals", date = "intervals"),
						  bgcol = list(fact = "categorical", num = "intervals", int = "discrete", datetime = "intervals", date = "intervals"),
						  lwd = list(fact = "categorical", num = "continuous", int = "discrete", datetime = "intervals", date = "intervals"),
						  lty = list(fact = "categorical", num = "intervals", datetime = "intervals", date = "intervals"),
						  shape = list(fact = "categorical", num = "intervals", datetime = "intervals", date = "intervals"),
						  size = list(fact = "continuous", num = "continuous", datetime = "continuous", date = "continuous"),
						  fill_alpha = list(fact = "categorical", num = "intervals", datetime = "continuous", date = "continuous"),
						  col_alpha = list(fact = "categorical", num = "intervals", datetime = "continuous", date = "continuous"),
						  bgcol_alpha = list(fact = "categorical", num = "intervals", datetime = "continuous", date = "continuous"),
						  area = list(fact = "categorical", num = "continuous", datetime = "continuous", date = "continuous"),
						  xmod = list(fact = "asis", num = "asis"),
						  ymod = list(fact = "asis", num = "asis"),
						  angle = list(fact = "asis", num = "asis"),
						  text = list(fact = "asis", num = "asis"),
						  fontface = list(fact = "categorical", num = "categorical", datetime = "categorical", date = "categorical")),

		scale.misc.args = list(continuous = list(n = c(fill = 5, col = 5, 5), outliers.trunc = c(FALSE, FALSE), trans = "identity", limits = list(fill = NA, col = NA, 0)),
							   continuous_log = list(n = c(fill = 5, col = 5, 5), outliers.trunc = c(FALSE, FALSE),	limits = list(fill = NA, col = NA, NA)),
							   continuous_log2 = list(n = c(fill = 5, col = 5, 5), outliers.trunc = c(FALSE, FALSE),	limits = list(fill = NA, col = NA, NA)),
							   continuous_log10 = list(n = c(fill = 5, col = 5, 5), outliers.trunc = c(FALSE, FALSE),	limits = list(fill = NA, col = NA, NA)),
							   continuous_log1p = list(n = c(fill = 5, col = 5, 5), outliers.trunc = c(FALSE, FALSE),	limits = list(fill = NA, col = NA, NA)),
							   continuous_sqrt = list(n = c(fill = 5, col = 5, 5), outliers.trunc = c(FALSE, FALSE),	limits = list(fill = NA, col = NA, 0)),
							   continuous_pseudo_log = list(n = c(fill = 5, col = 5, 5), outliers.trunc = c(FALSE, FALSE),	limits = list(fill = NA, col = NA, NA)),
							   rank = list(n = 5,
							   			unit = "rank")), # NA means take data range, 0 means include 0


		continuous.nclass_per_legend_break = 50, # the number of continuous legend breaks within one 'unit' (label).
		continuous.nclasses = 101, # the number of classes of a continuous scale. Should be odd

		# labels
		label.format = list(
			fun = NULL,
			scientific = FALSE,
			digits = NA,
			interval.disjoint = TRUE,
			big.num.abbr = c(mln = 6, bln = 9, trn = 12, qdn = 15, qnt = 18),
			prefix = "",
			suffix = "",
			text.separator = "-",
			text.less.than = "<",
			text.less.than_as.prefix = TRUE,
			text.or.more = "\u2265",
			text.or.more_as.prefix = TRUE,
			text.align = NA,
			text.to.columns = FALSE,
			html.escape = TRUE
		),
		label.na = "Missing",


		###############################3
		# tm_layout options
		###############################3
		scale = 1,
		asp = NA,

		# background
		bg = TRUE,
		bg.color = "white",
		outer.bg = FALSE,
		outer.bg.color = "white",

		# frame
		frame = TRUE,
		frame.color = NA,
		frame.alpha = 1,
		frame.lwd = 1,
		frame.r = NA,
		frame.double_line = FALSE,


		# margins
		outer.margins = rep(0.02, 4),
		inner.margins = list(stars = rep(0, 4), SpatRaster = rep(0, 4), rep(0.02, 4)),
		inner.margins.extra = c(0, 0, 0, 0),
		meta.margins = NA,
		meta.auto_margins = c(0.4, 0.4, 0.4, 0.4),
		between_margin = 0.5,
		panel.margin = c(xtab = 0.4, 0),

		# xlab, ylab
		xlab.show = FALSE,
		xlab.text = "",
		xlab.size = 1,
		xlab.color = "black",
		xlab.rotation = 0,
		xlab.space = 0.5,
		xlab.fontface = NULL,
		xlab.fontfamily = NULL,
		xlab.alpha = 1,
		xlab.side = "bottom",

		ylab.show = FALSE,
		ylab.text = "",
		ylab.size = 1,
		ylab.color = "black",
		ylab.rotation = 0,
		ylab.space = 0.5,
		ylab.fontface = NULL,
		ylab.fontfamily = NULL,
		ylab.alpha = 1,
		ylab.side = "left",


		# panel
		panel.type = NA, # "wrap" or "xtab",
		panel.wrap.pos = "top", # or "left", "right", "bottom"
		panel.xtab.pos = c("left", "top"),

		# data
		unit = "metric",

		# general visual settings

		# colors
		color.sepia_intensity = 0,
		color.saturation = 1,
		color_vision_deficiency_sim = "none",

		# text
		text.fontface = "plain",
		text.fontfamily = "",
		r = 2,

		component.position = list('in' = tm_pos_in(pos.h = "left", pos.v = "top",
											  align.h = "left", align.v = "top", just.h = "left", just.v = "top"),
								  out = tm_pos_out(cell.h = "right", cell.v = "center",
								  		   pos.h = "left", pos.v = "top",
								  		   align.h = "left", align.v = "top", just.h = "left", just.v = "top")),
		component.offset = c(inside = 0.75, INSIDE = 0, outside = 0, OUTSIDE = 0),
		component.stack_margin = c(combined = 0, apart = 0.5),
		component.autoscale = TRUE,
		component.resize_as_group = FALSE,
		component.frame_combine = TRUE,
		component.stack = "vertical",
		legend.stack = c(all = "vertical", per_row = "horizontal", per_col = "horizontal", all_row = "vertical", all_col = "horizontal", manual = "vertical"),
		chart.stack = c(all = "vertical", per_row = "horizontal", per_col = "horizontal", all_row = "vertical", all_col = "horizontal", manual = "vertical"),
		component.equalize = TRUE,

		component.frame = FALSE,
		component.frame.color = NA,
		component.frame.alpha = 1,
		component.frame.lwd = 1,
		component.frame.r = NA,

		component.bg = TRUE,
		component.bg.color = "white",
		component.bg.alpha = 1,

		# legend
		legend.show = TRUE,
		legend.orientation = "portrait",
		legend.position = tm_pos_auto_out(cell.h = "right", cell.v = "bottom",
										  pos.h = "left", pos.v = "top",
										  align.h = "left", align.v = "top", just.h = "left", just.v = "top"),
		legend.width = NA,
		legend.height = NA,
		legend.reverse = FALSE,
		legend.na.show = NA,
		legend.title.color = NA,
		legend.title.size = 0.9,
		legend.title.fontface = NULL,
		legend.title.fontfamily = NULL,
		legend.title.alpha = 1,
		legend.xlab.color = NA,
		legend.xlab.size = 0.9,
		legend.xlab.fontface = NULL,
		legend.xlab.fontfamily = NULL,
		legend.xlab.alpha = 1,
		legend.ylab.color = NA,
		legend.ylab.size = 0.9,
		legend.ylab.fontface = NULL,
		legend.ylab.fontfamily = NULL,
		legend.xlab.rot = 0,
		legend.ylab.rot = 0,
		legend.ylab.alpha = 1,
		legend.text.color = NA,
		legend.text.size = 0.7,
		legend.text.fontface = NULL,
		legend.text.fontfamily = NULL,
		legend.text.alpha = 1,
		legend.frame = TRUE,
		legend.frame.color = NA,
		legend.frame.alpha = 1,
		legend.frame.lwd = 1,
		legend.frame.r = NA,
		legend.bg = TRUE,
		legend.bg.color = "white",
		legend.bg.alpha = 1,
		legend.only = FALSE,
		legend.absolute_fontsize = 14,
		legend.settings.portrait = list(item.height = c(rect = 1.2, symbols = 1, gradient = 3, lines = 1.2, text = 1.2, bivariate = 1.2),
												 item.width = c(rect = 1.2, symbols = 1, gradient = 1.2, lines = 1.2, text = 3, bivariate = 1.2),
												 item.r = NA,
												 item.space = c(rect = 0.2, symbols = 0.2, gradient = 0, lines = 0.2, text = 0.2, bivariate = 0),
												 item.na.height = c(rect = NA, symbols = NA, gradient = 1.2, lines = NA, text = NA, bivariate = NA),
												 item.na.width = c(rect = NA, symbols = NA, gradient = 1.2, lines = NA, text = NA, bivariate = NA),
												 item.na.space = c(rect = 0.2, symbols = 0.3, gradient = 1, lines = 0.2, text = 0.2, bivariate = 0.2),
												 item.shape = 107,
												 title.padding  = c(0, 0, 0.25, 0),
												 xlab.padding = c(0, 0, 0.25, 0),
												 ylab.padding = c(0, 0, 0.25, 0),
												 title.align = c(bivariate = "right", "left"),
												 xlab.align = "left",
												 ylab.align = "center",
												 ticks = list(rect = list(), symbols = list(), gradient = list(c(0.85, 1.1)), lines = list(), text = list(), bivariate = list()),
												 ticks.disable.na = c(rect = FALSE, symbols = FALSE, gradient = TRUE, lines = FALSE, text = FALSE, bivariate = TRUE),
												 ticks.col = NA,
												 ticks.lwd = 1,
												 margins = c(0.4, 0.4, 0.4, 0.4),
												 item_text.margin = c(gradient = 0.5, 0.25)),
		legend.settings.landscape = list(item.height = c(rect = 1, symbols = 1, gradient = 1.2, lines = 1, text = 1),
												  item.width = c(rect = 6, symbols = 3, gradient = 6, lines = 6, text = 6),
												  item.r = NA,
												  item.space = c(rect = 0.2, symbols = 0.3, gradient = 0, lines = 0.2, text = 0.2),
												  item.na.height = c(rect = NA, symbols = NA, gradient = 2, lines = NA, text = NA),
												  item.na.width = c(rect = NA, symbols = NA, gradient = 4, lines = NA, text = NA),
												  item.na.space = c(rect = 0.2, symbols = 0.3, gradient = 0.3, lines = 0.2, text = 0.2),
												  item.shape = 107,

												  title.padding  = c(0, 0, 0.25, 0),
												  xlab.padding = c(0, 0, 0.25, 0),
												  ylab.padding = c(0, 0, 0.25, 0),
												  title.align = c(bivariate = "right", "left"),
												  xlab.align = "left",
												  ylab.align = "center",
												  ticks = list(rect = list(), symbols = list(), gradient = list(c(0.85, 1.1)), lines = list(), text = list()),
												  ticks.disable.na = c(rect = FALSE, symbols = FALSE, gradient = TRUE, lines = FALSE, text = FALSE),
												  ticks.col = NA,
												  ticks.lwd = 1,
												  margins = c(0.4, 0.4, 0.4, 0.4),
												  item_text.margin = 0.25),

		#
		add_legend.position = NULL,

		# charts
		chart.show = TRUE,
		chart.plot.axis.x = FALSE,
		chart.plot.axis.y = TRUE,
		chart.position = tm_pos_auto_out(cell.h = "right", cell.v = "bottom",
										 pos.h = "left", pos.v = "top",
										 align.h = "left", align.v = "bottom", just.h = "left", just.v = "top"),
		chart.width = c(histogram.min = 10,
						histogram.max = 20,
						bar.min = 10,
						bar.max = 20,
						donut.min = 10,
						donut.max = 10,
						heatmap.min = 10,
						heatmap.max = 15),
		chart.height = c(histogram.min = 10,
						 histogram.max = 10,
						 bar.min = 10,
						 bar.max = 10,
						 donut.min = 10,
						 donut.max = 10,
						 heatmap.min = 10,
						 heatmap.max = 10),
		chart.reverse = FALSE,
		chart.na.show = NA,
		chart.title.color = NA,
		chart.title.size = 0.9,
		chart.title.fontface = NULL,
		chart.title.fontfamily = NULL,
		chart.title.alpha = NULL,
		chart.xlab.color = NA,
		chart.xlab.size = 0.9,
		chart.xlab.fontface = NULL,
		chart.xlab.fontfamily = NULL,
		chart.xlab.alpha = NULL,
		chart.ylab.color = NA,
		chart.ylab.size = 0.9,
		chart.ylab.fontface = NULL,
		chart.ylab.fontfamily = NULL,
		chart.ylab.alpha = 1,
		chart.text.color = NA,
		chart.text.size = 0.7,
		chart.text.fontface = NULL,
		chart.text.fontfamily = NULL,
		chart.text.alpha = 1,
		chart.frame = TRUE,
		chart.frame.color = NA,
		chart.frame.alpha = 1,
		chart.frame.lwd = 1,
		chart.frame.r = NA,
		chart.bg = TRUE,
		chart.bg.color = "white",
		chart.bg.alpha = 1,
		chart.object.color = "#DDDDDD",

		# components
		title.size = 1.3,
		title.color = NA,
		title.fontface = NULL,
		title.fontfamily = NULL,
		title.alpha = 1,
		title.padding = c(0.25, 0.25, 0.25, 0.25),
		title.frame = FALSE,
		title.frame.color = NA,
		title.frame.alpha = 1,
		title.frame.lwd = 1,
		title.frame.r = NA,
		title.position = tm_pos_out(cell.h = "center", cell.v ="top", pos.h = "left", pos.v = "top", align.h = "left", align.v = "top", just.h = "left", just.v = "top"),
		title.width = NA,
		credits.size = .7,
		credits.color = NA,
		credits.fontface = NULL,
		credits.fontfamily = NULL,
		credits.alpha = 1,
		credits.padding = c(0.25, 0.25, 0.25, 0.25),
		credits.position = tm_pos_in(pos.h = "right", pos.v = "bottom", align.h = "right", align.v = "top", just.h = "left", just.v = "top"),
		credits.width = NA,
		credits.height = NA,
		compass.north=0,
		compass.type="arrow",
		compass.text.size=.8,
		compass.size=NA,
		compass.show.labels=1,
		compass.cardinal.directions=c("N", "E", "S", "W"),
		compass.text.color=NA,
		compass.color.dark=NA,
		compass.color.light=NA,
		compass.lwd=1,
		compass.margins = c(0.25, 0.25, 0.25, 0.25),
		compass.position = tm_pos_in(pos.h = "right", pos.v = "bottom", align.h = "right", align.v = "top", just.h = "left", just.v = "top"),
		# standard arguments:
		inset.position = tm_pos_in(pos.h = "right", pos.v = "bottom", align.h = "right", align.v = "top", just.h = "left", just.v = "top"),

		logo.height = 3,
		logo.margins = c(0.2, 0.2, 0.2, 0.2),
		logo.between_margin = 0.2,
		logo.position = tm_pos_in(pos.h = "right", pos.v = "bottom", align.h = "right", align.v = "top", just.h = "left", just.v = "top"),


		inset.height = 3,
		inset.width = 3,
		inset.margins = c(0.2, 0.2, 0.2, 0.2),
		inset.between_margin = 0.2,
		inset.position = tm_pos_in(pos.h = "right", pos.v = "bottom", align.h = "right", align.v = "top", just.h = "left", just.v = "top"),
		inset.frame = TRUE,
		inset.bg = TRUE,
		inset.bg.color = "#ffffff",
		inset.bg.alpha = 1,

		inset.box_frame = TRUE,
		inset.box_frame.color = "#ee2211",
		inset.box_frame.alpha = 1,
		inset.box_frame.lwd = 2,
		inset.box_frame.lty = "solid",
		inset.box_bg = FALSE,
		inset.box_bg.color = "#ffffff",
		inset.box_bg.alpha = 1,

		inset.main_frame = TRUE,
		inset.main_frame.r = NA,
		inset.main_frame.color = NA,
		inset.main_frame.alpha = 1,
		inset.main_frame.lwd = 2,

		inset_map.height = 7,
		inset_map.width = 7,
		inset_map.margins = c(0.2, 0.2, 0.2, 0.2),
		inset_map.between_margin = 0.2,
		inset_map.position = tm_pos_in(pos.h = "right", pos.v = "bottom", align.h = "right", align.v = "top", just.h = "left", just.v = "top"),
		inset_map.frame = FALSE,

		inset_tmap.height = 7,
		inset_tmap.width = 7,

		inset_grob.height = 7,
		inset_grob.width = 7,

		inset_gg.height = 7,
		inset_gg.width = 7,

		scalebar.breaks = NULL,
		scalebar.width = 12,
		scalebar.allow_clipping = FALSE,
		scalebar.text.size = .5,
		scalebar.text.color = NA,
		scalebar.text.fontface = NULL,
		scalebar.text.fontfamily = NULL,
		scalebar.color.dark = NA,
		scalebar.color.light = NA,
		scalebar.lwd = 1,
		scalebar.size = NULL,
		scalebar.margins = c(0.01,0.01,0.01,0.01),

		# standard arguments:
		scalebar.position = tm_pos_in(pos.h = "right", pos.v = "bottom", align.h = "right", align.v = "top", just.h = "left", just.v = "top"),

		grid.show = FALSE,
		grid.labels.pos = c("left", "bottom"),
		grid.x=NA,
		grid.y=NA,
		grid.n.x=NA,
		grid.n.y=NA,
		grid.crs=NA,
		grid.col=NA,
		grid.lwd=1,
		grid.alpha=NA,
		grid.labels.show=TRUE,
		grid.labels.size=.6,
		grid.labels.col=NA,
		grid.labels.fontface=NULL,
		grid.labels.fontfamily=NULL,
		grid.labels.rot = c(0, 0),
		grid.labels.format = list(big.mark = ","),
		grid.labels.cardinal = FALSE,
		grid.labels.margin.x=0,
		grid.labels.margin.y=0,
		grid.labels.space.x=NA,
		grid.labels.space.y=NA,
		grid.labels.inside_frame=FALSE,
		grid.ticks = TRUE, #labels.show & !labels.inside_frame,
		grid.lines = TRUE,
		grid.ndiscr = 100,


		# standard arguments:
		# mouse.stack = "vertical",
		# mouse.position = tm_pos_in(pos.h = "right", pos.v = "bottom", align.h = "left", align.v = "top", just.h = "left", just.v = "top"),
		# mouse.frame = FALSE,
		# mouse.frame.lwd = 1,
		# mouse.frame.r = 2,
		# mouse.group.frame = TRUE,
		# mouse.resize_as_group = FALSE,
		#

		mouse_coordinates.position = tm_pos_in(pos.h = "right", pos.v = "bottom", align.h = "right", align.v = "top", just.h = "left", just.v = "top"),

		minimap.server = NA,
		minimap.toggle = TRUE,
		minimap.position = tm_pos_in(pos.h = "right", pos.v = "bottom", align.h = "right", align.v = "top", just.h = "left", just.v = "top"),
		minimap.height = 7,
		minimap.width = 7,
		minimap.margins = c(0.2, 0.2, 0.2, 0.2),
		minimap.between_margin = 0.2,
		minimap.frame = FALSE,
		minimap.bg = TRUE,


		panel.show = NA, # depends on type: TRUE for facets
		panel.labels = NA,
		panel.label.size = 1,
		panel.label.color = "black",
		panel.label.fontface = NULL,
		panel.label.fontfamily = NULL,
		panel.label.alpha = 1,
		panel.label.bg = TRUE,
		panel.label.bg.color = "grey80",
		panel.label.bg.alpha = 1,
		panel.label.frame = TRUE,
		panel.label.frame.color = NA,
		panel.label.frame.alpha = 1,
		panel.label.frame.lwd = 1,
		panel.label.frame.r = NA,
		panel.label.height = 1,
		panel.label.rot = c(90, 0, 270, 0),

		qtm.scalebar = FALSE,
		qtm.minimap = FALSE,
		qtm.mouse_coordinates = FALSE,

		earth_boundary = FALSE,
		earth_boundary.color = NA,
		earth_boundary.lwd = 1,
		earth_datum = "OGC:CRS84",
		space = TRUE,
		space.color = "#ffffff",
		space_overlay = NA,

		check_and_fix = FALSE,

		basemap.show = FALSE,
		basemap.server = c("Esri.WorldGrayCanvas", "OpenStreetMap", "Esri.WorldTopoMap"),
		basemap.alpha = 1,
		basemap.zoom = NA,

		tiles.show = FALSE,
		tiles.server = "",
		tiles.alpha = 1,
		tiles.zoom = NA,

		attr.color = "black",

		crs_extra = "",
		crs_global = "+proj=eqearth"
	),
	style = "white",
	specified = character()
)

styles = list(
	v3 = list(
		modes = list(view = list(legend.bg.alpha = 0.8)),
		value.na = list(
			fill = "grey75",
			col = "grey75",
			col.raster = "#00000000",
			lty = "solid",
			lwd = NA,
			text = "Unknown",
			fontface = "plain",
			fill_alpha = 1,
			col_alpha = 1,
			col_alpha.raster = 0
		),
		value.null = list(
			fill = "grey95",
			col = "grey95",
			col.polygons = "grey40",
			lty = "solid",
			lwd = 0.2,
			text = "",
			fontface = "plain",
			fill_alpha = 1,
			col_alpha = 1,
			size = 0.2
		),
		value.blank = list(
			fill = "#00000000",
			col = "#00000000",
			lty = "blank",
			lwd = 0,
			text = "",
			fontface = "plain",
			fill_alpha = 0,
			col_alpha = 0
		),
		values.var = list(fill = list(seq = "brewer.yl_or_br", div = "brewer.rd_yl_gn", unord = "brewer.set3", ord = "brewer.yl_or_br"),
						  col = list(seq = "brewer.yl_or_br", div = "brewer.rd_yl_gn", unord = "brewer.set3", ord = "brewer.yl_or_br")),
		frame.lwd = 1,
		frame.r = 0,
		title.position = tm_pos_in(pos.h = "left", pos.v = "top", align.h = "left", align.v = "top", just.h = "left", just.v = "top"),
		title.size = 1.3,
		legend.position = tm_pos_auto_in(align.h = "left", align.v = "top", just.h = "left", just.v = "top"),
		legend.text.size = 0.7,
		legend.title.size = 0.9,
		legend.frame = FALSE,
		legend.frame.r = 0,
		legend.settings.portrait = list(item.height = c(rect = 1, symbols = 1, gradient = 1, bivariate = 1),
												 item.width = c(rect = 1, symbols = 1, gradient = 1.2, bivariate = 1),
												 item.r = 0,
												 item.space = c(rect = 0, symbols = 0, gradient = 0, bivariate = 0),
												 item.na.height = c(rect = NA, symbols = NA, gradient = 1.2, bivariate = 1),
												 item.na.width = c(rect = NA, symbols = NA, gradient = 1.2, bivariate = 1),
												 item.na.space = c(rect = 0, symbols = 0, gradient = 0, bivariate = 0),
												 title.padding  = c(0, 0, 0.25, 0),
												 ticks = list(rect = list(), symbols = list(), gradient = list(), bivariate = list()),
												 ticks.disable.na = c(rect = FALSE, symbols = FALSE, gradient = TRUE, bivariate = TRUE),
												 ticks.col = NA,
												 ticks.lwd = 1.5,
												 margins = c(0.4, 0.4, 0.4, 0.4),
												 item_text.margin = 0.25),
		legend.settings.landscape = list(item.height = c(rect = 1, symbols = 1, gradient = 1.2),
												  item.width = c(rect = 6, symbols = 3, gradient = 6),
												  item.r = 0,
												  item.space = c(rect = 0.2, symbols = 0.3, gradient = 0),
												  item.na.height = c(rect = NA, symbols = NA, gradient = 2),
												  item.na.width = c(rect = NA, symbols = NA, gradient = 6),
												  item.na.space = c(rect = 0.2, symbols = 0.3, gradient = 0.3),
												  title.padding  = c(0, 0, 0.25, 0),
												  ticks = list(rect = list(), symbols = list(), gradient = list(c(0.8, 1))),
												  ticks.disable.na = c(rect = FALSE, symbols = FALSE, gradient = TRUE),
												  ticks.col = NA,
												  ticks.lwd = 1.5,
												  margins = c(0.4, 0.4, 0.4, 0.4),
												  item_text.margin = 0.25)
	),
	gray = list(
		bg.color = "grey85",
		value.const = list(fill = "grey70",
						   fill.dots = "black",
						   fill = "grey50",
						   col.polygons = "grey20",
						   col = "black")
	),
	grey = list(
		bg.color = "grey85",
		value.const = list(fill = "grey70",
						   fill.dots = "black",
						   fill = "grey50",
						   col.polygons = "grey20",
						   col = "black")
	),
	natural = list(
		bg.color = "lightskyblue1",
		value.const = list(fill.polygons = "darkolivegreen3",
						   fill.dots = "firebrick",
						   fill = "tomato2",
						   col.lines = "steelblue",
						   col.text = "white",
						   col = "black"),
		value.na = list(
			fill = "white",
			col = "white",
			col.raster = "white"),
		value.null = list(
			fill = "grey70",
			col = "grey70",
			col.polygons = "grey70"),
		values.var = list(fill = list(seq = "brewer.yl_gn", div = "brewer.rd_yl_gn", unord = "brewer.set3", ord = "brewer.yl_gn"),
						  col = list(seq = "brewer.yl_gn", div = "brewer.rd_yl_gn", unord = "brewer.set3", ord = "brewer.yl_gn")),
		attr.color = "black",
		space.color = "white",
		legend.frame = TRUE,
		legend.bg.color = "grey90",
		frame = FALSE,
		earth_boundary = FALSE,
		basemap.server = "Esri.NatGeoWorldMap",
		basemap.alpha = 1),
	cobalt = list(bg.color = "#002240",
				  outer.bg = TRUE,
				  outer.bg.color = "#002240",
				  space.color = "#002240",
				  value.const = list(fill.polygons = "#0088FF",
				  				   fill = "#FF9D00",
				  				   col.text = "white",
				  				   col = "#FFEE80",
				  				   bgcol.labels_highlighted = "#002240",
				  				   bgcol = "#00000000"),
				  value.na = list(
				  	fill = "grey60",
				  	col = "grey60",
				  	col.raster = "grey60"),
				  value.null = list(
				  	fill = "grey40",
				  	col = "grey40",
				  	col.polygons = "grey40"),
				  values.var = list(fill = list(seq = "brewer.yl_gn", div = "brewer.rd_yl_gn", unord = "brewer.set3", ord = "brewer.yl_gn"),
				  				  col = list(seq = "brewer.yl_gn", div = "brewer.rd_yl_gn", unord = "brewer.set3", ord = "brewer.yl_gn")),
				  attr.color = "white",
				  legend.bg.color = "#002240",
				  chart.text.color = "white",
				  chart.title.color = "white",
				  basemap.server = "CartoDB.DarkMatter",
				  basemap.alpha = .5),
	albatross = list(bg.color = "#00007F",
					 outer.bg = TRUE,
					 outer.bg.color = "#000000",
					 space.color = "#000000",
					 value.const = list(fill.polygons = "#4C4C88",
					 				   fill = "#BFBFFF",
					 				   col.lines = "#BFBFFF",
					 				   col.text = "#FFE700",
					 				   col = "#00004C",
					 				   bgcol.labels_highlighted = "#00007F",
					 				   bgcol = "#00000000"),
					 value.na = list(
					 	fill = "grey60",
					 	col = "grey60",
					 	col.raster = "grey60"),
					 value.null = list(
					 	fill = "#4C4C88",
					 	col = "#4C4C88",
					 	col.polygons = "#4C4C88"),
					 values.var = list(fill = list(seq = "brewer.yl_or_rd", div = "brewer.rd_yl_gn", unord = "brewer.set3", ord = "brewer.yl_or_rd"),
					 				  col = list(seq = "brewer.yl_or_rd", div = "brewer.rd_yl_gn", unord = "brewer.set3", ord = "brewer.yl_or_rd")),
					 attr.color = "#BFBFFF",
					 legend.bg.color = "#00007F",
					 basemap.alpha = .5),
	beaver = list(bg.color = "#FFFFFF",
				  value.const = list(fill.polygons = "#FFE200",
				  				   fill = "#A30000",
				  				   col.lines = "#A30000",
				  				   col = "#00004C",
				  				   bgcol.labels_highlighted = "#FFFFFF",
				  				   bgcol = "#00000000"),
				  value.na = list(
				  	fill = "grey80",
				  	col = "grey80",
				  	col.raster = "grey80"),
				  value.null = list(
				  	fill = "grey95",
				  	col = "grey95",
				  	col.polygons = "grey95"),
				  values.var = list(fill = list(seq = "brewer.yl_or_br", div = "brewer.rd_yl_gn", unord = "brewer.dark2", ord = "brewer.yl_or_br"),
				  				  col = list(seq = "brewer.yl_or_br", div = "brewer.rd_yl_gn", unord = "brewer.dark2", ord = "brewer.yl_or_br")),
				  attr.color = "black"),
	bw = list(color.saturation = 0),
	classic = list(color.sepia_intensity = .7,
				   text.fontfamily = "serif",
				   frame = TRUE,
				   frame.double_line = TRUE,
				   inset.type = "rose"),
	watercolor = list(value.const = list(fill = "#D95F02",
										 fill.dots = "red",
										 col.lines = "red",
										 col = "black",
										 bgcol.labels_highlighted = "white",
										 bgcol = "#00000000"),
					  value.na = list(
					  	fill = "grey80",
					  	col = "grey80",
					  	col.raster = "grey80"),
					  value.null = list(
					  	fill = "#FDCDAC",
					  	col = "#FDCDAC",
					  	col.polygons = "#FDCDAC"),
					  values.var = list(fill = list(seq = "brewer.greens", div = "brewer.pi_yg", unord = "brewer.pastel1", ord = "brewer.greens"),
					  				  col = list(seq = "brewer.greens", div = "brewer.pi_yg", unord = "brewer.pastel1", ord = "brewer.greens")),
					  basemap.server = "Stadia.StamenWatercolor")
)

.defaultTmapStyles = list(
	gray = styles$gray,
	grey = styles$grey,
	natural = styles$natural,
	cobalt = styles$cobalt,
	albatross = styles$albatross,
	beaver = styles$beaver,
	bw = styles$bw,
	classic = styles$classic,
	watercolor = styles$watercolor,
	v3 = styles$v3,
	gray_v3 = c(styles$v3[setdiff(names(styles$v3), names(styles$gray))], styles$gray),
	grey_v3 = c(styles$v3[setdiff(names(styles$v3), names(styles$grey))], styles$grey),
	natural_v3 = c(styles$v3[setdiff(names(styles$v3), names(styles$natural))], styles$natural),
	cobalt_v3 = c(styles$v3[setdiff(names(styles$v3), names(styles$cobalt))], styles$cobalt),
	albatross_v3 = c(styles$v3[setdiff(names(styles$v3), names(styles$albatross))], styles$albatross),
	beaver_v3 = c(styles$v3[setdiff(names(styles$v3), names(styles$beaver))], styles$beaver),
	bw_v3 = c(styles$v3[setdiff(names(styles$v3), names(styles$bw))], styles$bw),
	classic_v3 = c(styles$v3[setdiff(names(styles$v3), names(styles$classic))], styles$classic),
	watercolor_v3 = c(styles$v3[setdiff(names(styles$v3), names(styles$watercolor))], styles$watercolor)
)

#' @export
#' @rdname tmapGridLeaflet
tmapGridDataPlot.tm_data_symbols = function(a, shpTM, dt, gp, bbx, facet_row, facet_col, facet_page, id, pane, group, glid, o, ...) {
	rc_text = frc(facet_row, facet_col)

	g = get("g", .TMAP_GRID)

	# COPIED FROM tmapGridText -> function
	# calculate native per line
	wIn = g$colsIn[g$cols_facet_ids[facet_col]]
	hIn = g$rowsIn[g$rows_facet_ids[facet_row]]

	wNative = bbx[3] - bbx[1]
	hNative = bbx[4] - bbx[2]

	xIn = wNative / wIn
	yIn = hNative / hIn

	lineIn = convertHeight(unit(1, "lines"), "inch", valueOnly = TRUE)


	res = select_sf(shpTM, dt[!is.na(dt$size) & !is.na(dt$shape), ])
	shp = res$shp
	dt = res$dt

	gp = impute_gp(gp, dt)
	gp = swap_pch_15_20(gp)

	#gp = get_pch_1000p(gp)

	gp = rescale_gp(gp, o$scale_down)

	#gp = gp_to_gpar(gp, sel = "all")

	diffAlpha = !any(is.na(c(gp$fill_alpha, gp$col_alpha))) && !(length(gp$fill_alpha) == length(gp$col_alpha) && all(gp$fill_alpha == gp$col_alpha))

	coords = sf::st_coordinates(shp)

	# in case shp is a multipoint (point_per == "segment"), expand gp:
	cp = expand_coords_gp(coords, gp, ndt = nrow(dt))
	coords = cp$coords
	gp = cp$gp

	if (diffAlpha) {
		gp1 = gp_to_gpar(gp, sel = "fill", o = o, type = "symbols")
		gp2 = gp_to_gpar(gp, sel = "col", o = o, type = "symbols")

		grb1 = grid::pointsGrob(x = grid::unit(coords[,1], "native"), y = grid::unit(coords[,2], "native"), pch = gp$shape, size = grid::unit(gp$size, "lines"), gp = gp1, name = paste0("symbols_", id))
		grb2 = grid::pointsGrob(x = grid::unit(coords[,1], "native"), y = grid::unit(coords[,2], "native"), pch = gp$shape, size = grid::unit(gp$size, "lines"), gp = gp2, name = paste0("symbols_borders_", id))
		grb = grid::grobTree(grb1, grb2)
	} else {
		gp = gp_to_gpar(gp, sel = "all", o = o, type = "symbols")

		if (any(!is.na(gp$shape) & gp$shape>999)) {
			shapeLib = get("shapeLib", envir = .TMAP)
			justLib = get("justLib", envir = .TMAP)

			grobs <- lapply(1:length(gp$shape), function(i) {
				shi = gp$shape[[i]]

				just = if (is.na(a$just[1])) c(0.5, 0.5) else a$just
				jst = if (!is.na(shi) && shi>999) {
					js = justLib[[shi-999]]
					if (!is.na(js[1])) {
						js
					} else if (!is.na(a$just[1])) {
						a$just
					} else {
						just
					}
				} else just

				justs.x = jst[1]
				justs.y = jst[2]
				justx = xIn * lineIn * (justs.x-.5)
				justy = yIn * lineIn * (justs.y-.5)



				if (!is.na(shi) && shi>999) {
					size = gp$size[i]*9/10 * a$icon.scale
					lwdid = if (length(gp$lwd) == 1L) 1 else i
					grbs <- gList(shapeLib[[shi-999]])
					gTree(children=grbs, vp=viewport(x=grid::unit(coords[i,1] + justx * size, "native"),
													 y=grid::unit(coords[i,2] + justy * size, "native"),
													 width=unit(size, "lines"),
													 height=unit(size, "lines")))
				} else {
					gpi = structure(lapply(gp, function(x) {
						if (length(x) == 1L) {
							x
						} else {
							x[i]
						}
					}), class = "gpar")
					pointsGrob(x=grid::unit(coords[i,1] + justx * gp$size[i], "native"),
							   y=grid::unit(coords[i,2] + justx * gp$size[i], "native"),
							   size=unit(gp$size[i], "lines"),
							   pch=shi,
							   gp=gpi)
				}
			})
			grb = gTree(children=do.call(gList, grobs), name=paste0("symbols_", id))
		} else {
			grb = grid::pointsGrob(x = grid::unit(coords[,1], "native"), y = grid::unit(coords[,2], "native"), pch = gp$shape, size = grid::unit(gp$size, "lines"), gp = gp, name = paste0("symbols_", id))
		}


	}


	gts = get("gts", .TMAP_GRID)
	gt = gts[[facet_page]]

	gt_name = paste0("gt_facet_", rc_text)

	gt = grid::addGrob(gt, grb, gPath = grid::gPath(gt_name))

	gts[[facet_page]] = gt
	assign("gts", gts, envir = .TMAP_GRID)
	NULL

}


#' @export
#' @rdname tmapGridLeaflet
tmapGridDataPlot.tm_data_dots = function(a, shpTM, dt, gp, bbx, facet_row, facet_col, facet_page, id, pane, group, glid, o, ...) {
	NextMethod()
}


#' @export
#' @rdname tmapGridLeaflet
tmapGridDataPlot.tm_data_bubbles = function(a, shpTM, dt, gp, bbx, facet_row, facet_col, facet_page, id, pane, group, glid, o, ...) {
	NextMethod()
}


#' @export
#' @rdname tmapGridLeaflet
tmapGridDataPlot.tm_data_squares = function(a, shpTM, dt, gp, bbx, facet_row, facet_col, facet_page, id, pane, group, glid, o, ...) {
	NextMethod()
}

#' @export
#' @rdname tmapGridLeaflet
tmapGridDataPlot.tm_data_markers = function(a, shpTM, dt, gp, bbx, facet_row, facet_col, facet_page, id, pane, group, glid, o, ...) {
	NextMethod()
}

#' Export tmap to the format of the used graphics mode
#'
#' * `tmap_grob()` returns a [`grob`][grid::grob()] object (`"plot" mode`)
#' * `tmap_leaflet()` a [`leaflet`][leaflet::leaflet()] object (`"view"` mode).
#'
#' @param x a tmap object.
#' @param asp,scale the desired aspect ratio and scale of the map. Only applicable for `"plot"` mode.
#' @param show show the map?
#' @inheritDotParams print.tmap
#' @return
#' * `tmap_grob()` returns a [`grob`][grid::grob()] object (`"plot"` mode)
#' * `tmap_leaflet()` a [`leaflet`][leaflet::leaflet()] object (`"view"` mode).
#'   In case small multiples are shown, a list is returned.
#' @export
#' @examples
#' map = tm_shape(World) + tm_polygons()
#' tmap_leaflet(map, show = TRUE)
tmap_leaflet = function(x,
						#mode = "view",
						show = FALSE,
						#add.titles = TRUE,
						#in.shiny = FALSE,
						...) {
	current_mode = getOption("tmap.mode")
	on.exit({
		options(tmap.mode = current_mode)
	})
	options(tmap.mode = "view")
	print.tmap(x, show = show, ...)
}

#' @rdname tmap_leaflet
#' @export
tmap_grob = function(x,
					 asp = NA,
					 scale = NA,
					 show = FALSE,
					 ...) {
	current_mode = getOption("tmap.mode")
	current_dev = getOption("tmap.devel.mode")
	on.exit({
		options(tmap.mode = current_mode, tmap.devel.mode = current_dev)
	})
	options(tmap.mode = "plot")

	opt = if (is.na(scale)) {
		tm_options(asp = asp)
	} else {
		tm_options(asp = asp, scale = scale)
	}

	print.tmap(x + opt, show = show, ...)
}

process_color <- function(col, alpha=NA, sepia_intensity=0, saturation=1, color_vision_deficiency_sim="none") {
	#if (length(col)>100) browser()

	isFactor <- is.factor(col)

	if (isFactor) {
		x <- as.integer(col)
		col <- levels(col)
	}

	res <- t(col2rgb(col, alpha=TRUE))

	# set alpha values
	if (!is.na(alpha)) res[res[,4] != 0, 4] <- alpha * 255

	# convert to sepia
	if (sepia_intensity!=0) {
		conv_matrix <- matrix(c(.393, .769, .189,
								.349, .686, .168,
								.272, .534, .131), ncol=3, byrow=FALSE)
		res[,1:3] <-  (res[,1:3] %*% conv_matrix) * sepia_intensity + res[,1:3] * (1-sepia_intensity)
		res[res>255] <- 255
		res[res<0] <- 0
	}

	# convert to black&white
	if (saturation!=1) {
		res[,1:3] <- (res[,1:3] %*% matrix(c(.299, .587, .114), nrow=3, ncol=3))  * (1-saturation) + res[,1:3] * saturation
		res[res>255] <- 255
		res[res<0] <- 0
	}
	if (all(res[,4]==255)) res <- res[,-4, drop=FALSE]

	new_cols <- do.call("rgb", c(unname(as.data.frame(res)), list(maxColorValue=255)))

	rlang::check_installed("colorspace")
	# color blind sim
	sim_colors = switch(color_vision_deficiency_sim,
		deutan = colorspace::deutan,
		protan = colorspace::protan,
		tritan = colorspace::tritan,
		function(x) x)

	new_cols2 = sim_colors(new_cols)

	if (isFactor) {
		new_cols2[x]
	} else {
		new_cols2
	}
}

is_light <- function(col) {
	colrgb <- col2rgb(col)
	apply(colrgb * c(.299, .587, .114), MARGIN=2, sum) >= 128
}

get_light <- function(col) {
	colrgb <- col2rgb(col)
	apply(colrgb * c(.299, .587, .114), MARGIN=2, sum) / 255
}


darker <- function(colour, rate, alpha=NA) {
	col <- col2rgb(colour, TRUE)/255
	col[1:3] <- col[1:3] * (1-rate)
	if (is.na(alpha)) alpha <- col[4,]
	rgb(col[1, ], col[2, ], col[3, ], alpha)
}

lighter <- function(colour, rate, alpha=NA) {
	col <- col2rgb(colour, TRUE)/255
	col[1:3] <- col[1:3] + (1-col[1:3]) * rate
	if (is.na(alpha)) alpha <- col[4,]
	rgb(col[1, ], col[2, ], col[3, ], alpha)
}

col2hex <- function(x) {
	y <- apply(col2rgb(x), MARGIN=2, FUN=function(y)do.call(rgb, c(as.list(y), list(maxColorValue=255))))
	y[is.na(x)] <- NA
	y
}


# determine palette type
# palettes of length 1,2 or 3 are cat
# palette of length 4+ are seq if luminance is increasing or decreaing
# palette of length 5+ are div if luminance is increasing in first half and decreasing in second half, or the other way round.
palette_type <- function(palette) {
	k <- length(palette)
	if (k<4) return("cat")

	m1 <- ceiling(k/2) - 1
	m2 <- floor(k/2) + 1

	colpal_light <- get_light(palette)

	s <- sign(colpal_light[-1] - colpal_light[-k])

	if (all(s==1) || all(s==-1)) {
		return("seq")
	} else if (k>4 && ((all(s[1:m1]==1) && all(s[m2:(k-1)]==-1)) ||
		(all(s[1:m1]==-1) && all(s[m2:(k-1)]==1)))) {
		return("div")
	} else {
		return("cat")
	}
}


valid_colors <- function(x) {
	is.na(x) | (x %in% colors()) |	(vapply(gregexpr("^#(([[:xdigit:]]){6}|([[:xdigit:]]){8})$", x), "[[", integer(1), 1) == 1L)
}



# get_alpha_col <- function(colour, alpha=NA) {
# 	col <- col2rgb(colour, TRUE)/255
# 	if (is.na(alpha)) alpha <- col[4,]
# 	new_col <- rgb(col[1, ], col[2, ], col[3, ], alpha)
# 	new_col
# }
#
# get_sepia_col <- function(col, intensity=1) {
# 	conv_matrix <- matrix(c(.393, .769, .189,
# 							.349, .686, .168,
# 							.272, .534, .131), ncol=3, byrow=FALSE)
# 	colM <- t(col2rgb(col))
# 	res <-  (colM %*% conv_matrix) * intensity + colM * (1-intensity)
# 	res[res>255] <- 255
# 	do.call("rgb", c(unname(as.data.frame(res)), list(maxColorValue=255)))
# }
#

#' Draw thematic map
#'
#' @param x tmap object.
#' @param return.asp should the aspect ratio be returned?
#' @param show show the map
#' @param vp viewport (for `"plot"` mode)
#' @param knit A logical, should knit?
#' @param in.shiny A logical, is the map drawn in **shiny**?
#' @param proxy A logical, if `in.shiny`, is [tmapProxy()] used?
#' @param options A vector of options
#' @param ... passed on internally (for developers: in `"view"` mode, the proxy leaflet object is passed to `tmapLeafletInit`).
#' @export
print.tmap = function(x, return.asp = FALSE, show = TRUE, vp = NULL, knit = FALSE, options = NULL, in.shiny = FALSE, proxy = FALSE, ...) {
	args = list(...)

	.TMAP$in.shiny = in.shiny
	.TMAP$proxy = proxy
	.TMAP$set_s2 = NA
	.TMAP$animate = NULL
	.TMAP$raster_wrap = FALSE # needed for space_overlay (#1170)

	.TMAP$stamp = as.numeric(Sys.time())

	# view mode will use panes, in principle one for each layer. They start at 400, unless shiny proxy is used

	dev = getOption("tmap.devel.mode")
	if (dev) timing_init()
	x2 = step1_rearrange(x, knit_opts = options)
	if (dev) timing_add("step 1")
	x3 = step2_data(x2)
	if (dev) timing_add("step 2")
	x4 = step3_trans(x3)
	if (dev) timing_add("step 3")
	res = step4_plot(x4, vp = vp, return.asp = return.asp, show = show, in.shiny = in.shiny, knit = knit, knit_opts = options, args)
	if (dev) timing_add("step 4")
	if (dev) timing_eval()

	v3_reset_flag()
	if (!is.na(.TMAP$set_s2)) suppressMessages(sf::sf_use_s2(.TMAP$set_s2))

	#if (return.asp) return(asp) else invisible(NULL)
	if (knit && tmap_graphics_name() != "Grid") {
		kp = get("knit_print", asNamespace("knitr"))
		return(do.call(kp, c(list(x=res), args, list(options=options))))
	} else if (inherits(res, "knit_asis")) {
		return(res)
	} else {
		invisible(res)
	}
}

#' @rdname print.tmap
#' @exportS3Method knitr::knit_print
knit_print.tmap = function(x, ..., options=NULL) {
	print.tmap(x, knit=TRUE, options=options, ...)
}


timing_init = function() {
	ts = data.table(s1 = "---------", s2 = "---------", s3 = "---------", s4 = "---------", t = Sys.time())
	assign("timings", ts, envir = .TMAP)
}

timing_add = function(s1 = "", s2 = "", s3 = "", s4 = "") {
	tsx = data.table(s1 = s1, s2 = s2, s3 = s3, s4 = s4, t = Sys.time())
	ts = data.table::rbindlist(list(get("timings", envir = .TMAP)[, c("s1", "s2", "s3", "s4", "t")], tsx))
	assign("timings", ts, envir = .TMAP)
}

timing_eval = function() {
	ts = get("timings", envir = .TMAP)

	ts[, total := round(as.numeric(difftime(ts$t, ts$t[1], units = "secs")), 3)]

	i1 = ts$s1 != ""
	i2 = ts$s2 != "" | i1
	i3 = ts$s3 != "" | i2
	i4 = ts$s4 != "" | i3

	ts[, ':='(t1=0,t2=0,t3=0,t4=0)]
	ts[i1, t1:= c(0, ts$total[i1][-1] - head(ts$total[i1], -1))]
	ts[i2, t2:= c(0, ts$total[i2][-1] - head(ts$total[i2], -1))]
	ts[i3, t3:= c(0, ts$total[i3][-1] - head(ts$total[i3], -1))]
	ts[i4, t4:= c(0, ts$total[i4][-1] - head(ts$total[i4], -1))]

	ts[s2 == "", t2 := 0]
	ts[s3 == "", t3 := 0]
	ts[s4 == "", t4 := 0]

	form = function(l, x) {
		zero = (x==0)
		y = sprintf("%.3f", x)
		z = paste0(l, " (", y, ")")
		z[zero] = ""
		z
	}

	ts[, ':='(s1=form(s1, t1),
			  s2=form(s2, t2),
			  s3=form(s3, t3),
			  s4=form(s4, t4))]

	print(ts[, c("s1", "s2", "s3", "s4", "total")])
}

#' Scales: bivariate scale
#'
#' Scales in tmap are configured by the family of functions with prefix `tm_scale`.
#' Such function should be used for the input of the `.scale` arguments in the
#' layer functions (e.g. `fill.scale` in [tm_polygons()]).
#' The function `tm_scale_bivariate()` is used for `bivariate.scales`.
#'
#' @param scale1,scale2 two `tm_scale` objects. Currently, all `tm_scale_*()` functions are supported except `tm_scale_continuous()`.
#' @param values (generic scale argument) The visual values. For colors
#'   (e.g. `fill` or `col` for `tm_polygons()`) this is a palette name from the `cols4all` package (see [cols4all::c4a()])
#'   or vector of colors, for size (e.g. `size` for `tm_symbols()`) these are a set of sizes (if two values are specified they are interpret as range),
#'   for symbol shapes (e.g. `shape` for [tm_symbols()]) these are a set of symbols, etc.
#'   The tmap option `values.var` contains the default values per visual variable and in some cases also per data type.
#' @param values.repeat (generic scale argument) Should the values be repeated in case there are more categories?
#' @param values.range (generic scale argument) Range of the values. Vector of two numbers (both between 0 and 1) where the first determines the minimum and the second the maximum.
#'   Full range, which means that all values are used, is encoded as `c(0, 1)`.
#'   For instance, when a grey scale is used for color (from black to white), `c(0,1)`
#'   means that all colors are used, `0.25, 0.75` means that only colors from dark grey to light
#'   grey are used (more precisely `"grey25"` to `"grey75"`), and `0, 0.5` means
#'   that only colors are used from black to middle grey (`"grey50"`).
#'   When only one number is specified, this is interpreted as the second number
#'   (where the first is set to 0). Default values can be set via the tmap option `values.range`.
#' @param values.scale (generic scale argument) Scaling of the values. Only useful for size-related visual variables, such as `size` of [tm_symbols()] and `lwd` of [tm_lines()].
#' @param value.na (generic scale argument) Value used for missing values. See tmap option `"value.na"` for defaults per visual variable.
#' @param value.null (generic scale argument) Value used for NULL values.
#'   See tmap option `"value.null"` for defaults per visual variable. Null data
#'   values occur when out-of-scope features are shown (e.g. for a map of Europe showing a data variable per country,
#'   the null values are applied to countries outside Europe).
#' @param value.neutral (generic scale argument) Value that can be considered neutral.
#'   This is used for legends of other visual variables of the same map layer.
#'   E.g. when both `fill` and `size` are used for [tm_symbols()] (using filled circles),
#'   the size legend items are filled with the `value.neutral` color from the `fill.scale` scale,
#'   and fill legend items are bubbles of size `value.neutral` from the `size.scale` scale.
#' @param labels (generic scale argument) Labels
#' @param label.na (generic scale argument) Label for missing values
#' @param label.null (generic scale argument) Label for null (out-of-scope) values
#' @seealso [tm_scale()]
#' @export
tm_scale_bivariate = function(scale1 = tm_scale(),
							  scale2 = tm_scale(),
							  values = NA,
							  values.repeat = FALSE,
							  values.range  = NA,
							  values.scale = 1,
							  value.na = NA,
							  value.null = NA,
							  value.neutral = NA,
							  labels = NULL,
							  label.na = NA,
							  label.null = NA) {
	structure(c(list(FUN = "tmapScaleBivariate"), as.list(environment())), class = c("tm_scale_bivariate", "tm_scale", "list"))
}

.defaultTmapOptions = structure(
	list(
		# mode specific options or default values
		modes = list(plot = list(name = "Grid",
								 use_gradient = FALSE,
								 crs_basemap = 3857,
								 limit_latitude_3857 = c(-84, 84)),
								 #crs = "auto"),
					 view = list(name = "Leaflet",
					 			use_browser = FALSE,
					 			use_WebGL = NA,
					 			legend.position = tm_pos_in(pos.h = "right", pos.v = "bottom", align.h = "left", align.v = "top", just.h = "left", just.v = "top"),
					 			crs_basemap = list(dimensions = 3857, 4326),
					 			facet.max = 16,
					 			#legend.bg.alpha = 0.8,
					 			#view.legend.position = c("right", "top"),
					 			control.position = c("left", "top"),
					 			control.collapse = TRUE,
					 			basemap.show = TRUE,
					 			set_bounds = FALSE,
					 			set_view = NA,
					 			set_zoom_limits = NA,
					 			qtm.scalebar = TRUE,
					 			qtm.minimap = FALSE,
					 			qtm.mouse_coordinates = TRUE,
					 			use_circle_markers = TRUE,
					 			leaflet.options = list())),

		crs = NA,
		bbox = NULL,

		# facets
		facet.max = 64, # was max.facets
		facet_levels.max = 1e5, # maximum number of facet levels
		free.scales = NULL, # for backward compatibility: if this value is set, it will be used to impute the free arguments in the layer functions


		# spatial object class specific options
		raster.max_cells = 1e7, # was max.raster
		raster.warp = TRUE,

		# general
		show.messages = TRUE,
		show.warnings = TRUE,

		# output
		output.format = "png",
		output.size = 49,
		output.dpi = 300,
		animation.dpi = 100,


		# default visual variable values
		value.const = list(fill.symbols = "grey60",
						   fill.dots = "black",
						   fill = "grey85",
						   col.polygons = "grey25",
						   col.symbols = "grey25",
						   col.raster = "grey40",
						   col.text = "black",
						   col = "black",
						   bgcol.labels_highlighted = "white",
						   bgcol = "#00000000",
						   lwd = 1,
						   lty = "solid",
						   text = "Abc",
						   text.labels = "",
						   fontface = NULL,
						   shape.squares = 22,
						   shape.dots = 19,
						   shape.markers = NA,
						   shape = 21,
						   size.circles = 1000,
						   size.bubbles = 1.3333,
						   size.squares = 1.3333,
						   size.dots = .3,
						   size = 1,
						   fill_alpha = 1,
						   col_alpha = 1,
						   bgcol_alpha = 1,
						   angle = 0,
						   num = 1),
		value.na = list(
			fill = "grey75",
			col = "grey75",
			col.raster = "#00000000",
			bgcol = "grey75",
			lty = "solid",
			text = "Unknown",
			fontface = NULL,
			fill_alpha = 1,
			col_alpha = 1,
			bgcol_alpha = 1,
			col_alpha.raster = 0,
			angle = 0
		),
		value.null = list(
			fill = "grey95",
			col = "grey95",
			col.polygons = "grey40",
			bgcol = "grey95",
			lty = "solid",
			lwd = 1,
			text = "",
			fontface = NULL,
			fill_alpha = 1,
			col_alpha = 1,
			bgcol_alpha = 1,
			shape.squares = 22,
			shape.dots = 19,
			shape = 21,
			size.circles = 1000,
			size.bubbles = 1.3333,
			size.squares = 1.3333,
			size.dots = .3,
			size = 1,
			angle = 0,
			num = 0
		),
		value.blank = list(
			fill = "#00000000",
			col = "#00000000",
			bgcol = "#00000000",
			lty = "blank",
			lwd = NA,
			text = "",
			fontface = NULL,
			fill_alpha = 0,
			col_alpha = 0,
			bgcol_alpha = 0,
			angle = 0,
			num = 0
		),
		values.var = list(fill = list(seq = "-hcl.blues3", div = "pu_gn_div",
									  unord = "cols4all.area7", ord = "-hcl.blues3", cyc = "tol.rainbow_pu_rd", biv = "pu_gn_bivs"),
						  col = list(seq = "-hcl.blues3", div = "pu_gn_div",
						  		   unord = "cols4all.line7", ord = "-hcl.blues3", cyc = "tol.rainbow_pu_rd", biv = "pu_gn_bivs"),
						  fill.dots = list(seq = "-hcl.blues3", div = "pu_gn_div",
						  				 unord = "cols4all.line7", ord = "-hcl.blues3", cyc = "tol.rainbow_pu_rd", biv = "pu_gn_bivs"),
						  bgcol = list(seq = "-hcl.blues3", div = "pu_gn_div",
						  			 unord = "cols4all.area7", ord = "-hcl.blues3", cyc = "tol.rainbow_pu_rd", biv = "pu_gn_bivs"),
						  size = tm_seq(0, 1, power = "sqrt"),
						  size.bubbles = tm_seq(0, 1, power = "sqrt"),
						  lwd = c(0, 3),
						  lty = c("dashed", "dotted", "dotdash", "longdash", "twodash"),
						  text = LETTERS,
						  fontface = c("plain", "italic", "bold"),
						  fill_alpha = c(0.25, 1),
						  col_alpha = c(0.25, 1),
						  bgcol_alpha = c(0.25, 1),
						  shape = 21:25,
						  area = c(0, 1),
						  xmod = c(0, 1),
						  ymod = c(0, 1),
						  angle = c(0, 360),
						  num = c(0, 1)),
		values.range = list(fill = NA, col = NA, size = c(0, 1), lwd = c(0, 1),
							lty = NA, text = NA, fontface = NA, fill_alpha = NA,
							col_alpha = NA, shape = NA, angle = NA, num = c(0,1)), # NA = automatic, NULL is not applicable
		value.neutral = list(size = 0.75,
							 lwd = 2,
							 lty = "solid",
							 fill_alpha = 1,
							 col_alpha = 1,
							 bgcol_alpha = 1,
							 text = "Abc",
							 fontface = "plain",
							 angle = 0,
							 num = 1),
		values.scale = list(
			1,
			lwd.lines = 1,
			size.symbols = 1,
			size.bubbles = 1.3333,
			size.squares = 1.3333
		),

		# scales
		scales.var = list(fill = list(fact = "categorical", num = "intervals", int = "discrete", datetime = "intervals", date = "intervals"),
						  col = list(fact = "categorical", num = "intervals", int = "discrete", datetime = "intervals", date = "intervals"),
						  bgcol = list(fact = "categorical", num = "intervals", int = "discrete", datetime = "intervals", date = "intervals"),
						  lwd = list(fact = "categorical", num = "continuous", int = "discrete", datetime = "intervals", date = "intervals"),
						  lty = list(fact = "categorical", num = "intervals", datetime = "intervals", date = "intervals"),
						  shape = list(fact = "categorical", num = "intervals", datetime = "intervals", date = "intervals"),
						  size = list(fact = "continuous", num = "continuous", datetime = "continuous", date = "continuous"),
						  fill_alpha = list(fact = "categorical", num = "intervals", datetime = "continuous", date = "continuous"),
						  col_alpha = list(fact = "categorical", num = "intervals", datetime = "continuous", date = "continuous"),
						  bgcol_alpha = list(fact = "categorical", num = "intervals", datetime = "continuous", date = "continuous"),
						  area = list(fact = "categorical", num = "continuous", datetime = "continuous", date = "continuous"),
						  xmod = list(fact = "asis", num = "asis"),
						  ymod = list(fact = "asis", num = "asis"),
						  angle = list(fact = "asis", num = "asis"),
						  text = list(fact = "asis", num = "asis"),
						  fontface = list(fact = "categorical", num = "categorical", datetime = "categorical", date = "categorical")),

		scale.misc.args = list(continuous = list(n = c(fill = 5, col = 5, 5), outliers.trunc = c(FALSE, FALSE), trans = "identity", limits = list(fill = NA, col = NA, 0)),
							   continuous_log = list(n = c(fill = 5, col = 5, 5), outliers.trunc = c(FALSE, FALSE),	limits = list(fill = NA, col = NA, NA)),
							   continuous_log2 = list(n = c(fill = 5, col = 5, 5), outliers.trunc = c(FALSE, FALSE),	limits = list(fill = NA, col = NA, NA)),
							   continuous_log10 = list(n = c(fill = 5, col = 5, 5), outliers.trunc = c(FALSE, FALSE),	limits = list(fill = NA, col = NA, NA)),
							   continuous_log1p = list(n = c(fill = 5, col = 5, 5), outliers.trunc = c(FALSE, FALSE),	limits = list(fill = NA, col = NA, NA)),
							   continuous_sqrt = list(n = c(fill = 5, col = 5, 5), outliers.trunc = c(FALSE, FALSE),	limits = list(fill = NA, col = NA, 0)),
							   continuous_pseudo_log = list(n = c(fill = 5, col = 5, 5), outliers.trunc = c(FALSE, FALSE),	limits = list(fill = NA, col = NA, NA)),
							   rank = list(n = 5,
							   			unit = "rank")), # NA means take data range, 0 means include 0


		continuous.nclass_per_legend_break = 50, # the number of continuous legend breaks within one 'unit' (label).
		continuous.nclasses = 101, # the number of classes of a continuous scale. Should be odd

		# labels
		label.format = list(
			fun = NULL,
			scientific = FALSE,
			digits = NA,
			interval.disjoint = TRUE,
			big.num.abbr = c(mln = 6, bln = 9, trn = 12, qdn = 15, qnt = 18),
			prefix = "",
			suffix = "",
			text.separator = "-",
			text.less.than = "<",
			text.less.than_as.prefix = TRUE,
			text.or.more = "\u2265",
			text.or.more_as.prefix = TRUE,
			text.align = NA,
			text.to.columns = FALSE,
			html.escape = TRUE
		),
		label.na = "Missing",


		###############################3
		# tm_layout options
		###############################3
		scale = 1,
		asp = NA,

		# background
		bg = TRUE,
		bg.color = "white",
		outer.bg = FALSE,
		outer.bg.color = "white",

		# frame
		frame = TRUE,
		frame.color = NA,
		frame.alpha = 1,
		frame.lwd = 1,
		frame.r = NA,
		frame.double_line = FALSE,


		# margins
		outer.margins = rep(0.02, 4),
		inner.margins = list(stars = rep(0, 4), SpatRaster = rep(0, 4), rep(0.02, 4)),
		inner.margins.extra = c(0, 0, 0, 0),
		meta.margins = NA,
		meta.auto_margins = c(0.4, 0.4, 0.4, 0.4),
		between_margin = 0.5,
		panel.margin = c(xtab = 0.4, 0),

		# xlab, ylab
		xlab.show = FALSE,
		xlab.text = "",
		xlab.size = 1,
		xlab.color = "black",
		xlab.rotation = 0,
		xlab.space = 0.5,
		xlab.fontface = NULL,
		xlab.fontfamily = NULL,
		xlab.alpha = 1,
		xlab.side = "bottom",

		ylab.show = FALSE,
		ylab.text = "",
		ylab.size = 1,
		ylab.color = "black",
		ylab.rotation = 0,
		ylab.space = 0.5,
		ylab.fontface = NULL,
		ylab.fontfamily = NULL,
		ylab.alpha = 1,
		ylab.side = "left",


		# panel
		panel.type = NA, # "wrap" or "xtab",
		panel.wrap.pos = "top", # or "left", "right", "bottom"
		panel.xtab.pos = c("left", "top"),

		# data
		unit = "metric",

		# general visual settings

		# colors
		color.sepia_intensity = 0,
		color.saturation = 1,
		color_vision_deficiency_sim = "none",

		# text
		text.fontface = "plain",
		text.fontfamily = "",
		r = 2,

		component.position = list('in' = tm_pos_in(pos.h = "left", pos.v = "top",
											  align.h = "left", align.v = "top", just.h = "left", just.v = "top"),
								  out = tm_pos_out(cell.h = "right", cell.v = "center",
								  		   pos.h = "left", pos.v = "top",
								  		   align.h = "left", align.v = "top", just.h = "left", just.v = "top")),
		component.offset = c(inside = 0.75, INSIDE = 0, outside = 0, OUTSIDE = 0),
		component.stack_margin = c(combined = 0, apart = 0.5),
		component.autoscale = TRUE,
		component.resize_as_group = FALSE,
		component.frame_combine = TRUE,
		component.stack = "vertical",
		legend.stack = c(all = "vertical", per_row = "horizontal", per_col = "horizontal", all_row = "vertical", all_col = "horizontal", manual = "vertical"),
		chart.stack = c(all = "vertical", per_row = "horizontal", per_col = "horizontal", all_row = "vertical", all_col = "horizontal", manual = "vertical"),
		component.equalize = TRUE,

		component.frame = FALSE,
		component.frame.color = NA,
		component.frame.alpha = 1,
		component.frame.lwd = 1,
		component.frame.r = NA,

		component.bg = TRUE,
		component.bg.color = "white",
		component.bg.alpha = 1,

		# legend
		legend.show = TRUE,
		legend.orientation = "portrait",
		legend.position = tm_pos_auto_out(cell.h = "right", cell.v = "bottom",
										  pos.h = "left", pos.v = "top",
										  align.h = "left", align.v = "top", just.h = "left", just.v = "top"),
		legend.width = NA,
		legend.height = NA,
		legend.reverse = FALSE,
		legend.na.show = NA,
		legend.title.color = NA,
		legend.title.size = 0.9,
		legend.title.fontface = NULL,
		legend.title.fontfamily = NULL,
		legend.title.alpha = 1,
		legend.xlab.color = NA,
		legend.xlab.size = 0.9,
		legend.xlab.fontface = NULL,
		legend.xlab.fontfamily = NULL,
		legend.xlab.alpha = 1,
		legend.ylab.color = NA,
		legend.ylab.size = 0.9,
		legend.ylab.fontface = NULL,
		legend.ylab.fontfamily = NULL,
		legend.xlab.rot = 0,
		legend.ylab.rot = 0,
		legend.ylab.alpha = 1,
		legend.text.color = NA,
		legend.text.size = 0.7,
		legend.text.fontface = NULL,
		legend.text.fontfamily = NULL,
		legend.text.alpha = 1,
		legend.frame = TRUE,
		legend.frame.color = NA,
		legend.frame.alpha = 1,
		legend.frame.lwd = 1,
		legend.frame.r = NA,
		legend.bg = TRUE,
		legend.bg.color = "white",
		legend.bg.alpha = 1,
		legend.only = FALSE,
		legend.absolute_fontsize = 14,
		legend.settings.portrait = list(item.height = c(rect = 1.2, symbols = 1, gradient = 3, lines = 1.2, text = 1.2, bivariate = 1.2),
												 item.width = c(rect = 1.2, symbols = 1, gradient = 1.2, lines = 1.2, text = 3, bivariate = 1.2),
												 item.r = NA,
												 item.space = c(rect = 0.2, symbols = 0.2, gradient = 0, lines = 0.2, text = 0.2, bivariate = 0),
												 item.na.height = c(rect = NA, symbols = NA, gradient = 1.2, lines = NA, text = NA, bivariate = NA),
												 item.na.width = c(rect = NA, symbols = NA, gradient = 1.2, lines = NA, text = NA, bivariate = NA),
												 item.na.space = c(rect = 0.2, symbols = 0.3, gradient = 1, lines = 0.2, text = 0.2, bivariate = 0.2),
												 item.shape = 107,
												 title.padding  = c(0, 0, 0.25, 0),
												 xlab.padding = c(0, 0, 0.25, 0),
												 ylab.padding = c(0, 0, 0.25, 0),
												 title.align = c(bivariate = "right", "left"),
												 xlab.align = "left",
												 ylab.align = "center",
												 ticks = list(rect = list(), symbols = list(), gradient = list(c(0.85, 1.1)), lines = list(), text = list(), bivariate = list()),
												 ticks.disable.na = c(rect = FALSE, symbols = FALSE, gradient = TRUE, lines = FALSE, text = FALSE, bivariate = TRUE),
												 ticks.col = NA,
												 ticks.lwd = 1,
												 margins = c(0.4, 0.4, 0.4, 0.4),
												 item_text.margin = c(gradient = 0.5, 0.25)),
		legend.settings.landscape = list(item.height = c(rect = 1, symbols = 1, gradient = 1.2, lines = 1, text = 1),
												  item.width = c(rect = 6, symbols = 3, gradient = 6, lines = 6, text = 6),
												  item.r = NA,
												  item.space = c(rect = 0.2, symbols = 0.3, gradient = 0, lines = 0.2, text = 0.2),
												  item.na.height = c(rect = NA, symbols = NA, gradient = 2, lines = NA, text = NA),
												  item.na.width = c(rect = NA, symbols = NA, gradient = 4, lines = NA, text = NA),
												  item.na.space = c(rect = 0.2, symbols = 0.3, gradient = 0.3, lines = 0.2, text = 0.2),
												  item.shape = 107,

												  title.padding  = c(0, 0, 0.25, 0),
												  xlab.padding = c(0, 0, 0.25, 0),
												  ylab.padding = c(0, 0, 0.25, 0),
												  title.align = c(bivariate = "right", "left"),
												  xlab.align = "left",
												  ylab.align = "center",
												  ticks = list(rect = list(), symbols = list(), gradient = list(c(0.85, 1.1)), lines = list(), text = list()),
												  ticks.disable.na = c(rect = FALSE, symbols = FALSE, gradient = TRUE, lines = FALSE, text = FALSE),
												  ticks.col = NA,
												  ticks.lwd = 1,
												  margins = c(0.4, 0.4, 0.4, 0.4),
												  item_text.margin = 0.25),

		#
		add_legend.position = NULL,

		# charts
		chart.show = TRUE,
		chart.plot.axis.x = FALSE,
		chart.plot.axis.y = TRUE,
		chart.position = tm_pos_auto_out(cell.h = "right", cell.v = "bottom",
										 pos.h = "left", pos.v = "top",
										 align.h = "left", align.v = "bottom", just.h = "left", just.v = "top"),
		chart.width = c(histogram.min = 10,
						histogram.max = 20,
						bar.min = 10,
						bar.max = 20,
						donut.min = 10,
						donut.max = 10,
						heatmap.min = 10,
						heatmap.max = 15),
		chart.height = c(histogram.min = 10,
						 histogram.max = 10,
						 bar.min = 10,
						 bar.max = 10,
						 donut.min = 10,
						 donut.max = 10,
						 heatmap.min = 10,
						 heatmap.max = 10),
		chart.reverse = FALSE,
		chart.na.show = NA,
		chart.title.color = NA,
		chart.title.size = 0.9,
		chart.title.fontface = NULL,
		chart.title.fontfamily = NULL,
		chart.title.alpha = NULL,
		chart.xlab.color = NA,
		chart.xlab.size = 0.9,
		chart.xlab.fontface = NULL,
		chart.xlab.fontfamily = NULL,
		chart.xlab.alpha = NULL,
		chart.ylab.color = NA,
		chart.ylab.size = 0.9,
		chart.ylab.fontface = NULL,
		chart.ylab.fontfamily = NULL,
		chart.ylab.alpha = 1,
		chart.text.color = NA,
		chart.text.size = 0.7,
		chart.text.fontface = NULL,
		chart.text.fontfamily = NULL,
		chart.text.alpha = 1,
		chart.frame = TRUE,
		chart.frame.color = NA,
		chart.frame.alpha = 1,
		chart.frame.lwd = 1,
		chart.frame.r = NA,
		chart.bg = TRUE,
		chart.bg.color = "white",
		chart.bg.alpha = 1,
		chart.object.color = "#DDDDDD",

		# components
		title.size = 1.3,
		title.color = NA,
		title.fontface = NULL,
		title.fontfamily = NULL,
		title.alpha = 1,
		title.padding = c(0.25, 0.25, 0.25, 0.25),
		title.frame = FALSE,
		title.frame.color = NA,
		title.frame.alpha = 1,
		title.frame.lwd = 1,
		title.frame.r = NA,
		title.position = tm_pos_out(cell.h = "center", cell.v ="top", pos.h = "left", pos.v = "top", align.h = "left", align.v = "top", just.h = "left", just.v = "top"),
		title.width = NA,
		credits.size = .7,
		credits.color = NA,
		credits.fontface = NULL,
		credits.fontfamily = NULL,
		credits.alpha = 1,
		credits.padding = c(0.25, 0.25, 0.25, 0.25),
		credits.position = tm_pos_in(pos.h = "right", pos.v = "bottom", align.h = "right", align.v = "top", just.h = "left", just.v = "top"),
		credits.width = NA,
		credits.height = NA,
		compass.north=0,
		compass.type="arrow",
		compass.text.size=.8,
		compass.size=NA,
		compass.show.labels=1,
		compass.cardinal.directions=c("N", "E", "S", "W"),
		compass.text.color=NA,
		compass.color.dark=NA,
		compass.color.light=NA,
		compass.lwd=1,
		compass.margins = c(0.25, 0.25, 0.25, 0.25),
		compass.position = tm_pos_in(pos.h = "right", pos.v = "bottom", align.h = "right", align.v = "top", just.h = "left", just.v = "top"),
		# standard arguments:
		inset.position = tm_pos_in(pos.h = "right", pos.v = "bottom", align.h = "right", align.v = "top", just.h = "left", just.v = "top"),

		logo.height = 3,
		logo.margins = c(0.2, 0.2, 0.2, 0.2),
		logo.between_margin = 0.2,
		logo.position = tm_pos_in(pos.h = "right", pos.v = "bottom", align.h = "right", align.v = "top", just.h = "left", just.v = "top"),


		inset.height = 3,
		inset.width = 3,
		inset.margins = c(0.2, 0.2, 0.2, 0.2),
		inset.between_margin = 0.2,
		inset.position = tm_pos_in(pos.h = "right", pos.v = "bottom", align.h = "right", align.v = "top", just.h = "left", just.v = "top"),
		inset.frame = TRUE,
		inset.bg = TRUE,
		inset.bg.color = "#ffffff",
		inset.bg.alpha = 1,

		inset.box_frame = TRUE,
		inset.box_frame.color = "#ee2211",
		inset.box_frame.alpha = 1,
		inset.box_frame.lwd = 2,
		inset.box_frame.lty = "solid",
		inset.box_bg = FALSE,
		inset.box_bg.color = "#ffffff",
		inset.box_bg.alpha = 1,

		inset.main_frame = TRUE,
		inset.main_frame.r = NA,
		inset.main_frame.color = NA,
		inset.main_frame.alpha = 1,
		inset.main_frame.lwd = 2,

		inset_map.height = 7,
		inset_map.width = 7,
		inset_map.margins = c(0.2, 0.2, 0.2, 0.2),
		inset_map.between_margin = 0.2,
		inset_map.position = tm_pos_in(pos.h = "right", pos.v = "bottom", align.h = "right", align.v = "top", just.h = "left", just.v = "top"),
		inset_map.frame = FALSE,

		inset_tmap.height = 7,
		inset_tmap.width = 7,

		inset_grob.height = 7,
		inset_grob.width = 7,

		inset_gg.height = 7,
		inset_gg.width = 7,

		scalebar.breaks = NULL,
		scalebar.width = 12,
		scalebar.allow_clipping = FALSE,
		scalebar.text.size = .5,
		scalebar.text.color = NA,
		scalebar.text.fontface = NULL,
		scalebar.text.fontfamily = NULL,
		scalebar.color.dark = NA,
		scalebar.color.light = NA,
		scalebar.lwd = 1,
		scalebar.size = NULL,
		scalebar.margins = c(0.01,0.01,0.01,0.01),

		# standard arguments:
		scalebar.position = tm_pos_in(pos.h = "right", pos.v = "bottom", align.h = "right", align.v = "top", just.h = "left", just.v = "top"),

		grid.show = FALSE,
		grid.labels.pos = c("left", "bottom"),
		grid.x=NA,
		grid.y=NA,
		grid.n.x=NA,
		grid.n.y=NA,
		grid.crs=NA,
		grid.col=NA,
		grid.lwd=1,
		grid.alpha=NA,
		grid.labels.show=TRUE,
		grid.labels.size=.6,
		grid.labels.col=NA,
		grid.labels.fontface=NULL,
		grid.labels.fontfamily=NULL,
		grid.labels.rot = c(0, 0),
		grid.labels.format = list(big.mark = ","),
		grid.labels.cardinal = FALSE,
		grid.labels.margin.x=0,
		grid.labels.margin.y=0,
		grid.labels.space.x=NA,
		grid.labels.space.y=NA,
		grid.labels.inside_frame=FALSE,
		grid.ticks = TRUE, #labels.show & !labels.inside_frame,
		grid.lines = TRUE,
		grid.ndiscr = 100,


		# standard arguments:
		# mouse.stack = "vertical",
		# mouse.position = tm_pos_in(pos.h = "right", pos.v = "bottom", align.h = "left", align.v = "top", just.h = "left", just.v = "top"),
		# mouse.frame = FALSE,
		# mouse.frame.lwd = 1,
		# mouse.frame.r = 2,
		# mouse.group.frame = TRUE,
		# mouse.resize_as_group = FALSE,
		#

		mouse_coordinates.position = tm_pos_in(pos.h = "right", pos.v = "bottom", align.h = "right", align.v = "top", just.h = "left", just.v = "top"),

		minimap.server = NA,
		minimap.toggle = TRUE,
		minimap.position = tm_pos_in(pos.h = "right", pos.v = "bottom", align.h = "right", align.v = "top", just.h = "left", just.v = "top"),
		minimap.height = 7,
		minimap.width = 7,
		minimap.margins = c(0.2, 0.2, 0.2, 0.2),
		minimap.between_margin = 0.2,
		minimap.frame = FALSE,
		minimap.bg = TRUE,


		panel.show = NA, # depends on type: TRUE for facets
		panel.labels = NA,
		panel.label.size = 1,
		panel.label.color = "black",
		panel.label.fontface = NULL,
		panel.label.fontfamily = NULL,
		panel.label.alpha = 1,
		panel.label.bg = TRUE,
		panel.label.bg.color = "grey80",
		panel.label.bg.alpha = 1,
		panel.label.frame = TRUE,
		panel.label.frame.color = NA,
		panel.label.frame.alpha = 1,
		panel.label.frame.lwd = 1,
		panel.label.frame.r = NA,
		panel.label.height = 1,
		panel.label.rot = c(90, 0, 270, 0),

		qtm.scalebar = FALSE,
		qtm.minimap = FALSE,
		qtm.mouse_coordinates = FALSE,

		earth_boundary = FALSE,
		earth_boundary.color = NA,
		earth_boundary.lwd = 1,
		earth_datum = "OGC:CRS84",
		space = TRUE,
		space.color = "#ffffff",
		space_overlay = NA,

		check_and_fix = FALSE,

		basemap.show = FALSE,
		basemap.server = c("Esri.WorldGrayCanvas", "OpenStreetMap", "Esri.WorldTopoMap"),
		basemap.alpha = 1,
		basemap.zoom = NA,

		tiles.show = FALSE,
		tiles.server = "",
		tiles.alpha = 1,
		tiles.zoom = NA,

		attr.color = "black",

		crs_extra = "",
		crs_global = "+proj=eqearth"
	),
	style = "white",
	specified = character()
)

styles = list(
	v3 = list(
		modes = list(view = list(legend.bg.alpha = 0.8)),
		value.na = list(
			fill = "grey75",
			col = "grey75",
			col.raster = "#00000000",
			lty = "solid",
			lwd = NA,
			text = "Unknown",
			fontface = "plain",
			fill_alpha = 1,
			col_alpha = 1,
			col_alpha.raster = 0
		),
		value.null = list(
			fill = "grey95",
			col = "grey95",
			col.polygons = "grey40",
			lty = "solid",
			lwd = 0.2,
			text = "",
			fontface = "plain",
			fill_alpha = 1,
			col_alpha = 1,
			size = 0.2
		),
		value.blank = list(
			fill = "#00000000",
			col = "#00000000",
			lty = "blank",
			lwd = 0,
			text = "",
			fontface = "plain",
			fill_alpha = 0,
			col_alpha = 0
		),
		values.var = list(fill = list(seq = "brewer.yl_or_br", div = "brewer.rd_yl_gn", unord = "brewer.set3", ord = "brewer.yl_or_br"),
						  col = list(seq = "brewer.yl_or_br", div = "brewer.rd_yl_gn", unord = "brewer.set3", ord = "brewer.yl_or_br")),
		frame.lwd = 1,
		frame.r = 0,
		title.position = tm_pos_in(pos.h = "left", pos.v = "top", align.h = "left", align.v = "top", just.h = "left", just.v = "top"),
		title.size = 1.3,
		legend.position = tm_pos_auto_in(align.h = "left", align.v = "top", just.h = "left", just.v = "top"),
		legend.text.size = 0.7,
		legend.title.size = 0.9,
		legend.frame = FALSE,
		legend.frame.r = 0,
		legend.settings.portrait = list(item.height = c(rect = 1, symbols = 1, gradient = 1, bivariate = 1),
												 item.width = c(rect = 1, symbols = 1, gradient = 1.2, bivariate = 1),
												 item.r = 0,
												 item.space = c(rect = 0, symbols = 0, gradient = 0, bivariate = 0),
												 item.na.height = c(rect = NA, symbols = NA, gradient = 1.2, bivariate = 1),
												 item.na.width = c(rect = NA, symbols = NA, gradient = 1.2, bivariate = 1),
												 item.na.space = c(rect = 0, symbols = 0, gradient = 0, bivariate = 0),
												 title.padding  = c(0, 0, 0.25, 0),
												 ticks = list(rect = list(), symbols = list(), gradient = list(), bivariate = list()),
												 ticks.disable.na = c(rect = FALSE, symbols = FALSE, gradient = TRUE, bivariate = TRUE),
												 ticks.col = NA,
												 ticks.lwd = 1.5,
												 margins = c(0.4, 0.4, 0.4, 0.4),
												 item_text.margin = 0.25),
		legend.settings.landscape = list(item.height = c(rect = 1, symbols = 1, gradient = 1.2),
												  item.width = c(rect = 6, symbols = 3, gradient = 6),
												  item.r = 0,
												  item.space = c(rect = 0.2, symbols = 0.3, gradient = 0),
												  item.na.height = c(rect = NA, symbols = NA, gradient = 2),
												  item.na.width = c(rect = NA, symbols = NA, gradient = 6),
												  item.na.space = c(rect = 0.2, symbols = 0.3, gradient = 0.3),
												  title.padding  = c(0, 0, 0.25, 0),
												  ticks = list(rect = list(), symbols = list(), gradient = list(c(0.8, 1))),
												  ticks.disable.na = c(rect = FALSE, symbols = FALSE, gradient = TRUE),
												  ticks.col = NA,
												  ticks.lwd = 1.5,
												  margins = c(0.4, 0.4, 0.4, 0.4),
												  item_text.margin = 0.25)
	),
	gray = list(
		bg.color = "grey85",
		value.const = list(fill = "grey70",
						   fill.dots = "black",
						   fill = "grey50",
						   col.polygons = "grey20",
						   col = "black")
	),
	grey = list(
		bg.color = "grey85",
		value.const = list(fill = "grey70",
						   fill.dots = "black",
						   fill = "grey50",
						   col.polygons = "grey20",
						   col = "black")
	),
	natural = list(
		bg.color = "lightskyblue1",
		value.const = list(fill.polygons = "darkolivegreen3",
						   fill.dots = "firebrick",
						   fill = "tomato2",
						   col.lines = "steelblue",
						   col.text = "white",
						   col = "black"),
		value.na = list(
			fill = "white",
			col = "white",
			col.raster = "white"),
		value.null = list(
			fill = "grey70",
			col = "grey70",
			col.polygons = "grey70"),
		values.var = list(fill = list(seq = "brewer.yl_gn", div = "brewer.rd_yl_gn", unord = "brewer.set3", ord = "brewer.yl_gn"),
						  col = list(seq = "brewer.yl_gn", div = "brewer.rd_yl_gn", unord = "brewer.set3", ord = "brewer.yl_gn")),
		attr.color = "black",
		space.color = "white",
		legend.frame = TRUE,
		legend.bg.color = "grey90",
		frame = FALSE,
		earth_boundary = FALSE,
		basemap.server = "Esri.NatGeoWorldMap",
		basemap.alpha = 1),
	cobalt = list(bg.color = "#002240",
				  outer.bg = TRUE,
				  outer.bg.color = "#002240",
				  space.color = "#002240",
				  value.const = list(fill.polygons = "#0088FF",
				  				   fill = "#FF9D00",
				  				   col.text = "white",
				  				   col = "#FFEE80",
				  				   bgcol.labels_highlighted = "#002240",
				  				   bgcol = "#00000000"),
				  value.na = list(
				  	fill = "grey60",
				  	col = "grey60",
				  	col.raster = "grey60"),
				  value.null = list(
				  	fill = "grey40",
				  	col = "grey40",
				  	col.polygons = "grey40"),
				  values.var = list(fill = list(seq = "brewer.yl_gn", div = "brewer.rd_yl_gn", unord = "brewer.set3", ord = "brewer.yl_gn"),
				  				  col = list(seq = "brewer.yl_gn", div = "brewer.rd_yl_gn", unord = "brewer.set3", ord = "brewer.yl_gn")),
				  attr.color = "white",
				  legend.bg.color = "#002240",
				  chart.text.color = "white",
				  chart.title.color = "white",
				  basemap.server = "CartoDB.DarkMatter",
				  basemap.alpha = .5),
	albatross = list(bg.color = "#00007F",
					 outer.bg = TRUE,
					 outer.bg.color = "#000000",
					 space.color = "#000000",
					 value.const = list(fill.polygons = "#4C4C88",
					 				   fill = "#BFBFFF",
					 				   col.lines = "#BFBFFF",
					 				   col.text = "#FFE700",
					 				   col = "#00004C",
					 				   bgcol.labels_highlighted = "#00007F",
					 				   bgcol = "#00000000"),
					 value.na = list(
					 	fill = "grey60",
					 	col = "grey60",
					 	col.raster = "grey60"),
					 value.null = list(
					 	fill = "#4C4C88",
					 	col = "#4C4C88",
					 	col.polygons = "#4C4C88"),
					 values.var = list(fill = list(seq = "brewer.yl_or_rd", div = "brewer.rd_yl_gn", unord = "brewer.set3", ord = "brewer.yl_or_rd"),
					 				  col = list(seq = "brewer.yl_or_rd", div = "brewer.rd_yl_gn", unord = "brewer.set3", ord = "brewer.yl_or_rd")),
					 attr.color = "#BFBFFF",
					 legend.bg.color = "#00007F",
					 basemap.alpha = .5),
	beaver = list(bg.color = "#FFFFFF",
				  value.const = list(fill.polygons = "#FFE200",
				  				   fill = "#A30000",
				  				   col.lines = "#A30000",
				  				   col = "#00004C",
				  				   bgcol.labels_highlighted = "#FFFFFF",
				  				   bgcol = "#00000000"),
				  value.na = list(
				  	fill = "grey80",
				  	col = "grey80",
				  	col.raster = "grey80"),
				  value.null = list(
				  	fill = "grey95",
				  	col = "grey95",
				  	col.polygons = "grey95"),
				  values.var = list(fill = list(seq = "brewer.yl_or_br", div = "brewer.rd_yl_gn", unord = "brewer.dark2", ord = "brewer.yl_or_br"),
				  				  col = list(seq = "brewer.yl_or_br", div = "brewer.rd_yl_gn", unord = "brewer.dark2", ord = "brewer.yl_or_br")),
				  attr.color = "black"),
	bw = list(color.saturation = 0),
	classic = list(color.sepia_intensity = .7,
				   text.fontfamily = "serif",
				   frame = TRUE,
				   frame.double_line = TRUE,
				   inset.type = "rose"),
	watercolor = list(value.const = list(fill = "#D95F02",
										 fill.dots = "red",
										 col.lines = "red",
										 col = "black",
										 bgcol.labels_highlighted = "white",
										 bgcol = "#00000000"),
					  value.na = list(
					  	fill = "grey80",
					  	col = "grey80",
					  	col.raster = "grey80"),
					  value.null = list(
					  	fill = "#FDCDAC",
					  	col = "#FDCDAC",
					  	col.polygons = "#FDCDAC"),
					  values.var = list(fill = list(seq = "brewer.greens", div = "brewer.pi_yg", unord = "brewer.pastel1", ord = "brewer.greens"),
					  				  col = list(seq = "brewer.greens", div = "brewer.pi_yg", unord = "brewer.pastel1", ord = "brewer.greens")),
					  basemap.server = "Stadia.StamenWatercolor")
)

.defaultTmapStyles = list(
	gray = styles$gray,
	grey = styles$grey,
	natural = styles$natural,
	cobalt = styles$cobalt,
	albatross = styles$albatross,
	beaver = styles$beaver,
	bw = styles$bw,
	classic = styles$classic,
	watercolor = styles$watercolor,
	v3 = styles$v3,
	gray_v3 = c(styles$v3[setdiff(names(styles$v3), names(styles$gray))], styles$gray),
	grey_v3 = c(styles$v3[setdiff(names(styles$v3), names(styles$grey))], styles$grey),
	natural_v3 = c(styles$v3[setdiff(names(styles$v3), names(styles$natural))], styles$natural),
	cobalt_v3 = c(styles$v3[setdiff(names(styles$v3), names(styles$cobalt))], styles$cobalt),
	albatross_v3 = c(styles$v3[setdiff(names(styles$v3), names(styles$albatross))], styles$albatross),
	beaver_v3 = c(styles$v3[setdiff(names(styles$v3), names(styles$beaver))], styles$beaver),
	bw_v3 = c(styles$v3[setdiff(names(styles$v3), names(styles$bw))], styles$bw),
	classic_v3 = c(styles$v3[setdiff(names(styles$v3), names(styles$classic))], styles$classic),
	watercolor_v3 = c(styles$v3[setdiff(names(styles$v3), names(styles$watercolor))], styles$watercolor)
)

preprocess_meta_step1 = function(o, knit_opts = NULL) {
	within(o, {
		# if (!is.null(knit_opts)) {
		# 	scale = scale / knit_opts$fig.retina
		# }

		pc = list(sepia_intensity=color.sepia_intensity, saturation=color.saturation, color_vision_deficiency_sim=color_vision_deficiency_sim)
		color.sepia_intensity = NULL
		color.saturation = NULL
		color_vision_deficiency_sim = NULL

		# attr color is the standard for all component colors
		# attr.color.light is the 'opposite', only used in compass and scalebar
		#   it is white if the attr.color is dark and black if it is light
		attr.color.light = is_light(attr.color)
		attr.color = do.call(process_color, c(list(col = attr.color), pc))
		attr.color_light = if (attr.color.light) "#000000" else "#ffffff"
		attr.color_light = do.call(process_color, c(list(col = attr.color_light), pc))

		# bg.color = NA was used to disable background
		# as of 4.2, bg = FALSE will be used (#1119)

		# however, if a color is explicitly specified, override bg (also apply this to frame)
		for (nm in c("bg", "outer.bg", "frame")) {
			nmc = paste0(nm, ".color")
			if (nmc %in% o$calls && !is.null(get(nmc))) {
				assign(nm, TRUE)
			}
		}

		for (nm in names(o)[grep("bg\\.color$", names(o))]) {
			tmp = get(nm)
			if (!is.null(tmp) && is.na(tmp)) {
				nm2 = gsub("\\.color", "", nm)
				assign(nm, "white") #dummy
				assign(nm2, FALSE)
				cli::cli_alert("{.field [layout options]} use {.code {nm2} = FALSE} instead of {.code {nm} = NA}")
			}
			rm(tmp)
		}


		# color options: replace NA with attr.color
		# process colors: apply sepia and cvd sim
		for (nm in setdiff(names(o)[grep("color(\\.light|\\.dark)?$", names(o))], "attr.color")) {
			assign(nm, local({
				x = get(nm)
				if (is.na(x)) {
					if (length(grep("color(\\.light)$", nm))) attr.color_light else attr.color
				} else do.call("process_color", c(list(col=x), pc))
			}))
		}

		# radius: replace NA with r
		for (nm in names(o)[grep("\\.r$", names(o))]) {
			assign(nm, local({
				x = get(nm)
				if (is.na(x)) r else x
			}))
		}
		if (is.na(legend.settings.landscape$item.r)) legend.settings.landscape$item.r = r
		if (is.na(legend.settings.portrait$item.r)) legend.settings.portrait$item.r = r

		outer.margins = rep(outer.margins, length.out = 4)

		inner.margins.extra = rep(inner.margins.extra, length.out = 4)

		inner.margins = if (is.list(inner.margins)) {
			lapply(inner.margins, function(im) {
				rep(im, length.out = 4) + inner.margins.extra
			})
		} else {
			rep(inner.margins, length.out = 4) + inner.margins.extra
		}
		inner.margins.extra = NULL


		panel.label.rot = rep_len(panel.label.rot, 4L)

		earth_boundary.lwd = earth_boundary.lwd * scale

		# fontface
		for (nm in names(o)[grep("fontface", names(o), fixed = TRUE)]) if (is.null(get(nm))) assign(nm, text.fontface)
		for (nm in names(o)[grep("fontfamily", names(o), fixed = TRUE)]) if (is.null(get(nm))) assign(nm, text.fontfamily)

		# special case: fontface is a visual variable. Therefore, check if value.const etc is NULL if so, replace
		if (is.null(value.const$fontface)) value.const$fontface = text.fontface
		if (is.null(value.blank$fontface)) value.blank$fontface = text.fontface
		if (is.null(value.na$fontface)) value.na$fontface = text.fontface
		if (is.null(value.null$fontface)) value.null$fontface = text.fontface



	})
}

step2_data = function(tm) {
	dev = getOption("tmap.devel.mode")

	tmo = tm$tmo
	o = tm$o
	aux = tm$aux
	cmp = tm$cmp
	prx = tm$prx

	if (is.null(tmo)) {
		return(	list(tmo = NULL, aux = aux, cmp = cmp, prx = prx, o = o))
	}

	groupnames = paste0("group", seq_along(tmo))
	fl = list(1L, 1L, 1L)
	assign("fl", fl, envir = .TMAP)

	# to reset the legends (which are  temporarily stored in .TMAP environment)
	legends_init()
	charts_init()

	grps = mapply(function(tmg, groupname) {
		tmf = tmg$tmf
		dt = tmg$tms$dt

		type_ids = tmg$tms$type_ids
		type_vars = tmg$tms$type_vars # used to split dt vars in the layer functions, e.g. for networks with edges and nodes

		if ("by1__" %in% names(dt) && o$rev1) dt[, by1__ := (o$fn[1]+1L)-by1__]
		if ("by2__" %in% names(dt) && o$rev2) dt[, by2__ := (o$fn[2]+1L)-by2__]
		if ("by3__" %in% names(dt) && o$rev3) dt[, by3__ := (o$fn[3]+1L)-by3__]

		shpvars = tmg$tms$smeta$vars

		# step2_data_grp_prepare
		#tmf_meta = step2_data_grp_prepare(tmg$tmf, grpvars, dt)

		layernames = paste0("layer", seq_along(tmg$tmls))
		lrs = mapply(function(tml, layername) {
			glid = paste0(groupname, "_", layername)

			#cat("step2_grp_lyr======================\n")

			if (!is.null(type_ids)) {
				ids = tml$layer %in% names(type_ids)
				if (any(ids)) {
					lid = tml$layer[[which(ids)[1]]]
					tmapids = type_ids[[lid]]
					dt = data.table::copy(dt[dt$tmapID__ %in% tmapids, c(type_vars[[lid]], "sel__"), with = FALSE])
				}

			}

			gp = tml$gpar
			tp = tml$tpar

			plot.order = tml$plot.order

			group = if (is.na(tml$group)) tmg$tms$shp_name else as.character(tml$group)
			group.control = as.character(tml$group.control)

			.TMAP$popup.format = tml$popup.format

			mfun = tml$mapping.fun

			# args will be passed on to the scale functions (in case needed)
			# they also will be used in step 3 (trans) and step 4 (mapping)
			if (is.null(dt)) {
				# bypass (e.g. in case of pmtiles)
				trans = list()
				mapping = mapply(getdts, tml$mapping.aes, names(tml$mapping.aes), SIMPLIFY = FALSE, MoreArgs = list(p = gp, q = tmf, o = o, dt = dt, shpvars = shpvars, layer = tml$layer, group = group, glid = glid, mfun = mfun, args = tml$mapping.args, plot.order = plot.order))
				#mapping = list()
			} else {
				trans = mapply(getdts, tml$trans.aes, names(tml$trans.aes), SIMPLIFY = FALSE, MoreArgs = list(p = tp, q = tmf, o = o, dt = dt, shpvars = shpvars, layer = tml$layer, group = group, glid = glid,mfun = mfun, args = tml$trans.args, plot.order = plot.order))

				mapping = mapply(getdts, tml$mapping.aes, names(tml$mapping.aes), SIMPLIFY = FALSE, MoreArgs = list(p = gp, q = tmf, o = o, dt = dt, shpvars = shpvars, layer = tml$layer, group = group, glid = glid,mfun = mfun, args = tml$mapping.args, plot.order = plot.order))
			}


			dts_trans = cbind_dts(lapply(trans, function(x) x$dt), plot.order)
			trans_legend = lapply(trans, function(x) x$leg)

			dts_mapping = cbind_dts(lapply(mapping, function(x) x$dt), plot.order)
			mapping_legend = lapply(mapping, function(x) x$leg)
			if (dev) timing_add(s4 = "combine")
			if (dev) timing_add(s3 = paste0("layer ", tml$layer))


			if (!length(tml$popup.vars) || is.null(dt)) {
				popup.data = NULL
			} else {
				# tml$popup.vars = tml$popup.vars
				# 	intersect(tml$popup.vars, names(dt)) # in case type_ids !NULL

				popupvars = intersect(tml$popup.vars, names(dt))

				if (!length(popupvars)) {
					popup.data = NULL
				} else {
					if (!is.null(names(tml$popup.vars))) {
						popupvars = structure(popupvars, names = names(tml$popup.vars)[match(popupvars, tml$popup.vars)])
					}

					popup.data = copy(dt[, c(popupvars, "tmapID__"), with = FALSE])
					if (!is.null(names(popupvars))) popup.data = data.table::setnames(popup.data, old = unname(popupvars), new = names(popupvars))
				}
			}
			hover.data = if (tml$hover == "" || is.null(dt)) {
				NULL
			} else {
				data.table(hover = as.character(dt[[tml$hover]]), tmapID__ = dt$tmapID__)
			}
			id.data = if (tml$id == "" || is.null(dt)) {
				NULL
			} else {
				data.table(id = as.character(dt[[tml$id]]), tmapID__ = dt$tmapID__)
			}

			format_called = attr(tml$popup.format, "called")
			if (length(format_called) > 0L) {
				.TMAP$popup.format[format_called] = tml$popup.format[format_called]
			}


			list(trans_dt = dts_trans,
				 trans_legend = trans_legend,
				 trans_fun = tml$trans.fun,
				 trans_args = tml$trans.args,
				 trans_apply = tml$trans.apply_to,
				 mapping_dt = dts_mapping,
				 mapping_legend = mapping_legend,
				 mapping_fun = tml$mapping.fun,
				 mapping_args = tml$mapping.args,
				 lid = tml$lid,
				 group = group,
				 glid = glid,
				 group.control = group.control,
				 popup.data = popup.data,
				 popup.format = .TMAP$popup.format,
				 hover.data = hover.data,
				 id.data = id.data,
				 plot.order = plot.order, # passed on for step 3 non-data driven transformation
				 gp = gp,
				 tp = tp)
		}, tmg$tmls, layernames, SIMPLIFY = FALSE, USE.NAMES = FALSE)
		if (length(lrs)) names(lrs) = layernames

		shpDT = data.table(shpTM = list(tmg$tms$shpTM))
		if (dev) timing_add(s2 = "group")

		list(layers = lrs, shpDT = shpDT)
	}, tmo, groupnames, SIMPLIFY = FALSE, USE.NAMES = FALSE)
	names(grps) = groupnames
	#attr(grps, "fl") = fl
	#attr(grps, "main") = attr(tmo, "main")
	#attr(grps, "crs") = attr(tmo, "crs")

	#tmf = get_tmf(lapply(tmo, function(tmoi) tmoi$tmf))



	# facet labels "_old' were obtained in step2, and still determine data levels
	# currently, facet labels are determined in step1, but they need to be the same as the data labels
	# o$fl_old = get("fl", envir = .TMAP)
	# o$fn_old = sapply(o$fl_old, function(f) {
	# 	if (is.character(f)) length(f) else f
	# })
	# o$fl_old = lapply(o$fl_old, function(f) {
	# 	if (is.character(f)) f else NULL
	# })
	#
	# if (!identical(o$fn, o$fn_old) || !identical(o$fl, o$fl_old)) {
	# 	po(o$fn_old, o$fl_old, o$fn, o$fl)
	# 	stop("fl not identical")
	# }



	#cat("fl old:\n")
	#print(o$fl_old)


	#o = c(o, tmf)

	# attr(grps, "is.wrap") = tmo[[1]]$tmf$is.wrap
	# attr(grps, "nrows") = tmo[[1]]$tmf$nrows
	# attr(grps, "ncols") = tmo[[1]]$tmf$ncols

	o = within(o, {
		if (rev1) fl[[1]][] = rev(fl[[1]][])
		if (rev2) fl[[2]][] = rev(fl[[2]][])
		if (rev3) fl[[3]][] = rev(fl[[3]][])
	})

	list(tmo = grps, aux = aux, cmp = cmp, prx = prx, o = o)
}

leaflet_pos = function(pos) {
	if (pos$type %in% c("out", "autoout")) {
		sel = c("cell.v", "cell.h")
	} else {
		sel = c("pos.v", "pos.h")
	}
	x = tolower(unlist(pos[sel]))

	if (x[1] %in% c("center", "centre")) x[1] = "top"
	if (x[2] %in% c("center", "centre")) x[2] = "left"

  paste(x, collapse = "")
}


gp_to_lpar = function(gp, mfun, shape = 20, pick_middle = TRUE, size_factor = 20, rename_prop = TRUE) {
	# create a list of gp elements

	lst = c(list(fill = {if (!all(is.na(gp$fill))) gp$fill else "#000000"},
				 col = {if (!all(is.na(gp$col))) gp$col else "#000000"},
				 fill_alpha = {if (!all(is.na(gp$fill_alpha))) gp$fill_alpha else 0},
				 col_alpha = {if (!all(is.na(gp$col_alpha))) gp$col_alpha else 0},
				 lwd = {if (!all(is.na(gp$lwd))) gp$lwd else 0},
				 lty = {if (!all(is.na(gp$lty))) lty2dash(gp$lty, gp$lwd) else "none"},
				 size = {if (!all(is.na(gp$size))) gp$size else 1},
				 shape = {if (!all(is.na(gp$shape))) gp$shape else shape}))

	lst_isnum = c(fill = FALSE,
				  col = FALSE,
				  fill_alpha = TRUE,
				  col_alpha = TRUE,
				  lwd = TRUE,
				  lty = FALSE,
				  size = TRUE,
				  shape = TRUE)

	if (rename_prop) {
		newNames = c("fillColor",
					  "color",
					  "fillOpacity",
					  "opacity",
					  "stroke-width",
					  "stroke-dash",
					  "size",
					  "shape")
		names(lst) = newNames
		names(lst_isnum) = newNames
	}


	lst = mapply(function(lsti, nm, isnum) {
		if (!is.character(lsti)) return(lsti)

		if (nchar(lsti[1]) > 50) {
			x = cont_split(lsti)
			x = lapply(x, function(i) {
				i[i=="NA"] <- NA
				i
			})
			if (isnum) x = lapply(x, as.numeric)
			if (nm %in% c("fillColor", "color")) {
				# create ramp of 10
				not1 = which(vapply(x, length, FUN.VALUE = integer(1)) != 1L)
				if (length(not1) != length(x)) {
					# add NA
					xNA = x[[length(x)]]
				} else {
					xNA = NULL
				}
				x[not1[-1]] = lapply(x[not1[-1]], tail, -1)
				x = na.omit(unlist(x[not1]))
				x = c(x[seq(1, length(x), length.out = length(not1) * 2 + 1)], xNA)
			} else if (pick_middle) {
				x = sapply(x, function(i) {
					if (all(is.na(i))) NA else {
						sq = c(5,6,4,7,3,8,2,9,1,10) # priority for middle values
						i[sq[which(!is.na(i)[sq])[1]]]
					}
				})
			}
			return(x)

		} else {
			return(lsti)
		}
	}, lst, names(lst), lst_isnum[names(lst)], SIMPLIFY = FALSE)

	pch2shp = c("rect", "circle", "triangle", "plus", "cross", "diamond", "triangle",
				"cross", "star", "diamond", "circle", "polygon", "plus", "cross",
				"triangle", "rect", "circle", "triangle", "diamond", "circle",
				"circle", "circle", "rect", "diamond", "triangle", "polygon", "stadium") # shapes for pch 0:25 + 26 for stadium (NOTE: last one is a triangle upside-down. Since 21:25 are the defaults, and a polygon is chosen to differentiate from the other triangle)
	lst$shape = get_pch_names(lst$shape)

	if ("tm_data_lines" %in% mfun) lst$shape = "line"

	lst$width = lst$size * size_factor
	lst$height = lst$size * size_factor
	#lst$width[]
	lst$size = NULL
	lst
}



make_equal_list = function(x) {
	cls = class(x)
	n = max(vapply(x, length, integer(1)))
	structure(lapply(x, rep, length.out = n), class = cls)
}


tmapLeaflet_legend = function(cmp, lf, o, orientation) {
	group = cmp$group
	leg_className = paste("info legend", gsub(" ", "", group, fixed = TRUE))



	if ("layerId" %in% names(cmp)) {
		layerId = cmp$layerId
		group = "always_on"
	} else {
		layerId =  paste0("legend", sprintf("%02d", .TMAP_LEAFLET$leg_id)) # "legend401" #todo
		.TMAP_LEAFLET$leg_id = .TMAP_LEAFLET$leg_id + 1
	}



	lab = cmp$labels
	val = c(cmp$dvalues)
	title = if (nonempty_text(cmp$title)) expr_to_char(cmp$title) else NULL

	if (!is.null(title) && !is.null(cmp$title.color)) {

		if (cmp$type == "gradient") {
			title = htmltools::tags$div(
				title,
				style = paste0('color: ', cmp$title.color, ';')
			)
		} else {
			title = htmltools::tags$div(
				title,
				style = paste0('font-size: 14px; text-align: left; margin-bottom: 5px; color: ', cmp$title.color, ';')
			)
		}

	}

	legpos = leaflet_pos(cmp$position)

	lf2 = if (cmp$type == "none") {
		#message("Text based legends not supported in view mode")
		lf
	} else if (cmp$type == "gradient") {
		breaks_show =  ("is_discrete" %in% names(cmp)) && cmp$is_discrete


		nbins = length(val)

		incl.na = cmp$na.show

		if (incl.na) {
			sel = head(cmp$labels_select, -1)
		} else {
			sel = cmp$labels_select
		}

		bins = val[sel]
		val = val[sel]

		if (!breaks_show) {
			if (cmp$scale == "intervals") cli::cli_warn("{.field tm_scale_intervals} {.code label.style = {.str continuous}} implementation in view mode work in progress")
			if (!head(sel, 1)) {
				val = c(cmp$limits[1], val)
			} else {
				val = c(val[1] - (val[2] - val[1]) * 0.5, val)
			}

			if (!tail(sel, 1)) {
				val = c(val, cmp$limits[2])
			} else {
				val = c(val, val[length(val)] + diff(tail(val, 2))/2)
			}
		} else {
			brks = val
		}

		vary = if ("fill" %in% cmp$varying) "fillColor" else "color"
		#vary_alpha = paste0(vary, "_alpha")

		incl.na = cmp$na.show
		if (incl.na) {
			pal = head(cmp$gp2[[vary]], -1)
			colNA = tail(cmp$gp2[[vary]], 1)
			textNA = lab[length(lab)]
			labs = head(lab, -1)[sel]
			val = c(val, NA)
			#labs = c(labs, "")
		} else {
			pal = cmp$gp2[[vary]]
			colNA = NA
			textNA = NA
			labs = lab[sel]
		}

		pfun = function(x) {
			r = brks[c(1, length(brks))]
			d = (x * (r[2] - r[1])) + r[1]
			ids = cut(d, breaks = brks, include.lowest = TRUE, right = TRUE, labels = FALSE)
			pal2[ids]
		}
		# pfun = function(x) {
		# 	ifelse(x < 0.5, "blue", "red")
		# }

		if (breaks_show) {
			# when labels_cutpoints is enabled in tm_scale_intervals
			pal2 = unname(head(pal, -1)[-1][seq(2L, length(pal)-2L, by = 2L)])
			pal = colorNumeric(palette = pfun,
							   domain = val,
							   na.color=colNA,
							   alpha = FALSE)
		} else {
			pal = colorNumeric(palette = pal,
							   domain = val,
							   na.color=colNA,
							   alpha = FALSE)
		}


		brks = pretty(cmp$limits, 7)

		trns = function(x) {
			f = (x-cmp$limits[1]) / diff(cmp$limits)
			y = cmp$tr$fun(cmp$limits)
			x2 = y[1] + diff(y) * f
			cmp$tr$rev(x2)
		}

		opacity = if (vary == "fillColor") {
			cmp$gp2$fillOpacity
		} else {
			cmp$gp2$opacity
		}

		if (orientation == "horizontal") {
			cli::cli_inform("{.field [landscape legend in view mode]} doesn't support labels yet",
							.frequency_id = "landscape_legend_view",
							.frequency = "once")
			labs = NULL

		}

		lf %>% leaflegend::addLegendNumeric(position=legpos,
										   orientation = orientation,
										   group = group,
										   height = cmp$height,
										   width = cmp$width,
										   pal=pal,
										   values=val,
										   # numberFormat = function(x) {
										   # 	prettyNum(trns(x), format = "f", big.mark = ",", digits =
										   # 			  	3, scientific = FALSE)
										   # },
										   labels = labs,
										   bins = bins,
										   naLabel = textNA,
										   title=title,
										   fillOpacity=opacity,
										   layerId = layerId,
										   className = leg_className)

	} else {
		vary = if ("fill" %in% cmp$varying) "fill" else if ("col" %in% cmp$varying) "col" else NA



		gp2 = make_equal_list(cmp$gp2)
		#gp2$baseSize = gp2$baseSize[1]

		sn = suppressWarnings(as.numeric(gp2$shape))
		sid = which(!is.na(sn))

		#symbols = do.call(Map, c(list(f = leaflegend::makeSymbol), cmp$gp2))

		# alternative:
		if (length(sid)) {
			gp2$shape[sid] = "circle" # as dummy
		}

		names(gp2)[names(gp2) == 'stroke-width'] = "strokeWidth"
		names(gp2)[names(gp2) == 'stroke-dash'] = "dashArray"
		gp2$baseSize = 20
		#symbols = do.call(leaflegend::makeSymbolIcons, gp2)#$iconUrl


		symbols = do.call(makeSymbolIcons2, gp2)
		#po(sort(gp2$width, decreasing = T))

		symbols$iconWidth = gp2$width#, length(symbols$iconUrl))
		symbols$iconHeight = gp2$height#, length(symbols$iconUrl))

		if (length(sid)) {
			iconLib <- get("shapeLib", envir = .TMAP)[sn[sid]-999]
			symbols_icons <- merge_icons(iconLib)

			size = gp2$width[sid] / gp2$baseSize * cmp$layer_args$icon.scale/3

			for (i in seq_along(sid)) {
				symbols$iconUrl[sid[i]] = symbols_icons$iconUrl[i]
				symbols$iconWidth[sid[i]] <- symbols_icons$iconWidth[i] * size[i]
				symbols$iconHeight[sid[i]] <- symbols_icons$iconHeight[i] * size[i]
				if (all(c("iconAnchorX", "iconAnchorY") %in% names(symbols_icons))) {
					symbols$iconAnchorX[sid[i]] <- symbols_icons$iconAnchorX[i] * size[i]
					symbols$iconAnchorY[sid[i]] <- symbols_icons$iconAnchorY[i] * size[i]

				}
			}
		}

		#symbols = symbols$iconUrl


		lf %>% leaflegend::addLegendImage(symbols$iconUrl,
										  group = group,
										 labels = lab,
										 width = symbols$iconWidth + max(2*gp2$strokeWidth),
										 height = symbols$iconHeight + 2*gp2$strokeWidth,
										 position = legpos,
										 orientation = orientation,
										 labelStyle = "font-size: 14px; vertical-align: middle; margin: 0px;",
										 title = title,
										 layerId = layerId,
										 className = leg_className)
	}
	backg <- htmltools::tags$style(paste0("#", layerId, " { background: ", substr(cmp$bg.color,1,7), "; opacity: ", cmp$bg.alpha, "}"))
	if (!.TMAP$in.shiny) {
		htmlwidgets::prependContent(lf2, backg)
	} else {
		lf2
	}
}

#' @export
tmapLeafletCompPlot.tm_legend_portrait = function(comp, lf, o) {
	tmapLeaflet_legend(comp, lf, o, orientation = "vertical")
}

#' @export
tmapLeafletCompPlot.tm_legend_landscape = function(comp, lf, o) {
	tmapLeaflet_legend(comp, lf, o, orientation = "horizontal")
}


tmapLeafletComp = function(comp, o, facet_row = NULL, facet_col = NULL, facet_page, class, stack, stack_auto, pos.h, pos.v, bbox) {
	lf = get_lf(facet_row, facet_col, facet_page)

	rc_text = frc(facet_row, facet_col)

	# grp = tmapGetCompGroupArgs(comp) none of the tm_components args are used yet in view mode

	for (cmp in comp) {
		lf = tmapLeafletCompPlot(cmp, lf, o)
	}



	assign_lf(lf, facet_row, facet_col, facet_page)
	NULL
}

#' @export
tmapGridCompPrepare.tm_legend_portrait = function(comp, o) {
	within(comp, {
		bg.color = do.call("process_color", c(list(col=bg.color), o$pc))
		title.color = do.call("process_color", c(list(col=title.color), o$pc))
		text.color = do.call("process_color", c(list(col=text.color), o$pc))

		type = if ("biv" %in% names(attributes(gp$fill))) {
			"bivariate"
		} else if (!is.na(gp$fill[1]) && any(nchar(gp$fill) > 20) || !is.na(gp$fill_alpha[1]) && any(nchar(gp$fill_alpha) > 20) ||
				   !is.na(gp$col[1]) && any(nchar(gp$col) > 20) || !is.na(gp$col_alpha[1]) && any(nchar(gp$col_alpha) > 20)) {
			"gradient"
		} else if (!is.na(gp$shape[1])) {
			"symbols"
		} else if ("tm_data_lines" %in% mfun) {
			"lines"
		} else if ("tm_data_text" %in% mfun) {
			"text"
		} else {
			"rect"
		}

		text.size = text.size * o$scale
		title.size = title.size * o$scale

		ticks.lwd = ticks.lwd * o$scale

		title.align = get_vector_id(title.align, type)
		xlab.align = get_vector_id(xlab.align, type)
		ylab.align = get_vector_id(ylab.align, type)

		if (!is.na(height) && height < 0) {
			height = -height / absolute_fontsize
		}
		if (!is.na(width) && width < 0) {
			width = -width / absolute_fontsize
		}

		# determine empty areas
		if (type == "gradient") {
			var = if (!is.na(gp$fill[1]) && any(nchar(gp$fill) > 20)) {
				"fill"
			} else if (!is.na(gp$fill_alpha[1]) && any(nchar(gp$fill_alpha) > 20)) {
				"fill_alpha"
			} else if (!is.na(gp$col[1]) && any(nchar(gp$col) > 20)) {
				"col"
			} else {
				"col_alpha"
			}
			gradient_ht_fracs = local({
				values  = cont_split(gp[[var]])
				k = comp$nitems - comp$na.show
				c(sum(values[[1]] == "NA") / length(values[[1]]),
				  sum(values[[k]] == "NA") / length(values[[k]]))
			})
		}


		gpar = gp_to_gpar(gp, o = o, type = comp$type)
	})
}


#' @export
tmapGridCompHeight.tm_legend_portrait = function(comp, o) {

	nlev = if (comp$type == "bivariate") attr(comp$gp$fill, "m") + 1L else comp$nitems




	textS = comp$text.size #* o$scale
	titleS = if (comp$title == "") 0 else comp$title.size * number_text_lines(comp$title) # * o$scale

	space = get_legend_option(comp$item.space, comp$type)
	spaceNA = get_legend_option(comp$item.na.space, comp$type)
	height = get_legend_option(comp$item.height, comp$type)
	heightNA = get_legend_option(comp$item.na.height, comp$type)


	if (is.na(heightNA)) heightNA = height

	# legend height can be 3: NA (components determine height), finite number (number of text lines), or Inf (whole height)
	# for the latter two, there are 3 ways of stretching the legend: padding (space between items), items (heights of all items), or itemsNNA (heights of non-NA items)

	if (comp$type == "symbols") {
		shps = rep(comp$gpar$shape, length.out = nlev)
		item_heights = pmax(height, rep(comp$gpar$size / textS, length.out = nlev) * ifelse(shps > 999, comp$layer_args$icon.scale, 1))
		comp$stretch = if (!is.na(comp$height)) "padding" else "none"
	} else if (comp$type %in% c("rect", "lines", "bivariate")) {
		item_heights = rep(height, nlev)
		if (comp$na.show) item_heights[nlev] = heightNA
		comp$stretch = if (!is.na(comp$height)) "items" else "none"
	} else if (comp$type == "gradient") {
		item_heights = rep(height, nlev)



		if (comp$na.show) item_heights[nlev] = heightNA
		comp$stretch = if (!is.na(comp$height)) "itemsNNA" else "none"
	} else if (comp$type == "text") {
		item_heights = pmax(height, rep(comp$gpar$size / textS, length.out = nlev))
		comp$stretch = if (!is.na(comp$height)) "padding" else "none"
	}
	titleP = comp$title.padding[c(3,1)] * titleS * o$lin
	titleH = titleS * o$lin

	textH = textS * o$lin

	if (comp$type == "bivariate") {
		xlabS = if (comp$xlab == "") 0 else comp$xlab.size * number_text_lines(comp$xlab)
		xlabP = comp$xlab.padding[c(3,1)] * xlabS * o$lin
		xlabH = xlabS * o$lin

	} else {
		xlabP = c(0, 0)
		xlabH = 0
	}



	marH = comp$margins[c(3,1)] * textS * o$lin

	if (nlev == 1) {
		items_all = item_heights
		if (comp$stretch == "padding") comp$stretch = "items"
	} else {
		item_space = c(rep(space, nlev - 1 - comp$na.show), {if (comp$na.show) spaceNA else NULL})
		items_all = c(rbind(item_heights[-nlev], item_space), item_heights[nlev])
	}

	if (comp$type == "gradient") {
		# Cut empty space
		if (comp$gradient_ht_fracs[1] != 0) {
			# use top margin
			if (comp$labels_select[1]) {
				ih1 = min((item_heights[1] - 1)/2, item_heights[1] * comp$gradient_ht_fracs[1])
			} else {
				ih1 = item_heights[1] * comp$gradient_ht_fracs[1]
			}

			marH[1] = marH[1] - (ih1 * textS * o$lin)
		}
		if (comp$gradient_ht_fracs[2] != 0) {
			k = nlev - comp$na.show
			if (comp$labels_select[k]) {
				ihk = min((item_heights[k] - 1)/2, item_heights[k] * comp$gradient_ht_fracs[2])
			} else {
				ihk = item_heights[k] * comp$gradient_ht_fracs[2]
			}
			if (comp$na.show) {
				items_all[k * 2] = items_all[k * 2] - ihk
			} else {
				marH[2] = marH[2] - (ihk * textS * o$lin)
			}
		}
	}

	hs = c(titleP[1], titleH, titleP[2], marH[1], items_all * textS * o$lin, xlabP[1], xlabH, xlabP[2], marH[2])

	item_ids = seq(5, by = 2, length.out = nlev)
	pad_ids = seq(6, by = 2, length.out = nlev - 1L)


	hsu = if (comp$stretch == "padding") {
		set_unit_with_stretch(hs, pad_ids)
	} else if (comp$stretch == "items") {
		set_unit_with_stretch(hs, item_ids)
	} else if (comp$stretch == "itemsNNA") {
		if (comp$na.show) set_unit_with_stretch(hs, head(item_ids, -1)) else set_unit_with_stretch(hs, item_ids)
	} else {
		sides = switch(comp$position$align.v, top = "second", bottom = "first", "both")
		set_unit_with_stretch(hs, sides = sides)
	}


	Hin = if (comp$stretch == "none") sum(hs) else comp$height * textS * o$lin

	comp$flexRow = NA
	comp$Hin = Hin
	#comp$hs = hs
	comp$hsu = hsu

	comp$item_ids = item_ids + 1L

	comp

}

fontface2nr = function(face) {
	# needed for strwidth
	switch(face,
		   "plain" = 1,
		   "bold" = 2,
		   "italic" = 3,
		   "bold.italic" = 4,
		   "oblique" = 3,
		   "cyrillic.oblique" = 3,
		   1)
}


#' @export
tmapGridCompWidth.tm_legend_portrait = function(comp, o) {

	ni = if (comp$type == "bivariate") attr(comp$gp$fill, "n") else 1L
	mi = if (comp$type == "bivariate") attr(comp$gp$fill, "m") else 1L


	textS = comp$text.size #* o$scale
	titleS = if (comp$title == "") 0 else comp$title.size #* o$scale

	marW = comp$margins[c(2,4)] * textS * o$lin

	width = get_legend_option(comp$item.width, comp$type)
	item_text.margin = get_vector_id(comp$item_text.margin, comp$type)


	item_widths = if (comp$type == "symbols") {
		shps = rep(comp$gpar$shape, length.out = comp$nitems)
		pmax(width, rep(comp$gpar$size / textS, length.out = comp$nitems) * ifelse(shps > 999, comp$layer_args$icon.scale, 1))
	} else rep(width, length.out = ni)


	if (comp$type == "bivariate") {
		item_widths_max = max(item_widths)
	} else {
		item_widths_max = max(item_widths)
	}




	tW = ifelse(comp$title == "", 0, graphics::strwidth(comp$title, units = "inch", cex = titleS, family = comp$title.fontfamily, font = fontface2nr(comp$title.fontface)) + sum(comp$title.padding[c(2,4)]) * o$lin * titleS)

	if (comp$type == "bivariate") {
		txtRowW = ifelse(comp$ylab == "", 0, graphics::strwidth(comp$ylab, units = "inch", cex = comp$ylab.size, family = comp$ylab.fontfamily, font = fontface2nr(comp$ylab.fontface)))

		margRowW = ifelse(txtRowW == 0, 0, item_text.margin * comp$ylab.size * o$lin)
	}

	if (comp$type == "bivariate") {
		iW = graphics::strwidth(comp$labels[1:mi], units = "inch", cex = textS, family = comp$text.fontfamily, font = fontface2nr(comp$text.fontface)) + (item_widths_max * ni + item_text.margin) * textS * o$lin
	} else {
		iW = graphics::strwidth(comp$labels, units = "inch", cex = textS, family = comp$text.fontfamily, font = fontface2nr(comp$text.fontface)) + (item_widths_max + item_text.margin) * textS * o$lin
	}

	colW = max(tW, iW)

	if (comp$type == "bivariate") {
		txtW = colW - (item_widths_max * ni + item_text.margin) * textS * o$lin
	} else {
		txtW = colW - (item_widths_max * ni + item_text.margin) * textS * o$lin
	}

	n = switch(comp$position$align.h, left = c(0, 1), right = c(1, 0), c(0.5, 0.5))

	if (comp$type == "bivariate") {
		wsu = grid::unit(c(marW[1],
						   n[1],
						   txtRowW,
						   margRowW,
						   txtW,
						   item_text.margin * textS * o$lin,
						   rep(item_widths_max * textS * o$lin, ni),
						   n[2],
						   marW[2]), units = c("inch", "null", rep("inch", 4+ni), "null", "inch"))
		comp$flexCol = 5 #column that will shrink if the null columns will otherwise be sizes < 0. (happens if legend.width < minimal width)
		Win = if (is.na(comp$width)) sum(as.numeric(wsu)[c(1, 3:(6+ni), 8+ni)]) else comp$width * textS * o$lin

	} else {
		wsu = grid::unit(c(marW[1],
						   n[1],
						   item_widths_max * textS * o$lin,
						   item_text.margin * textS * o$lin,
						   txtW,
						   n[2],
						   marW[2]), units = c("inch", "null", rep("inch", 2+ni), "null", "inch"))
		comp$flexCol = 5 #column that will shrink if the null columns will otherwise be sizes < 0. (happens if legend.width < minimal width)
		Win = if (is.na(comp$width)) sum(as.numeric(wsu)[c(1, 3:5, 7)]) else comp$width * textS * o$lin
	}


	comp$Win = Win
	comp$wsu = wsu

	comp

}

add_user_specified_values = function(gp, usr) {
	if (!length(usr)) return(gp)

	for (v in names(usr)) {
		fun = paste0("tmapValuesCheck_", v)
		val = usr[[v]]
		if (!do.call(fun, list(x = val, is_var = TRUE))) stop("Values assigned to visual variable ", v, " in tm_legend incorrect")

		gp[[v]] = val
	}
	gp
}



#' @export
tmapGridCompPlot.tm_legend_portrait = function(comp, o, fH, fW) {

	if (comp$type != "gradient") comp$labels_select = TRUE # labels_select only needed for continuous legends (#1039)

	icons = all(comp$gp$shape >= 1000)

	# replace gp visual values with user-specified used (e.g. tm_shape(World) + tm_polygons("HPI", fill.legend = tm_legend(col = "red")))
	comp$gp = add_user_specified_values(comp$gp, comp[intersect(names(comp), names(comp$gp))])

	# icons replaced by 'normal' shapes: apply icon.scale
	if (icons && all(comp$gp$shape < 1000) && !is.null(comp$layer_args$icon.scale)) {
		comp$gp$size = comp$gp$size * comp$layer_args$icon.scale
	}

	textS = comp$text.size * comp$scale #* o$scale

	titleS = if (comp$title == "") 0 else comp$title.size * comp$scale #* o$scale

	n = if (comp$type == "bivariate") attr(comp$gp$fill, "n") else 1
	m = if (comp$type == "bivariate") attr(comp$gp$fill, "m") else 1

	nlev = comp$nitems

	wsu = comp$wsu
	hsu = comp$hsu


	vp = grid::viewport(layout = grid::grid.layout(ncol = length(wsu),
												   nrow = length(hsu),
												   widths = wsu,
												   heights = hsu))

	grobBG = if (getOption("tmap.design.mode")) rectGrob(gp=gpar(fill="#CAB2D6")) else NULL

	if (is.na(comp$title.align)) comp$title.align = comp$position$align.h

	shiftCol = if (comp$type == "bivariate") 2L else 0L

	titleGP = grid::gpar(col = comp$title.color, cex = titleS, fontface = comp$title.fontface, fontfamily = comp$title.fontfamily, alpha = comp$title.alpha)

	if (comp$title.align == "left") {
		grTitle = gridCell(3, (2 + shiftCol):(length(comp$wsu)-1), grid::textGrob(comp$title, x = grid::unit(comp$title.padding[2] * titleS * o$lin, units = "inch"), just = "left", gp = titleGP))
	} else if (comp$title.align == "right") {
		grTitle = gridCell(3, (2 + shiftCol):(length(comp$wsu)-1), grid::textGrob(comp$title, x = grid::unit(1, "npc") - grid::unit(comp$title.padding[4] * titleS * o$lin, units = "inch"), just = "right", gp = titleGP))
	} else {
		grTitle = gridCell(3, (2 + shiftCol):(length(comp$wsu)-1), grid::textGrob(comp$title, x = 0.5, just = "center", gp = titleGP))
	}


	if (comp$type == "bivariate") {
		if (comp$ylab != "") {
			# ignoring just for now, assuming "left"
			ylabS = comp$ylab.size * comp$scale
			grYlab = gridCell(comp$item_ids[1:m], 3,
							  grid::textGrob(comp$ylab,
							  			   rot = comp$ylab.rot,
							  			   x = grid::unit(comp$ylab.padding[2] * ylabS * o$lin, units = "inch"),
							  			   just = "left",
							  			   gp = grid::gpar(col = comp$ylab.color, cex = ylabS, fontface = comp$ylab.fontface, fontfamily = comp$ylab.fontfamily)))
		} else {
			grYlab = NULL
		}

		if (comp$xlab != "") {
			# ignoring just for now, assuming "left"
			xlabS = comp$xlab.size * comp$scale
			grXlab = gridCell(comp$item_ids[m] + 4, 7:(6+n),
							  grid::textGrob(comp$xlab,
							  			   rot = comp$xlab.rot,
							  			   x = grid::unit(comp$xlab.padding[2] * xlabS * o$lin, units = "inch"),
							  			   just = "left",
							  			   gp = grid::gpar(col = comp$xlab.color, cex = xlabS, fontface = comp$xlab.fontface, fontfamily = comp$xlab.fontfamily))
							  )
		} else {
			grXlab = NULL
		}

	} else {
		grXlab = NULL
		grYlab = NULL
	}


	if (comp$type == "bivariate") {
		textW = graphics::strwidth(comp$labels[1:m], units = "inch", cex = textS, family = comp$text.fontfamily, font = fontface2nr(comp$text.fontface))
		scale_labels = max(textW / grid::convertUnit(wsu[5], unitTo = "inch", valueOnly = TRUE), 1)
		grText1 = mapply(function(i, id) gridCell(id, 5, grid::textGrob(comp$labels[i], x = 0, just = "left", gp = grid::gpar(col = comp$text.color, cex = textS / scale_labels, fontface = comp$text.fontface, fontfamily = comp$text.fontfamily, alpha = comp$text.alpha))), m:1L, comp$item_ids[1L:m], SIMPLIFY = FALSE)

		colLabels = comp$labels[(m+1L):(n+m)]

		# easy solution: take first letter. Alternatives: rotation, ...
		colLabs = substr(colLabels, 1, 1)
		if (anyDuplicated(colLabs)) {
			colLabs = substr(colLabels, 1, 2)
			if (anyDuplicated(colLabs)) {
				cli::cli_inform("{.field bivariate legend} Labels abbreviated by the first two letters, e.g.: {.str {colLabels[1]}} => {.str {colLabs[1]}}. However, there are duplicated abbreviations.", .frequency_id = "biv_leg1",
								.frequency = "once")
			} else {
				cli::cli_inform("{.field bivariate legend} Labels abbreviated by the first two letters, e.g.: {.str {colLabels[1]}} => {.str {colLabs[1]}}.", .frequency_id = "biv_leg2",
								.frequency = "once")
			}
		} else {
			cli::cli_inform("{.field bivariate legend} Labels abbreviated by the first two letters, e.g.: {.str {colLabels[1]}} => {.str {colLabs[1]}}.", .frequency_id = "biv_leg2",
							.frequency = "once")
		}



		grText2 = mapply(function(i, j, id) gridCell(comp$item_ids[m+1], j, grid::textGrob(colLabs[i], x = 0.5, just = "center", gp = grid::gpar(col = comp$text.color, cex = textS / scale_labels, fontface = comp$text.fontface, fontfamily = comp$text.fontfamily, alpha = comp$text.alpha))), 1L:n, 7:(6+n), comp$item_ids[1L:n], SIMPLIFY = FALSE)

		grText = c(grText1, grText2)
	} else {
		textW = graphics::strwidth(comp$labels, units = "inch", cex = textS, family = comp$text.fontfamily, font = fontface2nr(comp$text.fontface))
		scale_labels = max(textW / grid::convertUnit(wsu[5], unitTo = "inch", valueOnly = TRUE), 1)

		grText = mapply(function(i, id) gridCell(id, 5, grid::textGrob(comp$labels[i], x = 0, just = "left", gp = grid::gpar(col = comp$text.color, cex = textS / scale_labels, fontface = comp$text.fontface, fontfamily = comp$text.fontfamily, alpha = comp$text.alpha))), (1L:nlev)[comp$labels_select], comp$item_ids[comp$labels_select], SIMPLIFY = FALSE)



	}

	ticks = get_legend_option(comp$ticks, comp$type)
	ticks.disable.na = get_legend_option(comp$ticks.disable.na, comp$type)
	if (length(ticks)) {
		tick_col = if (is.na(comp$ticks.col)) comp$text.color else comp$ticks.col
		if (is.na(tick_col)) tick_col = "white"

		ticks_in_margin = sapply(ticks, function(l) all(l>=1))

		ni = nlev - (comp$na.show && ticks.disable.na)



		tck_ids = (1L:ni)[rep(comp$labels_select, length.out = ni)]

		# remove tick ids that are too close to edges
		# only if color box frame color is there
		if (!(("col" %in% names(comp) && is.na(comp$col)) || split_alpha_channel(comp$gpar$col[1], alpha = 1)$opacity == 0)) {
			ih = get_legend_option(comp$item.height, "gradient")
			na_share = ((ih - 0.25)/2)/ih
			if (comp$gradient_ht_fracs[1] >= na_share) {
				tck_ids = setdiff(tck_ids, 1)
			}
			if (comp$gradient_ht_fracs[2] > na_share) {
				tck_ids = setdiff(tck_ids, comp$nitems - comp$na.show)
			}
		}




		# tick marks in margin (specified with x coordinates between 1 and 2)
		if (any(ticks_in_margin)) {
			grTicksMargin = do.call(c, lapply(ticks[ticks_in_margin], function(x) {
				mapply(function(i, id) gridCell(id, 4, grid::linesGrob(x = x - 1, y = c(0.5, 0.5), gp = grid::gpar(col = tick_col, lwd = comp$ticks.lwd * comp$scale * o$scale_down))), tck_ids, comp$item_ids[tck_ids], SIMPLIFY = FALSE)
			}))
		} else {
			grTicksMargin = NULL
		}

		if (!all(ticks_in_margin)) {
			grTicksItem = do.call(c, lapply(ticks[!ticks_in_margin], function(x) {
				mapply(function(i, id) gridCell(id, 3, grid::linesGrob(x = x, y = c(0.5, 0.5), gp = grid::gpar(col = tick_col, lwd = comp$ticks.lwd * comp$scale * o$scale_down))), tck_ids, comp$item_ids[tck_ids], SIMPLIFY = FALSE)
			}))
		} else {
			grTicksItem = NULL
		}

		grTicks = c(grTicksMargin, grTicksItem)
	} else {
		grTicks = NULL
	}



	if (getOption("tmap.design.mode")) {
		df = expand.grid(col = 1:length(comp$wsu),
						 row = 1:length(comp$hsu))

		grDesign = lapply(1:nrow(df), function(i) gridCell(df$row[i], df$col[i], grid::rectGrob(gp=gpar(fill=NA,col="#000000", lwd=1))))
	} else {
		grDesign = NULL
	}

	gp = comp$gp



	if (comp$type == "gradient") {
		# for borders
		gpars = gp_to_gpar(gp, id = 1L, sel = "col", o = o, type = comp$type)

		# for gradient fill
		nlev2 = (nlev-comp$na.show) # nlev without na
		lvs = 1:nlev2

		vary_fill = (length(gp$fill) > 1)
		vary_fill_alpha = (length(gp$fill_alpha) > 1)

		vary_col = (length(gp$col) > 1)
		vary_col_alpha = (length(gp$col_alpha) > 1)


		# vary fill color
		if (vary_fill) {
			fill_list = cont_split(gp$fill[lvs])
			fill_list = lapply(fill_list, function(i) {
				i[i=="NA"] <- NA
				i
			})
		} else if (vary_col) {
			fill_list = cont_split(gp$col[lvs])
			fill_list = lapply(fill_list, function(i) {
				i[i=="NA"] <- NA
				i
			})
			gpars$col = NA
		} else {
			fill_list = rep(gp$fill[1], nlev2)
		}

		# vary fill alpha
		if (vary_fill_alpha) {
			alpha_list = cont_split(gp$fill_alpha[lvs])
			alpha_list = lapply(alpha_list, function(i) {
				i[i=="NA"] <- 0
				as.numeric(i)
			})
			if (!vary_fill) {
				fill_list = mapply(rep.int, fill_list, vapply(alpha_list, length, FUN.VALUE = integer(1)), SIMPLIFY = FALSE, USE.NAMES = FALSE)
			}

		} else if (vary_col_alpha) {
			alpha_list = cont_split(gp$col_alpha[lvs])
			alpha_list = lapply(alpha_list, function(i) {
				i[i=="NA"] <- 0
				as.numeric(i)
			})

			if (!vary_col) {
				fill_list = mapply(rep.int, fill_list, vapply(alpha_list, length, FUN.VALUE = integer(1)), SIMPLIFY = FALSE, USE.NAMES = FALSE)
			}
		} else {
			if (!is.na(gp$fill_alpha[1])) {
				alpha_list = rep(gp$fill_alpha[1], nlev2)
			} else {
				alpha_list = rep(gp$col_alpha[1], nlev2)
			}
		}



		nvv = o$continuous.nclass_per_legend_break

		if (o$use_gradient) {
			id1 = which(!is.na(fill_list[[1]]))[1]
			id2 = tail(which(!is.na(fill_list[[nlev2]])), 1)

			y1 = 1 - ((id1-1) / nvv) / nlev2
			y2 = 1 - ((id2 / nvv) / nlev2 + ((nlev2-1)/nlev2))
			h = y1 - y2

			if (vary_fill) {
				cols = unlist(fill_list)[id1:(nvv*(nlev2-1) + id2)]
				cols_alph = paste0(cols, num_to_hex(gp$fill_alpha[1] * 255))
			} else {
				alph = unlist(alpha_list)[id1:(nvv*(nlev2-1) + id2)]
				cols_alph = paste0(col2hex(gp$fill[1]), num_to_hex(alph * 255))
			}
			grItems1 = list(gridCell(comp$item_ids[lvs], 3, grid::rectGrob(y = y2 + 0.5*h, height= h, gp=gpar(fill = grid::linearGradient(colours = rev(cols_alph)), col = NA))))
		} else {
			grItems1 = mapply(function(id, f, a) {
				h = 1 / length(f)
				ys = seq(1-.5*h, by = -h, length.out = length(f))
				#f[!is.na(f)] = "red"
				gpi = grid::gpar(fill = f, alpha = a, col = NA)
				gridCell(id, 3, grid::rectGrob(y = ys, height = h, gp = gpi))
			}, comp$item_ids[lvs], fill_list, alpha_list, SIMPLIFY = FALSE)
		}



		if (vary_fill) {
			y1 = (sum(is.na(fill_list[[1]])) /nvv) / nlev2
			y2 = (sum(is.na(fill_list[[nlev2]])) /nvv) / nlev2
		} else {
			y1 = (sum(is.na(alpha_list[[1]])) /nvv) / nlev2
			y2 = (sum(is.na(alpha_list[[nlev2]])) /nvv) / nlev2
		}

		grItems2 = list(gridCell(comp$item_ids[lvs], 3, rndrectGrob(y = grid::unit(y2, "npc"), just = c("center", "bottom"), height = grid::unit(1-(y1+y2), "npc"), gp = gpars, r = comp$item.r)))
		#grItems2 = list(gridCell(comp$item_ids[lvs], 3, rndrectGrob(y = grid::unit(y2, "npc"), just = c("center", "bottom"), height = grid::unit(1, "npc"), gp = gpars, r = comp$item.r)))

		if (comp$na.show) {
			gpars$fill = gp$fill[ifelse(length(gp$fill), nlev, 1)]
			gpars$fill_alpha = gp$fill[ifelse(length(gp$fill_alpha), nlev, 1)]
			# fill and border for NA
			grItems1 = c(grItems1, list(gridCell(comp$item_ids[nlev], 3, rndrectGrob(gp = gpars, r = comp$item.r))))
		}

		# borders
		grItems = c(grItems1, grItems2)


	} else if (comp$type == "rect") {
		#gps = split_gp(gp, n = nlev)

		diffAlpha = !anyNA(c(gp$fill_alpha, gp$col_alpha)) && !(length(gp$fill_alpha) == length(gp$col_alpha) && all(gp$fill_alpha == gp$col_alpha))


		if (diffAlpha) {
			gpars1 = gp_to_gpar(gp, sel = "fill", split_to_n = nlev, o = o, type = comp$type) #lapply(gps, gp_to_gpar_fill)
			gpars2 = gp_to_gpar(gp, sel = "col", split_to_n = nlev, o = o, type = comp$type) #lapply(gps, gp_to_gpar_borders)

			#grItems = mapply(function(i, gpari) gridCell(i+3, 2, rndrectGrob(gp = gpari)), 1:nlev, gpars, SIMPLIFY = FALSE)
			grItems = mapply(function(id, gpar1i, gpar2i) gridCell(id, 3, {
				grid::grobTree(
					rndrectGrob(width = grid::unit(o$lin* textS, "inches"), height = grid::unit(o$lin* textS, "inches"), gp = gpar1i, r = comp$item.r),
					rndrectGrob(width = grid::unit(o$lin* textS, "inches"), height = grid::unit(o$lin* textS, "inches"), gp = gpar2i, r = comp$item.r))
			}), comp$item_ids, gpars1, gpars2, SIMPLIFY = FALSE)

		} else {
			gpars = gp_to_gpar(gp, sel = "all", split_to_n = nlev, o = o, type = comp$type)#lapply(gps, gp_to_gpar)
			#grItems = mapply(function(i, gpari) gridCell(i+3, 2, grid::rectGrob(gp = gpari)), 1:nlev, gpars, SIMPLIFY = FALSE)
			grItems = mapply(function(id, gpari) gridCell(id, 3, rndrectGrob(gp = gpari, r = comp$item.r)), comp$item_ids, gpars, SIMPLIFY = FALSE)
		}
	} else if (comp$type == "lines") {
		#gps = split_gp(gp, n = nlev)

		gpars = gp_to_gpar(gp, sel = "col", split_to_n = nlev, o = o, type = comp$type)#lapply(gps, gp_to_gpar)
			#grItems = mapply(function(i, gpari) gridCell(i+3, 2, grid::rectGrob(gp = gpari)), 1:nlev, gpars, SIMPLIFY = FALSE)
		grItems = mapply(function(id, gpari) gridCell(id, 3, grid::linesGrob(y = grid::unit(c(0.5,0.5), "npc"), gp = gpari)), comp$item_ids, gpars, SIMPLIFY = FALSE)


	} else if (comp$type == "symbols") {
		if (length(gp$size) == 1) gp$size = min(gp$size, min(get_legend_option(comp$item.height, "symbols"),
															 get_legend_option(comp$item.width, "symbols")) * comp$textS)

		gp = swap_pch_15_20(gp)

		gpars = gp_to_gpar(gp, split_to_n = nlev, o = o, type = comp$type)

		diffAlpha = !anyNA(c(gp$fill_alpha, gp$col_alpha)) && !(length(gp$fill_alpha) == length(gp$col_alpha) && all(gp$fill_alpha == gp$col_alpha))
		if (diffAlpha) {
			gpars1 = gp_to_gpar(gp, split_to_n = nlev, o = o, type = comp$type, sel = "fill")
			gpars2 = gp_to_gpar(gp, split_to_n = nlev, o = o, type = comp$type, sel = "col")
		} else {
			gpars1 = vector(mode = "list", length = nlev)
			gpars2 = vector(mode = "list", length = nlev)
		}

		# scale down (due to facet use)
		gpars = lapply(gpars, rescale_gp, scale = o$scale_down)

		shapeLib = get("shapeLib", envir = .TMAP)
		justLib = get("justLib", envir = .TMAP)
#browser()
		grItems = mapply(function(id, gpari, gpari1, gpari2) {
			grobs = if (gpari$shape > 999) {
				grbs = gList(shapeLib[[gpari$shape-999]])
				grid::gTree(children=grbs, vp=viewport(x=0.5,
												 y=0.5,
												 width=unit(gpari$size*9/10 * comp$layer_args$icon.scale, "lines"),
												 height=unit(gpari$size*9/10 * comp$layer_args$icon.scale, "lines")))
			} else {
				if (diffAlpha) {
					grid::grobTree(
						grid::pointsGrob(x=0.5, y=0.5, pch = gpari1$shape, size = grid::unit(gpari1$size, "lines"), gp = gpari1),
						grid::pointsGrob(x=0.5, y=0.5, pch = gpari2$shape, size = grid::unit(gpari2$size, "lines"), gp = gpari2)
					)
				} else {
					grid::pointsGrob(x=0.5, y=0.5, pch = gpari$shape, size = grid::unit(gpari$size, "lines"), gp = gpari)
				}
			}
			#grb = gTree(children=do.call(gList, grobs), name=paste0("symbols_", id))
			gridCell(id, 3, grobs)
		}, comp$item_ids, gpars, gpars1, gpars2, SIMPLIFY = FALSE)
	} else if (comp$type == "text") {
		gp$text[is.na(gp$text)] = getAesOption("value.const", o, aes = "text", layer = "text")

		if (length(gp$cex) == 1) gp$cex = min(gp$cex, min(get_legend_option(comp$item.height, "text"),
															 get_legend_option(comp$item.width, "text")) * comp$textS)

		bgcols = rep(gp$bgcol, length.out = nlev)
		bgcols_alpha = rep(gp$bgcol_alpha, length.out = nlev)

		# in case size is a continuous scale
		if (is.character(gp$cex)) gp$cex = vapply(cont_split(gp$cex), FUN = function(x) {
			as.numeric(x[round(length(x)/2)])
		}, FUN.VALUE = numeric(1))


		gpars = gp_to_gpar(gp, split_to_n = nlev, o = o, type = comp$type)

		# scale down (due to facet use)
		gpars = lapply(gpars, rescale_gp, scale = o$scale_down)



		grItems = mapply(function(id, gpari, txt, bgcol, bgcol_alpha) {
			gridCell(id, 3, {
				grid::gList(
					rndrectGrob(width = grid::unit(grid::convertWidth(grid::stringWidth(txt), "inches")* gpari$cex, "inches"), height = grid::unit(o$lin* textS, "inches"), gp = grid::gpar(fill = bgcol, alpha = bgcol_alpha, col = NA), r = comp$item.r),
					grid::textGrob(x=0.5, y=0.5, label = txt, gp = gpari))
			})
		}, comp$item_ids, gpars, gp$text, bgcols, bgcols_alpha, SIMPLIFY = FALSE)

	} else if (comp$type == "bivariate") {
		gpars = gp_to_gpar(gp, sel = "all", split_to_n = n*m, o = o, type = comp$type)#lapply(gps, gp_to_gpar)
		#grItems = mapply(function(i, gpari) gridCell(i+3, 2, grid::rectGrob(gp = gpari)), 1:nlev, gpars, SIMPLIFY = FALSE)

		ind = expand.grid(row = comp$item_ids[m:1], col = 7:(6+n))

		grItems = mapply(function(i, j, gpari) gridCell(i, j, rndrectGrob(gp = gpari, r = comp$item.r)), ind$row, ind$col, gpars, SIMPLIFY = FALSE)
	}


	g = do.call(grid::grobTree, c(list(grobBG, grTitle, grXlab, grYlab), grText, grItems, grTicks, grDesign, list(vp = vp)))

	g

}

message_init = function() {
	.TMAP$messages = character()
	invisible(NULL)
}

message_thrown = function(id) {
	id %in% .TMAP$messages
}

message_reg = function(id) {
	.TMAP$messages = c(.TMAP$messages, id)
	NULL
}

message_comp_scale = function() {
	if (!message_thrown("comp_scale")) {
		cli::cli_inform(c(
			"{.field [plot mode]} fit legend/component: Some legend items or map compoments do not fit well, and are therefore rescaled.",
			i = "Set the tmap option {.code component.autoscale =  FALSE} to disable rescaling."
			),
			.frequency_id = "comp_scale"
		)
		message_reg("comp_scale")
	}
	NULL
}

message_comp_high_wide = function(stack) {
	if (!message_thrown("comp_scale")) {
		cli::cli_inform(c(
			"{.field [plot mode]} legend/component: Some components or legends are too {.val {ifelse(stack == 'vertical', 'high', 'wide')}} and are therefore rescaled.",
			"i" = "Set the tmap option {.code component.autoscale = FALSE} to disable rescaling."
			),
			.frequency_id = "comp_scale"
		)
		message_reg("comp_scale")
	}
	NULL
}

message_c4a = function(old_palette_name, info, fullname = FALSE) {
	new1 = info$fullname
	new2 = info$name
	mess = 	paste0("c4a_", old_palette_name)

	if (!message_thrown(mess)) {
		if (fullname) {
			cli::cli_inform(
				"{.field [cols4all]} color palettes: use palettes from the R package cols4all. Run {.run cols4all::c4a_gui()} to explore them. The old palette name {.str {old_palette_name}} is named {.val {new1}}",
				.frequency_id = "cols4all"
			)
		} else {
			cli::cli_inform(
				"{.field [cols4all]} color palettes: use palettes from the R package cols4all. Run {.run cols4all::c4a_gui()} to explore them. The old palette name {.str {old_palette_name}} is named {.str {new2}} (in long format {.str {new1}})",
				.frequency_id = "cols4all"
			)

		}
		message_reg(mess)
	}
}

message_nothing_to_show = function(any_groups) {
	if (any_groups) {
		cli::cli_inform("{.field [nothing to show]} no data layers defined after {.fn tm_shape}",
						.frequency_id = "nothing")
	} else {
		cli::cli_inform("{.field [nothing to show]} no data layers defined",
						.frequency_id = "nothing")
	}
	NULL
}

message_wrapstack = function(horizontal = TRUE) {
	if (horizontal) {
		cli::cli_inform("{.field [facets]} use {.fn tm_facets_hstack} instead of {.fn tm_facets_wrap} to put the legends next to and aligned with the facets.")
	} else {
		cli::cli_inform("{.field [facets]} use {.fn tm_facets_vstack} instead of {.fn tm_facets_wrap} to put the legends next to and aligned with the facets.")
	}
	NULL
}

message_pos_auto = function(type) {
	if (!message_thrown("pos_auto")) {
		fun = if (type == "autoout") "tm_pos_auto_out" else "tm_pos_auto_in"
		fun2 = if (type == "autoout") "tm_pos_out" else "tm_pos_in"
		cli::cli_inform("{.field [position]} use {.fn {fun2}} instead of {.fn {fun}}. The latter should be used with {.fn tmap_options}.")
		message_reg("pos_auto")
	}
	NULL
}

error_dimvalues = function() {
	cli::cli_abort(
		"{.val dimvalues} has been used, but the shape object does not contain any dimensions; please use {.arg x} to specify variables",
		call = call("tm_vars")
	)
}

message_webgl_vars = function(supported, vary) {
	sup_text = paste(supported, collapse = ", ")
	var_text = paste(names(vary)[vary], collapse = ", ")
	var_sel = names(vary)[vary]

	cli::cli_inform("{.field [view mode]} WebGL enabled, but the only supported visual variables are: {.val {supported}}. The visual variable(s) {.val {var_sel}} are not supported. Set {.code use_WebGL = FALSE} to support them.")

}

message_webgl_hover = function(type) {
	cli::cli_inform("{.field [view mode]} WegGL enabled, but it does not support hover labels for layer type {.val {type}}. Set {.code use_WebGL = FALSE} to support them.",
					.frequency_id = "webGL_hover",
					.frequency = "once")
}


message_webgl_checks = function(checks, checkif) {
	vals = paste(paste(names(checkif)[!checks], checkif[!checks], sep = " = "), collapse = ", ")
	cli::cli_inform("{.field [view mode]} WebGL enabled, but the following visual variable only accept one value {.arg {vals}}. Set {.code use_WebGL = FALSE} to support them.",
					.frequency_id = "webGL_vv",
					.frequency = "once")
}

message_webgl_crs_simple = function() {
	cli::cli_inform("{.field [view mode]} WebGL does not work (yet) with projected map projections, so it has been disabled.",
					.frequency_id = "webGL_crs",
					.frequency = "once")

}


message_qtm_empty = function() {
	cli::cli_inform(
		"{.field [qtm()]} nothing to show. Either specify {.arg shp} or {.arg basemap}. Alternatively, switch to view mode with {.code tmap_mode(\"view\")}"
	)
}

message_layer_unused_args = function(layer_fun, args) {
	cli::cli_inform(
		"{.field [{layer_fun}()]} Argument{ifelse(length(args)>1, 's', '')} {.arg {args}} unknown."
	)
}

message_crs_property_unknown = function() {
	cli::cli_inform(
		"{.field [shiny]} some css styling (used for background and markers) cannot be used yet in shiny apps",
		.frequency_id = "shiny_css",
		.frequency = "once")
}

message_crs_property_not_used = function() {
	cli::cli_inform(
		"{.field [tm_crs()]} {.arg crs} specified, so {.arg property} is ignored"
	)
}

message_crs_ll = function() {
	cli::cli_inform(
		"{.field [tip]} Consider a suitable map projection, e.g. by adding {.code + tm_crs({.str auto})}.",
		.frequency_id = "crs",
		.frequency = "once"
	)
}

message_scale_interval_value0 = function(aes, values, layer) {
	suggestion = if (inherits(values, "tmapSeq")) {
		to = values$to
		power = values$power
		if (is.character(power)) power = paste0("\"", power, "\"")
		paste0("values = tm_seq(0.25, ", to, ", power = ", power, ")")
	} else {
		to = values[2]
		paste0("values = c(0.25, ", to, ")")
	}


	cli::cli_inform(
		"{.field [layer {layer}, tm_scale_intervals()]} By default the value of the visual variable {.code {aes}} for the first interval is 0. Consider {.code {suggestion}} (or another lower bound number than 0.25). Or alternatively, use {.fun tm_scale_continuous}.",
		.frequency_id = "size0"
	)
}

message_shiny_css = function() {
	cli::cli_inform(
		"{.field [layer {layer}, tm_scale_intervals()]} By default the value of the visual variable {.code {aes}} for the first interval is 0. Consider {.code {suggestion}} (or another lower bound number than 0.25). Or alternatively, use {.fun tm_scale_continuous}.",
		.frequency_id = "size0"
	)
}

message_basemaps_none = function(serv, z) {
	cli::cli_inform(
		"{.field [basemaps]} Tiles from {.str {serv}} at zoom level {z} couldn't be loaded",
		.frequency_id = "basemap_none",
		.frequency = "once"
	)

}

message_basemaps_blurry = function(serv) {
	cli::cli_inform(
		"{.field [basemaps]} Tiles from {.str {serv}} will be projected so details (e.g. text) could appear blurry",
		.frequency_id = "basemap_blurry",
		.frequency = "once"
	)
}


message_basemaps = function(is_stadia) {
	provider = ifelse(is_stadia, "Stadia", "Thunderforest")
	cli::cli_inform(
		"{.field [basemaps]} An API key is requierd for tiles from {.str {provider}}",
		.frequency_id = "basemap_API",
		.frequency = "once"
	)
}

#' Scales: bivariate scale
#'
#' Scales in tmap are configured by the family of functions with prefix `tm_scale`.
#' Such function should be used for the input of the `.scale` arguments in the
#' layer functions (e.g. `fill.scale` in [tm_polygons()]).
#' The function `tm_scale_bivariate()` is used for `bivariate.scales`.
#'
#' @param scale1,scale2 two `tm_scale` objects. Currently, all `tm_scale_*()` functions are supported except `tm_scale_continuous()`.
#' @param values (generic scale argument) The visual values. For colors
#'   (e.g. `fill` or `col` for `tm_polygons()`) this is a palette name from the `cols4all` package (see [cols4all::c4a()])
#'   or vector of colors, for size (e.g. `size` for `tm_symbols()`) these are a set of sizes (if two values are specified they are interpret as range),
#'   for symbol shapes (e.g. `shape` for [tm_symbols()]) these are a set of symbols, etc.
#'   The tmap option `values.var` contains the default values per visual variable and in some cases also per data type.
#' @param values.repeat (generic scale argument) Should the values be repeated in case there are more categories?
#' @param values.range (generic scale argument) Range of the values. Vector of two numbers (both between 0 and 1) where the first determines the minimum and the second the maximum.
#'   Full range, which means that all values are used, is encoded as `c(0, 1)`.
#'   For instance, when a grey scale is used for color (from black to white), `c(0,1)`
#'   means that all colors are used, `0.25, 0.75` means that only colors from dark grey to light
#'   grey are used (more precisely `"grey25"` to `"grey75"`), and `0, 0.5` means
#'   that only colors are used from black to middle grey (`"grey50"`).
#'   When only one number is specified, this is interpreted as the second number
#'   (where the first is set to 0). Default values can be set via the tmap option `values.range`.
#' @param values.scale (generic scale argument) Scaling of the values. Only useful for size-related visual variables, such as `size` of [tm_symbols()] and `lwd` of [tm_lines()].
#' @param value.na (generic scale argument) Value used for missing values. See tmap option `"value.na"` for defaults per visual variable.
#' @param value.null (generic scale argument) Value used for NULL values.
#'   See tmap option `"value.null"` for defaults per visual variable. Null data
#'   values occur when out-of-scope features are shown (e.g. for a map of Europe showing a data variable per country,
#'   the null values are applied to countries outside Europe).
#' @param value.neutral (generic scale argument) Value that can be considered neutral.
#'   This is used for legends of other visual variables of the same map layer.
#'   E.g. when both `fill` and `size` are used for [tm_symbols()] (using filled circles),
#'   the size legend items are filled with the `value.neutral` color from the `fill.scale` scale,
#'   and fill legend items are bubbles of size `value.neutral` from the `size.scale` scale.
#' @param labels (generic scale argument) Labels
#' @param label.na (generic scale argument) Label for missing values
#' @param label.null (generic scale argument) Label for null (out-of-scope) values
#' @seealso `r .doc_see_also_scales()` `r .doc_see_also_legend(biv = TRUE)` [tm_scale()]
#' @export
tm_scale_bivariate = function(scale1 = tm_scale(),
							  scale2 = tm_scale(),
							  values = NA,
							  values.repeat = FALSE,
							  values.range  = NA,
							  values.scale = 1,
							  value.na = NA,
							  value.null = NA,
							  value.neutral = NA,
							  labels = NULL,
							  label.na = NA,
							  label.null = NA) {
	structure(c(list(FUN = "tmapScaleBivariate"), as.list(environment())), class = c("tm_scale_bivariate", "tm_scale", "list"))
}

#' @export
tmapReproject.stars = function(shp, tmapID, bbox = NULL, ..., crs) {
	shp[[1]] = tmapID

	args = list(...)
	raster.warp = if ("raster.warp" %in% names(args)) args$raster.warp else TRUE

	shp2 = transwarp(shp, crs, raster.warp = raster.warp)
	tmapID2 = shp2[[1]]
	shp2[[1]] = NA

	if (!is.null(bbox$x)) {
		if (identical(bbox$x, "FULL")) {
			bbox = list(x = full_bbox(crs))
		} else {
			bbox = list(x = do.call(tmaptools::bb, c(bbox, list(projection = crs))))
		}
	}

	shapeTM(shp2, tmapID2, bbox, ...)
}

#' @export
tmapReproject.dimensions = function(shp, tmapID, bbox = NULL, ..., crs) {
	s = structure(list(tmapID = matrix(tmapID, nrow = nrow(shp))), dimensions = shp, class = "stars")

	args = list(...)
	raster.warp = if ("raster.warp" %in% names(args)) args$raster.warp else TRUE

	if (is.na(sf::st_crs(shp))) {
		shp2 = s
	} else {
		shp2 = transwarp(s, crs, raster.warp = raster.warp)
	}

	tmapID2 = shp2[[1]]

	d2 = stars::st_dimensions(shp2)

	if (!is.null(bbox$x)) {
		if (identical(bbox$x, "FULL")) {
			bbox = list(x = full_bbox(crs))
		} else {
			bbox = list(x = do.call(tmaptools::bb, c(bbox, list(projection = crs))))
		}
	}

	shapeTM(d2, tmapID2, bbox, ...)
}


#' @export
tmapShape.stars = function(shp, is.main, crs, bbox, unit, filter, layer, shp_name, smeta, o, tmf) {
	if (identical(crs, "auto")) crs = auto_crs(shp, crs_extra = o$crs_extra, crs_global = o$crs_global)

	dev = getOption("tmap.devel.mode")

	if (dev) cat("-- stars object:", shp_name, "--\n")


	if (!has_raster(shp)) {
		dimnms = dimnames(shp)

		dimvals = lapply(1:length(dimnms), function(i) stars::st_get_dimension_values(shp, i))
		dimsfc = vapply(dimvals, inherits, what = "sfc", FUN.VALUE = logical(1))

		if (!any(dimsfc)) {
			stop("stars object ", shp_name, " is a stars object without raster and doens't have a geometry dimension")
		} else {
			dimid = which(dimsfc)
			geoms = dimvals[[dimid]]
			dimnms_new = dimnms
			dimnms_new[dimid] = "tmapID__"
			dimcols = dimnms_new[-dimid] # columns names, used for default facetting
			dimvls = lapply(dimcols, function(d) stars::st_get_dimension_values(shp, d))
			shpnames = names(shp)
			shp = stars::st_set_dimensions(shp, dimnms[dimid], values = seq_along(geoms))
			shp = stars::st_set_dimensions(shp, names = dimnms_new)
		}

		dt = as.data.table(shp)

		if (!is.null(crs) && sf::st_crs(geoms) != crs) {
			shp = sf::st_transform(shp, crs)
		} else {
			shp = geoms
		}
		shpTM = shapeTM(shp = shp, tmapID = seq_along(shp), bbox = bbox)


		shpclass = "sfc"
	} else {
		shp = downsample_stars(shp, max.raster = max_cells(o$raster.max_cells) / (o$fn[1] * o$fn[2]))
		if (!is.null(crs) && sf::st_crs(shp) != crs) {
			shp = transwarp(shp, crs, raster.warp = o$raster.warp)
		}

		dims = stars::st_dimensions(shp)
		rst = attr(dims, "raster")

		dim_xy = get_xy_dim(shp)
		dimsxy = dims[names(dim_xy)]

		if (rst$curvilinear) {
			shp3 = shp
			attr(shp3, "dimensions")[[rst$dimensions[1]]]$values = 1L:nrow(shp)
			attr(shp3, "dimensions")[[rst$dimensions[2]]]$values = 1L:ncol(shp)
			attr(attr(shp3, "dimensions"), "raster")$curvilinear = FALSE
		} else {
			shp2 = stars::st_set_dimensions(shp, rst$dimensions[1], values = 1L:nrow(shp))
			shp3 = stars::st_set_dimensions(shp2, rst$dimensions[2], values = 1L:ncol(shp))
		}


		dt = as.data.table(shp3, center = FALSE)
		names(dt)[1:2] = c("X__", "Y__")

		if (!all(names(shp3) %in% names(dt))) {
			subst_names = tail(names(dt), length(shp3))
			for (i in 1L:length(shp3)) {
				setnames(dt, subst_names[i], names(shp3)[i])
			}
		}

		dt[, tmapID__ := as.integer((Y__-1) * nrow(shp) + X__)]
		dt[, X__:= NULL]
		dt[, Y__:= NULL]

		data.table::setorder(dt, cols = "tmapID__")

		shp = dimsxy
		shpclass = "stars"


		shpTM = shapeTM(shp = shp, tmapID = 1L:(nrow(shp) * ncol(shp)), bbox = bbox)

	}

	make_by_vars(dt, tmf, smeta)


	dtcols = setdiff(names(dt), "tmapID__")

	filter = if (is.null(filter)) {
		rep(TRUE, nrow(dt))
	} else filter[dt$tmapID__]
	dt[, ':='(sel__ = filter)] # tmapID__ = 1L:nrow(dt),

	for (nm in dtcols) {
		if (is.factor(dt[[nm]])) {
			lvls = levels(dt[[nm]])
			cols = attr(dt[[nm]], "colors")
			ids = seq_along(lvls)[lvls!="NA"]
			if (!is.null(cols)) {
				#if (any(lvls=="NA")) {
				dt[[nm]] = factor(as.integer(dt[[nm]]), levels = ids, labels = paste(lvls[ids], cols[ids], sep = "=<>="))
				#}
				#dt[[nm]] = factor()
			}
		}
	}

	structure(list(shpTM = shpTM, dt = dt, is.main = is.main, dtcols = dtcols, shpclass = shpclass, bbox = bbox, unit = unit, shp_name = shp_name, smeta = smeta, type_ids = NULL, type_vars = NULL), class = "tmapShape")
}


#' @export
tmapSubsetShp.stars = function(shp, vars) {
	ids = which(names(shp) %in% vars)
	nom = setdiff(vars, names(shp))
	if (length(nom)) cli::cli_abort("dimension values {.arg {nom}} not found in shape. Options are {.arg {names(shp)}}")
	shp2 = shp[ids]
	if (!length(vars)) {
		shp2$dummy__ = TRUE
	}
	shp2
}


#' @export
tmapGetShapeMeta1.stars = function(shp, layer, o) {
	d = stars::st_dimensions(shp)

	if (!has_raster(shp)) {
		d_non_xy = local({
			dimvals = lapply(seq_along(d), function(i) stars::st_get_dimension_values(shp, i))
			dimsfc = vapply(dimvals, inherits, what = "sfc", FUN.VALUE = logical(1))
			d[!dimsfc]
		})
	} else {
		d_non_xy = local({
			dxy = attr(d, "raster")$dimensions
			d[setdiff(names(d), dxy)]
		})
	}

	dims = names(d_non_xy)
	dims_vals = lapply(dims, function(d) stars::st_get_dimension_values(shp, d))
	names(dims_vals) = dims

	vars_orig = names(shp)
	vars = vars_orig #make.names(vars_orig)
	names(vars) = vars_orig
	list(vars = vars,
		 dims = dims,
		 dims_vals = dims_vals)
}

#' @export
tmapGetShapeMeta2.stars = function(shp, smeta, o) {
	smeta$vars_levs = lapply(seq_len(length(shp)), function(i) {
		get_fact_levels_na(shp[[i]], o)
	})
	names(smeta$vars_levs) = names(shp)
	smeta
}

submit_symbols = function(x, grid, args) {
	if (any(vapply(x, is.null, logical(1)))) stop("one of more shapes are NULL")
	shapeLib = get("shapeLib", envir = .TMAP)
	justLib = get("justLib", envir = .TMAP)
	n = length(x)
	id = 999 + length(shapeLib)
	if (grid) {
		# symbols as grobs
		just_items = lapply(x, function(xs) {
			if (args$just.override) {
				args$just
			} else if ("iconUrl" %in% names(xs)) {
				if (all(c("iconWidth", "iconHeight", "iconAnchorX", "iconAnchorY") %in% names(xs))) {
					c(1-(xs$iconAnchorX / xs$iconWidth), xs$iconAnchorY / xs$iconHeight)
				} else NA
			} else NA
		})

		items = lapply(x, function(xs) {
			if ("iconUrl" %in% names(xs)) {
				grb = icon2grob(xs)
				# take first one
				if (is.grob(grb)) grb else grb[[1]]
			} else if (is.grob(xs)) {
				xs
			} else NA
		})
	} else {
		# symbols as images
		items = lapply(x, function(xs) {
			ic = if ("iconUrl" %in% names(xs)) {
				split_icon(xs)[[1]]
			} else if (is.grob(xs)) {
				grob2icon(xs, args$grob.dim, args$just)
			} else NA

			# add anchor based on just specs
			if (all(c("iconWidth", "iconHeight") %in% names(ic)) &&
				((!any(c("iconAnchorX", "iconAnchorY") %in% names(ic))) || args$just.override)) {
				ic$iconAnchorX = ic$iconWidth * (1-args$just[1])
				ic$iconAnchorY = ic$iconHeight * args$just[2]
			}
			ic
		})
		just_items = as.list(rep(NA, n))
	}

	numbers = is.na(items)

	if (all(numbers)) return(unlist(x, use.names = FALSE))

	new_id = id + 1:sum(!numbers)

	x2 = integer(n)
	x2[numbers] = unlist(x[numbers], use.names = FALSE)
	x2[!numbers] = new_id

	shapeLib = c(shapeLib, items[!numbers])
	justLib = c(justLib, just_items[!numbers])
	assign("shapeLib", shapeLib, envir = .TMAP)
	assign("justLib", justLib, envir = .TMAP)
	names(x2) = names(x)
	x2
}

step2_data = function(tm) {
	dev = getOption("tmap.devel.mode")

	tmo = tm$tmo
	o = tm$o
	aux = tm$aux
	cmp = tm$cmp
	prx = tm$prx

	if (is.null(tmo)) {
		return(	list(tmo = NULL, aux = aux, cmp = cmp, prx = prx, o = o))
	}

	groupnames = paste0("group", seq_along(tmo))
	fl = list(1L, 1L, 1L)
	assign("fl", fl, envir = .TMAP)

	# to reset the legends (which are  temporarily stored in .TMAP environment)
	legends_init()
	charts_init()

	grps = mapply(function(tmg, groupname) {
		tmf = tmg$tmf
		dt = tmg$tms$dt

		type_ids = tmg$tms$type_ids
		type_vars = tmg$tms$type_vars # used to split dt vars in the layer functions, e.g. for networks with edges and nodes

		if ("by1__" %in% names(dt) && o$rev1) dt[, by1__ := (o$fn[1]+1L)-by1__]
		if ("by2__" %in% names(dt) && o$rev2) dt[, by2__ := (o$fn[2]+1L)-by2__]
		if ("by3__" %in% names(dt) && o$rev3) dt[, by3__ := (o$fn[3]+1L)-by3__]

		shpvars = tmg$tms$smeta$vars

		# step2_data_grp_prepare
		#tmf_meta = step2_data_grp_prepare(tmg$tmf, grpvars, dt)

		layernames = paste0("layer", seq_along(tmg$tmls))
		lrs = mapply(function(tml, layername) {
			glid = paste0(groupname, "_", layername)

			#cat("step2_grp_lyr======================\n")

			if (!is.null(type_ids)) {
				ids = tml$layer %in% names(type_ids)
				if (any(ids)) {
					lid = tml$layer[[which(ids)[1]]]
					tmapids = type_ids[[lid]]
					dt = data.table::copy(dt[dt$tmapID__ %in% tmapids, c(type_vars[[lid]], "sel__"), with = FALSE])
				}

			}

			gp = tml$gpar
			tp = tml$tpar

			plot.order = tml$plot.order

			group = if (is.na(tml$group)) tmg$tms$shp_name else as.character(tml$group)
			group.control = as.character(tml$group.control)

			.TMAP$popup.format = tml$popup.format

			mfun = tml$mapping.fun

			# args will be passed on to the scale functions (in case needed)
			# they also will be used in step 3 (trans) and step 4 (mapping)
			if (is.null(dt)) {
				# bypass (e.g. in case of pmtiles)
				trans = list()
				mapping = mapply(getdts, tml$mapping.aes, names(tml$mapping.aes), SIMPLIFY = FALSE, MoreArgs = list(p = gp, q = tmf, o = o, dt = dt, shpvars = shpvars, layer = tml$layer, group = group, glid = glid, mfun = mfun, args = tml$mapping.args, plot.order = plot.order))
				#mapping = list()
			} else {
				trans = mapply(getdts, tml$trans.aes, names(tml$trans.aes), SIMPLIFY = FALSE, MoreArgs = list(p = tp, q = tmf, o = o, dt = dt, shpvars = shpvars, layer = tml$layer, group = group, glid = glid,mfun = mfun, args = tml$trans.args, plot.order = plot.order))

				mapping = mapply(getdts, tml$mapping.aes, names(tml$mapping.aes), SIMPLIFY = FALSE, MoreArgs = list(p = gp, q = tmf, o = o, dt = dt, shpvars = shpvars, layer = tml$layer, group = group, glid = glid,mfun = mfun, args = tml$mapping.args, plot.order = plot.order))
			}


			dts_trans = cbind_dts(lapply(trans, function(x) x$dt), plot.order)
			trans_legend = lapply(trans, function(x) x$leg)

			dts_mapping = cbind_dts(lapply(mapping, function(x) x$dt), plot.order)
			mapping_legend = lapply(mapping, function(x) x$leg)
			if (dev) timing_add(s4 = "combine")
			if (dev) timing_add(s3 = paste0("layer ", tml$layer))


			if (!length(tml$popup.vars) || is.null(dt)) {
				popup.data = NULL
			} else {
				# tml$popup.vars = tml$popup.vars
				# 	intersect(tml$popup.vars, names(dt)) # in case type_ids !NULL

				popupvars = intersect(tml$popup.vars, names(dt))

				if (!length(popupvars)) {
					popup.data = NULL
				} else {
					if (!is.null(names(tml$popup.vars))) {
						popupvars = structure(popupvars, names = names(tml$popup.vars)[match(popupvars, tml$popup.vars)])
					}

					popup.data = copy(dt[, c(popupvars, "tmapID__"), with = FALSE])
					if (!is.null(names(popupvars))) popup.data = data.table::setnames(popup.data, old = unname(popupvars), new = names(popupvars))
				}
			}
			hover.data = if (tml$hover == "" || is.null(dt)) {
				NULL
			} else {
				data.table(hover = as.character(dt[[tml$hover]]), tmapID__ = dt$tmapID__)
			}
			id.data = if (tml$id == "" || is.null(dt)) {
				NULL
			} else {
				data.table(id = as.character(dt[[tml$id]]), tmapID__ = dt$tmapID__)
			}

			# popup.title: a data variable used as the popup header. Overrules
			# `id` for the popup title (analogous to how `hover` overrules `id`
			# for hover labels). `tml$popup.title` is NULL for layers that do
			# not set it (e.g. extension layers), or NA when not specified.
			popup.title = tml$popup.title
			popup.title.data = if (is.null(popup.title) || (length(popup.title) == 1L && is.na(popup.title)) || is.null(dt)) {
				NULL
			} else if (!popup.title %in% names(dt)) {
				NULL
			} else {
				ptt = data.table(title = as.character(dt[[popup.title]]), tmapID__ = dt$tmapID__)
				if (!is.null(names(popup.title)) && names(popup.title) != "") {
					attr(ptt, "label") = names(popup.title)
				}
				ptt
			}

			format_called = attr(tml$popup.format, "called")
			if (length(format_called) > 0L) {
				.TMAP$popup.format[format_called] = tml$popup.format[format_called]
			}


			list(trans_dt = dts_trans,
				 trans_legend = trans_legend,
				 trans_fun = tml$trans.fun,
				 trans_args = tml$trans.args,
				 trans_apply = tml$trans.apply_to,
				 mapping_dt = dts_mapping,
				 mapping_legend = mapping_legend,
				 mapping_fun = tml$mapping.fun,
				 mapping_args = tml$mapping.args,
				 lid = tml$lid,
				 group = group,
				 glid = glid,
				 group.control = group.control,
				 popup.data = popup.data,
				 popup.format = .TMAP$popup.format,
				 popup.title.data = popup.title.data,
				 popup.layout = tml$popup.layout,
				 hover.data = hover.data,
				 id.data = id.data,
				 plot.order = plot.order, # passed on for step 3 non-data driven transformation
				 gp = gp,
				 tp = tp)
		}, tmg$tmls, layernames, SIMPLIFY = FALSE, USE.NAMES = FALSE)
		if (length(lrs)) names(lrs) = layernames

		shpDT = data.table(shpTM = list(tmg$tms$shpTM))
		if (dev) timing_add(s2 = "group")

		list(layers = lrs, shpDT = shpDT)
	}, tmo, groupnames, SIMPLIFY = FALSE, USE.NAMES = FALSE)
	names(grps) = groupnames
	#attr(grps, "fl") = fl
	#attr(grps, "main") = attr(tmo, "main")
	#attr(grps, "crs") = attr(tmo, "crs")

	#tmf = get_tmf(lapply(tmo, function(tmoi) tmoi$tmf))



	# facet labels "_old' were obtained in step2, and still determine data levels
	# currently, facet labels are determined in step1, but they need to be the same as the data labels
	# o$fl_old = get("fl", envir = .TMAP)
	# o$fn_old = sapply(o$fl_old, function(f) {
	# 	if (is.character(f)) length(f) else f
	# })
	# o$fl_old = lapply(o$fl_old, function(f) {
	# 	if (is.character(f)) f else NULL
	# })
	#
	# if (!identical(o$fn, o$fn_old) || !identical(o$fl, o$fl_old)) {
	# 	po(o$fn_old, o$fl_old, o$fn, o$fl)
	# 	stop("fl not identical")
	# }



	#cat("fl old:\n")
	#print(o$fl_old)


	#o = c(o, tmf)

	# attr(grps, "is.wrap") = tmo[[1]]$tmf$is.wrap
	# attr(grps, "nrows") = tmo[[1]]$tmf$nrows
	# attr(grps, "ncols") = tmo[[1]]$tmf$ncols

	o = within(o, {
		if (rev1) fl[[1]][] = rev(fl[[1]][])
		if (rev2) fl[[2]][] = rev(fl[[2]][])
		if (rev3) fl[[3]][] = rev(fl[[3]][])
	})

	list(tmo = grps, aux = aux, cmp = cmp, prx = prx, o = o)
}

# Helper: convert per-feature metre radii to native CRS units, for use with
# grid::circleGrob(r = unit(r_native, "native")).
#
# For geographic CRS the approximation is 111 320 m/degree.
# For projected CRS (including Web Mercator) we derive the conversion by
# displacing a reference point 1 degree north in geographic space and measuring
# how many native units that corresponds to. This correctly accounts for scale
# distortion (e.g. Mercator stretching at higher latitudes) without assuming
# any specific projection.
.circles_m_to_native = function(radius_m, shp) {
	if (isTRUE(sf::st_is_longlat(shp))) {
		return(radius_m / 111320)
	}

	# Back-project the centroid to geographic, displace 1 degree northward,
	# re-project both points to the display CRS, and compute the dy in native units.
	crs_disp = sf::st_crs(shp)
	ctr = sf::st_centroid(sf::st_union(shp))
	ctr_geo = sf::st_transform(ctr, 4326)

	ctr_geo_n = sf::st_sfc(
		sf::st_point(sf::st_coordinates(ctr_geo) + c(0, 1)),
		crs = 4326
	)

	dy_native = abs(
		sf::st_coordinates(sf::st_transform(ctr_geo_n, crs_disp))[, 2] -
		sf::st_coordinates(sf::st_transform(ctr_geo,   crs_disp))[, 2]
	)

	# dy_native = native units per degree; 111 320 m ≈ 1 degree latitude
	m_to_native = dy_native / 111320
	radius_m * m_to_native
}


#' @export
#' @rdname tmapGridLeaflet
tmapGridDataPlot.tm_data_circles = function(a, shpTM, dt, gp, bbx,
											facet_row, facet_col, facet_page,
											id, pane, group, glid, o, ...) {
	rc_text = frc(facet_row, facet_col)

	# Filter to rows where size is known (no shape filter: circles have none)
	res = select_sf(shpTM, dt[!is.na(dt$size), ])
	shp = res$shp
	dt  = res$dt

	gp = impute_gp(gp, dt)
	# NOTE: do NOT call rescale_gp — that converts size to "lines" units.
	# For tm_circles, gp$size is already in metres (the direct scale output).

	coords = sf::st_coordinates(shp)
	cp     = expand_coords_gp(coords, gp, ndt = nrow(dt))
	coords = cp$coords
	gp     = cp$gp

	# gp$size is in metres; convert to native CRS units for grid, correcting for
	# any projection scale distortion (e.g. Web Mercator at higher latitudes).
	radius_native = .circles_m_to_native(gp$size, shp)

	gp_par = gp_to_gpar(gp, sel = "all", o = o, type = "symbols")

	grobs = lapply(seq_along(radius_native), function(i) {
		gpi = structure(lapply(gp_par, function(x) if (length(x) == 1L) x else x[i]),
						class = "gpar")
		grid::circleGrob(
			x  = grid::unit(coords[i, 1], "native"),
			y  = grid::unit(coords[i, 2], "native"),
			r  = grid::unit(radius_native[i], "native"),
			gp = gpi
		)
	})

	grb = grid::gTree(children = do.call(grid::gList, grobs),
					  name     = paste0("circles_", id))

	gts = get("gts", .TMAP_GRID)
	gt  = gts[[facet_page]]
	gt  = grid::addGrob(gt, grb, gPath = grid::gPath(paste0("gt_facet_", rc_text)))
	gts[[facet_page]] = gt
	assign("gts", gts, envir = .TMAP_GRID)
	NULL
}

leaflet_pos = function(pos, collapse = "") {
	if (pos$type %in% c("out", "autoout")) {
		sel = c("cell.v", "cell.h")
	} else {
		sel = c("pos.v", "pos.h")
	}
	x = tolower(unlist(pos[sel]))

	if (x[1] %in% c("center", "centre")) x[1] = "top"
	if (x[2] %in% c("center", "centre")) x[2] = "left"

  paste(x, collapse = collapse)
}


gp_to_lpar = function(gp, mfun, shape = 20, pick_middle = TRUE, size_factor = 20, rename_prop = TRUE) {
	# create a list of gp elements

	lst = c(list(fill = {if (!all(is.na(gp$fill))) gp$fill else "#000000"},
				 col = {if (!all(is.na(gp$col))) gp$col else "#000000"},
				 fill_alpha = {if (!all(is.na(gp$fill_alpha))) gp$fill_alpha else 0},
				 col_alpha = {if (!all(is.na(gp$col_alpha))) gp$col_alpha else 0},
				 lwd = {if (!all(is.na(gp$lwd))) gp$lwd else 0},
				 lty = {if (!all(is.na(gp$lty))) lty2dash(gp$lty, gp$lwd) else "none"},
				 size = {if (!all(is.na(gp$size))) gp$size else 1},
				 shape = {if (!all(is.na(gp$shape))) gp$shape else shape}))

	lst_isnum = c(fill = FALSE,
				  col = FALSE,
				  fill_alpha = TRUE,
				  col_alpha = TRUE,
				  lwd = TRUE,
				  lty = FALSE,
				  size = TRUE,
				  shape = TRUE)

	if (rename_prop) {
		newNames = c("fillColor",
					  "color",
					  "fillOpacity",
					  "opacity",
					  "stroke-width",
					  "stroke-dash",
					  "size",
					  "shape")
		names(lst) = newNames
		names(lst_isnum) = newNames
	}


	lst = mapply(function(lsti, nm, isnum) {
		if (!is.character(lsti)) return(lsti)

		if (nchar(lsti[1]) > 50) {
			x = cont_split(lsti)
			x = lapply(x, function(i) {
				i[i=="NA"] <- NA
				i
			})
			if (isnum) x = lapply(x, as.numeric)
			if (nm %in% c("fillColor", "color")) {
				# create ramp of 10
				not1 = which(vapply(x, length, FUN.VALUE = integer(1)) != 1L)
				if (length(not1) != length(x)) {
					# add NA
					xNA = x[[length(x)]]
				} else {
					xNA = NULL
				}
				x[not1[-1]] = lapply(x[not1[-1]], tail, -1)
				x = na.omit(unlist(x[not1]))
				x = c(x[seq(1, length(x), length.out = length(not1) * 2 + 1)], xNA)
			} else if (pick_middle) {
				x = sapply(x, function(i) {
					if (all(is.na(i))) NA else {
						sq = c(5,6,4,7,3,8,2,9,1,10) # priority for middle values
						i[sq[which(!is.na(i)[sq])[1]]]
					}
				})
			}
			return(x)

		} else {
			return(lsti)
		}
	}, lst, names(lst), lst_isnum[names(lst)], SIMPLIFY = FALSE)

	pch2shp = c("rect", "circle", "triangle", "plus", "cross", "diamond", "triangle",
				"cross", "star", "diamond", "circle", "polygon", "plus", "cross",
				"triangle", "rect", "circle", "triangle", "diamond", "circle",
				"circle", "circle", "rect", "diamond", "triangle", "polygon", "stadium") # shapes for pch 0:25 + 26 for stadium (NOTE: last one is a triangle upside-down. Since 21:25 are the defaults, and a polygon is chosen to differentiate from the other triangle)
	lst$shape = get_pch_names(lst$shape)

	if ("tm_data_lines" %in% mfun) lst$shape = "line"

	lst$width = lst$size * size_factor
	lst$height = lst$size * size_factor
	#lst$width[]
	lst$size = NULL
	lst
}



make_equal_list = function(x) {
	cls = class(x)
	n = max(vapply(x, length, integer(1)))
	structure(lapply(x, rep, length.out = n), class = cls)
}


tmapLeaflet_legend = function(cmp, lf, o, orientation) {
	group = cmp$group
	leg_className = paste("info legend", gsub(" ", "", group, fixed = TRUE))



	if ("layerId" %in% names(cmp)) {
		layerId = cmp$layerId
		group = "always_on"
	} else {
		layerId =  paste0("legend", sprintf("%02d", .TMAP_LEAFLET$leg_id)) # "legend401" #todo
		.TMAP_LEAFLET$leg_id = .TMAP_LEAFLET$leg_id + 1
	}



	lab = cmp$labels
	val = c(cmp$dvalues)
	title = if (nonempty_text(cmp$title)) expr_to_char(cmp$title) else NULL

	if (!is.null(title) && !is.null(cmp$title.color)) {

		if (cmp$type == "gradient") {
			title = htmltools::tags$div(
				title,
				style = paste0('color: ', cmp$title.color, ';')
			)
		} else {
			title = htmltools::tags$div(
				title,
				style = paste0('font-size: 14px; text-align: left; margin-bottom: 5px; color: ', cmp$title.color, ';')
			)
		}

	}

	legpos = leaflet_pos(cmp$position)

	lf2 = if (cmp$type == "none") {
		#message("Text based legends not supported in view mode")
		lf
	} else if (cmp$type == "gradient") {
		breaks_show =  ("is_discrete" %in% names(cmp)) && cmp$is_discrete


		nbins = length(val)

		incl.na = cmp$na.show

		if (incl.na) {
			sel = head(cmp$labels_select, -1)
		} else {
			sel = cmp$labels_select
		}

		bins = val[sel]
		val = val[sel]

		if (!breaks_show) {
			if (cmp$scale == "intervals") cli::cli_warn("{.field tm_scale_intervals} {.code label.style = {.str continuous}} implementation in view mode work in progress")
			if (!head(sel, 1)) {
				val = c(cmp$limits[1], val)
			} else {
				val = c(val[1] - (val[2] - val[1]) * 0.5, val)
			}

			if (!tail(sel, 1)) {
				val = c(val, cmp$limits[2])
			} else {
				val = c(val, val[length(val)] + diff(tail(val, 2))/2)
			}
		} else {
			brks = val
		}

		vary = if ("fill" %in% cmp$varying) "fillColor" else "color"
		#vary_alpha = paste0(vary, "_alpha")

		incl.na = cmp$na.show
		if (incl.na) {
			pal = head(cmp$gp2[[vary]], -1)
			colNA = tail(cmp$gp2[[vary]], 1)
			textNA = lab[length(lab)]
			labs = head(lab, -1)[sel]
			val = c(val, NA)
			#labs = c(labs, "")
		} else {
			pal = cmp$gp2[[vary]]
			colNA = NA
			textNA = NA
			labs = lab[sel]
		}

		pfun = function(x) {
			r = brks[c(1, length(brks))]
			d = (x * (r[2] - r[1])) + r[1]
			ids = cut(d, breaks = brks, include.lowest = TRUE, right = TRUE, labels = FALSE)
			pal2[ids]
		}
		# pfun = function(x) {
		# 	ifelse(x < 0.5, "blue", "red")
		# }

		if (breaks_show) {
			# when labels_cutpoints is enabled in tm_scale_intervals
			pal2 = unname(head(pal, -1)[-1][seq(2L, length(pal)-2L, by = 2L)])
			pal = colorNumeric(palette = pfun,
							   domain = val,
							   na.color=colNA,
							   alpha = FALSE)
		} else {
			pal = colorNumeric(palette = pal,
							   domain = val,
							   na.color=colNA,
							   alpha = FALSE)
		}


		brks = pretty(cmp$limits, 7)

		trns = function(x) {
			f = (x-cmp$limits[1]) / diff(cmp$limits)
			y = cmp$tr$fun(cmp$limits)
			x2 = y[1] + diff(y) * f
			cmp$tr$rev(x2)
		}

		opacity = if (vary == "fillColor") {
			cmp$gp2$fillOpacity
		} else {
			cmp$gp2$opacity
		}

		if (orientation == "horizontal") {
			cli::cli_inform("{.field [landscape legend in view mode]} doesn't support labels yet",
							.frequency_id = "landscape_legend_view",
							.frequency = "once")
			labs = NULL

		}

		lf %>% leaflegend::addLegendNumeric(position=legpos,
										   orientation = orientation,
										   group = group,
										   height = cmp$height,
										   width = cmp$width,
										   pal=pal,
										   values=val,
										   # numberFormat = function(x) {
										   # 	prettyNum(trns(x), format = "f", big.mark = ",", digits =
										   # 			  	3, scientific = FALSE)
										   # },
										   labels = labs,
										   bins = bins,
										   naLabel = textNA,
										   title=title,
										   fillOpacity=opacity,
										   layerId = layerId,
										   className = leg_className)

	} else {
		vary = if ("fill" %in% cmp$varying) "fill" else if ("col" %in% cmp$varying) "col" else NA



		gp2 = make_equal_list(cmp$gp2)
		#gp2$baseSize = gp2$baseSize[1]

		sn = suppressWarnings(as.numeric(gp2$shape))
		sid = which(!is.na(sn))

		#symbols = do.call(Map, c(list(f = leaflegend::makeSymbol), cmp$gp2))

		# alternative:
		if (length(sid)) {
			gp2$shape[sid] = "circle" # as dummy
		}

		names(gp2)[names(gp2) == 'stroke-width'] = "strokeWidth"
		names(gp2)[names(gp2) == 'stroke-dash'] = "dashArray"
		gp2$baseSize = 20
		#symbols = do.call(leaflegend::makeSymbolIcons, gp2)#$iconUrl


		symbols = do.call(makeSymbolIcons2, gp2)
		#po(sort(gp2$width, decreasing = T))

		symbols$iconWidth = gp2$width#, length(symbols$iconUrl))
		symbols$iconHeight = gp2$height#, length(symbols$iconUrl))

		if (length(sid)) {
			iconLib <- get("shapeLib", envir = .TMAP)[sn[sid]-999]
			symbols_icons <- merge_icons(iconLib)

			size = gp2$width[sid] / gp2$baseSize * cmp$layer_args$icon.scale/3

			for (i in seq_along(sid)) {
				symbols$iconUrl[sid[i]] = symbols_icons$iconUrl[i]
				symbols$iconWidth[sid[i]] <- symbols_icons$iconWidth[i] * size[i]
				symbols$iconHeight[sid[i]] <- symbols_icons$iconHeight[i] * size[i]
				if (all(c("iconAnchorX", "iconAnchorY") %in% names(symbols_icons))) {
					symbols$iconAnchorX[sid[i]] <- symbols_icons$iconAnchorX[i] * size[i]
					symbols$iconAnchorY[sid[i]] <- symbols_icons$iconAnchorY[i] * size[i]

				}
			}
		}

		#symbols = symbols$iconUrl


		lf %>% leaflegend::addLegendImage(symbols$iconUrl,
										  group = group,
										 labels = lab,
										 width = symbols$iconWidth + max(2*gp2$strokeWidth),
										 height = symbols$iconHeight + 2*gp2$strokeWidth,
										 position = legpos,
										 orientation = orientation,
										 labelStyle = "font-size: 14px; vertical-align: middle; margin: 0px;",
										 title = title,
										 layerId = layerId,
										 className = leg_className)
	}
	backg <- htmltools::tags$style(paste0("#", layerId, " { background: ", substr(cmp$bg.color,1,7), "; opacity: ", cmp$bg.alpha, "}"))
	if (!.TMAP$in.shiny) {
		htmlwidgets::prependContent(lf2, backg)
	} else {
		lf2
	}
}

#' @export
tmapLeafletCompPlot.tm_legend_portrait = function(comp, lf, o) {
	tmapLeaflet_legend(comp, lf, o, orientation = "vertical")
}

#' @export
tmapLeafletCompPlot.tm_legend_landscape = function(comp, lf, o) {
	tmapLeaflet_legend(comp, lf, o, orientation = "horizontal")
}


tmapLeafletComp = function(comp, o, facet_row = NULL, facet_col = NULL, facet_page, class, stack, stack_auto, pos.h, pos.v, bbox) {
	lf = get_lf(facet_row, facet_col, facet_page)

	rc_text = frc(facet_row, facet_col)

	# grp = tmapGetCompGroupArgs(comp) none of the tm_components args are used yet in view mode

	for (cmp in comp) {
		lf = tmapLeafletCompPlot(cmp, lf, o)
	}



	assign_lf(lf, facet_row, facet_col, facet_page)
	NULL
}

#' @export
#' @rdname tmap_internal
tmapValuesCheck_col = function(x, is_var = TRUE) {
	isnum = is.numeric(x)
	if (isnum) {
		structure(FALSE,
				  info = {if (is_var) "Variable should be data variable name or color name" else " Values should be numeric (between -50 and 50)."}
		)
	} else {
		is_c4a = !is.null(getPalMeta(x[1])) && length(x) == 1L && !valid_colors(x[1])
		if (is_c4a) {
			if (is_var) {
				structure(FALSE,
						  info = "Variable should be a data variable name or a single color (not a color palette).")
			} else {
				TRUE
			}
		} else {
			all_cols = all(valid_colors(x))
			if (!all_cols) {
				structure(FALSE,
						  info = if (is_var) "Variable should a data variable name or a single color." else "Values should be color names or a color palette (run {.run cols4all::c4a_palettes()} for available ones.")
			} else {
				TRUE
			}
		}
	}
}

#' @export
#' @rdname tmap_internal
tmapValuesCheck_fill = function(x, is_var = TRUE) {
	tmapValuesCheck_col(x, is_var)
}

#' @export
#' @rdname tmap_internal
tmapValuesCheck_bgcol = function(x, is_var = TRUE) {
	tmapValuesCheck_col(x, is_var)
}

isSymbol = function(s) {
	inherits(s, "grob") || any(vapply(s, inherits, FUN.VALUE = logical(1), "grob")) || ("iconUrl" %in% names(s))
}
isUrl <- function(string) {
	any(grepl("(https?|ftp)://[^\\s/$.?#].[^\\s]*", string))
}

isLocal = function(string) {
	file.exists(string)
}

#' @export
#' @rdname tmap_internal
tmapValuesCheck_shape = function(x, is_var = TRUE) {

	if (inherits(x, "tmap_icons")) {
		TRUE
	} else if (all(is.numeric(x))) {
		TRUE
	} else if (is.list(x)) {
	 	if (isSymbol(x)) {
	 		TRUE
	 	} else {
	 		all(vapply(x, isSymbol, FUN.VALUE = logical(1)) | vapply(x, is.numeric, FUN.VALUE = logical(1)))
	 	}
	} else if (all(is.character(x))) {
		if (all(isUrl(x) | isLocal(x))) {
			TRUE
		} else {
			structure(FALSE, info = {if (is_var) " Variable should be a data variable name or a symbol (see {.help [tm_symbols](tmap::tm_symbols)} - details section)." else "  Values should be symbols (see {.help [tm_symbols](tmap::tm_symbols)} - details section)."})
		}
	}
}



#' @export
#' @rdname tmap_internal
tmapValuesCheck_size = function(x, is_var = TRUE) {
	res = inherits(x, "tmapSeq") || (is.numeric(x) && (all(x>=0) || all(x<=0)))

	if (!res) {
		structure(FALSE, info = {if (is_var) " Variable should be a data variable name or a numeric value." else "  Values should be numeric."})
	} else {
		TRUE
	}
}


#' @export
#' @rdname tmap_internal
tmapValuesCheck_area = function(x, is_var = TRUE) {
	tmapValuesCheck_size(x, is_var)
}

#' @export
#' @rdname tmap_internal
tmapValuesCheck_lwd = function(x, is_var = TRUE) {
	tmapValuesCheck_size(x, is_var)
}

#' @export
#' @rdname tmap_internal
tmapValuesCheck_lty = function(x, is_var = TRUE) {
	if (all(x %in% c("blank", "solid", "dashed", "dotted", "dotdash", "longdash", "twodash")) ||
		(is.numeric(x) && all(x %in% 0:6)) ||
		(all(grepl("^[1-9A-F]{1,8}$", x)))) {
		TRUE
	} else {
		structure(FALSE, info = {if (is_var) " Variable should be a data variable name or a line type (see documentation of graphics::par - lty)." else "  Values should be a line type (see documentation of graphics::par - lty). numeric."})
	}
}

#' @export
#' @rdname tmap_internal
tmapValuesCheck_xmod = tmapValuesCheck_ymod = function(x, is_var = TRUE) {
	if (!is.numeric(x)) {
		structure(FALSE,
				  info = {if (is_var) " It is neither a data variable name nor a numeric (between -50 and 50)." else " Values should be numeric (between -50 and 50)."}
				  )
	} else {
		res = all(x >= -50 & x <= 50)
		if (!res) {
			structure(FALSE,
					  info = {if (is_var) " Variable should be in the [-50,50] range." else " Values found that are outside the [-50,50] range. Note that the default scale for xmod and ymod is tm_scale_asis."})
		} else {
			TRUE
		}
	}
}

#' @export
#' @rdname tmap_internal
tmapValuesCheck_angle = function(x, is_var = TRUE) {
	# to do
	if (!is.numeric(x)) {
		structure(FALSE, info = {if (is_var) " Variable should be a data variable name or a numeric value." else "  Values should be numeric."})
	} else {
		TRUE
	}
}

#' @export
#' @rdname tmap_internal
tmapValuesCheck_col_alpha= function(x, is_var = TRUE) {
	tmapValuesCheck_size(x, is_var)
}

#' @export
#' @rdname tmap_internal
tmapValuesCheck_fill_alpha = function(x, is_var = TRUE) {
	tmapValuesCheck_size(x, is_var)
}

#' @export
#' @rdname tmap_internal
tmapValuesCheck_bgcol_alpha= function(x, is_var = TRUE) {
	tmapValuesCheck_size(x, is_var)
}

#' @export
#' @rdname tmap_internal
tmapValuesCheck_area = function(x, is_var = TRUE) {
	tmapValuesCheck_size(x, is_var)
}

#' @export
#' @rdname tmap_internal
tmapValuesCheck_text = function(x, is_var = TRUE) {
	TRUE
}

#' @export
#' @rdname tmap_internal
tmapValuesCheck_fontface = function(x, is_var = TRUE) {

	if 	((is.numeric(x) && (all(x %in% 1:5))) || (is.character(x) && (all(x %in% c("plain", "bold", "italic", "oblique", "bold.italic", "cyrillic", "cyrillic.oblique", "EUC"))))) {
		TRUE
	} else {
		structure(FALSE, info = {if (is_var) " Variable should be a data variable name or a font face (see documentation of graphics::par font)." else "  Values should be a font face (see documentation of graphics::par - font). numeric."})
	}
}

#' @export
#' @rdname tmap_internal
tmapValuesIsDiv_fill = function(x) {
	m = getPalMeta(x[1])
	ispal = !is.null(m)

	if (ispal) {
		m$type == "div"
	} else {
		(palette_type(x) == "div")
	}
}

#' @export
#' @rdname tmap_internal
tmapValuesIsDiv_col = function(x) {
	tmapValuesIsDiv_fill(x)
}

#' @export
#' @rdname tmap_internal
tmapValuesIsDiv_bgcol = function(x) {
	tmapValuesIsDiv_fill(x)
}

#' @export
#' @rdname tmap_internal
tmapValuesIsDiv_size = function(x) {
	inherits(x, "tmapSeq") && (x$from < 0) && (x$to > 1) || (is.numeric(x) && (any(x < 0) && any(x> 0)))
}

#' @export
#' @rdname tmap_internal
tmapValuesIsDiv_area = function(x) {
	tmapValuesIsDiv_size(x)
}

#' @export
#' @rdname tmap_internal
tmapValuesIsDiv_lwd = function(x) {
	tmapValuesIsDiv_size(x)
}

#' @export
#' @rdname tmap_internal
tmapValuesIsDiv_lty = function(x) {
	FALSE
}

#' @export
#' @rdname tmap_internal
tmapValuesIsDiv_col_alpha = function(x) {
	tmapValuesIsDiv_size(x)
}

#' @export
#' @rdname tmap_internal
tmapValuesIsDiv_fill_alpha = function(x) {
	tmapValuesIsDiv_size(x)
}

#' @export
#' @rdname tmap_internal
tmapValuesIsDiv_bgcol_alpha = function(x) {
	tmapValuesIsDiv_size(x)
}

#' @export
#' @rdname tmap_internal
tmapValuesIsDiv_area = function(x) {
	tmapValuesIsDiv_size(x)
}

#' @export
#' @rdname tmap_internal
tmapValuesIsDiv_xmod = tmapValuesIsDiv_ymod = function(x) {
	tmapValuesIsDiv_size(x)
}

#' @export
#' @rdname tmap_internal
tmapValuesIsDiv_angle = function(x) {
	FALSE
}

#' @export
#' @rdname tmap_internal
tmapValuesIsDiv_shape = function(x) {
	FALSE
}

#' @export
#' @rdname tmap_internal
tmapValuesIsDiv_text = function(x) {
	FALSE
}

#' @export
#' @rdname tmap_internal
tmapValuesIsDiv_fontface = function(x) {
	FALSE
}


#' @export
#' @rdname tmap_internal
tmapValuesRange_fill = function(x, n, isdiv) {
	m = getPalMeta(x[1])

	if (!is.null(m)) {
		NA # in c4a palette definition
	} else c(0, 1)
}

#' @export
#' @rdname tmap_internal
tmapValuesRange_col = function(x, n, isdiv) {
	tmapValuesRange_fill(x, n, isdiv)
}

#' @export
#' @rdname tmap_internal
tmapValuesRange_bgcol = function(x, n, isdiv) {
	tmapValuesRange_fill(x, n, isdiv)
}

#' @export
#' @rdname tmap_internal
tmapValuesRange_shape = function(x, n, isdiv) {
	c(0, 1)
}

#' @export
#' @rdname tmap_internal
tmapValuesRange_lty = function(x, n, isdiv) {
	c(0, 1)
}


#' @export
#' @rdname tmap_internal
tmapValuesRange_size = function(x, n, isdiv) {
	#print(c(.5/n, 1 - .5/n))
#	c(.5/n, 1 - .5/n)
	c(0, 1)
}

#' @export
#' @rdname tmap_internal
tmapValuesRange_area = function(x, n, isdiv) {
	tmapValuesRange_size(x, n, isdiv)
}

#' @export
#' @rdname tmap_internal
tmapValuesRange_lwd = function(x, n, isdiv) {
	tmapValuesRange_size(x, n, isdiv)
}

#' @export
#' @rdname tmap_internal
tmapValuesRange_col_alpha = function(x, n, isdiv) {
	tmapValuesRange_size(x, n, isdiv)
}

#' @export
#' @rdname tmap_internal
tmapValuesRange_fill_alpha = function(x, n, isdiv) {
	tmapValuesRange_size(x, n, isdiv)
}

#' @export
#' @rdname tmap_internal
tmapValuesRange_bgcol_alpha = function(x, n, isdiv) {
	tmapValuesRange_size(x, n, isdiv)
}

#' @export
#' @rdname tmap_internal
tmapValuesRange_area = function(x, n, isdiv) {
	tmapValuesRange_size(x, n, isdiv)
}

#' @export
#' @rdname tmap_internal
tmapValuesRange_xmod = tmapValuesRange_ymod = function(x, n, isdiv) {
	tmapValuesRange_size(x, n, isdiv)
}

#' @export
#' @rdname tmap_internal
tmapValuesRange_angle = function(x, n, isdiv) {
	c(0, 1)
}

#' @export
#' @rdname tmap_internal
tmapValuesRange_text = function(x, n, isdiv) {
	c(0, 1)
}

#' @export
#' @rdname tmap_internal
tmapValuesRange_fontface = function(x, n, isdiv) {
	c(0, 1)
}

#' @export
#' @rdname tmap_internal
tmapValuesVV_fill = function(x, value.na, isdiv, n, dvalues, are_breaks, midpoint, range, scale, rep, o, aes = "fill") {

	#palid = tmapPalId(x[1])
	if (x[1] %in% c("seq", "div", "unord", "ord", "biv")) {
		# cols4all can also take "div", but does not take into account tmap style
		x = getAesOption("values.var", o, aes = aes, layer = NA, cls = x[1])
	}

	m = if (length(x) > 1) NULL else getPalMeta(x[1])


	scale_ids = function(ids, n) {
		1 + ((ids - 1) / (n - 1)) * 100
	}

	map_ids = function(i, s, n) {
		di = i[2] - i[1]
		seq(i[1] + di * s[1], i[1] + di * s[2], length.out = n)
	}


	if (isdiv) {
		if (are_breaks) {
			# scale_interval
			outside = (all(midpoint < dvalues) || all(midpoint > dvalues))
			cat0 = !any(dvalues==midpoint) && !outside

			nneg = max(0L, sum(dvalues < midpoint) - cat0 - outside) # max 0L needed when midpoint is outside range (and cat0 is true)
			npos = max(0L, sum(dvalues > midpoint) - cat0 - outside)

			nmax = max(nneg, npos)

			ntot = 2L * nmax + cat0

			ids = (1L + max(0L, (npos-nneg))):(ntot - max(0L, (nneg-npos)))
		} else {
			# scale_discrete
			outside = (all(midpoint < dvalues) || all(midpoint > dvalues))
			cat0 = any(dvalues==midpoint)
			between = !outside && !cat0

			if (between) {
				# suppose 4 break values, and m is between 1st and 2nd:
				# x m x    x    x
				# create palette of half steps, and take odd numbers at the end
				nneg = max(0L, (sum(dvalues < midpoint) * 2 - 1L))
				npos = max(0L, (sum(dvalues > midpoint) * 2 - 1L))
				nmax = max(nneg, npos)
				ntot = 2L * nmax + 1L
				ids = (1L + max(0L, (npos-nneg))):(ntot - max(0L, (nneg-npos)))
				ids = ids[seq(1L, length(ids), by = 2L)]
			} else {
				nneg = max(0L, sum(dvalues < midpoint))
				npos = max(0L, sum(dvalues > midpoint))
				nmax = max(nneg, npos)
				ntot = 2L * nmax + cat0
				ids = (1L + max(0L, (npos-nneg))):(ntot - max(0L, (nneg-npos)))
			}

		}
	} else {
		ntot = n
		ids = 1L:n
	}

	if (!is.null(m)) {
		if (x[1] != tolower(x[1])) message_c4a(x[1], info = m)
		vvalues = getPal(m$fullname, n = ntot, range = range, reversed = m$reverse)[ids]
	} else {
		pal =colorRampPalette(x)
		if (is.na(range[1])) range = c(0, 1)
		if (range[1] != 0 || range[2] != 1) {
			ids_scaled = scale_ids(ids, ntot)
			if (isdiv) {
				ids_after_range = c({if (nneg > 0) head(map_ids(ids_scaled[c(1L, (nneg+cat0))],
																1-rev(range),
																n = nneg + cat0),
														nneg) else NULL},
									{if (cat0) ids_scaled[1L + nneg] else NULL},
									if (npos > 0) tail(map_ids(ids_scaled[c(nneg+1, n)],
															   range,
															   n = npos + cat0),
													   npos) else NULL)
			} else {
				ids_after_range = map_ids(ids_scaled[c(1L, ntot)], range, ntot)
			}
			vvalues = grDevices::colorRampPalette(x)(o$continuous.nclasses)[ids_after_range]
		} else {
			vvalues = grDevices::colorRampPalette(x)(ntot)[ids]
		}
	}


	if (isdiv) {
		if (cat0) {
			value.neutral = vvalues[1L + nneg]
		} else {
			value.neutral = vvalues[1L + nneg] # first positive
		}
	} else {
		value.neutral = vvalues[n/2]
	}
	vvalues = do.call(process_color, c(list(col = vvalues), o$pc))
	value.neutral = do.call(process_color, c(list(col = value.neutral), o$pc))
	value.na = do.call(process_color, c(list(col = value.na), o$pc))
	list(vvalues = vvalues, value.neutral = value.neutral, value.na = value.na)

}

#' @export
#' @rdname tmap_internal
tmapValuesVV_col = function(...) {
	tmapValuesVV_fill(..., aes = "col")
}

#' @export
#' @rdname tmap_internal
tmapValuesVV_bgcol = function(...) {
	tmapValuesVV_fill(..., aes = "bgcol")
}

#' @export
#' @rdname tmap_internal
tmapValuesVV_shape = function(x, value.na, isdiv, n, dvalues, are_breaks, midpoint, range, scale, rep, o) {
	if (inherits(x, "tmap_icons") && "iconUrl" %in% names(x)) x = split_icon(x)
	list(vvalues = rep(x, length.out = n), value.neutral = x[1], value.na = value.na)
}

#' @export
#' @rdname tmap_internal
tmapValuesVV_lty = function(x, value.na, isdiv, n, dvalues, are_breaks, midpoint, range, scale, rep, o) {
	list(vvalues = rep(x, length.out = n), value.neutral = x[1], value.na = value.na)
}


#' @export
#' @rdname tmap_internal
tmapValuesVV_size = function(x, value.na, isdiv, n, dvalues, are_breaks, midpoint, range, scale, rep, o) {
	vvalues = if (is.numeric(x) && length(x) == n) {
		if (range[1] !=0 || range[2] != 1) {
			warning("values.range not used because the individual values have been specified (instead of a sequence)", call. = FALSE)
		}
		x
	} else {
		if (is.numeric(x)) {
			x = tm_seq(x[1], x[length(x)], power = "lin")
		}

		if (range[1] !=0 || range[2] != 1) {
			p = if (is.numeric(x$power)) x$power else switch(x$power, lin = 1, sqrt = 0.5, sqrt_perceptual = 0.5716, quadratic = 2)
			x$from = range[1] ^ (1/p)
			x$to = range[2] ^ (1/p)
		}
		tmapSeq(x, n)

	}
	value.neutral = vvalues[round((n+1)/2)]
	list(vvalues = vvalues * scale, value.neutral = value.neutral * scale, value.na = value.na * scale)
}



#' @export
#' @rdname tmap_internal
tmapValuesVV_area = function(...) {
	do.call(tmapValuesVV_size, args = list(...))
}

#' @export
#' @rdname tmap_internal
tmapValuesVV_lwd = function(...) {
	do.call(tmapValuesVV_size, args = list(...))
}

#' @export
#' @rdname tmap_internal
tmapValuesVV_col_alpha = function(...) {
	do.call(tmapValuesVV_size, args = list(...))
}

#' @export
#' @rdname tmap_internal
tmapValuesVV_fill_alpha = function(...) {
	do.call(tmapValuesVV_size, args = list(...))
}

#' @export
#' @rdname tmap_internal
tmapValuesVV_bgcol_alpha = function(...) {
	do.call(tmapValuesVV_size, args = list(...))
}

#' @export
#' @rdname tmap_internal
tmapValuesVV_area = function(...) {
	do.call(tmapValuesVV_size, args = list(...))
}

#' @export
#' @rdname tmap_internal
tmapValuesVV_xmod = tmapValuesVV_ymod = function(...) {
	do.call(tmapValuesVV_size, args = list(...))
}


#' @export
#' @rdname tmap_internal
tmapValuesVV_angle = function(...) {
	do.call(tmapValuesVV_size, args = list(...))
}

#' @export
#' @rdname tmap_internal
tmapValuesVV_text = function(x, value.na, isdiv, n, dvalues, are_breaks, midpoint, range, scale, rep, o) {
	list(vvalues = rep(x, length.out = n), value.neutral = x[1], value.na = value.na)
}

#' @export
#' @rdname tmap_internal
tmapValuesVV_fontface = function(x, value.na, isdiv, n, dvalues, are_breaks, midpoint, range, scale, rep, o) {
	list(vvalues = rep(x, length.out = n), value.neutral = x[1], value.na = value.na)
}

# for symbols, which need to be 'submitted' (content replaced by an integer)
#' @export
#' @rdname tmap_internal
tmapValuesSubmit_col = function(x, args) x
#' @export
#' @rdname tmap_internal
tmapValuesSubmit_fill = function(x, args) x
#' @export
#' @rdname tmap_internal
tmapValuesSubmit_bgcol = function(x, args) x
#' @export
#' @rdname tmap_internal
tmapValuesSubmit_size = function(x, args) x
#' @export
#' @rdname tmap_internal
tmapValuesSubmit_area = function(x, args) x
#' @export
#' @rdname tmap_internal
tmapValuesSubmit_xmod = function(x, args) x
#' @export
#' @rdname tmap_internal
tmapValuesSubmit_ymod = function(x, args) x
#' @export
#' @rdname tmap_internal
tmapValuesSubmit_angle = function(x, args) x
#' @export
#' @rdname tmap_internal
tmapValuesSubmit_lwd = function(x, args) x
#' @export
#' @rdname tmap_internal
tmapValuesSubmit_lty = function(x, args) x
#' @export
#' @rdname tmap_internal
tmapValuesSubmit_shape = function(x, args) {
	if (all(is.na(x))) return(x)
	if (!inherits(x, c("tmapStandard", "tmapSpecial"))) x = tmapVV(x)
	if (inherits(x, "tmapStandard")) {
		xvec = unlist(x, use.names = FALSE)
		if (is.character(xvec) && all(isUrl(xvec) | isLocal(xvec))) {
			x = tmapVV(tmap_icons(xvec, merge = FALSE))
		}
	}

	if (inherits(x, "tmapSpecial")) {
		isGrid = (tmap_graphics_name() == "Grid")


		# symbols just specification:
		# copy-pasted from v3, but not the best place
		# improvement of just needed (-> trans?)
		args = within(args, {
			if (anyNA(just)) {
				just = c(.5, .5)
				just.override = FALSE
			} else {
				just = c(ifelse(is_num_string(just[1]), as.numeric(just[1]), ifelse(just[1]=="left", 1, ifelse(just[1]=="right", 0, .5))),
						  ifelse(is_num_string(just[2]), as.numeric(just[2]), ifelse(just[2]=="bottom", 1, ifelse(just[2]=="top", 0, .5))))
				just.override = TRUE
			}
		})

		do.call(submit_symbols, args = list(x = x, grid = isGrid, args = args))
	} else {
		unlist(x)
	}
}

#' @export
#' @rdname tmap_internal
tmapValuesSubmit_col_alpha = function(x, args) x
#' @export
#' @rdname tmap_internal
tmapValuesSubmit_fill_alpha = function(x, args) x
#' @export
#' @rdname tmap_internal
tmapValuesSubmit_bgcol_alpha = function(x, args) x
#' @export
#' @rdname tmap_internal
tmapValuesSubmit_text = function(x, args) x
#' @export
#' @rdname tmap_internal
tmapValuesSubmit_fontface = function(x, args) x


#' @export
#' @rdname tmap_internal
tmapValuesScale_col = function(x, scale) x
#' @export
#' @rdname tmap_internal
tmapValuesScale_fill = function(x, scale) x
#' @export
#' @rdname tmap_internal
tmapValuesScale_bgcol = function(x, scale) x
#' @export
#' @rdname tmap_internal
tmapValuesScale_size = function(x, scale) x * scale
#' @export
#' @rdname tmap_internal
tmapValuesScale_area = function(x, scale) x
#' @export
#' @rdname tmap_internal
tmapValuesScale_lwd = function(x, scale) x * scale
#' @export
#' @rdname tmap_internal
tmapValuesScale_lty = function(x, scale) x
#' @export
#' @rdname tmap_internal
tmapValuesScale_shape = function(x, scale) x
#' @export
#' @rdname tmap_internal
tmapValuesScale_col_alpha = function(x, scale) x
#' @export
#' @rdname tmap_internal
tmapValuesScale_fill_alpha = function(x, scale) x
#' @export
#' @rdname tmap_internal
tmapValuesScale_bgcol_alpha = function(x, scale) x
#' @export
#' @rdname tmap_internal
tmapValuesScale_text = function(x, scale) x
#' @export
#' @rdname tmap_internal
tmapValuesScale_fontface = function(x, scale) x
#' @export
#' @rdname tmap_internal
tmapValuesScale_xmod = tmapValuesScale_ymod = function(x, scale) x
#' @export
#' @rdname tmap_internal
tmapValuesScale_angle = function(x, scale) x

#' @export
#' @rdname tmap_internal
tmapValuesColorize_col = function(x, pc) do.call(process_color, c(list(col = x), pc))
#' @export
#' @rdname tmap_internal
tmapValuesColorize_fill = function(x, pc) do.call(process_color, c(list(col = x), pc))
#' @export
#' @rdname tmap_internal
tmapValuesColorize_bgcol = function(x, pc) do.call(process_color, c(list(col = x), pc))
#' @export
#' @rdname tmap_internal
tmapValuesColorize_size = function(x, pc) x
#' @export
#' @rdname tmap_internal
tmapValuesColorize_area = function(x, pc) x
#' @export
#' @rdname tmap_internal
tmapValuesColorize_lwd = function(x, pc) x
#' @export
#' @rdname tmap_internal
tmapValuesColorize_lty = function(x, pc) x
#' @export
#' @rdname tmap_internal
tmapValuesColorize_shape = function(x, pc) x
#' @export
#' @rdname tmap_internal
tmapValuesColorize_col_alpha = function(x, pc) x
#' @export
#' @rdname tmap_internal
tmapValuesColorize_fill_alpha = function(x, pc) x
#' @export
#' @rdname tmap_internal
tmapValuesColorize_bgcol_alpha = function(x, pc) x
#' @export
#' @rdname tmap_internal
tmapValuesColorize_text = function(x, pc) x
#' @export
#' @rdname tmap_internal
tmapValuesColorize_fontface = function(x, pc) x
#' @export
#' @rdname tmap_internal
tmapValuesColorize_xmod = tmapValuesColorize_ymod = function(x, pc) x
#' @export
#' @rdname tmap_internal
tmapValuesColorize_angle = function(x, pc) x


#' Specify a numeric sequence
#'
#' Specify a numeric sequence, for numeric scales like [tm_scale_continuous()]. This function is needed when there is a non-linear relationship between the numeric data values and the visual variables. E.g. to make relationship with the area of bubbles linear, the square root of input variables should be used to calculate the radius of the bubbles.
#'
#' The perceived area of larger symbols is often underestimated. Flannery (1971) experimentally derived a method to compensate this for symbols. This compensation is obtained by using the power exponent of 0.5716 instead of 0.5, or by setting `power` to `"sqrt_perceptual"`
#'
#' @param from,to The numeric range, default 0 and 1 respectively
#' @param power The power component, a number or or one of `"lin"`, `"sqrt"`, `"sqrt_perceptual"`, `"quadratic"`, which correspond to 1, 0.5, 0.5716, 2 respectively. See details.
#' @export
tm_seq = function(from = 0, to = 1, power = c("lin", "sqrt", "sqrt_perceptual", "quadratic")) {
	if (length(power) > 1) power = power[1]
	structure(as.list(environment()), class = "tmapSeq")
}

# x is vector, rng is its range
norm_vector = function(x, rng) {
	(x - rng[1]) / diff(rng)
}

scale_vector = function(x, new_rng) {
	(x + new_rng[1]) * diff(new_rng)
}

#' @export
#' @rdname tmap_internal
tmapSeq = function(s, n = NULL) {
	if (is.null(n) && is.null(s$values)) stop("One of n or s$values should be provided")
	if (is.null(s$values)) s["values"] = list(NULL)
	with(s, {
		p = if (is.numeric(power)) power else switch(power, lin = 1, sqrt = 0.5, sqrt_perceptual = 0.5716, quadratic = 2)
		r = seq(from = from, to = to, length.out = n) ^ p
	})
}

#' @export
#' @rdname tmap_internal
transform_values = function(x, lim, rng, power, scale, include.neutral = TRUE) {
	p = if (is.numeric(power)) power else switch(power, lin = 1, sqrt = 0.5, sqrt_perceptual = 0.5716, quadratic = 2)
	if (p != 1) rng = rng ^ (1/p)

	x2 = norm_vector(x, lim)
	x3 = if (rng[1] != 0 || rng[2] != 1) scale_vector(x2, rng) else x2

	if (include.neutral) neutral = mean(rng)
	if (p != 1) {
		x3 = x3 ^ p
		if (include.neutral) neutral = neutral ^ p
	}
	x4 = x3 * scale
	if (include.neutral) neutral = neutral * scale

	if (include.neutral) {
		list(x = x4,
			 neutral = neutral)
	} else {
		x4
	}
}


#' @export
#' @rdname tmap_internal
tmapValuesCVV_fill = function(x, value.na, n, range, scale, rep, o) {


	# process values
	#palid = tmapPalId(x[1])

	arecolors = valid_colors(x[1])
	m = getPalMeta(x[1])

	ispalette = !is.null(m) && !arecolors # the latter in case of ambiguity (e.g. "blue")

	if (ispalette) if (x[1] != tolower(x[1])) message_c4a(x[1], info = m)


	values = if (!ispalette && !arecolors) {
		rep(x, length.out = n)
	} else if (ispalette) {
		getPal(m$fullname, n, rep = rep, range = range, reversed = m$reverse)
	} else if (!rep && (length(x) < n)) {
		grDevices::colorRampPalette(x)(n)
	} else {
		rep(x, length.out=n)
	}

	nms = names(values)
	values = do.call(process_color, c(list(col = values), o$pc))
	names(values) = nms
	value.neutral = do.call(process_color, c(list(col = values[1]), o$pc))
	value.na = do.call(process_color, c(list(col = value.na), o$pc))


	list(vvalues = values, value.neutral = value.neutral, value.na = value.na)

	# } else if (arenumbers) {
	# 	values = if (length(scale$values) == 2) seq(scale$values[1], scale$values[2], length.out = n) else rep(scale$values, length.out = n)
	# } else {
	# 	values = rep(scale$values, length.out = n)
	# }

}


#' @export
#' @rdname tmap_internal
tmapValuesCVV_col = function(x, value.na, n, range, scale, rep, o) {
	tmapValuesCVV_fill(x, value.na, n, range, scale, rep, o)
}

#' @export
#' @rdname tmap_internal
tmapValuesCVV_bgcol = function(x, value.na, n, range, scale, rep, o) {
	tmapValuesCVV_fill(x, value.na, n, range, scale, rep, o)
}

#' @export
#' @rdname tmap_internal
tmapValuesCVV_size = function(x, value.na, n, range, scale, rep, o) {
	tmapValuesVV_size(x = x, value.na = value.na, isdiv = FALSE, n = n, dvalues = NA, are_breaks = FALSE, midpoint = NA, range = range, scale = scale, rep = rep)
}

#' @export
#' @rdname tmap_internal
tmapValuesCVV_area = function(x, value.na, n, range, scale, rep, o) {
	tmapValuesVV_area(x = x, value.na = value.na, isdiv = FALSE, n = n, dvalues = NA, are_breaks = FALSE, midpoint = NA, range = range, scale = scale, rep = rep)
}

#' @export
#' @rdname tmap_internal
tmapValuesCVV_lwd = function(x, value.na, n, range, scale, rep, o) {
	tmapValuesVV_lwd(x = x, value.na = value.na, isdiv = FALSE, n = n, dvalues = NA, are_breaks = FALSE, midpoint = NA, range = range, scale = scale, rep = rep)
}

#' @export
#' @rdname tmap_internal
tmapValuesCVV_col_alpha = function(x, value.na, n, range, scale, rep, o) {
	tmapValuesVV_col_alpha(x = x, value.na = value.na, isdiv = FALSE, n = n, dvalues = NA, are_breaks = FALSE, midpoint = NA, range = range, scale = scale, rep = rep)
}

#' @export
#' @rdname tmap_internal
tmapValuesCVV_fill_alpha = function(x, value.na, n, range, scale, rep, o) {
	tmapValuesVV_fill_alpha(x = x, value.na = value.na, isdiv = FALSE, n = n, dvalues = NA, are_breaks = FALSE, midpoint = NA, range = range, scale = scale, rep = rep)
}

#' @export
#' @rdname tmap_internal
tmapValuesCVV_bgcol_alpha = function(x, value.na, n, range, scale, rep, o) {
	tmapValuesVV_bgcol_alpha(x = x, value.na = value.na, isdiv = FALSE, n = n, dvalues = NA, are_breaks = FALSE, midpoint = NA, range = range, scale = scale, rep = rep)
}

#' @export
#' @rdname tmap_internal
tmapValuesCVV_area = function(x, value.na, n, range, scale, rep, o) {
	tmapValuesVV_area(x = x, value.na = value.na, isdiv = FALSE, n = n, dvalues = NA, are_breaks = FALSE, midpoint = NA, range = range, scale = scale, rep = rep)
}

#' @export
#' @rdname tmap_internal
tmapValuesCVV_xmod = function(x, value.na, n, range, scale, rep, o) {
	tmapValuesVV_xmod(x = x, value.na = value.na, isdiv = FALSE, n = n, dvalues = NA, are_breaks = FALSE, midpoint = NA, range = range, scale = scale, rep = rep)
}

#' @export
#' @rdname tmap_internal
tmapValuesCVV_ymod = function(x, value.na, n, range, scale, rep, o) {
	tmapValuesVV_ymod(x = x, value.na = value.na, isdiv = FALSE, n = n, dvalues = NA, are_breaks = FALSE, midpoint = NA, range = range, scale = scale, rep = rep)
}

#' @export
#' @rdname tmap_internal
tmapValuesCVV_angle = function(x, value.na, n, range, scale, rep, o) {
	tmapValuesVV_angle(x = x, value.na = value.na, isdiv = FALSE, n = n, dvalues = NA, are_breaks = FALSE, midpoint = NA, range = range, scale = scale, rep = rep)
}

#' @export
#' @rdname tmap_internal
tmapValuesCVV_shape = function(x, value.na, n, range, scale, rep, o) {
	tmapValuesVV_shape(x = x, value.na = value.na, isdiv = FALSE, n = n, dvalues = NA, are_breaks = FALSE, midpoint = NA, range = range, scale = scale, rep = rep)
}

#' @export
#' @rdname tmap_internal
tmapValuesCVV_lty = function(x, value.na, n, range, scale, rep, o) {
	tmapValuesVV_lty(x = x, value.na = value.na, isdiv = FALSE, n = n, dvalues = NA, are_breaks = FALSE, midpoint = NA, range = range, scale = scale, rep = rep)
}

#' @export
#' @rdname tmap_internal
tmapValuesCVV_text = function(x, value.na, n, range, scale, rep, o) {
	tmapValuesVV_text(x = x, value.na = value.na, isdiv = FALSE, n = n, dvalues = NA, are_breaks = FALSE, midpoint = NA, range = range, scale = scale, rep = rep)
}

#' @export
#' @rdname tmap_internal
tmapValuesCVV_fontface = function(x, value.na, n, range, scale, rep, o) {
	tmapValuesVV_fontface(x = x, value.na = value.na, isdiv = FALSE, n = n, dvalues = NA, are_breaks = FALSE, midpoint = NA, range = range, scale = scale, rep = rep)
}

# bivariate visual variables
#' @export
#' @rdname tmap_internal
tmapValuesBVV_fill = function(x, value.na, m, n, scale, rep, o) {
	#palid = tmapPalId(x[1])

	meta = getPalMeta(x[1])

	if (!is.null(meta)) {
		values = getPalBiv(x, m = m, n = n)
	} else if (!is.matrix(x)) {
		values = matrix(grDevices::colorRampPalette(x)(m*n), nrow = m)
	} else if (is.matrix(x)) {
		if (ncol(x) != n) {
			x = do.call(rbind, apply(x, function(xi) {
				grDevices::colorRampPalette(xi)(n)
			}, MARGIN = 1, simplify = FALSE))
		}

		if (nrow(x) != m) {
			x = do.call(cbind, apply(x, function(xi) {
				grDevices::colorRampPalette(xi)(m)
			}, MARGIN = 2, simplify = FALSE))
		}
		values = x
	}

	values[] = do.call(process_color, c(list(col = values), o$pc))
	value.neutral = do.call(process_color, c(list(col = values[1]), o$pc))
	value.na = do.call(process_color, c(list(col = value.na), o$pc))


	list(vvalues = values, value.neutral = value.neutral, value.na = value.na)

}

#' @export
#' @rdname tmap_internal
tmapValuesBVV_col = function(x, value.na, m, n, scale, rep, o) {
	tmapValuesBVV_fill(x, value.na, m, n, scale, rep, o)
}

#' @export
#' @rdname tmap_internal
tmapValuesBVV_bgcol = function(x, value.na, m, n, scale, rep, o) {
	tmapValuesBVV_fill(x, value.na, m, n, scale, rep, o)
}


#' @export
#' @rdname tmap_internal
tmapValuesCheck_num = tmapValuesCheck_size
#' @export
#' @rdname tmap_internal
tmapValuesIsDiv_num = tmapValuesIsDiv_size
#' @export
#' @rdname tmap_internal
tmapValuesRange_num = tmapValuesRange_size
#' @export
#' @rdname tmap_internal
tmapValuesVV_num = tmapValuesVV_size
#' @export
#' @rdname tmap_internal
tmapValuesSubmit_num = tmapValuesSubmit_size
#' @export
#' @rdname tmap_internal
tmapValuesScale_num = tmapValuesScale_size
#' @export
#' @rdname tmap_internal
tmapValuesColorize_num = tmapValuesColorize_size
#' @export
#' @rdname tmap_internal
tmapValuesCVV_num = tmapValuesCVV_size

#' @export
#' @rdname tmap_internal
tmapValuesCheck_skip = function(x, is_var) TRUE
#' @export
#' @rdname tmap_internal
tmapValuesIsDiv_skip = function(x) FALSE
#' @export
#' @rdname tmap_internal
tmapValuesRange_skip = function(x, n, isdiv) c(0, 1)
#' @export
#' @rdname tmap_internal
tmapValuesVV_skip = function(x, value.na, isdiv, n, dvalues, are_breaks, midpoint, range, scale, rep, o) {
	list(vvalues = rep(x, length.out = n), value.neutral = x[1], value.na = value.na)
}
#' @export
#' @rdname tmap_internal
tmapValuesSubmit_skip = function(x, args) x
#' @export
#' @rdname tmap_internal
tmapValuesScale_skip = function(x, scale) x
#' @export
#' @rdname tmap_internal
tmapValuesColorize_skip = function(x, pc) x
#' @export
#' @rdname tmap_internal
tmapValuesCVV_skip = function(x, value.na, n, range, scale, rep, o) {
	tmapValuesVV_shape(x = x, value.na = value.na, isdiv = FALSE, n = n, dvalues = NA, are_breaks = FALSE, midpoint = NA, range = range, scale = scale, rep = rep)
}

message_init = function() {
	.TMAP$messages = character()
	invisible(NULL)
}

message_thrown = function(id) {
	id %in% .TMAP$messages
}

message_reg = function(id) {
	.TMAP$messages = c(.TMAP$messages, id)
	NULL
}

message_comp_scale = function() {
	if (!message_thrown("comp_scale")) {
		cli::cli_inform(c(
			"{.field [plot mode]} fit legend/component: Some legend items or map compoments do not fit well, and are therefore rescaled.",
			i = "Set the tmap option {.code component.autoscale =  FALSE} to disable rescaling."
			),
			.frequency_id = "comp_scale"
		)
		message_reg("comp_scale")
	}
	NULL
}

message_comp_high_wide = function(stack) {
	if (!message_thrown("comp_scale")) {
		cli::cli_inform(c(
			"{.field [plot mode]} legend/component: Some components or legends are too {.val {ifelse(stack == 'vertical', 'high', 'wide')}} and are therefore rescaled.",
			"i" = "Set the tmap option {.code component.autoscale = FALSE} to disable rescaling."
			),
			.frequency_id = "comp_scale"
		)
		message_reg("comp_scale")
	}
	NULL
}

message_c4a = function(old_palette_name, info, fullname = FALSE) {
	new1 = info$fullname
	new2 = info$name
	mess = 	paste0("c4a_", old_palette_name)

	if (!message_thrown(mess)) {
		if (fullname) {
			cli::cli_inform(
				"{.field [cols4all]} color palettes: use palettes from the R package cols4all. Run {.run cols4all::c4a_gui()} to explore them. The old palette name {.str {old_palette_name}} is named {.val {new1}}",
				.frequency_id = "cols4all"
			)
		} else {
			cli::cli_inform(
				"{.field [cols4all]} color palettes: use palettes from the R package cols4all. Run {.run cols4all::c4a_gui()} to explore them. The old palette name {.str {old_palette_name}} is named {.str {new2}} (in long format {.str {new1}})",
				.frequency_id = "cols4all"
			)

		}
		message_reg(mess)
	}
}

message_nothing_to_show = function(any_groups) {
	if (any_groups) {
		cli::cli_inform("{.field [nothing to show]} no data layers defined after {.fn tm_shape}",
						.frequency_id = "nothing")
	} else {
		cli::cli_inform("{.field [nothing to show]} no data layers defined",
						.frequency_id = "nothing")
	}
	NULL
}

message_wrapstack = function(horizontal = TRUE) {
	if (horizontal) {
		cli::cli_inform("{.field [facets]} use {.fn tm_facets_hstack} instead of {.fn tm_facets_wrap} to put the legends next to and aligned with the facets.")
	} else {
		cli::cli_inform("{.field [facets]} use {.fn tm_facets_vstack} instead of {.fn tm_facets_wrap} to put the legends next to and aligned with the facets.")
	}
	NULL
}

message_pos_auto = function(type) {
	if (!message_thrown("pos_auto")) {
		fun = if (type == "autoout") "tm_pos_auto_out" else "tm_pos_auto_in"
		fun2 = if (type == "autoout") "tm_pos_out" else "tm_pos_in"
		cli::cli_inform("{.field [position]} use {.fn {fun2}} instead of {.fn {fun}}. The latter should be used with {.fn tmap_options}.")
		message_reg("pos_auto")
	}
	NULL
}

error_dimvalues = function() {
	cli::cli_abort(
		"{.val dimvalues} has been used, but the shape object does not contain any dimensions; please use {.arg x} to specify variables",
		call = call("tm_vars")
	)
}

message_webgl_vars = function(supported, vary) {
	sup_text = paste(supported, collapse = ", ")
	var_text = paste(names(vary)[vary], collapse = ", ")
	var_sel = names(vary)[vary]

	cli::cli_inform("{.field [view mode]} WebGL enabled, but the only supported visual variables are: {.val {supported}}. The visual variable(s) {.val {var_sel}} are not supported. Set {.code use_WebGL = FALSE} to support them.")

}

message_webgl_hover = function(type) {
	cli::cli_inform("{.field [view mode]} WegGL enabled, but it does not support hover labels for layer type {.val {type}}. Set {.code use_WebGL = FALSE} to support them.",
					.frequency_id = "webGL_hover",
					.frequency = "once")
}


message_webgl_checks = function(checks, checkif) {
	vals = paste(paste(names(checkif)[!checks], checkif[!checks], sep = " = "), collapse = ", ")
	cli::cli_inform("{.field [view mode]} WebGL enabled, but the following visual variable only accept one value {.arg {vals}}. Set {.code use_WebGL = FALSE} to support them.",
					.frequency_id = "webGL_vv",
					.frequency = "once")
}

message_webgl_crs_simple = function() {
	cli::cli_inform("{.field [view mode]} WebGL does not work (yet) with projected map projections, so it has been disabled.",
					.frequency_id = "webGL_crs",
					.frequency = "once")

}


message_qtm_empty = function() {
	cli::cli_inform(
		"{.field [qtm()]} nothing to show. Either specify {.arg shp} or {.arg basemap}. Alternatively, switch to view mode with {.code tmap_mode(\"view\")}"
	)
}

message_layer_unused_args = function(layer_fun, args) {
	cli::cli_inform(
		"{.field [{layer_fun}()]} Argument{ifelse(length(args)>1, 's', '')} {.arg {args}} unknown."
	)
}

message_crs_property_unknown = function() {
	cli::cli_inform(
		"{.field [shiny]} some css styling (used for background and markers) cannot be used yet in shiny apps",
		.frequency_id = "shiny_css",
		.frequency = "once")
}

message_crs_property_not_used = function() {
	cli::cli_inform(
		"{.field [tm_crs()]} {.arg crs} specified, so {.arg property} is ignored"
	)
}

message_crs_ll = function() {
	cli::cli_inform(
		"{.field [tip]} Consider a suitable map projection, e.g. by adding {.code + tm_crs({.str auto})}.",
		.frequency_id = "crs",
		.frequency = "once"
	)
}

message_scale_interval_value0 = function(aes, values, layer) {
	suggestion = if (inherits(values, "tmapSeq")) {
		to = values$to
		power = values$power
		if (is.character(power)) power = paste0("\"", power, "\"")
		paste0("values = tm_seq(0.25, ", to, ", power = ", power, ")")
	} else {
		to = values[2]
		paste0("values = c(0.25, ", to, ")")
	}


	cli::cli_inform(
		"{.field [layer {layer}, tm_scale_intervals()]} By default the value of the visual variable {.code {aes}} for the first interval is 0. Consider {.code {suggestion}} (or another lower bound number than 0.25). Or alternatively, use {.fun tm_scale_continuous}.",
		.frequency_id = "size0"
	)
}

message_shiny_css = function() {
	cli::cli_inform(
		"{.field [layer {layer}, tm_scale_intervals()]} By default the value of the visual variable {.code {aes}} for the first interval is 0. Consider {.code {suggestion}} (or another lower bound number than 0.25). Or alternatively, use {.fun tm_scale_continuous}.",
		.frequency_id = "size0"
	)
}

message_basemaps_none = function(serv, z) {
	cli::cli_inform(
		"{.field [basemaps]} Tiles from {.str {serv}} at zoom level {z} couldn't be loaded",
		.frequency_id = "basemap_none",
		.frequency = "once"
	)

}

message_basemaps_invalid_provider = function(serv, mode, new) {
	cli::cli_warn(
		"{.field [basemaps]} Provider {.str {serv}} does not exist for mode {.str {mode}}. Run {.fun tmap_providers} to see which ones are available. Basemap is set to {.str {new}}."
	)

}

message_basemaps_blurry = function(serv) {
	cli::cli_inform(
		"{.field [basemaps]} Tiles from {.str {serv}} will be projected so details (e.g. text) could appear blurry",
		.frequency_id = "basemap_blurry",
		.frequency = "once"
	)
}


message_basemaps = function(is_stadia) {
	provider = ifelse(is_stadia, "Stadia", "Thunderforest")
	cli::cli_inform(
		"{.field [basemaps]} An API key is requierd for tiles from {.str {provider}}",
		.frequency_id = "basemap_API",
		.frequency = "once"
	)
}

message_midpoint = function(aes, scale) {
	cli::cli_inform(
		"{.field [{.fun {scale}}]} Variable(s) {.str {aes}} contains positive and negative values, so midpoint is set to 0. Set midpoint = NA to show the full range of visual values.",
		.frequency_id = "midpoint",
		.frequency = "once"
	)
}

#' @export
tmapReproject.stars = function(shp, tmapID, bbox = NULL, ..., crs) {
	shp[[1]] = tmapID

	args = list(...)
	raster.warp = if ("raster.warp" %in% names(args)) args$raster.warp else TRUE

	shp2 = transwarp(shp, crs, raster.warp = raster.warp)
	tmapID2 = shp2[[1]]
	shp2[[1]] = NA

	if (!is.null(bbox$x)) {
		if (identical(bbox$x, "FULL")) {
			bbox = list(x = full_bbox(crs))
		} else {
			bbox = list(x = do.call(tmaptools::bb, c(bbox, list(projection = crs))))
		}
	}

	shapeTM(shp2, tmapID2, bbox, ...)
}

#' @export
tmapReproject.dimensions = function(shp, tmapID, bbox = NULL, ..., crs) {
	s = structure(list(tmapID = matrix(tmapID, nrow = nrow(shp))), dimensions = shp, class = "stars")

	args = list(...)
	raster.warp = if ("raster.warp" %in% names(args)) args$raster.warp else TRUE

	if (is.na(sf::st_crs(shp))) {
		shp2 = s
	} else {
		shp2 = transwarp(s, crs, raster.warp = raster.warp)
	}

	tmapID2 = shp2[[1]]

	d2 = stars::st_dimensions(shp2)

	if (!is.null(bbox$x)) {
		if (identical(bbox$x, "FULL")) {
			bbox = list(x = full_bbox(crs))
		} else {
			bbox = list(x = do.call(tmaptools::bb, c(bbox, list(projection = crs))))
		}
	}

	shapeTM(d2, tmapID2, bbox, ...)
}


#' @export
tmapShape.stars = function(shp, is.main, crs, bbox, unit, filter, layer, shp_name, smeta, o, tmf) {
	if (identical(crs, "auto")) crs = auto_crs(shp, crs_extra = o$crs_extra, crs_global = o$crs_global)

	dev = getOption("tmap.devel.mode")

	if (dev) cat("-- stars object:", shp_name, "--\n")


	if (!has_raster(shp)) {
		dimnms = dimnames(shp)

		dimvals = lapply(1:length(dimnms), function(i) stars::st_get_dimension_values(shp, i))
		dimsfc = vapply(dimvals, inherits, what = "sfc", FUN.VALUE = logical(1))

		if (!any(dimsfc)) {
			stop("stars object ", shp_name, " is a stars object without raster and doens't have a geometry dimension")
		} else {
			dimid = which(dimsfc)
			geoms = dimvals[[dimid]]
			dimnms_new = dimnms
			dimnms_new[dimid] = "tmapID__"
			dimcols = dimnms_new[-dimid] # columns names, used for default facetting
			dimvls = lapply(dimcols, function(d) stars::st_get_dimension_values(shp, d))
			shpnames = names(shp)
			shp = stars::st_set_dimensions(shp, dimnms[dimid], values = seq_along(geoms))
			shp = stars::st_set_dimensions(shp, names = dimnms_new)
		}

		dt = as.data.table(shp)

		if (!is.null(crs) && sf::st_crs(geoms) != crs) {
			shp = sf::st_transform(shp, crs)
		} else {
			shp = geoms
		}
		shpTM = shapeTM(shp = shp, tmapID = seq_along(shp), bbox = bbox)


		shpclass = "sfc"
	} else {
		shp = downsample_stars(shp, max.raster = max_cells(o$raster.max_cells) / (o$fn[1] * o$fn[2]))
		if (!is.null(crs) && sf::st_crs(shp) != crs) {
			shp = transwarp(shp, crs, raster.warp = o$raster.warp)
		}

		dims = stars::st_dimensions(shp)
		rst = attr(dims, "raster")

		dim_xy = get_xy_dim(shp)
		dimsxy = dims[names(dim_xy)]

		if (rst$curvilinear) {
			shp3 = shp
			attr(shp3, "dimensions")[[rst$dimensions[1]]]$values = 1L:nrow(shp)
			attr(shp3, "dimensions")[[rst$dimensions[2]]]$values = 1L:ncol(shp)
			attr(attr(shp3, "dimensions"), "raster")$curvilinear = FALSE
		} else {
			shp2 = stars::st_set_dimensions(shp, rst$dimensions[1], values = 1L:nrow(shp))
			shp3 = stars::st_set_dimensions(shp2, rst$dimensions[2], values = 1L:ncol(shp))
		}


		dt = as.data.table(shp3, center = FALSE)
		names(dt)[1:2] = c("X__", "Y__")

		if (!all(names(shp3) %in% names(dt))) {
			subst_names = tail(names(dt), length(shp3))
			for (i in 1L:length(shp3)) {
				setnames(dt, subst_names[i], names(shp3)[i])
			}
		}

		dt[, tmapID__ := as.integer((Y__-1) * nrow(shp) + X__)]
		dt[, X__:= NULL]
		dt[, Y__:= NULL]

		data.table::setorder(dt, cols = "tmapID__")

		shp = dimsxy
		shpclass = "stars"


		shpTM = shapeTM(shp = shp, tmapID = 1L:(nrow(shp) * ncol(shp)), bbox = bbox)

	}

	make_by_vars(dt, tmf, smeta)


	dtcols = setdiff(names(dt), "tmapID__")

	filter = if (is.null(filter)) {
		rep(TRUE, nrow(dt))
	} else filter[dt$tmapID__]
	dt[, ':='(sel__ = filter)] # tmapID__ = 1L:nrow(dt),

	for (nm in dtcols) {
		if (is.factor(dt[[nm]])) {
			lvls = levels(dt[[nm]])
			cols = attr(dt[[nm]], "colors")
			ids = seq_along(lvls)[lvls!="NA"]
			if (!is.null(cols)) {
				#if (any(lvls=="NA")) {
				dt[[nm]] = factor(as.integer(dt[[nm]]), levels = ids, labels = paste(lvls[ids], cols[ids], sep = "=<>="))
				#}
				#dt[[nm]] = factor()
			}
		}
	}

	structure(list(shpTM = shpTM, dt = dt, is.main = is.main, dtcols = dtcols, shpclass = shpclass, bbox = bbox, unit = unit, shp_name = shp_name, smeta = smeta, type_ids = NULL, type_vars = NULL), class = "tmapShape")
}


#' @export
tmapSubsetShp.stars_proxy = function(shp, vars) {
	shp
}

#' @export
tmapSubsetShp.stars = function(shp, vars) {
	ids = which(names(shp) %in% vars)
	nom = setdiff(vars, names(shp))
	if (length(nom)) cli::cli_abort("dimension values {.arg {nom}} not found in shape. Options are {.arg {names(shp)}}")
	shp2 = shp[ids]
	if (!length(vars)) {
		shp2$dummy__ = TRUE
	}
	shp2
}

#' @export
tmapGetShapeMeta1.stars_proxy = function(shp, layer, o) {
	shp = stars::st_downsample(shp, floor(dim(shp)[1:2])/2)
	tmapGetShapeMeta1.stars(shp, layer, o)
}

#' @export
tmapGetShapeMeta1.stars = function(shp, layer, o) {
	d = stars::st_dimensions(shp)

	if (!has_raster(shp)) {
		d_non_xy = local({
			dimvals = lapply(seq_along(d), function(i) stars::st_get_dimension_values(shp, i))
			dimsfc = vapply(dimvals, inherits, what = "sfc", FUN.VALUE = logical(1))
			d[!dimsfc]
		})
	} else {
		d_non_xy = local({
			dxy = attr(d, "raster")$dimensions
			d[setdiff(names(d), dxy)]
		})
	}

	dims = names(d_non_xy)
	dims_vals = lapply(dims, function(d) stars::st_get_dimension_values(shp, d))
	names(dims_vals) = dims

	vars_orig = names(shp)
	vars = vars_orig #make.names(vars_orig)
	names(vars) = vars_orig
	list(vars = vars,
		 dims = dims,
		 dims_vals = dims_vals)
}

#' @export
tmapGetShapeMeta2.stars_proxy = function(shp, smeta, o) {
	vars = character(0)
	smeta$vars_levs = list()
	smeta
}

#' @export
tmapGetShapeMeta2.stars = function(shp, smeta, o) {
	smeta$vars_levs = lapply(seq_len(length(shp)), function(i) {
		get_fact_levels_na(shp[[i]], o)
	})
	names(smeta$vars_levs) = names(shp)
	smeta
}

#' Specify facets
#'
#' @description
#' * `tm_facets_wrap()` specify facets for one grouping variable (so one faceting dimension)
#' * `tm_facets_(hv)stack()` stacks the facets either horizontally or vertically (one grouping variable).
#' * `tm_facets_grid()` specify facets for two grouping variables in a grid of rows and columns.
#' * `tm_facets_pagewise()` same as wrap, but the facets are drawn on different plots (pages). Replaces the `along` argument from version 3.
#' * `tm_facets()` is the core function, but it is recommended to use the other functions.
#'
#' @param by Group by variable (only for a facet wrap or facet stack)
#' @param rows Variable that specifies the rows (only for a facet grid)
#' @param columns Variable that specifies the columns (only for a facet grid)
#' @param pages Variable that specifies the pages (only for a facet grid)
#' @param as.layers show facets as layers?
#' @param nrow Number of rows
#' @param ncol Number of columns
#' @param byrow Should facets be wrapped by row?
#' @param orientation For facet stack: horizontal or vertical?
#' @param free.coords Logical. If the `by` argument is specified, should each
#'   map has its own coordinate ranges? By default `TRUE`, unless facets are shown
#'   in as different layers (`as.layers = TRUE`)
#' @param drop.units Logical. If the `by` argument is specified, should
#'   non-selected spatial units be dropped? If `FALSE`, they are plotted where
#'   mapped aesthetics are regarded as missing values. Not applicable for
#'   raster shapes. By default `TRUE`.
#' @param drop.empty.facets Logical. If the `by` argument is specified, should
#'   empty facets be dropped? Empty facets occur when the `by`-variable contains
#'   unused levels. When `TRUE` and two `by`-variables are specified, empty rows
#'   and columns are dropped.
#' @param drop.NA.facets Logical. If the `by` argument is specified, and all
#'   data values for specific facets are missing, should these facets be dropped?
#'   `FALSE` by default. In v3, it was called `showNA`.
#' @param sync Logical. Should the navigation in view mode (zooming and panning)
#'   be synchronized? By default `TRUE` if the facets have the same bounding box.
#'   This is generally the case when rasters are plotted, or when `free.coords` is
#'   `FALSE`.
#' @param swipe Logical. If \code{TRUE}, the two maps are overlaid with a draggable divider to reveal either map. Requires exactly two maps, and only works in the \code{"mapbox"} and \code{"maplibre"} modes provided by the \pkg{tmap.mapgl} package. Default is \code{FALSE}.
#' @param na.text Text used for facets of missing values. In v3, it was `textNA`.
#' @param scale.factor Number that determines how the elements (e.g. font sizes,
#'   symbol sizes, line widths) of the small multiples are scaled in relation to
#'   the scaling factor of the shapes. The elements are scaled to the `scale.factor`th
#'   root of the scaling factor of the shapes. So, for `scale.factor = 1`,
#'   they are scaled proportional to the scaling of the shapes. Since elements,
#'   especially text, are often too small to read, a higher value is recommended.
#'   By default, `scale.factor = 2`.
#' @param type `"grid"`, `"wrap"` or `"stack"`
#' @param free.scales deprecated. Please use the `.free` arguments in the layer functions, e.g. `fill.free` in `tm_polygons`.
#' @param ... used to catch deprecated arguments
#' @seealso [tm_animate()]
#' @example ./examples/tm_facets.R
#' @note In older versions (< 4.1) [tm_facets()] with `page` specification was used to create animations frames and [tmap_animation()] to create the animation itself using inputs like the frame rate specification. As of version 4.2, the whole animation, including frame rate, is specified in [tm_animate()]. The animation can still be saved via [tmap_animation()].
#' @seealso \href{https://r-tmap.github.io/tmap/articles/basics_facets}{Vignette about facets}
#' @export
tm_facets = function(by = NULL,
					 rows = NULL,
					 columns = NULL,
					 pages = NULL,
					 as.layers = FALSE,
					 nrow = NA,
					 ncol = NA,
					 byrow = TRUE,
					 orientation = NA,
					 free.coords = NA,
					 drop.units = TRUE,
					 drop.empty.facets = TRUE,
					 drop.NA.facets = FALSE,
					 sync = TRUE,
					 swipe = FALSE,
					 na.text = NA,
					 scale.factor=2,
					 type = NA,
					 free.scales = NULL,
					 ...) {

	args = list(...)
	args_called = names(rlang::call_match(dots_expand = TRUE)[-1])

	if (any(v3_only("tm_facets") %in% names(args))) {
		layer_fun = "facets"
		v3_start_message()

		v3_list_init()
		drop.empty.facets = v3_impute(args, "showNA", TRUE, paste0("drop.empty.facets = ", !args$showNA))
		na.text = v3_impute(args, "textNA", NA, "na.text")
		drop.units = v3_impute(args, "drop.shapes", TRUE, "drop.units")
		v3_tm_facets(arg_list = v3_list_get())

		if (any(substr(names(args), 1, 12) == "free.scales.")) {
			v3_tm_facets_free_scales()
		}

	}

	if (!("animate" %in% args_called)) {
		args$animate = FALSE
		#dummy values:
		args$nframes = 60L
		args$fps = 30L
		args$play = "loop"
		args$dpr = 2
	}


	if ("along" %in% args_called) {
		warning("The 'along' argument of 'tm_facets()' is deprecated as of tmap 4.0. Please use 'pages' instead.", call. = FALSE)
		pages = args$along
	}

	if (!is.null(rows) || !is.null(columns)) {
		type = "grid"
	}

	x = tm_element_list(tm_element(
		type = type,
		by = by,
		rows = rows,
		columns = columns,
		pages = pages,
		as.layers = as.layers,
		pages = pages,
		nrows = nrow,
		ncols = ncol,
		byrow = byrow,
		orientation = orientation,
		free.coords = free.coords,
		drop.units = drop.units,
		drop.empty.facets = drop.empty.facets,
		drop.NA.facets = drop.NA.facets,
		sync = sync,
		swipe = swipe,
		na.text = na.text,
		scale.factor = scale.factor,
		animate = args$animate,
		nframes = as.integer(args$nframes),
		fps = as.integer(args$fps),
		play = args$play,
		dpr = args$dpr,
		calls = args_called,
		subclass = "tm_facets"))

	if (!is.null(free.scales)) {
		message("tm_facets(): the argument free.scales is deprecated. Specify this via the layer functions (e.g. fill.free in tm_polygons)")
		x + tm_options(free.scales = free.scales)
	} else {
		x
	}
}

#' @export
#' @rdname tm_facets
tm_facets_grid = function(rows = NULL,
						  columns = NULL,
						  pages = NULL,
						  ...) {
	args = list(...)
	args_called = names(rlang::call_match()[-1])
	tm = do.call("tm_facets", c(list(by = NULL, rows = rows, columns = columns, pages = pages, type = "grid"), args[setdiff(names(args), "type")]))
	tm[[1]]$calls = args_called
	tm
}


#' @export
#' @rdname tm_facets
#' @param ... passed on to `tm_facets()`
tm_facets_wrap = function(by = "VARS__",
						  nrow = NA,
						  ncol = NA,
						  byrow = TRUE,
						  ...) {
	args = list(...)
	args_called = names(rlang::call_match()[-1])
	tm = do.call("tm_facets", c(list(by = by, nrow = nrow, ncol = ncol, byrow = byrow, type = "wrap"), args[setdiff(names(args), "type")]))
	tm[[1]]$calls = args_called
	tm
}

#' @export
#' @rdname tm_facets
tm_facets_pagewise = function(by = "VARS__",
							  byrow = TRUE,
							  ...) {
	args = list(...)
	args_called = names(rlang::call_match()[-1])

	if (any(c("nrow", "ncol") %in% args_called) && (args$nrow != 1 || args$ncol != 1)) {
		cli::cli_warn("tm_facets_pagewise does not support multiple facets per page. Please use tm_facets instead")
	}

	tm = do.call("tm_facets", c(list(pages = by, byrow = byrow, type = NA), args[setdiff(names(args), "type")]))
	tm[[1]]$calls = args_called
	tm
}

#' @export
#' @rdname tm_facets
tm_facets_stack = function(by = "VARS__",
						   orientation = NA,
						  ...) {
	args = list(...)
	args_called = names(rlang::call_match()[-1])
	tm = do.call("tm_facets", c(list(by = by, orientation = orientation, type = "stack"), args[setdiff(names(args), "type")]))
	tm[[1]]$calls = args_called
	tm
}

#' @export
#' @rdname tm_facets
tm_facets_hstack = function(by = "VARS__",
						   ...) {
	do.call("tm_facets_stack", c(list(by = by, orientation = "horizontal"), list(...)))
}

#' @export
#' @rdname tm_facets
tm_facets_vstack = function(by = "VARS__",
							...) {
	do.call("tm_facets_stack", c(list(by = by, orientation = "vertical"), list(...)))
}

#' @export
#' @rdname tm_facets
tm_facets_flip = function(...) {
	cli::cli_warn("{.field [tm_facets_flip]} please use {.code tm_facets(byrow = FALSE)} instead of {.fun tm_facets_flip}")
	do.call(tm_facets, c(list(byrow = FALSE), list(...)))
}

step1_rearrange = function(tmel, knit_opts) {
	message_init()
	dev = getOption("tmap.devel.mode")

	# get options (mode specific)
	o = tmap_options_mode(mode.specific = FALSE)

	# reset symbols
	assign("shapeLib", list(), envir = .TMAP)
	assign("justLib", list(), envir = .TMAP)

	# find shape, (aesthetic) layer, facet, and other elements
	is_tms = vapply(tmel, inherits, "tm_shape", FUN.VALUE = logical(1))
	is_tml = vapply(tmel, inherits, "tm_layer", FUN.VALUE = logical(1))
	is_tmf = vapply(tmel, inherits, "tm_facets", FUN.VALUE = logical(1))
	is_other = !is_tml & !is_tms & !is_tmf

	is_aux = vapply(tmel, inherits, "tm_aux_layer", FUN.VALUE = logical(1))
	is_proxy = vapply(tmel, inherits, "tm_proxy", FUN.VALUE = logical(1))

	# find layer id numbers (needed to plot layers in correct order, which is not trivial due to the two layer types)
	lay_id = cumsum(is_tml | is_aux)
	lay_id[!is_tml & !is_aux] = 0


	if (any(is_proxy)) {
		prx = tmel[is_proxy]
		proxy_z = lapply(prx, function(px) px$zindex)
	} else {
		prx = list()
		proxy_z = numeric(0)
	}

	# remove layers from q (via tm_remove_layer)
	# and determine highest pane number (400 default)
	if (.TMAP$proxy) {
		q = .TMAP$q
		.TMAP$pane_ids = setdiff(unique(c(.TMAP$pane_ids, q$lid)), 0)
		q = q[!(q$lid %in% proxy_z), ]
		.TMAP$q = q
		if (any(q$lid != 0)) {
			.TMAP$start_pane_id = max(q$lid)
		} else {
			.TMAP$start_pane_id = 400
		}
	} else {
		.TMAP$start_pane_id = 400
		.TMAP$pane_ids = NULL
	}

	# create groups, for each group: tms (tmap shape), tmls (tmap layers), tmf (tmap facets)
	ids = cumsum(is_tms)
	ids[is_other] = 0L
	tmel_spl = split(tmel, f = ids)
	lay_id_spl = split(lay_id, f = ids)
	if (any(is_other)) {
		oth = tmel_spl[[1]]
		oth_lay_id = lay_id_spl[[1]]
		tmel_spl = tmel_spl[-1]
		lay_id_spl = lay_id_spl[-1]
	} else {
		oth = list()
		oth_lay_id = list()
	}


	# warning("v3 code detected")
	# args$v3 = NULL
	# names(args) = paste0("legend.", names(args))
	# do.call(tm_options, args)
	#

	# update options with tm_option elements
	#sapply(oth, inherits, "tm_legend")

	is_opt = sapply(oth, inherits, "tm_options")
	crs_opt_called = FALSE
	if (any(is_opt)) for (id in which(is_opt)) {
		o2 = oth[[id]]

		if ("crs" %in% o2$calls) {
			crs_opt_called = TRUE
			o$crs = "" # to prevent merging list(dimensions = ... , ...) with leaflet crs
		}

		cls = class(o2)[1]
		if (substr(cls, nchar(cls)-2,nchar(cls)) == "_v3") cls = substr(cls, 1, nchar(cls) - 3)
		if ("v3_tm_legend" %in% o2$calls) v3_tm_legend_general(fun = cls)

		# special case: position, in case c("left", "top") is used
		pids = grep(".position", names(o2), fixed = TRUE)

		if (length(pids) > 0L) {
			for (i in 1L:length(pids)) {
				if (is.character(o2[[pids[i]]]) || is.numeric(o2[[pids[i]]])) o2[[pids[i]]] = tm_pos_in(o2[[pids[i]]][1], o2[[pids[i]]][2])
			}
		}

		if ("style" %in% names(o2) && !is.na(o2$style)) { #() {
			check_style(o2$style)
			o = tmap_options_mode(style = o2$style, mode.specific = FALSE)
			o2$style = NULL
		}
		o = complete_options(o2, o)
	}
	o = preprocess_meta_step1(o, knit_opts)


	any_data_layer = length(tmel_spl) > 0L

	if (any_data_layer) {
		# organize groups, 1 tm_shape, at least 1 tm_layers, 1 tm_facets
		tmo = mapply(function(tmg, lid) {
			is_tms = vapply(tmg, inherits, "tm_shape", FUN.VALUE = logical(1))
			is_tml = vapply(tmg, inherits, "tm_layer", FUN.VALUE = logical(1))
			is_tmf = vapply(tmg, inherits, "tm_facets", FUN.VALUE = logical(1))

			# make sure there is exactly one tm_facets per group (if there are none, add one, if there are mutple, take last)
			if (!any(is_tmf)) {
				tmf = tm_facets()[[1]]
				tmf$calls = NULL
			} else {
				# get last tm_facets element
				k = sum(is_tmf)

				is_tmf_ani = vapply(tmg, inherits, "tm_animate", FUN.VALUE = logical(1))

				ka = sum(is_tmf_ani)

				if (k == ka || ka == 0) {
					tmf = tmg[[which(is_tmf)[k]]]
				} else {
					# take last tm_facets and update with last tm_animate
					tmf = tmg[[tail(which(is_tmf & !is_tmf_ani), 1L)]]
					ani_mns = c("pages", "animate", "nframes", "fps", "play", "dpr")
					tmf[ani_mns] = tmg[[which(is_tmf_ani)[ka]]][ani_mns]
				}

				if ((k-ka) > 1) warning("Multiple tm_facets defined per layer group. Only the last one is processed", call. = FALSE)
				if (ka > 1) warning("Multiple tm_animate defined per layer group. Only the last one is processed", call. = FALSE)
			}

			# extract layers and add layer id number
			tmls = mapply(function(l, i) {
				l$lid = if (is.na(l$zindex)) i + .TMAP$start_pane_id else l$zindex
				l
			}, tmg[is_tml], lid[is_tml], SIMPLIFY = FALSE)

			structure(list(tms = tmg[[1]], tmls = tmls, tmf = tmf), class = c("tmapGroup", "list"))
		}, tmel_spl, lay_id_spl, SIMPLIFY = FALSE)

		tmo = step1_rearrange_facets(tmo, o) # save smeta's and keep track of group id (to obtain smeta)
		tmf = tmo$tmf_global # global facetting options, to be appended to options o
		tmo$tmf_global = NULL

		# find the 'main' group: this is the group for which tm_shape is used for CRS and bbox. By default, take the first, unless is.main is set to TRUE.
		# is.main can be set multiple times: the CRS will be taken from the first, but the bbox from all
		ids = get_main_ids(tmo)
		is_main_defined = !is.na(ids[1])
		if (!is_main_defined) ids = 1L

		tms = tmo[[ids[1]]]$tms


		# determine whether to animate
		tmf$show_gif_ani = is.null(o$animate_disable) && (tmf$animate || tmf$trans_animate) # animate is multiple variables/facets, trans_animate for transition animation only

		if (identical(tms$bbox$x, "FULL") || identical(o$bbox, "FULL") || o$earth_boundary) {
			o$earth_bbox = TRUE
		} else {
			o$earth_bbox = FALSE
		}

		#if (tmf$show_gif_ani) o$scale = o$scale * 2

	} else {
		tmo = NULL
		tmf = tm_facets()[[1]]
		tmf$calls = NULL
		tmf$fn = c(1, 1, 1)
		tmf$n = 1
		tmf$fl = list(NULL, NULL, NULL)
		tmf$type = "wrap"
		tmf$npp = 1
		tmf$show_gif_ani = FALSE
		ids = 0

		if (identical(o$bbox, "FULL") || o$earth_boundary) {
			o$earth_bbox = TRUE
		} else {
			o$earth_bbox = FALSE
		}
	}

	# add basemaps
	basemaps_added = any(vapply(oth, inherits, "tm_basemap", FUN.VALUE = logical(1)))
	if (o$basemap.show && !.TMAP$proxy) {
		if (!basemaps_added) {
			oth = c(oth, tm_basemap())
			oth_lay_id = c(oth_lay_id, 0L)
		}
	}

	# check for disabled aux layers (so far only working for tm_basemap/tm_tiles):
	is_aux = vapply(oth, inherits, "tm_aux_layer", FUN.VALUE = logical(1))
	if (any(is_aux)) {

		aux = mapply(function(l, i) {
			l$lid = if (is.na(l$zindex)) i + .TMAP$start_pane_id else l$zindex

			cls = class(l)[1]
			ot = get_prefix_opt(class = cls, o = o)
			l$args = complete_options(l$args, ot)
			l
		}, oth[is_aux], oth_lay_id[is_aux], SIMPLIFY = FALSE)

		# disable aux layers (so far only working for tm_basemap/tm_tiles):
		sel_aux = local({
			cls1 = vapply(aux, function(ai) {
				class(ai)[1]
			}, FUN.VALUE = character(1))
			dis = vapply(aux, function(ai) {
				"disable" %in% names(ai$args) && ai$args$disable
			}, FUN.VALUE = logical(1))
			clsu = unique(cls1)
			sel_aux = rep(TRUE, length(aux))
			for (cl in clsu) {
				ids = which(cls1 == cl)
				d = dis[ids]
				if (any(d)) {
					latest = tail(which(d), 1)
					ids_dis = 1:latest
					sel_aux[ids[ids_dis]] = FALSE
				}
			}
			sel_aux
		})
		if (!any(sel_aux)) {
			#is_aux = FALSE
			aux = list()
		} else if (!all(sel_aux)) {
			aux = aux[sel_aux]
		}
	} else {
		aux = list()
	}


	crs_step4 = o$crs

	crs_opt_defined = !is.na(crs_step4[1])

	# if basemaps are used AND tm_options(crs = ...) is not used AND tm_shape(crs = ... / is.main = TRUE) is not used:
	basemaps_defined = length(aux) && any(vapply(aux, inherits, c("tm_basemap", "tm_tiles"), FUN.VALUE = logical(1)))
	if ((is.na(crs_step4[1]) || identical(crs_step4, "auto")) && basemaps_defined && !crs_opt_called) {
		crs_step4 = o$crs_basemap
	}

	crs_step3 = if (any_data_layer) get_crs(tms, is_auto = identical(crs_step4, "auto"), crs_extra = o$crs_extra, crs_global = o$crs_global, basemaps_defined = basemaps_defined) else NA

	if (!is.na(crs_step3) && crs_is_local(crs_step3)) {
		if (crs_opt_defined) {
			cli::cli_warn("local crs used in {.fun tm_shape} which cannot be transformed to another crs")
		}
		crs_step4 = NA
		crs_opt_defined = FALSE
	}

	if (identical(crs_step4, "auto")) {
		if (is.na(crs_step3[1])) {
			# no data layer:
			crs_step4 = auto_crs(TRUE, crs_extra = o$crs_extra, crs_global = o$crs_global)
		} else {
			crs_step4 = crs_step3
		}
	}

	if (!.TMAP$proxy) {
		.TMAP$shiny_crs = crs_step4
	} else {
		crs_step4 = .TMAP$shiny_crs
	}


	if (inherits(crs_step3, "leaflet_crs")) crs_step3 = leaflet2crs(crs_step3)

	if (inherits(crs_step4, "leaflet_crs")) {
		crs_leaflet = crs_step4
		crs_step4 = sf::st_crs(4326) #leaflet2crs(crs_leaflet)
	} else if (any_data_layer && (is.na(crs_step4[1]) || (is.numeric(crs_step4) && crs_step4 == 0))) {#  || (is_main_defined && !crs_opt_called))) {
		crs_step4 = if (inherits(crs_step3, "leaflet_crs")) leaflet2crs(crs_step3) else crs_step3
		crs_leaflet = leafletSimple
	} else if (is.na(crs_step4[1])) {
		crs_step4 = sf::st_crs(4326)
		crs_leaflet = crs2leaflet(get_option_class(crs_step4, "dimensions"))
	} else {
		crs_leaflet = crs2leaflet(get_option_class(crs_step4, "dimensions"))
	}

	if (any_data_layer) {
		main_class = get_class(tms)
	} else {
		main_class = "stars" # basemaps
	}

	if (dev) timing_add(s2 = "facet meta")

	#so(crs_step4, crs_step3, crs_leaflet)


	is_comp = sapply(oth, inherits, "tm_component")
	if (any(is_comp)) {
		cmp = oth[is_comp]
		cmp = lapply(cmp, impute_comp, o = o)
	} else {
		cmp = list()
	}

	# remove basemaps and scalebars if final crs is local (so no units)
	if (basemaps_defined && crs_is_local(crs_step4)) {
		if (length(aux)) {
			is_bm = vapply(aux, inherits, c("tm_basemap", "tm_tiles"), FUN.VALUE = logical(1))
			if (any(is_bm)) aux = aux[!is_bm]
		}

		if (length(cmp)) {
			is_sb = vapply(cmp, inherits, "tm_scalebar", FUN.VALUE = logical(1))
			if (any(is_sb)) cmp = cmp[!is_sb]
		}
	}



	# to be used later
	o$main = ids # to determine total bounding box in step 4
	o$main_class = main_class # inner.margins are class specific (preprecess_meta)
	o$crs_step4 = crs_step4 # in step 3, when other shapes are transformed to this crs (prepare for plotting in step4)
	o$crs_leaflet = crs_leaflet
	o$crs_step3 = crs_step3 # step 3, transformation

	o = c(o, tmf)
	# process shapes: put non-spatial data in data.table, keep spatial data separately


	# disable s2 in case earth.boundaries are drawn
	if ((o$earth_boundary || o$earth_bbox) && sf::sf_use_s2()) {
		suppressMessages(sf::sf_use_s2(FALSE))
		.TMAP$set_s2 = TRUE
	}

	if (any_data_layer) {
		tmo = structure(lapply(tmo, function(tmg) {
			tmg$tms = do.call(tmapShape, c(tmg$tms, list(o = o, tmf = tmg$tmf)))
			tmg
		}), names = paste0("group", seq_len(length(tmo))), class = c("tmapObject", "list"))
	}

	if (dev) timing_add(s2 = "prep shape")

	if (any_data_layer && !is.null(tms$unit)) o$unit = tms$unit

	list(tmo = tmo, aux = aux, cmp = cmp, prx = prx, o = o)
}

# see above
get_main_ids = function(tmo) {
	is_main = vapply(tmo, function(tmg) {
		identical(tmg$tms$is.main, TRUE)
	}, FUN.VALUE = logical(1), USE.NAMES = FALSE)

	if (any(is_main)) which(is_main) else NA_integer_
}


get_crs = function(tms, is_auto, crs_extra, crs_global, basemaps_defined) {
	if (is.character(tms$shp)) return(sf::st_crs(4326))
	if (is.na(sf::st_crs(tms$shp))) return(sf::st_crs("ENGCRS[\"Cartesian (Meter)\",\n    EDATUM[\"Local Datum\"],\n    CS[Cartesian,2],\n        AXIS[\"x\",east,\n            ORDER[1],\n            LENGTHUNIT[\"Meter\",1]],\n        AXIS[\"y\",north,\n            ORDER[2],\n            LENGTHUNIT[\"Meter\",1]]]"))
	if (is.null(tms$crs)) {
		crs = sf::st_crs(tms$shp)
		is_ll = sf::st_is_longlat(crs) || crs == sf::st_crs(3857)
		if (is_ll && is_auto) {
			auto_crs(tms$shp, crs_extra = crs_extra, crs_global = crs_global)
		} else {
			if (is_ll && !basemaps_defined) {
				if (inherits(tms$shp, "sf") && consider_global(tms$shp)) {
					message_crs_ll()
				}

			}
			crs
		}
	} else {
		crs = tms$crs
		if (identical(crs, "auto") || is_auto) {
			auto_crs(tms$shp, crs_extra = crs_extra, crs_global = crs_global)
		} else crs
	}
}

get_class = function(tms) {
	class(tms$shp)
}

str2pos = function(x) {
	if (length(x) == 1L) {
		if (tolower(x) %in% c("left", "right")) {
			args = list(pos.h = x)
		} else if (tolower(x) %in% c("top", "bottom")) {
			args = list(pos.v = x)
		} else if (tolower(x) %in% c("center", "centre")) {
			args = list(pos.h = "center", pos.v = "center")
		} else {
			stop("position: incorrect specification")
		}
	} else {
		if (tolower(x[1]) %in% c("top", "bottom")) {
			args = list(pos.h = x[2], pos.v = x[1])
		} else {
			args = list(pos.h = x[1], pos.v = x[2])
		}
	}
	do.call(tm_pos_in, args)
}

num2pos = function(x) {
	if (length(x) == 1L) stop("position: please provide two values: the horizontal and the vertical")
	tm_pos_in(x[1], x[2])
}

process_position = function(position, o) {
	if (is.null(position)) return(NULL)

	# special case: position, in case c("left", "top") is used
	if (is.character(position)) position = str2pos(position)
	if (is.numeric(position)) position = num2pos(position)
	if (inherits(position, "tm_pos")) {
		position = complete_options(position, o$component.position[[position$type]])
		if (position$type %in% c("autoin", "autoout")) message_pos_auto(position$type)
	} else {
		stop("position should be a tm_pos object")
	}

	# checking
	# with(position, {
	# 	if (cell.h %in% c(""))
	# })

	within.list(position, {
		if (exists("cell.h")) cell.h = check_h(cell.h, "cell")
		if (exists("cell.v")) cell.v = check_v(cell.v, "cell", h_is_num = is.numeric(cell.h))
		if (exists("pos.h")) pos.h = check_h(pos.h, "pos")
		if (exists("pos.v")) pos.v = check_v(pos.v, "pos", h_is_num = is.numeric(pos.h))
		if (exists("align.h")) align.h = check_h(align.h, "align")
		if (exists("align.v")) align.v = check_v(align.v, "align", h_is_num = is.numeric(align.h))
		if (exists("just.h")) just.h = check_h(just.h, "just")
		if (exists("just.v")) just.v = check_v(just.v, "just", h_is_num = is.numeric(just.h))
	})

}

process_padding = function(padding) {
	if (is.null(padding)) return(NULL)
	rep(padding, length.out = 4)
}

check_h = function(x, var) {
	if (is.numeric(x)) {
		x
	} else {
		y = tolower(x)
		is_upper = (x != y)
		y2 = if (y %in% c("center", "centre")) {
			"center"
		} else if (y %in% c("left", "right")) {
			y
		} else {
			stop("position arguments: incorrect ", var, ".h; it should be 'left', 'center', 'right' or a numeric value", call. = FALSE)
		}
		if (is_upper) toupper(y2) else y
	}
}

check_v = function(x, var, h_is_num) {
	if (is.numeric(x)) {
		if (!h_is_num) stop("position argument ", var, ".h and .v should both be numeric or character values", call. = FALSE)
		x
	} else {
		y = tolower(x)
		is_upper = (x != y)
		if (h_is_num) stop("position argument ", var, ".h and .v should both be numeric or character values", call. = FALSE)
		y2 = if (y %in% c("center", "centre")) {
			"center"
		} else if (y %in% c("top", "bottom")) {
			y
		} else {
			stop("position argument: incorrect ", var, ".v; it should be 'top', 'center', 'bottom' or a numeric value", call. = FALSE)
		}
		if (is_upper) toupper(y2) else y
	}
}

#' @export
#' @rdname tmap_internal
tmapScaleAsIs = function(x1, scale, legend, chart, o, aes, layer, layer_args, sortRev, bypass_ord, submit_legend = TRUE) {
	legend = list(title = NA,
				  nitems = 0,
				  labels = NA,
				  dvalues = NA,
				  vvalues = NA,
				  vneutral = NA,
				  na.show = NA,
				  scale = "AsIs",
				  show = FALSE,
				  active = FALSE)

	x1h = head(x1, 100)
	check_values(layer, aes, x1h)


	scale$values.scale = if (is.na(scale$values.scale)) getAesOption("values.scale", o, aes, layer) else scale$values.scale

	isna = is.na(x1)
	if (any(isna) && is.na(scale$value.na)) {
		x1[isna] = getAesOption("value.na", o, aes, layer)
	}

	sfun = paste0("tmapValuesScale_", aes)
	cfun = paste0("tmapValuesColorize_", aes)

	sc = o$scale * scale$values.scale

	x2 = do.call(sfun, list(x = x1, scale = sc))
	values = do.call(cfun, list(x = x2, pc = o$pc))

	vneutral = if (is.na(scale$value.neutral)) {
		vna = getAesOption("value.neutral", o, aes, layer)
		if (is.na(vna)) {
			x1[1]
		} else vna
	} else {
		scale$value.neutral
	}


	legend$vneutral = do.call(sfun, list(x = do.call(cfun, list(x = vneutral, pc = o$pc)), scale = o$scale))

	if (submit_legend) {
		if (bypass_ord) {
			format_aes_results(values, legend = legend, chart = chart)
		} else {
			format_aes_results(values, ord = 1L, legend = legend, chart = chart)
		}
	} else {
		list(vals = values, ids = 1L, legend = legend, chart = chart, bypass_ord = bypass_ord)
	}
}

#' @export
#' @rdname tmap_internal
tmapValuesCheck_col = function(x, is_var = TRUE) {
	isnum = is.numeric(x)
	if (isnum) {
		structure(FALSE,
				  info = {if (is_var) "Variable should be data variable name or color name" else " Values should be numeric (between -50 and 50)."}
		)
	} else {
		is_c4a = !is.null(getPalMeta(x[1])) && length(x) == 1L && !valid_colors(x[1])
		if (is_c4a) {
			if (is_var) {
				structure(FALSE,
						  info = "Variable should be a data variable name or a single color (not a color palette).")
			} else {
				TRUE
			}
		} else {
			all_cols = all(valid_colors(x))
			if (!all_cols) {
				structure(FALSE,
						  info = if (is_var) "Variable should a data variable name or a single color." else "Values should be color names or a color palette (run {.run cols4all::c4a_palettes()} for available ones.")
			} else {
				TRUE
			}
		}
	}
}

#' @export
#' @rdname tmap_internal
tmapValuesCheck_fill = function(x, is_var = TRUE) {
	tmapValuesCheck_col(x, is_var)
}

#' @export
#' @rdname tmap_internal
tmapValuesCheck_bgcol = function(x, is_var = TRUE) {
	tmapValuesCheck_col(x, is_var)
}

isSymbol = function(s) {
	inherits(s, "grob") || any(vapply(s, inherits, FUN.VALUE = logical(1), "grob")) || ("iconUrl" %in% names(s))
}
isUrl <- function(string) {
	any(grepl("(https?|ftp)://[^\\s/$.?#].[^\\s]*", string))
}

isLocal = function(string) {
	file.exists(string)
}

#' @export
#' @rdname tmap_internal
tmapValuesCheck_shape = function(x, is_var = TRUE) {

	if (inherits(x, "tmap_icons")) {
		TRUE
	} else if (all(is.numeric(x))) {
		TRUE
	} else if (is.list(x)) {
	 	if (isSymbol(x)) {
	 		TRUE
	 	} else {
	 		all(vapply(x, isSymbol, FUN.VALUE = logical(1)) | vapply(x, is.numeric, FUN.VALUE = logical(1)))
	 	}
	} else if (all(is.character(x))) {
		if (all(isUrl(x) | isLocal(x))) {
			TRUE
		} else {
			structure(FALSE, info = {if (is_var) " Variable should be a data variable name or a symbol (see {.help [tm_symbols](tmap::tm_symbols)} - details section)." else "  Values should be symbols (see {.help [tm_symbols](tmap::tm_symbols)} - details section)."})
		}
	}
}



#' @export
#' @rdname tmap_internal
tmapValuesCheck_size = function(x, is_var = TRUE) {
	# units objects (from the units package) are accepted; they will be
	# converted to plain numeric metres in tmapValuesScale_size.
	if (inherits(x, "units")) return(TRUE)

	res = inherits(x, "tmapSeq") || (is.numeric(x) && (all(x>=0) || all(x<=0)))

	if (!res) {
		structure(FALSE, info = {if (is_var) " Variable should be a data variable name or a numeric value." else "  Values should be numeric."})
	} else {
		TRUE
	}
}


#' @export
#' @rdname tmap_internal
tmapValuesCheck_area = function(x, is_var = TRUE) {
	tmapValuesCheck_size(x, is_var)
}

#' @export
#' @rdname tmap_internal
tmapValuesCheck_lwd = function(x, is_var = TRUE) {
	tmapValuesCheck_size(x, is_var)
}

#' @export
#' @rdname tmap_internal
tmapValuesCheck_lty = function(x, is_var = TRUE) {
	if (all(x %in% c("blank", "solid", "dashed", "dotted", "dotdash", "longdash", "twodash")) ||
		(is.numeric(x) && all(x %in% 0:6)) ||
		(all(grepl("^[1-9A-F]{1,8}$", x)))) {
		TRUE
	} else {
		structure(FALSE, info = {if (is_var) " Variable should be a data variable name or a line type (see documentation of graphics::par - lty)." else "  Values should be a line type (see documentation of graphics::par - lty). numeric."})
	}
}

#' @export
#' @rdname tmap_internal
tmapValuesCheck_xmod = tmapValuesCheck_ymod = function(x, is_var = TRUE) {
	if (!is.numeric(x)) {
		structure(FALSE,
				  info = {if (is_var) " It is neither a data variable name nor a numeric (between -50 and 50)." else " Values should be numeric (between -50 and 50)."}
				  )
	} else {
		res = all(x >= -50 & x <= 50)
		if (!res) {
			structure(FALSE,
					  info = {if (is_var) " Variable should be in the [-50,50] range." else " Values found that are outside the [-50,50] range. Note that the default scale for xmod and ymod is tm_scale_asis."})
		} else {
			TRUE
		}
	}
}

#' @export
#' @rdname tmap_internal
tmapValuesCheck_angle = function(x, is_var = TRUE) {
	# to do
	if (!is.numeric(x)) {
		structure(FALSE, info = {if (is_var) " Variable should be a data variable name or a numeric value." else "  Values should be numeric."})
	} else {
		TRUE
	}
}

#' @export
#' @rdname tmap_internal
tmapValuesCheck_col_alpha= function(x, is_var = TRUE) {
	tmapValuesCheck_size(x, is_var)
}

#' @export
#' @rdname tmap_internal
tmapValuesCheck_fill_alpha = function(x, is_var = TRUE) {
	tmapValuesCheck_size(x, is_var)
}

#' @export
#' @rdname tmap_internal
tmapValuesCheck_bgcol_alpha= function(x, is_var = TRUE) {
	tmapValuesCheck_size(x, is_var)
}

#' @export
#' @rdname tmap_internal
tmapValuesCheck_area = function(x, is_var = TRUE) {
	tmapValuesCheck_size(x, is_var)
}

#' @export
#' @rdname tmap_internal
tmapValuesCheck_text = function(x, is_var = TRUE) {
	TRUE
}

#' @export
#' @rdname tmap_internal
tmapValuesCheck_fontface = function(x, is_var = TRUE) {

	if 	((is.numeric(x) && (all(x %in% 1:5))) || (is.character(x) && (all(x %in% c("plain", "bold", "italic", "oblique", "bold.italic", "cyrillic", "cyrillic.oblique", "EUC"))))) {
		TRUE
	} else {
		structure(FALSE, info = {if (is_var) " Variable should be a data variable name or a font face (see documentation of graphics::par font)." else "  Values should be a font face (see documentation of graphics::par - font). numeric."})
	}
}

#' @export
#' @rdname tmap_internal
tmapValuesIsDiv_fill = function(x) {
	m = getPalMeta(x[1])
	ispal = !is.null(m)

	if (ispal) {
		m$type == "div"
	} else {
		(palette_type(x) == "div")
	}
}

#' @export
#' @rdname tmap_internal
tmapValuesIsDiv_col = function(x) {
	tmapValuesIsDiv_fill(x)
}

#' @export
#' @rdname tmap_internal
tmapValuesIsDiv_bgcol = function(x) {
	tmapValuesIsDiv_fill(x)
}

#' @export
#' @rdname tmap_internal
tmapValuesIsDiv_size = function(x) {
	inherits(x, "tmapSeq") && (x$from < 0) && (x$to > 1) || (is.numeric(x) && (any(x < 0) && any(x> 0)))
}

#' @export
#' @rdname tmap_internal
tmapValuesIsDiv_area = function(x) {
	tmapValuesIsDiv_size(x)
}

#' @export
#' @rdname tmap_internal
tmapValuesIsDiv_lwd = function(x) {
	tmapValuesIsDiv_size(x)
}

#' @export
#' @rdname tmap_internal
tmapValuesIsDiv_lty = function(x) {
	FALSE
}

#' @export
#' @rdname tmap_internal
tmapValuesIsDiv_col_alpha = function(x) {
	tmapValuesIsDiv_size(x)
}

#' @export
#' @rdname tmap_internal
tmapValuesIsDiv_fill_alpha = function(x) {
	tmapValuesIsDiv_size(x)
}

#' @export
#' @rdname tmap_internal
tmapValuesIsDiv_bgcol_alpha = function(x) {
	tmapValuesIsDiv_size(x)
}

#' @export
#' @rdname tmap_internal
tmapValuesIsDiv_area = function(x) {
	tmapValuesIsDiv_size(x)
}

#' @export
#' @rdname tmap_internal
tmapValuesIsDiv_xmod = tmapValuesIsDiv_ymod = function(x) {
	tmapValuesIsDiv_size(x)
}

#' @export
#' @rdname tmap_internal
tmapValuesIsDiv_angle = function(x) {
	FALSE
}

#' @export
#' @rdname tmap_internal
tmapValuesIsDiv_shape = function(x) {
	FALSE
}

#' @export
#' @rdname tmap_internal
tmapValuesIsDiv_text = function(x) {
	FALSE
}

#' @export
#' @rdname tmap_internal
tmapValuesIsDiv_fontface = function(x) {
	FALSE
}


#' @export
#' @rdname tmap_internal
tmapValuesRange_fill = function(x, n, isdiv) {
	m = getPalMeta(x[1])

	if (!is.null(m)) {
		NA # in c4a palette definition
	} else c(0, 1)
}

#' @export
#' @rdname tmap_internal
tmapValuesRange_col = function(x, n, isdiv) {
	tmapValuesRange_fill(x, n, isdiv)
}

#' @export
#' @rdname tmap_internal
tmapValuesRange_bgcol = function(x, n, isdiv) {
	tmapValuesRange_fill(x, n, isdiv)
}

#' @export
#' @rdname tmap_internal
tmapValuesRange_shape = function(x, n, isdiv) {
	c(0, 1)
}

#' @export
#' @rdname tmap_internal
tmapValuesRange_lty = function(x, n, isdiv) {
	c(0, 1)
}


#' @export
#' @rdname tmap_internal
tmapValuesRange_size = function(x, n, isdiv) {
	#print(c(.5/n, 1 - .5/n))
#	c(.5/n, 1 - .5/n)
	c(0, 1)
}

#' @export
#' @rdname tmap_internal
tmapValuesRange_area = function(x, n, isdiv) {
	tmapValuesRange_size(x, n, isdiv)
}

#' @export
#' @rdname tmap_internal
tmapValuesRange_lwd = function(x, n, isdiv) {
	tmapValuesRange_size(x, n, isdiv)
}

#' @export
#' @rdname tmap_internal
tmapValuesRange_col_alpha = function(x, n, isdiv) {
	tmapValuesRange_size(x, n, isdiv)
}

#' @export
#' @rdname tmap_internal
tmapValuesRange_fill_alpha = function(x, n, isdiv) {
	tmapValuesRange_size(x, n, isdiv)
}

#' @export
#' @rdname tmap_internal
tmapValuesRange_bgcol_alpha = function(x, n, isdiv) {
	tmapValuesRange_size(x, n, isdiv)
}

#' @export
#' @rdname tmap_internal
tmapValuesRange_area = function(x, n, isdiv) {
	tmapValuesRange_size(x, n, isdiv)
}

#' @export
#' @rdname tmap_internal
tmapValuesRange_xmod = tmapValuesRange_ymod = function(x, n, isdiv) {
	tmapValuesRange_size(x, n, isdiv)
}

#' @export
#' @rdname tmap_internal
tmapValuesRange_angle = function(x, n, isdiv) {
	c(0, 1)
}

#' @export
#' @rdname tmap_internal
tmapValuesRange_text = function(x, n, isdiv) {
	c(0, 1)
}

#' @export
#' @rdname tmap_internal
tmapValuesRange_fontface = function(x, n, isdiv) {
	c(0, 1)
}

#' @export
#' @rdname tmap_internal
tmapValuesVV_fill = function(x, value.na, isdiv, n, dvalues, are_breaks, midpoint, range, scale, rep, o, aes = "fill") {

	#palid = tmapPalId(x[1])
	if (x[1] %in% c("seq", "div", "unord", "ord", "biv")) {
		# cols4all can also take "div", but does not take into account tmap style
		x = getAesOption("values.var", o, aes = aes, layer = NA, cls = x[1])
	}

	m = if (length(x) > 1) NULL else getPalMeta(x[1])


	scale_ids = function(ids, n) {
		1 + ((ids - 1) / (n - 1)) * 100
	}

	map_ids = function(i, s, n) {
		di = i[2] - i[1]
		seq(i[1] + di * s[1], i[1] + di * s[2], length.out = n)
	}


	if (isdiv) {
		if (are_breaks) {
			# scale_interval
			outside = (all(midpoint < dvalues) || all(midpoint > dvalues))
			cat0 = !any(dvalues==midpoint) && !outside

			nneg = max(0L, sum(dvalues < midpoint) - cat0 - outside) # max 0L needed when midpoint is outside range (and cat0 is true)
			npos = max(0L, sum(dvalues > midpoint) - cat0 - outside)

			nmax = max(nneg, npos)

			ntot = 2L * nmax + cat0

			ids = (1L + max(0L, (npos-nneg))):(ntot - max(0L, (nneg-npos)))
		} else {
			# scale_discrete
			outside = (all(midpoint < dvalues) || all(midpoint > dvalues))
			cat0 = any(dvalues==midpoint)
			between = !outside && !cat0

			if (between) {
				# suppose 4 break values, and m is between 1st and 2nd:
				# x m x    x    x
				# create palette of half steps, and take odd numbers at the end
				nneg = max(0L, (sum(dvalues < midpoint) * 2 - 1L))
				npos = max(0L, (sum(dvalues > midpoint) * 2 - 1L))
				nmax = max(nneg, npos)
				ntot = 2L * nmax + 1L
				ids = (1L + max(0L, (npos-nneg))):(ntot - max(0L, (nneg-npos)))
				ids = ids[seq(1L, length(ids), by = 2L)]
			} else {
				nneg = max(0L, sum(dvalues < midpoint))
				npos = max(0L, sum(dvalues > midpoint))
				nmax = max(nneg, npos)
				ntot = 2L * nmax + cat0
				ids = (1L + max(0L, (npos-nneg))):(ntot - max(0L, (nneg-npos)))
			}

		}
	} else {
		ntot = n
		ids = 1L:n
	}

	if (!is.null(m)) {
		if (x[1] != tolower(x[1])) message_c4a(x[1], info = m)
		vvalues = getPal(m$fullname, n = ntot, range = range, reversed = m$reverse)[ids]
	} else {
		pal =colorRampPalette(x)
		if (is.na(range[1])) range = c(0, 1)
		if (range[1] != 0 || range[2] != 1) {
			ids_scaled = scale_ids(ids, ntot)
			if (isdiv) {
				ids_after_range = c({if (nneg > 0) head(map_ids(ids_scaled[c(1L, (nneg+cat0))],
																1-rev(range),
																n = nneg + cat0),
														nneg) else NULL},
									{if (cat0) ids_scaled[1L + nneg] else NULL},
									if (npos > 0) tail(map_ids(ids_scaled[c(nneg+1, n)],
															   range,
															   n = npos + cat0),
													   npos) else NULL)
			} else {
				ids_after_range = map_ids(ids_scaled[c(1L, ntot)], range, ntot)
			}
			vvalues = grDevices::colorRampPalette(x)(o$continuous.nclasses)[ids_after_range]
		} else {
			vvalues = grDevices::colorRampPalette(x)(ntot)[ids]
		}
	}


	if (isdiv) {
		if (cat0) {
			value.neutral = vvalues[1L + nneg]
		} else {
			value.neutral = vvalues[1L + nneg] # first positive
		}
	} else {
		value.neutral = vvalues[n/2]
	}
	vvalues = do.call(process_color, c(list(col = vvalues), o$pc))
	value.neutral = do.call(process_color, c(list(col = value.neutral), o$pc))
	value.na = do.call(process_color, c(list(col = value.na), o$pc))
	list(vvalues = vvalues, value.neutral = value.neutral, value.na = value.na)

}

#' @export
#' @rdname tmap_internal
tmapValuesVV_col = function(...) {
	tmapValuesVV_fill(..., aes = "col")
}

#' @export
#' @rdname tmap_internal
tmapValuesVV_bgcol = function(...) {
	tmapValuesVV_fill(..., aes = "bgcol")
}

#' @export
#' @rdname tmap_internal
tmapValuesVV_shape = function(x, value.na, isdiv, n, dvalues, are_breaks, midpoint, range, scale, rep, o) {
	if (inherits(x, "tmap_icons") && "iconUrl" %in% names(x)) x = split_icon(x)
	list(vvalues = rep(x, length.out = n), value.neutral = x[1], value.na = value.na)
}

#' @export
#' @rdname tmap_internal
tmapValuesVV_lty = function(x, value.na, isdiv, n, dvalues, are_breaks, midpoint, range, scale, rep, o) {
	list(vvalues = rep(x, length.out = n), value.neutral = x[1], value.na = value.na)
}


#' @export
#' @rdname tmap_internal
tmapValuesVV_size = function(x, value.na, isdiv, n, dvalues, are_breaks, midpoint, range, scale, rep, o) {
	vvalues = if (is.numeric(x) && length(x) == n) {
		if (range[1] !=0 || range[2] != 1) {
			warning("values.range not used because the individual values have been specified (instead of a sequence)", call. = FALSE)
		}
		x
	} else {
		if (is.numeric(x)) {
			x = tm_seq(x[1], x[length(x)], power = "lin")
		}

		if (range[1] !=0 || range[2] != 1) {
			p = if (is.numeric(x$power)) x$power else switch(x$power, lin = 1, sqrt = 0.5, sqrt_perceptual = 0.5716, quadratic = 2)
			x$from = range[1] ^ (1/p)
			x$to = range[2] ^ (1/p)
		}
		tmapSeq(x, n)

	}
	value.neutral = vvalues[round((n+1)/2)]
	list(vvalues = vvalues * scale, value.neutral = value.neutral * scale, value.na = value.na * scale)
}



#' @export
#' @rdname tmap_internal
tmapValuesVV_area = function(...) {
	do.call(tmapValuesVV_size, args = list(...))
}

#' @export
#' @rdname tmap_internal
tmapValuesVV_lwd = function(...) {
	do.call(tmapValuesVV_size, args = list(...))
}

#' @export
#' @rdname tmap_internal
tmapValuesVV_col_alpha = function(...) {
	do.call(tmapValuesVV_size, args = list(...))
}

#' @export
#' @rdname tmap_internal
tmapValuesVV_fill_alpha = function(...) {
	do.call(tmapValuesVV_size, args = list(...))
}

#' @export
#' @rdname tmap_internal
tmapValuesVV_bgcol_alpha = function(...) {
	do.call(tmapValuesVV_size, args = list(...))
}

#' @export
#' @rdname tmap_internal
tmapValuesVV_area = function(...) {
	do.call(tmapValuesVV_size, args = list(...))
}

#' @export
#' @rdname tmap_internal
tmapValuesVV_xmod = tmapValuesVV_ymod = function(...) {
	do.call(tmapValuesVV_size, args = list(...))
}


#' @export
#' @rdname tmap_internal
tmapValuesVV_angle = function(...) {
	do.call(tmapValuesVV_size, args = list(...))
}

#' @export
#' @rdname tmap_internal
tmapValuesVV_text = function(x, value.na, isdiv, n, dvalues, are_breaks, midpoint, range, scale, rep, o) {
	list(vvalues = rep(x, length.out = n), value.neutral = x[1], value.na = value.na)
}

#' @export
#' @rdname tmap_internal
tmapValuesVV_fontface = function(x, value.na, isdiv, n, dvalues, are_breaks, midpoint, range, scale, rep, o) {
	list(vvalues = rep(x, length.out = n), value.neutral = x[1], value.na = value.na)
}

# for symbols, which need to be 'submitted' (content replaced by an integer)
#' @export
#' @rdname tmap_internal
tmapValuesSubmit_col = function(x, args) x
#' @export
#' @rdname tmap_internal
tmapValuesSubmit_fill = function(x, args) x
#' @export
#' @rdname tmap_internal
tmapValuesSubmit_bgcol = function(x, args) x
#' @export
#' @rdname tmap_internal
tmapValuesSubmit_size = function(x, args) x
#' @export
#' @rdname tmap_internal
tmapValuesSubmit_area = function(x, args) x
#' @export
#' @rdname tmap_internal
tmapValuesSubmit_xmod = function(x, args) x
#' @export
#' @rdname tmap_internal
tmapValuesSubmit_ymod = function(x, args) x
#' @export
#' @rdname tmap_internal
tmapValuesSubmit_angle = function(x, args) x
#' @export
#' @rdname tmap_internal
tmapValuesSubmit_lwd = function(x, args) x
#' @export
#' @rdname tmap_internal
tmapValuesSubmit_lty = function(x, args) x
#' @export
#' @rdname tmap_internal
tmapValuesSubmit_shape = function(x, args) {
	if (all(is.na(x))) return(x)
	if (!inherits(x, c("tmapStandard", "tmapSpecial"))) x = tmapVV(x)
	if (inherits(x, "tmapStandard")) {
		xvec = unlist(x, use.names = FALSE)
		if (is.character(xvec) && all(isUrl(xvec) | isLocal(xvec))) {
			x = tmapVV(tmap_icons(xvec, merge = FALSE))
		}
	}

	if (inherits(x, "tmapSpecial")) {
		isGrid = (tmap_graphics_name() == "Grid")


		# symbols just specification:
		# copy-pasted from v3, but not the best place
		# improvement of just needed (-> trans?)
		args = within(args, {
			if (anyNA(just)) {
				just = c(.5, .5)
				just.override = FALSE
			} else {
				just = c(ifelse(is_num_string(just[1]), as.numeric(just[1]), ifelse(just[1]=="left", 1, ifelse(just[1]=="right", 0, .5))),
						  ifelse(is_num_string(just[2]), as.numeric(just[2]), ifelse(just[2]=="bottom", 1, ifelse(just[2]=="top", 0, .5))))
				just.override = TRUE
			}
		})

		do.call(submit_symbols, args = list(x = x, grid = isGrid, args = args))
	} else {
		unlist(x)
	}
}

#' @export
#' @rdname tmap_internal
tmapValuesSubmit_col_alpha = function(x, args) x
#' @export
#' @rdname tmap_internal
tmapValuesSubmit_fill_alpha = function(x, args) x
#' @export
#' @rdname tmap_internal
tmapValuesSubmit_bgcol_alpha = function(x, args) x
#' @export
#' @rdname tmap_internal
tmapValuesSubmit_text = function(x, args) x
#' @export
#' @rdname tmap_internal
tmapValuesSubmit_fontface = function(x, args) x


#' @export
#' @rdname tmap_internal
tmapValuesScale_col = function(x, scale) x
#' @export
#' @rdname tmap_internal
tmapValuesScale_fill = function(x, scale) x
#' @export
#' @rdname tmap_internal
tmapValuesScale_bgcol = function(x, scale) x
#' @export
#' @rdname tmap_internal
tmapValuesScale_size = function(x, scale) {
	# Convert units objects (from the units package) to plain numeric metres
	# so that e.g. units::as_units(1, "mi") works transparently with tm_circles.
	if (inherits(x, "units")) {
		x = as.numeric(units::set_units(x, "m", mode = "standard"))
	}
	x * scale
}
#' @export
#' @rdname tmap_internal
tmapValuesScale_area = function(x, scale) x
#' @export
#' @rdname tmap_internal
tmapValuesScale_lwd = function(x, scale) x * scale
#' @export
#' @rdname tmap_internal
tmapValuesScale_lty = function(x, scale) x
#' @export
#' @rdname tmap_internal
tmapValuesScale_shape = function(x, scale) x
#' @export
#' @rdname tmap_internal
tmapValuesScale_col_alpha = function(x, scale) x
#' @export
#' @rdname tmap_internal
tmapValuesScale_fill_alpha = function(x, scale) x
#' @export
#' @rdname tmap_internal
tmapValuesScale_bgcol_alpha = function(x, scale) x
#' @export
#' @rdname tmap_internal
tmapValuesScale_text = function(x, scale) x
#' @export
#' @rdname tmap_internal
tmapValuesScale_fontface = function(x, scale) x
#' @export
#' @rdname tmap_internal
tmapValuesScale_xmod = tmapValuesScale_ymod = function(x, scale) x
#' @export
#' @rdname tmap_internal
tmapValuesScale_angle = function(x, scale) x

#' @export
#' @rdname tmap_internal
tmapValuesColorize_col = function(x, pc) do.call(process_color, c(list(col = x), pc))
#' @export
#' @rdname tmap_internal
tmapValuesColorize_fill = function(x, pc) do.call(process_color, c(list(col = x), pc))
#' @export
#' @rdname tmap_internal
tmapValuesColorize_bgcol = function(x, pc) do.call(process_color, c(list(col = x), pc))
#' @export
#' @rdname tmap_internal
tmapValuesColorize_size = function(x, pc) x
#' @export
#' @rdname tmap_internal
tmapValuesColorize_area = function(x, pc) x
#' @export
#' @rdname tmap_internal
tmapValuesColorize_lwd = function(x, pc) x
#' @export
#' @rdname tmap_internal
tmapValuesColorize_lty = function(x, pc) x
#' @export
#' @rdname tmap_internal
tmapValuesColorize_shape = function(x, pc) x
#' @export
#' @rdname tmap_internal
tmapValuesColorize_col_alpha = function(x, pc) x
#' @export
#' @rdname tmap_internal
tmapValuesColorize_fill_alpha = function(x, pc) x
#' @export
#' @rdname tmap_internal
tmapValuesColorize_bgcol_alpha = function(x, pc) x
#' @export
#' @rdname tmap_internal
tmapValuesColorize_text = function(x, pc) x
#' @export
#' @rdname tmap_internal
tmapValuesColorize_fontface = function(x, pc) x
#' @export
#' @rdname tmap_internal
tmapValuesColorize_xmod = tmapValuesColorize_ymod = function(x, pc) x
#' @export
#' @rdname tmap_internal
tmapValuesColorize_angle = function(x, pc) x


#' Specify a numeric sequence
#'
#' Specify a numeric sequence, for numeric scales like [tm_scale_continuous()]. This function is needed when there is a non-linear relationship between the numeric data values and the visual variables. E.g. to make relationship with the area of bubbles linear, the square root of input variables should be used to calculate the radius of the bubbles.
#'
#' The perceived area of larger symbols is often underestimated. Flannery (1971) experimentally derived a method to compensate this for symbols. This compensation is obtained by using the power exponent of 0.5716 instead of 0.5, or by setting `power` to `"sqrt_perceptual"`
#'
#' @param from,to The numeric range, default 0 and 1 respectively
#' @param power The power component, a number or or one of `"lin"`, `"sqrt"`, `"sqrt_perceptual"`, `"quadratic"`, which correspond to 1, 0.5, 0.5716, 2 respectively. See details.
#' @export
tm_seq = function(from = 0, to = 1, power = c("lin", "sqrt", "sqrt_perceptual", "quadratic")) {
	if (length(power) > 1) power = power[1]
	structure(as.list(environment()), class = "tmapSeq")
}

# x is vector, rng is its range
norm_vector = function(x, rng) {
	(x - rng[1]) / diff(rng)
}

scale_vector = function(x, new_rng) {
	(x + new_rng[1]) * diff(new_rng)
}

#' @export
#' @rdname tmap_internal
tmapSeq = function(s, n = NULL) {
	if (is.null(n) && is.null(s$values)) stop("One of n or s$values should be provided")
	if (is.null(s$values)) s["values"] = list(NULL)
	with(s, {
		p = if (is.numeric(power)) power else switch(power, lin = 1, sqrt = 0.5, sqrt_perceptual = 0.5716, quadratic = 2)
		r = seq(from = from, to = to, length.out = n) ^ p
	})
}

#' @export
#' @rdname tmap_internal
transform_values = function(x, lim, rng, power, scale, include.neutral = TRUE) {
	p = if (is.numeric(power)) power else switch(power, lin = 1, sqrt = 0.5, sqrt_perceptual = 0.5716, quadratic = 2)
	if (p != 1) rng = rng ^ (1/p)

	x2 = norm_vector(x, lim)
	x3 = if (rng[1] != 0 || rng[2] != 1) scale_vector(x2, rng) else x2

	if (include.neutral) neutral = mean(rng)
	if (p != 1) {
		x3 = x3 ^ p
		if (include.neutral) neutral = neutral ^ p
	}
	x4 = x3 * scale
	if (include.neutral) neutral = neutral * scale

	if (include.neutral) {
		list(x = x4,
			 neutral = neutral)
	} else {
		x4
	}
}


#' @export
#' @rdname tmap_internal
tmapValuesCVV_fill = function(x, value.na, n, range, scale, rep, o) {


	# process values
	#palid = tmapPalId(x[1])

	arecolors = valid_colors(x[1])
	m = getPalMeta(x[1])

	ispalette = !is.null(m) && !arecolors # the latter in case of ambiguity (e.g. "blue")

	if (ispalette) if (x[1] != tolower(x[1])) message_c4a(x[1], info = m)


	values = if (!ispalette && !arecolors) {
		rep(x, length.out = n)
	} else if (ispalette) {
		getPal(m$fullname, n, rep = rep, range = range, reversed = m$reverse)
	} else if (!rep && (length(x) < n)) {
		grDevices::colorRampPalette(x)(n)
	} else {
		rep(x, length.out=n)
	}

	nms = names(values)
	values = do.call(process_color, c(list(col = values), o$pc))
	names(values) = nms
	value.neutral = do.call(process_color, c(list(col = values[1]), o$pc))
	value.na = do.call(process_color, c(list(col = value.na), o$pc))


	list(vvalues = values, value.neutral = value.neutral, value.na = value.na)

	# } else if (arenumbers) {
	# 	values = if (length(scale$values) == 2) seq(scale$values[1], scale$values[2], length.out = n) else rep(scale$values, length.out = n)
	# } else {
	# 	values = rep(scale$values, length.out = n)
	# }

}


#' @export
#' @rdname tmap_internal
tmapValuesCVV_col = function(x, value.na, n, range, scale, rep, o) {
	tmapValuesCVV_fill(x, value.na, n, range, scale, rep, o)
}

#' @export
#' @rdname tmap_internal
tmapValuesCVV_bgcol = function(x, value.na, n, range, scale, rep, o) {
	tmapValuesCVV_fill(x, value.na, n, range, scale, rep, o)
}

#' @export
#' @rdname tmap_internal
tmapValuesCVV_size = function(x, value.na, n, range, scale, rep, o) {
	tmapValuesVV_size(x = x, value.na = value.na, isdiv = FALSE, n = n, dvalues = NA, are_breaks = FALSE, midpoint = NA, range = range, scale = scale, rep = rep)
}

#' @export
#' @rdname tmap_internal
tmapValuesCVV_area = function(x, value.na, n, range, scale, rep, o) {
	tmapValuesVV_area(x = x, value.na = value.na, isdiv = FALSE, n = n, dvalues = NA, are_breaks = FALSE, midpoint = NA, range = range, scale = scale, rep = rep)
}

#' @export
#' @rdname tmap_internal
tmapValuesCVV_lwd = function(x, value.na, n, range, scale, rep, o) {
	tmapValuesVV_lwd(x = x, value.na = value.na, isdiv = FALSE, n = n, dvalues = NA, are_breaks = FALSE, midpoint = NA, range = range, scale = scale, rep = rep)
}

#' @export
#' @rdname tmap_internal
tmapValuesCVV_col_alpha = function(x, value.na, n, range, scale, rep, o) {
	tmapValuesVV_col_alpha(x = x, value.na = value.na, isdiv = FALSE, n = n, dvalues = NA, are_breaks = FALSE, midpoint = NA, range = range, scale = scale, rep = rep)
}

#' @export
#' @rdname tmap_internal
tmapValuesCVV_fill_alpha = function(x, value.na, n, range, scale, rep, o) {
	tmapValuesVV_fill_alpha(x = x, value.na = value.na, isdiv = FALSE, n = n, dvalues = NA, are_breaks = FALSE, midpoint = NA, range = range, scale = scale, rep = rep)
}

#' @export
#' @rdname tmap_internal
tmapValuesCVV_bgcol_alpha = function(x, value.na, n, range, scale, rep, o) {
	tmapValuesVV_bgcol_alpha(x = x, value.na = value.na, isdiv = FALSE, n = n, dvalues = NA, are_breaks = FALSE, midpoint = NA, range = range, scale = scale, rep = rep)
}

#' @export
#' @rdname tmap_internal
tmapValuesCVV_area = function(x, value.na, n, range, scale, rep, o) {
	tmapValuesVV_area(x = x, value.na = value.na, isdiv = FALSE, n = n, dvalues = NA, are_breaks = FALSE, midpoint = NA, range = range, scale = scale, rep = rep)
}

#' @export
#' @rdname tmap_internal
tmapValuesCVV_xmod = function(x, value.na, n, range, scale, rep, o) {
	tmapValuesVV_xmod(x = x, value.na = value.na, isdiv = FALSE, n = n, dvalues = NA, are_breaks = FALSE, midpoint = NA, range = range, scale = scale, rep = rep)
}

#' @export
#' @rdname tmap_internal
tmapValuesCVV_ymod = function(x, value.na, n, range, scale, rep, o) {
	tmapValuesVV_ymod(x = x, value.na = value.na, isdiv = FALSE, n = n, dvalues = NA, are_breaks = FALSE, midpoint = NA, range = range, scale = scale, rep = rep)
}

#' @export
#' @rdname tmap_internal
tmapValuesCVV_angle = function(x, value.na, n, range, scale, rep, o) {
	tmapValuesVV_angle(x = x, value.na = value.na, isdiv = FALSE, n = n, dvalues = NA, are_breaks = FALSE, midpoint = NA, range = range, scale = scale, rep = rep)
}

#' @export
#' @rdname tmap_internal
tmapValuesCVV_shape = function(x, value.na, n, range, scale, rep, o) {
	tmapValuesVV_shape(x = x, value.na = value.na, isdiv = FALSE, n = n, dvalues = NA, are_breaks = FALSE, midpoint = NA, range = range, scale = scale, rep = rep)
}

#' @export
#' @rdname tmap_internal
tmapValuesCVV_lty = function(x, value.na, n, range, scale, rep, o) {
	tmapValuesVV_lty(x = x, value.na = value.na, isdiv = FALSE, n = n, dvalues = NA, are_breaks = FALSE, midpoint = NA, range = range, scale = scale, rep = rep)
}

#' @export
#' @rdname tmap_internal
tmapValuesCVV_text = function(x, value.na, n, range, scale, rep, o) {
	tmapValuesVV_text(x = x, value.na = value.na, isdiv = FALSE, n = n, dvalues = NA, are_breaks = FALSE, midpoint = NA, range = range, scale = scale, rep = rep)
}

#' @export
#' @rdname tmap_internal
tmapValuesCVV_fontface = function(x, value.na, n, range, scale, rep, o) {
	tmapValuesVV_fontface(x = x, value.na = value.na, isdiv = FALSE, n = n, dvalues = NA, are_breaks = FALSE, midpoint = NA, range = range, scale = scale, rep = rep)
}

# bivariate visual variables
#' @export
#' @rdname tmap_internal
tmapValuesBVV_fill = function(x, value.na, m, n, scale, rep, o) {
	#palid = tmapPalId(x[1])

	meta = getPalMeta(x[1])

	if (!is.null(meta)) {
		values = getPalBiv(x, m = m, n = n)
	} else if (!is.matrix(x)) {
		values = matrix(grDevices::colorRampPalette(x)(m*n), nrow = m)
	} else if (is.matrix(x)) {
		if (ncol(x) != n) {
			x = do.call(rbind, apply(x, function(xi) {
				grDevices::colorRampPalette(xi)(n)
			}, MARGIN = 1, simplify = FALSE))
		}

		if (nrow(x) != m) {
			x = do.call(cbind, apply(x, function(xi) {
				grDevices::colorRampPalette(xi)(m)
			}, MARGIN = 2, simplify = FALSE))
		}
		values = x
	}

	values[] = do.call(process_color, c(list(col = values), o$pc))
	value.neutral = do.call(process_color, c(list(col = values[1]), o$pc))
	value.na = do.call(process_color, c(list(col = value.na), o$pc))


	list(vvalues = values, value.neutral = value.neutral, value.na = value.na)

}

#' @export
#' @rdname tmap_internal
tmapValuesBVV_col = function(x, value.na, m, n, scale, rep, o) {
	tmapValuesBVV_fill(x, value.na, m, n, scale, rep, o)
}

#' @export
#' @rdname tmap_internal
tmapValuesBVV_bgcol = function(x, value.na, m, n, scale, rep, o) {
	tmapValuesBVV_fill(x, value.na, m, n, scale, rep, o)
}


#' @export
#' @rdname tmap_internal
tmapValuesCheck_num = tmapValuesCheck_size
#' @export
#' @rdname tmap_internal
tmapValuesIsDiv_num = tmapValuesIsDiv_size
#' @export
#' @rdname tmap_internal
tmapValuesRange_num = tmapValuesRange_size
#' @export
#' @rdname tmap_internal
tmapValuesVV_num = tmapValuesVV_size
#' @export
#' @rdname tmap_internal
tmapValuesSubmit_num = tmapValuesSubmit_size
#' @export
#' @rdname tmap_internal
tmapValuesScale_num = tmapValuesScale_size
#' @export
#' @rdname tmap_internal
tmapValuesColorize_num = tmapValuesColorize_size
#' @export
#' @rdname tmap_internal
tmapValuesCVV_num = tmapValuesCVV_size

#' @export
#' @rdname tmap_internal
tmapValuesCheck_skip = function(x, is_var) TRUE
#' @export
#' @rdname tmap_internal
tmapValuesIsDiv_skip = function(x) FALSE
#' @export
#' @rdname tmap_internal
tmapValuesRange_skip = function(x, n, isdiv) c(0, 1)
#' @export
#' @rdname tmap_internal
tmapValuesVV_skip = function(x, value.na, isdiv, n, dvalues, are_breaks, midpoint, range, scale, rep, o) {
	list(vvalues = rep(x, length.out = n), value.neutral = x[1], value.na = value.na)
}
#' @export
#' @rdname tmap_internal
tmapValuesSubmit_skip = function(x, args) x
#' @export
#' @rdname tmap_internal
tmapValuesScale_skip = function(x, scale) x
#' @export
#' @rdname tmap_internal
tmapValuesColorize_skip = function(x, pc) x
#' @export
#' @rdname tmap_internal
tmapValuesCVV_skip = function(x, value.na, n, range, scale, rep, o) {
	tmapValuesVV_shape(x = x, value.na = value.na, isdiv = FALSE, n = n, dvalues = NA, are_breaks = FALSE, midpoint = NA, range = range, scale = scale, rep = rep)
}

#' Legend charts
#'
#' Legend charts are small charts that are added to the map, usually in addition to legends.
#'
#' Note that these charts are different from charts drawn inside the map. Those are called glyphs (to be implemented).
#'
#' @param breaks The breaks of the bins (for histograms)
#' @param plot.axis.x,plot.axis.y Should the x axis and y axis be plot?
#' @param extra.ggplot2 Extra ggplot2 code
#' @inheritParams tm_title
#' @example examples/tm_chart.R
#' @seealso \href{https://r-tmap.github.io/tmap/articles/basics_charts}{Vignette about charts}
#' @name tm_chart
#' @export
tm_chart_histogram = function(breaks,
							  plot.axis.x,
							  plot.axis.y,
							  extra.ggplot2,
							  position,
							  group_id,
							  width,
							  height,
							  stack,
							  z,
							  ...) {

	args = lapply(as.list(rlang::call_match()[-1]), eval, envir = parent.frame())
	args = warning_group_args(args)

	args$z = args$z %||% NA_integer_
	args$group_id = args$group_id %||% NA_character_
	args$group_type = "tm_chart"

	args$show = TRUE
	args$type = "histogram"
	args$summary = "binned"
	structure(args, class = c("tm_chart_histogram", "tm_chart", "tm_component", "list"))
}

#' @rdname tm_chart
#' @export
tm_chart_bar = function(plot.axis.x,
						plot.axis.y,
						extra.ggplot2,
						position,
						group_id,
						width,
						height,
						stack,
						z,
						...) {
	args = lapply(as.list(rlang::call_match()[-1]), eval, envir = parent.frame())
	args = warning_group_args(args)

	args$z = args$z %||% NA_integer_
	args$group_id = args$group_id %||% NA_character_
	args$group_type = "tm_chart"

	args$show = TRUE
	args$type = "bar"
	args$summary = "binned"
	structure(args, class = c("tm_chart_bar", "tm_chart", "tm_component", "list"))
}


#' @rdname tm_chart
#' @export
tm_chart_donut = function(position,
						  group_id,
						  width,
						  height,
						  stack,
						  z,
						  ...) {
	args = lapply(as.list(rlang::call_match()[-1]), eval, envir = parent.frame())
	args = warning_group_args(args)


	args$z = args$z %||% NA_integer_
	args$group_id = args$group_id %||% NA_character_
	args$group_type = "tm_chart"

	args$show = TRUE
	args$type = "donut"
	args$summary = "binned"
	structure(args, class = c("tm_chart_donut", "tm_chart", "tm_component", "list"))
}

#' @rdname tm_chart
#' @export
tm_chart_violin = function(position,
						   group_id,
						   width,
						   height,
						   stack,
						   z,
						   ...) {
	args = lapply(as.list(rlang::call_match()[-1]), eval, envir = parent.frame())
	args = warning_group_args(args)

	args$z = args$z %||% NA_integer_
	args$group_id = args$group_id %||% NA_character_
	args$group_type = "tm_chart"

	args$show = TRUE
	args$type = "violin"
	args$summary = "raw_nna"
	structure(args, class = c("tm_chart_violin", "tm_chart", "tm_component", "list"))
}

#' @rdname tm_chart
#' @export
tm_chart_box = function(position,
						group_id,
					    width,
					    height,
					    stack,
					    z,
						...) {
	args = lapply(as.list(rlang::call_match()[-1]), eval, envir = parent.frame())
	args = warning_group_args(args)

	args$z = args$z %||% NA_integer_
	args$group_id = args$group_id %||% NA_character_
	args$group_type = "tm_chart"

	args$show = TRUE
	args$type = "box"
	args$summary = "raw_nna"
	structure(args, class = c("tm_chart_box", "tm_chart", "tm_component", "list"))
}

#' @rdname tm_chart
#' @export
tm_chart_none = function() {
	structure(list(show = FALSE, group_id = NA_character_, z = NA_integer_, summary = "none"), class = c("tm_chart_none", "tm_chart", "tm_component", "list"))
}



#' @rdname tm_chart
#' @export
tm_chart_heatmap = function(position,
							group_id,
							width,
							height,
							stack,
							z,
							...){
	args = lapply(as.list(rlang::call_match()[-1]), eval, envir = parent.frame())
	args = warning_group_args(args)

	args$z = args$z %||% NA_integer_
	args$group_id = args$group_id %||% NA_character_
	args$group_type = "tm_chart"

	args$show = TRUE
	args$type = "heatmap"
	args$summary = "binned2D"
	structure(args, class = c("tm_chart_heatmap", "tm_chart", "tm_component", "list"))
}

# Cross-mode basemap equivalence registry
# ----------------------------------------
# Different modes expose different basemap providers (see tmap_providers()).
# When a map made in one mode is reproduced in another, a provider that is valid
# in the first mode may not exist in the second. This registry maps providers to
# a small set of visual "concepts" (gray canvas, dark, satellite, ...) so the
# closest available provider in the target mode can be substituted, rather than
# falling back to the generic default. The mappings are necessarily approximate.
#
# Mode extensions (e.g. tmap.mapgl) declare their equivalences with
# tmapSubmitBasemapEquivalents(). Each concept is a named list with mode names
# (e.g. "plot", "view", "maplibre", "mapbox") as names and provider name(s) as
# values. When several providers are listed for a mode, the first is the
# representative used when substituting into that mode; all of them are matched
# when identifying the concept of an incoming provider.
#
# The registry lives in the package state environment as .TMAP$basemap_equiv (a
# named list of concepts). It is populated once, from the extension's .onLoad.

#' Register cross-mode basemap equivalences
#'
#' Used by tmap mode extensions to declare which basemap providers are visually
#' equivalent across modes. This lets a basemap be approximated when a map is
#' reproduced in a different mode, instead of falling back to the mode default.
#'
#' @param equivalents a named list of concept entries. Each entry is itself a
#'   named list with mode names (e.g. `plot`, `view`, `maplibre`, `mapbox`) as
#'   names and provider name(s) as values.
#' @param replace if `TRUE`, replace the whole registry; otherwise the entries
#'   are merged by concept name (existing names are overwritten), so repeated
#'   submissions are idempotent and multiple extensions compose.
#' @export
#' @keywords internal
tmapSubmitBasemapEquivalents = function(equivalents, replace = FALSE) {
	if (replace || is.null(.TMAP$basemap_equiv)) {
		.TMAP$basemap_equiv = equivalents
	} else {
		.TMAP$basemap_equiv = utils::modifyList(.TMAP$basemap_equiv, equivalents)
	}
	invisible(NULL)
}

# Closest equivalent of `serv` in `to_mode`, or NA_character_ if `serv` is not
# known to the registry. `serv` may be a provider from any mode.
basemap_equivalent = function(serv, to_mode) {
	concepts = .TMAP$basemap_equiv
	if (is.null(concepts)) return(NA_character_)
	for (concept in concepts) {
		if (serv %in% unlist(concept, use.names = FALSE)) {
			q = concept[[to_mode]]
			if (!is.null(q) && length(q)) return(q[[1]])
		}
	}
	NA_character_
}

# Gentle message used when a basemap is substituted by its cross-mode
# equivalent (as opposed to a genuinely unknown provider, which warns).
message_basemaps_equivalent = function(serv, mode, eq) {
	cli::cli_inform(c(
		"i" = "{.field [basemaps]} Provider {.str {serv}} is not available in mode {.str {mode}}; using the closest equivalent {.str {eq}}."
	))
}

tmapGridRun = function(o, q, show, knit, knit_opts, args) {
	gts = get("gts", .TMAP_GRID)

	if (show) {
		if (o$show_gif_ani) {
			d <- paste(tempdir(), "/tmap_plots", sep="/")
			dir.create(d, showWarnings = FALSE)
			devsize = c(grid::convertWidth(grid::unit(1, "npc"), unitTo = "inch", valueOnly = TRUE) * 72 * o$dpr,
						grid::convertHeight(grid::unit(1, "npc"), unitTo = "inch", valueOnly = TRUE) * 72 * o$dpr)
		}

		mapply(function(gt,i) {
			if (o$show_gif_ani) {
				png(paste0(d, "/plot", sprintf("%03d", i), ".png"), width = devsize[1], height = devsize[2], res = 72*o$dpr, type = "cairo-png")
				grid::grid.newpage()
			} else {
				if (is.null(o$vp) && i != 1L) grid::grid.newpage()
			}

			tryCatch({
				grid::grid.draw(gt)
			}, error = function(e) {
				stop("Plot error. Try adding + tm_check_fix()", call. = FALSE)
			})
			if (o$show_gif_ani) {
				dev.off()
			}

		}, gts, seq_along(gts), SIMPLIFY = FALSE)
		if (!is.null(o$vp)) upViewport(1)
	}
	if (length(gts) == 1) gts = gts[[1]]

	if (show && o$show_gif_ani) {
		files = list.files(path = d, pattern = "^plot[0-9]{3}\\.png$", full.names = TRUE)
		if (o$play == "pingpong") files = c(files, rev(files))
		filename = tempfile(fileext = ".gif")
		create_animation(filename = filename, files = files, width = devsize[1], height = devsize[2], delay = NA, fps = o$fps, loop = o$play != "once", progress = FALSE, gif = TRUE, showAni = TRUE, dpr = o$dpr, knit = knit, knit_opts = knit_opts)
		if (knit && !knitr::is_latex_output()) {
			knitr_path <- knitr::fig_path('.gif', knit_opts)
			dir.create(dirname(knitr_path), showWarnings = FALSE, recursive = TRUE)
			file.copy(filename, knitr_path, overwrite = TRUE)
			if (is.null(knit_opts$out.width)) {
				knit_opts$out.width <- knit_opts$fig.width * knit_opts$dpi / (knit_opts$fig.retina %||% 1)
			}
			return(do.call(knitr::knit_print, c(list(x = knitr::include_graphics(knitr_path), options = knit_opts), args)))
		}
	}

	gts
}

get_panel_grob = function(label, o, g, rot, row, col) {
	panel.show = o$panel.label.frame || o$panel.label.bg
	frame.col = if (o$panel.label.frame) o$panel.label.frame.color else NA
	bg.col = if (o$panel.label.bg && !getOption("tmap.design.mode")) o$panel.label.bg.color else NA
	if (panel.show) {
		if (o$panel.label.bg.alpha != o$panel.label.frame.alpha) {
			gpar_panel = grid::gpar(fill = bg.col, lwd=0, col = NA, alpha = o$panel.label.bg.alpha)
			gpar_panel_frame = grid::gpar(fill = NA, lwd=o$panel.label.frame.lwd * o$scale, col = frame.col, alpha = o$panel.label.frame.alpha)
		} else {
			gpar_panel = grid::gpar(fill = bg.col, lwd=o$panel.label.frame.lwd * o$scale, col = frame.col, alpha = o$panel.label.bg.alpha)
			gpar_panel_frame = NULL
		}
	}
	gpar_text = rescale_gp(grid::gpar(cex = o$panel.label.size * o$scale, col = o$panel.label.color, fontfamily = o$panel.label.fontfamily, fontface = o$panel.label.fontface, alpha = o$panel.label.alpha), o$scale_down)

	# resize due to not fitting
	gpar_text$cex = determine_scale(label = label, rot = rot, row = row, col = col, g = g, scale = gpar_text$cex)

	if (!is.na(label)) label = decode_expr(label)


	grid::grobTree(
		if (panel.show) {
			rndrectGrob(gp = gpar_panel, r = o$panel.label.frame.r * o$scale)
		} else NULL,
		if (panel.show && !is.null(gpar_panel_frame)) {
			rndrectGrob(gp = gpar_panel_frame, r = o$panel.label.frame.r * o$scale)
		} else NULL,
		grid::textGrob(label = label, rot = rot, gp = gpar_text)
	)
}

tmapGridWrap = function(label, facet_row, facet_col, facet_page, o) {
	gts = get("gts", envir = .TMAP_GRID)
	g = get("g", envir = .TMAP_GRID)

	gt = gts[[facet_page]]

	rot = g$panel_rot
	row = g$rows_panel_ids[facet_row]
	col = g$cols_panel_ids[facet_col]

	grb = get_panel_grob(label, o, g, rot, row, col)

	gt = add_to_gt(gt, grb, row = row, col = col)

	gts[[facet_page]] = gt

	assign("gts", gts, envir = .TMAP_GRID)


}

tmapGridXtab = function(label, facet_row = NULL, facet_col = NULL, facet_page, o) {
	gts = get("gts", envir = .TMAP_GRID)
	g = get("g", envir = .TMAP_GRID)

	gt = gts[[facet_page]]

	if (is.null(facet_row)) {
		rot = g$panel_col_rot
		row = g$cols_panel_row_id
		col = g$cols_panel_col_ids[facet_col]
	} else {
		rot = g$panel_row_rot
		row = g$rows_panel_row_ids[facet_row]
		col = g$rows_panel_col_id
	}

	grb = get_panel_grob(label, o, g, rot, row, col)

	gt = add_to_gt(gt, grb, row = row, col = col)

	gts[[facet_page]] = gt

	assign("gts", gts, envir = .TMAP_GRID)


}

tmapScaleDiscrete = function(x1, scale, legend, chart, o, aes, layer, layer_args, sortRev, bypass_ord, submit_legend = TRUE) {
	cls = data_class(x1, midpoint_enabled = !is.null(scale$midpoint))
	maincls = class(scale)[1]


	#if (cls[1] == "na") stop("data contain only NAs, so tm_scale_discrete cannot be applied", call. = FALSE)
	#if (cls[1] != "num") stop("tm_scale_discrete can only be used for numeric data", call. = FALSE)

	if (cls[1] != "num") {
		if (!is.factor(x1)) x1 = as.factor(x1)
		x1 = as.integer(x1)
		warning(maincls, " is supposed to be applied to numerical data", call. = FALSE)
	}

	x1 = without_units(x1)

	if (aes %in% c("lty", "shape", "pattern")) stop("tm_scale_discrete cannot be used for layer ", layer, ", aesthetic ", aes, call. = FALSE)

	scale = get_scale_defaults(scale, o, aes, layer, cls)

	show.messages <- o$show.messages
	show.warnings <- o$show.warnings

	if (all(is.na(x1))) return(tmapScale_returnNA(n = length(x1), legend = legend, chart = chart, value.na = scale$value.na, label.na = scale$label.na, label.show = scale$label.show, na.show = legend$na.show, sortRev = sortRev, bypass_ord = bypass_ord))

	with(scale, {

		u = sort(unique(x1))
		#u = na.omit(u)
		rng = range(u)

		if (!is.na(ticks[1])) {
			if (!all(u %in% ticks)) stop("Values have been found for which no ticks have been specified", call. = FALSE)
		} else {
			for (n in c(1, 2, 3, 5, 10, 15, 20, 30, 50)) {
				ticks_candidate = pretty(rng, n = n)

				found_ticks = all(u %in% ticks_candidate)
				if (found_ticks) {
					ticks = ticks_candidate
					break
				}
			}
			if (!found_ticks) stop("Unable to fit a discrete scale, probably due to too many values", call. = FALSE)
		}

		n = length(ticks)

		d_isdiv = rng[1] < 0 && rng[2] > 0

		check_values(layer, aes, values)

		fun_isdiv = paste0("tmapValuesIsDiv_", aes)

		isdiv = !is.null(midpoint) || do.call(fun_isdiv, args = list(x = values))


		# determine midpoint
		if ((is.null(midpoint) || is.na(midpoint)) && isdiv) {
			if (d_isdiv && is.null(midpoint)) {

				if (show.messages) message("Variable(s) \"", paste(aes, collapse = "\", \""), "\" contains positive and negative values, so midpoint is set to 0. Set midpoint = NA to show the full spectrum of the color palette.")
				midpoint = 0
			} else {
				if ((n %% 2) == 1) {
					# number of classes is odd, so take middle class (average of those breaks)
					midpoint <- mean.default(ticks[c((n+1) / 2, (n+3) / 2)])
				} else {
					midpoint <- ticks[(n+2) / 2]
				}
			}
		}

		# update range if NA (automatic)
		if (is.na(values.range[1])) {
			fun_range = paste0("tmapValuesRange_", aes)
			values.range = do.call(fun_range, args = list(x = values, n = n, isdiv = isdiv))
		}
		if (length(values.range) == 1 && !is.na(values.range[1])) values.range = c(0, values.range)


		fun_getVV = paste0("tmapValuesVV_", aes)
		VV = do.call(fun_getVV, list(x = values, value.na = value.na, isdiv = isdiv, n = n, dvalues = ticks, are_breaks = FALSE, midpoint = midpoint, range = values.range, scale = values.scale * o$scale, rep = values.repeat, o = o))

		vvalues = VV$vvalues
		value.na = VV$value.na

		sfun = paste0("tmapValuesScale_", aes)
		cfun = paste0("tmapValuesColorize_", aes)
		if (is.na(value.neutral)) value.neutral = VV$value.neutral else value.neutral = do.call(sfun, list(x = do.call(cfun, list(x = value.neutral, pc = o$pc)), scale = values.scale))

		mfun = paste0("tmapValuesSubmit_", aes)
		vvalues = do.call(mfun, list(x = vvalues, args = layer_args))
		value.na = do.call(mfun, list(x = value.na, args = layer_args))
		value.neutral = do.call(mfun, list(x = value.neutral, args = layer_args))


		ids = match(x1, ticks)
		vals = vvalues[ids]

		isna = is.na(vals)
		anyNA = any(isna)

		na.show = update_na.show(label.show, legend$na.show, anyNA)

		if (is.null(sortRev)) {
			ids = NULL
		} else if (is.na(sortRev)) {
			ids[] = 1L
		} else if (sortRev) {
			ids = (as.integer(n) + 1L) - ids
		}

		if (anyNA) {
			vals[isna] = value.na
			if (!is.null(sortRev)) ids[isna] = 0L
		}


		if (is.null(labels)) {
			labels = do.call("fancy_breaks", c(list(vec=ticks, as.count = FALSE, intervals=FALSE), label.format))
		} else {
			if (length(labels)!=n && show.warnings) warning("number of legend labels should be ", n, call. = FALSE)
			labels = rep(labels, length.out=n)
			attr(labels, "align") <- label.format$text.align
		}

		if (legend$reverse) {
			#labels.brks <- attr(labels, "brks")
			labels.align <- attr(labels, "align")
			labels <- rev(labels)
			# if (!is.null(labels.brks)) {
			# 	attr(labels, "brks") = labels.brks[length(labels):1L,]
			# }
			attr(labels, "align") = labels.align
			vvalues = rev(vvalues)
		}



		if (na.show) {
			#labels.brks = attr(labels, "brks")
			labels.align = attr(labels, "align")
			#labels <- c(labels, legend.NA.text)
			# if (!is.null(labels.brks)) {
			# 	labels <- c(labels, paste(label.na, " ", sep = ""))
			# 	attr(labels, "brks") = rbind(labels.brks, rep(nchar(label.na) + 2, 2))
			# } else {
			labels = c(labels, label.na)

			attr(labels, "align") = labels.align
			vvalues = c(vvalues, value.na)
		}

		# temp solution to include NA, to do update num2pal
		# if (length(labels) == length(vvalues)) {
		# 	la = attributes(labels)
		# 	labels = c(labels, label.na)
		# 	attributes(labels) = la
		# 	values = c(values, NA)
		# }
		#

		#type = ifelse(nchar(vvalues[1]) > 50, "color_cont", "color_cls")

		legend = within(legend, {
			nitems = length(labels)
			labels = labels
			dvalues = ticks
			vvalues = vvalues
			vneutral = value.neutral
			na.show = get("na.show", envir = parent.env(environment()))
			scale = "discrete"
			layer_args = layer_args
		})

		chartFun = paste0("tmapChart", toTitleCase(chart$summary))

		chart = do.call(chartFun, list(chart,
									   bin_colors = vvalues,
									   breaks_def = NULL,
									   na.show = na.show,
									   x1 = x1))



		if (submit_legend) {
			if (bypass_ord) {
				format_aes_results(vals, legend = legend, chart = chart)
			} else {
				format_aes_results(vals, ids, legend, chart = chart)
			}
		} else {
			list(vals = vals, ids = ids, legend = legend, chart = chart, bypass_ord = bypass_ord)
		}
	})
}

downsample_stars = function(x, max.raster) {
	k = length(dim(x))
	xy_dim = get_xy_dim(x)
	
	fact = round(sqrt(prod(xy_dim) / max.raster) - 1)
	
	n = dim(x)
	n[] = 0L
	n[names(xy_dim)] = fact
	
	if (inherits(x, "stars_proxy") || (fact > 0)) {
		y = stars::st_downsample(x, n)
		message("stars object downsampled to ", paste(get_xy_dim(y), collapse = " by "), " cells.")
	} else {
		y = x
	}
	y
}

downsample_SpatRaster = function(x, max.raster) {
	xy_dim = dim(x)[1:2]
	
	downsample = prod(xy_dim) > max.raster
	
	y = if (downsample) {
		terra::spatSample(x, max.raster, method="regular", as.raster=TRUE)
	} else x
	
	
	if (downsample) message("SpatRaster object downsampled to ", paste(dim(y)[1:2], collapse = " by "), " cells.")
	
	y
}

tmapGridCompPrepare_text = function(comp, o) {
	if (is.null(comp$text)) {
		cmp = within(comp, {
			show = FALSE
		})
		return(cmp)
	}
	n = length(comp$text)
	lst = lapply(seq_len(n), function(i){
		within(comp, {
			color[is.na(color)] = o$attr.color
			color = do.call("process_color", c(list(col=color), o$pc))
			size = size * o$scale
			if (!is.numeric(fontface)) fontface[is.na(fontface)] = o$text.fontface
			fontfamily[is.na(fontfamily)] =o$text.fontfamily
			#text = lapply(text, rep, length.out=o$n)
			text = text[i]
			if (is.na(text)) text = ""
			show = nonempty_text(text)
		})
	})
	class(lst) = c("tm_multi_comp", "list")
	lst
}


#' @export
tmapGridCompPrepare.tm_title = function(comp, o) {
	tmapGridCompPrepare_text(comp, o)
}

#' @export
tmapGridCompHeight.tm_title = function(comp, o) {
	tmapGridCompHeight_text(comp, o)
}

#' @export
tmapGridCompWidth.tm_title = function(comp, o) {
	tmapGridCompWidth_text(comp, o)
}

#' @export
tmapGridCompPlot.tm_title = function(comp, o, fH, fW) {
	tmapGridCompPlot_text(comp, o)
}

#' @export
tmapGridCompPrepare.tm_credits = function(comp, o) {
	tmapGridCompPrepare_text(comp, o)
}

#' @export
tmapGridCompHeight.tm_credits = function(comp, o) {
	tmapGridCompHeight_text(comp, o)
}

#' @export
tmapGridCompWidth.tm_credits = function(comp, o) {
	tmapGridCompWidth_text(comp, o)
}

#' @export
tmapGridCompPlot.tm_credits = function(comp, o, fH, fW) {
	tmapGridCompPlot_text(comp, o, fH, fW)
}

#' @export
tmapGridCompPrepare.tm_compass = function(comp, o) {
	typs = c("arrow", "4star", "8star", "radar", "rose")
	within(comp, {

		if (!(type %in% typs)) cli::cli_abort("{.field [tm_compass]} wrong compass type, should be one of: {.str {typs}}")
		text.size = text.size * o$scale

		asp = if (type == "arrow") 0.5 else 1

		show = TRUE
		if (is.na(type)) type = o$type
		if (is.na(size)) size = switch(type, arrow=2, radar=6, rose=6, 4)
		nlines = size + ifelse(show.labels==0, 0, ifelse(show.labels==1, 1, 2))
	})
}

#' @export
tmapGridCompHeight.tm_compass = function(comp, o) {


	textS = comp$text.size #* o$scale
	#textP = comp$padding[c(3,1)] * textS * o$lin

	marH = comp$margins[c(3,1)] * textS * o$lin
	hs = c(marH[1], comp$nlines * textS * o$lin, marH[2])


	sides = switch(comp$position$align.v, top = "second", bottom = "first", "both")
	hsu = set_unit_with_stretch(hs, sides = sides)

	Hin = sum(hs)
	comp$flexRow = NA
	comp$Hin = Hin #  sum(textP[1], textH, textP[2])
	comp$hsu = hsu
	comp
}

#' @export
tmapGridCompWidth.tm_compass = function(comp, o) {



	textS = comp$text.size #* o$scale
	#textP = comp$padding[c(3,1)] * textS * o$lin


	marW = comp$margins[c(2,4)] * textS * o$lin
	ws = c(marW[1], comp$nlines * textS * o$lin * comp$asp, marW[2])

	sides = switch(comp$position$align.h, left = "second", right = "first", "both")
	wsu = set_unit_with_stretch(ws, sides = sides)

	comp$flexCol = NA
	comp$Win = sum(ws)
	comp$wsu = wsu
	comp
}

#' @export
tmapGridCompPlot.tm_compass = function(comp, o, fH, fW) {

	u = 1/(comp$nlines)
	#vpComp = viewport(x=u, y=u, height=1-2*u, width=1-2*u, just=c("left", "bottom"))

	light = comp$color.light
	dark = comp$color.dark


	wsu = comp$wsu
	hsu = comp$hsu

	vp = grid::viewport(layout = grid::grid.layout(ncol = length(wsu),
												   nrow = length(hsu),
												   widths = wsu,
												   heights = hsu))


	if (comp$type=="4star") {
		s = c(.5, .5, .57, .5, .5, .43, 0, .5, .43, 1, .5, .57)
		x = list(rep.int(s, 2))
		y = list(s[c(10:12, 10:12, 1:3, 1:3, 7:9, 7:9, 4:6, 4:6)])
		id = rep(1:8, each=3)
		fill = c(dark, light, dark, light, light, dark, light, dark)
	} else if (comp$type=="8star") {
		s = c(.5, .5, .56, .5, .5, .44, 0, .5, .38, 1, .5, .62)
		s2 = c(.5, .62, .7, .5, .56, .7, .5, .38, .3, .5, .44, .3)
		x = list(c(rep.int(s, 2), rep.int(s2, 2)))
		y = list(c(s[c(10:12, 10:12, 1:3, 1:3, 7:9, 7:9, 4:6, 4:6)], s2[c(4:6, 1:3, 10:12, 7:9, 10:12, 7:9, 4:6, 1:3)]))
		id = rep(1:16, each=3)
		fill = c(dark, light, dark, light, light, dark, light, dark)
	} else if (comp$type=="arrow") {
		x = list(c(.5, .9, .5, .5, .1, .5))
		y = list(c(1, 0, .2, 1, 0, .2))
		id = rep(1:2, each=3)
		fill = c(dark, light)
	} else if (comp$type=="radar") {
		cr = c(.45, .42, .2, .17, .1)
		LWD = (o$lineH * 24) * comp$lwd

		cd = seq(1/8, 15/8, by=.25) * pi
		cd2 = seq(1/4, 7/4, by=.5) * pi
		cd3 = seq(0, 1.75, by=.25) * pi

		x = list(.5,
				  unlist(lapply(.5 + sin(cd) * cr[1], c, .5), use.names = FALSE),
				  .5 + c(0, cr[1]-.005, 0, -cr[1]+.005, 0, 0, 0, 0),
				  unlist(lapply(.5 + sin(cd2) * cr[1], c, .5), use.names = FALSE),
				  .5 + unlist(mapply(c, sin(cd3) * cr[4], sin(cd3) * cr[5], SIMPLIFY=FALSE), use.names = FALSE))

		y = list(.5,
				  unlist(lapply(.5 + cos(cd) * cr[1], c, .5), use.names = FALSE),
				  .5 + c(0, 0, 0, 0, 0, cr[1]-.005, 0, -cr[1]+.005),
				  unlist(lapply(.5 + cos(cd2) * cr[1], c, .5), use.names = FALSE),
				  .5 + unlist(mapply(c, cos(cd3) * cr[4], cos(cd3) * cr[5], SIMPLIFY=FALSE), use.names = FALSE))

	} else if (comp$type=="rose") {
		cr = c(.45, .42, .2, .17, .1)
		LWD = (o$lineH * 24) * comp$lwd
		cd = seq(1/8, 15/8, by=.25) * pi
		cd2 = seq(1/4, 7/4, by=.5) * pi
		cd3 = seq(0, 1.75, by=.25) * pi

		b = cr[4]
		a = 0.4142136 * b # 1/16th circleL
		s = c(.5, .5, .5+a, .5, .5, .5-a, 0, .5, .5-b, 1, .5, .5+b)
		s2 = c(.5, .5+b, .78, .5, .5+a, .78, .5, .5-b, .22, .5, .5-a, .22)

		id = rep(1:16, each=3)
		fill = c(dark, light, dark, light, light, dark, light, dark)


		x = list(.5,
				  unlist(lapply(.5 + sin(cd) * cr[1], c, .5), use.names = FALSE),
				  .5 + unlist(mapply(c, sin(cd3) * cr[4], sin(cd3) * cr[5], SIMPLIFY=FALSE), use.names = FALSE),
				  c(rep.int(s, 2), rep.int(s2, 2)))

		y = list(.5,
				  unlist(lapply(.5 + cos(cd) * cr[1], c, .5), use.names = FALSE),
				  .5 + unlist(mapply(c, cos(cd3) * cr[4], cos(cd3) * cr[5], SIMPLIFY=FALSE), use.names = FALSE),
				  c(s[c(10:12, 10:12, 1:3, 1:3, 7:9, 7:9, 4:6, 4:6)], s2[c(4:6, 1:3, 10:12, 7:9, 10:12, 7:9, 4:6, 1:3)]))

	}


	# rescale
	resc = function(a) (a-.5)*(comp$size/comp$nlines) + .5

	x = lapply(x, resc)
	y = lapply(y, resc)
	if (comp$type %in% c("radar", "rose")) cr = cr * (comp$size/comp$nlines)

	if (comp$north!=0) {
		drotate = comp$north/180*pi - .5*pi

		xy = mapply(function(a,b){
			d = atan2(b-.5, a-.5)
			r = sqrt((a-.5)^2 + (b-.5)^2)

			list(x=r * sin(d+drotate) + .5,
				 y=r * cos(d+drotate) + .5)
		}, x, y, SIMPLIFY=FALSE)
		x = lapply(xy, "[[", 1)
		y = lapply(xy, "[[", 2)
	} else drotate = -.5*pi


	# shift compass to south direction
	if (comp$show.labels==1) {
		x = lapply(x, function(a) a - (u/2) * sin(drotate + .5*pi))
		y = lapply(y, function(b) b - (u/2) * cos(drotate + .5*pi))
	}



	grobBG = if (getOption("tmap.design.mode")) rectGrob(gp=gpar(fill="#CAB2D6")) else NULL

	grobLabels = if (comp$show.labels==0) {
		NULL
	} else {
		selection = if (comp$show.labels==1) {
			c(TRUE, rep.int(FALSE, 7))
		} else if (comp$show.labels==2) {
			rep.int(c(TRUE, FALSE), 4)
		} else rep.int(TRUE, 8)

		labels = comp$cardinal.directions[c(1, 1, 2, 3, 3, 3, 4, 1)]
		labels[c(2,4,6,8)] = paste(labels[c(2,4,6,8)], comp$cardinal.directions[c(2, 2, 4, 4)], sep="")
		labels = labels[selection]

		lr = (1-u)/2
		ld = (seq(0, 1.75, by=.25) * pi)[selection]

		lx = lr * sin(ld+drotate + .5*pi) + .5
		ly = lr * cos(ld+drotate + .5*pi) + .5
		textGrob(labels, x=lx, y=ly, just=c("center", "center"), rot=0, gp=gpar(col=comp$text.color, cex=comp$text.size, fontface=o$text.fontface, fontfamily=o$text.fontfamily)) # -drotate/pi*180 - 90
	}

	grobComp = if (comp$type %in% c("arrow", "4star", "8star")) {
		polygonGrob(x=x[[1]], y=y[[1]], id=id, gp=gpar(fill=fill, lwd=comp$lwd, col=dark))
	} else if (comp$type=="radar") {
		gTree(children = gList(
			circleGrob(x=x[[1]], y=y[[1]], r = cr[1], gp=gpar(lwd=2*LWD, col=dark, fill=light)),
			polylineGrob(x=x[[2]], y=y[[2]], id=rep(1:8, each=2), gp=gpar(lwd=1*LWD, col=dark)),
			polylineGrob(x=x[[3]], y=y[[3]], id=rep(1:4, each=2), gp=gpar(lwd=2*LWD, col=dark)),
			polylineGrob(x=x[[4]], y=y[[4]], id=rep(1:4, each=2), gp=gpar(lwd=1*LWD, col=dark)),
			circleGrob(x=x[[1]], y=y[[1]], r = cr[2], gp=gpar(lwd=1*LWD, col=dark, fill=NA)),
			circleGrob(x=x[[1]], y=y[[1]], r = cr[3], gp=gpar(lwd=2*LWD, col=dark, fill=light)),
			circleGrob(x=x[[1]], y=y[[1]], r = cr[4], gp=gpar(lwd=1*LWD, col=NA, fill=dark)),
			circleGrob(x=x[[1]], y=y[[1]], r = cr[5], gp=gpar(lwd=1*LWD, col=NA, fill=light)),
			polylineGrob(x=x[[5]], y=y[[5]], id=rep(1:8, each=2), gp=gpar(lwd=2*LWD, col=light))))
	} else if (comp$type=="rose") {
		gTree(children = gList(
			circleGrob(x=x[[1]], y=y[[1]], r = cr[1], gp=gpar(lwd=2*LWD, col=dark, fill=light)),
			polygonGrob(x=x[[4]], y=y[[4]], id=id, gp=gpar(lwd=1*LWD, col=dark, fill=fill)),
			polylineGrob(x=x[[2]], y=y[[2]], id=rep(1:8, each=2), gp=gpar(lwd=1*LWD, col=dark)),
			circleGrob(x=x[[1]], y=y[[1]], r = cr[2], gp=gpar(lwd=1*LWD, col=dark, fill=NA)),
			circleGrob(x=x[[1]], y=y[[1]], r = cr[3], gp=gpar(lwd=2*LWD, col=dark, fill=light)),
			circleGrob(x=x[[1]], y=y[[1]], r = cr[4], gp=gpar(lwd=1*LWD, col=NA, fill=dark)),
			circleGrob(x=x[[1]], y=y[[1]], r = cr[5], gp=gpar(lwd=1*LWD, col=NA, fill=light)),
			polylineGrob(x=x[[3]], y=y[[3]], id=rep(1:8, each=2), gp=gpar(lwd=2*LWD, col=light))))
	}


	# other grid cells are aligns (1 and 5) and margins (2 and 4)
	compass = gridCell(3,3, {
		gTree(children=gList(grobBG,
						 grobComp,
						 grobLabels),
		  name="compass")
	})
	grid::grobTree(compass, vp = vp)
}






#' @export
tmapGridCompPrepare.tm_scalebar = function(comp, o) {
	show.messages = o$show.messages
	show.warnings = o$show.warnings
	within(comp, {
		if (all(c("breaks", "width") %in% call) && show.warnings) {
			cli::cli_inform("For {.fun tm_scalebar}, {.arg breaks} and {.arg width} are not supposed to be used together; normally, setting the exact width is not needed when breaks have been specified, only for fine-tuning")
		}
		if ("breaks" %in% call) {
			if (breaks[1] != 0) {
				if (show.warnings) warning("First scalebar breaks value should be 0.", call. = FALSE)
				breaks = c(0, breaks)
			}
		}

		# if (is.na(width))
		# 	width = .25
		# else if (width > 1) {
		# 	if (show.messages) message("Scale bar width set to 0.25 of the map width")
		# 	width = .25
		# }

		if (is.na(text.color)) text.color = o$attr.color
		text.size = text.size * o$scale
		lwd = lwd * o$scale
		show = TRUE
	})
}

#' @export
tmapGridCompHeight.tm_scalebar = function(comp, o) {
	h = 2.75 * o$lin * comp$text.size

	textS = comp$text.size #* o$scale
	#textP = comp$padding[c(3,1)] * textS * o$lin


	marH = comp$margins[c(3,1)] * textS * o$lin
	hs = c(marH[1], h, marH[2])


	sides = switch(comp$position$align.v, top = "second", bottom = "first", "both")
	hsu = set_unit_with_stretch(hs, sides = sides)

	Hin = sum(hs)
	comp$flexRow = NA
	comp$Hin = Hin #  sum(textP[1], textH, textP[2])
	comp$hsu = hsu
	comp
}

#' @export
tmapGridCompWidth.tm_scalebar = function(comp, o) {
	#w = comp$width * o$lin * comp$text.size

	textS = comp$text.size #* o$scale
	#textP = comp$padding[c(3,1)] * textS * o$lin

	marW = comp$margins[c(2,4)] * textS * o$lin

	W = comp$width * textS * o$lin

	ws = c(marW[1], W, marW[2])

	sides = switch(comp$position$align.h, left = "second", right = "first", "both")
	wsu = set_unit_with_stretch(ws, sides = sides)


	comp$Win = sum(ws)
	comp$wsu = wsu

	# in case breaks are used: adjust the legend width later (in tmapGridComp)
	comp$WnativeID = 3
	if (!is.null(comp$breaks)) {
		comp$WnativeRange = tail(comp$breaks, 1) - comp$breaks[1]# + (comp$breaks[2] - comp$breaks[1]) * 2
		#comp$Wextra_text_inch = text_width_inch(paste0("  ", cmp$units$unit)) + text_width_inch(paste0(tail(comp$breaks, 1), comp$breaks[1])) / 2
	}

	comp
}

#' @exportS3Method tmapGridCompPlot tm_scalebar
#' @export
tmapGridCompPlot.tm_scalebar = function(comp, o, fH, fW) {
	light = comp$color.light
	dark = comp$color.dark

	if (any(is.na(comp$dst))) {
		cli::cli_inform("{.fun tm_scalebar} the scalebar cannot be plotted because the crs does not have measurable units")
		return(NULL)
	}

	wsu = comp$wsu
	hsu = comp$hsu

	vp = grid::viewport(layout = grid::grid.layout(ncol = length(wsu),
												   nrow = length(hsu),
												   widths = wsu,
												   heights = hsu))


	unit = comp$units$unit
	unit.size = 1/comp$units$to

	#xrange = fW * comp$cpi
	xrange = fW * (comp$dst[comp$which_side_of_map] / comp$bb_facet)

	# xrange is the range of the viewport in terms of coordinates
	# xrange2 is the same but with units (e.g. km instead of m)
	# W is the targeted space for the scalebar


	W = as.numeric(wsu[3])

	crop_factor = W / fW
	just = 0

	compare_scale_pos = function(x) {
		curr = comp$dst[x]
		oth = comp$dst[names(comp$dst) != x]
		lapply(1:length(oth), function(i) {
			wh = names(oth)[i]
			o = oth[i]
			if (o < curr) {
				sign = "large"
				scale = round(curr/o, 1)
			} else {
				sign = "large"
				scale = round(o/curr, 1)
			}
			loc = paste0("at the ", ifelse(wh == "mid", "center", wh))
			list(sign = sign, scale = scale, loc = loc)
		})
	}




	if (o$show.warnings) {
		res = compare_scale_pos(comp$which_side_of_map)

		tms = vapply(res, function(r)r$scale, FUN.VALUE = numeric(1))
		if (any(tms > 1.1)) {
			cli::cli_warn("The scale bar would be {res[[1]]$scale} times as {res[[1]]$sign} when plotted {res[[1]]$loc} and {res[[2]]$scale} times as {res[[2]]$sign} when plotted {res[[2]]$loc}")
		}
	}

	xrange2 = as.numeric(units::set_units(xrange, "km", mode = "standard"))#xrange/unit.size

	if (is.null(comp$breaks)) {
		# determine resolution only (unselect steps that do not fit later (with 'sel'))
		for (i in 10:1) {
			tcks = pretty(c(0, xrange2*crop_factor), i)
			tcks3 = (tcks / xrange2) * fW
			tcksL = format(tcks, trim=TRUE)
			labW = text_width_inch(tcksL) * comp$text.size
			tickW = tcks3[-1] - head(tcks3, -1)
			if (all(tickW > labW[-1])) {
				sbW = W - labW
				break
			}
		}
		ticks2 = tcks
	} else {
		ticks2 = comp$breaks
		tcksL = format(ticks2, trim=TRUE)

		labW = text_width_inch(tcksL) * comp$text.size
		sbW = W - labW
	}

	ticks3 = ticks2 / xrange2 * fW
	sel = which(ticks3 <= sbW)

	if (!is.null(comp$breaks) && length(sel) != length(ticks3)) {
		cli::cli_warn("{.field tm_scalebar} Not all scale bar breaks could be plotted. Try increasing the scale bar width or decreasing the font size")
	}

	ticks3 = ticks3[sel]
	ticks2 = ticks2[sel]

	ticks2Labels = format(ticks2, trim=TRUE)
	ticksWidths = text_width_inch(ticks2Labels)
	unitWidth = text_width_inch(unit)

	labels = c(ticks2Labels, unit)
	labelsW = c(ticksWidths, unitWidth)

	n = length(ticks2)

	if (length(ticks3) == 1) {
		cli::cli_warn("{.field tm_scalebar} Only one scale bar break could be plotted. Try decreasing the font size or increasing the scale bar width")
		widths = sbW
	} else {
		widths = ticks3[2:n] - ticks3[1:(n-1)]
	}
	size = min(comp$text.size, widths/max(ticksWidths))
	x = ticks3[1:(n-1)]  + ticksWidths[1]*size #+ .5*widths[1]

	lineHeight = convertHeight(unit(1, "lines"), "inch", valueOnly=TRUE) * size

	unitWidth = text_width_inch(unit) * size

	xtext = x[1] + c(ticks3, ticks3[n] + .5*ticksWidths[n]*size + .5*unitWidth)# + widths*.5 + unitWidth*.5) #+ position[1]


	# if "unit" text is clipped, remove last label and move unit to previous label
	if (!comp$allow_clipping) {
		xright = xtext + labelsW / 2
		if (tail(xright, 1) > W) {
			labels = c(head(labels, -2), unit)
			xtext = x[1] + c(head(ticks3, -1), ticks3[n-1] + .5*ticksWidths[n-1]*size + .5*unitWidth)# + widths*.5 + unitWidth*.5) #+ position[1]
		}
	}


	grobBG = if (getOption("tmap.design.mode")) rectGrob(gp=gpar(fill="#CAB2D6")) else NULL


	# other grid cells are aligns (1 and 5) and margins (2 and 4)
	scalebar = gridCell(3,3, {
		gTree(children=gList(

			grobBG,
			rectGrob(x=unit(x, "inch"), y=unit(1.5*lineHeight, "inch"), width = unit(widths, "inch"), height=unit(lineHeight*.5, "inch"), just=c("left", "bottom"), gp=gpar(col=dark, fill=c(light, dark), lwd=comp$lwd)),
			textGrob(label=labels, x = unit(xtext, "inch"), y = unit(lineHeight, "inch"), just=c("center", "center"), gp=gpar(col=comp$text.color, cex=size, fontface=comp$text.fontface, fontfamily=comp$text.fontfamily))
			), name="scalebar")
	})

	grid::grobTree(scalebar, vp = vp)
	#scalebar
}










tmapGridCompPlot_text = function(comp, o, fH, fW) {

	textS = if (comp$text == "") 0 else comp$size * comp$scale #* o$scale

	padding = grid::unit(comp$padding[c(3,4,1,2)] * textS * o$lin, units = "inch")

	if (comp$position$align.h == "left") {
		#x = grid::unit(0, "npc")
		x = grid::unit(comp$padding[2] * textS * o$lin, units = "inch")
		halign = 0
		hjust = 1
		just = "left"
	} else if (comp$position$align.h == "right") {
		#x = grid::unit(1, "npc")
		x = grid::unit(1, "npc") - grid::unit(comp$padding[4] * textS * o$lin, units = "inch")
		halign = 1
		hjust = 0
		just = "right"
	} else {
		x = grid::unit(0.5, "npc")
		halign = 0.5
		hjust = 0.5
		just = "center"
	}

	# grtext = gridtext::richtext_grob(comp$text,
	# 								  x = x,
	# 								  box_gp = gpar(col = frame.col, fill = bg.color, alpha = bg.alpha, lwd = frame.lwd),
	# 								  r = grid::unit(frame.r, "pt"),
	# 								  halign = halign,
	# 								  hjust = hjust,
	# 								  gp = grid::gpar(cex = textS))
	grtext = grid::textGrob(comp$text,
							 x = x,
							 just = just,
							 gp = grid::gpar(col = comp$color, cex = textS, fontface = comp$fontface, fontfamily = comp$fontfamily, alpha = comp$alpha))

	if (getOption("tmap.design.mode")) {
		grDesign = grid::rectGrob(gp=gpar(fill=NA,col="#000000", lwd=1))
	} else {
		grDesign = NULL
	}

	#grtext = rectGrob(gp=gpar(col = "green", fill= NA))


	g = do.call(grid::grobTree, c(list(grtext), list(grDesign))) #, list(vp = vp)

	g

}


correct_nlines = function(n) {
	# Linear model applied based on this empirical data:
	# (results may depend on device)
	#
	# y = sapply(1:50, function(i) {
	# 	s = paste(rep("text", i), collapse = "\n")
	# 	convertHeight(stringHeight(s), "inch", valueOnly = TRUE)
	# })
	# df = data.frame(x = 1:50, y = y / 0.2) #0.2 is the lineheight (par "cin")
	# lm(y~x, df)
	#
	-.6035 + n * 1.2
}


tmapGridCompHeight_text = function(comp, o) {
	textS = if (comp$text == "") 0 else comp$size #* o$scale
	textP = comp$padding[c(3,1)] * textS * o$lin
	textH = textS * o$lin

	nlines = number_text_lines(comp$text)

	nlines2 = correct_nlines(nlines)
	comp$flexRow = NA

	hs = c(textP[1], textH * nlines2, textP[2])
	h = sum(hs)

	sides = switch(comp$position$align.v, top = "second", bottom = "first", "both")
	hsu = set_unit_with_stretch(hs, sides = sides)

	comp$Hin = h
	comp$hsu = hsu
	comp
}

# borrowed from treemap (wraps text to 1-5 lines)
wrapText = function(txt, nlines) {
	if (nlines == 1) {
		txt
	} else {
		# create some wrappings, with slightly different widths:
		results <- lapply(1:5, FUN=function(pos, nlines, txt) {
			strwrap(txt, width = pos+(nchar(txt)/nlines))}, nlines, txt)
		lengths = sapply(results, length)

		# find the best match
		diff = nlines - lengths
		diff[diff < 0] = 1000
		id = which.min(diff)[1]

		paste(results[[id]], collapse = "\n")
	}
}




tmapGridCompWidth_text = function(comp, o) {
	textS = if (comp$text == "") 0 else comp$size #* o$scale
	textP = comp$padding[c(2,4)] * textS * o$lin
	textW = textS * graphics::strwidth(comp$text, units = "inch", family = comp$fontfamily, font = fontface2nr(comp$fontface))


	if (!is.na(comp$width)) {
		textPgs = strsplit(comp$text, "\n")[[1]]
		text2 = do.call(paste, c(lapply(textPgs, function(p) {
			textW = textS * graphics::strwidth(p, units = "inch", family = comp$fontfamily, font = fontface2nr(comp$fontface))
			w = sum(textP[1], textW, textP[2])
			nlines = round(w / (comp$width * textS * o$lin))
			wrapText(p, nlines)
		}), list(sep = "\n")))

		textW2 = textS * graphics::strwidth(text2, units = "inch", family = comp$fontfamily, font = fontface2nr(comp$fontface))
		wsu2 = c(textP[1], textW2, textP[2])
		ws = sum(textP[1], textW2, textP[2])
		comp$text = text2
	} else {
		ws = c(textP[1], textW, textP[2])
	}


	sides = switch(comp$position$align.h, left = "second", right = "first", "both")
	wsu = set_unit_with_stretch(ws, sides = sides)

	comp$Win = sum(ws)
	comp$wsu = wsu

	comp$flexCol = NA
	comp
}






#' @export
tmapGridCompPrepare.tm_logo = function(comp, o) {
	comp$logo = lapply(comp$file, function(lf){
		tmap_icons(lf)
	})
	comp$asp = vapply(comp$logo, function(lg) {
		lg$iconWidth / lg$iconHeight
	}, FUN.VALUE = numeric(1))
	comp$show = TRUE
	comp
}

#' @export
tmapGridCompHeight.tm_logo = function(comp, o) {
	marH = comp$margins[c(3,1)] * o$lin
	hs = c(marH[1], comp$height * o$lin, marH[2])

	sides = switch(comp$position$align.v, top = "second", bottom = "first", "both")
	hsu = set_unit_with_stretch(hs, sides = sides)

	Hin = sum(hs)
	comp$flexRow = NA
	comp$Hin = Hin #  sum(textP[1], textH, textP[2])
	comp$hsu = hsu
	comp
}

#' @export
tmapGridCompWidth.tm_logo = function(comp, o) {
	k = length(comp$asp)
	comp$width = comp$height * comp$asp

	marW = comp$margins[c(2,4)] * o$lin
	ws = c(marW[1],
		   comp$width[1] * o$lin,
		   {if (k > 1) unlist(lapply(comp$width[-1], function(w) c(comp$between_margin, w) * o$lin)) else NULL},
		   marW[2])

	sides = switch(comp$position$align.h, left = "second", right = "first", "both")
	wsu = set_unit_with_stretch(ws, sides = sides)
	comp$col_ids = seq(3L, by = 2L, length.out = k)
	comp$flexCol = NA
	comp$Win = sum(ws)
	comp$wsu = wsu
	comp
}

#' @export
tmapGridCompPlot.tm_logo = function(comp, o, fH, fW) {

	k = length(comp$logo)

	wsu = comp$wsu
	hsu = comp$hsu

	vp = grid::viewport(layout = grid::grid.layout(ncol = length(wsu),
												   nrow = length(hsu),
												   widths = wsu,
												   heights = hsu))
	gLogos = mapply(function(logo, col) {
		grobLogo = pngGrob(logo$iconUrl, fix.borders = TRUE, n=2, height.inch=as.numeric(comp$hsu[3]), target.dpi=96)
		rdim = dim(grobLogo$raster)
		grobLogo$raster = matrix(do.call("process_color", c(list(as.vector(grobLogo$raster)), o$pc)), nrow = rdim[1], ncol=rdim[2])
		gridCell(3L, col, grobLogo)
	}, comp$logo, comp$col_ids, SIMPLIFY = FALSE)

	grobBG = if (getOption("tmap.design.mode")) rectGrob(gp=gpar(fill="#CAB2D6")) else NULL
	gBG = gridCell(3L, 3L:(length(wsu) - 2L), grobBG)

	do.call(grid::grobTree, c(list(gBG), gLogos, list(vp = vp)))

}


#' @export
tmapGridCompPrepare.tm_inset_tmap = function(comp, o) {
	asp = comp$width / comp$height

	comp$x = tmap_grob(comp$x, asp = asp)
	.TMAP$is_first_inset = FALSE
	class(comp)[1] = "tm_inset_grob"

	b = .TMAP$geo_ref$bbx

	comp$bbox = .TMAP$geo_ref$bbx_frame
	comp$show = TRUE
	comp
}


#' @export
tmapGridCompPrepare.tm_inset_grob = function(comp, o) {
	comp$show = TRUE
	comp
}

#' @export
tmapGridCompPrepare.tm_inset_gg = function(comp, o) {
	rlang::check_installed("ggplot2", reason = "for plotting ggplot2 charts")
	comp$x = ggplot2::ggplotGrob(comp$x)
	class(comp)[1] = "tm_inset_grob"
	comp$show = TRUE
	comp
}



#' @export
tmapGridCompHeight.tm_inset_grob = function(comp, o) {
	marH = comp$margins[c(3,1)] * o$lin
	hs = c(marH[1], comp$height * o$lin, marH[2])

	sides = switch(comp$position$align.v, top = "second", bottom = "first", "both")
	hsu = set_unit_with_stretch(hs, sides = sides)

	Hin = sum(hs)
	comp$flexRow = NA
	comp$Hin = Hin #  sum(textP[1], textH, textP[2])
	comp$hsu = hsu
	comp
}

#' @export
tmapGridCompWidth.tm_inset_grob = function(comp, o) {
	marW = comp$margins[c(2,4)] * o$lin

	ws = c(marW[1], comp$width * o$lin, marW[2])

	sides = switch(comp$position$align.h, left = "second", right = "first", "both")
	wsu = set_unit_with_stretch(ws, sides = sides)

	comp$flexCol = NA
	comp$Win = sum(ws)
	comp$wsu = wsu
	comp
}

#' @export
tmapGridCompPrepare.tm_inset_image = function(comp, o) {
	comp$file = comp$x
	comp = tmapGridCompPrepare.tm_logo(comp, o)
	class(comp) = c("tm_logo", class(comp))
	comp
}



#' @export
tmapGridCompPlot.tm_inset_grob = function(comp, o, fH, fW) {


	wsu = comp$wsu
	hsu = comp$hsu

	vp = grid::viewport(layout = grid::grid.layout(ncol = length(wsu),
												   nrow = length(hsu),
												   widths = wsu,
												   heights = hsu))

	grobBG = if (getOption("tmap.design.mode")) rectGrob(gp=gpar(fill="#CAB2D6")) else NULL
	gBG = gridCell(3L, 3L:(length(wsu) - 2L), grobBG)

	g = gridCell(3L, 3L, comp$x)

	do.call(grid::grobTree, c(list(gBG, g), list(vp = vp)))

}



#' @export
tmapGridCompPrepare.tm_inset_map = function(comp, o) {
	limit_lat = if ("limit_latitude_3857" %in% names(o)) o$limit_latitude_3857 else c(-90, 90)

	if (is.null(comp$x)) {
		comp$x = sf::st_bbox(c(xmin = -180, xmax = 180, ymin = limit_lat[1], ymax = limit_lat[2]), crs = 4326)
		comp$crs = tmap_options()$crs_global
	}

	bbpoly = sf::st_transform(tmaptools::bb_poly(comp$x), o$crs_step4)

	b = tmaptools::bb(bbpoly, asp.target = comp$width / comp$height)

	comp$bbox = b
	comp$show = TRUE
	comp
}

#' @export
tmapGridCompHeight.tm_inset_map = function(comp, o) {
	marH = comp$margins[c(3,1)] * o$lin
	hs = c(marH[1], comp$height * o$lin, marH[2])

	sides = switch(comp$position$align.v, top = "second", bottom = "first", "both")
	hsu = set_unit_with_stretch(hs, sides = sides)

	Hin = sum(hs)
	comp$flexRow = NA
	comp$Hin = Hin #  sum(textP[1], textH, textP[2])
	comp$hsu = hsu
	comp
}

#' @export
tmapGridCompWidth.tm_inset_map = function(comp, o) {
	marW = comp$margins[c(2,4)] * o$lin

	ws = c(marW[1], comp$width * o$lin, marW[2])

	sides = switch(comp$position$align.h, left = "second", right = "first", "both")
	wsu = set_unit_with_stretch(ws, sides = sides)

	comp$flexCol = NA
	comp$Win = sum(ws)
	comp$wsu = wsu
	comp
}

#' @export
tmapGridCompPlot.tm_inset_map = function(comp, o, fH, fW) {


	wsu = comp$wsu
	hsu = comp$hsu

	vp = grid::viewport(layout = grid::grid.layout(ncol = length(wsu),
												   nrow = length(hsu),
												   widths = wsu,
												   heights = hsu))

	grobBG = if (getOption("tmap.design.mode")) rectGrob(gp=gpar(fill="#CAB2D6")) else NULL
	gBG = gridCell(3L, 3L:(length(wsu) - 2L), grobBG)

	vp = grid::viewport(layout.pos.row = 3L, layout.pos.col = 3L)

	#comp$tm$o = prepreprocess_meta(comp$tm$o, vp)
	comp$tm$o$asp = as.numeric(wsu[3]) / as.numeric(hsu[3])

	#comp$tm$tmo$group1$layers$layer1$shpDT$shpTM[[1]]$bbox$x = comp$bbox
	comp$tm$o$inset = TRUE
	comp$tm$o$bbox = comp$bbox

	g = step4_plot(comp$tm, vp, return.asp = FALSE, show = FALSE, in.shiny = FALSE, knit = FALSE, knit_opts = list(), args = list())

	#g = gridCell(3L, 3L, grid::rectGrob(gp=gpar(fill="red")))#comp$x)

	do.call(grid::grobTree, c(list(gBG, g), list(vp = vp)))

}



#' @export
tmapGridCompPrepare.tm_minimap = function(comp, o) {
	comp$show = TRUE

	comp
}

#' @export
tmapGridCompHeight.tm_minimap = function(comp, o) {
	tmapGridCompHeight.tm_inset_grob(comp, o)
}

#' @export
tmapGridCompWidth.tm_minimap = function(comp, o) {
	tmapGridCompWidth.tm_inset_grob(comp, o)
}

#' @export
tmapGridCompPlot.tm_minimap = function(comp, o, fH, fW) {
	poly = tmaptools::bb_poly(comp$bbox)

	if (sf::st_is_longlat(comp$bbox)) {
		center = c(mean(comp$bbox[c("xmin", "xmax")]),
				   mean(comp$bbox[c("ymin", "ymax")]))
	} else {
		center = suppressWarnings(round(sf::st_transform(sf::st_centroid(poly), crs = 4326)[[1]][], 3))
	}

	# bound lat to -30,-10 or 10,30
	center[2] = if (center[2] >= 0) {
		max(min(center[2], 30), 10)
	} else {
		max(min(center[2], -10), -30)
	}



	#"+proj=ortho +lat_0=10 +lon_0=0"

	lastcalln = x = get("last_map_new", envir = .TMAP)

	tm = tm_shape(World) +
		tm_crs(ortho_lonlat(center[1], center[2]), bbox = "FULL")+
		tm_polygons(col = NULL, fill = "#2CA02C") +
	tm_shape(poly) +
		tm_polygons(fill = NULL, col = "#EE8866", lwd = 3) +
		#tm_polygons(fill = NULL, col = "#EE8866", lwd = 3, lty = "dotted") +
		tm_graticules(labels.show = FALSE, col = "#000000", lwd = 0.5) +
		tm_layout(bg.color = "#88CCEE",
				  earth_boundary = TRUE,
				  frame = FALSE,
				  space = FALSE) +
		tm_options(space_overlay = o$space_overlay)

	asp = comp$width / comp$height


	comp$x = tmap_grob(tm, asp = asp)

	assign("last_map_new", lastcalln, envir = .TMAP)


	class(comp)[1] = "tm_inset_grob"
	tmapGridCompPlot.tm_inset_grob(comp, o, fH, fW)
}

#' Specify facets
#'
#' @description
#' * `tm_facets_wrap()` specify facets for one grouping variable (so one faceting dimension)
#' * `tm_facets_(hv)stack()` stacks the facets either horizontally or vertically (one grouping variable).
#' * `tm_facets_grid()` specify facets for two grouping variables in a grid of rows and columns.
#' * `tm_facets_pagewise()` same as wrap, but the facets are drawn on different plots (pages). Replaces the `along` argument from version 3.
#' * `tm_facets()` is the core function, but it is recommended to use the other functions.
#'
#' @param by Group by variable (only for a facet wrap or facet stack)
#' @param rows Variable that specifies the rows (only for a facet grid)
#' @param columns Variable that specifies the columns (only for a facet grid)
#' @param pages Variable that specifies the pages (only for a facet grid)
#' @param as.layers show facets as layers?
#' @param nrow Number of rows
#' @param ncol Number of columns
#' @param byrow Should facets be wrapped by row?
#' @param orientation For facet stack: horizontal or vertical?
#' @param free.coords Logical. If the `by` argument is specified, should each
#'   map has its own coordinate ranges? By default `TRUE`, unless facets are shown
#'   in as different layers (`as.layers = TRUE`)
#' @param drop.units Logical. If the `by` argument is specified, should
#'   non-selected spatial units be dropped? If `FALSE`, they are plotted where
#'   mapped aesthetics are regarded as missing values. Not applicable for
#'   raster shapes. By default `TRUE`.
#' @param drop.empty.facets Logical. If the `by` argument is specified, should
#'   empty facets be dropped? Empty facets occur when the `by`-variable contains
#'   unused levels. When `TRUE` and two `by`-variables are specified, empty rows
#'   and columns are dropped.
#' @param drop.NA.facets Logical. If the `by` argument is specified, and all
#'   data values for specific facets are missing, should these facets be dropped?
#'   `FALSE` by default. In v3, it was called `showNA`.
#' @param sync Logical. Should the navigation in view mode (zooming and panning)
#'   be synchronized? By default `TRUE` if the facets have the same bounding box.
#'   This is generally the case when rasters are plotted, or when `free.coords` is
#'   `FALSE`.
#' @param swipe Logical. If \code{TRUE}, the two maps are overlaid with a draggable divider to reveal either map. Requires exactly two maps, and only works in the \code{"mapbox"} and \code{"maplibre"} modes provided by the \pkg{tmap.mapgl} package. Default is \code{FALSE}.
#' @param na.text Text used for facets of missing values. In v3, it was `textNA`.
#' @param scale.factor Number that determines how the elements (e.g. font sizes,
#'   symbol sizes, line widths) of the small multiples are scaled in relation to
#'   the scaling factor of the shapes. The elements are scaled to the `scale.factor`th
#'   root of the scaling factor of the shapes. So, for `scale.factor = 1`,
#'   they are scaled proportional to the scaling of the shapes. Since elements,
#'   especially text, are often too small to read, a higher value is recommended.
#'   By default, `scale.factor = 2`.
#' @param type `"grid"`, `"wrap"` or `"stack"`
#' @param free.scales deprecated. Please use the `.free` arguments in the layer functions, e.g. `fill.free` in `tm_polygons`.
#' @param ... used to catch deprecated arguments
#' @seealso [tm_animate()]
#' @example ./examples/tm_facets.R
#' @note In older versions (< 4.1) [tm_facets()] with `page` specification was used to create animations frames and [tmap_animation()] to create the animation itself using inputs like the frame rate specification. As of version 4.2, the whole animation, including frame rate, is specified in [tm_animate()]. The animation can still be saved via [tmap_animation()].
#' @seealso `r .doc_see_also_facets()`
#' @export
tm_facets = function(by = NULL,
					 rows = NULL,
					 columns = NULL,
					 pages = NULL,
					 as.layers = FALSE,
					 nrow = NA,
					 ncol = NA,
					 byrow = TRUE,
					 orientation = NA,
					 free.coords = NA,
					 drop.units = TRUE,
					 drop.empty.facets = TRUE,
					 drop.NA.facets = FALSE,
					 sync = TRUE,
					 swipe = FALSE,
					 na.text = NA,
					 scale.factor=2,
					 type = NA,
					 free.scales = NULL,
					 ...) {

	args = list(...)
	args_called = names(rlang::call_match(dots_expand = TRUE)[-1])

	if (any(v3_only("tm_facets") %in% names(args))) {
		layer_fun = "facets"
		v3_start_message()

		v3_list_init()
		drop.empty.facets = v3_impute(args, "showNA", TRUE, paste0("drop.empty.facets = ", !args$showNA))
		na.text = v3_impute(args, "textNA", NA, "na.text")
		drop.units = v3_impute(args, "drop.shapes", TRUE, "drop.units")
		v3_tm_facets(arg_list = v3_list_get())

		if (any(substr(names(args), 1, 12) == "free.scales.")) {
			v3_tm_facets_free_scales()
		}

	}

	if (!("animate" %in% args_called)) {
		args$animate = FALSE
		#dummy values:
		args$nframes = 60L
		args$fps = 30L
		args$play = "loop"
		args$dpr = 2
	}


	if ("along" %in% args_called) {
		warning("The 'along' argument of 'tm_facets()' is deprecated as of tmap 4.0. Please use 'pages' instead.", call. = FALSE)
		pages = args$along
	}

	if (!is.null(rows) || !is.null(columns)) {
		type = "grid"
	}

	x = tm_element_list(tm_element(
		type = type,
		by = by,
		rows = rows,
		columns = columns,
		pages = pages,
		as.layers = as.layers,
		pages = pages,
		nrows = nrow,
		ncols = ncol,
		byrow = byrow,
		orientation = orientation,
		free.coords = free.coords,
		drop.units = drop.units,
		drop.empty.facets = drop.empty.facets,
		drop.NA.facets = drop.NA.facets,
		sync = sync,
		swipe = swipe,
		na.text = na.text,
		scale.factor = scale.factor,
		animate = args$animate,
		nframes = as.integer(args$nframes),
		fps = as.integer(args$fps),
		play = args$play,
		dpr = args$dpr,
		calls = args_called,
		subclass = "tm_facets"))

	if (!is.null(free.scales)) {
		message("tm_facets(): the argument free.scales is deprecated. Specify this via the layer functions (e.g. fill.free in tm_polygons)")
		x + tm_options(free.scales = free.scales)
	} else {
		x
	}
}

#' @export
#' @rdname tm_facets
tm_facets_grid = function(rows = NULL,
						  columns = NULL,
						  pages = NULL,
						  ...) {
	args = list(...)
	args_called = names(rlang::call_match()[-1])
	tm = do.call("tm_facets", c(list(by = NULL, rows = rows, columns = columns, pages = pages, type = "grid"), args[setdiff(names(args), "type")]))
	tm[[1]]$calls = args_called
	tm
}


#' @export
#' @rdname tm_facets
#' @param ... passed on to `tm_facets()`
tm_facets_wrap = function(by = "VARS__",
						  nrow = NA,
						  ncol = NA,
						  byrow = TRUE,
						  ...) {
	args = list(...)
	args_called = names(rlang::call_match()[-1])
	tm = do.call("tm_facets", c(list(by = by, nrow = nrow, ncol = ncol, byrow = byrow, type = "wrap"), args[setdiff(names(args), "type")]))
	tm[[1]]$calls = args_called
	tm
}

#' @export
#' @rdname tm_facets
tm_facets_pagewise = function(by = "VARS__",
							  byrow = TRUE,
							  ...) {
	args = list(...)
	args_called = names(rlang::call_match()[-1])

	if (any(c("nrow", "ncol") %in% args_called) && (args$nrow != 1 || args$ncol != 1)) {
		cli::cli_warn("tm_facets_pagewise does not support multiple facets per page. Please use tm_facets instead")
	}

	tm = do.call("tm_facets", c(list(pages = by, byrow = byrow, type = NA), args[setdiff(names(args), "type")]))
	tm[[1]]$calls = args_called
	tm
}

#' @export
#' @rdname tm_facets
tm_facets_stack = function(by = "VARS__",
						   orientation = NA,
						  ...) {
	args = list(...)
	args_called = names(rlang::call_match()[-1])
	tm = do.call("tm_facets", c(list(by = by, orientation = orientation, type = "stack"), args[setdiff(names(args), "type")]))
	tm[[1]]$calls = args_called
	tm
}

#' @export
#' @rdname tm_facets
tm_facets_hstack = function(by = "VARS__",
						   ...) {
	do.call("tm_facets_stack", c(list(by = by, orientation = "horizontal"), list(...)))
}

#' @export
#' @rdname tm_facets
tm_facets_vstack = function(by = "VARS__",
							...) {
	do.call("tm_facets_stack", c(list(by = by, orientation = "vertical"), list(...)))
}

#' @export
#' @rdname tm_facets
tm_facets_flip = function(...) {
	cli::cli_warn("{.field [tm_facets_flip]} please use {.code tm_facets(byrow = FALSE)} instead of {.fun tm_facets_flip}")
	do.call(tm_facets, c(list(byrow = FALSE), list(...)))
}

grid_nonoverlap = function(x, s) {
	n = length(x)
	x = c(-x[1], x, 1 + (1 - x[n]))

	m_left = x[2:(n+1)] - x[1:n]
	m_right = x[3:(n+2)] - x[2:(n+1)]
	m = pmin(m_left, m_right)

	m > s
}

pretty30 = function(x, n, longlat) {
	p = pretty(x, n)
	if (!longlat) return(p)
	step = p[2] - p[1]
	if (step > 50) {
		x = p %/% 60 * 60
		seq(min(x), max(x), by = 60)
	} else if (step > 10) {
		x = p %/% 30 * 30
		seq(min(x), max(x), by = 30)
	} else {
		p
	}
}

#' @export
#' @keywords internal
#' @rdname tmapGridLeaflet
tmapGridAuxPrepare.tm_aux_graticules = function(a, bs, id, o) {
	tmapGridAuxPrepare.tm_aux_grid(a, bs, id, o)
}

#' @export
#' @keywords internal
#' @rdname tmapGridLeaflet
tmapGridAuxPrepare.tm_aux_grid = function(a, bs, id, o) {
	g = get("g", envir = .TMAP_GRID)




	#alpha <- labels.inside_frame <- labels.rot <- NULL
	a2 = within(a, {
		if (labels.inside_frame && !identical(labels.pos, c("left", "bottom"))) cli::cli_warn("label.pos not implemented yet for {.code labels.inside_frame==TRUE} (only {.val left} and {.val bottom})")
		show <- TRUE
		if (is.na(col)) col <- ifelse(o$attr.color.light, darker(o$attr.color, .5), lighter(o$attr.color, .5))
		if (is.na(labels.col)) labels.col <- ifelse(o$attr.color.light, darker(o$attr.color, .2), lighter(o$attr.color, .2))
		if (!is.numeric(labels.rot) || length(labels.rot) != 2) stop("labels.rot should be a numeric vector of length two")
		col <- do.call("process_color", c(list(col=col, alpha=alpha), o$pc))
		labels.col <- do.call("process_color", c(list(col=labels.col), o$pc))
		lwd <- lwd * o$scale
		is.projected <- !(crs=="longlat" || tryCatch(sf::st_is_longlat(crs), error = function(e) TRUE))

		crs <- sf::st_crs(crs)

		if (!labels.inside_frame && any(o$outer.margins[1:2]==0)) stop("When grid labels are plotted outside the frame, outer.margins (the bottom and the left) should be greater than 0. When using tmap_save, notice that outer.margins are set to 0 by default, unless set to NA.")
		if (!"scientific" %in% names(labels.format)) labels.format$scientific <- FALSE
		if (!"digits" %in% names(labels.format)) labels.format$digits <- NA

		labels.show <- rep(labels.show, length.out = 2)
		ticks <- rep(ticks, length.out = 2)

		add.labels = labels.inside_frame & labels.show


	})

	o$scale.extra = 1

	a3s = lapply(bs, function(bbx) {
		crs_bb = sf::st_crs(bbx)
		within(a2, {
			if (!is.na(crs)) {
				bbx_orig <- bbx
				bbx <- suppressWarnings(bb(bbx, current.projection = crs_bb, projection = crs))
			}
			sasp = get_asp_ratio(bbx)

			## automatically determine number of grid lines
			if (is.na(n.x) && !is.na(n.y)) {
				n.x <- n.y * sasp
			} else if (!is.na(n.x) && is.na(n.y)) {
				n.y <- n.x / sasp
			} else if (is.na(n.x) && is.na(n.y)) {
				n.lines <- 15 / (o$scale / o$scale.extra)
				n.x <- round(sasp * (n.lines/(1+sasp)))
				n.y <- round(n.lines / (1+sasp))
			}

			## find natural breaks
			custom.x = !is.na(x[1])
			custom.y = !is.na(y[1])

			if (!custom.x) x <- pretty30(bbx[c(1,3)], n=n.x, longlat = !is.na(crs) && sf::st_is_longlat(crs_bb))
			if (!custom.y) y <- pretty30(bbx[c(2,4)], n=n.y, longlat = !is.na(crs) && sf::st_is_longlat(crs_bb))

			## copy x and y
			x.orig <- x
			y.orig <- y

			## crop
			x <- x[x > bbx[1] & x < bbx[3]]
			y <- y[y > bbx[2] & y < bbx[4]]

			## project grid lines
			if (!is.na(crs)) {
				## add extra grid lines to make sure the warped grid is full
				if (custom.x) {
					x2 <- x.orig
				} else {
					gnx2 <- floor(length(x))
					if (gnx2==1) {
						x2 <- x
					} else if (gnx2>1) {
						x2 <- c(rev(seq(x[1], by=-diff(x[1:2]), length.out = gnx2)),
								x[-c(1, length(x))],
								seq(x[length(x)], by=diff(x[1:2]), length.out = gnx2))
					} else x2 <- NA
				}
				if (custom.y) {
					y2 <- y.orig
				} else {
					gny2 <- floor(length(y))
					if (gny2==1) {
						y2 <- y
					} else if (gny2>1) {
						y2 <- c(rev(seq(y[1], by=-diff(y[1:2]), length.out = gny2)),
								y[-c(1, length(y))],
								seq(y[length(y)], by=diff(y[1:2]), length.out = gny2))
					} else y2 <- NA
				}
				if (!is.projected) {
					# x2[abs(x2-180)<1e-9] <- 180
					# x2[abs(x2- -180)<1e-9] <- -180
					# y2[abs(y2-90)<1e-9] <- 90
					# y2[abs(y2- -90)<1e-9] <- -90
					x2 <- x2[x2>=-180 & x2<=180]
					y2 <- y2[y2>=-90 & y2<=90]
				}
				gnx2 <- gny2 <- NULL

				## determine limits
				x2.min <- min(min(x2), bbx[1], na.rm=TRUE)
				x2.max <- max(max(x2), bbx[3], na.rm=TRUE)
				y2.min <- min(min(y2), bbx[2], na.rm=TRUE)
				y2.max <- max(max(y2), bbx[4], na.rm=TRUE)

				lnsSel <- c(length(x2) && !is.na(x2[1]),
							length(y2) && !is.na(y2[1]))

				if (lnsSel[1]) {
					lnsX = sf::st_sfc(lapply(x2, function(x) {
						sf::st_linestring(matrix(c(rep(x,ndiscr), seq(y2.min, y2.max, length.out=ndiscr)), ncol=2))
					}), crs = crs)


					lnsX_proj_res <- transform_ortho(lnsX, crs = crs_bb, tmapID = 1:length(x2))

					# keeping longest grid line, however some CRS's may have MULTILINESTRING grid lines
					lnsX_proj = to_longest_linestring(lnsX_proj_res$shp)

					lnsX_proj_emp = sf::st_is_empty(lnsX_proj)
					lnsX_proj <- lnsX_proj[!lnsX_proj_emp]

					lnsX_emp = rep(TRUE, length(x2))
					lnsX_emp[lnsX_proj_res$tmapID] <- lnsX_proj_emp

					x2 <- x2[!lnsX_emp]

					xco <- sf::st_coordinates(lnsX_proj)
					# co.x.lns
					co.x <- lapply(unique(xco[,3]), function(i) {
						lco <- xco[xco[,3]==i, 1:2,drop = FALSE]
						lco[, 1] <- (lco[, 1]-bbx_orig[1]) / (bbx_orig[3] - bbx_orig[1])
						lco[, 2] <- (lco[, 2]-bbx_orig[2]) / (bbx_orig[4] - bbx_orig[2])
						lco
					})
					lnsX <- NULL
					lnsX_proj <- NULL
					lnsX_emp <- NULL
					lnsX_proj_res <- NULL

					sel.x <- which(x2 %in% x)
				} else {
					co.x <- numeric(0)
				}

				if (lnsSel[2]) {
					lnsY = sf::st_sfc(lapply(y2, function(y) {
						sf::st_linestring(matrix(c(seq(x2.min, x2.max, length.out=ndiscr), rep(y,ndiscr)), ncol=2))
					}), crs = crs)
					lnsY_proj_res <- transform_ortho(lnsY, crs = crs_bb, tmapID = 1:length(y2))

					lnsY_proj = to_longest_linestring(lnsY_proj_res$shp)

					lnsY_proj_emp = sf::st_is_empty(lnsY_proj)
					lnsY_proj <- lnsY_proj[!lnsY_proj_emp]

					lnsY_emp = rep(TRUE, length(y2))
					lnsY_emp[lnsY_proj_res$tmapID] <- lnsY_proj_emp

					y2 <- y2[!lnsY_emp]

					yco <- sf::st_coordinates(lnsY_proj)
					co.y <- lapply(unique(yco[,3]), function(i) {
						lco <- yco[yco[,3]==i, 1:2,drop=FALSE]
						lco[, 1] <- (lco[, 1]-bbx_orig[1]) / (bbx_orig[3] - bbx_orig[1])
						lco[, 2] <- (lco[, 2]-bbx_orig[2]) / (bbx_orig[4] - bbx_orig[2])
						lco
					})
					lnsY <- NULL
					lnsY_proj <- NULL
					lnsY_emp <- NULL
					lnsY_proj_res <- NULL

					sel.y <- which(y2 %in% y)
				}	else {
					co.y <- numeric(0)
				}


				labels.show <- labels.show & lnsSel # update needed for plot_n
			} else {
				# normalize coordinates
				co.x <- (x-bbx[1]) / (bbx[3] - bbx[1])
				co.y <- (y-bbx[2]) / (bbx[4] - bbx[2])
			}

			## format grid labels


			labels.x <- local({
				if (labels.cardinal) {
					xneg <- x < 0
					xpos <- x > 0

					xlab <- do.call("fancy_breaks", c(list(vec=abs(x), intervals=FALSE), o$labels.format)) #format(x, big.mark = ",")

					xlab[xpos] <- paste0(xlab[xpos], "\u00B0E")
					xlab[xneg] <- paste0(xlab[xneg], "\u00B0W")
					xlab
				} else {
					do.call("fancy_breaks", c(list(vec=x, intervals=FALSE), o$labels.format)) #format(x, big.mark = ",")
				}
			})


			labels.y <- local({
				if (labels.cardinal) {
					yneg <- y < 0
					ypos <- y > 0

					ylab <- do.call("fancy_breaks", c(list(vec=abs(y), intervals=FALSE), o$labels.format))

					ylab[ypos] <- paste0(ylab[ypos], "\u00B0N")
					ylab[yneg] <- paste0(ylab[yneg], "\u00B0S")
					ylab
				} else {
					do.call("fancy_breaks", c(list(vec=y, intervals=FALSE), o$labels.format))
				}
			})

			if (!is.na(crs)) {
				# remove lines with one coordinate-row, and update related list items
				cox_sel = vapply(co.x, nrow, FUN.VALUE = integer(1)) >= 2
				co.x = co.x[cox_sel]
				x2 = x2[cox_sel]
				sel.x = which(x2 %in% x)

				coy_sel = vapply(co.y, nrow, FUN.VALUE = integer(1)) >= 2
				co.y = co.y[coy_sel]
				y2 = y2[coy_sel]
				sel.y = which(y2 %in% y)
			}

		})

	})

	g$grid_comp_per_bbx = a3s

	assign("g", g, envir = .TMAP_GRID)

	return("grid")
}

#' @export
#' @keywords internal
#' @rdname tmapGridLeaflet
tmapGridAuxPlot.tm_aux_basemap = function(a, bi, bbx, facet_row, facet_col, facet_page, id, pane, group, o) {
	tmapGridAuxPlot.tm_aux_tiles(a, bi, bbx, facet_row, facet_col, facet_page, id, pane, group, o)
}

#' @export
#' @keywords internal
#' @rdname tmapGridLeaflet
tmapGridAuxPlot.tm_aux_tiles = function(a, bi, bbx, facet_row, facet_col, facet_page, id, pane, group, o) {
	g = get("g", envir = .TMAP_GRID)

	# id is on aux level
	id2 = which(names(g$bmaps_dts) == as.character(id))

	dt = g$bmaps_dts[[id2]][[bi]]
	shpTM = g$bmaps_shpTHs[[id2]][[bi]]
	gp = list()

	class(a) = "tm_data_raster"

	a2 = structure(list(interpolate = TRUE), class = "tm_data_raster")
	if (!is.null(dt)) tmapGridDataPlot(a2, shpTM, dt, gp, bbx, facet_row, facet_col, facet_page, id, pane, group, glid = 0, o)
}


tmapGridGridXLab = function(bi, bbx, facet_row, facet_col, facet_page, o) {
	gts = get("gts", envir = .TMAP_GRID)
	g = get("g", envir = .TMAP_GRID)
	gt = gts[[facet_page]]

	is_top = o$grid.labels.pos[2] == "top"


	rc_text = frc(facet_row, facet_col)

	rowid = g$rows_facet_ids[facet_row] + ifelse(is_top, -3, 3)
	colid = g$cols_facet_ids[facet_col]

	H = g$rowsIn[rowid]
	W = g$colsIn[colid]


	a = g$grid_comp_per_bbx[[bi]]



	labelsx <- a$labels.x

	# find coordinates for projected grid labels
	if (!is.na(a$crs)) {
		glabelsx <- get_gridline_labels(lco=a$co.x[a$sel.x], xax = as.integer(is_top))
		cogridx <- glabelsx$cogrid
		idsx <- glabelsx$ids
		labelsx <- labelsx[idsx]
	} else {
		cogridx <- a$co.x
	}

	cex <- a$labels.size*o$scale

	# remove overlapping grid labels
	widths = text_width_inch(labelsx) * cex / W
	selx2 = grid_nonoverlap(cogridx, widths)
	if (!any(selx2)) return(NULL)
	labelsx2 = labelsx[selx2]
	cogridx2 = cogridx[selx2]


	spacerX <- grid::convertHeight(unit(.5, "lines"), unitTo="inch", valueOnly=TRUE) * cex / H
	marginX <- grid::convertHeight(unit(a$labels.margin.x, "lines"), unitTo="inch", valueOnly=TRUE) * cex / H

	just <- ifelse(a$labels.rot[1] == 90, "right",
				   ifelse(a$labels.rot[1] == 270, "left",
				   	   ifelse(a$labels.rot[1] == 180, "bottom", "top")))

	if (a$ticks[1]) {
		if (is_top) {
			ticks <- grid::polylineGrob(x=rep(cogridx2, each = 2), y = rep(c(0, spacerX*.5+marginX), length(cogridx2)), id = rep(1:length(cogridx2), each = 2), gp=grid::gpar(col=a$col, lwd=a$lwd))
		} else {
			ticks <- grid::polylineGrob(x=rep(cogridx2, each = 2), y = rep(c(1-spacerX*.5-marginX,1), length(cogridx2)), id = rep(1:length(cogridx2), each = 2), gp=grid::gpar(col=a$col, lwd=a$lwd))
		}
	} else {
		ticks <- NULL
	}

	if (is_top) {
		labels <- grid::textGrob(labelsx2, y=1-spacerX/2, x=cogridx2, just=just, rot=a$labels.rot[1], gp=grid::gpar(col=a$labels.col, cex=cex, fontface=a$labels.fontface, fontfamily=a$labels.fontfamily))
	} else {
		labels <- grid::textGrob(labelsx2, y=1-spacerX-marginX, x=cogridx2, just=just, rot=a$labels.rot[1], gp=grid::gpar(col=a$labels.col, cex=cex, fontface=a$labels.fontface, fontfamily=a$labels.fontfamily))
	}

	res = grid::gTree(children = gList(ticks, labels), name = "gridTicksLabelsX")

	#res = gTree(children = gList(rectGrob(gp=gpar(fill="red"))), name = "gridTicksLabelsX")

	gt = add_to_gt(gt, res, row = rowid, col = colid)

	gts[[facet_page]] = gt

	assign("gts", gts, envir = .TMAP_GRID)
	NULL
}


tmapGridGridYLab = function(bi, bbx, facet_row, facet_col, facet_page, o) {
	gts = get("gts", envir = .TMAP_GRID)
	g = get("g", envir = .TMAP_GRID)
	gt = gts[[facet_page]]

	is_left = o$grid.labels.pos[1] == "left"

	rc_text = frc(facet_row, facet_col)

	rowid = g$rows_facet_ids[facet_row]
	colid = g$cols_facet_ids[facet_col] + ifelse(is_left, -3, 3)

	H = g$rowsIn[rowid]
	W = g$colsIn[colid]


	a = g$grid_comp_per_bbx[[bi]]


	labelsy <- a$labels.y

	# find coordinates for projected grid labels
	if (!is.na(a$crs)) {
		glabelsy <- get_gridline_labels(lco=a$co.y[a$sel.y], yax = 0)
		cogridy <- glabelsy$cogrid
		idsy <- glabelsy$ids
		labelsy <- labelsy[idsy]
	} else {
		cogridy <- a$co.y
	}

	cex <- a$labels.size*o$scale

	# remove overlapping grid labels
	heights = text_height_npc(labelsy) * cex
	sely2 = grid_nonoverlap(cogridy, heights)
	if (!any(sely2)) return(NULL)
	labelsy2 = labelsy[sely2]
	cogridy2 = cogridy[sely2]

	spacerY <- grid::convertWidth(grid::unit(.5, "lines"), unitTo="inch", valueOnly=TRUE) * cex / W
	marginY <- grid::convertWidth(grid::unit(a$labels.margin.y, "lines"), unitTo="inch", valueOnly=TRUE) * cex / W

	if (is_left) {
		just <- ifelse(a$labels.rot[2] == 90, "bottom", ifelse(a$labels.rot[2] == 270, "top", ifelse(a$labels.rot[2] == 180, "left", "right")))
	} else {
		just <- ifelse(a$labels.rot[2] == 90, "bottom", ifelse(a$labels.rot[2] == 270, "top", ifelse(a$labels.rot[2] == 180, "right", "left")))
	}

	if (a$labels.rot[2] %in% c(90, 270)) {
		x = 0.5 - (is_left - 0.5) * 2 * (spacerY + 0.5* marginY)
	} else {
		if (is_left) {
			x = 1 - spacerY - marginY
		} else {
			x = spacerY + marginY
		}
	}

	if (a$ticks[2]) {
		if (is_left) {
			ticks <- grid::polylineGrob(x = rep(c(1-spacerY*.5-marginY, 1), length(cogridy2)), y = rep(cogridy2, each = 2), id = rep(1:length(cogridy2), each = 2), gp=gpar(col=a$col, lwd=a$lwd))
		} else {
			ticks <- grid::polylineGrob(x = rep(c(0, spacerY*.5+marginY), length(cogridy2)), y = rep(cogridy2, each = 2), id = rep(1:length(cogridy2), each = 2), gp=gpar(col=a$col, lwd=a$lwd))
		}
	} else {
		ticks <- NULL
	}

	labels = grid::textGrob(labelsy2, y=cogridy2, x=x,
							just=just, rot=a$labels.rot[2], gp=gpar(col=a$labels.col, cex=cex, fontface=a$labels.fontface, fontfamily=a$labels.fontfamily))
	res = gTree(children = gList(ticks, labels), name = "gridTicksLabelsY")


	gt = add_to_gt(gt, res, row = rowid, col = colid)

	gts[[facet_page]] = gt

	assign("gts", gts, envir = .TMAP_GRID)
	NULL


}

#' @export
#' @keywords internal
#' @rdname tmapGridLeaflet
tmapGridAuxPlot.tm_aux_graticules = function(a, bi, bbx, facet_row, facet_col, facet_page, id, pane, group, o) {
	tmapGridAuxPlot.tm_aux_grid(a, bi, bbx, facet_row, facet_col, facet_page, id, pane, group, o)
}

#' @export
#' @keywords internal
#' @rdname tmapGridLeaflet
tmapGridAuxPlot.tm_aux_grid = function(a, bi, bbx, facet_row, facet_col, facet_page, id, pane, group, o) {
	rc_text = frc(facet_row, facet_col)

	g = get("g", envir = .TMAP_GRID)

	a = g$grid_comp_per_bbx[[bi]]
	gp = list()

	fH = g$mapRowsIn[facet_row]
	fW = g$mapColsIn[facet_col]


	if (a$labels.inside_frame && any(a$ticks) && o$show.warnings) warning("Grid ticks are not supported when labels.inside_frame = TRUE", call. = FALSE)


	## might be confusing: gridx are grid lines for the x-axis, so they are vertical
	cogridx <- a$co.x
	cogridy <- a$co.y
	labelsx <- a$labels.x
	labelsy <- a$labels.y


	cex <- a$labels.size*o$scale

	selx <- (length(cogridx) > 0)
	sely <- (length(cogridy) > 0)

	# find margins due to grid labels
	if (o$frame) {
		if (o$frame.double_line) {
			fw <- (6 * grid::convertWidth(unit(1, "points"), unitTo = "inch", valueOnly = TRUE) * o$frame.lwd) / fW
			fh <- (6 * grid::convertHeight(unit(1, "points"), unitTo = "inch", valueOnly = TRUE) * o$frame.lwd) / fH
		} else {
			fw <- (grid::convertWidth(unit(1, "points"), unitTo = "inch", valueOnly = TRUE) * o$frame.lwd) / fW
			fh <- (grid::convertHeight(unit(1, "points"), unitTo = "inch", valueOnly = TRUE) * o$frame.lwd) / fH
		}
	} else {
		fw <- 0
		fh <- 0
	}



	if (a$add.labels[1]) {
		if (a$labels.rot[1] %in% c(0, 180)) {
			labelsXw <- if (selx) (max(text_height_inch(labelsx))  * cex + fh) / fH else 0
		} else {
			labelsXw <- if (selx) (max(text_width_inch(labelsx, space = FALSE))  * cex + fh) / fH else 0
		}
		spacerX <- grid::convertHeight(unit(.5, "lines"), unitTo="inch", valueOnly = TRUE) * cex / fH
		marginX <- grid::convertWidth(unit(a$labels.margin.x, "lines"), unitTo="inch", valueOnly = TRUE) * cex / fW
	} else {
		labelsXw <- spacerX <- marginX <- 0
	}


	if (a$add.labels[2]) {
		if (a$labels.rot[2] %in% c(0, 180)) {
			labelsYw <- if (sely) (max(text_width_inch(labelsy, space=FALSE))  * cex + fw) / fW else 0
		} else {
			labelsYw <- if (sely) (max(text_height_inch(labelsy, to_width = TRUE))  * cex + fw) / fW else 0
		}
		spacerY <- grid::convertWidth(grid::unit(.5, "lines"), unitTo="inch", valueOnly=TRUE) * cex / fW
		marginY <- grid::convertWidth(grid::unit(a$labels.margin.y, "lines"), unitTo="inch", valueOnly=TRUE) * cex / fW
	} else {
		labelsYw <- spacerY <- marginY <- 0
	}

	# find coordinates for projected grid labels
	if (!is.na(a$crs)) {
		if (selx) {
			glabelsx <- get_gridline_labels(lco=a$co.x[a$sel.x], xax = labelsXw + spacerX+marginX)
			cogridx <- glabelsx$cogrid
			idsx <- glabelsx$ids
			labelsx <- labelsx[idsx]
			cogridx_frst <- cogridx[sapply(1:max(idsx), function(i) which(i==idsx)[1])]
		}
		# } else {
		# 	glabelsx <- numeric(0)
		# }

		if (sely) {
			glabelsy <- get_gridline_labels(lco=a$co.y[a$sel.y], yax = labelsYw + spacerY+marginY)
			cogridy <- glabelsy$cogrid
			idsy <- glabelsy$ids
			labelsy <- labelsy[idsy]
			cogridy_frst <- cogridy[sapply(1:max(idsy), function(i) which(i==idsy)[1])]
		}

	} else {
		cogridx_frst <- cogridx
		cogridy_frst <- cogridy
	}

	# select grid labels to print
	selx2 <- if (selx) (cogridx >= labelsYw + spacerY + marginY & cogridx <= 1 - spacerY) else selx
	sely2 <- if (sely) (cogridy >= labelsXw + spacerX + marginX & cogridy <= 1 - spacerX) else sely


	# remove overlapping grid labels
	widths = text_width_inch(labelsx) * cex / fW
	heights = text_height_inch(labelsy) * cex / fH


	selx2 = selx2 & grid_nonoverlap(cogridx, widths)
	sely2 = sely2 & grid_nonoverlap(cogridy, heights)


	# select grid lines to draw
	selx <- if (selx) (cogridx_frst >= labelsYw + spacerY + marginY & cogridx_frst <= 1) else selx
	sely <- if (sely) (cogridy_frst >= labelsXw + spacerX + marginX & cogridy_frst <= 1) else sely

	# crop projected grid lines, and extract polylineGrob ingredients
	if (!is.na(a$crs)) {
		lnsList <- list(
			if (any(selx)) sf::st_multilinestring(a$co.x) else NULL,
			if (any(sely)) sf::st_multilinestring(a$co.y) else NULL
		)
		lnsSel <- !vapply(lnsList, is.null, logical(1))
		if (!any(lnsSel)) {
			grid.co.x <- numeric(0)
			grid.co.y <- numeric(0)
		} else {
			lns <- sf::st_sf(ID=c("x", "y")[lnsSel], geometry = sf::st_sfc(lnsList[lnsSel], crs = 4326)) # trick for 0-1 coordinates
			sf_bbox <- tmaptools::bb_poly(bb(c(labelsYw + spacerY + marginY, labelsXw + spacerX + marginX, 1, 1)), projection = 4326)
			lns_crop <- suppressWarnings(suppressMessages(sf::st_intersection(lns, sf_bbox)))

			# quick fix for #564
			if (!all(sf::st_geometry_type(lns_crop) == "MULTILINESTRING")) {
				lns_crop_lns = sf::st_collection_extract(lns_crop, "LINESTRING")
				stopifnot(nrow(lns_crop) == length(lns_crop_lns))
				lns_crop$geometry = sf::st_geometry(lns_crop_lns)
			}

			if (any(selx)) {
				cogridxlns <- as.data.frame(sf::st_coordinates(sf::st_geometry(lns_crop)[1])[,1:3])
				names(cogridxlns) <- c("x", "y", "ID")
			} else {
				cogridxlns <- numeric(0)
			}

			if (any(sely)) {
				cogridylns <- as.data.frame(sf::st_coordinates(sf::st_geometry(lns_crop)[sum(lnsSel)])[,1:3])
				names(cogridylns) <- c("x", "y", "ID")
			} else {
				cogridylns <- numeric(0)
			}
		}

	}

	## process x-axis grid lines and labels
	if (any(selx)) {
		cogridx2 <- cogridx_frst[selx]
		cogridx3 <- cogridx[selx2]
		labelsx <- labelsx[selx2]

		if (!a$lines) {
			grobGridX <- NULL
		} else if (is.na(a$crs)) {
			grobGridX <- grid::polylineGrob(x=rep(cogridx2, each=2), y=rep(c(labelsXw+spacerX+marginX,1), length(cogridx2)),
											id=rep(1:length(cogridx2), each=2), gp=grid::gpar(col=a$col, lwd=a$lwd))
		} else {
			grobGridX <- grid::polylineGrob(x=cogridxlns$x, y=cogridxlns$y, id=cogridxlns$ID, gp=grid::gpar(col=a$col, lwd=a$lwd))
		}

		grobGridTextX <- if (a$add.labels[1] && any(selx2)) {
			just <- ifelse(a$labels.rot[1] == 90, "right", ifelse(a$labels.rot[1] == 270, "left", ifelse(a$labels.rot[1] == 180, "bottom", "top")))

			grid::textGrob(labelsx, y=labelsXw+spacerX*.5+marginX, x=cogridx3, just=just, rot=a$labels.rot[1], gp=grid::gpar(col=a$labels.col, cex=cex, fontface=a$labels.fontface, fontfamily=a$labels.fontfamily))
		} else NULL
	} else {
		grobGridX <- NULL
		grobGridTextX <- NULL
	}


	## process y-axis grid lines and labels
	if (any(sely)) {
		cogridy2 <- cogridy_frst[sely]
		cogridy3 <- cogridy[sely2]
		labelsy <- labelsy[sely2]

		if (!a$lines) {
			grobGridY <- NULL
		} else if (is.na(a$crs)) {
			grobGridY <- grid::polylineGrob(y=rep(cogridy2, each=2), x=rep(c(labelsYw+spacerY+marginY,1), length(cogridy2)),
											id=rep(1:length(cogridy2), each=2), gp=grid::gpar(col=a$col, lwd=a$lwd))
		} else {
			grobGridY <- grid::polylineGrob(x=cogridylns$x, y=cogridylns$y, id=cogridylns$ID, gp=grid::gpar(col=a$col, lwd=a$lwd))
		}

		grobGridTextY <- if (a$add.labels[2] && any(sely2)) {
			just <- ifelse(a$labels.rot[2] == 90, "bottom", ifelse(a$labels.rot[2] == 270, "top", ifelse(a$labels.rot[2] == 180, "left", "right")))

			grid::textGrob(labelsy, x=labelsYw+spacerY*.5+marginY, y=cogridy3, just=just, rot=a$labels.rot[2], gp=grid::gpar(col=a$labels.col, cex=a$labels.size*o$scale, fontface=a$labels.fontface, fontfamily=a$labels.fontfamily))
		} else NULL
	} else {
		grobGridY <- NULL
		grobGridTextY <- NULL
	}
	res = list(treeGridLines=gTree(children=gList(grobGridX, grobGridY), name="grid_lines"),
			   treeGridLabels=gTree(children=gList(grobGridTextX, grobGridTextY), name="grid_labels"),
			   metaX=labelsYw+spacerY,
			   metaY=labelsXw+spacerX)





	grb = grid::grobTree(gList(res$treeGridLines, res$treeGridLabels))

	gts = get("gts", .TMAP_GRID)
	gt = gts[[facet_page]]

	gt_name = paste0("gt_facet_", rc_text)

	gt = grid::addGrob(gt, grb, gPath = grid::gPath(gt_name))

	gts[[facet_page]] = gt
	assign("gts", gts, envir = .TMAP_GRID)
	NULL


}





get_gridline_labels = function(lco, xax = NA, yax = NA) {
	k = length(lco)
	d = ifelse(is.na(xax), 2, 1)


	lns = sf::st_sf(geometry = sf::st_sfc(lapply(lco, function(l) {
		sf::st_linestring(l)
	})), crs = 4326) # trick for 0-1 coordinates


	if (!is.na(xax)) {
		ax = sf::st_sf(geometry = sf::st_sfc(sf::st_linestring(matrix(c(0, 1, xax, xax), nrow=2))), crs = 4326)
		#ax <- SpatialLines(list(Lines(Line(matrix(c(0, 1, xax, xax), nrow=2)), ID="base")))
	} else {
		ax = sf::st_sf(geometry = sf::st_sfc(sf::st_linestring(matrix(c(yax, yax, 0, 1), nrow=2))), crs = 4326)
		#ax <- SpatialLines(list(Lines(Line(matrix(c(yax, yax, 0, 1), nrow=2)), ID="base")))
	}
	gint = suppressMessages(sf::st_intersects(lns, ax, sparse = FALSE, prepared = FALSE))[ ,1]

	ins = vapply(lco, function(m) {
		l = m[1,]
		if (!is.na(xax)) {
			res = l[1] >= 0 && l[1] <= 1 && l[2] >= xax && l[2] <= 1
		} else {
			res = l[1] >= yax && l[1] <= 1 && l[2] >= 0 && l[2] <= 1
		}
		if (res) l[d] else -1
	}, numeric(1))
	cogrid = ifelse(gint, 0, ins)
	ids = seq_along(cogrid) # number of grid labels per grid line (could be 2 for warped grid lines)
	if (any(gint)) {
		cogrid = as.list(cogrid)
		ids = as.list(ids)

		wgint = which(gint)
		gints = suppressMessages(sf::st_intersection(lns[gint, ], ax))

		# count number of intersections per coordinate
		gnrs = vapply(sf::st_geometry(gints), function(g) {
			nr = nrow(g)
			if (is.null(nr)) 1L else nr
		}, integer(1))
		gints = suppressWarnings(sf::st_cast(x = sf::st_cast(gints, "MULTIPOINT"), to = "POINT"))

		coor = sf::st_coordinates(gints)[,d]

		ids2 = unlist(mapply(rep, 1:length(wgint), gnrs, SIMPLIFY = FALSE), use.names = FALSE)
		j = 1L
		for (i in which(gint)) {
			cogrid[[i]] = unname(coor[ids2 == j])
			ids[[i]] = rep(ids[[i]], gnrs[j])
			j = j + 1L
		}

		# remove non-selected items (#938)
		cogrid = cogrid[which(gint)]
		ids = ids[which(gint)]

		cogrid = unlist(cogrid, use.names = FALSE)
		ids = unlist(ids, use.names = FALSE)
	}
	list(cogrid = cogrid, ids = ids)
}

#' Wrapper functions for using **tmap** in **shiny**
#'
#' @description
#' * `tmapOutput()` creates a UI element
#' * `renderTmap()` renders a tmap map
#' * `tmapProxy()` updates a tmap map in `view` mode
#'
#' Adding layers is as usual via the map layer functions like [tm_polygons()].
#' Removing layers can be done, removing with `tm_remove_layer()`.
#'
#' @details
#' Two features from tmap are not (yet) supported in Shiny: small multiples (facets) and colored backgrounds (argument `bg.color` of [tm_layout()]).
#' Workarounds for small multiples: create multiple independent maps or specify `as.layers = TRUE` in [tm_facets()].
#'
#' @param expr A tmap object. A tmap object is created with [qtm()] or by stacking [`tmap-element`]s.
#' @param mode tmap mode, see [tmap_mode()] If not defined, the current mode is used
#' @param env The environment in which to evaluate expr
#' @param quoted Is `expr` a quoted expression (with `quote()`)? This is useful if you want to save an expression in a variable
#' @param outputId Output variable to read from
#' @param width,height the width and height of the map
#' @param mapId single-element character vector indicating the output ID of the map to modify (if invoked from a Shiny module, the namespace will be added automatically)
#' @param session the Shiny session object to which the map belongs; usually the default value will suffice
#' @param x the tmap object that specifies the added and removed layers.
#' @param zindex the z index of the pane in which the layer is contained that is going to be removed. It is recommended to specify the `zindex` for this layer when creating the map (inside `renderTmap()`).
#' @param execOnResize If `TRUE` (default), when the plot is resized, the map is regenerated. When set to `FALSE` the map is rescaled: the aspect ratio is kept, but the layout will be less desirable.
#' @importFrom htmlwidgets shinyWidgetOutput
#' @example ./examples/tmapOutput.R
#' @export
renderTmap <- function(expr, env = parent.frame(), quoted = FALSE, execOnResize = TRUE, mode = NA) {
	if (is.na(mode)) mode = getOption("tmap.mode")
	gs = tmap_graphics_name(mode)


	if (gs == "Grid") {
		expr = substitute(getFromNamespace("print.tmap", "tmap")(expr))
		shiny::renderPlot(expr, env = env, quoted = TRUE, execOnResize = execOnResize)
	} else if (gs == "Leaflet") {
		if (!quoted)
			expr = substitute(expr)
		expr = substitute(getFromNamespace("print.tmap", "tmap")(expr, in.shiny = TRUE))
		htmlwidgets::shinyRenderWidget(expr, leafletOutput, env,
									   quoted = TRUE)



		# fun = paste0("renderTmap", gs)
		# do.call(fun, list(expr = expr, env = env, quoted = quoted, execOnResize = execOnResize))
	} else {
		stop('renderTmap does not work yet for modes other than "plot" and "view"')
	}

}


#' @rdname renderTmap
#' @export
tmapOutput <- function(outputId, width = "100%", height = 400, mode = NA) {
	if (is.na(mode)) mode = getOption("tmap.mode")
	gs = tmap_graphics_name(mode)
	if (gs == "Grid") {
		shiny::plotOutput(outputId = outputId, width = width, height = height)
	} else {
		fun = paste0("tmapOutput", gs)
		do.call(fun, list(outputId = outputId, width = width, height = height))
	}
}


#' @rdname renderTmap
#' @export
tmapProxy <- function(mapId, session = shiny::getDefaultReactiveDomain(), x, mode = NA) {
	if (is.na(mode)) mode = getOption("tmap.mode")
	gs = tmap_graphics_name(mode)

	if (gs == "Grid") {
		message("tmapProxy not working for plot mode (yet)")
		print.tmap(x, show = TRUE)
	} else {
		#fun = paste0("tmapProxy", gs)
		#do.call(fun, list(mapId = mapId, session = session, x = x))
		print.tmap(x, lf = leaflet::leafletProxy(mapId, session), show = FALSE, in.shiny = TRUE, proxy = TRUE)
	}

}


#' @rdname renderTmap
#' @export
tm_remove_layer <- function(zindex) {
	tm_element_list(tm_element(zindex = zindex,
							   subclass = c("tm_proxy", "tm_remove_layer")))
}

#' Legend charts
#'
#' Legend charts are small charts that are added to the map, usually in addition to legends.
#'
#' Note that these charts are different from charts drawn inside the map. Those are called glyphs (to be implemented).
#'
#' @param breaks The breaks of the bins (for histograms)
#' @param plot.axis.x,plot.axis.y Should the x axis and y axis be plot?
#' @param extra.ggplot2 Extra ggplot2 code
#' @inheritParams tm_title
#' @example examples/tm_chart.R
#' @seealso `r .doc_see_also_chart()`
#' @name tm_chart
#' @export
tm_chart_histogram = function(breaks,
							  plot.axis.x,
							  plot.axis.y,
							  extra.ggplot2,
							  position,
							  group_id,
							  width,
							  height,
							  stack,
							  z,
							  ...) {

	args = lapply(as.list(rlang::call_match()[-1]), eval, envir = parent.frame())
	args = warning_group_args(args)

	args$z = args$z %||% NA_integer_
	args$group_id = args$group_id %||% NA_character_
	args$group_type = "tm_chart"

	args$show = TRUE
	args$type = "histogram"
	args$summary = "binned"
	structure(args, class = c("tm_chart_histogram", "tm_chart", "tm_component", "list"))
}

#' @rdname tm_chart
#' @export
tm_chart_bar = function(plot.axis.x,
						plot.axis.y,
						extra.ggplot2,
						position,
						group_id,
						width,
						height,
						stack,
						z,
						...) {
	args = lapply(as.list(rlang::call_match()[-1]), eval, envir = parent.frame())
	args = warning_group_args(args)

	args$z = args$z %||% NA_integer_
	args$group_id = args$group_id %||% NA_character_
	args$group_type = "tm_chart"

	args$show = TRUE
	args$type = "bar"
	args$summary = "binned"
	structure(args, class = c("tm_chart_bar", "tm_chart", "tm_component", "list"))
}


#' @rdname tm_chart
#' @export
tm_chart_donut = function(position,
						  group_id,
						  width,
						  height,
						  stack,
						  z,
						  ...) {
	args = lapply(as.list(rlang::call_match()[-1]), eval, envir = parent.frame())
	args = warning_group_args(args)


	args$z = args$z %||% NA_integer_
	args$group_id = args$group_id %||% NA_character_
	args$group_type = "tm_chart"

	args$show = TRUE
	args$type = "donut"
	args$summary = "binned"
	structure(args, class = c("tm_chart_donut", "tm_chart", "tm_component", "list"))
}

#' @rdname tm_chart
#' @export
tm_chart_violin = function(position,
						   group_id,
						   width,
						   height,
						   stack,
						   z,
						   ...) {
	args = lapply(as.list(rlang::call_match()[-1]), eval, envir = parent.frame())
	args = warning_group_args(args)

	args$z = args$z %||% NA_integer_
	args$group_id = args$group_id %||% NA_character_
	args$group_type = "tm_chart"

	args$show = TRUE
	args$type = "violin"
	args$summary = "raw_nna"
	structure(args, class = c("tm_chart_violin", "tm_chart", "tm_component", "list"))
}

#' @rdname tm_chart
#' @export
tm_chart_box = function(position,
						group_id,
					    width,
					    height,
					    stack,
					    z,
						...) {
	args = lapply(as.list(rlang::call_match()[-1]), eval, envir = parent.frame())
	args = warning_group_args(args)

	args$z = args$z %||% NA_integer_
	args$group_id = args$group_id %||% NA_character_
	args$group_type = "tm_chart"

	args$show = TRUE
	args$type = "box"
	args$summary = "raw_nna"
	structure(args, class = c("tm_chart_box", "tm_chart", "tm_component", "list"))
}

#' @rdname tm_chart
#' @export
tm_chart_none = function() {
	structure(list(show = FALSE, group_id = NA_character_, z = NA_integer_, summary = "none"), class = c("tm_chart_none", "tm_chart", "tm_component", "list"))
}



#' @rdname tm_chart
#' @export
tm_chart_heatmap = function(position,
							group_id,
							width,
							height,
							stack,
							z,
							...){
	args = lapply(as.list(rlang::call_match()[-1]), eval, envir = parent.frame())
	args = warning_group_args(args)

	args$z = args$z %||% NA_integer_
	args$group_id = args$group_id %||% NA_character_
	args$group_type = "tm_chart"

	args$show = TRUE
	args$type = "heatmap"
	args$summary = "binned2D"
	structure(args, class = c("tm_chart_heatmap", "tm_chart", "tm_component", "list"))
}

tmapGridRun = function(o, q, show, knit, knit_opts, args) {
	gts = get("gts", .TMAP_GRID)

	if (show) {
		if (o$show_gif_ani) {
			d <- paste(tempdir(), "/tmap_plots", sep="/")
			dir.create(d, showWarnings = FALSE)
			devsize = c(grid::convertWidth(grid::unit(1, "npc"), unitTo = "inch", valueOnly = TRUE) * 72 * o$dpr,
						grid::convertHeight(grid::unit(1, "npc"), unitTo = "inch", valueOnly = TRUE) * 72 * o$dpr)
		}

		mapply(function(gt,i) {
			if (o$show_gif_ani) {
				png(paste0(d, "/plot", sprintf("%03d", i), ".png"), width = devsize[1], height = devsize[2], res = 72*o$dpr, type = "cairo-png")
				grid::grid.newpage()
			} else {
				if (is.null(o$vp) && i != 1L) grid::grid.newpage()
			}

			withCallingHandlers(
				tryCatch({
					grid::grid.draw(gt)
				}, error = function(e) {
					msg <- conditionMessage(e)
					stop(msg, "\nIf this is a spatial data issue, try adding + tm_check_fix()", call. = FALSE)
				}),
				warning = function(w) {
					if (grepl("Unknown group", conditionMessage(w))) {
						warning("Layer blending (blend) is not supported by the current graphics device on this run. Try re-plotting.", call. = FALSE)
						invokeRestart("muffleWarning")
					}
				}
			)

			if (o$show_gif_ani) {
				dev.off()
			}

		}, gts, seq_along(gts), SIMPLIFY = FALSE)
	}
	if (!is.null(o$vp)) upViewport(1)
	if (length(gts) == 1) gts = gts[[1]]

	if (show && o$show_gif_ani) {
		files = list.files(path = d, pattern = "^plot[0-9]{3}\\.png$", full.names = TRUE)
		if (o$play == "pingpong") files = c(files, rev(files))
		filename = tempfile(fileext = ".gif")
		create_animation(filename = filename, files = files, width = devsize[1], height = devsize[2], delay = NA, fps = o$fps, loop = o$play != "once", progress = FALSE, gif = TRUE, showAni = TRUE, dpr = o$dpr, knit = knit, knit_opts = knit_opts)
		if (knit && !knitr::is_latex_output()) {
			knitr_path <- knitr::fig_path('.gif', knit_opts)
			dir.create(dirname(knitr_path), showWarnings = FALSE, recursive = TRUE)
			file.copy(filename, knitr_path, overwrite = TRUE)
			if (is.null(knit_opts$out.width)) {
				knit_opts$out.width <- knit_opts$fig.width * knit_opts$dpi / (knit_opts$fig.retina %||% 1)
			}
			return(do.call(knitr::knit_print, c(list(x = knitr::include_graphics(knitr_path), options = knit_opts), args)))
		}
	}

	gts
}

tmapScaleDiscrete = function(x1, scale, legend, chart, o, aes, layer, layer_args, sortRev, bypass_ord, submit_legend = TRUE) {
	cls = data_class(x1, midpoint_enabled = !is.null(scale$midpoint))
	maincls = class(scale)[1]


	#if (cls[1] == "na") stop("data contain only NAs, so tm_scale_discrete cannot be applied", call. = FALSE)
	#if (cls[1] != "num") stop("tm_scale_discrete can only be used for numeric data", call. = FALSE)

	if (cls[1] != "num") {
		if (!is.factor(x1)) x1 = as.factor(x1)
		x1 = as.integer(x1)
		warning(maincls, " is supposed to be applied to numerical data", call. = FALSE)
	}

	x1 = without_units(x1)

	if (aes %in% c("lty", "shape", "pattern")) stop("tm_scale_discrete cannot be used for layer ", layer, ", aesthetic ", aes, call. = FALSE)

	scale = get_scale_defaults(scale, o, aes, layer, cls)

	show.messages <- o$show.messages
	show.warnings <- o$show.warnings

	if (all(is.na(x1))) return(tmapScale_returnNA(n = length(x1), legend = legend, chart = chart, value.na = scale$value.na, label.na = scale$label.na, label.show = scale$label.show, na.show = legend$na.show, sortRev = sortRev, bypass_ord = bypass_ord))

	with(scale, {

		u = sort(unique(x1))
		#u = na.omit(u)
		rng = range(u)

		if (!is.na(ticks[1])) {
			if (!all(u %in% ticks)) stop("Values have been found for which no ticks have been specified", call. = FALSE)
		} else {
			for (n in c(1, 2, 3, 5, 10, 15, 20, 30, 50)) {
				ticks_candidate = pretty(rng, n = n)

				found_ticks = all(u %in% ticks_candidate)
				if (found_ticks) {
					ticks = ticks_candidate
					break
				}
			}
			if (!found_ticks) stop("Unable to fit a discrete scale, probably due to too many values", call. = FALSE)
		}

		n = length(ticks)

		d_isdiv = rng[1] < 0 && rng[2] > 0

		check_values(layer, aes, values)

		fun_isdiv = paste0("tmapValuesIsDiv_", aes)

		isdiv = !is.null(midpoint) || do.call(fun_isdiv, args = list(x = values))


		# determine midpoint
		if ((is.null(midpoint) || is.na(midpoint)) && isdiv) {
			if (d_isdiv && is.null(midpoint)) {

				if (show.messages) message_midpoint(aes, "tm_scale_discrete")
				midpoint = 0
			} else {
				if ((n %% 2) == 1) {
					# number of classes is odd, so take middle class (average of those breaks)
					midpoint <- mean.default(ticks[c((n+1) / 2, (n+3) / 2)])
				} else {
					midpoint <- ticks[(n+2) / 2]
				}
			}
		}

		# update range if NA (automatic)
		if (is.na(values.range[1])) {
			fun_range = paste0("tmapValuesRange_", aes)
			values.range = do.call(fun_range, args = list(x = values, n = n, isdiv = isdiv))
		}
		if (length(values.range) == 1 && !is.na(values.range[1])) values.range = c(0, values.range)


		fun_getVV = paste0("tmapValuesVV_", aes)
		VV = do.call(fun_getVV, list(x = values, value.na = value.na, isdiv = isdiv, n = n, dvalues = ticks, are_breaks = FALSE, midpoint = midpoint, range = values.range, scale = values.scale * o$scale, rep = values.repeat, o = o))

		vvalues = VV$vvalues
		value.na = VV$value.na

		sfun = paste0("tmapValuesScale_", aes)
		cfun = paste0("tmapValuesColorize_", aes)
		if (is.na(value.neutral)) value.neutral = VV$value.neutral else value.neutral = do.call(sfun, list(x = do.call(cfun, list(x = value.neutral, pc = o$pc)), scale = values.scale))

		mfun = paste0("tmapValuesSubmit_", aes)
		vvalues = do.call(mfun, list(x = vvalues, args = layer_args))
		value.na = do.call(mfun, list(x = value.na, args = layer_args))
		value.neutral = do.call(mfun, list(x = value.neutral, args = layer_args))


		ids = match(x1, ticks)
		vals = vvalues[ids]

		isna = is.na(vals)
		anyNA = any(isna)

		na.show = update_na.show(label.show, legend$na.show, anyNA)

		if (is.null(sortRev)) {
			ids = NULL
		} else if (is.na(sortRev)) {
			ids[] = 1L
		} else if (sortRev) {
			ids = (as.integer(n) + 1L) - ids
		}

		if (anyNA) {
			vals[isna] = value.na
			if (!is.null(sortRev)) ids[isna] = 0L
		}


		if (is.null(labels)) {
			labels = do.call("fancy_breaks", c(list(vec=ticks, as.count = FALSE, intervals=FALSE), label.format))
		} else {
			if (length(labels)!=n && show.warnings) warning("number of legend labels should be ", n, call. = FALSE)
			labels = rep(labels, length.out=n)
			attr(labels, "align") <- label.format$text.align
		}

		if (legend$reverse) {
			#labels.brks <- attr(labels, "brks")
			labels.align <- attr(labels, "align")
			labels <- rev(labels)
			# if (!is.null(labels.brks)) {
			# 	attr(labels, "brks") = labels.brks[length(labels):1L,]
			# }
			attr(labels, "align") = labels.align
			vvalues = rev(vvalues)
		}



		if (na.show) {
			#labels.brks = attr(labels, "brks")
			labels.align = attr(labels, "align")
			#labels <- c(labels, legend.NA.text)
			# if (!is.null(labels.brks)) {
			# 	labels <- c(labels, paste(label.na, " ", sep = ""))
			# 	attr(labels, "brks") = rbind(labels.brks, rep(nchar(label.na) + 2, 2))
			# } else {
			labels = c(labels, label.na)

			attr(labels, "align") = labels.align
			vvalues = c(vvalues, value.na)
		}

		# temp solution to include NA, to do update num2pal
		# if (length(labels) == length(vvalues)) {
		# 	la = attributes(labels)
		# 	labels = c(labels, label.na)
		# 	attributes(labels) = la
		# 	values = c(values, NA)
		# }
		#

		#type = ifelse(nchar(vvalues[1]) > 50, "color_cont", "color_cls")

		legend = within(legend, {
			nitems = length(labels)
			labels = labels
			dvalues = ticks
			vvalues = vvalues
			vneutral = value.neutral
			na.show = get("na.show", envir = parent.env(environment()))
			scale = "discrete"
			layer_args = layer_args
		})

		chartFun = paste0("tmapChart", toTitleCase(chart$summary))

		chart = do.call(chartFun, list(chart,
									   bin_colors = vvalues,
									   breaks_def = NULL,
									   na.show = na.show,
									   x1 = x1))



		if (submit_legend) {
			if (bypass_ord) {
				format_aes_results(vals, legend = legend, chart = chart)
			} else {
				format_aes_results(vals, ids, legend, chart = chart)
			}
		} else {
			list(vals = vals, ids = ids, legend = legend, chart = chart, bypass_ord = bypass_ord)
		}
	})
}

tmapGridCompPrepare_text = function(comp, o) {
	if (is.null(comp$text)) {
		cmp = within(comp, {
			show = FALSE
		})
		return(cmp)
	}
	n = length(comp$text)
	lst = lapply(seq_len(n), function(i){
		within(comp, {
			color[is.na(color)] = o$attr.color
			color = do.call("process_color", c(list(col=color), o$pc))
			size = size * o$scale
			if (!is.numeric(fontface)) fontface[is.na(fontface)] = o$text.fontface
			fontfamily[is.na(fontfamily)] =o$text.fontfamily
			#text = lapply(text, rep, length.out=o$n)
			text = text[i]
			if (is.na(text)) text = ""
			show = nonempty_text(text)
		})
	})
	class(lst) = c("tm_multi_comp", "list")
	lst
}


#' @export
tmapGridCompPrepare.tm_title = function(comp, o) {
	tmapGridCompPrepare_text(comp, o)
}

#' @export
tmapGridCompHeight.tm_title = function(comp, o) {
	tmapGridCompHeight_text(comp, o)
}

#' @export
tmapGridCompWidth.tm_title = function(comp, o) {
	tmapGridCompWidth_text(comp, o)
}

#' @export
tmapGridCompPlot.tm_title = function(comp, o, fH, fW) {
	tmapGridCompPlot_text(comp, o)
}

#' @export
tmapGridCompPrepare.tm_credits = function(comp, o) {
	tmapGridCompPrepare_text(comp, o)
}

#' @export
tmapGridCompHeight.tm_credits = function(comp, o) {
	tmapGridCompHeight_text(comp, o)
}

#' @export
tmapGridCompWidth.tm_credits = function(comp, o) {
	tmapGridCompWidth_text(comp, o)
}

#' @export
tmapGridCompPlot.tm_credits = function(comp, o, fH, fW) {
	tmapGridCompPlot_text(comp, o, fH, fW)
}

#' @export
tmapGridCompPrepare.tm_compass = function(comp, o) {
	typs = c("arrow", "4star", "8star", "radar", "rose")
	within(comp, {

		if (!(type %in% typs)) cli::cli_abort("{.field [tm_compass]} wrong compass type, should be one of: {.str {typs}}")
		text.size = text.size * o$scale

		asp = if (type == "arrow") 0.5 else 1

		show = TRUE
		if (is.na(type)) type = o$type
		if (is.na(size)) size = switch(type, arrow=2, radar=6, rose=6, 4)
		nlines = size + ifelse(show.labels==0, 0, ifelse(show.labels==1, 1, 2))
	})
}

#' @export
tmapGridCompHeight.tm_compass = function(comp, o) {


	textS = comp$text.size #* o$scale
	#textP = comp$padding[c(3,1)] * textS * o$lin

	marH = comp$margins[c(3,1)] * textS * o$lin
	hs = c(marH[1], comp$nlines * textS * o$lin, marH[2])


	sides = switch(comp$position$align.v, top = "second", bottom = "first", "both")
	hsu = set_unit_with_stretch(hs, sides = sides)

	Hin = sum(hs)
	comp$flexRow = NA
	comp$Hin = Hin #  sum(textP[1], textH, textP[2])
	comp$hsu = hsu
	comp
}

#' @export
tmapGridCompWidth.tm_compass = function(comp, o) {



	textS = comp$text.size #* o$scale
	#textP = comp$padding[c(3,1)] * textS * o$lin


	marW = comp$margins[c(2,4)] * textS * o$lin
	ws = c(marW[1], comp$nlines * textS * o$lin * comp$asp, marW[2])

	sides = switch(comp$position$align.h, left = "second", right = "first", "both")
	wsu = set_unit_with_stretch(ws, sides = sides)

	comp$flexCol = NA
	comp$Win = sum(ws)
	comp$wsu = wsu
	comp
}

#' @export
tmapGridCompPlot.tm_compass = function(comp, o, fH, fW) {

	u = 1/(comp$nlines)
	#vpComp = viewport(x=u, y=u, height=1-2*u, width=1-2*u, just=c("left", "bottom"))

	light = comp$color.light
	dark = comp$color.dark


	wsu = comp$wsu
	hsu = comp$hsu

	vp = grid::viewport(layout = grid::grid.layout(ncol = length(wsu),
												   nrow = length(hsu),
												   widths = wsu,
												   heights = hsu))


	if (comp$type=="4star") {
		s = c(.5, .5, .57, .5, .5, .43, 0, .5, .43, 1, .5, .57)
		x = list(rep.int(s, 2))
		y = list(s[c(10:12, 10:12, 1:3, 1:3, 7:9, 7:9, 4:6, 4:6)])
		id = rep(1:8, each=3)
		fill = c(dark, light, dark, light, light, dark, light, dark)
	} else if (comp$type=="8star") {
		s = c(.5, .5, .56, .5, .5, .44, 0, .5, .38, 1, .5, .62)
		s2 = c(.5, .62, .7, .5, .56, .7, .5, .38, .3, .5, .44, .3)
		x = list(c(rep.int(s, 2), rep.int(s2, 2)))
		y = list(c(s[c(10:12, 10:12, 1:3, 1:3, 7:9, 7:9, 4:6, 4:6)], s2[c(4:6, 1:3, 10:12, 7:9, 10:12, 7:9, 4:6, 1:3)]))
		id = rep(1:16, each=3)
		fill = c(dark, light, dark, light, light, dark, light, dark)
	} else if (comp$type=="arrow") {
		x = list(c(.5, .9, .5, .5, .1, .5))
		y = list(c(1, 0, .2, 1, 0, .2))
		id = rep(1:2, each=3)
		fill = c(dark, light)
	} else if (comp$type=="radar") {
		cr = c(.45, .42, .2, .17, .1)
		LWD = (o$lineH * 24) * comp$lwd

		cd = seq(1/8, 15/8, by=.25) * pi
		cd2 = seq(1/4, 7/4, by=.5) * pi
		cd3 = seq(0, 1.75, by=.25) * pi

		x = list(.5,
				  unlist(lapply(.5 + sin(cd) * cr[1], c, .5), use.names = FALSE),
				  .5 + c(0, cr[1]-.005, 0, -cr[1]+.005, 0, 0, 0, 0),
				  unlist(lapply(.5 + sin(cd2) * cr[1], c, .5), use.names = FALSE),
				  .5 + unlist(mapply(c, sin(cd3) * cr[4], sin(cd3) * cr[5], SIMPLIFY=FALSE), use.names = FALSE))

		y = list(.5,
				  unlist(lapply(.5 + cos(cd) * cr[1], c, .5), use.names = FALSE),
				  .5 + c(0, 0, 0, 0, 0, cr[1]-.005, 0, -cr[1]+.005),
				  unlist(lapply(.5 + cos(cd2) * cr[1], c, .5), use.names = FALSE),
				  .5 + unlist(mapply(c, cos(cd3) * cr[4], cos(cd3) * cr[5], SIMPLIFY=FALSE), use.names = FALSE))

	} else if (comp$type=="rose") {
		cr = c(.45, .42, .2, .17, .1)
		LWD = (o$lineH * 24) * comp$lwd
		cd = seq(1/8, 15/8, by=.25) * pi
		cd2 = seq(1/4, 7/4, by=.5) * pi
		cd3 = seq(0, 1.75, by=.25) * pi

		b = cr[4]
		a = 0.4142136 * b # 1/16th circleL
		s = c(.5, .5, .5+a, .5, .5, .5-a, 0, .5, .5-b, 1, .5, .5+b)
		s2 = c(.5, .5+b, .78, .5, .5+a, .78, .5, .5-b, .22, .5, .5-a, .22)

		id = rep(1:16, each=3)
		fill = c(dark, light, dark, light, light, dark, light, dark)


		x = list(.5,
				  unlist(lapply(.5 + sin(cd) * cr[1], c, .5), use.names = FALSE),
				  .5 + unlist(mapply(c, sin(cd3) * cr[4], sin(cd3) * cr[5], SIMPLIFY=FALSE), use.names = FALSE),
				  c(rep.int(s, 2), rep.int(s2, 2)))

		y = list(.5,
				  unlist(lapply(.5 + cos(cd) * cr[1], c, .5), use.names = FALSE),
				  .5 + unlist(mapply(c, cos(cd3) * cr[4], cos(cd3) * cr[5], SIMPLIFY=FALSE), use.names = FALSE),
				  c(s[c(10:12, 10:12, 1:3, 1:3, 7:9, 7:9, 4:6, 4:6)], s2[c(4:6, 1:3, 10:12, 7:9, 10:12, 7:9, 4:6, 1:3)]))

	}


	# rescale
	resc = function(a) (a-.5)*(comp$size/comp$nlines) + .5

	x = lapply(x, resc)
	y = lapply(y, resc)
	if (comp$type %in% c("radar", "rose")) cr = cr * (comp$size/comp$nlines)

	if (comp$north!=0) {
		drotate = comp$north/180*pi - .5*pi

		xy = mapply(function(a,b){
			d = atan2(b-.5, a-.5)
			r = sqrt((a-.5)^2 + (b-.5)^2)

			list(x=r * sin(d+drotate) + .5,
				 y=r * cos(d+drotate) + .5)
		}, x, y, SIMPLIFY=FALSE)
		x = lapply(xy, "[[", 1)
		y = lapply(xy, "[[", 2)
	} else drotate = -.5*pi


	# shift compass to south direction
	if (comp$show.labels==1) {
		x = lapply(x, function(a) a - (u/2) * sin(drotate + .5*pi))
		y = lapply(y, function(b) b - (u/2) * cos(drotate + .5*pi))
	}



	grobBG = if (getOption("tmap.design.mode")) rectGrob(gp=gpar(fill="#CAB2D6")) else NULL

	grobLabels = if (comp$show.labels==0) {
		NULL
	} else {
		selection = if (comp$show.labels==1) {
			c(TRUE, rep.int(FALSE, 7))
		} else if (comp$show.labels==2) {
			rep.int(c(TRUE, FALSE), 4)
		} else rep.int(TRUE, 8)

		labels = comp$cardinal.directions[c(1, 1, 2, 3, 3, 3, 4, 1)]
		labels[c(2,4,6,8)] = paste(labels[c(2,4,6,8)], comp$cardinal.directions[c(2, 2, 4, 4)], sep="")
		labels = labels[selection]

		lr = (1-u)/2
		ld = (seq(0, 1.75, by=.25) * pi)[selection]

		lx = lr * sin(ld+drotate + .5*pi) + .5
		ly = lr * cos(ld+drotate + .5*pi) + .5
		textGrob(labels, x=lx, y=ly, just=c("center", "center"), rot=0, gp=gpar(col=comp$text.color, cex=comp$text.size, fontface=o$text.fontface, fontfamily=o$text.fontfamily)) # -drotate/pi*180 - 90
	}

	grobComp = if (comp$type %in% c("arrow", "4star", "8star")) {
		polygonGrob(x=x[[1]], y=y[[1]], id=id, gp=gpar(fill=fill, lwd=comp$lwd, col=dark))
	} else if (comp$type=="radar") {
		gTree(children = gList(
			circleGrob(x=x[[1]], y=y[[1]], r = cr[1], gp=gpar(lwd=2*LWD, col=dark, fill=light)),
			polylineGrob(x=x[[2]], y=y[[2]], id=rep(1:8, each=2), gp=gpar(lwd=1*LWD, col=dark)),
			polylineGrob(x=x[[3]], y=y[[3]], id=rep(1:4, each=2), gp=gpar(lwd=2*LWD, col=dark)),
			polylineGrob(x=x[[4]], y=y[[4]], id=rep(1:4, each=2), gp=gpar(lwd=1*LWD, col=dark)),
			circleGrob(x=x[[1]], y=y[[1]], r = cr[2], gp=gpar(lwd=1*LWD, col=dark, fill=NA)),
			circleGrob(x=x[[1]], y=y[[1]], r = cr[3], gp=gpar(lwd=2*LWD, col=dark, fill=light)),
			circleGrob(x=x[[1]], y=y[[1]], r = cr[4], gp=gpar(lwd=1*LWD, col=NA, fill=dark)),
			circleGrob(x=x[[1]], y=y[[1]], r = cr[5], gp=gpar(lwd=1*LWD, col=NA, fill=light)),
			polylineGrob(x=x[[5]], y=y[[5]], id=rep(1:8, each=2), gp=gpar(lwd=2*LWD, col=light))))
	} else if (comp$type=="rose") {
		gTree(children = gList(
			circleGrob(x=x[[1]], y=y[[1]], r = cr[1], gp=gpar(lwd=2*LWD, col=dark, fill=light)),
			polygonGrob(x=x[[4]], y=y[[4]], id=id, gp=gpar(lwd=1*LWD, col=dark, fill=fill)),
			polylineGrob(x=x[[2]], y=y[[2]], id=rep(1:8, each=2), gp=gpar(lwd=1*LWD, col=dark)),
			circleGrob(x=x[[1]], y=y[[1]], r = cr[2], gp=gpar(lwd=1*LWD, col=dark, fill=NA)),
			circleGrob(x=x[[1]], y=y[[1]], r = cr[3], gp=gpar(lwd=2*LWD, col=dark, fill=light)),
			circleGrob(x=x[[1]], y=y[[1]], r = cr[4], gp=gpar(lwd=1*LWD, col=NA, fill=dark)),
			circleGrob(x=x[[1]], y=y[[1]], r = cr[5], gp=gpar(lwd=1*LWD, col=NA, fill=light)),
			polylineGrob(x=x[[3]], y=y[[3]], id=rep(1:8, each=2), gp=gpar(lwd=2*LWD, col=light))))
	}


	# other grid cells are aligns (1 and 5) and margins (2 and 4)
	compass = gridCell(3,3, {
		gTree(children=gList(grobBG,
						 grobComp,
						 grobLabels),
		  name="compass")
	})
	grid::grobTree(compass, vp = vp)
}






#' @export
tmapGridCompPrepare.tm_scalebar = function(comp, o) {
	show.messages = o$show.messages
	show.warnings = o$show.warnings
	within(comp, {
		if (all(c("breaks", "width") %in% call) && show.warnings) {
			cli::cli_inform("For {.fun tm_scalebar}, {.arg breaks} and {.arg width} are not supposed to be used together; normally, setting the exact width is not needed when breaks have been specified, only for fine-tuning")
		}
		if ("breaks" %in% call) {
			if (breaks[1] != 0) {
				if (show.warnings) warning("First scalebar breaks value should be 0.", call. = FALSE)
				breaks = c(0, breaks)
			}
		}

		# if (is.na(width))
		# 	width = .25
		# else if (width > 1) {
		# 	if (show.messages) message("Scale bar width set to 0.25 of the map width")
		# 	width = .25
		# }

		if (is.na(text.color)) text.color = o$attr.color
		text.size = text.size * o$scale
		lwd = lwd * o$scale
		show = TRUE
	})
}

#' @export
tmapGridCompHeight.tm_scalebar = function(comp, o) {
	h = 2.75 * o$lin * comp$text.size

	textS = comp$text.size #* o$scale
	#textP = comp$padding[c(3,1)] * textS * o$lin


	marH = comp$margins[c(3,1)] * textS * o$lin
	hs = c(marH[1], h, marH[2])


	sides = switch(comp$position$align.v, top = "second", bottom = "first", "both")
	hsu = set_unit_with_stretch(hs, sides = sides)

	Hin = sum(hs)
	comp$flexRow = NA
	comp$Hin = Hin #  sum(textP[1], textH, textP[2])
	comp$hsu = hsu
	comp
}

#' @export
tmapGridCompWidth.tm_scalebar = function(comp, o) {
	#w = comp$width * o$lin * comp$text.size

	textS = comp$text.size #* o$scale
	#textP = comp$padding[c(3,1)] * textS * o$lin

	marW = comp$margins[c(2,4)] * textS * o$lin

	W = comp$width * textS * o$lin

	ws = c(marW[1], W, marW[2])

	sides = switch(comp$position$align.h, left = "second", right = "first", "both")
	wsu = set_unit_with_stretch(ws, sides = sides)


	comp$Win = sum(ws)
	comp$wsu = wsu

	# in case breaks are used: adjust the legend width later (in tmapGridComp)
	comp$WnativeID = 3
	if (!is.null(comp$breaks)) {
		comp$WnativeRange = tail(comp$breaks, 1) - comp$breaks[1]# + (comp$breaks[2] - comp$breaks[1]) * 2
		#comp$Wextra_text_inch = text_width_inch(paste0("  ", cmp$units$unit)) + text_width_inch(paste0(tail(comp$breaks, 1), comp$breaks[1])) / 2
	}

	comp
}

#' @exportS3Method tmapGridCompPlot tm_scalebar
#' @export
tmapGridCompPlot.tm_scalebar = function(comp, o, fH, fW) {
	light = comp$color.light
	dark = comp$color.dark

	if (any(is.na(comp$dst))) {
		cli::cli_inform("{.fun tm_scalebar} the scalebar cannot be plotted because the crs does not have measurable units")
		return(NULL)
	}

	wsu = comp$wsu
	hsu = comp$hsu

	vp = grid::viewport(layout = grid::grid.layout(ncol = length(wsu),
												   nrow = length(hsu),
												   widths = wsu,
												   heights = hsu))


	unit = comp$units$unit
	#unit.size = 1/comp$units$to

	#xrange = fW * comp$cpi
	xrange = fW * (comp$dst[comp$which_side_of_map] / comp$bb_facet)

	# xrange is the range of the viewport in terms of coordinates
	# xrange2 is the same but with units (e.g. km instead of m)
	# W is the targeted space for the scalebar


	W = as.numeric(wsu[3])

	crop_factor = W / fW
	just = 0

	compare_scale_pos = function(x) {
		curr = comp$dst[x]
		oth = comp$dst[names(comp$dst) != x]
		lapply(1:length(oth), function(i) {
			wh = names(oth)[i]
			o = oth[i]
			if (o < curr) {
				sign = "large"
				scale = round(curr/o, 1)
			} else {
				sign = "large"
				scale = round(o/curr, 1)
			}
			loc = paste0("at the ", ifelse(wh == "mid", "center", wh))
			list(sign = sign, scale = scale, loc = loc)
		})
	}




	if (o$show.warnings) {
		res = compare_scale_pos(comp$which_side_of_map)

		tms = vapply(res, function(r)r$scale, FUN.VALUE = numeric(1))
		if (any(tms > 1.1)) {
			cli::cli_warn("The scale bar would be {res[[1]]$scale} times as {res[[1]]$sign} when plotted {res[[1]]$loc} and {res[[2]]$scale} times as {res[[2]]$sign} when plotted {res[[2]]$loc}")
		}
	}

	xrange2 = as.numeric(units::set_units(xrange, unit, mode = "standard"))#xrange/unit.size

	if (is.null(comp$breaks)) {
		# determine resolution only (unselect steps that do not fit later (with 'sel'))
		for (i in 10:1) {
			tcks = pretty(c(0, xrange2*crop_factor), i)
			tcks3 = (tcks / xrange2) * fW
			tcksL = format(tcks, trim=TRUE)
			labW = text_width_inch(tcksL) * comp$text.size
			tickW = tcks3[-1] - head(tcks3, -1)
			if (all(tickW > labW[-1])) {
				sbW = W - labW
				break
			}
		}
		ticks2 = tcks
	} else {
		ticks2 = comp$breaks
		tcksL = format(ticks2, trim=TRUE)

		labW = text_width_inch(tcksL) * comp$text.size
		sbW = W - labW
	}

	ticks3 = ticks2 / xrange2 * fW
	sel = which(ticks3 <= sbW)

	if (!is.null(comp$breaks) && length(sel) != length(ticks3)) {
		cli::cli_warn("{.field tm_scalebar} Not all scale bar breaks could be plotted. Try increasing the scale bar width or decreasing the font size")
	}

	ticks3 = ticks3[sel]
	ticks2 = ticks2[sel]

	ticks2Labels = format(ticks2, trim=TRUE)
	ticksWidths = text_width_inch(ticks2Labels)
	unitWidth = text_width_inch(unit)

	labels = c(ticks2Labels, unit)
	labelsW = c(ticksWidths, unitWidth)

	n = length(ticks2)

	if (length(ticks3) == 1) {
		cli::cli_warn("{.field tm_scalebar} Only one scale bar break could be plotted. Try decreasing the font size or increasing the scale bar width")
		widths = sbW
	} else {
		widths = ticks3[2:n] - ticks3[1:(n-1)]
	}
	size = min(comp$text.size, widths/max(ticksWidths))
	x = ticks3[1:(n-1)]  + ticksWidths[1]*size #+ .5*widths[1]

	lineHeight = convertHeight(unit(1, "lines"), "inch", valueOnly=TRUE) * size

	unitWidth = text_width_inch(unit) * size

	xtext = x[1] + c(ticks3, ticks3[n] + .5*ticksWidths[n]*size + .5*unitWidth)# + widths*.5 + unitWidth*.5) #+ position[1]


	# if "unit" text is clipped, remove last label and move unit to previous label
	if (!comp$allow_clipping) {
		xright = xtext + labelsW / 2
		if (tail(xright, 1) > W) {
			labels = c(head(labels, -2), unit)
			xtext = x[1] + c(head(ticks3, -1), ticks3[n-1] + .5*ticksWidths[n-1]*size + .5*unitWidth)# + widths*.5 + unitWidth*.5) #+ position[1]
		}
	}


	grobBG = if (getOption("tmap.design.mode")) rectGrob(gp=gpar(fill="#CAB2D6")) else NULL


	# other grid cells are aligns (1 and 5) and margins (2 and 4)
	scalebar = gridCell(3,3, {
		gTree(children=gList(

			grobBG,
			rectGrob(x=unit(x, "inch"), y=unit(1.5*lineHeight, "inch"), width = unit(widths, "inch"), height=unit(lineHeight*.5, "inch"), just=c("left", "bottom"), gp=gpar(col=dark, fill=c(light, dark), lwd=comp$lwd)),
			textGrob(label=labels, x = unit(xtext, "inch"), y = unit(lineHeight, "inch"), just=c("center", "center"), gp=gpar(col=comp$text.color, cex=size, fontface=comp$text.fontface, fontfamily=comp$text.fontfamily))
			), name="scalebar")
	})

	grid::grobTree(scalebar, vp = vp)
	#scalebar
}










tmapGridCompPlot_text = function(comp, o, fH, fW) {

	textS = if (comp$text == "") 0 else comp$size * comp$scale #* o$scale

	padding = grid::unit(comp$padding[c(3,4,1,2)] * textS * o$lin, units = "inch")

	if (comp$position$align.h == "left") {
		#x = grid::unit(0, "npc")
		x = grid::unit(comp$padding[2] * textS * o$lin, units = "inch")
		halign = 0
		hjust = 1
		just = "left"
	} else if (comp$position$align.h == "right") {
		#x = grid::unit(1, "npc")
		x = grid::unit(1, "npc") - grid::unit(comp$padding[4] * textS * o$lin, units = "inch")
		halign = 1
		hjust = 0
		just = "right"
	} else {
		x = grid::unit(0.5, "npc")
		halign = 0.5
		hjust = 0.5
		just = "center"
	}

	# grtext = gridtext::richtext_grob(comp$text,
	# 								  x = x,
	# 								  box_gp = gpar(col = frame.col, fill = bg.color, alpha = bg.alpha, lwd = frame.lwd),
	# 								  r = grid::unit(frame.r, "pt"),
	# 								  halign = halign,
	# 								  hjust = hjust,
	# 								  gp = grid::gpar(cex = textS))
	grtext = grid::textGrob(comp$text,
							 x = x,
							 just = just,
							 gp = grid::gpar(col = comp$color, cex = textS, fontface = comp$fontface, fontfamily = comp$fontfamily, alpha = comp$alpha))

	if (getOption("tmap.design.mode")) {
		grDesign = grid::rectGrob(gp=gpar(fill=NA,col="#000000", lwd=1))
	} else {
		grDesign = NULL
	}

	#grtext = rectGrob(gp=gpar(col = "green", fill= NA))


	g = do.call(grid::grobTree, c(list(grtext), list(grDesign))) #, list(vp = vp)

	g

}


correct_nlines = function(n) {
	# Linear model applied based on this empirical data:
	# (results may depend on device)
	#
	# y = sapply(1:50, function(i) {
	# 	s = paste(rep("text", i), collapse = "\n")
	# 	convertHeight(stringHeight(s), "inch", valueOnly = TRUE)
	# })
	# df = data.frame(x = 1:50, y = y / 0.2) #0.2 is the lineheight (par "cin")
	# lm(y~x, df)
	#
	-.6035 + n * 1.2
}


tmapGridCompHeight_text = function(comp, o) {
	textS = if (comp$text == "") 0 else comp$size #* o$scale
	textP = comp$padding[c(3,1)] * textS * o$lin
	textH = textS * o$lin

	nlines = number_text_lines(comp$text)

	nlines2 = correct_nlines(nlines)
	comp$flexRow = NA

	hs = c(textP[1], textH * nlines2, textP[2])
	h = sum(hs)

	sides = switch(comp$position$align.v, top = "second", bottom = "first", "both")
	hsu = set_unit_with_stretch(hs, sides = sides)

	comp$Hin = h
	comp$hsu = hsu
	comp
}

# borrowed from treemap (wraps text to 1-5 lines)
wrapText = function(txt, nlines) {
	if (nlines == 1) {
		txt
	} else {
		# create some wrappings, with slightly different widths:
		results <- lapply(1:5, FUN=function(pos, nlines, txt) {
			strwrap(txt, width = pos+(nchar(txt)/nlines))}, nlines, txt)
		lengths = sapply(results, length)

		# find the best match
		diff = nlines - lengths
		diff[diff < 0] = 1000
		id = which.min(diff)[1]

		paste(results[[id]], collapse = "\n")
	}
}




tmapGridCompWidth_text = function(comp, o) {
	textS = if (comp$text == "") 0 else comp$size #* o$scale
	textP = comp$padding[c(2,4)] * textS * o$lin
	textW = textS * graphics::strwidth(comp$text, units = "inch", family = comp$fontfamily, font = fontface2nr(comp$fontface))


	if (!is.na(comp$width)) {
		textPgs = strsplit(comp$text, "\n")[[1]]
		text2 = do.call(paste, c(lapply(textPgs, function(p) {
			textW = textS * graphics::strwidth(p, units = "inch", family = comp$fontfamily, font = fontface2nr(comp$fontface))
			w = sum(textP[1], textW, textP[2])
			nlines = round(w / (comp$width * textS * o$lin))
			wrapText(p, nlines)
		}), list(sep = "\n")))

		textW2 = textS * graphics::strwidth(text2, units = "inch", family = comp$fontfamily, font = fontface2nr(comp$fontface))
		wsu2 = c(textP[1], textW2, textP[2])
		ws = sum(textP[1], textW2, textP[2])
		comp$text = text2
	} else {
		ws = c(textP[1], textW, textP[2])
	}


	sides = switch(comp$position$align.h, left = "second", right = "first", "both")
	wsu = set_unit_with_stretch(ws, sides = sides)

	comp$Win = sum(ws)
	comp$wsu = wsu

	comp$flexCol = NA
	comp
}






#' @export
tmapGridCompPrepare.tm_logo = function(comp, o) {
	comp$logo = lapply(comp$file, function(lf){
		tmap_icons(lf)
	})
	comp$asp = vapply(comp$logo, function(lg) {
		lg$iconWidth / lg$iconHeight
	}, FUN.VALUE = numeric(1))
	comp$show = TRUE
	comp
}

#' @export
tmapGridCompHeight.tm_logo = function(comp, o) {
	marH = comp$margins[c(3,1)] * o$lin
	hs = c(marH[1], comp$height * o$lin, marH[2])

	sides = switch(comp$position$align.v, top = "second", bottom = "first", "both")
	hsu = set_unit_with_stretch(hs, sides = sides)

	Hin = sum(hs)
	comp$flexRow = NA
	comp$Hin = Hin #  sum(textP[1], textH, textP[2])
	comp$hsu = hsu
	comp
}

#' @export
tmapGridCompWidth.tm_logo = function(comp, o) {
	k = length(comp$asp)
	comp$width = comp$height * comp$asp

	marW = comp$margins[c(2,4)] * o$lin
	ws = c(marW[1],
		   comp$width[1] * o$lin,
		   {if (k > 1) unlist(lapply(comp$width[-1], function(w) c(comp$between_margin, w) * o$lin)) else NULL},
		   marW[2])

	sides = switch(comp$position$align.h, left = "second", right = "first", "both")
	wsu = set_unit_with_stretch(ws, sides = sides)
	comp$col_ids = seq(3L, by = 2L, length.out = k)
	comp$flexCol = NA
	comp$Win = sum(ws)
	comp$wsu = wsu
	comp
}

#' @export
tmapGridCompPlot.tm_logo = function(comp, o, fH, fW) {

	k = length(comp$logo)

	wsu = comp$wsu
	hsu = comp$hsu

	vp = grid::viewport(layout = grid::grid.layout(ncol = length(wsu),
												   nrow = length(hsu),
												   widths = wsu,
												   heights = hsu))
	gLogos = mapply(function(logo, col) {
		grobLogo = pngGrob(logo$iconUrl, fix.borders = TRUE, n=2, height.inch=as.numeric(comp$hsu[3]), target.dpi=96)
		rdim = dim(grobLogo$raster)
		grobLogo$raster = matrix(do.call("process_color", c(list(as.vector(grobLogo$raster)), o$pc)), nrow = rdim[1], ncol=rdim[2])
		gridCell(3L, col, grobLogo)
	}, comp$logo, comp$col_ids, SIMPLIFY = FALSE)

	grobBG = if (getOption("tmap.design.mode")) rectGrob(gp=gpar(fill="#CAB2D6")) else NULL
	gBG = gridCell(3L, 3L:(length(wsu) - 2L), grobBG)

	do.call(grid::grobTree, c(list(gBG), gLogos, list(vp = vp)))

}

#' @export
tmapGridCompPrepare.tm_inset_tmap = function(comp, o) {
	asp = comp$width / comp$height

	# Pass inset dimensions as a viewport so prepreprocess_meta computes
	# cw/ch correctly (dasp=asp → cw=asp, ch=1 for landscape)
	vp = grid::viewport(
		width  = grid::unit(comp$width * o$lin,  "inches"),
		height = grid::unit(comp$height * o$lin, "inches")
	)

	comp$x = tmap_grob(comp$x, asp = asp, vp = vp)

	.TMAP$is_first_inset = FALSE
	class(comp)[1] = "tm_inset_grob"
	comp$bbox = .TMAP$geo_ref$bbx_frame
	comp$show = TRUE
	comp
}

#' @export
tmapGridCompPrepare.tm_inset_grob = function(comp, o) {
	comp$show = TRUE
	comp
}

#' @export
tmapGridCompPrepare.tm_inset_gg = function(comp, o) {
	rlang::check_installed("ggplot2", reason = "for plotting ggplot2 charts")
	comp$x = ggplot2::ggplotGrob(comp$x)
	class(comp)[1] = "tm_inset_grob"
	comp$show = TRUE
	comp
}



#' @export
tmapGridCompHeight.tm_inset_grob = function(comp, o) {
	marH = comp$margins[c(3,1)] * o$lin
	hs = c(marH[1], comp$height * o$lin, marH[2])

	sides = switch(comp$position$align.v, top = "second", bottom = "first", "both")
	hsu = set_unit_with_stretch(hs, sides = sides)

	Hin = sum(hs)
	comp$flexRow = NA
	comp$Hin = Hin #  sum(textP[1], textH, textP[2])
	comp$hsu = hsu
	comp
}

#' @export
tmapGridCompWidth.tm_inset_grob = function(comp, o) {
	marW = comp$margins[c(2,4)] * o$lin

	ws = c(marW[1], comp$width * o$lin, marW[2])

	sides = switch(comp$position$align.h, left = "second", right = "first", "both")
	wsu = set_unit_with_stretch(ws, sides = sides)

	comp$flexCol = NA
	comp$Win = sum(ws)
	comp$wsu = wsu
	comp
}

#' @export
tmapGridCompPrepare.tm_inset_image = function(comp, o) {
	comp$file = comp$x
	comp = tmapGridCompPrepare.tm_logo(comp, o)
	class(comp) = c("tm_logo", class(comp))
	comp
}



#' @export
tmapGridCompPlot.tm_inset_grob = function(comp, o, fH, fW) {


	wsu = comp$wsu
	hsu = comp$hsu

	vp = grid::viewport(layout = grid::grid.layout(ncol = length(wsu),
												   nrow = length(hsu),
												   widths = wsu,
												   heights = hsu))

	grobBG = if (getOption("tmap.design.mode")) rectGrob(gp=gpar(fill="#CAB2D6")) else NULL
	gBG = gridCell(3L, 3L:(length(wsu) - 2L), grobBG)
#comp$x = grid::rectGrob(gp=gpar(fill="gold"))
	g = gridCell(3L, 3L, comp$x)

	do.call(grid::grobTree, c(list(gBG, g), list(vp = vp)))

}



#' @export
tmapGridCompPrepare.tm_inset_map = function(comp, o) {
	limit_lat = if ("limit_latitude_3857" %in% names(o)) o$limit_latitude_3857 else c(-90, 90)

	if (is.null(comp$x)) {
		comp$x = sf::st_bbox(c(xmin = -180, xmax = 180, ymin = limit_lat[1], ymax = limit_lat[2]), crs = 4326)
		comp$crs = tmap_options()$crs_global
	}

	bbpoly = sf::st_transform(tmaptools::bb_poly(comp$x), o$crs_step4)

	b = tmaptools::bb(bbpoly, asp.target = comp$width / comp$height)

	comp$bbox = b
	comp$show = TRUE
	comp
}

#' @export
tmapGridCompHeight.tm_inset_map = function(comp, o) {
	marH = comp$margins[c(3,1)] * o$lin
	hs = c(marH[1], comp$height * o$lin, marH[2])

	sides = switch(comp$position$align.v, top = "second", bottom = "first", "both")
	hsu = set_unit_with_stretch(hs, sides = sides)

	Hin = sum(hs)
	comp$flexRow = NA
	comp$Hin = Hin #  sum(textP[1], textH, textP[2])
	comp$hsu = hsu
	comp
}

#' @export
tmapGridCompWidth.tm_inset_map = function(comp, o) {
	marW = comp$margins[c(2,4)] * o$lin

	ws = c(marW[1], comp$width * o$lin, marW[2])

	sides = switch(comp$position$align.h, left = "second", right = "first", "both")
	wsu = set_unit_with_stretch(ws, sides = sides)

	comp$flexCol = NA
	comp$Win = sum(ws)
	comp$wsu = wsu
	comp
}

#' @export
tmapGridCompPlot.tm_inset_map = function(comp, o, fH, fW) {


	wsu = comp$wsu
	hsu = comp$hsu

	vp = grid::viewport(layout = grid::grid.layout(ncol = length(wsu),
												   nrow = length(hsu),
												   widths = wsu,
												   heights = hsu))

	grobBG = if (getOption("tmap.design.mode")) rectGrob(gp=gpar(fill="#CAB2D6")) else NULL
	gBG = gridCell(3L, 3L:(length(wsu) - 2L), grobBG)

	vp = grid::viewport(layout.pos.row = 3L, layout.pos.col = 3L)

	#comp$tm$o = prepreprocess_meta(comp$tm$o, vp)
	comp$tm$o$asp = as.numeric(wsu[3]) / as.numeric(hsu[3])

	#comp$tm$tmo$group1$layers$layer1$shpDT$shpTM[[1]]$bbox$x = comp$bbox
	comp$tm$o$inset = TRUE
	comp$tm$o$bbox = comp$bbox

	g = step4_plot(comp$tm, vp, return.asp = FALSE, show = FALSE, in.shiny = FALSE, knit = FALSE, knit_opts = list(), args = list())

	#g = gridCell(3L, 3L, grid::rectGrob(gp=gpar(fill="red")))#comp$x)

	do.call(grid::grobTree, c(list(gBG, g), list(vp = vp)))

}



#' @export
tmapGridCompPrepare.tm_minimap = function(comp, o) {
	comp$show = TRUE

	comp
}

#' @export
tmapGridCompHeight.tm_minimap = function(comp, o) {
	tmapGridCompHeight.tm_inset_grob(comp, o)
}

#' @export
tmapGridCompWidth.tm_minimap = function(comp, o) {
	tmapGridCompWidth.tm_inset_grob(comp, o)
}

#' @export
tmapGridCompPlot.tm_minimap = function(comp, o, fH, fW) {
	poly = tmaptools::bb_poly(comp$bbox)

	if (sf::st_is_longlat(comp$bbox)) {
		center = c(mean(comp$bbox[c("xmin", "xmax")]),
				   mean(comp$bbox[c("ymin", "ymax")]))
	} else {
		center = suppressWarnings(round(sf::st_transform(sf::st_centroid(poly), crs = 4326)[[1]][], 3))
	}

	# bound lat to -30,-10 or 10,30
	center[2] = if (center[2] >= 0) {
		max(min(center[2], 30), 10)
	} else {
		max(min(center[2], -10), -30)
	}



	#"+proj=ortho +lat_0=10 +lon_0=0"

	lastcalln = x = get("last_map_new", envir = .TMAP)

	tm = tm_shape(World) +
		tm_crs(ortho_lonlat(center[1], center[2]), bbox = "FULL")+
		tm_polygons(col = NULL, fill = "#2CA02C") +
	tm_shape(poly) +
		tm_polygons(fill = NULL, col = "#EE8866", lwd = 3) +
		#tm_polygons(fill = NULL, col = "#EE8866", lwd = 3, lty = "dotted") +
		tm_graticules(labels.show = FALSE, col = "#000000", lwd = 0.5) +
		tm_layout(bg.color = "#88CCEE",
				  earth_boundary = TRUE,
				  frame = FALSE,
				  space = FALSE) +
		tm_options(space_overlay = o$space_overlay)

	asp = comp$width / comp$height


	comp$x = tmap_grob(tm, asp = asp)

	assign("last_map_new", lastcalln, envir = .TMAP)


	class(comp)[1] = "tm_inset_grob"
	tmapGridCompPlot.tm_inset_grob(comp, o, fH, fW)
}

#' Set or get the default tmap style
#'
#' Set or get the default tmap style. Without arguments, the current style is returned.
#' Also the available styles are displayed. When a style is set, the corresponding tmap
#' options (see [tmap_options()]) will be set accordingly.
#' The default style (i.e. when loading the package) is `"white"`.
#'
#' Note that [tm_style()] is used within a plot call (so it only affects that plot),
#' whereas `tmap_style()` sets the style globally.
#'
#' After loading a style, the options that defined this style
#' (i.e. the difference with the default `"white"` style) can be obtained by [tmap_options_diff()].
#'
#' The documentation of [tmap_options()] (details and the examples) shows how to create a new style.
#'
#' @param style Name of the style. When omitted, `tmap_style()` returns the current style
#'   and also shows all available styles. When the style is specified,`tmap_style()`
#'   sets the style accordingly. Note that in that case, all tmap options (see [tmap_options()])
#'   will be reset according to the style definition.
#'   See [tm_layout()] for predefined styles, and `tmap_style_catalogue()` for creating a catalogue.
#' @return The style before changing
#' @seealso
#' * [tmap_options()] for tmap options
#' * [tmap_style_catalogue()]  to create a style catalogue of all available styles.
#' @example ./examples/tmap_style.R
#' @export
#' @rdname tmap_style
tmap_style = function(style) {
	current.style = getOption("tmap.style")
	show.messages = get("tmapOptions", envir = .TMAP)$show.messages

	if (missing(style) && show.messages) {
		message("current tmap style is ", style_names(current.style))
		message("other available styles are: ", print_text_vector(get_style_names(current.style, v3 = "no")))
		message("tmap v3 styles: ", print_text_vector(get_style_names(current.style, v3 = "only")))
	} else {
		.tmapOptions = .TMAP$defaultTmapOptions
		check_style(style)
		options(tmap.style=style)
		if (style == "white") {
			.tmapOptions = .TMAP$defaultTmapOptions
		} else {
			styleOptions = get("tmapStyles", envir = .TMAP)[[style]]
			.tmapOptions = complete_options(styleOptions, .tmapOptions)
			attr(.tmapOptions, "style") = style
		}

		assign("tmapOptions", .tmapOptions, envir = .TMAP)

		if (show.messages) {
			message("style set to ", style_names(style))
			message("other available styles are: ", print_text_vector(get_style_names(style, v3 = "no")))
			message("tmap v3 styles: ", print_text_vector(get_style_names(style, v3 = "only")))
		}
	}
	invisible(current.style)
}



print_text_vector = function(x) {
	x2 = style_names(x)

	paste(x2, collapse = ", ")
}

style_names = function(x) {
	x2 = paste0("\"", x, "\"")
	x2[x2 == "\"white\""] = "\"white\" (tmap default)"
	x2[x2 == "\"v3\""] = "\"v3\" (tmap v3 default)"
	x2
}

get_style_names = function(except_style = NULL, remove_grey = TRUE, v3 = "yes") {
	styles = c("white", names(get("tmapStyles", envir = .TMAP)))
	if (!is.null(except_style)) {
		styles = setdiff(styles, except_style)
	}

	# remove double name gray/grey
	if (remove_grey) {
		if (!is.null(except_style) && (except_style %in% c("gray", "grey", "gray_v3", "grey_v3"))) {
			styles = setdiff(styles, c("gray", "grey", "gray_v3", "grey_v3"))
		} else {
			styles = setdiff(styles, c("grey", "grey_v3"))
		}
	}

	is_v3 = substr(styles, nchar(styles) - 1, nchar(styles)) == "v3"

	if (v3 == "no") {
		styles = styles[!is_v3]
	} else if (v3 == "only") {
		styles = styles[is_v3]
	}

	styles
}




check_style = function(style) {
	styles = get_style_names(remove_grey = FALSE)
	if (!style %in% styles) {
		stop("style \"" , style, "\" unknown. The available styles are: ",
			 print_text_vector(get_style_names()), call. = FALSE)
	}
}

#' Map layer: rgb images
#'
#' Map layer that an rgb image.. The used (multivariate) visual variable is `col`,
#' which should be specified with 3 or 4 variables for `tm_rgb()` and `tm_rgba()` respectively.
#' The first three correspond to the red, green, and blue channels. The optional
#' fourth is the alpha transparency channel.
#'
#' @param col,col.scale,col.legend,col.chart,col.free Visual variable that determines
#'   the color. `col` is a multivariate variable, with 3 (`tm_rgb`) or 4 (`tm_rgba`) numeric data variables. These can be specified via [tm_vars()] with `multivariate = TRUE`
#' @param col_alpha,col_alpha.scale,col_alpha.legend,col_alpha.chart,col_alpha.free  `r .doc_vv("col_alpha")`
#' @param options options passed on to the corresponding `opt_<layer_function>` function
#' @param ... to catch deprecated arguments from version < 4.0
#' @example ./examples/tm_rgb.R
#' @export
tm_rgb = function(col = tm_vars(n = 3, multivariate = TRUE),
				  col.scale = tm_scale_rgb(),
				  col.legend = tm_legend(),
				  col.chart = tm_chart_none(),
				  col.free = NA,
				  col_alpha = tm_const(),
				  col_alpha.scale = tm_scale(),
				  col_alpha.legend = tm_legend(),
				  col_alpha.chart = tm_chart_none(),
				  col_alpha.free = NA,
				  options = opt_tm_rgb(),
				  ...) {
	args = list(...)

	layer_fun = if ("called_from" %in% names(args)) {
		args$called_from
	} else {
		"tm_rgb"
	}

	if (any(v3_only("tm_rgb") %in% names(args)) || is.numeric(col.scale)) {
		v3_start_message()

		# second condition needed to catch tm_rgb(1, 2, 3)


		if (all(c("r", "g", "b") %in% names(args))) {
			# v3_tm_rgb(args$r, args$g, args$b) moved to step1_helper_facets
			col = tm_vars(c(args$r, args$g, args$b), n = -1, multivariate = TRUE)
		}
		if (is.numeric(col) && is.numeric(col.scale) && is.numeric(col.legend)) {
			col = tm_vars(c(col, col.scale, col.legend), n = -1, multivariate = TRUE)
			col.scale = tm_scale_rgb()
			col.legend = tm_legend()
		}


		if ("alpha" %in% names(args) && layer_fun == "tm_rgb") {
			col_alpha = args$alpha
			v3_message_rgb_alpha(layer_fun = layer_fun)
		}
		if ("saturation" %in% names(args)) {
			options$saturation = args$saturation
			v3_message_rgb_opt(layer_fun = layer_fun, opt = "saturation", args$saturation)
		}
		if ("interpolate" %in% names(args)) {
			options$interpolate = args$interpolate
			v3_message_rgb_opt(layer_fun = layer_fun, opt = "interpolate", args$interpolate)
		}
		if ("max.value" %in% names(args)) {
			col.scale$max_color_value = args$max.value
			v3_message_rgb_maxV(layer_fun = layer_fun, opt = "interpolate", args$max.value)
		}

		v3_start_message()

	}

	# unused arguments: typos?
	unused = setdiff(names(args), c(v3_only("tm_rgb"), "called_from"))

	if (length(unused)) {
		message_layer_unused_args(layer_fun, unused)
	}



	do.call(tm_raster, args = list(col = col, col.scale = col.scale, col.legend = col.legend, col.chart = col.chart, col.free = col.free,
								   col_alpha = col_alpha, col_alpha.scale = col_alpha.scale, col_alpha.legend = col_alpha.legend, col_alpha.chart = col_alpha.chart, col_alpha.free = col_alpha.free,
								   options = options))
}

#' @rdname tm_rgb
#' @export
tm_rgba = function(col = tm_vars(n = 4, multivariate = TRUE),
				  col.scale = tm_scale_rgba(),
				  col.legend = tm_legend(),
				  col.chart = tm_chart_none(),
				  col.free = NA,
				  options = opt_tm_rgb()) {
	do.call(tm_raster, args = list(col = col, col.scale = col.scale, col.legend = col.legend, col.chart = col.chart, col.free = col.free, options = options))
}

#' @rdname tm_rgb
#' @param interpolate Should the raster image be interpolated? Currently only applicable in view mode (passed on to [`grid`][grid::rasterGrob()])
#' @param saturation `r .doc_opt("rgb.saturation")`
#' @export
opt_tm_rgb = function(interpolate = FALSE, saturation = 1) {
	list(trans.args = list(),
		 mapping.args = list(interpolate = interpolate, saturation = saturation))
}

#' Wrapper functions for using **tmap** in **shiny**
#'
#' @description
#' * `tmapOutput()` creates a UI element
#' * `renderTmap()` renders a tmap map
#' * `tmapProxy()` updates a tmap map in `view` mode
#'
#' Adding layers is as usual via the map layer functions like [tm_polygons()].
#' Removing layers can be done, removing with `tm_remove_layer()`.
#'
#' @details
#' Two features from tmap are not (yet) supported in Shiny: small multiples (facets) and colored backgrounds (argument `bg.color` of [tm_layout()]).
#' Workarounds for small multiples: create multiple independent maps or specify `as.layers = TRUE` in [tm_facets()].
#'
#' @param expr A tmap object. A tmap object is created with [qtm()] or by stacking [`tmap-element`]s.
#' @param mode tmap mode, see [tmap_mode()] If not defined, the current mode is used
#' @param env The environment in which to evaluate expr
#' @param quoted Is `expr` a quoted expression (with `quote()`)? This is useful if you want to save an expression in a variable
#' @param outputId Output variable to read from
#' @param width,height the width and height of the map
#' @param mapId single-element character vector indicating the output ID of the map to modify (if invoked from a Shiny module, the namespace will be added automatically)
#' @param session the Shiny session object to which the map belongs; usually the default value will suffice
#' @param x the tmap object that specifies the added and removed layers.
#' @param zindex The stacking number of the layer to be removed. It is recommended to specify the `zindex` for this layer when creating the map (inside `renderTmap()`).
#' @param execOnResize If `TRUE` (default), when the plot is resized, the map is regenerated. When set to `FALSE` the map is rescaled: the aspect ratio is kept, but the layout will be less desirable.
#' @param ... passed on to the mode-specific methods
#' @importFrom htmlwidgets shinyWidgetOutput
#' @example ./examples/tmapOutput.R
#' @seealso `r .doc_see_also_shiny()`
#' @export
renderTmap <- function(expr, env = parent.frame(), quoted = FALSE, execOnResize = TRUE, mode = NA) {
	if (!quoted) expr <- substitute(expr)            # capture once, in this frame
	renderTmapGS(tmap_graphics_class(mode),
				 expr = expr, env = env, execOnResize = execOnResize)
}

#' @rdname renderTmap
#' @export
tmapOutput <- function(outputId, width = "100%", height = 400, mode = NA) {
	tmapOutputGS(tmap_graphics_class(mode),
				 outputId = outputId, width = width, height = height)
}



#' @rdname renderTmap
#' @export
tmapProxy <- function(mapId, session = shiny::getDefaultReactiveDomain(), x, mode = NA) {
	tmapProxyGS(tmap_graphics_class(mode),
				mapId = mapId, session = session, tmobj = x)   # `x` (tmap obj) -> tmobj
}


#' @rdname renderTmap
#' @export
tm_remove_layer <- function(zindex) {
	tm_element_list(tm_element(zindex = zindex,
							   subclass = c("tm_proxy", "tm_remove_layer")))
}

# function to update the levels fl per facet dimension
update_fl = function(k, lev = NULL, m = NULL) {
	fl = get("fl", envir = .TMAP)
	# print("test")
	# print(fl)
	fk = fl[[k]]
	n =  if (is.character(fk)) length(fk) else fk

	if (!missing(lev)) {
		m = length(lev)
		if (m == 1L && n == 1L) {
			fk = lev
		} else if (m > 1L && n > 1L && m != n) {
			stop("Inconsistent number of facets in the ", k, " dimension.", call. = FALSE)
		} else if (m > n) {
			fk = lev
		}
	} else if (!missing(m)) {
		if (m > 1L && n > 1L && m != n) {
			stop("Inconsistent number of facets in the ", k, " dimension.", call. = FALSE)
		} else if (m > n) {
			fk = m
		}
	}
	fl[[k]] = fk
	assign("fl", fl, envir = .TMAP)
}



get_tmf = function(tmfs) {
	# Get tmf object: start with the one that is called, and add 'calls'

	nf = length(tmfs)

	gns = vapply(tmfs, "[[", FUN.VALUE = integer(3), "gn", USE.NAMES = FALSE)

	grpid = apply(gns, which.max, MARGIN = 1)

	gn = apply(gns, max, MARGIN = 1)

	for (i in seq_len(3L)) if (any(gns[i, ] > 1L & gns[i, ] != gn[i])) stop("Number of facets inconsistent over groups", call. = FALSE)

	gls = lapply(tmfs, "[[", "gl")

	gl = lapply(seq_len(3L), function(i) gls[[grpid[i]]][[i]])


	#gnl = lapply(tmfs, "[[", FUN.VALUE = integer(3), "gn", USE.NAMES = FALSE)

	# find first tmf that has been called
	fid = which(vapply(tmfs, function(tmf){
		"calls" %in% names(tmf)
	}, FUN.VALUE = logical(1)))[1]

	if (is.na(fid)) fid = 1L

	tmf = tmfs[[fid]]

	if (fid < nf) {
		for (i in (fid+1):nf) {
			args = tmfs[[i]]$calls
			tmf[args] = tmfs[[i]][args]
		}
	}
	tmf$fl = gl
	tmf$fn = gn
	tmf$n = prod(gn)
	tmf$npp = prod(gn[1L:2L])


	if (tmf$type %in% "wrapstack") {
		tmf$type = if ((!is.na(tmf$ncols) && tmf$ncols %in% c(tmf$n, 1)) ||
			(!is.na(tmf$nrows) && tmf$nrows %in% c(tmf$n, 1))) {
			"stack"
		} else if (!is.na(tmf$ncols) || !is.na(tmf$nrows)) {
			"wrap"
		} else if (tmf$npp > 3) "wrap" else "stack"
	}

	tmf
}

cbind_dts = function(dts, plot.order) {
	if (!length(dts)) return(list())

	bypass_ord = plot.order$aes == "DATA"

	max_values <- tryCatch(
		vapply(dts, FUN =  ncol, FUN.VALUE = integer(1)),
		error = function(e) {
			warning("The data.table with the most group-by columns (others are joined). This is an internal error.")
			NULL
		})
	if (is.null(max_values)) {
		return(NULL)
	}
	id = which.max(max_values) # data.table with the most group-by columns (others are joined)

	dt = dts[[id]]

	dev = getOption("tmap.devel.mode")

	if (length(dts) > 1L) {
		for (i in setdiff(seq_along(dts), id)) {
			dti = dts[[i]]

			id_cols = ncol(dti) - 1L - as.integer(!bypass_ord) #minus one aes and one ord
			id_nams = names(dti)[seq.int(id_cols)]

			#dt = dt[dti, on = names(dti)[1L:(ncol(dti)-2L)]]
			dt = dt[dti, on = id_nams]
		}
	}



	ord_cols = which(subStr(names(dt), -5) == "__ord")
	#m = as.matrix(dt[, ord_cols, with = FALSE])


	#dt[, ord_cols, with = FALSE][, ]

	if (!bypass_ord) {
		dt_rep = function(old, new) {
			dt[, (ord_cols) := replace(.SD, .SD == old, new), .SDcols = ord_cols]
		}


		if (plot.order$na.order == "mix") dt_rep(0L, 1L)
		if (plot.order$null.order == "mix") dt_rep(-1L, 1L)

		if (!plot.order$null.below.na) {
			if (plot.order$na.order == "top") dt_rep(0L, 2147483646L) else dt_rep(0L, -2L)
			if (plot.order$null.order == "top") dt_rep(-1L, 2147483647L)
		} else {
			if (plot.order$na.order == "top") dt_rep(0L, 2147483647L)
			if (plot.order$null.order == "top") dt_rep(-1L, 2147483646L)
		}

		dt[, ord__ := do.call(pmin, .SD), .SDcols = ord_cols]
		dt[ord__ > 0L, ord__ := do.call(pmax, .SD), .SDcols = ord_cols]
		dt[, (ord_cols) := NULL]
	}


	dt
}


subStr = function(x, k) {
	if (k > 0L) {
		substr(x, 1L, k)
	} else {
		n = nchar(x)
		substr(x, n + k + 1L, n)
	}
}



limit_nx = function(nx, ox = NA) {
	# bound number of facets
	mode = getOption("tmap.mode")

	if (is.na(ox)) {
		tmapOptions <- tmap::tmap_options_mode(mode.specific = FALSE)
		ox = tmapOptions$facet.max
	}
	if (ox < nx) cli::cli_warn("{.field number of facets}. Number of facets for mode {.str {mode}} is limited to {ox}. Change the option facet.max to allow more facets, with {.code tmap_options(facet.max = {nx})} or {.code + tm_options(facet.max = {nx})}")
	min(ox, nx)
}

#' Map layer: geographic circles
#'
#' Map layer that draws circles with geographically fixed radii — i.e. the
#' radius is expressed in meters and the circles scale with zoom in interactive
#' (view) mode. This is in contrast to [tm_bubbles()], where the symbol size
#' is a fixed number of screen pixels.
#'
#' Supported visual variables: `fill` (fill colour), `col` (border colour),
#' `size` (radius in meters), `lwd` (line width), `lty` (line type),
#' `fill_alpha` (fill transparency), `col_alpha` (border transparency).
#'
#' @param fill,fill.scale,fill.legend,fill.chart,fill.free `r .doc_vv("fill")` `r .doc_unit("fill")`
#' @param col,col.scale,col.legend,col.chart,col.free `r .doc_vv("col")` `r .doc_unit("col")`
#' @param size,size.scale,size.legend,size.chart,size.free `r .doc_vv("size")` `r .doc_unit("size.circles")`
#'   `r .doc_unit("size.circles")`
#' @param lwd,lwd.scale,lwd.legend,lwd.chart,lwd.free `r .doc_vv("lwd")` `r .doc_unit("lwd")`
#' @param lty,lty.scale,lty.legend,lty.chart,lty.free `r .doc_vv("lty")` `r .doc_unit("lty")`
#' @param fill_alpha,fill_alpha.scale,fill_alpha.legend,fill_alpha.chart,fill_alpha.free `r .doc_vv("fill_alpha")` `r .doc_unit("fill_alpha")`
#' @param col_alpha,col_alpha.scale,col_alpha.legend,col_alpha.chart,col_alpha.free `r .doc_vv("col_alpha")` `r .doc_unit("col_alpha")`
#' @param plot.order Specification in which order the spatial features are
#'   drawn. See [tm_plot_order()] for details.
#' @param options Options passed on to [opt_tm_circles()].
#' @param points_only Should only point geometries of the shape object be
#'   plotted? Default `"ifany"` means `TRUE` whenever the geometry collection
#'   contains points.
#' @param point_per How spatial points are generated for non-point geometries.
#'   One of `"feature"` (one point per feature, default), `"segment"` (one per
#'   sub-feature), or `"largest"` (only the largest sub-feature).
#' @param on_surface For polygon inputs, should the centroid be guaranteed to
#'   lie on the surface? If `TRUE` (slower), centroids outside the polygon are
#'   replaced via [sf::st_point_on_surface()].
#' @inheritParams tm_polygons
#' @inherit tm_polygons details
#' @param ... To catch deprecated arguments from version < 4.0.
#' @seealso
#'   * [tm_bubbles()] for screen-fixed proportional circles (pixel radius).
#'   * [tm_symbols()] for the general symbol layer with configurable shapes.
#'   * [tm_scale_asis()] to pass data values directly as metre radii.
#' @example examples/tm_circles.R
#' @export
tm_circles = function(size = tm_const(),
					  size.scale = tm_scale(),
					  size.legend = tm_legend(),
					  size.chart = tm_chart_none(),
					  size.free = NA,
					  fill = tm_const(),
					  fill.scale = tm_scale(),
					  fill.legend = tm_legend(),
					  fill.chart = tm_chart_none(),
					  fill.free = NA,
					  col = tm_const(),
					  col.scale = tm_scale(),
					  col.legend = tm_legend(),
					  col.chart = tm_chart_none(),
					  col.free = NA,
					  lwd = tm_const(),
					  lwd.scale = tm_scale(),
					  lwd.legend = tm_legend(),
					  lwd.chart = tm_chart_none(),
					  lwd.free = NA,
					  lty = tm_const(),
					  lty.scale = tm_scale(),
					  lty.legend = tm_legend(),
					  lty.chart = tm_chart_none(),
					  lty.free = NA,
					  fill_alpha = tm_const(),
					  fill_alpha.scale = tm_scale(),
					  fill_alpha.legend = tm_legend(),
					  fill_alpha.chart = tm_chart_none(),
					  fill_alpha.free = NA,
					  col_alpha = tm_const(),
					  col_alpha.scale = tm_scale(),
					  col_alpha.legend = tm_legend(),
					  col_alpha.chart = tm_chart_none(),
					  col_alpha.free = NA,
					  plot.order = tm_plot_order("size"),
					  zindex = NA,
					  group = NA,
					  group.control = "check",
					  popup = tm_popup(),
					  popup.vars = NA,
					  popup.format = tm_label_format(),
					  hover = NA,
					  id = "",
					  blend = "over",
					  options = opt_tm_circles(),
					  ...) {

	args = c(as.list(environment()), list(...))

	# tm_symbols expects a shape argument; fix it to a filled circle (21).
	# The "circles" layer class drives dispatch to the geographic-circle
	# DataPlot methods, which ignore shape entirely.
	args$shape = 21
	args$called_from = "tm_circles"
	args$options = complete_options(options, opt_tm_circles())

	tm = do.call(tm_symbols, args)
	tm[[1]]$layer = c("circles", "symbols")
	tm
}


#' @rdname tm_circles
#' @export
opt_tm_circles = function(points_only = "ifany",
						  point_per = "feature",
						  on_surface = FALSE) {
	list(
		trans.args = list(
			points_only = points_only,
			point_per   = point_per,
			on_surface  = on_surface,
			along_lines = FALSE
		),
		mapping.args = list()
	)
}

#' @exportS3Method
renderTmapGS.Grid <- function(x, expr, env, execOnResize, ...) {
	expr <- bquote(getFromNamespace("print.tmap", "tmap")(.(expr)))
	shiny::renderPlot(expr, env = env, quoted = TRUE, execOnResize = execOnResize)
}

#' @exportS3Method
tmapOutputGS.Grid <- function(x, outputId, width, height, ...) {
	shiny::plotOutput(outputId = outputId, width = width, height = height)
}

#' @exportS3Method
tmapProxyGS.Grid <- function(x, mapId, session, tmobj, ...) {
	message("tmapProxy not working for plot mode (yet)")
	print.tmap(tmobj, show = TRUE)
}

#' Layer group control
#'
#' Controls the layer groups in interactive maps (view mode): the layer control box (radio buttons or check boxes) and at which zoom levels the layers are displayed at.
#'
#' @param name group name that corresponds with the group name specified in the layer functions (e.g.  [tm_polygons()])
#' @param control The group control determines how
#'   layer groups can be switched on and off. Options: `"radio"` for radio
#'   buttons (meaning only one group can be shown), `"check"` for check boxes
#'   (so multiple groups can be shown), and `"none"` for no control
#'   (the group cannot be (de)selected).
#' @param zoom_levels The zoom levels at which the group is displays at. When specified `control` will be set to `"none"`.
#' @seealso \href{https://r-tmap.github.io/tmap/articles/adv_groups}{vignette about layer groups}
#' @export
tm_group = function(name, control = NA, zoom_levels = NA) {
	optname = paste0("view_group_", name)
	x = structure(list(list(name = name, control = control, zoom_levels = zoom_levels)), names= optname)
	do.call(tm_options, x)
}

#' Internal method that split shape objects
#' 
#' Internal method that split shape objects. So far, only used to split stars object
#' (from dimension to attributes)
#' 
#' @param shp shape
#' @param split_stars_dim name of the dimension to split (`""` to skip)
#' @export
#' @keywords internal
tmapSplitShp = function(shp, split_stars_dim, smeta) {
	UseMethod("tmapSplitShp")
}

#' @export
tmapSplitShp.stars = function(shp, split_stars_dim, smeta) {
	if (split_stars_dim != "") {
		if (length(smeta$vars) > 1L) shp = merge(shp, name = "DIMVARS__")
		
		vals = stars::st_get_dimension_values(shp, split_stars_dim)
		shp = split(shp, split_stars_dim)
		names(shp) = vals
	}
	shp
}

#' @export
tmapSplitShp.default = function(shp, split_stars_dim, smeta) {
	shp
}

#' @export
#' @keywords internal
#' @rdname tmapGridLeaflet
tmapGridDataPlot = function(a, shpTM, dt, gp, bbx, facet_row, facet_col, facet_page, id, pane, group, glid, o, ...) {
	UseMethod("tmapGridDataPlot")
}

#' @export
#' @keywords internal
#' @rdname tmapGridLeaflet
tmapGridDataPlot.default = function(a, shpTM, dt, gp, bbx, facet_row, facet_col, facet_page, id, pane, group, glid, o, ...) {
	NULL
}

#' Map component: title
#'
#' Map component that adds a title
#'
#' @param text text
#' @param size font size
#' @param color font color
#' @param padding padding
#' @param fontface font face, bold, italic
#' @param fontfamily font family
#' @param alpha alpha transparency of the text
#' @param stack stack with other map components, either `"vertical"` or `"horizontal"`.
#' @param just just
#' @param frame frame should a frame be drawn?
#' @param frame.color frame color
#' @param frame.alpha frame alpha transparancy
#' @param frame.lwd frame line width
#' @param frame.r Radius of the rounded frame corners. 0 means no rounding.
#' @param bg Show background?
#' @param bg.color Background color
#' @param bg.alpha Background transparency
#' @param position The position specification of the component: an object created with `tm_pos_in()` or `tm_pos_out()`. Or, as a shortcut, a vector of two values, specifying the x and y coordinates. The first is `"left"`, `"center"` or `"right"` (or upper case, meaning tighter to the map frame), the second `"top"`, `"center"` or `"bottom"`. Numeric values are also supported, where 0, 0 means left bottom and 1, 1 right top. See also \href{https://r-tmap.github.io/tmap/articles/adv_positions}{vignette about positioning}. In case multiple components should be combined (stacked), use `group_id` and specify `component` in [tm_components()].
#' @param group_id Component group id name. All components (e.g. legends, titles, etc) with the same `group_id` will be grouped. The specifications of how they are placed (e.g. stacking, margins etc.) are determined in [tm_components()] where its argument `id` should correspond to `group_id`.
#' @param width,height width and height of the component.
#' @param z z index, e.g. the place of the component relative to the other componets
#' @seealso \href{https://r-tmap.github.io/tmap/articles/basics_components}{Vignette about components}
#' @export
tm_title = function(text, size, color, padding, fontface, fontfamily, alpha, stack, just, frame, frame.color, frame.alpha, frame.lwd, frame.r, bg, bg.color, bg.alpha, position, group_id, width, height, z) {
	args = lapply(as.list(rlang::call_match()[-1]), eval, envir = parent.frame())
	args$called = names(args)
	args$group_id = args$group_id %||% NA_character_
	args$group_type = c("tm_<other>", "tm_title")
	args$z = args$z %||% NA_integer_
	tm_element_list(do.call(tm_element, c(args, list(subclass = c("tm_title", "tm_component")))))
}

#' @export
#' @param ... passed on to `tm_title()`
#' @rdname tm_title
tm_title_in = function(text, ..., position = tm_pos_in("left", "top")) {
	args = lapply(as.list(rlang::call_match()[-1]), eval, envir = parent.frame())
	args$called = names(args)
	args$group_id = args$group_id %||% NA_character_
	args$group_type = c("tm_<other>", "tm_title_in")
	args$z = args$z %||% NA_integer_
	args$position = position
	tm_element_list(do.call(tm_element, c(args, list(subclass = c("tm_title", "tm_component"
	)))))
}

#' @export
#' @rdname tm_title
tm_title_out = function(text, ..., position = tm_pos_out("center", "top")) {
	args = lapply(as.list(rlang::call_match()[-1]), eval, envir = parent.frame())
	args$called = names(args)
	args$group_id = args$group_id %||% NA_character_
	args$group_type = c("tm_<other>", "tm_title_out")
	args$z = args$z %||% NA_integer_
	args$position = position
	tm_element_list(do.call(tm_element, c(args, list(subclass = c("tm_title", "tm_component"
	)))))
}

#' Map component: (credits) text
#'
#' Map component that adds a text, typically used as credits. This function is the same as [tm_title()] but with different default values.
#'
#' @inheritParams tm_title
#' @param ... to catch deprecated arguments
#' @seealso \href{https://r-tmap.github.io/tmap/articles/basics_components}{Vignette about components}
#' @export
tm_credits = function(text, size, color, padding, fontface, fontfamily, alpha, stack, just, frame, frame.lwd, frame.r, bg, bg.color, bg.alpha, position, group_id, width, height, z, ...) {
	args = lapply(as.list(rlang::call_match(dots_expand = TRUE)[-1]), eval, envir = parent.frame())
	args$called = names(args)
	args$group_id = args$group_id %||% NA_character_
	args$group_type = c("tm_<other>", "tm_credits")
	args$z = args$z %||% NA_integer_
	if ("align" %in% names(args)) {
		args$position = tm_pos_in(pos.h = "right", pos.v = "bottom", align.h = args$align, align.v = "top", just.h = "left", just.v = "bottom")
		args$align = NULL
		v3_instead_message("align", "position = tm_pos_in(align.h)", "tm_credits")
	}
	tm_element_list(do.call(tm_element, c(args, list(subclass = c("tm_credits", "tm_component")))))
}


#' Map component: compass
#'
#' Map component that adds a compass
#'
#' @param north north
#' @param type compass type, one of: \code{"arrow"}, \code{"4star"}, \code{"8star"}, \code{"radar"}, \code{"rose"}. The default is controlled by \code{\link{tm_layout}} (which uses \code{"arrow"} for the default style)
#' @param text.size text.size
#' @param size size
#' @param show.labels show.labels
#' @param cardinal.directions cardinal.directions
#' @param text.color text.color
#' @param color.dark color.dark
#' @param color.light color.light
#' @param lwd lwd
#' @inheritParams tm_title
#' @param stack stack with other map components, either `"vertical"` or `"horizontal"`.
#' @param just just
#' @param margins margins
#' @param ... to catch deprecated arguments (alpha)
#' @seealso \href{https://r-tmap.github.io/tmap/articles/basics_components}{Vignette about components}
#' @export
tm_compass <- function(north,
					   type,
					   text.size,
					   size,
					   show.labels,
					   cardinal.directions,
					   text.color,
					   color.dark,
					   color.light,
					   lwd,
					   position,
					   group_id,
					   bg,
					   bg.color,
					   bg.alpha,
					   stack,
					   just,
					   frame,
					   frame.color,
					   frame.alpha,
					   frame.lwd,
					   frame.r,
					   margins,
					   z,
					   ...) {
	args = lapply(as.list(rlang::call_match()[-1]), eval, envir = parent.frame())
	args$called = names(args)
	args = warning_group_args(args)

	args$group_id = args$group_id %||% NA_character_
	args$group_type = c("tm_<other>", "tm_compass")
	args$z = args$z %||% NA_integer_
	tm_element_list(do.call(tm_element, c(args, list(subclass = c("tm_compass", "tm_component")))))
}

#' Map component: scale bar
#'
#' Map component that adds a scale bar.
#'
#' @param breaks breaks. E.g. `c(0, 10, 50)` places scale bar breaks at 0, 10, and 50 units. These units are specified in [tm_shape()].
#' @param width width of the scale bar. Units are number of text line heights, which is similar to the number of characters. In case `beaks` are specified, the `width` is only handy to finetune the approximated width, e.g. in case clipping of the labels occurs, or there is too much whitespace.
#' @param allow_clipping should clipping of the last label by allowed? If `TRUE` (default), the last break label including unit is printed even when it doesn't fit the frame. If `FALSE` it will not be printed. Instead the unit suffix is added to the second last label.
#' @param text.size text size
#' @param text.color text.color
#' @param color.dark color.dark
#' @param color.light color.light
#' @param lwd linewidth
#' @inheritParams tm_title
#' @param size Deprecated (use `text.size` instead)
#' @param stack stack with other map components, either `"vertical"` or `"horizontal"`.
#' @param margins margins
#' @seealso \href{https://r-tmap.github.io/tmap/articles/basics_components}{Vignette about components}
#' @export
tm_scalebar = function(breaks,
						width,
						allow_clipping,
						text.size,
						text.color,
						color.dark,
						color.light,
						lwd,
						position,
						group_id,
						bg,
						bg.color,
						bg.alpha,
						size = "deprecated",
						stack,
						frame,
						frame.color,
						frame.alpha,
						frame.lwd,
						frame.r,
						margins,
						z) {

	if (!identical(size, "deprecated")) {
		cli::cli_inform(c(
			"!" = "The 'size' argument of {.fn tm_scalebar} is deprecated as of tmap 4.0.",
			"i" = "Please use 'text.size' instead."
		))
	}

	args = lapply(as.list(rlang::call_match()[-1]), eval, envir = parent.frame())
	args$called = names(args)
	args$group_id = args$group_id %||% NA_character_
	args$group_type = c("tm_<other>", "tm_scalebar")
	args$z = args$z %||% NA_integer_
	tm_element_list(do.call(tm_element, c(args, list(subclass = c("tm_scalebar", "tm_component")))))
}


#' Map component: mouse coordinates
#'
#' Map component that adds mouse coordinates
#'
#' @seealso \href{https://r-tmap.github.io/tmap/articles/basics_components}{Vignette about components}
#' @inheritParams tm_title
#' @export
tm_mouse_coordinates <- function(stack,
								 position,
								 group_id,
								 z) {
	args = lapply(as.list(rlang::call_match()[-1]), eval, envir = parent.frame())
	args$called = names(args)
	args$group_id = args$group_id %||% NA_character_
	args$group_type = c("tm_<other>", "tm_mouse_coordinates")
	args$z = args$z %||% NA_integer_
	tm_element_list(do.call(tm_element, c(args, list(subclass = c("tm_mouse_coordinates", "tm_component")))))
}

#' Map component: minimap
#'
#' Map component that adds a [minimap][leaflet::addMiniMap()] in view mode.
#'
#' @param server name of the provider or an URL (see \code{\link{tm_tiles}}).
#'   By default, it shows the same map as the basemap, and moreover, it will automatically change when the user switches basemaps.
#'   Note the latter does not happen when `server` is specified.
#' @param toggle should the minimap have a button to minimise it? By default \code{TRUE}.
#' @inheritParams tm_title
#' @inheritParams tm_inset
#' @inheritDotParams leaflet::addMiniMap
#' @seealso \href{https://r-tmap.github.io/tmap/articles/basics_components}{Vignette about components}
#' @export
tm_minimap <- function(server,
					   toggle,
					   height,
					   width,
					   margins,
					   between_margin,
					   position,
					   group_id,
					   frame, frame.color, frame.alpha, frame.lwd, frame.r, bg, bg.color, bg.alpha,
					   z,
					   ...) {
	args = lapply(as.list(rlang::call_match()[-1]), eval, envir = parent.frame())
	args$called = names(args)
	args$group_id = args$group_id %||% NA_character_
	args$group_type = c("tm_<other>", "tm_minimap")
	args$z = args$z %||% NA_integer_
	tm_element_list(do.call(tm_element, c(args, list(subclass = c("tm_minimap", "tm_component")))))
}


#' Map component: logo
#'
#' Map component that adds a logo.
#'
#' @param file either a filename or url of a png image. If multiple files/urls are provided with a character vector, the logos are placed near each other.
#'   To specify logos for small multiples use a list of character values/vectors. In order to stack logos vertically, multiple \code{tm_logo} elements can be stacked.
#' @param height height of the logo in number of text line heights. The width is scaled based the height and the aspect ratio of the logo.
#'   If multiple logos are specified by a vector or list, the heights can be specified accordingly.
#' @param margins margins
#' @param between_margin Margin between
#' @inheritParams tm_title
#' @example ./examples/tm_logo.R
#' @seealso \href{https://r-tmap.github.io/tmap/articles/basics_components}{Vignette about components}
#' @export
tm_logo = function(file,
				   height,
				   margins,
				   between_margin,
				   stack,
				   position,
				   group_id,
				   frame,
				   frame.color,
				   frame.alpha,
				   frame.lwd,
				   frame.r,
				   z) {
	args = lapply(as.list(rlang::call_match()[-1]), eval, envir = parent.frame())
	args$called = names(args)
	args$group_id = args$group_id %||% NA_character_
	args$group_type = c("tm_<other>", "tm_logo")
	args$z = args$z %||% NA_integer_
	tm_element_list(do.call(tm_element, c(args, list(subclass = c("tm_logo", "tm_component")))))
}

check_values = function(layer, aes, values) {
	fun_check = paste0("tmapValuesCheck_", aes)

	are_valid = do.call(fun_check, args = list(x = values, is_var = FALSE))
	if (!are_valid) {
		info = attr(are_valid, "info")
		cli::cli_abort(c(
			"x" = "Incorrect values for layer {layer} aesthetic {aes}",
			"!" = "values should conform visual variable {.val {aes}}.",
			"i" = info
		),
		call = call(paste0("tm_", layer[1])))
	}
}

getValuesNames = function(values) {
	if (inherits(values, "tmap_icons")) {
		if ("iconUrl" %in% names(values)) {
			rep("", length(values$iconUrl))
		} else {
			if (!is.null(names(values))) {
				names(values)
			} else {
				rep("", length(values))
			}
		}
	} else {
		if (!is.null(names(values))) {
			names(values)
		} else {
			rep("", length(values))
		}
	}
}

#' Internal tmap function get scale values
#'
#' Internal tmap function get scale values
#'
#' @param scale scale
#' @param o o
#' @param aes aes
#' @param layer layer
#' @param cls cls
#' @param ct ct
#' @export
#' @keywords internal
get_scale_defaults = function(scale, o, aes, layer, cls, ct = NULL) {
	within(scale, {
		values = if (is.na(values[1])) {
			if (is.null(ct)) {
				getAesOption("values.var", o, aes, layer, cls = cls)
			} else {
				ct
			}
		} else values

		value.na = if (is.na(value.na) || isTRUE(value.na)) {
			m = if (aes %in% c("col", "fill")) getPalMeta(as.character(values[1])) else NULL
			ona = getAesOption("value.na", o, aes, layer, cls = cls)

			# take option value.na instead of cols4all palette na-color in these two cases:
			if (is.null(m) || substr(ona, 8, 9) == "00") {
				ona
			} else{
				getPalNA(as.character(values[1]))
			}
		} else {
			value.na
		}

		value.null = if (is.na(value.null)) getAesOption("value.null", o, aes, layer, cls = cls) else value.null
		value.neutral = if (identical(value.neutral, NA)) getAesOption("value.neutral", o, aes, layer, cls = cls) else value.neutral
		values.range = if (is.na(values.range[1])) getAesOption("values.range", o, aes, layer, cls = cls) else values.range
		values.scale = if (is.na(values.scale)) getAesOption("values.scale", o, aes, layer, cls = cls) else values.scale

		value.blank = getAesOption("value.blank", o, aes, layer, cls = cls)
		if (is.na(value.na) || identical(value.na, value.blank)) label.na = ""

		# label.na TRUE: always show NA's, but use option
		# label.na FALSE or "": never show NA's
		# label.na NA: show NA is there are any
		# label.na "qwerty" always snow NA's
		label.show = !isFALSE(label.na) && (isTRUE(label.na) || (!is.na(label.na) && label.na != ""))
		if (is.na(label.na)) label.show = NA # will be TRUE if there are NAs
		if (is.logical(label.na)) label.na = getAesOption("label.na", o, aes, layer, cls = cls)
	})
}

update_na.show = function(label.show, na.show, anyNA) {
	if (is.na(label.show)) {
		if (is.na(na.show)) anyNA else na.show
	} else {
		label.show
	}
}



tmapScale_returnNA = function(n, legend, chart, value.na, label.na, label.show, na.show, sortRev, bypass_ord) {

	ids = if (is.null(sortRev)) {
		NULL
	} else  {
		rep(0L, n)
	}

	if (isFALSE(label.show)) {
		legend = within(legend, {
			title = NA
			nitems = 0
			labels = NA
			dvalues = NA
			vvalues = NA
			vneutral = value.na
			na.show = na.show
			show = FALSE
		})
	} else {
		legend = within(legend, {
			title = legend$title
			nitems = 1
			labels = label.na
			dvalues = NA
			vvalues = value.na
			vneutral = value.na
			na.show = TRUE
		})
	}

	vals = rep(value.na, n)

	if (bypass_ord) {
		format_aes_results(vals, legend = legend, chart = chart)
	} else {
		format_aes_results(vals, ids, legend, chart = chart)
	}

}

tmapLeafletShape = function(bbx, facet_row, facet_col, facet_page, o) {
	bbx = unname(bbx)
	if (!.TMAP$proxy) {
		get_lf(facet_row, facet_col, facet_page) %>%
			leaflet::fitBounds(bbx[1], bbx[2], bbx[3], bbx[4]) %>%
			view_set_bounds(bbx, o) %>%
			assign_lf(facet_row, facet_col, facet_page)
	}
	NULL
}

tmapLeafletOverlay = function(bbx, facet_row, facet_col, facet_page, o) {
	NULL
}

#' Map layer: rgb images
#'
#' Map layer that an rgb image.. The used (multivariate) visual variable is `col`,
#' which should be specified with 3 or 4 variables for `tm_rgb()` and `tm_rgba()` respectively.
#' The first three correspond to the red, green, and blue channels. The optional
#' fourth is the alpha transparency channel.
#'
#' @param col,col.scale,col.legend,col.chart,col.free Visual variable that determines
#'   the color. `col` is a multivariate variable, with 3 (`tm_rgb`) or 4 (`tm_rgba`) numeric data variables. These can be specified via [tm_vars()] with `multivariate = TRUE`
#' @param col_alpha,col_alpha.scale,col_alpha.legend,col_alpha.chart,col_alpha.free  `r .doc_vv("col_alpha")` `r .doc_unit("col_alpha")`
#' @param blend Compositing operator for layer blending. Default `"over"` applies no blending. See the "Layer blending" section of [tm_polygons()] for supported values.
#' @param options options passed on to the corresponding `opt_<layer_function>` function
#' @param ... to catch deprecated arguments from version < 4.0
#' @example ./examples/tm_rgb.R
#' @export
tm_rgb = function(col = tm_vars(n = 3, multivariate = TRUE),
				  col.scale = tm_scale_rgb(),
				  col.legend = tm_legend(),
				  col.chart = tm_chart_none(),
				  col.free = NA,
				  col_alpha = tm_const(),
				  col_alpha.scale = tm_scale(),
				  col_alpha.legend = tm_legend(),
				  col_alpha.chart = tm_chart_none(),
				  col_alpha.free = NA,
				  blend = "over",
				  options = opt_tm_rgb(),
				  ...) {
	args = list(...)

	layer_fun = if ("called_from" %in% names(args)) {
		args$called_from
	} else {
		"tm_rgb"
	}

	if (any(v3_only("tm_rgb") %in% names(args)) || is.numeric(col.scale)) {
		v3_start_message()

		# second condition needed to catch tm_rgb(1, 2, 3)


		if (all(c("r", "g", "b") %in% names(args))) {
			# v3_tm_rgb(args$r, args$g, args$b) moved to step1_helper_facets
			col = tm_vars(c(args$r, args$g, args$b), n = -1, multivariate = TRUE)
		}
		if (is.numeric(col) && is.numeric(col.scale) && is.numeric(col.legend)) {
			col = tm_vars(c(col, col.scale, col.legend), n = -1, multivariate = TRUE)
			col.scale = tm_scale_rgb()
			col.legend = tm_legend()
		}


		if ("alpha" %in% names(args) && layer_fun == "tm_rgb") {
			col_alpha = args$alpha
			v3_message_rgb_alpha(layer_fun = layer_fun)
		}
		if ("saturation" %in% names(args)) {
			options$saturation = args$saturation
			v3_message_rgb_opt(layer_fun = layer_fun, opt = "saturation", args$saturation)
		}
		if ("interpolate" %in% names(args)) {
			options$interpolate = args$interpolate
			v3_message_rgb_opt(layer_fun = layer_fun, opt = "interpolate", args$interpolate)
		}
		if ("max.value" %in% names(args)) {
			col.scale$max_color_value = args$max.value
			v3_message_rgb_maxV(layer_fun = layer_fun, opt = "interpolate", args$max.value)
		}

		v3_start_message()

	}

	# unused arguments: typos?
	unused = setdiff(names(args), c(v3_only("tm_rgb"), "called_from"))

	if (length(unused)) {
		message_layer_unused_args(layer_fun, unused)
	}



	do.call(tm_raster, args = list(col = col, col.scale = col.scale, col.legend = col.legend, col.chart = col.chart, col.free = col.free,
								   col_alpha = col_alpha, col_alpha.scale = col_alpha.scale, col_alpha.legend = col_alpha.legend, col_alpha.chart = col_alpha.chart, col_alpha.free = col_alpha.free,
								   blend = blend,
								   options = options))
}

#' @rdname tm_rgb
#' @export
tm_rgba = function(col = tm_vars(n = 4, multivariate = TRUE),
				  col.scale = tm_scale_rgba(),
				  col.legend = tm_legend(),
				  col.chart = tm_chart_none(),
				  col.free = NA,
				  blend = "over",
				  options = opt_tm_rgb()) {
	do.call(tm_raster, args = list(col = col, col.scale = col.scale, col.legend = col.legend, col.chart = col.chart, col.free = col.free, blend = blend, options = options))
}

#' @rdname tm_rgb
#' @param interpolate Should the raster image be interpolated? Currently only applicable in view mode (passed on to [`grid`][grid::rasterGrob()])
#' @param saturation `r .doc_opt("rgb.saturation")`
#' @export
opt_tm_rgb = function(interpolate = FALSE, saturation = 1) {
	list(trans.args = list(),
		 mapping.args = list(interpolate = interpolate, saturation = saturation))
}

#' tmap function to define a constant visual value
#'
#' tmap function to define a constant visual value
#'
#' @export
tm_const = function() {
	tmapOption("value.const")
}

#' Scales: automatic scale
#'
#' Scales in tmap are configured by the family of functions with prefix `tm_scale`.
#' Such function should be used for the input of the `.scale` arguments in the
#' layer functions (e.g. `fill.scale` in [tm_polygons()]). The function `tm_scale()`
#' is a scale that is set automatically given by the data type (factor, numeric, and integer)
#' and the visual variable. The tmap option `scales.var` contains information
#' which scale is applied when.
#'
#' @param ... arguments passed on to the applied scale function `tm_scale_*()`
#' @seealso [tm_scale_asis()], [tm_scale_ordinal()], [tm_scale_categorical()],
#' [tm_scale_intervals()], [tm_scale_discrete()], [tm_scale_continuous()],
#' [tm_scale_rank()], [tm_scale_continuous_log()], [tm_scale_continuous_log2()],
#' [tm_scale_continuous_log10()], [tm_scale_continuous_log1p()], [tm_scale_continuous_sqrt()],
#' [tm_scale_continuous_pseudo_log()], [tm_scale_rgb()], [tm_scale_bivariate()]
#' @export
tm_scale = function(...) {
	# maybe add the generic scales parameters after ... here?
	structure(c(list(FUN = "tmapScaleAuto"), list(...)), class = c("tm_scale_auto", "tm_scale", "list"))
}

#' Scales: as is
#'
#' Scales in tmap are configured by the family of functions with prefix `tm_scale`.
#' Such function should be used for the input of the `.scale` arguments in the
#' layer functions (e.g. `fill.scale` in [tm_polygons()]).
#' The function `tm_scale_asis()` is used to take data values as they are and use them as such for the visual variable.
#'
#' @inheritParams tm_scale_categorical
#' @param ... Arguments caught (and not used) from the automatic function [tm_scale()]
#' @seealso [tm_scale()]
#' @export
tm_scale_asis = function(values.scale = NA, value.neutral = NA, value.na = NA, ...) {
	structure(c(list(FUN = "tmapScaleAsIs"), c(list(values.scale = values.scale, value.neutral = value.neutral, value.na = value.na), list(...))), class = c("tm_scale_asis", "tm_scale", "list"))
}

#' @export
#' @rdname tm_scale_categorical
tm_scale_ordinal = function(n.max = 30,
							values = NA, # e.g. palette, shape numbers, set of sizes (if two values are specified they are interpret as range)
							values.repeat = FALSE,
							values.range  = 1,
							values.scale = NA,
							value.na = NA, # value for NA
							value.null = NA, # value for NULL (needed?)
							value.neutral = NA, # value for neutral (used in other legends)
							levels = NULL, # levels to show (other values are treated as missing)
							levels.drop = FALSE,
							labels = NULL, # labels for the categories (by default the factor levels)
							label.na = NA, # label for missing values
							label.null = NA,
							label.format = list()) {
	structure(c(list(FUN = "tmapScaleCategorical"), as.list(environment())), class = c("tm_scale_ordinal", "tm_scale", "list"))
}

#' Scales: categorical and ordinal scale
#'
#' Scales in tmap are configured by the family of functions with prefix `tm_scale`.
#' Such function should be used for the input of the `.scale` arguments in the
#' layer functions (e.g. `fill.scale` in [tm_polygons()]).
#' The functions `tm_scale_categorical()` and `tm_scale_ordinal()` are used
#' for categorical data. The only difference between these functions is that the
#' former assumes unordered categories whereas the latter assumes ordered categories.
#' For colors (the visual variable `fill` or `col`), different default color
#' palettes are used (see the tmap option `values.var`).
#'
#' @param n.max Maximum number of categories (factor levels). In case there are more, they are grouped into `n.max` groups.
#' @param values (generic scale argument) The visual values. For colors (e.g. `fill` or `col` for `tm_polygons()`) this is a palette name from the `cols4all` package (see [cols4all::c4a()]) or vector of colors, for size (e.g. `size` for `tm_symbols()`) these are a set of sizes (if two values are specified they are interpret as range), for symbol shapes (e.g. `shape` for [tm_symbols()]) these are a set of symbols, etc. The tmap option `values.var` contains the default values per visual variable and in some cases also per data type.
#' @param values.repeat (generic scale argument) Should the values be repeated in case there are more categories?
#' @param values.range (generic scale argument) Range of the values. Vector of two numbers (both between 0 and 1) where the first determines the minimum and the second the maximum. Full range, which means that all values are used, is encoded as `c(0, 1)`. For instance, when a gray scale is used for color (from black to white), `c(0,1)` means that all colors are used, `0.25, 0.75` means that only colors from dark gray to light gray are used (more precisely `"grey25"` to `"grey75"`), and `0, 0.5` means that only colors are used from black to middle gray (`"gray50"`). When only one number is specified, this is interpreted as the second number (where the first is set to 0). Default values can be set via the tmap option `values.range`.
#' @param values.scale (generic scale argument) Scaling of the values. Only useful for size-related visual variables, such as `size` of [tm_symbols()] and `lwd` of [tm_lines()].
#' @param value.na (generic scale argument) Value used for missing values. See tmap option `"value.na"` for defaults per visual variable.
#' @param value.null (generic scale argument) Value used for NULL values. See tmap option `"value.null"` for defaults per visual variable. Null data values occur when out-of-scope features are shown (e.g. for a map of Europe showing a data variable per country, the null values are applied to countries outside Europe).
#' @param value.neutral (generic scale argument) Value that can be considered neutral. This is used for legends of other visual variables of the same map layer. E.g. when both `fill` and `size` are used for [tm_symbols()] (using filled circles), the size legend items are filled with the `value.neutral` color from the `fill.scale` scale, and fill legend items are bubbles of size `value.neutral` from the `size.scale` scale.
#' @param levels Levels to show. Other values are treated as missing.
#' @param levels.drop Should unused levels be dropped (and therefore are not assigned to a visual value and shown in the legend)?
#' @param labels (generic scale argument) Labels
#' @param label.na (generic scale argument) Label for missing values
#' @param label.null (generic scale argument) Label for null (out-of-scope) values
#' @param label.format (generic scale argument) Label formatting. Output of [tm_label_format()]
#' @seealso [tm_scale()]
#' @export
tm_scale_categorical = function(n.max = 30,
								values = NA, # e.g. palette, shape numbers, set of sizes (if two values are specified they are interpret as range)
								values.repeat = TRUE,
								values.range  = NA,
								values.scale = NA,
								value.na = NA, # value for NA
								value.null = NA, # value for NULL (needed?)
								value.neutral = NA, # value for neutral (used in other legends)
								levels = NULL, # levels to show (other values are treated as missing)
								levels.drop = FALSE,
								labels = NULL, # labels for the categories (by default the factor levels)
								label.na = NA, # label for missing values
								label.null = NA,
								label.format = list()) { # label for null values (needed?)
	structure(c(list(FUN = "tmapScaleCategorical"), as.list(environment())), class = c("tm_scale_categorical", "tm_scale", "list"))
}

#' Scales: interval scale
#'
#' Scales in tmap are configured by the family of functions with prefix `tm_scale`.
#' Such function should be used for the input of the `.scale` arguments in the
#' layer functions (e.g. `fill.scale` in [tm_polygons()]).
#' The function `tm_scale_intervals()` is used for numerical data.
#'
#' @param n Number of intervals. For some styles (see argument `style` below) it is the preferred number rather than the exact number.
#' @param style Method to create intervals. Options are `"cat"`, `"fixed"`, `"sd"`, `"equal"`, `"pretty"`, `"quantile"`, `"kmeans"`, `"hclust"`, `"bclust"`, `"fisher"`, `"jenks"`, `"dpih"`, `"headtails"`, and `"log10_pretty"`. See the details in [classInt::classIntervals()] (extra arguments can be passed on via `style.args`).
#' @param style.args List of extra arguments passed on to [classInt::classIntervals()].
#' @param breaks Interval breaks (only used and required when `style = "fixed"`)
#' @param interval.closure value that determines whether where the intervals are closed: `"left"` or `"right"`. If `as.count = TRUE`, `inverval.closure` is always set to `"left"`.
#' @param label.style Either `"discrete"` or `"continuous"`. If `discrete` (default) intervals will be labeled. If `continuous`, the legend will be similar to a continuous scale legend (but with discrete colors), where the breaks are labeled.
#' @param label.select Which labels are shown? A logical vector is expected, which is repeated along the number of labels. By default `TRUE`, which means all.
#' @param midpoint The data value that is interpreted as the midpoint. By default it is set to 0 if negative and positive values are present. Useful when values are diverging colors. In that case, the two sides of the color palette are assigned to negative respectively positive values. If all values are positive or all values are negative, then the midpoint is set to `NA`, which means that the value that corresponds to the middle color class (see `style`) is mapped to the middle color. If it is specified for sequential color palettes (e.g. `"Blues"`), then this color palette will be treated as a diverging color palette.
#' @param as.count Should the data variable be processed as a count variable? For instance, if `style = "pretty"`, `n = 2`, and the value range of the variable is 0 to 10, then the column classes for `as.count = TRUE` are 0; 1 to 5; 6 to 10 (note that 0 is regarded as an own category) whereas for `as.count = FALSE` they are 0 to 5; 5 to 10. Only applicable if `style` is `"pretty"`, `"fixed"`, or `"log10_pretty"`. By default `FALSE`.
#' @inheritParams tm_scale_categorical
#' @param values (generic scale argument) The visual values. For colors (e.g. `fill` or `col` for [tm_polygons()]) this is a palette name from the `cols4all` package (see [cols4all::c4a()]) or vector of colors, for size (e.g. `size` for `tm_symbols`) these are a set of sizes (if two values are specified they are interpret as range), for symbol shapes (e.g. `shape` for `tm_symbols`) these are a set of symbols, etc. The tmap option `values.var` contains the default values per visual variable and in some cases also per data type.
#' @param values.range (generic scale argument) Range of the values. Vector of two numbers (both between 0 and 1) where the first determines the minimum and the second the maximum. Full range, which means that all values are used, is encoded as `c(0, 1)`. For instance, when a gray scale is used for color (from black to white), `c(0,1)` means that all colors are used, `0.25, 0.75` means that only colors from dark gray to light gray are used (more precisely `"gray25"` to `"gray75"`), and `0, 0.5` means that only colors are used from black to middle grey (`"grey50"`). When only one number is specified, this is interpreted as the second number (where the first is set to 0). Default values can be set via the tmap option `values.range`.
#' @param labels (generic scale argument) Labels
#' @param label.na (generic scale argument) Label for missing values
#' @param label.null (generic scale argument) Label for null (out-of-scope) values
#' @param label.format (generic scale argument) Label formatting. Output of [tm_label_format()]
#' @seealso \href{https://r-tmap.github.io/tmap/articles/basics_scales}{tmap basics: scales}
#' @seealso [tm_scale()]
#' @export
tm_scale_intervals = function(n = 5,
							  style = ifelse(is.null(breaks), "pretty", "fixed"),
							  style.args = list(),
							  breaks = NULL,
							  interval.closure = "left",
							  label.style = "discrete",
							  label.select = TRUE,
							  midpoint = NULL,
							  as.count = FALSE,
							  values = NA,
							  values.repeat = FALSE,
							  values.range  = NA,
							  values.scale = NA,
							  value.na = NA,
							  value.null = NA,
							  value.neutral = NA,
							  labels = NULL,
							  label.na = NA,
							  label.null = NA,
							  label.format = tm_label_format()) {
	structure(c(list(FUN = "tmapScaleIntervals"), as.list(environment())), class = c("tm_scale_intervals", "tm_scale", "list"))
}

#' Scales: discrete scale
#'
#' Scales in tmap are configured by the family of functions with prefix `tm_scale`.
#' Such function should be used for the input of the `.scale` arguments in the
#' layer functions (e.g. `fill.scale` in [tm_polygons()]).
#' The function [tm_scale_discrete()] is used for discrete numerical data, such as integers.
#'
#' @param ticks Discrete values. If not specified, it is determined automatically: unique values are put on a discrete scale.
#' @param midpoint The data value that is interpreted as the midpoint. By default it is set to 0 if negative and positive values are present. Useful when values are diverging colors. In that case, the two sides of the color palette are assigned to negative respectively positive values. If all values are positive or all values are negative, then the midpoint is set to `NA`, which means that the value that corresponds to the middle color class (see `style`) is mapped to the middle color. If it is specified for sequential color palettes (e.g. `"Blues"`), then this color palette will be treated as a diverging color palette.
#' @param values (generic scale argument) The visual values. For colors (e.g. `fill` or `col` for [tm_polygons()]) this is a palette name from the `cols4all` package (see [cols4all::c4a()]) or vector of colors, for size (e.g. `size` for `tm_symbols`) these are a set of sizes (if two values are specified they are interpret as range), for symbol shapes (e.g. `shape` for [tm_symbols()]) these are a set of symbols, etc. The tmap option `values.var` contains the default values per visual variable and in some cases also per data type.
#' @param values.repeat (generic scale argument) Should the values be repeated in case there are more categories?
#' @param values.range (generic scale argument) Range of the values. Vector of two numbers (both between 0 and 1) where the first determines the minimum and the second the maximum. Full range, which means that all values are used, is encoded as `c(0, 1)`. For instance, when a gray scale is used for color (from black to white), `c(0,1)` means that all colors are used, `0.25, 0.75` means that only colors from dark gray to light gray are used (more precisely `"grey25"` to `"grey75"`), and `0, 0.5` means that only colors are used from black to middle grey (`"grey50"`). When only one number is specified, this is interpreted as the second number (where the first is set to 0). Default values can be set via the tmap option `values.range`.
#' @param values.scale (generic scale argument) Scaling of the values. Only useful for size-related visual variables, such as `size` of [tm_symbols()] and `lwd` of [tm_lines()].
#' @param value.na (generic scale argument) Value used for missing values. See tmap option `"value.na"` for defaults per visual variable.
#' @param value.null (generic scale argument) Value used for NULL values. See tmap option `"value.null"` for defaults per visual variable. Null data values occur when out-of-scope features are shown (e.g. for a map of Europe showing a data variable per country, the null values are applied to countries outside Europe).
#' @param value.neutral (generic scale argument) Value that can be considered neutral. This is used for legends of other visual variables of the same map layer. E.g. when both `fill` and `size` are used for [tm_symbols()] (using filled circles), the size legend items are filled with the `value.neutral` color from the `fill.scale` scale, and fill legend items are bubbles of size `value.neutral` from the `size.scale` scale.
#' @param labels (generic scale argument) Labels
#' @param label.na (generic scale argument) Label for missing values
#' @param label.null (generic scale argument) Label for null (out-of-scope) values
#' @param label.format (generic scale argument) Label formatting. Output of [tm_label_format()]
#' @seealso [tm_scale()]
#' @export
tm_scale_discrete = function(ticks = NA,
							 #step = NA,
							 midpoint = NULL,
							 values = NA,
							 values.repeat = FALSE,
							 values.range  = NA,
							 values.scale = NA,
							 value.na = NA,
							 value.null = NA,
							 value.neutral = NA,
							 labels = NULL,
							 label.na = NA,
							 label.null = NA,
							 label.format = list()) {
	structure(c(list(FUN = "tmapScaleDiscrete"), as.list(environment())), class = c("tm_scale_discrete", "tm_scale", "list"))
}

#' Scales: continuous scale
#'
#' Scales in tmap are configured by the family of functions with prefix `tm_scale`.
#' Such function should be used for the input of the `.scale` arguments in the layer
#' functions (e.g. `fill.scale` in [tm_polygons()]).
#' The function `tm_scale_continuous()` is used for continuous data.
#' The functions `tm_scale_continuous_<x>()` use transformation functions x.
#'
#' @param n Preferred number of tick labels. Only used if `ticks` is not specified
#' @param limits Limits of the data values that are mapped to the continuous scale. When `NA`, the range of data values is taken. When only one value is provided, the range of data values with this provided value is taken. The default depends on the visual variable: it is 0 for all visual variables other than color when `tm_scale_continuous` is used. For the transformation scale functions, it is `NA`.
#' @param outliers.trunc Should outliers be truncated? An outlier is a data value that is below or above the respectively lower and upper limit. A logical vector of two values is expected. The first and second value determines whether values lower than the lower limit respectively higher than the upper limit are truncated to the lower respectively upper limit. If `FALSE` (default), they are considered as missing values.
#' @param ticks Tick values. If not specified, it is determined automatically with `n`
#' @param trans Transformation function. One of `"identity"` (default), `"log"`, and `"log1p"`. Note: the base of the log scale is irrelevant, since the log transformed values are normalized before mapping to visual values.
#' @param midpoint The data value that is interpreted as the midpoint. By default it is set to 0 if negative and positive values are present. Useful when values are diverging colors. In that case, the two sides of the color palette are assigned to negative respectively positive values. If all values are positive or all values are negative, then the midpoint is set to `NA`, which means that the value that corresponds to the middle color class (see `style`) is mapped to the middle color. If it is specified for sequential color palettes (e.g. `"Blues"`), then this color palette will be treated as a diverging color palette.
#' @param values (generic scale argument) The visual values. For colors (e.g. `fill` or `col` for [tm_polygons()]) this is a palette name from the `cols4all` package (see [cols4all::c4a()]) or vector of colors, for size (e.g. `size` for [tm_symbols()]) these are a set of sizes (if two values are specified they are interpret as range), for symbol shapes (e.g. `shape` for [tm_symbols()]) these are a set of symbols, etc. The tmap option `values.var` contains the default values per visual variable and in some cases also per data type.
#' @param values.repeat (generic scale argument) Should the values be repeated in case there are more categories?
#' @param values.range (generic scale argument) Range of the values, especially useful for color palettes. Vector of two numbers (both between 0 and 1) where the first determines the minimum and the second the maximum. Full range, which means that all values are used, is encoded as `c(0, 1)`. For instance, when a gray scale is used for color (from black to white), `c(0,1)` means that all colors are used, `0.25, 0.75` means that only colors from dark gray to light gray are used (more precisely `"grey25"` to `"grey75"`), and `0, 0.5` means that only colors are used from black to middle gray (`"grey50"`). When only one number is specified, this is interpreted as the second number (where the first is set to 0). Default values can be set via the tmap option `values.range`.
#' @param values.scale (generic scale argument) Scaling of the values. Only useful for size-related visual variables, such as `size` of [tm_symbols()] and `lwd` of [tm_lines()].
#' @param value.na (generic scale argument) Value used for missing values. See tmap option `"value.na"` for defaults per visual variable.
#' @param value.null (generic scale argument) Value used for NULL values. See tmap option `"value.null"` for defaults per visual variable. Null data values occur when out-of-scope features are shown (e.g. for a map of Europe showing a data variable per country, the null values are applied to countries outside Europe).
#' @param value.neutral (generic scale argument) Value that can be considered neutral. This is used for legends of other visual variables of the same map layer. E.g. when both `fill` and `size` are used for [tm_symbols()] (using filled circles), the size legend items are filled with the `value.neutral` color from the `fill.scale` scale, and fill legend items are bubbles of size `value.neutral` from the `size.scale` scale.
#' @param labels (generic scale argument) Labels
#' @param label.na (generic scale argument) Label for missing values
#' @param label.null (generic scale argument) Label for null (out-of-scope) values
#' @param label.format (generic scale argument) Label formatting. Output of [tm_label_format()]
#' @param trans.args list of additional argument for the transformation (generic transformation arguments)
#' @inheritParams scales::transform_pseudo_log
#' @example ./examples/tm_scale_continuous.R
#' @seealso [tm_scale()]
#' @export
#' @name tm_scale_continuous
tm_scale_continuous = function(n = NULL,
							   limits = NULL,
							   outliers.trunc = NULL,
							   ticks = NULL,
							   trans = NULL,
							   midpoint = NULL,
							   values = NA,
							   values.repeat = FALSE,
							   values.range  = NA,
							   values.scale = NA,
							   value.na = NA,
							   value.null = NA,
							   value.neutral = NA,
							   labels = NULL,
							   label.na = NA,
							   label.null = NA,
							   label.format = tm_label_format(),
							   trans.args = list()) {

	structure(c(list(FUN = "tmapScaleContinuous"), as.list(environment())), class = c("tm_scale_continuous", "tm_scale", "list"))
}


#' Scales: rank scale
#'
#' Scales in tmap are configured by the family of functions with prefix `tm_scale`.
#' Such function should be used for the input of the `.scale` arguments in the layer
#' functions (e.g. `fill.scale` in [tm_polygons()]).
#' The function [tm_scale_rank()] is used to rank numeric data.
#'
#' @param n Preferred number of tick labels. Only used if `ticks` is not specified
#' @param ticks Tick values. If not specified, it is determined automatically with `n`
#' @param values (generic scale argument) The visual values. For colors (e.g. `fill` or `col` for [tm_polygons()]) this is a palette name from the `cols4all` package (see [cols4all::c4a()]) or vector of colors, for size (e.g. `size` for [tm_symbols()]) these are a set of sizes (if two values are specified they are interpret as range), for symbol shapes (e.g. `shape` for [tm_symbols()]) these are a set of symbols, etc. The tmap option `values.var` contains the default values per visual variable and in some cases also per data type.
#' @param values.repeat (generic scale argument) Should the values be repeated in case there are more categories?
#' @param values.range (generic scale argument) Range of the values, especially useful for color palettes. Vector of two numbers (both between 0 and 1) where the first determines the minimum and the second the maximum. Full range, which means that all values are used, is encoded as `c(0, 1)`. For instance, when a gray scale is used for color (from black to white), `c(0,1)` means that all colors are used, `0.25, 0.75` means that only colors from dark gray to light gray are used (more precisely `"grey25"` to `"grey75"`), and `0, 0.5` means that only colors are used from black to middle gray (`"grey50"`). When only one number is specified, this is interpreted as the second number (where the first is set to 0). Default values can be set via the tmap option `values.range`.
#' @param values.scale (generic scale argument) Scaling of the values. Only useful for size-related visual variables, such as `size` of [tm_symbols()] and `lwd` of [tm_lines()].
#' @param value.na (generic scale argument) Value used for missing values. See tmap option `"value.na"` for defaults per visual variable.
#' @param value.null (generic scale argument) Value used for NULL values. See tmap option `"value.null"` for defaults per visual variable. Null data values occur when out-of-scope features are shown (e.g. for a map of Europe showing a data variable per country, the null values are applied to countries outside Europe).
#' @param value.neutral (generic scale argument) Value that can be considered neutral. This is used for legends of other visual variables of the same map layer. E.g. when both `fill` and `size` are used for [tm_symbols()] (using filled circles), the size legend items are filled with the `value.neutral` color from the `fill.scale` scale, and fill legend items are bubbles of size `value.neutral` from the `size.scale` scale.
#' @param labels (generic scale argument) Labels
#' @param label.na (generic scale argument) Label for missing values
#' @param label.null (generic scale argument) Label for null (out-of-scope) values
#' @param label.format (generic scale argument) Label formatting. Output of [tm_label_format()]
#' @param unit The unit name of the values. By default `"rank"`.
#' @seealso [tm_scale()]
#' @export
tm_scale_rank = function(n = NULL,
						 ticks = NULL,
						 values = NA,
						 values.repeat = FALSE,
						 values.range  = NA,
						 values.scale = NA,
						 value.na = NA,
						 value.null = NA,
						 value.neutral = NA,
						 labels = NULL,
						 label.na = NA,
						 label.null = NA,
						 label.format = tm_label_format(),
						 unit = "rank") {
	structure(c(list(FUN = "tmapScaleRank"), as.list(environment())), class = c("tm_scale_rank", "tm_scale", "list"))
}


#' #' @export
#' tm_scale_rank = function(...) {
#' 	tmc = tm_scale_continuous(...)
#' 	class(tmc) = c("tm_scale_rank", "tm_scale", "list")
#' 	tmc
#' }

#' @export
#' @param ... passed on to [tm_scale_continuous()]
#' @rdname tm_scale_continuous
tm_scale_continuous_log = function(..., base = exp(1)) {
	tm_scale_continuous(trans = "log", trans.args = list(base = base), ...)
}

#' @export
#' @rdname tm_scale_continuous
tm_scale_continuous_log2 = function(...) {
	tm_scale_continuous(trans = "log", trans.args = list(base = 2), ...)
}

#' @export
#' @rdname tm_scale_continuous
tm_scale_continuous_log10 = function(...) {
	x = tm_scale_continuous(trans = "log", trans.args = list(base = 10), ...)
	class(x) = c("tm_scale_continuous_log10", "tm_scale", "list")
	x
}


#' @export
#' @rdname tm_scale_continuous
tm_scale_continuous_log1p = function(...) {
	tm_scale_continuous(trans = "log1p", ...)
}

#' @export
#' @rdname tm_scale_continuous
tm_scale_continuous_sqrt = function(...) {
	tm_scale_continuous(trans = "sqrt", ...)
}

#' @export
#' @rdname tm_scale_continuous
tm_scale_continuous_pseudo_log = function(..., base = exp(1), sigma = 1) {
	tm_scale_continuous(trans = "pseudo_log", trans.args = list(base = base, sigma = sigma), ...)
}

#
# #' @export
# #' @rdname tm_scale_continuous
# tm_scale_continuous_logistic = function(...) {
# 	tm_scale_continuous(trans = "logistic", ...)
# }


#' Scales: RGB
#'
#' Scales in tmap are configured by the family of functions with prefix `tm_scale`.
#' Such function should be used for the input of the `.scale` arguments in the layer
#' functions (e.g. `fill.scale` in [tm_polygons()]).
#' The function [tm_scale_rgb()] is used to transform r, g, b band variables to colors. This function is adopted from (and works similar as) [stars::st_rgb()]
#'
#' @param value.na value for missing values
#' @param stretch should each (r, g, b) band be stretched? Possible values: `"percent"` (same as `TRUE`), `"histogram"`, `FALSE`.
#'   In the first case, the values are stretched to `probs[1]...probs[2]`. In the second case, a histogram equalization is performed
#' @param probs probability (quantile) values when `stretch = "percent"`
#' @param max_color_value maximum value
#' @seealso [tm_scale()] and [stars::st_rgb()]
#' @example ./examples/tm_scale_rgb.R
#' @export
tm_scale_rgb = function(value.na = NA,
						stretch = FALSE,
						probs = c(0, 1),
						max_color_value = 255L) {
	structure(c(list(FUN = "tmapScaleRGB"), as.list(environment())), class = c("tm_scale_rgb", "tm_scale", "list"))
}

#' @rdname tm_scale_rgb
#' @export
tm_scale_rgba = function(value.na = NA,
						 stretch = FALSE,
						 probs = c(0, 1),
						 max_color_value = 255) {
	structure(c(list(FUN = "tmapScaleRGBA"), as.list(environment())), class = c("tm_scale_rgba", "tm_scale", "list"))
}

# tm_scale_na = function() {
# 	structure(c(list(FUN = "tmapScaleNA"), as.list(environment())), class = c("tm_scale_na", "tm_scale", "list"))
# }

# tme_graphics = function(init, polygons, symbols, lines, raster) {
# 	structure(list(init = init, polygons = polygons, symbols = symbols, lines = lines, raster = raster), class = "tme_graphics")
# }
# 
# 
# 
# tme_graphics.grid = tme_graphics(
# 	require = "grid",
# 	init = function(bbx) {
# 		if (!requireNamespace("grid")) stop("grid package required but not installed yet.")
# 		grid.newpage()
# 		grid::pushViewport(viewport(xscale = st_bbox(World)[c(1,3)], yscale = st_bbox(World)[c(2,4)]))
# 	},
# 	polygons = function(x, col, alpha, border.col, border.lwd, border.lty) {
# 		geoms = sf::st_geometry(x)
# 		gp = gpar(fill=col, col=border.col, lwd=border.lwd, lty=border.lty)
# 		sf::st_as_grob(geoms, gp = gp)
# 	},
# 	symbols = function(x, size, col, shape, alpha, border.col, border.lwd, border.lty) {
# 		geoms = sf::st_centroid(sf::st_geometry(x))
# 		co = sf::st_coordinates(geoms)
# 		gp = gpar(fill=col, col=border.col, lwd=border.lwd, lty=border.lty)
# 		
# 		pointsGrob(x=co[,1], y = co[,2],
# 				   size = size,
# 				   pch = shape,
# 				   gp = get_symbol_gpar(x=symbol.shape2,
# 				   				   fill=symbol.col2,
# 				   				   col=bordercol,
# 				   				   lwd=symbol.border.lwd), 
# 				   name=idName)
# 	}
# 	
# 	
# )
# 
# 
# tme_grid = function(init, )
# 
# 
# 
# 
# t_polygon = function(x, col, alpha, border.col, border.lwd, border.lty) {
# 	
# 	geoms = sf::st_geometry(x)
# 	
# 	gp=gpar(fill=col, col=border.col, lwd=border.lwd, lty=border.lty)
# 	
# 	st_as_grob(geoms, gp = gp)
# }
# 

# function to update the levels fl per facet dimension
update_grp_vars = function(lev = NULL, m = NULL) {
	vl = get("vl", envir = .TMAP)
	vn = get("vn", envir = .TMAP)

	if (!missing(lev)) {
		m = length(lev)
		if (m == 1L && vn == 1L) {
			if (is.null(vl)) {
				vl = lev
				vn = length(lev)
			}
		} else if (m > 1L && vn > 1L && m != vn) {
			stop("Inconsistent number of map variable values.", call. = FALSE)
		} else if (m > vn) {
			vl = lev
			vn = length(lev)
		}
	} else if (!missing(m)) {
		if (m > 1L && vn > 1L && m != vn) {
			stop("Inconsistent number of map variable values.", call. = FALSE)
		} else if (m > vn) {
			vn = m
			vl = NULL
		}
	}
	assign("vl", vl, envir = .TMAP)
	assign("vn", vn, envir = .TMAP)
}

add_used_vars = function(v, convert2density = FALSE, is_title = FALSE) {
	if (convert2density) {
		c2d_vars = get("c2d_vars", envir = .TMAP)
		c2d_vars = unique(c(c2d_vars, v))
		assign("c2d_vars", c2d_vars, envir = .TMAP)
	}

	used_vars = get("used_vars", envir = .TMAP)
	used_vars = unique(c(used_vars, v))

	if (is.null(attr(used_vars, "is_title"))) {
		attr(used_vars, "is_title") = is_title
	} else {
		attr(used_vars, "is_title") = attr(used_vars, "is_title") || is_title
	}

	assign("used_vars", used_vars, envir = .TMAP)
}

get_split_stars_dim = function(lst) {
	if (length(lst)) {
		y = vapply(lst, function(x) x$split_stars_dim, FUN.VALUE = character(1))
		if (any(y != "")) y[which(y != "")[1]] else ""
	} else {
		""
	}
}

is_rev = function(x) !is.null(x) && substr(x, 1, 1) == "-"
remove_min = function(x) substr(x, 2, nchar(x))

# ## estimate number of facets
step1_rearrange_facets = function(tmo, o) {
	nframes = NULL

	#o = tmap_options_mode()
	dev = getOption("tmap.devel.mode")


	# get the final tm_faets object (ignoring group specific args: is.wrap, by, rows, columns, pages)

	#fl = list(1L, 1L, 1L)
	#assign("fl", fl, envir = .TMAP)

	#assign("fl", list(NULL, NULL, NULL), envir = .TMAP)
	#assign("fn", c(1L, 1L, 1L), envir = .TMAP)

	tmo = lapply(tmo, function(tmg) {

		shp = tmg$tms$shp
		layer = tmg$tms$layer

		shp_is_pointer = inherits(shp, "character")

		smeta = tmapGetShapeMeta1(shp, layer, c(o, tmg$tmf))

		if (dev) timing_add(s3 = "get_shape_meta1")


		assign("vl", NULL, envir = .TMAP)
		assign("vn", 1L, envir = .TMAP)

		assign("used_vars", character(0), envir = .TMAP)
		assign("c2d_vars", character(0), envir = .TMAP)

		precheck_aes = function(a, layer, shpvars, args) {
			b = within(a, {
				if (inherits(value, "tmapAsIs")) {
					if (inherits(scale, "tm_scale_auto")) {
						class(scale) = c("tm_scale_asis", "tm_scale", "list")
						scale$FUN = tmapScaleAsIs
					}
				}

				# check if dimension is or should be used
				if (inherits(value, "tmapVars") && (length(smeta$dims) != 0) && (!is.null(value$dimvalues) || (!is.na(value$x[1]) && length(value$x) == 1L) || (value$multivariate && length(shpvars) == 1L))) {
					if (is.na(value$x)[1]) {
						value$x = smeta$dims[1]
					} else {
						if (length(value$x) > 1L) {
							if (identical(value$n, -1)) { # shortcut used for v3 compat: tm_rgb(r=3,g=2,=b1) when not known if shp had a band
								v3_tm_rgb(value$x[1], value$x[2], value$x[3], dim = smeta$dims[1])

								value$dimvalues = value$x
								value$x = smeta$dims[1]
								value$n = NA
							} else {
								warning("dimvalues specified while more than one dimension name is specified. Only the first will be used", call. = FALSE)
								value$x = value$x[1]
							}
						}
						if (!(value$x %in% smeta$dims)) {
							stop("dimvalues specified, but dimension \"" , x, "\" not found. Available dimension name(s): ", paste(smeta$dims,collapse = ", "), call. = FALSE)
						}
					}
					if (is.null(value$dimvalues)) {
						if (!is.na(value$n)) {
							value$dimvalues = smeta$dims_vals[[value$x]][1L:value$n]
						} else {
							if (!all(value$dimvalues) %in% smeta$dims_vals[[value$x]]) stop("Incorrect dimvalues", call. = FALSE)
						}
					}

					split_stars_dim = value$x
					add_used_vars(value$dimvalues)

					# redefine value for step 2
					if (value$multivariate) {
						update_grp_vars(lev = value$x)
						value = structure(list(as.character(value$dimvalues)), names = paste(as.character(value$dimvalues), collapse = "_"), class = "tmapVars")
					} else {
						update_grp_vars(lev = value$dimvalues)
						value = structure(as.list(as.character(value$dimvalues)), names = as.character(value$dimvalues), class = "tmapVars")
					}

					data_vars = TRUE
					bypass_vars = FALSE
					geo_vars = FALSE

				} else {
					split_stars_dim = ""

					#value_orig = value # just for the case of L156
					if (!inherits(value, c("tmapOption", "tmapVars", "tmapAsIs", "tmapSpecial")) && length(value) && is.na(value[[1]][1])) {
						# NA -> value.blank
						value = tmapVV(getAesOption("value.blank", o, aes = aes, layer = layer))
					}

					if (inherits(value, "tmapOption")) {
						#value_orig = tmapVV(getAesOption(value[[1]], o, aes = aes, layer = layer))
						value = tmapVV(getAesOption(value[[1]], o, aes = aes, layer = layer))
						data_vars = FALSE
						bypass_vars = FALSE
						geo_vars = FALSE
						#if (!is.list(value_orig)) value = list(value_orig)
						#value = value_orig
						#names(value) = sapply(value, "[", 1)
					} else if (inherits(value, "tmapVars")) {
						if (identical(value$n, -1)) { # shortcut used for v3 compat: tm_rgb(r=3,g=2,=b1) when not known if shp had a band
							v3_tm_rgb(value$x[1], value$x[2], value$x[3])
						}

						if (!is.null(value$dimvalues) && length(smeta$dims) == 0L) {
							error_dimvalues()
						}
						if (!is.na(value$x[1])) {
							if (is.character(value$x)) {
								if (!all(value$x %in% shpvars)) stop("not all variables specified in tm_vars are found", call. = FALSE)
								vars = value$x
							} else {
								if (!all(value$x %in% 1L:length(shpvars))) stop("tm_vars defined for x = ", paste(value$ids, collapse = ", "), " while there are only ", length(shpvars), " variables", call. = FALSE)
								vars = shpvars[value$x]
							}
						} else if (!is.na(value$n)) {
							if (length(shpvars) < value$n) {
								stop("tm_vars defined for n = ", value$n, " while there are only ", length(shpvars), " variables", call. = FALSE)
							}
							vars = shpvars[1L:value$n]
						} else {
							vars = shpvars
						}
						names(vars) = vars

						if (value$animate) {
							.TMAP$animate = c(.TMAP$animate, vars)
						}

						if (value$multivariate) {
							value = structure(list(unname(vars)), names = paste(vars, collapse = "_"), class = "tmapStandard")
						} else {
							value = structure(as.list(vars), class = "tmapStandard")
						}

						data_vars = !shp_is_pointer
						bypass_vars = shp_is_pointer
						geo_vars = FALSE
					} else {
						if (inherits(value, "tmapStandard")) {
							uvalue = unlist(value)
							data_vars = all(uvalue %in% shpvars) && !shp_is_pointer

							bypass_vars = (length(shpvars) == 0 || all(uvalue %in% shpvars)) && shp_is_pointer
							geo_vars = all(uvalue %in% c("AREA", "LENGTH", "MAP_COLORS")) && !data_vars
							if (geo_vars) {
								w = which(uvalue == "AREA")[1]
								if (!is.na(w)) {
									uvalue[w] = "AREA_"
									names(uvalue)[w] = "AREA_"
									w2 = which(names(value) == "AREA")[1]
									value[[w2]] = "AREA_"
									names(value)[w2] = "AREA_"
								}
								w = which(uvalue == "LENGTH")[1]
								if (!is.na(w)) {
									uvalue[w] = "LENGTH_"
									names(uvalue)[w] = "LENGTH_"
									w2 = which(names(value) == "LENGTH")[1]
									value[[w2]] = "LENGTH_"
									names(value)[w2] = "LENGTH_"
								}
							}

							if (data_vars || geo_vars) vars = uvalue else vars = character(0)
						} else {
							data_vars = FALSE
							bypass_vars = FALSE
							geo_vars = FALSE
							vars = character(0)
						}
					}

					nvars = length(value) #m
					nvari = vapply(value, length, integer(1))

					convert2density = "convert2density" %in% names(scale) && scale$convert2density

					nflvar = nvars
					if (data_vars) {
						flvar = names(value)
						update_grp_vars(lev = flvar)
						add_used_vars(vars, convert2density = convert2density)
					} else if (geo_vars) {
						flvar = names(value)
						update_grp_vars(lev = flvar)
						add_used_vars(vars)
					} else {
						mfun = paste0("tmapValuesSubmit_", aes)
						if (exists(mfun)) {
							value = do.call(mfun, list(x = value, args = args))
						}
						nvars = length(value)
						nflvar = nvars
						nvari = vapply(value, length, integer(1))
						#value =

						#value = value_orig
						update_grp_vars(m = nflvar)
					}
				}
			})
			b
		}

		gs = tmap_graphics_name()

		# preprocess layers: check aes values
		tmg$tmls = lapply(tmg$tmls, function(tml) {
			within(tml, {
				mapping.fun = paste0("tm_data_", layer)
				if (!(mapping.fun[1] %in% get_method_classes(paste0("tmap", gs, "DataPlot")))) {

					mode = getOption("tmap.mode")

					modes = get_modes()
					gss = vapply(modes, tmap_graphics_name, FUN.VALUE = character(1))

					exs = vapply(1L:length(modes), function(i) {
						mapping.fun[1] %in% get_method_classes(paste0("tmap", gss[i], "DataPlot"))
					}, FUN.VALUE = logical(1))
					modes2 = modes[exs]

					if (length(modes2)) {
						cli::cli_inform("{.field [{mode} mode]} map layer {.code tm_{layer[1]}} not available for mode {.str {mode}}. This map layer is only available for modes {.str {modes2}}")
					} else {
						cli::cli_inform("{.field [{mode} mode]} map layer {.code tm_{layer[1]}} not available for mode {.str {mode}} and also not for the other modes")
					}
					remove = TRUE
				} else {
					remove = FALSE
					if (length(trans.aes)) trans.aes = lapply(trans.aes, precheck_aes, layer = tml$layer, shpvars = smeta$vars, args = trans.args)
					if (length(mapping.aes)) mapping.aes = lapply(mapping.aes, precheck_aes, layer = tml$layer, shpvars = smeta$vars, args = mapping.args)

					# get first non-empty split_stars_dim, the dimension specificied by tm_vars_dim
					split_stars_dim = get_split_stars_dim(mapping.aes)

					if (length(hover) > 1) {
						cli::cli_abort("hover should have length <= 1, not {length(hover)}.", call = NULL)
					}

					if (is.na(hover)) {
						hover = id
					} else if (is.logical(hover)) {
						hover = ifelse(hover, id, "")
					}

					if (hover != "" && !hover %in% smeta$vars) rlang::arg_match0(hover, smeta$vars, arg_nm = "hover", error_call = NULL)
					if (hover != "") add_used_vars(hover, is_title = TRUE)
					if (id != "" && !id %in% smeta$vars) rlang::arg_match0(id, smeta$vars, arg_nm = "id", error_call = NULL)
					if (id != "") add_used_vars(id, is_title = TRUE)

					if (isTRUE(popup.vars)) {
						popup.vars = smeta$vars
					} else if (isFALSE(popup.vars)) {
						popup.vars = character(0)
					} else if (is.na(popup.vars[1])) {
						popup.vars = smeta$vars
					}

					if (!all(popup.vars %in% smeta$vars)) {
						rlang::arg_match(popup.vars, values = smeta$vars, multiple = TRUE)
					}
					if (length(popup.vars)) add_used_vars(popup.vars)


					if (length(popup.format) != 0 && !is.null(names(popup.format)) && all(names(popup.format) %in% popup.vars)) {
						popup.called = names(popup.format)
						popup.format = lapply(popup.vars, function(pv) {
							if (pv %in% names(popup.format)) {
								process_label_format(popup.format[[pv]], o$label.format)
							} else {
								process_label_format(list(), o$label.format)
							}
						})
					} else {
						popup.called = character(0)
						one.popup.format = process_label_format(popup.format, o$label.format)
						popup.format = lapply(popup.vars, function(pv) {
							one.popup.format
						})

					}
					if (!is.null(names(popup.vars))) {
						names(popup.format) = names(popup.vars)
					} else {
						names(popup.format) = popup.vars
					}
					attr(popup.format, "called") = popup.called				}
			})
		})

		tmg$tmls = tmg$tmls[vapply(tmg$tmls, function(l) {
			!l$remove
		}, FUN.VALUE = logical(1))]


		vl = get("vl", envir = .TMAP)
		vn = get("vn", envir = .TMAP)

		# split stars if needed (dimension -> attributes)
		split_stars_dim = get_split_stars_dim(tmg$tmls)
		shp = tmapSplitShp(shp, split_stars_dim, smeta)
		if (split_stars_dim != "") {
			smeta = tmapGetShapeMeta1(shp, layer, o)
			if (dev) timing_add(s3 = "get_shape_meta1_2")
		}

		nrsd = length(smeta$dims) # number of required shape dimensions
		nrvd = as.integer(vn > 1L) # number of required variable dimensions (0 or 1)
		nrd = nrsd + nrvd # number of required by-dimensions


		tmg$tmf = within(tmg$tmf, {
				by1 = rows
				by2 = columns
				by3 = pages

				trans_animate = !is.null(.TMAP$animate)
				nad = as.integer(trans_animate)

				if (trans_animate) {
					if (is.null(by3)) {
						by3 = "FRAME__"
					} else {
						if (by3 == "VARS__") {
							if (nrvd == 0) {
								by3 = "FRAME__"
							} else {
								if (is.null(by1)) {
									by1 = "VARS__"
									by3 = "FRAME__"
								} else if (is.null(by2)) {
									by2 = "VARS__"
									by3 = "FRAME__"
								} else {
									ani = names(.TMAP$animate)
									cli::cli_abort("transformation animation for variable {.val {ani}} cannot be used when pages is already specified to multiple variables")
								}

							}
						} else {
							ani = names(.TMAP$animate)
							cli::cli_abort("transformation animation for variable {.val {ani}} cannot be used when pages is already specified to {.val {by3}}")
						}
					}
				}

				# if (animate) {
				# 	if (!is.null(by3) {
				#
				# 		if (nrvd == 1) {
				#
				# 	} else if ()
				#
				# 	}
				# }


				if (!is.null(by)) {
					if (is.null(by1)) {
						by1 = by
					} else if (is.null(by2)) {
						by2 = by
					} else if (is.null(by3)) {
						by3 = by
					} else {
						stop("Too many dimensions") # quick&dirty, to do: move to other checks
					}
				}


				rev1 = is_rev(by1)
				rev2 = is_rev(by2)
				rev3 = is_rev(by3)

				if (rev1) by1 = remove_min(by1)
				if (rev2) by2 = remove_min(by2)
				if (rev3) by3 = remove_min(by3)




				if (nrd > 3L) {
					if (nrsd > 3L) cli::cli_abort("The shape object has more than 3 dimensions, so even {.fn tm_facets_grid} cannot be used.")
					nrvd = 0L
					nrd = 3L
					limitvars = TRUE
				} else {
					limitvars = FALSE
					#todo
					if (nrvd == 1L && !identical(by1, "VARS__") && !identical(by2, "VARS__") && !identical(by3, "VARS__")) {
						if (is.null(by1)) {
							by1 = "VARS__"
						} else if (is.null(by2)) {
							by2 = "VARS__"
						} else if (is.null(by3)) {
							by3 = "VARS__"
						}
					}
				}
				unsigned = setdiff(smeta$dims, c(by1, by2, by3))

				if (length(unsigned)) {
					if (is.null(by1)) by1 = unsigned[1] else if (is.null(by2)) by2 = unsigned[1] else by3 = unsigned[1]
					if (length(unsigned) > 1) {
						if (is.null(by2)) by2 = unsigned[2] else by3 = unsigned[2]
					}
					if (length(unsigned) == 3) {
						by3 = unsigned[3]
					}
				}

			# }

			bys = c(by1, by2, by3)
			if (length(bys)) {
				byvars = intersect(smeta$vars, bys)
				if (length(byvars)) add_used_vars(byvars)
			}
			if (!all(bys %in% c("VARS__", "FRAME__", smeta$vars, smeta$dims))) stop("unknown facet variables", call. = FALSE)
			if (is.na(na.text)) na.text = o$label.na

			if (is.na(type)) {
				type = if (!is.null(by1) && !is.null(by2)) "grid" else "wrapstack"
			}

		})

		smeta$vars = get("used_vars", envir = .TMAP)
		convert2density = get("c2d_vars", envir = .TMAP)
		smvars = if (length(convert2density)) unique(c(smeta$vars, "AREA")) else smeta$vars
		shp = tmapSubsetShp(shp, smvars)

		if (length(convert2density)) {
			for (v in convert2density) {
				sunit = tmg$tms$unit
				if (is.null(sunit)) sunit = o$unit
				shape.unit <- ifelse(sunit == "metric", "km", ifelse(sunit == "imperial", "mi", sunit))
				u = paste(shape.unit, shape.unit)
				if (is.numeric(shp[[v]])) shp[[v]] = shp[[v]] / units::set_units(shp$AREA, u, mode = "standard")
			}
		}


		if (dev) timing_add(s3 = "subset_shp")

		smeta = tmapGetShapeMeta2(shp, smeta, c(o, tmg$tmf))
		if (dev) timing_add(s3 = "get_shape_meta2")

		tmg$tmf = within(tmg$tmf, {

			gl = list(NULL, NULL, NULL)
			gn = c(1L, 1L, 1L)

			for (i in 1L:3L) {
				byi = get(paste0("by", i))
				if (!is.null(byi)) {
					if (byi == "FRAME__") {
						gl[i] = list(NULL)
						gn[i] = as.integer(nframes)
					} else
					if (byi == "VARS__") {
						gl[i] = list(vl)
						gn[i] = vn
					} else if (byi %in% smeta$vars) {
						gl[i] = list(smeta$vars_levs[[byi]])
						gn[i] = length(gl[[i]])
					} else if (byi %in% smeta$dims) {
						gl[[i]] = smeta$dims_val[[match(byi, smeta$dims)]]
						gn[i] = length(gl[[i]])
					}
				}
			}

			if (is.na(free.coords)) {
				if (type %in% c("wrapstack", "wrap", "stack")) {
					free.coords = rep(!any(c(by1, by2, by3) == "VARS__"), 3)
				} else {
					free.coords = c((!is.null(rows) && (rows != "VARS__")), (!is.null(columns)) && (columns != "VARS__"), (!is.null(pages)) && (pages != "VARS__"))
				}
			} else {
				free.coords = rep(free.coords, length.out = 3)
			}


			v = which(vapply(list(by1, by2, by3), identical, FUN.VALUE = logical(1), "VARS__"))

			b = setdiff(which(!vapply(list(by1, by2, by3), is.null, FUN.VALUE = logical(1))), v)

			by123 = paste0("by", 1L:3L)
			by123__ = paste0("by", 1L:3L, "__")

			var__ = by123__[v]
			by__ = by123__[b]


			#fl1 =
		})
		tmg$tms$shp = shp
		tmg$tms$smeta = smeta

		tmg
	})

	tmf = get_tmf(lapply(tmo, function(tmoi) tmoi$tmf))

	tmo$tmf_global = tmf
	tmo
}

complete_with_comp_group = function(comp, o) {
	# complete all non-called options from tm_legend and tm_<component> with the tm_components specs

	comp_grps = names(o)[grepl("component_", names(o))]
	grps = gsub(".*component_", "", comp_grps)
	grps_lst = strsplit(grps, "^", fixed = TRUE)
	grps_lst = lapply(grps_lst, function(l) {
		if (length(l) == 0) "" else l
	})
	if (is.na(comp$group_id)) {
		comp$group_id = paste(comp$position$type, comp$position$cell.h, comp$position$cell.v, comp$position$pos.h, comp$position$pos.v, comp$position$just.h, comp$position$just.v, sep = "_")
	}

	ids = c("", comp$group_id, comp$group_type)
	comp_ids = paste("component", ids, sep = "_")


	# take called from component itself (component function call)
	if ("called" %in% names(comp)) {
		comp$called_via_comp_group = comp$called
	} else {
		comp$called_via_comp_group = character()
	}

	for (i in seq_along(grps)) {
		if (length(intersect(grps_lst[[i]], ids))) {
			oc = o[[comp_grps[i]]]
			if ("position" %in% names(oc)) {
				oc$position = process_position(oc$position, o)
			}
			if (length(oc)) {
				comp[names(oc)] = oc
				comp$called_via_comp_group = unique(c(comp$called_via_comp_group, names(oc)))
			}
		}
	}

	# TO DO:
	# - loop through grp_ids(2)
	# - if any ids is equal to grp, do old script

	comp
}


impute_comp = function(a, o) {
	ca = class(a)

	call = names(a)
	a$position = process_position(a$position, o)
	a$padding = process_padding(a$padding)

	ot = get_prefix_opt(class = ca[1], o = o)
	if (!is.null(a$position)) ot$position = NULL
	a = complete_options(a, ot)

	# add options from parent classes
	if (length(ca) > 1L) {
		for (i in 2L:max(2L, length(ca)-2L)) {
			oti = get_prefix_opt(class = ca[i], o = o)
			if (!is.null(a$position) || ("tm_add_legend" %in% ca)) oti$position = NULL
			# otherwise the position lists (in and out) get mixed
			# for tm_add_legend, keep it empty because it will be imputed in update_l (using legend. defaults)
			a = complete_options(a, oti)
		}
	}

	if (!("tm_add_legend" %in% ca)) a = complete_with_comp_group(a, o)

	a$call = call

	#a$position = NULL

	if ("margins" %in% names(a)) a$margins = rep_len(a$margins, 4L)

	class(a) = ca
	a
}



update_l = function(o, l, v, mfun, unm, active) {
	# update legend options
	oltype = o["legend.orientation"]
	names(oltype) = "orientation"
	if (all(v %in% c("AREA_", "LENGTH_", "MAP_COLORS")) && is.null(l$show)) {
		l$show = FALSE
	}

	call = names(l)

	l = complete_options(l, oltype)
	oleg = o[names(o)[substr(names(o), 1, 6) == "legend" & substr(names(o), 1, 15) != "legend.settings"]]
	names(oleg) = substr(names(oleg), 8, nchar(names(oleg)))
	settings_name = paste0("legend.settings.", l$orientation)
	oleg = c(oleg, o[[settings_name]])


	if ("position" %in% names(l) && !is.null(l$position)) {
		l$position = process_position(l$position, o)
		oleg$position = NULL
	}


	l = complete_options(l, oleg)


	# general component options
	ot2 = get_prefix_opt(class = "tm_component", o = o)
	ot2$position = NULL
	l = complete_options(l, ot2)


	l$call = call
	l$mfun = mfun
	l$unm = unm
	l$active = active

	l = complete_with_comp_group(l, o)

	# update legend class
	class(l) = c(paste0("tm_legend", ifelse(!is.null(l$orientation), paste0("_", l$orientation), "")), "tm_legend", "tm_component", class(l))
	l
}

update_crt = function(o, crt, v, mfun, unm, active) {

	#crt_options
	cls = class(crt)

	call = names(crt)

	ocrt = o[substr(names(o), 1, 6) == "chart."]
	names(ocrt) = substr(names(ocrt), 7, nchar(names(ocrt)))


	if ("position" %in% names(crt)) crt$position = process_position(crt$position, o)

	crt = complete_options(crt, ocrt)


	# general component options
	ot2 = get_prefix_opt(class = "tm_component", o = o)
	ot2$position = NULL
	crt = complete_options(crt, ot2)

	crt = complete_with_comp_group(crt, o)


	crt$call = call
	crt$mfun = mfun
	crt$unm = unm
	crt$active = active


	# update legend class
	#class(l) = c(paste0("tm_legend_", l$design, ifelse(!is.null(l$orientation), paste0("_", l$orientation), "")), class(l))
	#l
	class(crt) = cls
	crt
}

warning_group_args = function(args) {
	old = c("group.frame", "resize_as_group", "stack")
	new = c("frame_combine", "resize_as_group", "stack")

	x = intersect(old, names(args))

	if (length(x)) {
		s = do.call(paste, c(mapply(function(o,n,v) {
			cli::format_inline(paste0("{n} = {.val {v}}"))
		}, old[match(x, old)], new[match(x, old)], args[x], SIMPLIFY = FALSE), list(sep = ", ")))

		cli::cli_warn(paste0("Component group arguments, such as {.var {x}}, are deprecated as of 4.1. Please use {.code group_id = {.val ID}} in combination with {.code tm_components(", s, ")} instead."))

		if ("group.frame" %in% names(args)) {
			args$frame_combine = args$group.frame
			args$group.frame = NULL
		}
	}
	args
}

#' Layer group control
#'
#' Controls the layer groups in interactive maps (view mode): the layer control box (radio buttons or check boxes) and at which zoom levels the layers are displayed at.
#'
#' @param name group name that corresponds with the group name specified in the layer functions (e.g.  [tm_polygons()])
#' @param control The group control determines how
#'   layer groups can be switched on and off. Options: `"radio"` for radio
#'   buttons (meaning only one group can be shown), `"check"` for check boxes
#'   (so multiple groups can be shown), and `"none"` for no control
#'   (the group cannot be (de)selected).
#' @param zoom_levels The zoom levels at which the group is displays at. When specified `control` will be set to `"none"`.
#' @seealso `r .doc_see_also_groups()`
#' @export
tm_group = function(name, control = NA, zoom_levels = NA) {
	optname = paste0("view_group_", name)
	x = structure(list(list(name = name, control = control, zoom_levels = zoom_levels)), names= optname)
	do.call(tm_options, x)
}

tmapGridInit = function(o, show = TRUE, newpage = TRUE, return.asp = FALSE, vp, prx, ...) {
	rlang::check_installed("grid")

	rows = with(o, {
		x = c(outer.margins.top = outer.margins[3],
			  meta.buffers.top.out = meta.buffers[3],
			  meta.margins.top = meta.margins[3],
			  meta.buffers.top.in = meta.buffers[3],
			  xylab.margins.top = xylab.margins[3],

			  panel.xtab.top = panel.xtab.size[3],
			  panel.xtab.margin.top = panel.xtab.margin[3],
			  grid.buffers.top = grid.buffers[3],
			  grid.margins.top = grid.margins[3],

			  {if (o$nrows > 1) rep(c(panel.wrap.size[3], panel.wrap.margin[3], 0, panel.wrap.margin[1], panel.wrap.size[1], between_marginH), o$nrows -1) else NULL},
			                          panel.wrap.size[3], panel.wrap.margin[3], 0, panel.wrap.margin[1], panel.wrap.size[1],

			  grid.margins.bottom = grid.margins[1],
			  grid.buffers.bottom = grid.buffers[1],
			  panel.xtab.margin.bottom = panel.xtab.margin[1],
			  panel.xtab.bottom = panel.xtab.size[1],

			  xylab.margins.bottom = xylab.margins[1],
			  meta.buffers.bottom.in = meta.buffers[1],
			  meta.margins.bottom = meta.margins[1],
			  meta.buffers.bottom.out = meta.buffers[1],
			  outer.margins.bottom = outer.margins[1])

		u = grid::unit(x, "npc")
		names(u) = names(x)
		u
	})

	cols = with(o, {
		x = c(outer.margins.left = outer.margins[2],
			  meta.buffers.left.out = meta.buffers[2],
			  meta.margins.left = meta.margins[2],
			  meta.buffers.left.in = meta.buffers[2],
			  xylab.margins.left = xylab.margins[2],

			  panel.xtab.left = panel.xtab.size[2],
			  panel.xtab.margin.left = panel.xtab.margin[2],
			  grid.buffers.left = grid.buffers[2],
			  grid.margins.left = grid.margins[2],

			  {if (o$ncols > 1) rep(c(panel.wrap.size[2], panel.wrap.margin[2], 0, panel.wrap.margin[4], panel.wrap.size[4], between_marginW), o$ncols -1) else NULL},
			                          panel.wrap.size[2], panel.wrap.margin[2], 0, panel.wrap.margin[4], panel.wrap.size[4],

			  grid.margins.left = grid.margins[4],
			  grid.buffers.left = grid.buffers[4],
			  panel.xtab.margin.right = panel.xtab.margin[4],
			  panel.xtab.right = panel.xtab.size[4],

			  xylab.margins.right = xylab.margins[4],
			  meta.buffers.right.in = meta.buffers[4],
			  meta.margins.right = meta.margins[4],
			  meta.buffers.right.out = meta.buffers[4],
			  outer.margins.right = outer.margins[4])

		u = grid::unit(x, "npc")
		names(u) = names(x)
		u
	})

	nr = length(rows)
	nc = length(cols)

	cols_facet_ids = 1:o$ncols * 6 + 6
	rows_facet_ids = 1:o$nrows * 6 + 6

	#if (o$panel.type == "xtab") {
	cols_panel_col_ids = cols_facet_ids
	cols_panel_row_id = ifelse(o$panel.xtab.pos[2] == "top", 6, nc - 5)

	rows_panel_row_ids = rows_facet_ids
	rows_panel_col_id = ifelse(o$panel.xtab.pos[1] == "left", 6, nc - 5)
	#} else if (o$panel.type == "wrap") {
	cols_panel_ids = cols_facet_ids + ifelse(o$panel.wrap.pos  == "left", -2, ifelse(o$panel.wrap.pos  == "right", 2, 0))
	rows_panel_ids = rows_facet_ids + ifelse(o$panel.wrap.pos  == "top", -2, ifelse(o$panel.wrap.pos  == "bottom", 2, 0))

	panel_col_rot = ifelse(o$panel.xtab.pos[2] == "top", o$panel.label.rot[2], o$panel.label.rot[4])
	panel_row_rot = ifelse(o$panel.xtab.pos[1] == "left", o$panel.label.rot[1], o$panel.label.rot[3])
	panel_rot = ifelse(o$panel.wrap.pos  == "left", o$panel.label.rot[1], ifelse(o$panel.wrap.pos  == "right", o$panel.label.rot[3], ifelse(o$panel.wrap.pos  == "top", o$panel.label.rot[2], o$panel.label.rot[4])))

	#}
	#####
	### x and y lab
	xlab_row_id = if (o$xlab.side == "top") 5 else nr - 4
	xlab_col_ids = 6:(nc-5)

	ylab_col_id = if (o$ylab.side == "left") 5 else nc - 4
	ylab_row_ids = 6:(nr-5)


	prows = as.numeric(rows)
	pcols = as.numeric(cols)

	if (sum(prows) >= 1 || sum(pcols) >= 1) stop("Margins are too large, or too many facets.", call. = FALSE)

	fasp = ((1-sum(pcols)) / (1-sum(prows))) * o$dasp / o$ncols * o$nrows # asp per facet (with original outer margins)
	gasp = ((1-sum(pcols)) / (1-sum(prows))) * o$dasp # asp total facets (with original outer margins)

	if (!o$legend.only) {
		if (!is.na(o$asp) && o$asp != 0) {
			# follow device
			fasp = o$asp
		} else if (is.na(o$asp) && !is.na(o$sasp)) {
			fasp = o$sasp
		}
	}

	gasp2 = fasp * o$ncols / o$nrows # target gasp

	# needed for tmap save and arrange
	if (return.asp) return(gasp2)


	if (gasp2 > gasp) {
		extra.height =   (1 - ((1 - sum(pcols))/(gasp2/o$dasp))) - sum(prows)
		rows[c(1, length(rows))] = rows[c(1, length(rows))] + grid::unit(extra.height / 2, "npc")
	} else if (gasp2 < gasp) {
		extra.width =   (1 - ((1 - sum(prows)) * (gasp2/o$dasp))) - sum(pcols)
		cols[c(1, length(cols))] = cols[c(1, length(cols))] + grid::unit(extra.width / 2, "npc")
	}

	cols[cols_facet_ids] = (grid::unit(1, "npc") - sum(cols)) / o$ncols
	rows[rows_facet_ids] = (grid::unit(1, "npc") - sum(rows)) / o$nrows

	colsIn = as.numeric(cols) * o$devsize[1]
	rowsIn = as.numeric(rows) * o$devsize[2]

	#sum(grid::convertWidth(cols, "inches", valueOnly = TRUE))


	vp_tree = grid::vpStack(grid::viewport(width = grid::unit(o$cw, "snpc"), height = grid::unit(o$ch, "snpc"), name = "vp_asp"),
							grid::viewport(layout = grid::grid.layout(nrow = length(rows), ncol = length(cols), widths = cols, heights = rows), name = "vp_main")
	)

	bgcol = o$outer.bg.color

	outerRect = if (o$outer.bg) rndrectGrob(gp=grid::gpar(lwd = NA, fill = bgcol), name = "outer_rect") else NULL

	gts = lapply(1L:o$npages, function(ip) {
		grid::grobTree(
			outerRect,
			grid::grobTree(name = "gt_main"),
			vp = vp_tree, name = "tmap_grob_tree")
	})

	g = list(
		rows_facet_ids = rows_facet_ids,
		cols_facet_ids = cols_facet_ids,

		rows_panel_ids = rows_panel_ids,
		cols_panel_ids = cols_panel_ids,

		rows_panel_row_ids = rows_panel_row_ids,
		rows_panel_col_id = rows_panel_col_id,
		cols_panel_row_id = cols_panel_row_id,
		cols_panel_col_ids = cols_panel_col_ids,

		panel_col_rot = panel_col_rot,
		panel_row_rot = panel_row_rot,
		panel_rot = panel_rot,

		xlab_row_id = xlab_row_id,
		xlab_col_ids = xlab_col_ids,

		ylab_col_id = ylab_col_id,
		ylab_row_ids = ylab_row_ids,

		meta_rows = c(3, (nr-2)),
		meta_cols = c(3, (nc-2)),

		fasp = fasp,

		colsIn = colsIn,
		rowsIn = rowsIn,

		nc = nc,
		nr = nr
	)


	# design mode: widths and heights (and asp) for subset of cols/rows
	res = do.call(rbind, lapply(1:7, function(i) {
		tot_col = sum(g$colsIn[i:(nc-(i-1))])
		tot_row = sum(g$rowsIn[i:(nr-(i-1))])
		c(tot_col, tot_row, tot_col/ tot_row)
	}))


	# take make rows and columns
	g$mapColsIn = g$colsIn[g$cols_facet_ids]
	g$mapRowsIn = g$rowsIn[g$rows_facet_ids]

	# add w/h/asp for first map
	res = rbind(res, c(g$mapColsIn[1], g$mapRowsIn[1], g$mapColsIn[1]/ g$mapRowsIn[1]))

	## show aspect ratios in design mode
	if (getOption("tmap.design.mode")) {
		posttext <- apply(format(res[c(1,2,4, 8),], digits = 3), FUN = paste, MARGIN = 1, collapse = " ")
		pretext <- c("device", "plot area", "facets area", "map area")

		lns <- nchar(pretext) + nchar(posttext)
		l <- max(max(nchar(pretext)) + max(nchar(posttext)) + 1, 25)
		medtext <- vapply(l-lns, function(i)paste(rep(" ", i), collapse=""), character(1))

		texts <- c("---------------W (in)-H (in)-asp---",
				   paste("|", pretext, medtext, posttext, "|"),
				   paste(rep("-", l+6), collapse=""))

		texts_colors = c("Color codings:",
						 "- light blue  outer margins",
						 "- dark blue   buffers around outside cells",
						 "- light green outside cells",
						 "- dark green  x and ylab cells",
						 "- pink        panels",
						 "- red         margins for outside grid labels",
						 "- orange      margins around maps for grid labels",
						 "- yellow      map area",
						 "- lavender    component areas",
						 "Guide lines:",
						 "- thick       component position (legend, scalebar, etc.)",
						 "- thin        component-element position (e.g. legend items)"
						 )

		for (tx in c(texts, texts_colors)) message(tx)

		gts = lapply(gts, function(gt) {

			p = rep(cols4all::c4a("brewer.paired"), 3)

			gt = gt %>%
				add_to_gt(grid::rectGrob(gp=grid::gpar(fill = p[1])), row = 1:(nr), col = 1:(nc)) %>%  # outer
				add_to_gt(grid::rectGrob(gp=grid::gpar(fill = p[2])), row = 2:(nr-1), col = 2:(nc-1)) %>%   # meta buffer out
				add_to_gt(grid::rectGrob(gp=grid::gpar(fill = p[3])), row = 3:(nr-2), col = 3:(nc-2)) %>%   # meta margins
				add_to_gt(grid::rectGrob(gp=grid::gpar(fill = p[2])), row = 4:(nr-3), col = 4:(nc-3)) %>%   # meta buffer in
				add_to_gt(grid::rectGrob(gp=grid::gpar(fill = p[4])), row = 5:(nr-4), col = 5:(nc-4))  # xylab
			if (o$panel.type == "xtab") {
				#add_to_gt(grid::rectGrob(gp=grid::gpar(fill = p[5])), row = 6:(nr-5), col = 6:(nc-5)) # panel buffer
				gt = add_to_gt(gt, grid::rectGrob(gp=grid::gpar(fill = p[5])), row = 6:(nr-5), col = 6:(nc-5)) # panel
			}

			gt = gt %>%
				add_to_gt(grid::rectGrob(gp=grid::gpar(fill = p[6])), row = 7:(nr-6), col = 7:(nc-6)) %>%  # grid buffer
				add_to_gt(grid::rectGrob(gp=grid::gpar(fill = p[7])), row = 8:(nr-7), col = 8:(nc-7))  # grid


			for (i in 1:o$nrows) {
				for (j in 1:o$ncols) {
					gt = add_to_gt(gt, grid::rectGrob(gp=grid::gpar(fill = p[11])), row = g$rows_facet_ids[i], col = g$cols_facet_ids[j])
					if (o$panel.type == "wrap") {
						gt = add_to_gt(gt, grid::rectGrob(gp=grid::gpar(fill = p[5])), row = g$rows_panel_ids[i], col = g$cols_panel_ids[j])
					}

				}
			}

			gt
		})
	}

	# margins around first map (needed for georeferencing)
	rid = g$rows_facet_ids[1]
	cid = g$cols_facet_ids[1]
	map_all_margins = c(
		sum(g$rowsIn[(rid+1L):nr]),
		sum(g$colsIn[1:(cid-1L)]),
		sum(g$rowsIn[1:(rid-1L)]),
		sum(g$colsIn[(cid+1L):nc]))




	if (is.null(vp) && show && newpage) {
		grid.newpage()
	}# else {
	#	if (is.character(vp)) seekViewport(vp) else pushViewport(vp)
	#}
	assign("gts", gts, envir = .TMAP_GRID)
	assign("g", g, envir = .TMAP_GRID)
	assign("gasp2", gasp2, envir = .TMAP_GRID)
	.TMAP$start_pane_id = 401
	list(dev = res[1,],map = res[8,], margins = map_all_margins)
}

tmapGridAux = function(o, q) {
	NULL
}

#' Legend
#'
#' Legend specification
#'
#' @param title Legend title
#' @param show Show legend?
#' @param orientation Orientation of the legend: `"portrait"` or `"landscape"`
#' @param reverse Should the legend items be reversed?
#' @param na.show Show `NA` values in legend?
#' @param width Width of the legend. Units are 'text line heights'. In case a negative number is specified, the units are (approximate) pixels. The relation between these two is configured via the option `absolute_fontsize`.
#' @param height Height of the legend. Units are 'text line heights'. In case a negative number is specified, the units are (approximate) pixels. The relation between these two is configured via the option `absolute_fontsize`.
#' @inheritParams tm_title
#' @param title.color `r .doc_opt("legend.title.color")`
#' @param title.size `r .doc_opt("legend.title.size")`
#' @param title.fontface `r .doc_opt("legend.title.fontface")`
#' @param title.fontfamily `r .doc_opt("legend.title.fontfamily")`
#' @param title.alpha `r .doc_opt("legend.title.alpha")`
#' @param title.padding `r .doc_opt("legend.title.padding")`
#' @param title.align `r .doc_opt("legend.title.align")`
#' @param text.color `r .doc_opt("legend.text.color")`
#' @param text.size `r .doc_opt("legend.text.size")`
#' @param text.fontface `r .doc_opt("legend.text.fontface")`
#' @param text.fontfamily `r .doc_opt("legend.text.fontfamily")`
#' @param text.alpha `r .doc_opt("legend.text.alpha")`
#' @param format Not used anymore: use the format argument of the `tm_scale_*()` functions instead.
#' @param bg.color `r .doc_opt("legend.bg.color")`
#' @param bg.alpha `r .doc_opt("legend.bg.alpha")`
#' @param absolute_fontsize `r .doc_opt("legend.absolute_fontsize")`
#' @param item.height `r .doc_opt("legend.item.height")`
#' @param item.width `r .doc_opt("legend.item.width")`
#' @param item.space `r .doc_opt("legend.item.space")`
#' @param item.na.height `r .doc_opt("legend.na item.height")`
#' @param item.na.width `r .doc_opt("legend.na item.width")`
#' @param item.na.space `r .doc_opt("legend.na item.space")`
#' @param item.shape `r .doc_opt("legend.item.shape")`
#' @param ticks List of vectors of size 2 that determines a tick mark line (for portrait legends). The values are the y-values of begin and endpoint of a tick mark. For a solid line, only one vector is required, for dashed lines one for each dash. See \href{https://r-tmap.github.io/tmap/articles/versus_ggplot2#mimicking-ggplot2-layout}{ggplot2 style example}.
#' @param ticks.disable.na Remove ticks for `NA` values
#' @param ticks.col `r .doc_opt("legend.ticks.col")`
#' @param ticks.lwd `r .doc_opt("legend.ticks.lwd")`
#' @param margins `r .doc_opt("legend.margins")`
#' @param item_text.margin `r .doc_opt("legend.item_text.margin")`
#' @param ... visual values, e.g. `col`, `fill`, `lwd`, can be specified. If so, they overrule the default visual values, which are determined by the drawn map objects (e.g. polygons)
#' @param variable visual (or transformation) variable to combine the legend with: e.g. `"fill"` or `"size"`
#' @seealso \href{https://r-tmap.github.io/tmap/articles/basics_legends}{Vignette about legends}
#' @return A tm_legend component
#' @export
#' @example ./examples/tm_legend.R
tm_legend = function(title,
					 show,
					 orientation,
					 reverse,
					 na.show,
					 position,
					 group_id,
					 width,
					 height,
					 z,
					 title.color,
					 title.size,
					 title.fontface,
					 title.fontfamily,
					 title.alpha,
					 title.padding,
					 title.align,
					 text.color,
					 text.size,
					 text.fontface,
					 text.fontfamily,
					 text.alpha,
					 format,
					 frame,
					 frame.lwd,
					 frame.r,
					 bg,
					 bg.color,
					 bg.alpha,
					 absolute_fontsize,
					 item.height,
					 item.width,
					 item.space,
					 item.na.height,
					 item.na.width,
					 item.na.space,
					 item.shape,
					 ticks,
					 ticks.disable.na,
					 ticks.col,
					 ticks.lwd,
					 margins,
					 item_text.margin,
					 ...) {
	args = lapply(as.list(rlang::call_match()[-1]), eval, envir = parent.frame())
	args$called = names(args)
	args = warning_group_args(args)

	args$title = args$title %||% NA
	args$xlab = args$xlab %||% NA
	args$ylab = args$ylab %||% NA
	args$group_id = args$group_id %||% NA_character_
	args$group_type = "tm_legend"
	args$z = args$z %||% NA_integer_
	structure(args, class = c("tm_legend", "tm_component", "list"))
}

#' @rdname tm_legend
#' @export
tm_legend_hide = function() {
	tm_legend(show = FALSE)
}

#' @rdname tm_legend
#' @export
tm_legend_combine = function(variable) {
	structure(list(FUN = "tmapLegend", title = NA, reverse = FALSE, show = FALSE, aes = variable, group_id = NA_character_), class = c("tm_legend", "tm_component", "list"))
}

#' @rdname tm_legend
#' @param xlab label for the x dimension (rows)
#' @param ylab label for the y dimension (columns)
#' @param xlab.color `r .doc_opt("legend.xlab.color")`
#' @param xlab.size `r .doc_opt("legend.xlab.size")`
#' @param xlab.rot `r .doc_opt("legend.xlab.rot")`
#' @param xlab.fontface `r .doc_opt("legend.xlab.fontface")`
#' @param xlab.fontfamily `r .doc_opt("legend.xlab.fontfamily")`
#' @param xlab.alpha `r .doc_opt("legend.xlab.alpha")`
#' @param xlab.padding `r .doc_opt("legend.xlab.padding")`
#' @param xlab.align `r .doc_opt("legend.xlab.align")`
#' @param ylab.color `r .doc_opt("legend.ylab.color")`
#' @param ylab.size `r .doc_opt("legend.ylab.size")`
#' @param ylab.rot `r .doc_opt("legend.ylab.rot")`
#' @param ylab.fontface `r .doc_opt("legend.ylab.fontface")`
#' @param ylab.fontfamily `r .doc_opt("legend.ylab.fontfamily")`
#' @param ylab.alpha `r .doc_opt("legend.ylab.alpha")`
#' @param ylab.padding `r .doc_opt("legend.ylab.padding")`
#' @param ylab.align `r .doc_opt("legend.ylab.align")`
#' @export
tm_legend_bivariate = function(xlab,
							   ylab,
							   xlab.color,
							   xlab.size,
							   xlab.rot,
							   xlab.fontface,
							   xlab.fontfamily,
							   xlab.alpha,
							   xlab.padding,
							   xlab.align,
							   ylab.color,
							   ylab.size,
							   ylab.rot,
							   ylab.fontface,
							   ylab.fontfamily,
							   ylab.alpha,
							   ylab.padding,
							   ylab.align,
							   ...) {
	args = lapply(as.list(rlang::call_match()[-1]), eval, envir = parent.frame())

	args$title = args$title %||% NA
	args$xlab = args$xlab %||% NA
	args$ylab = args$ylab %||% NA
	args$group_id = args$group_id %||% NA_character_
	args$group_type = "tm_legend_bivariate"
	args$z = args$z %||% NA_integer_
	args$orientation = "portrait"
	structure(args, class = c("tm_legend", "tm_component", "list"))
}

#' Map component: manual legend
#'
#' Map component that adds a manual legend.
#'
#' @param ... visual variables and arguments passed on to `tm_legend()`.
#'   By default, the argument `type` is set to `"symbols"`, which means that the
#'   supported visual variables are: `"fill"`, `"col"`, `"shape"`, `"size"`,
#'   `"fill_alpha"`, `"col_alpha"`, `"lty"`, `"lwd"`, `"linejoin"`, and `"lineend"`.
#'   The number of legend items will be equal to the maximum number of specific values (and specified labels.)
#' @param labels labels by default `""` (so omitted)
#' @param type the layer type from which the visual variables (see `...`) are taken.
#'   Options: `"symbols"` (default), `"lines"`, `"polygons"`, and `"text"`.
#' @param title `r .doc_opt("legend.title")`
#' @param orientation `r .doc_opt("legend.orientation")`
#' @param position `r .doc_opt("legend.position")`
#' @inheritParams tm_title
#' @param group Name of the group to which this layer belongs. This is only
#'   relevant in view mode, where layer groups can be switched (see `group.control`)
#' @param group.control In view mode, the group control determines how
#'   layer groups can be switched on and off. Options: `"radio"` for radio
#'   buttons (meaning only one group can be shown), `"check"` for check boxes
#'   (so multiple groups can be shown), and `"none"` for no control
#'   (the group cannot be (de)selected).
#' @example ./examples/tm_add_legend.R
#' @export
tm_add_legend = function(...,
						 labels = "",
						 type = "symbols",
						 title = "",
						 orientation = NULL,
						 position = NULL,
						 group_id = NA_character_,
						 group = NA,
						 group.control = "check",
						 z = NA_integer_) {

	args = lapply(as.list(rlang::call_match(defaults = TRUE)[-1]), eval, envir = parent.frame())
	args$called = names(rlang::call_match(dots_expand = TRUE)[-1])

	args = warning_group_args(args)
	args$group_id = args$group_id %||% NA_character_
	args$group_type = "tm_legend"
	args$z = args$z %||% NA_integer_

	if (type %in% c("fill", "symbol", "line")) {
		args$type = v3_add_legend(type, names(args))
		if ("col" %in% names(args) && !c("fill" %in% names(args)) && type != "line") {
			args$fill = args$col
			args$col = NULL
		}
		if ("border.col" %in% names(args)) {
			args$col = args$border.col
			args$border.col = NULL
		}
	}
	tm_element_list(do.call(tm_element, c(args, list(subclass = c("tm_add_legend", "tm_component")))))
}

#' @export
#' @rdname tmap_internal
toTitleCase = function(x) {
	paste0(toupper(substr(x,1,1)), tolower(substr(x,2, nchar(x))))
}


tmapAddedLegend = function(comp, o) {
	#message("tm_mouse_coordinates ignored for 'plot' mode")
	l = update_l(o = o, l = comp, v = "", mfun = paste0("tm_data_", comp$type), unm = "", active = FALSE)

	fun = paste0("tm_", comp$type)
	if (!exists(fun)) {
		stop(paste0("type \"", comp$type, "\" not supported because tm_", comp$type,  " not found"), call. = FALSE)
	}
	res = do.call(fun, args = list())
	gp = res[[1]]$gpar

	for (gpi in names(gp)) {
		if (gpi %in% names(l)) {
			gp[[gpi]] = l[[gpi]]
		} else if (!is.na(gp[[gpi]])) {
			gp[[gpi]] = getAesOption("value.const", o, aes = gpi, layer = comp$type)
		}
	}

	if (comp$type == "text") {
		gp$cex = gp$size
		gp$size = NA
	}

	l$gp = gp

	l2 = within(l, {
		nitems = max(length(labels), vapply(gp, length, FUN.VALUE = integer(1), USE.NAMES = FALSE))
		labels = rep(labels, length.out = nitems)
		dvalues = 1:nitems
		vvalues = 1:nitems
		vneutral = NA
		na.show = FALSE

	})
	l2
}

#' Map component: title
#'
#' Map component that adds a title
#'
#' @param text text
#' @param size font size
#' @param color font color
#' @param padding padding
#' @param fontface font face, bold, italic
#' @param fontfamily font family
#' @param alpha alpha transparency of the text
#' @param stack stack with other map components, either `"vertical"` or `"horizontal"`.
#' @param just just
#' @param frame frame should a frame be drawn?
#' @param frame.color frame color
#' @param frame.alpha frame alpha transparancy
#' @param frame.lwd frame line width
#' @param frame.r Radius of the rounded frame corners. 0 means no rounding.
#' @param bg Show background?
#' @param bg.color Background color
#' @param bg.alpha Background transparency
#' @param position The position specification of the component: an object created with `tm_pos_in()` or `tm_pos_out()`. Or, as a shortcut, a vector of two values, specifying the x and y coordinates. The first is `"left"`, `"center"` or `"right"` (or upper case, meaning tighter to the map frame), the second `"top"`, `"center"` or `"bottom"`. Numeric values are also supported, where 0, 0 means left bottom and 1, 1 right top. See also `r .doc_see_vig("adv_positions")`. In case multiple components should be combined (stacked), use `group_id` and specify `component` in [tm_components()].
#' @param group_id Component group id name. All components (e.g. legends, titles, etc) with the same `group_id` will be grouped. The specifications of how they are placed (e.g. stacking, margins etc.) are determined in [tm_components()] where its argument `id` should correspond to `group_id`.
#' @param width,height width and height of the component.
#' @param z z index, e.g. the place of the component relative to the other componets
#' @seealso `r .doc_see_also_comp()`
#' @export
tm_title = function(text, size, color, padding, fontface, fontfamily, alpha, stack, just, frame, frame.color, frame.alpha, frame.lwd, frame.r, bg, bg.color, bg.alpha, position, group_id, width, height, z) {
	args = lapply(as.list(rlang::call_match()[-1]), eval, envir = parent.frame())
	args$called = names(args)
	args$group_id = args$group_id %||% NA_character_
	args$group_type = c("tm_<other>", "tm_title")
	args$z = args$z %||% NA_integer_
	tm_element_list(do.call(tm_element, c(args, list(subclass = c("tm_title", "tm_component")))))
}

#' @export
#' @param ... passed on to `tm_title()`
#' @rdname tm_title
tm_title_in = function(text, ..., position = tm_pos_in("left", "top")) {
	args = lapply(as.list(rlang::call_match()[-1]), eval, envir = parent.frame())
	args$called = names(args)
	args$group_id = args$group_id %||% NA_character_
	args$group_type = c("tm_<other>", "tm_title_in")
	args$z = args$z %||% NA_integer_
	args$position = position
	tm_element_list(do.call(tm_element, c(args, list(subclass = c("tm_title", "tm_component"
	)))))
}

#' @export
#' @rdname tm_title
tm_title_out = function(text, ..., position = tm_pos_out("center", "top")) {
	args = lapply(as.list(rlang::call_match()[-1]), eval, envir = parent.frame())
	args$called = names(args)
	args$group_id = args$group_id %||% NA_character_
	args$group_type = c("tm_<other>", "tm_title_out")
	args$z = args$z %||% NA_integer_
	args$position = position
	tm_element_list(do.call(tm_element, c(args, list(subclass = c("tm_title", "tm_component"
	)))))
}

#' Map component: (credits) text
#'
#' Map component that adds a text, typically used as credits. This function is the same as [tm_title()] but with different default values.
#'
#' @inheritParams tm_title
#' @param ... to catch deprecated arguments
#' @seealso `r .doc_see_also_comp()`
#' @export
tm_credits = function(text, size, color, padding, fontface, fontfamily, alpha, stack, just, frame, frame.lwd, frame.r, bg, bg.color, bg.alpha, position, group_id, width, height, z, ...) {
	args = lapply(as.list(rlang::call_match(dots_expand = TRUE)[-1]), eval, envir = parent.frame())
	args$called = names(args)
	args$group_id = args$group_id %||% NA_character_
	args$group_type = c("tm_<other>", "tm_credits")
	args$z = args$z %||% NA_integer_
	if ("align" %in% names(args)) {
		args$position = tm_pos_in(pos.h = "right", pos.v = "bottom", align.h = args$align, align.v = "top", just.h = "left", just.v = "bottom")
		args$align = NULL
		v3_instead_message("align", "position = tm_pos_in(align.h)", "tm_credits")
	}
	tm_element_list(do.call(tm_element, c(args, list(subclass = c("tm_credits", "tm_component")))))
}


#' Map component: compass
#'
#' Map component that adds a compass
#'
#' @param north north
#' @param type compass type, one of: \code{"arrow"}, \code{"4star"}, \code{"8star"}, \code{"radar"}, \code{"rose"}. The default is controlled by \code{\link{tm_layout}} (which uses \code{"arrow"} for the default style)
#' @param text.size text.size
#' @param size size
#' @param show.labels show.labels
#' @param cardinal.directions cardinal.directions
#' @param text.color text.color
#' @param color.dark color.dark
#' @param color.light color.light
#' @param lwd lwd
#' @inheritParams tm_title
#' @param stack stack with other map components, either `"vertical"` or `"horizontal"`.
#' @param just just
#' @param margins margins
#' @param ... to catch deprecated arguments (alpha)
#' @seealso `r .doc_see_also_comp()`
#' @export
tm_compass <- function(north,
					   type,
					   text.size,
					   size,
					   show.labels,
					   cardinal.directions,
					   text.color,
					   color.dark,
					   color.light,
					   lwd,
					   position,
					   group_id,
					   bg,
					   bg.color,
					   bg.alpha,
					   stack,
					   just,
					   frame,
					   frame.color,
					   frame.alpha,
					   frame.lwd,
					   frame.r,
					   margins,
					   z,
					   ...) {
	args = lapply(as.list(rlang::call_match()[-1]), eval, envir = parent.frame())
	args$called = names(args)
	args = warning_group_args(args)

	args$group_id = args$group_id %||% NA_character_
	args$group_type = c("tm_<other>", "tm_compass")
	args$z = args$z %||% NA_integer_
	tm_element_list(do.call(tm_element, c(args, list(subclass = c("tm_compass", "tm_component")))))
}

#' Map component: scale bar
#'
#' Map component that adds a scale bar.
#'
#' @param breaks Scale bar break positions. E.g. `c(0, 10, 50)` places breaks
#'   at 0, 10, and 50 units. The unit is controlled by the `unit` argument from [tm_shape()].
#'   When `NULL` (default), break positions are chosen automatically.
#' @param width Width of the scale bar, in number of text line heights (roughly
#'   equivalent to character widths). When `breaks` are specified, `width` is
#'   only useful for fine-tuning, e.g. to prevent label clipping or reduce
#'   excess whitespace.
#' @param allow_clipping Should clipping of the last label be allowed? If
#'   `TRUE` (default), the last break label including its unit suffix is printed
#'   even when it extends beyond the frame. If `FALSE`, that label is suppressed
#'   and the unit suffix is appended to the second-to-last label instead.
#' @param text.size text size
#' @param text.color text.color
#' @param color.dark color.dark
#' @param color.light color.light
#' @param lwd linewidth
#' @inheritParams tm_title
#' @param size Deprecated (use `text.size` instead)
#' @param stack stack with other map components, either `"vertical"` or `"horizontal"`.
#' @param margins margins
#' @seealso `r .doc_see_also_comp()`
#' @export
tm_scalebar = function(breaks,
						width,
						allow_clipping,
						text.size,
						text.color,
						color.dark,
						color.light,
						lwd,
						position,
						group_id,
						bg,
						bg.color,
						bg.alpha,
						size = "deprecated",
						stack,
						frame,
						frame.color,
						frame.alpha,
						frame.lwd,
						frame.r,
						margins,
						z) {

	if (!identical(size, "deprecated")) {
		cli::cli_inform(c(
			"!" = "The 'size' argument of {.fn tm_scalebar} is deprecated as of tmap 4.0.",
			"i" = "Please use 'text.size' instead."
		))
	}

	args = lapply(as.list(rlang::call_match()[-1]), eval, envir = parent.frame())
	args$called = names(args)
	args$group_id = args$group_id %||% NA_character_
	args$group_type = c("tm_<other>", "tm_scalebar")
	args$z = args$z %||% NA_integer_
	tm_element_list(do.call(tm_element, c(args, list(subclass = c("tm_scalebar", "tm_component")))))
}


#' Map component: mouse coordinates
#'
#' Map component that adds mouse coordinates
#'
#' @seealso `r .doc_see_also_comp()`
#' @inheritParams tm_title
#' @export
tm_mouse_coordinates <- function(stack,
								 position,
								 group_id,
								 z) {
	args = lapply(as.list(rlang::call_match()[-1]), eval, envir = parent.frame())
	args$called = names(args)
	args$group_id = args$group_id %||% NA_character_
	args$group_type = c("tm_<other>", "tm_mouse_coordinates")
	args$z = args$z %||% NA_integer_
	tm_element_list(do.call(tm_element, c(args, list(subclass = c("tm_mouse_coordinates", "tm_component")))))
}

#' Map component: minimap
#'
#' Map component that adds a [minimap][leaflet::addMiniMap()] in view mode.
#'
#' @param server name of the provider or an URL (see \code{\link{tm_tiles}}).
#'   By default, it shows the same map as the basemap, and moreover, it will automatically change when the user switches basemaps.
#'   Note the latter does not happen when `server` is specified.
#' @param toggle should the minimap have a button to minimise it? By default \code{TRUE}.
#' @inheritParams tm_title
#' @inheritParams tm_inset
#' @inheritDotParams leaflet::addMiniMap
#' @seealso `r .doc_see_also_comp()` `.doc_see_also_insets()`
#' @export
tm_minimap <- function(server,
					   toggle,
					   height,
					   width,
					   margins,
					   between_margin,
					   position,
					   group_id,
					   frame, frame.color, frame.alpha, frame.lwd, frame.r, bg, bg.color, bg.alpha,
					   z,
					   ...) {
	args = lapply(as.list(rlang::call_match()[-1]), eval, envir = parent.frame())
	args$called = names(args)
	args$group_id = args$group_id %||% NA_character_
	args$group_type = c("tm_<other>", "tm_minimap")
	args$z = args$z %||% NA_integer_
	tm_element_list(do.call(tm_element, c(args, list(subclass = c("tm_minimap", "tm_component")))))
}


#' Map component: logo
#'
#' Map component that adds a logo.
#'
#' @param file either a filename or url of a png image. If multiple files/urls are provided with a character vector, the logos are placed near each other.
#'   To specify logos for small multiples use a list of character values/vectors. In order to stack logos vertically, multiple \code{tm_logo} elements can be stacked.
#' @param height height of the logo in number of text line heights. The width is scaled based the height and the aspect ratio of the logo.
#'   If multiple logos are specified by a vector or list, the heights can be specified accordingly.
#' @param margins margins
#' @param between_margin Margin between
#' @inheritParams tm_title
#' @example ./examples/tm_logo.R
#' @seealso `r .doc_see_also_comp()`
#' @export
tm_logo = function(file,
				   height,
				   margins,
				   between_margin,
				   stack,
				   position,
				   group_id,
				   frame,
				   frame.color,
				   frame.alpha,
				   frame.lwd,
				   frame.r,
				   z) {
	args = lapply(as.list(rlang::call_match()[-1]), eval, envir = parent.frame())
	args$called = names(args)
	args$group_id = args$group_id %||% NA_character_
	args$group_type = c("tm_<other>", "tm_logo")
	args$z = args$z %||% NA_integer_
	tm_element_list(do.call(tm_element, c(args, list(subclass = c("tm_logo", "tm_component")))))
}

check_fix = function(sfc, shp_name, reproj, messages) {
	s2 = sf::sf_use_s2()

	isLarge = (inherits(sfc, c("sfc_MULTIPOLYGON", "sfc_MULTILINESTRING", "sfc_GEOMETRYCOLLECTION")) && length(sfc) >= 5000)
	if (isLarge && messages) message("Checking shape object \"", shp_name, " \". It is large, so it may take a while")

	if (!all(sf::st_is_valid(sfc))) {
		checkAgain = FALSE
		tryCatch({
			if (messages) message("The shape object \"", shp_name, "\" is invalid", ifelse(reproj, " (after reprojection). ", ". "), "Trying to fix it...")
			sfc = sf::st_make_valid(sfc)
		}, error = function(e) {
			if (messages) message("Unsuccesful attempt with sf::st_make_valid")
			checkAgain = TRUE
		})
		if (checkAgain || !all(sf::st_is_valid(sfc))) {
			suppressMessages(sf::sf_use_s2(!s2))
			.TMAP$set_s2 = s2
			sfc = tryCatch({
				sf::st_make_valid(sfc)
			}, error = function(e) {
				suppressMessages(sf::sf_use_s2(s2))
				cli::cli_warn(c(
					"Unable to make {.code {shp_name}} valid.",
					i = "Invalid geometries are left out"
				))
				isv = which(sf::st_is_valid(sfc))
				sfc = sfc[isv]
				attr(sfc, "is_valid") = isv
				sfc
			})
			if (messages) {
				cli::cli_inform(c(
					"Shape {shp_name} has been fixed with {.code s2 = {!s2}.",
					"If the map doesn't look correct, please run {.code sf::sf_use_s2({!s2})} before running the tmap code again."
				))
			}

		}
	}
	sfc
}

add_to_gt = function(gt, grb, row, col) {
	vp = grid::viewport(layout.pos.col = col, layout.pos.row = row)
	gtr = grid::grobTree(grb, vp = vp)

	grid::addGrob(gt, gtr, gPath = grid::gPath("gt_main"))
}





get_legend_option = function(x, type) {
	if (length(x) == 1 || (!type %in% names(x))) x else if (is.list(x)) x[[type]] else x[type]
}


gp_to_gpar = function(gp, id = NULL, sel = "all", split_to_n = NULL, pick_middle = TRUE, o, type) {
	type = tolower(type)
	if (sel == "all") {
		if (is.na(gp$fill_alpha[1]) && !is.na(gp$col_alpha[1])) sel = "col"
		if (!is.na(gp$fill_alpha[1]) && is.na(gp$col_alpha[1])) sel = "fill"
	}


	# get alpha value (sel: "all" means fill and col, "fill" and "col" mean fill and col only respectively)
	alpha = if (sel == "fill") {
		if (!is.na(gp$fill_alpha[1])) gp$fill_alpha else 1
	} else {
		if (!is.na(gp$col_alpha[1])) gp$col_alpha else 1
	}

	# create a list of gp elements
	lst = c(list(fill = {if (sel == "col") NA else gp$fill},
			   col = {if (sel == "fill") NA else gp$col},
			   alpha = alpha,
			   lty = if (sel == "fill") getAesOption("value.blank", o, aes = "lty", layer = type) else if (!is.na(gp$lty[1])) gp$lty else getAesOption("value.const", o, aes = "lty", layer = type) ,
			   lwd = {if (!all(is.na(gp$lwd))) gp$lwd else getAesOption("value.blank", o, aes = "lwd", layer = type)},
			   lineend = {if (!all(is.na(gp$lineend))) gp$lineend else "round"},
			   linejoin = {if (!all(is.na(gp$linejoin))) gp$linejoin else "round"},
			   size = {if (!all(is.na(gp$size))) gp$size else 1},
			   cex = {if (!all(is.na(gp$cex))) gp$cex else 1},
			   fontface = {if (!all(is.na(gp$cex))) gp$fontface else getAesOption("value.blank", o, aes = "fontface", layer = type)},
			   fontfamily = {if (!all(is.na(gp$fontfamily))) gp$fontfamily else o$text.fontfamily},
			   shape = {if (!all(is.na(gp$shape))) gp$shape else getAesOption("value.const", o, aes = "shape", layer = type)}))

	#
	if (!is.null(id)) {
		lst = lapply(lst, "[", id)
	}

	lst_isnum = c(fill = FALSE,
				  col = FALSE,
				  alpha = TRUE,
				  lty = FALSE,
				  lwd = TRUE,
				  lineend = FALSE,
				  linejoin = FALSE,
				  size = TRUE,
				  cex = TRUE,
				  fontface = FALSE,
				  fontfamily = FALSE,
				  shape = TRUE)

	lst = mapply(function(lsti, isnum) {
		if (!is.character(lsti)) return(lsti)

		if (nchar(lsti[1]) > 50) {
			x = cont_split(lsti)
			x = lapply(x, function(i) {
				i[i == "NA"] <- NA
				i
			})
			if (isnum) x = lapply(x, as.numeric)
			if (pick_middle) {
				x = sapply(x, function(i) {
					if (all(is.na(i))) NA else {
						sq = (o$continuous.nclass_per_legend_break/2) + (rep(0:5,each=2) * c(1,-1))[-1] # priority for middle values
						i[sq[which(!is.na(i)[sq])[1]]]
					}
				})
			}
			return(x)

		} else {
			return(lsti)
		}
	}, lst, lst_isnum[names(lst)], SIMPLIFY = FALSE)

	if (!is.null(split_to_n)) {
		split_gp(lst, split_to_n)
	} else {
		do.call(grid::gpar, lst)
	}
}

split_gp = function(gp, n) {
	lst = lapply(1L:n, function(i) {
		lapply(gp, function(gpi) {
			if (length(gpi) == n) gpi[i] else gpi[1]
		})
	})
	lapply(lst, function(lsti) {
		do.call(grid::gpar, lsti)
	})
}


determine_scale = function(label, rot, row, col, g, scale = 1) {

	w = sum(g$colsIn[col])
	h = sum(g$rowsIn[row])

	labwidth = graphics::strwidth(label, units = "inches")
	labheight = graphics::strheight(label, units = "inches")


	scale = min(scale, {if (rot %in% c(0, 180)) {
		min(w / labwidth, h / labheight)
	} else {
		min(h / labwidth, w / labheight)
	}})
}

frc = function(row, col) paste0(sprintf("%02d", row), "_", sprintf("%02d", col))

impute_gp = function(gp, dt) {
	dtn = setdiff(names(dt), c("tmapID__", paste0("by", 1L:3L, "__")))

	cols = paste0("__", dtn)
	gp1 = sapply(gp, "[[", 1)
	gpids = which(gp1 %in% cols)
	#gp[gpids] = as.list(dt[, dtn, with = FALSE])

	for (i in gpids) gp[i] = as.list(dt[, dtn[match(gp1[i], cols)], with = FALSE])
	gp
}

rndrectGrob = function(...) {
	args = list(...)
	if ("r" %in% names(args) && args$r > 0) {
		args$r = grid::unit(args$r, "pt")
		do.call(grid::roundrectGrob, args)
	} else {
		args$r = NULL
		do.call(grid::rectGrob, args)
	}
}

# numeric vector (interpreted as inches) will be cast to a unit vector of inches with nulls on the head and tail. The values of the nulls depend on the justification
set_unit_with_stretch = function(x, ids = NULL, sides = c("both", "first", "second")) {
	u = rep("inch", length(x))
	if (!is.null(ids)) {
		x[ids] = 1/length(ids)
		u[ids] = "null"
		x = c(0, x, 0)
	} else {
		sides = match.arg(sides)
		x = if (sides == "both") c(0.5, x, 0.5) else if (sides == "first") c(1, x, 0) else c(0, x, 1)
	}
	u = c("null", u, "null")
	grid::unit(x, units = u)
}

# numeric vector for which spacers will be added between
unit_add_between = function(x, y) {
	n = length(x)
	if (n == 1L) return(x)
	x = rep(x, each = 2)[-(2L*n)]
	x[seq(2, (2*n)-1, by = 2)] = y
	x
}

# numeric vector for which head and tail will be added
unit_add_sides = function(x, y) {
	if (length(y) == 1L) {
		c(y, x, y)
	} else {
		c(y[1], x, y[2])
	}
}



# u is unit vector, tot = is the total size (width or height). The null units are distributed such that the total equals tot
distr_space_over_nulls = function(u, tot, stretchID = NA) {
	u0 = grid::unitType(u) == "null"

	un = grid::convertUnit(u, unitTo = "inch", valueOnly = TRUE)
	tn = as.numeric(tot)

	un_not0 = sum(un[!u0])
	un_dist_over0 = tn - un_not0

	normalize = function(x, s = 1) (x / sum(x)) * s

	if (un_dist_over0 < 0) {
		un[u0] = 0
		if (!is.na(stretchID)) un[stretchID] = max(0, un[stretchID] + un_dist_over0)
		un = normalize(un, s = tn)
	} else {
		un[u0] = normalize(as.numeric(u[u0]), un_dist_over0)
	}
	#print(un)
	grid::unit(un, units = "inch")
}

rescale_gp = function(gp, scale, skip = character()) {
	if ("lwd" %in% names(gp) && (!"lwd" %in% skip)) gp$lwd = gp$lwd * scale
	if ("size" %in% names(gp) && (!"size" %in% skip)) gp$size = gp$size * sqrt(scale)
	if ("cex" %in% names(gp) && (!"cex" %in% skip)) gp$cex = gp$cex * sqrt(scale)
	gp
}




appendGlist = function(glist, x) {
	glist = grid::gList(glist, x)
	names(glist)[length(glist)] = x$name
	glist
}

swap_pch_15_20 = function(gp) {
	# swap fill and col for pch 15-20
	if (any(gp$shape %in% 15:20)) {
		gp = make_equal_list(gp)
		pch15_20 = which(gp$shape %in% 15:20)
		fill = gp$col[pch15_20]
		fill_alpha = gp$col_alpha[pch15_20]
		gp$col[pch15_20] = gp$fill[pch15_20]
		gp$col_alpha[pch15_20] = gp$fill_alpha[pch15_20]
		gp$fill[pch15_20] = fill
		gp$fill_alpha[pch15_20] = fill_alpha
	}
	gp
}


# zero_one_to_hex = function(x) {
# 	# using indexing
# 	u = unique(x)
#
# 	x255 = round(u * 255)
#
# 	nc = c(0:9, LETTERS[1:6])
#
# 	y1 = (x255 %/% 16) + 1
# 	y2 = (x255 - (y1 - 1) * 16) + 1
#
# 	r = paste0(nc[y1], nc[y2])
# 	r[match(x, u)]
# }

# 255 to 2digit hex number
num_to_hex = function(x) {

	nc = c(0:9, LETTERS[1:6])

	y1 = (x %/% 16) + 1
	y2 = (x - (y1 - 1) * 16) + 1

	paste0(nc[y1], nc[y2])
}

hex_to_num = function(h) {
	nc = c(0:9, LETTERS[1:6])
	y1 = match(substr(h, 1, 1), nc)
	y2 = match(substr(h, 2, 2), nc)
	(y1 - 1) * 16 + (y2 - 1)
}

merge_alpha = function(dt, name) {
	name_a = paste0(name, "_alpha")
	f = function(d) {
		d1 = d[1,]
		col = d1[[name]]
		alpha = d1[[name_a]]

		if (nchar(col) == 9) {
			a = hex_to_num(substr(col, 8, 9)) * alpha
			cl = substr(col, 1, 7)
		} else {
			a = alpha * 255
			cl = col
		}
		ac = paste0(cl, num_to_hex(round(a)))
		rep(ac, nrow(d))
	}

	dt[, ca:=f(.SD), by = c(name, name_a), .SDcols = c(name, name_a)]
}





npc_to_native = function(x, scale) {
	x * (scale[2] - scale[1])# + scale[1]
}

v3_only = function(fun) {
	setdiff(funs_v3[[fun]], funs_v4[[fun]])
}

v3_start_message = function() {
	if (!.TMAP$v3) {
		cli::cli_h1("tmap v3 code detected")
		.TMAP$v3 = TRUE
	}
	invisible(NULL)
}

v3_reset_flag = function() {
	.TMAP$v3 = FALSE
	invisible(NULL)
}


v3_use_component = function(arg = "title", comp = "tm_title", container = "tm_format") {
	v3_start_message()
	id = paste0(arg, comp, container)
	cli::cli_inform("{.field [v3->v4]} {.fn {container}}: use {.fn {comp}} instead of the {.arg {arg}} argument.", .frequency_id = id, .frequency = "always")
}

v3_title = function(fun) {
	id = paste0(fun, "title")
	cli::cli_inform("{.field [v3->v4]} {.fn {fun}}: use {.fn tm_title} instead of {.code {fun}(title = )}", .frequency_id = id, .frequency = "always")
}

v3_main_title = function(fun) {
	id = paste0(fun, "main.title")

	cli::cli_inform("{.field [v3->v4]} {.fn {fun}}: use {.fn tm_title} instead of {.code {fun}(main.title = )}", .frequency_id = id, .frequency = "always")
}

v3_convert2density = function(layer_fun) {
	id = paste0(layer_fun, "convert2density")
	cli::cli_inform(c(
		"{.field [v3->v4]} {.fn {layer_fun}} {.arg convert2density} is deprecated.",
		"i" = "Divide the variable values by the polygon areas manually (obtain the areas with 'sf::st_area()')."
	),
	.id = id,
	.frequency_id = "always"
	)
}

v3_tm_scale_instead_of_style = function(style, scale_fun, vv, layer_fun, arg_list) {
	m = paste0("scale_", scale_fun, "_vv_", vv, "style_", style)
	if (!message_thrown(m)) {
		if (scale_fun == "intervals") {
			arg_list$old = c("style", arg_list$old)
			arg_list$new = c("style", arg_list$new)
		}
		if (length(arg_list$old)) {
			x = if (arg_list$mult) {
				paste0("For small multiples, specify a 'tm_scale_' for each multiple, and put them in a list: '{vv}'.scale = list(<scale1>, <scale2>, ...)'")
			} else {
				NULL
			}
			al = do.call(paste, c(mapply(function(x,y) {
				if (x == y) paste0("'", x, "'") else paste0("'", x, "' (rename to '", y, "')")
			}, arg_list$old, arg_list$new, SIMPLIFY = FALSE), sep = ", "))
			xtra = paste0("Migrate the argument(s) ", al, " to 'tm_scale_", scale_fun, "(<HERE>)'")
		} else {
			xtra = NULL
			x = NULL
		}

		cli::cli_inform(c(
			"{.field [v3->v4]} {.fn {layer_fun}}: instead of {.code style = {.str {style}}}, use {vv}.scale = {.fn tm_scale_{scale_fun}}.",
			i = xtra,
			x
		))
		message_reg(m)
	}
	NULL
}

v3_tm_scale = function(scale_fun, vv, layer_fun, arg_list) {
	m = paste0("scale_", scale_fun, "_vv_", vv)

	if (!message_thrown(m)) {
		scale_fun = if (scale_fun == "") {
			"tm_scale"
		} else {
			paste0("tm_scale_", scale_fun)
		}
		if (length(arg_list$old)) {
			x = if (arg_list$mult) {
				paste0("For small multiples, specify a 'tm_scale_' for each multiple, and put them in a list: '", vv, ".scale = list(<scale1>, <scale2>, ...)'")
			} else {
				NULL
			}

			al = v3_list_text(olds = arg_list$old, news = arg_list$new)
			cli::cli_inform(c(
				"{.field [v3->v4]} {.fn tm_{layer_fun}}: migrate the argument(s) related to the scale of the visual variable {.var {vv}} namely {al} to {vv}.scale = {scale_fun}(<HERE>).",
				"i" = x
			))
		}
		message_reg(m)
	}
	NULL
}

v3_instead_message = function(arg_old, arg_new, fun) {
	v3_start_message()
	id <- paste0(fun, arg_old, arg_new)
	cli::cli_inform(c(
		"{.field [v3->v4]} {.fn {fun}}: use {.arg {arg_new}} instead of {.arg {arg_old}}."
	),
	.frequency_id = id,
	.frequency = "always"
	)
}

v3_instead = function(args_called, old, new, fun, extra_called = character()) {
	args = args_called$args
	called = args_called$called

	if (old %in% called) {
		args[[new]] = args[[old]]
		args[[old]] = NULL

		# may be needed to trigger something (e.g. "shape" to use symbols instead of dots)
		if (length(extra_called)) {
			called = unique(c(called, extra_called))
		}
		v3_instead_message(arg_old = old, arg_new = new, fun = fun)
	}
	list(args = args, called = called)
}

v3_instead_value = function(args_called, old, new, fun, value_old, value_new) {
	args = args_called$args
	called = args_called$called

	if (old %in% called) {
		if (identical(args[[old]], value_old)) {
			args[[old]] = NULL
			args[[new]] = value_new
			value_old = value_old %||% NULL
			cli::cli_inform(c(
				"{.field [v3->v4]} {.fn {fun}}: use {.code {new} = {value_new}} instead of {.code {old} = {value_old}}."
			))
			list(args = args, called = called)
		} else {
			v3_instead(args_called, old, new, fun)
		}
	} else {
		list(args = args, called = called)
	}
}

v3_list_init = function() {
	.TMAP$v3_list = list(old = character(), new = character(), mult = FALSE)
	invisible(NULL)
}
v3_list_impute_item = function(name, new_name, mult = FALSE) {
	.TMAP$v3_list$old = c(.TMAP$v3_list$old, name)
	.TMAP$v3_list$new = c(.TMAP$v3_list$new, new_name)
	.TMAP$v3_list$mult = .TMAP$v3_list$mult || mult
}
v3_list_get = function() {
	.TMAP$v3_list
}

v3_impute = function(args, name, value, new_name = name) {
	if (name %in% names(args)) {
		res = args[[name]]
		if (is.list(res) && !any(grepl("format", name, fixed = TRUE)) && !is.list(res[[1]])) {
			res = res[[1]]
			mult = TRUE
		} else if (new_name %in% c("title", "na.show", "orientation", "reverse", "interval.closure", "drop.levels", "midpoint", "as.count", "values.repeat", "values.scale", "value.na", "value.null", "value.neutral", "label.na", "label.null")) {
			res = res[1]
			mult = TRUE
		} else {
			mult = FALSE
		}
		v3_list_impute_item(name, new_name, mult)
		res
	} else value
}

v3_list_text = function(olds, news) {
	do.call(paste, c(mapply(function(x,y) {
		if (x == y) {
			paste0("'", x, "'")
		} else if (nchar(y) < 20) {
			paste0("'", x, "' (rename to '", y, "')")
		} else {
			paste0("'", x, "' (use '", y, "')")
		}
	}, olds , news, SIMPLIFY = FALSE), sep = ", "))
}

v3_tm_legend = function(fun, vv, arg_list) {
	if (length(arg_list$old)) {
		al = v3_list_text(olds = arg_list$old, news = arg_list$new)
		cli::cli_inform("{.field [v3->v4]} {.fn {fun}}: migrate the argument(s) related to the legend of the visual variable {.var {vv}} namely {al} to '{vv}.legend = tm_legend(<HERE>)'")
	}
	NULL
}

v3_tm_facets = function(arg_list) {
	if (length(arg_list$old)) {
		al = v3_list_text(olds = arg_list$old, news = arg_list$new)
		cli::cli_inform(c(
			"{.field [v3->v4]} {.fn tm_facets}: rename the following argument(s): {al}."
		),
		.frequency_id = "facets2",
		.frequency = "always"
		)
	}
	NULL
}

v3_tm_facets_free_scales = function() {
	cli::cli_inform(c(
		"{.field [v3->v4]} {.fn tm_facets}: migrate each  'free.scales.<X>' argument to the argument '<X>.free' of the corresponding layer function."
		),
		.frequency_id = "facets",
		.frequency = "always"
	)
}



v3_tm_legend_hide = function(fun, arg, vv) {
	cli::cli_inform(c(
		"{.field [v3->v4]} {.fn {fun}}: use {.code {vv}.legend = tm_legend_hide()} instead of {.code {arg} = FALSE}."
	),
	.frequency_id = "hide-legend",
	.frequency = "always"
	)
}



v3_tm_legend_general = function(fun) {
	v3_start_message()
	cli::cli_inform(c(
		"{.field [v3->v4]} {.fn {fun}}: use 'tm_legend()' inside a layer function, e.g. 'tm_polygons(..., fill.legend = tm_legend())'"
	),
	.frequency_id = "legend-general",
	.frequency = "regularly"
	)
}




v3_tm_chart_hist = function(layer_fun, vv, arg) {
	cli::cli_inform(c(
		"{.field [v3->v4]} {.fn {layer_fun}}: use {.code {vv}.chart = tm_chart_histogram()} instead of {.code {arg} = TRUE}."
	),
	.frequency_id = "chart-hist",
	.frequency = "regularly"
	)
}

v3_message_col_fill = function(layer_fun) {
	cli::cli_inform(c(
		"{.field [v3->v4]} {.fn {layer_fun}}: use 'fill' for the fill color of polygons/symbols (instead of 'col'), and 'col' for the outlines (instead of 'border.col')."
	),
	.frequency_id = "col-fill",
	.frequency = "regularly"
	)
}

v3_message_vv_null = function(layer_fun) {
	cli::cli_inform(c(
		"{.field [v3->v4]} {.fn {layer_fun}}: use {.val {NA}} instead of {.code NULL} to hide a visual variable."
	),
	.frequency_id = "hide-vv",
	.frequency = "regularly"
	)
}


v3_message_fill_alpha = function(layer_fun = layer_fun) {
	v3_instead_message("alpha", "fill_alpha", layer_fun)
}

v3_message_col_alpha = function(layer_fun = layer_fun, orig = "border.alpha") {
	v3_instead_message(arg_old = orig, arg_new = "col_alpha", layer_fun)
}

v3_add_legend = function(type, args) {
	newtype = c(fill = "polygons", symbol = "symbols", line = "lines")
	cli::cli_inform(
		"{.field [v3->v4]} {.fn tm_add_legend}:  use {.code type = {.val {newtype[type]}}} instead of {.code type = {.val {type}}}.",
		.frequency_id = "legend-type"
	)
	if ("col" %in% args && !c("fill" %in% args)) {
		cli::cli_inform(
			"{.field [v3->v4]} {.fn tm_add_legend}:  use {.arg fill} instead of {.arg col} for the fill color of {newtype[type]}.",
			.frequency_id = "legend-fill"
		)
	}
	if ("border.col" %in% args) {
		cli::cli_inform(
			"{.field [v3->v4]} {.fn tm_add_legend}:  use {.arg col} instead of {.arg border.col} for the outline color of {newtype[type]}.",
			.frequency_id = "legend-border-col"
		)
	}
	newtype[type]
}

v3_layer_opt = function(olds, news, layer_fun) {
	x = v3_list_text(olds, news)
	cli::cli_inform(
		"{.field [v3->v4]} {.fn {layer_fun}}: migrate the layer options {x} to 'options = opt_{layer_fun}(<HERE>)'",
		.frequency_id = "opt-tm"
	)
}


v3_tm_rgb = function(r, g, b, dim = NA) {
	if (is.na(dim)) {
		cli::cli_inform(
			"{.field [v3->v4]} {.fn tm_rgb}: instead of using r = {r}, g = {g} and b = {b} , please use col = tm_vars(c({r}, {g}, {b}), multivariate = TRUE)",
			.frequency_id = "tm-rgb"
		)
	} else {
		cli::cli_inform(
			"{.field [v3->v4]} {.fn tm_rgb}: instead of using r = {r}, g = {g} and b = {b} , please use col = tm_vars(dimvalues = c({r}, {g}, {b}), multivariate = TRUE)",
			.frequency_id = "tm-rgb"
		)
	}
}


v3_opt = function(fun, old, new) {
	cli::cli_inform(
		"{.field [v3->v4]} {.fn {fun}}: use {new} instead of {old}",
		.frequency_id = "tm-opt"
	)
}

v3_message_rgb_alpha = function(layer_fun) {
	cli::cli_inform(
		"{.field [v3->v4]} {.fn {layer_fun}}: use {.arg col_alpha} instead of {.arg alpha}",
		.frequency_id = "tm-rgb"
	)
}


v3_message_rgb_opt = function(layer_fun, opt, value) {
	cli::cli_inform(
		"{.field [v3->v4]} {.fn {layer_fun}}: use {.code options = opt_tm_rgb({opt} = {value})} instead of {.arg {opt} = {value}}",
		.frequency_id = "tm-rgb"
	)
}

v3_message_rgb_maxV = function(layer_fun, opt, value) {
	cli::cli_inform(
		"{.field [v3->v4]} {.fn {layer_fun}}: use {.code col.scale = tm_scale_rgb(max_color_value = {value})} instead of {.arg max.value = {value}}",
		.frequency_id = "tm-rgb"
	)
}

v3_rivers = function() {
	cli::cli_inform(
		"{.field [v3->v4]} tmap data: use {.code World_rivers} instead of {.code rivers}",
		.frequency_id = "rivers"
	)
}

v3_simplify = function(name, args) {
	names(args)[names(args) == "simplify"] = "fact"

	txt = paste(mapply(paste, names(args), unname(args), MoreArgs = list(sep = " = "), USE.NAMES = FALSE), collapse = ", ")

	cli::cli_inform(
		"{.field [v3->v4]} {.fn tm_shape}: {.arg simplify} cannot be used anymore. Please use {.code tmaptools::simplify_shape({name}, {txt})} instead",
		.frequency_id = "simplify"
	)
}

#' tmap function to define a constant visual value
#'
#' tmap function to define a constant visual value
#'
#' @export
tm_const = function() {
	tmapOption("value.const")
}

#' Scales: automatic scale
#'
#' Scales in tmap are configured by the family of functions with prefix `tm_scale`.
#' Such function should be used for the input of the `.scale` arguments in the
#' layer functions (e.g. `fill.scale` in [tm_polygons()]). The function `tm_scale()`
#' is a scale that is set automatically given by the data type (factor, numeric, and integer)
#' and the visual variable. The tmap option `scales.var` contains information
#' which scale is applied when.
#'
#' @param ... arguments passed on to the applied scale function `tm_scale_*()`
#' @seealso `r .doc_see_also_scales()` [tm_scale_asis()], [tm_scale_ordinal()], [tm_scale_categorical()],
#' [tm_scale_intervals()], [tm_scale_discrete()], [tm_scale_continuous()],
#' [tm_scale_rank()], [tm_scale_continuous_log()], [tm_scale_continuous_log2()],
#' [tm_scale_continuous_log10()], [tm_scale_continuous_log1p()], [tm_scale_continuous_sqrt()],
#' [tm_scale_continuous_pseudo_log()], [tm_scale_rgb()], [tm_scale_bivariate()]
#' @export
tm_scale = function(...) {
	# maybe add the generic scales parameters after ... here?
	structure(c(list(FUN = "tmapScaleAuto"), list(...)), class = c("tm_scale_auto", "tm_scale", "list"))
}

#' Scales: as is
#'
#' Scales in tmap are configured by the family of functions with prefix `tm_scale`.
#' Such function should be used for the input of the `.scale` arguments in the
#' layer functions (e.g. `fill.scale` in [tm_polygons()]).
#' The function `tm_scale_asis()` is used to take data values as they are and use them as such for the visual variable.
#'
#' @inheritParams tm_scale_categorical
#' @param ... Arguments caught (and not used) from the automatic function [tm_scale()]
#' @seealso `r .doc_see_also_scales()` [tm_scale()]
#' @export
tm_scale_asis = function(values.scale = NA, value.neutral = NA, value.na = NA, ...) {
	structure(c(list(FUN = "tmapScaleAsIs"), c(list(values.scale = values.scale, value.neutral = value.neutral, value.na = value.na), list(...))), class = c("tm_scale_asis", "tm_scale", "list"))
}

#' @export
#' @rdname tm_scale_categorical
tm_scale_ordinal = function(n.max = 30,
							values = NA, # e.g. palette, shape numbers, set of sizes (if two values are specified they are interpret as range)
							values.repeat = FALSE,
							values.range  = 1,
							values.scale = NA,
							value.na = NA, # value for NA
							value.null = NA, # value for NULL (needed?)
							value.neutral = NA, # value for neutral (used in other legends)
							levels = NULL, # levels to show (other values are treated as missing)
							levels.drop = FALSE,
							labels = NULL, # labels for the categories (by default the factor levels)
							label.na = NA, # label for missing values
							label.null = NA,
							label.format = list()) {
	structure(c(list(FUN = "tmapScaleCategorical"), as.list(environment())), class = c("tm_scale_ordinal", "tm_scale", "list"))
}

#' Scales: categorical and ordinal scale
#'
#' Scales in tmap are configured by the family of functions with prefix `tm_scale`.
#' Such function should be used for the input of the `.scale` arguments in the
#' layer functions (e.g. `fill.scale` in [tm_polygons()]).
#' The functions `tm_scale_categorical()` and `tm_scale_ordinal()` are used
#' for categorical data. The only difference between these functions is that the
#' former assumes unordered categories whereas the latter assumes ordered categories.
#' For colors (the visual variable `fill` or `col`), different default color
#' palettes are used (see the tmap option `values.var`).
#'
#' @param n.max Maximum number of categories (factor levels). In case there are more, they are grouped into `n.max` groups.
#' @param values (generic scale argument) The visual values. For colors (e.g. `fill` or `col` for `tm_polygons()`) this is a palette name from the `cols4all` package (see [cols4all::c4a()]) or vector of colors, for size (e.g. `size` for `tm_symbols()`) these are a set of sizes (if two values are specified they are interpret as range), for symbol shapes (e.g. `shape` for [tm_symbols()]) these are a set of symbols, etc. The tmap option `values.var` contains the default values per visual variable and in some cases also per data type.
#' @param values.repeat (generic scale argument) Should the values be repeated in case there are more categories?
#' @param values.range (generic scale argument) Range of the values. Vector of two numbers (both between 0 and 1) where the first determines the minimum and the second the maximum. Full range, which means that all values are used, is encoded as `c(0, 1)`. For instance, when a gray scale is used for color (from black to white), `c(0,1)` means that all colors are used, `0.25, 0.75` means that only colors from dark gray to light gray are used (more precisely `"grey25"` to `"grey75"`), and `0, 0.5` means that only colors are used from black to middle gray (`"gray50"`). When only one number is specified, this is interpreted as the second number (where the first is set to 0). Default values can be set via the tmap option `values.range`.
#' @param values.scale (generic scale argument) Scaling of the values. Only useful for size-related visual variables, such as `size` of [tm_symbols()] and `lwd` of [tm_lines()].
#' @param value.na (generic scale argument) Value used for missing values. See tmap option `"value.na"` for defaults per visual variable.
#' @param value.null (generic scale argument) Value used for NULL values. See tmap option `"value.null"` for defaults per visual variable. Null data values occur when out-of-scope features are shown (e.g. for a map of Europe showing a data variable per country, the null values are applied to countries outside Europe).
#' @param value.neutral (generic scale argument) Value that can be considered neutral. This is used for legends of other visual variables of the same map layer. E.g. when both `fill` and `size` are used for [tm_symbols()] (using filled circles), the size legend items are filled with the `value.neutral` color from the `fill.scale` scale, and fill legend items are bubbles of size `value.neutral` from the `size.scale` scale.
#' @param levels Levels to show. Other values are treated as missing.
#' @param levels.drop Should unused levels be dropped (and therefore are not assigned to a visual value and shown in the legend)?
#' @param labels (generic scale argument) Labels
#' @param label.na (generic scale argument) Label for missing values
#' @param label.null (generic scale argument) Label for null (out-of-scope) values
#' @param label.format (generic scale argument) Label formatting. Output of [tm_label_format()]
#' @seealso `r .doc_see_also_scales()` [tm_scale()]
#' @export
tm_scale_categorical = function(n.max = 30,
								values = NA, # e.g. palette, shape numbers, set of sizes (if two values are specified they are interpret as range)
								values.repeat = TRUE,
								values.range  = NA,
								values.scale = NA,
								value.na = NA, # value for NA
								value.null = NA, # value for NULL (needed?)
								value.neutral = NA, # value for neutral (used in other legends)
								levels = NULL, # levels to show (other values are treated as missing)
								levels.drop = FALSE,
								labels = NULL, # labels for the categories (by default the factor levels)
								label.na = NA, # label for missing values
								label.null = NA,
								label.format = list()) { # label for null values (needed?)
	structure(c(list(FUN = "tmapScaleCategorical"), as.list(environment())), class = c("tm_scale_categorical", "tm_scale", "list"))
}

#' Scales: interval scale
#'
#' Scales in tmap are configured by the family of functions with prefix `tm_scale`.
#' Such function should be used for the input of the `.scale` arguments in the
#' layer functions (e.g. `fill.scale` in [tm_polygons()]).
#' The function `tm_scale_intervals()` is used for numerical data.
#'
#' @param n Number of intervals. For some styles (see argument `style` below) it is the preferred number rather than the exact number.
#' @param style Method to create intervals. Options are `"cat"`, `"fixed"`, `"sd"`, `"equal"`, `"pretty"`, `"quantile"`, `"kmeans"`, `"hclust"`, `"bclust"`, `"fisher"`, `"jenks"`, `"dpih"`, `"headtails"`, and `"log10_pretty"`. See the details in [classInt::classIntervals()] (extra arguments can be passed on via `style.args`).
#' @param style.args List of extra arguments passed on to [classInt::classIntervals()].
#' @param breaks Interval breaks (only used and required when `style = "fixed"`)
#' @param interval.closure value that determines whether where the intervals are closed: `"left"` or `"right"`. If `as.count = TRUE`, `inverval.closure` is always set to `"left"`.
#' @param label.style Either `"discrete"` or `"continuous"`. If `discrete` (default) intervals will be labeled. If `continuous`, the legend will be similar to a continuous scale legend (but with discrete colors), where the breaks are labeled.
#' @param label.select Which labels are shown? A logical vector is expected, which is repeated along the number of labels. By default `TRUE`, which means all.
#' @param midpoint The data value that is interpreted as the midpoint. By default it is set to 0 if negative and positive values are present. Useful when values are diverging colors. In that case, the two sides of the color palette are assigned to negative respectively positive values. If all values are positive or all values are negative, then the midpoint is set to `NA`, which means that the value that corresponds to the middle color class (see `style`) is mapped to the middle color. If it is specified for sequential color palettes (e.g. `"Blues"`), then this color palette will be treated as a diverging color palette.
#' @param as.count Should the data variable be processed as a count variable? For instance, if `style = "pretty"`, `n = 2`, and the value range of the variable is 0 to 10, then the column classes for `as.count = TRUE` are 0; 1 to 5; 6 to 10 (note that 0 is regarded as an own category) whereas for `as.count = FALSE` they are 0 to 5; 5 to 10. Only applicable if `style` is `"pretty"`, `"fixed"`, or `"log10_pretty"`. By default `FALSE`.
#' @inheritParams tm_scale_categorical
#' @param values (generic scale argument) The visual values. For colors (e.g. `fill` or `col` for [tm_polygons()]) this is a palette name from the `cols4all` package (see [cols4all::c4a()]) or vector of colors, for size (e.g. `size` for `tm_symbols`) these are a set of sizes (if two values are specified they are interpret as range), for symbol shapes (e.g. `shape` for `tm_symbols`) these are a set of symbols, etc. The tmap option `values.var` contains the default values per visual variable and in some cases also per data type.
#' @param values.range (generic scale argument) Range of the values. Vector of two numbers (both between 0 and 1) where the first determines the minimum and the second the maximum. Full range, which means that all values are used, is encoded as `c(0, 1)`. For instance, when a gray scale is used for color (from black to white), `c(0,1)` means that all colors are used, `0.25, 0.75` means that only colors from dark gray to light gray are used (more precisely `"gray25"` to `"gray75"`), and `0, 0.5` means that only colors are used from black to middle grey (`"grey50"`). When only one number is specified, this is interpreted as the second number (where the first is set to 0). Default values can be set via the tmap option `values.range`.
#' @param labels (generic scale argument) Labels
#' @param label.na (generic scale argument) Label for missing values
#' @param label.null (generic scale argument) Label for null (out-of-scope) values
#' @param label.format (generic scale argument) Label formatting. Output of [tm_label_format()]
#' @seealso `r .doc_see_also_scales()` [tm_scale()]
#' @export
tm_scale_intervals = function(n = 5,
							  style = ifelse(is.null(breaks), "pretty", "fixed"),
							  style.args = list(),
							  breaks = NULL,
							  interval.closure = "left",
							  label.style = "discrete",
							  label.select = TRUE,
							  midpoint = NULL,
							  as.count = FALSE,
							  values = NA,
							  values.repeat = FALSE,
							  values.range  = NA,
							  values.scale = NA,
							  value.na = NA,
							  value.null = NA,
							  value.neutral = NA,
							  labels = NULL,
							  label.na = NA,
							  label.null = NA,
							  label.format = tm_label_format()) {
	structure(c(list(FUN = "tmapScaleIntervals"), as.list(environment())), class = c("tm_scale_intervals", "tm_scale", "list"))
}

#' Scales: discrete scale
#'
#' Scales in tmap are configured by the family of functions with prefix `tm_scale`.
#' Such function should be used for the input of the `.scale` arguments in the
#' layer functions (e.g. `fill.scale` in [tm_polygons()]).
#' The function [tm_scale_discrete()] is used for discrete numerical data, such as integers.
#'
#' @param ticks Discrete values. If not specified, it is determined automatically: unique values are put on a discrete scale.
#' @param midpoint The data value that is interpreted as the midpoint. By default it is set to 0 if negative and positive values are present. Useful when values are diverging colors. In that case, the two sides of the color palette are assigned to negative respectively positive values. If all values are positive or all values are negative, then the midpoint is set to `NA`, which means that the value that corresponds to the middle color class (see `style`) is mapped to the middle color. If it is specified for sequential color palettes (e.g. `"Blues"`), then this color palette will be treated as a diverging color palette.
#' @param values (generic scale argument) The visual values. For colors (e.g. `fill` or `col` for [tm_polygons()]) this is a palette name from the `cols4all` package (see [cols4all::c4a()]) or vector of colors, for size (e.g. `size` for `tm_symbols`) these are a set of sizes (if two values are specified they are interpret as range), for symbol shapes (e.g. `shape` for [tm_symbols()]) these are a set of symbols, etc. The tmap option `values.var` contains the default values per visual variable and in some cases also per data type.
#' @param values.repeat (generic scale argument) Should the values be repeated in case there are more categories?
#' @param values.range (generic scale argument) Range of the values. Vector of two numbers (both between 0 and 1) where the first determines the minimum and the second the maximum. Full range, which means that all values are used, is encoded as `c(0, 1)`. For instance, when a gray scale is used for color (from black to white), `c(0,1)` means that all colors are used, `0.25, 0.75` means that only colors from dark gray to light gray are used (more precisely `"grey25"` to `"grey75"`), and `0, 0.5` means that only colors are used from black to middle grey (`"grey50"`). When only one number is specified, this is interpreted as the second number (where the first is set to 0). Default values can be set via the tmap option `values.range`.
#' @param values.scale (generic scale argument) Scaling of the values. Only useful for size-related visual variables, such as `size` of [tm_symbols()] and `lwd` of [tm_lines()].
#' @param value.na (generic scale argument) Value used for missing values. See tmap option `"value.na"` for defaults per visual variable.
#' @param value.null (generic scale argument) Value used for NULL values. See tmap option `"value.null"` for defaults per visual variable. Null data values occur when out-of-scope features are shown (e.g. for a map of Europe showing a data variable per country, the null values are applied to countries outside Europe).
#' @param value.neutral (generic scale argument) Value that can be considered neutral. This is used for legends of other visual variables of the same map layer. E.g. when both `fill` and `size` are used for [tm_symbols()] (using filled circles), the size legend items are filled with the `value.neutral` color from the `fill.scale` scale, and fill legend items are bubbles of size `value.neutral` from the `size.scale` scale.
#' @param labels (generic scale argument) Labels
#' @param label.na (generic scale argument) Label for missing values
#' @param label.null (generic scale argument) Label for null (out-of-scope) values
#' @param label.format (generic scale argument) Label formatting. Output of [tm_label_format()]
#' @seealso `r .doc_see_also_scales()` [tm_scale()]
#' @export
tm_scale_discrete = function(ticks = NA,
							 #step = NA,
							 midpoint = NULL,
							 values = NA,
							 values.repeat = FALSE,
							 values.range  = NA,
							 values.scale = NA,
							 value.na = NA,
							 value.null = NA,
							 value.neutral = NA,
							 labels = NULL,
							 label.na = NA,
							 label.null = NA,
							 label.format = list()) {
	structure(c(list(FUN = "tmapScaleDiscrete"), as.list(environment())), class = c("tm_scale_discrete", "tm_scale", "list"))
}

#' Scales: continuous scale
#'
#' Scales in tmap are configured by the family of functions with prefix `tm_scale`.
#' Such function should be used for the input of the `.scale` arguments in the layer
#' functions (e.g. `fill.scale` in [tm_polygons()]).
#' The function `tm_scale_continuous()` is used for continuous data.
#' The functions `tm_scale_continuous_<x>()` use transformation functions x.
#'
#' @param n Preferred number of tick labels. Only used if `ticks` is not specified
#' @param limits Limits of the data values that are mapped to the continuous scale. When `NA`, the range of data values is taken. When only one value is provided, the range of data values with this provided value is taken. The default depends on the visual variable: it is 0 for all visual variables other than color when `tm_scale_continuous` is used. For the transformation scale functions, it is `NA`.
#' @param outliers.trunc Should outliers be truncated? An outlier is a data value that is below or above the respectively lower and upper limit. A logical vector of two values is expected. The first and second value determines whether values lower than the lower limit respectively higher than the upper limit are truncated to the lower respectively upper limit. If `FALSE` (default), they are considered as missing values.
#' @param ticks Tick values. If not specified, it is determined automatically with `n`
#' @param trans Transformation function. One of `"identity"` (default), `"log"`, and `"log1p"`. Note: the base of the log scale is irrelevant, since the log transformed values are normalized before mapping to visual values.
#' @param midpoint The data value that is interpreted as the midpoint. By default it is set to 0 if negative and positive values are present. Useful when values are diverging colors. In that case, the two sides of the color palette are assigned to negative respectively positive values. If all values are positive or all values are negative, then the midpoint is set to `NA`, which means that the value that corresponds to the middle color class (see `style`) is mapped to the middle color. If it is specified for sequential color palettes (e.g. `"Blues"`), then this color palette will be treated as a diverging color palette.
#' @param values (generic scale argument) The visual values. For colors (e.g. `fill` or `col` for [tm_polygons()]) this is a palette name from the `cols4all` package (see [cols4all::c4a()]) or vector of colors, for size (e.g. `size` for [tm_symbols()]) these are a set of sizes (if two values are specified they are interpret as range), for symbol shapes (e.g. `shape` for [tm_symbols()]) these are a set of symbols, etc. The tmap option `values.var` contains the default values per visual variable and in some cases also per data type.
#' @param values.repeat (generic scale argument) Should the values be repeated in case there are more categories?
#' @param values.range (generic scale argument) Range of the values, especially useful for color palettes. Vector of two numbers (both between 0 and 1) where the first determines the minimum and the second the maximum. Full range, which means that all values are used, is encoded as `c(0, 1)`. For instance, when a gray scale is used for color (from black to white), `c(0,1)` means that all colors are used, `0.25, 0.75` means that only colors from dark gray to light gray are used (more precisely `"grey25"` to `"grey75"`), and `0, 0.5` means that only colors are used from black to middle gray (`"grey50"`). When only one number is specified, this is interpreted as the second number (where the first is set to 0). Default values can be set via the tmap option `values.range`.
#' @param values.scale (generic scale argument) Scaling of the values. Only useful for size-related visual variables, such as `size` of [tm_symbols()] and `lwd` of [tm_lines()].
#' @param value.na (generic scale argument) Value used for missing values. See tmap option `"value.na"` for defaults per visual variable.
#' @param value.null (generic scale argument) Value used for NULL values. See tmap option `"value.null"` for defaults per visual variable. Null data values occur when out-of-scope features are shown (e.g. for a map of Europe showing a data variable per country, the null values are applied to countries outside Europe).
#' @param value.neutral (generic scale argument) Value that can be considered neutral. This is used for legends of other visual variables of the same map layer. E.g. when both `fill` and `size` are used for [tm_symbols()] (using filled circles), the size legend items are filled with the `value.neutral` color from the `fill.scale` scale, and fill legend items are bubbles of size `value.neutral` from the `size.scale` scale.
#' @param labels (generic scale argument) Labels
#' @param label.na (generic scale argument) Label for missing values
#' @param label.null (generic scale argument) Label for null (out-of-scope) values
#' @param label.format (generic scale argument) Label formatting. Output of [tm_label_format()]
#' @param trans.args list of additional argument for the transformation (generic transformation arguments)
#' @inheritParams scales::transform_pseudo_log
#' @example ./examples/tm_scale_continuous.R
#' @seealso `r .doc_see_also_scales()` [tm_scale()]
#' @export
#' @name tm_scale_continuous
tm_scale_continuous = function(n = NULL,
							   limits = NULL,
							   outliers.trunc = NULL,
							   ticks = NULL,
							   trans = NULL,
							   midpoint = NULL,
							   values = NA,
							   values.repeat = FALSE,
							   values.range  = NA,
							   values.scale = NA,
							   value.na = NA,
							   value.null = NA,
							   value.neutral = NA,
							   labels = NULL,
							   label.na = NA,
							   label.null = NA,
							   label.format = tm_label_format(),
							   trans.args = list()) {

	structure(c(list(FUN = "tmapScaleContinuous"), as.list(environment())), class = c("tm_scale_continuous", "tm_scale", "list"))
}


#' Scales: rank scale
#'
#' Scales in tmap are configured by the family of functions with prefix `tm_scale`.
#' Such function should be used for the input of the `.scale` arguments in the layer
#' functions (e.g. `fill.scale` in [tm_polygons()]).
#' The function [tm_scale_rank()] is used to rank numeric data.
#'
#' @param n Preferred number of tick labels. Only used if `ticks` is not specified
#' @param ticks Tick values. If not specified, it is determined automatically with `n`
#' @param values (generic scale argument) The visual values. For colors (e.g. `fill` or `col` for [tm_polygons()]) this is a palette name from the `cols4all` package (see [cols4all::c4a()]) or vector of colors, for size (e.g. `size` for [tm_symbols()]) these are a set of sizes (if two values are specified they are interpret as range), for symbol shapes (e.g. `shape` for [tm_symbols()]) these are a set of symbols, etc. The tmap option `values.var` contains the default values per visual variable and in some cases also per data type.
#' @param values.repeat (generic scale argument) Should the values be repeated in case there are more categories?
#' @param values.range (generic scale argument) Range of the values, especially useful for color palettes. Vector of two numbers (both between 0 and 1) where the first determines the minimum and the second the maximum. Full range, which means that all values are used, is encoded as `c(0, 1)`. For instance, when a gray scale is used for color (from black to white), `c(0,1)` means that all colors are used, `0.25, 0.75` means that only colors from dark gray to light gray are used (more precisely `"grey25"` to `"grey75"`), and `0, 0.5` means that only colors are used from black to middle gray (`"grey50"`). When only one number is specified, this is interpreted as the second number (where the first is set to 0). Default values can be set via the tmap option `values.range`.
#' @param values.scale (generic scale argument) Scaling of the values. Only useful for size-related visual variables, such as `size` of [tm_symbols()] and `lwd` of [tm_lines()].
#' @param value.na (generic scale argument) Value used for missing values. See tmap option `"value.na"` for defaults per visual variable.
#' @param value.null (generic scale argument) Value used for NULL values. See tmap option `"value.null"` for defaults per visual variable. Null data values occur when out-of-scope features are shown (e.g. for a map of Europe showing a data variable per country, the null values are applied to countries outside Europe).
#' @param value.neutral (generic scale argument) Value that can be considered neutral. This is used for legends of other visual variables of the same map layer. E.g. when both `fill` and `size` are used for [tm_symbols()] (using filled circles), the size legend items are filled with the `value.neutral` color from the `fill.scale` scale, and fill legend items are bubbles of size `value.neutral` from the `size.scale` scale.
#' @param labels (generic scale argument) Labels
#' @param label.na (generic scale argument) Label for missing values
#' @param label.null (generic scale argument) Label for null (out-of-scope) values
#' @param label.format (generic scale argument) Label formatting. Output of [tm_label_format()]
#' @param unit The unit name of the values. By default `"rank"`.
#' @seealso `r .doc_see_also_scales()` [tm_scale()]
#' @export
tm_scale_rank = function(n = NULL,
						 ticks = NULL,
						 values = NA,
						 values.repeat = FALSE,
						 values.range  = NA,
						 values.scale = NA,
						 value.na = NA,
						 value.null = NA,
						 value.neutral = NA,
						 labels = NULL,
						 label.na = NA,
						 label.null = NA,
						 label.format = tm_label_format(),
						 unit = "rank") {
	structure(c(list(FUN = "tmapScaleRank"), as.list(environment())), class = c("tm_scale_rank", "tm_scale", "list"))
}


#' #' @export
#' tm_scale_rank = function(...) {
#' 	tmc = tm_scale_continuous(...)
#' 	class(tmc) = c("tm_scale_rank", "tm_scale", "list")
#' 	tmc
#' }

#' @export
#' @param ... passed on to [tm_scale_continuous()]
#' @rdname tm_scale_continuous
tm_scale_continuous_log = function(..., base = exp(1)) {
	tm_scale_continuous(trans = "log", trans.args = list(base = base), ...)
}

#' @export
#' @rdname tm_scale_continuous
tm_scale_continuous_log2 = function(...) {
	tm_scale_continuous(trans = "log", trans.args = list(base = 2), ...)
}

#' @export
#' @rdname tm_scale_continuous
tm_scale_continuous_log10 = function(...) {
	x = tm_scale_continuous(trans = "log", trans.args = list(base = 10), ...)
	class(x) = c("tm_scale_continuous_log10", "tm_scale", "list")
	x
}


#' @export
#' @rdname tm_scale_continuous
tm_scale_continuous_log1p = function(...) {
	tm_scale_continuous(trans = "log1p", ...)
}

#' @export
#' @rdname tm_scale_continuous
tm_scale_continuous_sqrt = function(...) {
	tm_scale_continuous(trans = "sqrt", ...)
}

#' @export
#' @rdname tm_scale_continuous
tm_scale_continuous_pseudo_log = function(..., base = exp(1), sigma = 1) {
	tm_scale_continuous(trans = "pseudo_log", trans.args = list(base = base, sigma = sigma), ...)
}

#
# #' @export
# #' @rdname tm_scale_continuous
# tm_scale_continuous_logistic = function(...) {
# 	tm_scale_continuous(trans = "logistic", ...)
# }


#' Scales: RGB
#'
#' Scales in tmap are configured by the family of functions with prefix `tm_scale`.
#' Such function should be used for the input of the `.scale` arguments in the layer
#' functions (e.g. `fill.scale` in [tm_polygons()]).
#' The function [tm_scale_rgb()] is used to transform r, g, b band variables to colors. This function is adopted from (and works similar as) [stars::st_rgb()]
#'
#' @param value.na value for missing values
#' @param stretch should each (r, g, b) band be stretched? Possible values: `"percent"` (same as `TRUE`), `"histogram"`, `FALSE`.
#'   In the first case, the values are stretched to `probs[1]...probs[2]`. In the second case, a histogram equalization is performed
#' @param probs probability (quantile) values when `stretch = "percent"`
#' @param max_color_value maximum value
#' @seealso `r .doc_see_also_scales()` [tm_scale()] [stars::st_rgb()]
#' @example ./examples/tm_scale_rgb.R
#' @export
tm_scale_rgb = function(value.na = NA,
						stretch = FALSE,
						probs = c(0, 1),
						max_color_value = 255L) {
	structure(c(list(FUN = "tmapScaleRGB"), as.list(environment())), class = c("tm_scale_rgb", "tm_scale", "list"))
}

#' @rdname tm_scale_rgb
#' @export
tm_scale_rgba = function(value.na = NA,
						 stretch = FALSE,
						 probs = c(0, 1),
						 max_color_value = 255) {
	structure(c(list(FUN = "tmapScaleRGBA"), as.list(environment())), class = c("tm_scale_rgba", "tm_scale", "list"))
}

# tm_scale_na = function() {
# 	structure(c(list(FUN = "tmapScaleNA"), as.list(environment())), class = c("tm_scale_na", "tm_scale", "list"))
# }

# function to update the levels fl per facet dimension
update_grp_vars = function(lev = NULL, m = NULL) {
	vl = get("vl", envir = .TMAP)
	vn = get("vn", envir = .TMAP)

	if (!missing(lev)) {
		m = length(lev)
		if (m == 1L && vn == 1L) {
			if (is.null(vl)) {
				vl = lev
				vn = length(lev)
			}
		} else if (m > 1L && vn > 1L && m != vn) {
			stop("Inconsistent number of map variable values.", call. = FALSE)
		} else if (m > vn) {
			vl = lev
			vn = length(lev)
		}
	} else if (!missing(m)) {
		if (m > 1L && vn > 1L && m != vn) {
			stop("Inconsistent number of map variable values.", call. = FALSE)
		} else if (m > vn) {
			vn = m
			vl = NULL
		}
	}
	assign("vl", vl, envir = .TMAP)
	assign("vn", vn, envir = .TMAP)
}

add_used_vars = function(v, convert2density = FALSE, is_title = FALSE) {
	if (convert2density) {
		c2d_vars = get("c2d_vars", envir = .TMAP)
		c2d_vars = unique(c(c2d_vars, v))
		assign("c2d_vars", c2d_vars, envir = .TMAP)
	}

	used_vars = get("used_vars", envir = .TMAP)
	used_vars = unique(c(used_vars, v))

	if (is.null(attr(used_vars, "is_title"))) {
		attr(used_vars, "is_title") = is_title
	} else {
		attr(used_vars, "is_title") = attr(used_vars, "is_title") || is_title
	}

	assign("used_vars", used_vars, envir = .TMAP)
}

get_split_stars_dim = function(lst) {
	if (length(lst)) {
		y = vapply(lst, function(x) x$split_stars_dim, FUN.VALUE = character(1))
		if (any(y != "")) y[which(y != "")[1]] else ""
	} else {
		""
	}
}

is_rev = function(x) !is.null(x) && substr(x, 1, 1) == "-"
remove_min = function(x) substr(x, 2, nchar(x))

# ## estimate number of facets
step1_rearrange_facets = function(tmo, o) {
	nframes = NULL

	#o = tmap_options_mode()
	dev = getOption("tmap.devel.mode")


	# get the final tm_faets object (ignoring group specific args: is.wrap, by, rows, columns, pages)

	#fl = list(1L, 1L, 1L)
	#assign("fl", fl, envir = .TMAP)

	#assign("fl", list(NULL, NULL, NULL), envir = .TMAP)
	#assign("fn", c(1L, 1L, 1L), envir = .TMAP)

	tmo = lapply(tmo, function(tmg) {

		shp = tmg$tms$shp
		layer = tmg$tms$layer

		shp_is_pointer = inherits(shp, "character")

		smeta = tmapGetShapeMeta1(shp, layer, c(o, tmg$tmf))

		if (dev) timing_add(s3 = "get_shape_meta1")


		assign("vl", NULL, envir = .TMAP)
		assign("vn", 1L, envir = .TMAP)

		assign("used_vars", character(0), envir = .TMAP)
		assign("c2d_vars", character(0), envir = .TMAP)

		precheck_aes = function(a, layer, shpvars, args) {
			b = within(a, {
				if (inherits(value, "tmapAsIs")) {
					if (inherits(scale, "tm_scale_auto")) {
						class(scale) = c("tm_scale_asis", "tm_scale", "list")
						scale$FUN = tmapScaleAsIs
					}
				}

				# check if dimension is or should be used
				if (inherits(value, "tmapVars") && (length(smeta$dims) != 0) && (!is.null(value$dimvalues) || (!is.na(value$x[1]) && length(value$x) == 1L) || (value$multivariate && length(shpvars) == 1L))) {
					if (is.na(value$x)[1]) {
						value$x = smeta$dims[1]
					} else {
						if (length(value$x) > 1L) {
							if (identical(value$n, -1)) { # shortcut used for v3 compat: tm_rgb(r=3,g=2,=b1) when not known if shp had a band
								v3_tm_rgb(value$x[1], value$x[2], value$x[3], dim = smeta$dims[1])

								value$dimvalues = value$x
								value$x = smeta$dims[1]
								value$n = NA
							} else {
								warning("dimvalues specified while more than one dimension name is specified. Only the first will be used", call. = FALSE)
								value$x = value$x[1]
							}
						}
						if (!(value$x %in% smeta$dims)) {
							stop("dimvalues specified, but dimension \"" , x, "\" not found. Available dimension name(s): ", paste(smeta$dims,collapse = ", "), call. = FALSE)
						}
					}
					if (is.null(value$dimvalues)) {
						if (!is.na(value$n)) {
							value$dimvalues = smeta$dims_vals[[value$x]][1L:value$n]
						} else {
							if (!all(value$dimvalues) %in% smeta$dims_vals[[value$x]]) stop("Incorrect dimvalues", call. = FALSE)
						}
					}

					split_stars_dim = value$x
					add_used_vars(value$dimvalues)

					# redefine value for step 2
					if (value$multivariate) {
						update_grp_vars(lev = value$x)
						value = structure(list(as.character(value$dimvalues)), names = paste(as.character(value$dimvalues), collapse = "_"), class = "tmapVars")
					} else {
						update_grp_vars(lev = value$dimvalues)
						value = structure(as.list(as.character(value$dimvalues)), names = as.character(value$dimvalues), class = "tmapVars")
					}

					data_vars = TRUE
					bypass_vars = FALSE
					geo_vars = FALSE

				} else {
					split_stars_dim = ""

					#value_orig = value # just for the case of L156
					if (!inherits(value, c("tmapOption", "tmapVars", "tmapAsIs", "tmapSpecial")) && length(value) && is.na(value[[1]][1])) {
						# NA -> value.blank
						value = tmapVV(getAesOption("value.blank", o, aes = aes, layer = layer))
					}

					if (inherits(value, "tmapOption")) {
						#value_orig = tmapVV(getAesOption(value[[1]], o, aes = aes, layer = layer))
						value = tmapVV(getAesOption(value[[1]], o, aes = aes, layer = layer))
						data_vars = FALSE
						bypass_vars = FALSE
						geo_vars = FALSE
						#if (!is.list(value_orig)) value = list(value_orig)
						#value = value_orig
						#names(value) = sapply(value, "[", 1)
					} else if (inherits(value, "tmapVars")) {
						if (identical(value$n, -1)) { # shortcut used for v3 compat: tm_rgb(r=3,g=2,=b1) when not known if shp had a band
							v3_tm_rgb(value$x[1], value$x[2], value$x[3])
						}

						if (!is.null(value$dimvalues) && length(smeta$dims) == 0L) {
							error_dimvalues()
						}
						if (!is.na(value$x[1])) {
							if (is.character(value$x)) {
								if (!all(value$x %in% shpvars)) stop("not all variables specified in tm_vars are found", call. = FALSE)
								vars = value$x
							} else {
								if (!all(value$x %in% 1L:length(shpvars))) stop("tm_vars defined for x = ", paste(value$ids, collapse = ", "), " while there are only ", length(shpvars), " variables", call. = FALSE)
								vars = shpvars[value$x]
							}
						} else if (!is.na(value$n)) {
							if (length(shpvars) < value$n) {
								stop("tm_vars defined for n = ", value$n, " while there are only ", length(shpvars), " variables", call. = FALSE)
							}
							vars = shpvars[1L:value$n]
						} else {
							vars = shpvars
						}
						names(vars) = vars

						if (value$animate) {
							.TMAP$animate = c(.TMAP$animate, vars)
						}

						if (value$multivariate) {
							value = structure(list(unname(vars)), names = paste(vars, collapse = "_"), class = "tmapStandard")
						} else {
							value = structure(as.list(vars), class = "tmapStandard")
						}

						data_vars = !shp_is_pointer
						bypass_vars = shp_is_pointer
						geo_vars = FALSE
					} else {
						if (inherits(value, "tmapStandard")) {
							uvalue = unlist(value)
							data_vars = all(uvalue %in% shpvars) && !shp_is_pointer

							bypass_vars = (length(shpvars) == 0 || all(uvalue %in% shpvars)) && shp_is_pointer
							geo_vars = all(uvalue %in% c("AREA", "LENGTH", "MAP_COLORS")) && !data_vars
							if (geo_vars) {
								w = which(uvalue == "AREA")[1]
								if (!is.na(w)) {
									uvalue[w] = "AREA_"
									names(uvalue)[w] = "AREA_"
									w2 = which(names(value) == "AREA")[1]
									value[[w2]] = "AREA_"
									names(value)[w2] = "AREA_"
								}
								w = which(uvalue == "LENGTH")[1]
								if (!is.na(w)) {
									uvalue[w] = "LENGTH_"
									names(uvalue)[w] = "LENGTH_"
									w2 = which(names(value) == "LENGTH")[1]
									value[[w2]] = "LENGTH_"
									names(value)[w2] = "LENGTH_"
								}
							}

							if (data_vars || geo_vars) vars = uvalue else vars = character(0)
						} else {
							data_vars = FALSE
							bypass_vars = FALSE
							geo_vars = FALSE
							vars = character(0)
						}
					}

					nvars = length(value) #m
					nvari = vapply(value, length, integer(1))

					convert2density = "convert2density" %in% names(scale) && scale$convert2density

					nflvar = nvars
					if (data_vars) {
						flvar = names(value)
						update_grp_vars(lev = flvar)
						add_used_vars(vars, convert2density = convert2density)
					} else if (geo_vars) {
						flvar = names(value)
						update_grp_vars(lev = flvar)
						add_used_vars(vars)
					} else {
						mfun = paste0("tmapValuesSubmit_", aes)
						if (exists(mfun)) {
							value = do.call(mfun, list(x = value, args = args))
						}
						nvars = length(value)
						nflvar = nvars
						nvari = vapply(value, length, integer(1))
						#value =

						#value = value_orig
						update_grp_vars(m = nflvar)
					}
				}
			})
			b
		}

		gs = tmap_graphics_name()

		# preprocess layers: check aes values
		tmg$tmls = lapply(tmg$tmls, function(tml) {
			within(tml, {
				mapping.fun = paste0("tm_data_", layer)
				if (!(mapping.fun[1] %in% get_method_classes(paste0("tmap", gs, "DataPlot")))) {

					mode = getOption("tmap.mode")

					modes = get_modes()
					gss = vapply(modes, tmap_graphics_name, FUN.VALUE = character(1))

					exs = vapply(1L:length(modes), function(i) {
						mapping.fun[1] %in% get_method_classes(paste0("tmap", gss[i], "DataPlot"))
					}, FUN.VALUE = logical(1))
					modes2 = modes[exs]

					if (length(modes2)) {
						cli::cli_inform("{.field [{mode} mode]} map layer {.code tm_{layer[1]}} not available for mode {.str {mode}}. This map layer is only available for modes {.str {modes2}}")
					} else {
						cli::cli_inform("{.field [{mode} mode]} map layer {.code tm_{layer[1]}} not available for mode {.str {mode}} and also not for the other modes")
					}
					remove = TRUE
				} else {
					remove = FALSE
					if (length(trans.aes)) trans.aes = lapply(trans.aes, precheck_aes, layer = tml$layer, shpvars = smeta$vars, args = trans.args)
					if (length(mapping.aes)) mapping.aes = lapply(mapping.aes, precheck_aes, layer = tml$layer, shpvars = smeta$vars, args = mapping.args)

					# get first non-empty split_stars_dim, the dimension specificied by tm_vars_dim
					split_stars_dim = get_split_stars_dim(mapping.aes)

					if (length(hover) > 1) {
						cli::cli_abort("hover should have length <= 1, not {length(hover)}.", call = NULL)
					}

					if (is.na(hover)) {
						hover = id
					} else if (is.logical(hover)) {
						hover = ifelse(hover, id, "")
					}

					if (hover != "" && !hover %in% smeta$vars) rlang::arg_match0(hover, smeta$vars, arg_nm = "hover", error_call = NULL)
					if (hover != "") add_used_vars(hover, is_title = TRUE)
					if (id != "" && !id %in% smeta$vars) rlang::arg_match0(id, smeta$vars, arg_nm = "id", error_call = NULL)
					if (id != "") add_used_vars(id, is_title = TRUE)

					# popup.title: like id/hover, a title-type variable that must be
					# registered so its column is included in dt in step2. It may be
					# NULL (layers that don't set it, e.g. extensions), NA (default,
					# meaning the popup title falls back to id), or a (possibly named)
					# length-1 variable name. Validation uses the value (names stripped).

					if (!exists("popup.title", inherits = FALSE)) popup.title = NULL
					if (!is.null(popup.title) && !(length(popup.title) == 1L && is.na(popup.title))) {
						if (length(popup.title) > 1) {
							cli::cli_abort("popup.title should have length <= 1, not {length(popup.title)}.", call = NULL)
						}
						if (!unname(popup.title) %in% smeta$vars) rlang::arg_match0(unname(popup.title), smeta$vars, arg_nm = "popup.title", error_call = NULL)
						add_used_vars(unname(popup.title), is_title = TRUE)
					}

					if (isTRUE(popup.vars)) {
						popup.vars = smeta$vars
					} else if (isFALSE(popup.vars)) {
						popup.vars = character(0)
					} else if (is.na(popup.vars[1])) {
						popup.vars = smeta$vars
					}

					if (!all(popup.vars %in% smeta$vars)) {
						rlang::arg_match(popup.vars, values = smeta$vars, multiple = TRUE)
					}
					if (length(popup.vars)) add_used_vars(popup.vars)


					if (length(popup.format) != 0 && !is.null(names(popup.format)) && all(names(popup.format) %in% popup.vars)) {
						popup.called = names(popup.format)
						popup.format = lapply(popup.vars, function(pv) {
							if (pv %in% names(popup.format)) {
								process_label_format(popup.format[[pv]], o$label.format)
							} else {
								process_label_format(list(), o$label.format)
							}
						})
					} else {
						popup.called = character(0)
						one.popup.format = process_label_format(popup.format, o$label.format)
						popup.format = lapply(popup.vars, function(pv) {
							one.popup.format
						})

					}
					if (!is.null(names(popup.vars))) {
						names(popup.format) = names(popup.vars)
					} else {
						names(popup.format) = popup.vars
					}
					attr(popup.format, "called") = popup.called				}
			})
		})

		tmg$tmls = tmg$tmls[vapply(tmg$tmls, function(l) {
			!l$remove
		}, FUN.VALUE = logical(1))]


		vl = get("vl", envir = .TMAP)
		vn = get("vn", envir = .TMAP)

		# split stars if needed (dimension -> attributes)
		split_stars_dim = get_split_stars_dim(tmg$tmls)
		shp = tmapSplitShp(shp, split_stars_dim, smeta)
		if (split_stars_dim != "") {
			smeta = tmapGetShapeMeta1(shp, layer, o)
			if (dev) timing_add(s3 = "get_shape_meta1_2")
		}

		nrsd = length(smeta$dims) # number of required shape dimensions
		nrvd = as.integer(vn > 1L) # number of required variable dimensions (0 or 1)
		nrd = nrsd + nrvd # number of required by-dimensions


		tmg$tmf = within(tmg$tmf, {
				by1 = rows
				by2 = columns
				by3 = pages

				trans_animate = !is.null(.TMAP$animate)
				nad = as.integer(trans_animate)

				if (trans_animate) {
					if (is.null(by3)) {
						by3 = "FRAME__"
					} else {
						if (by3 == "VARS__") {
							if (nrvd == 0) {
								by3 = "FRAME__"
							} else {
								if (is.null(by1)) {
									by1 = "VARS__"
									by3 = "FRAME__"
								} else if (is.null(by2)) {
									by2 = "VARS__"
									by3 = "FRAME__"
								} else {
									ani = names(.TMAP$animate)
									cli::cli_abort("transformation animation for variable {.val {ani}} cannot be used when pages is already specified to multiple variables")
								}

							}
						} else {
							ani = names(.TMAP$animate)
							cli::cli_abort("transformation animation for variable {.val {ani}} cannot be used when pages is already specified to {.val {by3}}")
						}
					}
				}

				# if (animate) {
				# 	if (!is.null(by3) {
				#
				# 		if (nrvd == 1) {
				#
				# 	} else if ()
				#
				# 	}
				# }


				if (!is.null(by)) {
					if (is.null(by1)) {
						by1 = by
					} else if (is.null(by2)) {
						by2 = by
					} else if (is.null(by3)) {
						by3 = by
					} else {
						stop("Too many dimensions") # quick&dirty, to do: move to other checks
					}
				}


				rev1 = is_rev(by1)
				rev2 = is_rev(by2)
				rev3 = is_rev(by3)

				if (rev1) by1 = remove_min(by1)
				if (rev2) by2 = remove_min(by2)
				if (rev3) by3 = remove_min(by3)




				if (nrd > 3L) {
					if (nrsd > 3L) cli::cli_abort("The shape object has more than 3 dimensions, so even {.fn tm_facets_grid} cannot be used.")
					nrvd = 0L
					nrd = 3L
					limitvars = TRUE
				} else {
					limitvars = FALSE
					#todo
					if (nrvd == 1L && !identical(by1, "VARS__") && !identical(by2, "VARS__") && !identical(by3, "VARS__")) {
						if (is.null(by1)) {
							by1 = "VARS__"
						} else if (is.null(by2)) {
							by2 = "VARS__"
						} else if (is.null(by3)) {
							by3 = "VARS__"
						}
					}
				}
				unsigned = setdiff(smeta$dims, c(by1, by2, by3))

				if (length(unsigned)) {
					if (is.null(by1)) by1 = unsigned[1] else if (is.null(by2)) by2 = unsigned[1] else by3 = unsigned[1]
					if (length(unsigned) > 1) {
						if (is.null(by2)) by2 = unsigned[2] else by3 = unsigned[2]
					}
					if (length(unsigned) == 3) {
						by3 = unsigned[3]
					}
				}

			# }

			bys = c(by1, by2, by3)
			if (length(bys)) {
				byvars = intersect(smeta$vars, bys)
				if (length(byvars)) add_used_vars(byvars)
			}
			if (!all(bys %in% c("VARS__", "FRAME__", smeta$vars, smeta$dims))) stop("unknown facet variables", call. = FALSE)
			if (is.na(na.text)) na.text = o$label.na

			if (is.na(type)) {
				type = if (!is.null(by1) && !is.null(by2)) "grid" else "wrapstack"
			}

		})

		smeta$vars = get("used_vars", envir = .TMAP)
		convert2density = get("c2d_vars", envir = .TMAP)
		smvars = if (length(convert2density)) unique(c(smeta$vars, "AREA")) else smeta$vars
		shp = tmapSubsetShp(shp, smvars)

		if (length(convert2density)) {
			for (v in convert2density) {
				sunit = tmg$tms$unit
				if (is.null(sunit)) sunit = o$unit
				shape.unit <- ifelse(sunit == "metric", "km", ifelse(sunit == "imperial", "mi", sunit))
				u = paste(shape.unit, shape.unit)
				if (is.numeric(shp[[v]])) shp[[v]] = shp[[v]] / units::set_units(shp$AREA, u, mode = "standard")
			}
		}


		if (dev) timing_add(s3 = "subset_shp")

		smeta = tmapGetShapeMeta2(shp, smeta, c(o, tmg$tmf))
		if (dev) timing_add(s3 = "get_shape_meta2")

		tmg$tmf = within(tmg$tmf, {

			gl = list(NULL, NULL, NULL)
			gn = c(1L, 1L, 1L)

			for (i in 1L:3L) {
				byi = get(paste0("by", i))
				if (!is.null(byi)) {
					if (byi == "FRAME__") {
						gl[i] = list(NULL)
						gn[i] = as.integer(nframes)
					} else if (byi == "VARS__") {
						gl[i] = list(vl)
						gn[i] = vn
					} else if (byi %in% smeta$vars) {
						gl[i] = list(smeta$vars_levs[[byi]])
						gn[i] = length(gl[[i]])
					} else if (byi %in% smeta$dims) {
						gl[[i]] = smeta$dims_val[[match(byi, smeta$dims)]]
						gn[i] = length(gl[[i]])
					}
				}
			}

			if (is.na(free.coords)) {
				if (type %in% c("wrapstack", "wrap", "stack")) {
					free.coords = rep(!any(c(by1, by2, by3) == "VARS__"), 3)
				} else {
					free.coords = c((!is.null(rows) && (rows != "VARS__")), (!is.null(columns)) && (columns != "VARS__"), (!is.null(pages)) && (pages != "VARS__"))
				}
			} else {
				free.coords = rep(free.coords, length.out = 3)
			}


			v = which(vapply(list(by1, by2, by3), identical, FUN.VALUE = logical(1), "VARS__"))

			b = setdiff(which(!vapply(list(by1, by2, by3), is.null, FUN.VALUE = logical(1))), v)

			by123 = paste0("by", 1L:3L)
			by123__ = paste0("by", 1L:3L, "__")

			var__ = by123__[v]
			by__ = by123__[b]


			#fl1 =
		})
		tmg$tms$shp = shp
		tmg$tms$smeta = smeta

		tmg
	})

	tmf = get_tmf(lapply(tmo, function(tmoi) tmoi$tmf))

	tmo$tmf_global = tmf
	tmo
}

#' ggplot2 theme for proportional symbols
#' 
#' ggplot2 theme for proportional symbols. By default, this theme only shows the
#' plotting area, so without titles, axes, and legend.
#' 
#' @param base_size base size
#' @param base_family base family
#' @param plot.axes should the axes be shown?
#' @param plot.legend should the legend(s) be shown?
#' @export
theme_ps <- function(base_size = 12, base_family = "", plot.axes = FALSE, plot.legend = FALSE) {
  rlang::check_installed("ggplot2", reason = "to create plot themes.")
  replace <- function(e1, e2) {
    e1[names(e2)] <- e2
    e1
  }
  tps <- replace(
    ggplot2::theme_minimal(base_size = base_size, base_family = base_family),
    ggplot2::theme(
      panel.background = ggplot2::element_blank(),
      panel.border = ggplot2::element_blank(),
      panel.grid = ggplot2::element_blank(),
      strip.background = ggplot2::element_blank(),
      plot.background = ggplot2::element_blank(),
      plot.title = ggplot2::element_blank(),
      strip.text = ggplot2::element_blank(),
      panel.spacing = unit(0, "null"),
      plot.margin = rep(unit(0, "null"), 4)
    )
  )
  if (!plot.legend) {
    tps <- replace(tps, ggplot2::theme(legend.position = "none"))
  }
  if (plot.axes) {
    tps
  } else {
    replace(
      tps,
      ggplot2::theme(
        axis.text = ggplot2::element_blank(),
        axis.title = ggplot2::element_blank(),
        axis.ticks = ggplot2::element_line(),
        axis.ticks.length = unit(0, "lines")
      )
    )
  }
}


theme_chart <- function(base_size = 12 * scale, base_family = "", plot.axes = FALSE, plot.axis.x = FALSE, plot.axis.y = FALSE, plot.grid.x = FALSE, plot.grid.y = TRUE, plot.legend = FALSE, scale = 1, text.color = "#000000", text.size = 1) {
	rlang::check_installed("ggplot2", reason = "to create plot themes.")
	replace <- function(e1, e2) {
		e1[names(e2)] <- e2
		e1
	}
	
	tps <- replace(
		ggplot2::theme_minimal(base_size = base_size, base_family = base_family),
		ggplot2::theme(
			text = ggplot2::element_text(color = text.color, size = base_size * text.size),
			axis.text = ggplot2::element_text(color = text.color, size = base_size * text.size),
			panel.background = ggplot2::element_blank(),
			panel.border = ggplot2::element_blank(),
			panel.grid = ggplot2::element_blank(),
			strip.background = ggplot2::element_blank(),
			plot.background = ggplot2::element_blank(),
			plot.title = ggplot2::element_blank(),
			strip.text = ggplot2::element_blank(),
			panel.spacing = unit(0, "null"),
			plot.margin = unit(rep(.25, 4), "lines")
		)
	)
	if (!plot.legend) {
		tps <- replace(tps, ggplot2::theme(legend.position = "none"))
	}
	
	if (plot.axes) {
		plot.axis.x = TRUE
		plot.axis.y = TRUE
	}
	
	
	if (plot.grid.x) {
		tps = replace(
			tps,
			ggplot2::theme(
				panel.grid.major.x = ggplot2::element_line(colour = "#888888", linewidth = lwd_to_mm(scale))	
			)
		)
	}
	if (plot.grid.y) {
		tps = replace(
			tps,
			ggplot2::theme(
				panel.grid.major.y = ggplot2::element_line(colour = "#888888", linewidth = lwd_to_mm(scale))	
			)
		)
	}
	
	if (plot.axis.x) {
		tps = replace(
			tps,
			ggplot2::theme(
				#axis.text.x = ggplot2::element_blank(),
				axis.title.x = ggplot2::element_blank(),
				axis.ticks.x = ggplot2::element_line(),
				axis.ticks.length.x = unit(0, "lines")
			)
		)
	} else {
		tps = replace(
			tps,
			ggplot2::theme(
				axis.text.x = ggplot2::element_blank(),
				axis.title.x = ggplot2::element_blank(),
				axis.ticks.x = ggplot2::element_line(),
				axis.ticks.length.x = unit(0, "lines")
			)
		)
	}
	
	if (plot.axis.y) {
		tps = replace(
			tps,
			ggplot2::theme(
				#axis.text.y = ggplot2::element_blank(),
				axis.title.y = ggplot2::element_blank(),
				axis.ticks.y = ggplot2::element_line(),
				axis.ticks.length.y = unit(0, "lines")
			)
		)
	} else {
		tps = replace(
			tps,
			ggplot2::theme(
				axis.text.y = ggplot2::element_blank(),
				axis.title.y = ggplot2::element_blank(),
				axis.ticks.y = ggplot2::element_line(),
				axis.ticks.length.y = unit(0, "lines")
			)
		)
	}
		
	tps
}

# NEED TO SYNC THIS WITH tmap_options
# generate all options with:
# args = setdiff(intersect(names(tmap_options_mode("view", mode.specific = FALSE)),names(tmap_options_mode("plot", mode.specific = FALSE))),  "name")
# cat(paste(args, collapse = ", "))
# manual: tm_layout: start with scale (the options before that are tm_options only)
#' Layout options
#'
#' Specify the layout of the maps. [tm_layout()] is identical as [tm_options()] but only contain the tmap options that are directly related to the layout. [tm_style()] sets the style for the map. A style is a specified set of options (that can be changed afterwards with [tm_layout()]). These functions are used within a plot a plot call (stacked with the `+` operator). Their counterparts [tmap_options()] and [tmap_style()] can be used to set the (layout) options globally.
#'
#' @seealso \href{https://r-tmap.github.io/tmap/articles/basics_layout}{Vignette about layout}, \href{https://r-tmap.github.io/tmap/articles/adv_margins}{vignette about margins and aspect ratio} and \href{https://r-tmap.github.io/tmap/articles/adv_options}{vignette about options}
#' @inheritParams tmap_options
#' @example ./examples/tm_layout.R
#' @export
tm_layout = function(
		scale, asp, legend.only, bg, bg.color, outer.bg, outer.bg.color, frame, frame.color, frame.alpha, frame.lwd, frame.r, frame.double_line, outer.margins, inner.margins, inner.margins.extra, meta.margins, meta.auto_margins, between_margin,
		text.fontfamily, text.fontface, r, attr.color,
		panel.margin,
		panel.type, panel.wrap.pos, panel.xtab.pos, color.sepia_intensity, color.saturation, color_vision_deficiency_sim,
		panel.show, panel.labels, panel.label.size, panel.label.color, panel.label.fontface, panel.label.fontfamily, panel.label.alpha, panel.label.bg, panel.label.bg.color, panel.label.bg.alpha, panel.label.frame, panel.label.frame.color, panel.label.frame.alpha, panel.label.frame.lwd, panel.label.frame.r, panel.label.height, panel.label.rot, earth_boundary, earth_boundary.color, earth_boundary.lwd, earth_datum, space, space.color,
		...
	) {

	args = lapply(as.list(rlang::call_match(dots_expand = TRUE)[-1]), eval, envir = parent.frame())
	args$called_from = "tm_layout"

	# 1106: need to use this for other texts as well
	if ("panel.labels" %in% names(args)) {
		if (is.list(args$panel.labels)) {
			args$panel.labels = lapply(args$panel.labels, encode_expr)
		} else {
			args$panel.labels = encode_expr(args$panel.labels)
		}
	}

	do.call(tm_options, args)
}

#' View mode options
#'
#' View mode options. These options are specific to the view mode.
#'
#' @param use_browser If `TRUE` it opens an external browser, and `FALSE` (default) it opens the internal IDEs (e.g. RStudio) browser.
#' @param use_WebGL use webGL for points, lines, and polygons. For large spatial objects, this is much faster than the standard leaflet layer functions. However, it can not always be used for two reasons. First, the number of visual variables are limited; only fill, size, and color (for lines) are supported. Second, projected CRS's are not supported. Furthermore, it has the drawback that polygon borders are not as sharp. By default only `TRUE` for large spatial objects (1000 or more features) when the mentioned criteria are met.
#'  By default `TRUE` if no other visual variables are used.
#' @param control.position `r .doc_opt("control.position")`
#' @param control.bases base layers
#' @param control.overlays overlay layers
#' @param control.collapse Should the box be collapsed or expanded?
#' @param set_bounds logical that determines whether maximum bounds are set,
#'   or a bounding box. Not applicable in plot mode.
#'   In view mode, this is passed on to [setMaxBounds()][leaflet::setMaxBounds()]
#' @param set_view numeric vector that determines the view.
#'   Either a vector of three: `lng`, `lat`, and `zoom`, or a single value:
#'    `zoom`. See [setView()][leaflet::setView()].
#'    Only applicable if `bbox` is not specified
#' @param set_zoom_limits numeric vector of two that set the minimum and maximum
#'   zoom levels (see [tileOptions()][leaflet::tileOptions()]).
#' @param use_circle_markers If `TRUE` (default) circle shaped symbols (e.g. `tm_dots()` and `tm_symbols()`) will be rendered as [addCircleMarkers()][leaflet::addCircleMarkers()] instead of [addMarkers()][leaflet::addMarkers()]. The former is faster, the latter can support any symbol since it is based on icons
#' @param leaflet.options options passed on to
#'   [leafletOptions()][leaflet::leafletOptions()]
#' @param ... to catch deprecated arguments
#' @export
#' @seealso [tm_group()]
tm_view = function(use_browser,
				   use_WebGL,
				   control.position,
				   control.bases,
				   control.overlays,
				   control.collapse,
				   set_bounds,
				   set_view,
				   set_zoom_limits,
				   use_circle_markers,
				   leaflet.options,
				   ...) {

	args = lapply(as.list(rlang::call_match(dots_expand = TRUE)[-1]), eval, envir = parent.frame())
	args$called_from = "tm_view"
	do.call(tm_options, args)

}

#' Plot mode options
#'
#' Plot mode options. This option is specific to the plot mode.
#'
#' @param use_gradient Use gradient fill using [linearGradient()][grid::linearGradient()]
#' @param limit_latitude_3857 Vector of two limit latitude values for maps printed in Web Mercator projection (EPSG 3857). If `c(-90, 90)` the poles will be inflated too much. The Web Mercator is defines as `c(-85.06, 85.06)`, but the default setting in tmap is `c(-84, 84)`.
#' @export
tm_plot = function(use_gradient, limit_latitude_3857) {
	args = lapply(as.list(rlang::call_match()[-1]), eval, envir = parent.frame())
	do.call(tm_options, args)
}


#' @rdname tm_extra_innner_margin
#' @export
tm_place_legends_right = function(width = NA) {
	if (is.na(width)) {
		tm_options(legend.position = tm_pos_out("right", "center"))
	} else {
		tm_options(meta.margins = c(0, 0, 0, width), legend.position = tm_pos_out("right", "center"))
	}
}

#' @rdname tm_extra_innner_margin
#' @param width width
#' @export
tm_place_legends_left = function(width = NA) {
	if (is.na(width)) {
		tm_options(legend.position = tm_pos_out("left", "center"))
	} else {
		tm_options(meta.margins = c(0, 0, 0, width), legend.position = tm_pos_out("right", "center"))
	}
}

#' @rdname tm_extra_innner_margin
#' @param height height
#' @export
tm_place_legends_bottom = function(height = NA) {
	if (is.na(height)) {
		tm_options(legend.position = tm_pos_out("center", "bottom"))
	} else {
		tm_options(meta.margins = c(height, 0, 0, 0), legend.position = tm_pos_out("center", "bottom"))
	}
}

#' @rdname tm_extra_innner_margin
#' @export
tm_place_legends_top = function(height = NA) {
	if (is.na(height)) {
		tm_options(legend.position = tm_pos_out("center", "top"))
	} else {
		tm_options(meta.margins = c(height, 0, 0, 0), legend.position = tm_pos_out("center", "top"))
	}
}

#' @rdname tm_extra_innner_margin
#' @param pos.h,pos.v position (horizontal and vertical)
#' @export
tm_place_legends_inside = function(pos.h = NULL, pos.v = NULL) {
	if (is.null(pos.h) || is.null(pos.v)) {
		if (!is.null(pos.h)) {
			warning("tm_place_legends_inside also requires pos.v", call. = FALSE)
		} else if (!is.null(pos.v)) {
			warning("tm_place_legends_inside also requires pos.h", call. = FALSE)
		}
		tm_options(legend.position = tm_pos_auto_in())
	} else {
		tm_options(legend.position = tm_pos_in(pos.h = pos.h, pos.v = pos.v))
	}
}

#' tmap layout: helper functions
#'
#' @param left,right,top,bottom extra margins
#' @export
#' @name tm_extra_innner_margin
tm_extra_inner_margin = function(left = 0, right = 0, top = 0, bottom = 0) {
	tm_options(inner.margins.extra = c(bottom, left, top, right))
}

#' @rdname tmap_options
#' @export
tm_check_fix = function() {
	tm_options(check_and_fix = TRUE)
}


#' @rdname tm_layout
#' @param style name of the style
#' @export
#' @order 1
tm_style = function(style, ...) {
	args = list(...)

	check_style(style)

	args$style = style
	args$called_from = "tm_style"
	do.call(tm_options, args)
}

# Will need investigating, maybe some are true flags
# i.e. a function that needs to be prefixed, a variable that is unused, or misnamed
# although most seem to come from data.table variables.
# But this will reduce clutter in R CMD CHECK
utils::globalVariables(c(
  ".", "..cols", "X__", "Y__", "asp", "attr.color", "bbox",
  "between_margin", "bi", "blue", "by1", "by1__", "by2", "by2__", "by2b__",
  "by3", "by3__", "ca", "cell.h", "cell.v", "columns", "comp",
  "crs", "crs_parameters", "dasp", "devsize", "facet_col", "facet_row", "fl",
  "fn", "gp", "green", "grid.labels.format",
  "grid.labels.inside_frame", "grid.labels.margin.x", "grid.labels.margin.y",
  "grid.labels.pos", "grid.labels.rot", "grid.labels.size",
  "grid.labels.space.x", "grid.labels.space.y", "hover", "id",
  "ids", "label.show", "labels.cardinal", "labels.inside_frame",
  "labels.pos", "labels.rot", "legH", "legW",
  "legend.present.auto", "legend.present.fix", "lfmv", "lineH", "lineW",
  "main_class", "mapping.args", "mfun", "nby", "ndiscr", "ord__",
  "outer.margins", "pages", "panel.label.height",
  "panel.labels", "panel.type", "panel.wrap.pos", "panel.xtab.pos", "pos.h",
  "pos.h.id", "pos.v", "pos.v.id", "red", "rows", "s1", "s2", "s3", "s4",
  "scale.factor", "set_bounds",
  "split_geometry_collection", "stack_auto", "t1",
  "t2", "t3", "t4", "text.fontface", "text.fontfamily", "title.bg.alpha",
  "tmapID__", "vneutral",
  "panel.show", "set_to_stack_message", "rev1", "rev2", "rev3", "panel.show",
  "zauto",
  "calledCV", "calledCH"
))

# Add more to silence R CMD CHECK (see if some are false positive)
utils::globalVariables(c(
	"aes", "alpha", "col_alpha", "frame", "grid.show", "label.na", "legend",
	"legend.bg.alpha", "lin", "m", "n", "overlays_tiles", "show", "show.labels",
	"show.warnings", "total", "trans.args", "type", "values", "xlab.rotation",
	"xlab.show", "xlab.side", "xlab.space", "xlab.text", "ylab.rotation",
	"ylab.show", "ylab.side", "ylab.space", "ylab.text", "z",
	"x", "y", "color", "freq", "bin", "bin1", "bin2",
	"nitems", "item.width", "page", "stack",
  "item.height", "absolute_fontsize", "rc_text", "show", "gp2", "trns", "nbins",
  "World", "World_rivers", "NLD_muni", "NLD_prov", ".N"
))

#' Legend
#'
#' Legend specification
#'
#' @param title Legend title
#' @param show Show legend?
#' @param orientation Orientation of the legend: `"portrait"` or `"landscape"`
#' @param reverse Should the legend items be reversed?
#' @param na.show Show `NA` values in legend?
#' @param width Width of the legend. Units are 'text line heights'. In case a negative number is specified, the units are (approximate) pixels. The relation between these two is configured via the option `absolute_fontsize`.
#' @param height Height of the legend. Units are 'text line heights'. In case a negative number is specified, the units are (approximate) pixels. The relation between these two is configured via the option `absolute_fontsize`.
#' @inheritParams tm_title
#' @param title.color `r .doc_opt("legend.title.color")`
#' @param title.size `r .doc_opt("legend.title.size")`
#' @param title.fontface `r .doc_opt("legend.title.fontface")`
#' @param title.fontfamily `r .doc_opt("legend.title.fontfamily")`
#' @param title.alpha `r .doc_opt("legend.title.alpha")`
#' @param title.padding `r .doc_opt("legend.title.padding")`
#' @param title.align `r .doc_opt("legend.title.align")`
#' @param text.color `r .doc_opt("legend.text.color")`
#' @param text.size `r .doc_opt("legend.text.size")`
#' @param text.fontface `r .doc_opt("legend.text.fontface")`
#' @param text.fontfamily `r .doc_opt("legend.text.fontfamily")`
#' @param text.alpha `r .doc_opt("legend.text.alpha")`
#' @param format Not used anymore: use the format argument of the `tm_scale_*()` functions instead.
#' @param bg.color `r .doc_opt("legend.bg.color")`
#' @param bg.alpha `r .doc_opt("legend.bg.alpha")`
#' @param absolute_fontsize `r .doc_opt("legend.absolute_fontsize")`
#' @param item.height `r .doc_opt("legend.item.height")`
#' @param item.width `r .doc_opt("legend.item.width")`
#' @param item.space `r .doc_opt("legend.item.space")`
#' @param item.na.height `r .doc_opt("legend.na item.height")`
#' @param item.na.width `r .doc_opt("legend.na item.width")`
#' @param item.na.space `r .doc_opt("legend.na item.space")`
#' @param item.shape `r .doc_opt("legend.item.shape")`
#' @param ticks List of vectors of size 2 that determines a tick mark line (for portrait legends). The values are the y-values of begin and endpoint of a tick mark. For a solid line, only one vector is required, for dashed lines one for each dash. See `r .doc_see_vig("versus_ggplot2")`.
#' @param ticks.disable.na Remove ticks for `NA` values
#' @param ticks.col `r .doc_opt("legend.ticks.col")`
#' @param ticks.lwd `r .doc_opt("legend.ticks.lwd")`
#' @param margins `r .doc_opt("legend.margins")`
#' @param item_text.margin `r .doc_opt("legend.item_text.margin")`
#' @param ... visual values, e.g. `col`, `fill`, `lwd`, can be specified. If so, they overrule the default visual values, which are determined by the drawn map objects (e.g. polygons)
#' @param variable visual (or transformation) variable to combine the legend with: e.g. `"fill"` or `"size"`
#' @seealso `r .doc_see_also_legend(advanced = TRUE, biv = TRUE)` `r .doc_see_also_comp()`
#' @return A tm_legend component
#' @export
#' @example ./examples/tm_legend.R
tm_legend = function(title,
					 show,
					 orientation,
					 reverse,
					 na.show,
					 position,
					 group_id,
					 width,
					 height,
					 z,
					 title.color,
					 title.size,
					 title.fontface,
					 title.fontfamily,
					 title.alpha,
					 title.padding,
					 title.align,
					 text.color,
					 text.size,
					 text.fontface,
					 text.fontfamily,
					 text.alpha,
					 format,
					 frame,
					 frame.color,
					 frame.alpha,
					 frame.lwd,
					 frame.r,
					 bg,
					 bg.color,
					 bg.alpha,
					 absolute_fontsize,
					 item.height,
					 item.width,
					 item.space,
					 item.na.height,
					 item.na.width,
					 item.na.space,
					 item.shape,
					 ticks,
					 ticks.disable.na,
					 ticks.col,
					 ticks.lwd,
					 margins,
					 item_text.margin,
					 ...) {
	args = lapply(as.list(rlang::call_match()[-1]), eval, envir = parent.frame())
	args$called = names(args)
	args = warning_group_args(args)

	args$title = args$title %||% NA
	args$xlab = args$xlab %||% NA
	args$ylab = args$ylab %||% NA
	args$group_id = args$group_id %||% NA_character_
	args$group_type = "tm_legend"
	args$z = args$z %||% NA_integer_
	structure(args, class = c("tm_legend", "tm_component", "list"))
}

#' @rdname tm_legend
#' @export
tm_legend_hide = function() {
	tm_legend(show = FALSE)
}

#' @rdname tm_legend
#' @export
tm_legend_combine = function(variable) {
	structure(list(FUN = "tmapLegend", title = NA, reverse = FALSE, show = FALSE, aes = variable, group_id = NA_character_), class = c("tm_legend", "tm_component", "list"))
}

#' @rdname tm_legend
#' @param xlab label for the x dimension (rows)
#' @param ylab label for the y dimension (columns)
#' @param xlab.color `r .doc_opt("legend.xlab.color")`
#' @param xlab.size `r .doc_opt("legend.xlab.size")`
#' @param xlab.rot `r .doc_opt("legend.xlab.rot")`
#' @param xlab.fontface `r .doc_opt("legend.xlab.fontface")`
#' @param xlab.fontfamily `r .doc_opt("legend.xlab.fontfamily")`
#' @param xlab.alpha `r .doc_opt("legend.xlab.alpha")`
#' @param xlab.padding `r .doc_opt("legend.xlab.padding")`
#' @param xlab.align `r .doc_opt("legend.xlab.align")`
#' @param ylab.color `r .doc_opt("legend.ylab.color")`
#' @param ylab.size `r .doc_opt("legend.ylab.size")`
#' @param ylab.rot `r .doc_opt("legend.ylab.rot")`
#' @param ylab.fontface `r .doc_opt("legend.ylab.fontface")`
#' @param ylab.fontfamily `r .doc_opt("legend.ylab.fontfamily")`
#' @param ylab.alpha `r .doc_opt("legend.ylab.alpha")`
#' @param ylab.padding `r .doc_opt("legend.ylab.padding")`
#' @param ylab.align `r .doc_opt("legend.ylab.align")`
#' @export
tm_legend_bivariate = function(xlab,
							   ylab,
							   xlab.color,
							   xlab.size,
							   xlab.rot,
							   xlab.fontface,
							   xlab.fontfamily,
							   xlab.alpha,
							   xlab.padding,
							   xlab.align,
							   ylab.color,
							   ylab.size,
							   ylab.rot,
							   ylab.fontface,
							   ylab.fontfamily,
							   ylab.alpha,
							   ylab.padding,
							   ylab.align,
							   ...) {
	args = lapply(as.list(rlang::call_match()[-1]), eval, envir = parent.frame())

	args$title = args$title %||% NA
	args$xlab = args$xlab %||% NA
	args$ylab = args$ylab %||% NA
	args$group_id = args$group_id %||% NA_character_
	args$group_type = "tm_legend_bivariate"
	args$z = args$z %||% NA_integer_
	args$orientation = "portrait"
	structure(args, class = c("tm_legend", "tm_component", "list"))
}

#' Map component: manual legend
#'
#' Map component that adds a manual legend.
#'
#' @param ... visual variables and arguments passed on to `tm_legend()`.
#'   By default, the argument `type` is set to `"symbols"`, which means that the
#'   supported visual variables are: `"fill"`, `"col"`, `"shape"`, `"size"`,
#'   `"fill_alpha"`, `"col_alpha"`, `"lty"`, `"lwd"`, `"linejoin"`, and `"lineend"`.
#'   The number of legend items will be equal to the maximum number of specific values (and specified labels.)
#' @param labels labels by default `""` (so omitted)
#' @param type the layer type from which the visual variables (see `...`) are taken.
#'   Options: `"symbols"` (default), `"lines"`, `"polygons"`, and `"text"`.
#' @param title `r .doc_opt("legend.title")`
#' @param orientation `r .doc_opt("legend.orientation")`
#' @param position `r .doc_opt("legend.position")`
#' @inheritParams tm_title
#' @param group Name of the group to which this layer belongs. This is only
#'   relevant in view mode, where layer groups can be switched (see `group.control`)
#' @param group.control In view mode, the group control determines how
#'   layer groups can be switched on and off. Options: `"radio"` for radio
#'   buttons (meaning only one group can be shown), `"check"` for check boxes
#'   (so multiple groups can be shown), and `"none"` for no control
#'   (the group cannot be (de)selected).
#' @example ./examples/tm_add_legend.R
#' @seealso `r .doc_see_also_legend(advanced = TRUE)`
#' @export
tm_add_legend = function(...,
						 labels = "",
						 type = "symbols",
						 title = "",
						 orientation = NULL,
						 position = NULL,
						 group_id = NA_character_,
						 group = NA,
						 group.control = "check",
						 z = NA_integer_) {

	args = lapply(as.list(rlang::call_match(defaults = TRUE)[-1]), eval, envir = parent.frame())
	args$called = names(rlang::call_match(dots_expand = TRUE)[-1])

	args = warning_group_args(args)
	args$group_id = args$group_id %||% NA_character_
	args$group_type = "tm_legend"
	args$z = args$z %||% NA_integer_

	if (type %in% c("fill", "symbol", "line")) {
		args$type = v3_add_legend(type, names(args))
		if ("col" %in% names(args) && !c("fill" %in% names(args)) && type != "line") {
			args$fill = args$col
			args$col = NULL
		}
		if ("border.col" %in% names(args)) {
			args$col = args$border.col
			args$border.col = NULL
		}
	}
	tm_element_list(do.call(tm_element, c(args, list(subclass = c("tm_add_legend", "tm_component")))))
}

#' @export
#' @rdname tmap_internal
toTitleCase = function(x) {
	paste0(toupper(substr(x,1,1)), tolower(substr(x,2, nchar(x))))
}


tmapAddedLegend = function(comp, o) {
	#message("tm_mouse_coordinates ignored for 'plot' mode")
	l = update_l(o = o, l = comp, v = "", mfun = paste0("tm_data_", comp$type), unm = "", active = FALSE)

	fun = paste0("tm_", comp$type)
	if (!exists(fun)) {
		stop(paste0("type \"", comp$type, "\" not supported because tm_", comp$type,  " not found"), call. = FALSE)
	}
	res = do.call(fun, args = list())
	gp = res[[1]]$gpar

	for (gpi in names(gp)) {
		if (gpi %in% names(l)) {
			gp[[gpi]] = l[[gpi]]
		} else if (!is.na(gp[[gpi]])) {
			gp[[gpi]] = getAesOption("value.const", o, aes = gpi, layer = comp$type)
		}
	}

	if (comp$type == "text") {
		gp$cex = gp$size
		gp$size = NA
	}

	l$gp = gp

	l2 = within(l, {
		nitems = max(length(labels), vapply(gp, length, FUN.VALUE = integer(1), USE.NAMES = FALSE))
		labels = rep(labels, length.out = nitems)
		dvalues = 1:nitems
		vvalues = 1:nitems
		vneutral = NA
		na.show = FALSE

	})
	l2
}

#' tmap function to specify variables
#'
#' tmap function to specify all variables in the shape object
#'
#' @param x variable names, variable indices, or a dimension name
#' @param dimvalues dimension values
#' @param n if specified the first `n` variables are taken (or the first `n` dimension values)
#' @param multivariate in case multiple variables are specified, should they serve as facets (FALSE) or as a multivariate visual variable?
#' @param animate should the variable(s) be animated? (experimental)
#' @export
tm_vars = function(x = NA, dimvalues = NULL, n = NA, multivariate = FALSE, animate = FALSE) {
	structure(list(x = x, dimvalues = dimvalues, n = n, multivariate = multivariate, animate = animate), class = c("tmapVars", "list"))
}


# process visual variable specification. Can either be tmapVars (output of tm_vars) or a list of values.
tmapVV = function(x) {
	if (is.null(x)) {
		x = structure(list("value.blank"), class = "tmapOption")

	} else if (!is.list(x) && length(x) == 1 && is.na(x)) {
		x = structure(list("value.const"), class = "tmapOption")
	}

	if (inherits(x, c("tmapOption", "tmapVars"))) return(x)

	# if (inherits(x, "tm_shape_vars")) return(structure(list(ids = x$ids, n = x$n), class = "tmapShpVars"))
	# if (inherits(x, "tm_mv_shape_vars")) return(structure(list(ids = x$ids, n = x$n), class = "tmapMVShpVars"))
	# if (inherits(x, "tmapDimVars")) return(x)

	cls = if (inherits(x, "AsIs")) "tmapAsIs" else if (inherits(x, "tmapUsrCls")) "tmapUsrCls" else "tbd"


	isL = is.list(x)

	if (cls == "tbd" && !isL) {
		if (is.character(x) && all(substr(x, 1, 1) == "*")) {
			x = substr(x, 2, nchar(x))
			return(tm_vars(x, animate = TRUE))
		}
	}

	isNestedL = isL && any(vapply(x, is.list, FUN.VALUE = logical(1)))
	isSpecialL = isL && !setequal(class(x), "list")
	isSpecialNestedL = isL && is.list(x[[1]]) &&  !setequal(class(x[[1]]), "list")
	if (!isL) {
		x = as.list(x)
	} else if (isSpecialL) {
		x = list(x)
	}

	if (cls == "tbd") cls = if (isSpecialL || isSpecialNestedL || isNestedL) "tmapSpecial" else "tmapStandard"

	if (cls == "tmapSpecial") {
		nms = seq_along(x)
	} else {
		nms = x
	}

	structure(x, names = nms, class = cls)
}

add_to_gt = function(gt, grb, row, col) {
	vp = grid::viewport(layout.pos.col = col, layout.pos.row = row)
	gtr = grid::grobTree(grb, vp = vp)

	grid::addGrob(gt, gtr, gPath = grid::gPath("gt_main"))
}





get_legend_option = function(x, type) {
	if (length(x) == 1 || (!type %in% names(x))) x else if (is.list(x)) x[[type]] else x[type]
}


gp_to_gpar = function(gp, id = NULL, sel = "all", split_to_n = NULL, pick_middle = TRUE, o, type) {
	type = tolower(type)
	if (sel == "all") {
		if (is.na(gp$fill_alpha[1]) && !is.na(gp$col_alpha[1])) sel = "col"
		if (!is.na(gp$fill_alpha[1]) && is.na(gp$col_alpha[1])) sel = "fill"
	}


	# get alpha value (sel: "all" means fill and col, "fill" and "col" mean fill and col only respectively)
	alpha = if (sel == "fill") {
		if (!is.na(gp$fill_alpha[1])) gp$fill_alpha else 1
	} else {
		if (!is.na(gp$col_alpha[1])) gp$col_alpha else 1
	}

	# create a list of gp elements
	lst = c(list(fill = {if (sel == "col") NA else gp$fill},
			   col = {if (sel == "fill") NA else gp$col},
			   alpha = alpha,
			   lty = if (sel == "fill") getAesOption("value.blank", o, aes = "lty", layer = type) else if (!is.na(gp$lty[1])) gp$lty else getAesOption("value.const", o, aes = "lty", layer = type) ,
			   lwd = {if (!all(is.na(gp$lwd))) gp$lwd else getAesOption("value.blank", o, aes = "lwd", layer = type)},
			   lineend = {if (!all(is.na(gp$lineend))) gp$lineend else "round"},
			   linejoin = {if (!all(is.na(gp$linejoin))) gp$linejoin else "round"},
			   size = {if (!all(is.na(gp$size))) gp$size else 1},
			   cex = {if (!all(is.na(gp$cex))) gp$cex else 1},
			   fontface = {if (!all(is.na(gp$cex))) gp$fontface else getAesOption("value.blank", o, aes = "fontface", layer = type)},
			   fontfamily = {if (!all(is.na(gp$fontfamily))) gp$fontfamily else o$text.fontfamily},
			   shape = {if (!all(is.na(gp$shape))) gp$shape else getAesOption("value.const", o, aes = "shape", layer = type)}))

	#
	if (!is.null(id)) {
		lst = lapply(lst, "[", id)
	}

	lst_isnum = c(fill = FALSE,
				  col = FALSE,
				  alpha = TRUE,
				  lty = FALSE,
				  lwd = TRUE,
				  lineend = FALSE,
				  linejoin = FALSE,
				  size = TRUE,
				  cex = TRUE,
				  fontface = FALSE,
				  fontfamily = FALSE,
				  shape = TRUE)

	lst = mapply(function(lsti, isnum) {
		if (!is.character(lsti)) return(lsti)

		if (nchar(lsti[1]) > 50) {
			x = cont_split(lsti)
			x = lapply(x, function(i) {
				i[i == "NA"] <- NA
				i
			})
			if (isnum) x = lapply(x, as.numeric)
			if (pick_middle) {
				x = sapply(x, function(i) {
					if (all(is.na(i))) NA else {
						sq = (o$continuous.nclass_per_legend_break/2) + (rep(0:5,each=2) * c(1,-1))[-1] # priority for middle values
						i[sq[which(!is.na(i)[sq])[1]]]
					}
				})
			}
			return(x)

		} else {
			return(lsti)
		}
	}, lst, lst_isnum[names(lst)], SIMPLIFY = FALSE)

	if (!is.null(split_to_n)) {
		split_gp(lst, split_to_n)
	} else {
		do.call(grid::gpar, lst)
	}
}

split_gp = function(gp, n) {
	lst = lapply(1L:n, function(i) {
		lapply(gp, function(gpi) {
			if (length(gpi) == n) gpi[i] else gpi[1]
		})
	})
	lapply(lst, function(lsti) {
		do.call(grid::gpar, lsti)
	})
}


determine_scale = function(label, rot, row, col, g, scale = 1) {

	w = sum(g$colsIn[col])
	h = sum(g$rowsIn[row])

	labwidth = graphics::strwidth(label, units = "inches")
	labheight = graphics::strheight(label, units = "inches")


	scale = min(scale, {if (rot %in% c(0, 180)) {
		min(w / labwidth, h / labheight)
	} else {
		min(h / labwidth, w / labheight)
	}})
}

frc = function(row, col) paste0(sprintf("%02d", row), "_", sprintf("%02d", col))

impute_gp = function(gp, dt) {
	dtn = setdiff(names(dt), c("tmapID__", paste0("by", 1L:3L, "__")))

	cols = paste0("__", dtn)
	gp1 = sapply(gp, "[[", 1)
	gpids = which(gp1 %in% cols)
	#gp[gpids] = as.list(dt[, dtn, with = FALSE])

	for (i in gpids) gp[i] = as.list(dt[, dtn[match(gp1[i], cols)], with = FALSE])
	gp
}

rndrectGrob = function(...) {
	args = list(...)
	if ("r" %in% names(args) && args$r > 0) {
		args$r = grid::unit(args$r, "pt")
		do.call(grid::roundrectGrob, args)
	} else {
		args$r = NULL
		do.call(grid::rectGrob, args)
	}
}

# numeric vector (interpreted as inches) will be cast to a unit vector of inches with nulls on the head and tail. The values of the nulls depend on the justification
set_unit_with_stretch = function(x, ids = NULL, sides = c("both", "first", "second")) {
	u = rep("inch", length(x))
	if (!is.null(ids)) {
		x[ids] = 1/length(ids)
		u[ids] = "null"
		x = c(0, x, 0)
	} else {
		sides = match.arg(sides)
		x = if (sides == "both") c(0.5, x, 0.5) else if (sides == "first") c(1, x, 0) else c(0, x, 1)
	}
	u = c("null", u, "null")
	grid::unit(x, units = u)
}

# numeric vector for which spacers will be added between
unit_add_between = function(x, y) {
	n = length(x)
	if (n == 1L) return(x)
	x = rep(x, each = 2)[-(2L*n)]
	x[seq(2, (2*n)-1, by = 2)] = y
	x
}

# numeric vector for which head and tail will be added
unit_add_sides = function(x, y) {
	if (length(y) == 1L) {
		c(y, x, y)
	} else {
		c(y[1], x, y[2])
	}
}



# u is unit vector, tot = is the total size (width or height). The null units are distributed such that the total equals tot
distr_space_over_nulls = function(u, tot, stretchID = NA) {
	u0 = grid::unitType(u) == "null"

	un = grid::convertUnit(u, unitTo = "inch", valueOnly = TRUE)
	tn = as.numeric(tot)

	un_not0 = sum(un[!u0])
	un_dist_over0 = tn - un_not0

	normalize = function(x, s = 1) (x / sum(x)) * s

	if (un_dist_over0 < 0) {
		un[u0] = 0
		if (!is.na(stretchID)) un[stretchID] = max(0, un[stretchID] + un_dist_over0)
		un = normalize(un, s = tn)
	} else {
		un[u0] = normalize(as.numeric(u[u0]), un_dist_over0)
	}
	#print(un)
	grid::unit(un, units = "inch")
}

rescale_gp = function(gp, scale, skip = character()) {
	if ("lwd" %in% names(gp) && (!"lwd" %in% skip)) gp$lwd = gp$lwd * scale
	if ("size" %in% names(gp) && (!"size" %in% skip)) gp$size = gp$size * sqrt(scale)
	if ("cex" %in% names(gp) && (!"cex" %in% skip)) gp$cex = gp$cex * sqrt(scale)
	gp
}




appendGlist = function(glist, x) {
	glist = grid::gList(glist, x)
	names(glist)[length(glist)] = x$name
	glist
}

swap_pch_15_20 = function(gp) {
	# swap fill and col for pch 15-20
	if (any(gp$shape %in% 15:20)) {
		gp = make_equal_list(gp)
		pch15_20 = which(gp$shape %in% 15:20)
		fill = gp$col[pch15_20]
		fill_alpha = gp$col_alpha[pch15_20]
		gp$col[pch15_20] = gp$fill[pch15_20]
		gp$col_alpha[pch15_20] = gp$fill_alpha[pch15_20]
		gp$fill[pch15_20] = fill
		gp$fill_alpha[pch15_20] = fill_alpha
	}
	gp
}


# zero_one_to_hex = function(x) {
# 	# using indexing
# 	u = unique(x)
#
# 	x255 = round(u * 255)
#
# 	nc = c(0:9, LETTERS[1:6])
#
# 	y1 = (x255 %/% 16) + 1
# 	y2 = (x255 - (y1 - 1) * 16) + 1
#
# 	r = paste0(nc[y1], nc[y2])
# 	r[match(x, u)]
# }

# 255 to 2digit hex number
num_to_hex = function(x) {

	nc = c(0:9, LETTERS[1:6])

	y1 = (x %/% 16) + 1
	y2 = (x - (y1 - 1) * 16) + 1

	paste0(nc[y1], nc[y2])
}

hex_to_num = function(h) {
	nc = c(0:9, LETTERS[1:6])
	y1 = match(substr(h, 1, 1), nc)
	y2 = match(substr(h, 2, 2), nc)
	(y1 - 1) * 16 + (y2 - 1)
}

merge_alpha = function(dt, name) {
	name_a = paste0(name, "_alpha")
	f = function(d) {
		d1 = d[1,]
		col = d1[[name]]
		alpha = d1[[name_a]]

		if (nchar(col) == 9) {
			a = hex_to_num(substr(col, 8, 9)) * alpha
			cl = substr(col, 1, 7)
		} else {
			a = alpha * 255
			cl = col
		}
		ac = paste0(cl, num_to_hex(round(a)))
		rep(ac, nrow(d))
	}

	dt[, ca:=f(.SD), by = c(name, name_a), .SDcols = c(name, name_a)]
}





npc_to_native = function(x, scale) {
	x * (scale[2] - scale[1])# + scale[1]
}

blend_grobs = function(grb, blend, dst = NULL) {
	if (getRversion() < "4.2.0") {
		warning("blend requires R >= 4.2. Falling back to no blending.")
		return(grb)
	}
	caps <- grDevices::dev.capabilities()$compositing
	if (!is.null(caps) && !(blend %in% caps)) {
		warning("Compositing operator '", blend, "' is not supported by the current graphics device. ",
				"Falling back to no blending. Try png(type = \"cairo\") or svg().")
		return(grb)
	}
	grid::groupGrob(src = grb, op = blend, dst = dst, name = grb$name)
}

#' @export
tmapShape.Spatial = function(shp, is.main, crs, bbox, unit, filter, layer, shp_name, smeta, o, tmf) {
	tmapShape.sf(as(shp, "sf"), is.main, crs, bbox, unit, filter, shp_name, smeta, o, tmf)
}

#' @export
tmapSubsetShp.Spatial = function(shp, vars) {
	tmapSubsetShp.sf(as(shp, "sf"), vars)
}


#' @export
tmapGetShapeMeta1.Spatial = function(shp, layer, o) {
	tmapGetShapeMeta1.SpatRaster(as(shp, "sf"), o)
}

#' Map layer: text
#'
#' Map layer that draws symbols Supported visual variables are: `text`
#' (the text itself) `col` (color), `size` (font size), and `fontface` (font face).
#'
#' @param text,text.scale,text.legend,text.chart,text.free `r .doc_vv("text")`
#' @param size,size.scale,size.legend,size.chart,size.free `r .doc_vv("size")`
#' @param col,col.scale,col.legend,col.chart,col.free `r .doc_vv("col")`
#' @param col_alpha,col_alpha.scale,col_alpha.legend,col_alpha.chart,col_alpha.free `r .doc_vv("col_alpha")`
#' @param fontface,fontface.scale,fontface.legend,fontface.chart,fontface.free `r .doc_vv("fontface")`
#' @param fontfamily The font family. See [gpar()][grid::gpar()] for details.
#' @param bgcol,bgcol.scale,bgcol.legend,bgcol.chart,bgcol.free Visual variable that determines
#'   the background color. See Details.
#' @param bgcol_alpha,bgcol_alpha.scale,bgcol_alpha.legend,bgcol_alpha.chart,bgcol_alpha.free Visual variable that determines
#'   the background color transparency. See Details.
#' @param xmod,xmod.scale,xmod.legend,xmod.chart,xmod.free Transformation variable that determines the x offset. See details.
#' @param ymod,ymod.scale,ymod.legend,ymod.chart,ymod.free Transformation variable that determines the y offset. See details.
#'   the text. See details.
#' @param angle,angle.scale,angle.legend,angle.chart,angle.free Rotation angle
#' @param plot.order Specification in which order the spatial features are drawn.
#'   See [tm_plot_order()] for details.
#' @inherit tm_polygons details
#' @inheritParams tm_polygons
#' @seealso \href{https://r-tmap.github.io/tmap/articles/examples_terrain}{Terrain map example}
#' @param options options passed on to the corresponding `opt_<layer_function>` function
#' @param ... to catch deprecated arguments from version < 4.0
#' @example ./examples/tm_text.R
#' @seealso \href{https://r-tmap.github.io/tmap/articles/examples_topo_Africa}{Topographic map}
#' @export
tm_text = function(text = tm_const(),
				   text.scale = tm_scale(),
				   text.legend = tm_legend(),
				   text.chart = tm_chart_none(),
				   text.free = NA,
				   size = tm_const(),
				   size.scale = tm_scale(),
				   size.legend = tm_legend(),
				   size.chart = tm_chart_none(),
				   size.free = NA,
				   col = tm_const(),
				   col.scale = tm_scale(),
				   col.legend = tm_legend(),
				   col.chart = tm_chart_none(),
				   col.free = NA,
				   col_alpha = tm_const(),
				   col_alpha.scale = tm_scale(),
				   col_alpha.legend = tm_legend(),
				   col_alpha.chart = tm_chart_none(),
				   col_alpha.free = NA,
				   fontface = tm_const(),
				   fontface.scale = tm_scale(),
				   fontface.legend = tm_legend(),
				   fontface.chart = tm_chart_none(),
				   fontface.free = NA,
				   fontfamily = NA,
				   bgcol = tm_const(),
				   bgcol.scale = tm_scale(),
				   bgcol.legend = tm_legend(),
				   bgcol.chart = tm_chart_none(),
				   bgcol.free = NA,
				   bgcol_alpha = tm_const(),
				   bgcol_alpha.scale = tm_scale(),
				   bgcol_alpha.legend = tm_legend(),
				   bgcol_alpha.chart = tm_chart_none(),
				   bgcol_alpha.free = NA,
				   xmod = 0,
				   xmod.scale = tm_scale(),
				   xmod.legend = tm_legend_hide(),
				   xmod.chart = tm_chart_none(),
				   xmod.free = NA,
				   ymod = 0,
				   ymod.scale = tm_scale(),
				   ymod.legend = tm_legend_hide(),
				   ymod.chart = tm_chart_none(),
				   ymod.free = NA,
				   angle = 0,
				   angle.scale = tm_scale(),
				   angle.legend = tm_legend_hide(),
				   angle.chart = tm_chart_none(),
				   angle.free = NA,
				   plot.order = tm_plot_order("size", reverse = FALSE),
				   zindex = NA,
				   group = NA,
				   group.control = "check",
				   options = opt_tm_text(),
				   ...) {

	#if (FALSE) {
	args = list(...)

	layer_fun = if ("called_from" %in% names(args)) {
		args$called_from
	} else {
		"tm_text"
	}

	if (any(v3_only("tm_text") %in% names(args))) {
		v3_start_message()
		if (!("style" %in% names(args))) {
			if (!"breaks" %in% names(args)) {
				style = "pretty"
			} else {
				style = "fixed"
			}
		} else {
			style = args$style
		}

		v3_list_init()
		if (length(style) > 1) {
			style = style[1]
			.TMAP$v3_list$mult = TRUE
		}
		col.scale.args = c(list(n = v3_impute(args, "n", 5),
								style = style,
								style.args = v3_impute(args, "style.args", list())),
							if (style %in% c("cont", "log10")) {
								c({
									if (!is.null(args$breaks)) {
										if (length(args$breaks) > 2) {
											list(ticks = v3_impute(args, "breaks", NULL, "ticks"))
										} else {
											list(ticks = v3_impute(args, "breaks", NULL, "limits"))
										}
									} else {
										list()
									}
								},
								list(outliers.trunc = c(TRUE, FALSE)))
							} else {
								list(breaks = v3_impute(args, "breaks", NULL),
									 interval.closure = v3_impute(args, "interval.closure", "left"),
									 drop.levels = v3_impute(args, "drop.levels", FALSE))
							},
							list(midpoint = v3_impute(args, "midpoint", NULL),
								 as.count = v3_impute(args, "as.count", NA),
								 values = v3_impute(args, "palette", NA, "values"),
								 values.repeat = !v3_impute(args, "stretch.palette", TRUE, "values.repeat"),
								 values.range = v3_impute(args, "contrast", NA, "values.range"),
								 values.scale = 1,
								 value.na = v3_impute(args, "colorNA", NA, "value.na"),
								 value.null = v3_impute(args, "colorNULL", NA, "value.null"),
								 value.neutral = NA,
								 labels = v3_impute(args, "labels", NULL),
								 label.na = v3_impute(args, "textNA", "Missing", "label.na"),
								 label.null = NA,
								 label.format = v3_impute(args, "legend.format", list(), "label.format")))

		col.scale.args$fun_pref = if (style == "cat") {
			"categorical"
		} else if (style %in% c("fixed", "sd", "equal", "pretty", "quantile", "kmeans", "hclust", "bclust", "fisher", "jenks", "dpih", "headtails", "log10_pretty")) {
			"intervals"
		} else if (style == "cont") {
			"continuous"
		} else if (style == "log10") {
			"continuous_log10"
		} else {
			stop("unknown style")
		}

		if ("style" %in% names(args)) {
			v3_tm_scale_instead_of_style(style, scale_fun = col.scale.args$fun_pref, vv = "col", layer_fun = layer_fun, arg_list = v3_list_get())
		} else {
			v3_tm_scale(scale_fun = "", vv = "col", layer_fun = layer_fun, arg_list = v3_list_get())
		}

		col.scale = do.call("tm_scale", args = col.scale.args)

		v3_list_init()
		col.legend.args = alist(title = v3_impute(args, "title.col", NA, "title"),
								show = v3_impute(args, "legend.col.show", NULL, "show"),
								na.show = v3_impute(args, "showNA", NA, "na.show"),
								format = v3_impute(args, "legend.format", list(), "format"),
								orientation = ifelse(v3_impute(args, "legend.col.is.portrait", TRUE, "orientation"), "portrait", "landscape"),
								reverse = v3_impute(args, "legend.col.reverse", FALSE, "reverse"))

		v3_tm_legend(fun = layer_fun, vv = "col", arg_list = v3_list_get())
		col.legend = do.call("tm_legend", col.legend.args)

		v3_list_init()
		text.scale = tm_scale_asis(value.neutral = v3_impute(args, "sizes.legend.text", NA, "value.neutral"))
		v3_tm_scale(scale_fun = "asis", vv = "text", layer_fun = layer_fun, arg_list = v3_list_get())

		if ("legend.hist" %in% names(args) && args$legend.hist) {
			col.chart = tm_chart_histogram()
			v3_tm_chart_hist(layer_fun = layer_fun, vv = "col", arg = "legend.hist")
			# to do: histogram title
		}

		v3_list_init()
		size.scale.args = list(values = tm_seq(v3_impute(args, "size.lowerbound", 0, "values = tm_seq(<size.lowerbound>, 1)"), 1, power = 1/v3_impute(args, "root", 3, "values = tm_seq(<size.lowerbound>, 1, power = 1/<root>)")),
							   values.scale = v3_impute(args, "scale", 1, "values.scale"),
							   limits = v3_impute(args, "size.lim", NULL, "limits"),
							   outliers.trunc = c(v3_impute(args, "print.tiny", FALSE, paste0("outliers.trunc = c(", {if ("print.tiny" %in% names(args)) !args$print.tiny else TRUE}, ", TRUE)") ), TRUE),
							   ticks = v3_impute(args, "breaks", NULL, "ticks"),
							   midpoint = v3_impute(args, "midpoint", NULL),
							   labels = v3_impute(args, "sizes.legend.labels", NULL, "labels"))
		v3_tm_scale(scale_fun = "continuous", vv = "size", layer_fun = layer_fun, arg_list = v3_list_get())
		size.scale = do.call("tm_scale_continuous", size.scale.args)


		v3_list_init()
		size.legend.args = alist(title = v3_impute(args, "title.size", NA, "title"),
							   show = v3_impute(args, "legend.size.show", NULL, "show"),
							   na.show = v3_impute(args, "showNA", NA, "na.show"),
							   format = v3_impute(args, "legend.format", list(), "format"),
							   orientation = ifelse(v3_impute(args, "legend.size.is.portrait", TRUE, "orientation"), "portrait", "landscape"),
							   reverse = v3_impute(args, "legend.size.reverse", FALSE, "reverse"))
		v3_tm_legend(fun = "tm_text", vv = "size", arg_list = v3_list_get())
		size.legend = do.call("tm_legend", size.legend.args)



		v4_opt_args = c("along_lines", "shadow", "just", "clustering", "point.label", "remove_overlap")
		v3_opt_args = c("along_lines", "shadow", "just", "clustering", "auto.placement", "remove_overlap")
		osel = which(v3_opt_args %in% names(args))
		if (length(osel)) {
			o3 = v3_opt_args[osel]
			o4 = v4_opt_args[osel]
			v3_layer_opt(o3, o4, layer_fun)
		}


	#}
	}

	# unused arguments: typos?
	unused = setdiff(names(args), c(v3_only("tm_text"), "called_from"))

	if (length(unused)) {
		message_layer_unused_args(layer_fun, unused)
	}

	tm_element_list(tm_element(
		layer = "text",
		trans.fun = tmapTransCentroid,
		trans.aes = list(),
		trans.args = options$trans.args,
		trans.apply_to = "this",
		mapping.aes = list(
						   text = tmapScale(aes = "text",
						   				value = text,
						   				scale = text.scale,
						   				legend = text.legend,
						   				chart = text.chart,
						   				free = text.free),
						   size = tmapScale(aes = "size",
						   				value = size,
						   				scale = size.scale,
						   				legend = size.legend,
						   				chart = size.chart,
						   				free = size.free),
						   col = tmapScale(aes = "col",
						   				value = col,
						   				scale = col.scale,
						   				legend = col.legend,
						   				chart = col.chart,
						   				free = col.free),
						   col_alpha = tmapScale(aes = "col_alpha",
						   					  value = col_alpha,
						   					  scale = col_alpha.scale,
						   					  legend = col_alpha.legend,
						   					  chart = col_alpha.chart,
						   					  free = col_alpha.free),
						   bgcol = tmapScale(aes = "bgcol",
						   				value = bgcol,
						   				scale = bgcol.scale,
						   				legend = bgcol.legend,
						   				chart = bgcol.chart,
						   				free = bgcol.free),
						   bgcol_alpha = tmapScale(aes = "bgcol_alpha",
						   					  value = bgcol_alpha,
						   					  scale = bgcol_alpha.scale,
						   					  legend = bgcol_alpha.legend,
						   					  chart = bgcol_alpha.chart,
						   					  free = bgcol_alpha.free),
						   angle = tmapScale(aes = "angle",
						   						value = angle,
						   						scale = angle.scale,
						   						legend = angle.legend,
						   						chart = angle.chart,
						   						free = angle.free),
						   fontface = tmapScale(aes = "fontface",
						   				value = fontface,
						   				scale = fontface.scale,
						   				legend = fontface.legend,
						   				chart = fontface.chart,
						   				free = fontface.free),
						   xmod = tmapScale(aes = "xmod",
						   				 value = xmod,
						   				 scale = xmod.scale,
						   				 legend = xmod.legend,
						   				 chart = xmod.chart,
						   				 free = xmod.free),
						   ymod = tmapScale(aes = "ymod",
						   				 value = ymod,
						   				 scale = ymod.scale,
						   				 legend = ymod.legend,
						   				 chart = ymod.chart,
						   				 free = ymod.free)),

		gpar = tmapGpar(fill = NA,
						col = "__col",
						shape = NA,
						size = NA,
						cex = "__size",
						fill_alpha = NA,
						col_alpha = "__col_alpha",
						pattern = NA,
						text = "__text",
						fontface = "__fontface",
						fontfamily = fontfamily,
						lty = NA,
						lwd = NA,
						linejoin = NA,
						lineend = NA,
						bgcol = "__bgcol",
						bgcol_alpha = "__bgcol_alpha",
						xmod = "__xmod",
						ymod = "__ymod",
						angle = "__angle"),
		tpar = tmapTpar(),
		plot.order = plot.order,
		mapping.args = options$mapping.args,
		zindex = zindex,
		group = group,
		group.control = group.control,
		popup.vars = FALSE,
		popup.format = list(),
		hover = "",
		id = "",
		subclass = c("tm_aes_layer", "tm_layer")))
}


#' @rdname tm_text
#' @export
tm_labels = function(text = tm_const(),
					text.scale = tm_scale(),
					text.legend = tm_legend(),
					text.chart = tm_chart_none(),
					text.free = NA,
					size = tm_const(),
					size.scale = tm_scale(),
					size.legend = tm_legend(),
					size.chart = tm_chart_none(),
					size.free = NA,
					col = tm_const(),
					col.scale = tm_scale(),
					col.legend = tm_legend(),
					col.chart = tm_chart_none(),
					col.free = NA,
					col_alpha = tm_const(),
					col_alpha.scale = tm_scale(),
					col_alpha.legend = tm_legend(),
					col_alpha.chart = tm_chart_none(),
					col_alpha.free = NA,
					fontface = tm_const(),
					fontface.scale = tm_scale(),
					fontface.legend = tm_legend(),
					fontface.chart = tm_chart_none(),
					fontface.free = NA,
					fontfamily = "",
					bgcol = tm_const(),
					bgcol.scale = tm_scale(),
					bgcol.legend = tm_legend(),
					bgcol.chart = tm_chart_none(),
					bgcol.free = NA,
					bgcol_alpha = tm_const(),
					bgcol_alpha.scale = tm_scale(),
					bgcol_alpha.legend = tm_legend(),
					bgcol_alpha.chart = tm_chart_none(),
					bgcol_alpha.free = NA,
					xmod = 0,
					xmod.scale = tm_scale(),
					xmod.legend = tm_legend_hide(),
					xmod.chart = tm_chart_none(),
					xmod.free = NA,
					ymod = 0,
					ymod.scale = tm_scale(),
					ymod.legend = tm_legend_hide(),
					ymod.chart = tm_chart_none(),
					ymod.free = NA,
					angle = 0,
					angle.scale = tm_scale(),
					angle.legend = tm_legend_hide(),
					angle.chart = tm_chart_none(),
					angle.free = NA,
					plot.order = tm_plot_order("AREA", reverse = FALSE, na.order = "bottom"),
					zindex = NA,
					group = NA,
					group.control = "check",
					options = opt_tm_labels(),
					...) {
	args = c(as.list(environment()), list(...))
	tm = do.call(tm_text, args)
	tm[[1]]$layer = c("labels", "text")
	tm
}



#' @rdname tm_text
#' @export
tm_labels_highlighted = function(text = tm_const(),
								 text.scale = tm_scale(),
								 text.legend = tm_legend(),
								 text.chart = tm_chart_none(),
								 text.free = NA,
								 size = tm_const(),
								 size.scale = tm_scale(),
								 size.legend = tm_legend(),
								 size.chart = tm_chart_none(),
								 size.free = NA,
								 col = tm_const(),
								 col.scale = tm_scale(),
								 col.legend = tm_legend(),
								 col.chart = tm_chart_none(),
								 col.free = NA,
								 col_alpha = tm_const(),
								 col_alpha.scale = tm_scale(),
								 col_alpha.legend = tm_legend(),
								 col_alpha.chart = tm_chart_none(),
								 col_alpha.free = NA,
								 fontface = tm_const(),
								 fontface.scale = tm_scale(),
								 fontface.legend = tm_legend(),
								 fontface.chart = tm_chart_none(),
								 fontface.free = NA,
								 fontfamily = "",
								 bgcol = tm_const(),
								 bgcol.scale = tm_scale(),
								 bgcol.legend = tm_legend(),
								 bgcol.chart = tm_chart_none(),
								 bgcol.free = NA,
								 bgcol_alpha = tm_const(),
								 bgcol_alpha.scale = tm_scale(),
								 bgcol_alpha.legend = tm_legend(),
								 bgcol_alpha.chart = tm_chart_none(),
								 bgcol_alpha.free = NA,
								 xmod = 0,
								 xmod.scale = tm_scale(),
								 xmod.legend = tm_legend_hide(),
								 xmod.chart = tm_chart_none(),
								 xmod.free = NA,
								 ymod = 0,
								 ymod.scale = tm_scale(),
								 ymod.legend = tm_legend_hide(),
								 ymod.chart = tm_chart_none(),
								 ymod.free = NA,
								 angle = 0,
								 angle.scale = tm_scale(),
								 angle.legend = tm_legend_hide(),
								 angle.chart = tm_chart_none(),
								 angle.free = NA,
								 plot.order = tm_plot_order("AREA", reverse = FALSE, na.order = "bottom"),
								 zindex = NA,
								 group = NA,
								 group.control = "check",
								 options = opt_tm_labels(),
								 ...) {
	args = c(as.list(environment()), list(...))
	tm = do.call(tm_text, args)
	tm[[1]]$layer = c("labels_highlighted", "labels", "text")
	tm
}

#' @param points_only should only point geometries of the shape object (defined in [tm_shape()]) be plotted? By default `"ifany"`, which means `TRUE` in case a geometry collection is specified.
#' @param point_per specification of how spatial points are mapped when the geometry is a multi line or a multi polygon. One of \code{"feature"}, \code{"segment"} or \code{"largest"}. The first generates a spatial point for every feature, the second for every segment (i.e. subfeature), the third only for the largest segment (subfeature). Note that the last two options can be significant slower.
#' @param on_surface In case of polygons, centroids are computed. Should the points be on the surface? If `TRUE`, which is slower than the default `FALSE`, centroids outside the surface are replaced with points computed with [sf::st_point_on_surface()].
#' @param shadow Shadow behind the text. Logical.
#' @param shadow.col Color of the shadow.
#' @param shadow.offset.x,shadow.offset.y Shadow offset in line heights
#' @param halo Halo behind the text. In plot mode, it is just an outline, in view mode also a subtle glow.
#' @param halo.col Color of the halo.
#' @param halo.width Width (thickness) of the halo outline. In line heights
#' @param halo.blur Blur radius of the halo glow (view mode only). In line heights. Should be sufficiently larger than \code{halo.width} in order to see the effect.
#' @param halo.alpha Alpha transparency of the halo
#' @param just justification of the text relative to the point coordinates. Either one of the following values: \code{"left"} , \code{"right"}, \code{"center"}, \code{"bottom"}, and \code{"top"}, or a vector of two values where first value specifies horizontal and the second value vertical justification. Besides the mentioned values, also numeric values between 0 and 1 can be used. 0 means left justification for the first value and bottom justification for the second value. Note that in view mode, only one value is used.
#' @param along_lines logical that determines whether labels are rotated along the spatial lines. Only applicable if a spatial lines shape is used.
#' @param bg.padding The padding of the background in terms of line heights.
#' @param bg.border Should the background have borders?
#' @param bg.border.col Color of the borders
#' @param bg.border.lwd Line width of the borders
#' @param clustering in interactive modes (e.g. \code{"view"} mode), should clustering be applied at lower zoom levels? Either `FALSE` (default), `TRUE`, or a mode specific specification, e.g. for \code{"view"} mode \code{\link[leaflet:markerClusterOptions]{markerClusterOptions}}.
#' @param point.label logical that determines whether the labels are placed automatically. By default `FALSE` for `tm_text`, and `TRUE` for `tm_labels` if the number of labels is less than 500 (otherwise it will be too slow).
#' @param point.label.gap numeric that determines the gap between the point and label
#' @param point.label.method the optimization method, either `"SANN"` for simulated annealing (the default) or `"GA"` for a genetic algorithm.
#' @param remove_overlap logical that determines whether the overlapping labels are removed
#' @rdname tm_text
#' @export
opt_tm_text = function(points_only = "ifany",
					   point_per = "feature",
					   on_surface = FALSE,
					   shadow = FALSE,
					   shadow.col = NA,
					   shadow.offset.x = 0.05,
					   shadow.offset.y = 0.05,
					   halo = FALSE,
					   halo.col = NA,
					   halo.width = 0.05,
					   halo.blur = 0,
					   halo.alpha = 0.8,
					   just = "center",
					   along_lines = FALSE,
					   bg.padding = 0.4,
					   bg.border = FALSE,
					   bg.border.col = "black",
					   bg.border.lwd = 1,
					   clustering = FALSE,
					   point.label = FALSE,
					   point.label.gap = 0,
					   point.label.method = "SANN",
					   remove_overlap = FALSE) {
	if (!is.logical(shadow)) {
		cli::cli_warn("{.fun opt_tm_text} {.arg shadow} should be {.code TRUE} or {.code FALSE}. Please use shadow.col to specify a color")
		shadow.col = shadow
		shadow = TRUE
	}
	list(trans.args = list(points_only = points_only,
						   point_per = point_per,
						   on_surface = on_surface,
						   along_lines = along_lines),
		 mapping.args = list(shadow = shadow,
		 					shadow.col = shadow.col,
		 					shadow.offset.x = shadow.offset.x,
		 					shadow.offset.y = shadow.offset.y,
		 					halo = halo,
		 					halo.col = halo.col,
		 					halo.width = halo.width,
		 					halo.blur = halo.blur,
		 					halo.alpha = halo.alpha,
		 					just = just,
		 					along_lines = along_lines,
		 					bg.padding = bg.padding,
		 					bg.border = bg.border,
		 					bg.border.col = bg.border.col,
		 					bg.border.lwd = bg.border.lwd,
		 					clustering = clustering,
		 					point.label = point.label,
		 					point.label.gap = point.label.gap,
		 					point.label.method = point.label.method,
		 					remove_overlap = remove_overlap))
}

#' @rdname tm_text
#' @export
opt_tm_labels = function(points_only = "ifany",
						 point_per = "feature",
						 on_surface = FALSE,
						 shadow = FALSE,
						 shadow.col = NA,
						 shadow.offset.x = 0.05,
						 shadow.offset.y = 0.05,
						 halo = FALSE,
						 halo.col = NA,
						 halo.width = 0.05,
						 halo.blur = 0,
						 halo.alpha = 0.8,
						 just = "center",
						 along_lines = TRUE,
						 bg.padding = 0.4,
						 bg.border = FALSE,
						 bg.border.col = "black",
						 bg.border.lwd = 1,
						 clustering = FALSE,
						 point.label = NA,
						 point.label.gap = 0.4,
						 point.label.method = "SANN",
						 remove_overlap = FALSE) {
	if (!is.logical(shadow)) {
		cli::cli_warn("{.fun opt_tm_labels} {.arg shadow} should be {.code TRUE} or {.code FALSE}. Please use shadow.col to specify a color")
		shadow.col = shadow
		shadow = TRUE
	}

	list(trans.args = list(points_only = points_only,
						   point_per = point_per,
						   on_surface = on_surface,
						   along_lines = along_lines),
		 mapping.args = list(shadow = shadow,
		 					shadow.col = shadow.col,
		 					shadow.offset.x = shadow.offset.x,
		 					shadow.offset.y = shadow.offset.y,
		 					halo = halo,
		 					halo.col = halo.col,
		 					halo.width = halo.width,
		 					halo.blur = halo.blur,
		 					halo.alpha = halo.alpha,
		 					just = just,
		 					along_lines = along_lines,
		 					bg.padding = bg.padding,
		 					bg.border = bg.border,
		 					bg.border.col = bg.border.col,
		 					bg.border.lwd = bg.border.lwd,
		 					clustering = clustering,
		 					point.label = point.label,
		 					point.label.gap = point.label.gap,
		 					point.label.method = point.label.method,
		 					remove_overlap = remove_overlap))
}

scale_init = function() {
	assign("scls", list(), envir = .TMAP)
}

scale_save = function(scale) {
	if (!exists("scls", envir = .TMAP)) scale_init()
	scls = get("scls", envir = .TMAP)
	scls = c(scls, (list(scale)))
	assign("scls", scls, envir = .TMAP)
	length(scls)
}

scale_load = function(nr) {
	if (!exists("scls", envir = .TMAP)) return(NULL)
	scls = get("scls", envir = .TMAP)
	scls[[nr]]
}

#' Map component: inset maps and other objects
#'
#' Map component that adds an inset object, e.g. a mini map
#'
#' @param x object to draw. Can be: bounding box, tmap object, ggplot2 object, grob object, image file name.
#' @param height height of the component in number of text line heights.
#' @param width width of the component in number of text line heights.
#' @param margins margins
#' @param between_margin Margin between
#' @param box_frame Should a box frame be drawn in the main map that shows where the inset is? `TRUE` by default
#' @param box_frame.color,box_frame.alpha,box_frame.lwd,box_frame.lty Properties of the box frame
#' @param box_bg Should the frame box have a background? `FALSE` by default
#' @param box_bg.color,box_bg.alpha Properties of the box background
#' @param main_frame Should a frame be drawn around the inset map? Note that this is different from the general map component frame (the argument `frame`)
#' @param main_frame.r,main_frame.color,main_frame.alpha,main_frame.lwd Properties of the main frame
#' @inheritParams tm_title
#' @example ./examples/tm_inset.R
#' @export
tm_inset = function(x = NULL,
					height,
					width,
					margins,
					between_margin,
					position,
					group_id,
					frame, frame.color, frame.alpha, frame.lwd, frame.r, bg, bg.color, bg.alpha,
					box_frame, box_frame.color, box_frame.alpha, box_frame.lwd, box_frame.lty, box_bg, box_bg.color, box_bg.alpha,
					main_frame, main_frame.r, main_frame.color, main_frame.alpha, main_frame.lwd,
					z) {
	args = lapply(as.list(rlang::call_match()[-1]), eval, envir = parent.frame())
	args$called = names(args)
	args$group_id = args$group_id %||% NA_character_
	args$group_type = c("tm_inset")
	args$z = args$z %||% NA_integer_

	cls = if (is.null(x) || (inherits(x, "bbox"))) {
		"map"
	} else if (inherits(x, "tmap")) {
		"tmap"
	} else if (inherits(x, "ggplot")) {
		"gg"
	} else if (inherits(x, c("grob", "gList"))) {
		"grob"
	} else if (is.character(x)) {
		"image"
	} else {
		stop("Unsupported object")
	}

	cls2 = paste0("tm_inset_", cls)

	tm_element_list(do.call(tm_element, c(args, list(subclass = c(cls2, "tm_inset", "tm_component")))))
}

#' Internal tmap function to create by variables (used for faceting)
#'
#' Internal tmap function to create by variables (used for faceting)
#'
#' @param dt data.table
#' @param tmf tmf object
#' @param smeta smeta object
#' @export
#' @keywords internal
make_by_vars = function(dt, tmf, smeta) {
	by123 = paste0("by", 1L:3L)
	by123__ = paste0("by", 1L:3L, "__")
	with(tmf, {
		if (length(b)) {
			for (w in b) {
				byvar = by123[w]
				byname = by123__[w]
				var = tmf[[byvar]]

				if (var == "FRAME__") {
					dt[, (byname):= NA_integer_] # dummy
				} else if (var %in% smeta$vars) {
					levs = smeta$vars_levs[[var]]
					if (is.null(levs)) {
						cli::cli_abort("Variable {.val {var}} is used as facetting variable. Therefore, it should be a factor.")
					}
					if (attr(levs, "showNA")) levs[length(levs)] = NA

					dt[, (byname):= match(get(get(..byvar)), levs)]
				} else if (tmf[[byvar]] %in% smeta$dims) {
					dt[, (byname):= match(get(get(..byvar)), smeta$dims_vals[[var]])]
				}
			}
		}
	})
	invisible(NULL)
}



#' Internal tmap function that gets factor levels with NA's
#'
#' Internal tmap function that gets factor levels with NA's
#'
#' @param x vector
#' @param o options
#' @export
#' @keywords internal
get_fact_levels_na = function(x, o) {
	if (inherits(x, "sfc") || is.list(x)) {
		levs = NULL
	} else if (is.factor(x)) {
		if (o$drop.empty.facets) {
			tab = tabulate(x, nbins = nlevels(x))
			anyna = (sum(tab) != length(x)) # note that NA can already be included in the levels (in that case anyna = FALSE)
			levs = levels(x)[tab != 0]
		} else {
			anyna = anyNA(x)
			levs = levels(x)
		}

		if (!o$drop.NA.facets && anyna) {
			showNA = TRUE
			levs = c(levs, o$na.text)
		} else if (!o$drop.NA.facets && any(is.na(levs))) {
			showNA = TRUE
			levs[is.na(levs)] = o$na.text
		} else if (o$drop.NA.facets && any(is.na(levs))) {
			showNA = FALSE
			levs = levs[!is.na(levs)]
		} else {
			showNA = FALSE
		}
	} else {
		u = unique(as.vector(x))
		if (length(u) > o$facet.max) {
			levs = NULL
		} else {
			levs = as.character(sort(u))
			if (!o$drop.NA.facets && any(is.na(u))) {
				showNA = TRUE
				levs = c(levs, o$na.text)
			} else {
				showNA = FALSE
			}
		}
	}
	if (!is.null(levs)) attr(levs, "showNA") = showNA
	levs
}

#' Determine plotting order of features
#' 
#' Determine plotting order of features.
#' 
#' @param aes Visual variable for which the values determine the plotting order.
#'   Example: bubble map where the `"size"` aesthetic is used.
#'   A data variable (say population) is mapped via a continuous scale ([tm_scale_continuous()])
#'   to bubble sizes. The bubbles are plotted in order of size.
#'   How is determined by the other arguments. Use `"DATA"` to keep the same
#'   order as in the data. Another special value are `"AREA"` and `"LENGTH"`
#'   which are preserved for polygons and lines respectively: rather than a data
#'   variable the polygon area / line lengths determines the plotting order.
#' @param reverse Logical that determines whether the visual values are plotted
#'   in reversed order. The visual values (specified with tmap option `"values.var"`)
#'   are by default reversed, so plotted starting from the last value.
#'   In the bubble map example, this means that large bubbles are plotted first,
#'   hence at the bottom.
#' @param na.order Where should features be plotted that have an `NA` value for
#'   (at least) one other aesthetic variable? In the (order) `"mix"`, at the `"bottom"`,
#'   or on `"top"`? In the bubble map example: if fill color is missing for some bubble,
#'   where should those bubbles be plotted?
#' @param null.order Where should non-selected (aka null) features be plotted?
#' @param null.below.na Should null features be plotted below NA features?
#' @export
tm_plot_order = function(aes, reverse = TRUE, na.order = c("mix", "bottom", "top"), null.order = c("bottom", "mix", "top"), null.below.na = TRUE) {
	na.order = match.arg(na.order)
	null.order = match.arg(null.order)
	structure(list(aes = aes, reverse = reverse, na.order = na.order, null.order = null.order, null.below.na = null.below.na), class = "tm_plot_order")
}

# function adapted from car::pointLabel
# also used in tmap3
# changes are documented
# input: x, y, width and height of rectangles, bounding box (in 'sf' format')
pointLabel2 <- function (x, y, width, height, bbx,
						 method = c("SANN", "GA"),
						 gap = 0,
						 trace = FALSE, ...) {

	asp = tmaptools::get_asp_ratio(bbx)

	toUnityCoords <- function(xy) {
		if (asp > 1) {
			list(x = (xy$x - bbx[1])/(bbx[3] - bbx[1]) * asp,
				 y = (xy$y - bbx[2])/(bbx[4] - bbx[2]))
		} else {
			list(x = (xy$x - bbx[1])/(bbx[3] - bbx[1]),
				 y = (xy$y - bbx[2])/(bbx[4] - bbx[2])/asp)
		}


	}
	toUserCoords <- function(xy) {
		if (asp > 1) {
			list(x = bbx[1] + xy$x/asp * (bbx[3] - bbx[1]),
				 y = bbx[2] + xy$y * (bbx[4] - bbx[2]))
		} else {
			list(x = bbx[1] + xy$x * (bbx[3] - bbx[1]),
				 y = bbx[2] + xy$y * asp * (bbx[4] - bbx[2]))
		}

	}

	z <- grDevices::xy.coords(x, y, recycle = TRUE)
	z <- toUnityCoords(z)
	x <- z$x
	y <- z$y

	# CHANGED: width and height are specified by user
	width <- ((width + gap) / (bbx[3] - bbx[1])) * asp
	height <- ((height + gap) / (bbx[4] - bbx[2])) / asp


	method <- match.arg(method)
	n_labels <- length(x)

	gen_offset <- function(code) c(-1, -1, -1, 0, 0, 1, 1, 1)[code] *
		(width/2) + (0+1i) * c(-1, 0, 1, -1, 1, -1, 0, 1)[code] *
		(height/2)
	rect_intersect <- function(xy1, offset1, xy2, offset2) {
		w <- pmin(Re(xy1 + offset1/2), Re(xy2 + offset2/2)) -
			pmax(Re(xy1 - offset1/2), Re(xy2 - offset2/2))
		h <- pmin(Im(xy1 + offset1/2), Im(xy2 + offset2/2)) -
			pmax(Im(xy1 - offset1/2), Im(xy2 - offset2/2))
		w[w <= 0] <- 0
		h[h <= 0] <- 0
		w * h
	}

	objective <- function(gene) {
		offset <- gen_offset(gene)
		if (!is.null(rectidx1))
			area <- sum(rect_intersect(xy[rectidx1] + offset[rectidx1],
									   rectv[rectidx1], xy[rectidx2] + offset[rectidx2],
									   rectv[rectidx2]))
		else area <- 0
		n_outside <- sum(Re(xy + offset - rectv/2) < 0 | Re(xy +
																offset + rectv/2) > asp | Im(xy + offset - rectv/2) <
						 	0 | Im(xy + offset + rectv/2) > 1/asp)
		res <- 1000 * area + n_outside
		res
	}
	xy <- x + (0+1i) * y
	rectv <- width + (0+1i) * height
	rectidx1 <- rectidx2 <- array(0, (length(x)^2 - length(x))/2)
	k <- 0
	for (i in 1:length(x)) for (j in seq_len(i - 1)) {
		k <- k + 1
		rectidx1[k] <- i
		rectidx2[k] <- j
	}
	canIntersect <- rect_intersect(xy[rectidx1], 2 * rectv[rectidx1],
								   xy[rectidx2], 2 * rectv[rectidx2]) > 0
	rectidx1 <- rectidx1[canIntersect]
	rectidx2 <- rectidx2[canIntersect]
	if (trace)
		cat("possible intersects =", length(rectidx1), "\n")
	if (trace)
		cat("portion covered =", sum(rect_intersect(xy, rectv,
													xy, rectv)), "\n")
	GA <- function() {
		n_startgenes <- 1000
		n_bestgenes <- 30
		prob <- 0.2
		mutate <- function(gene) {
			offset <- gen_offset(gene)
			doesIntersect <- rect_intersect(xy[rectidx1] + offset[rectidx1],
											rectv[rectidx1], xy[rectidx2] + offset[rectidx2],
											rectv[rectidx2]) > 0
			for (i in which(doesIntersect)) {
				gene[rectidx1[i]] <- sample(1:8, 1)
			}
			for (i in seq(along = gene)) if (runif(1) <= prob)
				gene[i] <- sample(1:8, 1)
			gene
		}
		crossbreed <- function(g1, g2) ifelse(sample(c(0, 1),
													 length(g1), replace = TRUE) > 0.5, g1, g2)
		genes <- matrix(sample(1:8, n_labels * n_startgenes,
							   replace = TRUE), n_startgenes, n_labels)
		for (i in 1:10) {
			scores <- array(0, NROW(genes))
			for (j in 1:NROW(genes)) scores[j] <- objective(genes[j,
																  ])
			rankings <- order(scores)
			genes <- genes[rankings, ]
			bestgenes <- genes[1:n_bestgenes, ]
			bestscore <- scores[rankings][1]
			if (bestscore == 0) {
				if (trace)
					cat("overlap area =", bestscore, "\n")
				break
			}
			genes <- matrix(0, n_bestgenes^2, n_labels)
			for (j in 1:n_bestgenes) for (k in 1:n_bestgenes) genes[n_bestgenes *
																		(j - 1) + k, ] <- mutate(crossbreed(bestgenes[j,
																													  ], bestgenes[k, ]))
			genes <- rbind(bestgenes, genes)
			if (trace)
				cat("overlap area =", bestscore, "\n")
		}
		nx <- Re(xy + gen_offset(bestgenes[1, ]))
		ny <- Im(xy + gen_offset(bestgenes[1, ]))
		list(x = nx, y = ny)
	}
	SANN <- function() {
		gene <- rep(8, n_labels)
		score <- objective(gene)
		bestgene <- gene
		bestscore <- score
		T <- 2.5
		for (i in 1:50) {
			k <- 1
			for (j in 1:50) {
				newgene <- gene
				newgene[sample(1:n_labels, 1)] <- sample(1:8,
														 1)
				newscore <- objective(newgene)
				if (newscore <= score || runif(1) < exp((score -
														 newscore)/T)) {
					k <- k + 1
					score <- newscore
					gene <- newgene
				}
				if (score <= bestscore) {
					bestscore <- score
					bestgene <- gene
				}
				if (bestscore == 0 || k == 10)
					break
			}
			if (bestscore == 0)
				break
			if (trace)
				cat("overlap area =", bestscore, "\n")
			T <- 0.9 * T
		}
		if (trace)
			cat("overlap area =", bestscore, "\n")
		nx <- Re(xy + gen_offset(bestgene))
		ny <- Im(xy + gen_offset(bestgene))
		list(x = nx, y = ny)
	}
	if (method == "SANN")
		xy <- SANN()
	else xy <- GA()
	xy <- toUserCoords(xy)
	invisible(xy)
}

# NEED TO SYNC THIS WITH tmap_options
# generate all options with:
# args = setdiff(intersect(names(tmap_options_mode("view", mode.specific = FALSE)),names(tmap_options_mode("plot", mode.specific = FALSE))),  "name")
# cat(paste(args, collapse = ", "))
# manual: tm_layout: start with scale (the options before that are tm_options only)
#' Layout options
#'
#' Specify the layout of the maps. [tm_layout()] is identical as [tm_options()] but only contain the tmap options that are directly related to the layout. [tm_style()] sets the style for the map. A style is a specified set of options (that can be changed afterwards with [tm_layout()]). These functions are used within a plot call (stacked with the `+` operator). Their counterparts [tmap_options()] and [tmap_style()] can be used to set the (layout) options globally.
#'
#' @details
#' When [tm_style()] is called, all layout options are reset according to that
#' style. To override specific layout options for a chosen style, call
#' [tm_layout()] *after* [tm_style()].
#' @note In earlier versions of tmap, the `"natural"` style enabled the earth
#'   boundary automatically. This is no longer the case, because it did not work
#'   well for many projections. The earth boundary can still be enabled
#'   explicitly via the `earth_boundary` argument of [tm_layout()] (see the
#'   example above).
#' @seealso `r .doc_see_also_layout()`
#' @inheritParams tmap_options
#' @example ./examples/tm_layout.R
#' @export
tm_layout = function(
		scale, asp, legend.only, bg, bg.color, outer.bg, outer.bg.color, frame, frame.color, frame.alpha, frame.lwd, frame.r, frame.double_line, outer.margins, inner.margins, inner.margins.extra, meta.margins, meta.auto_margins, between_margin,
		text.fontfamily, text.fontface, r, attr.color,
		panel.margin,
		panel.type, panel.wrap.pos, panel.xtab.pos, color.sepia_intensity, color.saturation, color_vision_deficiency_sim,
		panel.show, panel.labels, panel.label.size, panel.label.color, panel.label.fontface, panel.label.fontfamily, panel.label.alpha, panel.label.bg, panel.label.bg.color, panel.label.bg.alpha, panel.label.frame, panel.label.frame.color, panel.label.frame.alpha, panel.label.frame.lwd, panel.label.frame.r, panel.label.height, panel.label.rot, earth_boundary, earth_boundary.color, earth_boundary.lwd, earth_datum, space, space.color,
		...
	) {

	args = lapply(as.list(rlang::call_match(dots_expand = TRUE)[-1]), eval, envir = parent.frame())
	args$called_from = "tm_layout"

	# 1106: need to use this for other texts as well
	if ("panel.labels" %in% names(args)) {
		if (is.list(args$panel.labels)) {
			args$panel.labels = lapply(args$panel.labels, encode_expr)
		} else {
			args$panel.labels = encode_expr(args$panel.labels)
		}
	}

	do.call(tm_options, args)
}

#' View mode options
#'
#' View mode options. These options are specific to the view mode.
#'
#' @param use_browser If `TRUE` it opens an external browser, and `FALSE` (default) it opens the internal IDEs (e.g. RStudio) browser.
#' @param use_WebGL use webGL for points, lines, and polygons. For large spatial objects, this is much faster than the standard leaflet layer functions. However, it can not always be used for two reasons. First, the number of visual variables are limited; only fill, size, and color (for lines) are supported. Second, projected CRS's are not supported. Furthermore, it has the drawback that polygon borders are not as sharp. By default only `TRUE` for large spatial objects (1000 or more features) when the mentioned criteria are met.
#'  By default `TRUE` if no other visual variables are used.
#' @param control.position `r .doc_opt("control.position")`
#' @param control.bases base layers
#' @param control.overlays overlay layers
#' @param control.collapse Should the box be collapsed or expanded?
#' @param set_bounds logical that determines whether maximum bounds are set,
#'   or a bounding box. Not applicable in plot mode.
#'   In view mode, this is passed on to [setMaxBounds()][leaflet::setMaxBounds()]
#' @param set_view numeric vector that determines the view.
#'   Either a vector of three: `lng`, `lat`, and `zoom`, or a single value:
#'    `zoom`. See [setView()][leaflet::setView()].
#'    Only applicable if `bbox` is not specified
#' @param set_zoom_limits numeric vector of two that set the minimum and maximum
#'   zoom levels (see [tileOptions()][leaflet::tileOptions()]).
#' @param use_circle_markers If `TRUE` (default) circle shaped symbols (e.g. `tm_dots()` and `tm_symbols()`) will be rendered as [addCircleMarkers()][leaflet::addCircleMarkers()] instead of [addMarkers()][leaflet::addMarkers()]. The former is faster, the latter can support any symbol since it is based on icons
#' @param leaflet.options options passed on to
#'   [leafletOptions()][leaflet::leafletOptions()]
#' @param ... to catch deprecated arguments
#' @export
#' @seealso `r .doc_see_also_modes()` `r .doc_see_also_groups()`
tm_view = function(use_browser,
				   use_WebGL,
				   control.position,
				   control.bases,
				   control.overlays,
				   control.collapse,
				   set_bounds,
				   set_view,
				   set_zoom_limits,
				   use_circle_markers,
				   leaflet.options,
				   ...) {

	args = lapply(as.list(rlang::call_match(dots_expand = TRUE)[-1]), eval, envir = parent.frame())
	args$called_from = "tm_view"
	do.call(tm_options, args)

}

#' Plot mode options
#'
#' Plot mode options. This option is specific to the plot mode.
#'
#' @param use_gradient Use gradient fill using [linearGradient()][grid::linearGradient()]
#' @param limit_latitude_3857 Vector of two limit latitude values for maps printed in Web Mercator projection (EPSG 3857). If `c(-90, 90)` the poles will be inflated too much. The Web Mercator is defines as `c(-85.06, 85.06)`, but the default setting in tmap is `c(-84, 84)`.
#' @seealso `r .doc_see_also_modes()`
#' @export
tm_plot = function(use_gradient, limit_latitude_3857) {
	args = lapply(as.list(rlang::call_match()[-1]), eval, envir = parent.frame())
	do.call(tm_options, args)
}


#' @rdname tm_extra_innner_margin
#' @export
tm_place_legends_right = function(width = NA) {
	if (is.na(width)) {
		tm_options(legend.position = tm_pos_out("right", "center"))
	} else {
		tm_options(meta.margins = c(0, 0, 0, width), legend.position = tm_pos_out("right", "center"))
	}
}

#' @rdname tm_extra_innner_margin
#' @param width width
#' @export
tm_place_legends_left = function(width = NA) {
	if (is.na(width)) {
		tm_options(legend.position = tm_pos_out("left", "center"))
	} else {
		tm_options(meta.margins = c(0, 0, 0, width), legend.position = tm_pos_out("right", "center"))
	}
}

#' @rdname tm_extra_innner_margin
#' @param height height
#' @export
tm_place_legends_bottom = function(height = NA) {
	if (is.na(height)) {
		tm_options(legend.position = tm_pos_out("center", "bottom"))
	} else {
		tm_options(meta.margins = c(height, 0, 0, 0), legend.position = tm_pos_out("center", "bottom"))
	}
}

#' @rdname tm_extra_innner_margin
#' @export
tm_place_legends_top = function(height = NA) {
	if (is.na(height)) {
		tm_options(legend.position = tm_pos_out("center", "top"))
	} else {
		tm_options(meta.margins = c(height, 0, 0, 0), legend.position = tm_pos_out("center", "top"))
	}
}

#' @rdname tm_extra_innner_margin
#' @param pos.h,pos.v position (horizontal and vertical)
#' @export
tm_place_legends_inside = function(pos.h = NULL, pos.v = NULL) {
	if (is.null(pos.h) || is.null(pos.v)) {
		if (!is.null(pos.h)) {
			warning("tm_place_legends_inside also requires pos.v", call. = FALSE)
		} else if (!is.null(pos.v)) {
			warning("tm_place_legends_inside also requires pos.h", call. = FALSE)
		}
		tm_options(legend.position = tm_pos_auto_in())
	} else {
		tm_options(legend.position = tm_pos_in(pos.h = pos.h, pos.v = pos.v))
	}
}

#' tmap layout: helper functions
#'
#' @param left,right,top,bottom extra margins
#' @export
#' @name tm_extra_innner_margin
tm_extra_inner_margin = function(left = 0, right = 0, top = 0, bottom = 0) {
	tm_options(inner.margins.extra = c(bottom, left, top, right))
}

#' @rdname tmap_options
#' @export
tm_check_fix = function() {
	tm_options(check_and_fix = TRUE)
}


#' @rdname tm_layout
#' @param style name of the style
#' @export
#' @order 1
tm_style = function(style, ...) {
	args = list(...)

	check_style(style)

	args$style = style
	args$called_from = "tm_style"
	do.call(tm_options, args)
}

crs_reproject_shpTM = function(s, crs, raster.warp) {
	# crs can be a list per class (due to leaflet EPSG:3857 requirement for raster images)

	cls = class(s$shp)
	if (cls[1] != "character") {
		crs2 = sf::st_crs(get_option_class(crs, class = class(s$shp)))
		if (sf::st_crs(s$shp) != crs2) {
			s = do.call(tmapReproject, c(s, list(crs = crs2, raster.warp = raster.warp)))
		}	}

	s
}

trans_sf = function(sf1, sf2, nframes, ease) {
	df1 = data.frame(geo = sf::st_geometry(sf1))
	df2 = data.frame(geo = sf::st_geometry(sf2))
	x = transformr::tween_sf(df1, df2, ease = ease, nframes = nframes)
	sf::st_sf(.frame = x$.frame, geometry = x$geometry, crs = sf::st_crs(sf1))
}

trans_shpTM = function(shpTM1, shpDT, nframes, ease) {
	rlang::check_installed(c("transformr", "gifski"))
	shpTM2 = shpDT$shpTM[[1 ]]
	s = trans_sf(shpTM1$shp, shpTM2$shp[match(shpTM1$tmapID, shpTM2$tmapID)], nframes, ease)

	data.table::rbindlist(lapply(1:nframes, function(i) {
		x = shpDT
		x$shpTM[[1]]$shp = s$geometry[s$.frame == i]
		x$shpTM[[1]]$tmapID = shpTM1$tmapID
		x$by3__ = i
		x
	}))

	# data.table(shpTM = lapply(1:nframes, function(i) {
	# 	shpTM1$shp = s$geometry[s$.frame == i]
	# 	shpTM1
	# }), by3__ = 1:nframes)
}

step3_trans = function(tm) {
	ad = tm$tmo
	o = tm$o
	aux = tm$aux
	cmp = tm$cmp
	prx = tm$prx

	bd = lapply(ad, function(adi) {
		shpDT = adi$shpDT


		trans_shp = function(al, shpDT) {
			transDT = al$trans_dt

			plot.order = al$plot.order
			bycols = names(transDT)[substr(names(transDT), 1, 2) == "by"]
			sdcols = setdiff(names(transDT), "frame")

			if (length(transDT)) {
				y = transDT[, .(shp = do.call(do_trans, list(tdt = .SD, FUN = al$trans_fun, shpDT = shpDT, plot.order = plot.order, args = al$trans_args, scale = o$scale))), by = bycols, .SDcols = sdcols]

				if ("by3__" %in% bycols && all(is.na(y$by3__))) {
					shpDT = data.table::rbindlist(lapply(y$shp, function(yi) {
						trans_shpTM(shpDT$shpTM[[1]], yi, nframes = o$nframes, ease = "cubic-in-out")
					}))


				} else {
					shpDT = data.table(rbindlist(y$shp))
				}
			} else {
				shpDT$shpTM = lapply(shpDT$shpTM, function(s) do.call(al$trans_fun, list(shpTM = s, plot.order = plot.order, args = al$trans_args)))
			}
			shpDT
		}

		crs_step4 = o$crs_step4
		crs_step3 = o$crs_step3



		# step 3.a : reproject crs to main_crs
		shpDT$shpTM = lapply(shpDT$shpTM, crs_reproject_shpTM, crs = crs_step3, raster.warp = o$raster.warp)

		# step 3.b: apply all global transformation functions
		appl = vapply(adi$layers, "[[", "trans_apply", FUN.VALUE = character(1))

		is_global = which(appl == "all")

		if (length(is_global) > 1) {
			stop("Number of 'global transformation' layers is larger than 1", call. = FALSE)
		} else if (length(is_global) == 1) {
			shpDT = trans_shp(adi$layers[[is_global]], shpDT)
		}

		e = environment()

		shpDT_along = shpDT

		adi$layers = lapply(adi$layers, function(al) {
			if (al$trans_apply == "all") {
				al$shpDT = shpDT
			} else {
				al$shpDT = trans_shp(al, shpDT_along)
			}
			if (al$trans_apply == "this_following") {
				assign("shpDT_along", al$shpDT, envir = e)
			}

			# step 3.c2 : reproject crs to crs (for plotting)
			al$shpDT$shpTM = lapply(al$shpDT$shpTM, crs_reproject_shpTM, crs = crs_step4, raster.warp = o$raster.warp)

			al[c("trans_dt", "trans_args", "trans_apply", "tp")] = NULL
			al
		})
		if (!length(adi$layers)) {
			adi$bbx = tmaptools::bb(shpDT$shpTM[[1]]$shp)
		}
		adi$shpDT = NULL

		adi
	})

	list(tmo = bd, aux = aux, cmp = cmp, prx = prx, o = o)
}

#' tmap function to specify variables
#'
#' tmap function to specify all variables in the shape object
#'
#' @param x variable names, variable indices, or a dimension name
#' @param dimvalues dimension values
#' @param n if specified the first `n` variables are taken (or the first `n` dimension values)
#' @param multivariate in case multiple variables are specified, should they serve as facets (FALSE) or as a multivariate visual variable?
#' @param animate should the variable(s) be animated? (experimental)
#' @seealso `r .doc_see_also_scales()`
#' @export
tm_vars = function(x = NA, dimvalues = NULL, n = NA, multivariate = FALSE, animate = FALSE) {
	structure(list(x = x, dimvalues = dimvalues, n = n, multivariate = multivariate, animate = animate), class = c("tmapVars", "list"))
}


# process visual variable specification. Can either be tmapVars (output of tm_vars) or a list of values.
tmapVV = function(x) {
	if (is.null(x)) {
		x = structure(list("value.blank"), class = "tmapOption")

	} else if (!is.list(x) && length(x) == 1 && is.na(x)) {
		x = structure(list("value.const"), class = "tmapOption")
	}

	if (inherits(x, c("tmapOption", "tmapVars"))) return(x)

	# if (inherits(x, "tm_shape_vars")) return(structure(list(ids = x$ids, n = x$n), class = "tmapShpVars"))
	# if (inherits(x, "tm_mv_shape_vars")) return(structure(list(ids = x$ids, n = x$n), class = "tmapMVShpVars"))
	# if (inherits(x, "tmapDimVars")) return(x)

	cls = if (inherits(x, "AsIs")) "tmapAsIs" else if (inherits(x, "tmapUsrCls")) "tmapUsrCls" else "tbd"


	isL = is.list(x)

	if (cls == "tbd" && !isL) {
		if (is.character(x) && all(substr(x, 1, 1) == "*")) {
			x = substr(x, 2, nchar(x))
			return(tm_vars(x, animate = TRUE))
		}
	}

	isNestedL = isL && any(vapply(x, is.list, FUN.VALUE = logical(1)))
	isSpecialL = isL && !setequal(class(x), "list")
	isSpecialNestedL = isL && is.list(x[[1]]) &&  !setequal(class(x[[1]]), "list")
	if (!isL) {
		x = as.list(x)
	} else if (isSpecialL) {
		x = list(x)
	}

	if (cls == "tbd") cls = if (isSpecialL || isSpecialNestedL || isNestedL) "tmapSpecial" else "tmapStandard"

	if (cls == "tmapSpecial") {
		nms = seq_along(x)
	} else {
		nms = x
	}

	structure(x, names = nms, class = cls)
}

#' @export
#' @rdname tmapGridLeaflet
tmapLeafletDataPlot.tm_data_lines = function(a, shpTM, dt, pdt, popup.format, hdt, idt, gp, bbx, facet_row, facet_col, facet_page, id, pane, group, glid, o, ...) {
	lf = get_lf(facet_row, facet_col, facet_page)

	rc_text = frc(facet_row, facet_col)

	res = select_sf(shpTM, dt[!is.na(dt$lwd), ])
	shp = res$shp
	if (o$crs_leaflet$crsClass  == "L.CRS.Simple") {
		shp = sf::st_set_crs(shp, NA)
	}

	names(shp) = NULL # also required for other layers?
	dt = res$dt

	if (!is.null(idt)) {
		idt = idt$id[match(dt$tmapID__, idt$tmapID__)]
	}
	if (!is.null(hdt)) {
		hdt = hdt$hover[match(dt$tmapID__, hdt$tmapID__)]
		hdt = lapply(hdt, htmltools::HTML, FUN.VALUE = character(1))
	}

	if (is.null(pdt)) {
		popups = NULL
	} else {
		mtch = match(dt$tmapID__, pdt$tmapID__)
		pdt = pdt[mtch][, tmapID__ := NULL]


		popups = view_format_popups(id = idt, titles = names(pdt), values = pdt, format = popup.format)
	}

	gp = impute_gp(gp, dt)
	gp = rescale_gp(gp, o$scale_down)

	interactive = (!is.null(pdt) || !is.null(hdt))


	# process hitbox option:
	hitbox = if (a$hitbox == "auto") {
		thin = median(gp$lwd) < 4
		light = length(shp) < 10000
		if (thin && light) "pmax8" else "none"
	} else {
		a$hitbox
	}

	o$use_WebGL = impute_webgl(o$use_WebGL, dt, supported = "col", checkif = list(lty = "solid"), type = "lines", hover = !is.null(hdt), popup = !is.null(pdt), crs_class = o$crs_leaflet$crsClass)


	k = nrow(dt)
	if (hitbox == "none") {
		opt1 = leaflet::pathOptions(interactive = interactive, pane = pane, lineCap = gp$lineend, lineJoin = gp$linejoin)

	} else {
		opt1 = leaflet::pathOptions(interactive = FALSE, pane = pane, lineCap = gp$lineend, lineJoin = gp$linejoin)
		opt2 = leaflet::pathOptions(interactive = interactive, pane = pane, lineCap = gp$lineend, lineJoin = gp$linejoin)

		hb = parse_hitbox(hitbox)
	}


	if (o$use_WebGL) {
		shp2 = sf::st_sf(id = seq_along(shp), geom = shp)
		shp3 = suppressWarnings(sf::st_cast(shp2, "LINESTRING"))
		gp3 = lapply(gp, function(gpi) {if (length(gpi) == 1) gpi else gpi[shp3$id]})
		popups = popups[shp3$id]

		if (hitbox == "none") {

			idt = {if (is.null(idt)) dt$tmapID__[1] else idt[1]} |>
				submit_labels("linesGL", pane, group)

			lf %>%
				leafgl::addGlPolylines(
					data = shp3,
					color = gp3$col,
					opacity = gp3$col_alpha[1],
					weight = lwd_to_leafgl(gp3$lwd[1]),
					pane = pane,
					group = group,
					layerId = idt[1],
					label = hdt,
					popup= popups) %>%
				assign_lf(facet_row, facet_col, facet_page)
		} else {
			idt = {if (is.null(idt)) dt$tmapID__[c(1, 1:k)] else idt[c(1, 1:k)]} |>
				submit_labels("linesGL_hb", pane, group)

			weights = gp3$lwd + hb$plus
			if (hb$pmax != 0) weights = pmax(weights, hb$pmax)

			lf %>%
				# Fast WebGL rendering
				leafgl::addGlPolylines(
					data    = shp3,
					color   = gp3$col,
					opacity = gp3$col_alpha[1],
					weight  = lwd_to_leafgl(gp3$lwd[1]),
					pane    = pane,
					group   = group,
					popup = NULL,
					layerId = idt[1]
				) %>%

				# Invisible interaction layer
				leaflet::addPolylines(
					data    = shp3,
					opacity = 0,
					weight  = weights,
					group   = group,
					layerId = idt[-1],
					label   = hdt,
					options = opt2,
					popup   = popups
				) %>%

				assign_lf(facet_row, facet_col, facet_page)
		}

	} else {
		if (hitbox == "none") {
			idt = if (is.null(idt)) dt$tmapID__ else idt |>
				submit_labels("lines", pane, group)

			lf %>%
				leaflet::addPolylines(
					data = shp,
					layerId = idt,
					label = hdt,
					color = gp$col,
					opacity = gp$col_alpha,
					weight = gp$lwd,
					group = group,
					options = opt1,
					dashArray = lty2dash(gp$lty, gp$lwd),
					popup = popups) %>%
				assign_lf(facet_row, facet_col, facet_page)
		} else {
			idt = rep({if (is.null(idt)) dt$tmapID__ else idt}, 2) |>
				submit_labels("lines", pane, group)


			weights = gp$lwd + hb$plus
			if (hb$pmax != 0) weights = pmax(weights, hb$pmax)

			lf %>%
				# Visible styled layer (no popup here)
				leaflet::addPolylines(
					data      = shp,
					layerId = idt[1:k],
					color     = gp$col,
					opacity   = gp$col_alpha,
					weight    = gp$lwd,
					group     = group,
					options   = opt1,
					dashArray = lty2dash(gp$lty, gp$lwd)
				) %>%

				# Invisible interaction layer (fatter hitbox)
				leaflet::addPolylines(
					data    = shp,
					layerId = idt[(k+1):(2*k)],
					label   = hdt,
					popup   = popups,
					options = opt2,
					opacity = 0,
					weight  = weights,
					group   = group
				) %>%

				assign_lf(facet_row, facet_col, facet_page)

		}


		# lf %>%
		# 	leaflet::addPolylines(data = shp, layerId = idt, label = hdt, color = gp$col, opacity = gp$col_alpha, weight = gp$lwd, group = group, options = opt, dashArray = lty2dash(gp$lty, gp$lwd), popup = popups) %>%
		# 	assign_lf(facet_row, facet_col, facet_page)

	}
	NULL
}

#' @export
#' @rdname tmapGridLeaflet
tmapLeafletDataPlot.tm_data_iso = function(a, shpTM, dt, pdt, popup.format, hdt, idt, gp, bbx, facet_row, facet_col, facet_page, id, pane, group, glid, o, ...) {
	# isn't called, but needed to make tm_iso visible in tmap_overview
	NextMethod()
}

parse_hitbox <- function(hitbox) {

	# defaults
	plus  <- 0
	pmax_ <- 0

	if (is.null(hitbox) || hitbox == "none") {
		return(list(plus = 0, pmax = 0))
	}

	# extract plus
	m_plus <- regexpr("plus[0-9]+", hitbox)
	if (m_plus > 0) {
		plus <- as.numeric(sub("plus", "", regmatches(hitbox, m_plus)))
	}

	# extract pmax
	m_pmax <- regexpr("pmax[0-9]+", hitbox)
	if (m_pmax > 0) {
		pmax_ <- as.numeric(sub("pmax", "", regmatches(hitbox, m_pmax)))
	}

	list(plus = plus, pmax = pmax_)
}

lwd_to_leafgl <- function(lwd) {
	# empirical formula, based on observations: 1 (very small), 2 (2/8), 3 (3/6), 4 (4/4), 6 (6/3.5), 8 (8/3), 16 (16/8)
	ifelse(lwd <= 4,
		   lwd / 2^(4 - log2(lwd)),
	ifelse(lwd > 8,
		   lwd / 3,
		   lwd / (6 - log2(lwd))
		   ))
}

submit_labels = function(labels, cls, pane, group) {
	layerIds = get("layerIds", envir = .TMAP_LEAFLET)
# 	print("** voor **")
# po(layerIds)
	if (length(layerIds)) {
		labels = local({
			labels_all = unlist(lapply(layerIds, function(l) l$Lid), use.names = FALSE)
			pos <- length(labels_all)
			labels_all = gsub("_", ".", labels_all,fixed = TRUE)
			labels_all = make.names(c(labels_all, labels), unique = TRUE)
			labels_all = gsub(".", "_", labels_all,fixed = TRUE)
			labels_all[(pos + 1): length(labels_all)]
		})
	} else {
		labels = make.names(labels, unique = TRUE)
		labels = gsub(".", "_", labels,fixed = TRUE)
	}

	layerIds = c(layerIds, list(list(name = pane, type = cls, group = group, Lid = labels, stamp = .TMAP$stamp)))
	# print("-- na --")
	# po(layerIds)
	assign("layerIds", layerIds, envir = .TMAP_LEAFLET)
	labels
}

impute_webgl = function(use_WebGL, dt, supported, checkif = NULL, type, hover, popup, crs_class) {
	if (!identical(use_WebGL, FALSE)) {
		if (crs_class == "L.CRS.Simple") {
			message_webgl_crs_simple()
			use_WebGL = FALSE
		} else {

			vary = vapply(dt, function(x) {
				any(is.na(x)) || any(x!=x[1])
			}, FUN.VALUE = logical(1))

			vary = vary[setdiff(names(vary), c(supported, "tmapID__", "ord__"))]

			if (!is.null(checkif)) {
				checks = vapply(seq_along(checkif), function(i) {
					if (!(checkif[i] %in% names(dt))) {
						TRUE
					} else {
						dt[[names(checkif)[i]]][1] %in% checkif[[i]]
					}
				}, FUN.VALUE = logical(1))
			} else {
				checks = TRUE
			}

			if (any(vary)) {
				if (is.na(use_WebGL)) {
					use_WebGL = FALSE
				} else {
					message_webgl_vars(supported, vary)
				}
			} else if (!all(checks)) {
				if (is.na(use_WebGL)) {
					use_WebGL = FALSE
				} else {
					checkif = lapply(checkif, function(x) {
						if (is.character(x)) paste0("\"", x, "\"") else x
					})

					checkif = lapply(checkif, function(x) {
						if (length(x) == 1) {
							x
						} else {
							k = length(x)
							paste(paste(head(x, -1), collapse = ", "), tail(x, 1), sep = " or ")
						}
					})

					message_webgl_checks(checks, checkif)

				}
			} else if (type != "symbols" && hover) {
				if (!is.na(use_WebGL)) {
					message_webgl_hover(type)
				} else {
					use_WebGL = FALSE
				}
			} else if ((is.na(use_WebGL))) {
				n = nrow(dt)
				use_WebGL = (n >= 1000)
			}
		}
	}
	use_WebGL
}


expand_coords_gp = function(coords, gp, ndt) {
	expanded = (ncol(coords) == 3L)
	if  (expanded) {
		gp = lapply(gp, function(gpi) {
			if (is.list(gpi)) {
				unlist(gpi)
			} else if (length(gpi) == ndt) {
				gpi[coords[,3L]]
			} else {
				gpi
			}
		})
		coords = coords[, 1:2, drop=FALSE]
	}
	list(coords = coords, gp = gp, expanded = expanded)
}

#' @export
#' @keywords internal
#' @rdname tmapGridLeaflet
tmapLeafletDataPlot = function(a, shpTM, dt, pdt, popup.format, hdt, idt, gp, bbx, facet_row, facet_col, facet_page, id, pane, group, glid, o, ...) {
	UseMethod("tmapLeafletDataPlot")
}

#' @export
#' @keywords internal
#' @rdname tmapGridLeaflet
tmapLeafletDataPlot.default = function(a, shpTM, dt, pdt, popup.format, hdt, idt, gp, bbx, facet_row, facet_col, facet_page, id, pane, group, glid, o, ...) {
	NULL
}

#	preset = lapply(list(dashed = c(4, 4), dotted = c(1,3), dotdash = c(1, 3, 4, 3), longdash = c(7, 3), twodash = c(2, 2, 6, 2)), function(x) {
#		x * scale
#	})
lty2dash = function(lty, lwd) {
	k = max(length(lty), length(lwd))
	lty = rep(lty, length.out = k)
	lwd = rep(lwd, length.out = k)



	tab = c(solid = "", dashed = "4 4", dotted = "1 3", dotdash = "1 3 4 3", longdash = "7 3", twodash = "2 2 6 2")
	are_words = (lty %in% names(tab))
	x = if (all(are_words)) {
		unname(tab[lty])
	} else {
		are_letters = (suppressWarnings(!is.na(as.numeric(lty))))

		if (!all(are_letters | are_words)) {
			stop("Incorrect lty specification: ", lty[which(!are_letters & !are_words)[1]])
		} else {
			lty[are_words] = unname(tab[lty[are_words]])
			lty[are_letters] = vapply(strsplit(lty[are_letters], ""), FUN = function(x) paste(x, collapse = " "), FUN.VALUE = character(1))
		}
		lty
	}

	# multiply by lwd
	split_x <- strsplit(x, " ")

	z = mapply(
		function(vals, mult) paste(as.numeric(vals) * mult, collapse = " "),
		split_x, lwd, USE.NAMES = FALSE
	)

	z[x == ""] = ""   # restore empties if needed
	z
}

makeSymbolIcons2  = function (shape, color, fillColor = color, opacity, fillOpacity = opacity,
		  strokeWidth = 1, dashArray = "none", width, height = width, ...)
{
	symbols <- Map(leaflegend::makeSymbol, shape = shape, width = width,
				   height = height, color = color, fillColor = fillColor,
				   opacity = opacity, fillOpacity = fillOpacity, `stroke-width` = strokeWidth,
				   `stroke-dasharray` = dashArray,
				   ...)
	leaflet::icons(iconUrl = unname(symbols), iconAnchorX = width/2 + strokeWidth,
				   iconAnchorY = height/2 + strokeWidth)
}


split_alpha_channel <- function(x, alpha) {
	if (is.null(x)) {
		list(col=NULL, opacity=0)
	} else {
		RGBA <- col2rgb(x, alpha = TRUE)
		col <- rgb(RGBA[1,], RGBA[2,], RGBA[3,], maxColorValue = 255)
		opacity <- unname(RGBA[4,]/255 * alpha)
		list(col=col, opacity=opacity)
	}
}

#' Map layer: text
#'
#' Map layer that draws symbols Supported visual variables are: `text`
#' (the text itself) `col` (color), `size` (font size), and `fontface` (font face).
#'
#' @param text,text.scale,text.legend,text.chart,text.free `r .doc_vv("text")` `r .doc_unit("text")`
#' @param size,size.scale,size.legend,size.chart,size.free `r .doc_vv("size")` `r .doc_unit("size.text")`
#' @param col,col.scale,col.legend,col.chart,col.free `r .doc_vv("col")` `r .doc_unit("col")`
#' @param col_alpha,col_alpha.scale,col_alpha.legend,col_alpha.chart,col_alpha.free `r .doc_vv("col_alpha")` `r .doc_unit("col_alpha")`
#' @param fontface,fontface.scale,fontface.legend,fontface.chart,fontface.free `r .doc_vv("fontface")` `r .doc_unit("fontface")`
#' @param fontfamily The font family. See [gpar()][grid::gpar()] for details.
#' @param bgcol,bgcol.scale,bgcol.legend,bgcol.chart,bgcol.free Visual variable that determines
#'   the background color. See Details. `r .doc_unit("bgcol")`
#' @param bgcol_alpha,bgcol_alpha.scale,bgcol_alpha.legend,bgcol_alpha.chart,bgcol_alpha.free Visual variable that determines
#'   the background color transparency. See Details. `r .doc_unit("bgcol_alpha")`
#' @param xmod,xmod.scale,xmod.legend,xmod.chart,xmod.free Transformation variable that determines the x offset. See details. `r .doc_unit("xmod")`
#' @param ymod,ymod.scale,ymod.legend,ymod.chart,ymod.free Transformation variable that determines the y offset. See details. `r .doc_unit("ymod")`
#'   the text. See details.
#' @param angle,angle.scale,angle.legend,angle.chart,angle.free Rotation angle `r .doc_unit("angle")`
#' @param plot.order Specification in which order the spatial features are drawn.
#'   See [tm_plot_order()] for details.
#' @inherit tm_polygons details
#' @inheritParams tm_polygons
#' @seealso `r .doc_see_also_layers("terrain")`
#' @param options options passed on to the corresponding `opt_<layer_function>` function
#' @param ... to catch deprecated arguments from version < 4.0
#' @example ./examples/tm_text.R
#' @seealso `r .doc_see_also_layers("text")`
#' @export
tm_text = function(text = tm_const(),
				   text.scale = tm_scale(),
				   text.legend = tm_legend(),
				   text.chart = tm_chart_none(),
				   text.free = NA,
				   size = tm_const(),
				   size.scale = tm_scale(),
				   size.legend = tm_legend(),
				   size.chart = tm_chart_none(),
				   size.free = NA,
				   col = tm_const(),
				   col.scale = tm_scale(),
				   col.legend = tm_legend(),
				   col.chart = tm_chart_none(),
				   col.free = NA,
				   col_alpha = tm_const(),
				   col_alpha.scale = tm_scale(),
				   col_alpha.legend = tm_legend(),
				   col_alpha.chart = tm_chart_none(),
				   col_alpha.free = NA,
				   fontface = tm_const(),
				   fontface.scale = tm_scale(),
				   fontface.legend = tm_legend(),
				   fontface.chart = tm_chart_none(),
				   fontface.free = NA,
				   fontfamily = NA,
				   bgcol = tm_const(),
				   bgcol.scale = tm_scale(),
				   bgcol.legend = tm_legend(),
				   bgcol.chart = tm_chart_none(),
				   bgcol.free = NA,
				   bgcol_alpha = tm_const(),
				   bgcol_alpha.scale = tm_scale(),
				   bgcol_alpha.legend = tm_legend(),
				   bgcol_alpha.chart = tm_chart_none(),
				   bgcol_alpha.free = NA,
				   xmod = 0,
				   xmod.scale = tm_scale(),
				   xmod.legend = tm_legend_hide(),
				   xmod.chart = tm_chart_none(),
				   xmod.free = NA,
				   ymod = 0,
				   ymod.scale = tm_scale(),
				   ymod.legend = tm_legend_hide(),
				   ymod.chart = tm_chart_none(),
				   ymod.free = NA,
				   angle = 0,
				   angle.scale = tm_scale(),
				   angle.legend = tm_legend_hide(),
				   angle.chart = tm_chart_none(),
				   angle.free = NA,
				   plot.order = tm_plot_order("size", reverse = FALSE),
				   zindex = NA,
				   group = NA,
				   group.control = "check",
				   blend = "over",
				   options = opt_tm_text(),
				   ...) {

	#if (FALSE) {
	args = list(...)

	layer_fun = if ("called_from" %in% names(args)) {
		args$called_from
	} else {
		"tm_text"
	}

	if (any(v3_only("tm_text") %in% names(args))) {
		v3_start_message()
		if (!("style" %in% names(args))) {
			if (!"breaks" %in% names(args)) {
				style = "pretty"
			} else {
				style = "fixed"
			}
		} else {
			style = args$style
		}

		v3_list_init()
		if (length(style) > 1) {
			style = style[1]
			.TMAP$v3_list$mult = TRUE
		}
		col.scale.args = c(list(n = v3_impute(args, "n", 5),
								style = style,
								style.args = v3_impute(args, "style.args", list())),
							if (style %in% c("cont", "log10")) {
								c({
									if (!is.null(args$breaks)) {
										if (length(args$breaks) > 2) {
											list(ticks = v3_impute(args, "breaks", NULL, "ticks"))
										} else {
											list(ticks = v3_impute(args, "breaks", NULL, "limits"))
										}
									} else {
										list()
									}
								},
								list(outliers.trunc = c(TRUE, FALSE)))
							} else {
								list(breaks = v3_impute(args, "breaks", NULL),
									 interval.closure = v3_impute(args, "interval.closure", "left"),
									 drop.levels = v3_impute(args, "drop.levels", FALSE))
							},
							list(midpoint = v3_impute(args, "midpoint", NULL),
								 as.count = v3_impute(args, "as.count", NA),
								 values = v3_impute(args, "palette", NA, "values"),
								 values.repeat = !v3_impute(args, "stretch.palette", TRUE, "values.repeat"),
								 values.range = v3_impute(args, "contrast", NA, "values.range"),
								 values.scale = 1,
								 value.na = v3_impute(args, "colorNA", NA, "value.na"),
								 value.null = v3_impute(args, "colorNULL", NA, "value.null"),
								 value.neutral = NA,
								 labels = v3_impute(args, "labels", NULL),
								 label.na = v3_impute(args, "textNA", "Missing", "label.na"),
								 label.null = NA,
								 label.format = v3_impute(args, "legend.format", list(), "label.format")))

		col.scale.args$fun_pref = if (style == "cat") {
			"categorical"
		} else if (style %in% c("fixed", "sd", "equal", "pretty", "quantile", "kmeans", "hclust", "bclust", "fisher", "jenks", "dpih", "headtails", "log10_pretty")) {
			"intervals"
		} else if (style == "cont") {
			"continuous"
		} else if (style == "log10") {
			"continuous_log10"
		} else {
			stop("unknown style")
		}

		if ("style" %in% names(args)) {
			v3_tm_scale_instead_of_style(style, scale_fun = col.scale.args$fun_pref, vv = "col", layer_fun = layer_fun, arg_list = v3_list_get())
		} else {
			v3_tm_scale(scale_fun = "", vv = "col", layer_fun = layer_fun, arg_list = v3_list_get())
		}

		col.scale = do.call("tm_scale", args = col.scale.args)

		v3_list_init()
		col.legend.args = alist(title = v3_impute(args, "title.col", NA, "title"),
								show = v3_impute(args, "legend.col.show", NULL, "show"),
								na.show = v3_impute(args, "showNA", NA, "na.show"),
								format = v3_impute(args, "legend.format", list(), "format"),
								orientation = ifelse(v3_impute(args, "legend.col.is.portrait", TRUE, "orientation"), "portrait", "landscape"),
								reverse = v3_impute(args, "legend.col.reverse", FALSE, "reverse"))

		v3_tm_legend(fun = layer_fun, vv = "col", arg_list = v3_list_get())
		col.legend = do.call("tm_legend", col.legend.args)

		v3_list_init()
		text.scale = tm_scale_asis(value.neutral = v3_impute(args, "sizes.legend.text", NA, "value.neutral"))
		v3_tm_scale(scale_fun = "asis", vv = "text", layer_fun = layer_fun, arg_list = v3_list_get())

		if ("legend.hist" %in% names(args) && args$legend.hist) {
			col.chart = tm_chart_histogram()
			v3_tm_chart_hist(layer_fun = layer_fun, vv = "col", arg = "legend.hist")
			# to do: histogram title
		}

		v3_list_init()
		size.scale.args = list(values = tm_seq(v3_impute(args, "size.lowerbound", 0, "values = tm_seq(<size.lowerbound>, 1)"), 1, power = 1/v3_impute(args, "root", 3, "values = tm_seq(<size.lowerbound>, 1, power = 1/<root>)")),
							   values.scale = v3_impute(args, "scale", 1, "values.scale"),
							   limits = v3_impute(args, "size.lim", NULL, "limits"),
							   outliers.trunc = c(v3_impute(args, "print.tiny", FALSE, paste0("outliers.trunc = c(", {if ("print.tiny" %in% names(args)) !args$print.tiny else TRUE}, ", TRUE)") ), TRUE),
							   ticks = v3_impute(args, "breaks", NULL, "ticks"),
							   midpoint = v3_impute(args, "midpoint", NULL),
							   labels = v3_impute(args, "sizes.legend.labels", NULL, "labels"))
		v3_tm_scale(scale_fun = "continuous", vv = "size", layer_fun = layer_fun, arg_list = v3_list_get())
		size.scale = do.call("tm_scale_continuous", size.scale.args)


		v3_list_init()
		size.legend.args = alist(title = v3_impute(args, "title.size", NA, "title"),
							   show = v3_impute(args, "legend.size.show", NULL, "show"),
							   na.show = v3_impute(args, "showNA", NA, "na.show"),
							   format = v3_impute(args, "legend.format", list(), "format"),
							   orientation = ifelse(v3_impute(args, "legend.size.is.portrait", TRUE, "orientation"), "portrait", "landscape"),
							   reverse = v3_impute(args, "legend.size.reverse", FALSE, "reverse"))
		v3_tm_legend(fun = "tm_text", vv = "size", arg_list = v3_list_get())
		size.legend = do.call("tm_legend", size.legend.args)



		v4_opt_args = c("along_lines", "shadow", "just", "clustering", "point.label", "remove_overlap")
		v3_opt_args = c("along_lines", "shadow", "just", "clustering", "auto.placement", "remove_overlap")
		osel = which(v3_opt_args %in% names(args))
		if (length(osel)) {
			o3 = v3_opt_args[osel]
			o4 = v4_opt_args[osel]
			v3_layer_opt(o3, o4, layer_fun)
		}


	#}
	}

	# unused arguments: typos?
	unused = setdiff(names(args), c(v3_only("tm_text"), "called_from"))

	if (length(unused)) {
		message_layer_unused_args(layer_fun, unused)
	}

	# blend has been migrated to the layer function root; it is still passed on
	# internally via mapping.args (read as a$blend).
	options$mapping.args$blend = blend

	tm_element_list(tm_element(
		layer = "text",
		trans.fun = tmapTransCentroid,
		trans.aes = list(),
		trans.args = options$trans.args,
		trans.apply_to = "this",
		mapping.aes = list(
						   text = tmapScale(aes = "text",
						   				value = text,
						   				scale = text.scale,
						   				legend = text.legend,
						   				chart = text.chart,
						   				free = text.free),
						   size = tmapScale(aes = "size",
						   				value = size,
						   				scale = size.scale,
						   				legend = size.legend,
						   				chart = size.chart,
						   				free = size.free),
						   col = tmapScale(aes = "col",
						   				value = col,
						   				scale = col.scale,
						   				legend = col.legend,
						   				chart = col.chart,
						   				free = col.free),
						   col_alpha = tmapScale(aes = "col_alpha",
						   					  value = col_alpha,
						   					  scale = col_alpha.scale,
						   					  legend = col_alpha.legend,
						   					  chart = col_alpha.chart,
						   					  free = col_alpha.free),
						   bgcol = tmapScale(aes = "bgcol",
						   				value = bgcol,
						   				scale = bgcol.scale,
						   				legend = bgcol.legend,
						   				chart = bgcol.chart,
						   				free = bgcol.free),
						   bgcol_alpha = tmapScale(aes = "bgcol_alpha",
						   					  value = bgcol_alpha,
						   					  scale = bgcol_alpha.scale,
						   					  legend = bgcol_alpha.legend,
						   					  chart = bgcol_alpha.chart,
						   					  free = bgcol_alpha.free),
						   angle = tmapScale(aes = "angle",
						   						value = angle,
						   						scale = angle.scale,
						   						legend = angle.legend,
						   						chart = angle.chart,
						   						free = angle.free),
						   fontface = tmapScale(aes = "fontface",
						   				value = fontface,
						   				scale = fontface.scale,
						   				legend = fontface.legend,
						   				chart = fontface.chart,
						   				free = fontface.free),
						   xmod = tmapScale(aes = "xmod",
						   				 value = xmod,
						   				 scale = xmod.scale,
						   				 legend = xmod.legend,
						   				 chart = xmod.chart,
						   				 free = xmod.free),
						   ymod = tmapScale(aes = "ymod",
						   				 value = ymod,
						   				 scale = ymod.scale,
						   				 legend = ymod.legend,
						   				 chart = ymod.chart,
						   				 free = ymod.free)),

		gpar = tmapGpar(fill = NA,
						col = "__col",
						shape = NA,
						size = NA,
						cex = "__size",
						fill_alpha = NA,
						col_alpha = "__col_alpha",
						pattern = NA,
						text = "__text",
						fontface = "__fontface",
						fontfamily = fontfamily,
						lty = NA,
						lwd = NA,
						linejoin = NA,
						lineend = NA,
						bgcol = "__bgcol",
						bgcol_alpha = "__bgcol_alpha",
						xmod = "__xmod",
						ymod = "__ymod",
						angle = "__angle"),
		tpar = tmapTpar(),
		plot.order = plot.order,
		mapping.args = options$mapping.args,
		zindex = zindex,
		group = group,
		group.control = group.control,
		popup.vars = FALSE,
		popup.format = list(),
		hover = "",
		id = "",
		subclass = c("tm_aes_layer", "tm_layer")))
}


#' @rdname tm_text
#' @export
tm_labels = function(text = tm_const(),
					text.scale = tm_scale(),
					text.legend = tm_legend(),
					text.chart = tm_chart_none(),
					text.free = NA,
					size = tm_const(),
					size.scale = tm_scale(),
					size.legend = tm_legend(),
					size.chart = tm_chart_none(),
					size.free = NA,
					col = tm_const(),
					col.scale = tm_scale(),
					col.legend = tm_legend(),
					col.chart = tm_chart_none(),
					col.free = NA,
					col_alpha = tm_const(),
					col_alpha.scale = tm_scale(),
					col_alpha.legend = tm_legend(),
					col_alpha.chart = tm_chart_none(),
					col_alpha.free = NA,
					fontface = tm_const(),
					fontface.scale = tm_scale(),
					fontface.legend = tm_legend(),
					fontface.chart = tm_chart_none(),
					fontface.free = NA,
					fontfamily = "",
					bgcol = tm_const(),
					bgcol.scale = tm_scale(),
					bgcol.legend = tm_legend(),
					bgcol.chart = tm_chart_none(),
					bgcol.free = NA,
					bgcol_alpha = tm_const(),
					bgcol_alpha.scale = tm_scale(),
					bgcol_alpha.legend = tm_legend(),
					bgcol_alpha.chart = tm_chart_none(),
					bgcol_alpha.free = NA,
					xmod = 0,
					xmod.scale = tm_scale(),
					xmod.legend = tm_legend_hide(),
					xmod.chart = tm_chart_none(),
					xmod.free = NA,
					ymod = 0,
					ymod.scale = tm_scale(),
					ymod.legend = tm_legend_hide(),
					ymod.chart = tm_chart_none(),
					ymod.free = NA,
					angle = 0,
					angle.scale = tm_scale(),
					angle.legend = tm_legend_hide(),
					angle.chart = tm_chart_none(),
					angle.free = NA,
					plot.order = tm_plot_order("AREA", reverse = FALSE, na.order = "bottom"),
					zindex = NA,
					group = NA,
					group.control = "check",
					options = opt_tm_labels(),
					...) {
	args = c(as.list(environment()), list(...))
	tm = do.call(tm_text, args)
	tm[[1]]$layer = c("labels", "text")
	tm
}



#' @rdname tm_text
#' @export
tm_labels_highlighted = function(text = tm_const(),
								 text.scale = tm_scale(),
								 text.legend = tm_legend(),
								 text.chart = tm_chart_none(),
								 text.free = NA,
								 size = tm_const(),
								 size.scale = tm_scale(),
								 size.legend = tm_legend(),
								 size.chart = tm_chart_none(),
								 size.free = NA,
								 col = tm_const(),
								 col.scale = tm_scale(),
								 col.legend = tm_legend(),
								 col.chart = tm_chart_none(),
								 col.free = NA,
								 col_alpha = tm_const(),
								 col_alpha.scale = tm_scale(),
								 col_alpha.legend = tm_legend(),
								 col_alpha.chart = tm_chart_none(),
								 col_alpha.free = NA,
								 fontface = tm_const(),
								 fontface.scale = tm_scale(),
								 fontface.legend = tm_legend(),
								 fontface.chart = tm_chart_none(),
								 fontface.free = NA,
								 fontfamily = "",
								 bgcol = tm_const(),
								 bgcol.scale = tm_scale(),
								 bgcol.legend = tm_legend(),
								 bgcol.chart = tm_chart_none(),
								 bgcol.free = NA,
								 bgcol_alpha = tm_const(),
								 bgcol_alpha.scale = tm_scale(),
								 bgcol_alpha.legend = tm_legend(),
								 bgcol_alpha.chart = tm_chart_none(),
								 bgcol_alpha.free = NA,
								 xmod = 0,
								 xmod.scale = tm_scale(),
								 xmod.legend = tm_legend_hide(),
								 xmod.chart = tm_chart_none(),
								 xmod.free = NA,
								 ymod = 0,
								 ymod.scale = tm_scale(),
								 ymod.legend = tm_legend_hide(),
								 ymod.chart = tm_chart_none(),
								 ymod.free = NA,
								 angle = 0,
								 angle.scale = tm_scale(),
								 angle.legend = tm_legend_hide(),
								 angle.chart = tm_chart_none(),
								 angle.free = NA,
								 plot.order = tm_plot_order("AREA", reverse = FALSE, na.order = "bottom"),
								 zindex = NA,
								 group = NA,
								 group.control = "check",
								 options = opt_tm_labels(),
								 ...) {
	args = c(as.list(environment()), list(...))
	tm = do.call(tm_text, args)
	tm[[1]]$layer = c("labels_highlighted", "labels", "text")
	tm
}

#' @param points_only should only point geometries of the shape object (defined in [tm_shape()]) be plotted? By default `"ifany"`, which means `TRUE` in case a geometry collection is specified.
#' @param point_per specification of how spatial points are mapped when the geometry is a multi line or a multi polygon. One of \code{"feature"}, \code{"segment"} or \code{"largest"}. The first generates a spatial point for every feature, the second for every segment (i.e. subfeature), the third only for the largest segment (subfeature). Note that the last two options can be significant slower.
#' @param on_surface In case of polygons, centroids are computed. Should the points be on the surface? If `TRUE`, which is slower than the default `FALSE`, centroids outside the surface are replaced with points computed with [sf::st_point_on_surface()].
#' @param shadow Shadow behind the text. Logical.
#' @param shadow.col Color of the shadow.
#' @param shadow.offset.x,shadow.offset.y Shadow offset in line heights
#' @param halo Halo behind the text. In plot mode, it is just an outline, in view mode also a subtle glow.
#' @param halo.col Color of the halo.
#' @param halo.width Width (thickness) of the halo outline. In line heights
#' @param halo.blur Blur radius of the halo glow (view mode only). In line heights. Should be sufficiently larger than \code{halo.width} in order to see the effect.
#' @param halo.alpha Alpha transparency of the halo
#' @param just justification of the text relative to the point coordinates. Either one of the following values: \code{"left"} , \code{"right"}, \code{"center"}, \code{"bottom"}, and \code{"top"}, or a vector of two values where first value specifies horizontal and the second value vertical justification. Besides the mentioned values, also numeric values between 0 and 1 can be used. 0 means left justification for the first value and bottom justification for the second value. Note that in view mode, only one value is used.
#' @param along_lines logical that determines whether labels are rotated along the spatial lines. Only applicable if a spatial lines shape is used.
#' @param bg.padding The padding of the background in terms of line heights.
#' @param bg.border Should the background have borders?
#' @param bg.border.col Color of the borders
#' @param bg.border.lwd Line width of the borders
#' @param clustering in interactive modes (e.g. \code{"view"} mode), should clustering be applied at lower zoom levels? Either `FALSE` (default), `TRUE`, or a mode specific specification, e.g. for \code{"view"} mode \code{\link[leaflet:markerClusterOptions]{markerClusterOptions}}.
#' @param point.label logical that determines whether the labels are placed automatically. By default `FALSE` for `tm_text`, and `TRUE` for `tm_labels` if the number of labels is less than 500 (otherwise it will be too slow).
#' @param point.label.gap numeric that determines the gap between the point and label
#' @param point.label.method the optimization method, either `"SANN"` for simulated annealing (the default) or `"GA"` for a genetic algorithm.
#' @param remove_overlap logical that determines whether the overlapping labels are removed
#' @rdname tm_text
#' @export
opt_tm_text = function(points_only = "ifany",
					   point_per = "feature",
					   on_surface = FALSE,
					   shadow = FALSE,
					   shadow.col = NA,
					   shadow.offset.x = 0.05,
					   shadow.offset.y = 0.05,
					   halo = FALSE,
					   halo.col = NA,
					   halo.width = 0.05,
					   halo.blur = 0,
					   halo.alpha = 0.8,
					   just = "center",
					   along_lines = FALSE,
					   bg.padding = 0.4,
					   bg.border = FALSE,
					   bg.border.col = "black",
					   bg.border.lwd = 1,
					   clustering = FALSE,
					   point.label = FALSE,
					   point.label.gap = 0,
					   point.label.method = "SANN",
					   remove_overlap = FALSE) {
	if (!is.logical(shadow)) {
		cli::cli_warn("{.fun opt_tm_text} {.arg shadow} should be {.code TRUE} or {.code FALSE}. Please use shadow.col to specify a color")
		shadow.col = shadow
		shadow = TRUE
	}
	list(trans.args = list(points_only = points_only,
						   point_per = point_per,
						   on_surface = on_surface,
						   along_lines = along_lines),
		 mapping.args = list(shadow = shadow,
		 					shadow.col = shadow.col,
		 					shadow.offset.x = shadow.offset.x,
		 					shadow.offset.y = shadow.offset.y,
		 					halo = halo,
		 					halo.col = halo.col,
		 					halo.width = halo.width,
		 					halo.blur = halo.blur,
		 					halo.alpha = halo.alpha,
		 					just = just,
		 					along_lines = along_lines,
		 					bg.padding = bg.padding,
		 					bg.border = bg.border,
		 					bg.border.col = bg.border.col,
		 					bg.border.lwd = bg.border.lwd,
		 					clustering = clustering,
		 					point.label = point.label,
		 					point.label.gap = point.label.gap,
		 					point.label.method = point.label.method,
		 					remove_overlap = remove_overlap))
}

#' @rdname tm_text
#' @export
opt_tm_labels = function(points_only = "ifany",
						 point_per = "feature",
						 on_surface = FALSE,
						 shadow = FALSE,
						 shadow.col = NA,
						 shadow.offset.x = 0.05,
						 shadow.offset.y = 0.05,
						 halo = FALSE,
						 halo.col = NA,
						 halo.width = 0.05,
						 halo.blur = 0,
						 halo.alpha = 0.8,
						 just = "center",
						 along_lines = TRUE,
						 bg.padding = 0.4,
						 bg.border = FALSE,
						 bg.border.col = "black",
						 bg.border.lwd = 1,
						 clustering = FALSE,
						 point.label = NA,
						 point.label.gap = 0.4,
						 point.label.method = "SANN",
						 remove_overlap = FALSE) {
	if (!is.logical(shadow)) {
		cli::cli_warn("{.fun opt_tm_labels} {.arg shadow} should be {.code TRUE} or {.code FALSE}. Please use shadow.col to specify a color")
		shadow.col = shadow
		shadow = TRUE
	}

	list(trans.args = list(points_only = points_only,
						   point_per = point_per,
						   on_surface = on_surface,
						   along_lines = along_lines),
		 mapping.args = list(shadow = shadow,
		 					shadow.col = shadow.col,
		 					shadow.offset.x = shadow.offset.x,
		 					shadow.offset.y = shadow.offset.y,
		 					halo = halo,
		 					halo.col = halo.col,
		 					halo.width = halo.width,
		 					halo.blur = halo.blur,
		 					halo.alpha = halo.alpha,
		 					just = just,
		 					along_lines = along_lines,
		 					bg.padding = bg.padding,
		 					bg.border = bg.border,
		 					bg.border.col = bg.border.col,
		 					bg.border.lwd = bg.border.lwd,
		 					clustering = clustering,
		 					point.label = point.label,
		 					point.label.gap = point.label.gap,
		 					point.label.method = point.label.method,
		 					remove_overlap = remove_overlap))
}

#' Map component: inset maps and other objects
#'
#' Map component that adds an inset object, e.g. a mini map
#'
#' @param x object to draw. Can be: bounding box, tmap object, ggplot2 object, grob object, image file name.
#' @param height height of the component in number of text line heights.
#' @param width width of the component in number of text line heights.
#' @param margins margins
#' @param between_margin Margin between
#' @param box_frame Should a box frame be drawn in the main map that shows where the inset is? `TRUE` by default
#' @param box_frame.color,box_frame.alpha,box_frame.lwd,box_frame.lty Properties of the box frame
#' @param box_bg Should the frame box have a background? `FALSE` by default
#' @param box_bg.color,box_bg.alpha Properties of the box background
#' @param main_frame Should a frame be drawn around the inset map? Note that this is different from the general map component frame (the argument `frame`)
#' @param main_frame.r,main_frame.color,main_frame.alpha,main_frame.lwd Properties of the main frame
#' @inheritParams tm_title
#' @example ./examples/tm_inset.R
#' @seealso `r .doc_see_also_insets()` `r .doc_see_also_comp()`
#' @export
tm_inset = function(x = NULL,
					height,
					width,
					margins,
					between_margin,
					position,
					group_id,
					frame, frame.color, frame.alpha, frame.lwd, frame.r, bg, bg.color, bg.alpha,
					box_frame, box_frame.color, box_frame.alpha, box_frame.lwd, box_frame.lty, box_bg, box_bg.color, box_bg.alpha,
					main_frame, main_frame.r, main_frame.color, main_frame.alpha, main_frame.lwd,
					z) {
	args = lapply(as.list(rlang::call_match()[-1]), eval, envir = parent.frame())
	args$called = names(args)
	args$group_id = args$group_id %||% NA_character_
	args$group_type = c("tm_inset")
	args$z = args$z %||% NA_integer_

	cls = if (is.null(x) || (inherits(x, "bbox"))) {
		"map"
	} else if (inherits(x, "tmap")) {
		"tmap"
	} else if (inherits(x, "ggplot")) {
		"gg"
	} else if (inherits(x, c("grob", "gList"))) {
		"grob"
	} else if (is.character(x)) {
		"image"
	} else {
		stop("Unsupported object")
	}

	cls2 = paste0("tm_inset_", cls)

	tm_element_list(do.call(tm_element, c(args, list(subclass = c(cls2, "tm_inset", "tm_component")))))
}

paste2 = function(x, y, flipped) {
	if (flipped) {
		paste(y, x, sep = " ")
	} else {
		paste(x, y, sep = " ")
	}
}

fancy_breaks <- function(vec, as.count = FALSE, interval.disjoint = FALSE, intervals=FALSE, interval.closure="left", fun=NULL, scientific=FALSE, big.num.abbr = c("mln" = 6, "bln" = 9), prefix = "", suffix = "", text.separator="to", text.less.than=c("less", "than"), text.less.than_as.prefix = TRUE, text.or.more=c("or", "more"), text.or.more_as.prefix = FALSE, text.align="left", text.to.columns=FALSE, digits=NA, html.escape = TRUE, ...) {
	args <- list(...)
	n <- length(vec)

	args$called = NULL

	digits_defined = !is.na(digits)

	if (!intervals) {
		as.count = FALSE
		interval.disjoint = FALSE
	}

	if (inherits(vec, c("POSIXct", "POSIXlt", "Date"))) {
		x = format(vec)
	} else if (!is.null(fun) && !interval.disjoint) {
		if (as.count) {
			steps <- (vec[-1] - vec[-n])
			vec <- c(vec, vec - 1L, vec + 1L) # needed for: {1, 2, ... 9}
			digits <- 0
		}
		x <- do.call(fun, list(vec))
		if (as.count) {
			x1 <- x[1:(n-1)]
			x2 <- x[(n+2):(2*n)]
			x1p1 <- x[(2*n+1):(3*n-1)]
		}
	} else if (all(is.infinite(vec))) {
		x <- as.character(vec)
	} else {
		# calculate magnitude, needed to determine digits and big number abbreviations
		vec_fin <- unique(vec[!is.infinite(vec)])
		frm <- gsub(" ", "", sprintf("%20.10f", abs(vec_fin)))
		mag <- max(nchar(frm)-11)

		if (as.count) {
			steps <- (vec[-1] - vec[-n])
			vec <- c(vec, vec - 1L, vec + 1L) # needed for: {1, 2, ... 9}
			digits <- 0
		} else {
			# get number of decimals (which is number of decimals in vec, which is reduced when mag is large)
			ndec <- max(10 - nchar(frm) + nchar(sub("0+$","",frm)))
			if (!digits_defined) {
				digits <- max(min(ndec, 4-mag), 0)

				# add sign to frm
				frm_sign <- unique(paste0(ifelse(vec_fin<0, "-", "+"), frm))

				# test if number of digits is sufficient for unique labels
				if (!scientific) {
					while (anyDuplicated(substr(frm_sign, 1, nchar(frm_sign)-10 + digits)) && (digits < 10)) {
						digits <- digits + 1
					}

					# if (interval.disjoint) {
					# 	## NEW
					# 	#update vec_fin
					# 	# add 'to' numbers and see if digits should be increased
					# 	vec_fin2 = c(vec_fin, vec_fin[-1] - 10^-digits)
					# 	while (anyDuplicated(na.omit(vec_fin2))) {
					# 		digits = digits + 1
					# 		vec_fin2 = c(vec_fin, vec_fin[-1] - 10^-digits)
					# 	}
					# 	vec = c(vec, vec - 10^-digits)
					# }

				}

			# } else {
			# 	if (interval.disjoint) {
			# 		vec = c(vec, vec - 10^-digits)
			# 	}
			}
		}


		if ((!is.null(fun) && interval.disjoint)) {

			if (!digits_defined) {
				cli::cli_inform("{.fun tm_format_label} when {.arg fun} is defined and {.arg interval.disjoint} is enabled, it is unclear how to format the {.str to} numbers, and moreover how to determine the significant number of digits. Please use {.arg digits} to set this.")

				vec_fin = unique(vec[!is.infinite(vec)])
				vec_fin2 = c(vec_fin, vec_fin[-1] - 10^-digits)
				xtmp = do.call(fun, list(vec_fin2))
				while (anyDuplicated(na.omit(xtmp))) {
					digits = digits + 1
					vec_fin2 = c(vec_fin, vec_fin[-1] - 10^-digits)
					xtmp = do.call(fun, list(vec_fin2))
				}
			}
			vec = c(vec, head(vec, -1) - 10^-digits, tail(vec, 1))
			x <- do.call(fun, list(vec))
		} else if (!scientific || as.count) {

			# check whether big number abbrevations should be used
			ext <- ""
			if (!is.na(big.num.abbr[1])) {
				big.num.abbr <- sort(big.num.abbr, decreasing = TRUE)
				for (i in 1:length(big.num.abbr)) {
					o <- unname(big.num.abbr[i])
					if ((mag > o && all(vec - floor(vec/(10^o))*(10^o) < 1))) { #mag>(o+2) || removed
						vec <- vec / (10^o)
						ext <- paste0(" ", names(big.num.abbr)[i])
						break
					}
				}
			}

			if (interval.disjoint) {

				if (!digits_defined) {
					# update digits
					vec_fin = unique(vec[!is.infinite(vec)])
					vec_fin2 = c(vec_fin, vec_fin[-1] - 10^-digits)
					while (anyDuplicated(na.omit(vec_fin2))) {
						digits = digits + 1
						vec_fin2 = c(vec_fin, vec_fin[-1] - 10^-digits)
					}
				}
				# add to points
				vec = c(vec, head(vec, -1) - 10^-digits, tail(vec, 1))
			}

			# set default values
			if (!("big.mark" %in% names(args))) args$big.mark <- ","
			if (!("format" %in% names(args))) args$format <- "f"
			if (!("preserve.width" %in% names(args))) args$preserve.width <- "none"

			x = do.call("formatC", c(list(x=vec, digits=digits), args))
			if (as.count || interval.disjoint) {
				x[(n+2):(2*n)] = paste0(x[(n+2):(2*n)], ext)
			} else {
				x = paste0(x, ext)
			}
			x <- paste0(prefix, x, suffix)
		} else {
			if (!("format" %in% names(args))) args$format <- "g"
			x <- do.call("formatC", c(list(x=vec, digits=digits), args))
		}

		if (as.count) {
			x1 <- x[1:(n-1)]
			x2 <- x[(n+2):(2*n)]
			x1p1 <- x[(2*n+1):(3*n-1)]
		} else if (interval.disjoint) {
			x1 = x[1:(n-1)]
			x2 = x[(n+2):(2*n)]


		}
		# x <- formatC(vec, format = "f", digits = 0)
		# x1 <- x[-n]
		# x2 <- formatC(vec[-1] - 1L, format = "f", digits = 0)
		# xs <- (vec[-1] - vec[-n])
		# x1p1 <- formatC(vec[-n] + 1L, format = "f", digits = 0)
	}

	if (intervals) {
		if (scientific) {
			if (as.count) {
				# discrete
				lbls <- paste("{", x1, "}", sep = "")
				lbls[steps == 2] <- paste("{", x1[steps == 2], ", ", x2[steps == 2], "}", sep="")
				lbls[steps > 2] <- paste("{", x1[steps > 2], ", ", x1p1[steps > 2], ", ..., ", x2[steps > 2], "}", sep="")
			} else {
				# continuous
				if (interval.closure=="left") {
					lbls <- paste("[", x[-n], ", ", x[-1], ")", sep="")
					lbls[n-1] <- paste(substr(lbls[n-1], 1, nchar(lbls[n-1])-1), "]", sep="")
				} else {
					lbls <- paste("(", x[-n], ", ", x[-1], "]", sep="")
					lbls[1] <- paste("[", substr(lbls[1], 2, nchar(lbls[1])), sep="")
				}
			}


		} else {
			if (as.count) {
				lbls <- x1
				lbls[steps>1] <- paste(x1[steps>1], x2[steps>1], sep = paste0(" ", text.separator, " "))
				if (vec[n]==Inf) lbls[n-1] <- paste2(x1[n-1], paste(text.or.more, collapse = " "), flipped = text.or.more_as.prefix)
			} else if (interval.disjoint) {
				lbls <- paste(x1, x2, sep = paste0(" ", text.separator, " "))
				if (vec[1]==-Inf) lbls[1] <- paste2(x1[2], paste(text.less.than, collapse = " "), flipped = text.less.than_as.prefix)
				if (vec[n]==Inf) lbls[n-1] <- paste2(x1[n-1], paste(text.or.more, collapse = " "), flipped = text.or.more_as.prefix)
			} else {
				x[vec==-Inf] <- ""

				lbls <- paste(x[-n], x[-1], sep = paste0(" ", text.separator, " "))
				if (vec[1]==-Inf) lbls[1] <- paste2(x[2], paste(text.less.than, collapse = " "), flipped = text.less.than_as.prefix)
				if (vec[n]==Inf) lbls[n-1] <- paste2(x[n-1], paste(text.or.more, collapse = " "), flipped = text.or.more_as.prefix)
			}

			if (text.to.columns) {
				#xtra <- as.numeric(!is.na(text.align) && text.align=="right")


				nc1 <- nchar(paste(x[-n], " ", sep = "")) + 1
				nc2 <- rep(nchar(paste(text.separator, " ", sep = "")), n-1)

				lbls_breaks <- matrix(c(nc1, nc1+nc2), ncol=2)

				if (vec[1]==-Inf) {
					if (length(text.less.than)==1) {
						lbls_breaks[1,] <- rep(nchar(paste(text.less.than[1], " ", sep = "")) + 1, 2)
					} else {
						lbls_breaks[1,] <- cumsum(c(nchar(paste(text.less.than[1], " ", sep = "")) + 1, nchar(text.less.than[2])+1))
					}
				}
				if (vec[n]==Inf) {
					if (length(text.or.more)==1) {
						lbls_breaks[n-1,] <- rep(nchar(paste(x[n-1], " ", sep = "")) + 1, 2)
					} else {
						lbls_breaks[n-1,] <- cumsum(c(nchar(paste(x[n-1], " ", sep = "")) + 1, nchar(text.or.more[1])+1))
					}

				}
				attr(lbls, "brks") <- lbls_breaks
			}




		}
	}

	y <- if (intervals) lbls else x
	attr(y, "align") <- text.align
	y
}


num2breaks <- function(x, n, style, breaks, approx=FALSE, interval.closure="left", var = NULL, as.count = FALSE, args = list()) {

	tmapOptions = tmap_options_mode(mode.specific = FALSE)
	show.warnings <- tmapOptions$show.warnings

	nobs <- sum(!is.na(x))
	# create intervals and assign colors
	if (style=="fixed") {
		x2 = na.omit(x)
		tooLow = (x2 < min(breaks))
		tooHigh = (x2 > max(breaks))
		if (any(tooLow) && show.warnings) warning("Values have found that are less than the lowest break. They are assigned to the lowest interval", call. = FALSE)
		if (any(tooHigh) && show.warnings) warning("Values have found that are higher than the highest break. They are assigned to the highest interval", call. = FALSE)

		q <- list(var=x,
				  brks=breaks)
		attr(q, "style") <- "fixed"
		attr(q, "nobs") <- nobs
		attr(q, "intervalClosure") <- interval.closure
		class(q) <- "classIntervals"
	} else {
		if (nobs == 0) {
			if (!is.null(var)) {
				cli::cli_abort("Numerical variable {.var {var}} only contains missing values.")
			} else {
				cli::cli_abort("Numerical variable only contains missing values.")
			}
		}

		nunique <- length(na.omit(unique(x)))


		if (nunique == 1 && style != "pretty" && show.warnings) {
			if (is.null(var)) {
				bullet <- "Single value found."
			} else {
				bullet <- "Single unique value found for variable {.var {var}}"
			}
			cli::cli_warn(c(
				"!" = bullet,
				"i" = "Setting {.code style = {.str pretty}}"
			))
		}

		tempx <- nunique <= n

		if (tempx) {
			x_orig <- x
			if (length(na.omit(unique(x))) == 1) x <- pretty(x)
			x <- seq(min(x, na.rm = TRUE), max(x, na.rm = TRUE), length.out = n + 1)
		}

		q <- tryCatch({
			suppressWarnings(do.call(classInt::classIntervals, c(list(x, n, style = style, intervalClosure = interval.closure), args)))
		}, error = function(e) {
			cli::cli_abort(c(
				"Calculating interval classes failed for the variable {.var {var}} with {.code style = {.val {style}}}."
			    ),
			    parent = e
			)
		})



		if (tempx) q$var <- x_orig

	}

	if (approx && style != "fixed") {
		if (n >= length(unique(x)) && style == "equal") {
			# to prevent classIntervals to set style to "unique"
			q <- list(var = x, brks = seq(min(x, na.rm = TRUE), max(x, na.rm = TRUE), length.out = n))
			attr(q, "intervalClosure") <- interval.closure
			class(q) <- "classIntervals"
		} else {
			brks <- q$brks

			# to prevent ugly rounded breaks such as -.5, .5, ..., 100.5 for n=101
			qm1 <- suppressWarnings(do.call(classInt::classIntervals, c(list(x, n - 1, style = style, intervalClosure = interval.closure), args)))
			brksm1 <- qm1$brks
			qp1 <- suppressWarnings(do.call(classInt::classIntervals, c(list(x, n + 1, style = style, intervalClosure = interval.closure), args)))
			brksp1 <- qp1$brks
			if (min(brksm1) > min(brks) && max(brksm1) < max(brks)) {
				q <- qm1
			} else if (min(brksp1) > min(brks) && max(brksp1) < max(brks)) {
				q <- qp1
			}
		}
	}
	q
}


are_breaks_diverging = function(brks) {
	# if !divx then c-Inf, 2, 5, 10) is considered sequential
	negb <- any(brks[brks != -Inf] < 0) || (brks[1] == -Inf && brks[2]<=0)
	nb <- length(brks)
	posb <- any(brks[brks != Inf] > 0) || (brks[nb] == Inf && brks[nb-1]>=0)
	negb && posb
}








# function to determine whether a diverging of sequential palette is used given the values and the breaks
# use_diverging_palette <- function(v, brks, midpoint = NULL) {
# 	if (!is.null(midpoint) && !is.na(midpoint)) return(TRUE)
# 	x <- na.omit(v)
# 	divx <- any(x<0) && any(x>0)
#
# 	if (divx || is.null(brks)) {
# 		return(divx)
# 	} else {
# 		are_breaks_diverging(brks)
# 	}
# }
#
use_div = function(brks, midpoint = NULL) {
	if (!is.null(midpoint) && !is.na(midpoint)) return(TRUE)

	if (is.null(brks)) {
		return(NA)
	} else {
		are_breaks_diverging(brks)
	}
}



# https://stackoverflow.com/questions/12688717/round-up-from-5
#rnd <- function(x) trunc(x + sign(x) * 0.5)
round2 = function(x, n = 0) {
	posneg = sign(x)
	z = abs(x)*10^n
	z = z + 0.5 + sqrt(.Machine$double.eps)
	z = trunc(z)
	z = z/10^n
	z*posneg
}
# a = seq(0.5, 10.5, by = 0.5)
# names(a) = a
# round(a)
# round2(a, 0)
# rnd(a)


# round x to nearest multiple of y
round_num = function(x, y) round2(x / y) * y


# x = c(12.777156207494, 20.7613848999599, 28.7456135924257, 36.7298422848916, 44.7140709773574)
prettyTicks = function(x, dev = 0.1) {
	is_equi = local({
		dff = max(x) - min(x)
		if (dff == 0) return(TRUE)
		steps = (x[-1] - head(x,-1)) / dff
		all(abs(steps-steps[1]) < 1e-3)
	})

	if (is_equi) {
		pretty(x, n = length(x))
	} else {
		s = x[-1] - head(x,-1)
		s = c(s[1], s, tail(s,1))
		s = pmin(head(s,-1), tail(s,-1))

		mapply(function(xi, si) {
			for (r in rev(.TMAP$round_to)) {
				xir = round_num(xi, r)
				if (abs(xi - xir) < (si * dev)) break
			}
			xir
		}, x, s, SIMPLIFY = TRUE)	}
}
# x = c(3.654, 4.65, 5.1, 7.5)
# x = sort(c(0.004324, 0.00324, 0.00227, 0.00745435))
# x = c(3, 4,6,7,9)
# x = c(3.1,3.354,3.6)
# x = c(3.1,4.354,5.6)
# x = c(1043,2045,4005, 6765)
# x = c(265, 280, 520, 1000)
# x = c(265, 520, 1000, 1280)
# prettyTicks(x)

#' Internal tmap function to create by variables (used for faceting)
#'
#' Internal tmap function to create by variables (used for faceting)
#'
#' @param dt data.table
#' @param tmf tmf object
#' @param smeta smeta object
#' @export
#' @keywords internal
make_by_vars = function(dt, tmf, smeta) {
	by123 = paste0("by", 1L:3L)
	by123__ = paste0("by", 1L:3L, "__")
	with(tmf, {
		if (length(b)) {
			for (w in b) {
				byvar = by123[w]
				byname = by123__[w]
				var = tmf[[byvar]]

				if (var == "FRAME__") {
					dt[, (byname):= NA_integer_] # dummy
				} else if (var %in% smeta$vars) {
					levs = smeta$vars_levs[[var]]
					if (is.null(levs)) {
						cli::cli_abort("Variable {.val {var}} is used as facetting variable. Therefore, it should be a factor.")
					}
					if (attr(levs, "showNA")) levs[length(levs)] = NA

					# `levs` are always character labels (built via as.character()
					# or levels() in get_fact_levels_na). match() does coerce the
					# left-hand side to character, but for classed columns it relies
					# on as.character() S3 dispatch: Date has an as.character method
					# and formats correctly, whereas POSIXct (and other date/time
					# classes lacking such a method) fall back to the raw underlying
					# double (e.g. "1672574400") and never match the formatted labels.
					# Coercing here with as.character() uses the same representation
					# that produced `levs`, so the join is correct for every class.
					dt[, (byname):= match(as.character(get(get(..byvar))), levs)]
				} else if (tmf[[byvar]] %in% smeta$dims) {
					dt[, (byname):= match(get(get(..byvar)), smeta$dims_vals[[var]])]
				}
			}
		}
	})
	invisible(NULL)
}



#' Internal tmap function that gets factor levels with NA's
#'
#' Internal tmap function that gets factor levels with NA's
#'
#' @param x vector
#' @param o options
#' @export
#' @keywords internal
get_fact_levels_na = function(x, o) {
	# POSIXlt is stored as a list; normalise to POSIXct so it is treated as a
	# date/time vector below instead of being caught by the is.list() guard
	if (inherits(x, "POSIXlt")) x = as.POSIXct(x)

	if (inherits(x, "sfc") || is.list(x) || is.expression(x)) {
		levs = NULL
	} else if (is.factor(x)) {
		if (o$drop.empty.facets) {
			tab = tabulate(x, nbins = nlevels(x))
			anyna = (sum(tab) != length(x)) # note that NA can already be included in the levels (in that case anyna = FALSE)
			levs = levels(x)[tab != 0]
		} else {
			anyna = anyNA(x)
			levs = levels(x)
		}

		if (!o$drop.NA.facets && anyna) {
			showNA = TRUE
			levs = c(levs, o$na.text)
		} else if (!o$drop.NA.facets && any(is.na(levs))) {
			showNA = TRUE
			levs[is.na(levs)] = o$na.text
		} else if (o$drop.NA.facets && any(is.na(levs))) {
			showNA = FALSE
			levs = levs[!is.na(levs)]
		} else {
			showNA = FALSE
		}
	} else if (inherits(x, c("POSIXct", "Date", "difftime"))) {
		# Date/time classes: keep the class. The plain branch below uses
		# as.vector(), which would strip the class and turn dates into raw
		# day/second counts. Order chronologically and format the labels with
		# the class's own as.character() method; make_by_vars() coerces the
		# column with the same as.character(), so the facet join lines up.
		u = unique(x)
		if (length(u) > o$facet_levels.max) {
			levs = NULL
		} else {
			levs = as.character(sort(u))
			if (!o$drop.NA.facets && any(is.na(u))) {
				showNA = TRUE
				levs = c(levs, o$na.text)
			} else {
				showNA = FALSE
			}
		}
	} else {
 		u = unique(as.vector(x))
		if (length(u) > o$facet_levels.max) {
			levs = NULL
		} else {
			levs = as.character(sort(u))
			if (!o$drop.NA.facets && any(is.na(u))) {
				showNA = TRUE
				levs = c(levs, o$na.text)
			} else {
				showNA = FALSE
			}
		}
	}
	if (!is.null(levs)) attr(levs, "showNA") = showNA
	levs
}

tmapLeafletArrange = function(tms, nx, ncol, nrow, opts, knit, show, args, options) {
	res = lapply(tms, function(tm) {
		print(tm, show = FALSE)
	})
	res2 = do.call(leafsync::latticeView, c(res, list(ncol=ncol, sync=ifelse(identical(opts$sync, TRUE), "all", "none"), no.initial.sync = FALSE)))

	#
	# lfs <- lapply(tms, function(tm) {
	# 	tmap_leaflet(tm, add.titles = FALSE)
	# })
	# lfmv <- do.call(leafsync::latticeView, c(lfs, list(ncol=ncol, sync=ifelse(opts$sync, "all", "none"))))
	#
	# if (add.titles) lfmv <- view_add_leaflet_titles(lfmv)

	if (show) {
		if (knit) {
			kp <- get("knit_print", asNamespace("knitr"))
			return(do.call(kp, c(list(x=res2), args, list(options=options))))
		} else {
			return(print(res2))
		}
	} else res2
}

#' @export
tmapGetShapeMeta1.default = function(shp, layer, o) {
	stop("Specified shp argument of tm_shape is a ", class(shp)[1], ", which is not a recognized/supported spatial data class.", call. = FALSE)
}

#' @export
tmapGetShapeMeta2.default = function(shp, smeta, o) {
	stop("Specified shp argument of tm_shape is a ", class(shp)[1], ", which is not a recognized/supported spatial data class.", call. = FALSE)
}

#' @export
tmapSubsetShp.default = function(shp, vars) {
	stop("Specified shp argument of tm_shape is a ", class(shp)[1], ", which is not a recognized/supported spatial data class.", call. = FALSE)
}

#' @export
tmapShape.default = function(shp, is.main, crs, bbox, unit, filter, layer, shp_name, smeta, o, tmf) {
	stop("Specified shp argument of tm_shape is a ", class(shp)[1], ", which is not a recognized/supported spatial data class.", call. = FALSE)
}

# split_alpha_channel <- function(x, alpha) {
# 	if (is.null(x)) {
# 		list(col=NULL, opacity=0)
# 	} else {
# 		RGBA <- col2rgb(x, alpha = TRUE)
# 		col <- rgb(RGBA[1,], RGBA[2,], RGBA[3,], maxColorValue = 255)
# 		opacity <- unname(RGBA[4,]/255 * alpha)
# 		list(col=col, opacity=opacity)
# 	}
# }
# get_x_name <- function(type) {
# 	if (type=="fill") {
# 		"xfill"
# 	} else if (type=="symbol") {
# 		c("xsize", "xcol", "xshape")
# 	} else if (type=="raster") {
# 		"xraster"
# 	} else if (type=="line") {
# 		c("xline", "xlwd")
# 	} else if (type=="text") {
# 		c("xtext", "xtsize", "xtcol")
# 	}
# }
# 
# get_aes_name <- function(type) {
# 	if (type=="fill") {
# 		"fill"
# 	} else if (type=="symbol") {
# 		c("symbol.size", "symbol.col", "symbol.shape")
# 	} else if (type=="raster") {
# 		"raster"
# 	} else if (type=="line") {
# 		c("line.col", "line.lwd")
# 	} else if (type=="text") {
# 		c("text", "text.size", "text.color")
# 	}
# }
# 
# get_labels <- function(gpl, type) {
# 	var_names <- paste(type, "names", sep=".")
# 	gpl$data[[gpl[[var_names]]]]
# }

# 
# working_internet <- function(url = "https://www.google.com") {
# 	# test the http capabilities of the current R build
# 	if (!capabilities(what = "http/ftp")) return(FALSE)
# 	
# 	# test connection by trying to read first line of url
# 	test <- try(suppressWarnings(readLines(url, n = 1)), silent = TRUE)
# 	# return FALSE if test inherits 'try-error' class
# 	!inherits(test, "try-error")
# }
# 
# bbx_per_line <- function(bbx) {
# 	max_lines <- 60
# 	(bbx[4] - bbx[2]) / max_lines
# }
# 
# units_per_line <- function(bbx) {
# 	max_lines <- 60
# 	
# 	# calculate top-center to bottom-center
# 	vdist <- suppressWarnings({tmaptools::approx_distances(bbx, projection = 4326, target = "m")$vdist})
# 	vdist/max_lines
# }
# 
# lty2dashArray <- function(lty) {
# 	numlty <- switch(lty,
# 					 solid=0,
# 					 blank=0,
# 					 # These numbers taken from ?par
# 					 dashed=c(4, 4),
# 					 dotted=c(1, 3),
# 					 dotdash=c(1, 3, 4, 3),
# 					 longdash=c(7, 3),
# 					 twodash=c(2, 2, 6, 2),
# 					 # Otherwise we're a hex string
# 					 as.numeric(as.hexmode(strsplit(lty, "")[[1]])))
# 	paste(ifelse(numlty == 0,
# 				 "none",
# 				 numlty),
# 		  collapse=",")
# }
# 
# get_epsg_number <- function(proj) {
# 	if (inherits(proj, "crs")) {
# 		if (!is.na(proj$epsg)) {
# 			return(proj$epsg)	
# 		} else {
# 			proj <- proj$proj4string
# 		}
# 	}
# 	if (inherits(proj, "CRS")) proj <- attr(proj, "projargs")
# 	
# 	pat <- "^.*\\=epsg ?: ?(\\S*)(.*)$"
# 	epsg <- as.numeric(sub(pat, "\\1", proj[grepl(pat, proj)]))
# 	if (length(epsg)==0) NA else epsg
# }
# 
# 
# submit_labels <- function(labels, cls, pane, group_name, e) {
# 	
# 	layerIds <- get("layerIds", envir = e)
# 	
# 	
# 	types <- attr(layerIds, "types")
# 	groups <- attr(layerIds, "groups")
# 	
# 	labels_all <- unlist(layerIds, use.names = FALSE)
# 	
# 	pos <- length(labels_all)
# 	
# 	labels_all <- make.names(c(labels_all, labels), unique = TRUE)
# 	
# 	labels <- labels_all[(pos + 1): length(labels_all)]	
# 	
# 	labelsList <- list(labels)
# 	names(labelsList) <- pane
# 	
# 	layerIds <- c(layerIds, labelsList)
# 	
# 	#layerIds[[cls]] <- labels_all
# 	
# 	attr(layerIds, "types") <- c(types, cls)
# 	attr(layerIds, "groups") <- c(types, group_name)
# 	
# 	assign("layerIds", layerIds, envir = e)
# 	labels
# }
# 
# paneName <- function(x) {
# 	paste0("tmap", sprintf("%03d", x))
# }
# 
# legendName <- function(x) {
# 	paste0("legend", sprintf("%03d", x))
# }

#' Specify an animation
#'
#' Specify an animation from a tmap plot. This is similar to creating facets with [tm_facets()]. The animation subsequently can be exported to a gif or video file (e.g. mp4) with [tmap_animation()]. If the tmap plot with [tm_animate()] is printed, the animation will be previewed. The default [tm_animate()] will show the individual frames slowly (frame per seconds (fps) set to 2) whereas [tm_animate_fast()] will show them like a movie (with a fps set to 24).
#'
#' @param frames group by variable used to create the animation frames. This is similar to the `by` argument of [tm_facets_wrap()]. Instead of showing facets next to each other, they are shown as animation frames. However, under the hood `frames` will be used to specify `pages` of [tm_facets()]. This makes it possible to create an animation of regular facets.
#' @param nframes number of animation frames. So far, this only applied experimentally in transition map variables. See the extension package tmap.cartogram.
#' @param fps frames per second. Default: 30 for `tm_facets_animate` and 2 for `tm_facets_animate_slow`.
#' @param play how should the animation be played? One of `"loop"` (default), `"pingpong"`, and `"once"`, where `"loop"` means that the animation will loop indefinitely, `"pingpong"` means that it will play forward and then backward, and `"once"` means that it will play only once.
#' @param dpr device pixel ratio. The ratio between the physical pixel density of a device and its logical pixel density.
#' @param ... passed on to [tm_facets()]. Note that for animated facets, `by` can be specified to create animated facet wraps, and `rows` and `cols` to created animated facet grids.
#' @export
#' @note In older versions (< 4.1) [tm_facets()] with `page` specification was used to create animations frames and [tmap_animation()] to create the animation itself using inputs like the frame rate specification. As of version 4.2, the whole animation, including frame rate, is specified in [tm_animate()]. The animation can still be saved via [tmap_animation()].
#' @seealso [tm_facets()] which is the core function, and [tmap_animation()] used to save the animation
#' @rdname tm_animate
tm_animate_fast = function(frames = "VARS__",
							 nframes = 60L,
							 fps = 24L,
							 play = c("loop", "pingpong", "once"),
							 dpr = 2,
							 ...) {
	args = list(...)
	args_called = names(rlang::call_match()[-1])

	play = match.arg(play)

	tm = do.call("tm_facets", c(list(pages = frames, type = NA, animate = TRUE, nframes = nframes, fps = fps, play = play, dpr = dpr), args[setdiff(names(args), "type")]))
	tm[[1]]$calls = args_called
	class(tm[[1]]) = c("tm_animate", class(tm[[1]]))
	tm
}

#' @export
#' @rdname tm_animate
tm_animate = function(frames = "VARS__",
								  nframes = 60L,
								  fps = 2L,
								  play = c("loop", "pingpong", "once"),
								dpr = 2,
								  ...) {
	args = list(...)
	args_called = names(rlang::call_match()[-1])

	play = match.arg(play)

	tm = do.call("tm_facets", c(list(pages = frames, type = NA, animate = TRUE, nframes = nframes, fps = fps, play = play, dpr = dpr), args[setdiff(names(args), "type")]))
	tm[[1]]$calls = unique(c(args_called, "frames"))
	class(tm[[1]]) = c("tm_animate", class(tm[[1]]))
	tm
}

v3_tm_format = function(format, code) {
	id = "tm_format"
	ct = if (code == "") {
		paste0("Moreover, the format \"", format, "\" is unknown (also in tmap < 4.0)")
	} else {
		paste0("Instead, please use `", code, "`")
	}

	cli::cli_inform(c(
		"{.field [deprecated]} {.code tm_format(\"{format}\")} is deprecated as of tmap 4.0. {ct}"
	),
	.frequency_id = id,
	.frequency = "always"
	)
}

v3_tmap_format = function(format) {
	id = "tmap_format"

	ct = if (format %in% c("World", "World_wide", "NLD", "NLD_wide")) {
		paste0("Moreover, the format \"", format, "\" is unknown (also in tmap < 4.0)")
	} else {
		""
	}

	cli::cli_inform(c(
		"{.field [deprecated]} {.code tmap_format({.str {format}})} is deprecated as of tmap 4.0. {ct}"
	),
	.frequency_id = id,
	.frequency = "always"
	)
}

v3_tmap_format_add = function(format) {
	id = "tmap_format_add"
	cli::cli_inform(c(
		"{.field [deprecated]} {.fn tmap_format_add} is deprecated as of tmap 4.0. Please use {.code tm_layout(...)} instead, or create a style with {.fn tmap_options_save}"
	),
	.frequency_id = id,
	.frequency = "always"
	)
}

#' @export
#' @rdname tmap_internal
tmapScaleCategorical = function(x1, scale, legend, chart, o, aes, layer, layer_args, sortRev, bypass_ord, submit_legend = TRUE) {
	cls = if (inherits(scale, "tm_scale_categorical")) c("fact", "unord") else c("fact", "ord")

	if (is.factor(x1)) {
		defcols_cats = grepl("=<>=", levels(x1)[1], fixed = TRUE)
		defcols_nocats = grepl("=><=", levels(x1)[1], fixed = TRUE)

		if (defcols_cats || defcols_nocats) {
			res = strsplit(levels(x1), {if (defcols_cats) "=<>=" else "=><="}, fixed = TRUE)
			levels(x1) = vapply(res, "[", 1, FUN.VALUE = character(1))
			ct = vapply(res, "[", 2, FUN.VALUE = character(1))

			if (defcols_nocats && !legend$legend_called) {
				legend$show = FALSE
			}
		} else {
			ct = NULL
		}
	} else {
		ct = NULL
	}


	scale = get_scale_defaults(scale, o, aes, layer, cls, ct)

	show.messages <- o$show.messages
	show.warnings <- o$show.warnings


	with(scale, {
		check_values(layer, aes, values)

		nms = getValuesNames(values)

		nv = length(nms)
		if (all(nms == "")) nms = NULL

		# cast to factor if needed
		if (!is.factor(x1)) {
			su = sort(unique(x1))
			x1 = tryCatch({
				factor(x1, levels=su)
			}, error = function(e) {
				stop("tm_scale_categorical in layer \"tm_", layer, "\", visual variable \"", aes, "\" cannot be applied due to an error categorization of the data", call. = FALSE)
			})

			if (is.numeric(su)) levels(x1) <- do.call("fancy_breaks", c(list(vec=su, intervals=FALSE, as.count = FALSE), label.format))
		}

		# select levels
		if (!is.null(levels)) {
			x1 = factor(x1, levels = levels)
		}

		# drop levels
		if (levels.drop) {
			y = droplevels(x1)
			matching = match(levels(y), levels(x1))
			if (length(values) == nlevels(x1) && is.null(nms)) {
				values = values[matching]
			}
			if (!is.null(labels) && (length(labels) == nlevels(x1)) && is.null(names(labels))) {
				labels = labels[matching]
			}
			x1 = y
		}

		lvls = levels(x1)
		n = nlevels(x1)

		if (is.null(labels)) {
			labs = lvls
		} else {
			if (is.null(names(labels))) {
				if (length(labels) != n) warning("labels do not have the same length as levels, so they are repeated", call. = FALSE)
				labs = rep(labels, length.out = n)
			} else {
				nms = names(labels)
				labs = structure(lvls, names = lvls)

				nms2 = intersect(nms, lvls)
				labs[nms2] = unname(labels[nms2])
			}
		}
		names(labs) = NULL

		if (!is.null(nms)) {
			xlev = levels(x1)
			if (!all(xlev %in% nms)) {
				stop("All levels should occur in the vector names of values: ", paste(setdiff(xlev, nms), collapse = ", "), " are missing", call. = FALSE)
			} else {
				values = values[match(xlev, nms)]
			}
		}



		# combine levels
		if (n.max < n) {
			# only show warning if n.max is the default
			if (show.warnings && n.max == 30) warning("Number of levels of the variable assigned to the aesthetic \"",aes ,"\" of the layer \"", layer, "\" is ", n, ", which is larger than n.max (which is ", n.max, "), so levels are combined.", call. = FALSE)

			mapping = as.numeric(cut(seq.int(n), breaks=n.max))
			to = c(which(mapping[-n] - mapping[-1]!=0), n)
			from = c(0, to[-n.max]) + 1

			new_lvls = paste0(labs[from], "...", labs[to])

			x1 = factor(mapping[as.integer(x1)], levels=1L:n.max, labels=new_lvls)
			labs = new_lvls
		}
		n = nlevels(x1)

		# update range if NA (automatic)
		if (is.na(values.range[1])) {
			fun_range = paste0("tmapValuesRange_", aes)
			values.range = do.call(fun_range, args = list(x = values, n = n, isdiv = FALSE))
		}
		if (length(values.range) == 1 && !is.na(values.range[1])) values.range = c(0, values.range)

		fun_getCVV = paste0("tmapValuesCVV_", aes)
		VV = do.call(fun_getCVV, list(x = values, value.na = value.na, n = n, range = values.range, scale = values.scale, rep = values.repeat, o = o))

		values = VV$vvalues
		value.na = VV$value.na

		sfun = paste0("tmapValuesScale_", aes)
		cfun = paste0("tmapValuesColorize_", aes)
		if (identical(value.neutral, NA)) value.neutral = VV$value.neutral else value.neutral = do.call(sfun, list(x = do.call(cfun, list(x = value.neutral, pc = o$pc)), scale = values.scale))

		mfun = paste0("tmapValuesSubmit_", aes)
		values = do.call(mfun, list(x = values, args = layer_args))
		value.na = do.call(mfun, list(x = value.na, args = layer_args))
		value.neutral = do.call(mfun, list(x = value.neutral, args = layer_args))


		# legend.palette <- do.call("process_color", c(list(col=legend.palette), process.colors))
		# colorNA <- do.call("process_color", c(list(col=colorNA), process.colors))

		ids = as.integer(x1)
		vals = values[ids]

		isna = is.na(vals)
		anyNA = any(isna)

		na.show = update_na.show(label.show, legend$na.show, anyNA)




		if (is.null(sortRev)) {
			ids = NULL
		} else if (is.na(sortRev)) {
			ids[] = 1L
		} else if (sortRev) {
			ids = (as.integer(n) + 1L) - ids
		}

		if (anyNA) {
			vals[isna] = value.na
			if (!is.null(sortRev)) ids[isna] = 0L
		}




		if (legend$reverse) {
			labs = rev(labs)
			values_rev = rev(values)
		} else {
			values_rev = values
		}

		values_orig = unname(values_rev)
		labels_orig = labs
		levels_orig = levels

		if (na.show) {
			labs = c(labs, label.na)
			values = c(values, value.na)
			values_rev = c(values_rev, value.na)
		}
		attr(labs, "align") = label.format$text.align

		# add info for source shapes (e.g. PMTiles)
		layer_args = c(layer_args, list(mapping = list(values_orig = values_orig, levels_orig = levels_orig, labels_orig = labels_orig, label_na = label.na, value_na = value.na)))


		# SPECIAL CASE: if icons are used, specify this information in the symbol legend, such that it can be taken (in step4_plot_collect_legends) by other legends (e.g. for symbol sizes)
		icon_scale = if ((aes == "shape") && any(values_rev > 999) && getOption("tmap.mode") == "plot") layer_args$icon.scale else 1

		legend = within(legend, {
			nitems = length(labs)
			labels = labs
			dvalues = vals
			vvalues = values_rev
			vneutral = value.neutral
			icon_scale = icon_scale
			na.show = get("na.show", envir = parent.env(environment()))
			scale = "categorical"
			layer_args = layer_args
		})


		chartFun = paste0("tmapChart", toTitleCase(chart$summary))
		chart = do.call(chartFun, list(chart,
									   bin_colors = values,
									   breaks_def = NULL,
									   na.show = na.show,
									   x1 = x1))



		if (submit_legend) {
			if (bypass_ord) {
				format_aes_results(vals, legend = legend, chart = chart)
			} else {
				format_aes_results(vals, ids, legend, chart = chart)
			}
		} else {
			list(vals = vals, ids = ids, legend = legend, chart = chart, bypass_ord = bypass_ord)
		}

	})
}

#' @export
#' @rdname tmapGridLeaflet
tmapGridDataPlot.tm_data_polygons = function(a, shpTM, dt, gp, bbx, facet_row, facet_col, facet_page, id, pane, group, glid, o, ...) {

	rc_text = frc(facet_row, facet_col)

	res = select_sf(shpTM, dt)
	shp = res$shp

	shp_is_pointer = inherits(shp, "character")

	if (shp_is_pointer) {
		cli::cli_inform("{.field plot mode} polygons of pmtiles are not supported in plot mode yet")
		return(NULL)
	}


	dt = res$dt
	gp = impute_gp(gp, dt)
	gp = rescale_gp(gp, o$scale_down)

	# none should contain NA's && (length or content should be different)
	diffAlpha = !anyNA(c(gp$fill_alpha, gp$col_alpha)) && !(length(gp$fill_alpha) == length(gp$col_alpha) && all(gp$fill_alpha == gp$col_alpha))


	if (diffAlpha) {
		gp1 = gp_to_gpar(gp, sel = "fill", o = o, type = "polygons")
		gp2 = gp_to_gpar(gp, sel = "col", o = o, type = "polygons")
		grb1 = sf::st_as_grob(shp, gp = gp1, name = paste0("polygons_", id))
		grb2 = sf::st_as_grob(shp, gp = gp2, name = paste0("polygon_borders_", id))
		grb = grid::grobTree(grb1, grb2)
	} else {
		gp = gp_to_gpar(gp, sel = "all", o = o, type = "polygons")
		grb = sf::st_as_grob(shp, gp = gp, name = paste0("polygons_", id))
	}


	gts = get("gts", .TMAP_GRID)
	gt = gts[[facet_page]]

	gt_name = paste0("gt_facet_", rc_text)

	gt = grid::addGrob(gt, grb, gPath = grid::gPath(gt_name))

	gts[[facet_page]] = gt

	assign("gts", gts, envir = .TMAP_GRID)
	NULL
}


#' @export
#' @rdname tmapGridLeaflet
tmapGridDataPlot.tm_data_fill = function(a, shpTM, dt, gp, bbx, facet_row, facet_col, facet_page, id, pane, group, glid, o, ...) {
	NextMethod()
}


#' @export
#' @rdname tmapGridLeaflet
tmapGridDataPlot.tm_data_borders = function(a, shpTM, dt, gp, bbx, facet_row, facet_col, facet_page, id, pane, group, glid, o, ...) {
	NextMethod()
}

#' @export
#' @rdname tmapGridLeaflet
tmapLeafletDataPlot.tm_data_lines = function(a, shpTM, dt, pdt, popup.format, hdt, idt, gp, bbx, facet_row, facet_col, facet_page, id, pane, group, glid, o, ...) {
	# ptdt (popup.title data) and popup.layout are passed via `...` to keep this
	# method's signature identical to the generic (see tm_data_polygons).
	dots = list(...)
	ptdt = dots$ptdt
	popup.layout = dots$popup.layout

	lf = get_lf(facet_row, facet_col, facet_page)

	rc_text = frc(facet_row, facet_col)

	res = select_sf(shpTM, dt[!is.na(dt$lwd), ])
	shp = res$shp
	if (o$crs_leaflet$crsClass  == "L.CRS.Simple") {
		shp = sf::st_set_crs(shp, NA)
	}

	names(shp) = NULL # also required for other layers?
	dt = res$dt

	if (!is.null(idt)) {
		idt = idt$id[match(dt$tmapID__, idt$tmapID__)]
	}
	if (!is.null(hdt)) {
		hdt = hdt$hover[match(dt$tmapID__, hdt$tmapID__)]
		hdt = lapply(hdt, htmltools::HTML, FUN.VALUE = character(1))
	}

	if (!is.null(ptdt)) {
		pttl = ptdt$title[match(dt$tmapID__, ptdt$tmapID__)]
	} else {
		pttl = NULL
	}

	if (is.null(pdt)) {
		popups = NULL
	} else {
		mtch = match(dt$tmapID__, pdt$tmapID__)
		pdt = pdt[mtch][, tmapID__ := NULL]


		popups = view_format_popups(id = if (!is.null(pttl)) pttl else idt, titles = names(pdt), values = pdt, format = popup.format, layout = popup.layout)
	}

	gp = impute_gp(gp, dt)
	gp = rescale_gp(gp, o$scale_down)

	interactive = (!is.null(pdt) || !is.null(hdt))


	# process hitbox option:
	hitbox = if (isTRUE(a$hitbox == "auto")) {
		thin = median(gp$lwd) < 4
		light = length(shp) < 10000
		if (thin && light) "pmax8" else "none"
	} else {
		a$hitbox %||% "none"
	}

	o$use_WebGL = impute_webgl(o$use_WebGL, dt, supported = "col", checkif = list(lty = "solid"), type = "lines", hover = !is.null(hdt), popup = !is.null(pdt), crs_class = o$crs_leaflet$crsClass)


	k = nrow(dt)
	if (hitbox == "none") {
		opt1 = leaflet::pathOptions(interactive = interactive, pane = pane, lineCap = gp$lineend, lineJoin = gp$linejoin)

	} else {
		opt1 = leaflet::pathOptions(interactive = FALSE, pane = pane, lineCap = gp$lineend, lineJoin = gp$linejoin)
		opt2 = leaflet::pathOptions(interactive = interactive, pane = pane, lineCap = gp$lineend, lineJoin = gp$linejoin)

		hb = parse_hitbox(hitbox)
	}


	if (o$use_WebGL) {
		shp2 = sf::st_sf(id = seq_along(shp), geom = shp)
		shp3 = suppressWarnings(sf::st_cast(shp2, "LINESTRING"))
		gp3 = lapply(gp, function(gpi) {if (length(gpi) == 1) gpi else gpi[shp3$id]})
		popups = popups[shp3$id]

		if (hitbox == "none") {

			idt = {if (is.null(idt)) dt$tmapID__[1] else idt[1]} |>
				submit_labels("linesGL", pane, group)

			lf %>%
				leafgl::addGlPolylines(
					data = shp3,
					color = gp3$col,
					opacity = gp3$col_alpha[1],
					weight = lwd_to_leafgl(gp3$lwd[1]),
					pane = pane,
					group = group,
					layerId = idt[1],
					label = hdt,
					popup= popups) %>%
				assign_lf(facet_row, facet_col, facet_page)
		} else {
			idt = {if (is.null(idt)) dt$tmapID__[c(1, 1:k)] else idt[c(1, 1:k)]} |>
				submit_labels("linesGL_hb", pane, group)

			weights = gp3$lwd + hb$plus
			if (hb$pmax != 0) weights = pmax(weights, hb$pmax)

			lf %>%
				# Fast WebGL rendering
				leafgl::addGlPolylines(
					data    = shp3,
					color   = gp3$col,
					opacity = gp3$col_alpha[1],
					weight  = lwd_to_leafgl(gp3$lwd[1]),
					pane    = pane,
					group   = group,
					popup = NULL,
					layerId = idt[1]
				) %>%

				# Invisible interaction layer
				leaflet::addPolylines(
					data    = shp3,
					opacity = 0,
					weight  = weights,
					group   = group,
					layerId = idt[-1],
					label   = hdt,
					options = opt2,
					popup   = popups
				) %>%

				assign_lf(facet_row, facet_col, facet_page)
		}

	} else {
		if (hitbox == "none") {
			idt = if (is.null(idt)) dt$tmapID__ else idt |>
				submit_labels("lines", pane, group)

			lf %>%
				leaflet::addPolylines(
					data = shp,
					layerId = idt,
					label = hdt,
					color = gp$col,
					opacity = gp$col_alpha,
					weight = gp$lwd,
					group = group,
					options = opt1,
					dashArray = lty2dash(gp$lty, gp$lwd),
					popup = popups) %>%
				assign_lf(facet_row, facet_col, facet_page)
		} else {
			idt = rep({if (is.null(idt)) dt$tmapID__ else idt}, 2) |>
				submit_labels("lines", pane, group)


			weights = gp$lwd + hb$plus
			if (hb$pmax != 0) weights = pmax(weights, hb$pmax)

			lf %>%
				# Visible styled layer (no popup here)
				leaflet::addPolylines(
					data      = shp,
					layerId = idt[1:k],
					color     = gp$col,
					opacity   = gp$col_alpha,
					weight    = gp$lwd,
					group     = group,
					options   = opt1,
					dashArray = lty2dash(gp$lty, gp$lwd)
				) %>%

				# Invisible interaction layer (fatter hitbox)
				leaflet::addPolylines(
					data    = shp,
					layerId = idt[(k+1):(2*k)],
					label   = hdt,
					popup   = popups,
					options = opt2,
					opacity = 0,
					weight  = weights,
					group   = group
				) %>%

				assign_lf(facet_row, facet_col, facet_page)

		}


		# lf %>%
		# 	leaflet::addPolylines(data = shp, layerId = idt, label = hdt, color = gp$col, opacity = gp$col_alpha, weight = gp$lwd, group = group, options = opt, dashArray = lty2dash(gp$lty, gp$lwd), popup = popups) %>%
		# 	assign_lf(facet_row, facet_col, facet_page)

	}
	NULL
}

#' @export
#' @rdname tmapGridLeaflet
tmapLeafletDataPlot.tm_data_iso = function(a, shpTM, dt, pdt, popup.format, hdt, idt, gp, bbx, facet_row, facet_col, facet_page, id, pane, group, glid, o, ...) {
	# isn't called, but needed to make tm_iso visible in tmap_overview
	NextMethod()
}

parse_hitbox <- function(hitbox) {

	# defaults
	plus  <- 0
	pmax_ <- 0

	if (is.null(hitbox) || hitbox == "none") {
		return(list(plus = 0, pmax = 0))
	}

	# extract plus
	m_plus <- regexpr("plus[0-9]+", hitbox)
	if (m_plus > 0) {
		plus <- as.numeric(sub("plus", "", regmatches(hitbox, m_plus)))
	}

	# extract pmax
	m_pmax <- regexpr("pmax[0-9]+", hitbox)
	if (m_pmax > 0) {
		pmax_ <- as.numeric(sub("pmax", "", regmatches(hitbox, m_pmax)))
	}

	list(plus = plus, pmax = pmax_)
}

lwd_to_leafgl <- function(lwd) {
	# empirical formula, based on observations: 1 (very small), 2 (2/8), 3 (3/6), 4 (4/4), 6 (6/3.5), 8 (8/3), 16 (16/8)
	ifelse(lwd <= 4,
		   lwd / 2^(4 - log2(lwd)),
	ifelse(lwd > 8,
		   lwd / 3,
		   lwd / (6 - log2(lwd))
		   ))
}

#' Set the map projection (CRS)
#'
#' This function sets the map projection. It can also be set via [tm_shape()], but `tm_crs` is generally recommended for most cases. It can also be determined automatically (see details); however, this is still work-in-progress.
#'
#' @param crs Map projection (CRS). Can be set to an `crs` object (see [sf::st_crs()]), a proj4string, an EPSG number, the value `"auto"` (automatic crs recommendation), or one the the following generic projections: `c("laea", "aeqd", "utm", "pconic", "eqdc", "stere")`. See details.
#' @param property Which property should the projection have? One of: `"global"`, `"area"` (equal-area), `"distance"` (equidistant), `"shape"` (conformal). Only applicable if `crs = "auto"`. See details.
#' @param bbox bounding box. Three options: a [sf::st_bbox()] object, an Open Street Map query (passed on to [tmaptools::geocode_OSM()]), or `"FULL"`, which means the whole earth , which means the whole earth (this also guarantees that transformations to another CRS keep the whole earth, unlike \code{\link[sf:st_bbox]{sf::st_bbox()}}).
#' @note Plans are to migrate the functionality regarding generic crs and automatic crs recommendation to a separate package.
#' @inherit tm_shape details
#' @example ./examples/tm_crs.R
#' @seealso \href{https://r-tmap.github.io/tmap/articles/foundations_crs}{vignette about CRS}
#' @export
tm_crs = function(crs = NA, property = NA, bbox = NULL) {
	if (is.na(crs)) {
		if (is.na(property)) {
			return(NULL)
		} else {
			crs = "auto"
		}
	}
	if (identical(crs, "auto") && !is.na(property)) {
		if (!(property %in% c("global", "area", "distance", "shape"))) {
			message_crs_property_unknown()
		}
		extra = property
	} else {
		extra = ""
	}

	if (is.character(crs) && crs %in% c("laea", "aeqd", "utm", "pconic", "eqdc")) {
		extra = crs
		crs = "auto"
	}

	tm_options(crs = crs, crs_extra = extra, bbox = bbox)
}


consider_global = function(x, th = 0.6) {
	b = sf::st_bbox(x)
	# in case margins are applied
	if (sf::st_is_longlat(b)) {
		b["xmin"] = max(b["xmin"], -180)
		b["xmax"] = min(b["xmax"], 180)
		b["ymin"] = max(b["ymin"], -90)
		b["ymax"] = min(b["ymax"], 90)
	}
	if (b$xmin == b$xmax || b$ymin == b$ymax) return(FALSE)
	earth_surface = 5.1e14
	area = b |>
		sf::st_as_sfc() |>
		sf::st_area() |>
		as.numeric()
	area > (earth_surface * 0.6)
}

auto_crs = function(x, crs_extra, crs_global) {

	if (identical(x, TRUE)) return(crs_global)




    # check if global should be used
	is_global = if (crs_extra == "global") {
		TRUE
	} else if (crs_extra != "") {
		FALSE
	} else {
		consider_global(x)
	}

	y = if (is_global) {
		crs_global
	} else {
		# impute family
		proj = switch(crs_extra,
			area = "laea",
			distance = "aeqd",
			shape = "stere",
			aeqd = "aeqd",
			utm = "utm",
			pconic = "pconic",
			eqdc = "eqdc",
			stere = "stere",
			"laea"
		)
		to_generic_projected(x, proj = proj, return_as = "crs")
	}
	sf::st_crs(y)
}


to_generic_projected <- function(
		x,
		proj = c("laea", "aeqd", "utm", "pconic", "eqdc", "stere"),
		ellps = "WGS84",
		no_defs = TRUE,
		opts = "",
		return_as = c("sf", "proj4", "wkt", "crs")) {
	# arg assertions
	if (!rlang::is_true(rlang::inherits_any(x, c("sf", "sfc", "stars")))) {
		rlang::abort(
			"`x` must be either an sf object or an sfc object or a stars object",
			class = "to_generic_projected_incorrect_input"
		)
	}

	proj <- rlang::arg_match(proj)
	ellps <- rlang::arg_match(ellps, sf::sf_proj_info(type = "ellps")$name)

	if (!rlang::is_logical(no_defs)) {
		rlang::abort("`no_defs` must be a logical value",
					 class = "to_generic_projected_incorrect_input"
		)
	}

	if (!rlang::is_character(opts) && !nchar(opts)) {
		rlang::abort("`opts` must be a character vector",
					 class = "to_generic_projected_incorrect_input"
		)
	}

	return_as <- rlang::arg_match(return_as)

	# was centroid
	cent_coor <- sf::sf_project(
		sf::st_crs(x), "EPSG:4326",
		sf::st_bbox(x) |>
			sf::st_as_sfc() |>
			sf::st_centroid() |>
			sf::st_coordinates()
	)

	# configure proj args
	n_or_s <- ifelse(cent_coor[2] == 0, "",
					 ifelse(cent_coor[2] > 0, "+north", "+south")
	)

	no_defs <- ifelse(no_defs, "+no_defs", "")

	if (proj %in% c("pconic", "eqdc")) {
		bounds <- sf::st_bbox(sf::st_transform(x, 4326))
		lat_1 <- bounds$ymax
		lat_2 <- bounds$ymin
	}

	# base R replacement of glue::glue to reduce dependencies
	glue = function(..., .sep) {
		args = list(...)
		paste(lapply(args, gluestick, src = parent.frame()), collapse = .sep)
	}


	# construct proj4 string
	prj <- trimws(switch(proj,
						 laea = glue(
						 	"+proj=laea +lon_0={cent_coor[1]} +lat_0={cent_coor[2]}",
						 	"+ellps={ellps} {no_defs}",
						 	opts,
						 	.sep = " "
						 ),
						 utm = glue(
						 	"+proj=utm +zone={round((180 + cent_coor[1]) / 6)} {n_or_s}",
						 	"+ellps={ellps} {no_defs}",
						 	opts,
						 	.sep = " "
						 ),
						 aeqd = glue(
						 	"+proj=aeqd +lon_0={cent_coor[1]} +lat_0={cent_coor[2]}",
						 	"+ellps={ellps} {no_defs}",
						 	opts,
						 	.sep = " "
						 ),
						 pconic = glue(
						 	"+proj=pconic +lon_0={cent_coor[1]} +lat_0={cent_coor[2]}",
						 	"+lat_1={lat_1} +lat_2={lat_2}",
						 	"+ellps={ellps} {no_defs}",
						 	opts,
						 	.sep = " "
						 ),
						 eqdc = glue(
						 	"+proj=eqdc +lon_0={cent_coor[1]}",
						 	"+lat_1={lat_1} +lat_2={lat_2}",
						 	"+ellps={ellps} {no_defs}",
						 	opts,
						 	.sep = " "
						 ),
						 stere = glue(
						 	"+proj=stere +lon_0={cent_coor[1]} +lat_0={cent_coor[2]}",
						 	"+ellps={ellps} {no_defs}",
						 	opts,
						 	.sep = " "
						 )
	))

	switch(return_as,
		   sf = sf::st_transform(x, prj),
		   proj4 = prj,
		   wkt = sf::st_crs(prj)$wkt,
		   crs = sf::st_crs(prj)
	)
}




# Source: https://github.com/coolbutuseless/gluestick
gluestick <- function(fmt, src = parent.frame(), open = "{", close = "}", eval = TRUE) {

	nchar_open  <- nchar(open)
	nchar_close <- nchar(close)

	#~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
	# Sanity checks
	#~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
	stopifnot(exprs = {
		is.character(fmt)
		length(fmt) == 1L
		is.character(open)
		length(open) == 1L
		nchar_open > 0L
		is.character(close)
		length(close) == 1
		nchar_close > 0
	})

	#~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
	# Brute force the open/close characters into a regular expression for
	# extracting the expressions from the format string
	#~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
	open  <- gsub("(.)", "\\\\\\1", open ) # Escape everything!!
	close <- gsub("(.)", "\\\\\\1", close) # Escape everything!!
	re    <- paste0(open, ".*?", close)

	#~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
	# Extract the delimited expressions
	#~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
	matches  <- gregexpr(re, fmt)
	exprs    <- regmatches(fmt, matches)[[1]]

	#~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
	# Remove the delimiters
	#~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
	exprs <- substr(exprs, nchar_open + 1L, nchar(exprs) - nchar_close)

	#~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
	# create a valid sprintf fmt string.
	#  - replace all "{expr}" strings with "%s"
	#  - escape any '%' so sprintf() doesn't try and use them for formatting
	#    but only if the '%' is NOT followed by an 's'
	#
	# gluestick() doesn't deal with any pathological cases
	#~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
	fmt_sprintf <- gsub(re      , "%s", fmt)
	fmt_sprintf <- gsub("%(?!s)", "%%", fmt_sprintf, perl=TRUE)

	#~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
	# Evaluate
	#~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
	if (eval) {
		args <- lapply(exprs, function(expr) {eval(parse(text = expr), envir = src)})
	} else {
		args <- unname(mget(exprs, envir = as.environment(src)))
	}

	#~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
	# Create the string(s)
	#~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
	do.call(sprintf, c(list(fmt_sprintf), args))
}

findZoom = function(b) {
	## calculate zoom level
	# borrowed from https://github.com/dkahle/ggmap/blob/master/R/calc_zoom.r
	lon_diff = b[3] - b[1]
	lat_diff = b[4] - b[2]

	zoomlon = ceiling(log2(360 * 2/lon_diff))
	zoomlat = ceiling(log2(180 * 2/lat_diff))
	zoom = as.integer(min(zoomlon, zoomlat))
}


#' @export
#' @keywords internal
#' @rdname tmapGridLeaflet
tmapGridAuxPrepare = function(a, bs, id, o) {
	UseMethod("tmapGridAuxPrepare")
}

#' @export
#' @keywords internal
#' @rdname tmapGridLeaflet
tmapGridAuxPrepare.default = function(a, bs, id, o) {
	""
}


#' @export
#' @keywords internal
#' @rdname tmapGridLeaflet
tmapGridAuxPlot = function(a, bi, bbx, facet_row, facet_col, facet_page, id, pane, group, o) {
	UseMethod("tmapGridAuxPlot")
}

#' @export
#' @keywords internal
#' @rdname tmapGridLeaflet
tmapGridAuxPlot.default = function(a, bi, bbx, facet_row, facet_col, facet_page, id, pane, group, o) {
	NULL
}

paste2 = function(x, y, flipped) {
	if (flipped) {
		paste(y, x, sep = " ")
	} else {
		paste(x, y, sep = " ")
	}
}

fancy_breaks <- function(vec, as.count = FALSE, interval.disjoint = FALSE, intervals=FALSE, interval.closure="left", fun=NULL, scientific=FALSE, big.num.abbr = c("mln" = 6, "bln" = 9), prefix = "", suffix = "", text.separator="to", text.less.than=c("less", "than"), text.less.than_as.prefix = TRUE, text.or.more=c("or", "more"), text.or.more_as.prefix = FALSE, text.align="left", text.to.columns=FALSE, digits=NA, html.escape = TRUE, ...) {
	args <- list(...)
	n <- length(vec)

	args$called = NULL

	digits_defined = !is.na(digits)

	if (!intervals) {
		as.count = FALSE
		interval.disjoint = FALSE
	}

	if (inherits(vec, c("POSIXct", "POSIXlt", "Date"))) {
		x = format(vec)
	} else if (!is.null(fun) && !interval.disjoint) {
		if (as.count) {
			steps <- (vec[-1] - vec[-n])
			vec <- c(vec, vec - 1L, vec + 1L) # needed for: {1, 2, ... 9}
			digits <- 0
		}
		x <- do.call(fun, list(vec))
		if (as.count) {
			x1 <- x[1:(n-1)]
			x2 <- x[(n+2):(2*n)]
			x1p1 <- x[(2*n+1):(3*n-1)]
		}
	} else if (all(is.infinite(vec))) {
		x <- as.character(vec)
	} else {
		# calculate magnitude, needed to determine digits and big number abbreviations
		vec_fin <- unique(vec[!is.infinite(vec)])
		frm <- gsub(" ", "", sprintf("%20.10f", abs(vec_fin)))
		mag <- max(nchar(frm)-11)

		if (as.count) {
			steps <- (vec[-1] - vec[-n])
			vec <- c(vec, vec - 1L, vec + 1L) # needed for: {1, 2, ... 9}
			digits <- 0
		} else {
			# get number of decimals (which is number of decimals in vec, which is reduced when mag is large)
			ndec <- max(10 - nchar(frm) + nchar(sub("0+$","",frm)))
			if (!digits_defined) {
				digits <- max(min(ndec, 4-mag), 0)

				# add sign to frm
				frm_sign <- unique(paste0(ifelse(vec_fin<0, "-", "+"), frm))

				# test if number of digits is sufficient for unique labels
				if (!scientific) {
					while (anyDuplicated(substr(frm_sign, 1, nchar(frm_sign)-10 + digits)) && (digits < 10)) {
						digits <- digits + 1
					}

					# if (interval.disjoint) {
					# 	## NEW
					# 	#update vec_fin
					# 	# add 'to' numbers and see if digits should be increased
					# 	vec_fin2 = c(vec_fin, vec_fin[-1] - 10^-digits)
					# 	while (anyDuplicated(na.omit(vec_fin2))) {
					# 		digits = digits + 1
					# 		vec_fin2 = c(vec_fin, vec_fin[-1] - 10^-digits)
					# 	}
					# 	vec = c(vec, vec - 10^-digits)
					# }

				}

			# } else {
			# 	if (interval.disjoint) {
			# 		vec = c(vec, vec - 10^-digits)
			# 	}
			}
		}


		if ((!is.null(fun) && interval.disjoint)) {

			if (!digits_defined) {
				cli::cli_inform("{.fun tm_format_label} when {.arg fun} is defined and {.arg interval.disjoint} is enabled, it is unclear how to format the {.str to} numbers, and moreover how to determine the significant number of digits. Please use {.arg digits} to set this.")

				vec_fin = unique(vec[!is.infinite(vec)])
				vec_fin2 = c(vec_fin, vec_fin[-1] - 10^-digits)
				xtmp = do.call(fun, list(vec_fin2))
				while (anyDuplicated(na.omit(xtmp))) {
					digits = digits + 1
					vec_fin2 = c(vec_fin, vec_fin[-1] - 10^-digits)
					xtmp = do.call(fun, list(vec_fin2))
				}
			}
			vec = c(vec, head(vec, -1) - 10^-digits, tail(vec, 1))
			x <- do.call(fun, list(vec))
		} else if (!scientific || as.count) {

			# check whether big number abbrevations should be used
			ext <- ""
			if (!is.na(big.num.abbr[1])) {
				big.num.abbr <- sort(big.num.abbr, decreasing = TRUE)
				vec_fin = vec[!is.infinite(vec)]
				for (i in 1:length(big.num.abbr)) {
					o <- unname(big.num.abbr[i])

					if ((mag > o && all(vec_fin - floor(vec_fin/(10^o))*(10^o) < 1))) { #mag>(o+2) || removed
						vec <- vec / (10^o)
						ext <- paste0(" ", names(big.num.abbr)[i])
						break
					}
				}
			}

			if (interval.disjoint) {

				if (!digits_defined) {
					# update digits
					vec_fin = unique(vec[!is.infinite(vec)])
					vec_fin2 = c(vec_fin, vec_fin[-1] - 10^-digits)
					while (anyDuplicated(na.omit(vec_fin2))) {
						digits = digits + 1
						vec_fin2 = c(vec_fin, vec_fin[-1] - 10^-digits)
					}
				}
				# add to points
				vec = c(vec, head(vec, -1) - 10^-digits, tail(vec, 1))
			}

			# set default values
			if (!("big.mark" %in% names(args))) args$big.mark <- ","
			if (!("format" %in% names(args))) args$format <- "f"
			if (!("preserve.width" %in% names(args))) args$preserve.width <- "none"

			x = do.call("formatC", c(list(x=vec, digits=digits), args))
			if (as.count || interval.disjoint) {
				x[(n+2):(2*n)] = paste0(x[(n+2):(2*n)], ext)
			} else {
				x = paste0(x, ext)
			}
			x <- paste0(prefix, x, suffix)
		} else {
			if (!("format" %in% names(args))) args$format <- "g"
			x <- do.call("formatC", c(list(x=vec, digits=digits), args))
		}

		if (as.count) {
			x1 <- x[1:(n-1)]
			x2 <- x[(n+2):(2*n)]
			x1p1 <- x[(2*n+1):(3*n-1)]
		} else if (interval.disjoint) {
			x1 = x[1:(n-1)]
			x2 = x[(n+2):(2*n)]


		}
		# x <- formatC(vec, format = "f", digits = 0)
		# x1 <- x[-n]
		# x2 <- formatC(vec[-1] - 1L, format = "f", digits = 0)
		# xs <- (vec[-1] - vec[-n])
		# x1p1 <- formatC(vec[-n] + 1L, format = "f", digits = 0)
	}

	if (intervals) {
		if (scientific) {
			if (as.count) {
				# discrete
				lbls <- paste("{", x1, "}", sep = "")
				lbls[steps == 2] <- paste("{", x1[steps == 2], ", ", x2[steps == 2], "}", sep="")
				lbls[steps > 2] <- paste("{", x1[steps > 2], ", ", x1p1[steps > 2], ", ..., ", x2[steps > 2], "}", sep="")
			} else {
				# continuous
				if (interval.closure=="left") {
					lbls <- paste("[", x[-n], ", ", x[-1], ")", sep="")
					lbls[n-1] <- paste(substr(lbls[n-1], 1, nchar(lbls[n-1])-1), "]", sep="")
				} else {
					lbls <- paste("(", x[-n], ", ", x[-1], "]", sep="")
					lbls[1] <- paste("[", substr(lbls[1], 2, nchar(lbls[1])), sep="")
				}
			}


		} else {
			if (as.count) {
				lbls <- x1
				lbls[steps>1] <- paste(x1[steps>1], x2[steps>1], sep = paste0(" ", text.separator, " "))
				if (vec[n]==Inf) lbls[n-1] <- paste2(x1[n-1], paste(text.or.more, collapse = " "), flipped = text.or.more_as.prefix)
			} else if (interval.disjoint) {
				lbls <- paste(x1, x2, sep = paste0(" ", text.separator, " "))
				if (vec[1]==-Inf) lbls[1] <- paste2(x1[2], paste(text.less.than, collapse = " "), flipped = text.less.than_as.prefix)
				if (vec[n]==Inf) lbls[n-1] <- paste2(x1[n-1], paste(text.or.more, collapse = " "), flipped = text.or.more_as.prefix)
			} else {
				x[vec==-Inf] <- ""

				lbls <- paste(x[-n], x[-1], sep = paste0(" ", text.separator, " "))
				if (vec[1]==-Inf) lbls[1] <- paste2(x[2], paste(text.less.than, collapse = " "), flipped = text.less.than_as.prefix)
				if (vec[n]==Inf) lbls[n-1] <- paste2(x[n-1], paste(text.or.more, collapse = " "), flipped = text.or.more_as.prefix)
			}

			if (text.to.columns) {
				#xtra <- as.numeric(!is.na(text.align) && text.align=="right")


				nc1 <- nchar(paste(x[-n], " ", sep = "")) + 1
				nc2 <- rep(nchar(paste(text.separator, " ", sep = "")), n-1)

				lbls_breaks <- matrix(c(nc1, nc1+nc2), ncol=2)

				if (vec[1]==-Inf) {
					if (length(text.less.than)==1) {
						lbls_breaks[1,] <- rep(nchar(paste(text.less.than[1], " ", sep = "")) + 1, 2)
					} else {
						lbls_breaks[1,] <- cumsum(c(nchar(paste(text.less.than[1], " ", sep = "")) + 1, nchar(text.less.than[2])+1))
					}
				}
				if (vec[n]==Inf) {
					if (length(text.or.more)==1) {
						lbls_breaks[n-1,] <- rep(nchar(paste(x[n-1], " ", sep = "")) + 1, 2)
					} else {
						lbls_breaks[n-1,] <- cumsum(c(nchar(paste(x[n-1], " ", sep = "")) + 1, nchar(text.or.more[1])+1))
					}

				}
				attr(lbls, "brks") <- lbls_breaks
			}




		}
	}

	y <- if (intervals) lbls else x
	attr(y, "align") <- text.align
	y
}


num2breaks <- function(x, n, style, breaks, approx=FALSE, interval.closure="left", var = NULL, as.count = FALSE, args = list()) {

	tmapOptions = tmap_options_mode(mode.specific = FALSE)
	show.warnings <- tmapOptions$show.warnings

	nobs <- sum(!is.na(x))
	# create intervals and assign colors
	if (style=="fixed") {
		x2 = na.omit(x)
		tooLow = (x2 < min(breaks))
		tooHigh = (x2 > max(breaks))
		if (any(tooLow) && show.warnings) warning("Values have found that are less than the lowest break. They are assigned to the lowest interval", call. = FALSE)
		if (any(tooHigh) && show.warnings) warning("Values have found that are higher than the highest break. They are assigned to the highest interval", call. = FALSE)

		q <- list(var=x,
				  brks=breaks)
		attr(q, "style") <- "fixed"
		attr(q, "nobs") <- nobs
		attr(q, "intervalClosure") <- interval.closure
		class(q) <- "classIntervals"
	} else {
		if (nobs == 0) {
			if (!is.null(var)) {
				cli::cli_abort("Numerical variable {.var {var}} only contains missing values.")
			} else {
				cli::cli_abort("Numerical variable only contains missing values.")
			}
		}

		nunique <- length(na.omit(unique(x)))


		if (nunique == 1 && style != "pretty" && show.warnings) {
			if (is.null(var)) {
				bullet <- "Single value found."
			} else {
				bullet <- "Single unique value found for variable {.var {var}}"
			}
			cli::cli_warn(c(
				"!" = bullet,
				"i" = "Setting {.code style = {.str pretty}}"
			))
		}

		tempx <- nunique <= n

		if (tempx) {
			x_orig <- x
			if (length(na.omit(unique(x))) == 1) x <- pretty(x)
			x <- seq(min(x, na.rm = TRUE), max(x, na.rm = TRUE), length.out = n + 1)
		}

		q <- tryCatch({
			suppressWarnings(do.call(classInt::classIntervals, c(list(x, n, style = style, intervalClosure = interval.closure), args)))
		}, error = function(e) {
			cli::cli_abort(c(
				"Calculating interval classes failed for the variable {.var {var}} with {.code style = {.val {style}}}."
			    ),
			    parent = e
			)
		})



		if (tempx) q$var <- x_orig

	}

	if (approx && style != "fixed") {
		if (n >= length(unique(x)) && style == "equal") {
			# to prevent classIntervals to set style to "unique"
			q <- list(var = x, brks = seq(min(x, na.rm = TRUE), max(x, na.rm = TRUE), length.out = n))
			attr(q, "intervalClosure") <- interval.closure
			class(q) <- "classIntervals"
		} else {
			brks <- q$brks

			# to prevent ugly rounded breaks such as -.5, .5, ..., 100.5 for n=101
			qm1 <- suppressWarnings(do.call(classInt::classIntervals, c(list(x, n - 1, style = style, intervalClosure = interval.closure), args)))
			brksm1 <- qm1$brks
			qp1 <- suppressWarnings(do.call(classInt::classIntervals, c(list(x, n + 1, style = style, intervalClosure = interval.closure), args)))
			brksp1 <- qp1$brks
			if (min(brksm1) > min(brks) && max(brksm1) < max(brks)) {
				q <- qm1
			} else if (min(brksp1) > min(brks) && max(brksp1) < max(brks)) {
				q <- qp1
			}
		}
	}
	q
}


are_breaks_diverging = function(brks) {
	# if !divx then c-Inf, 2, 5, 10) is considered sequential
	negb <- any(brks[brks != -Inf] < 0) || (brks[1] == -Inf && brks[2]<=0)
	nb <- length(brks)
	posb <- any(brks[brks != Inf] > 0) || (brks[nb] == Inf && brks[nb-1]>=0)
	negb && posb
}








# function to determine whether a diverging of sequential palette is used given the values and the breaks
# use_diverging_palette <- function(v, brks, midpoint = NULL) {
# 	if (!is.null(midpoint) && !is.na(midpoint)) return(TRUE)
# 	x <- na.omit(v)
# 	divx <- any(x<0) && any(x>0)
#
# 	if (divx || is.null(brks)) {
# 		return(divx)
# 	} else {
# 		are_breaks_diverging(brks)
# 	}
# }
#
use_div = function(brks, midpoint = NULL) {
	if (!is.null(midpoint) && !is.na(midpoint)) return(TRUE)

	if (is.null(brks)) {
		return(NA)
	} else {
		are_breaks_diverging(brks)
	}
}



# https://stackoverflow.com/questions/12688717/round-up-from-5
#rnd <- function(x) trunc(x + sign(x) * 0.5)
round2 = function(x, n = 0) {
	posneg = sign(x)
	z = abs(x)*10^n
	z = z + 0.5 + sqrt(.Machine$double.eps)
	z = trunc(z)
	z = z/10^n
	z*posneg
}
# a = seq(0.5, 10.5, by = 0.5)
# names(a) = a
# round(a)
# round2(a, 0)
# rnd(a)


# round x to nearest multiple of y
round_num = function(x, y) round2(x / y) * y


# x = c(12.777156207494, 20.7613848999599, 28.7456135924257, 36.7298422848916, 44.7140709773574)
prettyTicks = function(x, dev = 0.1) {
	is_equi = local({
		dff = max(x) - min(x)
		if (dff == 0) return(TRUE)
		steps = (x[-1] - head(x,-1)) / dff
		all(abs(steps-steps[1]) < 1e-3)
	})

	if (is_equi) {
		pretty(x, n = length(x))
	} else {
		s = x[-1] - head(x,-1)
		s = c(s[1], s, tail(s,1))
		s = pmin(head(s,-1), tail(s,-1))

		mapply(function(xi, si) {
			for (r in rev(.TMAP$round_to)) {
				xir = round_num(xi, r)
				if (abs(xi - xir) < (si * dev)) break
			}
			xir
		}, x, s, SIMPLIFY = TRUE)	}
}
# x = c(3.654, 4.65, 5.1, 7.5)
# x = sort(c(0.004324, 0.00324, 0.00227, 0.00745435))
# x = c(3, 4,6,7,9)
# x = c(3.1,3.354,3.6)
# x = c(3.1,4.354,5.6)
# x = c(1043,2045,4005, 6765)
# x = c(265, 280, 520, 1000)
# x = c(265, 520, 1000, 1280)
# prettyTicks(x)

tmapLeafletArrange = function(tms, nx, ncol, nrow, opts, knit, show, args, options) {
	res = lapply(tms, function(tm) {
		print(tm, show = FALSE)
	})

	# Apply the requested height (interactive modes only). leafsync::latticeView
	# sizes its grid from the widgets themselves, so set each leaflet widget's
	# height. A numeric height (pixels) is the height of the WHOLE arrangement,
	# so split it across the rows; a string (e.g. "80vh") is applied per map.
	if (!is.null(opts$height)) {
		per = if (is.numeric(opts$height)) opts$height / nrow else opts$height
		res = lapply(res, function(w) {
			if (inherits(w, "htmlwidget")) w$height = per
			w
		})
	}

	res2 = do.call(leafsync::latticeView, c(res, list(ncol = ncol, sync = ifelse(identical(opts$sync, TRUE), "all", "none"), no.initial.sync = FALSE)))

	if (show) {
		if (knit) {
			kp = get("knit_print", asNamespace("knitr"))
			return(do.call(kp, c(list(x = res2), args, list(options = options))))
		} else {
			return(print(res2))
		}
	} else res2
}

#' @export
#' @rdname tmap_internal
tmapChartBinned = function(chart, bin_colors, breaks_def, na.show, x1) {
	if (is.numeric(x1)) {
		tmapChartBinned_numeric(chart, bin_colors, breaks_def, na.show, x1)
	} else {
		tmapChartBinned_categorical(chart, bin_colors, breaks_def, na.show, x1)
	}
}

#' @export
#' @rdname tmap_internal
tmapChartBinned2d = function(chart, chart1, chart2) {
	if (is.numeric(chart1$x1) && is.numeric(chart2$x1)) {
		tmapChartBinned2d_numnum(chart, chart1, chart2)
	} else if (is.numeric(chart1$x1) && !is.numeric(chart2$x1)) {
		tmapChartBinned2d_numcat(chart, chart1, chart2)
	} else if (!is.numeric(chart1$x1) && is.numeric(chart2$x1)) {
		res = tmapChartBinned2d_numcat(chart, chart2, chart1)
		names(res$tab) = names(res$tab)[c(2,1,3)]
		res
	} else {
		tmapChartBinned2d_catcat(chart, chart1, chart2)
	}
}

#' @export
#' @rdname tmap_internal
tmapChartBinned2d_numcat = function(chart, chart1, chart2) {
	res = bin_num(chart1$x1, chart1$breaks_def, chart1)
	chart$tab = as.data.frame(table(bin1 = res$xcat, bin2 = chart2$x1, useNA = "no"), responseName = "freq")
	chart$pal = NA
	chart$datatype = "categorized"
	chart$predefined = FALSE
	chart
}

#' @export
#' @rdname tmap_internal
tmapChartBinned2d_numnum = function(chart, chart1, chart2) {
	res1 = bin_num(chart1$x1, chart1$breaks_def, chart1)
	res2 = bin_num(chart2$x1, chart2$breaks_def, chart2)
	chart$tab = as.data.frame(table(bin1 = res1$xcat, bin2 = res2$xcat, useNA = "no"), responseName = "freq")
	chart$pal = NA
	chart$datatype = "categorized"
	chart$predefined = FALSE
	chart
}


#' @export
#' @rdname tmap_internal
tmapChartBinned2d_catcat = function(chart, chart1, chart2) {
	chart$tab = as.data.frame(table(bin1 = chart1$x1, bin2 = chart2$x1, useNA = "no"), responseName = "freq")
	chart$pal = NA
	chart$datatype = "categorized"
	chart$predefined = FALSE
	chart
}


#' @export
#' @rdname tmap_internal
tmapChartRaw = function(chart, na.show, x1, ...) {
	if (!na.show) x1 = na.omit(x1)

	chart$df = data.frame(x = 1L, y = x1)
	chart$datatype = "raw"
	chart$predefined = FALSE
	chart
}

#' @export
#' @rdname tmap_internal
tmapChartRaw_nna = function(chart, na.show, x1, ...) {
	x1 = na.omit(x1)

	chart$df = data.frame(x = 1L, y = x1)
	chart$datatype = "raw"
	chart$predefined = FALSE
	chart
}

#' @export
#' @rdname tmap_internal
tmapChartNone = function(chart, na.show, x1, ...) {
	chart
}

#' @export
#' @rdname tmap_internal
tmapChartPass = function(chart, na.show, x1, ...) {
	args = c(list(na.show = na.show, x1 = x1), list(...))
	chart[names(args)] = args
	chart
}

#' @export
#' @rdname tmap_internal
tmapChartBinned_categorical = function(chart, bin_colors, breaks_def, na.show, x1) {
	if (chart$type == "histogram") {
		message("histograms are supposed to be used for numerical data, a bar chart will be shown instead (tm_chart_bar)")
	}

	if (na.show) {
		tab = as.data.frame(table(bin = x1, useNA = "always"), responseName = "freq")
		tab$color = factor(1L:nrow(tab))
		pal = structure(bin_colors, names = levels(tab$color))
	} else {
		tab = as.data.frame(table(bin = x1, useNA = "no"), responseName = "freq")
		tab$color = factor(1L:nrow(tab))
		pal = structure(bin_colors, names = levels(tab$color))
	}
	chart$tab = tab
	chart$pal = pal
	chart$datatype = "categorized"
	chart$predefined = TRUE
	chart
}


bin_num = function(x1, breaks_def, chart) {
	# are breaks (and bin_colors)
	predefined = !is.null(breaks_def)

	if (is.null(chart$breaks_def)) {
		if (!predefined) {
			breaks = pretty(x1)
			ids = rep(1L, length(breaks) - 1)

		} else {
			breaks = breaks_def
			ids = 1L:(length(breaks) - 1L)
		}
	} else {
		breaks = chart$breaks_def
		subbreaks = (all(breaks_def %in% breaks))

		break_mids = (breaks[-1] + head(breaks, -1)) / 2

		if (predefined) {
			ids = as.integer(cut(break_mids, breaks_def, include.lowest = TRUE, right = FALSE))
		} else {
			ids = rep(1L, length(breaks) - 1)
		}

	}
	xcat = cut(x1, breaks = breaks, include.lowest = TRUE, right = FALSE)
	list(xcat = xcat, ids = ids)
}


#' @export
#' @rdname tmap_internal
tmapChartBinned_numeric = function(chart, bin_colors, breaks_def, na.show, x1) {

	# are breaks (and bin_colors)
	predefined = !is.null(breaks_def)

	res = bin_num(x1, breaks_def, chart)
	xcat = res$xcat
	ids = res$ids

	if (!predefined || (length(bin_colors) != (length(breaks_def) -1+na.show))) {
		bin_colors = chart$object.color
	}

	if (na.show) {
		tab = as.data.frame(table(bin=xcat, useNA = "always"), responseName = "freq")
		tab$color = factor(c(ids, length(bin_colors)), levels = seq_along(bin_colors))
		pal = structure(bin_colors, names = levels(tab$color))
	} else {
		tab = as.data.frame(table(bin=xcat, useNA = "no"), responseName = "freq")
		tab$color = factor(ids, levels = seq_along(bin_colors))
		pal = structure(bin_colors, names = levels(tab$color))
	}
	chart$tab = tab
	chart$pal = pal
	chart$datatype = "binned"
	chart$predefined = predefined
	chart
}

#' @export
tmapShape.Raster = function(shp, is.main, crs, bbox, unit, filter, layer, shp_name, smeta, o, tmf) {
	tmapShape.SpatRaster(terra::rast(shp), is.main, crs, bbox, unit, filter, shp_name, smeta, o, tmf)
}

#' @export
tmapSubsetShp.Raster = function(shp, vars) {
	tmapSubsetShp.SpatRaster(terra::rast(shp), vars)
}

#' @export
tmapGetShapeMeta1.Raster = function(shp, layer, o) {
	tmapGetShapeMeta1.SpatRaster(terra::rast(shp), o)
}

select_band = function(x, band) {
	nms = stars::st_get_dimension_values(x, "band")
	if (!(band %in% nms)) stop("band not found")
	r = x[, , , which(band == nms), drop = TRUE]
	names(r) = band
	r
}

trim_stars = function(x) {
	dim1 = apply(x[[1]], 1, function(x) !all(is.na(x)))
	dim1 = which(dim1)
	dim1 = dim1[1]:dim1[length(dim1)]
	dim2 = apply(x[[1]], 2, function(x) !all(is.na(x)))
	dim2 = which(dim2)
	dim2 = dim2[1]:dim2[length(dim2)]
	x = x[, dim1, dim2]
	x = sf::st_normalize(x)
	return(x)
}

# from stars package
is_regular_grid = function (x) {
	has_raster(x) && !(has_rotate_or_shear(x) || is_rectilinear(x) ||
					   	is_curvilinear(x))
}

has_raster = function (x) {
	if (inherits(x, "stars"))
		x = stars::st_dimensions(x)
	!is.null(r <- attr(x, "raster")) && all(r$dimensions %in%
												names(x))
}

has_rotate_or_shear = function (x) {
	dimensions = stars::st_dimensions(x)
	if (has_raster(x)) {
		r = attr(dimensions, "raster")
		!anyNA(r$affine) && any(r$affine != 0)
	}
	else FALSE
}

is_curvilinear = function (x) {
	d = stars::st_dimensions(x)
	has_raster(x) && isTRUE(attr(d, "raster")$curvilinear)
}

is_rectilinear = function (x) {
	d = stars::st_dimensions(x)
	if (has_raster(x) && !is_curvilinear(x)) {
		xy = attr(d, "raster")$dimensions
		dimx = d[[xy[1]]]
		dimy = d[[xy[2]]]
		(is.na(dimx$delta) || is.na(dimy$delta)) && (!regular_intervals(dimx$values) ||
													 	!regular_intervals(dimy$values))
	}
	else FALSE
}

regular_intervals = function (x, epsilon = 1e-10) {
	if (length(x) <= 1) {
		FALSE
	} else {
		ud = if (is.atomic(x))
			unique(diff(x))
		else {
			if (identical(tail(x$end, -1), head(x$start, -1)))
				x$end - x$start
			else return(FALSE)
		}
		abs(diff(range(ud))/mean.default(ud)) < epsilon
	}
}

get_downsample = function(dims, px = round(dev.size("px") * (graphics::par("fin")[1] / dev.size()[1]))) {
	floor(sqrt(prod(dims) / prod(px)))
}

# st_downsample = function (x, n, fill_out = TRUE) {
# 	stopifnot(all(n >= 0))
# 	d = dim(x)
# 	n = rep(n, length.out = length(d))
# 	dims = stars::st_dimensions(x)
# 	regular = is_regular_grid(x)
# 	if (!all(n <= 1)) {
# 		args = rep(list(rlang::missing_arg()), length(d) + 1)
# 		for (i in seq_along(d)) {
# 			if (n[i] > 1) {
# 				sq = seq(1, d[i], n[i])
# 				args[[i + 1]] = sq
# 				if (!is.null(dims[[i]]$values))
# 					dims[[i]]$values = dims[[i]]$values[sq]
# 			}
# 		}
# 		x = eval(rlang::expr(x[!!!args]))
# 		if (fill_out && regular) {
# 			d_new = stars::st_dimensions(x)
# 			for (i in seq_along(d)) {
# 				dims[[i]]$delta = dims[[i]]$delta * n[i]
# 				dims[[i]]$from = d_new[[i]]$from
# 				dims[[i]]$to = d_new[[i]]$to
# 			}
# 			x = structure(x, dimensions = dims)
# 		}
# 	}
# 	x
# }

############# other functions


# st_is_merc = function(x) {
# 	crs = sf::st_crs(x)
# 	if (is.na(crs)) {
# 		NA
# 	} else {
# 		isTRUE(crs$proj == "merc") || isTRUE(crs$epsg == 3857)
# 	}
# }

get_xy_dim = function(x) {
	d = stars::st_dimensions(x)
	dxy = attr(d, "raster")$dimensions
	dim(x)[dxy]
}

transwarp = function(x, crs, raster.warp) {
	# NOTE: dropped colors after st_warp fixed in stars 0.4-2
	shpcolors = attr(x[[1]], "colors")
	if (raster.warp) {
		y = tryCatch({
			if (is_curvilinear(x)) {
				sf::st_transform(x, crs = crs)
			} else {
				z = stars::st_warp(x, crs = crs)
				.TMAP$raster_wrap = TRUE
				z
			}
		}, error = function(e) {
			cli::cli_warn(c("!" = "Unable to warp stars. Stars will be transformed now (which will take some time)."))
			tryCatch({
				sf::st_transform(x, crs = crs)
			}, error = function(e) {
				cli::cli_abort("Unable to warp or transform stars", call = NULL)
			})
		})
	} else {
		y = sf::st_transform(x, crs = crs)
	}

	if (!is.null(shpcolors)) attr(y[[1]], "colors") = shpcolors
	y
}

has_rotate_or_shear = function (x)
{
	dimensions = stars::st_dimensions(x)
	if (has_raster(x)) {
		r = attr(dimensions, "raster")
		!anyNA(r$affine) && any(r$affine != 0)
	}
	else FALSE
}

do_trans = function(tdt, FUN, shpDT, plot.order, args, scale) {

	shpDT = copy(shpDT)

	# copy by columns from tdt that do not yet exist in shpDT
	t_by = names(tdt)[substr(names(tdt), 1, 2) == "by"]
	s_by = names(shpDT)[substr(names(shpDT), 1, 2) == "by"]
	n_by = setdiff(t_by, s_by)
	if (length(n_by)) {
		shpDT[, (n_by) := tdt[1, n_by, with = FALSE]]
	}

	aesvars = setdiff(names(tdt), c("tmapID__", paste0("by", 1:3, "__")))

	apply_trans = function(shpTM) {
		# todo: stars
		ids = intersect(shpTM$tmapID, tdt$tmapID__)

		shp = shpTM$shp[match(ids, shpTM$tmapID)]

		shpX = list(shp = shp, tmapID = ids, bbox = shpTM$bbox)

		x = as.list(tdt[match(tmapID__, ids), aesvars, with = FALSE])

		res = do.call(FUN, c(list(shpTM = shpX), x, list(plot.order = plot.order, args = args, scale = scale)))
	}
	shpDT$shpTM = lapply(shpDT$shpTM, apply_trans)
	list(shpDT)
}

delta_per_lineheight = function(x, n = 20, scale = 1) {
	b = unname(sf::st_bbox(x))
	(b[4] - b[2]) / n * scale
}

get_midpoint_angle = function(shp, bbx) {
	lShp = sf::st_cast(shp, "MULTILINESTRING")

	coor = sf::st_coordinates(lShp)
	coors = split.data.frame(coor[,1:2], f = coor[, ncol(coor)], drop=FALSE)
	co = do.call(rbind, lapply(coors, get_midpoint))
	pShp = sf::st_as_sf(as.data.frame(co), coords = c("X", "Y"), crs = sf::st_crs(shp))

	deltax = bbx[3] - bbx[1]
	deltay = bbx[4] - bbx[2]
	delta = max(deltax, deltay)

	pbShp = sf::st_cast(sf::st_geometry(sf::st_buffer(pShp, dist = delta / 100)), "MULTILINESTRING")

	iShps <- mapply(function(x,y,z) {
		res = sf::st_intersection(x,y)
		if (!sf::st_is_empty(res)) {
			sf::st_cast(res, "MULTIPOINT")
		} else {
			z
		}
	}, lShp, pbShp,pShp, SIMPLIFY = FALSE)

	angles = vapply(iShps, function(x) {
		if (length(x) == 0) 0 else .get_direction_angle(sf::st_coordinates(x)[,1:2, drop = FALSE])
	}, numeric(1))

	angles[angles > 90 & angles < 270] = angles[angles > 90 & angles < 270] - 180

	list(shp = sf::st_geometry(pShp), angles = angles)
}

sf_expand <- function(shp, cast = c("MULTILINESTRING", "MULTIPOLYGON", "MULTIPOINT")) {
	x = mapply(function(tp, ge) {
		if (tp %in% cast) {
			if (tp == "MULTILINESTRING") {
				sf::st_cast(st_geometry(ge), "LINESTRING")
			} else if (tp == "MULTIPOLYGON") {
				sf::st_cast(st_geometry(ge), "POLYGON")
			} else if (tp == "MULTIPOINT") {
				sf::st_cast(st_geometry(ge), "POINT")
			} else {
				sf::st_geometry(ge)
			}
		} else {
			sf::st_geometry(ge)
		}
	}, sf::st_geometry_type(shp), sf::st_geometry(shp), SIMPLIFY = FALSE)
	ids = rep(1L:length(x), vapply(x, length, integer(1)))
	shp3 = sf::st_sf(geometry=sf::st_sfc(do.call(c, x), crs = sf::st_crs(shp)))
	shp3$split__id = ids
	shp3
}

#' @param shpTM shpTM
#' @param xmod,ymod xmod and ymod
#' @param ord__ ord
#' @param plot.order plot.order
#' @param args args
#' @param scale scale
#' @export
#' @rdname tmap_internal
tmapTransCentroid = function(shpTM, xmod = NULL, ymod = NULL, ord__, plot.order, args, scale) {
	within(shpTM, {
		if (!is.character(shp)) {

			is_stars = inherits(shp, "dimensions")
			if (is_stars && args$points_only == "no") {
				### stars

				s = structure(list(values = matrix(TRUE, nrow = nrow(shp))), dimensions = shp, class = "stars")
				strs = stars::st_as_stars(list(values = m), dimensions = shp)
				shp = sf::st_as_sfc(s, as_points = TRUE)
			} else if (is_stars) {
				shp = sf::st_sfc()
				tmapID = integer(0L)
			} else {
				geom_all = sf::st_geometry_type(shp, by_geometry = FALSE)
				if (geom_all == "GEOMETRY") {
					geom_types = sf::st_geometry_type(shp)
				} else {
					geom_types = rep(geom_all, length(shp))
				}


				if (any(geom_types %in% c("MULTILINESTRING", "MULTIPOINT", "MULTIPOLYGON"))) {
					if (args$point_per=="feature" && any(geom_types == "MULTIPOINT")) {
						shp = sf_expand(shp, cast = "MULTIPOINT")
						tmapID = tmapID[shp$split__id]
						shp = sf::st_geometry(shp)
					} else if (args$point_per=="segment") {
						shp = sf_expand(shp)
						tmapID = tmapID[shp$split__id]
						shp = sf::st_geometry(shp)
					} else if (args$point_per == "largest") {
						ids_multiline = which(geom_types == "MULTILINESTRING")
						if (length(ids_multiline)) {
							shp[ids_multiline] = local({
								shp_ml = sf_expand(shp[ids_multiline])
								lengths <- sf::st_length(shp_ml)
								id_max <- vapply(split(lengths, f=shp_ml$split__id), which.max, integer(1))
								id_max2 <- vapply(seq_along(ids_multiline), function(i) {
									which(shp_ml$split__id==i)[id_max[i]]
								}, integer(1))
								sf::st_geometry(shp_ml[id_max2, ])
							})
						}
					}
				}

				geom_all2 = sf::st_geometry_type(shp, by_geometry = FALSE)
				if (geom_all2 == "GEOMETRY") {
					geom_types2 = sf::st_geometry_type(shp)
				} else {
					geom_types2 = rep(geom_all, length(shp))
				}

				ids_poly = which(geom_types2 %in% c("POLYGON", "MULTIPOLYGON"))
				ids_line = which(geom_types2 %in% c("LINESTRING", "MULTILINESTRING"))
				ids_point = which(geom_types2 %in% c("POINT", "MULTIPOINT"))

				if (args$points_only == "yes" || (args$points_only == "ifany" && length(ids_point))) {
					shp = shp[ids_point]
					tmapID = tmapID[ids_point]
				} else {
					if (length(ids_line)) {
						if (args$along_lines) {
							bb_crs = stm_bbox(shpTM, shpTM$tmapID, crs = sf::st_crs(shp))
							res = get_midpoint_angle(shp[ids_line], bbx = bb_crs)
							shp[ids_line] = res$shp
							prop_angle = rep(0, length(shp))
							prop_angle[ids_line] = res$angles
							rm(res)
						} else {
							shp[ids_line] = suppressWarnings({
								sf::st_centroid(shp[ids_line])
							})
						}
					}
					if (length(ids_poly)) {
						ctds = suppressWarnings({
							sf::st_centroid(shp[ids_poly], of_largest_polygon = (args$point_per == "largest"))
						})

						if (args$on_surface) {
							ctds_in = unlist(mapply(function(x,y) {
								length(sf::st_contains(y,x)[[1]]) > 0L
							}, ctds, shp[ids_poly], SIMPLIFY = FALSE, USE.NAMES = FALSE))


							if (!all(ctds_in)) {
								shp[ids_poly[!ctds_in]] = sf::st_point_on_surface(shp[ids_poly[!ctds_in]])
								shp[ids_poly[ctds_in]] = ctds[ctds_in]
							} else {
								shp[ids_poly] = ctds
							}
						} else {
							shp[ids_poly] = ctds
						}
						rm(ctds)
					}
				}

				if (!is.null(xmod) || !is.null(ymod)) {
					shp = local({
						d = delta_per_lineheight(shp, scale = scale)

						if (is.null(xmod)) xmod = rep(0, length(shp))
						if (is.null(ymod)) ymod = rep(0, length(shp))
						mod = mapply(c, xmod * d, ymod * d, SIMPLIFY = FALSE)
						shp + mod
					})
				}

				# in case of expansion: cast to multipoint (expanded again in step4)

				if (args$point_per=="segment" || (args$point_per=="feature" && any(geom_types == "MULTIPOINT"))) {
					tmapID_expanded = tmapID
					u = sort(unique(tmapID))
					tmapID_1n = match(tmapID, u)

					shp = sf::st_cast(shp, "MULTIPOINT", ids = tmapID_1n)
					tmapID = u

					if (length(ids_line) && args$along_lines) {
						prop_angle = split(prop_angle, f = tmapID_1n)
					}

					rm(u)
					rm(tmapID_1n)
				}

				rm(geom_types, geom_types2, ids_point, ids_line, ids_poly, geom_all, geom_all2, is_stars)
			}
		}
	})
}
# args:
# - points_only: "yes", "no", "ifany"
#
# prop_ vectors, e.g. prop_angle can be added to shpTM. These are later put into dt (and used in step 4 (plotting))


#' @export
#' @rdname tmap_internal
tmapTransRaster = function(shpTM, ord__, plot.order, args) {
	if (!is.character(shpTM$shp) && !inherits(shpTM$shp, "dimensions")) cli::cli_abort("Stars or terra object expected in {.fun tm_shape} for {.fun tm_raster}")
	shpTM
}


#' @export
#' @rdname tmap_internal
tmapTransPolygons = function(shpTM, ord__, plot.order, args, scale) {
	within(shpTM, {
		if (!is.character(shp)) {
			is_stars = inherits(shp, "dimensions")
			if (is_stars && args$polygons.only != "yes") {
				### stars
				s = structure(list(values = matrix(TRUE, nrow = nrow(shp), ncol = ncol(shp))), dimensions = shp, class = "stars")
				shp = sf::st_as_sfc(s, as_points = FALSE)
			} else if (is_stars) {
				shp = sf::st_sfc()
				tmapID = integer(0)
			} else {

				### sf
				geom_types = sf::st_geometry_type(shp)
				#crs = sf::st_crs(shp)

				ids_poly = which(geom_types %in% c("POLYGON", "MULTIPOLYGON"))
				ids_line = which(geom_types %in% c("LINESTRING", "MULTILINESTRING"))
				ids_point = which(geom_types %in% c("POINT", "MULTIPOINT"))


				if (args$polygons.only == "yes" || (args$polygons.only == "ifany" && length(ids_poly))) {
					shp = shp[ids_poly]
					tmapID = tmapID[ids_poly]
				} else {
					if (length(ids_line)) {
						tryCatch({
							shp[ids_line] = sf::st_cast(sf::st_cast(shp[ids_line], "MULTILINESTRING"), "MULTIPOLYGON")
						}, error = function(e) {
							stop("Unable to cast lines to polygon. Error from st_cast: \"", e$message, "\"", call. = FALSE)
						})
					}
					if (length(ids_point)) {
						dist = if (sf::st_is_longlat(shp)) 0.01 else 100
						shp[ids_point] = sf::st_buffer(shp[ids_point], dist = dist)
					}

				}
				rm(geom_types)
			}

			if (plot.order$aes == "AREA" && !is_stars) {
				o = order(without_units(sf::st_area(shp)), decreasing = !plot.order$reverse)
				shp = shp[o]
				tmapID = tmapID[o]
			}
		}
	})
}
# args:
# - polygons.only: "yes", "no", "ifany"

#' @export
#' @rdname tmap_internal
tmapTransLines = function(shpTM, ord__, plot.order, args, scale) {
	within(shpTM, {
		if (!is.character(shp)) {
			is_stars = inherits(shp, "dimensions")
			if (is_stars) {
				shp = sf::st_sfc()
				tmapID = integer(0)
			} else {

				### sf
				geom_types = sf::st_geometry_type(shp)
				#crs = sf::st_crs(shp)

				ids_poly = which(geom_types %in% c("POLYGON", "MULTIPOLYGON"))
				ids_line = which(geom_types %in% c("LINESTRING", "MULTILINESTRING"))
				ids_point = which(geom_types %in% c("POINT", "MULTIPOINT"))


				if (args$lines.only == "yes" || (args$lines.only == "ifany" && length(ids_line))) {
					shp = shp[ids_line]
					tmapID = tmapID[ids_line]
				} else {
					if (length(ids_poly)) {
						tryCatch({
							shp[ids_poly] = sf::st_cast(sf::st_cast(shp[ids_poly], "MULTIPOLYGON"), "MULTILINESTRING")
						}, error = function(e) {
							stop("Unable to cast to polygon. Error from st_cast: \"", e$message, "\"", call. = FALSE)
						})
					}
					if (length(ids_point)) {
						ids_not_point = which(!(geom_types %in% c("POINT", "MULTIPOINT")))
						shp = shp[ids_not_point]
						tmapID = tmapID[ids_not_point]
					}

				}
				rm(geom_types)
			}

			if (plot.order$aes == "LENGTH") {
				o = order(without_units(sf::st_length(shp)), decreasing = !plot.order$reverse)
				shp = shp[o]
				tmapID = tmapID[o]
			}
		}
	})
}
# args:
# - lines.only: "yes", "no", "ifany"

#' Specify an animation
#'
#' Specify an animation from a tmap plot. This is similar to creating facets with [tm_facets()]. The animation subsequently can be exported to a gif or video file (e.g. mp4) with [tmap_animation()]. If the tmap plot with [tm_animate()] is printed, the animation will be previewed. The default [tm_animate()] will show the individual frames slowly (frame per seconds (fps) set to 2) whereas [tm_animate_fast()] will show them like a movie (with a fps set to 24).
#'
#' @param frames group by variable used to create the animation frames. This is similar to the `by` argument of [tm_facets_wrap()]. Instead of showing facets next to each other, they are shown as animation frames. However, under the hood `frames` will be used to specify `pages` of [tm_facets()]. This makes it possible to create an animation of regular facets.
#' @param nframes number of animation frames. So far, this only applied experimentally in transition map variables. See the extension package tmap.cartogram.
#' @param fps frames per second. Default: 30 for `tm_facets_animate` and 2 for `tm_facets_animate_slow`.
#' @param play how should the animation be played? One of `"loop"` (default), `"pingpong"`, and `"once"`, where `"loop"` means that the animation will loop indefinitely, `"pingpong"` means that it will play forward and then backward, and `"once"` means that it will play only once.
#' @param dpr device pixel ratio. The ratio between the physical pixel density of a device and its logical pixel density.
#' @param ... passed on to [tm_facets()]. Note that for animated facets, `by` can be specified to create animated facet wraps, and `rows` and `cols` to created animated facet grids.
#' @export
#' @note In older versions (< 4.1) [tm_facets()] with `page` specification was used to create animations frames and [tmap_animation()] to create the animation itself using inputs like the frame rate specification. As of version 4.2, the whole animation, including frame rate, is specified in [tm_animate()]. The animation can still be saved via [tmap_animation()].
#' @seealso `r .doc_see_also_animations()`
#' @rdname tm_animate
#' @example examples/tm_animate.R
tm_animate_fast = function(frames = "VARS__",
							 nframes = 60L,
							 fps = 24L,
							 play = c("loop", "pingpong", "once"),
							 dpr = 2,
							 ...) {
	args = list(...)
	args_called = names(rlang::call_match()[-1])

	play = match.arg(play)

	tm = do.call("tm_facets", c(list(pages = frames, type = NA, animate = TRUE, nframes = nframes, fps = fps, play = play, dpr = dpr), args[setdiff(names(args), "type")]))
	tm[[1]]$calls = args_called
	class(tm[[1]]) = c("tm_animate", class(tm[[1]]))
	tm
}

#' @export
#' @rdname tm_animate
tm_animate = function(frames = "VARS__",
								  nframes = 60L,
								  fps = 2L,
								  play = c("loop", "pingpong", "once"),
								dpr = 2,
								  ...) {
	args = list(...)
	args_called = names(rlang::call_match()[-1])

	play = match.arg(play)

	tm = do.call("tm_facets", c(list(pages = frames, type = NA, animate = TRUE, nframes = nframes, fps = fps, play = play, dpr = dpr), args[setdiff(names(args), "type")]))
	tm[[1]]$calls = unique(c(args_called, "frames"))
	class(tm[[1]]) = c("tm_animate", class(tm[[1]]))
	tm
}

#' @export
#' @rdname tmapGridLeaflet
tmapGridDataPlot.tm_data_polygons = function(a, shpTM, dt, gp, bbx, facet_row, facet_col, facet_page, id, pane, group, glid, o, ...) {

	rc_text = frc(facet_row, facet_col)

	res = select_sf(shpTM, dt)
	shp = res$shp

	shp_is_pointer = inherits(shp, "character")

	if (shp_is_pointer) {
		cli::cli_inform("{.field plot mode} polygons of pmtiles are not supported in plot mode yet")
		return(NULL)
	}


	dt = res$dt
	gp = impute_gp(gp, dt)
	gp = rescale_gp(gp, o$scale_down)

	# none should contain NA's && (length or content should be different)
	diffAlpha = !anyNA(c(gp$fill_alpha, gp$col_alpha)) && !(length(gp$fill_alpha) == length(gp$col_alpha) && all(gp$fill_alpha == gp$col_alpha))


	if (diffAlpha) {
		gp1 = gp_to_gpar(gp, sel = "fill", o = o, type = "polygons")
		gp2 = gp_to_gpar(gp, sel = "col", o = o, type = "polygons")
		grb1 = sf::st_as_grob(shp, gp = gp1, name = paste0("polygons_", id))
		grb2 = sf::st_as_grob(shp, gp = gp2, name = paste0("polygon_borders_", id))
		grb = grid::grobTree(grb1, grb2)
	} else {
		gp = gp_to_gpar(gp, sel = "all", o = o, type = "polygons")
		grb = sf::st_as_grob(shp, gp = gp, name = paste0("polygons_", id))
	}

	gts = get("gts", .TMAP_GRID)
	gt_name = paste0("gt_facet_", rc_text)

	if (!is.null(a$blend) && a$blend != "over") {
		existing = gts[[facet_page]]$children$gt_main$children[[gt_name]]$children
		dst = do.call(grid::grobTree, existing)
		grb = blend_grobs(grb, a$blend, dst = dst)
	}

	gt = grid::addGrob(gts[[facet_page]], grb, gPath = grid::gPath(gt_name))
	gts[[facet_page]] = gt
	assign("gts", gts, envir = .TMAP_GRID)

	NULL
}


#' @export
#' @rdname tmapGridLeaflet
tmapGridDataPlot.tm_data_fill = function(a, shpTM, dt, gp, bbx, facet_row, facet_col, facet_page, id, pane, group, glid, o, ...) {
	NextMethod()
}


#' @export
#' @rdname tmapGridLeaflet
tmapGridDataPlot.tm_data_borders = function(a, shpTM, dt, gp, bbx, facet_row, facet_col, facet_page, id, pane, group, glid, o, ...) {
	NextMethod()
}

get_option_class = function(o, class = NULL, spatial_class = TRUE) {
	is_spatial = !spatial_class || (any(names(o) %in% c("stars", "sf", "sfc", "raster", "terra", "sp", "dimensions")))
	if (!is.null(class) && is_spatial) { # && is.list(o)
		mtch = which(names(o) %in% class)
		if (!length(mtch)) mtch = which(names(o) == "")[1]
		o = o[[mtch]]
	}
	o
}




# add/merge options x to the full option set o: x can be style options
complete_options = function(x, o, erase_style = TRUE) {
	nmx = names(x)
	nmo = names(o)
	if ("calls" %in% nmx) {
		o$calls = unique(c(x$calls, o$calls))
		x$calls = NULL
		nmx = setdiff(nmx, "calls")
	}
	if (length(x) == 0L) return(o)
	if (is.null(nmo) || is.null(nmx)) return(x)
	d = setdiff(nmx, nmo)
	e = intersect(nmx, nmo)
	if (length(d)) o = c(o, x[d])
	if (length(e)) {
		for (i in e) {
			if (i %in% c("value.const", "value.na", "value.null", "value.blank", "values.var") && erase_style) {
				# special case to cover the following issue
				#    if o = list(value.const = list(fill = "red", fill.polygons = "blue", fill.dots = "black)), and
				#       x = list(value.const = list(fill = "white", fill.polygons = "grey"))
				#    the new option set should be x (so dot fill color should be white)
				o[[i]] = complete_value_list(x[[i]], o[[i]])
			} else {
				if (i != "") {
					tmp = complete_options(x[[i]], o[[i]])
					if (is.null(tmp)) {
						o[i] = list(NULL)
					} else {
						o[[i]] = tmp
					}
				}else {
					o[[which("" == nmo)]] = x[[which("" == nmx)]]
				}
			}

		}
	}
	o
}

complete_value_list = function(x, o) {
	aes_x = gsub("\\..*", "", names(x))
	aes_o = gsub("\\..*", "", names(o))

	aes_o_not_x = setdiff(aes_o, aes_x)

	c(x, o[aes_o %in% aes_o_not_x])
}




getAesOption = function(x, o, aes, layer, cls = NULL) {
	y = o[[x]]
	al = paste(aes, layer, sep = ".")



	if (any(al %in% names(y))) {
		id = which(al %in% names(y))[1] # take first, most specific layer, e.g. when layer = c("dots", "symbols"), take dots if exists
		z = y[[al[id]]]
	} else if (aes %in% names(y)) {
		# take matching visual variable (regardless what layer)
		z = y[[aes]]
	} else if (is.list(y)) {
		# check if there are non-named list items, if so take the first one
		eid = which(names(y) == "")[1]
		if (!is.na(eid)) {
			z = y[[eid]]
		} else {
			return(NA)
		}
	} else {
		return(y)
	}

	if (!is.null(cls) && is.list(z)) {
		mid = vapply(names(z), FUN = "%in%", FUN.VALUE = logical(1), cls)
		if (any(mid)) {
			z = z[[which(mid)[1]]]
		}
	}
	z
}

getAesValue = function(x, aes) {
	nms = names(x)

	if (is.null(nms)) {
		x
	} else if (any(nms %in% c("fill", "col", "size", "shape", "lwd", "lty", "fontsize", "fontface"))) {
		if (aes %in% nms) {
			x[[aes]]
		} else {
			if (any(nms == "")) {
				x[[which(nms == "")[1]]]
			} else {
				x
			}
		}
	} else {
		x
	}
}


# get options with a prefic
get_prefix_opt = function(prefix, class, o) {
	if (missing(prefix)) prefix = substr(class, 4, nchar(class))
	prefixdot = paste0(prefix, ".")
	ot = o[names(o)[substr(names(o), 1, nchar(prefixdot)) == prefixdot]]
	names(ot) = substr(names(ot), nchar(prefix)+2, nchar(names(ot)))
	ot
}

# (partly) named vector: get 1st name match or otherwise 1st non-named argument
# used in tm_scale_continuous, but similar function should exists for options? (todo: check)
get_vector_id = function(x, id) {
	if (is.null(names(x))) {
		x[1]
	} else if (id %in% names(x)) {
		unname(x[id][1L])
	} else if (any("" %in% names(x))) {
		unname(x[which(names(x) == "")[1]])
	} else {
		x[1]
	}
}

# raster.max_cell can be mode specific: e.g. c(plot = 3000, view = 1000, 1000),
max_cells = function(raster.max_cells) {
	mode = getOption("tmap.mode")

	if (mode %in% names(raster.max_cells)) {
		raster.max_cells[mode]
	} else {
		raster.max_cells[length(raster.max_cells)]
	}
}

get_facet_id = function(row, col, nrow, ncol) {
	col + (row - 1L) * ncol
}

get_lf = function(facet_row, facet_col, facet_page) {
	lfs = get("lfs", envir = .TMAP_LEAFLET)
	nrow = get("nrow", envir = .TMAP_LEAFLET)
	ncol = get("ncol", envir = .TMAP_LEAFLET)

	lfsi = lfs[[facet_page]]

	fr = max(1, facet_row) # facet_row can be -1 or -2
	fc = max(1, facet_col) # facet_row can be -1 or -2

	lfid = get_facet_id(fr, fc, nrow, ncol)

	lfsi[[lfid]]
}

assign_lf = function(lf, facet_row, facet_col, facet_page) {
	lfs = get("lfs", envir = .TMAP_LEAFLET)
	nrow = get("nrow", envir = .TMAP_LEAFLET)
	ncol = get("ncol", envir = .TMAP_LEAFLET)


	fr = max(1, facet_row) # facet_row can be -1 or -2
	fc = max(1, facet_col) # facet_row can be -1 or -2



	lfid = get_facet_id(fr, fc, nrow, ncol)

	lfs[[facet_page]][[lfid]] = lf
	assign("lfs", lfs, envir = .TMAP_LEAFLET)
	NULL
}

getClusterOpts = function(clustering) {
	if (identical(clustering, TRUE)) {
		clusterOpts = leaflet::markerClusterOptions()
	} else if (identical(clustering, FALSE)) {
		clusterOpts = NULL
	} else {
		clusterArgs = setdiff(names(formals(leaflet::markerClusterOptions)), "...")
		if (!all(clusterArgs %in% names(clustering))) cli::cli_abort("{.field clustering} Cluster options should have the same options as {.fun leaflet::markerClusterOptions}")
		clusterOpts = clustering
	}
}


tmapLeafletWrap = function(label, facet_row, facet_col, facet_page, o) {

	NULL
}

tmapLeafletXtab = function(label, facet_row, facet_col, facet_page, o) {
}

#' Set the map projection (CRS)
#'
#' This function sets the map projection. It can also be set via [tm_shape()], but `tm_crs` is generally recommended for most cases. It can also be determined automatically (see details); however, this is still work-in-progress.
#'
#' @param crs Map projection (CRS). Can be set to an `crs` object (see [sf::st_crs()]), a proj4string, an EPSG number, the value `"auto"` (automatic crs recommendation), or one the the following generic projections: `c("laea", "aeqd", "utm", "pconic", "eqdc", "stere")`. See details.
#' @param property Which property should the projection have? One of: `"global"`, `"area"` (equal-area), `"distance"` (equidistant), `"shape"` (conformal). Only applicable if `crs = "auto"`. See details.
#' @param bbox bounding box. Three options: a [sf::st_bbox()] object, an Open Street Map query (passed on to [tmaptools::geocode_OSM()]), or `"FULL"`, which means the whole earth , which means the whole earth (this also guarantees that transformations to another CRS keep the whole earth, unlike \code{\link[sf:st_bbox]{sf::st_bbox()}}).
#' @param ... passed on to \code{\link[tmaptools:bb]{bb}} (e.g. \code{ext} can be used to enlarge or shrink a bounding box)
#' @note Plans are to migrate the functionality regarding generic crs and automatic crs recommendation to a separate package.
#' @inherit tm_shape details
#' @example ./examples/tm_crs.R
#' @seealso `r .doc_see_also_shape()`
#' @export
tm_crs = function(crs = NA, property = NA, bbox = NULL, ...) {
	bb_args = list(...)
	if (is.na(crs)) {
		if (is.na(property)) {
			return(tm_options(bbox = bbox, bbox_args = bb_args))
		} else {
			crs = "auto"
		}
	}
	if (identical(crs, "auto") && !is.na(property)) {
		if (!(property %in% c("global", "area", "distance", "shape"))) {
			message_crs_property_unknown()
		}
		extra = property
	} else {
		extra = ""
	}

	if (is.character(crs) && crs %in% c("laea", "aeqd", "utm", "pconic", "eqdc")) {
		extra = crs
		crs = "auto"
	}

	tm_options(crs = crs, crs_extra = extra, bbox = bbox, bbox_args = bb_args)
}


consider_global = function(x, th = 0.6) {
	b = sf::st_bbox(x)
	# in case margins are applied
	if (sf::st_is_longlat(b)) {
		b["xmin"] = max(b["xmin"], -180)
		b["xmax"] = min(b["xmax"], 180)
		b["ymin"] = max(b["ymin"], -90)
		b["ymax"] = min(b["ymax"], 90)
	}
	if (b$xmin == b$xmax || b$ymin == b$ymax) return(FALSE)
	earth_surface = 5.1e14
	area = b |>
		sf::st_as_sfc() |>
		sf::st_area() |>
		as.numeric()
	area > (earth_surface * 0.6)
}

auto_crs = function(x, crs_extra, crs_global) {

	if (identical(x, TRUE)) return(crs_global)




    # check if global should be used
	is_global = if (crs_extra == "global") {
		TRUE
	} else if (crs_extra != "") {
		FALSE
	} else {
		consider_global(x)
	}

	y = if (is_global) {
		crs_global
	} else {
		# impute family
		proj = switch(crs_extra,
			area = "laea",
			distance = "aeqd",
			shape = "stere",
			aeqd = "aeqd",
			utm = "utm",
			pconic = "pconic",
			eqdc = "eqdc",
			stere = "stere",
			"laea"
		)
		to_generic_projected(x, proj = proj, return_as = "crs")
	}
	sf::st_crs(y)
}


to_generic_projected <- function(
		x,
		proj = c("laea", "aeqd", "utm", "pconic", "eqdc", "stere"),
		ellps = "WGS84",
		no_defs = TRUE,
		opts = "",
		return_as = c("sf", "proj4", "wkt", "crs")) {
	# arg assertions
	if (!rlang::is_true(rlang::inherits_any(x, c("sf", "sfc", "stars")))) {
		rlang::abort(
			"`x` must be either an sf object or an sfc object or a stars object",
			class = "to_generic_projected_incorrect_input"
		)
	}

	proj <- rlang::arg_match(proj)
	ellps <- rlang::arg_match(ellps, sf::sf_proj_info(type = "ellps")$name)

	if (!rlang::is_logical(no_defs)) {
		rlang::abort("`no_defs` must be a logical value",
					 class = "to_generic_projected_incorrect_input"
		)
	}

	if (!rlang::is_character(opts) && !nchar(opts)) {
		rlang::abort("`opts` must be a character vector",
					 class = "to_generic_projected_incorrect_input"
		)
	}

	return_as <- rlang::arg_match(return_as)

	# was centroid
	cent_coor <- sf::sf_project(
		sf::st_crs(x), "EPSG:4326",
		sf::st_bbox(x) |>
			sf::st_as_sfc() |>
			sf::st_centroid() |>
			sf::st_coordinates()
	)

	# configure proj args
	n_or_s <- ifelse(cent_coor[2] == 0, "",
					 ifelse(cent_coor[2] > 0, "+north", "+south")
	)

	no_defs <- ifelse(no_defs, "+no_defs", "")

	if (proj %in% c("pconic", "eqdc")) {
		bounds <- sf::st_bbox(sf::st_transform(x, 4326))
		lat_1 <- bounds$ymax
		lat_2 <- bounds$ymin
	}

	# base R replacement of glue::glue to reduce dependencies
	glue = function(..., .sep) {
		args = list(...)
		paste(lapply(args, gluestick, src = parent.frame()), collapse = .sep)
	}


	# construct proj4 string
	prj <- trimws(switch(proj,
						 laea = glue(
						 	"+proj=laea +lon_0={cent_coor[1]} +lat_0={cent_coor[2]}",
						 	"+ellps={ellps} {no_defs}",
						 	opts,
						 	.sep = " "
						 ),
						 utm = glue(
						 	"+proj=utm +zone={round((180 + cent_coor[1]) / 6)} {n_or_s}",
						 	"+ellps={ellps} {no_defs}",
						 	opts,
						 	.sep = " "
						 ),
						 aeqd = glue(
						 	"+proj=aeqd +lon_0={cent_coor[1]} +lat_0={cent_coor[2]}",
						 	"+ellps={ellps} {no_defs}",
						 	opts,
						 	.sep = " "
						 ),
						 pconic = glue(
						 	"+proj=pconic +lon_0={cent_coor[1]} +lat_0={cent_coor[2]}",
						 	"+lat_1={lat_1} +lat_2={lat_2}",
						 	"+ellps={ellps} {no_defs}",
						 	opts,
						 	.sep = " "
						 ),
						 eqdc = glue(
						 	"+proj=eqdc +lon_0={cent_coor[1]}",
						 	"+lat_1={lat_1} +lat_2={lat_2}",
						 	"+ellps={ellps} {no_defs}",
						 	opts,
						 	.sep = " "
						 ),
						 stere = glue(
						 	"+proj=stere +lon_0={cent_coor[1]} +lat_0={cent_coor[2]}",
						 	"+ellps={ellps} {no_defs}",
						 	opts,
						 	.sep = " "
						 )
	))

	switch(return_as,
		   sf = sf::st_transform(x, prj),
		   proj4 = prj,
		   wkt = sf::st_crs(prj)$wkt,
		   crs = sf::st_crs(prj)
	)
}




# Source: https://github.com/coolbutuseless/gluestick
gluestick <- function(fmt, src = parent.frame(), open = "{", close = "}", eval = TRUE) {

	nchar_open  <- nchar(open)
	nchar_close <- nchar(close)

	#~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
	# Sanity checks
	#~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
	stopifnot(exprs = {
		is.character(fmt)
		length(fmt) == 1L
		is.character(open)
		length(open) == 1L
		nchar_open > 0L
		is.character(close)
		length(close) == 1
		nchar_close > 0
	})

	#~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
	# Brute force the open/close characters into a regular expression for
	# extracting the expressions from the format string
	#~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
	open  <- gsub("(.)", "\\\\\\1", open ) # Escape everything!!
	close <- gsub("(.)", "\\\\\\1", close) # Escape everything!!
	re    <- paste0(open, ".*?", close)

	#~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
	# Extract the delimited expressions
	#~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
	matches  <- gregexpr(re, fmt)
	exprs    <- regmatches(fmt, matches)[[1]]

	#~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
	# Remove the delimiters
	#~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
	exprs <- substr(exprs, nchar_open + 1L, nchar(exprs) - nchar_close)

	#~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
	# create a valid sprintf fmt string.
	#  - replace all "{expr}" strings with "%s"
	#  - escape any '%' so sprintf() doesn't try and use them for formatting
	#    but only if the '%' is NOT followed by an 's'
	#
	# gluestick() doesn't deal with any pathological cases
	#~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
	fmt_sprintf <- gsub(re      , "%s", fmt)
	fmt_sprintf <- gsub("%(?!s)", "%%", fmt_sprintf, perl=TRUE)

	#~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
	# Evaluate
	#~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
	if (eval) {
		args <- lapply(exprs, function(expr) {eval(parse(text = expr), envir = src)})
	} else {
		args <- unname(mget(exprs, envir = as.environment(src)))
	}

	#~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
	# Create the string(s)
	#~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
	do.call(sprintf, c(list(fmt_sprintf), args))
}

is_rstudio = function () {
	if (requireNamespace("rstudioapi", quietly = TRUE)) {
		rstudioapi::isAvailable() && rstudioapi::versionInfo()$mode !=
			"vscode"
	}
	else {
		FALSE
	}
}

get_ide = function () {
	if (is_rstudio()) {
		return("rstudio")
	}
	else if (is_vscode()) {
		return("vscode")
	}
	else {
		"other"
	}
}

## need to assign global variable .vsc
utils::globalVariables(c(".vsc"))

is_vscode = function() {
	exists(".vsc") && exists("attach", envir = .vsc)
}

get_url_dir = function (url) gsub("file://|/index.html", "", url)

#' @export
#' @rdname tmapGridLeaflet
tmapLeafletDataPlot.tm_data_symbols = function(a, shpTM, dt, pdt, popup.format, hdt, idt, gp, bbx, facet_row, facet_col, facet_page, id, pane, group, glid, o, ...) {
	lf = get_lf(facet_row, facet_col, facet_page)

	rc_text = frc(facet_row, facet_col)

	res = select_sf(shpTM, dt[!is.na(dt$size) & !is.na(dt$shape), ])
	shp = res$shp
	if (o$crs_leaflet$crsClass  == "L.CRS.Simple") {
		shp = sf::st_set_crs(shp, NA)
	}
	dt = res$dt

	gp = impute_gp(gp, dt)
	gp = rescale_gp(gp, o$scale_down)

	coords = sf::st_coordinates(shp)

	# in case shp is a multipoint (point_per == "segment"), expand gp:
	cp = expand_coords_gp(coords, gp, ndt = nrow(dt))
	coords = cp$coords
	gp = cp$gp

	if (cp$expanded) {
		shpTM_match = match(shpTM$tmapID_expanded, shpTM$tmapID)
	} else {
		shpTM_match = TRUE
	}

	idt_null = is.null(idt)

	if (!idt_null) {
		idt = idt$id[match(dt$tmapID__[shpTM_match], idt$tmapID__)]
	} else {
		idt = sprintf("%07d", dt$tmapID__)[shpTM_match]
	}
	if (!is.null(hdt)) {
		hdt = hdt$hover[match(dt$tmapID__[shpTM_match], hdt$tmapID__)]
		hdt = lapply(hdt, htmltools::HTML, FUN.VALUE = character(1))
	}
	if (is.null(pdt)) {
		popups = NULL
	} else {
		mtch = match(dt$tmapID__[shpTM_match], pdt$tmapID__)
		pdt = pdt[mtch][, tmapID__ := NULL]

		if (idt_null) {
			popups = view_format_popups(titles = names(pdt), values = pdt, format = popup.format)
		} else {
			popups = view_format_popups(id = idt, titles = names(pdt), values = pdt, format = popup.format)
		}
	}

	clusterOpts = getClusterOpts(a$clustering)



	interactive = (!is.null(pdt) || !is.null(hdt))

	opt = leaflet::pathOptions(interactive = interactive, pane = pane)

	o$use_WebGL = impute_webgl(o$use_WebGL, dt, supported = c("fill", "size"), checkif = list(shape = c(1,16,19,20,21)), type = "symbols", hover = !is.null(hdt), popup = !is.null(pdt), crs_class = o$crs_leaflet$crsClass)

	use_circleMarkers = o$use_circle_markers && all(gp$shape %in% c(1, 16, 19, 20, 21))


	gp2 = gp_to_lpar(gp, mfun = "Symbols", size_factor = 14) # 14 based on similarity with plot mode and consistency with tmap3
	#gp = gp2leafgp(gp)
	names(gp2)[names(gp2) == 'stroke-width'] = "strokeWidth"
	gp2$baseSize = 20

	#po(sort(gp2$width, decreasing = T))

	shp2 = sf::st_as_sf(as.data.frame(coords), coords = c("X", "Y"), crs = st_crs(shp))

	if (o$use_WebGL) {
		idt = submit_labels(idt[1], "symbolsGL", pane, group)

		lf |>  leafgl::addGlPoints(shp2, fillColor = gp2$fillColor, radius = gp2$width, fillOpacity = gp2$fillOpacity[1], pane = pane, group = group, label = hdt, popup = popups, layerId = idt) %>%
			assign_lf(facet_row, facet_col, facet_page)
	} else if (use_circleMarkers) {
		idt = submit_labels(idt, "symbols", pane, group)
		cidt = if (!is.null(clusterOpts)) submit_labels(idt[1], "symbolsCluster", pane, group) else NULL

		gp2$strokeWidth[gp2$shape %in% c("solid-circle-md", "solid-circle-bg", "solid-circle-sm")] = 0
		gp2$fillOpacity[gp2$shape %in% "open-circle"] = 0
		multiplier = ifelse(gp2$shape == "solid-circle-md", 0.28, ifelse(gp2$shape == "solid-circle-sm", 0.25, 0.5)) # manually calibrated with 4k screen

		lf |>  leaflet::addCircleMarkers(data = shp2, layerId = idt,
										 color = gp2$color,
										 opacity = gp2$opacity,
										 weight = gp2$strokeWidth,
										 dashArray = gp2[["stroke-dasharray"]],
										 fillColor = gp2$fillColor,
										 fillOpacity = gp2$fillOpacity,
										 clusterOptions = clusterOpts,
										 clusterId = cidt,
										 radius = gp2$width * multiplier,
										 options = opt,
										 group = group, label = hdt, popup = popups) |>
			assign_lf(facet_row, facet_col, facet_page)

	} else {
		idt = submit_labels(idt, "symbols", pane, group)
		cidt = if (!is.null(clusterOpts)) submit_labels(idt[1], "cluster", pane, group) else NULL

		sn = suppressWarnings(as.numeric(gp2$shape))

		is_num = !is.na(sn)
		sid = which(is_num)
		nid = which(!is_num)


		# ="circle" to make makeSymbolsIcons2 work
		# shape_orig to let unique pick unique rows (one for each )
		gp2$shape_orig = gp2$shape
		gp2$shape[sid] = "circle"

		# faster than symbols2 = do.call(makeSymbolIcons2, gp2)
		gp2df = as.data.table(gp2)
		gp2dfU = unique(gp2df)

		k = nrow(gp2dfU)

		symbols = do.call(makeSymbolIcons2, as.list(gp2dfU[,-(ncol(gp2dfU)),with=F]))

		gp2dfU[, id:=1L:.N]
		gp2join = gp2df[gp2dfU, id:= id, on=names(gp2df)]
		ids = gp2join$id

		coords_grps = split.data.frame(coords, ids)
		idt_grps = split(idt, ids)
		if (!is.null(hdt)) {
			hdt_grps = split(hdt, ids)
		} else {
			hdt_grps = replicate(k, list(NULL))
		}

		if (is.null(popups)) {
			popups_grps =  rep(list(NULL), k)
		} else {
			popups_grps = split(popups, ids)
		}


		symbols$iconWidth = rep(NA, k)
		symbols$iconHeight = rep(NA, k)


		if (length(sid)) {
			sym_shapes = suppressWarnings(as.numeric(gp2dfU$shape_orig))
			sid2 = which(!is.na(sym_shapes))

			iconLib <- get("shapeLib", envir = .TMAP)[sym_shapes[sid2]-999]
			symbols_icons <- merge_icons(iconLib)

			size = gp2dfU$width[sid2] / gp2dfU$baseSize[sid2]
			size = size * a$icon.scale/3

			for (i in seq_along(sid2)) {
				symbols$iconUrl[sid2[i]] = symbols_icons$iconUrl[i]
				symbols$iconWidth[sid2[i]] <- symbols_icons$iconWidth[i] * size[i]
				symbols$iconHeight[sid2[i]] <- symbols_icons$iconHeight[i] * size[i]
				if (all(c("iconAnchorX", "iconAnchorY") %in% names(symbols_icons))) {
					symbols$iconAnchorX[sid2[i]] <- symbols_icons$iconAnchorX[i] * size[i]
					symbols$iconAnchorY[sid2[i]] <- symbols_icons$iconAnchorY[i] * size[i]

				}
			}
		}

		for (i in 1L:k) {
			opt$zIndexOffset = i * 1e5
			.TMAP_LEAFLET$markerLayers = c(.TMAP_LEAFLET$markerLayers, opt$pane) # register for additional css
			lf = lf |>
				leaflet::addMarkers(lng = coords_grps[[i]][, 1],
									lat = coords_grps[[i]][, 2],
									icon = lapply(symbols, "[", i),
									group = group,
									layerId = idt_grps[[i]],
									label = hdt_grps[[i]],
									clusterOptions = clusterOpts,
									clusterId = cidt,
									popup = popups_grps[[i]],
									options = opt
				)
		}
		lf |> assign_lf(facet_row, facet_col, facet_page)

	}


	# if (o$use_WebGL) {
	# 	lf %>%
	# 		leafgl::addGlPoints(sf::st_sf(shp), fillColor = gp$fill, radius = gp$size*10, fillOpacity = gp$fill_alpha, color = gp$col, opacity = gp$color_alpha, weight = gp$lwd, pane = pane, group = group) %>%
	# 		assign_lf(facet_row, facet_col, facet_page)
	# } else {
	# 	lf %>%
	# 		leaflet::addCircleMarkers(lng = coords[, 1], lat = coords[, 2], fillColor = gp$fill, radius = gp$size*4, fillOpacity = gp$fill_alpha, color = gp$col, opacity = gp$color_alpha, weight = gp$lwd, group = group, options = opt) %>%
	# 		assign_lf(facet_row, facet_col, facet_page)
	# }

	NULL
}


#' @export
#' @rdname tmapGridLeaflet
tmapLeafletDataPlot.tm_data_dots = function(a, shpTM, dt, pdt, popup.format, hdt, idt, gp, bbx, facet_row, facet_col, facet_page, id, pane, group, glid, o, ...) {
	NextMethod()
}

#' @export
#' @rdname tmapGridLeaflet
tmapLeafletDataPlot.tm_data_bubbles = function(a, shpTM, dt, pdt, popup.format, hdt, idt, gp, bbx, facet_row, facet_col, facet_page, id, pane, group, glid, o, ...) {
	NextMethod()
}

#' @export
#' @rdname tmapGridLeaflet
tmapLeafletDataPlot.tm_data_squares = function(a, shpTM, dt, pdt, popup.format, hdt, idt, gp, bbx, facet_row, facet_col, facet_page, id, pane, group, glid, o, ...) {
	NextMethod()
}

#' @export
#' @rdname tmapGridLeaflet
tmapLeafletDataPlot.tm_data_markers = function(a, shpTM, dt, pdt, popup.format, hdt, idt, gp, bbx, facet_row, facet_col, facet_page, id, pane, group, glid, o, ...) {
	NextMethod()
}

#' @export
#' @rdname tmap_internal
tmapChartBinned = function(chart, bin_colors, breaks_def, na.show, x1) {
	if (is.numeric(x1)) {
		tmapChartBinned_numeric(chart, bin_colors, breaks_def, na.show, x1)
	} else {
		tmapChartBinned_categorical(chart, bin_colors, breaks_def, na.show, x1)
	}
}

#' @export
#' @rdname tmap_internal
tmapChartBinned2d = function(chart, chart1, chart2) {
	if (is.numeric(chart1$x1) && is.numeric(chart2$x1)) {
		tmapChartBinned2d_numnum(chart, chart1, chart2)
	} else if (is.numeric(chart1$x1) && !is.numeric(chart2$x1)) {
		tmapChartBinned2d_numcat(chart, chart1, chart2)
	} else if (!is.numeric(chart1$x1) && is.numeric(chart2$x1)) {
		res = tmapChartBinned2d_numcat(chart, chart2, chart1)
		names(res$tab) = names(res$tab)[c(2,1,3)]
		res
	} else {
		tmapChartBinned2d_catcat(chart, chart1, chart2)
	}
}

#' @export
#' @rdname tmap_internal
tmapChartBinned2d_numcat = function(chart, chart1, chart2) {
	res = bin_num(chart1$x1, chart1$breaks_def, chart1)
	chart$tab = as.data.frame(table(bin1 = res$xcat, bin2 = chart2$x1, useNA = "no"), responseName = "freq")
	chart$pal = NA
	chart$datatype = "categorized"
	chart$predefined = FALSE
	chart
}

#' @export
#' @rdname tmap_internal
tmapChartBinned2d_numnum = function(chart, chart1, chart2) {
	res1 = bin_num(chart1$x1, chart1$breaks_def, chart1)
	res2 = bin_num(chart2$x1, chart2$breaks_def, chart2)
	chart$tab = as.data.frame(table(bin1 = res1$xcat, bin2 = res2$xcat, useNA = "no"), responseName = "freq")
	chart$pal = NA
	chart$datatype = "categorized"
	chart$predefined = FALSE
	chart
}


#' @export
#' @rdname tmap_internal
tmapChartBinned2d_catcat = function(chart, chart1, chart2) {
	chart$tab = as.data.frame(table(bin1 = chart1$x1, bin2 = chart2$x1, useNA = "no"), responseName = "freq")
	chart$pal = NA
	chart$datatype = "categorized"
	chart$predefined = FALSE
	chart
}


#' @export
#' @rdname tmap_internal
tmapChartRaw = function(chart, na.show, x1, ...) {
	if (!na.show) x1 = na.omit(x1)

	chart$df = data.frame(x = 1L, y = unname(x1))
	chart$datatype = "raw"
	chart$predefined = FALSE
	chart
}

#' @export
#' @rdname tmap_internal
tmapChartRaw_nna = function(chart, na.show, x1, ...) {
	x1 = na.omit(x1)

	chart$df = data.frame(x = 1L, y = unname(x1))
	chart$datatype = "raw"
	chart$predefined = FALSE
	chart
}

#' @export
#' @rdname tmap_internal
tmapChartNone = function(chart, na.show, x1, ...) {
	chart
}

#' @export
#' @rdname tmap_internal
tmapChartPass = function(chart, na.show, x1, ...) {
	args = c(list(na.show = na.show, x1 = x1), list(...))
	chart[names(args)] = args
	chart
}

#' @export
#' @rdname tmap_internal
tmapChartBinned_categorical = function(chart, bin_colors, breaks_def, na.show, x1) {
	if (chart$type == "histogram") {
		message("histograms are supposed to be used for numerical data, a bar chart will be shown instead (tm_chart_bar)")
	}

	if (na.show) {
		t = table(bin = x1, useNA = "always")
		tab = as.data.frame(t, responseName = "freq")
		levels(tab$bin) = c(levels(tab$bin), "NA")
		tab$bin[is.na(tab$bin)] = "NA"
		tab$color = factor(1L:nrow(tab))
		pal = structure(bin_colors, names = levels(tab$color))
	} else {
		tab = as.data.frame(table(bin = x1, useNA = "no"), responseName = "freq")
		tab$color = factor(1L:nrow(tab))
		pal = structure(bin_colors, names = levels(tab$color))
	}
	chart$tab = tab
	chart$pal = pal
	chart$datatype = "categorized"
	chart$predefined = TRUE
	chart
}


bin_num = function(x1, breaks_def, chart) {
	# are breaks (and bin_colors)
	predefined = !is.null(breaks_def)

	if (is.null(chart$breaks_def)) {
		if (!predefined) {
			breaks = pretty(x1)
			ids = rep(1L, length(breaks) - 1)

		} else {
			breaks = breaks_def
			ids = 1L:(length(breaks) - 1L)
		}
	} else {
		breaks = chart$breaks_def
		subbreaks = (all(breaks_def %in% breaks))

		break_mids = (breaks[-1] + head(breaks, -1)) / 2

		if (predefined) {
			ids = as.integer(cut(break_mids, breaks_def, include.lowest = TRUE, right = FALSE))
		} else {
			ids = rep(1L, length(breaks) - 1)
		}

	}
	xcat = cut(x1, breaks = breaks, include.lowest = TRUE, right = FALSE)
	list(xcat = xcat, ids = ids)
}


#' @export
#' @rdname tmap_internal
tmapChartBinned_numeric = function(chart, bin_colors, breaks_def, na.show, x1) {

	# are breaks (and bin_colors)
	predefined = !is.null(breaks_def)

	res = bin_num(x1, breaks_def, chart)
	xcat = res$xcat
	ids = res$ids

	if (!predefined || (length(bin_colors) != (length(breaks_def) -1+na.show))) {
		bin_colors = chart$object.color
	}

	if (na.show) {
		t = table(bin = xcat, useNA = "always")
		tab = as.data.frame(t, responseName = "freq")
		levels(tab$bin) = c(levels(tab$bin), "NA")
		tab$bin[is.na(tab$bin)] = "NA"
		tab$color = factor(c(ids, length(bin_colors)), levels = seq_along(bin_colors))
		pal = structure(bin_colors, names = levels(tab$color))
	} else {
		tab = as.data.frame(table(bin=xcat, useNA = "no"), responseName = "freq")
		tab$color = factor(ids, levels = seq_along(bin_colors))
		pal = structure(bin_colors, names = levels(tab$color))
	}
	chart$tab = tab
	chart$pal = pal
	chart$datatype = "binned"
	chart$predefined = predefined
	chart
}

#' Set tmap mode to static plotting or interactive viewing
#'
#' @description
#' * `tmap_mode()` informs of the current mode (if called without argument).
#' * `ttm()` toggles between the most recent two modes.
#' * `ttmp()` same as `ttm()` and calls [tmap_last()] to display the last map in the new mode.
#' * `rtm()` rotate between between all modes
#' * `rtmp()` same as `rtm()` and calls [tmap_last()] to display the last map in the new mode.
#'
#' Set tmap mode to static plotting or interactive viewing.
#' The global option `tmap.mode` determines the whether thematic maps are plot
#' in the graphics device, or shown as an interactive leaflet map (see also [tmap_options()].
#' The function `tmap_mode()` is a wrapper to set this global option.
#' The convenient function `ttm()`, which stands for toggle thematic map,
#' is a toggle switch between the two modes. The function `ttmp()` stands for
#' toggle thematic map and print last map: it does the same as `ttm()` followed
#' by [tmap_last()]; in order words, it shows the last map in the other mode.
#' It is recommended to use `tmap_mode()` in scripts and `ttm()`/`ttmp()` in the console.
#'
#' @details
#' # `mode = "plot"`
#'
#' Thematic maps are shown in the graphics device.
#' This is the default mode, and supports all tmap's features,
#' such as small multiples (see [tm_facets()]) and extensive layout settings (see [tm_layout()]).
#' It is recommended to use [tmap_save()] for saving static maps.
#'
#' # `mode = "view"`
#'
#' Thematic maps are viewed interactively in the web browser or RStudio's Viewer pane.
#' Maps are fully interactive with tiles from OpenStreetMap or other map providers
#' (see [tm_tiles()]). See also [tm_view()] for options related to the `"view"` mode.
#' This mode generates a [leaflet::leaflet()] widget, which can also be directly
#' obtained with [tmap_leaflet()].
#' With R Markdown, it is possible to publish it to an HTML page.
#'
#' However, there are a couple of constraints in comparison to `"plot"`:
#'
#' @param mode One of `"plot"` or `"view"`. See Details for more info.
#' @param silent Should the mode be switched silently (by default `FALSE`)
#' @return
#' * `tmap_mode()` returns the current tmap mode invisibly (when called without argument).
#'   Otherwise, returns the previous mode.
#' * `ttm()` switches mode and returns previous tmap mode invisibly.
#' The previous tmap mode before switching.
#' @example ./examples/tmap_mode.R
#' @seealso \href{https://r-tmap.github.io/tmap/articles/basics_modes}{vignette about modes}
#' * [tmap_last()] to show the last map
#' * [tm_view()] for viewing options
#' * [tmap_leaflet()] for obtaining a leaflet widget
#' * [tmap_options()] for tmap options
#' @references Tennekes, M., 2018, {tmap}: Thematic Maps in {R},
#' Journal of Statistical Software, 84(6), 1-39, \doi{10.18637/jss.v084.i06}
#' @export
tmap_mode = function(mode = NULL, silent = FALSE) {
	type = if (is.null(mode)) "get" else "set"
	set_mode(mode, show.messages = !silent && get("tmapOptions", envir = .TMAP)$show.messages, type = type)
}


#' @rdname tmap_mode
#' @export
ttm = function() {
	set_mode(show.messages = get("tmapOptions", envir = .TMAP)$show.messages, type = "toggle")
}

#' @rdname tmap_mode
#' @export
rtm = function() {
	set_mode(show.messages = get("tmapOptions", envir = .TMAP)$show.messages, type = "rotate")
}


set_mode = function(mode = NULL, show.messages = FALSE, type = "set") {
	mode_now = getOption("tmap.mode")

	modes = get_modes()
	id = match(mode_now, modes)

	arrow = length(modes) > 2

	if (type == "get") {
		mode_next = mode_now
		id_next = id
	} else if (type == "set") {
		mode_next = rlang::arg_match0(mode, modes)
		id_next = match(mode_next, modes)
	} else if (type == "toggle") {
		mode_next = .TMAP$mode_last
		id_next = match(mode_next, modes)
	} else {
		id_next = id + 1L
		if (id_next > length(modes)) id_next = 1L
		mode_next = modes[id_next]
	}

	modes_str = paste0("{.val ", modes, "}")
	modes_str[id_next] = paste0("{.strong ", modes_str[id_next], "}")
	rotate_str = paste(modes_str, collapse = ifelse(arrow, " -> ", " - "))

	if (mode_now != mode_next) {
		.TMAP$mode_last = mode_now
		options(tmap.mode = mode_next)
		fill_providers()
	}

	if (show.messages) {
		str1 = paste0("tmap modes ", rotate_str)
		str2 = paste0(ifelse(arrow, "rotate with {.run tmap::rtm()}", ""),
					 ifelse(arrow, "switch to {.str {mode_now}} ", "toggle "),
					 "with {.run tmap::ttm()}")

		cli::cli_inform(c("i" = str1))
		cli::cli_inform(c("i" = str2), .frequency = "once", .frequency_id = paste0("mode", ifelse(arrow, "_rotate", "toggle")))
	}

	invisible(mode_now)
}


# tmap_graphics = function(mode = NULL) {
# 	if (is.null(mode)) mode = getOption("tmap.mode")
# 	get("tmapOptions", envir = .TMAP)$graphics[[mode]]
# }
#
# tmap_graphics_name = function(mode = NULL) {
# 	tmap_graphics(mode = mode)$name
# }

get_modes = function() {
	names(get("tmapOptions", envir = .TMAP)$modes)
}

#' Set the design mode
#'
#' When the so-called "design mode" is enabled,
#' the composition of the plot is shown explicitly in plot mode.
#' The used color codings is printed in the console as well as information about plot size and aspect ratio.
#'
#' This function sets the global option `tmap.design.mode`.
#' It can be used as toggle function without arguments.
#'
#' @seealso [tmap_options()]
#' @param design.mode Logical value that determines the design mode.
#'   If omitted then the design mode is toggled.
#' @export
tmap_design_mode = function(design.mode) {
	dm = if (missing(design.mode)) {
		!getOption("tmap.design.mode")
	} else {
		if (!is.logical(design.mode)) stop("design.mode is not a logical")
		design.mode[1]
	}

	options(tmap.design.mode = dm)
	message(
		"design.mode: ", if (!dm) "OFF" else "ON",
		if (dm && getOption("tmap.mode") == "view") " (only effective in plot mode)" else "")
}

#' Set the development mode
#'
#' When the so-called "development mode" is enabled, helpful messages and timings
#' are printed in the console
#'
#' @param devel.mode logical value that determines the development mode.
#'   If omitted then the development mode is toggled.
#' @export
tmap_devel_mode = function(devel.mode) {
	dm = if (missing(devel.mode)) {
		!getOption("tmap.devel.mode")
	} else {
		if (!is.logical(devel.mode)) stop("devel.mode is not a logical")
		devel.mode[1]
	}

	options(tmap.devel.mode = dm)
	message("devel.mode: ", if (!dm) "OFF" else "ON")
}

pm = function(message) {
	cat("<><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><>\n")
	cat(message, "\n")
}

po = function(...) {
	e = substitute(list(...))
	nms = sapply(e, deparse)[-1]

	x = list(...)

	for (i in seq_along(x)) {
		cat("<==================== ", nms[i], "===============>\n")
		print(x[[i]])
		if (i == length(x)) {
			cat("</============================================>\n")
		}
	}

	invisible()
}

so = function(...) {
	e = substitute(list(...))
	nms = sapply(e, deparse)[-1]

	x = list(...)

	for (i in seq_along(x)) {
		cat("<==================== ", nms[i], "===============>\n")
		str(x[[i]])
		if (i == length(x)) {
			cat("</============================================>\n")
		}
	}

	invisible()
}

#' @rdname tmap_mode
#' @export
ttmp = function() {
	ttm()
	tmap_last()
}

#' @rdname tmap_mode
#' @export
rtmp = function() {
	rtm()
	tmap_last()
}


check_unit = function(unit) {
	if (!unit %in% c("metric", "imperial", "km", "m", "mi", "miles", "ft", "feet")) stop("incorrect unit", call. = FALSE)
}

#' @export
#' @rdname tmap_internal
tmapScaleRank = function(x1, scale, legend, chart, o, aes, layer, layer_args, sortRev, bypass_ord, submit_legend = TRUE) {

	# update misc argument from tmap option scale.misc.args
	scale = update_scale_args("rank", scale, aes, o)

	cls = data_class(x1, midpoint_enabled = !is.null(scale$midpoint))
	maincls = class(scale)[1]

	if (attr(cls, "unique") && is.null(scale$limits) && is.null(scale$ticks)) stop("Unique value, so cannot determine continuous scale range. Please specify limits and/or ticks.", call. = FALSE)

	#if (cls[1] == "na") stop("data contain only NAs, so ", maincls, " cannot be applied", call. = FALSE)

	if (cls[1] != "num") {
		if (!is.factor(x1)) x1 = as.factor(x1)
		x1 = as.integer(x1)
		warning(maincls, " is supposed to be applied to numerical data", call. = FALSE)
	}


	x1 = rank(without_units(x1), na.last = "keep")

	if (aes %in% c("lty", "shape", "pattern")) stop("tm_scale_rank cannot be used for layer ", layer, ", aesthetic ", aes, call. = FALSE)

	scale = structure(c(scale, list(limits = NULL, outliers.trunc = NULL, trans = NULL, midpoint = NULL,
						  trans.args = list())), class = c("tm_scale_continuous", "tm_scale", "list"))

	if (is.null(scale$ticks)) {
		scale$ticks = generate_rank_labels(max_rank = max(x1,na.rm=T), num_ticks = scale$n)
	}


	tmapScaleContinuous(x1, scale, legend, chart, o, aes, layer, layer_args, sortRev, bypass_ord, submit_legend)

}

generate_rank_labels = function(max_rank, num_ticks = 5) {
	ticks = c(1, max_rank)

	if (num_ticks > 2) {
		probs = seq(0, 1, length.out = num_ticks)
		probs = probs[-c(1, length(probs))]

		intermediate_ticks = quantile(1:max_rank, probs = probs)
		ticks = c(ticks, intermediate_ticks)
	}

	# Ensure the order
	ticks = sort(unique(round(ticks)))

	return(ticks)
}

#rm = rownames(tmaptools::tmap.pal.info)
#dput(sapply(rm, cols4all:::c4a_name_convert))

pals_v3 = c(BrBG = "brewer.br_bg", PiYG = "brewer.pi_yg", PRGn = "brewer.prgn",
			PuOr = "brewer.pu_or", RdBu = "brewer.rd_bu", RdGy = "brewer.rd_gy",
			RdYlBu = "brewer.rd_yl_bu", RdYlGn = "brewer.rd_yl_gn", Spectral = "brewer.spectral",
			Accent = "brewer.accent", Dark2 = "brewer.dark2", Paired = "brewer.paired",
			Pastel1 = "brewer.pastel1", Pastel2 = "brewer.pastel2", Set1 = "brewer.set1",
			Set2 = "brewer.set2", Set3 = "brewer.set3", Blues = "brewer.blues",
			BuGn = "brewer.bu_gn", BuPu = "brewer.bu_pu", GnBu = "brewer.gn_bu",
			Greens = "brewer.greens", Greys = "brewer.greys", Oranges = "brewer.oranges",
			OrRd = "brewer.or_rd", PuBu = "brewer.pu_bu", PuBuGn = "brewer.pu_bu_gn",
			PuRd = "brewer.pu_rd", Purples = "brewer.purples", RdPu = "brewer.rd_pu",
			Reds = "brewer.reds", YlGn = "brewer.yl_gn", YlGnBu = "brewer.yl_gn_bu",
			YlOrBr = "brewer.yl_or_br", YlOrRd = "brewer.yl_or_rd")


getPal = function(name, n = NA, rep = TRUE, range = NA, reversed = FALSE) {
	if (name %in% c("cat", "seq", "div")) {
		name = cols4all::c4a_options("defaults")$defaults[[name]]
	}

	cols4all::c4a(name, n = n, nm_invalid = {if (rep) "repeat" else "interpolate"}, range = range, reverse = reversed)
}

getPalNA = function(name) {
	if (name %in% c("cat", "seq", "div")) {
		name = cols4all::c4a_options("defaults")$defaults[[name]]
	}
	cols4all::c4a_na(name)
}

getPalBiv = function(name, m = NA, n = NA, rep = TRUE) {
	cols4all::c4a(name, m = m, n = n, nm_invalid = {if (rep) "repeat" else "interpolate"})
}

getPalMeta = function(name, no.match = "null") {
	if (!is.character(name)) return(NULL)
	if (name %in% c("cat", "seq", "div")) {
		name = cols4all::c4a_options("defaults")$defaults[[name]]
	}
	if (name %in% names(pals_v3)) {
		oldname = name
		name = unname(pals_v3[oldname])
		info = cols4all::c4a_info(name, verbose = FALSE)
		message_c4a(oldname, info, fullname = info$series == "brewer")
	} else {
		info = cols4all::c4a_info(name, no.match = no.match, verbose = FALSE)
	}
	info
}

do_trans = function(tdt, FUN, shpDT, plot.order, args, scale) {

	shpDT = copy(shpDT)

	# copy by columns from tdt that do not yet exist in shpDT
	t_by = names(tdt)[substr(names(tdt), 1, 2) == "by"]
	s_by = names(shpDT)[substr(names(shpDT), 1, 2) == "by"]
	n_by = setdiff(t_by, s_by)
	if (length(n_by)) {
		shpDT[, (n_by) := tdt[1, n_by, with = FALSE]]
	}

	aesvars = setdiff(names(tdt), c("tmapID__", paste0("by", 1:3, "__")))

	apply_trans = function(shpTM) {
		# todo: stars
		ids = intersect(shpTM$tmapID, tdt$tmapID__)

		shp = shpTM$shp[match(ids, shpTM$tmapID)]

		shpX = list(shp = shp, tmapID = ids, bbox = shpTM$bbox)

		x = as.list(tdt[match(tmapID__, ids), aesvars, with = FALSE])

		res = do.call(FUN, c(list(shpTM = shpX), x, list(plot.order = plot.order, args = args, scale = scale)))
	}
	shpDT$shpTM = lapply(shpDT$shpTM, apply_trans)
	list(shpDT)
}

delta_per_lineheight = function(x, n = 20, scale = 1) {
	b = unname(sf::st_bbox(x))
	(b[4] - b[2]) / n * scale
}

get_midpoint_angle = function(shp, bbx) {
	lShp = sf::st_cast(shp, "MULTILINESTRING")

	coor = sf::st_coordinates(lShp)
	coors = split.data.frame(coor[,1:2], f = coor[, ncol(coor)], drop=FALSE)
	co = do.call(rbind, lapply(coors, get_midpoint))
	pShp = sf::st_as_sf(as.data.frame(co), coords = c("X", "Y"), crs = sf::st_crs(shp))

	deltax = bbx[3] - bbx[1]
	deltay = bbx[4] - bbx[2]
	delta = max(deltax, deltay)

	pbShp = sf::st_cast(sf::st_geometry(sf::st_buffer(pShp, dist = delta / 100)), "MULTILINESTRING")

	iShps <- mapply(function(x,y,z) {
		res = sf::st_intersection(x,y)
		if (!sf::st_is_empty(res)) {
			sf::st_cast(res, "MULTIPOINT")
		} else {
			z
		}
	}, lShp, pbShp,pShp, SIMPLIFY = FALSE)

	angles = vapply(iShps, function(x) {
		if (length(x) == 0) 0 else .get_direction_angle(sf::st_coordinates(x)[,1:2, drop = FALSE])
	}, numeric(1))

	angles[angles > 90 & angles < 270] = angles[angles > 90 & angles < 270] - 180

	list(shp = sf::st_geometry(pShp), angles = angles)
}

sf_expand <- function(shp, cast = c("MULTILINESTRING", "MULTIPOLYGON", "MULTIPOINT")) {
	x = mapply(function(tp, ge) {
		if (tp %in% cast) {
			if (tp == "MULTILINESTRING") {
				sf::st_cast(st_geometry(ge), "LINESTRING")
			} else if (tp == "MULTIPOLYGON") {
				sf::st_cast(st_geometry(ge), "POLYGON")
			} else if (tp == "MULTIPOINT") {
				sf::st_cast(st_geometry(ge), "POINT")
			} else {
				sf::st_geometry(ge)
			}
		} else {
			sf::st_geometry(ge)
		}
	}, sf::st_geometry_type(shp), sf::st_geometry(shp), SIMPLIFY = FALSE)
	ids = rep(1L:length(x), vapply(x, length, integer(1)))
	shp3 = sf::st_sf(geometry=sf::st_sfc(do.call(c, x), crs = sf::st_crs(shp)))
	shp3$split__id = ids
	shp3
}

#' @param shpTM shpTM
#' @param xmod,ymod xmod and ymod
#' @param ord__ ord
#' @param plot.order plot.order
#' @param args args
#' @param scale scale
#' @export
#' @rdname tmap_internal
tmapTransCentroid = function(shpTM, xmod = NULL, ymod = NULL, ord__, plot.order, args, scale) {
	within(shpTM, {
		if (!is.character(shp)) {

			is_stars = inherits(shp, "dimensions")
			if (is_stars && args$points_only == "no") {
				### stars

				s = structure(list(values = matrix(TRUE, nrow = nrow(shp))), dimensions = shp, class = "stars")
				strs = stars::st_as_stars(list(values = m), dimensions = shp)
				shp = sf::st_as_sfc(s, as_points = TRUE)
			} else if (is_stars) {
				shp = sf::st_sfc()
				tmapID = integer(0L)
			} else {
				geom_all = sf::st_geometry_type(shp, by_geometry = FALSE)
				if (geom_all == "GEOMETRY") {
					geom_types = sf::st_geometry_type(shp)
				} else {
					geom_types = rep(geom_all, length(shp))
				}


				if (any(geom_types %in% c("MULTILINESTRING", "MULTIPOINT", "MULTIPOLYGON"))) {
					if (args$point_per=="feature" && any(geom_types == "MULTIPOINT")) {
						shp = sf_expand(shp, cast = "MULTIPOINT")
						tmapID = tmapID[shp$split__id]
						shp = sf::st_geometry(shp)
					} else if (args$point_per=="segment") {
						shp = sf_expand(shp)
						tmapID = tmapID[shp$split__id]
						shp = sf::st_geometry(shp)
					} else if (args$point_per == "largest") {
						ids_multiline = which(geom_types == "MULTILINESTRING")
						if (length(ids_multiline)) {
							shp[ids_multiline] = local({
								shp_ml = sf_expand(shp[ids_multiline])
								lengths <- sf::st_length(shp_ml)
								id_max <- vapply(split(lengths, f=shp_ml$split__id), which.max, integer(1))
								id_max2 <- vapply(seq_along(ids_multiline), function(i) {
									which(shp_ml$split__id==i)[id_max[i]]
								}, integer(1))
								sf::st_geometry(shp_ml[id_max2, ])
							})
						}
					}
				}

				geom_all2 = sf::st_geometry_type(shp, by_geometry = FALSE)
				if (geom_all2 == "GEOMETRY") {
					geom_types2 = sf::st_geometry_type(shp)
				} else {
					geom_types2 = rep(geom_all, length(shp))
				}

				ids_poly = which(geom_types2 %in% c("POLYGON", "MULTIPOLYGON"))
				ids_line = which(geom_types2 %in% c("LINESTRING", "MULTILINESTRING"))
				ids_point = which(geom_types2 %in% c("POINT", "MULTIPOINT"))

				if (args$points_only == "yes" || (args$points_only == "ifany" && length(ids_point))) {
					shp = shp[ids_point]
					tmapID = tmapID[ids_point]
				} else {
					if (length(ids_line)) {
						if (args$along_lines) {
							bb_crs = stm_bbox(shpTM, shpTM$tmapID, crs = sf::st_crs(shp))
							res = get_midpoint_angle(shp[ids_line], bbx = bb_crs)
							shp[ids_line] = res$shp
							prop_angle = rep(0, length(shp))
							prop_angle[ids_line] = res$angles
							rm(res)
						} else {
							shp[ids_line] = suppressWarnings({
								sf::st_centroid(shp[ids_line])
							})
						}
					}
					if (length(ids_poly)) {
						ctds = suppressWarnings({
							sf::st_centroid(shp[ids_poly], of_largest_polygon = (args$point_per == "largest"))
						})

						if (args$on_surface) {
							ctds_in = unlist(mapply(function(x,y) {
								length(sf::st_contains(y,x)[[1]]) > 0L
							}, ctds, shp[ids_poly], SIMPLIFY = FALSE, USE.NAMES = FALSE))


							if (!all(ctds_in)) {
								shp[ids_poly[!ctds_in]] = sf::st_point_on_surface(shp[ids_poly[!ctds_in]])
								shp[ids_poly[ctds_in]] = ctds[ctds_in]
							} else {
								shp[ids_poly] = ctds
							}
						} else {
							shp[ids_poly] = ctds
						}
						rm(ctds)
					}
				}

				if (!is.null(xmod) || !is.null(ymod)) {
					shp = local({
						d = delta_per_lineheight(shp, scale = scale)

						if (is.null(xmod)) xmod = rep(0, length(shp))
						if (is.null(ymod)) ymod = rep(0, length(shp))
						mod = mapply(c, xmod * d, ymod * d, SIMPLIFY = FALSE)
						shp + mod
					})
				}

				# in case of expansion: cast to multipoint (expanded again in step4)

				if (args$point_per=="segment" || (args$point_per=="feature" && any(geom_types == "MULTIPOINT"))) {
					tmapID_expanded = tmapID
					u = sort(unique(tmapID))
					tmapID_1n = match(tmapID, u)

					shp = sf::st_cast(shp, "MULTIPOINT", ids = tmapID_1n)
					tmapID = u

					if (length(ids_line) && args$along_lines) {
						prop_angle = split(prop_angle, f = tmapID_1n)
					}

					rm(u)
					rm(tmapID_1n)
				}

				rm(geom_types, geom_types2, ids_point, ids_line, ids_poly, geom_all, geom_all2, is_stars)
			}
		}
	})
}
# args:
# - points_only: "yes", "no", "ifany"
#
# prop_ vectors, e.g. prop_angle can be added to shpTM. These are later put into dt (and used in step 4 (plotting))


#' @export
#' @rdname tmap_internal
tmapTransRaster = function(shpTM, ord__, plot.order, args) {
	if (!is.character(shpTM$shp) && !inherits(shpTM$shp, "dimensions")) cli::cli_abort("Stars or terra object expected in {.fun tm_shape} for {.fun tm_raster}")
	shpTM
}


#' @export
#' @rdname tmap_internal
tmapTransPolygons = function(shpTM, ord__, plot.order, args, scale) {
	within(shpTM, {
		if (!is.character(shp)) {
			is_stars = inherits(shp, "dimensions")
			if (is_stars && args$polygons.only != "yes") {
				### stars
				s = structure(list(values = matrix(TRUE, nrow = nrow(shp), ncol = ncol(shp))), dimensions = shp, class = "stars")
				shp = sf::st_as_sfc(s, as_points = FALSE)
				s = NULL
			} else if (is_stars) {
				shp = sf::st_sfc()
				tmapID = integer(0)
			} else {

				### sf
				geom_types = sf::st_geometry_type(shp)
				#crs = sf::st_crs(shp)

				ids_poly = which(geom_types %in% c("POLYGON", "MULTIPOLYGON"))
				ids_line = which(geom_types %in% c("LINESTRING", "MULTILINESTRING"))
				ids_point = which(geom_types %in% c("POINT", "MULTIPOINT"))


				if (args$polygons.only == "yes" || (args$polygons.only == "ifany" && length(ids_poly))) {
					shp = shp[ids_poly]
					tmapID = tmapID[ids_poly]
				} else {
					if (length(ids_line)) {
						tryCatch({
							shp[ids_line] = sf::st_cast(sf::st_cast(shp[ids_line], "MULTILINESTRING"), "MULTIPOLYGON")
						}, error = function(e) {
							stop("Unable to cast lines to polygon. Error from st_cast: \"", e$message, "\"", call. = FALSE)
						})
					}
					if (length(ids_point)) {
						dist = if (sf::st_is_longlat(shp)) 0.01 else 100
						shp[ids_point] = sf::st_buffer(shp[ids_point], dist = dist)
					}

				}
				rm(geom_types)
			}

			if (plot.order$aes == "AREA" && !is_stars) {
				o = order(without_units(sf::st_area(shp)), decreasing = !plot.order$reverse)
				shp = shp[o]
				tmapID = tmapID[o]
			}
		}
	})
}
# args:
# - polygons.only: "yes", "no", "ifany"

#' @export
#' @rdname tmap_internal
tmapTransLines = function(shpTM, ord__, plot.order, args, scale) {
	within(shpTM, {
		if (!is.character(shp)) {
			is_stars = inherits(shp, "dimensions")
			if (is_stars) {
				shp = sf::st_sfc()
				tmapID = integer(0)
			} else {

				### sf
				geom_types = sf::st_geometry_type(shp)
				#crs = sf::st_crs(shp)

				ids_poly = which(geom_types %in% c("POLYGON", "MULTIPOLYGON"))
				ids_line = which(geom_types %in% c("LINESTRING", "MULTILINESTRING"))
				ids_point = which(geom_types %in% c("POINT", "MULTIPOINT"))


				if (args$lines.only == "yes" || (args$lines.only == "ifany" && length(ids_line))) {
					shp = shp[ids_line]
					tmapID = tmapID[ids_line]
				} else {
					if (length(ids_poly)) {
						tryCatch({
							shp[ids_poly] = sf::st_cast(sf::st_cast(shp[ids_poly], "MULTIPOLYGON"), "MULTILINESTRING")
						}, error = function(e) {
							stop("Unable to cast to polygon. Error from st_cast: \"", e$message, "\"", call. = FALSE)
						})
					}
					if (length(ids_point)) {
						ids_not_point = which(!(geom_types %in% c("POINT", "MULTIPOINT")))
						shp = shp[ids_not_point]
						tmapID = tmapID[ids_not_point]
					}

				}
				rm(geom_types)
			}

			if (plot.order$aes == "LENGTH") {
				o = order(without_units(sf::st_length(shp)), decreasing = !plot.order$reverse)
				shp = shp[o]
				tmapID = tmapID[o]
			}
		}
	})
}
# args:
# - lines.only: "yes", "no", "ifany"

#' Map layer: symbols
#'
#' Map layer that draws symbols Supported visual variables are:
#' `fill` (the fill color), `col` (the border color), `size` the symbol size,
#' `shape` the symbol shape, `lwd` (line width), `lty` (line type), `fill_alpha`
#' (fill color alpha transparency) and `col_alpha` (border color alpha transparency).
#'
#' A symbol shape specification is one of the following three options.
#' \enumerate{
#'  \item{A numeric value that specifies the plotting character of the symbol. See parameter \code{pch} of \code{\link[graphics:points]{points}} and the last example to create a plot with all options. Note that this is not supported for the \code{"view" mode.}}
#'  \item{A \code{\link[grid:grid.grob]{grob}} object, which can be a ggplot2 plot object created with \code{\link[ggplot2:ggplotGrob]{ggplotGrob}}. To specify multiple shapes, a list of grob objects is required. Tip: for proportional symbols, such as donuts or pies, see the extension package \href{https://r-tmap.github.io/tmap/articles/ext_glyphs}{tmap.glyphs}}.
#'  \item{An icon specification, which can be created with \code{\link{tmap_icons}}.}
#'  }
#'  To specify multiple shapes (needed for the \code{shapes} argument), a vector or list of these shape specification is required. The shape specification options can also be mixed. For the \code{shapes} argument, it is possible to use a named vector or list, where the names correspond to the value of the variable specified by the \code{shape} argument.
#'  For small multiples, a list of these shape specification(s) should be provided.
#'
#' @param fill,fill.scale,fill.legend,fill.chart,fill.free `r .doc_vv("fill")`
#' @param col,col.scale,col.legend,col.chart,col.free `r .doc_vv("col")`
#' @param size,size.scale,size.legend,size.chart,size.free `r .doc_vv("size")`
#' @param shape,shape.scale,shape.legend,shape.chart,shape.free `r .doc_vv("shape")`
#' @param lwd,lwd.scale,lwd.legend,lwd.chart,lwd.free `r .doc_vv("lwd")`
#' @param lty,lty.scale,lty.legend,lty.chart,lty.free `r .doc_vv("lty")`
#' @param fill_alpha,fill_alpha.scale,fill_alpha.legend,fill_alpha.chart,fill_alpha.free `r .doc_vv("fill_alpha")`
#'   the fill color alpha transparency See details.
#' @param col_alpha,col_alpha.scale,col_alpha.legend,col_alpha.chart,col_alpha.free `r .doc_vv("col_alpha")`
#' @param plot.order Specification in which order the spatial features are drawn.
#'   See [tm_plot_order()] for details.
#' @param options options passed on to the corresponding `opt_<layer_function>` function
#' @inherit tm_polygons details
#' @inheritParams tm_polygons
#' @seealso \href{https://r-tmap.github.io/tmap/articles/examples_bubble}{Bubble map example} and \href{https://r-tmap.github.io/tmap/articles/examples_terrain}{terrain map example}
#' @param ... to catch deprecated arguments from version < 4.0
#' @example ./examples/tm_symbols.R
#' @export
tm_symbols = function(size = tm_const(),
					  size.scale = tm_scale(),
					  size.legend = tm_legend(),
					  size.chart = tm_chart_none(),
					  size.free = NA,
					  fill = tm_const(),
					  fill.scale = tm_scale(),
					  fill.legend = tm_legend(),
					  fill.chart = tm_chart_none(),
					  fill.free = NA,
					  col = tm_const(),
					  col.scale = tm_scale(),
					  col.legend = tm_legend(),
					  col.chart = tm_chart_none(),
					  col.free = NA,
					  shape = tm_const(),
					  shape.scale = tm_scale(),
					  shape.legend = tm_legend(),
					  shape.chart = tm_chart_none(),
					  shape.free = NA,
					  lwd = tm_const(),
					  lwd.scale = tm_scale(),
					  lwd.legend = tm_legend(),
					  lwd.chart = tm_chart_none(),
					  lwd.free = NA,
					  lty = tm_const(),
					  lty.scale = tm_scale(),
					  lty.legend = tm_legend(),
					  lty.chart = tm_chart_none(),
					  lty.free = NA,
					  fill_alpha = tm_const(),
					  fill_alpha.scale = tm_scale(),
					  fill_alpha.legend = tm_legend(),
					  fill_alpha.chart = tm_chart_none(),
					  fill_alpha.free = NA,
					  col_alpha = tm_const(),
					  col_alpha.scale = tm_scale(),
					  col_alpha.legend = tm_legend(),
					  col_alpha.chart = tm_chart_none(),
					  col_alpha.free = NA,
					  plot.order = tm_plot_order("size"),
					  zindex = NA,
					  group = NA,
					  group.control = "check",
					  popup.vars = NA,
					  popup.format = tm_label_format(),
					  hover = NA,
					  id = "",
					  options = opt_tm_symbols(),
					  ...) {

	args = list(...)
	args_called = names(rlang::call_match()[-1])


	res = v3_symbols(args, args_called, col)
	if (!is.null(res)) {
		fill = res$fill
		col = res$col
		col_alpha = res$col_alpha
		fill_alpha = res$fill_alpha
		fill.scale = res$fill.scale
		fill.legend = res$fill.legend
		size.scale = res$size.scale
		size.legend = res$size.legend
		shape.scale = res$shape.scale
		shape.legend = res$shape.legend
		fill.chart = res$fill.chart
	}

	# unused arguments: typos?
	unused = setdiff(names(args), c(v3_only("tm_symbols"), "called_from"))

	if (length(unused)) {
		layer_fun = if ("called_from" %in% names(args)) {
			args$called_from
		} else {
			"tm_symbols"
		}

		message_layer_unused_args(layer_fun, unused)
	}

	# make sure required options are there
	options = complete_options(options, opt_tm_symbols())


	tm_element_list(tm_element(
		layer = "symbols",
		trans.fun = tmapTransCentroid,
		trans.aes = list(),
		trans.args = options$trans.args,
		trans.apply_to = "this",
		mapping.aes = list(size = tmapScale(aes = "size",
											value = size,
											scale = size.scale,
											legend = size.legend,
											chart = size.chart,
											free = size.free),

						   fill = tmapScale(aes = "fill",
						   				 value = fill,
						   				 scale = fill.scale,
						   				 legend = fill.legend,
						   				 chart = fill.chart,
						   				 free = fill.free),
						   col = tmapScale(aes = "col",
						   				value = col,
						   				scale = col.scale,
						   				legend = col.legend,
						   				chart = col.chart,
						   				free = col.free),
						   shape = tmapScale(aes = "shape",
						   				  value = shape,
						   				  scale = shape.scale,
						   				  legend = shape.legend,
						   				  chart = shape.chart,
						   				  free = shape.free),
						   lwd = tmapScale(aes = "lwd",
						   				value = lwd,
						   				scale = lwd.scale,
						   				legend = lwd.legend,
						   				chart = lwd.chart,
						   				free = lwd.free),
						   lty = tmapScale(aes = "lty",
						   				value = lty,
						   				scale = lty.scale,
						   				legend = lty.legend,
						   				chart = lty.chart,
						   				free = lty.free),
						   fill_alpha = tmapScale(aes = "fill_alpha",
						   					   value = fill_alpha,
						   					   scale = fill_alpha.scale,
						   					   legend = fill_alpha.legend,
						   					   chart = fill_alpha.chart,
						   					   free = fill_alpha.free),
						   col_alpha = tmapScale(aes = "col_alpha",
						   					  value = col_alpha,
						   					  scale = col_alpha.scale,
						   					  legend = col_alpha.legend,
						   					  chart = col_alpha.chart,
						   					  free = col_alpha.free)),

		gpar = tmapGpar(fill = "__fill",
						col = "__col",
						shape = "__shape",
						size = "__size",
						fill_alpha = "__fill_alpha",
						col_alpha = "__col_alpha",
						pattern = "fill",
						lty = "__lty",
						lwd = "__lwd",
						linejoin = NA,
						lineend = NA),
		tpar = tmapTpar(),
		plot.order = plot.order,
		mapping.args = options$mapping.args,
		zindex = zindex,
		group = group,
		group.control = group.control,
		popup.vars = popup.vars,
		popup.format = popup.format,
		hover = hover,
		id = id,
		subclass = c("tm_aes_layer", "tm_layer")))
}

v3_symbols = function(args, args_called, arg_col = NULL) {
	if (any(v3_only("tm_symbols") %in% names(args))) {
		layer_fun = if ("called_from" %in% names(args)) {
			args$called_from
		} else {
			"symbols"
		}

		v3_start_message()
		if (!("style" %in% names(args))) {
			if (!"breaks" %in% names(args)) {
				style = "pretty"
			} else {
				style = "fixed"
			}
		} else {
			style = args$style
		}

		v3_list_init()
		if (length(style) > 1) {
			style = style[1]
			.TMAP$v3_list$mult = TRUE
		}

		# v3 visual variable: fill
		fill.scale.args = c(list(n = v3_impute(args, "n", 5),
								 style = style,
								 style.args = v3_impute(args, "style.args", list())),
							if (style %in% c("cont", "log10")) {
								c({
									if (!is.null(args$breaks)) {
										if (length(args$breaks) > 2) {
											list(ticks = v3_impute(args, "breaks", NULL, "ticks"))
										} else {
											list(ticks = v3_impute(args, "breaks", NULL, "limits"))
										}
									} else {
										list()
									}
								},
								list(outliers.trunc = c(TRUE, FALSE)))
							} else {
								list(breaks = v3_impute(args, "breaks", NULL),
									 interval.closure = v3_impute(args, "interval.closure", "left"),
									 drop.levels = v3_impute(args, "drop.levels", FALSE))
							},
							list(midpoint = v3_impute(args, "midpoint", NULL),
								 as.count = v3_impute(args, "as.count", NA),
								 values = v3_impute(args, "palette", NA, "values"),
								 values.repeat = !v3_impute(args, "stretch.palette", TRUE, "values.repeat"),
								 values.range = v3_impute(args, "contrast", NA, "values.range"),
								 values.scale = 1,
								 value.na = v3_impute(args, "colorNA", NA, "value.na"),
								 value.null = v3_impute(args, "colorNULL", NA, "value.null"),
								 value.neutral = NA,
								 labels = v3_impute(args, "labels", NULL),
								 label.na = v3_impute(args, "textNA", "Missing", "label.na"),
								 label.null = NA,
								 label.format = v3_impute(args, "legend.format", list(), "label.format")))
		fill.scale.args$fun_pref = if (style == "cat") {
			"categorical"
		} else if (style %in% c("fixed", "sd", "equal", "pretty", "quantile",
								"kmeans", "hclust", "bclust", "fisher", "jenks",
								"dpih", "headtails", "log10_pretty")) {
			"intervals"
		} else if (style == "cont") {
			"continuous"
		} else if (style == "log10") {
			"continuous_log"
		} else if (style == "order") {
			"rank"
		} else {
			stop("unknown style")
		}

		if ("style" %in% names(args)) {
			v3_tm_scale_instead_of_style(style, scale_fun = fill.scale.args$fun_pref, vv = "fill", layer_fun = layer_fun, arg_list = v3_list_get())
		} else {
			v3_tm_scale(scale_fun = "", vv = "fill", layer_fun = layer_fun, arg_list = v3_list_get())
		}
		fill.scale = do.call("tm_scale", args = fill.scale.args)

		if ("col" %in% args_called) {
			fill = arg_col
			isn_fill = is.null(fill)
			v3_message_col_fill(layer_fun = layer_fun)
			if (isn_fill) {
				v3_message_vv_null(layer_fun = layer_fun)
				fill = NA
			}
			col = tm_const()
		} else {
			fill = tm_const()
			col = tm_const()
			isn_fill = FALSE
		}

		if ("border.col" %in% names(args)) {
			col = args$border.col
			isn_col = is.null(col)
			if (!("col" %in% args_called)) v3_message_col_fill(layer_fun = layer_fun)
			if (isn_col) {
				if (!isn_fill) v3_message_vv_null(layer_fun = layer_fun)
				col = NA
			}
		}


		if ("alpha" %in% names(args)) {
			v3_message_fill_alpha(layer_fun = layer_fun)
			fill_alpha = args$alpha
		} else {
			fill_alpha = tm_const()
		}

		if ("border.alpha" %in% names(args)) {
			v3_message_col_alpha(layer_fun = layer_fun)
			col_alpha = args$border.alpha
		} else {
			col_alpha = tm_const()
		}

		v3_list_init()
		if ("legend.show" %in% names(args) && !args$legend.show) {
			v3_tm_legend_hide(layer_fun, arg = "legend.show", vv = "fill")
			fill.legend = tm_legend_hide()
		} else {
			fill.legend.args = list(atitle = v3_impute(args, "title.col", NA, "title"),
									na.show = v3_impute(args, "showNA", NA, "na.show"),
									format = v3_impute(args, "legend.format", list(), "format"),
									orientation = ifelse(v3_impute(args, "legend.is.portrait", TRUE, "orientation"), "portrait", "landscape"),
									reverse = v3_impute(args, "legend.reverse", FALSE, "reverse"))
			fill.legend = do.call("tm_legend", fill.legend.args)
			v3_tm_legend(fun = layer_fun, vv = "fill", arg_list = v3_list_get())
		}


		# v3 visual variable: size
		v3_list_init()
		size.scale.args = list(ticks = v3_impute(args, "sizes.legend", NULL, "ticks"),
							   values.range = {if ("size.lim" %in% names(args)) {
							   	c(args[["size.lim"]][1] / args[["size.lim"]][2], 1)
							   } else c(0, 1)},
							   values.scale = v3_impute(args, "scale", 1, "values.scale"),
							   limits = v3_impute(args, "size.lim", NULL, "limits"),
							   outliers.trunc = c(TRUE, FALSE),
							   labels = v3_impute(args, "sizes.legend.labels", NULL, "labels"),
							   fun_pref = "continuous")
		if ("size" %in% args_called) v3_tm_scale(scale_fun = "continuous", vv = "size", layer_fun = layer_fun, arg_list = v3_list_get())
		size.scale = do.call("tm_scale", args = size.scale.args)

		v3_list_init()
		if ("legend.size.show" %in% names(args) && !args$legend.size.show) {
			v3_tm_legend_hide(layer_fun, arg = "legend.size.show", vv = "size")
			size.legend = tm_legend_hide()
		} else {
			size.legend.args = alist(title = v3_impute(args, "title.size", NA, "title"),
									na.show = v3_impute(args, "showNA", NA),
									format = v3_impute(args, "legend.format", list(), "format"),
									orientation = ifelse(v3_impute(args, "legend.size.is.portrait", FALSE), "portrait", "landscape"),
									reverse = v3_impute(args, "legend.size.reverse", FALSE))
			if ("size" %in% args_called) v3_tm_legend(fun = layer_fun, vv = "size", arg_list = v3_list_get())
			size.legend = do.call("tm_legend", size.legend.args)
		}

		# v3 visual variable: shape
		if (!("shapes.style" %in% names(args))) {
			if (!"shapes.breaks" %in% names(args)) {
				shapes.style = "pretty"
			} else {
				shapes.style = "fixed"
			}
		} else {
			shapes.style = args$shapes.style
		}

		v3_list_init()
		shape.scale.args = list(n = v3_impute(args, "shapes.n", 5, "n"),
								style = shapes.style,
								style.args = v3_impute(args, "shapes.style.args", list(), "style.args"),
								breaks = v3_impute(args, "shapes.breaks", NULL, "breaks"),
								interval.closure = v3_impute(args, "shapes.interval.closure", "left", "interval.closure"),
								drop.levels = v3_impute(args, "drop.levels", FALSE),
								midpoint = v3_impute(args, "midpoint", NULL),
								as.count = v3_impute(args, "as.count", NA),
								value.neutral = v3_impute(args, "shapes.legend", NA, "value.neutral"),
								values = v3_impute(args, "shapes", 21:25, "values"),
								labels = v3_impute(args, "shapes.labels", NULL, "labels"),
								label.na = v3_impute(args, "shape.textNA", NA, "label.na"),
								label.null = NA,
								label.format = v3_impute(args, "legend.format", list(), "label.format"),
								fun_pref = "intervals")
		shape.scale = do.call("tm_scale", args = shape.scale.args)
		if ("shape" %in% args_called) {
			if ("shapes.style" %in% args) {
				v3_tm_scale_instead_of_style(shapes.style, scale_fun = shape.scale.args$fun_pref, vv = "shape", layer_fun = layer_fun, arg_list = v3_list_get())
			} else {
				v3_tm_scale(scale_fun = shape.scale.args$fun_pref, vv = "shape", layer_fun = layer_fun, arg_list = v3_list_get())
			}
		}

		v3_list_init()
		if ("legend.shape.show" %in% names(args) && !args$legend.shape.show) {
			v3_tm_legend_hide(layer_fun, arg = "legend.shape.show", vv = "shape")
			shape.legend = tm_legend_hide()
		} else {
			shape.legend.args = alist(title = v3_impute(args, "title.shape", NA),
									 na.show = v3_impute(args, "shape.showNA ", NA),
									 format = v3_impute(args, "legend.format", list(), "format"),
									 orientation = ifelse(v3_impute(args, "legend.shape.is.portrait", TRUE), "portrait", "landscape"),
									 reverse = v3_impute(args, "legend.shape.reverse", FALSE))
			if ("shape" %in% args_called)v3_tm_legend(fun = layer_fun, vv = "shape", arg_list = v3_list_get())
			shape.legend = do.call("tm_legend", shape.legend.args)
		}

		if ("legend.hist" %in% names(args) && args$legend.hist) {
			fill.chart = tm_chart_histogram()
			v3_tm_chart_hist(layer_fun = layer_fun, vv = "fill", arg = "legend.hist")

			# to do: histogram title
		} else {
			fill.chart = tm_chart_none()
		}


		v4_opt_args = c("icon.scale", "just", "grob.dim")
		v3_opt_args = c("icon.scale", "just", "grob.dim")
		osel = which(v3_opt_args %in% names(args))
		if (length(osel)) {
			o3 = v3_opt_args[osel]
			o4 = v4_opt_args[osel]
			v3_layer_opt(o3, o4, layer_fun)
		}

		list(fill = fill,
			 col = col,
			 col_alpha = col_alpha,
			 fill_alpha = fill_alpha,
			 fill.scale = fill.scale,
			 fill.legend = fill.legend,
			 size.scale = size.scale,
			 size.legend = size.legend,
			 shape.scale = shape.scale,
			 shape.legend = shape.legend,
			 fill.chart = fill.chart)
	} else {
		NULL
	}
}



#' @export
#' @rdname tm_symbols
tm_dots = function(fill = tm_const(),
				   fill.scale = tm_scale(),
				   fill.legend = tm_legend(),
				   fill.free = NA,
				   size = tm_const(),
				   size.scale = tm_scale(),
				   size.legend = tm_legend(),
				   size.free = NA,
				   lwd = tm_const(),
				   lwd.scale = tm_scale(),
				   lwd.legend = tm_legend(),
				   lwd.free = NA,
				   lty = tm_const(),
				   lty.scale = tm_scale(),
				   lty.legend = tm_legend(),
				   lty.free = NA,
				   fill_alpha = tm_const(),
				   fill_alpha.scale = tm_scale(),
				   fill_alpha.legend = tm_legend(),
				   fill_alpha.free = NA,
				   plot.order = tm_plot_order("DATA"),
				   zindex = NA,
				   group = NA,
				   group.control = "check",
				   options = opt_tm_dots(),
				   ...) {

		args = c(as.list(environment()), list(...))
		args$called_from = "tm_dots"

		tm = do.call(tm_symbols, args)
		tm[[1]]$layer = c("dots", "symbols")
		tm
}


#' @export
#' @rdname tm_symbols
tm_bubbles = function(size = tm_const(),
					  size.scale = tm_scale(),
					  size.legend = tm_legend(),
					  size.free = NA,
					  fill = tm_const(),
					  fill.scale = tm_scale(),
					  fill.legend = tm_legend(),
					  fill.free = NA,
					  col = tm_const(),
					  col.scale = tm_scale(),
					  col.legend = tm_legend(),
					  col.free = NA,
					  lwd = tm_const(),
					  lwd.scale = tm_scale(),
					  lwd.legend = tm_legend(),
					  lwd.free = NA,
					  lty = tm_const(),
					  lty.scale = tm_scale(),
					  lty.legend = tm_legend(),
					  lty.free = NA,
					  fill_alpha = tm_const(),
					  fill_alpha.scale = tm_scale(),
					  fill_alpha.legend = tm_legend(),
					  fill_alpha.free = NA,
					  col_alpha = tm_const(),
					  col_alpha.scale = tm_scale(),
					  col_alpha.legend = tm_legend(),
					  col_alpha.free = NA,
					  plot.order = tm_plot_order("size"),
					  zindex = NA,
					  group = NA,
					  group.control = "check",
					  options = opt_tm_bubbles(),
					  ...) {

	args = c(as.list(environment()), list(...))
	args$called_from = "tm_bubbles"

	tm = do.call(tm_symbols, args)
	tm[[1]]$layer = c("bubbles", "symbols")
	tm
}


#' @export
#' @rdname tm_symbols
tm_squares = function(size = tm_const(),
					  size.scale = tm_scale(),
					  size.legend = tm_legend(),
					  size.free = NA,
					  fill = tm_const(),
					  fill.scale = tm_scale(),
					  fill.legend = tm_legend(),
					  fill.free = NA,
					  col = tm_const(),
					  col.scale = tm_scale(),
					  col.legend = tm_legend(),
					  col.free = NA,
					  lwd = tm_const(),
					  lwd.scale = tm_scale(),
					  lwd.legend = tm_legend(),
					  lwd.free = NA,
					  lty = tm_const(),
					  lty.scale = tm_scale(),
					  lty.legend = tm_legend(),
					  lty.free = NA,
					  fill_alpha = tm_const(),
					  fill_alpha.scale = tm_scale(),
					  fill_alpha.legend = tm_legend(),
					  fill_alpha.free = NA,
					  col_alpha = tm_const(),
					  col_alpha.scale = tm_scale(),
					  col_alpha.legend = tm_legend(),
					  col_alpha.free = NA,
					  plot.order = tm_plot_order("size"),
					  zindex = NA,
					  group = NA,
					  group.control = "check",
					  options = opt_tm_squares(),
					  ...) {

	args = c(as.list(environment()), list(...))
	args$called_from = "tm_squares"
	tm = do.call(tm_symbols, args)
	tm[[1]]$layer = c("squares", "symbols")
	tm
}

#' @rdname tm_symbols
#' @inheritParams tm_text
#' @export
tm_markers = function(text = tm_const(),
					  text.scale = tm_scale(),
					  text.legend = tm_legend(),
					  text.chart = tm_chart_none(),
					  text.free = NA,
					  size = tm_const(),
					  size.scale = tm_scale(),
					  size.legend = tm_legend(),
					  size.chart = tm_chart_none(),
					  size.free = NA,
					  col = tm_const(),
					  col.scale = tm_scale(),
					  col.legend = tm_legend(),
					  col.chart = tm_chart_none(),
					  col.free = NA,
					  col_alpha = tm_const(),
					  col_alpha.scale = tm_scale(),
					  col_alpha.legend = tm_legend(),
					  col_alpha.chart = tm_chart_none(),
					  col_alpha.free = NA,
					  fontface = tm_const(),
					  fontface.scale = tm_scale(),
					  fontface.legend = tm_legend(),
					  fontface.chart = tm_chart_none(),
					  fontface.free = NA,
					  fontfamily = "",
					  bgcol = tm_const(),
					  bgcol.scale = tm_scale(),
					  bgcol.legend = tm_legend(),
					  bgcol.chart = tm_chart_none(),
					  bgcol.free = NA,
					  bgcol_alpha = tm_const(),
					  bgcol_alpha.scale = tm_scale(),
					  bgcol_alpha.legend = tm_legend(),
					  bgcol_alpha.chart = tm_chart_none(),
					  bgcol_alpha.free = NA,
					  xmod = 0,
					  xmod.scale = tm_scale(),
					  xmod.legend = tm_legend_hide(),
					  xmod.chart = tm_chart_none(),
					  xmod.free = NA,
					  ymod = 0,
					  ymod.scale = tm_scale(),
					  ymod.legend = tm_legend_hide(),
					  ymod.chart = tm_chart_none(),
					  ymod.free = NA,
					  angle = 0,
					  angle.scale = tm_scale(),
					  angle.legend = tm_legend_hide(),
					  angle.chart = tm_chart_none(),
					  angle.free = NA,
					  plot.order = tm_plot_order("AREA", reverse = FALSE, na.order = "bottom"),
					  zindex = NA,
					  group = NA,
					  group.control = "check",
					  options = opt_tm_markers(),
					  ...) {

	if (!setequal(names(options), c("markers", "text", "dots"))) cli::cli_abort("{.field [tm_markers]} options should have {.str markers}, {.str text} and {.str dots} items")
	e = as.list(environment())
	e$options = options$text
	a = list(...)

	sel = substr(names(a), 1, 5) == "dots_"
	a_dots = a[sel]
	a_other = a[!sel]
	names(a_dots) = substr(names(a_dots), 6, nchar(names(a_dots)))

	args_dots = c(a_dots, list(plot.order = plot.order, zindex = zindex, group = group, group.control = group.control, options = options$dots))
	args_text = c(e, a_other, list(called_from = "tm_markers"))

	tm_t = do.call(tm_text, args_text)
	tm_t[[1]]$layer = c("labels", "text")

	tm_d = do.call(tm_dots, args_dots)
	tm_d[[1]]$layer = c("markers", "symbols")

	if (options$markers$markers_on_top_of_text) tm_t + tm_d else tm_d + tm_t

}

#' @inheritParams opt_tm_labels
#' @param markers_on_top_of_text should markers be plot on top of the text (by default `FALSE`)
#' @param dots.icon.scale scaling number that determines how large the icons (or grobs) are in plot mode in comparison to proportional symbols (such as bubbles). In view mode, the size is determined by the icon specification (see \code{\link{tmap_icons}}) or, if grobs are specified by \code{grob.width} and \code{grob.height}
#' @param dots.just justification of the text relative to the point coordinates. Either one of the following values: \code{"left"} , \code{"right"}, \code{"center"}, \code{"bottom"}, and \code{"top"}, or a vector of two values where first value specifies horizontal and the second value vertical justification. Besides the mentioned values, also numeric values between 0 and 1 can be used. 0 means left justification for the first value and bottom justification for the second value. Note that in view mode, only one value is used.
#' @param dots.grob.dim vector of four values that determine how grob objects (see details) are shown in view mode. The first and second value are the width and height of the displayed icon. The third and fourth value are the width and height of the rendered png image that is used for the icon. Generally, the third and fourth value should be large enough to render a ggplot2 graphic successfully. Only needed for the view mode.
#' @rdname tm_symbols
#' @export
opt_tm_markers = function(markers_on_top_of_text = FALSE,
						  points_only = "ifany",
						  point_per = "feature",
						  on_surface = FALSE,
						  shadow = FALSE,
						  shadow.offset.x = 0.1,
						  shadow.offset.y = 0.1,
						  just = "center",
						  along_lines = TRUE,
						  bg.padding = 0.4,
						  clustering = TRUE,
						  point.label = TRUE,
						  point.label.gap = 0.4,
						  point.label.method = "SANN",
						  remove_overlap = FALSE,
						  dots.just = NA,
						  dots.icon.scale = 3,
						  dots.grob.dim = c(width=48, height=48, render.width=256, render.height=256)) {
	list(markers = list(markers_on_top_of_text = markers_on_top_of_text),
		 text = opt_tm_labels(points_only = points_only,
		 					 point_per = point_per,
		 					 on_surface = on_surface,
		 					 shadow = shadow,
		 					 shadow.offset.x = shadow.offset.x,
		 					 shadow.offset.y = shadow.offset.y,
		 					 just = just,
		 					 along_lines = along_lines,
		 					 bg.padding = bg.padding,
		 					 clustering = clustering,
		 					 point.label = point.label,
		 					 point.label.gap = point.label.gap,
		 					 point.label.method = point.label.method,
		 					 remove_overlap = remove_overlap),
		 dots = opt_tm_dots(points_only = points_only,
		 				   point_per = point_per,
		 				   on_surface = on_surface,
		 				   clustering = clustering,
		 				   icon.scale = dots.icon.scale,
		 				   just = dots.just,
		 				   grob.dim = dots.grob.dim))
}


#' @param icon.scale scaling number that determines how large the icons (or grobs) are in plot mode in comparison to proportional symbols (such as bubbles). For view mode, use the argument `grob.dim`
#' @param grob.dim vector of four values that determine how grob objects (see details) are shown in view mode. The first and second value are the width and height of the displayed icon. The third and fourth value are the width and height of the rendered png image that is used for the icon. Generally, the third and fourth value should be large enough to render a ggplot2 graphic successfully. Only needed for the view mode.
#' @param clustering in interactive modes (e.g. \code{"view"} mode), should clustering be applied at lower zoom levels? Either `FALSE` (default), `TRUE`, or a mode specific specification, e.g. for \code{"view"} mode \code{\link[leaflet:markerClusterOptions]{markerClusterOptions}}.
#' @rdname tm_symbols
#' @export
opt_tm_symbols = function(points_only = "ifany",
						  point_per = "feature",
						  on_surface = FALSE,
						  clustering = FALSE,
						  icon.scale = 3,
						  just = NA,
						  grob.dim = c(width=48, height=48, render.width=256, render.height=256)) {
	list(trans.args = list(points_only = points_only, point_per = point_per, on_surface = on_surface, along_lines = FALSE),
		 mapping.args = list(icon.scale = icon.scale,
		 					just = just,
		 					clustering = clustering,
		 					grob.dim = grob.dim))
}

#' @rdname tm_symbols
#' @export
opt_tm_dots = opt_tm_symbols


#' @rdname tm_symbols
#' @export
opt_tm_bubbles = opt_tm_symbols

#' @rdname tm_symbols
#' @export
opt_tm_squares = opt_tm_symbols

# not working yet, see #1024
renderTmapLeaflet = function(expr, env, quoted, execOnResize) {
	if (!quoted)
		expr = substitute(expr)
	expr = substitute(getFromNamespace("print.tmap", "tmap")(expr, in.shiny = TRUE))
	htmlwidgets::shinyRenderWidget(expr, leafletOutput, env,
								   quoted = TRUE)
}

tmapOutputLeaflet = function(outputId, width, height) {
	leaflet::leafletOutput(outputId, width = width, height = height)
}


tmapProxyLeaflet = function(mapId, session, x) {
	print.tmap(x, lf = leaflet::leafletProxy(mapId, session), show = FALSE, in.shiny = TRUE, proxy = TRUE)
}

#' World dataset
#'
#' World dataset, class [`sf`][`sf::sf`]
#'
#' @details
#' | **Variable**        | **Source**	| **Description** |
#' | ------         	| -----------    	| ----------- |
#' | `iso_a3`		    | NED | ISO 3166-1 alpha-3 three-letter country code (see below) |
#' | `name`		    	| NED | Country name |
#' | `sovereignt`		| NED | Sovereignt country name |
#' | `continent`		| NED | Continent (primary; some countries are transcontinental) |
#' | `area`				| NED | Area in km2 |
#' | `pop_est`			| NED | Population estimation |
#' | `pop_est_dens`		| NED | Population estimation per km2|
#' | `economy`			| NED | Economy class |
#' | `income_grp`		| NED | Income group |
#' | `gdp_cap_est`		| NED | GDP per capita (estimated) |
#' | `life_exp`			| HPI | Life expectancy. The average number of years an infant born in that country is expected to live  |
#' | `well_being`		| HPI | Well being. Self-reported from 0 (worst) to 10 (best) |
#' | `footprint`		| HPI | Carbon footprint. Per capita greendwelling gas emissions associated with consumption and economic activity |
#' | `HPI`				| HPI | Happy Planet Indicator. An index of human well-being and environmental impact that was introduced by the New Economics Foundation in 2006. Approximate formula: `(life_exp * well_being) / footprint` |
#' | `inequality`		| WB | Income inequality: Gini coefficient (World Bank variable SI.POV.GINI) A value of 0 represents perfect equality, while 100 implies perfect inequality |
#' | `gender`			| UNDP/OWiD | Gender Inequality Index (GII) Composite metric using reproductive health, empowerment and the labour market. A low value indicates low inequality between women and men, and vice-versa |
#' | `press`			| RSF | World Press Freedom Index. Degree of freedom that journalists, news organizations and netizens have |
#'
#' See sources for more detailed information about the variables.
#'
#' This dataset, created Noveber 2024, is an update from the old version, which has been created around 2016. All variables from the old version are included, but updated. Furthermore, gender ineuqlity and press freedom have been added.
#'
#' ISO country-code: two countries have user-assigned codes, namely: XKX is used for Kosovo (conform European Union and World Bank) (was UNK in the old version); XNC is used for Northern Cyprus (was CYN in the old version).
#'
#' For some variables data were available from multiple years, but availability was different across countries. In those cases, the most recent values were taken.
#' @rdname World
#' @docType data
#' @format NULL
#' @source NED: Natural Earth Data <https://www.naturalearthdata.com/>
#' @source HPI: Happy Planet Index <https://happyplanetindex.org/>
#' @source UNDP: Human Development Report (2024) <https://hdr.undp.org/content/human-development-report-2023-24>
#' @source WB: World Bank <https://data.worldbank.org>
#' @source OWiD: Our World in Data <https://ourworldindata.org>
#' @source RSF: Reporters Without Borders <https://rsf.org/en/index>
"World"



#' Netherlands datasets
#'
#' Datasets of the Netherlands for 2022 at three levels: `NLD_prov` (12) provinces, `NLD_muni` (345) municipalities and `NLD_dist` (3340) districts , all class [`sf`][`sf::sf`]
#'
#' @details
#' The data variables for `NLD_muni` and `NLD_dist` are identical:
#'
#' | **Variable**        | **Description** |
#' | ------          | ----------- |
#' | `code`		     | Code. Format is "GMaaaa" (municipality/'**g**e**m**eente') and "WKaaaabb" (district/**w**ij**k**). Here, "aaaa" represents the municipality id number, and "bb" the additional district number.
#' | `name`		     | Name. |
#' | `province`	 | Province name. |
#' | `area`		     | Total area in km2. This area corresponds to the area of the polygons (including inland waters, excluding coastal waters), but is more precise because it is based on non-simplified geometries. |
#' | `urbanity`		 | Level of urbanity. Five classes, determined by the number of addresses per km2 (break values are 2500, 1500, 1000, and 500). |
#' | `population`	 | The total population count at 2022-01-01. |
#' | `pop_0_14`		 | Percentage (rounded) of people between 0 and 15. |
#' | `pop_15_24`	 | Percentage (rounded) of people between 15 and 25.  |
#' | `pop_25_44`	 | Percentage (rounded) of people between 25 and 45.  |
#' | `pop_45_64`	 | Percentage (rounded) of people between 45 and 65.  |
#' | `pop_65plus`	 | Percentage (rounded) of people of 65 and older. |
#' | `dwelling_total`	 | Number of dwellings. |
#' | `dwelling_value`	 | Average dwelling value (Dutch: WOZ-value). |
#' | `dwelling_ownership`| Percentage of dwellings owned by the residents. |
#' | `employment_rate`|  Share of the employed population within the total population from 15 to 75 years old. |
#' | `income_low`	 | Percentage of individuals in private households belonging to the lowest 40% of personal income nationwide. |
#' | `income_high`	 | Percentage of individuals in private households belonging to the highest 20% of personal income nationwide. |
#' | `edu_appl_sci`	 | Percentage of people aged 15 to 75 with a university of applied sciences (Dutch: HBO) or university (Dutch: WO) degree. |
#'
#' See source for detailed information about the variables.
#'
#' This dataset, created Noveber 2024, is an update from the datasets `NLD_muni` and `NLD_prov` used in tmap <= 3, which has been created around 2016. Note that the number of municipalities have been reduced (due to mergings). All old variables are included, except for variables related to ethnicity. Many new variable have been added, and moreover, district (Dutch: wijk) level data have added: `NLD_dist`.
#'
#' The CRS (coordinate reference system) used is the Rijksdriekhoekstelsel New, EPSG 28992. Coordinates have been rounded to meters to reduce file size.
#'
#' @rdname Netherlands
#' @docType data
#' @format NULL
#' @source https://www.cbs.nl/nl-nl/maatwerk/2024/11/kerncijfers-wijken-en-buurten-2022
#' @references Statistics Netherlands (2024), The Hague/Heerlen, Netherlands, <https://www.cbs.nl/>.
"NLD_prov"

#' @rdname Netherlands
#' @docType data
#' @format NULL
"NLD_muni"

#' @rdname Netherlands
#' @docType data
#' @format NULL
"NLD_dist"

#' Spatial data of rivers
#'
#' @source <https://www.naturalearthdata.com>
#' @note In tmap <= 3, this dataset was called `rivers`.
"World_rivers"

#' Spatial data of metropolitan areas
#'
#' `metro` includes a population time series from 1950 to (forecasted) 2030. All metro areas with over 1 million inhabitants in 2010 are included.
#'
#' @references United Nations, Department of Economic and Social Affairs, Population Division (2014). World Urbanization Prospects: The 2014 Revision, CD-ROM Edition.
#' @source <https://population.un.org/wup/>
"metro"

#' Spatial data of global land cover
#'
#' Spatial data of global land cover, percent tree cover, and elevation of class [`stars`][stars::st_as_stars()].
#' Two attributes in this object relates to global land cover.
#' The cover layer classifies the status of land cover of the whole globe into 20 categories, while
#' the cover_cls layer uses 8 simplified categories.
#' Percent Tree Cover (trees) represents the density of trees on the ground, and the last attribute represents elevation.
#'
#' **Important:** publication of these maps is only allowed when cited to Tateishi et al. (2014), and when "Geospatial Information Authority of Japan, Chiba University and collaborating organizations." is shown.
#'
#' @references Production of Global Land Cover Data - GLCNMO2008, Tateishi, R., Thanh Hoan, N., Kobayashi, T., Alsaaideh, B., Tana, G., Xuan Phong, D. (2014), Journal of Geography and Geology, 6 (3).
"land"

get_facet_id = function(row, col, nrow, ncol) {
	col + (row - 1L) * ncol
}

get_lf = function(facet_row, facet_col, facet_page) {
	lfs = get("lfs", envir = .TMAP_LEAFLET)
	nrow = get("nrow", envir = .TMAP_LEAFLET)
	ncol = get("ncol", envir = .TMAP_LEAFLET)

	lfsi = lfs[[facet_page]]

	fr = max(1, facet_row) # facet_row can be -1 or -2
	fc = max(1, facet_col) # facet_row can be -1 or -2

	lfid = get_facet_id(fr, fc, nrow, ncol)

	lfsi[[lfid]]
}

assign_lf = function(lf, facet_row, facet_col, facet_page) {
	lfs = get("lfs", envir = .TMAP_LEAFLET)
	nrow = get("nrow", envir = .TMAP_LEAFLET)
	ncol = get("ncol", envir = .TMAP_LEAFLET)


	fr = max(1, facet_row) # facet_row can be -1 or -2
	fc = max(1, facet_col) # facet_row can be -1 or -2



	lfid = get_facet_id(fr, fc, nrow, ncol)

	lfs[[facet_page]][[lfid]] = lf
	assign("lfs", lfs, envir = .TMAP_LEAFLET)
	NULL
}

getClusterOpts = function(clustering) {
	if (identical(clustering, TRUE)) {
		clusterOpts = leaflet::markerClusterOptions()
	} else if (identical(clustering, FALSE)) {
		clusterOpts = NULL
	} else {
		clusterArgs = setdiff(names(formals(leaflet::markerClusterOptions)), "...")
		if (!all(clusterArgs %in% names(clustering))) cli::cli_abort("{.field clustering} Cluster options should have the same options as {.fun leaflet::markerClusterOptions}")
		clusterOpts = clustering
	}
}


tmapLeafletWrap = function(label, facet_row, facet_col, facet_page, o) {

	NULL
}

tmapLeafletXtab = function(label, facet_row, facet_col, facet_page, o) {
}

blend_lf = function(lf, blend, pane) {
	if (is.null(blend) || blend == "over") return(lf)
	htmlwidgets::onRender(lf, sprintf("
        function(el, x) {
            var pane = el.querySelector('.leaflet-%s-pane');
            if (pane) pane.style.mixBlendMode = '%s';
        }
    ", pane, blend))
}

#' @export
tmapShape.SpatRaster = function(shp, is.main, crs, bbox, unit, filter, layer, shp_name, smeta, o, tmf) {
	#tmapShape.stars(stars::st_as_stars(shp), is.main, crs, bbox, unit, filter, shp_name)
	rlang::check_installed("terra")

	shp = downsample_SpatRaster(shp, max.raster = max_cells(o$raster.max_cells) / (o$fn[1] * o$fn[2]))


	ctabs = terra::coltab(shp)
	cats = terra::levels(shp)

	minmax = as.list(as.data.frame(terra::minmax(shp, compute = all(!terra::hasMinMax(shp)))))

	dt = data.table::setDT(terra::as.data.frame(shp, na.rm=FALSE))
	dt[, tmapID__:=1L:.N]
	#dt = data.table::melt(dt, id.vars = "tmapID__", variable.name = "layer", value.name = "value")

	xy_dim = dim(shp)[1:2]
	b = sf::st_bbox(shp)

	crs = sf::st_crs(shp)

	shp = structure(list(x = structure(list(from = 1, to = xy_dim[2], offset = b[1], delta = (b[3] - b[1]) / xy_dim[2], refsys = crs, point = FALSE, values = NULL), class = "dimension"),
						 y = structure(list(from = 1, to = xy_dim[1], offset = b[4], delta = (b[2] - b[4]) / xy_dim[1], refsys = crs, point = FALSE, values = NULL), class = "dimension")), class = "dimensions")
	attr(shp, "raster") = structure(list(affine = c(0, 0), dimensions = c("x", "y"), curvilinear = FALSE), class = "stars_raster")


	shpclass = "stars"

	dtcols = setdiff(names(dt), "tmapID__")

	names(ctabs) = dtcols
	names(cats) = dtcols
	names(minmax) = dtcols

	if (!is.null(tmf)) make_by_vars(dt, tmf, smeta)


	if (is.null(filter)) filter = rep(TRUE, nrow(dt))
	dt[, ':='(sel__ = filter)] # tmapID__ = 1L:nrow(dt),

	shpTM = shapeTM(shp = shp, tmapID = 1L:(nrow(shp) * ncol(shp)), bbox = bbox)

	for (nm in dtcols) {
		if (!is.null(ctabs[[nm]])) {
			# in case of predefined colors: if categories are associated (cats), use tm_scale_categorical, otherwise tm_

			ct = ctabs[[nm]]
			lt = cats[[nm]]
			if (is.factor(dt[[nm]])) {
				#levels(dt[[nm]])

				ids = match(lt[match(levels(dt[[nm]]), lt[,2]), 1], ct$value)
				cti = ct[ids,]

				cls = rgb(cti$red, cti$green, cti$blue, cti$alpha, maxColorValue = 255)

				levels(dt[[nm]]) = paste(levels(dt[[nm]]), cls, sep = "=<>=")
			} else if ("value" %in% names(ct)) {
				cls = rgb(ct$red, ct$green, ct$blue, ct$alpha, maxColorValue = 255)

				rng = minmax[[nm]]
				sel = ct$value %in% seq(rng[1], rng[2])

				#until = anyDuplicated(cls) - 1L

				cls = cls[sel]
				ct = ct[sel, ]
				dt[[nm]] = factor(dt[[nm]], levels = ct$value, labels = paste(ct$value, cls, sep = "=><="))
			}

		}
	}

	structure(list(shpTM = shpTM, dt = dt, is.main = is.main, dtcols = dtcols, shpclass = shpclass, bbox = bbox, unit = unit, shp_name = shp_name, smeta = smeta, type_ids = NULL, type_vars = NULL), class = "tmapShape")

}

#' @export
tmapShape.SpatVector = function(shp, is.main, crs, bbox, unit, filter, layer, shp_name, smeta, o, tmf) {
	tmapShape.sf(sf::st_as_sf(shp), is.main, crs, bbox, unit, filter, layer, shp_name, smeta, o, tmf)
}

#' @export
tmapSubsetShp.SpatRaster = function(shp, vars) {
	#v = setdiff(vars, names(shp))
	#if (length(v)) stop("Variable(s) not found: \"", paste(v, collapse = "\",\""), "\"")
	if (!length(vars)) {
		terra::rast(extent = terra::ext(shp), crs = terra::crs(shp), vals = TRUE, names = "dummy__")
	} else {
		shp[[vars]]
	}
}

#' @export
tmapSubsetShp.SpatVector = function(shp, vars) {
	if ("AREA_" %in% vars && !("AREA_" %in% names(shp))) {
		shp$AREA_ = terra::expanse(shp)
	}
	if ("LENGTH_" %in% vars && !("LENGTH_" %in% names(shp))) {
		shp$LENGTH = terra::perim(shp)
	}
	tmapSubsetShp.sf(shp, vars)
}

#' @export
tmapGetShapeMeta2.SpatRaster = function(shp, smeta, o) {
	if (terra::ncell(shp) > max_cells(o$raster.max_cells)) {
		# NOTE: this option is not ideal, because categories may be undiscovered
		# NOTE2: maybe the same should be done with large stars?
		shp = terra::spatSample(shp, 1e5, method = "regular", as.raster = TRUE)
	}
	smeta$vars_levs = lapply(terra::values(shp, dataframe=TRUE), function(dat) {
		get_fact_levels_na(dat, o)
	})

	names(smeta$vars_levs) = names(shp)
	smeta
}

#' @export
tmapGetShapeMeta2.SpatVector = function(shp, smeta, o) {

	# slow, needs to be improved with terra functions, e.g. unique and levels
	smeta$vars_levs = lapply(terra::values(shp), function(dat) {
		get_fact_levels_na(dat, o)
	})
	names(smeta$vars_levs) = names(shp)
	smeta
}



#' @export
tmapGetShapeMeta1.SpatRaster = function(shp, layer, o) {
	vars = names(shp)
	names(vars) = vars

	dims = character(0)
	dims_vals = list()

	list(vars = vars,
		 dims = dims,
		 dims_vals = dims_vals)

}

#' @export
tmapGetShapeMeta1.SpatVector = function(shp, layer, o) {
	vars = names(shp)
	names(vars) = vars

	dims = character(0)
	dims_vals = list()

	list(vars = vars,
		 dims = dims,
		 dims_vals = dims_vals)

}

#' @export
#' @rdname tmapGridLeaflet
tmapLeafletDataPlot.tm_data_symbols = function(a, shpTM, dt, pdt, popup.format, hdt, idt, gp, bbx, facet_row, facet_col, facet_page, id, pane, group, glid, o, ...) {
	# ptdt (popup.title data) and popup.layout are passed via `...` to keep this
	# method's signature identical to the generic (see tm_data_polygons).
	dots = list(...)
	ptdt = dots$ptdt
	popup.layout = dots$popup.layout

	lf = get_lf(facet_row, facet_col, facet_page)

	rc_text = frc(facet_row, facet_col)

	res = select_sf(shpTM, dt[!is.na(dt$size) & !is.na(dt$shape), ])
	shp = res$shp
	if (o$crs_leaflet$crsClass  == "L.CRS.Simple") {
		shp = sf::st_set_crs(shp, NA)
	}
	dt = res$dt

	gp = impute_gp(gp, dt)
	gp = rescale_gp(gp, o$scale_down)

	coords = sf::st_coordinates(shp)

	# in case shp is a multipoint (point_per == "segment"), expand gp:
	cp = expand_coords_gp(coords, gp, ndt = nrow(dt))
	coords = cp$coords
	gp = cp$gp

	if (cp$expanded) {
		shpTM_match = match(shpTM$tmapID_expanded, shpTM$tmapID)
	} else {
		shpTM_match = TRUE
	}

	idt_null = is.null(idt)

	if (!idt_null) {
		idt = idt$id[match(dt$tmapID__[shpTM_match], idt$tmapID__)]
	} else {
		idt = sprintf("%07d", dt$tmapID__)[shpTM_match]
	}
	if (!is.null(hdt)) {
		hdt = hdt$hover[match(dt$tmapID__[shpTM_match], hdt$tmapID__)]
		hdt = lapply(hdt, htmltools::HTML, FUN.VALUE = character(1))
	}
	if (!is.null(ptdt)) {
		pttl = ptdt$title[match(dt$tmapID__[shpTM_match], ptdt$tmapID__)]
	} else {
		pttl = NULL
	}
	if (is.null(pdt)) {
		popups = NULL
	} else {
		mtch = match(dt$tmapID__[shpTM_match], pdt$tmapID__)
		pdt = pdt[mtch][, tmapID__ := NULL]

		if (!is.null(pttl)) {
			popups = view_format_popups(id = pttl, titles = names(pdt), values = pdt, format = popup.format, layout = popup.layout)
		} else if (idt_null) {
			popups = view_format_popups(titles = names(pdt), values = pdt, format = popup.format, layout = popup.layout)
		} else {
			popups = view_format_popups(id = idt, titles = names(pdt), values = pdt, format = popup.format, layout = popup.layout)
		}
	}

	clusterOpts = getClusterOpts(a$clustering)



	interactive = (!is.null(pdt) || !is.null(hdt))

	opt = leaflet::pathOptions(interactive = interactive, pane = pane)

	o$use_WebGL = impute_webgl(o$use_WebGL, dt, supported = c("fill", "size"), checkif = list(shape = c(1,16,19,20,21)), type = "symbols", hover = !is.null(hdt), popup = !is.null(pdt), crs_class = o$crs_leaflet$crsClass)

	use_circleMarkers = o$use_circle_markers && all(gp$shape %in% c(1, 16, 19, 20, 21))


	gp2 = gp_to_lpar(gp, mfun = "Symbols", size_factor = 14) # 14 based on similarity with plot mode and consistency with tmap3
	#gp = gp2leafgp(gp)
	names(gp2)[names(gp2) == 'stroke-width'] = "strokeWidth"
	gp2$baseSize = 20

	#po(sort(gp2$width, decreasing = T))

	shp2 = sf::st_as_sf(as.data.frame(coords), coords = c("X", "Y"), crs = st_crs(shp))

	# Screen-size factor for circle markers. The *visible* on-screen radius (in
	# leaflet px) is `gp2$width * multiplier` regardless of render path: the
	# WebGL path passes `radius = gp2$width` to addGlPoints and the circleMarkers
	# path passes `radius = gp2$width * multiplier`, but both render identically,
	# so addGlPoints' radius unit is effectively `multiplier` leaflet px. Using
	# this single quantity for hitbox sizing keeps the WebGL and non-WebGL
	# hitboxes identical. Only meaningful for circle shapes (the only shapes that
	# reach the WebGL / circleMarkers branches); harmless (-> 0.5) otherwise.
	multiplier = ifelse(gp2$shape == "solid-circle-md", 0.28, ifelse(gp2$shape == "solid-circle-sm", 0.25, 0.5)) # manually calibrated with 4k screen

	# process hitbox option (mirrors tm_lines):
	# an invisible, larger addCircleMarkers overlay carries the interaction
	# (label/popup/layerId) so that small symbols are easier to click/hover.
	#
	# The visible on-screen diameter (leaflet px) of a symbol is
	# `2 * gp2$width * multiplier` across all render paths (see note above;
	# for icons multiplier -> 0.5 so this degrades to gp2$width). The `auto`
	# rule fires only for small symbols (median visible diameter < 12, i.e.
	# below the 12px floor it would be raised to) when there are few enough
	# features, and floors the clickable diameter at 12px. pmax12 leaves
	# symbols already larger than 12px untouched.
	hitbox = if (isTRUE(a$hitbox == "auto")) {
		small = stats::median(2 * gp2$width * multiplier) < 12   # visible diameter (px)
		light = nrow(dt) < 10000
		if (small && light) "pmax12" else "none"
	} else {
		a$hitbox %||% "none"
	}

	# A hitbox is only meaningful when there is something to click (interactive)
	# and when markers are not clustered (clusters already provide a large
	# click target, and overlaying a second clustered layer would double-count).
	use_hitbox = interactive && hitbox != "none" && is.null(clusterOpts)
	if (use_hitbox) hb = parse_hitbox(hitbox)

	# helper: compute the invisible interaction radius (px) from a per-feature
	# *visible diameter* (px). Hitbox numbers (`plusX`, `pmaxX`) are interpreted
	# on the diameter, consistent with line `weight`; the overlay needs a radius.
	hitbox_radius = function(vis_diam) {
		d = vis_diam + hb$plus
		if (hb$pmax != 0) d = pmax(d, hb$pmax)
		d / 2
	}

	# helper: invisible interaction overlay (regular leaflet circle markers)
	add_hitbox = function(lf, layerId, radius) {
		leaflet::addCircleMarkers(lf,
								  data        = shp2,
								  layerId     = layerId,
								  radius      = radius,
								  stroke      = FALSE,
								  fill        = TRUE,
								  fillOpacity = 0,
								  options     = leaflet::pathOptions(interactive = TRUE, pane = pane),
								  group       = group,
								  label       = hdt,
								  popup       = popups)
	}

	if (o$use_WebGL) {
		if (!use_hitbox) {
			idt = submit_labels(idt[1], "symbolsGL", pane, group)

			lf |>  leafgl::addGlPoints(shp2, fillColor = gp2$fillColor, radius = gp2$width, fillOpacity = gp2$fillOpacity[1], pane = pane, group = group, label = hdt, popup = popups, layerId = idt) %>%
				assign_lf(facet_row, facet_col, facet_page)
		} else {
			# visible on-screen radius (leaflet px) = gp2$width * multiplier,
			# the same quantity the non-WebGL path uses -> identical hitboxes
			hb_radius = hitbox_radius(2 * gp2$width * multiplier)

			idt_gl = submit_labels(idt[1], "symbolsGL", pane, group)
			idt_hb = submit_labels(idt, "symbolsGL_hb", pane, group)

			lf |>
				# Fast WebGL rendering (no interaction; handled by overlay)
				leafgl::addGlPoints(shp2, fillColor = gp2$fillColor, radius = gp2$width, fillOpacity = gp2$fillOpacity[1], pane = pane, group = group, label = NULL, popup = NULL, layerId = idt_gl) |>
				# Invisible interaction layer (fatter hitbox)
				add_hitbox(layerId = idt_hb, radius = hb_radius) |>
				assign_lf(facet_row, facet_col, facet_page)
		}
	} else if (use_circleMarkers) {
		cidt = if (!is.null(clusterOpts)) submit_labels(idt[1], "symbolsCluster", pane, group) else NULL

		gp2$strokeWidth[gp2$shape %in% c("solid-circle-md", "solid-circle-bg", "solid-circle-sm")] = 0
		gp2$fillOpacity[gp2$shape %in% "open-circle"] = 0

		vis_radius = gp2$width * multiplier   # multiplier hoisted above

		if (!use_hitbox) {
			idt = submit_labels(idt, "symbols", pane, group)

			lf |>  leaflet::addCircleMarkers(data = shp2, layerId = idt,
											 color = gp2$color,
											 opacity = gp2$opacity,
											 weight = gp2$strokeWidth,
											 dashArray = gp2[["stroke-dasharray"]],
											 fillColor = gp2$fillColor,
											 fillOpacity = gp2$fillOpacity,
											 clusterOptions = clusterOpts,
											 clusterId = cidt,
											 radius = vis_radius,
											 options = opt,
											 group = group, label = hdt, popup = popups) |>
				assign_lf(facet_row, facet_col, facet_page)

		} else {
			idt_vis = submit_labels(idt, "symbols", pane, group)
			idt_hb  = submit_labels(idt, "symbols_hb", pane, group)

			hb_radius = hitbox_radius(vis_radius * 2)
			opt_vis = leaflet::pathOptions(interactive = FALSE, pane = pane)

			lf |>
				# Visible styled layer (non-interactive)
				leaflet::addCircleMarkers(data = shp2, layerId = idt_vis,
										  color = gp2$color,
										  opacity = gp2$opacity,
										  weight = gp2$strokeWidth,
										  dashArray = gp2[["stroke-dasharray"]],
										  fillColor = gp2$fillColor,
										  fillOpacity = gp2$fillOpacity,
										  radius = vis_radius,
										  options = opt_vis,
										  group = group) |>
				# Invisible interaction layer (fatter hitbox)
				add_hitbox(layerId = idt_hb, radius = hb_radius) |>
				assign_lf(facet_row, facet_col, facet_page)
		}

	} else {
		cidt = if (!is.null(clusterOpts)) submit_labels(idt[1], "cluster", pane, group) else NULL

		idt0 = idt # keep per-feature ids for the hitbox overlay
		idt = submit_labels(idt, "symbols", pane, group)
		idt_hb = if (use_hitbox) submit_labels(idt0, "symbols_hb", pane, group) else NULL

		sn = suppressWarnings(as.numeric(gp2$shape))

		is_num = !is.na(sn)
		sid = which(is_num)
		nid = which(!is_num)


		# ="circle" to make makeSymbolsIcons2 work
		# shape_orig to let unique pick unique rows (one for each )
		gp2$shape_orig = gp2$shape
		gp2$shape[sid] = "circle"

		# faster than symbols2 = do.call(makeSymbolIcons2, gp2)
		gp2df = as.data.table(gp2)
		gp2dfU = unique(gp2df)

		k = nrow(gp2dfU)

		symbols = do.call(makeSymbolIcons2, as.list(gp2dfU[,-(ncol(gp2dfU)),with=F]))

		gp2dfU[, id:=1L:.N]
		gp2join = gp2df[gp2dfU, id:= id, on=names(gp2df)]
		ids = gp2join$id

		coords_grps = split.data.frame(coords, ids)
		idt_grps = split(idt, ids)
		if (!is.null(hdt)) {
			hdt_grps = split(hdt, ids)
		} else {
			hdt_grps = replicate(k, list(NULL))
		}

		if (is.null(popups)) {
			popups_grps =  rep(list(NULL), k)
		} else {
			popups_grps = split(popups, ids)
		}


		symbols$iconWidth = rep(NA, k)
		symbols$iconHeight = rep(NA, k)


		if (length(sid)) {
			sym_shapes = suppressWarnings(as.numeric(gp2dfU$shape_orig))
			sid2 = which(!is.na(sym_shapes))

			iconLib <- get("shapeLib", envir = .TMAP)[sym_shapes[sid2]-999]
			symbols_icons <- merge_icons(iconLib)

			size = gp2dfU$width[sid2] / gp2dfU$baseSize[sid2]
			size = size * a$icon.scale/3

			for (i in seq_along(sid2)) {
				symbols$iconUrl[sid2[i]] = symbols_icons$iconUrl[i]
				symbols$iconWidth[sid2[i]] <- symbols_icons$iconWidth[i] * size[i]
				symbols$iconHeight[sid2[i]] <- symbols_icons$iconHeight[i] * size[i]
				if (all(c("iconAnchorX", "iconAnchorY") %in% names(symbols_icons))) {
					symbols$iconAnchorX[sid2[i]] <- symbols_icons$iconAnchorX[i] * size[i]
					symbols$iconAnchorY[sid2[i]] <- symbols_icons$iconAnchorY[i] * size[i]

				}
			}
		}

		# when a hitbox overlay handles interaction, the icon markers themselves
		# carry no label/popup and are non-interactive
		marker_label = if (use_hitbox) replicate(k, list(NULL)) else hdt_grps
		marker_popup = if (use_hitbox) rep(list(NULL), k) else popups_grps

		for (i in 1L:k) {
			opt$zIndexOffset = i * 1e5
			if (use_hitbox) opt$interactive = FALSE
			.TMAP_LEAFLET$markerLayers = c(.TMAP_LEAFLET$markerLayers, opt$pane) # register for additional css
			lf = lf |>
				leaflet::addMarkers(lng = coords_grps[[i]][, 1],
									lat = coords_grps[[i]][, 2],
									icon = lapply(symbols, "[", i),
									group = group,
									layerId = idt_grps[[i]],
									label = marker_label[[i]],
									clusterOptions = clusterOpts,
									clusterId = cidt,
									popup = marker_popup[[i]],
									options = opt
				)
		}

		if (use_hitbox) {
			# icon px size ~ gp2$width (per feature) -> use as visible diameter proxy
			hb_radius = hitbox_radius(gp2$width)
			lf = lf |> add_hitbox(layerId = idt_hb, radius = hb_radius)
		}

		lf |> assign_lf(facet_row, facet_col, facet_page)

	}


	# if (o$use_WebGL) {
	# 	lf %>%
	# 		leafgl::addGlPoints(sf::st_sf(shp), fillColor = gp$fill, radius = gp$size*10, fillOpacity = gp$fill_alpha, color = gp$col, opacity = gp$color_alpha, weight = gp$lwd, pane = pane, group = group) %>%
	# 		assign_lf(facet_row, facet_col, facet_page)
	# } else {
	# 	lf %>%
	# 		leaflet::addCircleMarkers(lng = coords[, 1], lat = coords[, 2], fillColor = gp$fill, radius = gp$size*4, fillOpacity = gp$fill_alpha, color = gp$col, opacity = gp$color_alpha, weight = gp$lwd, group = group, options = opt) %>%
	# 		assign_lf(facet_row, facet_col, facet_page)
	# }

	NULL
}


#' @export
#' @rdname tmapGridLeaflet
tmapLeafletDataPlot.tm_data_dots = function(a, shpTM, dt, pdt, popup.format, hdt, idt, gp, bbx, facet_row, facet_col, facet_page, id, pane, group, glid, o, ...) {
	NextMethod()
}

#' @export
#' @rdname tmapGridLeaflet
tmapLeafletDataPlot.tm_data_bubbles = function(a, shpTM, dt, pdt, popup.format, hdt, idt, gp, bbx, facet_row, facet_col, facet_page, id, pane, group, glid, o, ...) {
	NextMethod()
}

#' @export
#' @rdname tmapGridLeaflet
tmapLeafletDataPlot.tm_data_squares = function(a, shpTM, dt, pdt, popup.format, hdt, idt, gp, bbx, facet_row, facet_col, facet_page, id, pane, group, glid, o, ...) {
	NextMethod()
}

#' @export
#' @rdname tmapGridLeaflet
tmapLeafletDataPlot.tm_data_markers = function(a, shpTM, dt, pdt, popup.format, hdt, idt, gp, bbx, facet_row, facet_col, facet_page, id, pane, group, glid, o, ...) {
	NextMethod()
}

#' Set tmap mode
#'
#' @description
#' * `tmap_mode()` gets (no argument) or sets the current mode.
#' * `ttm()` toggles between the two most recent modes.
#' * `ttmp()` same as `ttm()`, then calls [tmap_last()].
#' * `rtm()` rotates through all modes in the pool.
#' * `rtmp()` same as `rtm()`, then calls [tmap_last()].
#' * `tmap_mode_pool()` restricts which modes are cycled by [ttm()] and [rtm()].
#'   Call without arguments to inspect the current pool, or pass `NULL` to reset.
#'
#' It is recommended to use `tmap_mode()` in scripts and `ttm()`/`ttmp()` in the console.
#'
#' @details
#' The default modes are `"plot"` (static, graphics device) and `"view"` (interactive,
#' browser or RStudio Viewer). Additional modes such as `"maplibre"` and `"mapbox"` become
#' available when \pkg{tmap.mapgl} is loaded.
#'
#' @param mode A string specifying the mode. See [tmap_options()] for available modes.
#' @param silent Should the mode be switched silently? Default `FALSE`.
#' @param modes Character vector of mode names (minimum 2), or `NULL` to reset.
#' @return
#' * `tmap_mode()` returns the current mode invisibly when called without argument,
#'   otherwise the previous mode.
#' * `ttm()`, `rtm()` return the previous mode invisibly.
#' * `tmap_mode_pool()` returns the previous pool invisibly.
#' @example ./examples/tmap_mode.R
#' @seealso
#' * .doc_see_also_modes()
#' * [tmap_last()] to show the last map
#' * [tm_view()] for viewing options
#' * [tmap_leaflet()] for obtaining a leaflet widget
#' * [tmap_options()] for tmap options
#' @references Tennekes, M., 2018, {tmap}: Thematic Maps in {R},
#' Journal of Statistical Software, 84(6), 1-39, \doi{10.18637/jss.v084.i06}
#' @export
tmap_mode = function(mode = NULL, silent = FALSE) {
	type = if (is.null(mode)) "get" else "set"
	set_mode(mode, show.messages = !silent && get("tmapOptions", envir = .TMAP)$show.messages, type = type)
}


#' @rdname tmap_mode
#' @export
ttm = function() {
	set_mode(show.messages = get("tmapOptions", envir = .TMAP)$show.messages, type = "toggle")
}

#' @rdname tmap_mode
#' @export
rtm = function() {
	set_mode(show.messages = get("tmapOptions", envir = .TMAP)$show.messages, type = "rotate")
}


set_mode = function(mode = NULL, show.messages = FALSE, type = "set") {
	mode_now = getOption("tmap.mode")

	modes = get_modes()
	id = match(mode_now, modes)

	arrow = length(modes) > 2

	if (type == "get") {
		mode_next = mode_now
		id_next = id
	} else if (type == "set") {
		if (!mode %in% modes) {
			if (mode %in% c("maplibre", "mapbox")) {
				cli::cli_abort(c(
					"Mode {.val {mode}} requires the {.pkg tmap.mapgl} package.",
					"i" = "Please install and load it first: {.run library(tmap.mapgl)}"
				), call = rlang::caller_env())
			} else {
				cli::cli_abort(c(
					"{.val {mode}} is not a valid tmap mode.",
					"i" = "Must be one of {.or {.val {modes}}}."
				), call = rlang::caller_env())
			}
		}

		mode_next = rlang::arg_match0(mode, modes)
		id_next = match(mode_next, modes)
	} else if (type == "toggle") {
		mode_next = .TMAP$mode_last
		if (is.null(mode_next) || !mode_next %in% modes) {
			mode_next = modes[modes != mode_now][1]
		}
		id_next = match(mode_next, modes)
	} else {
		id_next = id + 1L
		if (id_next > length(modes)) id_next = 1L
		mode_next = modes[id_next]
	}

	modes_str = paste0("{.val ", modes, "}")
	modes_str[id_next] = paste0("{.strong ", modes_str[id_next], "}")
	rotate_str = paste(modes_str, collapse = ifelse(arrow, " -> ", " - "))

	if (mode_now != mode_next) {
		.TMAP$mode_last = mode_now
		options(tmap.mode = mode_next)
		fill_providers()
	}

	if (show.messages) {
		str1 = paste0("tmap modes ", rotate_str)
		str2 = paste0(ifelse(arrow, "rotate with {.run tmap::rtm()}", ""),
					 ifelse(arrow, "switch to {.str {mode_now}} ", "toggle "),
					 "with {.run tmap::ttm()}")

		cli::cli_inform(c("i" = str1))
		cli::cli_inform(c("i" = str2), .frequency = "once", .frequency_id = paste0("mode", ifelse(arrow, "_rotate", "toggle")))
	}

	invisible(mode_now)
}


# tmap_graphics = function(mode = NULL) {
# 	if (is.null(mode)) mode = getOption("tmap.mode")
# 	get("tmapOptions", envir = .TMAP)$graphics[[mode]]
# }
#
# tmap_graphics_name = function(mode = NULL) {
# 	tmap_graphics(mode = mode)$name
# }

# Internal: always returns the full set of registered modes,
# regardless of any active pool. Used for validation.
get_all_modes = function() {
	names(get("tmapOptions", envir = .TMAP)$modes)
}

# Internal: returns the active pool if set, otherwise all modes.
# This is what ttm() and rtm() cycle through.
get_modes = function() {
	.TMAP$mode_pool %||% get_all_modes()
}

#' @rdname tmap_mode
#' @export
tmap_mode_pool = function(modes = NULL, silent = FALSE) {
	all_modes    = get_all_modes()
	prev_pool    = .TMAP$mode_pool %||% all_modes
	show_message = !silent && get("tmapOptions", envir = .TMAP)$show.messages

	# getter
	if (missing(modes)) {
		if (show_message) cli::cli_inform(c("i" = "Current mode pool: {.val {prev_pool}}"))
		return(invisible(prev_pool))
	}

	# reset
	if (is.null(modes)) {
		.TMAP$mode_pool = NULL
		if (show_message) cli::cli_inform(c("i" = "Mode pool reset to: {.val {all_modes}}"))
		return(invisible(prev_pool))
	}

	# validate
	if (length(modes) < 2L) {
		cli::cli_abort("The mode pool must contain at least 2 modes.",
					   call = rlang::caller_env())
	}

	bad = setdiff(modes, all_modes)
	if (length(bad) > 0L) {
		cli::cli_abort(c(
			"Invalid mode{?s}: {.val {bad}}",
			"i" = "Available modes: {.val {all_modes}}"
		), call = rlang::caller_env())
	}

	mode_now = getOption("tmap.mode")
	if (!mode_now %in% modes) {
		cli::cli_abort(c(
			"Current mode {.val {mode_now}} must be included in the pool.",
			"i" = "Either add it: {.run tmap_mode_pool(c({toString(shQuote(c(mode_now, modes)))}))}",
			"i" = "Or switch mode first: {.run tmap_mode(\"{modes[1]}\")}"
		), call = rlang::caller_env())
	}

	.TMAP$mode_pool = modes
	if (show_message) cli::cli_inform(c("i" = "Mode pool set to: {.val {modes}}"))
	invisible(prev_pool)
}

#' Set the design mode
#'
#' When the so-called "design mode" is enabled,
#' the composition of the plot is shown explicitly in plot mode.
#' The used color codings is printed in the console as well as information about plot size and aspect ratio.
#'
#' This function sets the global option `tmap.design.mode`.
#' It can be used as toggle function without arguments.
#'
#' @seealso [tmap_options()]
#' @param design.mode Logical value that determines the design mode.
#'   If omitted then the design mode is toggled.
#' @export
tmap_design_mode = function(design.mode) {
	dm = if (missing(design.mode)) {
		!getOption("tmap.design.mode")
	} else {
		if (!is.logical(design.mode)) stop("design.mode is not a logical")
		design.mode[1]
	}

	options(tmap.design.mode = dm)
	message(
		"design.mode: ", if (!dm) "OFF" else "ON",
		if (dm && getOption("tmap.mode") == "view") " (only effective in plot mode)" else "")
}

#' Set the development mode
#'
#' When the so-called "development mode" is enabled, helpful messages and timings
#' are printed in the console
#'
#' @param devel.mode logical value that determines the development mode.
#'   If omitted then the development mode is toggled.
#' @export
tmap_devel_mode = function(devel.mode) {
	dm = if (missing(devel.mode)) {
		!getOption("tmap.devel.mode")
	} else {
		if (!is.logical(devel.mode)) stop("devel.mode is not a logical")
		devel.mode[1]
	}

	options(tmap.devel.mode = dm)
	message("devel.mode: ", if (!dm) "OFF" else "ON")
}

pm = function(message) {
	cat("<><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><>\n")
	cat(message, "\n")
}

po = function(...) {
	e = substitute(list(...))
	nms = sapply(e, deparse)[-1]

	x = list(...)

	for (i in seq_along(x)) {
		cat("<==================== ", nms[i], "===============>\n")
		print(x[[i]])
		if (i == length(x)) {
			cat("</============================================>\n")
		}
	}

	invisible()
}

so = function(...) {
	e = substitute(list(...))
	nms = sapply(e, deparse)[-1]

	x = list(...)

	for (i in seq_along(x)) {
		cat("<==================== ", nms[i], "===============>\n")
		str(x[[i]])
		if (i == length(x)) {
			cat("</============================================>\n")
		}
	}

	invisible()
}

#' @rdname tmap_mode
#' @export
ttmp = function() {
	ttm()
	tmap_last()
}

#' @rdname tmap_mode
#' @export
rtmp = function() {
	rtm()
	tmap_last()
}

# similar as rlang::'%||%' but than for NA
'%--%' = function (x, y) if (is.na(x[1])) y else x

islistof = function(x, class) {
	is.list(x) && all(vapply(x, inherits, logical(1), what = class))
}

select_sf = function(shpTM, dt) {
	shp = shpTM$shp
	stid = shpTM$tmapID

	dtid = dt$tmapID__

	tid = intersect(stid, dtid)

	d = data.table(sord = seq_along(tid), ord = dt$ord__[match(tid, dtid)], tid = tid)
	if ("ord" %in% names(d)) {
		setkeyv(d, cols = c("ord", "sord"))
	} else {
		setkeyv(d, cols = "sord")
	}
	sid = match(d$tid, stid)

	shpSel = shp[sid] #sf::st_cast(shp[match(tid, tmapID)], "MULTIPOLYGON")

	# assign prop_ vectors to data dt (to be used in plotting) e.g. prop_angle is determined in tmapTransCentroid when along_lines = TRUE
	prop_vars = names(shpTM)[substr(names(shpTM), 1, 5) == "prop_"]
	if (length(prop_vars)) {
		for (p in prop_vars) {
			pname = substr(p, 6, nchar(p))
			dt[[pname]] = shpTM[[p]][sid]
		}
	}

	dt = dt[match(d$tid, dtid), ]
	list(shp = shpSel, dt = dt)
}

get_nby = function(fl) {
	vapply(fl, function(f) {
		if (is.integer(f)) f else length(f)
	}, integer(1))
}

get_row = function(i, nrows) {
	as.integer((i - 1) %% nrows + 1)
}

get_col = function(i, nrows, ncols) {
	as.integer((((i - 1) %/% nrows + 1) - 1) %% ncols + 1)
}

get_page = function(i, nrows, ncols) {
	as.integer(i - 1) %/% (nrows * ncols) + 1
}

get_i = function(ir, ic, ip, nby) {
	ir + (ic - 1) * nby[1] + (ip - 1) * prod(nby[1:2])
}

completeDT = function(DT, cols, defs = NULL){
	mDT = do.call(CJ, c(DT[, ..cols], list(unique=TRUE)))
	res = DT[mDT, on=names(mDT)]
	if (length(defs))
		res[, names(defs) := Map(replace, .SD, lapply(.SD, is.na), defs), .SDcols=names(defs)]
	res[]
}

completeDT2 = function(DT, cols, defs = NULL){
	mDT = do.call(CJ, cols)
	res = DT[mDT, on=names(mDT)]
	if (length(defs))
		res[, names(defs) := Map(replace, .SD, lapply(.SD, is.na), defs), .SDcols=names(defs)]
	res[]
}

cont_breaks = function(breaks, n=101) {
	x = round(seq(1, n, length.out=length(breaks)))

	unlist(lapply(1L:(length(breaks)-1L), function(i) {
		y = seq(breaks[i], breaks[i+1], length.out=x[i+1]-x[i]+1)
		if (i!=1) y[-1] else y
	}), use.names = FALSE)
}

prettyCount = function(x, n, ...) {
	x = na.omit(x)
	if (!length(x)) return(x)

	if (!is.integer(x)) x = as.integer(x)

	mn = min(x)
	mx = max(x)

	any0 = any(x==0)

	if (mn < 0) {
		n = floor(n / 2)
		pneg = -rev(prettyCount(-x[x<0], n = n, ...)) + 1L
		pneg = pneg[pneg != 0L]
		x = x[x>0]
		any0 = TRUE
	} else {
		pneg = integer()
	}

	if (any0) x = x[x!=0L]

	p = pretty(x - 1L, n = n, ...) + 1L

	p = p[(p %% 1) == 0]
	p = p[p!=0L]

	if (length(x) < 2) {
		if (any0) return(c(0L, p)) else return(p)
	}

	step = p[2] - p[1]
	if (p[length(p)] == mx) p = c(p, mx+step)

	if (any0) {
		c(pneg, 0L, p)
	} else {
		c(pneg, p)
	}
}

valid_colors = function(x) {
	is.na(x) | (x %in% colors()) |	(vapply(gregexpr("^#(([[:xdigit:]]){6}|([[:xdigit:]]){8})$", x), "[[", integer(1), 1) == 1L)
}

col2hex = function(x) {
	y = apply(col2rgb(x), MARGIN=2, FUN=function(y)do.call(rgb, c(as.list(y), list(maxColorValue=255))))
	y[is.na(x)] = NA
	y
}

# get aspect ratio of a shape
get_asp_ratio = function (x, width = 700, height = 700, res = 100)  {
	asp = if (inherits(x, "tmap")) {
		tmp = tempfile(fileext = ".png")
		png(tmp, width = width, height = height, res = res)
		mode = options(tmap.mode = "plot")
		on.exit({
			do.call(options, mode)
		})
		asp = print(x, return.asp = TRUE)
		dev.off()
		asp
	} else {
		bbx = sf::st_bbox(x)
		crs = sf::st_crs(x)

		ll = sf::st_is_longlat(crs)

		xlim = bbx[c(1, 3)]
		ylim = bbx[c(2, 4)]
		if (diff(xlim) == 0 || diff(ylim) == 0) {
			1
		} else unname((diff(xlim)/diff(ylim)) * ifelse(identical(ll, TRUE),cos((mean(ylim) * pi)/180), 1))
	}
	asp
}



# get aspect ratios of a list of bounding boxes
get_asp = function(bbxl) {
	vapply(bbxl, function(bbxi) {
		if (is.na(bbxi)) NA_real_ else get_asp_ratio(bbxi)
	}, FUN.VALUE = numeric(1))
}

leaflet2crs = function(x) {
	epsg = get_epsg(x)
	if (!is.na(epsg)) {
		sf::st_crs(epsg)
	} else if (!is.null(x$proj4def)) {
		sf::st_crs(x$proj4def)
	} else {
		cli::cli_abort("Unable to extract crs from leafletCRS object.")
	}
}

leafletSimple  = structure(list(crsClass = "L.CRS.Simple", code = NULL, proj4def = NULL,
								projectedBounds = NULL, options = structure(list(), .Names = character(0))), class = "leaflet_crs")

crs2leaflet = function(x) {
	epsg = get_epsg(x)
	if (!is.na(epsg) && (epsg %in% c(3857, 4326, 3395))) {
		epsg2 = 3857 # somehow L.CRS.EPSG4326 doesn't work well (?)
		structure(list(crsClass = paste0("L.CRS.EPSG", epsg2), code = NULL, proj4def = NULL,
					   projectedBounds = NULL, options = structure(list(), .Names = character(0))), class = "leaflet_crs")
	} else {
		leafletSimple
		#stop("Unable to extract leaflet crs from sf crs object")
	}
}

get_epsg = function (x) {
	if (is.numeric(x)) {
		x
	} else if (inherits(x, "leaflet_crs")) {
		if (grepl("EPSG", x$crsClass, fixed = TRUE)) {
			substr(x$crsClass, 11, 14) # one of L.CRS.EPSG3857, L.CRS.EPSG4326, L.CRS.EPSG3395
		} else {
			NA_integer_
		}
	} else {
		x = sf::st_crs(x)
		# from sf
		if (is.na(x))
			NA_integer_
		else if (grepl("^EPSG:", x[["input"]]))
			as.integer(gsub("^EPSG:(\\d+)\\b.*$", "\\1", x[["input"]]))
		else {
			# not exported by sf:
			# crs_parameters(x, with_units = FALSE)[["epsg"]]
			NA_integer_
		}
	}
}

number_text_lines = function(txt) {
	if (is.character(txt)) {
		length(strsplit(txt, "\n")[[1]])
	} else 1
}

nonempty_text = function(txt) {
	if (is.character(txt)) {
		nzchar(txt)
	} else rep(TRUE, length(txt))
}

cont_split = function(x) strsplit(x, split = "_", fixed=TRUE)
cont_collapse = function(x) sapply(x, paste, collapse="_")


without_units = function(x) {
	if (inherits(x, "units")) units::drop_units(x) else x
}


get_midpoint = function (coords) {
	dist = sqrt((diff(coords[, 1])^2 + (diff(coords[, 2]))^2))
	dist_mid = sum(dist)/2
	dist_cum = c(0, cumsum(dist))
	end_index = which(dist_cum > dist_mid)[1]
	start_index = end_index - 1
	start = coords[start_index, ]
	end = coords[end_index, ]
	dist_remaining = dist_mid - dist_cum[start_index]
	start + (end - start) * (dist_remaining/dist[start_index])

}

# copied from tmap3, may need updating
process_just = function(just, interactive) {
	show.messages = get("tmapOptions", envir = .TMAP)$show.messages
	show.warnings = get("tmapOptions", envir = .TMAP)$show.warnings

	n = length(just)
	isnum = is_num_string(just)

	if (!all(isnum | (just %in% c("left", "right", "top", "bottom", "center", "centre"))) && show.warnings) {
		cli::cli_warn("{.arg just} is not correctly specified. ")
	}

	just[just == "centre"] = "center"

	if (interactive) {
		just = just[1]
		if (n > 1 && show.messages) message("In interactive mode, the just argument should be one element")

		if (isnum[1]) {
			justnum = as.numeric(just)
			just = ifelse(justnum < .25, "left",
						  ifelse(justnum > .75, "right", "center"))
			if (show.messages) message("In interactive mode, just cannot be a numeric value. Therefore, ", justnum, " has been cenverted to \"", just, "\".")
		}
	} else {
		if (n > 2 && show.warnings) cli::cli_warn("{.arg just} should be a single value or a vector of 2 values.")
		if (n == 1) {
			if (just %in% c("top", "bottom")) {
				just = c("center", just)
				isnum = c(FALSE, isnum)
			} else {
				just = c(just, "center")
				isnum = c(isnum, FALSE)
			}
		}

		x = ifelse(isnum[1], as.numeric(just[1]),
					ifelse(just[1] == "left", 0,
						   ifelse(just[1] == "right", 1,
						   	   ifelse(just[1] == "center", .5, NA))))
		if (is.na(x)) {
			if (show.warnings) {
				cli::cli_warn("{.arg just} is not correctly specified. ")
			}
			x = 0.5
		}

		y = ifelse(isnum[2], as.numeric(just[2]),
				   ifelse(just[2] == "bottom", 0,
						  ifelse(just[2] == "top", 1,
						  	   ifelse(just[2] == "center", .5, NA))))
		if (is.na(y)) {
			if (show.warnings) {
				cli::cli_warn("{.arg just} is not correctly specified. ")
			}
			y = 0.5
		}
		just = c(x, y)
	}
	just
}






################!!!!! Functions below needed for Advanced text options !!!!####################

.grob2Poly = function(g) {
	x = convertX(g$x, unitTo = "native", valueOnly = TRUE)
	y = convertY(g$y, unitTo = "native", valueOnly = TRUE)
	if (inherits(g, "rect")) {
		w = convertWidth(g$width, unitTo = "native", valueOnly = TRUE)
		h = convertHeight(g$height, unitTo = "native", valueOnly = TRUE)
		x1 = x - .5*w
		x2 = x + .5*w
		y1 = y - .5*h
		y2 = y + .5*h
		polys = mapply(function(X1, X2, Y1, Y2) {
			sf::st_polygon(list(cbind(c(X1, X2, X2, X1, X1),
								  c(Y2, Y2, Y1, Y1, Y2))))
		}, x1, x2, y1, y2, SIMPLIFY = FALSE)
		sf::st_union(sf::st_sfc(polys))
	} else if (inherits(g, "polygon")) {
		xs = split(x, g$id)
		ys = split(y, g$id)

		polys = mapply(function(xi, yi) {
			co = cbind(xi, yi)
			sf::st_polygon(list(rbind(co, co[1,])))
		}, xs, ys, SIMPLIFY = FALSE)
		sf::st_union(sf::st_sfc(polys))
	} # else return(NULL)

}

polylineGrob2sfLines <- function(gL) {
	if (is.null(gL$id)) {
		ids = unlist(mapply(rep, 1L:length(gL$id.lengths), gL$id.lengths, SIMPLIFY = FALSE, USE.NAMES = FALSE))
	} else {
		ids = gL$id
	}
	coords = mapply(cbind, split(as.numeric(gL$x), ids), split(as.numeric(gL$y), ids), SIMPLIFY = FALSE)

	sf::st_sf(geometry = sf::st_sfc(sf::st_multilinestring(coords)))
}

npc_to_native <- function(x, scale) {
	x * (scale[2] - scale[1])# + scale[1]
}

native_to_npc_to_native <- function(x, scale) {
	#(x - scale[1]) / (scale[2] - scale[1])
	(x) / (scale[2] - scale[1])
}

.rectGrob2pathGrob <- function(rg, angles, bbx) {
	x = as.numeric(rg$x)
	y = as.numeric(rg$y)
	w = as.numeric(rg$width)
	h = as.numeric(rg$height)


	####################################
	### borrowed from pointLabel2
	####################################


	asp = tmaptools::get_asp_ratio(bbx)# * 1.25

	#xyAspect <- diff(boundary[c(1,2)]) / diff(boundary[c(3,4)])

	toUnityCoords <- function(xy) {
		if (asp > 1) {
			list(x = (xy$x - bbx[1])/(bbx[3] - bbx[1]) * asp,
				 y = (xy$y - bbx[2])/(bbx[4] - bbx[2]))
		} else {
			list(x = (xy$x - bbx[1])/(bbx[3] - bbx[1]),
				 y = (xy$y - bbx[2])/(bbx[4] - bbx[2])/asp)
		}


	}
	toUserCoords <- function(xy) {
		if (asp > 1) {
			list(x = bbx[1] + xy$x/asp * (bbx[3] - bbx[1]),
				 y = bbx[2] + xy$y * (bbx[4] - bbx[2]))
		} else {
			list(x = bbx[1] + xy$x * (bbx[3] - bbx[1]),
				 y = bbx[2] + xy$y * asp * (bbx[4] - bbx[2]))
		}

	}
	xy = grDevices::xy.coords(x, y, recycle = TRUE)
	z  = toUnityCoords(xy)
	x2 = z$x
	y2 = z$y

	# CHANGED: width and height are specified by user
	if (asp > 1) {
		w2 <- ((w) / (bbx[3] - bbx[1])) * asp
		h2 <- ((h) / (bbx[4] - bbx[2]))
	} else {
		w2 <- ((w) / (bbx[3] - bbx[1]))
		h2 <- ((h) / (bbx[4] - bbx[2])) / asp
	}


	####################################
	####################################
	####################################


	xs = c(x2 - w2/2, x2 + w2 / 2, x2 + w2 / 2, x2 - w2 / 2, x2 - w2/2)
	ys = c(y2 - h2/2, y2 - h2 / 2, y2 + h2 / 2, y2 + h2 / 2, y2 - h2/2)

	a <- atan2(h2, w2)
	#as <- as.vector(vapply(a, function(a)c(a,pi-a, pi+a,-a), numeric(4)))
	as <- as.vector(vapply(a, function(a)c(a,pi-a, pi+a,-a), numeric(4)))

	as2 <- as + rep(angles * pi / 180, each=4)

	dst <- rep(sqrt((w2/2)^2+(h2/2)^2), each=4)

	xs2 <- rep(x2, each=4) + cos(as2) * dst
	ys2 <- rep(y2, each=4) + sin(as2) * dst

	id <- rep(1:length(x), each=4)

	#w2 <- w + (h-w) * abs(cos(angles*pi/180))
	#h2 <- h + (w-h) * abs(sin(angles*pi/180))

	z2 <- grDevices::xy.coords(xs2, ys2, recycle = TRUE)
	xy2 <- toUserCoords(z2)

	list(poly = polygonGrob(unit(xy2$x, "native"), grid::unit(xy2$y, "native"), id = id, gp = rg$gp))
	#list(poly=rectGrob(unit(x, "native"), unit(y, "native"), width = unit(w, "native"), height=unit(h, "native"), gp = rg$gp))
}

.get_direction_angle <- function(co) {
	p1 <- co[1,]
	p2 <- co[nrow(co),]

	a <- atan2(p2[2] - p1[2], p2[1] - p1[1]) * 180 / pi
	if (a < 0) a <- a + 360
	a
}


.editGrob <- function(tg, sel, shiftX, shiftY, angles) {
	nt <- length(sel)
	angles <- rep(angles, length.out=nt)
	if (any(angles != 0)) {
		if (inherits(tg, "rect")) {
			tg <- .rectGrob2pathGrob(tg, angles)$poly
		}
	}
	tgx <- convertX(tg$x, "native", valueOnly = TRUE)
	tgy <- convertY(tg$y, "native", valueOnly = TRUE)

	if (inherits(tg, "polygon")) {
		sel4 <- rep(sel, each=4)
		tg$x <- unit(tgx + rep(shiftX, each=4), "native")[sel4]
		tg$y <- unit(tgy + rep(shiftY, each=4), "native")[sel4]
		tg$id <- rep(1:sum(sel), each=4)
	} else {
		tg$x <- unit(tgx + shiftX, "native")[sel]
		tg$y <- unit(tgy + shiftY, "native")[sel]
		if (inherits(tg, "rect")) {
			tg$height <- tg$height[sel]
			tg$width <- tg$width[sel]
		} else if (inherits(tg, "text")) {
			tg$label <- tg$label[sel]
			tg$rot <- angles[sel]
		}
	}
	tg$gp <- do.call("gpar", lapply(unclass(tg$gp)[names(tg$gp)!="font"], function(g) {
		if (length(g)==nt) g[sel] else g
	}))
	tg
}

# TODO remove function when we depend on R >= 4.4
`%||%` = function(x, y) {
	if (is.null(x)) y else x
}

get_method_classes <- function(generic) {
	meths <- methods(generic)
	# Remove generic. prefix and exclude default
	sub(paste0("^", generic, "\\."), "", meths[!grepl("\\.default$", meths)])
}

#' Internal methods for map components in plot and view mode
#'
#' Internal methods for map components
#' @export
#' @param comp component
#' @param o options
#' @keywords internal
#' @rdname tmapGridLeaflet
tmapGridCompPrepare = function(comp, o) {
	UseMethod("tmapGridCompPrepare")
}

#' @export
#' @keywords internal
#' @rdname tmapGridLeaflet
tmapGridCompHeight = function(comp, o) {
	UseMethod("tmapGridCompHeight")
}

#' @export
#' @keywords internal
#' @rdname tmapGridLeaflet
tmapGridCompWidth = function(comp, o) {
	UseMethod("tmapGridCompWidth")
}

#' @export
#' @keywords internal
#' @rdname tmapGridLeaflet
tmapGridCompPlot = function(comp, o, fH, fW) {
	UseMethod("tmapGridCompPlot")
}

#' @export
#' @keywords internal
#' @rdname tmapGridLeaflet
tmapGridCompPrepare.default = function(comp, o) {
	cls = class(comp)[1]
	id = paste("plot_mode", cls, sep = "_")
	cli::cli_inform("{.field [plot mode]} Map component {.fun {cls}} not supported in {.str plot} mode.",
					.frequency_id = id,
					.frequency = "once")
	comp$show = FALSE
	comp
}


#' @export
tmapGridCompPrepare.tm_legend_landscape = function(comp, o) {
	tmapGridCompPrepare.tm_legend_portrait(comp, o)
}




#' @export
tmapGridCompHeight.tm_legend_landscape = function(comp, o) {
	nlev = comp$nitems
	textS = comp$text.size #* o$scale
	titleS = if (comp$title == "") 0 else comp$title.size * number_text_lines(comp$title)#* o$scale

	height = get_legend_option(comp$item.height, comp$type)

	item_height = if (comp$type == "symbols") {
		shps = rep(comp$gpar$shape, length.out = nlev)
		max(height, rep(comp$gpar$size, length.out = nlev) / textS * ifelse(shps > 999, comp$layer_args$icon.scale, 1))
	} else height
	itemHsIn = grid::unit(item_height * textS * o$lin, units = "inch")

	titleP = comp$title.padding[c(3,1)] * titleS * o$lin
	titleH = titleS * o$lin

	marH = comp$margins[c(3,1)] * textS * o$lin


	hsinch = c(titleP[1], titleH, titleP[2], marH[1], item_height * textS * o$lin, c(comp$item_text.margin, 1) * textS * o$lin,marH[2])

	Hin = if (is.na(comp$height)) sum(hsinch) else comp$height * textS * o$lin

	sides = switch(comp$position$just.v, top = "second", bottom = "first", "both")
	hsu = set_unit_with_stretch(hsinch, sides = sides)



	comp$flexRow = NA

	comp$Hin = Hin
	comp$hsu = hsu

	comp

}


#' @export
tmapGridCompWidth.tm_legend_landscape = function(comp, o) {
	nlev = comp$nitems
	textS = comp$text.size #* o$scale
	titleS = if (comp$title == "") 0 else comp$title.size #* o$scale

	space = get_legend_option(comp$item.space, comp$type)
	spaceNA = get_legend_option(comp$item.na.space, comp$type)
	width = get_legend_option(comp$item.width, comp$type)
	widthNA = get_legend_option(comp$item.na.width, comp$type)
	if (is.na(widthNA)) widthNA = width

	labelW = graphics::strwidth(comp$labels, units = "inch", cex = textS, family = comp$text.fontfamily, font = fontface2nr(comp$text.fontface)) / o$lin

	# legend width can be 3: NA (components determine width), finite number (number of text lines), or Inf (whole width)
	# for the latter two, there are 3 ways of stretching the legend: padding (space between items), items (widths of all items), or itemsNNA (widths of non-NA items)

	if (comp$type == "symbols") {
		shps = rep(comp$gpar$shape, length.out = nlev)
		item_widths = pmax(width, rep(comp$gpar$size / textS, length.out = nlev) * ifelse(shps > 999, comp$layer_args$icon.scale, 1), labelW)
		comp$stretch = if (!is.na(comp$width)) "padding" else "none"
	} else if (comp$type %in% c("rect", "lines")) {
		item_widths = pmax(rep(width, nlev), labelW)
		if (comp$na.show) item_widths[nlev] = widthNA
		comp$stretch = if (!is.na(comp$width)) "items" else "none"
	} else if (comp$type == "gradient") {
		item_widths = rep(width, nlev)
		if (comp$na.show) item_widths[nlev] = widthNA
		comp$stretch = if (!is.na(comp$width)) "itemsNNA" else "none"
	} else if (comp$type == "text") {
		item_widths = pmax(width, rep(comp$gpar$size / textS, length.out = nlev), labelW)
		comp$stretch = if (!is.na(comp$width)) "padding" else "none"
	}


	titleP = comp$title.padding[c(2,4)] * titleS * o$lin
	titleW = if (titleS > 0) graphics::strwidth(comp$title, units = "inch", cex = titleS, family = comp$title.fontfamily, font = fontface2nr(comp$title.fontface)) * o$lin else 0

	marW = comp$margins[c(2,4)] * textS * o$lin



	item_space = c(rep(space, nlev - 1 - comp$na.show), {if (comp$na.show) spaceNA else NULL})
	items_all = c(rbind(item_widths[-nlev], item_space), item_widths[nlev])


	if (comp$type == "gradient") {
		# Cut empty space
		if (comp$gradient_ht_fracs[1] != 0) {
			labelW

			# use additional column
			if (comp$labels_select[1]) {
				ih1 = min(labelW[1], item_widths[1] * comp$gradient_ht_fracs[1])
			} else {
				ih1 = item_widths[1] * comp$gradient_ht_fracs[1]
			}

			corrW = - (ih1 * textS * o$lin)
		} else {
			corrW = 0
		}
		if (comp$gradient_ht_fracs[2] != 0) {
			k = nlev - comp$na.show
			if (comp$labels_select[k]) {
				ihk = min(labelW[k], item_widths[k] * comp$gradient_ht_fracs[2])
			} else {
				ihk = item_widths[k] * comp$gradient_ht_fracs[2]
			}
			if (comp$na.show) {
				items_all[k * 2] = items_all[k * 2] - ihk
			} else {
				marW[2] = marW[2] - (ihk * textS * o$lin)
			}
		}
	} else {
		corrW = 0
	}

	ws = c(marW[1], corrW, items_all * textS * o$lin, marW[2])


	item_ids = seq(3, by = 2, length.out = nlev)
	pad_ids = seq(4, by = 2, length.out = nlev - 1L)

	wsu = if (comp$stretch == "padding") {
		set_unit_with_stretch(ws, pad_ids)
	} else if (comp$stretch == "items") {
		set_unit_with_stretch(ws, item_ids)
	} else if (comp$stretch == "itemsNNA") {
		if (comp$na.show) set_unit_with_stretch(ws, head(item_ids, -1)) else set_unit_with_stretch(ws, item_ids)
	} else {
		sides = switch(comp$position$align.h, left = "second", right = "first", "both")
		set_unit_with_stretch(ws, sides = sides)
	}

	Win = if (comp$stretch == "none") sum(ws) else comp$width * textS * o$lin

	comp$flexCol = NA

	comp$Win = Win
	#comp$hs = hs
	comp$wsu = wsu

	comp$item_ids = item_ids + 1L

	comp


}


#' @export
tmapGridCompPlot.tm_legend_landscape = function(comp, o, fH, fW) {
	if (comp$type != "gradient") comp$labels_select = TRUE # labels_select only needed for continuous legends (#1039)


	icons = all(comp$gp$shape >= 1000)

	# replace gp visual values with user-specified used (e.g. tm_shape(World) + tm_polygons("HPI", fill.legend = tm_legend(col = "red")))
	comp$gp = add_user_specified_values(comp$gp, comp[intersect(names(comp), names(comp$gp))])

	# icons replaced by 'normal' shapes: apply icon.scale
	if (icons && all(comp$gp$shape < 1000) && !is.null(comp$layer_args$icon.scale)) {
		comp$gp$size = comp$gp$size * comp$layer_args$icon.scale
	}

	textS = comp$text.size * comp$scale
	titleS = if (comp$title == "") 0 else comp$title.size * comp$scale

	nlev = comp$nitems

	wsu = comp$wsu
	hsu = comp$hsu

	vp = grid::viewport(layout = grid::grid.layout(ncol = length(wsu),
												   nrow = length(hsu),
												   widths = wsu,
												   heights = hsu))

	grobBG = if (getOption("tmap.design.mode")) rectGrob(gp=gpar(fill="#CAB2D6")) else NULL

	if (is.na(comp$title.align)) comp$title.align = comp$position$align.h

	titleGP = grid::gpar(col = comp$title.color, cex = titleS, fontface = comp$title.fontface, fontfamily = comp$title.fontfamily, alpha = comp$title.alpha)

	if (comp$title.align == "left") {
		grTitle = gridCell(3, 3:(length(comp$wsu)-2), grid::textGrob(comp$title, x = grid::unit(comp$title.padding[2] * titleS * o$lin, units = "inch"), just = "left", gp = titleGP))
	} else if (comp$title.align == "right") {
		grTitle = gridCell(3, 3:(length(comp$wsu)-2), grid::textGrob(comp$title, x = grid::unit(1, "npc") - grid::unit(comp$title.padding[4] * titleS * o$lin, units = "inch"), just = "right", gp = titleGP))
	} else {
		grTitle = gridCell(3, 3:(length(comp$wsu)-2), grid::textGrob(comp$title, x = 0.5, just = "center", gp = titleGP))
	}

	textW = graphics::strwidth(comp$labels, units = "inch", cex = textS, family = comp$text.fontfamily, font = fontface2nr(comp$text.fontface))
	scale_labels = max(textW / grid::convertUnit(wsu[comp$item_ids], unitTo = "inch", valueOnly = TRUE), 1)

	grText = mapply(function(i, id) gridCell(8, id, grid::textGrob(comp$labels[i], x = 0.5, just = "center", gp = grid::gpar(col = comp$text.color, cex = textS/scale_labels, fontface = comp$text.fontface, fontfamily = comp$text.fontfamily, alpha = comp$text.alpha))), (1L:nlev)[comp$labels_select], comp$item_ids[comp$labels_select], SIMPLIFY = FALSE)

	ticks = get_legend_option(comp$ticks, comp$type)
	ticks.disable.na = get_legend_option(comp$ticks.disable.na, comp$type)
	if (length(ticks)) {
		tick_col = if (is.na(comp$ticks.col)) comp$text.color else comp$ticks.col
		if (is.na(tick_col)) tick_col = "white"

		ticks_in_margin = sapply(ticks, function(l) all(l>=1))

		ni = nlev - (comp$na.show && ticks.disable.na)

		tck_ids = (1L:ni)[rep(comp$labels_select, length.out = ni)]

		# remove tick ids that are too close to edges
		# only if color box frame color is there
		if (!(("col" %in% names(comp) && is.na(comp$col)) || split_alpha_channel(comp$gpar$col[1], alpha = 1)$opacity == 0)) {
			ih = get_legend_option(comp$item.height, "gradient")
			na_share = ((ih - 0.25)/2)/ih
			if (comp$gradient_ht_fracs[1] >= na_share) {
				tck_ids = setdiff(tck_ids, 1)
			}
			if (comp$gradient_ht_fracs[2] > na_share) {
				tck_ids = setdiff(tck_ids, comp$nitems - comp$na.show)
			}
		}

		# tick marks in margin (specified with y coordinates between 1 and 2)
		if (any(ticks_in_margin)) {
			grTicksMargin = do.call(c, lapply(ticks[ticks_in_margin], function(y) {
				mapply(function(i, id) gridCell(8, id, grid::linesGrob(x = c(0.5, 0.5), y = y, gp = grid::gpar(col = tick_col, lwd = comp$ticks.lwd * comp$scale * o$scale_down))), tck_ids, comp$item_ids[tck_ids], SIMPLIFY = FALSE)
			}))
		} else {
			grTicksMargin = NULL
		}

		if (!all(ticks_in_margin)) {
			grTicksItem = do.call(c, lapply(ticks[!ticks_in_margin], function(y) {
				mapply(function(i, id) gridCell(6, id, grid::linesGrob(x = c(0.5, 0.5), y = 1 - y, gp = grid::gpar(col = tick_col, lwd = comp$ticks.lwd * comp$scale * o$scale_down))), tck_ids, comp$item_ids[tck_ids], SIMPLIFY = FALSE)
			}))
		} else {
			grTicksItem = NULL
		}

		grTicks = c(grTicksMargin, grTicksItem)
	} else {
		grTicks = NULL
	}


	if (getOption("tmap.design.mode")) {
		df = expand.grid(col = 1:length(comp$wsu),
						 row = 1:length(comp$hsu))

		grDesign = lapply(1:nrow(df), function(i) gridCell(df$row[i], df$col[i], grid::rectGrob(gp=gpar(fill=NA,col="#000000", lwd=1))))
	} else {
		grDesign = NULL
	}

	gp = comp$gp

	if (comp$type == "gradient") {
		# for borders
		gpars = gp_to_gpar(gp, id = 1L, sel = "col", o = o, type = comp$type)

		# for gradient fill
		nlev2 = (nlev-comp$na.show) # nlev without na
		lvs = 1:nlev2

		vary_fill = (length(gp$fill) > 1)
		vary_fill_alpha = (length(gp$fill_alpha) > 1)

		vary_col = (length(gp$col) > 1)
		vary_col_alpha = (length(gp$col_alpha) > 1)


		# vary fill color
		if (vary_fill) {
			fill_list = cont_split(gp$fill[lvs])
			fill_list = lapply(fill_list, function(i) {
				i[i=="NA"] <- NA
				i
			})
		} else if (vary_col) {
			fill_list = cont_split(gp$col[lvs])
			fill_list = lapply(fill_list, function(i) {
				i[i=="NA"] <- NA
				i
			})
			gpars$col = NA
		} else {
			fill_list = rep(gp$fill[1], nlev2)
		}

		# vary fill alpha
		if (vary_fill_alpha) {
			alpha_list = cont_split(gp$fill_alpha[lvs])
			alpha_list = lapply(alpha_list, function(i) {
				i[i=="NA"] <- 0
				as.numeric(i)
			})
			if (!vary_fill) {
				fill_list = mapply(rep.int, fill_list, vapply(alpha_list, length, FUN.VALUE = integer(1)), SIMPLIFY = FALSE, USE.NAMES = FALSE)
			}

		} else if (vary_col_alpha) {
			alpha_list = cont_split(gp$col_alpha[lvs])
			alpha_list = lapply(alpha_list, function(i) {
				i[i=="NA"] <- 0
				as.numeric(i)
			})

			if (!vary_col) {
				fill_list = mapply(rep.int, fill_list, vapply(alpha_list, length, FUN.VALUE = integer(1)), SIMPLIFY = FALSE, USE.NAMES = FALSE)
			}
		} else {
			if (!is.na(gp$fill_alpha[1])) {
				alpha_list = rep(gp$fill_alpha[1], nlev2)
			} else {
				alpha_list = rep(gp$col_alpha[1], nlev2)
			}
		}



		# fill
		nvv = o$continuous.nclass_per_legend_break

		if (o$use_gradient) {
			id1 = which(!is.na(fill_list[[1]]))[1]
			id2 = tail(which(!is.na(fill_list[[nlev2]])), 1)

			x1 = ((id1-1) / nvv) / nlev2
			x2 = (id2 / nvv) / nlev2 + ((nlev2-1)/nlev2)
			w = x2 - x1

			if (vary_fill) {
				cols = unlist(fill_list)[id1:(nvv*(nlev2-1) + id2)]
				cols_alph = paste0(cols, num_to_hex(gp$fill_alpha[1] * 255))
			} else {
				alph = unlist(alpha_list)[id1:(nvv*(nlev2-1) + id2)]
				cols_alph = paste0(col2hex(gp$fill[1]), num_to_hex(alph * 255))
			}
			grItems1 = list(gridCell(6, comp$item_ids[lvs], grid::rectGrob(x = x1 + 0.5*w, width= w, gp=gpar(fill = grid::linearGradient(colours = cols_alph), col = NA))))
		} else {
			grItems1 = mapply(function(id, f, a) {
				w = 1 / length(f)
				xs = seq(.5*w, by = w, length.out = length(f))
				#f[!is.na(f)] = "red"
				gpi = grid::gpar(fill = f, alpha = a, col = NA)
				gridCell(6, id, grid::rectGrob(x = xs, width = w, gp = gpi))
			}, comp$item_ids[lvs], fill_list, alpha_list, SIMPLIFY = FALSE)
		}

		if (vary_fill) {
			x1 = (sum(is.na(fill_list[[1]])) / nvv) / nlev2
			x2 = (sum(is.na(fill_list[[nlev2]])) / nvv) / nlev2
		} else {
			x1 = (sum(is.na(alpha_list[[1]])) / nvv) / nlev2
			x2 = (sum(is.na(alpha_list[[nlev2]])) / nvv) / nlev2
		}


		grItems2 = list(gridCell(6, comp$item_ids[lvs], rndrectGrob(x = grid::unit(x1, "npc"), just = c("left", "center"), width = grid::unit(1-(x1+x2), "npc"), gp = gpars, r = comp$item.r)))
		if (comp$na.show) {
			gpars$fill = gp$fill[ifelse(length(gp$fill), nlev, 1)]
			gpars$fill_alpha = gp$fill[ifelse(length(gp$fill_alpha), nlev, 1)]
			# fill and border for NA
			grItems1 = c(grItems1, list(gridCell(6, comp$item_ids[nlev], rndrectGrob(gp = gpars, r = comp$item.r))))
		}

		# borders
		grItems = c(grItems1, grItems2)


	} else if (comp$type == "rect") {
		#gps = split_gp(gp, n = nlev)

		diffAlpha = !any(is.na(c(gp$fill_alpha, gp$col_alpha))) && !(length(gp$fill_alpha) == length(gp$col_alpha) && all(gp$fill_alpha == gp$col_alpha))


		if (diffAlpha) {
			gpars1 = gp_to_gpar(gp, sel = "fill", split_to_n = nlev, o = o, type = comp$type) #lapply(gps, gp_to_gpar_fill)
			gpars2 = gp_to_gpar(gp, sel = "col", split_to_n = nlev, o = o, type = comp$type) #lapply(gps, gp_to_gpar_borders)

			#grItems = mapply(function(i, gpari) gridCell(i+3, 2, grid::rectGrob(gp = gpari)), 1:nlev, gpars, SIMPLIFY = FALSE)
			grItems = mapply(function(i, gpar1i, gpar2i) gridCell(6, id, {
				grid::grobTree(
					rndrectGrob(width = grid::unit(o$lin* textS, "inches"), height = grid::unit(o$lin* textS, "inches"), gp = gpar1i, r = comp$item.r),
					rndrectGrob(width = grid::unit(o$lin* textS, "inches"), height = grid::unit(o$lin* textS, "inches"), gp = gpar2i, r = comp$item.r))
			}), comp$item_ids, gpars1, gpars2, SIMPLIFY = FALSE)

		} else {
			gpars = gp_to_gpar(gp, sel = "all", split_to_n = nlev, o = o, type = comp$type)#lapply(gps, gp_to_gpar)
			#grItems = mapply(function(i, gpari) gridCell(i+3, 2, rndrectGrob(gp = gpari)), 1:nlev, gpars, SIMPLIFY = FALSE)
			grItems = mapply(function(id, gpari) gridCell(6, id, rndrectGrob(gp = gpari, r = comp$item.r)), comp$item_ids, gpars, SIMPLIFY = FALSE)
		}

	} else if (comp$type == "lines") {
		gpars = gp_to_gpar(gp, sel = "col", split_to_n = nlev, o = o, type = comp$type)#lapply(gps, gp_to_gpar)
		#grItems = mapply(function(i, gpari) gridCell(i+3, 2, grid::rectGrob(gp = gpari)), 1:nlev, gpars, SIMPLIFY = FALSE)
		grItems = mapply(function(id, gpari) gridCell(6, id, grid::linesGrob(x = grid::unit(c(0.5,0.5), "npc"), gp = gpari)), comp$item_ids, gpars, SIMPLIFY = FALSE)
	} else if (comp$type == "symbols") {
		if (length(gp$size) == 1) gp$size = min(gp$size, min(get_legend_option(comp$item.height, "symbols"),
															 get_legend_option(comp$item.width, "symbols")) * comp$textS)

		gp = swap_pch_15_20(gp)

		gpars = gp_to_gpar(gp, split_to_n = nlev, o = o, type = comp$type)

		diffAlpha = !anyNA(c(gp$fill_alpha, gp$col_alpha)) && !(length(gp$fill_alpha) == length(gp$col_alpha) && all(gp$fill_alpha == gp$col_alpha))
		if (diffAlpha) {
			gpars1 = gp_to_gpar(gp, split_to_n = nlev, o = o, type = comp$type, sel = "fill")
			gpars2 = gp_to_gpar(gp, split_to_n = nlev, o = o, type = comp$type, sel = "col")
		} else {
			gpars1 = vector(mode = "list", length = nlev)
			gpars2 = vector(mode = "list", length = nlev)
		}

		# scale down (due to facet use)
		gpars = lapply(gpars, rescale_gp, scale = o$scale_down)


		shapeLib = get("shapeLib", envir = .TMAP)
		justLib = get("justLib", envir = .TMAP)

		grItems = mapply(function(id, gpari, gpari1, gpari2) {
			grobs = if (gpari$shape > 999) {
				grbs = gList(shapeLib[[gpari$shape-999]])
				grid::gTree(children=grbs, vp=viewport(x=0.5,
													   y=0.5,
													   width=unit(gpari$size*9/10, "lines"),
													   height=unit(gpari$size*9/10, "lines")))
			} else {
				if (diffAlpha) {
					grid::grobTree(
						grid::pointsGrob(x=0.5, y=0.5, pch = gpari1$shape, size = grid::unit(gpari1$size, "lines"), gp = gpari1),
						grid::pointsGrob(x=0.5, y=0.5, pch = gpari2$shape, size = grid::unit(gpari2$size, "lines"), gp = gpari2)
					)
				} else {
					grid::pointsGrob(x=0.5, y=0.5, pch = gpari$shape, size = grid::unit(gpari$size, "lines"), gp = gpari)
				}
			}
			#grb = gTree(children=do.call(gList, grobs), name=paste0("symbols_", id))
			gridCell(6, id, grobs)
		}, comp$item_ids, gpars, gpars1, gpars2, SIMPLIFY = FALSE)



	} else if (comp$type == "text") {
		gp$text[is.na(gp$text)] = getAesOption("value.const", o, aes = "text", layer = "text")

		if (length(gp$cex) == 1) gp$cex = min(gp$cex, min(get_legend_option(comp$item.height, "text"),
														  get_legend_option(comp$item.width, "text")) * comp$textS)

		bgcols = rep(gp$bgcol, length.out = nlev)
		bgcols_alpha = rep(gp$bgcol_alpha, length.out = nlev)

		# in case size is a continuous scale
		if (is.character(gp$cex)) gp$cex = vapply(cont_split(gp$cex), FUN = function(x) {
			as.numeric(x[round(length(x)/2)])
		}, FUN.VALUE = numeric(1))


		gpars = gp_to_gpar(gp, split_to_n = nlev, o = o, type = comp$type)

		# scale down (due to facet use)
		gpars = lapply(gpars, rescale_gp, scale = o$scale_down)

		grItems = mapply(function(id, gpari, txt, bgcol, bgcol_alpha, size) {
			gridCell(6, id, {
				grid::gList(
					rndrectGrob(width = grid::unit(grid::convertWidth(grid::stringWidth(txt), "inches")* size, "inches"), height = grid::unit(o$lin* textS, "inches"), gp = grid::gpar(fill = bgcol, alpha = bgcol_alpha, col = NA), r = comp$item.r),
					grid::textGrob(x=0.5, y=0.5, label = txt, gp = gpari))
			})
		}, comp$item_ids, gpars, gp$text, bgcols, bgcols_alpha, gp$cex, SIMPLIFY = FALSE)

	}


	g = do.call(grid::grobTree, c(list(grobBG, grTitle), grText, grItems, grTicks, grDesign, list(vp = vp)))

	g

}

#' tmap function to specify labels
#'
#' tmap function to specify labels used in the scale functions, e.g. via the argument `label.format` in [tm_scale_intervals()].
#'
#' @param fun Function to specify the labels. It should take a numeric vector, and should return a character vector of the same size. By default it is not specified. If specified, the list items \code{scientific}, \code{format}, and \code{digits} (see below) are not used.
#' @param scientific Should the labels be formatted scientifically? If so, square brackets are used, and the \code{format} of the numbers is \code{"g"}. Otherwise, \code{format="f"}, and \code{text.separator}, \code{text.less.than}, and \code{text.or.more} are used. Also, the numbers are automatically  rounded to millions or billions if applicable. By default, \code{FALSE}
#' @param format By default, \code{"f"}, i.e. the standard notation \code{xxx.xxx}, is used. If \code{scientific = TRUE} then \code{"g"}, which means that numbers are formatted scientifically, i.e. \code{n.dddE+nn} if needed to save space.
#' @param digits Number of digits after the decimal point if \code{format="f"}, and the number of significant digits otherwise. By default \code{NA}, meaning as many as needed to have distinct numbers
#' @param interval.disjoint In case of intervals (see [tm_scale_intervals()]), should the intervals appear disjoint, e.g. \code{0 to 999}, \code{1000 - 1999}, \code{2000 - 2999} (\code{TRUE}, default), or not: \code{0 - 1000}, \code{1000 - 2000}, \code{2000- 3000}.
#' @param big.num.abbr Vector that defines whether and which abbrevations are used for large numbers. It is a named numeric vector, where the name indicated the abbreviation, and the number the magnitude (in terms on numbers of zero). Numbers are only abbrevation when they are large enough. Set it to \code{NA} to disable abbrevations.  The default is \code{c("mln" = 6, "bln" = 9)}. For layers where \code{style} is set to \code{log10} or \code{log10_pretty}, the default is \code{NA}.
#' @param prefix Prefix of each number
#' @param suffix Suffix of each number
#' @param text.separator Character string to use to separate numbers in an interval legend (default: \code{"to"}).
#' @param text.less.than Character value(s) to use for 'less than'. Default \code{"Less than"}. When a character vector of length 2 is specified, one for each word, these words are aligned when \code{text.to.columns = TRUE}
#' @param text.less.than_as.prefix Should \code{text.less.than} be used as prefix?
#' @param text.or.more Character value(s) to use to  'or more'. Default is \code{"or more"}. When a character vector of length 2 is specified, one for each word, these words are aligned when \code{text.to.columns = TRUE}
#' @param text.or.more_as.prefix Should \code{text.or.more} be used as prefix?
#' @param text.align Not implemented in v4 (yet). Value that determines how the numbers are aligned, \code{"left"}, \code{"center"} or \code{"right"}. By default \code{"left"}.
#' @param text.to.columns Not implemented in v4 (yet). Logical that determines whether the text is aligned to three columns (from, text.separator, to). By default \code{FALSE}.
#' @param html.escape Logical that determins whther HTML code is escaped in the popups in view mode. By default \code{TRUE}. If set to \code{FALSE} HTML code can be added, e.g. to added white space via \code{&nbsp;}.
#' @param ... arguments passed on to \code{\link[base:formatC]{formatC}}
#' @return list with formatting options
#' @export
tm_label_format = function(fun,
						   scientific,
						   format,
						   digits,
						   interval.disjoint,
						   big.num.abbr,
						   prefix,
						   suffix,
						   text.separator,
						   text.less.than,
						   text.less.than_as.prefix,
						   text.or.more,
						   text.or.more_as.prefix,
						   text.align,
						   text.to.columns,
						   html.escape,
						   ...) {

	args = lapply(as.list(rlang::call_match(dots_expand = TRUE)[-1]), eval, envir = parent.frame())
	args$called = names(args)

	structure(args, class = c("tmapLabelFormat", "list"))
}

#' @export
#' @keywords internal
#' @rdname tmapGridLeaflet
tmapLeafletAuxPrepare = function(a, bs, id, o) {
	UseMethod("tmapLeafletAuxPrepare")
}


#' @export
#' @keywords internal
#' @rdname tmapGridLeaflet
tmapLeafletAuxPlot = function(a, bi, bbx, facet_row, facet_col, facet_page, id, pane, group, o) {
	UseMethod("tmapLeafletAuxPlot")
}

#' @export
#' @keywords internal
#' @rdname tmapGridLeaflet
tmapLeafletAuxPrepare.tm_aux_basemap = function(a, bs, id, o) {
	tmapLeafletAuxPrepare.tm_aux_tiles(a, bs, id, o)
}

#' @export
#' @keywords internal
#' @rdname tmapGridLeaflet
tmapLeafletAuxPrepare.tm_aux_tiles = function(a, bs, id, o) {
	nms = names(a$server)
	a$server = unname(a$server)
	tiles = lapply(1L:length(bs), function(i) a)
	.TMAP_LEAFLET$tiles[[id]] = tiles
	if (!is.null(nms)) {
		paste0(nms, collapse = "__")
	} else {
		paste0(a$server, collapse = "__")
	}
}

#' @export
#' @keywords internal
#' @rdname tmapGridLeaflet
tmapLeafletAuxPlot.tm_aux_basemap = function(a, bi, bbx, facet_row, facet_col, facet_page, id, pane, group, o) {
	tmapLeafletAuxPlot.tm_aux_tiles(a, bi, bbx, facet_row, facet_col, facet_page, id, pane, group, o)
}

#' @export
#' @keywords internal
#' @rdname tmapGridLeaflet
tmapLeafletAuxPlot.tm_aux_tiles = function(a, bi, bbx, facet_row, facet_col, facet_page, id, pane, group, o) {

	lf = get_lf(facet_row, facet_col, facet_page)

	rc_text = frc(facet_row, facet_col)


	tiles = .TMAP_LEAFLET$tiles[[id]][[bi]]

	if (o$credits.defined) {
		opt = leaflet::tileOptions(attribution = "", maxNativeZoom = tiles$max.native.zoom, pane = pane)
	} else {
		opt = leaflet::tileOptions(maxNativeZoom = tiles$max.native.zoom, pane = pane)
	}
	if (!is.na(o$set_zoom_limits[2])) opt$maxZoom = o$set_zoom_limits[2]

	k = length(tiles$server)

	groups = rep(strsplit(group, "__", fixed = TRUE)[[1]], length.out = k)

	for (i in 1L:k) {
		if (tiles$server[i] != "") {
			if ((substr(tiles$server[i], 1, 4) == "http")) {
				lf = leaflet::addTiles(lf, urlTemplate = tiles$server[i], group = groups[i], options = opt)
			} else {
				lf = leaflet::addProviderTiles(lf, provider = tiles$server[i], group = groups[i], options = opt)
			}
		}
	}

	assign_lf(lf, facet_row, facet_col, facet_page)
	NULL
}

#' @export
#' @keywords internal
#' @rdname tmapGridLeaflet
tmapLeafletAuxPrepare.tm_aux_graticules = function(a, bs, id, o) {
	tmapLeafletAuxPrepare.tm_aux_grid(a, bs, id, o)
}

#' @export
#' @keywords internal
#' @rdname tmapGridLeaflet
tmapLeafletAuxPrepare.tm_aux_grid = function(a, bs, id, o) {

	grid_intervals = vapply(bs, function(b) {
		# implementation similarish as plot mode but needs polishing
		dx = b[3] - b[1]
		dy = b[4] - b[2]
		# if not specified, aim for 15 lines in total
		n.x = if (!is.na(a$n.x)) a$n.x else if (is.na(a$x)) 7.5 else length(a$x)
		n.y = if (!is.na(a$n.y)) a$n.y else if (is.na(a$y)) 7.5 else length(a$x)

		x = pretty30(b[c(1,3)], n=n.x, longlat = !is.na(a$crs) && sf::st_is_longlat(a$crs))
		y = pretty30(b[c(2,4)], n=n.y, longlat = !is.na(a$crs) && sf::st_is_longlat(a$crs))

		max((x[2] - x[1]), (y[2]-y[1]))
	}, FUN.VALUE = numeric(1))
	assign("grid_intervals", grid_intervals, envir = .TMAP_LEAFLET)

	return("grid")
}

#' @export
#' @keywords internal
#' @rdname tmapGridLeaflet
tmapLeafletAuxPlot.tm_aux_graticules = function(a, bi, bbx, facet_row, facet_col, facet_page, id, pane, group, o) {
	tmapLeafletAuxPlot.tm_aux_grid(a, bi, bbx, facet_row, facet_col, facet_page, id, pane, group, o)
}

#' @export
#' @keywords internal
#' @rdname tmapGridLeaflet
tmapLeafletAuxPlot.tm_aux_grid = function(a, bi, bbx, facet_row, facet_col, facet_page, id, pane, group, o) {
	lf = get_lf(facet_row, facet_col, facet_page)
	rc_text = frc(facet_row, facet_col)

	grid_intervals = get("grid_intervals", envir = .TMAP_LEAFLET)

	lf = leaflet::addGraticule(lf, interval = grid_intervals[bi], options = pathOptions(pane = pane))


	assign_lf(lf, facet_row, facet_col, facet_page)
	NULL
}

tmapLeafletGridXLab = function(bi, bbx, facet_row, facet_col, facet_page, o) {
	NULL
}

tmapLeafletGridYLab = function(bi, bbx, facet_row, facet_col, facet_page, o) {
	NULL
}

#' Map layer: symbols
#'
#' Map layer that draws symbols Supported visual variables are:
#' `fill` (the fill color), `col` (the border color), `size` the symbol size,
#' `shape` the symbol shape, `lwd` (line width), `lty` (line type), `fill_alpha`
#' (fill color alpha transparency) and `col_alpha` (border color alpha transparency).
#'
#' A symbol shape specification is one of the following three options.
#' \enumerate{
#'  \item{A numeric value that specifies the plotting character of the symbol. See parameter \code{pch} of \code{\link[graphics:points]{points}} and the last example to create a plot with all options. Note that this is not supported for the \code{"view" mode.}}
#'  \item{A \code{\link[grid:grid.grob]{grob}} object, which can be a ggplot2 plot object created with \code{\link[ggplot2:ggplotGrob]{ggplotGrob}}. To specify multiple shapes, a list of grob objects is required. Tip: for proportional symbols, such as donuts or pies, see the extension package `r .doc_see_vig("ext_glyphs")`.}
#'  \item{An icon specification, which can be created with \code{\link{tmap_icons}}.}
#'  }
#'  To specify multiple shapes (needed for the \code{shapes} argument), a vector or list of these shape specification is required. The shape specification options can also be mixed. For the \code{shapes} argument, it is possible to use a named vector or list, where the names correspond to the value of the variable specified by the \code{shape} argument.
#'  For small multiples, a list of these shape specification(s) should be provided.
#'
#' @param fill,fill.scale,fill.legend,fill.chart,fill.free `r .doc_vv("fill")` `r .doc_unit("fill")`
#' @param col,col.scale,col.legend,col.chart,col.free `r .doc_vv("col")` `r .doc_unit("col")`
#' @param size,size.scale,size.legend,size.chart,size.free `r .doc_vv("size")` `r .doc_unit("size")`
#' @param shape,shape.scale,shape.legend,shape.chart,shape.free `r .doc_vv("shape")` `r .doc_unit("shape")`
#' @param lwd,lwd.scale,lwd.legend,lwd.chart,lwd.free `r .doc_vv("lwd")` `r .doc_unit("lwd")`
#' @param lty,lty.scale,lty.legend,lty.chart,lty.free `r .doc_vv("lty")` `r .doc_unit("lty")`
#' @param fill_alpha,fill_alpha.scale,fill_alpha.legend,fill_alpha.chart,fill_alpha.free `r .doc_vv("fill_alpha")` `r .doc_unit("fill_alpha")`
#'   the fill color alpha transparency See details.
#' @param col_alpha,col_alpha.scale,col_alpha.legend,col_alpha.chart,col_alpha.free `r .doc_vv("col_alpha")` `r .doc_unit("col_alpha")`
#' @param plot.order Specification in which order the spatial features are drawn.
#'   See [tm_plot_order()] for details.
#' @param options options passed on to the corresponding `opt_<layer_function>` function
#' @inherit tm_polygons details
#' @inheritParams tm_polygons
#' @seealso `r .doc_see_also_layers("symbols")`
#' @param ... to catch deprecated arguments from version < 4.0
#' @example ./examples/tm_symbols.R
#' @export
tm_symbols = function(size = tm_const(),
					  size.scale = tm_scale(),
					  size.legend = tm_legend(),
					  size.chart = tm_chart_none(),
					  size.free = NA,
					  fill = tm_const(),
					  fill.scale = tm_scale(),
					  fill.legend = tm_legend(),
					  fill.chart = tm_chart_none(),
					  fill.free = NA,
					  col = tm_const(),
					  col.scale = tm_scale(),
					  col.legend = tm_legend(),
					  col.chart = tm_chart_none(),
					  col.free = NA,
					  shape = tm_const(),
					  shape.scale = tm_scale(),
					  shape.legend = tm_legend(),
					  shape.chart = tm_chart_none(),
					  shape.free = NA,
					  lwd = tm_const(),
					  lwd.scale = tm_scale(),
					  lwd.legend = tm_legend(),
					  lwd.chart = tm_chart_none(),
					  lwd.free = NA,
					  lty = tm_const(),
					  lty.scale = tm_scale(),
					  lty.legend = tm_legend(),
					  lty.chart = tm_chart_none(),
					  lty.free = NA,
					  fill_alpha = tm_const(),
					  fill_alpha.scale = tm_scale(),
					  fill_alpha.legend = tm_legend(),
					  fill_alpha.chart = tm_chart_none(),
					  fill_alpha.free = NA,
					  col_alpha = tm_const(),
					  col_alpha.scale = tm_scale(),
					  col_alpha.legend = tm_legend(),
					  col_alpha.chart = tm_chart_none(),
					  col_alpha.free = NA,
					  plot.order = tm_plot_order("size"),
					  zindex = NA,
					  group = NA,
					  group.control = "check",
					  popup = tm_popup(),
					  popup.vars = NA,
					  popup.format = tm_label_format(),
					  hover = NA,
					  id = "",
					  blend = "over",
					  options = opt_tm_symbols(),
					  ...) {

	args = list(...)
	args_called = names(rlang::call_match()[-1])


	res = v3_symbols(args, args_called, col)
	if (!is.null(res)) {
		fill = res$fill
		col = res$col
		col_alpha = res$col_alpha
		fill_alpha = res$fill_alpha
		fill.scale = res$fill.scale
		fill.legend = res$fill.legend
		size.scale = res$size.scale
		size.legend = res$size.legend
		shape.scale = res$shape.scale
		shape.legend = res$shape.legend
		fill.chart = res$fill.chart
	}

	# unused arguments: typos?
	unused = setdiff(names(args), c(v3_only("tm_symbols"), "called_from"))

	if (length(unused)) {
		layer_fun = if ("called_from" %in% names(args)) {
			args$called_from
		} else {
			"tm_symbols"
		}

		message_layer_unused_args(layer_fun, unused)
	}

	# make sure required options are there
	options = complete_options(options, opt_tm_symbols())

	# blend has been migrated to the layer function root; it is still passed on
	# internally via mapping.args (read as a$blend).
	options$mapping.args$blend = blend

	# resolve popup specification (popup.vars/popup.format are deprecated in
	# favour of popup = tm_popup(...))
	popup = process_popup(popup = popup,
						  popup.vars = popup.vars,
						  popup.format = popup.format,
						  popup.called = "popup" %in% args_called,
						  popup.vars.called = "popup.vars" %in% args_called,
						  popup.format.called = "popup.format" %in% args_called,
						  layer_fun = if ("called_from" %in% names(args)) args$called_from else "tm_symbols")

	tm_element_list(tm_element(
		layer = "symbols",
		trans.fun = tmapTransCentroid,
		trans.aes = list(),
		trans.args = options$trans.args,
		trans.apply_to = "this",
		mapping.aes = list(size = tmapScale(aes = "size",
											value = size,
											scale = size.scale,
											legend = size.legend,
											chart = size.chart,
											free = size.free),

						   fill = tmapScale(aes = "fill",
						   				 value = fill,
						   				 scale = fill.scale,
						   				 legend = fill.legend,
						   				 chart = fill.chart,
						   				 free = fill.free),
						   col = tmapScale(aes = "col",
						   				value = col,
						   				scale = col.scale,
						   				legend = col.legend,
						   				chart = col.chart,
						   				free = col.free),
						   shape = tmapScale(aes = "shape",
						   				  value = shape,
						   				  scale = shape.scale,
						   				  legend = shape.legend,
						   				  chart = shape.chart,
						   				  free = shape.free),
						   lwd = tmapScale(aes = "lwd",
						   				value = lwd,
						   				scale = lwd.scale,
						   				legend = lwd.legend,
						   				chart = lwd.chart,
						   				free = lwd.free),
						   lty = tmapScale(aes = "lty",
						   				value = lty,
						   				scale = lty.scale,
						   				legend = lty.legend,
						   				chart = lty.chart,
						   				free = lty.free),
						   fill_alpha = tmapScale(aes = "fill_alpha",
						   					   value = fill_alpha,
						   					   scale = fill_alpha.scale,
						   					   legend = fill_alpha.legend,
						   					   chart = fill_alpha.chart,
						   					   free = fill_alpha.free),
						   col_alpha = tmapScale(aes = "col_alpha",
						   					  value = col_alpha,
						   					  scale = col_alpha.scale,
						   					  legend = col_alpha.legend,
						   					  chart = col_alpha.chart,
						   					  free = col_alpha.free)),

		gpar = tmapGpar(fill = "__fill",
						col = "__col",
						shape = "__shape",
						size = "__size",
						fill_alpha = "__fill_alpha",
						col_alpha = "__col_alpha",
						pattern = "fill",
						lty = "__lty",
						lwd = "__lwd",
						linejoin = NA,
						lineend = NA),
		tpar = tmapTpar(),
		plot.order = plot.order,
		mapping.args = options$mapping.args,
		zindex = zindex,
		group = group,
		group.control = group.control,
		popup.vars = popup$vars,
		popup.format = popup$format,
		popup.title = popup$title,
		popup.layout = popup$layout,
		hover = hover,
		id = id,
		subclass = c("tm_aes_layer", "tm_layer")))
}

v3_symbols = function(args, args_called, arg_col = NULL) {
	if (any(v3_only("tm_symbols") %in% names(args))) {
		layer_fun = if ("called_from" %in% names(args)) {
			args$called_from
		} else {
			"symbols"
		}

		v3_start_message()
		if (!("style" %in% names(args))) {
			if (!"breaks" %in% names(args)) {
				style = "pretty"
			} else {
				style = "fixed"
			}
		} else {
			style = args$style
		}

		v3_list_init()
		if (length(style) > 1) {
			style = style[1]
			.TMAP$v3_list$mult = TRUE
		}

		# v3 visual variable: fill
		fill.scale.args = c(list(n = v3_impute(args, "n", 5),
								 style = style,
								 style.args = v3_impute(args, "style.args", list())),
							if (style %in% c("cont", "log10")) {
								c({
									if (!is.null(args$breaks)) {
										if (length(args$breaks) > 2) {
											list(ticks = v3_impute(args, "breaks", NULL, "ticks"))
										} else {
											list(ticks = v3_impute(args, "breaks", NULL, "limits"))
										}
									} else {
										list()
									}
								},
								list(outliers.trunc = c(TRUE, FALSE)))
							} else {
								list(breaks = v3_impute(args, "breaks", NULL),
									 interval.closure = v3_impute(args, "interval.closure", "left"),
									 drop.levels = v3_impute(args, "drop.levels", FALSE))
							},
							list(midpoint = v3_impute(args, "midpoint", NULL),
								 as.count = v3_impute(args, "as.count", NA),
								 values = v3_impute(args, "palette", NA, "values"),
								 values.repeat = !v3_impute(args, "stretch.palette", TRUE, "values.repeat"),
								 values.range = v3_impute(args, "contrast", NA, "values.range"),
								 values.scale = 1,
								 value.na = v3_impute(args, "colorNA", NA, "value.na"),
								 value.null = v3_impute(args, "colorNULL", NA, "value.null"),
								 value.neutral = NA,
								 labels = v3_impute(args, "labels", NULL),
								 label.na = v3_impute(args, "textNA", "Missing", "label.na"),
								 label.null = NA,
								 label.format = v3_impute(args, "legend.format", list(), "label.format")))
		fill.scale.args$fun_pref = if (style == "cat") {
			"categorical"
		} else if (style %in% c("fixed", "sd", "equal", "pretty", "quantile",
								"kmeans", "hclust", "bclust", "fisher", "jenks",
								"dpih", "headtails", "log10_pretty")) {
			"intervals"
		} else if (style == "cont") {
			"continuous"
		} else if (style == "log10") {
			"continuous_log"
		} else if (style == "order") {
			"rank"
		} else {
			stop("unknown style")
		}

		if ("style" %in% names(args)) {
			v3_tm_scale_instead_of_style(style, scale_fun = fill.scale.args$fun_pref, vv = "fill", layer_fun = layer_fun, arg_list = v3_list_get())
		} else {
			v3_tm_scale(scale_fun = "", vv = "fill", layer_fun = layer_fun, arg_list = v3_list_get())
		}
		fill.scale = do.call("tm_scale", args = fill.scale.args)

		if ("col" %in% args_called) {
			fill = arg_col
			isn_fill = is.null(fill)
			v3_message_col_fill(layer_fun = layer_fun)
			if (isn_fill) {
				v3_message_vv_null(layer_fun = layer_fun)
				fill = NA
			}
			col = tm_const()
		} else {
			fill = tm_const()
			col = tm_const()
			isn_fill = FALSE
		}

		if ("border.col" %in% names(args)) {
			col = args$border.col
			isn_col = is.null(col)
			if (!("col" %in% args_called)) v3_message_col_fill(layer_fun = layer_fun)
			if (isn_col) {
				if (!isn_fill) v3_message_vv_null(layer_fun = layer_fun)
				col = NA
			}
		}


		if ("alpha" %in% names(args)) {
			v3_message_fill_alpha(layer_fun = layer_fun)
			fill_alpha = args$alpha
		} else {
			fill_alpha = tm_const()
		}

		if ("border.alpha" %in% names(args)) {
			v3_message_col_alpha(layer_fun = layer_fun)
			col_alpha = args$border.alpha
		} else {
			col_alpha = tm_const()
		}

		v3_list_init()
		if ("legend.show" %in% names(args) && !args$legend.show) {
			v3_tm_legend_hide(layer_fun, arg = "legend.show", vv = "fill")
			fill.legend = tm_legend_hide()
		} else {
			fill.legend.args = list(atitle = v3_impute(args, "title.col", NA, "title"),
									na.show = v3_impute(args, "showNA", NA, "na.show"),
									format = v3_impute(args, "legend.format", list(), "format"),
									orientation = ifelse(v3_impute(args, "legend.is.portrait", TRUE, "orientation"), "portrait", "landscape"),
									reverse = v3_impute(args, "legend.reverse", FALSE, "reverse"))
			fill.legend = do.call("tm_legend", fill.legend.args)
			v3_tm_legend(fun = layer_fun, vv = "fill", arg_list = v3_list_get())
		}


		# v3 visual variable: size
		v3_list_init()
		size.scale.args = list(ticks = v3_impute(args, "sizes.legend", NULL, "ticks"),
							   values.range = {if ("size.lim" %in% names(args)) {
							   	c(args[["size.lim"]][1] / args[["size.lim"]][2], 1)
							   } else c(0, 1)},
							   values.scale = v3_impute(args, "scale", 1, "values.scale"),
							   limits = v3_impute(args, "size.lim", NULL, "limits"),
							   outliers.trunc = c(TRUE, FALSE),
							   labels = v3_impute(args, "sizes.legend.labels", NULL, "labels"),
							   fun_pref = "continuous")
		if ("size" %in% args_called) v3_tm_scale(scale_fun = "continuous", vv = "size", layer_fun = layer_fun, arg_list = v3_list_get())
		size.scale = do.call("tm_scale", args = size.scale.args)

		v3_list_init()
		if ("legend.size.show" %in% names(args) && !args$legend.size.show) {
			v3_tm_legend_hide(layer_fun, arg = "legend.size.show", vv = "size")
			size.legend = tm_legend_hide()
		} else {
			size.legend.args = alist(title = v3_impute(args, "title.size", NA, "title"),
									na.show = v3_impute(args, "showNA", NA),
									format = v3_impute(args, "legend.format", list(), "format"),
									orientation = ifelse(v3_impute(args, "legend.size.is.portrait", FALSE), "portrait", "landscape"),
									reverse = v3_impute(args, "legend.size.reverse", FALSE))
			if ("size" %in% args_called) v3_tm_legend(fun = layer_fun, vv = "size", arg_list = v3_list_get())
			size.legend = do.call("tm_legend", size.legend.args)
		}

		# v3 visual variable: shape
		if (!("shapes.style" %in% names(args))) {
			if (!"shapes.breaks" %in% names(args)) {
				shapes.style = "pretty"
			} else {
				shapes.style = "fixed"
			}
		} else {
			shapes.style = args$shapes.style
		}

		v3_list_init()
		shape.scale.args = list(n = v3_impute(args, "shapes.n", 5, "n"),
								style = shapes.style,
								style.args = v3_impute(args, "shapes.style.args", list(), "style.args"),
								breaks = v3_impute(args, "shapes.breaks", NULL, "breaks"),
								interval.closure = v3_impute(args, "shapes.interval.closure", "left", "interval.closure"),
								drop.levels = v3_impute(args, "drop.levels", FALSE),
								midpoint = v3_impute(args, "midpoint", NULL),
								as.count = v3_impute(args, "as.count", NA),
								value.neutral = v3_impute(args, "shapes.legend", NA, "value.neutral"),
								values = v3_impute(args, "shapes", 21:25, "values"),
								labels = v3_impute(args, "shapes.labels", NULL, "labels"),
								label.na = v3_impute(args, "shape.textNA", NA, "label.na"),
								label.null = NA,
								label.format = v3_impute(args, "legend.format", list(), "label.format"),
								fun_pref = "intervals")
		shape.scale = do.call("tm_scale", args = shape.scale.args)
		if ("shape" %in% args_called) {
			if ("shapes.style" %in% args) {
				v3_tm_scale_instead_of_style(shapes.style, scale_fun = shape.scale.args$fun_pref, vv = "shape", layer_fun = layer_fun, arg_list = v3_list_get())
			} else {
				v3_tm_scale(scale_fun = shape.scale.args$fun_pref, vv = "shape", layer_fun = layer_fun, arg_list = v3_list_get())
			}
		}

		v3_list_init()
		if ("legend.shape.show" %in% names(args) && !args$legend.shape.show) {
			v3_tm_legend_hide(layer_fun, arg = "legend.shape.show", vv = "shape")
			shape.legend = tm_legend_hide()
		} else {
			shape.legend.args = alist(title = v3_impute(args, "title.shape", NA),
									 na.show = v3_impute(args, "shape.showNA ", NA),
									 format = v3_impute(args, "legend.format", list(), "format"),
									 orientation = ifelse(v3_impute(args, "legend.shape.is.portrait", TRUE), "portrait", "landscape"),
									 reverse = v3_impute(args, "legend.shape.reverse", FALSE))
			if ("shape" %in% args_called)v3_tm_legend(fun = layer_fun, vv = "shape", arg_list = v3_list_get())
			shape.legend = do.call("tm_legend", shape.legend.args)
		}

		if ("legend.hist" %in% names(args) && args$legend.hist) {
			fill.chart = tm_chart_histogram()
			v3_tm_chart_hist(layer_fun = layer_fun, vv = "fill", arg = "legend.hist")

			# to do: histogram title
		} else {
			fill.chart = tm_chart_none()
		}


		v4_opt_args = c("icon.scale", "just", "grob.dim")
		v3_opt_args = c("icon.scale", "just", "grob.dim")
		osel = which(v3_opt_args %in% names(args))
		if (length(osel)) {
			o3 = v3_opt_args[osel]
			o4 = v4_opt_args[osel]
			v3_layer_opt(o3, o4, layer_fun)
		}

		list(fill = fill,
			 col = col,
			 col_alpha = col_alpha,
			 fill_alpha = fill_alpha,
			 fill.scale = fill.scale,
			 fill.legend = fill.legend,
			 size.scale = size.scale,
			 size.legend = size.legend,
			 shape.scale = shape.scale,
			 shape.legend = shape.legend,
			 fill.chart = fill.chart)
	} else {
		NULL
	}
}



#' @export
#' @rdname tm_symbols
tm_dots = function(fill = tm_const(),
				   fill.scale = tm_scale(),
				   fill.legend = tm_legend(),
				   fill.free = NA,
				   size = tm_const(),
				   size.scale = tm_scale(),
				   size.legend = tm_legend(),
				   size.free = NA,
				   lwd = tm_const(),
				   lwd.scale = tm_scale(),
				   lwd.legend = tm_legend(),
				   lwd.free = NA,
				   lty = tm_const(),
				   lty.scale = tm_scale(),
				   lty.legend = tm_legend(),
				   lty.free = NA,
				   fill_alpha = tm_const(),
				   fill_alpha.scale = tm_scale(),
				   fill_alpha.legend = tm_legend(),
				   fill_alpha.free = NA,
				   plot.order = tm_plot_order("DATA"),
				   zindex = NA,
				   group = NA,
				   group.control = "check",
				   options = opt_tm_dots(),
				   ...) {

		args = c(as.list(environment()), list(...))
		args$called_from = "tm_dots"

		tm = do.call(tm_symbols, args)
		tm[[1]]$layer = c("dots", "symbols")
		tm
}


#' @export
#' @rdname tm_symbols
tm_bubbles = function(size = tm_const(),
					  size.scale = tm_scale(),
					  size.legend = tm_legend(),
					  size.free = NA,
					  fill = tm_const(),
					  fill.scale = tm_scale(),
					  fill.legend = tm_legend(),
					  fill.free = NA,
					  col = tm_const(),
					  col.scale = tm_scale(),
					  col.legend = tm_legend(),
					  col.free = NA,
					  lwd = tm_const(),
					  lwd.scale = tm_scale(),
					  lwd.legend = tm_legend(),
					  lwd.free = NA,
					  lty = tm_const(),
					  lty.scale = tm_scale(),
					  lty.legend = tm_legend(),
					  lty.free = NA,
					  fill_alpha = tm_const(),
					  fill_alpha.scale = tm_scale(),
					  fill_alpha.legend = tm_legend(),
					  fill_alpha.free = NA,
					  col_alpha = tm_const(),
					  col_alpha.scale = tm_scale(),
					  col_alpha.legend = tm_legend(),
					  col_alpha.free = NA,
					  plot.order = tm_plot_order("size"),
					  zindex = NA,
					  group = NA,
					  group.control = "check",
					  options = opt_tm_bubbles(),
					  ...) {

	args = c(as.list(environment()), list(...))
	args$called_from = "tm_bubbles"

	tm = do.call(tm_symbols, args)
	tm[[1]]$layer = c("bubbles", "symbols")
	tm
}


#' @export
#' @rdname tm_symbols
tm_squares = function(size = tm_const(),
					  size.scale = tm_scale(),
					  size.legend = tm_legend(),
					  size.free = NA,
					  fill = tm_const(),
					  fill.scale = tm_scale(),
					  fill.legend = tm_legend(),
					  fill.free = NA,
					  col = tm_const(),
					  col.scale = tm_scale(),
					  col.legend = tm_legend(),
					  col.free = NA,
					  lwd = tm_const(),
					  lwd.scale = tm_scale(),
					  lwd.legend = tm_legend(),
					  lwd.free = NA,
					  lty = tm_const(),
					  lty.scale = tm_scale(),
					  lty.legend = tm_legend(),
					  lty.free = NA,
					  fill_alpha = tm_const(),
					  fill_alpha.scale = tm_scale(),
					  fill_alpha.legend = tm_legend(),
					  fill_alpha.free = NA,
					  col_alpha = tm_const(),
					  col_alpha.scale = tm_scale(),
					  col_alpha.legend = tm_legend(),
					  col_alpha.free = NA,
					  plot.order = tm_plot_order("size"),
					  zindex = NA,
					  group = NA,
					  group.control = "check",
					  options = opt_tm_squares(),
					  ...) {

	args = c(as.list(environment()), list(...))
	args$called_from = "tm_squares"
	tm = do.call(tm_symbols, args)
	tm[[1]]$layer = c("squares", "symbols")
	tm
}

#' @rdname tm_symbols
#' @inheritParams tm_text
#' @export
tm_markers = function(text = tm_const(),
					  text.scale = tm_scale(),
					  text.legend = tm_legend(),
					  text.chart = tm_chart_none(),
					  text.free = NA,
					  size = tm_const(),
					  size.scale = tm_scale(),
					  size.legend = tm_legend(),
					  size.chart = tm_chart_none(),
					  size.free = NA,
					  col = tm_const(),
					  col.scale = tm_scale(),
					  col.legend = tm_legend(),
					  col.chart = tm_chart_none(),
					  col.free = NA,
					  col_alpha = tm_const(),
					  col_alpha.scale = tm_scale(),
					  col_alpha.legend = tm_legend(),
					  col_alpha.chart = tm_chart_none(),
					  col_alpha.free = NA,
					  fontface = tm_const(),
					  fontface.scale = tm_scale(),
					  fontface.legend = tm_legend(),
					  fontface.chart = tm_chart_none(),
					  fontface.free = NA,
					  fontfamily = "",
					  bgcol = tm_const(),
					  bgcol.scale = tm_scale(),
					  bgcol.legend = tm_legend(),
					  bgcol.chart = tm_chart_none(),
					  bgcol.free = NA,
					  bgcol_alpha = tm_const(),
					  bgcol_alpha.scale = tm_scale(),
					  bgcol_alpha.legend = tm_legend(),
					  bgcol_alpha.chart = tm_chart_none(),
					  bgcol_alpha.free = NA,
					  xmod = 0,
					  xmod.scale = tm_scale(),
					  xmod.legend = tm_legend_hide(),
					  xmod.chart = tm_chart_none(),
					  xmod.free = NA,
					  ymod = 0,
					  ymod.scale = tm_scale(),
					  ymod.legend = tm_legend_hide(),
					  ymod.chart = tm_chart_none(),
					  ymod.free = NA,
					  angle = 0,
					  angle.scale = tm_scale(),
					  angle.legend = tm_legend_hide(),
					  angle.chart = tm_chart_none(),
					  angle.free = NA,
					  plot.order = tm_plot_order("AREA", reverse = FALSE, na.order = "bottom"),
					  zindex = NA,
					  group = NA,
					  group.control = "check",
					  options = opt_tm_markers(),
					  ...) {

	if (!setequal(names(options), c("markers", "text", "dots"))) cli::cli_abort("{.field [tm_markers]} options should have {.str markers}, {.str text} and {.str dots} items")
	e = as.list(environment())
	e$options = options$text
	a = list(...)

	sel = substr(names(a), 1, 5) == "dots_"
	a_dots = a[sel]
	a_other = a[!sel]
	names(a_dots) = substr(names(a_dots), 6, nchar(names(a_dots)))

	args_dots = c(a_dots, list(plot.order = plot.order, zindex = zindex, group = group, group.control = group.control, options = options$dots))
	args_text = c(e, a_other, list(called_from = "tm_markers"))

	tm_t = do.call(tm_text, args_text)
	tm_t[[1]]$layer = c("labels", "text")

	tm_d = do.call(tm_dots, args_dots)
	tm_d[[1]]$layer = c("markers", "symbols")

	if (options$markers$markers_on_top_of_text) tm_t + tm_d else tm_d + tm_t

}

#' @inheritParams opt_tm_labels
#' @param markers_on_top_of_text should markers be plot on top of the text (by default `FALSE`)
#' @param dots.icon.scale scaling number that determines how large the icons (or grobs) are in plot mode in comparison to proportional symbols (such as bubbles). In view mode, the size is determined by the icon specification (see \code{\link{tmap_icons}}) or, if grobs are specified by \code{grob.width} and \code{grob.height}
#' @param dots.just justification of the text relative to the point coordinates. Either one of the following values: \code{"left"} , \code{"right"}, \code{"center"}, \code{"bottom"}, and \code{"top"}, or a vector of two values where first value specifies horizontal and the second value vertical justification. Besides the mentioned values, also numeric values between 0 and 1 can be used. 0 means left justification for the first value and bottom justification for the second value. Note that in view mode, only one value is used.
#' @param dots.grob.dim vector of four values that determine how grob objects (see details) are shown in view mode. The first and second value are the width and height of the displayed icon. The third and fourth value are the width and height of the rendered png image that is used for the icon. Generally, the third and fourth value should be large enough to render a ggplot2 graphic successfully. Only needed for the view mode.
#' @rdname tm_symbols
#' @export
opt_tm_markers = function(markers_on_top_of_text = FALSE,
						  points_only = "ifany",
						  point_per = "feature",
						  on_surface = FALSE,
						  shadow = FALSE,
						  shadow.offset.x = 0.1,
						  shadow.offset.y = 0.1,
						  just = "center",
						  along_lines = TRUE,
						  bg.padding = 0.4,
						  clustering = TRUE,
						  point.label = TRUE,
						  point.label.gap = 0.4,
						  point.label.method = "SANN",
						  remove_overlap = FALSE,
						  dots.just = NA,
						  dots.icon.scale = 3,
						  dots.grob.dim = c(width=48, height=48, render.width=256, render.height=256)) {
	list(markers = list(markers_on_top_of_text = markers_on_top_of_text),
		 text = opt_tm_labels(points_only = points_only,
		 					 point_per = point_per,
		 					 on_surface = on_surface,
		 					 shadow = shadow,
		 					 shadow.offset.x = shadow.offset.x,
		 					 shadow.offset.y = shadow.offset.y,
		 					 just = just,
		 					 along_lines = along_lines,
		 					 bg.padding = bg.padding,
		 					 clustering = clustering,
		 					 point.label = point.label,
		 					 point.label.gap = point.label.gap,
		 					 point.label.method = point.label.method,
		 					 remove_overlap = remove_overlap),
		 dots = opt_tm_dots(points_only = points_only,
		 				   point_per = point_per,
		 				   on_surface = on_surface,
		 				   clustering = clustering,
		 				   icon.scale = dots.icon.scale,
		 				   just = dots.just,
		 				   grob.dim = dots.grob.dim))
}


#' @param icon.scale scaling number that determines how large the icons (or grobs) are in plot mode in comparison to proportional symbols (such as bubbles). For view mode, use the argument `grob.dim`
#' @param grob.dim vector of four values that determine how grob objects (see details) are shown in view mode. The first and second value are the width and height of the displayed icon. The third and fourth value are the width and height of the rendered png image that is used for the icon. Generally, the third and fourth value should be large enough to render a ggplot2 graphic successfully. Only needed for the view mode.
#' @param clustering in interactive modes (e.g. \code{"view"} mode), should clustering be applied at lower zoom levels? Either `FALSE` (default), `TRUE`, or a mode specific specification, e.g. for \code{"view"} mode \code{\link[leaflet:markerClusterOptions]{markerClusterOptions}}.
#' @param hitbox Controls whether an invisible interaction layer with a
#'   larger clickable area (\"hitbox\") is added on top of the symbols.
#'
#'   This can improve click and popup behaviour for small or densely packed
#'   symbols by enlarging the effective mouse interaction area.
#'
#'   Possible values:
#'   \describe{
#'     \item{"none"}{No additional hitbox layer is added. Symbols are clickable
#'     only at their visible size.}
#'
#'     \item{"plusX"}{Adds \code{X} pixels to the visible symbol diameter to
#'     compute the interaction size. For example, \code{"plus8"} widens the
#'     clickable area by 4 pixels around each symbol edge.}
#'
#'     \item{"pmaxX"}{Ensures a minimum interaction diameter of \code{X} pixels:
#'     \code{pmax(symbol_diameter, X)}. Useful only for very small symbols, as
#'     it adds no margin to symbols already larger than \code{X}.}
#'
#'     \item{"auto"}{\code{"pmax12"} if and only if interactive features are
#'     enabled (popup or hover), symbols are small (median visible diameter
#'     < 12px), and there are fewer than 10000 features. Otherwise \code{"none"}.}
#'   }
#'
#'   \code{plus} and \code{pmax} can be combined, e.g. \code{"plus4pmax8"}.
#'   The hitbox is not applied when clustering is enabled.
#' @rdname tm_symbols
#' @export
opt_tm_symbols = function(points_only = "ifany",
						  point_per = "feature",
						  on_surface = FALSE,
						  clustering = FALSE,
						  icon.scale = 3,
						  just = NA,
						  hitbox = "auto",
						  grob.dim = c(width=48, height=48, render.width=256, render.height=256)) {
	list(trans.args = list(points_only = points_only, point_per = point_per, on_surface = on_surface, along_lines = FALSE),
		 mapping.args = list(icon.scale = icon.scale,
		 					just = just,
		 					clustering = clustering,
		 					hitbox = hitbox,
		 					grob.dim = grob.dim))
}

#' @rdname tm_symbols
#' @export
opt_tm_dots = opt_tm_symbols


#' @rdname tm_symbols
#' @export
opt_tm_bubbles = opt_tm_symbols

#' @rdname tm_symbols
#' @export
opt_tm_squares = opt_tm_symbols

tmapGridArrange = function(tms, nx, ncol, nrow, opts, knit, show, args, options) {
	widths <- opts$widths
	heights = opts$heights
	if (is.na(widths[1])) widths <- rep(1/ncol, ncol)
	if (is.na(heights[1])) heights <- rep(1/nrow, nrow)

	grid.newpage()
	vp <- viewport(layout=grid.layout(nrow=nrow, ncol=ncol, widths = widths, heights=heights), name = "tmap_arrangement")
	pushViewport(vp)

	if (!is.null(opts$asp) || !is.null(opts$outer.margins)) {
		layout_args <- list(asp=opts$asp, outer.margins=opts$outer.margins)
		layout_args <- layout_args[!vapply(layout_args, is.null, logical(1))]
		tml <- do.call(tm_options, layout_args)
	} else {
		tml <- NULL
	}

	nc <- 1
	nr <- 1
	for (i in 1:nx) {
		tm <- tms[[i]]
		if (!is.null(tml)) tm <- tm + tml
		print(tm, vp = viewport(layout.pos.col = nc, layout.pos.row = nr))
		nc <- nc + 1
		if (nc > ncol) {
			nc <- 1
			nr <- nr + 1
		}
	}
}

pchs = stats::setNames(c(seq(0L, 25L, 1L), seq(100L, 109L, 1L)),
					   c(c('open-rect', 'open-circle', 'open-triangle', 'simple-plus',
					     'simple-cross', 'open-diamond', 'open-down-triangle', 'cross-rect',
					     'simple-star', 'plus-diamond', 'plus-circle', 'hexagram', 'plus-rect',
					     'cross-circle', 'triangle-rect', 'solid-rect', 'solid-circle-md',
					     'solid-triangle', 'solid-diamond', 'solid-circle-bg', 'solid-circle-sm', 'circle',
					     'rect', 'diamond', 'triangle', 'down-triangle'
					   ),
					     c('rect', 'circle', 'triangle', 'plus', 'cross', 'diamond', 'star', 'stadium', 'line', 'polygon')
					   ))


get_pch_names = function(x) {
	if (is.numeric(x)) {
		if (!(all(x %in% pchs | x > 999))) {
			cli::cli_abort("Unknown symbol values")
		}
		y = names(pchs)[match(x, pchs)]
		y[x > 999] = x[x>999]
		y
	} else {
		if (!all(x %in% names(pchs))) {
			unknown <- unique(x[!x %in% names(pchs)])
			cli::cli_abort("Unknown symbol values: {.val {unknown}}")
		}
		x
	}
}
#
# get_pch_number = function(x) {
# 	if (is.numeric(x)) {
# 		if (!(all(x %in% pchs | x > 999))) stop("Unknown symbol values", call. = FALSE)
# 		x
# 	} else {
# 		if (!all(x %in% names(pchs))) stop("Unknown symbol values", call. = FALSE)
# 		unname(pchs[match(x, names(pchs))])
# 	}
# }

#' Map layer: basemap / overlay tiles
#'
#' Map layer that draws tiles from a tile server. `tm_basemap()` draws the tile
#' layer as basemap, i.e. as bottom layer. In contrast, `tm_tiles()` draws the
#' tile layer as overlay layer, where the stacking order corresponds with the
#' order in which this layer is called, just like other map layers.
#'
#'  API keys. For Stadia and Thunderforest maps, an API key is required.
#'  This can be set via the argument `api`. Alternatively they can be stored in environment variables `"STADIA_MAPS"` and `THUNDERFOREST_MAPS` with `Sys.setenv`
#'
#' @param server Name of the provider or an URL. Or a vector of multiple values. The list of available providers
#'   can be obtained with `providers` (tip: in RStudio, type `leaflet::providers$` to see
#'   the options). See <https://leaflet-extras.github.io/leaflet-providers/preview/>
#'   for a preview of those. When a URL is provided, it should be in template format,
#'   e.g. \code{"https://{s}.tile.openstreetmap.org/{z}/{x}/{y}.png"}.
#'   Use `NULL` in `tm_basemap()` to disable basemaps. It can be a named vector. In that case these names will be used a group names, as alternative to the argument `group`.
#' @param alpha Transparency level
#' @param zoom Zoom level (only used in plot mode)
#' @param api API key. Needed for `Stadia` and `Thunderforest` maps in plot mode. See details
#' @param max.native.zoom Maximum native zoom level (only used in view mode).
#'   The minimum and maximum zoom levels are determined in `tm_view()`.
#' @param sub subdomain of the tile server. Only used when `server` is a url template. The
#'   default is `"abc"` which works for most tile servers.
#' @param zindex zindex of the pane in view mode. By default, it is set to the
#'   layer number plus 400. By default, the tmap layers will therefore be placed
#'   in the custom panes `"tmap401"`, `"tmap402"`, etc., except for the base tile
#'   layers, which are placed in the standard `"tile"`. This parameter determines
#'   both the name of the pane and the z-index, which determines the pane order
#'   from bottom to top. For instance, if `zindex` is set to 500, the pane will
#'   be named `"tmap500"`.
#' @param group Name of the group to which this layer belongs. This is only
#'   relevant in view mode, where layer groups can be switched (see `group.control`)
#' @param group.control In view mode, the group control determines how layer
#'   groups can be switched on and off. Options: `"radio"` for radio buttons
#'   (meaning only one group can be shown), `"check"` for check boxes
#'   (so multiple groups can be shown), and `"none"` for no control
#'   (the group cannot be (de)selected).
#' @export
#' @seealso \href{https://r-tmap.github.io/tmap/articles/basics_basemaps}{Basemap examples}
#' @example ./examples/tm_basemap.R
tm_basemap = function(server = NA, alpha = NULL, zoom = NULL, api = NULL, max.native.zoom = 17, sub = "abc", zindex = 0, group = NA, group.control = "radio") {
	if (is.null(server)) {
		disable = TRUE
	} else {
		if (is.na(server[1])) {
			server = NULL # complete_options is used in step1 to find the default
		}
		disable = FALSE
	}

	group = check_basemap_group(server, group, called_from = "tm_basemap")


	tm_element_list(tm_element(
		args = list(server = server, alpha = alpha, zoom = zoom, api = api, max.native.zoom = max.native.zoom, sub = sub, type = "basemap", disable = disable),
		mapping.fun = "tm_aux_basemap",
		zindex = zindex,
		group = group,
		group.control = group.control,
		subclass = c("tm_basemap", "tm_aux_layer")))
}

#' @export
#' @rdname tm_basemap
tm_tiles = function(server = NA, alpha = NULL, zoom = NULL, max.native.zoom = 17, sub = "abc", zindex = NA, group = NA, group.control = "check") {
	if (is.null(server)) {
		disable = TRUE
	} else {
		if (is.na(server[1])) {
			server = NULL # complete_options is used in step1 to find the default
		}
		disable = FALSE
	}

	group = check_basemap_group(server, group, called_from = "tm_tiles")


	tm_element_list(tm_element(
		args = list(server = server, alpha = alpha, zoom = zoom, max.native.zoom = max.native.zoom, sub = sub, type = "overlay", disable = disable),
		mapping.fun = "tm_aux_tiles",
		zindex = zindex,
		group = group,
		group.control = group.control,
		subclass = c("tm_tiles", "tm_aux_layer")))
}

check_basemap_group = function(server, group, called_from) {
	if (is.null(server)) return(group)
	if (!is.na(group[1])) {
		if (length(group) != length(server)) {
			cli::cli_warn("{.field {.fun {called_from}}} multiple {.arg group} names specified, but unequal to the length of {.arg server}")
		} else {
			group = NA
		}

		if (!is.null(names(server))) {
			cli::cli_warn("{.field {.fun {called_from}}} {.arg group} specified, but {.arg server} is already a named vector, so {.arg group} will be ignored")
			group = NA
		}
	}
	group
}



#' @export
#' @rdname tm_grid
tm_graticules = function(x = NA,
						 y = NA,
						 n.x = NA,
						 n.y = NA,
						 crs = 4326,
						 labels.format = list(suffix = intToUtf8(176)),
						 labels.cardinal = TRUE,
						 ...) {
	do.call(
		tm_grid,
		c(list(x = x, y = y, n.x = n.x, n.y = n.y, crs = crs, labels.format = labels.format, labels.cardinal = labels.cardinal), list(...))
		)
}

#' Coordinate grid / graticule lines
#'
#' Draws horizontal and vertical lines according to a coordinate reference system (CRS). `tm_grid()` uses the CRS of the (master) shape object, and `tm_graticules()` uses latitude and longitude coordinates (EPSG 4326). It creates a [`tmap-element`] that draws coordinate grid lines. It serves as a
#' layer that can be drawn anywhere between other layers.
#'
#' @param x X coordinates for vertical grid lines. If `NA`, it is specified
#'   with a pretty scale and `n.x`.
#' @param y Y coordinates for horizontal grid lines. If `NA`, it is specified
#'   with a pretty scale and `n.y`.
#' @param n.x Preferred number of grid lines for the x axis. For the labels,
#'   a [pretty()] sequence is used, so the number of actual labels may be different than `n.x`.
#' @param n.y Preferred number of grid lines for the y axis. For the labels, a
#'   [pretty()] sequence is used, so the number of actual labels may be different than `n.y`.
#' @param crs Projection character. If specified, the grid lines are projected
#'   accordingly. Many world maps are projected, but still have latitude longitude
#'   (EPSG 4326) grid lines.
#' @param col Color of the grid lines.
#' @param lwd Line width of the grid lines
#' @param alpha Alpha transparency of the grid lines. Number between 0 and 1.
#'   By default, the alpha transparency of `col` is taken.
#' @param labels.show Show tick labels. Either one value for both `x` and `y` axis,
#'   or a vector two: the first for `x` and latter for `y`.
#' @param labels.pos position of the labels. Vector of two: the horizontal
#'   ("left" or "right") and the vertical ("top" or "bottom") position.
#' @param labels.size Font size of the tick labels
#' @param labels.col Font color of the tick labels
#' @param labels.rot Rotation angles of the labels. Vector of two values: the
#'   first is the rotation angle (in degrees) of the tick labels on the x axis
#'   and the second is the rotation angle of the tick labels on the y axis.
#'   Only `0`, `90`, `180`, and `270` are valid values.
#' @param labels.format List of formatting options for the grid labels. Parameters are:
#' \describe{
#' \item{fun}{Function to specify the labels. It should take a numeric vector,
#'   and should return a character vector of the same size.
#'   By default it is not specified. If specified, the list items `scientific`,
#'   `format`, and `digits` (see below) are not used.}
#' \item{scientific}{Should the labels be formatted scientifically?
#'   If so, square brackets are used, and the `format` of the numbers is `"g"`.
#'   Otherwise, `format="f"`, and `text.separator`, `text.less.than`, and
#'   `text.or.more` are used. Also, the numbers are automatically rounded to
#'   millions or billions if applicable.}
#' \item{format}{By default, `"f"`, i.e. the standard notation `xxx.xxx`, is used.
#'   If `scientific=TRUE` then `"g"`, which means that numbers are formatted scientifically,
#'   i.e. `n.dddE+nn` if needed to save space.}
#' \item{digits}{Number of digits after the decimal point if `format="f"`, and
#'   the number of significant digits otherwise.}
#' \item{...}{Other arguments passed on to [formatC()]}
#' }
#' @param labels.cardinal Add the four cardinal directions (N, E, S, W) to the labels,
#'   instead of using negative coordinates for west and south (so it assumes that
#'   the coordinates are positive in the north-east direction).
#' @param labels.margin.x Margin between tick labels of x axis and the frame.
#'   Note that when `labels.inside_frame = FALSE` and `ticks = TRUE`,
#'   the ticks will be adjusted accordingly.
#' @param labels.margin.y Margin between tick labels of y axis and the frame.
#'   Note that when `labels.inside_frame = FALSE` and `ticks = TRUE`,
#'   the ticks will be adjusted accordingly.
#' @param labels.space.x Space that is used for the labels and ticks for the x-axis
#'   when `labels.inside_frame = FALSE`. By default, it is determined automatically
#'   using the widths and heights of the tick labels. The unit of this parameter is text line height.
#' @param labels.space.y Space that is used for the labels and ticks for the y-axis
#'   when `labels.inside_frame = FALSE`. By default, it is determined automatically
#'   using the widths and heights of the tick labels. The unit of this parameter is text line height.
#' @param labels.inside_frame Show labels inside the frame? By default `FALSE`.
#' @param ticks If `labels.inside_frame = FALSE`, should ticks can be drawn between the labels and the frame?
#'   Either one value for both `x` and `y` axis, or a vector two: the first for `x` and latter for `y`.
#' @param lines If `labels.inside_frame = FALSE`, should grid lines can be drawn?
#' @param ndiscr Number of points to discretize a parallel or meridian
#'   (only applicable for curved grid lines)
#' @param zindex zindex of the pane in view mode. By default, it is set to the
#'   layer number plus 400. By default, the tmap layers will therefore be placed
#'   in the custom panes `"tmap401"`, `"tmap402"`, etc., except for the base tile
#'   layers, which are placed in the standard `"tile"`. This parameter determines
#'   both the name of the pane and the z-index, which determines the pane order
#'   from bottom to top. For instance, if `zindex` is set to 500,
#'   the pane will be named `"tmap500"`.
#' @param group Name of the group to which this layer belongs.
#'   This is only relevant in view mode, where layer groups can be
#'   switched (see `group.control`)
#' @param group.control In view mode, the group control determines how layer groups
#'   can be switched on and off. Options: `"radio"` for radio buttons (meaning only one group can be shown),
#'   `"check"` for check boxes (so multiple groups can be shown), and `"none"`
#'   for no control (the group cannot be (de)selected).
#' @param ... Used to catch deprecated arguments from tmap v3.
#' @export
#' @example ./examples/tm_grid.R
tm_grid = function(x = NA,
				   y = NA,
				   n.x = NA,
				   n.y = NA,
				   crs = NA,
				   col = NA,
				   lwd = 1,
				   alpha = NA,
				   labels.show = TRUE,
				   labels.pos = c("left", "bottom"),
				   labels.size = .6,
				   labels.col = NA,
				   labels.rot = c(0, 0),
				   labels.format = list(big.mark = ","),
				   labels.cardinal = FALSE,
				   labels.margin.x = 0,
				   labels.margin.y = 0,
				   labels.space.x = NA,
				   labels.space.y = NA,
				   labels.inside_frame = FALSE,
				   ticks = labels.show & !labels.inside_frame,
				   lines = TRUE,
				   ndiscr = 100,
				   zindex = NA,
				   group = NA,
				   group.control = "none",
				   ...) {
	args = list(...)
	if ("projection" %in% names(args)) {
		cli::cli_warn("The 'projection' argument of 'tm_grid()' is deprecated as of tmap 4.0. Pleaes use 'crs' instead.", call. = FALSE)
		crs = args$projection
	}
	tm_element_list(tm_element(
		args = c(list(show = TRUE), as.list(environment())),
		mapping.fun = "tm_aux_grid",
		zindex = zindex,
		group = group,
		group.control = group.control,
		subclass = c("tm_grid", "tm_aux_layer")))
}

#' @export
tmapShape.SpatRaster = function(shp, is.main, crs, bbox, unit, filter, layer, shp_name, smeta, o, tmf) {
	#tmapShape.stars(stars::st_as_stars(shp), is.main, crs, bbox, unit, filter, shp_name)
	rlang::check_installed("terra")

	shp = downsample_SpatRaster(shp, max.raster = max_cells(o$raster.max_cells) / (o$fn[1] * o$fn[2]))


	ctabs = terra::coltab(shp)
	cats = terra::levels(shp)

	minmax = as.list(as.data.frame(terra::minmax(shp, compute = all(!terra::hasMinMax(shp)))))

	dt = data.table::setDT(terra::as.data.frame(shp, na.rm=FALSE))
	dt[, tmapID__:=1L:.N]
	#dt = data.table::melt(dt, id.vars = "tmapID__", variable.name = "layer", value.name = "value")

	xy_dim = dim(shp)[1:2]
	b = sf::st_bbox(shp)

	crs = sf::st_crs(shp)

	shp = structure(list(x = structure(list(from = 1, to = xy_dim[2], offset = b[1], delta = (b[3] - b[1]) / xy_dim[2], refsys = crs, point = FALSE, values = NULL), class = "dimension"),
						 y = structure(list(from = 1, to = xy_dim[1], offset = b[4], delta = (b[2] - b[4]) / xy_dim[1], refsys = crs, point = FALSE, values = NULL), class = "dimension")), class = "dimensions")
	attr(shp, "raster") = structure(list(affine = c(0, 0), dimensions = c("x", "y"), curvilinear = FALSE), class = "stars_raster")


	shpclass = "stars"

	dtcols = setdiff(names(dt), "tmapID__")

	names(ctabs) = dtcols
	names(cats) = dtcols
	names(minmax) = dtcols

	if (!is.null(tmf)) make_by_vars(dt, tmf, smeta)


	if (is.null(filter)) filter = rep(TRUE, nrow(dt))
	dt[, ':='(sel__ = filter)] # tmapID__ = 1L:nrow(dt),

	shpTM = shapeTM(shp = shp, tmapID = 1L:(nrow(shp) * ncol(shp)), bbox = bbox)

	for (nm in dtcols) {
		if (!is.null(ctabs[[nm]])) {
			# in case of predefined colors: if categories are associated (cats), use tm_scale_categorical, otherwise tm_

			ct = ctabs[[nm]]
			lt = cats[[nm]]
			if (is.factor(dt[[nm]])) {
				#levels(dt[[nm]])

				ids = match(lt[match(levels(dt[[nm]]), lt[,2]), 1], ct$value)
				cti = ct[ids,]

				cls = rgb(cti$red, cti$green, cti$blue, cti$alpha, maxColorValue = 255)

				levels(dt[[nm]]) = paste(levels(dt[[nm]]), cls, sep = "=<>=")
			} else if ("value" %in% names(ct)) {
				cls = rgb(ct$red, ct$green, ct$blue, ct$alpha, maxColorValue = 255)

				rng = minmax[[nm]]
				sel = ct$value %in% seq(rng[1], rng[2])

				#until = anyDuplicated(cls) - 1L

				cls = cls[sel]
				ct = ct[sel, ]
				dt[[nm]] = factor(dt[[nm]], levels = ct$value, labels = paste(ct$value, cls, sep = "=><="))
			}

		}
	}

	structure(list(shpTM = shpTM, dt = dt, is.main = is.main, dtcols = dtcols, shpclass = shpclass, bbox = bbox, unit = unit, shp_name = shp_name, smeta = smeta, type_ids = NULL, type_vars = NULL), class = "tmapShape")

}

#' @export
tmapShape.SpatVector = function(shp, is.main, crs, bbox, unit, filter, layer, shp_name, smeta, o, tmf) {
	tmapShape.sf(sf::st_as_sf(shp), is.main, crs, bbox, unit, filter, layer, shp_name, smeta, o, tmf)
}

#' @export
tmapSubsetShp.SpatRaster = function(shp, vars) {
	#v = setdiff(vars, names(shp))
	#if (length(v)) stop("Variable(s) not found: \"", paste(v, collapse = "\",\""), "\"")
	if (!length(vars)) {
		terra::rast(extent = terra::ext(shp), crs = terra::crs(shp), vals = TRUE, names = "dummy__")
	} else {
		shp[[vars]]
	}
}

#' @export
tmapSubsetShp.SpatVector = function(shp, vars) {
	if ("AREA_" %in% vars && !("AREA_" %in% names(shp))) {
		shp$AREA_ = terra::expanse(shp)
	}
	if ("LENGTH_" %in% vars && !("LENGTH_" %in% names(shp))) {
		shp$LENGTH = terra::perim(shp)
	}
	tmapSubsetShp.sf(shp, vars)
}

#' @export
tmapGetShapeMeta2.SpatRaster = function(shp, smeta, o) {
	if (terra::ncell(shp) > max_cells(o$raster.max_cells)) {
		# NOTE: this option is not ideal, because categories may be undiscovered
		# NOTE2: maybe the same should be done with large stars?
		shp = terra::spatSample(shp, 1e5, method = "regular", as.raster = TRUE)
	}
	smeta$vars_levs = lapply(terra::values(shp, dataframe=TRUE), function(dat) {
		get_fact_levels_na(dat, o)
	})

	names(smeta$vars_levs) = names(shp)
	smeta
}

#' @export
tmapGetShapeMeta2.SpatVector = function(shp, smeta, o) {

	# slow, needs to be improved with terra functions, e.g. unique and levels
	smeta$vars_levs = lapply(terra::values(shp), function(dat) {
		get_fact_levels_na(dat, o)
	})
	names(smeta$vars_levs) = names(shp)
	smeta
}



#' @export
tmapGetShapeMeta1.SpatRaster = function(shp, layer, o) {
	vars = names(shp)
	names(vars) = vars

	dims = character(0)
	dims_vals = list()

	list(vars = vars,
		 dims = dims,
		 dims_vals = dims_vals)

}

#' @export
tmapGetShapeMeta1.SpatVector = function(shp, layer, o) {
	vars = names(shp)
	names(vars) = vars

	dims = character(0)
	dims_vals = list()

	list(vars = vars,
		 dims = dims,
		 dims_vals = dims_vals)

}


#' @export
tmapShape.PackedSpatRaster = function(shp, ...) {
	tmapShape.SpatRaster(terra::unwrap(shp), ...)
}

#' @export
tmapShape.PackedSpatVector = function(shp, ...) {
	tmapShape.SpatVector(terra::unwrap(shp), ...)
}

#' @export
tmapSubsetShp.PackedSpatRaster = function(shp, vars) {
	tmapSubsetShp.SpatRaster(terra::unwrap(shp), vars)
}

#' @export
tmapSubsetShp.PackedSpatVector = function(shp, vars) {
	tmapSubsetShp.SpatVector(terra::unwrap(shp), vars)
}

#' @export
tmapGetShapeMeta1.PackedSpatRaster = function(shp, layer, o) {
	tmapGetShapeMeta1.SpatRaster(terra::unwrap(shp), layer, o)
}

#' @export
tmapGetShapeMeta1.PackedSpatVector = function(shp, layer, o) {
	tmapGetShapeMeta1.SpatVector(terra::unwrap(shp), layer, o)
}

#' @export
tmapGetShapeMeta2.PackedSpatRaster = function(shp, smeta, o) {
	tmapGetShapeMeta2.SpatRaster(terra::unwrap(shp), smeta, o)
}

#' @export
tmapGetShapeMeta2.PackedSpatVector = function(shp, smeta, o) {
	tmapGetShapeMeta2.SpatVector(terra::unwrap(shp), smeta, o)
}

#' @export
#' @rdname tmapGridLeaflet
tmapGridDataPlot.tm_data_text = function(a, shpTM, dt, gp, bbx, facet_row, facet_col, facet_page, id, pane, group, glid, o, ...) {
	rc_text = frc(facet_row, facet_col)

	if (("prop_angle" %in% names(shpTM))) {
		a$point.label = FALSE
	}

	res = select_sf(shpTM, dt[!is.na(dt$size), ])

	shp = res$shp
	dt = res$dt



	# specials non-vv (later on lost after gp_to_gpar)
	shadow = a$shadow


	gp = impute_gp(gp, dt)
	gp = rescale_gp(gp, o$scale_down)

	coords = sf::st_coordinates(shp)

	# in case shp is a multipoint (point_per == "segment"), expand gp:
	cp = expand_coords_gp(coords, gp, ndt = nrow(dt))
	coords = cp$coords
	gp = cp$gp

	n = nrow(coords)

	# by default point.label = NA for tm_labels
	# for small n (< 500) it will be set to TRUE
	if (is.na(a$point.label)) a$point.label = (n < 500)

	if (cp$expanded) {
		shpTM_match = match(shpTM$tmapID_expanded, shpTM$tmapID)
		text = dt$text[shpTM_match]
	} else {
		text = dt$text
	}



	g = get("g", .TMAP_GRID)


	# calculate native per line
	wIn = g$colsIn[g$cols_facet_ids[facet_col]]
	hIn = g$rowsIn[g$rows_facet_ids[facet_row]]

	wNative = bbx[3] - bbx[1]
	hNative = bbx[4] - bbx[2]

	xIn = wNative / wIn
	yIn = hNative / hIn

	lineIn = convertHeight(unit(1, "lines"), "inch", valueOnly = TRUE)



	just = process_just(a$just, interactive = FALSE)
	if (a$point.label) {
		if (!all(just == 0.5)) {
			just = c(0.5, 0.5)
			if (get("tmapOptions", envir = .TMAP)$show.messages) message("Point labeling is enabled. Therefore, just will be ignored.")

		}
	}


	# apply xmod and ymod
	coords[,1] = coords[,1] + xIn * lineIn * gp$cex * gp$xmod
	coords[,2] = coords[,2] + yIn * lineIn * gp$cex * gp$ymod


	# specials vv (later on lost after gp_to_gpar)
	bgcol = gp$bgcol
	bgcol_alpha = gp$bgcol_alpha

	angle = gp$angle

	gp = gp_to_gpar(gp, sel = "col", o = o, type = "text")

	with_bg = any(bgcol_alpha != 0)
	with_shadow_halo = a$shadow || a$halo


	if (with_bg || with_shadow_halo || a$remove_overlap || a$point.label) {
		# grobs are processed seperately because of the order: backgrond1, shadow1, text1, background2, shadow2, text2, etc.
		# becaues it is less efficient when there is no background/shadow (majority of use cases), this is a separate routine

		gps = split_gp(gp, n)

		grobTextList = mapply(function(txt, x , y, gp, a) {
			grid::textGrob(x = grid::unit(x, "native"), y = grid::unit(y, "native"), label = txt, gp = gp, rot = a, just = just) #, name = paste0("text_", id))
		}, text, coords[,1], coords[,2], gps, angle, SIMPLIFY = FALSE, USE.NAMES = FALSE)

		if (with_shadow_halo) {
			gp_sh = gp

			if (a$halo) {
				gp_sh$col = ifelse(is.na(a$halo.col), ifelse(is_light(gp$col), "#000000", "#FFFFFF"), a$halo.col)
			} else {
				gp_sh$col = ifelse(is.na(a$shadow.col), ifelse(is_light(gp$col), "#000000", "#FFFFFF"), a$shadow.col)
			}
			gps_sh = split_gp(gp_sh, n)
			grobTextShList = mapply(function(x, y, txt, g, ai) {
				if (a$halo) {
					offsets <- rbind(
						c(-1, 0),
						c( 1, 0),
						c( 0,-1),
						c( 0, 1)
					)

					diag_offsets <- rbind(
						c(-1,-1),
						c( 1,-1),
						c(-1, 1),
						c( 1, 1)
					)

					# allow smaller diagonal distance for rounder halo
					diag_offsets <- diag_offsets * .5 * sqrt(2)

					offsets <- rbind(offsets, diag_offsets)

					grid::textGrob(x = grid::unit(x + a$halo.width * xIn * lineIn * g$cex * offsets[,1], "native"),
								   y = grid::unit(y -  a$halo.width * yIn * lineIn * g$cex * offsets[,2], "native"),
								   label = txt,
								   gp = g,
								   rot = ai,
								   just = just)
				} else {
					grid::textGrob(x = grid::unit(x + a$shadow.offset.x * xIn * lineIn * g$cex, "native"),
								   y = grid::unit(y -  a$shadow.offset.y * yIn * lineIn * g$cex, "native"),
								   label = txt,
								   gp = g,
								   rot = ai,
								   just = just)
				}
			}, coords[,1], coords[,2], text, gps_sh, angle, SIMPLIFY = FALSE, USE.NAMES = FALSE)
		} else {
			grobTextShList = NULL
		}

		if (with_bg || a$remove_overlap) {
			tGH = vapply(grobTextList, function(grb) {
				grb$rot = 0
				convertHeight(grobHeight(grb), "inch", valueOnly = TRUE)
			}, FUN.VALUE = numeric(1), USE.NAMES = FALSE) * yIn

			tGW = vapply(grobTextList, function(grb) {
				grb$rot = 0
				convertWidth(grobWidth(grb), "inch", valueOnly = TRUE)
			}, FUN.VALUE = numeric(1), USE.NAMES = FALSE) * xIn

			justx <- .5 - just[1]
			justy <- .5 - just[2]

			#tGX <- grobText$x + unit(tGW * justx, "native")
			#tGY <- grobText$y + unit(tGH * justy, "native")


			tGX = unit(coords[,1] + justx * tGW, "native")
			tGY = unit(coords[,2] + justy * tGH, "native")

			tGH = unit(tGH + (0.1 + a$bg.padding) * yIn * lineIn * gp$cex, "native")
			tGW = unit(tGW + a$bg.padding * xIn * lineIn * gp$cex, "native")

			if (a$bg.border) {
				bordercol = a$bg.border.col
				lwd = a$bg.border.lwd
			} else {
				bordercol = NA
				lwd = 0
			}

			grobTextBGList = mapply(function(x, y, w, h, b, a, rot) {
				rect = rectGrob(x=x, y=y, width=w, height=h, gp=gpar(fill=b, alpha = a, col=bordercol, lwd = lwd))
				if (rot != 0) {
					.rectGrob2pathGrob(rect, rot, bbx)$poly
				} else {
					rect
				}
			}, tGX, tGY, tGW, tGH, bgcol, bgcol_alpha, angle, SIMPLIFY = FALSE, USE.NAMES = FALSE)
		} else {
			grobTextBGList = NULL
		}

		#if (a$auto.placement || a$remove_overlap) {
		# grobs to sf
		s = do.call(c,lapply(grobTextBGList, .grob2Poly))
		#}

		if (a$point.label) {
			get_rect_coords = function(polygon) {
				co = sf::st_coordinates(polygon)
				xr = range(co[,1])
				yr = range(co[,2])
				c(x = mean(xr),
				  y = mean(yr),
				  width = xr[2] - xr[1],
				  height = yr[2] - yr[1])
			}

			rect = do.call(rbind, lapply(s, get_rect_coords))

			res = pointLabel2(x = rect[,1], y = rect[,2], width = rect[,3], height = rect[,4], bbx = bbx, gap = yIn * lineIn * a$point.label.gap, method = a$point.label.method)

			sx = res$x - rect[,1]
			sy = res$y - rect[,2]

			grobTextList = mapply(function(grb, sxi, syi) {
				grb$x = grb$x + grid::unit(sxi, "native")
				grb$y = grb$y + grid::unit(syi, "native")
				grb
			}, grobTextList, sx, sy, SIMPLIFY = FALSE)


			grobTextBGList = mapply(function(grb, sxi, syi) {
				grb$x = grb$x + grid::unit(sxi, "native")
				grb$y = grb$y + grid::unit(syi, "native")
				grb
			}, grobTextBGList, sx, sy, SIMPLIFY = FALSE)


			if (with_shadow_halo) {
				grobTextShList = mapply(function(grb, sxi, syi) {
					grb$x = grb$x + grid::unit(sxi, "native")
					grb$y = grb$y + grid::unit(syi, "native")
					grb
				}, grobTextShList, sx, sy, SIMPLIFY = FALSE)
			}
		}

		if (a$remove_overlap) {
			im = sf::st_intersects(s, sparse = FALSE)
			sel = rep(TRUE, length(s))
			rs = rowSums(im)
			while(any(rs>1)) {
				id = which.max(rs)[1]
				sel[id] = FALSE
				im[id,] = FALSE
				im[, id] = FALSE
				rs = rowSums(im)
			}
		} else {
			sel = TRUE
		}

		if (!with_bg && a$remove_overlap) {
			grobTextBGList = NULL
		}



		grobTextAll = list(grobTextBGList[sel], grobTextShList[sel], grobTextList[sel])
		grobTextAll2 = grobTextAll[!vapply(grobTextAll, is.null, FUN.VALUE = logical(1))]

		grb = grid::grobTree(do.call(grid::gList, do.call(c, do.call(mapply, c(list(FUN = list, SIMPLIFY = FALSE, USE.NAMES = FALSE), grobTextAll2)))))

	} else {
		grobText = grid::textGrob(x = grid::unit(coords[,1], "native"), y = grid::unit(coords[,2], "native"), label = text, gp = gp, name = paste0("text_", id), rot = angle)
		grb = grid::grobTree(gList(grobText))
	}


	gts = get("gts", .TMAP_GRID)
	gt = gts[[facet_page]]

	gt_name = paste0("gt_facet_", rc_text)

	gt = grid::addGrob(gt, grb, gPath = grid::gPath(gt_name))



	gts[[facet_page]] = gt
	assign("gts", gts, envir = .TMAP_GRID)
	NULL
}

#' @export
#' @rdname tmapGridLeaflet
tmapGridDataPlot.tm_data_labels = function(a, shpTM, dt, gp, bbx, facet_row, facet_col, facet_page, id, pane, group, glid, o, ...) {
	NextMethod()
}



#' @export
#' @rdname tmapGridLeaflet
tmapGridDataPlot.tm_data_labels_highlighted = function(a, shpTM, dt, gp, bbx, facet_row, facet_col, facet_page, id, pane, group, glid, o, ...) {
	NextMethod()
}

tmapLeafletInit = function(o, return.asp = FALSE, vp, prx, lf = NULL, ...) {
	args = list(...)
	if (return.asp) return(1)

	per_page = rep(o$ncols * o$nrows, o$npages)
	k = o$ncols * o$nrows * o$npages
	if (o$n < k) {
		per_page[o$npages] = per_page[o$npages] - (k - o$n)
	}

	#proxy = !is.null(lf)
	proxy2 = (length(prx) > 0)

	# leaflet options

	if (o$crs_leaflet$crsClass == "L.CRS.Simple") {
		if (is.na(o$set_zoom_limits)[1]) o$set_zoom_limits[1] = -1000
	}
	leaflet_opts = do.call(leaflet::leafletOptions, c(list(crs=o$crs_leaflet), o$leaflet.options))

	if (!is.na(o$set_zoom_limits[1])) leaflet_opts$minZoom = o$set_zoom_limits[1]
	if (!is.na(o$set_zoom_limits[2])) leaflet_opts$maxZoom = o$set_zoom_limits[2]

	leaflet_opts$attributionControl = TRUE

	if (!.TMAP$proxy) {
		.TMAP_LEAFLET$layerIds = NULL
		.TMAP_LEAFLET$markerLayers = character(0)
	}


	lfs = lapply(per_page, function(p) {
		lapply(seq_len(p), function(i) {
			if (!.TMAP$proxy) {
				lf = leaflet::leaflet(options = leaflet_opts)
			}

			if (proxy2) {
				L2 = list()
				L = .TMAP_LEAFLET$layerIds
				Lnames = vapply(L, function(l) {
					l$name
				}, FUN.VALUE = character(1))
				Lids = lapply(L, function(l) {
					l$Lid
				})
				Ltypes = vapply(L, function(l) {
					l$type
				}, FUN.VALUE = character(1))

				Lstamps = vapply(L, function(l) {
					l$stamp
				}, FUN.VALUE = numeric(1))

				old = Lstamps != .TMAP$stamp

				Lremove = pane_name(vapply(prx, function(p) p$zindex, FUN.VALUE = numeric(1)))

				ids = which(Lnames %in% Lremove & Ltypes %in% c("polygons", "symbols", "text", "lines", "polygonsGL", "linesGL", "linesGL_hb", "symbolsGL", "cluster"))
				L2 = c(L2, L[ids])
#po(L, L2)

				#.TMAP_LEAFLET$layerIds = L
				.TMAP_LEAFLET$layerIds2 = L2
			}

			if (!.TMAP$in.shiny) {
				appendContent(lf, {
					tags$head(
						tags$style(HTML(paste(".leaflet-container {background:", o$bg.color, ";}", sep="")))
					)
				})
			} else {
				lf
			}
		})
	})

    #.TMAP_LEAFLET$layerIds = layerIds

	.TMAP_LEAFLET$lfs = lfs
	.TMAP_LEAFLET$nrow = o$nrows
	.TMAP_LEAFLET$ncol = o$ncols
	.TMAP_LEAFLET$leg_id = 1
	NULL
}

tmapLeafletAux = function(o, q) {
	lfs = .TMAP_LEAFLET$lfs


	isTMAP = substr(q$pane, 1, 4) == "tmap"
	isNEW = q$new

	lids = setdiff(q$lid[isTMAP & isNEW], .TMAP$pane_ids)

	groups_radio = unique(unlist(strsplit(q$group[q$group.control == "radio"], split = "__", fixed = TRUE)))
	groups_check = unique(unlist(strsplit(q$group[q$group.control == "check"], split = "__", fixed = TRUE)))


	# remove radio button when there is only one
	if (is.null(groups_radio) || length(groups_radio) == 1) groups_radio = character(0)
	if (is.null(groups_check)) groups_check = character(0)

	lfs = lapply(lfs, function(lfp) {
		lapply(lfp, function(lf) {
			if (length(lids)) for (lid in lids) lf = leaflet::addMapPane(lf, pane_name(lid), zIndex = lid)
			if (length(groups_radio) > 0L || length(groups_check) > 0L) {
				lf = leaflet::addLayersControl(lf, baseGroups = groups_radio, overlayGroups = groups_check, position = leaflet_pos(str2pos(o$control.position)), options = leaflet::layersControlOptions(collapsed = o$control.collapse))
			}
			lf
		})
	})


	.TMAP_LEAFLET$lfs = lfs
	NULL
}




view_set_bounds = function(lf, bbx, o) {
	if (!is.logical(o$set_bounds)) {
		lims = unname(o$set_bounds)
	} else {
		lims = unname(bbx)
	}
	if (!(identical(o$set_bounds, FALSE))) {
		lf = lf %>% leaflet::setMaxBounds(lims[1], lims[2], lims[3],lims[4])
	}

	if (is.na(o$set_view[1])) {

		# check if zoom is used from tm_basemap, and if so, use that
		zm = NA
		for (x in .TMAP_LEAFLET$tiles) {
			for (xi in x) {
				if (!is.na(xi$zoom)) {
					zm = xi$zoom
				}
			}
		}

		if (!is.na(zm)) {
			set_view = c(mean.default(lims[c(1,3)]), mean.default(lims[c(2,4)]), zm)
		} else if (!is.na(o$set_zoom_limits[2])) { # 2nd is checked to bypass (-1000, NA) used for simple CRS
			set_view = c(mean.default(lims[c(1,3)]), mean.default(lims[c(2,4)]), o$set_zoom_limits[1])
		} else {
			set_view = NULL
		}
	} else if (length(o$set_view) == 1) {
		set_view = c(mean.default(lims[c(1,3)]), mean.default(lims[c(2,4)]), o$set_view)
	} else {
		set_view = o$set_view
	}

	if (!is.null(set_view)) {
		names(set_view) = NULL
		lf = lf %>% setView(set_view[1], set_view[2], set_view[3])
	}

	if (!is.null(o$center)) lf = lf %>% addMarkers(o$center$lon, o$center$lat, label = o$center$query)

	lf
}

#' Create a style catalogue
#'
#' Create a style catalogue for each predefined tmap style. The result is a set of png images, one for each style.
#'
#' @param path path where the png images are stored
#' @param styles vector of styles function names (see [tmap_style()]) for which a preview is generated. By default, a preview is generated for all loaded styles.
#' @import grid
#' @rdname tmap_style_catalogue
#' @export
tmap_style_catalogue <- function(path="./tmap_style_previews", styles=NA) {
	# get all styles

	styles <- get_style_names()

	ns <- length(styles)

	if (!file.exists(path)) dir.create(path, recursive = TRUE)


	# load and process data
	pb <- txtProgressBar()

	tm1 = tm_shape(NLD_prov) +
		tm_polygons() +
		tm_title_out("Constant polygon fill")

	NLD_muni$letter = factor(substr(NLD_muni$name, 1, 1))
	tm2 = tm_shape(NLD_muni) +
		tm_polygons(lwd = .5, fill = "letter", fill.scale = tm_scale_categorical(n.max = 7)) +
		tm_shape(NLD_prov) +
		tm_borders(lwd = 2) +
		tm_title_out("Categorical color scale") +
		tm_compass() +
		tm_scalebar()


	tm3 = tm_shape(NLD_muni) +
		tm_polygons(lwd = 0.5, fill = "income_high") +
		tm_shape(NLD_prov) +
		tm_borders(lwd = 2) +
		tm_title_out("Sequential color scale")

	tm4 = tm_shape(NLD_muni) +
		tm_polygons(lwd = 0.5, fill = "urbanity", fill.scale = tm_scale_categorical(values = "div")) +
		tm_shape(NLD_prov) +
		tm_borders(lwd = 2) +
		tm_title_out("Diverging color scale")


	print_style <- function(style, i) {
		on.exit(dev.off())
		png(file.path(path, paste0(style, ".png")), width=1920, height=1920)
		grid.newpage()
		pushViewport(viewport(layout = grid.layout(2,2)))

		# set progress bar init
		pbi <- (i-1)/ns
		pbii <- function(i) pbi + (i/4) * (1/ns)

		print(tm1 + tm_style(style) + tm_layout(scale = 2),
			  vp = viewport(layout.pos.row = 1, layout.pos.col = 1))
		setTxtProgressBar(pb, pbii(1))
		print(tm2 + tm_style(style) + tm_layout(scale = 2),
			  vp = viewport(layout.pos.row = 1, layout.pos.col = 2))
		setTxtProgressBar(pb, pbii(2))
		print(tm3 + tm_style(style) + tm_layout(scale = 2),
			  vp = viewport(layout.pos.row = 2, layout.pos.col = 1))
		setTxtProgressBar(pb, pbii(3))
		print(tm4 + tm_style(style) + tm_layout(scale = 2),
			  vp = viewport(layout.pos.row = 2, layout.pos.col = 2))
		setTxtProgressBar(pb, pbii(4))
		upViewport()
	}

	mapply(print_style, styles, 1:ns)
	if (get("tmapOptions", envir = .TMAP)$show.messages) message("\nCatalogue created in ", normalizePath(path))
	invisible()
}

#' @export
#' @rdname tmap_style_catalogue
tmap_style_catalog <- tmap_style_catalogue

#' Group components
#'
#' @param group_id id of the component group. By default set to `""`, which will apply to all components. There are two other options. 1) To use the same (self-chosen) name that corresponds to the `group_id` argument of a component function, such as [tm_legend()] and [tm_title()]. 2) To specify one (or more) component function names, e.g. `"tm_legend"` or `c("tm_scalebar", "tm_compass")`.
#' @param position The position specification of the components in this group: an object created with `tm_pos_in()` or `tm_pos_out()`. Or, as a shortcut, a vector of two values, specifying the x and y coordinates. The first is `"left"`, `"center"` or `"right"` (or upper case, meaning tighter to the map frame), the second `"top"`, `"center"` or `"bottom"`. Numeric values are also supported, where 0, 0 means left bottom and 1, 1 right top. See also \href{https://r-tmap.github.io/tmap/articles/adv_positions}{vignette about positioning}.
#' @param stack stacking `"horizontal"` or `"vertical"`
#' @param frame_combine put frame around all components that are drawn on the same location. Whether a frame is drawn is still decided by the `frame` argument of the  'main' (first) component.
#' @param equalize in case `frame_combine` is `FALSE`, should the separate frames be equalized, i.e. have the same width (when stacked vertically) or height (when stacked horizontally)?
#' @param resize_as_group in case a component if rescaled because of the limited space, rescale the other components proportionally?
#' @param stack_margin Margin between components
#' @param offset Offset margin between frame and the components block
#' @param frame Should a frame be drawn? By default `TRUE` for legends, charts and insets, and `FALSE` otherwise.
#' @param frame.color frame color
#' @param frame.alpha frame alpha transparancy
#' @param frame.lwd frame line width
#' @param frame.r Radius of the rounded frame corners. 0 means no rounding.
#' @param bg Background color the components block. Is usually set in each component function, but if specified here, it will overwrite them.
#' @param bg.color Background color the components block. Is usually set in each component function, but if specified here, it will overwrite them.
#' @param bg.alpha Background alpha transparency of the components block. Is usually set in each component function, but if specified here, it will overwrite them.
#' @return A [`tmap-element`]
#' @export
tm_components = function(
		group_id = "",
		position,
		stack,     # was 'stack' in each tm_legend or tm_<comp> function
		frame_combine,
		equalize,
		resize_as_group,        # from tm_legend/tm_<comp>.
		stack_margin,     # margin between components
		offset,            # offset margin
		frame ,
		frame.color,
		frame.alpha,
		frame.lwd,
		frame.r,
		bg,
		bg.color,
		bg.alpha) {
	args = lapply(as.list(rlang::call_match()[-1]), eval, envir = parent.frame())
	args$called = names(args)
	optname = paste0("component_", paste(group_id, collapse = "^"))
	x = structure(list(args), names= optname)
	do.call(tm_options, x)
}

#' @exportS3Method
renderTmapGS.Leaflet <- function(x, expr, env, ...) {
	expr <- bquote(getFromNamespace("print.tmap", "tmap")(.(expr), in.shiny = TRUE))
	htmlwidgets::shinyRenderWidget(expr, leaflet::leafletOutput, env, quoted = TRUE)
}

#' @exportS3Method
tmapOutputGS.Leaflet <- function(x, outputId, width, height, ...) {
	leaflet::leafletOutput(outputId = outputId, width = width, height = height)
}

#' @exportS3Method
tmapProxyGS.Leaflet <- function(x, mapId, session, tmobj, ...) {
	print.tmap(tmobj,
			   lf = leaflet::leafletProxy(mapId, session),
			   show = FALSE, in.shiny = TRUE, proxy = TRUE)
}

#' @export
#' @rdname tmapGridLeaflet
tmapLeafletDataPlot.tm_data_polygons = function(a, shpTM, dt, pdt, popup.format, hdt, idt, gp, bbx, facet_row, facet_col, facet_page, id, pane, group, glid, o, ...) {
	lf = get_lf(facet_row, facet_col, facet_page)

	rc_text = frc(facet_row, facet_col)

	res = select_sf(shpTM, dt)
	shp = res$shp
	if (o$crs_leaflet$crsClass  == "L.CRS.Simple") {
		shp = sf::st_set_crs(shp, NA)
	}

	dt = res$dt

	idt_null = is.null(idt)

	if (!idt_null) {
		idt = idt$id[match(dt$tmapID__, idt$tmapID__)]
	} else {
		idt = sprintf("%07d", dt$tmapID__)
	}
	if (!is.null(hdt)) {
		hdt = hdt$hover[match(dt$tmapID__, hdt$tmapID__)]
		hdt = lapply(hdt, htmltools::HTML, FUN.VALUE = character(1))
	}

	if (is.null(pdt)) {
		popups = NULL
	} else {
		mtch = match(dt$tmapID__, pdt$tmapID__)
		pdt = pdt[mtch][, tmapID__ := NULL]

		if (idt_null && !is.null(hdt)) {
			popups = view_format_popups(id = hdt, titles = names(pdt), values = pdt, format = popup.format)
		} else if (idt_null && is.null(hdt)) {
			popups = view_format_popups(id = NULL, titles = names(pdt), values = pdt, format = popup.format)
		} else {
			popups = view_format_popups(id = idt, titles = names(pdt), values = pdt, format = popup.format)
		}
	}


	gp = impute_gp(gp, dt)
	gp = rescale_gp(gp, o$scale_down)
	gp = gp_to_lpar(gp, mfun = "Polygons", rename_prop = FALSE)

	interactive = (!is.null(pdt) || !is.null(hdt))

	opt = leaflet::pathOptions(interactive = interactive, pane = pane)

	o$use_WebGL = impute_webgl(o$use_WebGL, dt, supported = c("fill", "col"), checkif = list(lty = "solid"), type = "polygons", hover = !is.null(hdt), popup = !is.null(pdt), crs_class = o$crs_leaflet$crsClass)

	if (o$use_WebGL) {
		idt = rep({if (is.null(idt))dt$tmapID__[1] else idt[1]}, 2) |>
			submit_labels("polygonsGL", pane, group)

		shp2 = sf::st_sf(id = seq_along(shp), geom = shp)
		shp3 = sf_expand(shp2)

		shp3lines = suppressWarnings(sf::st_cast(shp3, "LINESTRING"))
		gp3 = lapply(gp, function(gpi) {if (length(gpi) == 1) gpi else gpi[shp3$split__id]})
		popups2 = popups[shp3$split__id]

		# opacity channel from fill (e.g. "#FF000099") is ignored by addGlPolygons
		fill_alpha = split_alpha_channel(gp3$fill[1], alpha = gp3$fill_alpha[1])$opacity

		lf |>
			leafgl::addGlPolygons(data = shp3, layerId = idt[1], label = hdt,
								  fillColor = gp3$fill, fillOpacity = fill_alpha,
								  group = group, pane = pane, popup = popups2) %>%
			{if (gp3$lwd[1]!=0 && gp3$col[1] != "#00000000") leafgl::addGlPolylines(., data = shp3lines, color = gp3$col, opacity = gp3$col_alpha[1], weight = gp3$lwd[1]/4, pane = pane, group = group, layerId = idt[2]) else .} %>%
			assign_lf(facet_row, facet_col, facet_page)
	} else {
		idt = (if (is.null(idt))dt$tmapID__ else idt) |>
			submit_labels("polygons", pane, group)

		lf %>%
			leaflet::addPolygons(data = shp, layerId = idt, label = hdt, color = gp$col, opacity = gp$col_alpha, fillColor = gp$fill, fillOpacity = gp$fill_alpha, weight = gp$lwd, options = opt, group = group, dashArray = gp$lty, popup = popups) %>%
			assign_lf(facet_row, facet_col, facet_page)
	}
	NULL
}


#' @export
#' @rdname tmapGridLeaflet
tmapLeafletDataPlot.tm_data_fill = function(a, shpTM, dt, pdt, popup.format, hdt, idt, gp, bbx, facet_row, facet_col, facet_page, id, pane, group, glid, o, ...) {
	NextMethod()
}

#' @export
#' @rdname tmapGridLeaflet
tmapLeafletDataPlot.tm_data_borders = function(a, shpTM, dt, pdt, popup.format, hdt, idt, gp, bbx, facet_row, facet_col, facet_page, id, pane, group, glid, o, ...) {
	NextMethod()
}

tmapLeafletRun = function(o, q, show, knit, knit_opts, args) {
	lfs = get("lfs", envir = .TMAP_LEAFLET)

	if (show && o$show_gif_ani) {
		cli::cli_alert("{.field view mode} Animations are not implemented in view mode, so they are shown as facets")
	}

	# switchable zoom levels
	zids = which(!is.na(q$group.zoom_levels))
	if (length(zids)) {
		lfs = lapply(lfs, function(lfp) {
			lapply(lfp, function(lf) {
				for (zid in zids) lf = leaflet::groupOptions(lf, group = q$group[zid], zoomLevels = q$group.zoom_levels[[zid]])
				lf
			})
		})
	}

	lfs2 = lapply(lfs, function(lfsi) {
		x = if (o$nrows == 1 && o$ncols == 1) {
			lf = lfsi[[1]]
			# proxy remove
			if (!is.null(.TMAP_LEAFLET$layerIds2)) {
				L = .TMAP_LEAFLET$layerIds
				#po(L, .TMAP_LEAFLET$layerIds2)
				for (L2 in .TMAP_LEAFLET$layerIds2) {
					if (L2$type == "raster") {
						lf = leaflet::removeImage(lf, L2$Lid)
					} else if (L2$type %in% c("symbols", "text")) {
						lf = leaflet::removeMarker(lf, L2$Lid)
					} else if (L2$type == "symbolsGL") {
						lf = leafgl::removeGlPoints(lf, L2$Lid[1])
					} else if (L2$type == "linesGL") {
						lf = leafgl::removeGlPolylines(lf, L2$Lid[1])
					} else if (L2$type == "linesGL_hb") {
						lf = lf |>
							leafgl::removeGlPolylines(L2$Lid[1]) |>
							leaflet::removeShape(L2$Lid[-1])
					} else if (L2$type == "polygonsGL") {
						lf = lf |>
							leafgl::removeGlPolygons(L2$Lid[1]) |>
							leafgl::removeGlPolylines(L2$Lid[2])
					} else if (L2$type == "cluster") {
						lf = leaflet::removeMarkerCluster(lf, L2$Lid[1])
					} else {
						lf = leaflet::removeShape(lf, L2$Lid)
					}
					L = lapply(L, function(x) {
						if (any(L2$Lid %in% x$Lid)) NULL else x
					})
					L = L[!vapply(L, is.null, FUN.VALUE = logical(1))]
				}
				# to do: update group and type arguments
				.TMAP_LEAFLET$layerIds = L
				.TMAP_LEAFLET$layerIds2 = NULL
			}
			lf
		} else {
			fc = o$free.coords
			sync = if (identical(o$sync, FALSE)) {
				"none"
			} else if (identical(o$sync, TRUE) || all(!fc)) {
				"all"
			} else if (all(fc)) {
				"none"
			} else if (fc[1]) {
				asplit(matrix(1:(o$nrows*o$ncols), ncol = o$ncols, byrow = TRUE), 1)
			} else {
				asplit(matrix(1:(o$nrows*o$ncols), ncol = o$ncols, byrow = TRUE), 2)
			}
			marg = paste0(o$between_margin, "em")

			#print(do.call(leafsync::latticeView, c(lfsi, list(ncol = o$ncols, sync = sync, sync.cursor = all(!fc), no.initial.sync = FALSE, between = list(x = marg, y = marg)))))
			do.call(leafsync::latticeView, c(lfsi, list(ncol = o$ncols, sync = sync, sync.cursor = all(!fc), no.initial.sync = FALSE)))
		}


		if (length(.TMAP_LEAFLET$markerLayers)) {
			style_markers = paste(vapply(unique(.TMAP_LEAFLET$markerLayers), function(pane) {
				paste0(".leaflet-container .leaflet-", pane, "-pane img {
				max-width: none !important;
				max-height: none !important;
				width: auto;
				padding: 0;
				}\n")}, FUN.VALUE = character(1)), collapse = "\n")
		} else {

			style_markers = NULL
		}

		if (o$pc$sepia_intensity != 0) {
			col = process_color("#ffffff", sepia_intensity = o$pc$sepia_intensity)
			style_sepia = paste0(".leaflet-control-layers {background: ", col, ";}
				.leaflet-control-zoom-in {background: ", col, " !important;}
				.leaflet-control-zoom-out {background: ", col, " !important;}\n"
			)
		} else {
			style_sepia = NULL
		}

		if (is.null(style_markers) && is.null(style_sepia)) {
			x
		} else {
			if (.TMAP$in.shiny) {

				x
			} else {
				htmlwidgets::prependContent(x, htmltools::tags$style(paste0(style_markers, style_sepia)))
			}
		}
	})

	if (length(lfs2) == 1) lfs2 = lfs2[[1]]
	if (show && !knit && !.TMAP$in.shiny) {

		# borrowed from mapview
		viewer = getOption("viewer")
		ide = get_ide()
		if (!is.null(viewer)) {
			viewerFunc = function(url) {
				paneHeight = lfs2$sizingPolicy$viewer$paneHeight
				if (identical(paneHeight, "maximize")) {
					paneHeight = -1
				}
				if (ide == "vscode") {
					# VSCode's viewer can't ignore cross-origin requests. Need to serve the
					# map so assests can be read, e.g. .fgb files.
					server <- servr::httd(
						dir = get_url_dir(url),
						verbose = FALSE,
						browser = FALSE
					)
					url <- server$url

				}
				viewer(url, height = paneHeight)
			}
		} else {
			viewerFunc = function(url) {
				dir = get_url_dir(url)
				switch(ide,
					   "rstudio" = if (o$use_browser) {
					   	fl = file.path(dir, "index.html")
					   	utils::browseURL(fl)
					   } else {
					   	servr::httd(
					   		dir = dir,
					   		verbose = FALSE
					   	)
					   },
					   "vscode" = servr::httd(
					   	dir = dir,
					   	verbose = FALSE
					   ),
					   # default
					   servr::httd(
					   	dir = dir,
					   	verbose = FALSE
					   )
				)
			}
		}

		htmltools::html_print(
			htmltools::as.tags(lfs2, standalone = TRUE)
			, viewer = if (interactive()) viewerFunc
		)
	}
	lfs2
}

#' Shape (spatial object) specification
#'
#' Specify a shape, which is a spatial object from one of these spatial object
#' class packages: [`sf`][`sf::sf`], [`stars`][stars::st_as_stars()], or `terra`.
#'
#' The map projection (`crs`) determines in which coordinate system the spatial object is processed and plotted.
#' See \href{https://r-tmap.github.io/tmap/articles/foundations_crs}{vignette about CRS}. The `crs` can be specified in two places: 1) `tm_shape()` and `tm_crs()`.
#' In both cases, the map is plotted into the specified `crs`. The difference is that in the first option, the `crs` is also taken into account in spatial transformation functions, such as the calculation of centroids and cartograms. In the second option, the `crs` is only used in the plotting phase.
#'
#' The automatic crs recommendation (which is still work-in-progress) is the following:
#'
#' | **Property**        | **Recommendation** |
#' | ------          | ----------- |
#' | `global` (for world maps)		     | A pseudocylindrical projection tmap option `crs_global`, by default `"eqearth` (Equal Earth). See \url{https://r-tmap.github.io/tmap/articles/41_advanced_crs.html} for more options|
#' | `area` (equal area)		     | Lambert Azimuthal Equal Area (`laea`) |
#' | `distance`	(equidistant) | Azimuthal Equidistant (`aeqd`) |
#' | `shape`	(conformal) | Stereographic (`stere`) |
#'
#' For further info about the available "generic" projects see:
#' for utm: \url{https://proj.org/en/9.4/operations/projections/utm.html}
#' for laea: \url{https://proj.org/en/9.4/operations/projections/laea.html}
#' for aeqd: \url{https://proj.org/en/9.4/operations/projections/aeqd.html}
#' for pconic: \url{https://proj.org/en/9.4/operations/projections/pconic.html}
#' for eqdc: \url{https://proj.org/en/9.4/operations/projections/eqdc.html}
#'
#' @param shp Spatial data object. Typically an object from \pkg{sf}, \pkg{terra}, or \pkg{stars}.
#'   Additional spatial data types can be supported via extension packages, such as
#'   \pkg{tmap.networks} and \pkg{tmap.sources} (experimental). These may include, for example,
#'   remote or streaming data sources.
#' @param bbox Bounding box of the map. Only used when `shp` is the main shape (see `is.main`).
#'   Three options are supported:
#'   \itemize{
#'     \item a \code{\link[sf:st_bbox]{sf::st_bbox()}} object,
#'     \item a character string specifying a location, passed to \code{\link[tmaptools:geocode_OSM]{tmaptools::geocode_OSM()}},
#'     \item \code{"FULL"}, which represents the whole earth. This option ensures that reprojection
#'     retains the full global extent, unlike a regular bounding box.
#'   }
#' @param crs Map projection (CRS). Can be set to an `crs` object (see [sf::st_crs()]), a proj4string, an EPSG number, the value `"auto"` (automatic crs recommendation), or one the the following generic projections: `c("laea", "aeqd", "utm", "pconic", "eqdc", "stere")`. See details.
#' @param is.main Is `shp` the main shape, which determines the crs and
#'   bounding box of the map? By default, `TRUE` if it is the first `tm_shape` call
#' @param name of the spatial object
#' @param layer Name of the layer to use. This is primarily relevant for multi-layer or remote
#'   data sources (e.g. PMTiles or vector tiles), where multiple layers may be available.
#' @param unit Unit of the coordinates
#' @param filter Optional filter expression used to subset features. The exact syntax depends
#'   on the data source. For in-memory objects (e.g. \pkg{sf}), this is typically evaluated in R,
#'   whereas for remote sources it may be translated to a query and executed on the server side.
#' @param ... passed on to \code{\link[tmaptools:bb]{bb}} (e.g. \code{ext} can be used to enlarge or shrink a bounding box)
#' @import tmaptools
#' @importFrom sf st_geometry st_sf st_as_sf st_transform st_crs
#' @importFrom stars st_dimensions st_get_dimension_values st_as_stars st_set_dimensions st_warp
#' @import units
#' @import grid
#' @import cols4all
#' @import classInt
#' @import htmltools
#' @import htmlwidgets
#' @import leaflet
#' @import leafsync
#' @import leafem
#' @import leaflegend
#' @import stats
#' @importFrom methods as
#' @importFrom utils head tail download.file txtProgressBar str data capture.output setTxtProgressBar methods getS3method
#' @importFrom rlang missing_arg expr
#' @importFrom s2 s2_buffer_cells as_s2_geography s2_intersects
#' @importFrom grDevices col2rgb colorRampPalette colors dev.off dev.size png rgb
#' @note as of tmap 4.0, simplify has been removed. Please use [tmaptools::simplify_shape()] instead
#' @example ./examples/tm_shape.R
#' @seealso \href{https://r-tmap.github.io/tmap/articles/foundations_crs}{vignette about CRS}
#' @export
tm_shape = function(shp = NULL,
					bbox = NULL,
					crs = NULL,
					is.main = NA,
					layer = NULL,
					name = NULL,
					unit = NULL,
					filter = NULL,
					...) {

	args_called = names(rlang::call_match()[-1])
	args = lapply(as.list(rlang::call_match(defaults = TRUE)[-1]), eval, envir = parent.frame())

	if ("projection" %in% args_called) {
		v3_instead_message(arg_old = "projection", arg_new = "crs", fun = "tm_shape")
		crs = args$projection
	}
	if (any(c("crs", "projection") %in% args_called) && is.na(is.main)) {
		is.main = TRUE
	}



	bbox_list = c(list(x = bbox), args[intersect(args_called, c("ext", "cx", "cy", "width", "height", "xlim", "ylim", "relative", "asp.limit"))])

	shp_called = deparse(substitute(shp))[1]



	# rivers -> World_rivers
	if (shp_called %in% c("rivers", "c(735, 320, 325, 392, 524, 450, 1459, 135, 465, 600, 330, 336, ") && # last one to catch qtm calls
		is.numeric(shp) && length(shp) == 141) {
		data(World_rivers, envir = environment())
		shp = World_rivers
		shp_called = "World_rivers"
		v3_rivers()
	}

	if ("simplify" %in% args_called) {
		sim_args = args[intersect(args_called, c("simplify", "keep.units", "keep.subunits"))]
		v3_simplify(shp_called, sim_args)
	}


	shp_name = ifelse(is.null(name) == TRUE, shp_called, name)


	if (missing(shp)) {
		do.call(tm_options, args[intersect(args_called, c("bbox", "crs", "set_bounds", "set_view", "set_zoom_limits"))])
	} else {
		tm_element_list(tm_element(shp = shp,
								   is.main = is.main,
								   crs = crs,
								   bbox = bbox_list,
								   unit = unit,
								   filter = filter,
								   layer = layer,
								   shp_name = shp_name,
								   subclass = "tm_shape"))
	}
}

#' tmap function to specify labels
#'
#' tmap function to specify labels used in the scale functions, e.g. via the argument `label.format` in [tm_scale_intervals()].
#'
#' @param fun Function to specify the labels. It should take a numeric vector, and should return a character vector of the same size. By default it is not specified. If specified, the list items \code{scientific}, \code{format}, and \code{digits} (see below) are not used.
#' @param scientific Should the labels be formatted scientifically? If so, square brackets are used, and the \code{format} of the numbers is \code{"g"}. Otherwise, \code{format="f"}, and \code{text.separator}, \code{text.less.than}, and \code{text.or.more} are used. Also, the numbers are automatically  rounded to millions or billions if applicable. By default, \code{FALSE}
#' @param format By default, \code{"f"}, i.e. the standard notation \code{xxx.xxx}, is used. If \code{scientific = TRUE} then \code{"g"}, which means that numbers are formatted scientifically, i.e. \code{n.dddE+nn} if needed to save space.
#' @param digits Number of digits after the decimal point if \code{format="f"}, and the number of significant digits otherwise. By default \code{NA}, meaning as many as needed to have distinct numbers
#' @param interval.disjoint In case of intervals (see [tm_scale_intervals()]), should the intervals appear disjoint, e.g. \code{0 to 999}, \code{1000 - 1999}, \code{2000 - 2999} (\code{TRUE}, default), or not: \code{0 - 1000}, \code{1000 - 2000}, \code{2000- 3000}.
#' @param big.num.abbr Vector that defines whether and which abbrevations are used for large numbers. It is a named numeric vector, where the name indicated the abbreviation, and the number the magnitude (in terms on numbers of zero). Numbers are only abbrevation when they are large enough. Set it to \code{NA} to disable abbrevations.  The default is \code{c("mln" = 6, "bln" = 9)}. For layers where \code{style} is set to \code{log10} or \code{log10_pretty}, the default is \code{NA}.
#' @param prefix Prefix of each number
#' @param suffix Suffix of each number
#' @param text.separator Character string to use to separate numbers in an interval legend (default: \code{"to"}).
#' @param text.less.than Character value(s) to use for 'less than'. Default \code{"Less than"}. When a character vector of length 2 is specified, one for each word, these words are aligned when \code{text.to.columns = TRUE}
#' @param text.less.than_as.prefix Should \code{text.less.than} be used as prefix?
#' @param text.or.more Character value(s) to use to  'or more'. Default is \code{"or more"}. When a character vector of length 2 is specified, one for each word, these words are aligned when \code{text.to.columns = TRUE}
#' @param text.or.more_as.prefix Should \code{text.or.more} be used as prefix?
#' @param text.align Not implemented in v4 (yet). Value that determines how the numbers are aligned, \code{"left"}, \code{"center"} or \code{"right"}. By default \code{"left"}.
#' @param text.to.columns Not implemented in v4 (yet). Logical that determines whether the text is aligned to three columns (from, text.separator, to). By default \code{FALSE}.
#' @param html.escape Logical that determins whther HTML code is escaped in the popups in view mode. By default \code{TRUE}. If set to \code{FALSE} HTML code can be added, e.g. to added white space via \code{&nbsp;}.
#' @param ... arguments passed on to \code{\link[base:formatC]{formatC}}
#' @seealso `r .doc_see_also_legend(advanced = TRUE)`
#' @return list with formatting options
#' @export
tm_label_format = function(fun,
						   scientific,
						   format,
						   digits,
						   interval.disjoint,
						   big.num.abbr,
						   prefix,
						   suffix,
						   text.separator,
						   text.less.than,
						   text.less.than_as.prefix,
						   text.or.more,
						   text.or.more_as.prefix,
						   text.align,
						   text.to.columns,
						   html.escape,
						   ...) {

	args = lapply(as.list(rlang::call_match(dots_expand = TRUE)[-1]), eval, envir = parent.frame())
	args$called = names(args)

	structure(args, class = c("tmapLabelFormat", "list"))
}

#' @export
#' @keywords internal
#' @rdname tmapGridLeaflet
tmapLeafletAuxPrepare = function(a, bs, id, o) {
	UseMethod("tmapLeafletAuxPrepare")
}


#' @export
#' @keywords internal
#' @rdname tmapGridLeaflet
tmapLeafletAuxPlot = function(a, bi, bbx, facet_row, facet_col, facet_page, id, pane, group, o) {
	UseMethod("tmapLeafletAuxPlot")
}

#' @export
#' @keywords internal
#' @rdname tmapGridLeaflet
tmapLeafletAuxPrepare.tm_aux_basemap = function(a, bs, id, o) {
	tmapLeafletAuxPrepare.tm_aux_tiles(a, bs, id, o)
}

#' @export
#' @keywords internal
#' @rdname tmapGridLeaflet
tmapLeafletAuxPrepare.tm_aux_tiles = function(a, bs, id, o) {
	nms = names(a$server)
	a$server = unname(a$server)
	tiles = lapply(1L:length(bs), function(i) a)
	.TMAP_LEAFLET$tiles[[id]] = tiles
	if (!is.null(nms)) {
		paste0(nms, collapse = "__")
	} else {
		paste0(a$server, collapse = "__")
	}
}

#' @export
#' @keywords internal
#' @rdname tmapGridLeaflet
tmapLeafletAuxPlot.tm_aux_basemap = function(a, bi, bbx, facet_row, facet_col, facet_page, id, pane, group, o) {
	tmapLeafletAuxPlot.tm_aux_tiles(a, bi, bbx, facet_row, facet_col, facet_page, id, pane, group, o)
}

#' @export
#' @keywords internal
#' @rdname tmapGridLeaflet
tmapLeafletAuxPlot.tm_aux_tiles = function(a, bi, bbx, facet_row, facet_col, facet_page, id, pane, group, o) {

	lf = get_lf(facet_row, facet_col, facet_page)

	rc_text = frc(facet_row, facet_col)


	tiles = .TMAP_LEAFLET$tiles[[id]][[bi]]

	if (o$credits.defined) {
		opt = leaflet::tileOptions(attribution = "", maxNativeZoom = tiles$max.native.zoom, pane = pane)
	} else {
		opt = leaflet::tileOptions(maxNativeZoom = tiles$max.native.zoom, pane = pane)
	}
	if (!is.na(o$set_zoom_limits[2])) opt$maxZoom = o$set_zoom_limits[2]

	k = length(tiles$server)

	groups = rep(strsplit(group, "__", fixed = TRUE)[[1]], length.out = k)

	for (i in 1L:k) {
		serv = tiles$server[i]
		if (serv != "") {
			if ((substr(serv, 1, 4) == "http")) {
				lf = leaflet::addTiles(lf, urlTemplate = serv, group = groups[i], options = opt)
			} else {
				# Invalid provider: warn and fall back to the default view basemap
				# rather than adding a broken (blank) provider layer, matching the
				# fallback behaviour of the maplibre/mapbox modes.
				if (!(serv %in% tmap_providers("view"))) {
					eq = basemap_equivalent(serv, "view")
					if (!is.na(eq)) {
						message_basemaps_equivalent(serv, "view", eq)
						serv = eq
					} else {
						message_basemaps_invalid_provider(serv, "view", o$basemap.server[1])
						serv = o$basemap.server[1]
					}
				}
				lf = leaflet::addProviderTiles(lf, provider = serv, group = groups[i], options = opt)
			}
		}
	}

	assign_lf(lf, facet_row, facet_col, facet_page)
	NULL
}

#' @export
#' @keywords internal
#' @rdname tmapGridLeaflet
tmapLeafletAuxPrepare.tm_aux_graticules = function(a, bs, id, o) {
	tmapLeafletAuxPrepare.tm_aux_grid(a, bs, id, o)
}

#' @export
#' @keywords internal
#' @rdname tmapGridLeaflet
tmapLeafletAuxPrepare.tm_aux_grid = function(a, bs, id, o) {

	grid_intervals = vapply(bs, function(b) {
		# implementation similarish as plot mode but needs polishing
		dx = b[3] - b[1]
		dy = b[4] - b[2]
		# if not specified, aim for 15 lines in total
		n.x = if (!is.na(a$n.x)) a$n.x else if (is.na(a$x)) 7.5 else length(a$x)
		n.y = if (!is.na(a$n.y)) a$n.y else if (is.na(a$y)) 7.5 else length(a$x)

		x = pretty30(b[c(1,3)], n=n.x, longlat = !is.na(a$crs) && sf::st_is_longlat(a$crs))
		y = pretty30(b[c(2,4)], n=n.y, longlat = !is.na(a$crs) && sf::st_is_longlat(a$crs))

		max((x[2] - x[1]), (y[2]-y[1]))
	}, FUN.VALUE = numeric(1))
	assign("grid_intervals", grid_intervals, envir = .TMAP_LEAFLET)

	return("grid")
}

#' @export
#' @keywords internal
#' @rdname tmapGridLeaflet
tmapLeafletAuxPlot.tm_aux_graticules = function(a, bi, bbx, facet_row, facet_col, facet_page, id, pane, group, o) {
	tmapLeafletAuxPlot.tm_aux_grid(a, bi, bbx, facet_row, facet_col, facet_page, id, pane, group, o)
}

#' @export
#' @keywords internal
#' @rdname tmapGridLeaflet
tmapLeafletAuxPlot.tm_aux_grid = function(a, bi, bbx, facet_row, facet_col, facet_page, id, pane, group, o) {
	lf = get_lf(facet_row, facet_col, facet_page)
	rc_text = frc(facet_row, facet_col)

	grid_intervals = get("grid_intervals", envir = .TMAP_LEAFLET)

	lf = leaflet::addGraticule(lf, interval = grid_intervals[bi], options = pathOptions(pane = pane))


	assign_lf(lf, facet_row, facet_col, facet_page)
	NULL
}

tmapLeafletGridXLab = function(bi, bbx, facet_row, facet_col, facet_page, o) {
	NULL
}

tmapLeafletGridYLab = function(bi, bbx, facet_row, facet_col, facet_page, o) {
	NULL
}

#' Map layer: raster
#'
#' Map layer that draws rasters. Supported visual variable is: `col` (the  color).
#'
#' @param col,col.scale,col.legend,col.chart,col.free  `r .doc_vv("col")`
#' @param col_alpha,col_alpha.scale,col_alpha.legend,col_alpha.chart,col_alpha.free  `r .doc_vv("col_alpha")`
#' @inheritParams tm_polygons
#' @param options options passed on to the corresponding `opt_<layer_function>` function
#' @param ... to catch deprecated arguments from version < 4.0
#' @example ./examples/tm_raster.R
#' @inherit tm_polygons details
#' @export
tm_raster = function(col = tm_vars(),
					 col.scale = tm_scale(),
					 col.legend = tm_legend(),
					 col.chart = tm_chart_none(),
					 col.free = NA,
					 col_alpha = tm_const(),
					 col_alpha.scale = tm_scale(),
					 col_alpha.legend = tm_legend(),
					 col_alpha.chart = tm_chart_none(),
					 col_alpha.free = NA,
					 zindex = NA,
					 group = NA,
					 group.control = "check",
					 options = opt_tm_raster(),
					 ...) {


	args_called = names(rlang::call_match()[-1])
	args = list(...)

	layer_fun = if ("called_from" %in% names(args)) {
		args$called_from
	} else {
		"tm_raster"
	}

	if (any(v3_only("tm_raster") %in% names(args))) {


		v3_start_message()
		if (!("style" %in% names(args))) {
			if (!"breaks" %in% names(args)) {
				style = "pretty"
			} else {
				style = "fixed"
			}
		} else {
			style = args$style
		}

		v3_list_init()
		if (length(style) > 1) {
			style = style[1]
			.TMAP$v3_list$mult = TRUE
		}
		# v3 visual variable: col
		col.scale.args = c(list(n = v3_impute(args, "n", 5),
								style = style,
								style.args = v3_impute(args, "style.args", list())),
							if (style %in% c("cont", "log10")) {
								c({
									if (!is.null(args$breaks)) {
										if (length(args$breaks) > 2) {
											list(ticks = v3_impute(args, "breaks", NULL, "ticks"))
										} else {
											list(ticks = v3_impute(args, "breaks", NULL, "limits"))
										}
									} else {
										list()
									}
								},
								list(outliers.trunc = c(TRUE, FALSE)))
							} else {
								list(breaks = v3_impute(args, "breaks", NULL),
									 interval.closure = v3_impute(args, "interval.closure", "left"),
									 drop.levels = v3_impute(args, "drop.levels", FALSE))
							},
							list(midpoint = v3_impute(args, "midpoint", NULL),
								 as.count = v3_impute(args, "as.count", NA),
								 values = v3_impute(args, "palette", NA, "values"),
								 values.repeat = !v3_impute(args, "stretch.palette", TRUE, "values.repeat"),
								 values.range = v3_impute(args, "contrast", NA, "values.range"),
								 values.scale = 1,
								 value.na = v3_impute(args, "colorNA", NA, "value.na"),
								 value.null = v3_impute(args, "colorNULL", NA, "value.null"),
								 value.neutral = NA,
								 labels = v3_impute(args, "labels", NULL),
								 label.na = v3_impute(args, "textNA", "Missing", "label.na"),
								 label.null = NA,
								 label.format = v3_impute(args, "legend.format", list(), "label.format")))
		col.scale.args$fun_pref = if (style == "cat") {
			"categorical"
		} else if (style %in% c("fixed", "sd", "equal", "pretty", "quantile",
								"kmeans", "hclust", "bclust", "fisher", "jenks",
								"dpih", "headtails", "log10_pretty")) {
			"intervals"
		} else if (style == "cont") {
			"continuous"
		} else if (style == "log10") {
			"continuous_log"
		} else if (style == "order") {
			"rank"
		} else {
			stop("unknown style")
		}
		if ("style" %in% names(args)) {
			v3_tm_scale_instead_of_style(style, scale_fun = col.scale.args$fun_pref, vv = "col", layer_fun = layer_fun, arg_list = v3_list_get())
		} else {
			v3_tm_scale(scale_fun = "", vv = "col", layer_fun = layer_fun, arg_list = v3_list_get())
		}
		col.scale = do.call("tm_scale", args = col.scale.args)

		if ("alpha" %in% names(args)) {
			col_alpha = args$alpha
			v3_message_col_alpha(layer_fun = layer_fun, orig = "alpha")

		}

		v3_list_init()
		if ("legend.show" %in% names(args) && !args$legend.show) {
			v3_tm_legend_hide(layer_fun, arg = "legend.show", vv = "col")
			lwd.legend = tm_legend_hide()
		} else {
			col.legend.args = alist(title = v3_impute(args, "title", NA, "title"),
								   na.show = v3_impute(args, "showNA", NA),
								   format = v3_impute(args, "legend.format", list(), "format"),
								   orientation = ifelse(v3_impute(args, "legend.is.portrait", TRUE, "orientation"), "portrait", "landscape"),
								   reverse = v3_impute(args, "legend.reverse", FALSE, "reverse"))
			col.legend = do.call("tm_legend", col.legend.args)
			v3_tm_legend(fun = layer_fun, vv = "col", arg_list = v3_list_get())
		}

		if ("legend.hist" %in% names(args) && args$legend.hist) {
			col.chart = tm_chart_histogram()
			v3_tm_chart_hist(layer_fun = layer_fun, vv = "col", arg = "legend.hist")

			# to do: histogram title
		}
	}

	# unused arguments: typos?
	unused = setdiff(names(args), c(v3_only("tm_raster"), "called_from"))

	if (length(unused)) {
		message_layer_unused_args(layer_fun, unused)
	}



	# needed for color maps without categories (then tm_scale_categorical is used without legend, unless called)
	col.legend$legend_called = "col.legend" %in% args_called


	tm_element_list(tm_element(
		layer = "raster",
		trans.fun = tmapTransRaster,
		trans.aes = list(),
		trans.args = options$trans.args,
		trans.apply_to = "this",
		mapping.aes = list(col = tmapScale(aes = "col",
										   value = col,
										   scale = col.scale,
										   legend = col.legend,
										   chart = col.chart,
										   free = col.free),
						   col_alpha = tmapScale(aes = "col_alpha",
						   					  value = col_alpha,
						   					  scale = col_alpha.scale,
						   					  legend = col_alpha.legend,
						   					  chart = col_alpha.chart,
						   					  free = col_alpha.free)),

		gpar = tmapGpar(fill = "__col",
						col = NA,
						shape = NA,
						size = NA,
						fill_alpha = "__col_alpha",
						col_alpha = NA,
						pattern = "fill",
						lty = NA,
						lwd = NA,
						linejoin = NA,
						lineend = NA),
		tpar = tmapTpar(),
		plot.order = tm_plot_order("DATA"),
		mapping.args = options$mapping.args,
		zindex = zindex,
		group = group,
		group.control = group.control,
		popup.vars = FALSE,
		popup.format = list(),
		hover = "",
		id = "",
		subclass = c("tm_aes_layer", "tm_layer")))
}

#' @rdname tm_raster
#' @param interpolate Should the raster image be interpolated? Currently only applicable in view mode (passed on to [`grid`][grid::rasterGrob()])
#' @export
opt_tm_raster = function(interpolate = FALSE) {
	list(trans.args = list(),
		 mapping.args = list(interpolate = interpolate))
}

step4_plot_collect_charts = function(tmx) {
	dt_template = data.table::data.table(by1__ = integer(0), by2__ =  integer(0), by3__ =  integer(0), comp = list())
	if (!length(tmx)) {
		crts = dt_template
	} else {
		crts = data.table::rbindlist(c(list(dt_template), lapply(tmx, function(tmxi) {
			data.table::rbindlist(lapply(tmxi$layers, function(tml) {

				crts_cached = get("charts", .TMAP)

				legs = c(tml$trans_legend, tml$mapping_legend)

				legs = lapply(legs, function(l) {
					#l$comp = list(crts_cached[l$crtnr])
					l[, comp:=list(crts_cached[l$crtnr])]
					l[, intersect(names(l), names(dt_template)), with = FALSE]
				})
				data.table::rbindlist(legs, fill = TRUE)
			}), fill = TRUE)
		})), fill = TRUE)
		# remove empty legends
		crts = crts[vapply(crts$comp, length, FUN.VALUE = integer(1)) > 1, ]
	}
	crts
}

step4_plot_collect_legends = function(tmx) {
	# collect legends
	dt_template = data.table::data.table(by1__ = integer(0), by2__ =  integer(0), by3__ =  integer(0), legend = list())

	if (!length(tmx)) {
		legs = dt_template
	} else {
		legs = data.table::rbindlist(c(list(dt_template), lapply(tmx, function(tmxi) {
			data.table::rbindlist(lapply(tmxi$layers, function(tml) {

				legs_cached = get("legs", .TMAP)

				legs = c(tml$trans_legend, tml$mapping_legend)

				legs = lapply(legs, function(l) {
					l[, legend:=list(legs_cached[l$legnr])]
					l
				})

				# make vneutral explicit (as a new column)
				legs2 = lapply(legs, function(legs_aes) {
					legs_aes$vneutral = unlist(lapply(legs_aes$legend, function(l) l$vneutral), use.names = FALSE, recursive = FALSE)
					if ("icon_scale" %in% names(legs_aes$legend[[1]])) {
						legs_aes$icon_scale = legs_aes$legend[[1]]$icon_scale
					} else {
						legs_aes$icon_scale = 1
					}
					legs_aes
				})


				# are icons used? If so set icon_scale
				icon_scales = vapply(legs2, function(l) {
					as.numeric(l$icon_scale[1])
				}, FUN.VALUE = numeric(1))
				icon_scale = if (any(icon_scales != 1)) icon_scales[which(icon_scales != 1)[1]] else 1

				clones = vapply(legs2, function(l) {
					a = l$legend[[1]]$aes
					a %||% ""
				}, FUN.VALUE = character(1))


				# find shared legends
				legs2b = mapply(function(l, lnm) {
					w = which(clones == lnm)
					if (length(w) > 0L) {
						legsclone = lapply(legs2[w], function(li) {
							li$legend
						})
						num_legends = vapply(legsclone, length, numeric(1))
						if (any(num_legends != length(l$legend))) warning("legends could not be shared; the aesthetics need the same .free specification", call. = FALSE)


						l$legend = do.call(mapply, args = c(list(FUN = function(l, ...) {
							clns = list(...)
							k = length(l$vvalues)
							l$clones = lapply(clns, function(cl) {
								vv = cl$vvalues
								if (k != length(vv)) cli::cli_abort("legends could not be shared; the number of legend items is different")
								vv
							})
							names(l$clones) = names(clones[w])
							l
						}, SIMPLIFY = FALSE), c(list(l$legend), legsclone)))
					}
					l
				}, legs2, names(legs2), SIMPLIFY = FALSE)

				copy_neutral = (length(legs) > 1)

				legs3 = mapply(function(legs_aes, legnm, i) {
					bvars = names(legs_aes)[substr(names(legs_aes), 1, 2) == "by"]

					if (!is.null(legs_aes)) {
						for (k in 1:nrow(legs_aes)) {
							if (!legs_aes$legend[[k]]$active || !legs_aes$legend[[k]]$show) {
								legs_aes$legend[[k]] = list(NULL)
								next
							}

							bval = legs_aes[k, bvars, with = FALSE]
							gp = tml$gp

							# fill gp values with (scaled) aes data
							gpaid = which(paste0("__", legnm) == unlist(gp, use.names = FALSE))
							for (j in gpaid) gp[[j]] = legs_aes$legend[[k]]$vvalues

							# fill gp with neutral values for other graphical variables
							if (copy_neutral) {
								nvalues = mapply(function(lclone, lname) {
									bvars2 = names(lclone)[substr(names(lclone), 1, 2) == "by"]

									bvars_int = intersect(bvars, bvars2)
									if (!length(bvars_int)) {
										lclone$vneutral[1]
									} else  {
										lclone[bval, on = bvars_int]$vneutral
									}
								}, legs2[-i], names(legs2[-i]), SIMPLIFY = FALSE)


								for (j in seq_along(nvalues)) {
									gpid = which(paste0("__", names(nvalues)[j]) == sapply(gp, "[[", 1))
									for (jj in gpid) gp[[jj]] = nvalues[[j]]
								}

							}

							leleg = legs_aes$legend[[k]]



							# get gp values from shared legends (slave)
							if ("clones" %in% names(leleg)) {
								clist = leleg$clones
								for (j in 1L:length(clist)) {
									cnm = names(clist)[j]
									gp[[cnm]] = clist[[j]]
								}
							}

							# in case the size legend consists of symbols shapes that represented icons
							if (icon_scale > 1 && legnm == "size") {
								gp$size = gp$size * icon_scale
							}

							leleg$gp = gp
							leleg$vneutral = NULL
							leleg$vvalues = NULL

							if (length(leleg) == 1) {
								legs_aes$legend[[k]] = list(leleg)
							} else {
								legs_aes$legend[[k]] = leleg
							}
						}
					}
					legs_aes
				}, legs2b, names(legs2b), 1:length(legs2b), SIMPLIFY = FALSE)

				data.table::rbindlist(legs3, fill = TRUE)
			}), fill = TRUE)
		})), fill = TRUE)

		# remove empty legends
		legs = legs[vapply(legs$legend, length, FUN.VALUE = integer(1)) > 1, ][, vneutral := NULL]
	}


	# remove colums and rename legend to comp
	set(legs, j= 5:ncol(legs), value = NULL)
	data.table::setnames(legs, "legend", "comp")
	legs
}

#' Map layer: basemap / overlay tiles
#'
#' Map layer that draws tiles from a tile server. `tm_basemap()` draws the tile
#' layer as basemap, i.e. as bottom layer. In contrast, `tm_tiles()` draws the
#' tile layer as overlay layer, where the stacking order corresponds with the
#' order in which this layer is called, just like other map layers.
#'
#'  API keys. For Stadia and Thunderforest maps, an API key is required.
#'  This can be set via the argument `api`. Alternatively they can be stored in environment variables `"STADIA_MAPS"` and `THUNDERFOREST_MAPS` with `Sys.setenv`
#'
#' In view mode, each layer is rendered in a Leaflet pane named `"tmap{zindex}"`
#' (e.g., `"tmap401"`, `"tmap402"`), with base tile layers placed in the
#' standard `"tile"` pane.
#' @param server Name of the provider or an URL. Or a vector of multiple values. The list of available providers
#'   can be obtained with `providers` (tip: in RStudio, type `leaflet::providers$` to see
#'   the options). See <https://leaflet-extras.github.io/leaflet-providers/preview/>
#'   for a preview of those. When a URL is provided, it should be in template format,
#'   e.g. \code{"https://{s}.tile.openstreetmap.org/{z}/{x}/{y}.png"}.
#'   Use `NULL` in `tm_basemap()` to disable basemaps. It can be a named vector. In that case these names will be used a group names, as alternative to the argument `group`.
#' @param alpha Transparency level
#' @param zoom Zoom level (only used in plot mode)
#' @param api API key. Needed for `Stadia` and `Thunderforest` maps in plot mode, and for `Esri`/`MapTiler` basemaps in `"maplibre"`/`"mapbox"` mode (alternatively set the env var `ARCGIS_API_KEY` / `MAPTILER_API_KEY`). See details
#' @param max.native.zoom Maximum native zoom level (only used in view mode).
#'   The minimum and maximum zoom levels are determined in `tm_view()`.
#' @param sub subdomain of the tile server. Only used when `server` is a url template. The
#'   default is `"abc"` which works for most tile servers.
#' @param zindex Controls the stacking order of map layers. Should be set to a
#'   value above 400. By default, layers are stacked in call order, starting
#'   at 401. See details.
#' @param group Name of the group to which this layer belongs. This is only
#'   relevant in view mode, where layer groups can be switched (see `group.control`)
#' @param group.control In view mode, the group control determines how layer
#'   groups can be switched on and off. Options: `"radio"` for radio buttons
#'   (meaning only one group can be shown), `"check"` for check boxes
#'   (so multiple groups can be shown), and `"none"` for no control
#'   (the group cannot be (de)selected).
#' @export
#' @seealso `r .doc_see_also_basemaps()`
#' @example ./examples/tm_basemap.R
tm_basemap = function(server = NA, alpha = NULL, zoom = NULL, api = NULL, max.native.zoom = 17, sub = "abc", zindex = 0, group = NA, group.control = "radio") {
	if (is.null(server)) {
		disable = TRUE
	} else {
		if (is.na(server[1])) {
			server = NULL # complete_options is used in step1 to find the default
		}
		disable = FALSE
	}

	group = check_basemap_group(server, group, called_from = "tm_basemap")


	tm_element_list(tm_element(
		args = list(server = server, alpha = alpha, zoom = zoom, api = api, max.native.zoom = max.native.zoom, sub = sub, type = "basemap", disable = disable),
		mapping.fun = "tm_aux_basemap",
		zindex = zindex,
		group = group,
		group.control = group.control,
		subclass = c("tm_basemap", "tm_aux_layer")))
}

#' @export
#' @rdname tm_basemap
tm_tiles = function(server = NA, alpha = NULL, zoom = NULL, max.native.zoom = 17, sub = "abc", zindex = NA, group = NA, group.control = "check") {
	if (is.null(server)) {
		disable = TRUE
	} else {
		if (is.na(server[1])) {
			server = NULL # complete_options is used in step1 to find the default
		}
		disable = FALSE
	}

	group = check_basemap_group(server, group, called_from = "tm_tiles")


	tm_element_list(tm_element(
		args = list(server = server, alpha = alpha, zoom = zoom, max.native.zoom = max.native.zoom, sub = sub, type = "overlay", disable = disable),
		mapping.fun = "tm_aux_tiles",
		zindex = zindex,
		group = group,
		group.control = group.control,
		subclass = c("tm_tiles", "tm_aux_layer")))
}

check_basemap_group = function(server, group, called_from) {
	if (is.null(server)) return(group)
	if (!is.na(group[1])) {
		if (length(group) != length(server)) {
			cli::cli_warn("{.field {.fun {called_from}}} multiple {.arg group} names specified, but unequal to the length of {.arg server}")
		} else {
			group = NA
		}

		if (!is.null(names(server))) {
			cli::cli_warn("{.field {.fun {called_from}}} {.arg group} specified, but {.arg server} is already a named vector, so {.arg group} will be ignored")
			group = NA
		}
	}
	group
}



#' @export
#' @rdname tm_grid
tm_graticules = function(x = NA,
						 y = NA,
						 n.x = NA,
						 n.y = NA,
						 crs = 4326,
						 labels.format = list(suffix = intToUtf8(176)),
						 labels.cardinal = TRUE,
						 ...) {
	do.call(
		tm_grid,
		c(list(x = x, y = y, n.x = n.x, n.y = n.y, crs = crs, labels.format = labels.format, labels.cardinal = labels.cardinal), list(...))
		)
}

#' Coordinate grid / graticule lines
#'
#' Draws horizontal and vertical lines according to a coordinate reference system (CRS). `tm_grid()` uses the CRS of the (master) shape object, and `tm_graticules()` uses latitude and longitude coordinates (EPSG 4326). It creates a [`tmap-element`] that draws coordinate grid lines. It serves as a
#' layer that can be drawn anywhere between other layers.
#'
#'#' In view mode, each layer is rendered in a Leaflet pane named `"tmap{zindex}"`
#' (e.g., `"tmap401"`, `"tmap402"`), with base tile layers placed in the
#' standard `"tile"` pane.
#' @param x X coordinates for vertical grid lines. If `NA`, it is specified
#'   with a pretty scale and `n.x`.
#' @param y Y coordinates for horizontal grid lines. If `NA`, it is specified
#'   with a pretty scale and `n.y`.
#' @param n.x Preferred number of grid lines for the x axis. For the labels,
#'   a [pretty()] sequence is used, so the number of actual labels may be different than `n.x`.
#' @param n.y Preferred number of grid lines for the y axis. For the labels, a
#'   [pretty()] sequence is used, so the number of actual labels may be different than `n.y`.
#' @param crs Projection character. If specified, the grid lines are projected
#'   accordingly. Many world maps are projected, but still have latitude longitude
#'   (EPSG 4326) grid lines.
#' @param col Color of the grid lines.
#' @param lwd Line width of the grid lines
#' @param alpha Alpha transparency of the grid lines. Number between 0 and 1.
#'   By default, the alpha transparency of `col` is taken.
#' @param labels.show Show tick labels. Either one value for both `x` and `y` axis,
#'   or a vector two: the first for `x` and latter for `y`.
#' @param labels.pos position of the labels. Vector of two: the horizontal
#'   ("left" or "right") and the vertical ("top" or "bottom") position.
#' @param labels.size Font size of the tick labels
#' @param labels.col Font color of the tick labels
#' @param labels.rot Rotation angles of the labels. Vector of two values: the
#'   first is the rotation angle (in degrees) of the tick labels on the x axis
#'   and the second is the rotation angle of the tick labels on the y axis.
#'   Only `0`, `90`, `180`, and `270` are valid values.
#' @param labels.format List of formatting options for the grid labels. Parameters are:
#' \describe{
#' \item{fun}{Function to specify the labels. It should take a numeric vector,
#'   and should return a character vector of the same size.
#'   By default it is not specified. If specified, the list items `scientific`,
#'   `format`, and `digits` (see below) are not used.}
#' \item{scientific}{Should the labels be formatted scientifically?
#'   If so, square brackets are used, and the `format` of the numbers is `"g"`.
#'   Otherwise, `format="f"`, and `text.separator`, `text.less.than`, and
#'   `text.or.more` are used. Also, the numbers are automatically rounded to
#'   millions or billions if applicable.}
#' \item{format}{By default, `"f"`, i.e. the standard notation `xxx.xxx`, is used.
#'   If `scientific=TRUE` then `"g"`, which means that numbers are formatted scientifically,
#'   i.e. `n.dddE+nn` if needed to save space.}
#' \item{digits}{Number of digits after the decimal point if `format="f"`, and
#'   the number of significant digits otherwise.}
#' \item{...}{Other arguments passed on to [formatC()]}
#' }
#' @param labels.cardinal Add the four cardinal directions (N, E, S, W) to the labels,
#'   instead of using negative coordinates for west and south (so it assumes that
#'   the coordinates are positive in the north-east direction).
#' @param labels.margin.x Margin between tick labels of x axis and the frame.
#'   Note that when `labels.inside_frame = FALSE` and `ticks = TRUE`,
#'   the ticks will be adjusted accordingly.
#' @param labels.margin.y Margin between tick labels of y axis and the frame.
#'   Note that when `labels.inside_frame = FALSE` and `ticks = TRUE`,
#'   the ticks will be adjusted accordingly.
#' @param labels.space.x Space that is used for the labels and ticks for the x-axis
#'   when `labels.inside_frame = FALSE`. By default, it is determined automatically
#'   using the widths and heights of the tick labels. The unit of this parameter is text line height.
#' @param labels.space.y Space that is used for the labels and ticks for the y-axis
#'   when `labels.inside_frame = FALSE`. By default, it is determined automatically
#'   using the widths and heights of the tick labels. The unit of this parameter is text line height.
#' @param labels.inside_frame Show labels inside the frame? By default `FALSE`.
#' @param ticks If `labels.inside_frame = FALSE`, should ticks can be drawn between the labels and the frame?
#'   Either one value for both `x` and `y` axis, or a vector two: the first for `x` and latter for `y`.
#' @param lines If `labels.inside_frame = FALSE`, should grid lines can be drawn?
#' @param ndiscr Number of points to discretize a parallel or meridian
#'   (only applicable for curved grid lines)
#' @param zindex Controls the stacking order of map layers. Should be set to a
#'   value above 400. By default, layers are stacked in call order, starting
#'   at 401. See details.
#' @param group Name of the group to which this layer belongs.
#'   This is only relevant in view mode, where layer groups can be
#'   switched (see `group.control`)
#' @param group.control In view mode, the group control determines how layer groups
#'   can be switched on and off. Options: `"radio"` for radio buttons (meaning only one group can be shown),
#'   `"check"` for check boxes (so multiple groups can be shown), and `"none"`
#'   for no control (the group cannot be (de)selected).
#' @param ... Used to catch deprecated arguments from tmap v3.
#' @export
#' @seealso `r .doc_see_also_shape()`
#' @example ./examples/tm_grid.R
tm_grid = function(x = NA,
				   y = NA,
				   n.x = NA,
				   n.y = NA,
				   crs = NA,
				   col = NA,
				   lwd = 1,
				   alpha = NA,
				   labels.show = TRUE,
				   labels.pos = c("left", "bottom"),
				   labels.size = .6,
				   labels.col = NA,
				   labels.rot = c(0, 0),
				   labels.format = list(big.mark = ","),
				   labels.cardinal = FALSE,
				   labels.margin.x = 0,
				   labels.margin.y = 0,
				   labels.space.x = NA,
				   labels.space.y = NA,
				   labels.inside_frame = FALSE,
				   ticks = labels.show & !labels.inside_frame,
				   lines = TRUE,
				   ndiscr = 100,
				   zindex = NA,
				   group = NA,
				   group.control = "none",
				   ...) {
	args = list(...)
	if ("projection" %in% names(args)) {
		cli::cli_warn("The 'projection' argument of 'tm_grid()' is deprecated as of tmap 4.0. Pleaes use 'crs' instead.", call. = FALSE)
		crs = args$projection
	}
	tm_element_list(tm_element(
		args = c(list(show = TRUE), as.list(environment())),
		mapping.fun = "tm_aux_grid",
		zindex = zindex,
		group = group,
		group.control = group.control,
		subclass = c("tm_grid", "tm_aux_layer")))
}

#' Overview of tmap layers
#'
#' Overview of tmap layers, organized by layer type.
#'
#' @export
#' @return A list of three layer types are returned: data layers, aux layers, and components.
tmap_overview = function() {

	modes = get_modes()
	tmap_graphics_name()

	om = get("tmapOptions", envir = .TMAP)$modes

	modes = names(om)

	gs = vapply(modes, function(m) om[[m]]$name, FUN.VALUE = character(1))

	funs = c("DataPlot", "AuxPrepare", "CompPrepare")
	nms = c("data_layer", "aux_layer", "component")
	rem = c("data_", "aux_", "")

	# Helper to get package
	find_env <- function(generic, class) {
		# Get the actual method
		fun <- tryCatch(
			getS3method(generic, class),
			error = function(e) NULL
		)
		if (is.null(fun)) return(NA)
		# Get its environment
		env <- environment(fun)
		if (isNamespace(env)) {
			return(unname(getNamespaceName(env)))
		} else {
			return(unname(as.character(env)))
		}
	}

	res = rbindlist(mapply(function(funi, remi, ni) {
		dt = data.table::rbindlist(mapply(function(g, nm) {
			guni = paste0("tmap", g, funi)
			ms = utils::methods(guni)
			df = attr(ms, "info")

			cls = sub("^[^.]*\\.", "", rownames(df))
			x = data.table::data.table(element = sub(paste0("^[^.]*\\.(tm_)?(", remi, ")?"), "\\1", rownames(df)),
									   type = ni)
			x$package = vapply(cls, function(x) {
				find_env(guni, x)
			}, FUN.VALUE = character(1))

			x$mode = nm
			x[x$element != "default", ]
		}, gs, names(gs), SIMPLIFY = FALSE))

		wide_dt <- data.table::dcast(
			dt,
			element + type ~ mode,
			value.var = "package",fill = "----"
		)
		data.table::setcolorder(
			wide_dt,
			c("element", "type", "plot", "view", sort(setdiff(names(wide_dt), c("element", "type", "plot", "view"))))
		)
	}, funs, rem, nms, SIMPLIFY = FALSE))
	as.data.frame(res)
}

#' @export
#' @rdname tmap_internal
tmapScaleIntervals = function(x1, scale, legend, chart, o, aes, layer, layer_args, sortRev, bypass_ord, submit_legend = TRUE) {
	cls = data_class(x1, midpoint_enabled = !is.null(scale$midpoint))
	maincls = class(scale)[1]

	if (attr(cls, "unique") && is.null(scale$breaks)) {
		scale$ticks = NA
		cli::cli_inform("The visual variable {.arg {aes}} of the layer {.str {layer}} contains a unique value. Therefore a discrete scale is applied (tm_scale_discrete).")
		return(tmapScaleDiscrete(x1, scale, legend, chart, o, aes, layer, layer_args, sortRev, bypass_ord, submit_legend = submit_legend))

	}

	if (!(cls[1] %in% c("num", "datetime", "date"))) {
		if (!is.factor(x1)) x1 = as.factor(x1)
		x1 = as.integer(x1)
		cls = c("num", "int", "seq")
		warning(maincls, " is supposed to be applied to numerical or date/time data", call. = FALSE)
	}

	x1 = without_units(x1)

	if (aes %in% c("pattern")) stop("tm_scale_intervals cannot be used for layer ", layer, ", aesthetic ", aes, call. = FALSE)

	if (!(aes %in% c("fill", "color")) && scale$label.style == "continuous") {
		scale$label.style = "discrete"
		cli::cli_warn("{.field tm_scale_intervals} {.arg label.style} can only be {.str continuous} for fill and color, not for {aes}")
	}


	scale = get_scale_defaults(scale, o, aes, layer, cls)

	show.messages <- o$show.messages
	show.warnings <- o$show.warnings


	fun = paste0("interval_", cls[1])

	# pre-checks
	midpoint = NULL
	scale = within(scale, {
		allna = all(is.na(x1))
		if (anyDuplicated(breaks)) stop("breaks specified in the ", aes,  ".scale scaling function contains duplicates.", call. = FALSE)
		if (!is.null(breaks) && length(breaks) < 2 && cls[1] == "num") stop("breaks should contain at least 2 numbers", call. = FALSE)

		udiv = identical(use_div(breaks, midpoint), TRUE)

		if (allna && is.null(breaks)) {
			if (show.messages) message("Variable(s) \"", paste(aes, collapse = "\", \""), "\" only contains NAs. Legend disabled for tm_scale_intervals, unless breaks are specified")
			chart = within(chart, {
				labels = label.na
				vvalues = c(value.na, value.na)
				breaks = c(0, 1)
				na.show = TRUE
				x1 = x1[1]
			})
			return = tmapScale_returnNA(n = length(x1), legend = legend, chart = chart, value.na = value.na, label.na = label.na, label.show = label.show, na.show = legend$na.show, sortRev = sortRev, bypass_ord = bypass_ord)
		}
	})

	if (!is.null(scale$return)) return(scale$return)

	# data-type dependent processing
	scale = do.call(fun, list(scale = scale, x1 = x1, aes = aes, layer = layer, show.messages = show.messages, show.warnings = show.warnings))

	# visual variable processing
	with(scale, {
		fun_getVV = paste0("tmapValuesVV_", aes)
		VV = do.call(fun_getVV, list(x = values, value.na = value.na, isdiv = isdiv, n = n, dvalues = breaks, midpoint = midpoint, range = values.range, scale = values.scale * o$scale, are_breaks = TRUE, rep = values.repeat, o = o))

		vvalues = VV$vvalues
		value.na = VV$value.na

		sfun = paste0("tmapValuesScale_", aes)
		cfun = paste0("tmapValuesColorize_", aes)
		if (is.na(value.neutral)) value.neutral = VV$value.neutral else value.neutral = do.call(sfun, list(x = do.call(cfun, list(x = value.neutral, pc = o$pc)), scale = values.scale))

		# check for size[1] = 0
		if (aes %in% c("size", "lwd")) {
			if (vvalues[1] == 0) {
				message_scale_interval_value0(aes, values, layer)
			}
		}

		mfun = paste0("tmapValuesSubmit_", aes)
		vvalues = do.call(mfun, list(x = vvalues, args = layer_args))
		value.na = do.call(mfun, list(x = value.na, args = layer_args))
		value.neutral = do.call(mfun, list(x = value.neutral, args = layer_args))


		ids = classInt::findCols(q)
		vals = vvalues[ids]
		isna = is.na(vals)
		anyNA = any(isna)
		na.show = update_na.show(label.show, legend$na.show, anyNA)

		if (is.null(sortRev)) {
			ids = NULL
		} else if (is.na(sortRev)) {
			ids[] = 1L
		} else if (sortRev) {
			ids = (as.integer(n) + 1L) - ids
		}

		if (anyNA) {
			vals[isna] = value.na
			if (!is.null(sortRev)) ids[isna] = 0L
		}

		# detransform log
		if (is.log) {
			if (any((breaks %% 1) != 0)) message("non-rounded breaks occur, because style = \"log10_pretty\" is designed for large values")
			breaks <- 10^breaks
		}

		# create legend values
		#values = breaks[-nbrks]

		if (label.style == "discrete") {
			if (is.null(labels)) {
				labels = do.call("fancy_breaks", c(list(vec=breaks, as.count = as.count, intervals=TRUE, interval.closure=int.closure), label.format))
			} else {
				if (length(labels)!=nbrks-1 && show.warnings) warning("number of legend labels should be ", nbrks-1, call. = FALSE)
				labels = rep(labels, length.out=nbrks-1)
				attr(labels, "align") <- label.format$text.align
			}
		} else {
			if (is.null(labels)) {
				labels = do.call("fancy_breaks", c(list(vec=breaks, as.count = FALSE, intervals=FALSE, interval.closure=int.closure), label.format))
			} else {
				if (length(labels) != length(breaks)) cli::cli_abort("{.field tm_scale_intervals} {.arg labels} should have length {length(breaks)}")
				#labels = c(labels, {if (na.show) label.na else NULL})
			}
		}



		if (legend$reverse) {
			labels.brks <- attr(labels, "brks")
			labels.align <- attr(labels, "align")
			labels <- rev(labels)
			if (!is.null(labels.brks)) {
				attr(labels, "brks") = labels.brks[length(labels):1L,]
			}
			attr(labels, "align") = labels.align
			vvalues_rev = rev(vvalues)
		} else {
			vvalues_rev = vvalues
		}



		if (na.show) {
			labels.brks = attr(labels, "brks")
			labels.align = attr(labels, "align")
			#labels <- c(labels, legend.NA.text)
			if (!is.null(labels.brks)) {
				labels <- c(labels, paste(label.na, " ", sep = ""))
				attr(labels, "brks") = rbind(labels.brks, rep(nchar(label.na) + 2, 2))
			} else {
				labels = c(labels, label.na)
			}
			attr(labels, "align") = labels.align
			vvalues = c(vvalues, value.na)
			vvalues_rev = c(vvalues_rev, value.na)
		}

		if (label.style == "discrete") {
			legend = within(legend, {
				nitems = length(labels)
				labels = labels
				dvalues = breaks #not used?
				vvalues = vvalues_rev
				vneutral = value.neutral
				na.show = get("na.show", envir = parent.env(environment()))
				scale = "intervals"
				layer_args = layer_args
			})

		} else {
			if ((o$continuous.nclass_per_legend_break %% 2) != 0) cli::cli_abort("{.field options} the tmap option {.arg continuous.nclass_per_legend_break} should be even")
			labels_select = c(rep(label.select, length.out = length(breaks)), {if (na.show) TRUE else NULL})

			vvalues = c(unlist(mapply(function(hd, tl) {
				paste(c(rep(hd, o$continuous.nclass_per_legend_break/2),
						rep(tl, o$continuous.nclass_per_legend_break/2)),
						collapse = "_")
			}, c(NA_character_, vvalues_rev[1L:(length(breaks)-1L)]), c(vvalues_rev[1L:(length(breaks)-1L)], NA_character_))), {if (na.show) value.na else NULL})
			nitems = length(labels)


			legend = within(legend, {
				nitems = nitems
				labels = labels
				dvalues = breaks #not used?
				vvalues = vvalues
				vneutral = value.neutral
				na.show = get("na.show", envir = parent.env(environment()))
				scale = "intervals"
				layer_args = layer_args
				is_discrete = TRUE

				# continuous legend specific:
				labels_select = labels_select
				tr = trans_identity
				limits = range(breaks)
			})
		}

		chartFun = paste0("tmapChart", toTitleCase(chart$summary))

		chartArgs = list(chart,
						 bin_colors = vvalues,
						 breaks_def = breaks,
						 na.show = na.show,
						 x1 = x1)

		# if breaks are user specified via tm_chart, use them
		if ("breaks" %in% names(chart) && !is.null(chart$breaks)) {
			chartArgs["bin_colors"] = list(NULL)
			chartArgs$breaks_def = chart$breaks
		}

		chart = do.call(chartFun, chartArgs)

		if (submit_legend) {
			if (bypass_ord) {
				format_aes_results(vals, legend = legend, chart = chart)
			} else {
				format_aes_results(vals, ids, legend, chart = chart)
			}
		} else {
			list(vals = vals, ids = ids, legend = legend, chart = chart, bypass_ord = bypass_ord)
		}
	})
}

interval_num = function(scale, x1, aes, layer, show.messages, show.warnings) {
	within(scale, {
		local({
			styles = c("fixed", "sd", "equal", "pretty", "quantile", "kmeans", "hclust", "bclust", "fisher", "jenks", "dpih", "headtails", "maximum", "box", "log10", "log10_pretty")
			fn_call <-  call(paste0("tm_", layer[1]))
			if (!(style %in% styles)) {
				cli::cli_abort("Invalid style. Style should be one of {.str {styles}}",
					call = fn_call
				)
			}
		})


		if (!any(style == c("pretty", "log10_pretty", "fixed"))) {
			if (identical(as.count, TRUE) && show.warnings) warning("as.count not implemented for styles other than \"pretty\", \"log10_pretty\" and \"fixed\"", call. = FALSE)
			as.count = FALSE
		}
		if (is.na(as.count)) {
			as.count = is.integer(x1) && !any(!is.na(x1) & x1 < 0)
		}

		if (as.count) {
			if (interval.closure != "left" && show.warnings) warning("For as.count = TRUE, interval.closure will be set to \"left\"", call. = FALSE)
			interval.closure = "left"
		}

		#breaks.specified <- !is.null(breaks)
		is.log = (style == "log10_pretty")
		if (is.log && !attr(label.format, "big.num.abbr.set")) label.format$big.num.abbr = NA

		if (style == "log10_pretty") {
			x1 = log10(x1)
			style = "fixed"
			breaks = seq(floor(min(x1, na.rm = TRUE)), ceiling(max(x1, na.rm=TRUE)))
		} else if (as.count && style == "pretty") {
			breaks = prettyCount(x1, n=n)
			style <- "fixed"
		} else if (as.count && style == "fixed") {
			breaks[length(breaks)] = breaks[length(breaks)] + 1L
		}

		q = num2breaks(x=x1, n=n, style=style, breaks=breaks, interval.closure=interval.closure, var=paste(layer, aes, sep = "-"), as.count = as.count, args = style.args)

		breaks = q$brks
		nbrks = length(breaks)
		n = nbrks - 1

		int.closure <- attr(q, "intervalClosure")

		# update range if NA (automatic)
		if (is.na(values.range[1])) {
			fun_range = paste0("tmapValuesRange_", aes)
			values.range = do.call(fun_range, args = list(x = values, n = n, isdiv = udiv))
		}
		if (length(values.range) == 1 && !is.na(values.range[1])) values.range = c(0, values.range)

		check_values(layer, aes, values)

		fun_isdiv = paste0("tmapValuesIsDiv_", aes)

		isdiv = !is.null(midpoint) || do.call(fun_isdiv, args = list(x = values))

		# determine midpoint
		if ((is.null(midpoint) || is.na(midpoint)) && isdiv) {
			rng <- range(x1, na.rm = TRUE)
			if (rng[1] < 0 && rng[2] > 0 && is.null(midpoint)) {

				if (show.messages) message(paste0("[scale] tm_", layer[1], ":() the data variable assigned to '", aes, "' contains positive and negative values, so midpoint is set to 0. Set 'midpoint = NA' in 'fill.scale = tm_scale_intervals(<HERE>)' to use all visual values (e.g. colors)"))
				midpoint <- 0
			} else {
				if ((n %% 2) == 1) {
					# number of classes is odd, so take middle class (average of those breaks)
					midpoint <- mean.default(breaks[c((n+1) / 2, (n+3) / 2)])
				} else {
					midpoint <- breaks[(n+2) / 2]
				}
			}
		}
	})
}

interval_date = function(scale, x1, aes, layer, show.messages, show.warnings) {
	interval_datetime(scale, x1, aes, layer, show.messages, show.warnings)
}


interval_datetime = function(scale, x1, aes, layer, show.messages, show.warnings) {
	style = NULL
	interval.closure = NULL
	udiv = NULL
	show.messages = NULL
	within(scale, {
		local({
			styles = c("fixed", "pretty")
			fn_call <-  call(paste0("tm_", layer[1]))
			if (!(style %in% styles)) {
				cli::cli_abort("Invalid style. For date/time data variables, style should be one of {.str {styles}}",
							   call = fn_call
				)
			}
		})


		if (identical(as.count, TRUE) && show.warnings) {
			warning("as.count not applicable for data/time data variables", call. = FALSE)
			as.count = FALSE
		}

		#if (style == "pretty") {
		if (is.null(breaks)) {
			breaks = pretty_datetime_seq(x=x1, n = n)
		} else if (length(breaks) == 1L) {
			breaks = pretty_datetime_seq3(x=x1, by = breaks)
		}


		q <- list(var=x1,
				  brks=breaks)
		attr(q, "style") <- "fixed"
		attr(q, "nobs") <- nobs
		attr(q, "intervalClosure") <- interval.closure
		class(q) <- "classIntervals"

		breaks = q$brks
		nbrks = length(breaks)
		n = nbrks - 1

		is.log = FALSE

		int.closure <- attr(q, "intervalClosure")

		# update range if NA (automatic)
		if (is.na(values.range[1])) {
			fun_range = paste0("tmapValuesRange_", aes)
			values.range = do.call(fun_range, args = list(x = values, n = n, isdiv = udiv))
		}
		if (length(values.range) == 1 && !is.na(values.range[1])) values.range = c(0, values.range)

		check_values(layer, aes, values)

		fun_isdiv = paste0("tmapValuesIsDiv_", aes)

		isdiv = !is.null(midpoint) || do.call(fun_isdiv, args = list(x = values))

		# determine midpoint
		if ((is.null(midpoint) || is.na(midpoint)) && isdiv) {
			rng <- range(x1, na.rm = TRUE)

			if ((n %% 2) == 1) {
				# number of classes is odd, so take middle class (average of those breaks)
				midpoint <- mean(breaks[c((n+1) / 2, (n+3) / 2)])
			} else {
				midpoint <- breaks[(n+2) / 2]
			}

		}
	})

}

.doc_opt = function(a) {
	x = strsplit(a, split = ".", fixed = TRUE)[[1]]
	cls = head(x, -1)
	what = tail(x, 1)

	cls_lookup = c(inner = "inner space (inside the frame)",
				   outer = "outer space (outside the frame",
				   item_text = "space between item and text")

	what_lookup = c(alpha = "alpha transparency",
					asp = "asp ration",
					bbox = "bounding box",
					breaks = "break values",
					col = "color",
					color = "color",
					crs = "coordinate reference system (CRS)",
					double_line = "double line",
					dpi = "dots per inch",
					earth_boundary = "earth_boundary",
					earth_datum = "earth_datum",
					extra = "extra arguments",
					flip = "flip",
					fontface = "font face",
					fontfamily = "font family",
					format = "format",
					frame = "frame",
					height = "height",
					inside_frame = "inside_frame",
					labels = "labels",
					landscape = "landscape",
					light = "light",
					limits = "limits",
					lines = "lines",
					lwd = "line width",
					margin = "margin",
					margins = "margins",
					max = "max",
					max_cells = "max_cells",
					messages = "messages",
					minimap = "minimap",
					modes = "modes",
					mouse_coordinates = "mouse_coordinates",
					na = "na",
					nclass_per_legend_break = "nclass_per_legend_break",
					nclasses = "nclasses",
					ndiscr = "ndiscr",
					neutral = "neutral",
					north = "north",
					null = "null",
					offset = "offset",
					only = "only",
					orientation = "orientation",
					padding = "padding",
					portrait = "portrait",
					pos = "pos",
					position = "position",
					r = "r (radius)",
					range = "range",
					resize_as_group = "resize_as_group",
					reverse = "reverse",
					rot = "rot",
					rotation = "rotation",
					saturation = "saturation",
					scale = "scale",
					scalebar = "scalebar",
					scales = "scales",
					sepia_intensity = "sepia_intensity",
					server = "server",
					show = "visibility",
					side = "side",
					size = "size",
					space = "space",
					stack = "stack",
					stack_margin = "stack_margin",
					text = "text",
					ticks = "ticks",
					toggle = "toggle",
					type = "type",
					unit = "unit",
					var = "var",
					view = "view",
					warnings = "warnings",
					width = "width",
					absolute_fontsize = "absolute fontsize",
					x = "x",
					y = "y",
					xlab = "`x` label",
					ylab = "`y` label",
					zoom = "zoom")

	extra_lookup = c(margins = " A vector of 4 values: bottom, left, top, right. The unit is the device height (for bottom and top) or width (for left and right).",
					 double_line = " `TRUE` or `FALSE`.",
					 extra = " A list of arguments.",
					 padding = " A vector of 4 values: bottom, left, top, right. The unit is the device height (for bottom and top) or width (for left and right).",
					 position = " A tm_pos object, or a shortcut of two values: horizontal (left, center, right) and vertical (top, center, bottom). See tm_pos for details",
					 space = " In terms of number of text line heights.",
					 fontfamily = " See `graphics::par`, option 'family'.",
					 fontface = " See `graphics::par`, option 'font'.",
					 absolute_fontsize = " So far, only used to calculate legend dimensions",
					 lwd = " See `graphics::par`, option 'lwd'.",
					 show = " `TRUE` or `FALSE`.")

	what_text = unname(what_lookup[what])
	if (is.na(what_text)) what_text = what

	cls_text = sapply(cls, function(x) {
		res = unname(cls_lookup[x])
		if (is.na(res)) x else res
	}, USE.NAMES = FALSE)

	extra_text = unname(extra_lookup[what])
	if (is.na(extra_text)) extra_text = ""


	if (length(cls) == 0L) {
		cls_texts = " (overall)"
	} else {
		cls_texts = paste(" of the", paste(rev(cls_text), collapse = " of the "))
	}
	paste0("The ", what_text, cls_texts, ".", extra_text)
}

.doc_vv = function(x) {
	lookup = c(fill = "fill color",
			   col = "color",
			   fill_alpha = "fill color transparency",
			   col_alpha = "color transparency",
			   lwd = "line width",
			   lty = "line type",
			   size = "size",
			   text = "text",
			   fontface = "font face")
	txt = unname(lookup[x])
	if (is.na(txt)) txt = x
	paste0("Visual variable that determines the ", txt, ". See details.")
}

#' @export
#' @rdname tmapGridLeaflet
tmapGridDataPlot.tm_data_text = function(a, shpTM, dt, gp, bbx, facet_row, facet_col, facet_page, id, pane, group, glid, o, ...) {
	rc_text = frc(facet_row, facet_col)

	if (("prop_angle" %in% names(shpTM))) {
		a$point.label = FALSE
	}

	res = select_sf(shpTM, dt[!is.na(dt$size), ])

	shp = res$shp
	dt = res$dt



	# specials non-vv (later on lost after gp_to_gpar)
	shadow = a$shadow


	gp = impute_gp(gp, dt)
	gp = rescale_gp(gp, o$scale_down)

	coords = sf::st_coordinates(shp)

	# in case shp is a multipoint (point_per == "segment"), expand gp:
	cp = expand_coords_gp(coords, gp, ndt = nrow(dt))
	coords = cp$coords
	gp = cp$gp

	n = nrow(coords)

	# by default point.label = NA for tm_labels
	# for small n (< 500) it will be set to TRUE
	if (is.na(a$point.label)) a$point.label = (n < 500)

	if (cp$expanded) {
		shpTM_match = match(shpTM$tmapID_expanded, shpTM$tmapID)
		text = dt$text[shpTM_match]
	} else {
		text = dt$text
	}



	g = get("g", .TMAP_GRID)


	# calculate native per line
	wIn = g$colsIn[g$cols_facet_ids[facet_col]]
	hIn = g$rowsIn[g$rows_facet_ids[facet_row]]

	wNative = bbx[3] - bbx[1]
	hNative = bbx[4] - bbx[2]

	xIn = wNative / wIn
	yIn = hNative / hIn

	lineIn = convertHeight(unit(1, "lines"), "inch", valueOnly = TRUE)



	just = process_just(a$just, interactive = FALSE)
	if (a$point.label) {
		if (!all(just == 0.5)) {
			just = c(0.5, 0.5)
			if (get("tmapOptions", envir = .TMAP)$show.messages) message("Point labeling is enabled. Therefore, just will be ignored.")

		}
	}


	# apply xmod and ymod
	coords[,1] = coords[,1] + xIn * lineIn * gp$cex * gp$xmod
	coords[,2] = coords[,2] + yIn * lineIn * gp$cex * gp$ymod


	# specials vv (later on lost after gp_to_gpar)
	bgcol = gp$bgcol
	bgcol_alpha = gp$bgcol_alpha

	angle = gp$angle

	gp = gp_to_gpar(gp, sel = "col", o = o, type = "text")

	with_bg = any(bgcol_alpha != 0)
	with_shadow_halo = a$shadow || a$halo


	if (with_bg || with_shadow_halo || a$remove_overlap || a$point.label) {
		# grobs are processed seperately because of the order: backgrond1, shadow1, text1, background2, shadow2, text2, etc.
		# becaues it is less efficient when there is no background/shadow (majority of use cases), this is a separate routine

		gps = split_gp(gp, n)

		# decode (possibly) expression-encoded labels; each element is either a
		# character string or a language object, both accepted by textGrob()
		text_dec = lapply(text, decode_expr)

		grobTextList = mapply(function(txt, x , y, gp, a) {
			grid::textGrob(x = grid::unit(x, "native"), y = grid::unit(y, "native"), label = txt, gp = gp, rot = a, just = just) #, name = paste0("text_", id))
		}, text_dec, coords[,1], coords[,2], gps, angle, SIMPLIFY = FALSE, USE.NAMES = FALSE)

		if (with_shadow_halo) {
			gp_sh = gp

			if (a$halo) {
				gp_sh$col = ifelse(is.na(a$halo.col), ifelse(is_light(gp$col), "#000000", "#FFFFFF"), a$halo.col)
			} else {
				gp_sh$col = ifelse(is.na(a$shadow.col), ifelse(is_light(gp$col), "#000000", "#FFFFFF"), a$shadow.col)
			}
			gps_sh = split_gp(gp_sh, n)
			grobTextShList = mapply(function(x, y, txt, g, ai) {
				if (a$halo) {
					offsets <- rbind(
						c(-1, 0),
						c( 1, 0),
						c( 0,-1),
						c( 0, 1)
					)

					diag_offsets <- rbind(
						c(-1,-1),
						c( 1,-1),
						c(-1, 1),
						c( 1, 1)
					)

					# allow smaller diagonal distance for rounder halo
					diag_offsets <- diag_offsets * .5 * sqrt(2)

					offsets <- rbind(offsets, diag_offsets)

					grid::textGrob(x = grid::unit(x + a$halo.width * xIn * lineIn * g$cex * offsets[,1], "native"),
								   y = grid::unit(y -  a$halo.width * yIn * lineIn * g$cex * offsets[,2], "native"),
								   label = txt,
								   gp = g,
								   rot = ai,
								   just = just)
				} else {
					grid::textGrob(x = grid::unit(x + a$shadow.offset.x * xIn * lineIn * g$cex, "native"),
								   y = grid::unit(y -  a$shadow.offset.y * yIn * lineIn * g$cex, "native"),
								   label = txt,
								   gp = g,
								   rot = ai,
								   just = just)
				}
			}, coords[,1], coords[,2], text_dec, gps_sh, angle, SIMPLIFY = FALSE, USE.NAMES = FALSE)
		} else {
			grobTextShList = NULL
		}

		if (with_bg || a$remove_overlap) {
			tGH = vapply(grobTextList, function(grb) {
				grb$rot = 0
				convertHeight(grobHeight(grb), "inch", valueOnly = TRUE)
			}, FUN.VALUE = numeric(1), USE.NAMES = FALSE) * yIn

			tGW = vapply(grobTextList, function(grb) {
				grb$rot = 0
				convertWidth(grobWidth(grb), "inch", valueOnly = TRUE)
			}, FUN.VALUE = numeric(1), USE.NAMES = FALSE) * xIn

			justx <- .5 - just[1]
			justy <- .5 - just[2]

			#tGX <- grobText$x + unit(tGW * justx, "native")
			#tGY <- grobText$y + unit(tGH * justy, "native")


			tGX = unit(coords[,1] + justx * tGW, "native")
			tGY = unit(coords[,2] + justy * tGH, "native")

			tGH = unit(tGH + (0.1 + a$bg.padding) * yIn * lineIn * gp$cex, "native")
			tGW = unit(tGW + a$bg.padding * xIn * lineIn * gp$cex, "native")

			if (a$bg.border) {
				bordercol = a$bg.border.col
				lwd = a$bg.border.lwd
			} else {
				bordercol = NA
				lwd = 0
			}

			grobTextBGList = mapply(function(x, y, w, h, b, a, rot) {
				rect = rectGrob(x=x, y=y, width=w, height=h, gp=gpar(fill=b, alpha = a, col=bordercol, lwd = lwd))
				if (rot != 0) {
					.rectGrob2pathGrob(rect, rot, bbx)$poly
				} else {
					rect
				}
			}, tGX, tGY, tGW, tGH, bgcol, bgcol_alpha, angle, SIMPLIFY = FALSE, USE.NAMES = FALSE)
		} else {
			grobTextBGList = NULL
		}

		#if (a$auto.placement || a$remove_overlap) {
		# grobs to sf
		s = do.call(c,lapply(grobTextBGList, .grob2Poly))
		#}

		if (a$point.label) {
			get_rect_coords = function(polygon) {
				co = sf::st_coordinates(polygon)
				xr = range(co[,1])
				yr = range(co[,2])
				c(x = mean(xr),
				  y = mean(yr),
				  width = xr[2] - xr[1],
				  height = yr[2] - yr[1])
			}

			rect = do.call(rbind, lapply(s, get_rect_coords))

			res = pointLabel2(x = rect[,1], y = rect[,2], width = rect[,3], height = rect[,4], bbx = bbx, gap = yIn * lineIn * a$point.label.gap, method = a$point.label.method)

			sx = res$x - rect[,1]
			sy = res$y - rect[,2]

			grobTextList = mapply(function(grb, sxi, syi) {
				grb$x = grb$x + grid::unit(sxi, "native")
				grb$y = grb$y + grid::unit(syi, "native")
				grb
			}, grobTextList, sx, sy, SIMPLIFY = FALSE)


			grobTextBGList = mapply(function(grb, sxi, syi) {
				grb$x = grb$x + grid::unit(sxi, "native")
				grb$y = grb$y + grid::unit(syi, "native")
				grb
			}, grobTextBGList, sx, sy, SIMPLIFY = FALSE)


			if (with_shadow_halo) {
				grobTextShList = mapply(function(grb, sxi, syi) {
					grb$x = grb$x + grid::unit(sxi, "native")
					grb$y = grb$y + grid::unit(syi, "native")
					grb
				}, grobTextShList, sx, sy, SIMPLIFY = FALSE)
			}
		}

		if (a$remove_overlap) {
			im = sf::st_intersects(s, sparse = FALSE)
			sel = rep(TRUE, length(s))
			rs = rowSums(im)
			while(any(rs>1)) {
				id = which.max(rs)[1]
				sel[id] = FALSE
				im[id,] = FALSE
				im[, id] = FALSE
				rs = rowSums(im)
			}
		} else {
			sel = TRUE
		}

		if (!with_bg && a$remove_overlap) {
			grobTextBGList = NULL
		}



		grobTextAll = list(grobTextBGList[sel], grobTextShList[sel], grobTextList[sel])
		grobTextAll2 = grobTextAll[!vapply(grobTextAll, is.null, FUN.VALUE = logical(1))]

		grb = grid::grobTree(do.call(grid::gList, do.call(c, do.call(mapply, c(list(FUN = list, SIMPLIFY = FALSE, USE.NAMES = FALSE), grobTextAll2)))))

	} else {
		grobText = grid::textGrob(x = grid::unit(coords[,1], "native"), y = grid::unit(coords[,2], "native"), label = decode_expr_vec(text), gp = gp, name = paste0("text_", id), rot = angle)
		grb = grid::grobTree(gList(grobText))
	}


	gts = get("gts", .TMAP_GRID)
	gt = gts[[facet_page]]

	gt_name = paste0("gt_facet_", rc_text)

	gt = grid::addGrob(gt, grb, gPath = grid::gPath(gt_name))



	gts[[facet_page]] = gt
	assign("gts", gts, envir = .TMAP_GRID)
	NULL
}

#' @export
#' @rdname tmapGridLeaflet
tmapGridDataPlot.tm_data_labels = function(a, shpTM, dt, gp, bbx, facet_row, facet_col, facet_page, id, pane, group, glid, o, ...) {
	NextMethod()
}



#' @export
#' @rdname tmapGridLeaflet
tmapGridDataPlot.tm_data_labels_highlighted = function(a, shpTM, dt, gp, bbx, facet_row, facet_col, facet_page, id, pane, group, glid, o, ...) {
	NextMethod()
}

#' Print a random tip to the console
#'
#' @returns A message
#' @export
tmap_tip <- function() {
	nr = .TMAP$tip_nr
	tip = .TMAP$tips[nr]

	nr2 = if (nr == length(.TMAP$tips)) 1L else nr + 1L

	.TMAP$tip_nr = nr2

	cli::cli_inform(tip, .frequency_id = "tmap_tip", .frequency = "always")
}

#' Map layer: lines
#'
#' Map layer that draws lines. Supported visual variables are: `col` (the color), `lwd` (line width), `lty` (line type), and `col_alpha` (color alpha transparency).
#'
#' @param col,col.scale,col.legend,col.chart,col.free  `r .doc_vv("col")`
#' @param lwd,lwd.scale,lwd.legend,lwd.chart,lwd.free  `r .doc_vv("lwd")`
#' @param lty,lty.scale,lty.legend,lty.chart,lty.free  `r .doc_vv("lty")`
#' @param col_alpha,col_alpha.scale,col_alpha.legend,col_alpha.chart,col_alpha.free  `r .doc_vv("col_alpha")`
#' @param linejoin,lineend line join and line end. See [gpar()][grid::gpar()] for details.
#' @param plot.order Specification in which order the spatial features are drawn. See [tm_plot_order()] for details.
#' @param options options passed on to the corresponding `opt_<layer_function>` function
#' @inheritParams tm_polygons
#' @inherit tm_polygons details
#' @param ... to catch deprecated arguments from version < 4.0
#' @example ./examples/tm_lines.R
#' @seealso \href{https://r-tmap.github.io/tmap/articles/examples_terrain}{Terrain map example}
#' @export
tm_lines = function(col = tm_const(),
					col.scale = tm_scale(),
					col.legend = tm_legend(),
					col.chart = tm_chart_none(),
					col.free = NA,
					lwd = tm_const(),
					lwd.scale = tm_scale(),
					lwd.legend = tm_legend(),
					lwd.chart = tm_chart_none(),
					lwd.free = NA,
					lty = tm_const(),
					lty.scale = tm_scale(),
					lty.legend = tm_legend(),
					lty.chart = tm_chart_none(),
					lty.free = NA,
					col_alpha = tm_const(),
					col_alpha.scale = tm_scale(),
					col_alpha.legend = tm_legend(),
					col_alpha.chart = tm_chart_none(),
					col_alpha.free = NA,
					linejoin = "round",
					lineend = "round",
					plot.order = tm_plot_order("lwd", reverse = TRUE, na.order = "bottom"),
					zindex = NA,
					group = NA,
					group.control = "check",
					popup.vars = NA,
					popup.format = tm_label_format(),
					hover = NA,
					id = "",
					options = opt_tm_lines(),
					...) {

	args_called = names(rlang::call_match()[-1])
	args = list(...)

	layer_fun = if ("called_from" %in% names(args)) {
		args$called_from
	} else {
		"tm_lines"
	}

	if (any(v3_only("tm_lines") %in% names(args))) {

		v3_start_message()
		if (!("style" %in% names(args))) {
			if (!"breaks" %in% names(args)) {
				style = "pretty"
			} else {
				style = "fixed"
			}
		} else {
			style = args$style
		}

		v3_list_init()
		if (length(style) > 1) {
			style = style[1]
			.TMAP$v3_list$mult = TRUE
		}
		# v3 visual variable: col
		col.scale.args = c(list(n = v3_impute(args, "n", 5),
								style = style,
								style.args = v3_impute(args, "style.args", list())),
							if (style %in% c("cont", "log10")) {
								c({
									if (!is.null(args$breaks)) {
										if (length(args$breaks) > 2) {
											list(ticks = v3_impute(args, "breaks", NULL, "ticks"))
										} else {
											list(ticks = v3_impute(args, "breaks", NULL, "limits"))
										}
									} else {
										list()
									}
								},
								list(outliers.trunc = c(TRUE, FALSE)))
							} else {
								list(breaks = v3_impute(args, "breaks", NULL),
									 interval.closure = v3_impute(args, "interval.closure", "left"),
									 drop.levels = v3_impute(args, "drop.levels", FALSE))
							},
							list(midpoint = v3_impute(args, "midpoint", NULL),
								 as.count = v3_impute(args, "as.count", NA),
								 values = v3_impute(args, "palette", NA, "values"),
								 values.repeat = !v3_impute(args, "stretch.palette", TRUE, "values.repeat"),
								 values.range = v3_impute(args, "contrast", NA, "values.range"),
								 values.scale = 1,
								 value.na = v3_impute(args, "colorNA", NA, "value.na"),
								 value.null = v3_impute(args, "colorNULL", NA, "value.null"),
								 value.neutral = NA,
								 labels = v3_impute(args, "labels", NULL),
								 label.na = v3_impute(args, "textNA", "Missing", "label.na"),
								 label.null = NA,
								 label.format = v3_impute(args, "legend.format", list(), "label.format")))

		col.scale.args$fun_pref = if (style == "cat") {
			"categorical"
		} else if (style %in% c("fixed", "sd", "equal", "pretty", "quantile",
								"kmeans", "hclust", "bclust", "fisher", "jenks",
								"dpih", "headtails", "log10_pretty")) {
			"intervals"
		} else if (style == "cont") {
			"continuous"
		} else if (style == "log10") {
			"continuous_log"
		} else if (style == "order") {
			"rank"
		} else {
			stop("unknown style")
		}
		if (!(layer_fun == "qtm" && (!"col" %in% names(args)))) {
			if ("style" %in% names(args)) {
				v3_tm_scale_instead_of_style(style, scale_fun = col.scale.args$fun_pref, vv = "col", layer_fun = layer_fun, arg_list = v3_list_get())
			} else {
				v3_tm_scale(scale_fun = "", vv = "col", layer_fun = layer_fun, arg_list = v3_list_get())
			}
		}

		col.scale = do.call("tm_scale", args = col.scale.args)



		if ("alpha" %in% names(args)) {
			col_alpha = args$alpha
			v3_message_col_alpha(layer_fun = layer_fun, orig = "alpha")

		}


		v3_list_init()
		if ("legend.col.show" %in% names(args) && !args$legend.col.show) {
			v3_tm_legend_hide(layer_fun, arg = "legend.col.show", vv = "col")
			lwd.legend = tm_legend_hide()
		} else {
			col.legend.args = alist(title = v3_impute(args, "title.col", NA, "title"),
									na.show = v3_impute(args, "showNA", NA),
									format = v3_impute(args, "legend.format", list(), "format"),
									orientation = ifelse(v3_impute(args, "legend.col.is.portrait", TRUE, "orientation"), "portrait", "landscape"),
									reverse = v3_impute(args, "legend.reverse", FALSE, "reverse"))
			col.legend = do.call("tm_legend", col.legend.args)
			v3_tm_legend(fun = layer_fun, vv = "col", arg_list = v3_list_get())
		}

		# v3 visual variable: lwd
		# v3 visual variable: size
		v3_list_init()
		lwd.scale.args = list(ticks = v3_impute(args, "lwd.legend", NULL, "ticks"),
							  labels = v3_impute(args, "lwd.legend.labels", NULL, "labels"),
							  values.scale = v3_impute(args, "scale", 1, "values.scale"),
							  fun_pref = "continuous")
		if ("lwd" %in% args_called) v3_tm_scale(scale_fun = "continuous", vv = "lwd", layer_fun = layer_fun, arg_list = v3_list_get())
		lwd.scale = do.call("tm_scale", args = lwd.scale.args)

		v3_list_init()
		if ("legend.lwd.show" %in% names(args) && !args$legend.lwd.show) {
			v3_tm_legend_hide(layer_fun, arg = "legend.lwd.show", vv = "lwd")
			lwd.legend = tm_legend_hide()
		} else {
			lwd.legend.args = alist(title = v3_impute(args, "title.lwd", NA, "title"),
								   na.show = v3_impute(args, "showNA", NA),
								   format = v3_impute(args, "legend.format", list(), "format"),
								   orientation = ifelse(v3_impute(args, "legend.lwd.is.portrait", FALSE), "portrait", "landscape"),
								   reverse = v3_impute(args, "legend.lwd.reverse", FALSE))
			if ("lwd" %in% args_called) v3_tm_legend(fun = layer_fun, vv = "lwd", arg_list = v3_list_get())
			lwd.legend = do.call("tm_legend", lwd.legend.args)
		}


		if ("legend.hist" %in% names(args) && args$legend.hist) {
			col.chart = tm_chart_histogram()
			v3_tm_chart_hist(layer_fun = layer_fun, vv = "col", arg = "legend.hist")

			# to do: histogram title
		}
	}

	# unused arguments: typos?
	unused = setdiff(names(args), c(v3_only("tm_lines"), "called_from"))

	if (length(unused)) {
		message_layer_unused_args(layer_fun, unused)
	}

	# make sure required options are there
	options = complete_options(options, opt_tm_lines())


	tm_element_list(tm_element(
		layer = "lines",
		trans.fun = tmapTransLines,
		trans.aes = list(),
		trans.args = options$trans.args,
		trans.apply_to = "this",
		mapping.aes = list(col = tmapScale(aes = "col",
										   value = col,
										   scale = col.scale,
										   legend = col.legend,
										   chart = col.chart,
										   free = col.free),
						   lwd = tmapScale(aes = "lwd",
						   				value = lwd,
						   				scale = lwd.scale,
						   				legend = lwd.legend,
						   				chart = lwd.chart,
						   				free = lwd.free),
						   lty = tmapScale(aes = "lty",
						   				value = lty,
						   				scale = lty.scale,
						   				legend = lty.legend,
						   				chart = lty.chart,
						   				free = lty.free),
						   col_alpha = tmapScale(aes = "col_alpha",
						   					  value = col_alpha,
						   					  scale = col_alpha.scale,
						   					  legend = col_alpha.legend,
						   					  chart = col_alpha.chart,
						   					  free = col_alpha.free)),

		gpar = tmapGpar(fill = NA,
						col = "__col",
						shape = NA,
						size = NA,
						fill_alpha = NA,
						col_alpha = "__col_alpha",
						pattern = NA,
						lty = "__lty",
						lwd = "__lwd",
						linejoin = linejoin,
						lineend = lineend),
		tpar = tmapTpar(),
		plot.order = plot.order,
		mapping.args = options$mapping.args,
		zindex = zindex,
		group = group,
		group.control = group.control,
		popup.vars = popup.vars,
		popup.format = popup.format,
		hover = hover,
		id = id,
		subclass = c("tm_aes_layer", "tm_layer")))
}

#' @rdname tm_lines
#' @param lines.only should only line geometries of the shape object (defined in [tm_shape()]) be plotted, or also other geometry types (like polygons)? By default `"ifany"`, which means `TRUE` in case a geometry collection is specified.
#' @param hitbox Controls whether an invisible interaction layer with a
#'   larger clickable area (\"hitbox\") is added on top of the lines.
#'
#'   This can improve click and popup behaviour for thin or densely packed
#'   lines by increasing the effective mouse interaction width.
#'
#'   Possible values:
#'   \describe{
#'     \item{"none"}{No additional hitbox layer is added. Lines are clickable
#'     only at their visible width.}
#'
#'     \item{"pmaxX"}{Ensures a minimum interaction width of \code{X} pixels.
#'     The hitbox width is computed as \code{pmax(line_width, X)}.
#'     For example, \code{"pmax8"} guarantees at least 8 pixels.}
#'
#'     \item{"timesX"}{Multiplies the visible line width by \code{X} to compute
#'     the interaction width. For example, \code{"times2"} doubles the
#'     clickable width.}
#'
#'     \item{"auto"}{Automatically: \code{"pmax8"} by default if and only if interactive features are enabled (popup or hover) and lines are thin (median line width < 4) and there are less than 10000 features. Otherwise, \code{"none"} }
#'   }
#'
#'   Adding a hitbox improves usability for thin lines but may reduce
#'   performance for very large datasets, as an additional invisible
#'   layer is rendered.
#' @export
opt_tm_lines = function(lines.only = "ifany", hitbox = "auto") {
	list(trans.args = list(lines.only = lines.only),
		 mapping.args = list(hitbox = hitbox))
}

#' Group components
#'
#' @param group_id id of the component group. By default set to `""`, which will apply to all components. There are two other options. 1) To use the same (self-chosen) name that corresponds to the `group_id` argument of a component function, such as [tm_legend()] and [tm_title()]. 2) To specify one (or more) component function names, e.g. `"tm_legend"` or `c("tm_scalebar", "tm_compass")`.
#' @param position The position specification of the components in this group: an object created with `tm_pos_in()` or `tm_pos_out()`. Or, as a shortcut, a vector of two values, specifying the x and y coordinates. The first is `"left"`, `"center"` or `"right"` (or upper case, meaning tighter to the map frame), the second `"top"`, `"center"` or `"bottom"`. Numeric values are also supported, where 0, 0 means left bottom and 1, 1 right top. See also `r .doc_see_vig("adv_positions")`.
#' @param stack stacking `"horizontal"` or `"vertical"`
#' @param frame_combine put frame around all components that are drawn on the same location. Whether a frame is drawn is still decided by the `frame` argument of the  'main' (first) component.
#' @param equalize in case `frame_combine` is `FALSE`, should the separate frames be equalized, i.e. have the same width (when stacked vertically) or height (when stacked horizontally)?
#' @param resize_as_group in case a component if rescaled because of the limited space, rescale the other components proportionally?
#' @param stack_margin Margin between components. One or two values. If two values are provided, the first is the horizontal stacking margin and the second the vertical. The unit is text line heights.
#' @param offset Offset margin between frame and the components block. One or two values. If two values are provided, the first is the horizontal offset and the second the vertical. The unit is text line heights.
#' @param frame Should a frame be drawn? By default `TRUE` for legends, charts and insets, and `FALSE` otherwise.
#' @param frame.color frame color
#' @param frame.alpha frame alpha transparancy
#' @param frame.lwd frame line width
#' @param frame.r Radius of the rounded frame corners. 0 means no rounding.
#' @param bg Background color the components block. Is usually set in each component function, but if specified here, it will overwrite them.
#' @param bg.color Background color the components block. Is usually set in each component function, but if specified here, it will overwrite them.
#' @param bg.alpha Background alpha transparency of the components block. Is usually set in each component function, but if specified here, it will overwrite them.
#' @seealso `r .doc_see_also_comp()`
#' @return A [`tmap-element`]
#' @export
tm_components = function(
		group_id = "",
		position,
		stack,     # was 'stack' in each tm_legend or tm_<comp> function
		frame_combine,
		equalize,
		resize_as_group,        # from tm_legend/tm_<comp>.
		stack_margin,     # margin between components
		offset,            # offset margin
		frame ,
		frame.color,
		frame.alpha,
		frame.lwd,
		frame.r,
		bg,
		bg.color,
		bg.alpha) {
	args = lapply(as.list(rlang::call_match()[-1]), eval, envir = parent.frame())
	args$called = names(args)
	optname = paste0("component_", paste(group_id, collapse = "^"))
	x = structure(list(args), names= optname)
	do.call(tm_options, x)
}

#' Set the position of map components
#'
#' Set the position of map components, such as legends, title, compass, scale bar, etc.
#' `tm_pos()` is the function to position these components:
#' `tm_pos_out()` places the components outside the map area, `tm_pos_in()` inside the map area, and `tm_pos_on_top()` on top of the map
#' Each `position` argument of a map layer or component should be specified with
#' one of these functions.
#' The functions `tm_pos_auto_out()` and `tm_pos_auto_in()`
#' are used to set the components automatically, and should be used
#' via [tmap_options()]. See Details how the positioning works.
#'
#' @param cell.h,cell.v The plotting area is overlaid with a 3x3 grid, of which
#'   the middle grid cell is the map area. Components can be drawn into any cell.
#'   `cell.h` specifies the horizontal position (column) can take values
#'   `"left"`, `"center"`, and `"right"`. `cell.v` specifies the vertical position
#'   (row) and can take values `"top"`, `"center"`, and `"bottom"`.
#'   See details for a graphical explanation.
#' @param pos.h,pos.v The position of the component within the cell.
#'   The main options for `pos.h` are `"left"`, `"center"`, and `"right"`.
#'   For `cell.v` these are `"top"`, `"center"`, and `"bottom"`.
#'   These options can also be provided in upper case; in that case there is no offset
#'   (see the tmap option `component.offset`). Also numbers between 0 and 1 can be provided,
#'   which determine the position of the component inside the cell (with (0,0) being left bottom).
#'   The arguments `just.h` and `just.v` determine the justification point.
#' @param align.h,align.v The alignment of the component in case multiple components are stacked.
#'   When they are stacked horizontally, `align.v` determines how components
#'   that are smaller in height than the available height (determined by the outer.margins
#'   if specified and otherwise by the highest component) are justified:
#'   `"top"`, `"center"`, or `"bottom"`. Similarly, `align.h` determines how components
#'   are justified horizontally when they are stacked vertically: `"left"`, `"center"`, or `"right"`.
#' @param just.h,just.v The justification of the components.
#'   Only used in case `pos.h` and `pos.v` are numbers.
#' @details
#' `tm_pos_in()` sets the position of the component(s) inside the maps area,
#' which is equivalent to the center-center cell (in case there are facets,
#' these are all drawn in this center-center cell).
#'
#' `tm_pos_out()` sets the position of the component(s) outside the map.
#'
#' `tm_pos_on_top()` is the same as `tm_pos_out`, but with the cell set to the center cell
#' It may be therefore seem similar to `tm_pos_in()`, but with an essential difference:
#' `tm_pos_in()` takes the map frame into account whereas `tm_pos_on_top()` does not.
#' #'
#' The amount of space that the top and bottom rows, and left and right columns
#' occupy is determined by the [tm_layout()] arguments `meta.margins` and `meta.auto_margins`.
#' The former sets the relative space of the bottom, left, top, and right side.
#' In case these are set to `NA`, the space is set automatically based on 1)
#' the maximum relative space specified by `meta.auto_margins` and 2)
#' the presence and size of components in each cell.
#' For instance, if there is one landscape oriented legend in the center-bottom cell,
#' then the relative space of the bottom row is set to the height of that legend
#' (given that it is smaller than the corresponding value of `meta.auto_margins`),
#' while the other four sides are set to 0.
#'
#' `tm_pos_auto_out()` is more complex: the `cell.h` and `cell.v` arguments should be set
#' to one of the four corners. It does not mean that the components are drawn in a corner.
#' The corner represents the sides of the map that the components are drawn.
#' By default, legends are drawn either at the bottom or on the right-side of
#' the map by default (see `tmap_options("legend.position")`).
#' Only when there are row- and column-wise legends and a general legend (using [tm_facets_grid()]),
#' the general legend is drawn in the corner, but in practice this case will be rare.
#'
#' The arguments `pos.h` and `pos.v` determine where the components are drawn within the cell.
#' Again, with `"left"`, `"center"`, and `"right"` for `pos.h`
#' and `"top"`, `"center"`, and `"bottom"` for `pos.v`.
#' The values can also be specified in upper-case, which influences the offset
#' with the cell borders, which is determined by tmap option `component.offset`.
#' By default, there is a small offset when components are drawn inside and no
#' offset when they are drawn outside or with upper-case.
#'
#' `tm_pos_auto_in()` automatically determines `pos.h` and `pos.v` given the
#' available space inside the map. This is similar to the default positioning in tmap3.
#'
#' In case multiple components are draw in the same cell and the same position
#' inside that cell, they are stacked (determined which the `stack` argument in
#' the legend or component function). The `align.h` and `align.v` arguments
#' determine how these components will be justified with each other.
#'
#' Note that legends and components may be different for a facet row or column.
#' This is the case when [tm_facets_grid()] or [tm_facets_stack()] are applied
#' and when scales are set to free (with the `.free` argument of the map layer functions).
#' In case a legends or components are draw row- or column wise, and the position
#' of the legends (or components) is right next to the maps, these legends
#' (or components) will be aligned with the maps.
#'
#' @seealso \href{https://r-tmap.github.io/tmap/articles/adv_positions}{Vignette about positioning}
#' @export
tm_pos = function(cell.h, cell.v, pos.h, pos.v, align.h, align.v, just.h, just.v) {
	args = lapply(as.list(rlang::call_match()[-1]), eval, envir = parent.frame())
	args$called = names(args)
	args$type = "out"
	structure(args, class = "tm_pos")
}

#' @export
#' @rdname tm_pos
tm_pos_in = function(pos.h, pos.v, align.h, align.v, just.h, just.v) {
	args = lapply(as.list(rlang::call_match()[-1]), eval, envir = parent.frame())
	args$called = names(args)
	args$type = "in"
	structure(args, class = "tm_pos")
}


#' @export
#' @rdname tm_pos
tm_pos_out = function(cell.h, cell.v, pos.h, pos.v, align.h, align.v, just.h, just.v) {
	args = lapply(as.list(rlang::call_match()[-1]), eval, envir = parent.frame())
	args$called = names(args)
	args$type = "out"
	structure(args, class = "tm_pos")
}

#' @export
#' @rdname tm_pos
tm_pos_on_top = function(pos.h, pos.v, align.h, align.v, just.h, just.v) {
	args = lapply(as.list(rlang::call_match()[-1]), eval, envir = parent.frame())
	args$called = names(args)
	args$cell.h = "center"
	args$cell.v = "center"
	args$type = "out"
	structure(args, class = "tm_pos")
}

#' @export
#' @rdname tm_pos
tm_pos_auto_out = function(cell.h, cell.v, pos.h, pos.v, align.h, align.v, just.h, just.v) {
	args = lapply(as.list(rlang::call_match()[-1]), eval, envir = parent.frame())
	args$called = names(args)
	args$type = "autoout"
	structure(args, class = "tm_pos")
}

#' @export
#' @rdname tm_pos
tm_pos_auto_in = function(align.h, align.v, just.h, just.v) {
	args = lapply(as.list(rlang::call_match()[-1]), eval, envir = parent.frame())
	args$called = names(args)
	args$type = "autoin"
	structure(args, class = "tm_pos")
}

#' @export
#' @rdname tmapGridLeaflet
tmapLeafletDataPlot.tm_data_polygons = function(a, shpTM, dt, pdt, popup.format, hdt, idt, gp, bbx, facet_row, facet_col, facet_page, id, pane, group, glid, o, ...) {
	# popup.title data (ptdt) and popup layout are passed by the tmap core via
	# `...` rather than as named formals, so that this method's signature stays
	# identical to the tmapLeafletDataPlot generic (S3 consistency) and so that
	# extension methods that do not handle them keep working unchanged.
	dots = list(...)
	ptdt = dots$ptdt
	popup.layout = dots$popup.layout

	lf = get_lf(facet_row, facet_col, facet_page)

	rc_text = frc(facet_row, facet_col)

	res = select_sf(shpTM, dt)
	shp = res$shp
	if (o$crs_leaflet$crsClass  == "L.CRS.Simple") {
		shp = sf::st_set_crs(shp, NA)
	}

	dt = res$dt

	idt_null = is.null(idt)

	if (!idt_null) {
		idt = idt$id[match(dt$tmapID__, idt$tmapID__)]
	} else {
		idt = sprintf("%07d", dt$tmapID__)
	}
	if (!is.null(hdt)) {
		hdt = hdt$hover[match(dt$tmapID__, hdt$tmapID__)]
		hdt = lapply(hdt, htmltools::HTML, FUN.VALUE = character(1))
	}

	# popup title: explicit popup.title (ptdt) takes precedence; otherwise fall
	# back to the previous behaviour (id, or hover when no id).
	if (!is.null(ptdt)) {
		pttl = ptdt$title[match(dt$tmapID__, ptdt$tmapID__)]
	} else {
		pttl = NULL
	}

	if (is.null(pdt)) {
		popups = NULL
	} else {
		mtch = match(dt$tmapID__, pdt$tmapID__)
		pdt = pdt[mtch][, tmapID__ := NULL]

		if (!is.null(pttl)) {
			popups = view_format_popups(id = pttl, titles = names(pdt), values = pdt, format = popup.format, layout = popup.layout)
		} else if (idt_null && !is.null(hdt)) {
			popups = view_format_popups(id = hdt, titles = names(pdt), values = pdt, format = popup.format, layout = popup.layout)
		} else if (idt_null && is.null(hdt)) {
			popups = view_format_popups(id = NULL, titles = names(pdt), values = pdt, format = popup.format, layout = popup.layout)
		} else {
			popups = view_format_popups(id = idt, titles = names(pdt), values = pdt, format = popup.format, layout = popup.layout)
		}
	}


	gp = impute_gp(gp, dt)
	gp = rescale_gp(gp, o$scale_down)
	gp = gp_to_lpar(gp, mfun = "Polygons", rename_prop = FALSE)

	interactive = (!is.null(pdt) || !is.null(hdt))

	opt = leaflet::pathOptions(interactive = interactive, pane = pane)

	o$use_WebGL = impute_webgl(o$use_WebGL, dt, supported = c("fill", "col"), checkif = list(lty = "solid"), type = "polygons", hover = !is.null(hdt), popup = !is.null(pdt), crs_class = o$crs_leaflet$crsClass)

	if (o$use_WebGL) {
		idt = rep({if (is.null(idt))dt$tmapID__[1] else idt[1]}, 2) |>
			submit_labels("polygonsGL", pane, group)

		shp2 = sf::st_sf(id = seq_along(shp), geom = shp)
		shp3 = sf_expand(shp2)

		shp3lines = suppressWarnings(sf::st_cast(shp3, "LINESTRING"))
		gp3 = lapply(gp, function(gpi) {if (length(gpi) == 1) gpi else gpi[shp3$split__id]})
		popups2 = popups[shp3$split__id]

		# opacity channel from fill (e.g. "#FF000099") is ignored by addGlPolygons
		fill_alpha = split_alpha_channel(gp3$fill[1], alpha = gp3$fill_alpha[1])$opacity

		lf |>
			leafgl::addGlPolygons(data = shp3, layerId = idt[1], label = hdt,
								  color = gp3$fill, opacity = fill_alpha,
								  group = group, pane = pane, popup = popups2) %>%
			{if (gp3$lwd[1]!=0 && gp3$col[1] != "#00000000") leafgl::addGlPolylines(., data = shp3lines, color = gp3$col, opacity = gp3$col_alpha[1], weight = gp3$lwd[1]/4, pane = pane, group = group, layerId = idt[2]) else .} %>%
			blend_lf(a$blend, pane) %>%
			assign_lf(facet_row, facet_col, facet_page)
	} else {
		idt = (if (is.null(idt))dt$tmapID__ else idt) |>
			submit_labels("polygons", pane, group)

		lf %>%
			leaflet::addPolygons(data = shp, layerId = idt, label = hdt, color = gp$col, opacity = gp$col_alpha, fillColor = gp$fill, fillOpacity = gp$fill_alpha, weight = gp$lwd, options = opt, group = group, dashArray = gp$lty, popup = popups) %>%
			blend_lf(a$blend, pane) %>%
			assign_lf(facet_row, facet_col, facet_page)
	}
	NULL
}


#' @export
#' @rdname tmapGridLeaflet
tmapLeafletDataPlot.tm_data_fill = function(a, shpTM, dt, pdt, popup.format, hdt, idt, gp, bbx, facet_row, facet_col, facet_page, id, pane, group, glid, o, ...) {
	NextMethod()
}

#' @export
#' @rdname tmapGridLeaflet
tmapLeafletDataPlot.tm_data_borders = function(a, shpTM, dt, pdt, popup.format, hdt, idt, gp, bbx, facet_row, facet_col, facet_page, id, pane, group, glid, o, ...) {
	NextMethod()
}

pane_name = function(id) {
	paste0("tmap", sprintf("%03g", id))
}

process_components = function(cdt, o) {
	gs = tmap_graphics_name()


	# update added legends
	is_added_leg = vapply(cdt$comp, inherits, logical(1), "tm_add_legend")

	if (any(is_added_leg)) {
		cdt$comp[is_added_leg] = lapply(cdt$comp[is_added_leg], tmapAddedLegend, o = o)
	}

	# the option component.position is not a general fall-back option (in case the component is undefined for a mode),
	# but used for imputation of not-specified tm_pos arguments
	# in case there is no specific .position argument for the current mode specified, use the 'in' position for the time being
	# the missing methods for the component is catched later via the missing tmap<gs>CompPrepare method
	cdt$comp = lapply(cdt$comp, function(l) {
		if ("in" %in% names(l$position)) l$position = l$position[["in"]]
		l
	})

	cdt$class = sapply(cdt$comp, function(l) l$position$type)
	cdt$cell.h = sapply(cdt$comp, function(l) {x = l$position$cell.h; if (is.null(x)) NA_character_ else x})
	cdt$cell.v = sapply(cdt$comp, function(l) {x = l$position$cell.v; if (is.null(x)) NA_character_ else x})
	cdt$pos.h = sapply(cdt$comp, function(l) {x = l$position$pos.h; if (is.null(x)) NA_character_ else x})
	cdt$pos.v = sapply(cdt$comp, function(l) {x = l$position$pos.v; if (is.null(x)) NA_character_ else x})
	cdt$z = sapply(cdt$comp, function(l) {x = l$z; if (is.null(x)) NA_integer_ else x})
	cdt[, zauto := .I]

	if (gs != "Grid") {
		cdt[class == "out", class := "in"]
		cdt[class == "autoout", class := "in"]
	}

	funP = paste0("tmap", gs, "CompPrepare")
	funW = paste0("tmap", gs, "CompWidth")
	funH = paste0("tmap", gs, "CompHeight")

	cdt$comp = lapply(cdt$comp, function(comp) do.call(funP, list(comp = comp, o = o)))

	# split components into facets (if needed)
	is_multi = sapply(cdt$comp, inherits, what = "tm_multi_comp")

	if (any(is_multi)) {
		cdt_single = cdt[!is_multi, ]
		cdt_multi = cdt[is_multi, ]

		allby_temp = as.data.table(do.call(expand.grid, lapply(o$fn, seq_len)))
		names(allby_temp) = paste0("by", 1:3, "__")

		cdt_rows = data.table::rbindlist(lapply(seq_len(nrow(cdt_multi)), function(i) {
			if (cdt_multi$class[i] %in% c("out", "autoout")) {
				if (length(cdt_multi$comp[[i]]) > 1) {
					warning("Component is specified for multiple facets (e.g. title text), but since placement is outside, only the first is used", call. = FALSE)				}
				cdt_multi$comp[[i]] = cdt_multi$comp[[i]][[1]]
				cdt_multi[i, ]
			} else {
				cbind(allby_temp,
					  comp = rep(cdt_multi$comp[[i]], length.out = nrow(allby_temp)),
					  cdt_multi[i, setdiff(names(cdt_multi), c(names(allby_temp), "comp")), with = FALSE])
			}
		}))
		cdt = rbind(cdt_single, cdt_rows)
	}


	# filter components that are not shown (e.g. not implemented), which is determined in the Prepare step
	cdt$show = sapply(cdt$comp, function(l) l$show)
	cdt = cdt[cdt$show,][,show:=NULL]


	cdt$comp = lapply(cdt$comp, function(comp) do.call(funW, list(comp = comp, o = o)))
	cdt$comp = lapply(cdt$comp, function(comp) do.call(funH, list(comp = comp, o = o)))

	cdt[, ':='(facet_row = character(), facet_col = character())]
	cdt$stack_auto = vapply(cdt$comp, function(l) {
		s = l$stack
		length(s) > 1
	}, FUN.VALUE = logical(1))
	cdt$stack = vapply(cdt$comp, function(l) {
		s = l$stack
		if (length(s) > 1 && "manual" %in% names(s)) s["manual"] else s[1]
	}, FUN.VALUE = character(1))


	getLW = function(x) sapply(x, function(y) {
		yW = y$Win
		yW %||% 0
	})
	getLH = function(x) sapply(x, function(y) {
		yH = y$Hin
		yH %||% 0
	})
	# attempt to determine margins
	cdt[, legW := getLW(comp)]
	cdt[, legH := getLH(comp)]

	if (any(is.na(cdt$z))) {
		cdt[is.na(z), z := zauto]
	}
	cdt[, zauto:=NULL]
	if (nrow(cdt)>0L) {
		data.table::setorder(cdt, "z")
	}


	cdt
}

process_components2 = function(cdt, o) {
	gs = tmap_graphics_name()




	if (o$type != "grid" && o$n > 1) {
		#if (o$nrows == 1 && o$ncols == 1)
		if (o$nrows == 1) {
			# -use by2 and not by1 when they form a row
			cdt[, by2__ := by1__]
			cdt[, by1__ := NA]
		}
	}

	stacks = o$legend.stack


	# place legends inside if needed
	#if (o$ncols > 1 && o$nrows > 1) {
	if (o$type == "wrap") {
		# all free legends inside
		if (any((!is.na(cdt$by1) | !is.na(cdt$by2)) & cdt$class == "out")) cli::cli_inform("{.field {.fun tm_pos_out}} Legends per facets are placed outside. To align them with the facets, consider {.fun tm_facets_stack} or {.fun tmap_arrange}.")
		cdt[(!is.na(by1__) | !is.na(by2__)) & class == "autoout", ':='(class = "in")]
	} else if (o$type == "grid") {
		# all free-per-facet legends inside
		cdt[!is.na(by1__) & !is.na(by2__) & class == "autoout", ':='(class = "in")]
	}



	# update auto position (for 'all', 'rows', 'columns' legends)
	cdt[is.na(by1__) & is.na(by2__) & class == "autoout", ':='(cell.h = o$legend.position.all$cell.h, cell.v = o$legend.position.all$cell.v)]
	cdt[!is.na(by1__) & is.na(by2__) & class == "autoout", ':='(cell.h = o$legend.position.sides$cell.h, cell.v = "by")]
	cdt[is.na(by1__) & !is.na(by2__) & class == "autoout", ':='(cell.h = "by", cell.v = o$legend.position.sides$cell.v)]

	# in case tm_pos is called to update the 'non-by' cell
	cdt$calledCV = vapply(cdt$comp, function(comp) "cell.v" %in% comp$position$called, FUN.VALUE = logical(1))
	cdt$calledCH = vapply(cdt$comp, function(comp) "cell.h" %in% comp$position$called, FUN.VALUE = logical(1))
	cdt[!is.na(by1__) & is.na(by2__) & !calledCV, ':='(cell.v = "by")]
	cdt[is.na(by1__) & !is.na(by2__) & !calledCH, ':='(cell.h = "by")]

	cdt[!is.na(by1__) & is.na(by2__) & !calledCV, ':='(stack = ifelse(stack_auto, stacks["per_row"], stack))]
	cdt[is.na(by1__) & !is.na(by2__) & !calledCH, ':='(stack = ifelse(stack_auto, stacks["per_col"], stack))]

	cdt$calledCV = NULL
	cdt$calledCH = NULL

	cdt[is.na(by1__) & is.na(by2__) & class == "autoout", ':='(stack = ifelse(stack_auto, ifelse(cell.h == "center", stacks["all_col"], ifelse(cell.v == "center", stacks["all_row"], stacks["all"])), stack))]
	cdt[!is.na(by1__) & is.na(by2__) & class == "autoout", ':='(stack = ifelse(stack_auto, stacks["per_row"], stack))]
	cdt[is.na(by1__) & !is.na(by2__) & class == "autoout", ':='(stack = ifelse(stack_auto, stacks["per_col"], stack))]


	cdt[class == "autoout", class := "out"]


	lai = which(cdt$class == "autoin")
	if (length(lai)) {
		cdt[class == "autoin", ":="(pos.h = ifelse(is.na(pos.h), o$legend.autoin.pos[1], pos.h),
									pos.v = ifelse(is.na(pos.v), o$legend.autoin.pos[2], pos.v),
									class = "in")]
	}

	vby = any(cdt$cell.v == "by" & !is.na(cdt$cell.v))
	hby = any(cdt$cell.h == "by" & !is.na(cdt$cell.h))

	toC = function(x) {
		paste(x, collapse = "_")
	}


	# # manual outside legends -2 is top or left, -1 is bottom or right
	# cdt[class %in% c("autoout", "out"), ':='(facet_row =
	# 		ifelse(cell.v == "center", ifelse(vby, "1", toC(1:o$nrows)),
	# 		ifelse(cell.v == "by", as.character(by1__),
	# 		ifelse(cell.v == "top", as.character(-2), as.character(-1)))),
	# 	facet_col =
	# 		ifelse(cell.h == "center", ifelse(hby, "1", toC(1:o$ncols)),
	# 		ifelse(cell.h == "by", as.character(by2__),
	# 		ifelse(cell.h == "left", as.character(-2), as.character(-1)))))]
	# manual outside legends -2 is top or left, -1 is bottom or right
	cdt[class %in% c("autoout", "out"), ':='(facet_row =
											 	ifelse(cell.v == "center", toC(1:o$nrows),
											 		   ifelse(cell.v == "by", as.character(by1__),
											 		   	   ifelse(cell.v == "top", as.character(-2), as.character(-1)))),
											 facet_col =
											 	ifelse(cell.h == "center", toC(1:o$ncols),
											 		   ifelse(cell.h == "by", as.character(by2__),
											 		   	   ifelse(cell.h == "left", as.character(-2), as.character(-1)))))]


	cdt
}

step4_plot = function(tm, vp, return.asp, show, in.shiny, knit, knit_opts, args) {
	tmx = tm$tmo
	o = tm$o
	aux = tm$aux
	cmp = tm$cmp
	prx = tm$prx

	# check if mode is the same as in step1
	# tmap_mode() may be executed in between (#1082)
	gs = tmap_graphics_name()

	if (o$name != gs) cli::cli_abort("tmap mode changed during execution; did you run {.code tmap_mode()} inside a shiny app?")

	bbo = o$bbox

	# shortcut if no data layers are used, but only a tm_shape
	if (length(tmx)) {
		bbx_def = vapply(o$main, function(mid) {
			"bbx" %in% names(tmx[[mid]])
		}, FUN.VALUE = logical(1))
		if (any(bbx_def)) {
			bbxs = lapply(o$main[which(bbx_def)], FUN = function(mid) {
				tmx[[mid]]$bbx
			})
			bbm = stm_merge_bbox(bbxs)
		} else {
			bbm = NULL
		}
	} else {
		bbm = NULL
	}

	# remove empty data layers
	any_groups = (length(tmx) > 0L)

	tmx = lapply(tmx, function(tmxi) {
		tmxi$layers = lapply(tmxi$layers, function(tml) {
			empt = vapply(tml$shpDT$shpTM, function(sdt) {
				length(sdt$shp) == 0L
			}, FUN.VALUE = logical(1))
			if (all(empt)) NULL else tml
		})
		empt = vapply(tmxi$layers, is.null, logical(1))

		if (all(empt)) {
			NULL
		} else {
			tmxi$layers = tmxi$layers[!empt]
			tmxi
		}
	})
	empt = vapply(tmx, is.null, logical(1))
	tmx = tmx[!empt]


	any_data_layer = (length(tmx) > 0L)

	# split tm in case of as.layers in tm_facets
	# TODO

	# function to subset data
	get_dt = function(dt, by1, by2, by3, remove_by = FALSE) {
		b = list(by1, by2, by3)
		bynames = intersect(names(dt), paste0("by", 1:3, "__"))
		byids = as.integer(substr(bynames, 3, 3))

		sel = rep(TRUE, nrow(dt))
		if (length(bynames)) {
			for (i in 1:length(bynames)) {
				sel = sel & (is.na(dt[[bynames[i]]]) | dt[[bynames[i]]] %in% b[[byids[i]]])
			}
		}
		if (remove_by) {
			nms = setdiff(names(dt), bynames)
			dt[which(sel), nms, with = FALSE]
		} else {
			dt[which(sel),]
		}
	}

	if (o$as.layers) {
		bys = do.call(expand.grid, lapply(o$fn, seq, from = 1))

		bys_labs = lapply(o$fl, function(x) if (is.null(x)) "" else x)
		labs = do.call(mapply, c(list(FUN = paste0, USE.NAMES = FALSE), mapply("[", bys_labs, bys, SIMPLIFY = FALSE, USE.NAMES = FALSE)))
		# to do: take labels from o$fl
		# assign to tml$group
		# disable multiple basemaps

		e = environment()
		any_radio = FALSE
		tmx = lapply(tmx, function(tmxi) {
			lys = mapply(function(tml, tml_name) {
				lys2 = lapply(1:nrow(bys), function(i) {
					tml$mapping_dt = get_dt(tml$mapping_dt, bys$Var1[i], bys$Var2[i], bys$Var3[i], remove_by = TRUE)
					tml$mapping_legend = lapply(tml$mapping_legend, function(ml) {
						ml = get_dt(ml, bys$Var1[i], bys$Var2[i], bys$Var3[i], remove_by = TRUE)
						fixed = (nrow(ml) == 1L && tml$group.control == "check")
						for (j in 1:nrow(ml)) {
							id_old = ml$legnr[j]
							leg_old = .TMAP$legs[[ml$legnr[j]]]
							leg_old$layerId = paste0("fixed", ml$legnr[j])
							leg_old$group = labs[i]
							ml$legnr[j] = legend_save(leg_old)
						}
						ml
					})
					tml$group = labs[i]
					if (tml$group.control == "radio") assign("any_radio", value = TRUE, envir = e)
					tml$lid = tml$lid + i
					tml
				})
				names(lys2) = paste0(tml_name, "_", sprintf("%02d", 1:length(lys2)))
				lys2
			}, tmxi$layers, names(tmxi$layers), SIMPLIFY = FALSE, USE.NAMES = FALSE)
			tmxi$layers = do.call(c, lys)
			tmxi
		})

		if (length(aux) && any_radio) {
			multi_tile = any(vapply(aux, function(ai) {
				(ai$mapping.fun == "Tiles" && length(ai$args$server) > 1)
			}, FUN.VALUE = logical(1L), USE.NAMES = FALSE))
			if (multi_tile) cli::cli_inform("When {.code as.layers = TRUE} with radio buttons, it may be easier to have only one basemap, e.g. {.code + tm_basemap({.str OpenStreetMap})}")
		}

		o$n = 1
		o$fl = list(NULL,NULL,NULL)
		o$fn = rep(1L, 3)
		o$ncols = 1
		o$nrows = 1

	}



	# get name of graphics engine (for function names e.g. tmapGridInit)
	gs = tmap_graphics_name()

	o = prepreprocess_meta(o, vp)

	# get legends from layer data and put them in "components data.table" (cdt)

	if (o$legend.only) {
		cmp_sel = vapply(cmp, function(cp) {
			inherits(cp, c("tm_legend", "tm_add_legend"))
		}, FUN.VALUE = logical(1))
		cdt_cmp = if (any(cmp_sel)) {
			data.table::rbindlist(lapply(cmp[cmp_sel], function(cp) {
				data.table::data.table(by1__ = NA_integer_,
									   by2__ = NA_integer_,
									   by3__ = NA_integer_,
									   comp = list(cp))
			}))
		} else {
			data.table::data.table(by1__ = integer(0),
								   by2__ = integer(0),
								   by3__ = integer(0),
								   comp = list())
		}
	} else {
		cdt_cmp = if (length(cmp)) {
			data.table::rbindlist(lapply(cmp, function(cp) {
				data.table::data.table(by1__ = NA_integer_,
									   by2__ = NA_integer_,
									   by3__ = NA_integer_,
									   comp = list(cp))
			}))
		} else {
			data.table::data.table(by1__ = integer(0),
								   by2__ = integer(0),
								   by3__ = integer(0),
								   comp = list())
		}
	}

	cdt = if (any_data_layer) {
		cdt_legs = step4_plot_collect_legends(tmx)
		cdt_crts = step4_plot_collect_charts(tmx)
		data.table::rbindlist(list(cdt_legs, cdt_crts, cdt_cmp))
	} else {
		cdt_cmp
	}


	## place components top left
	if (o$legend.only) {
		cdt$comp = lapply(cdt$comp, function(cc) {
			cc$position = complete_options(tm_pos_in("left", "top"), o$legend.position)
			cc
		})
	}


	if (gs == "Grid" && is.null(vp) && show) {
		is_inset_tmap = sapply(cdt$comp, inherits, what = "tm_inset_tmap")
		if (any(is_inset_tmap)) {
			grid::grid.newpage()
			newpage = FALSE
		} else {
			newpage = TRUE
		}
	} else {
		newpage = TRUE
	}

	if (nrow(cdt)) cdt = process_components(cdt, o)

	# determine panel type, inner margins, and automatic comp placement
	if (!o$legend.only) {
		o = preprocess_meta(o, cdt)

		# add shape unit (needed for e.g. tm_scalebar)
		unit = ifelse(o$unit == "metric", "km", ifelse(o$unit == "imperial", "mi", o$unit))
		crs = get_option_class(o$crs_step4, "sf") #o$crs
		longlat = sf::st_is_longlat(crs)
	} else {
		o$npages = 1L
	}



	# function to get shape object
	get_shpTM = function(shpDT, by1, by2, by3) {
		b = list(by1, by2, by3)
		bynames = intersect(names(shpDT), paste0("by", 1:3, "__"))
		byids = as.integer(substr(bynames, 3, 3))

		sel = rep(TRUE, nrow(shpDT))
		if (length(bynames)) {
			for (i in 1L:length(bynames)) {
				sel = sel & shpDT[[bynames[i]]] %in% b[[byids[i]]]
			}
		}
		shpDT$shpTM[which(sel)]
	}


	# function to get bbox per facet, also take into account bbm (for groups without data-layers)
	get_bbox = function(by1, by2, by3) {
		bbxs = lapply(tmain, function(tmi) {
			shpTM = get_shpTM(tmi$shpDT, by1, by2, by3)
			mdt = get_dt(tmi$mapping_dt, by1, by2, by3)
			bbxs2 = lapply(shpTM, stm_bbox, tmapID = mdt$tmapID__, crs = crs)
			if (!is.null(bbm)) bbxs2 = c(list(bbm), bbxs2)
			bbx = stm_merge_bbox(bbxs2)
			if (is.na(bbx)) bbx else tmaptools::bb(bbx, asp.limit = 10)
		})
		bbm = stm_merge_bbox(bbxs)
		bbc = bbm

		bbe = bb_ext(bbc, o$inner.margins)

		bbe = crop_lat(bbe, crs, o$limit_latitude_3857)



		list(list(bbe))
	}

	# main group (that determines bounding box)
	# TODO take into account multiple main groups (see step1_rearrange and get_main_ids)


	if (o$legend.only) {
		d = NULL
	} else if (any_data_layer) {
		# create table with meta data for the facets (row, col id, bbox, asp)
		d = data.table::data.table(do.call(expand.grid, lapply(structure(o$nby, names = c("by1", "by2", "by3")), seq_len)))
		d[, i := seq_len(nrow(d))]
		grps = c("by1", "by2", "by3")[o$free.coords]

		grp_ids = as.integer(substr(names(tmx), 6, nchar(names(tmx))))
		mains_in_grp = intersect(o$main[!bbx_def], grp_ids)
		if (length(mains_in_grp) && !("inset" %in% names(o)) && !o$earth_bbox && is.null(bbo)) {
			lookup = match(mains_in_grp, grp_ids)
			tmain = unlist(unlist(tmx[lookup], recursive = FALSE, use.names = FALSE), recursive = FALSE, use.names = FALSE)
			d[, bbox:=do.call(get_bbox, as.list(.SD)), by = grps, .SDcols = c("by1", "by2", "by3")]
		} else {
			if (!is.null(bbo) || is.null(bbm)) {
				if (!is.null(bbo) && identical(bbo, "FULL") || (is.null(bbo) && is.null(bbm))) {
					bbm = full_bbox(crs)
				} else {
					if (inherits(bbo, "SpatExtent")) bbo = sf::st_bbox(bbo)
					bbm = tmaptools::bb(bbo, projection = crs)
				}
				if (o$earth_bbox) {
					bbm = bb_ext(bbm, o$inner.margins)
				}
			} else {
				bbm = sf::st_transform(bbm, crs = crs)
			}
			d[, bbox:=rep(list(bbm),nrow(d))]
		}
	} else {
		if (is.null(bbm)) {
			if (!is.null(bbo)) {
				bbm = tmaptools::bb(bbo, projection = crs)
			} else {
				bbm = full_bbox(crs)
			}
		} else {
			bbm = sf::st_transform(bbm, crs = crs)
		}

		bbm = crop_lat(bbm, crs, o$limit_latitude_3857)



		d = data.table::data.table(by1 = 1L, by2 = 1L, by3 = 1L, i = 1, bbox = list(bbm))
	}

	if (!o$legend.only) {
		d[, asp:=get_asp(bbox)]
		d = d[!is.na(asp)]


		if (o$type != "grid" && is.na(o$nrows) && is.na(o$ncols)) {
			# limit facets
			n_lim = limit_nx(o$n, o$facet.max)
			if (n_lim != o$n) {
				fn_lim = pmin(o$fn, n_lim)
				while(prod(fn_lim) > n_lim) {
					fn_lim[which.max(fn_lim)] = fn_lim[which.max(fn_lim)] - 1L
				}
				d = d[by1<= fn_lim[1] & by2<= fn_lim[2] & by3 <= fn_lim[3]]
				o$fl = mapply(function(a, b) a[1:b], o$fl, fn_lim, SIMPLIFY = FALSE)
				o$fn = fn_lim
				o$n = n_lim
			}
		}

		# not sure what this does:
		d[, bbox:=lapply(bbox, FUN = function(bbx) {
			if (!is.na(bbx) && !is.na(longlat) && longlat && !sf::st_is_longlat(bbx)) {
				sf::st_bbox(sf::st_transform(tmaptools::bb_poly(bbx), crs = 4326))
			} else {
				bbx
			}
		})]

		d[, units:=lapply(bbox, FUN = function(bbx) {
			if (is.na(bbx)) {
				list()
			} else {
				#browser()
				# if (!is.na(bbx) && !is.na(longlat) && longlat) {
				# 	latitude <- mean.default(bbx[c(2,4)])
				# 	bbxll <- c(xmin=0, ymin=latitude, xmax=1, ymax=latitude)
				# 	ad <- suppressWarnings({tmaptools::approx_distances(bbxll, projection=crs)})
				# 	to <- as.numeric(units::set_units(ad$hdist, units::as_units(unit), mode = "standard"))
				# } else {
				# 	ad <- suppressWarnings({tmaptools::approx_distances(bbx, projection=crs)})
				#
				# 	if (is.na(crs)) {
				# 		to <- ad$hdist
				# 	} else {
				# 		to <- as.numeric(units::set_units(units::set_units(1, attr(ad$hdist, "units")$numerator, mode = "standard"), units::as_units(unit), mode = "standard"))
				# 	}
				# }

				dst = distances_bbox_sides(bbx)
				units::set_units(dst, units::as_units(unit), mode = "standard")
				list(projection=crs, unit=unit, dst = dst, projected = !longlat)
			}
		})]

		# determine automatic position of inside comp
		if (!any(o$free.coords) && any(cdt$class == "autoin")) {
			shp = tmain[[1]]$shpDT$shpTM[[1]]$shp
			# TODO take into account multiple main shapes
			# TODO take use areas instead of coordinates for polygons
			if (inherits(shp, c("sf", "sfc"))) {
				bbx = d$bbox[[1]]
				co = sf::st_coordinates(sf::st_centroid(shp))

				xn = (co[,1]-bbx[1])/(bbx[3]-bbx[1])
				yn = (co[,2]-bbx[2])/(bbx[4]-bbx[2])
				cornerID = which.max(c(
					bl = min(sqrt((xn^2) + (yn^2)), na.rm = TRUE),
					tl = min(sqrt((xn^2) + ((1-yn)^2)), na.rm = TRUE),
					tr = min(sqrt(((1-xn)^2) + ((1-yn)^2)), na.rm = TRUE),
					br = min(sqrt(((xn-1)^2) + (yn^2)), na.rm = TRUE)))

				o$legend.autoin.pos = switch(names(cornerID), tl = c("left", "top"), tr = c("right", "top"), bl = c("left", "bottom"), br = c("right", "bottom"))
			} else {
				o$legend.autoin.pos = c("left", "top")
			}
		} else {
			o$legend.autoin.pos = c("left", "top")
		}
	}


	# calculate margins, number of rows and colums, etc.
	o = process_meta(o, d, cdt, aux)

	if (!o$legend.only) {
		# workaround to move panels to titles in view mode
		# TO DO: make this generic (e.g. component prep function?)
		if (gs != "Grid") {
			if (o$panel.type != "none") {
				cdt = rbindlist(c({if (nrow(cdt) == 0L) NULL else list(cdt)}, mapply(function(lab, i) {
					data.table::data.table(by1__ = i,
										   by2__ = NA_integer_,
										   by3__ = NA_integer_,
										   comp = list(impute_comp(tm_title(lab)[[1]], o)),
										   class=  "in",
										   cell.h = NA,
										   cell.v = NA,
										   pos.h = "left",
										   pos.v = "top",
										   z = 1,
										   facet_row = NA_character_,
										   facet_col = NA_character_,
										   stack_auto = TRUE,
										   stack = "vertical",
										   legW = 0,
										   legH = 0)
				}, o$panel.labels[[1]], seq_len(o$fn[1]), SIMPLIFY = FALSE)))
			}
			o$panel.type = "none"
		}

		o$ng = length(tmx)

		# determine row and col ids

		if (o$type == "grid") {
			d[, row := as.integer((i - 1) %% o$nrows + 1)]
			d[, col := as.integer((((i - 1) %/% o$nrows + 1) - 1) %% o$ncols + 1)]
			d[, page := as.integer(i - 1) %/% (o$nrows * o$ncols) + 1]
		} else {

			# wrap
			if (o$nby[3] == 1L) {
				if (!o$byrow) {
					d[, row := as.integer((i - 1) %% o$nrows + 1)]
					d[, col := as.integer((((i - 1) %/% o$nrows + 1) - 1) %% o$ncols + 1)]
				} else {
					d[, col := as.integer((i - 1) %% o$ncols + 1)]
					d[, row := as.integer((((i - 1) %/% o$ncols + 1) - 1) %% o$nrows + 1)]
				}
				d[, page := as.integer(i - 1) %/% (o$nrows * o$ncols) + 1]
			} else {
				d[, j:= rep(1L:o$npp, length.out = nrow(d))]
				if (!o$byrow) {
					d[, row := as.integer((j - 1) %% o$nrows + 1)]
					d[, col := as.integer((((j - 1) %/% o$nrows + 1) - 1) %% o$ncols + 1)]
				} else {
					d[, col := as.integer((j - 1) %% o$ncols + 1)]
					d[, row := as.integer((((j - 1) %/% o$ncols + 1) - 1) %% o$nrows + 1)]
				}
				d[, page := by3]
				d[, j:= NULL]
			}
		}

	}



	#####o$legend.autoin.pos = c("left", "top")

	# prepare function names
	FUNinit = paste0("tmap", gs, "Init")
	FUNaux = paste0("tmap", gs, "Aux")
	FUNrun = paste0("tmap", gs, "Run")
	FUNshape = paste0("tmap", gs, "Shape")
	FUNoverlay = paste0("tmap", gs, "Overlay")
	FUNwrap = paste0("tmap", gs, "Wrap")
	FUNxtab = paste0("tmap", gs, "Xtab")

	FUNxlab = paste0("tmap", gs, "Xlab")
	FUNylab = paste0("tmap", gs, "Ylab")


	FUNgridxlab = paste0("tmap", gs, "GridXLab")
	FUNgridylab = paste0("tmap", gs, "GridYLab")

	# for inset maps
	inset_frame_ids = if (nrow(cdt) == 0L) integer(0) else which(sapply(cdt$comp, function(cmp) "bbox" %in% names(cmp)))

	if (!o$legend.only) {
		# create table with bounding boxes (the only important property, apart from settings)
		db = data.table(bbox = unique(d$bbox[!is.na(d$asp)]))
		db[, i:=1L:nrow(db)]
		d[, bi:=db$i[match(d$bbox, db$bbox)]]

		## process components
		if (nrow(cdt)) cdt = process_components2(cdt, o)

		# init
		res = do.call(FUNinit, c(list(o = o, show = show, newpage = newpage, return.asp = return.asp, vp = vp, prx = prx), args))
		if (return.asp) {
			return(res)
		} else {
			.TMAP$geo_ref = get_geo_ref(bbx = db$bbox[[1]],
										crs = crs,
										inner_margins = o$inner.margins,
										dev_size = res$dev[1:2], map_size = res$map[1:2], offset = res$margins)
			.TMAP$animation = list(fps = o$fps, play = o$play)
		}

		## prepare aux layers
		if (length(aux)) {
			# prepare aux layers and return group label (in case it is not user specified)
			aux_group_def = mapply(function(a, id) {
				FUNaux_prep = paste0("tmap", gs, "AuxPrepare")
				a_args = structure(a$args, class = c(a$mapping.fun, "list"))

				do.call(FUNaux_prep, list(a = a_args, bs = db$bbox, id = id, o = o))
			}, aux, 1:length(aux))
			aux_group = mapply(function(a, agd) {
				ng = lengths(gregexpr("__", agd, fixed = TRUE))
				if (is.na(a$group)) agd else as.character(a$group)
			}, aux, aux_group_def, USE.NAMES = FALSE)

			aux_group.control = vapply(aux, function(a) {
				a$group.control
			}, FUN.VALUE = character(1))

			# find lid (layer plot id values) for aux layers
			aux_lid = vapply(aux, function(a) a$lid, FUN.VALUE = numeric(1))
		}

		if (!any_data_layer && !length(aux)) {
			message_nothing_to_show(any_groups)
			return(invisible(NULL))
		}



		# data frame for layer ids
		q = do.call(rbind, c(
			{if (any_data_layer) {
				lapply(1L:o$ng, function(ig) {
					tmxi = tmx[[ig]]
					nl = length(tmxi$layers)
					lid = vapply(tmxi$layers, function(l) {l$lid}, FUN.VALUE = numeric(1))
					group = vapply(tmxi$layers, function(l) {l$group}, FUN.VALUE = character(1))
					group.control = vapply(tmxi$layers, function(l) {l$group.control}, FUN.VALUE = character(1)) # used to determine control layer group (view mode)
					data.frame(gid = rep(ig, nl), g_lid = 1:nl, lid = lid, group = group, group.control = group.control, lid2 = rep(0, nl), pane = rep("", nl), new = rep(TRUE, nl), group.zoom_levels = I(rep(list(NA),nl)))
				})
			} else {
				NULL
			}},
			{if (length(aux)) list(data.frame(gid = 0, g_lid = 1L:length(aux), lid = aux_lid, group = aux_group, group.control = aux_group.control, lid2 = 0, pane = "", new = TRUE, group.zoom_levels = I(rep(list(NA),length(aux))))) else NULL}))

		#q$lid[q$lid != 0] = q$lid[q$lid != 0] + 400L

		if (.TMAP$proxy) {
			q0 = .TMAP$q
			q0$new = FALSE
			q = rbind(q0, q)
		}

		qnotnull = (q$lid != 0)
		if (any(qnotnull)) q$lid2[qnotnull] = rank(q$lid[qnotnull])

		q = q[order(q$lid2), ]
		q$pane = "tilePane"

		q$pane[q$lid2 > 0] = pane_name(q$lid[q$lid2 > 0])

		# via tm_group, control and zoom levels can be set
		# these are stored in tmap options view_group_{name}, and merged with q:
		og = o[substr(names(o), 1, 10) == "view_group"]
		ogdf = 	data.frame(name = vapply(og, "[[",FUN.VALUE = character(1), "name", USE.NAMES = FALSE),
						   control = vapply(og, function(x) {
						   	as.character(x[["control"]])
						   } , FUN.VALUE = character(1), USE.NAMES = FALSE),
						   zoom_levels = I(unname(lapply(og, "[[", "zoom_levels"))))
		q$group.control = mapply(function(gi, v) {
			id = which(ogdf$name == gi)
			if (!length(id)) return(v)
			v2 = ogdf$control[id]
			if (is.na(v2)) return(v) else return(v2)
		}, q$group, q$group.control, USE.NAMES = FALSE, SIMPLIFY = TRUE)
		q$group.zoom_levels = mapply(function(gi, v) {
			id = which(ogdf$name == gi)
			if (!length(id)) return(NA)
			return(ogdf$zoom_levels[[id]])
		}, q$group, q$group.control, USE.NAMES = FALSE, SIMPLIFY = FALSE)
		q$group.control[!is.na(q$group.zoom_levels)] = "none" # see leaflet::groupOptions



		# q data frame:
		# gid = tmap-group counter
		# g_lid = layer counter inside tmap-group
		# lid = possibly-user-defined layer order number
		# lid2 = same as lid, but 1,2,3,...
		# pane = pane name (for view mode)
		# group = group name (for selecting layers in view mode)
		# group.control = leaflet::addLayerControl: check, radio or none (view mode)
		# group.zoom_levels = leaflet::groupOptions: zoom levels displayed at

		do.call(FUNaux, list(o = o, q = q))


		# plot xtab headers
		if (o$panel.type == "xtab") {
			for (k in 1:o$npages) {
				labrows = o$panel.labels[[1]]
				labcols = o$panel.labels[[2]]
				if (length(labrows) == o$nrows) for (i in 1:o$nrows) do.call(FUNxtab, list(label = labrows[i], facet_row = i, facet_page = k, o = o))
				if (length(labcols) == o$ncols) for (j in 1:o$ncols) do.call(FUNxtab, list(label = labcols[j], facet_col = j, facet_page = k, o = o))

			}
		}


		# plot xlab and ylab
		if (o$xlab.show) {
			for (k in 1:o$npages) {
				do.call(FUNxlab, list(facet_page = k, o = o))
			}
		}
		if (o$ylab.show) {
			for (k in 1:o$npages) {
				do.call(FUNylab, list(facet_page = k, o = o))
			}
		}



		for (i in seq_len(nrow(d))) {
			bbx = d$bbox[[i]]
			if (o$panel.type == "wrap") do.call(FUNwrap, list(label = o$panel.labels[[o$panel.labels.dim]][d[[paste0("by", o$panel.labels.dim)]][i]], facet_row = d$row[i], facet_col = d$col[i], facet_page = d$page[i], o = o))
			if (is.na(d$asp[i])) next
			do.call(FUNshape, list(bbx = bbx, facet_row = d$row[i], facet_col = d$col[i], facet_page = d$page[i], o = o))

			# plot grid labels
			if (o$grid.show && !o$grid.labels.inside_frame) {
				if (o$grid.labels.show[2] && ((o$grid.labels.pos[1] == "left" && d$col[i] == 1) || (o$grid.labels.pos[1] == "right" && d$col[i] == o$ncols))) {
					do.call(FUNgridylab, list(bi = d$bi[i], bbx = bbx, facet_row = d$row[i], facet_col = d$col[i], facet_page = d$page[i], o = o))
				}
				if (o$grid.labels.show[1] && ((o$grid.labels.pos[2] == "top" && d$row[i] == 1) || (o$grid.labels.pos[2] == "bottom" && d$row[i] == o$nrows))) {
					do.call(FUNgridxlab, list(bi = d$bi[i], bbx = bbx, facet_row = d$row[i], facet_col = d$col[i], facet_page = d$page[i], o = o))
				}
			}

			for (qi in which(q$new)) {
				gid = q$gid[qi]
				g_lid = q$g_lid[qi]
				pane = q$pane[qi]
				group = q$group[qi]
				if (gid > 0) {
					# data layer
					bl = tmx[[gid]]$layers[[g_lid]]
					shpTM = get_shpTM(bl$shpDT, d$by1[i], d$by2[i], d$by3[i])[[1]]
					mdt = get_dt(bl$mapping_dt, d$by1[i], d$by2[i], d$by3[i])

					id = paste0("f", sprintf("%03d", i), "g", sprintf("%02d", gid), "l", sprintf("%02d", g_lid))

					if (nrow(mdt) != 0) {
						gp = bl$gp

						FUN = paste0("tmap", gs, "DataPlot")

						a = structure(bl$mapping_args, class = c(bl$mapping_fun, "list"))

						do.call(FUN, c(list(a = a, shpTM = shpTM, dt = mdt, pdt = bl$popup.data, popup.format = bl$popup.format, hdt = bl$hover.data, idt = bl$id.data, gp = gp, bbx = bbx, facet_col = d$col[i], facet_row = d$row[i], facet_page = d$page[i], id = id, pane = pane, group = group, glid = bl$glid, o = o)))
					}

				} else {
					g_lid = q$g_lid[qi]


					# aux layer
					a = aux[[g_lid]]
					a_args = structure(a$args, class = c(a$mapping.fun, "list"))

					FUNaux_plot = paste0("tmap", gs, "AuxPlot")

					id = g_lid # to do: test!
					do.call(FUNaux_plot, list(a = a_args, bi = d$bi[i], bbx = bbx, facet_col = d$col[i], facet_row = d$row[i], facet_page = d$page[i], id = id, pane = pane, group = group, o = o))

				}



			}

			# inset map frame
			if (length(inset_frame_ids)) {
				sfc_bbxs = do.call(c, lapply(inset_frame_ids, function(iid) {
					ic = cdt$comp[[iid]]
					sf::st_transform(tmaptools::bb_poly(ic$bbox), crs = crs)
				}))

				collect_properties = function(comps, props) {
					res = lapply(props, function(p) {
						sapply(comps, function(cmp) {
							cmp[[p]]
						})
					})
					names(res) = props
					res
				}

				prp = collect_properties(cdt$comp[inset_frame_ids],
										 c("box_frame", "box_frame.color", "box_frame.alpha", "box_frame.lwd", "box_frame.lty", "box_bg", "box_bg.color", "box_bg.alpha"))
				calld = lapply(cdt$comp[inset_frame_ids], function(cmp) cmp$called)

				prp$box_frame[vapply(calld, function(cl) "box_frame.color" %in% cl, FUN.VALUE = logical(1))] = TRUE
				prp$box_bg[vapply(calld, function(cl) "box_bg.color" %in% cl, FUN.VALUE = logical(1))] = TRUE

				prp$box_bg.color[!prp$box_bg] =  "#00000000"
				prp$box_frame.color[!prp$box_frame] =  "#00000000"


				tb = tm_polygons()[[1]]


				gp = tb$gpar

				FUN = paste0("tmap", gs, "DataPlot")

				a = structure(tb$mapping.args, class = c("tm_data_polygons", "list"))

				shapepTM = shapeTM(sfc_bbxs, tmapID = seq_along(sfc_bbxs))

				shpTM = tmapTransPolygons(shapepTM, args = tb$trans.args, plot.order = tb$plot.order)

				mdt = data.table(tmapID__ = seq_along(sfc_bbxs), fill = prp$box_bg.color, col = prp$box_frame.color, lwd = prp$box_frame.lwd, lty = prp$box_frame.lty, fill_alpha = prp$box_bg.alpha, col_alpha = prp$box_frame.alpha, ord__ = 1L)

				do.call(FUN, c(list(a = a, shpTM = shpTM, dt = mdt, pdt = NULL, popup.format = list(), hdt = NULL, idt = NULL, gp = gp, bbx = bbx, facet_col = d$col[i], facet_row = d$row[i], facet_page = d$page[i], id = "inset_frame", pane = "tmap500", group = NA, o = o)))

			}
			do.call(FUNoverlay, list(bbx = bbx, facet_row = d$row[i], facet_col = d$col[i], facet_page = d$page[i], o = o))
		}

	} else {
		# init
		asp = do.call(FUNinit, list(o = o, return.asp = return.asp, vp = vp))
	}


	if (o$legend.only) {
		cdt[, facet_row := "1"]
		cdt[, facet_col := "1"]
		cdt[, page := 1L]
		cdt[, bbox := list()]
	} else {
		is_in = cdt$class == "in"
		#is_in = rep(TRUE, nrow(cdt))
		if (any(is_in)) {
			legs_in = lapply(which(is_in), function(i) {
				d2 = data.table::copy(d)
				legsi = cdt[i, ]
				if (o$type != "grid" && o$nrows == 1) {
					# reverse above
					d2[, by2 := by1]
					d2[, by1 := 1]
				}

				if (is.na(legsi$by1__)) d2[, by1:= NA]
				if (is.na(legsi$by2__)) d2[, by2:= NA]
				if (is.na(legsi$by3__)) d2[, by3:= NA]
				legsi = merge(legsi, d2[, c("by1", "by2", "by3", "row", "col", "page", "bbox", "units"), with = FALSE], by.x = c("by1__", "by2__", "by3__"), by.y = c("by1", "by2", "by3"))
				legsi[, ':='(facet_row = as.character(row), facet_col = as.character(col), row = NULL, col = NULL)]
				legsi
			})
		} else {
			legs_in = NULL
		}


		legs_out = copy(cdt[!is_in])
		if (nrow(legs_out) >= 1L) {
			legs_out[, page:=NA_integer_]
			legs_out[!is.na(by3__), page := by3__]

			by1_clip =  (length(which(legs_out$by1__ > o$nrows)) > 0L) && all(legs_out$facet_row == legs_out$by1__)
			by2_clip =  (length(which(legs_out$by2__ > o$ncols)) > 0L) && all(legs_out$facet_col == legs_out$by2__)
			if (by1_clip && by2_clip) {
				cli::cli_abort("multiple pages cannot be created over two facet dimensions")
			}
			if (by1_clip) {
				legs_out[!is.na(by1__), ':='(page = ((by1__-1L) %/% o$nrows) + 1L, facet_row = ((by1__-1L) %% o$nrows) + 1L)]
			} else if (by2_clip) {
				legs_out[!is.na(by2__), ':='(page = ((by2__-1L) %/% o$ncols) + 1L, facet_col = ((by2__-1L) %% o$ncols) + 1L)]
			}

			# ad-hoc method: take first bbox and units
			bbox_nb = d$bbox[1]
			attr(bbox_nb, "borrow") = list(col = d$col[1], row = d$row[1])
			legs_out[, bbox:=list(bbox_nb)]
			legs_out[, units:=list(d$units[1])]
		} else {
			legs_out$page = integer()
			legs_out$bbox = list()
			legs_out$units = list()
		}



		cdt = data.table::rbindlist(c(list(legs_out), legs_in))

		cdt$comp = mapply(function(cmp, bbx, u) {
			if (!("bbox" %in% names(cmp))) cmp$bbox = bbx
			cmp$units = u
			cmp
		}, cdt$comp, cdt$bbox, cdt$units, SIMPLIFY = FALSE)

		leg_dummy = tm_legend_hide()
		leg_dummy$active = FALSE
		leg_dummy$vneutral = NA

		leg_nr_dummy = legend_save(leg_dummy)
		crt_nr_dummy = chart_save(tm_chart_none())

		# inset ids
		inset_ids = if (nrow(cdt) == 0L) integer(0) else which(sapply(cdt$comp, inherits, "tm_inset_map"))


		# inset maps: prepare input for step4
		if (length(inset_ids)) {
			cdt$comp[inset_ids] = lapply(cdt$comp[inset_ids], function(comp) {
					tmo_i = tm$tmo
					o_i = tm$o
					aux_i = tm$aux
					cmp_i = list()
					prx_i = list()

					if ("crs" %in% names(comp)) {
						crs_i = comp$crs
						o_i$crs_step4 = crs_i
					} else {
						crs_i = o_i$crs_step4
					}
					o_i$outer.bg = FALSE
					o_i$frame = comp$main_frame
					o_i$frame.color = comp$main_frame.color
					o_i$frame.alpha = comp$main_frame.alpha
					o_i$frame.lwd = comp$main_frame.lwd
					o_i$frame.r = comp$main_frame.r


					# set legends to inactive
					tmo_i = lapply(tmo_i, function(tmg) {
						lapply(tmg, function(tml) {
							lapply(tml, function(tmi) {
								tmi$shpDT$shpTM = lapply(tmi$shpDT$shpTM, crs_reproject_shpTM, crs = crs_i, raster.warp = o$raster.warp)
								tmi$mapping_legend = lapply(tmi$mapping_legend, function(l) {
									l$legnr = leg_nr_dummy
									l$crtnr = crt_nr_dummy
									l
								})
								tmi$trans_legend = lapply(tmi$trans_legend, function(l) {
									l$legnr = leg_nr_dummy
									l$crtnr = crt_nr_dummy
									l
								})
								tmi
							})
						})
					})

				#	(tm, vp, return.asp, show, in.shiny, knit, args)
					comp$tm = list(tmo = tmo_i,
								   o = o_i,
								   aux = aux_i,
								   cmp = cmp_i,
								   prx = prx_i)
					comp
			})
		}
	}

	legfun = paste0("tmap", gs, "Comp")

	toI = function(x) {
		as.integer(strsplit(x, split = "_")[[1]])
	}

	if (nrow(cdt) > 0L) for (k in seq_len(o$npages)) {
		klegs = cdt[is.na(page) | (page == k), ] # was by3__ instead of page
		#klegs[, pos.h.id := pos.h][pos.h %in% c("left", "center", "right"), pos.h.id:="lower"][pos.h %in% c("LEFT", "CENTER", "RIGHT"), pos.h.id:="upper"]
		#klegs[, pos.v.id := pos.v][pos.v %in% c("top", "center", "bottom"), pos.v.id:="lower"][pos.v %in% c("TOP", "CENTER", "BOTTOM"), pos.v.id:="upper"]

		# find group ids
		klegs$id = paste(klegs$class, klegs$cell.h, klegs$cell.v, klegs$pos.h, klegs$pos.v,
						 sapply(klegs$comp, function(comp) paste(comp$position$just.h, comp$position$just.v, sep = "_")), sep = "_")
		# just.id = NULL
		# klegs$just.id = sapply(klegs$comp, function(l) paste(l$position$just.h, l$position$just.v, sep = "."))
		# klegs[, id:=paste(pos.h.id, pos.v.id, just.id, class, sep = "__")]
		klegs[, do.call(legfun, args = list(comp = .SD$comp, o = o, facet_row = toI(.SD$facet_row[1]), facet_col = toI(.SD$facet_col[1]), facet_page = k, class = .SD$class[1], stack = .SD$stack, stack_auto = .SD$stack_auto, pos.h = .SD$pos.h, pos.v = .SD$pos.v, .SD$bbox)), by = list(facet_row, facet_col, id), .SDcols = c("comp", "facet_row", "facet_col", "class", "stack", "stack_auto", "pos.h", "pos.v", "bbox")]
	}

	if (show) save_last_map()

	if (.TMAP$in.shiny) {
		.TMAP$q = q
	}

	do.call(FUNrun, list(o = o, q = q, show = show, knit = knit, knit_opts = knit_opts, args))
}

# bb_ext reversing
bb_ext_rev = function(bbx, ext = c(0, 0, 0, 0)) {
	dx = (bbx[3] - bbx[1]) / (1 + ext[2] + ext[4])
	dy = (bbx[4] - bbx[2]) / (1 + ext[1] + ext[3])

	bbx[2] = bbx[2] + ext[1] * dy
	bbx[1] = bbx[1] + ext[2] * dx
	bbx[4] = bbx[4] - ext[3] * dy
	bbx[3] = bbx[3] - ext[4] * dx
	bbx
}

get_geo_ref = function(bbx, crs, inner_margins, dev_size, map_size, offset) {
	bbx_crop = bb_ext_rev(bbx, inner_margins)

	x1 = offset[2] / dev_size[1]
	x2 = (dev_size[1] - offset[4]) / dev_size[1]

	y1 = offset[1] / dev_size[2]
	y2 = (dev_size[2] - offset[3]) / dev_size[2]

	xy_crop = bb_ext_rev(c(x1, y1, x2, y2), inner_margins)
	names(xy_crop) = c("xmin", "ymin", "xmax", "ymax")
	list(crs = crs, bbx = bbx_crop, bbx_frame = bbx, rel_coords = xy_crop)
}

#' Shape (spatial object) specification
#'
#' Specify a shape, which is a spatial object from one of these spatial object
#' class packages: [`sf`][`sf::sf`], [`stars`][stars::st_as_stars()], or `terra`.
#'
#' The map projection (`crs`) determines in which coordinate system the spatial object is processed and plotted.
#' See `r .doc_see_vig("foundations_crs")`. The `crs` can be specified in two places: 1) `tm_shape()` and `tm_crs()`.
#' In both cases, the map is plotted into the specified `crs`. The difference is that in the first option, the `crs` is also taken into account in spatial transformation functions, such as the calculation of centroids and cartograms. In the second option, the `crs` is only used in the plotting phase.
#'
#' The automatic crs recommendation (which is still work-in-progress) is the following:
#'
#' | **Property**        | **Recommendation** |
#' | ------          | ----------- |
#' | `global` (for world maps)		     | A pseudocylindrical projection tmap option `crs_global`, by default `"eqearth` (Equal Earth). See `r .doc_see_vig("foundations_crs")` for more options|
#' | `area` (equal area)		     | Lambert Azimuthal Equal Area (`laea`) |
#' | `distance`	(equidistant) | Azimuthal Equidistant (`aeqd`) |
#' | `shape`	(conformal) | Stereographic (`stere`) |
#'
#' For further info about the available "generic" projects see:
#' for utm: \url{https://proj.org/en/9.4/operations/projections/utm.html}
#' for laea: \url{https://proj.org/en/9.4/operations/projections/laea.html}
#' for aeqd: \url{https://proj.org/en/9.4/operations/projections/aeqd.html}
#' for pconic: \url{https://proj.org/en/9.4/operations/projections/pconic.html}
#' for eqdc: \url{https://proj.org/en/9.4/operations/projections/eqdc.html}
#'
#' @param shp Spatial data object. Typically an object from \pkg{sf}, \pkg{terra}, or \pkg{stars}.
#'   Additional spatial data types can be supported via extension packages, such as
#'   \pkg{tmap.networks} and \pkg{tmap.sources} (experimental). These may include, for example,
#'   remote or streaming data sources.
#' @param bbox Bounding box of the map. Only used when `shp` is the main shape (see `is.main`).
#'   Three options are supported:
#'   \itemize{
#'     \item a \code{\link[sf:st_bbox]{sf::st_bbox()}} object,
#'     \item a character string specifying a location, passed to \code{\link[tmaptools:geocode_OSM]{tmaptools::geocode_OSM()}},
#'     \item \code{"FULL"}, which represents the whole earth. This option ensures that reprojection
#'     retains the full global extent, unlike a regular bounding box.
#'   }
#' @param crs Map projection (CRS). Can be set to an `crs` object (see [sf::st_crs()]), a proj4string, an EPSG number, the value `"auto"` (automatic crs recommendation), or one the the following generic projections: `c("laea", "aeqd", "utm", "pconic", "eqdc", "stere")`. See details.
#' @param is.main Is `shp` the main shape, which determines the crs and
#'   bounding box of the map? By default, `TRUE` if it is the first `tm_shape` call
#' @param name of the spatial object
#' @param layer Name of the layer to use. This is primarily relevant for multi-layer or remote
#'   data sources (e.g. PMTiles or vector tiles), where multiple layers may be available.
#' @param unit Unit of distance measurement, used by [tm_scalebar()]. Either a
#'   specific unit string such as `"km"`, `"m"`, `"mi"`, `"yd"`, `"ft"`, or
#'   `"in"` (see [units::valid_udunits()] for all options), or one of two
#'   automatic options: `"metric"` (default) selects the most readable unit
#'   from km, m, and mm; `"imperial"` selects from mi, yd, ft, and in.
#'   In both cases the unit is chosen as the largest one for which the
#'   map width expressed in that unit is at least 10.
#' @param filter Optional filter expression used to subset features. The exact syntax depends
#'   on the data source. For in-memory objects (e.g. \pkg{sf}), this is typically evaluated in R,
#'   whereas for remote sources it may be translated to a query and executed on the server side.
#' @param ... passed on to \code{\link[tmaptools:bb]{bb}} (e.g. \code{ext} can be used to enlarge or shrink a bounding box)
#' @import tmaptools
#' @importFrom sf st_geometry st_sf st_as_sf st_transform st_crs
#' @importFrom stars st_dimensions st_get_dimension_values st_as_stars st_set_dimensions st_warp
#' @import units
#' @import grid
#' @import cols4all
#' @import classInt
#' @import htmltools
#' @import htmlwidgets
#' @import leaflet
#' @import leafsync
#' @import leafem
#' @import leaflegend
#' @import stats
#' @importFrom methods as
#' @importFrom utils head tail download.file txtProgressBar str data capture.output setTxtProgressBar methods getS3method
#' @importFrom rlang missing_arg expr
#' @importFrom s2 s2_buffer_cells as_s2_geography s2_intersects
#' @importFrom grDevices col2rgb colorRampPalette colors dev.off dev.size png rgb
#' @note as of tmap 4.0, simplify has been removed. Please use [tmaptools::simplify_shape()] instead
#' @example ./examples/tm_shape.R
#' @seealso `r .doc_see_also_shape()`
#' @export
tm_shape = function(shp = NULL,
					bbox = NULL,
					crs = NULL,
					is.main = NA,
					layer = NULL,
					name = NULL,
					unit = NULL,
					filter = NULL,
					...) {

	args_called = names(rlang::call_match()[-1])
	args = lapply(as.list(rlang::call_match(defaults = TRUE)[-1]), eval, envir = parent.frame())

	if ("projection" %in% args_called) {
		v3_instead_message(arg_old = "projection", arg_new = "crs", fun = "tm_shape")
		crs = args$projection
	}
	if (any(c("crs", "projection") %in% args_called) && is.na(is.main)) {
		is.main = TRUE
	}



	bbox_list = c(list(x = bbox), args[intersect(args_called, c("ext", "cx", "cy", "width", "height", "xlim", "ylim", "relative", "asp.limit"))])

	shp_called = deparse(substitute(shp))[1]



	# rivers -> World_rivers
	if (shp_called %in% c("rivers", "c(735, 320, 325, 392, 524, 450, 1459, 135, 465, 600, 330, 336, ") && # last one to catch qtm calls
		is.numeric(shp) && length(shp) == 141) {
		data(World_rivers, envir = environment())
		shp = World_rivers
		shp_called = "World_rivers"
		v3_rivers()
	}

	if ("simplify" %in% args_called) {
		sim_args = args[intersect(args_called, c("simplify", "keep.units", "keep.subunits"))]
		v3_simplify(shp_called, sim_args)
	}


	shp_name = ifelse(is.null(name) == TRUE, shp_called, name)


	if (missing(shp)) {
		cli::cli_inform("{.field {.fun tm_shape}} Argument {.arg shp} not specified: please use {.fun tm_crs} to specify a bounding box and/or crs")
		do.call(tm_options, args[intersect(args_called, c("bbox", "crs"))])
	} else {
		tm_element_list(tm_element(shp = shp,
								   is.main = is.main,
								   crs = crs,
								   bbox = bbox_list,
								   unit = unit,
								   filter = filter,
								   layer = layer,
								   shp_name = shp_name,
								   subclass = "tm_shape"))
	}
}

#' Arrange small multiples in grid layout
#'
#' Arrange small multiples in a grid layout. Normally, small multiples are created
#' by specifying multiple variables for one aesthetic or by specifying the by argument
#' (see [tm_facets()]). This function can be used to arrange custom small multiples in a grid layout.
#'
#' The global option `tmap.limits` controls the limit of the number of facets that are plotted.
#' By default, `tmap_options(tmap.limits = c(facets.view=4, facets.plot=64))`.
#' The maximum number of interactive facets is set to four since otherwise it may become very slow.
#'
#' @param ... [`tmap`] objects or one list of [`tmap`] objects.
#'   The number of multiples that can be plot is limited (see details).
#' @param ncol number of columns
#' @param nrow number of rows
#' @param widths vector of column widths. It should add up to 1 and the length
#'   should be equal to `ncol`.
#' @param heights vector of row heights. It should add up to 1 and the length
#'   should be equal to `nrow`.
#' @param sync logical. Should the navigation in view mode (zooming and panning)
#'   be synchronized? By default `FALSE`.
#' @param asp aspect ratio. The aspect ratio of each map. Normally, this is
#'   controlled by the `asp` argument from [tm_layout()] (also a tmap option).
#'   This argument will overwrite it, unless set to `NULL`.
#'   The default value for `asp` is 0, which means that the aspect ratio is
#'   adjusted to the size of the device divided by the number of columns and rows.
#'   When `asp` is set to `NA`, which is also the default value for `tm_layout()`,
#'   the aspect ratio will be adjusted to the used shapes.
#' @param outer.margins outer.margins, numeric vector four or a single value.
#'   If defines the outer margins for each multiple. If will overwrite the
#'   `outer.margins` argument from [tm_layout()], unless set to `NULL`.
#' @param x a `tmap_arrange` object (returned from `tmap_arrange()`).
#' @param knit should [knitr::knit_print()] be enabled, or the normal [base::print()] function?
#' @param options options passed on to [knitr::knit_print()]
#' @example ./examples/tmap_arrange.R
#' @export
tmap_arrange <- function(..., ncol = NA, nrow = NA, widths = NA, heights = NA, sync = FALSE, asp = 0, outer.margins = .02) {
	tms <- list(...)
	if (!inherits(tms[[1]], "tmap")) {
		if (!is.list(tms[[1]])) stop("The first argument of tmap_arrange is neither a tmap object nor a list.")
		tms <- tms[[1]]
	}
	istmap <- vapply(tms, FUN = inherits, FUN.VALUE = logical(1), what = "tmap")
	if (!all(istmap)) stop("Not all arguments are tmap objects.")
	opts <- list(ncol=ncol, nrow=nrow, widths=widths, heights=heights, sync=sync, asp=asp, outer.margins=outer.margins)
	attr(tms, "opts") <- opts
	class(tms) <- c("tmap_arrange", class(tms))
	tms
}

#' @rdname tmap_arrange
#' @exportS3Method knitr::knit_print
knit_print.tmap_arrange <- function(x, ..., options = NULL) {
	print_tmap_arrange(x, knit=TRUE, ..., options = options)
}

#' @export
#' @rdname tmap_arrange
print.tmap_arrange <- function(x, knit = FALSE, ..., options = NULL) {
	print_tmap_arrange(x, knit=knit, ..., options = options)
}




print_tmap_arrange <- function(tms, knit = FALSE, show = TRUE, add.titles = TRUE, ..., options = options) {
	args <- list(...)

	opts <- attr(tms, "opts")

	tms_len <- length(tms)

	nx <- limit_nx(tms_len)
	if (nx != tms_len) tms <- tms[1:nx]

	if (is.na(opts$ncol) && is.na(opts$nrow)) {
		devsize <- dev.size()
		dasp <- devsize[1] / devsize[2]

		## determine 'overall' aspect ratio by overlaying the maps
		#tmp <- tempfile(fileext = ".png")
		#png( tmp, width=700, height=700, res = 100)

		curdev <- grDevices::dev.cur()
		rmc = tmap_options(raster.max_cells = 36)

		on.exit(tmap_options(rmc))

		tasps <- suppressMessages({
			vapply(tms, function(tm) {
				asp <- get_asp_ratio(tm, width = 700, height = 700, res = 100)
				#asp <- print_tmap(tm, return.asp = TRUE, mode = "plot")
				# dev.off()
				asp
			}, numeric(1))
		})

		grDevices::dev.set(curdev)
		tmap_options(rmc)

		#dev.off()
		#dev.off()
		hs <- vapply(tasps, function(tasp) ifelse(tasp>1, 1, 1/tasp), numeric(1))
		ws <- vapply(tasps, function(tasp) ifelse(tasp>1, tasp, 1), numeric(1))
		iasp <- max(ws) / max(hs)

		asp_ratio <- iasp / dasp

		nrowcol <- process_get_arrangement(nx = nx, asp_ratio = asp_ratio)
		nrow <- nrowcol[1]
		ncol <- nrowcol[2]


	} else {
		ncol <- if (is.na(opts$ncol)) ceiling(nx / opts$nrow) else opts$ncol
		nrow <- if (is.na(opts$nrow)) ceiling(nx / opts$ncol) else opts$nrow
	}


	interactive <- (getOption("tmap.mode")=="view")
	gs = tmap_graphics_name()

	fun = paste0("tmap", gs, "Arrange")

	do.call(fun, list(tms = tms, nx = nx, ncol = ncol, nrow = nrow, opts = opts, knit = knit, show = show, args = args, options = options))

}





process_get_arrangement <- function(nx, asp_ratio) {
	#           asp ~ nrow
	#       |--------------
	#   1   |
	# ~ncol |       nx
	#       |
	ncol_init <- sqrt(nx/asp_ratio)
	nrow_init <- nx / ncol_init

	# rounding:
	nrow_ceiling <- min(ceiling(nrow_init), nx)
	ncol_ceiling <- min(ceiling(ncol_init), nx)

	# find minimal change
	nrow_xtra <- abs(nrow_ceiling - nrow_init) * ncol_init
	ncol_xtra <- abs(ncol_ceiling - ncol_init) * nrow_init

	# calculaet the other, and subtract 1 when possible
	if (nrow_xtra < ncol_xtra) {
		nrow <- nrow_ceiling
		ncol <- ceiling(nx / nrow)
		if ((nrow-1) * ncol >= nx) nrow <- nrow - 1
	} else {
		ncol <- ncol_ceiling
		nrow <- ceiling(nx / ncol)
		if ((ncol-1) * nrow >= nx) ncol <- ncol - 1
	}
	c(nrow=nrow, ncol=ncol)
}

#' @export
#' @rdname tmapGridLeaflet
tmapGridDataPlot.tm_data_lines = function(a, shpTM, dt, gp, bbx, facet_row, facet_col, facet_page, id, pane, group, glid, o, ...) {

	rc_text = frc(facet_row, facet_col)

	res = select_sf(shpTM, dt[!is.na(dt$lwd), ])
	shp = res$shp
	dt = res$dt

	gp = impute_gp(gp, dt)
	gp = rescale_gp(gp, o$scale_down)

	gp = gp_to_gpar(gp, sel = "col", o = o, type = "lines")
	grb = sf::st_as_grob(shp, gp = gp, name = paste0("lines_", id))

	gts = get("gts", .TMAP_GRID)
	gt = gts[[facet_page]]

	gt_name = paste0("gt_facet_", rc_text)

	gt = grid::addGrob(gt, grb, gPath = grid::gPath(gt_name))

	gts[[facet_page]] = gt

	assign("gts", gts, envir = .TMAP_GRID)
	NULL
}


#' @export
#' @rdname tmapGridLeaflet
tmapGridDataPlot.tm_data_iso = function(a, shpTM, dt, gp, bbx, facet_row, facet_col, facet_page, id, pane, group, glid, o, ...) {
	# isn't called, but needed to make tm_iso visible in tmap_overview
	NextMethod()
}

#' Map layer: raster
#'
#' Map layer that draws rasters. Supported visual variable is: `col` (the  color).
#'
#' @param col,col.scale,col.legend,col.chart,col.free  `r .doc_vv("col")` `r .doc_unit("col")`
#' @param col_alpha,col_alpha.scale,col_alpha.legend,col_alpha.chart,col_alpha.free  `r .doc_vv("col_alpha")` `r .doc_unit("col_alpha")`
#' @inheritParams tm_polygons
#' @param options options passed on to the corresponding `opt_<layer_function>` function
#' @param ... to catch deprecated arguments from version < 4.0
#' @example ./examples/tm_raster.R
#' @inherit tm_polygons details
#' @export
tm_raster = function(col = tm_vars(),
					 col.scale = tm_scale(),
					 col.legend = tm_legend(),
					 col.chart = tm_chart_none(),
					 col.free = NA,
					 col_alpha = tm_const(),
					 col_alpha.scale = tm_scale(),
					 col_alpha.legend = tm_legend(),
					 col_alpha.chart = tm_chart_none(),
					 col_alpha.free = NA,
					 zindex = NA,
					 group = NA,
					 group.control = "check",
					 blend = "over",
					 options = opt_tm_raster(),
					 ...) {


	args_called = names(rlang::call_match()[-1])
	args = list(...)

	layer_fun = if ("called_from" %in% names(args)) {
		args$called_from
	} else {
		"tm_raster"
	}

	if (any(v3_only("tm_raster") %in% names(args))) {


		v3_start_message()
		if (!("style" %in% names(args))) {
			if (!"breaks" %in% names(args)) {
				style = "pretty"
			} else {
				style = "fixed"
			}
		} else {
			style = args$style
		}

		v3_list_init()
		if (length(style) > 1) {
			style = style[1]
			.TMAP$v3_list$mult = TRUE
		}
		# v3 visual variable: col
		col.scale.args = c(list(n = v3_impute(args, "n", 5),
								style = style,
								style.args = v3_impute(args, "style.args", list())),
							if (style %in% c("cont", "log10")) {
								c({
									if (!is.null(args$breaks)) {
										if (length(args$breaks) > 2) {
											list(ticks = v3_impute(args, "breaks", NULL, "ticks"))
										} else {
											list(ticks = v3_impute(args, "breaks", NULL, "limits"))
										}
									} else {
										list()
									}
								},
								list(outliers.trunc = c(TRUE, FALSE)))
							} else {
								list(breaks = v3_impute(args, "breaks", NULL),
									 interval.closure = v3_impute(args, "interval.closure", "left"),
									 drop.levels = v3_impute(args, "drop.levels", FALSE))
							},
							list(midpoint = v3_impute(args, "midpoint", NULL),
								 as.count = v3_impute(args, "as.count", NA),
								 values = v3_impute(args, "palette", NA, "values"),
								 values.repeat = !v3_impute(args, "stretch.palette", TRUE, "values.repeat"),
								 values.range = v3_impute(args, "contrast", NA, "values.range"),
								 values.scale = 1,
								 value.na = v3_impute(args, "colorNA", NA, "value.na"),
								 value.null = v3_impute(args, "colorNULL", NA, "value.null"),
								 value.neutral = NA,
								 labels = v3_impute(args, "labels", NULL),
								 label.na = v3_impute(args, "textNA", "Missing", "label.na"),
								 label.null = NA,
								 label.format = v3_impute(args, "legend.format", list(), "label.format")))
		col.scale.args$fun_pref = if (style == "cat") {
			"categorical"
		} else if (style %in% c("fixed", "sd", "equal", "pretty", "quantile",
								"kmeans", "hclust", "bclust", "fisher", "jenks",
								"dpih", "headtails", "log10_pretty")) {
			"intervals"
		} else if (style == "cont") {
			"continuous"
		} else if (style == "log10") {
			"continuous_log"
		} else if (style == "order") {
			"rank"
		} else {
			stop("unknown style")
		}
		if ("style" %in% names(args)) {
			v3_tm_scale_instead_of_style(style, scale_fun = col.scale.args$fun_pref, vv = "col", layer_fun = layer_fun, arg_list = v3_list_get())
		} else {
			v3_tm_scale(scale_fun = "", vv = "col", layer_fun = layer_fun, arg_list = v3_list_get())
		}
		col.scale = do.call("tm_scale", args = col.scale.args)

		if ("alpha" %in% names(args)) {
			col_alpha = args$alpha
			v3_message_col_alpha(layer_fun = layer_fun, orig = "alpha")

		}

		v3_list_init()
		if ("legend.show" %in% names(args) && !args$legend.show) {
			v3_tm_legend_hide(layer_fun, arg = "legend.show", vv = "col")
			lwd.legend = tm_legend_hide()
		} else {
			col.legend.args = alist(title = v3_impute(args, "title", NA, "title"),
								   na.show = v3_impute(args, "showNA", NA),
								   format = v3_impute(args, "legend.format", list(), "format"),
								   orientation = ifelse(v3_impute(args, "legend.is.portrait", TRUE, "orientation"), "portrait", "landscape"),
								   reverse = v3_impute(args, "legend.reverse", FALSE, "reverse"))
			col.legend = do.call("tm_legend", col.legend.args)
			v3_tm_legend(fun = layer_fun, vv = "col", arg_list = v3_list_get())
		}

		if ("legend.hist" %in% names(args) && args$legend.hist) {
			col.chart = tm_chart_histogram()
			v3_tm_chart_hist(layer_fun = layer_fun, vv = "col", arg = "legend.hist")

			# to do: histogram title
		}
	}

	# unused arguments: typos?
	unused = setdiff(names(args), c(v3_only("tm_raster"), "called_from"))

	if (length(unused)) {
		message_layer_unused_args(layer_fun, unused)
	}



	# needed for color maps without categories (then tm_scale_categorical is used without legend, unless called)
	col.legend$legend_called = "col.legend" %in% args_called

	# blend has been migrated from opt_tm_raster() to the layer function root;
	# it is still passed on internally via mapping.args (read as a$blend).
	options$mapping.args$blend = blend


	tm_element_list(tm_element(
		layer = "raster",
		trans.fun = tmapTransRaster,
		trans.aes = list(),
		trans.args = options$trans.args,
		trans.apply_to = "this",
		mapping.aes = list(col = tmapScale(aes = "col",
										   value = col,
										   scale = col.scale,
										   legend = col.legend,
										   chart = col.chart,
										   free = col.free),
						   col_alpha = tmapScale(aes = "col_alpha",
						   					  value = col_alpha,
						   					  scale = col_alpha.scale,
						   					  legend = col_alpha.legend,
						   					  chart = col_alpha.chart,
						   					  free = col_alpha.free)),

		gpar = tmapGpar(fill = "__col",
						col = NA,
						shape = NA,
						size = NA,
						fill_alpha = "__col_alpha",
						col_alpha = NA,
						pattern = "fill",
						lty = NA,
						lwd = NA,
						linejoin = NA,
						lineend = NA),
		tpar = tmapTpar(),
		plot.order = tm_plot_order("DATA"),
		mapping.args = options$mapping.args,
		zindex = zindex,
		group = group,
		group.control = group.control,
		popup.vars = FALSE,
		popup.format = list(),
		popup.title = NA,
		hover = "",
		id = "",
		subclass = c("tm_aes_layer", "tm_layer")))
}

#' @rdname tm_raster
#' @param interpolate Should the raster image be interpolated? Currently only applicable in view mode (passed on to [`grid`][grid::rasterGrob()])
#' @export
opt_tm_raster = function(interpolate = FALSE) {
	list(trans.args = list(),
		 mapping.args = list(interpolate = interpolate))
}

#' @export
#' @rdname tmap_internal
tmapUsrCls = function(x) {
	structure(x, class = "tmapUsrCls")
}



#' @export
#' @rdname tmap_internal
format_aes_results = function(values, ord = NULL, legend, chart) {
	legnr = vector(mode = "integer", length = length(values))
	legnr[1] = legend_save(legend)

	crtnr = vector(mode = "integer", length = length(values))
	crtnr[1] = chart_save(chart)

	if (is.null(ord)) {
		list(values = values,
			 legnr = legnr,
			 crtnr = crtnr)
	} else {
		list(values = values,
			 ord = ord,
			 legnr = legnr,
			 crtnr = crtnr)
	}
}





legends_init = function() {
	assign("legs", list(), envir = .TMAP)
}

charts_init = function() {
	assign("charts", list(), envir = .TMAP)
}


legend_save = function(legend) {
	if (!exists("legs", envir = .TMAP)) legends_init()
	legs = get("legs", envir = .TMAP)
	legs = c(legs, (list(legend)))
	assign("legs", legs, envir = .TMAP)
	length(legs)
}

#' @export
#' @rdname tmap_internal
chart_save = function(legend) {
	if (!exists("charts", envir = .TMAP)) charts_init()
	charts = get("charts", envir = .TMAP)
	charts = c(charts, (list(legend)))
	assign("charts", charts, envir = .TMAP)
	length(charts)
}

#' @export
#' @rdname tmap_internal
data_type = function(x) {
	if (all(is.na(x))) {
		"na"
	} else if (is.ordered(x)) {
		"order"
	} else if (is.logical(x) || is.character(x) || is.factor(x)) {
		"factor"
	} else if (is.numeric(x)) {
		if (any(x < 0 & !is.na(x)) && any(x > 0 & !is.na(x))) {
			"div"
		} else {
			"seq"
		}
	} else {
		"unknown"
	}
}

data_type_grp = function(x) {
	if (x %in% c("seq", "div")) {
		"num"
	} else {
		x
	}
}

#' @export
#' @rdname tmap_internal
data_class = function(x, check_for_color_class = FALSE, midpoint_enabled = FALSE) {
	# if (all(is.na(x))) {
	# 	"na"
	# } else
	cls = if (inherits(x, c("POSIXct", "POSIXlt"))) {
		c("datetime", "seq")
	} else if (inherits(x, "Date")) {
		c("date", "seq")
	} else if (is.numeric(x)) {
		y = without_units(x)
		subclass1 = if (is.integer(x)) "int" else "real"
		subclass2 = if ((any(y < 0 & !is.na(y)) && any(y > 0 & !is.na(y))) || midpoint_enabled) {
			"div"
		} else {
			"seq"
		}
		c("num", subclass1, subclass2)
	} else {
		if (check_for_color_class) {
			w = which(!is.na(x))
			if (length(w) && all(valid_colors(head(x[w], 100)))) {
				c("asis", "color")
			} else {
				subclass = if (is.ordered(x)) "ord" else "unord"
				c("fact", subclass)
			}
		} else {
			subclass = if (is.ordered(x)) "ord" else "unord"
			c("fact", subclass)
		}
	}

	attr(cls, "units") = if (inherits(x, "units")) {
		paste0(" [", units(x), "]")
	} else ""
	attr(cls, "unique") = (length(na.omit(unique(x)))==1)
	cls
}


#' @export
#' @rdname tmap_internal
tmapScale = function(aes, value, scale, legend, chart, free) {
	if (is.null(legend)) legend = tm_legend_hide()
	structure(list(aes = aes, value = tmapVV(value), scale = scale, legend = legend, chart = chart, free = free), class = c("tmapScale", "list"))
}

#' @export
#' @rdname tmap_internal
tmapScaleAuto = function(x1, scale, legend, chart, o, aes, layer, layer_args, sortRev, bypass_ord, submit_legend = TRUE, ...) {
	args = list(...)
	k = length(args) + 1L
	if (length(args)) {
		names(args) = paste0("x", 2L:(length(args)+1L))
	}

	cls = data_class(x1, check_for_color_class = aes %in% c("col", "fill"))

	#if (cls[1] == "na")
	sc_opt = getAesOption("scales.var", o, aes, layer, cls = cls)
#if (aes == "fill") browser()
	if (k == 2) {
		sc = "bivariate"
	} else if (k > 2) {
		stop("No default scale for multivariate variables", call. = FALSE)
		#sc = "composition"
	} else if (cls[1] == "asis") {
		sc = "asis"
	} else if (attr(cls, "unique") && !(sc_opt == "asis")) {
		if ("num" %in% cls) {
			sc = "ordinal"
			cli::cli_inform("The visual variable {.arg {aes}} of the layer {.str {layer}} contains a unique value. Therefore a discrete scale is applied (tm_scale_discrete).")
		} else {
			sc = "categorical"
			cli::cli_inform("The visual variable {.arg {aes}} of the layer {.str {layer}} contains a unique value. Therefore a categorical scale is applied (tm_scale_categorical).")
		}
	} else {

		sc_pref = scale$fun_pref

		if (!is.null(sc_pref)) {
			if (sc_pref %in% c("categorical", "continuous", "continuous_log", "rank")) {
				sc = sc_pref
			} else {
				sc = sc_opt
			}
		} else {
			sc = sc_opt
		}
	}


	tm_scalefun = paste0("tm_scale_", sc)

	scale = scale[names(scale) %in% names(formals(tm_scalefun))]

	scale_new = do.call(tm_scalefun, args = scale)

	FUN = scale_new$FUN
	scale_new$FUN = NULL

	if (sc == "bivariate") {
		do.call(FUN, list(x1 = x1, x2 = args[[1]], scale = scale_new, legend = legend, chart = chart, o = o, aes = aes, layer = layer, layer_args = layer_args, sortRev = sortRev, bypass_ord = bypass_ord, submit_legend = submit_legend))
	} else if (sc == "multi_continuous") {
		do.call(FUN, c(list(x1 = x1), args, list(scale = scale_new, legend = legend, chart = chart, o = o, aes = aes, layer = layer, layer_args = layer_args, sortRev = sortRev, bypass_ord = bypass_ord, submit_legend = submit_legend)))
	} else {
		do.call(FUN, list(x1 = x1, scale = scale_new, legend = legend, chart = chart, o = o, aes = aes, layer = layer, layer_args = layer_args, sortRev, bypass_ord, submit_legend))
	}

}

#' Map layer: polygons
#'
#' Map layer that draws polygons. Supported visual variables are: `fill` (the fill color),
#' `col` (the border color), `lwd` (line width), `lty` (line type),
#' `fill_alpha` (fill color alpha transparency) and `col_alpha` (border color alpha transparency).
#'
#' The visual variable arguments (e.g. `col`) can be specified with a data
#' variable name (e.g., a spatial vector attribute or a raster layer of the object
#' specified in [tm_shape()]), with a visual value (for `col`, a color is expected), or with a geometry-derived variable (see below).
#' See \href{https://r-tmap.github.io/tmap/articles/basics_vv}{vignette about visual variables}.
#'
#' Multiple values can be specified: in that case facets are created.
#' These facets can be combined with other faceting data variables, specified with [tm_facets()].
#' See \href{https://r-tmap.github.io/tmap/articles/basics_facets}{vignette about facets}.
#'
#' - The `*.scale` arguments determine the used scale to map the data values to
#' visual variable values. These can be specified with one of the available
#' `tm_scale_*()` functions. The default is specified by the tmap option ([tm_options()]) `scales.var`.
#' See \href{https://r-tmap.github.io/tmap/articles/basics_scales}{vignette about scales}.
#'
#' - The `*.legend` arguments determine the used legend, specified with [tm_legend()].
#' The default legend and its settings are determined by the tmap options ([tm_options()]) `legend.` .
#' See \href{https://r-tmap.github.io/tmap/articles/basics_legends}{vignette about legends}.
#'
#' - The `*.chart` arguments specify additional charts, specified with `tm_chart_`, e.g. [tm_chart_histogram()].
#' See \href{https://r-tmap.github.io/tmap/articles/basics_charts}{vignette about charts}.
#'
#' - The `*.free` arguments determine whether scales are applied freely across facets, or shared.
#' A logical value is required. They can also be specified with a vector of three
#' logical values; these determine whether scales are applied freely per facet dimension.
#' This is only useful when facets are applied (see [tm_facets()]).
#' There are maximally three facet dimensions: rows, columns, and pages. This only
#' applies for a facet grid ([tm_facets_grid()]). For instance, `col.free = c(TRUE, FALSE, FALSE)`
#' means that for the visual variable `col`, each row of facets will have its own
#' scale, and therefore its own legend. For facet wraps and stacks
#' ([tm_facets_wrap()] and [tm_facets_stack()]) there is only one facet dimension,
#' so the `*.free` argument requires only one logical value.
#'
#' Currently, three geometry-derived variables are implemented:
#'
#' - `"AREA"` (polygons only), which uses the feature area;
#'
#' - `"LENGTH"` (lines only), which uses the feature length; and
#'
#' - `"MAP_COLORS"`, which assigns values so that adjacent features receive
#' different values, making it particularly suitable for coloring
#' neighbouring polygons.
#'
#' Note that geometry-derived variables do not generate a legend automatically.
#' If a legend is required, compute the corresponding variable explicitly,
#' for example with [sf::st_area()], [sf::st_length()], or
#' [tmaptools::map_coloring()], and use the resulting values instead.
#'
#' @param fill,fill.scale,fill.legend,fill.chart,fill.free `r .doc_vv("fill")`
#' @param col,col.scale,col.legend,col.chart,col.free `r .doc_vv("col")`
#' @param lwd,lwd.scale,lwd.legend,lwd.chart,lwd.free `r .doc_vv("lwd")`
#' @param lty,lty.scale,lty.legend,lty.chart,lty.free `r .doc_vv("lty")`
#' @param fill_alpha,fill_alpha.scale,fill_alpha.chart,fill_alpha.legend,fill_alpha.free `r .doc_vv("fill_alpha")`
#' @param col_alpha,col_alpha.scale,col_alpha.legend,col_alpha.chart,col_alpha.free `r .doc_vv("col_alpha")`
#' @param linejoin,lineend Line join and line end. See [gpar()][grid::gpar()] for details.
#' @param plot.order Specification in which order the spatial features are drawn.
#'   See [tm_plot_order()] for details.
#' @param zindex Map layers are drawn on top of each other. The `zindex` numbers
#'   (one for each map layer) determines the stacking order.
#'   By default the map layers are drawn in the order they are called.
#' @param group Name of the group to which this layer belongs. This is only
#'   relevant in view mode, where layer groups can be switched (see `group.control`)
#' @param group.control In view mode, the group control determines how layer groups
#'   can be switched on and off. Options: `"radio"` for radio buttons
#'   (meaning only one group can be shown), `"check"` for check boxes
#'   (so multiple groups can be shown), and `"none"` for no control
#'   (the group cannot be (de)selected).
#' @param options options passed on to the corresponding `opt_<layer_function>` function
#' @param popup.vars names of data variables that are shown in the popups
#'   in `"view"` mode. Set popup.vars to `TRUE` to show all variables in the
#'   shape object. Set popup.vars to `FALSE` to disable popups. Set `popup.vars`
#'   to a character vector of variable names to those those variables in the popups.
#'   The default (`NA`) depends on whether visual variables (e.g.`fill`) are used.
#'   If so, only those are shown. If not all variables in the shape object are shown.
#' @param popup.format list of formatting options for the popup values. Output of [tm_label_format()]. Only applicable for
#'   numeric data variables. If one list of formatting options is provided,
#'   it is applied to all numeric variables of `popup.vars`. Also, a (named)
#'   list of lists can be provided. In that case, each list of formatting options
#'   is applied to the named variable.
#' @param hover name of the data variable that specifies the hover labels (view mode only). Set to `FALSE` to disable hover labels. By default `FALSE`, unless `id` is specified. In that case, it is set to `id`,
#' @param id name of the data variable that specifies the indices of the spatial
#'   features. Only used for `"view"` mode.
#' @param ... to catch deprecated arguments from version < 4.0
#' @example ./examples/tm_polygons.R
#' @seealso \href{https://r-tmap.github.io/tmap/articles/examples_choro_World}{Choropleth example (1)} and \href{https://r-tmap.github.io/tmap/articles/examples_choro_NLD}{choropleth example (2)}
#' @export
tm_polygons = function(fill = tm_const(),
					   fill.scale = tm_scale(),
					   fill.legend = tm_legend(),
					   fill.chart = tm_chart_none(),
					   fill.free = NA,
					   col = tm_const(),
					   col.scale = tm_scale(),
					   col.legend = tm_legend(),
					   col.chart = tm_chart_none(),
					   col.free = NA,
					   lwd = tm_const(),
					   lwd.scale = tm_scale(),
					   lwd.legend = tm_legend(),
					   lwd.chart = tm_chart_none(),
					   lwd.free = NA,
					   lty = tm_const(),
					   lty.scale = tm_scale(),
					   lty.legend = tm_legend(),
					   lty.chart = tm_chart_none(),
					   lty.free = NA,
					   fill_alpha = tm_const(),
					   fill_alpha.scale = tm_scale(),
					   fill_alpha.legend = tm_legend(),
					   fill_alpha.chart = tm_chart_none(),
					   fill_alpha.free = NA,
					   col_alpha = tm_const(),
					   col_alpha.scale = tm_scale(),
					   col_alpha.legend = tm_legend(),
					   col_alpha.chart = tm_chart_none(),
					   col_alpha.free = NA,
					   linejoin = "round",
					   lineend = "round",
					   plot.order = tm_plot_order("lwd", reverse = TRUE, na.order = "bottom"),
					   zindex = NA,
					   group = NA,
					   group.control = "check",
					   popup.vars = NA,
					   popup.format = tm_label_format(),
					   hover = NA,
					   id = "",
					   options = opt_tm_polygons(),
					   ...) {

	args_called = names(rlang::call_match()[-1])
	args = list(...)

	layer_fun = if ("called_from" %in% names(args)) {
		args$called_from
	} else {
		"tm_polygons"
	}

	if (any(v3_only("tm_polygons") %in% names(args))) {



		v3_start_message()


		if (!("style" %in% names(args))) {
			if (!"breaks" %in% names(args) || is.null(args$breaks)) {
				style = "pretty"
			} else {
				style = "fixed"
			}
		} else {
			style = args$style
		}


		v3_list_init()
		if (length(style) > 1) {
			style = style[1]
			.TMAP$v3_list$mult = TRUE
		}



		fill.scale.args = c(list(n = v3_impute(args, "n", 5),
								 style = style,
								 style.args = v3_impute(args, "style.args", list())),
							if (style %in% c("cont", "log10")) {
								c({
									if (!is.null(args$breaks)) {
										if (length(args$breaks) > 2) {
											list(ticks = v3_impute(args, "breaks", NULL, "ticks"))
										} else {
											list(ticks = v3_impute(args, "breaks", NULL, "limits"))
										}
									} else {
										list()
									}
								},
								list(outliers.trunc = c(TRUE, FALSE)))
							} else {
								list(breaks = v3_impute(args, "breaks", NULL),
									 interval.closure = v3_impute(args, "interval.closure", "left"),
									 drop.levels = v3_impute(args, "drop.levels", FALSE))
							},
							list(midpoint = v3_impute(args, "midpoint", NULL),
								 as.count = v3_impute(args, "as.count", NA),
								 values = v3_impute(args, "palette", NA, "values"),
								 values.repeat = !v3_impute(args, "stretch.palette", TRUE, "values.repeat"),
								 values.range = v3_impute(args, "contrast", NA, "values.range"),
								 values.scale = 1,
								 value.na = v3_impute(args, "colorNA", NA, "value.na"),
								 value.null = v3_impute(args, "colorNULL", NA, "value.null"),
								 value.neutral = NA,
								 labels = v3_impute(args, "labels", NULL),
								 label.na = v3_impute(args, "textNA", "Missing", "label.na"),
								 label.null = NA,
								 label.format = v3_impute(args, "legend.format", list(), "label.format")))


		fill.scale.args$fun_pref = if (style[1] == "cat") {
			"categorical"
		} else if (style %in% c("fixed", "sd", "equal", "pretty", "quantile", "kmeans", "hclust", "bclust", "fisher", "jenks", "dpih", "headtails", "log10_pretty")) {
			"intervals"
		} else if (style == "cont") {
			"continuous"
		} else if (style == "log10") {
			"continuous_log"
		} else if (style == "order") {
			"rank"
		} else {
			stop("unknown style")
		}

		if ("style" %in% names(args)) {
			v3_tm_scale_instead_of_style(style[1], scale_fun = fill.scale.args$fun_pref, vv = "fill", layer_fun = layer_fun, arg_list = v3_list_get())
		} else {
			v3_tm_scale(scale_fun = "", vv = "fill", layer_fun = layer_fun, arg_list = v3_list_get())
		}

		fill.scale = do.call("tm_scale", args = fill.scale.args)

		if ("convert2density" %in% names(args) && args$convert2density) {
			fill.scale$convert2density = TRUE
			v3_convert2density(layer_fun)
		}

		if ("col" %in% args_called && (!layer_fun %in% c("tm_fill", "qtm"))) {
			fill = col
			col = tm_const()
			v3_message_col_fill(layer_fun = layer_fun)
		}



		if ("border.col" %in% names(args)) {
			col = args$border.col
			if (!("col" %in% args_called)) v3_message_col_fill(layer_fun = layer_fun)
		}
		if (identical(args$called_from, "tm_borders")) {
			fill = NA
		}

		if ("alpha" %in% names(args)) {
			fill_alpha = args$alpha
			v3_message_fill_alpha(layer_fun = layer_fun)
		}

		if ("border.alpha" %in% names(args)) {
			col_alpha = args$border.alpha
			v3_message_col_alpha(layer_fun = layer_fun)
		}


		v3_list_init()
		if ("legend.show" %in% names(args) && !args$legend.show) {
			v3_tm_legend_hide(layer_fun, arg = "legend.show", vv = "fill")
			fill.legend = tm_legend_hide()
		} else {
			fill.legend.args = alist(title = v3_impute(args, "title", NA),
									na.show = v3_impute(args, "showNA", NA, "na.show"),
									format = v3_impute(args, "legend.format", list(), "format"),
									orientation = ifelse(v3_impute(args, "legend.is.portrait", TRUE, "orientation"), "portrait", "landscape"),
									reverse = v3_impute(args, "legend.reverse", FALSE, "reverse"))
			fill.legend = do.call("tm_legend", fill.legend.args)
			v3_tm_legend(fun = layer_fun, vv = "fill", arg_list = v3_list_get())
		}

		if ("legend.hist" %in% names(args) && args$legend.hist) {
			fill.chart = tm_chart_histogram()
			v3_tm_chart_hist(layer_fun = layer_fun, vv = "fill", arg = "legend.hist")

			# to do: histogram title
		}

	}

	# unused arguments: typos?
	unused = setdiff(names(args), c(v3_only("tm_polygons"), "called_from"))

	if (length(unused)) {
		message_layer_unused_args(layer_fun, unused)
	}

	# make sure required options are there
	options = complete_options(options, opt_tm_polygons())


	tm_element_list(tm_element(
		layer = "polygons",
		trans.fun = tmapTransPolygons,
		trans.args = options$trans.args,
		trans.aes = list(),
		trans.apply_to = "this",
		mapping.aes = list(fill = tmapScale(aes = "fill",
											value = fill,
											scale = fill.scale,
											legend = fill.legend,
											chart = fill.chart,
											free = fill.free),
						   col = tmapScale(aes = "col",
						   				value = col,
						   				scale = col.scale,
						   				legend = col.legend,
						   				chart = col.chart,
						   				free = col.free),
						   lwd = tmapScale(aes = "lwd",
						   				value = lwd,
						   				scale = lwd.scale,
						   				legend = lwd.legend,
						   				chart = lwd.chart,
						   				free = lwd.free),
						   lty = tmapScale(aes = "lty",
						   				value = lty,
						   				scale = lty.scale,
						   				legend = lty.legend,
						   				chart = lty.chart,
						   				free = lty.free),
						   fill_alpha = tmapScale(aes = "fill_alpha",
						   					   value = fill_alpha,
						   					   scale = fill_alpha.scale,
						   					   legend = fill_alpha.legend,
						   					   chart = fill_alpha.chart,
						   					   free = fill_alpha.free),
						   col_alpha = tmapScale(aes = "col_alpha",
						   					  value = col_alpha,
						   					  scale = col_alpha.scale,
						   					  legend = col_alpha.legend,
						   					  chart = col_alpha.chart,
						   					  free = col_alpha.free)),

		gpar = tmapGpar(fill = "__fill",
						col = "__col",
						shape = NA,
						size = NA,
						fill_alpha = "__fill_alpha",
						col_alpha = "__col_alpha",
						pattern = "fill",
						lty = "__lty",
						lwd = "__lwd",
						linejoin = linejoin,
						lineend = lineend),
		tpar = tmapTpar(area = "AREA"),
		plot.order = plot.order,
		mapping.args = options$mapping.args,
		zindex = zindex,
		group = group,
		group.control = group.control,
		popup.vars = popup.vars,
		popup.format = popup.format,
		hover = hover,
		id = id,
		subclass = c("tm_aes_layer", "tm_layer")))
}

#' @rdname tm_polygons
#' @export
tm_fill = function(...) {
	args = list(...)
	# tricks to make backward comp. work
	if (!("col" %in% names(args))) {
		args["col"] = list(NULL)
	}
	args$called_from = if (names(args)[1] == "") "tm_fill" else  "tm_polygons"
	tm = do.call(tm_polygons, args)
	tm[[1]]$layer = c("fill", "polygons")
	tm
}

#' @rdname tm_polygons
#' @export
tm_borders = function(col = tm_const(), ...) {
	args = list(...)
	if (!("fill" %in% names(args))) {
		args["fill"] = list(NULL)
	}
	args$called_from = "tm_borders"
	args$popup.vars = FALSE
	args$hover = FALSE
	tm = do.call(tm_polygons, c(list(col = col), args))
	tm[[1]]$layer = c("borders", "polygons")
	tm
}


#' @param polygons.only should only polygon geometries of the shape object (defined in [tm_shape()]) be plotted? By default `"ifany"`, which means `TRUE` in case a geometry collection is specified.
#' @rdname tm_polygons
#' @export
opt_tm_polygons = function(polygons.only = "ifany") {
	list(trans.args = list(polygons.only = polygons.only),
		 mapping.args = list())
}

#' Map: x and y labels
#'
#' The x and y labels for maps
#'
#' @param text text of the title
#' @param size font size of the title
#' @param color color
#' @param rotation rotation in degrees
#' @param space space between label and map in number of line heights
#' @param fontface font face
#' @param fontfamily font family
#' @param alpha alpha transparency of the text
#' @param side side: `"top"` or `"bottom"` for `tm_xlab` and `"left"` or `"right"` for `tm_ylab`
#' @export
#' @export
tm_xlab = function(text, size, color, rotation, space, fontface, fontfamily, alpha, side) {
	args = lapply(as.list(rlang::call_match()[-1]), eval, envir = parent.frame())
	args$show = TRUE
	names(args) = paste("xlab", names(args), sep = ".")
	do.call(tm_options, args)
}

#' @export
#' @rdname tm_xlab
tm_ylab = function(text, size, color, rotation, space, fontface, fontfamily, alpha, side) {
	args = lapply(as.list(rlang::call_match()[-1]), eval, envir = parent.frame())
	args$show = TRUE
	names(args) = paste("ylab", names(args), sep = ".")
	do.call(tm_options, args)
}

#' Quick thematic map plot
#'
#' Draw a thematic map quickly. This function is a convenient wrapper of the main
#' plotting method of stacking [`tmap-element`]s. Without arguments or with a
#' search term, this functions draws an interactive map.
#'
#' The first argument is a shape object (normally specified by [tm_shape()]).
#' The next arguments, from `fill` to `raster`, are the aesthetics from the main
#' layers. The remaining arguments are related to the map layout. Any argument
#' from any main layer function, such as [tm_polygons()], can be specified (see `...`).
#' It is also possible to stack [`tmap-element`]s on a `qtm` plot. See examples.
#'
#' By default, a scale bar is shown. This option can be set with [tmap_options()]
#' (argument `qtm.scalebar`). A minimap is shown by default when `qtm` is called
#' without arguments of with a search term. This option can be set with [tmap_options()]
#' (argument `qtm.minimap`).
#'
#' @param shp One of:
#'  * shape object, which is an object from a class defined by the [`sf`][`sf::sf`]
#'    or [`stars`][stars::st_as_stars()] package. Objects from the packages `sp`
#'    and `raster` are also supported, but discouraged.
#'  * Not specified, i.e. `qtm()` is executed. In this case a plain interactive
#'    map is shown.
#'  * An OpenStreetMap search string, e.g. `qtm("Amsterdam")`. In this case a
#'    plain interactive map is shown positioned according to the results of the
#'    search query (from OpenStreetMap nominatim)
#' @param fill,col,size,shape,lwd,lty,fill_alpha,col_alpha Visual variables.
#' @param text,text_col,text_size Visual variables.
#' @param by data variable name by which the data is split, or a vector of two
#'   variable names to split the data by two variables (where the first is used
#'   for the rows and the second for the columns). See also [tm_facets()].
#' @param scale numeric value that serves as the global scale parameter. All font
#'   sizes, symbol sizes, border widths, and line widths are controlled by this value.
#'   The parameters `symbols.size`, `text.size`, and `lines.lwd` can be scaled
#'   separately with respectively `symbols.scale`, `text.scale`, and
#'   `lines.scale`. See also `...`.
#' @param title main title. For legend titles, use `X.legend`, where X is the
#'   layer name (see `...`).
#' @param crs Either a [`crs`][sf::st_crs()] object or a character value
#'   (`PROJ.4` character string). By default, the projection is used that is
#'   defined in the `shp` object itself.
#' @param bbox bounding box. Argument passed on to [tm_shape()]
#' @param basemaps name(s) of the provider or an URL of a tiled basemap.
#'   It is a shortcut to [tm_basemap()]. Set to `NULL` to disable basemaps.
#'   By default, it is set to the tmap option `basemaps`.
#' @param overlays name(s) of the provider or an URL of a tiled overlay map.
#'   It is a shortcut to [tm_tiles()].
#' @param zindex zindex
#' @param group group
#' @param group.control group.control
#' @param style Layout options (see [tm_layout()]) that define the style.
#'   See [tmap_style()] for details.
#' @param format Deprecated, see [tm_format()] for alternatives
#' @param ... arguments associated with the visual variables are passed on
#'   to the layer functions [tm_polygons()], [tm_lines()], [tm_symbols()],
#'   and [tm_text()].
#'   For instance, `fill.scale` is the scale specifications of the fill color
#'   of polygons (see [tm_polygons()]).
#' @return A [`tmap-element`]
#' @example ./examples/qtm.R
#' @references Tennekes, M., 2018, {tmap}: Thematic Maps in {R},
#'   Journal of Statistical Software, 84(6), 1-39, \doi{10.18637/jss.v084.i06}
#' @export
qtm = function(shp = NULL,
				fill = tmap::tm_const(),
				col = tmap::tm_const(),
				size = tmap::tm_const(),
				shape = tmap::tm_const(),
				lwd = tmap::tm_const(),
				lty = tmap::tm_const(),
				fill_alpha = tmap::tm_const(),
				col_alpha = tmap::tm_const(),
				text = tmap::tm_const(),
				text_col = tmap::tm_const(),
				text_size = tmap::tm_const(),
				by = NULL,
				scale = NULL,
				title = NULL,
				crs = NULL,
				bbox = NULL,
				basemaps = NA,
				overlays = NA,
				zindex = NA,
				group = NA,
				group.control = "check",
				style = NULL,
				format = NULL,
				...) {

	args = lapply(as.list(rlang::call_match(defaults = TRUE, dots_expand = TRUE)[-1]), eval, envir = parent.frame())
	shp_name = deparse(substitute(shp))[1]
	args_called = names(rlang::call_match(dots_expand = TRUE)[-1])

	if (any(v3_only("qtm") %in% args_called)) {
		v3_start_message()
		args_new = list(args = args, called = args_called) |>
			v3_instead("symbols.size", "size", "qtm", extra_called = "shape") |>
			v3_instead("symbols.col", "size", "qtm", extra_called = "shape") |>
			v3_instead("dots.col", "fill", "qtm") |>
			v3_instead("lines.lwd", "lwd", "qtm") |>
			v3_instead("lines.col", "col", "qtm") |>
			v3_instead("raster", "col", "qtm") |>
			v3_instead_value("borders", "col", "qtm", value_old = NULL, value_new = NA) |>
			v3_instead("text.size", "text_size", "qtm") |>
			v3_instead("text.col", "text_col", "qtm") |>
			v3_instead("projection", "crs", "qtm")
		args = args_new$args
		args_called = args_new$called
	}

	o = tmap_options_mode(mode.specific = FALSE)
	show.warnings = o$show.warnings

	if (missing(shp) || is.character(shp)) {
		viewargs = args[intersect(names(args), names(formals(tm_view)))]
		if (!missing(shp)) viewargs$bbox = shp

		g = c(tm_basemap(basemaps), tm_tiles(overlays), do.call("tm_view", viewargs))
		if (o$qtm.minimap) g = c(g, tm_minimap())
		if (o$qtm.scalebar) g = c(g, tm_scalebar())
		if (o$qtm.mouse_coordinates) g = c(g, tm_mouse_coordinates())

		attr(g, "qtm_shortcut") = TRUE
		class(g) = "tmap"
		return(g)
	}


	if (!missing(shp)) {
		nms_shp = intersect(args_called, names(formals(tm_shape)))
		g = do.call(tm_shape, c(args[nms_shp], list(name = shp_name)))

		is_rst = inherits(shp, c("stars", "SpatRaster"))

		prefix = list(tm_polygons = "fill",
					  tm_lines = "lines",
					  tm_symbols = c("symbols", "dots"),
					  tm_raster = "raster")

		if (is_rst) {
			nms_rst = intersect(names(args), funs_v4$tm_raster)
			args_rst = args[nms_rst]

			if (!any(c("col", "raster") %in% args_called)) {
				args_rst$col = tm_vars()
			}

			nms_rst_v3 = names(args)[substr(names(args), 1, 7) == "raster."]
			args_rst_v3 = args[nms_rst_v3]
			names(args_rst_v3) = substr(names(args_rst_v3), 8, nchar(names(args_rst_v3)))

			g = g + do.call(tm_raster, c(args_rst, args_rst_v3))
		} else {

			for (f in c("tm_polygons", "tm_lines", "tm_symbols")) {
				nms_f = intersect(names(args), funs_v4[[f]])
				args_f = args[nms_f]

				nms_other = intersect(setdiff(names(args), c(nms_f, nms_shp, "basemaps", "overlays", "style", "format")), args_called)
				v3_nms = intersect(nms_other, unlist(lapply(prefix[[f]], function(p) paste(p, setdiff(v3_only(f), "scale"), sep = "."))))
				args_v3 = args[v3_nms]
				names(args_v3) = sub("^[^.]+[.]", "", names(args_v3))
				if (f == "tm_symbols") {
					if (!"shape" %in% args_called) args_f$shape = NULL
					if (!"col" %in% args_called) args_f$col = NULL
				}
				if (f == "tm_lines") {
					if (!"col" %in% args_called) args_f$col = NULL
				}

				args_f$called_from = "qtm"
				options = opt_tm_sf()[[c(tm_polygons = "polygons", tm_lines = "lines", tm_symbols = "points")[f]]]
				g = g + do.call(f, c(args_f, args_v3, list(options = options)))
			}
			if ("text" %in% args_called) {
				args_txt = args[substr(names(args), 1, 5) == "text_" | names(args) == "text"]
				names(args_txt) = sub("text_", "", names(args_txt))

				v3_nms = substr(names(args), 1, 5) == "text."
				args_v3 = args[v3_nms]
				names(args_v3) = sub("^[^.]+[.]", "", names(args_v3))

				args_f$called_from = "qtm"
				g = g + do.call(tm_text, c(args_txt, args_v3))
			}
		}

		nms_fct = intersect(names(args), names(formals(tm_facets)))
		if (length(nms_fct)) {
			g = g + do.call(tm_facets, args[nms_fct])
		}
	} else {
		g = NULL
	}


	if ("basemaps" %in% args_called || o$basemap.show) {
		g = g + do.call(tm_basemap, list(server = args$basemaps))
	}
	if ("tiles" %in% args_called || o$tiles.show) {
		g = g + do.call(tm_tiles, list(server = args$overlays))
	}
	if ("style" %in% args_called) {
		g = g + tm_style(args$style)
	}
	if ("title" %in% args_called) {
		g = g + tm_title(args$title)
	}

	if (o$qtm.scalebar) g = g + tm_scalebar()
	if (o$qtm.minimap) g = g + tm_minimap()
	if (o$qtm.mouse_coordinates) g = g + tm_mouse_coordinates()

	if (is.null(g)) {
		message_qtm_empty()
		invisible(NULL)
	} else {
		assign("last_map_new", rlang::call_match(), envir = .TMAP)
		attr(g, "qtm_shortcut") = FALSE
		g
	}
}

#' @export
#' @rdname tmapGridLeaflet
tmapLeafletDataPlot.tm_data_circles = function(a, shpTM, dt, pdt, popup.format,
											   hdt, idt, gp, bbx,
											   facet_row, facet_col, facet_page,
											   id, pane, group, glid, o, ...) {
	dots         = list(...)
	ptdt         = dots$ptdt
	popup.layout = dots$popup.layout

	lf      = get_lf(facet_row, facet_col, facet_page)
	rc_text = frc(facet_row, facet_col)

	# Filter to rows where size is known (no shape filter: circles have none)
	res = select_sf(shpTM, dt[!is.na(dt$size), ])
	shp = res$shp
	if (o$crs_leaflet$crsClass == "L.CRS.Simple") {
		shp = sf::st_set_crs(shp, NA)
	}
	dt = res$dt

	gp = impute_gp(gp, dt)
	# NOTE: do NOT call rescale_gp — that converts size to "lines" units.
	# For tm_circles, gp$size is already in metres (the direct scale output).

	coords = sf::st_coordinates(shp)
	cp     = expand_coords_gp(coords, gp, ndt = nrow(dt))
	coords = cp$coords
	gp     = cp$gp

	if (cp$expanded) {
		shpTM_match = match(shpTM$tmapID_expanded, shpTM$tmapID)
	} else {
		shpTM_match = TRUE
	}

	# ---- IDs / popups / hover ------------------------------------------------
	idt_null = is.null(idt)
	if (!idt_null) {
		idt = idt$id[match(dt$tmapID__[shpTM_match], idt$tmapID__)]
	} else {
		idt = sprintf("%07d", dt$tmapID__)[shpTM_match]
	}

	if (!is.null(hdt)) {
		hdt = hdt$hover[match(dt$tmapID__[shpTM_match], hdt$tmapID__)]
		hdt = lapply(hdt, htmltools::HTML, FUN.VALUE = character(1))
	}

	if (!is.null(ptdt)) {
		pttl = ptdt$title[match(dt$tmapID__[shpTM_match], ptdt$tmapID__)]
	} else {
		pttl = NULL
	}

	if (is.null(pdt)) {
		popups = NULL
	} else {
		mtch = match(dt$tmapID__[shpTM_match], pdt$tmapID__)
		pdt  = pdt[mtch][, tmapID__ := NULL]
		if (!is.null(pttl)) {
			popups = view_format_popups(id = pttl, titles = names(pdt), values = pdt,
										format = popup.format, layout = popup.layout)
		} else if (idt_null) {
			popups = view_format_popups(titles = names(pdt), values = pdt,
										format = popup.format, layout = popup.layout)
		} else {
			popups = view_format_popups(id = idt, titles = names(pdt), values = pdt,
										format = popup.format, layout = popup.layout)
		}
	}

	interactive = (!is.null(pdt) || !is.null(hdt))
	opt = leaflet::pathOptions(interactive = interactive, pane = pane)

	# ---- Radius in metres ----------------------------------------------------
	# gp$size is the direct scale output, interpreted as metres.
	# With tm_scale_asis() the raw data values are used as-is.
	# With other scales, users control the output range via values.range /
	# values.scale in the scale function.
	radius_m = gp$size

	# ---- Colour / stroke parameters -----------------------------------------
	gp2 = gp_to_lpar(gp, mfun = "Symbols", size_factor = 14)

	shp2 = sf::st_as_sf(as.data.frame(coords), coords = c("X", "Y"),
						 crs = sf::st_crs(shp))

	idt = submit_labels(idt, "circles", pane, group)

	lf |>
		leaflet::addCircles(
			data        = shp2,
			layerId     = idt,
			radius      = radius_m,           # metres — scales with zoom natively
			color       = gp2$color,          # stroke colour
			opacity     = gp2$opacity,
			weight      = gp2$strokeWidth,
			dashArray   = gp2[["stroke-dasharray"]],
			fillColor   = gp2$fillColor,
			fillOpacity = gp2$fillOpacity,
			options     = opt,
			group       = group,
			label       = hdt,
			popup       = popups
		) |>
		assign_lf(facet_row, facet_col, facet_page)

	NULL
}

#' @export
#' @rdname tmap_internal
tmapScaleIntervals = function(x1, scale, legend, chart, o, aes, layer, layer_args, sortRev, bypass_ord, submit_legend = TRUE) {
	cls = data_class(x1, midpoint_enabled = !is.null(scale$midpoint))
	maincls = class(scale)[1]

	if (attr(cls, "unique") && is.null(scale$breaks)) {
		scale$ticks = NA
		cli::cli_inform("The visual variable {.arg {aes}} of the layer {.str {layer}} contains a unique value. Therefore a discrete scale is applied (tm_scale_discrete).")
		return(tmapScaleDiscrete(x1, scale, legend, chart, o, aes, layer, layer_args, sortRev, bypass_ord, submit_legend = submit_legend))

	}

	if (!(cls[1] %in% c("num", "datetime", "date"))) {
		if (!is.factor(x1)) x1 = as.factor(x1)
		x1 = as.integer(x1)
		cls = c("num", "int", "seq")
		warning(maincls, " is supposed to be applied to numerical or date/time data", call. = FALSE)
	}

	x1 = without_units(x1)

	if (aes %in% c("pattern")) stop("tm_scale_intervals cannot be used for layer ", layer, ", aesthetic ", aes, call. = FALSE)

	if (!(aes %in% c("fill", "col")) && scale$label.style == "continuous") {
		scale$label.style = "discrete"
		cli::cli_warn("{.field tm_scale_intervals} {.arg label.style} can only be {.str continuous} for fill and color, not for {aes}")
	}


	scale = get_scale_defaults(scale, o, aes, layer, cls)

	show.messages <- o$show.messages
	show.warnings <- o$show.warnings


	fun = paste0("interval_", cls[1])

	# pre-checks
	midpoint = NULL
	scale = within(scale, {
		allna = all(is.na(x1))
		if (anyDuplicated(breaks)) stop("breaks specified in the ", aes,  ".scale scaling function contains duplicates.", call. = FALSE)
		if (!is.null(breaks) && length(breaks) < 2 && cls[1] == "num") stop("breaks should contain at least 2 numbers", call. = FALSE)

		udiv = identical(use_div(breaks, midpoint), TRUE)

		if (allna && is.null(breaks)) {
			if (show.messages) message("Variable(s) \"", paste(aes, collapse = "\", \""), "\" only contains NAs. Legend disabled for tm_scale_intervals, unless breaks are specified")
			chart = within(chart, {
				labels = label.na
				vvalues = c(value.na, value.na)
				breaks = c(0, 1)
				na.show = TRUE
				x1 = x1[1]
			})
			return = tmapScale_returnNA(n = length(x1), legend = legend, chart = chart, value.na = value.na, label.na = label.na, label.show = label.show, na.show = legend$na.show, sortRev = sortRev, bypass_ord = bypass_ord)
		}
	})

	if (!is.null(scale$return)) return(scale$return)

	# data-type dependent processing
	scale = do.call(fun, list(scale = scale, x1 = x1, aes = aes, layer = layer, show.messages = show.messages, show.warnings = show.warnings))

	# visual variable processing
	with(scale, {
		fun_getVV = paste0("tmapValuesVV_", aes)
		VV = do.call(fun_getVV, list(x = values, value.na = value.na, isdiv = isdiv, n = n, dvalues = breaks, midpoint = midpoint, range = values.range, scale = values.scale * o$scale, are_breaks = TRUE, rep = values.repeat, o = o))

		vvalues = VV$vvalues
		value.na = VV$value.na

		sfun = paste0("tmapValuesScale_", aes)
		cfun = paste0("tmapValuesColorize_", aes)
		if (is.na(value.neutral)) value.neutral = VV$value.neutral else value.neutral = do.call(sfun, list(x = do.call(cfun, list(x = value.neutral, pc = o$pc)), scale = values.scale))

		# check for size[1] = 0
		if (aes %in% c("size", "lwd")) {
			if (vvalues[1] == 0) {
				message_scale_interval_value0(aes, values, layer)
			}
		}

		mfun = paste0("tmapValuesSubmit_", aes)
		vvalues = do.call(mfun, list(x = vvalues, args = layer_args))
		value.na = do.call(mfun, list(x = value.na, args = layer_args))
		value.neutral = do.call(mfun, list(x = value.neutral, args = layer_args))


		ids = classInt::findCols(q)
		vals = vvalues[ids]
		isna = is.na(vals)
		anyNA = any(isna)
		na.show = update_na.show(label.show, legend$na.show, anyNA)

		if (is.null(sortRev)) {
			ids = NULL
		} else if (is.na(sortRev)) {
			ids[] = 1L
		} else if (sortRev) {
			ids = (as.integer(n) + 1L) - ids
		}

		if (anyNA) {
			vals[isna] = value.na
			if (!is.null(sortRev)) ids[isna] = 0L
		}

		# detransform log
		if (is.log) {
			if (any((breaks %% 1) != 0)) message("non-rounded breaks occur, because style = \"log10_pretty\" is designed for large values")
			breaks <- 10^breaks
		}

		# create legend values
		#values = breaks[-nbrks]

		if (label.style == "discrete") {
			if (is.null(labels)) {
				labels = do.call("fancy_breaks", c(list(vec=breaks, as.count = as.count, intervals=TRUE, interval.closure=int.closure), label.format))
			} else {
				if (length(labels)!=nbrks-1 && show.warnings) warning("number of legend labels should be ", nbrks-1, call. = FALSE)
				labels = rep(labels, length.out=nbrks-1)
				attr(labels, "align") <- label.format$text.align
			}
		} else {
			if (is.null(labels)) {
				labels = do.call("fancy_breaks", c(list(vec=breaks, as.count = FALSE, intervals=FALSE, interval.closure=int.closure), label.format))
			} else {
				if (length(labels) != length(breaks)) cli::cli_abort("{.field tm_scale_intervals} {.arg labels} should have length {length(breaks)}")
				#labels = c(labels, {if (na.show) label.na else NULL})
			}
		}



		if (legend$reverse) {
			labels.brks <- attr(labels, "brks")
			labels.align <- attr(labels, "align")
			labels <- rev(labels)
			if (!is.null(labels.brks)) {
				attr(labels, "brks") = labels.brks[length(labels):1L,]
			}
			attr(labels, "align") = labels.align
			vvalues_rev = rev(vvalues)
		} else {
			vvalues_rev = vvalues
		}



		if (na.show) {
			labels.brks = attr(labels, "brks")
			labels.align = attr(labels, "align")
			#labels <- c(labels, legend.NA.text)
			if (!is.null(labels.brks)) {
				labels <- c(labels, paste(label.na, " ", sep = ""))
				attr(labels, "brks") = rbind(labels.brks, rep(nchar(label.na) + 2, 2))
			} else {
				labels = c(labels, label.na)
			}
			attr(labels, "align") = labels.align
			vvalues = c(vvalues, value.na)
			vvalues_rev = c(vvalues_rev, value.na)
		}

		if (label.style == "discrete") {
			legend = within(legend, {
				nitems = length(labels)
				labels = labels
				dvalues = breaks #not used?
				vvalues = vvalues_rev
				vneutral = value.neutral
				na.show = get("na.show", envir = parent.env(environment()))
				scale = "intervals"
				layer_args = layer_args
			})

		} else {
			if ((o$continuous.nclass_per_legend_break %% 2) != 0) cli::cli_abort("{.field options} the tmap option {.arg continuous.nclass_per_legend_break} should be even")
			labels_select = c(rep(label.select, length.out = length(breaks)), {if (na.show) TRUE else NULL})

			vvalues = c(unlist(mapply(function(hd, tl) {
				paste(c(rep(hd, o$continuous.nclass_per_legend_break/2),
						rep(tl, o$continuous.nclass_per_legend_break/2)),
						collapse = "_")
			}, c(NA_character_, vvalues_rev[1L:(length(breaks)-1L)]), c(vvalues_rev[1L:(length(breaks)-1L)], NA_character_))), {if (na.show) value.na else NULL})
			nitems = length(labels)


			legend = within(legend, {
				nitems = nitems
				labels = labels
				dvalues = breaks #not used?
				vvalues = vvalues
				vneutral = value.neutral
				na.show = get("na.show", envir = parent.env(environment()))
				scale = "intervals"
				layer_args = layer_args
				is_discrete = TRUE

				# continuous legend specific:
				labels_select = labels_select
				tr = trans_identity
				limits = range(breaks)
			})
		}

		chartFun = paste0("tmapChart", toTitleCase(chart$summary))

		chartArgs = list(chart,
						 bin_colors = vvalues,
						 breaks_def = breaks,
						 na.show = na.show,
						 x1 = x1)

		# if breaks are user specified via tm_chart, use them
		if ("breaks" %in% names(chart) && !is.null(chart$breaks)) {
			chartArgs["bin_colors"] = list(NULL)
			chartArgs$breaks_def = chart$breaks
		}

		chart = do.call(chartFun, chartArgs)

		if (submit_legend) {
			if (bypass_ord) {
				format_aes_results(vals, legend = legend, chart = chart)
			} else {
				format_aes_results(vals, ids, legend, chart = chart)
			}
		} else {
			list(vals = vals, ids = ids, legend = legend, chart = chart, bypass_ord = bypass_ord)
		}
	})
}

interval_num = function(scale, x1, aes, layer, show.messages, show.warnings) {
	within(scale, {
		local({
			styles = c("fixed", "sd", "equal", "pretty", "quantile", "kmeans", "hclust", "bclust", "fisher", "jenks", "dpih", "headtails", "maximum", "box", "log10", "log10_pretty")
			fn_call <-  call(paste0("tm_", layer[1]))
			if (!(style %in% styles)) {
				cli::cli_abort("Invalid style. Style should be one of {.str {styles}}",
					call = fn_call
				)
			}
		})


		if (!any(style == c("pretty", "log10_pretty", "fixed"))) {
			if (identical(as.count, TRUE) && show.warnings) warning("as.count not implemented for styles other than \"pretty\", \"log10_pretty\" and \"fixed\"", call. = FALSE)
			as.count = FALSE
		}
		if (is.na(as.count)) {
			as.count = is.integer(x1) && !any(!is.na(x1) & x1 < 0)
		}

		if (as.count) {
			if (interval.closure != "left" && show.warnings) warning("For as.count = TRUE, interval.closure will be set to \"left\"", call. = FALSE)
			interval.closure = "left"
		}

		#breaks.specified <- !is.null(breaks)
		is.log = (style == "log10_pretty")
		if (is.log && !attr(label.format, "big.num.abbr.set")) label.format$big.num.abbr = NA

		if (style == "log10_pretty") {
			x1 = log10(x1)
			style = "fixed"
			breaks = seq(floor(min(x1, na.rm = TRUE)), ceiling(max(x1, na.rm=TRUE)))
		} else if (as.count && style == "pretty") {
			breaks = prettyCount(x1, n=n)
			style <- "fixed"
		} else if (as.count && style == "fixed") {
			breaks[length(breaks)] = breaks[length(breaks)] + 1L
		}

		q = num2breaks(x=x1, n=n, style=style, breaks=breaks, interval.closure=interval.closure, var=paste(layer, aes, sep = "-"), as.count = as.count, args = style.args)

		breaks = q$brks
		nbrks = length(breaks)
		n = nbrks - 1

		int.closure <- attr(q, "intervalClosure")

		# update range if NA (automatic)
		if (is.na(values.range[1])) {
			fun_range = paste0("tmapValuesRange_", aes)
			values.range = do.call(fun_range, args = list(x = values, n = n, isdiv = udiv))
		}
		if (length(values.range) == 1 && !is.na(values.range[1])) values.range = c(0, values.range)

		check_values(layer, aes, values)

		fun_isdiv = paste0("tmapValuesIsDiv_", aes)

		isdiv = !is.null(midpoint) || do.call(fun_isdiv, args = list(x = values))

		# determine midpoint
		if ((is.null(midpoint) || is.na(midpoint)) && isdiv) {
			rng <- range(x1, na.rm = TRUE)
			if (rng[1] < 0 && rng[2] > 0 && is.null(midpoint)) {

				if (show.messages)  message_midpoint(aes, "tm_scale_intervals")
				midpoint <- 0
			} else {
				if ((n %% 2) == 1) {
					# number of classes is odd, so take middle class (average of those breaks)
					midpoint <- mean.default(breaks[c((n+1) / 2, (n+3) / 2)])
				} else {
					midpoint <- breaks[(n+2) / 2]
				}
			}
		}
	})
}

interval_date = function(scale, x1, aes, layer, show.messages, show.warnings) {
	interval_datetime(scale, x1, aes, layer, show.messages, show.warnings)
}


interval_datetime = function(scale, x1, aes, layer, show.messages, show.warnings) {
	style = NULL
	interval.closure = NULL
	udiv = NULL
	show.messages = NULL
	within(scale, {
		local({
			styles = c("fixed", "pretty")
			fn_call <-  call(paste0("tm_", layer[1]))
			if (!(style %in% styles)) {
				cli::cli_abort("Invalid style. For date/time data variables, style should be one of {.str {styles}}",
							   call = fn_call
				)
			}
		})


		if (identical(as.count, TRUE) && show.warnings) {
			warning("as.count not applicable for data/time data variables", call. = FALSE)
			as.count = FALSE
		}

		#if (style == "pretty") {
		if (is.null(breaks)) {
			breaks = pretty_datetime_seq(x=x1, n = n)
		} else if (length(breaks) == 1L) {
			breaks = pretty_datetime_seq3(x=x1, by = breaks)
		}


		q <- list(var=x1,
				  brks=breaks)
		attr(q, "style") <- "fixed"
		attr(q, "nobs") <- nobs
		attr(q, "intervalClosure") <- interval.closure
		class(q) <- "classIntervals"

		breaks = q$brks
		nbrks = length(breaks)
		n = nbrks - 1

		is.log = FALSE

		int.closure <- attr(q, "intervalClosure")

		# update range if NA (automatic)
		if (is.na(values.range[1])) {
			fun_range = paste0("tmapValuesRange_", aes)
			values.range = do.call(fun_range, args = list(x = values, n = n, isdiv = udiv))
		}
		if (length(values.range) == 1 && !is.na(values.range[1])) values.range = c(0, values.range)

		check_values(layer, aes, values)

		fun_isdiv = paste0("tmapValuesIsDiv_", aes)

		isdiv = !is.null(midpoint) || do.call(fun_isdiv, args = list(x = values))

		# determine midpoint
		if ((is.null(midpoint) || is.na(midpoint)) && isdiv) {
			rng <- range(x1, na.rm = TRUE)

			if ((n %% 2) == 1) {
				# number of classes is odd, so take middle class (average of those breaks)
				midpoint <- mean(breaks[c((n+1) / 2, (n+3) / 2)])
			} else {
				midpoint <- breaks[(n+2) / 2]
			}

		}
	})

}

.doc_opt = function(a) {
	x = strsplit(a, split = ".", fixed = TRUE)[[1]]
	cls = head(x, -1)
	what = tail(x, 1)

	cls_lookup = c(inner = "inner space (inside the frame)",
				   outer = "outer space (outside the frame",
				   item_text = "space between item and text",
				   meta = "'meta' space (reserved for outside map components)")

	what_lookup = c(alpha = "alpha transparency",
					asp = "asp ration",
					bbox = "bounding box",
					breaks = "break values",
					col = "color",
					color = "color",
					crs = "coordinate reference system (CRS)",
					double_line = "double line",
					dpi = "dots per inch",
					earth_boundary = "earth_boundary",
					earth_datum = "earth_datum",
					extra = "extra arguments",
					flip = "flip",
					fontface = "font face",
					fontfamily = "font family",
					format = "format",
					frame = "frame",
					height = "height",
					inside_frame = "inside_frame",
					labels = "labels",
					landscape = "landscape",
					light = "light",
					limits = "limits",
					lines = "lines",
					lwd = "line width",
					margin = "margin",
					margins = "margins",
					max = "max",
					max_cells = "max_cells",
					messages = "messages",
					minimap = "minimap",
					modes = "modes",
					mouse_coordinates = "mouse_coordinates",
					na = "na",
					nclass_per_legend_break = "nclass_per_legend_break",
					nclasses = "nclasses",
					ndiscr = "ndiscr",
					neutral = "neutral",
					north = "north",
					null = "null",
					offset = "offset",
					only = "only",
					orientation = "orientation",
					padding = "padding",
					portrait = "portrait",
					pos = "pos",
					position = "position",
					r = "r (radius)",
					range = "range",
					resize_as_group = "resize_as_group",
					reverse = "reverse",
					rot = "rot",
					rotation = "rotation",
					saturation = "saturation",
					scale = "scale",
					scalebar = "scalebar",
					scales = "scales",
					sepia_intensity = "sepia_intensity",
					server = "server",
					show = "visibility",
					side = "side",
					size = "size",
					space = "space",
					stack = "stack",
					stack_margin = "stack_margin",
					text = "text",
					ticks = "ticks",
					toggle = "toggle",
					type = "type",
					unit = "unit",
					var = "var",
					view = "view",
					warnings = "warnings",
					width = "width",
					absolute_fontsize = "absolute fontsize",
					x = "x",
					y = "y",
					xlab = "`x` label",
					ylab = "`y` label",
					zoom = "zoom")

	extra_lookup = c(margins = " A vector of 4 values: bottom, left, top, right. The unit is the device height (for bottom and top) or width (for left and right).",
					 double_line = " `TRUE` or `FALSE`.",
					 extra = " A list of arguments.",
					 padding = " A vector of 4 values: bottom, left, top, right. The unit is the device height (for bottom and top) or width (for left and right).",
					 position = " A tm_pos object, or a shortcut of two values: horizontal (left, center, right) and vertical (top, center, bottom). See tm_pos for details",
					 space = " In terms of number of text line heights.",
					 fontfamily = " See `graphics::par`, option 'family'.",
					 fontface = " See `graphics::par`, option 'font'.",
					 absolute_fontsize = " So far, only used to calculate legend dimensions",
					 lwd = " See `graphics::par`, option 'lwd'.",
					 show = " `TRUE` or `FALSE`.")

	what_text = unname(what_lookup[what])
	if (is.na(what_text)) what_text = what

	cls_text = sapply(cls, function(x) {
		res = unname(cls_lookup[x])
		if (is.na(res)) x else res
	}, USE.NAMES = FALSE)

	extra_text = unname(extra_lookup[what])
	if (is.na(extra_text)) extra_text = ""


	if (length(cls) == 0L) {
		cls_texts = " (overall)"
	} else {
		cls_texts = paste(" of the", paste(rev(cls_text), collapse = " of the "))
	}
	paste0("The ", what_text, cls_texts, ".", extra_text)
}

.doc_vv = function(x) {
	lookup = c(fill = "fill color",
			   col = "color",
			   fill_alpha = "fill color transparency",
			   col_alpha = "color transparency",
			   lwd = "line width",
			   lty = "line type",
			   size = "size",
			   text = "text",
			   fontface = "font face")
	txt = unname(lookup[x])
	if (is.na(txt)) txt = x
	paste0("Visual variable that determines the ", txt, ". See details.")
}

.tmap_article_url = "https://r-tmap.github.io/tmap/articles/"

# Single source of truth: vignette slug -> link label
.tmap_vignettes = c(
	basics_vv            = "Visual variables",
	basics_scales        = "Scales",
	basics_modes         = "Modes",
	basics_facets        = "Facets",
	basics_legends       = "Legends",
	basics_basemaps      = "Basemaps",
	basics_components    = "Components",
	basics_charts        = "Charts",
	basics_layout        = "Layout",
	basics_exporting     = "Exporting",
	foundations_gg       = "Grammar of Graphics",
	foundations_crs      = "Map projections (CRS)",
	foundations_units    = "Units",
	examples_choro_World = "Choropleth (World)",
	examples_choro_NLD   = "Choropleth (Netherlands)",
	examples_bubble      = "Bubble map",
	examples_raster      = "Raster map",
	examples_topo_Africa = "Topographic map (Africa)",
	examples_terrain     = "Terrain map",
	examples_gridmaps    = "Gridmap",
	examples_biv_choro   = "Bivariate choropleth",
	versus_ggplot2       = "tmap vs ggplot2",
	versus_mapview       = "tmap vs mapview",
	versus_mapsf         = "tmap vs mapsf",
	ext_mapgl            = "tmap.mapgl",
	ext_cartogram        = "tmap.cartogram",
	ext_glyphs           = "tmap.glyphs",
	ext_networks         = "tmap.networks",
	adv_legends          = "Legends (advanced)",
	adv_inset_maps       = "Inset maps",
	adv_animations       = "Animations",
	adv_shiny            = "Shiny integration",
	adv_ggplot2          = "ggplot2 integration",
	adv_multivariate     = "Multivariate visual variables",
	adv_positions        = "Positioning of components",
	adv_comp_group       = "Grouping of components",
	adv_options          = "Options and styles",
	adv_margins          = "Margins and aspect ratio",
	adv_groups           = "Layer groups (view mode)",
	adv_extensions       = "Extending tmap",
	adv_blend            = "Layer blending"
)

# Build a comma-separated list of markdown links from vignette slugs
.doc_links = function(slugs) {
	labels = .tmap_vignettes[slugs]
	if (anyNA(labels)) {
		stop("Unknown tmap vignette slug(s): ",
			 paste(slugs[is.na(labels)], collapse = ", "), call. = FALSE)
	}
	paste0("[", labels, "](", .tmap_article_url, slugs, ")", collapse = ", ")
}

.doc_see_also_layers = function(layer) {
	common = c("basics_vv", "basics_scales", "basics_legends", "basics_facets", "foundations_units", "adv_blend")
	# layer-specific examples — review/adjust these mappings to taste
	examples = switch(layer,
					  polygons = c("examples_choro_World", "examples_choro_NLD", "examples_biv_choro", "examples_gridmaps"),
					  symbols  = c("examples_bubble", "examples_terrain"),
					  lines = "examples_terrain",
					  text = c("examples_topo_Africa", "examples_terrain"),
					  raster   = c("examples_raster", "examples_terrain"),
					  character(0)  # default: common links only (lines, text, ...)
	)
	.doc_links(c(common, examples))
}

.doc_see_vig = function(x, cap = FALSE) {
	init = ifelse(cap, "V", "v")
	paste0(init, "ignette: ", .doc_links(x))
}

.doc_see_also_comp = function() {
	.doc_links(c("basics_components", "adv_positions", "adv_comp_group", "examples_choro_NLD"))
}

.doc_see_also_scales = function() {
	.doc_links(c("basics_scales", "basics_vv", "adv_multivariate"))
}

.doc_see_also_legend = function(advanced = FALSE, biv = FALSE) {
	biv_txt = if (biv) "examples_biv_choro" else NULL
	if (advanced) {
		.doc_links(c("adv_legends", "adv_positions", biv_txt))
	} else {
		.doc_links(c("basics_legends", "adv_legends", "basics_components", "adv_positions", biv_txt))
	}
}

.doc_see_also_chart = function() {
	.doc_links(c("basics_charts", "basics_components", "adv_positions"))
}

.doc_see_also_layout = function() {
	.doc_links(c("basics_layout", "adv_margins", "adv_options", "examples_choro_World", "examples_terrain"))
}

.doc_see_also_facets = function() {
	.doc_links(c("basics_facets", "adv_animations", "examples_gridmaps"))
}

.doc_see_also_animations = function() {
	.doc_links(c("adv_animations",  "ext_cartogram"))
}

.doc_see_also_modes = function() {
	.doc_links(c("basics_modes"))
}

.doc_see_also_export = function() {
	.doc_links(c("basics_exporting"))
}

.doc_see_also_shiny = function() {
	.doc_links(c("adv_shiny"))
}

.doc_see_also_basemaps = function() {
	.doc_links(c("basics_basemaps", "basics_modes"))
}

.doc_see_also_groups = function() {
	.doc_links(c("adv_groups"))
}

.doc_see_also_shape = function() {
	.doc_links(c("foundations_crs", "examples_choro_NLD", "examples_choro_World", "examples_terrain"))
}

.doc_see_also_insets = function() {
	.doc_links(c("adv_inset_maps", "adv_ggplot2"))
}

# --- dispatch carrier -------------------------------------------------------

tmap_graphics_class <- function(mode = NA) {
	if (is.na(mode)) mode <- getOption("tmap.mode")
	gs <- tmap_graphics_name(mode)
	structure(list(mode = mode, gs = gs), class = c(gs, "tmapGS"))
}

# --- generics (exported so extension pkgs can add methods) ------------------

#' @rdname renderTmap
#' @export
renderTmapGS <- function(x, ...) UseMethod("renderTmapGS")

#' @rdname renderTmap
#' @export
tmapOutputGS <- function(x, ...) UseMethod("tmapOutputGS")

#' @rdname renderTmap
#' @export
tmapProxyGS <- function(x, ...) UseMethod("tmapProxyGS")

#' @exportS3Method
renderTmapGS.default <- function(x, ...)
	stop('renderTmap does not (yet) support mode "', x$mode, '"', call. = FALSE)

#' @exportS3Method
tmapOutputGS.default <- function(x, ...)
	stop('tmapOutput does not (yet) support mode "', x$mode, '"', call. = FALSE)

#' @exportS3Method
tmapProxyGS.default <- function(x, ...)
	stop('tmapProxy does not (yet) support mode "', x$mode, '"', call. = FALSE)

text_width_npc = function(txt, space = TRUE, to_height = FALSE) {
	brks = attr(txt, "brks")
	if (is.null(brks)) {
		if (space) txt = paste(txt, " ", sep = "")

		if (to_height) {
			convertHeight(convertWidth(stringWidth(txt), "inch"), "npc", TRUE)
		} else {
			convertWidth(stringWidth(txt), "npc", TRUE)
		}

	} else {
		txt_splits = split_legend_labels(txt, brks)
		spx = if (space) convertWidth(stringWidth(" "), "npc", TRUE) * 1.5 else 0
		res = lapply(txt_splits, function(tx) convertWidth(stringWidth(tx), "npc", TRUE) + spx)

		max1 = max(sapply(res, "[", 1))
		max2 = max(sapply(res, "[", 2))
		max3 = max(sapply(res, "[", 3))
		#r3 = sapply(res, "[", 3)
		widths = max1 + max2 + max3 #r3

		attr(widths, "cw") =  do.call(rbind, res)
		widths
	}
}

text_width_inch = function(txt, space = TRUE) {
	brks = attr(txt, "brks")
	if (is.null(brks)) {
		if (space) txt = paste(txt, " ", sep = "")

		convertWidth(stringWidth(txt), "inch", TRUE)

	} else {
		txt_splits = split_legend_labels(txt, brks)
		spx = if (space) convertWidth(stringWidth(" "), "inch", TRUE) * 1.5 else 0
		res = lapply(txt_splits, function(tx) convertWidth(stringWidth(tx), "inch", TRUE) + spx)

		max1 = max(sapply(res, "[", 1))
		max2 = max(sapply(res, "[", 2))
		max3 = max(sapply(res, "[", 3))
		#r3 = sapply(res, "[", 3)
		widths = max1 + max2 + max3 #r3

		attr(widths, "cw") =  do.call(rbind, res)
		widths
	}
}

text_height_npc = function(txt, to_width = FALSE) {
	if (to_width) {
		convertWidth(convertHeight(stringHeight(txt), "inch"), "npc", TRUE)
	} else {
		convertHeight(stringHeight(txt), "npc", TRUE)
	}
}

text_height_inch = function(txt, to_width = FALSE) {
	if (to_width) {
		grid::convertWidth(grid::convertHeight(stringHeight(txt), "inch"), "inch", TRUE)
	} else {
		grid::convertHeight(stringHeight(txt), "inch", TRUE)
	}
}

split_legend_labels = function(txt, brks) {
	lapply(1L:length(txt), function(i) {
		c(substr(txt[i], 1, brks[i,1]-2),
		  substr(txt[i], brks[i,1], brks[i,2]-2),
		  substr(txt[i], brks[i,2], nchar(txt[i])))
	})
}

is.ena = function(x) {
	if (is.expression(x)) {
		rep_len(FALSE, length(x))
	} else is.na(x)
}


nonempty_text = function(txt) {
	if (is.character(txt)) {
		nzchar(txt)
	} else rep_len(TRUE, length(txt))
}

number_text_lines = function(txt) {
	if (is.character(txt)) {
		length(strsplit(txt, "\n")[[1]])
	} else 1
}

expr_to_char = function(txt) {
	if (is.character(txt)) {
		txt
	} else {
		as.character(txt)
	}
}


encode_expr = function(txt) {
	if (is.character(txt)) {
		txt
	} else {
		paste0("__expr__", as.character(txt))
	}
}

decode_expr = function(txt) {
	if (substr(txt, 1, 8) == "__expr__") {
		txt = substr(txt, 9, nchar(txt))
		parse(text = txt)[[1]]
	} else {
		txt
	}
}


is_num_string = function(x) {
	suppressWarnings(!is.na(as.numeric(x)))
}

#' Map layer: lines
#'
#' Map layer that draws lines. Supported visual variables are: `col` (the color), `lwd` (line width), `lty` (line type), and `col_alpha` (color alpha transparency).
#'
#' @param col,col.scale,col.legend,col.chart,col.free  `r .doc_vv("col")` `r .doc_unit("col")`
#' @param lwd,lwd.scale,lwd.legend,lwd.chart,lwd.free  `r .doc_vv("lwd")` `r .doc_unit("lwd")`
#' @param lty,lty.scale,lty.legend,lty.chart,lty.free  `r .doc_vv("lty")` `r .doc_unit("lty")`
#' @param col_alpha,col_alpha.scale,col_alpha.legend,col_alpha.chart,col_alpha.free  `r .doc_vv("col_alpha")` `r .doc_unit("col_alpha")`
#' @param linejoin,lineend line join and line end. See [gpar()][grid::gpar()] for details.
#' @param plot.order Specification in which order the spatial features are drawn. See [tm_plot_order()] for details.
#' @param options options passed on to the corresponding `opt_<layer_function>` function
#' @inheritParams tm_polygons
#' @inherit tm_polygons details
#' @param ... to catch deprecated arguments from version < 4.0
#' @example ./examples/tm_lines.R
#' @seealso `r .doc_see_also_layers("lines")`
#' @export
tm_lines = function(col = tm_const(),
					col.scale = tm_scale(),
					col.legend = tm_legend(),
					col.chart = tm_chart_none(),
					col.free = NA,
					lwd = tm_const(),
					lwd.scale = tm_scale(),
					lwd.legend = tm_legend(),
					lwd.chart = tm_chart_none(),
					lwd.free = NA,
					lty = tm_const(),
					lty.scale = tm_scale(),
					lty.legend = tm_legend(),
					lty.chart = tm_chart_none(),
					lty.free = NA,
					col_alpha = tm_const(),
					col_alpha.scale = tm_scale(),
					col_alpha.legend = tm_legend(),
					col_alpha.chart = tm_chart_none(),
					col_alpha.free = NA,
					linejoin = "round",
					lineend = "round",
					plot.order = tm_plot_order("lwd", reverse = TRUE, na.order = "bottom"),
					zindex = NA,
					group = NA,
					group.control = "check",
					popup = tm_popup(),
					popup.vars = NA,
					popup.format = tm_label_format(),
					hover = NA,
					id = "",
					blend = "over",
					options = opt_tm_lines(),
					...) {

	args_called = names(rlang::call_match()[-1])
	args = list(...)

	layer_fun = if ("called_from" %in% names(args)) {
		args$called_from
	} else {
		"tm_lines"
	}

	if (any(v3_only("tm_lines") %in% names(args))) {

		v3_start_message()
		if (!("style" %in% names(args))) {
			if (!"breaks" %in% names(args)) {
				style = "pretty"
			} else {
				style = "fixed"
			}
		} else {
			style = args$style
		}

		v3_list_init()
		if (length(style) > 1) {
			style = style[1]
			.TMAP$v3_list$mult = TRUE
		}
		# v3 visual variable: col
		col.scale.args = c(list(n = v3_impute(args, "n", 5),
								style = style,
								style.args = v3_impute(args, "style.args", list())),
							if (style %in% c("cont", "log10")) {
								c({
									if (!is.null(args$breaks)) {
										if (length(args$breaks) > 2) {
											list(ticks = v3_impute(args, "breaks", NULL, "ticks"))
										} else {
											list(ticks = v3_impute(args, "breaks", NULL, "limits"))
										}
									} else {
										list()
									}
								},
								list(outliers.trunc = c(TRUE, FALSE)))
							} else {
								list(breaks = v3_impute(args, "breaks", NULL),
									 interval.closure = v3_impute(args, "interval.closure", "left"),
									 drop.levels = v3_impute(args, "drop.levels", FALSE))
							},
							list(midpoint = v3_impute(args, "midpoint", NULL),
								 as.count = v3_impute(args, "as.count", NA),
								 values = v3_impute(args, "palette", NA, "values"),
								 values.repeat = !v3_impute(args, "stretch.palette", TRUE, "values.repeat"),
								 values.range = v3_impute(args, "contrast", NA, "values.range"),
								 values.scale = 1,
								 value.na = v3_impute(args, "colorNA", NA, "value.na"),
								 value.null = v3_impute(args, "colorNULL", NA, "value.null"),
								 value.neutral = NA,
								 labels = v3_impute(args, "labels", NULL),
								 label.na = v3_impute(args, "textNA", "Missing", "label.na"),
								 label.null = NA,
								 label.format = v3_impute(args, "legend.format", list(), "label.format")))

		col.scale.args$fun_pref = if (style == "cat") {
			"categorical"
		} else if (style %in% c("fixed", "sd", "equal", "pretty", "quantile",
								"kmeans", "hclust", "bclust", "fisher", "jenks",
								"dpih", "headtails", "log10_pretty")) {
			"intervals"
		} else if (style == "cont") {
			"continuous"
		} else if (style == "log10") {
			"continuous_log"
		} else if (style == "order") {
			"rank"
		} else {
			stop("unknown style")
		}
		if (!(layer_fun == "qtm" && (!"col" %in% names(args)))) {
			if ("style" %in% names(args)) {
				v3_tm_scale_instead_of_style(style, scale_fun = col.scale.args$fun_pref, vv = "col", layer_fun = layer_fun, arg_list = v3_list_get())
			} else {
				v3_tm_scale(scale_fun = "", vv = "col", layer_fun = layer_fun, arg_list = v3_list_get())
			}
		}

		col.scale = do.call("tm_scale", args = col.scale.args)



		if ("alpha" %in% names(args)) {
			col_alpha = args$alpha
			v3_message_col_alpha(layer_fun = layer_fun, orig = "alpha")

		}


		v3_list_init()
		if ("legend.col.show" %in% names(args) && !args$legend.col.show) {
			v3_tm_legend_hide(layer_fun, arg = "legend.col.show", vv = "col")
			lwd.legend = tm_legend_hide()
		} else {
			col.legend.args = alist(title = v3_impute(args, "title.col", NA, "title"),
									na.show = v3_impute(args, "showNA", NA),
									format = v3_impute(args, "legend.format", list(), "format"),
									orientation = ifelse(v3_impute(args, "legend.col.is.portrait", TRUE, "orientation"), "portrait", "landscape"),
									reverse = v3_impute(args, "legend.reverse", FALSE, "reverse"))
			col.legend = do.call("tm_legend", col.legend.args)
			v3_tm_legend(fun = layer_fun, vv = "col", arg_list = v3_list_get())
		}

		# v3 visual variable: lwd
		# v3 visual variable: size
		v3_list_init()
		lwd.scale.args = list(ticks = v3_impute(args, "lwd.legend", NULL, "ticks"),
							  labels = v3_impute(args, "lwd.legend.labels", NULL, "labels"),
							  values.scale = v3_impute(args, "scale", 1, "values.scale"),
							  fun_pref = "continuous")
		if ("lwd" %in% args_called) v3_tm_scale(scale_fun = "continuous", vv = "lwd", layer_fun = layer_fun, arg_list = v3_list_get())
		lwd.scale = do.call("tm_scale", args = lwd.scale.args)

		v3_list_init()
		if ("legend.lwd.show" %in% names(args) && !args$legend.lwd.show) {
			v3_tm_legend_hide(layer_fun, arg = "legend.lwd.show", vv = "lwd")
			lwd.legend = tm_legend_hide()
		} else {
			lwd.legend.args = alist(title = v3_impute(args, "title.lwd", NA, "title"),
								   na.show = v3_impute(args, "showNA", NA),
								   format = v3_impute(args, "legend.format", list(), "format"),
								   orientation = ifelse(v3_impute(args, "legend.lwd.is.portrait", FALSE), "portrait", "landscape"),
								   reverse = v3_impute(args, "legend.lwd.reverse", FALSE))
			if ("lwd" %in% args_called) v3_tm_legend(fun = layer_fun, vv = "lwd", arg_list = v3_list_get())
			lwd.legend = do.call("tm_legend", lwd.legend.args)
		}


		if ("legend.hist" %in% names(args) && args$legend.hist) {
			col.chart = tm_chart_histogram()
			v3_tm_chart_hist(layer_fun = layer_fun, vv = "col", arg = "legend.hist")

			# to do: histogram title
		}
	}

	# unused arguments: typos?
	unused = setdiff(names(args), c(v3_only("tm_lines"), "called_from"))

	if (length(unused)) {
		message_layer_unused_args(layer_fun, unused)
	}

	# make sure required options are there
	options = complete_options(options, opt_tm_lines())

	# blend has been migrated to the layer function root; it is still passed on
	# internally via mapping.args (read as a$blend).
	options$mapping.args$blend = blend

	# resolve popup specification (popup.vars/popup.format are deprecated in
	# favour of popup = tm_popup(...))
	popup = process_popup(popup = popup,
						  popup.vars = popup.vars,
						  popup.format = popup.format,
						  popup.called = "popup" %in% args_called,
						  popup.vars.called = "popup.vars" %in% args_called,
						  popup.format.called = "popup.format" %in% args_called,
						  layer_fun = layer_fun)

	tm_element_list(tm_element(
		layer = "lines",
		trans.fun = tmapTransLines,
		trans.aes = list(),
		trans.args = options$trans.args,
		trans.apply_to = "this",
		mapping.aes = list(col = tmapScale(aes = "col",
										   value = col,
										   scale = col.scale,
										   legend = col.legend,
										   chart = col.chart,
										   free = col.free),
						   lwd = tmapScale(aes = "lwd",
						   				value = lwd,
						   				scale = lwd.scale,
						   				legend = lwd.legend,
						   				chart = lwd.chart,
						   				free = lwd.free),
						   lty = tmapScale(aes = "lty",
						   				value = lty,
						   				scale = lty.scale,
						   				legend = lty.legend,
						   				chart = lty.chart,
						   				free = lty.free),
						   col_alpha = tmapScale(aes = "col_alpha",
						   					  value = col_alpha,
						   					  scale = col_alpha.scale,
						   					  legend = col_alpha.legend,
						   					  chart = col_alpha.chart,
						   					  free = col_alpha.free)),

		gpar = tmapGpar(fill = NA,
						col = "__col",
						shape = NA,
						size = NA,
						fill_alpha = NA,
						col_alpha = "__col_alpha",
						pattern = NA,
						lty = "__lty",
						lwd = "__lwd",
						linejoin = linejoin,
						lineend = lineend),
		tpar = tmapTpar(),
		plot.order = plot.order,
		mapping.args = options$mapping.args,
		zindex = zindex,
		group = group,
		group.control = group.control,
		popup.vars = popup$vars,
		popup.format = popup$format,
		popup.title = popup$title,
		popup.layout = popup$layout,
		hover = hover,
		id = id,
		subclass = c("tm_aes_layer", "tm_layer")))
}

#' @rdname tm_lines
#' @param lines.only should only line geometries of the shape object (defined in [tm_shape()]) be plotted, or also other geometry types (like polygons)? By default `"ifany"`, which means `TRUE` in case a geometry collection is specified.
#' @param hitbox Controls whether an invisible interaction layer with a
#'   larger clickable area (\"hitbox\") is added on top of the lines.
#'
#'   This can improve click and popup behaviour for thin or densely packed
#'   lines by increasing the effective mouse interaction width.
#'
#'   Possible values:
#'   \describe{
#'     \item{"none"}{No additional hitbox layer is added. Lines are clickable
#'     only at their visible width.}
#'     \item{"plusX"}{Adds \code{X} pixels to the visible line width to compute
#'     the interaction width: \code{line_width + X}. For example, \code{"plus8"}
#'     widens the clickable area by 4 pixels on each side.}
#'     \item{"pmaxX"}{Ensures a minimum interaction width of \code{X} pixels:
#'     \code{pmax(line_width, X)}. For example, \code{"pmax8"} guarantees at
#'     least 8 pixels. Useful for very thin lines.}
#'     \item{"auto"}{\code{"pmax8"} if and only if interactive features are
#'     enabled (popup or hover), lines are thin (median line width < 4), and
#'     there are fewer than 10000 features. Otherwise \code{"none"}.}
#'   }
#'
#'   \code{plus} and \code{pmax} can be combined, e.g. \code{"plus4pmax8"} means
#'   \code{pmax(line_width + 4, 8)}.
#'
#'   Adding a hitbox improves usability for thin lines but may reduce
#'   performance for very large datasets, as an additional invisible
#'   layer is rendered.
#' @export
opt_tm_lines = function(lines.only = "ifany", hitbox = "auto") {
	list(trans.args = list(lines.only = lines.only),
		 mapping.args = list(hitbox = hitbox))
}

#' Set the position of map components
#'
#' Set the position of map components, such as legends, title, compass, scale bar, etc.
#' `tm_pos()` is the function to position these components:
#' `tm_pos_out()` places the components outside the map area, `tm_pos_in()` inside the map area, and `tm_pos_on_top()` on top of the map
#' Each `position` argument of a map layer or component should be specified with
#' one of these functions.
#' The functions `tm_pos_auto_out()` and `tm_pos_auto_in()`
#' are used to set the components automatically, and should be used
#' via [tmap_options()]. See Details how the positioning works.
#'
#' @param cell.h,cell.v The plotting area is overlaid with a 3x3 grid, of which
#'   the middle grid cell is the map area. Components can be drawn into any cell.
#'   `cell.h` specifies the horizontal position (column) can take values
#'   `"left"`, `"center"`, and `"right"`. `cell.v` specifies the vertical position
#'   (row) and can take values `"top"`, `"center"`, and `"bottom"`.
#'   See details for a graphical explanation.
#' @param pos.h,pos.v The position of the component within the cell.
#'   The main options for `pos.h` are `"left"`, `"center"`, and `"right"`.
#'   For `cell.v` these are `"top"`, `"center"`, and `"bottom"`.
#'   These options can also be provided in upper case; in that case there is no offset
#'   (see the tmap option `component.offset`). Also numbers between 0 and 1 can be provided,
#'   which determine the position of the component inside the cell (with (0,0) being left bottom).
#'   The arguments `just.h` and `just.v` determine the justification point.
#' @param align.h,align.v The alignment of the component in case multiple components are stacked.
#'   When they are stacked horizontally, `align.v` determines how components
#'   that are smaller in height than the available height (determined by the outer.margins
#'   if specified and otherwise by the highest component) are justified:
#'   `"top"`, `"center"`, or `"bottom"`. Similarly, `align.h` determines how components
#'   are justified horizontally when they are stacked vertically: `"left"`, `"center"`, or `"right"`.
#' @param just.h,just.v The justification of the components.
#'   Only used in case `pos.h` and `pos.v` are numbers.
#' @details
#' `tm_pos_in()` sets the position of the component(s) inside the maps area,
#' which is equivalent to the center-center cell (in case there are facets,
#' these are all drawn in this center-center cell).
#'
#' `tm_pos_out()` sets the position of the component(s) outside the map.
#'
#' `tm_pos_on_top()` is the same as `tm_pos_out`, but with the cell set to the center cell
#' It may be therefore seem similar to `tm_pos_in()`, but with an essential difference:
#' `tm_pos_in()` takes the map frame into account whereas `tm_pos_on_top()` does not.
#' #'
#' The amount of space that the top and bottom rows, and left and right columns
#' occupy is determined by the [tm_layout()] arguments `meta.margins` and `meta.auto_margins`.
#' The former sets the relative space of the bottom, left, top, and right side.
#' In case these are set to `NA`, the space is set automatically based on 1)
#' the maximum relative space specified by `meta.auto_margins` and 2)
#' the presence and size of components in each cell.
#' For instance, if there is one landscape oriented legend in the center-bottom cell,
#' then the relative space of the bottom row is set to the height of that legend
#' (given that it is smaller than the corresponding value of `meta.auto_margins`),
#' while the other four sides are set to 0.
#'
#' `tm_pos_auto_out()` is more complex: the `cell.h` and `cell.v` arguments should be set
#' to one of the four corners. It does not mean that the components are drawn in a corner.
#' The corner represents the sides of the map that the components are drawn.
#' By default, legends are drawn either at the bottom or on the right-side of
#' the map by default (see `tmap_options("legend.position")`).
#' Only when there are row- and column-wise legends and a general legend (using [tm_facets_grid()]),
#' the general legend is drawn in the corner, but in practice this case will be rare.
#'
#' The arguments `pos.h` and `pos.v` determine where the components are drawn within the cell.
#' Again, with `"left"`, `"center"`, and `"right"` for `pos.h`
#' and `"top"`, `"center"`, and `"bottom"` for `pos.v`.
#' The values can also be specified in upper-case, which influences the offset
#' with the cell borders, which is determined by tmap option `component.offset`.
#' By default, there is a small offset when components are drawn inside and no
#' offset when they are drawn outside or with upper-case.
#'
#' `tm_pos_auto_in()` automatically determines `pos.h` and `pos.v` given the
#' available space inside the map. This is similar to the default positioning in tmap3.
#'
#' In case multiple components are draw in the same cell and the same position
#' inside that cell, they are stacked (determined which the `stack` argument in
#' the legend or component function). The `align.h` and `align.v` arguments
#' determine how these components will be justified with each other.
#'
#' Note that legends and components may be different for a facet row or column.
#' This is the case when [tm_facets_grid()] or [tm_facets_stack()] are applied
#' and when scales are set to free (with the `.free` argument of the map layer functions).
#' In case a legends or components are draw row- or column wise, and the position
#' of the legends (or components) is right next to the maps, these legends
#' (or components) will be aligned with the maps.
#'
#' @seealso `r .doc_see_also_comp()`
#' @export
tm_pos = function(cell.h, cell.v, pos.h, pos.v, align.h, align.v, just.h, just.v) {
	args = lapply(as.list(rlang::call_match()[-1]), eval, envir = parent.frame())
	args$called = names(args)
	args$type = "out"
	structure(args, class = "tm_pos")
}

#' @export
#' @rdname tm_pos
tm_pos_in = function(pos.h, pos.v, align.h, align.v, just.h, just.v) {
	args = lapply(as.list(rlang::call_match()[-1]), eval, envir = parent.frame())
	args$called = names(args)
	args$type = "in"
	structure(args, class = "tm_pos")
}


#' @export
#' @rdname tm_pos
tm_pos_out = function(cell.h, cell.v, pos.h, pos.v, align.h, align.v, just.h, just.v) {
	args = lapply(as.list(rlang::call_match()[-1]), eval, envir = parent.frame())
	args$called = names(args)
	args$type = "out"
	structure(args, class = "tm_pos")
}

#' @export
#' @rdname tm_pos
tm_pos_on_top = function(pos.h, pos.v, align.h, align.v, just.h, just.v) {
	args = lapply(as.list(rlang::call_match()[-1]), eval, envir = parent.frame())
	args$called = names(args)
	args$cell.h = "center"
	args$cell.v = "center"
	args$type = "out"
	structure(args, class = "tm_pos")
}

#' @export
#' @rdname tm_pos
tm_pos_auto_out = function(cell.h, cell.v, pos.h, pos.v, align.h, align.v, just.h, just.v) {
	args = lapply(as.list(rlang::call_match()[-1]), eval, envir = parent.frame())
	args$called = names(args)
	args$type = "autoout"
	structure(args, class = "tm_pos")
}

#' @export
#' @rdname tm_pos
tm_pos_auto_in = function(align.h, align.v, just.h, just.v) {
	args = lapply(as.list(rlang::call_match()[-1]), eval, envir = parent.frame())
	args$called = names(args)
	args$type = "autoin"
	structure(args, class = "tm_pos")
}

pane_name = function(id) {
	paste0("tmap", sprintf("%03g", id))
}

process_components = function(cdt, o) {
	gs = tmap_graphics_name()


	# update added legends
	is_added_leg = vapply(cdt$comp, inherits, logical(1), "tm_add_legend")

	if (any(is_added_leg)) {
		cdt$comp[is_added_leg] = lapply(cdt$comp[is_added_leg], tmapAddedLegend, o = o)
	}

	# the option component.position is not a general fall-back option (in case the component is undefined for a mode),
	# but used for imputation of not-specified tm_pos arguments
	# in case there is no specific .position argument for the current mode specified, use the 'in' position for the time being
	# the missing methods for the component is catched later via the missing tmap<gs>CompPrepare method
	cdt$comp = lapply(cdt$comp, function(l) {
		if ("in" %in% names(l$position)) l$position = l$position[["in"]]
		l
	})

	cdt$class = sapply(cdt$comp, function(l) l$position$type)
	cdt$cell.h = sapply(cdt$comp, function(l) {x = l$position$cell.h; if (is.null(x)) NA_character_ else x})
	cdt$cell.v = sapply(cdt$comp, function(l) {x = l$position$cell.v; if (is.null(x)) NA_character_ else x})
	cdt$pos.h = sapply(cdt$comp, function(l) {x = l$position$pos.h; if (is.null(x)) NA_character_ else x})
	cdt$pos.v = sapply(cdt$comp, function(l) {x = l$position$pos.v; if (is.null(x)) NA_character_ else x})
	cdt$z = sapply(cdt$comp, function(l) {x = l$z; if (is.null(x)) NA_integer_ else x})
	cdt[, zauto := .I]

	if (gs != "Grid") {
		cdt[class == "out", class := "in"]
		cdt[class == "autoout", class := "in"]
	}

	funP = paste0("tmap", gs, "CompPrepare")
	funW = paste0("tmap", gs, "CompWidth")
	funH = paste0("tmap", gs, "CompHeight")

	cdt$comp = lapply(cdt$comp, function(comp) do.call(funP, list(comp = comp, o = o)))

	# split components into facets (if needed)
	is_multi = sapply(cdt$comp, inherits, what = "tm_multi_comp")

	if (any(is_multi)) {
		cdt_single = cdt[!is_multi, ]
		cdt_multi = cdt[is_multi, ]

		allby_temp = as.data.table(do.call(expand.grid, lapply(o$fn, seq_len)))
		names(allby_temp) = paste0("by", 1:3, "__")

		cdt_rows = data.table::rbindlist(lapply(seq_len(nrow(cdt_multi)), function(i) {
			if (cdt_multi$class[i] %in% c("out", "autoout")) {
				if (length(cdt_multi$comp[[i]]) > 1) {
					warning("Component is specified for multiple facets (e.g. title text), but since placement is outside, only the first is used", call. = FALSE)				}
				cdt_multi$comp[[i]] = cdt_multi$comp[[i]][[1]]
				cdt_multi[i, ]
			} else {
				cbind(allby_temp,
					  comp = rep(cdt_multi$comp[[i]], length.out = nrow(allby_temp)),
					  cdt_multi[i, setdiff(names(cdt_multi), c(names(allby_temp), "comp")), with = FALSE])
			}
		}))
		cdt = rbind(cdt_single, cdt_rows)
	}


	# filter components that are not shown (e.g. not implemented), which is determined in the Prepare step
	cdt$show = sapply(cdt$comp, function(l) l$show)
	cdt = cdt[cdt$show,][,show:=NULL]


	cdt$comp = lapply(cdt$comp, function(comp) do.call(funW, list(comp = comp, o = o)))
	cdt$comp = lapply(cdt$comp, function(comp) do.call(funH, list(comp = comp, o = o)))

	cdt[, ':='(facet_row = character(), facet_col = character())]
	cdt$stack_auto = vapply(cdt$comp, function(l) {
		s = l$stack
		length(s) > 1
	}, FUN.VALUE = logical(1))
	cdt$stack = vapply(cdt$comp, function(l) {
		s = l$stack
		if (length(s) > 1 && "manual" %in% names(s)) s["manual"] else s[1]
	}, FUN.VALUE = character(1))


	getLW = function(x) sapply(x, function(y) {
		yW = y$Win
		yW %||% 0
	})
	getLH = function(x) sapply(x, function(y) {
		yH = y$Hin
		yH %||% 0
	})
	# attempt to determine margins
	cdt[, legW := getLW(comp)]
	cdt[, legH := getLH(comp)]

	if (any(is.na(cdt$z))) {
		cdt[is.na(z), z := zauto]
	}
	cdt[, zauto:=NULL]
	if (nrow(cdt)>0L) {
		data.table::setorder(cdt, "z")
	}


	cdt
}

process_components2 = function(cdt, o) {
	gs = tmap_graphics_name()




	if (o$type != "grid" && o$n > 1) {
		#if (o$nrows == 1 && o$ncols == 1)
		if (o$nrows == 1) {
			# -use by2 and not by1 when they form a row
			cdt[, by2__ := by1__]
			cdt[, by1__ := NA]
		}
	}

	stacks = o$legend.stack


	# place legends inside if needed
	#if (o$ncols > 1 && o$nrows > 1) {
	if (o$type == "wrap") {
		# all free legends inside
		if (any((!is.na(cdt$by1) | !is.na(cdt$by2)) & cdt$class == "out")) cli::cli_inform("{.field {.fun tm_pos_out}} Legends per facets are placed outside. To align them with the facets, consider {.fun tm_facets_stack} or {.fun tmap_arrange}.")
		cdt[(!is.na(by1__) | !is.na(by2__)) & class == "autoout", ':='(class = "in")]
	} else if (o$type == "grid") {
		# all free-per-facet legends inside
		cdt[!is.na(by1__) & !is.na(by2__) & class == "autoout", ':='(class = "in")]
	}



	# update auto position (for 'all', 'rows', 'columns' legends)
	cdt[is.na(by1__) & is.na(by2__) & class == "autoout", ':='(cell.h = o$legend.position.all$cell.h, cell.v = o$legend.position.all$cell.v)]
	cdt[!is.na(by1__) & is.na(by2__) & class == "autoout", ':='(cell.h = o$legend.position.sides$cell.h, cell.v = "by")]
	cdt[is.na(by1__) & !is.na(by2__) & class == "autoout", ':='(cell.h = "by", cell.v = o$legend.position.sides$cell.v)]

	# in case tm_pos is called to update the 'non-by' cell
	cdt$calledCV = vapply(cdt$comp, function(comp) "cell.v" %in% comp$position$called, FUN.VALUE = logical(1))
	cdt$calledCH = vapply(cdt$comp, function(comp) "cell.h" %in% comp$position$called, FUN.VALUE = logical(1))
	cdt[!is.na(by1__) & is.na(by2__) & !calledCV, ':='(cell.v = "by")]
	cdt[is.na(by1__) & !is.na(by2__) & !calledCH, ':='(cell.h = "by")]

	cdt[!is.na(by1__) & is.na(by2__) & !calledCV, ':='(stack = ifelse(stack_auto, stacks["per_row"], stack))]
	cdt[is.na(by1__) & !is.na(by2__) & !calledCH, ':='(stack = ifelse(stack_auto, stacks["per_col"], stack))]

	cdt$calledCV = NULL
	cdt$calledCH = NULL

	cdt[is.na(by1__) & is.na(by2__) & class == "autoout", ':='(stack = ifelse(stack_auto, ifelse(cell.h == "center", stacks["all_col"], ifelse(cell.v == "center", stacks["all_row"], stacks["all"])), stack))]
	cdt[!is.na(by1__) & is.na(by2__) & class == "autoout", ':='(stack = ifelse(stack_auto, stacks["per_row"], stack))]
	cdt[is.na(by1__) & !is.na(by2__) & class == "autoout", ':='(stack = ifelse(stack_auto, stacks["per_col"], stack))]


	cdt[class == "autoout", class := "out"]


	lai = which(cdt$class == "autoin")
	if (length(lai)) {
		cdt[class == "autoin", ":="(pos.h = ifelse(is.na(pos.h), o$legend.autoin.pos[1], pos.h),
									pos.v = ifelse(is.na(pos.v), o$legend.autoin.pos[2], pos.v),
									class = "in")]
	}

	vby = any(cdt$cell.v == "by" & !is.na(cdt$cell.v))
	hby = any(cdt$cell.h == "by" & !is.na(cdt$cell.h))

	toC = function(x) {
		paste(x, collapse = "_")
	}


	# # manual outside legends -2 is top or left, -1 is bottom or right
	# cdt[class %in% c("autoout", "out"), ':='(facet_row =
	# 		ifelse(cell.v == "center", ifelse(vby, "1", toC(1:o$nrows)),
	# 		ifelse(cell.v == "by", as.character(by1__),
	# 		ifelse(cell.v == "top", as.character(-2), as.character(-1)))),
	# 	facet_col =
	# 		ifelse(cell.h == "center", ifelse(hby, "1", toC(1:o$ncols)),
	# 		ifelse(cell.h == "by", as.character(by2__),
	# 		ifelse(cell.h == "left", as.character(-2), as.character(-1)))))]
	# manual outside legends -2 is top or left, -1 is bottom or right
	cdt[class %in% c("autoout", "out"), ':='(facet_row =
											 	ifelse(cell.v == "center", toC(1:o$nrows),
											 		   ifelse(cell.v == "by", as.character(by1__),
											 		   	   ifelse(cell.v == "top", as.character(-2), as.character(-1)))),
											 facet_col =
											 	ifelse(cell.h == "center", toC(1:o$ncols),
											 		   ifelse(cell.h == "by", as.character(by2__),
											 		   	   ifelse(cell.h == "left", as.character(-2), as.character(-1)))))]


	cdt
}

step4_plot = function(tm, vp, return.asp, show, in.shiny, knit, knit_opts, args) {
	tmx = tm$tmo
	o = tm$o
	aux = tm$aux
	cmp = tm$cmp
	prx = tm$prx

	# check if mode is the same as in step1
	# tmap_mode() may be executed in between (#1082)
	gs = tmap_graphics_name()

	if (o$name != gs) cli::cli_abort("tmap mode changed during execution; did you run {.code tmap_mode()} inside a shiny app?")

	bbo = o$bbox
	bbo_args = o$bbox_args %||% list()

	# shortcut if no data layers are used, but only a tm_shape
	if (length(tmx)) {
		bbx_def = vapply(o$main, function(mid) {
			"bbx" %in% names(tmx[[mid]])
		}, FUN.VALUE = logical(1))
		if (any(bbx_def)) {
			bbxs = lapply(o$main[which(bbx_def)], FUN = function(mid) {
				tmx[[mid]]$bbx
			})
			bbm = stm_merge_bbox(bbxs)
		} else {
			bbm = NULL
		}
	} else {
		bbm = NULL
	}

	# remove empty data layers
	any_groups = (length(tmx) > 0L)

	tmx = lapply(tmx, function(tmxi) {
		tmxi$layers = lapply(tmxi$layers, function(tml) {
			empt = vapply(tml$shpDT$shpTM, function(sdt) {
				length(sdt$shp) == 0L
			}, FUN.VALUE = logical(1))
			if (all(empt)) NULL else tml
		})
		empt = vapply(tmxi$layers, is.null, logical(1))

		if (all(empt)) {
			NULL
		} else {
			tmxi$layers = tmxi$layers[!empt]
			tmxi
		}
	})
	empt = vapply(tmx, is.null, logical(1))
	tmx = tmx[!empt]


	any_data_layer = (length(tmx) > 0L)

	# split tm in case of as.layers in tm_facets
	# TODO

	# function to subset data
	get_dt = function(dt, by1, by2, by3, remove_by = FALSE) {
		b = list(by1, by2, by3)
		bynames = intersect(names(dt), paste0("by", 1:3, "__"))
		byids = as.integer(substr(bynames, 3, 3))

		sel = rep(TRUE, nrow(dt))
		if (length(bynames)) {
			for (i in 1:length(bynames)) {
				sel = sel & (is.na(dt[[bynames[i]]]) | dt[[bynames[i]]] %in% b[[byids[i]]])
			}
		}
		if (remove_by) {
			nms = setdiff(names(dt), bynames)
			dt[which(sel), nms, with = FALSE]
		} else {
			dt[which(sel),]
		}
	}

	if (o$as.layers) {
		bys = do.call(expand.grid, lapply(o$fn, seq, from = 1))

		bys_labs = lapply(o$fl, function(x) if (is.null(x)) "" else x)
		labs = do.call(mapply, c(list(FUN = paste0, USE.NAMES = FALSE), mapply("[", bys_labs, bys, SIMPLIFY = FALSE, USE.NAMES = FALSE)))
		# to do: take labels from o$fl
		# assign to tml$group
		# disable multiple basemaps

		e = environment()
		any_radio = FALSE
		tmx = lapply(tmx, function(tmxi) {
			lys = mapply(function(tml, tml_name) {
				lys2 = lapply(1:nrow(bys), function(i) {
					tml$mapping_dt = get_dt(tml$mapping_dt, bys$Var1[i], bys$Var2[i], bys$Var3[i], remove_by = TRUE)
					tml$mapping_legend = lapply(tml$mapping_legend, function(ml) {
						ml = get_dt(ml, bys$Var1[i], bys$Var2[i], bys$Var3[i], remove_by = TRUE)
						fixed = (nrow(ml) == 1L && tml$group.control == "check")
						for (j in 1:nrow(ml)) {
							id_old = ml$legnr[j]
							leg_old = .TMAP$legs[[ml$legnr[j]]]
							leg_old$layerId = paste0("fixed", ml$legnr[j])
							leg_old$group = labs[i]
							ml$legnr[j] = legend_save(leg_old)
						}
						ml
					})
					tml$group = labs[i]
					if (tml$group.control == "radio") assign("any_radio", value = TRUE, envir = e)
					tml$lid = tml$lid + i
					tml
				})
				names(lys2) = paste0(tml_name, "_", sprintf("%02d", 1:length(lys2)))
				lys2
			}, tmxi$layers, names(tmxi$layers), SIMPLIFY = FALSE, USE.NAMES = FALSE)
			tmxi$layers = do.call(c, lys)
			tmxi
		})

		if (length(aux) && any_radio) {
			multi_tile = any(vapply(aux, function(ai) {
				(ai$mapping.fun == "Tiles" && length(ai$args$server) > 1)
			}, FUN.VALUE = logical(1L), USE.NAMES = FALSE))
			if (multi_tile) cli::cli_inform("When {.code as.layers = TRUE} with radio buttons, it may be easier to have only one basemap, e.g. {.code + tm_basemap({.str OpenStreetMap})}")
		}

		o$n = 1
		o$fl = list(NULL,NULL,NULL)
		o$fn = rep(1L, 3)
		o$ncols = 1
		o$nrows = 1

	}



	# get name of graphics engine (for function names e.g. tmapGridInit)
	gs = tmap_graphics_name()

	o = prepreprocess_meta(o, vp)

	# get legends from layer data and put them in "components data.table" (cdt)

	if (o$legend.only) {
		cmp_sel = vapply(cmp, function(cp) {
			inherits(cp, c("tm_legend", "tm_add_legend"))
		}, FUN.VALUE = logical(1))
		cdt_cmp = if (any(cmp_sel)) {
			data.table::rbindlist(lapply(cmp[cmp_sel], function(cp) {
				data.table::data.table(by1__ = NA_integer_,
									   by2__ = NA_integer_,
									   by3__ = NA_integer_,
									   comp = list(cp))
			}))
		} else {
			data.table::data.table(by1__ = integer(0),
								   by2__ = integer(0),
								   by3__ = integer(0),
								   comp = list())
		}
	} else {
		cdt_cmp = if (length(cmp)) {
			data.table::rbindlist(lapply(cmp, function(cp) {
				data.table::data.table(by1__ = NA_integer_,
									   by2__ = NA_integer_,
									   by3__ = NA_integer_,
									   comp = list(cp))
			}))
		} else {
			data.table::data.table(by1__ = integer(0),
								   by2__ = integer(0),
								   by3__ = integer(0),
								   comp = list())
		}
	}

	cdt = if (any_data_layer) {
		cdt_legs = step4_plot_collect_legends(tmx)
		cdt_crts = step4_plot_collect_charts(tmx)
		data.table::rbindlist(list(cdt_legs, cdt_crts, cdt_cmp))
	} else {
		cdt_cmp
	}


	## place components top left
	if (o$legend.only) {
		cdt$comp = lapply(cdt$comp, function(cc) {
			cc$position = complete_options(tm_pos_in("left", "top"), o$legend.position)
			cc
		})
	}


	if (gs == "Grid" && is.null(vp) && show) {
		is_inset_tmap = sapply(cdt$comp, inherits, what = "tm_inset_tmap")
		if (any(is_inset_tmap)) {
			grid::grid.newpage()
			newpage = FALSE
		} else {
			newpage = TRUE
		}
	} else {
		newpage = TRUE
	}

	if (nrow(cdt)) cdt = process_components(cdt, o)

	# determine panel type, inner margins, and automatic comp placement
	if (!o$legend.only) {
		o = preprocess_meta(o, cdt)


		crs = get_option_class(o$crs_step4, "sf") #o$crs
		longlat = sf::st_is_longlat(crs)
	} else {
		o$npages = 1L
	}



	# function to get shape object
	get_shpTM = function(shpDT, by1, by2, by3) {
		b = list(by1, by2, by3)
		bynames = intersect(names(shpDT), paste0("by", 1:3, "__"))
		byids = as.integer(substr(bynames, 3, 3))

		sel = rep(TRUE, nrow(shpDT))
		if (length(bynames)) {
			for (i in 1L:length(bynames)) {
				sel = sel & shpDT[[bynames[i]]] %in% b[[byids[i]]]
			}
		}
		shpDT$shpTM[which(sel)]
	}


	# function to get bbox per facet, also take into account bbm (for groups without data-layers)
	get_bbox = function(by1, by2, by3) {
		bbxs = lapply(tmain, function(tmi) {
			shpTM = get_shpTM(tmi$shpDT, by1, by2, by3)
			mdt = get_dt(tmi$mapping_dt, by1, by2, by3)
			bbxs2 = lapply(shpTM, stm_bbox, tmapID = mdt$tmapID__, crs = crs)
			if (!is.null(bbm)) bbxs2 = c(list(bbm), bbxs2)
			bbx = stm_merge_bbox(bbxs2)
			if (is.na(bbx)) bbx else tmaptools::bb(bbx, asp.limit = 10)
		})
		bbm = stm_merge_bbox(bbxs)
		bbc = bbm

		bbe = bb_ext(bbc, o$inner.margins)

		bbe = crop_lat(bbe, crs, o$limit_latitude_3857)



		list(list(bbe))
	}

	# main group (that determines bounding box)
	# TODO take into account multiple main groups (see step1_rearrange and get_main_ids)


	if (o$legend.only) {
		d = NULL
	} else if (any_data_layer) {
		# create table with meta data for the facets (row, col id, bbox, asp)
		d = data.table::data.table(do.call(expand.grid, lapply(structure(o$nby, names = c("by1", "by2", "by3")), seq_len)))
		d[, i := seq_len(nrow(d))]
		grps = c("by1", "by2", "by3")[o$free.coords]

		grp_ids = as.integer(substr(names(tmx), 6, nchar(names(tmx))))
		mains_in_grp = intersect(o$main[!bbx_def], grp_ids)
		if (length(mains_in_grp) && !("inset" %in% names(o)) && !o$earth_bbox && is.null(bbo)) {
			lookup = match(mains_in_grp, grp_ids)
			tmain = unlist(unlist(tmx[lookup], recursive = FALSE, use.names = FALSE), recursive = FALSE, use.names = FALSE)
			d[, bbox:=do.call(get_bbox, as.list(.SD)), by = grps, .SDcols = c("by1", "by2", "by3")]
		} else {
			if (!is.null(bbo) || is.null(bbm)) {
				if (!is.null(bbo) && identical(bbo, "FULL") || (is.null(bbo) && is.null(bbm))) {
					bbm = full_bbox(crs)
					if (length(bbo_args) >= 0) bbm = do.call(tmaptools::bb, c(list(x = bbm), bbo_args))
				} else {
					if (inherits(bbo, "SpatExtent")) bbo = sf::st_bbox(bbo)
					bbm = do.call(tmaptools::bb, c(list(x = bbo, projection = crs), bbo_args))
				}
				if (o$earth_bbox) {
					bbm = bb_ext(bbm, o$inner.margins)
				}
			} else {
				bbm = sf::st_transform(bbm, crs = crs)
			}
			d[, bbox:=rep(list(bbm),nrow(d))]
		}
	} else {
		if (is.null(bbm)) {
			if (!is.null(bbo) && !identical(bbo, "FULL")) {
				bbm = do.call(tmaptools::bb, c(list(x = bbo, projection = crs), bbo_args))
			} else {
				bbm = bbm = do.call(tmaptools::bb, c(list(x = full_bbox(crs)), bbo_args))
			}
		} else {
			bbm = sf::st_transform(bbm, crs = crs)
		}

		bbm = crop_lat(bbm, crs, o$limit_latitude_3857)



		d = data.table::data.table(by1 = 1L, by2 = 1L, by3 = 1L, i = 1, bbox = list(bbm))
	}

	if (!o$legend.only) {
		d[, asp:=get_asp(bbox)]
		d = d[!is.na(asp)]


		if (o$type != "grid" && is.na(o$nrows) && is.na(o$ncols)) {
			# limit facets
			o$n_min_ani = if (o$animate) o$n / o$nframes else o$n
			n_lim = limit_nx(o$n_min_ani, o$facet.max)
			if (n_lim != o$n_min_ani) {
				fn_lim = pmin(o$fn, n_lim)
				while(prod(fn_lim) > n_lim) {
					fn_lim[which.max(fn_lim)] = fn_lim[which.max(fn_lim)] - 1L
				}
				d = d[by1<= fn_lim[1] & by2<= fn_lim[2] & by3 <= fn_lim[3]]
				o$fl = mapply(function(a, b) a[1:b], o$fl, fn_lim, SIMPLIFY = FALSE)
				o$fn = fn_lim
				o$n = n_lim
			}
		}

		# not sure what this does:
		d[, bbox:=lapply(bbox, FUN = function(bbx) {
			if (!is.na(bbx) && !is.na(longlat) && longlat && !sf::st_is_longlat(bbx)) {
				sf::st_bbox(sf::st_transform(tmaptools::bb_poly(bbx), crs = 4326))
			} else {
				bbx
			}
		})]

		d[, units:=lapply(bbox, FUN = function(bbx) {
			if (is.na(bbx)) {
				list()
			} else {
				#browser()
				# if (!is.na(bbx) && !is.na(longlat) && longlat) {
				# 	latitude <- mean.default(bbx[c(2,4)])
				# 	bbxll <- c(xmin=0, ymin=latitude, xmax=1, ymax=latitude)
				# 	ad <- suppressWarnings({tmaptools::approx_distances(bbxll, projection=crs)})
				# 	to <- as.numeric(units::set_units(ad$hdist, units::as_units(unit), mode = "standard"))
				# } else {
				# 	ad <- suppressWarnings({tmaptools::approx_distances(bbx, projection=crs)})
				#
				# 	if (is.na(crs)) {
				# 		to <- ad$hdist
				# 	} else {
				# 		to <- as.numeric(units::set_units(units::set_units(1, attr(ad$hdist, "units")$numerator, mode = "standard"), units::as_units(unit), mode = "standard"))
				# 	}
				# }

				dst = distances_bbox_sides(bbx)

				if (o$unit %in% c("metric", "imperial")) {
					# automatic

					uns = if (o$unit == "metric") c("km", "m", "mm") else c("mi", "yd", "ft", "in")
					res = vapply(uns, function(u) {
						as.numeric(units::set_units(dst[2], units::as_units(u), mode = "standard"))
					}, FUN.VALUE = numeric(1))

					if (is.na(res[2])) {
						unit = uns[1]
					} else {
						res_num = as.numeric(res)
						if (any(res_num >= 10)) {
							unit = names(res)[res_num >= 10][1]
						} else {
							unit = tail(names(res), 1)
						}
					}
				} else {
					unit = o$unit
				}

				units::set_units(dst, units::as_units(unit), mode = "standard")
				list(projection=crs, unit=unit, dst = dst, projected = !longlat)
			}
		})]

		# determine automatic position of inside comp
		if (!any(o$free.coords) && any(cdt$class == "autoin")) {
			shp = tmain[[1]]$shpDT$shpTM[[1]]$shp
			# TODO take into account multiple main shapes
			# TODO take use areas instead of coordinates for polygons
			if (inherits(shp, c("sf", "sfc"))) {
				bbx = d$bbox[[1]]
				co = sf::st_coordinates(sf::st_centroid(shp))

				xn = (co[,1]-bbx[1])/(bbx[3]-bbx[1])
				yn = (co[,2]-bbx[2])/(bbx[4]-bbx[2])
				cornerID = which.max(c(
					bl = min(sqrt((xn^2) + (yn^2)), na.rm = TRUE),
					tl = min(sqrt((xn^2) + ((1-yn)^2)), na.rm = TRUE),
					tr = min(sqrt(((1-xn)^2) + ((1-yn)^2)), na.rm = TRUE),
					br = min(sqrt(((xn-1)^2) + (yn^2)), na.rm = TRUE)))

				o$legend.autoin.pos = switch(names(cornerID), tl = c("left", "top"), tr = c("right", "top"), bl = c("left", "bottom"), br = c("right", "bottom"))
			} else {
				o$legend.autoin.pos = c("left", "top")
			}
		} else {
			o$legend.autoin.pos = c("left", "top")
		}
	}


	# calculate margins, number of rows and colums, etc.
	o = process_meta(o, d, cdt, aux)

	if (!o$legend.only) {
		# workaround to move panels to titles in view mode
		# TO DO: make this generic (e.g. component prep function?)
		if (gs != "Grid") {
			if (o$panel.type != "none") {
				cdt = rbindlist(c({if (nrow(cdt) == 0L) NULL else list(cdt)}, mapply(function(lab, i) {
					data.table::data.table(by1__ = i,
										   by2__ = NA_integer_,
										   by3__ = NA_integer_,
										   comp = list(impute_comp(tm_title(lab)[[1]], o)),
										   class=  "in",
										   cell.h = NA,
										   cell.v = NA,
										   pos.h = "left",
										   pos.v = "top",
										   z = 1,
										   facet_row = NA_character_,
										   facet_col = NA_character_,
										   stack_auto = TRUE,
										   stack = "vertical",
										   legW = 0,
										   legH = 0)
				}, o$panel.labels[[1]], seq_len(o$fn[1]), SIMPLIFY = FALSE)))
			}
			o$panel.type = "none"
		}

		o$ng = length(tmx)

		# determine row and col ids

		if (o$type == "grid") {
			d[, row := as.integer((i - 1) %% o$nrows + 1)]
			d[, col := as.integer((((i - 1) %/% o$nrows + 1) - 1) %% o$ncols + 1)]
			d[, page := as.integer(i - 1) %/% (o$nrows * o$ncols) + 1]
		} else {

			# wrap
			if (o$nby[3] == 1L) {
				if (!o$byrow) {
					d[, row := as.integer((i - 1) %% o$nrows + 1)]
					d[, col := as.integer((((i - 1) %/% o$nrows + 1) - 1) %% o$ncols + 1)]
				} else {
					d[, col := as.integer((i - 1) %% o$ncols + 1)]
					d[, row := as.integer((((i - 1) %/% o$ncols + 1) - 1) %% o$nrows + 1)]
				}
				d[, page := as.integer(i - 1) %/% (o$nrows * o$ncols) + 1]
			} else {
				d[, j:= rep(1L:o$npp, length.out = nrow(d))]
				if (!o$byrow) {
					d[, row := as.integer((j - 1) %% o$nrows + 1)]
					d[, col := as.integer((((j - 1) %/% o$nrows + 1) - 1) %% o$ncols + 1)]
				} else {
					d[, col := as.integer((j - 1) %% o$ncols + 1)]
					d[, row := as.integer((((j - 1) %/% o$ncols + 1) - 1) %% o$nrows + 1)]
				}
				d[, page := by3]
				d[, j:= NULL]
			}
		}

	}



	#####o$legend.autoin.pos = c("left", "top")

	# prepare function names
	FUNinit = paste0("tmap", gs, "Init")
	FUNaux = paste0("tmap", gs, "Aux")
	FUNrun = paste0("tmap", gs, "Run")
	FUNshape = paste0("tmap", gs, "Shape")
	FUNoverlay = paste0("tmap", gs, "Overlay")
	FUNwrap = paste0("tmap", gs, "Wrap")
	FUNxtab = paste0("tmap", gs, "Xtab")

	FUNxlab = paste0("tmap", gs, "Xlab")
	FUNylab = paste0("tmap", gs, "Ylab")


	FUNgridxlab = paste0("tmap", gs, "GridXLab")
	FUNgridylab = paste0("tmap", gs, "GridYLab")

	# for inset maps
	inset_frame_ids = if (nrow(cdt) == 0L) integer(0) else which(sapply(cdt$comp, function(cmp) "bbox" %in% names(cmp)))

	if (!o$legend.only) {
		# create table with bounding boxes (the only important property, apart from settings)
		db = data.table(bbox = unique(d$bbox[!is.na(d$asp)]))
		db[, i:=1L:nrow(db)]
		d[, bi:=db$i[match(d$bbox, db$bbox)]]

		## process components
		if (nrow(cdt)) cdt = process_components2(cdt, o)

		# init
		res = do.call(FUNinit, c(list(o = o, show = show, newpage = newpage, return.asp = return.asp, vp = vp, prx = prx), args))
		if (return.asp) {
			return(res)
		} else {
			.TMAP$geo_ref = get_geo_ref(bbx = db$bbox[[1]],
										crs = crs,
										inner_margins = o$inner.margins,
										dev_size = res$dev[1:2], map_size = res$map[1:2], offset = res$margins)
			.TMAP$animation = list(fps = o$fps, play = o$play)
		}

		## prepare aux layers
		if (length(aux)) {
			# prepare aux layers and return group label (in case it is not user specified)
			aux_group_def = mapply(function(a, id) {
				FUNaux_prep = paste0("tmap", gs, "AuxPrepare")
				a_args = structure(a$args, class = c(a$mapping.fun, "list"))

				do.call(FUNaux_prep, list(a = a_args, bs = db$bbox, id = id, o = o))
			}, aux, 1:length(aux))
			aux_group = mapply(function(a, agd) {
				ng = lengths(gregexpr("__", agd, fixed = TRUE))
				if (is.na(a$group)) agd else as.character(a$group)
			}, aux, aux_group_def, USE.NAMES = FALSE)

			aux_group.control = vapply(aux, function(a) {
				a$group.control
			}, FUN.VALUE = character(1))

			# find lid (layer plot id values) for aux layers
			aux_lid = vapply(aux, function(a) a$lid, FUN.VALUE = numeric(1))
		}

		if (!any_data_layer && !length(aux)) {
			message_nothing_to_show(any_groups)
			return(invisible(NULL))
		}



		# data frame for layer ids
		q = do.call(rbind, c(
			{if (any_data_layer) {
				lapply(1L:o$ng, function(ig) {
					tmxi = tmx[[ig]]
					nl = length(tmxi$layers)
					lid = vapply(tmxi$layers, function(l) {l$lid}, FUN.VALUE = numeric(1))
					group = vapply(tmxi$layers, function(l) {l$group}, FUN.VALUE = character(1))
					group.control = vapply(tmxi$layers, function(l) {l$group.control}, FUN.VALUE = character(1)) # used to determine control layer group (view mode)
					data.frame(gid = rep(ig, nl), g_lid = 1:nl, lid = unname(lid), group = unname(group), group.control = unname(group.control), lid2 = rep(0, nl), pane = rep("", nl), new = rep(TRUE, nl), group.zoom_levels = I(rep(list(NA),nl)))
				})
			} else {
				NULL
			}},
			{if (length(aux)) list(data.frame(gid = 0, g_lid = 1L:length(aux), lid = unname(aux_lid), group = unname(aux_group), group.control = unname(aux_group.control), lid2 = 0, pane = "", new = TRUE, group.zoom_levels = I(rep(list(NA),length(aux))))) else NULL}))

		#q$lid[q$lid != 0] = q$lid[q$lid != 0] + 400L

		if (.TMAP$proxy) {
			q0 = .TMAP$q
			q0$new = FALSE
			q = rbind(q0, q)
		}

		qnotnull = (q$lid != 0)
		if (any(qnotnull)) q$lid2[qnotnull] = rank(q$lid[qnotnull])

		q = q[order(q$lid2), ]
		q$pane = "tilePane"

		q$pane[q$lid2 > 0] = pane_name(q$lid[q$lid2 > 0])

		# via tm_group, control and zoom levels can be set
		# these are stored in tmap options view_group_{name}, and merged with q:
		og = o[substr(names(o), 1, 10) == "view_group"]
		ogdf = 	data.frame(name = vapply(og, "[[",FUN.VALUE = character(1), "name", USE.NAMES = FALSE),
						   control = vapply(og, function(x) {
						   	as.character(x[["control"]])
						   } , FUN.VALUE = character(1), USE.NAMES = FALSE),
						   zoom_levels = I(unname(lapply(og, "[[", "zoom_levels"))))
		q$group.control = mapply(function(gi, v) {
			id = which(ogdf$name == gi)
			if (!length(id)) return(v)
			v2 = ogdf$control[id]
			if (is.na(v2)) return(v) else return(v2)
		}, q$group, q$group.control, USE.NAMES = FALSE, SIMPLIFY = TRUE)
		q$group.zoom_levels = mapply(function(gi, v) {
			id = which(ogdf$name == gi)
			if (!length(id)) return(NA)
			return(ogdf$zoom_levels[[id]])
		}, q$group, q$group.control, USE.NAMES = FALSE, SIMPLIFY = FALSE)
		q$group.control[!is.na(q$group.zoom_levels)] = "none" # see leaflet::groupOptions



		# q data frame:
		# gid = tmap-group counter
		# g_lid = layer counter inside tmap-group
		# lid = possibly-user-defined layer order number
		# lid2 = same as lid, but 1,2,3,...
		# pane = pane name (for view mode)
		# group = group name (for selecting layers in view mode)
		# group.control = leaflet::addLayerControl: check, radio or none (view mode)
		# group.zoom_levels = leaflet::groupOptions: zoom levels displayed at

		do.call(FUNaux, list(o = o, q = q))


		# plot xtab headers
		if (o$panel.type == "xtab") {
			for (k in 1:o$npages) {
				labrows = o$panel.labels[[1]]
				labcols = o$panel.labels[[2]]
				if (length(labrows) == o$nrows) for (i in 1:o$nrows) do.call(FUNxtab, list(label = labrows[i], facet_row = i, facet_page = k, o = o))
				if (length(labcols) == o$ncols) for (j in 1:o$ncols) do.call(FUNxtab, list(label = labcols[j], facet_col = j, facet_page = k, o = o))

			}
		}


		# plot xlab and ylab
		if (o$xlab.show) {
			for (k in 1:o$npages) {
				do.call(FUNxlab, list(facet_page = k, o = o))
			}
		}
		if (o$ylab.show) {
			for (k in 1:o$npages) {
				do.call(FUNylab, list(facet_page = k, o = o))
			}
		}



		for (i in seq_len(nrow(d))) {
			bbx = d$bbox[[i]]
			if (o$panel.type == "wrap") do.call(FUNwrap, list(label = o$panel.labels[[o$panel.labels.dim]][d[[paste0("by", o$panel.labels.dim_data)]][i]], facet_row = d$row[i], facet_col = d$col[i], facet_page = d$page[i], o = o))
			if (is.na(d$asp[i])) next
			do.call(FUNshape, list(bbx = bbx, facet_row = d$row[i], facet_col = d$col[i], facet_page = d$page[i], o = o))

			# plot grid labels
			if (o$grid.show && !o$grid.labels.inside_frame) {
				if (o$grid.labels.show[2] && ((o$grid.labels.pos[1] == "left" && d$col[i] == 1) || (o$grid.labels.pos[1] == "right" && d$col[i] == o$ncols))) {
					do.call(FUNgridylab, list(bi = d$bi[i], bbx = bbx, facet_row = d$row[i], facet_col = d$col[i], facet_page = d$page[i], o = o))
				}
				if (o$grid.labels.show[1] && ((o$grid.labels.pos[2] == "top" && d$row[i] == 1) || (o$grid.labels.pos[2] == "bottom" && d$row[i] == o$nrows))) {
					do.call(FUNgridxlab, list(bi = d$bi[i], bbx = bbx, facet_row = d$row[i], facet_col = d$col[i], facet_page = d$page[i], o = o))
				}
			}

			for (qi in which(q$new)) {
				gid = q$gid[qi]
				g_lid = q$g_lid[qi]
				pane = q$pane[qi]
				group = q$group[qi]
				if (gid > 0) {
					# data layer
					bl = tmx[[gid]]$layers[[g_lid]]
					shpTM = get_shpTM(bl$shpDT, d$by1[i], d$by2[i], d$by3[i])[[1]]
					mdt = get_dt(bl$mapping_dt, d$by1[i], d$by2[i], d$by3[i])

					id = paste0("f", sprintf("%03d", i), "g", sprintf("%02d", gid), "l", sprintf("%02d", g_lid))

					if (nrow(mdt) != 0) {
						gp = bl$gp

						FUN = paste0("tmap", gs, "DataPlot")

						a = structure(bl$mapping_args, class = c(bl$mapping_fun, "list"))

						do.call(FUN, c(list(a = a, shpTM = shpTM, dt = mdt, pdt = bl$popup.data, popup.format = bl$popup.format, ptdt = bl$popup.title.data, popup.layout = bl$popup.layout, hdt = bl$hover.data, idt = bl$id.data, gp = gp, bbx = bbx, facet_col = d$col[i], facet_row = d$row[i], facet_page = d$page[i], id = id, pane = pane, group = group, glid = bl$glid, o = o)))
					}

				} else {
					g_lid = q$g_lid[qi]


					# aux layer
					a = aux[[g_lid]]
					a_args = structure(a$args, class = c(a$mapping.fun, "list"))

					FUNaux_plot = paste0("tmap", gs, "AuxPlot")

					id = g_lid # to do: test!
					do.call(FUNaux_plot, list(a = a_args, bi = d$bi[i], bbx = bbx, facet_col = d$col[i], facet_row = d$row[i], facet_page = d$page[i], id = id, pane = pane, group = group, o = o))

				}



			}

			# inset map frame
			if (length(inset_frame_ids)) {
				sfc_bbxs = do.call(c, lapply(inset_frame_ids, function(iid) {
					ic = cdt$comp[[iid]]
					sf::st_transform(tmaptools::bb_poly(ic$bbox), crs = crs)
				}))

				collect_properties = function(comps, props) {
					res = lapply(props, function(p) {
						sapply(comps, function(cmp) {
							cmp[[p]]
						})
					})
					names(res) = props
					res
				}

				prp = collect_properties(cdt$comp[inset_frame_ids],
										 c("box_frame", "box_frame.color", "box_frame.alpha", "box_frame.lwd", "box_frame.lty", "box_bg", "box_bg.color", "box_bg.alpha"))
				calld = lapply(cdt$comp[inset_frame_ids], function(cmp) cmp$called)

				prp$box_frame[vapply(calld, function(cl) "box_frame.color" %in% cl, FUN.VALUE = logical(1))] = TRUE
				prp$box_bg[vapply(calld, function(cl) "box_bg.color" %in% cl, FUN.VALUE = logical(1))] = TRUE

				prp$box_bg.color[!prp$box_bg] =  "#00000000"
				prp$box_frame.color[!prp$box_frame] =  "#00000000"


				tb = tm_polygons()[[1]]


				gp = tb$gpar

				FUN = paste0("tmap", gs, "DataPlot")

				a = structure(tb$mapping.args, class = c("tm_data_polygons", "list"))

				shapepTM = shapeTM(sfc_bbxs, tmapID = seq_along(sfc_bbxs))

				shpTM = tmapTransPolygons(shapepTM, args = tb$trans.args, plot.order = tb$plot.order)

				mdt = data.table(tmapID__ = seq_along(sfc_bbxs), fill = prp$box_bg.color, col = prp$box_frame.color, lwd = prp$box_frame.lwd, lty = prp$box_frame.lty, fill_alpha = prp$box_bg.alpha, col_alpha = prp$box_frame.alpha, ord__ = 1L)

				do.call(FUN, c(list(a = a, shpTM = shpTM, dt = mdt, pdt = NULL, popup.format = list(), hdt = NULL, idt = NULL, gp = gp, bbx = bbx, facet_col = d$col[i], facet_row = d$row[i], facet_page = d$page[i], id = "inset_frame", pane = "tmap500", group = NA, o = o)))

			}
			do.call(FUNoverlay, list(bbx = bbx, facet_row = d$row[i], facet_col = d$col[i], facet_page = d$page[i], o = o))
		}

	} else {
		# init
		asp = do.call(FUNinit, list(o = o, return.asp = return.asp, vp = vp))
	}


	if (o$legend.only) {
		cdt[, facet_row := "1"]
		cdt[, facet_col := "1"]
		cdt[, page := 1L]
		cdt[, bbox := list()]
	} else {
		is_in = cdt$class == "in"
		#is_in = rep(TRUE, nrow(cdt))
		if (any(is_in)) {
			legs_in = lapply(which(is_in), function(i) {
				d2 = data.table::copy(d)
				legsi = cdt[i, ]
				if (o$type != "grid" && o$nrows == 1) {
					# reverse above
					d2[, by2 := by1]
					d2[, by1 := 1]
				}

				if (is.na(legsi$by1__)) d2[, by1:= NA]
				if (is.na(legsi$by2__)) d2[, by2:= NA]
				if (is.na(legsi$by3__)) d2[, by3:= NA]
				legsi = merge(legsi, d2[, c("by1", "by2", "by3", "row", "col", "page", "bbox", "units"), with = FALSE], by.x = c("by1__", "by2__", "by3__"), by.y = c("by1", "by2", "by3"))
				legsi[, ':='(facet_row = as.character(row), facet_col = as.character(col), row = NULL, col = NULL)]
				legsi
			})
		} else {
			legs_in = NULL
		}


		legs_out = copy(cdt[!is_in])
		if (nrow(legs_out) >= 1L) {
			legs_out[, page:=NA_integer_]
			legs_out[!is.na(by3__), page := by3__]

			by1_clip =  (length(which(legs_out$by1__ > o$nrows)) > 0L) && all(legs_out$facet_row == legs_out$by1__)
			by2_clip =  (length(which(legs_out$by2__ > o$ncols)) > 0L) && all(legs_out$facet_col == legs_out$by2__)
			if (by1_clip && by2_clip) {
				cli::cli_abort("multiple pages cannot be created over two facet dimensions")
			}
			if (by1_clip) {
				legs_out[!is.na(by1__), ':='(page = ((by1__-1L) %/% o$nrows) + 1L, facet_row = ((by1__-1L) %% o$nrows) + 1L)]
			} else if (by2_clip) {
				legs_out[!is.na(by2__), ':='(page = ((by2__-1L) %/% o$ncols) + 1L, facet_col = ((by2__-1L) %% o$ncols) + 1L)]
			}

			# ad-hoc method: take first bbox and units
			bbox_nb = d$bbox[1]
			attr(bbox_nb, "borrow") = list(col = d$col[1], row = d$row[1])
			legs_out[, bbox:=list(bbox_nb)]
			legs_out[, units:=list(d$units[1])]
		} else {
			legs_out$page = integer()
			legs_out$bbox = list()
			legs_out$units = list()
		}



		cdt = data.table::rbindlist(c(list(legs_out), legs_in))

		cdt$comp = mapply(function(cmp, bbx, u) {
			if (!("bbox" %in% names(cmp))) cmp$bbox = bbx
			cmp$units = u
			cmp
		}, cdt$comp, cdt$bbox, cdt$units, SIMPLIFY = FALSE)

		leg_dummy = tm_legend_hide()
		leg_dummy$active = FALSE
		leg_dummy$vneutral = NA

		leg_nr_dummy = legend_save(leg_dummy)
		crt_nr_dummy = chart_save(tm_chart_none())

		# inset ids
		inset_ids = if (nrow(cdt) == 0L) integer(0) else which(sapply(cdt$comp, inherits, "tm_inset_map"))


		# inset maps: prepare input for step4
		if (length(inset_ids)) {
			cdt$comp[inset_ids] = lapply(cdt$comp[inset_ids], function(comp) {
					tmo_i = tm$tmo
					o_i = tm$o
					aux_i = tm$aux
					cmp_i = list()
					prx_i = list()

					if ("crs" %in% names(comp)) {
						crs_i = comp$crs
						o_i$crs_step4 = crs_i
					} else {
						crs_i = o_i$crs_step4
					}
					o_i$outer.bg = FALSE
					o_i$frame = comp$main_frame
					o_i$frame.color = comp$main_frame.color
					o_i$frame.alpha = comp$main_frame.alpha
					o_i$frame.lwd = comp$main_frame.lwd
					o_i$frame.r = comp$main_frame.r


					# set legends to inactive
					tmo_i = lapply(tmo_i, function(tmg) {
						lapply(tmg, function(tml) {
							lapply(tml, function(tmi) {
								tmi$shpDT$shpTM = lapply(tmi$shpDT$shpTM, crs_reproject_shpTM, crs = crs_i, raster.warp = o$raster.warp)
								tmi$mapping_legend = lapply(tmi$mapping_legend, function(l) {
									l$legnr = leg_nr_dummy
									l$crtnr = crt_nr_dummy
									l
								})
								tmi$trans_legend = lapply(tmi$trans_legend, function(l) {
									l$legnr = leg_nr_dummy
									l$crtnr = crt_nr_dummy
									l
								})
								tmi
							})
						})
					})

				#	(tm, vp, return.asp, show, in.shiny, knit, args)
					comp$tm = list(tmo = tmo_i,
								   o = o_i,
								   aux = aux_i,
								   cmp = cmp_i,
								   prx = prx_i)
					comp
			})
		}
	}

	legfun = paste0("tmap", gs, "Comp")

	toI = function(x) {
		as.integer(strsplit(x, split = "_")[[1]])
	}

	if (nrow(cdt) > 0L) for (k in seq_len(o$npages)) {
		klegs = cdt[is.na(page) | (page == k), ] # was by3__ instead of page
		#klegs[, pos.h.id := pos.h][pos.h %in% c("left", "center", "right"), pos.h.id:="lower"][pos.h %in% c("LEFT", "CENTER", "RIGHT"), pos.h.id:="upper"]
		#klegs[, pos.v.id := pos.v][pos.v %in% c("top", "center", "bottom"), pos.v.id:="lower"][pos.v %in% c("TOP", "CENTER", "BOTTOM"), pos.v.id:="upper"]

		# find group ids
		klegs$id = paste(klegs$class, klegs$cell.h, klegs$cell.v, klegs$pos.h, klegs$pos.v,
						 sapply(klegs$comp, function(comp) paste(comp$position$just.h, comp$position$just.v, sep = "_")), sep = "_")
		# just.id = NULL
		# klegs$just.id = sapply(klegs$comp, function(l) paste(l$position$just.h, l$position$just.v, sep = "."))
		# klegs[, id:=paste(pos.h.id, pos.v.id, just.id, class, sep = "__")]
		klegs[, do.call(legfun, args = list(comp = .SD$comp, o = o, facet_row = toI(.SD$facet_row[1]), facet_col = toI(.SD$facet_col[1]), facet_page = k, class = .SD$class[1], stack = .SD$stack, stack_auto = .SD$stack_auto, pos.h = .SD$pos.h, pos.v = .SD$pos.v, .SD$bbox)), by = list(facet_row, facet_col, id), .SDcols = c("comp", "facet_row", "facet_col", "class", "stack", "stack_auto", "pos.h", "pos.v", "bbox")]
	}

	if (show) save_last_map()

	if (.TMAP$in.shiny) {
		.TMAP$q = q
	}

	do.call(FUNrun, list(o = o, q = q, show = show, knit = knit, knit_opts = knit_opts, args))
}

# bb_ext reversing
bb_ext_rev = function(bbx, ext = c(0, 0, 0, 0)) {
	dx = (bbx[3] - bbx[1]) / (1 + ext[2] + ext[4])
	dy = (bbx[4] - bbx[2]) / (1 + ext[1] + ext[3])

	bbx[2] = bbx[2] + ext[1] * dy
	bbx[1] = bbx[1] + ext[2] * dx
	bbx[4] = bbx[4] - ext[3] * dy
	bbx[3] = bbx[3] - ext[4] * dx
	bbx
}

get_geo_ref = function(bbx, crs, inner_margins, dev_size, map_size, offset) {
	bbx_crop = bb_ext_rev(bbx, inner_margins)

	x1 = offset[2] / dev_size[1]
	x2 = (dev_size[1] - offset[4]) / dev_size[1]

	y1 = offset[1] / dev_size[2]
	y2 = (dev_size[2] - offset[3]) / dev_size[2]

	xy_crop = bb_ext_rev(c(x1, y1, x2, y2), inner_margins)
	names(xy_crop) = c("xmin", "ymin", "xmax", "ymax")
	list(crs = crs, bbx = bbx_crop, bbx_frame = bbx, rel_coords = xy_crop)
}

# Internal helper: generate a short "Unit: ..." sentence for a visual variable,
# for use as inline R code inside roxygen @param tags, e.g.:
#
#   #' @param size,size.scale,... `r .doc_vv("size")` `r .doc_unit("size")`
#
# Plain ASCII only -- no \uXXXX sequences, no Unicode dashes -- so the
# output survives the Rd processing pipeline on all platforms.
.doc_unit = function(aes) {
	lu = list(
		fill        = "Color -- a color name, hex string.",
		col         = "Color -- a color name, hex string.",
		bgcol       = "Color -- a color name, hex string.",
		fill_alpha  = "Proportion -- numeric 0-1 (0 = fully transparent, 1 = fully opaque).",
		col_alpha   = "Proportion -- numeric 0-1 (0 = fully transparent, 1 = fully opaque).",
		bgcol_alpha = "Proportion -- numeric 0-1 (0 = fully transparent, 1 = fully opaque).",

		# Symbol / bubble / square / dot size: typographic lines
		size        = paste0(
			"Typographic lines (\"lines\"); 1 line is approx. 1/6 inch.",
			" Controlled by `values.scale` and `tmap_options(values.scale = ...)`."
		),

		# Circle size: geographic meters
		size.circles = paste0(
			"Meters. Accepts a plain numeric vector (values already in meters) or a",
			" `units` object from the **units** package (any linear unit,",
			" e.g. `units::as_units(50, \"km\")`), which is converted to meters automatically."
		),

		# Text / label size: cex multiplier of the base font size.
		# In plot mode the base size is par("ps") = 12 pt (base R default).
		# In view mode: sizeChar = round(gp$cex * 12) px  (tmapLeafletDataPlot_text.R).
		# Both modes share the same base: 1 unit = 12 pt / 12 px.
		size.text   = paste0(
			"Multiplier of the base font size.",
			" `size = 1` renders at the default font size,",
			" which is 12 pt in plot mode (`par(\"ps\")`) and 12 px in",
			" view mode (consistent by design).",
			" `size = 1.5` renders at 18 pt / px, etc."
		),

		lwd         = paste0(
			"Base R line-width units; 1 lwd is approx. 0.75 pt at 96 dpi."
		),
		lty         = paste0(
			"Integer (1-6) or name: \"solid\", \"dashed\", \"dotted\",",
			" \"dotdash\", \"longdash\", \"twodash\"."
		),
		shape       = "Integer `pch` code (1-25) used as a plotting symbol. See example of `tm_symbols()`",
		angle       = "Degrees, clockwise from north (0-360).",
		fontface    = "\"plain\", \"bold\", \"italic\", or \"bold.italic\".",
		text        = "Character string.",
		xmod        = "Line heights, relative to the label anchor. Positive = right.",
		ymod        = "Line heights, relative to the label anchor. Positive = up."
	)

	info = lu[[aes]]
	if (is.null(info)) return("")
	paste0("*Unit:* ", info)
}


# -------------------------------------------------------------------------
# Details section to add to tm_polygons (inherited by all layer functions
# via  @inherit tm_polygons details).
#
# Paste the block below (without the leading '# ') into the tm_polygons
# roxygen header. All text is plain ASCII so it survives the Rd pipeline.
# -------------------------------------------------------------------------
#
# #' @details
# #' ## Visual variable units
# #'
# #' Every visual variable maps data values to a specific output unit.
# #' Knowing the unit matters when supplying constant values via [tm_const()],
# #' or output ranges via `values.range` / `values.scale` in the scale
# #' functions.
# #'
# #' | Variable | Output unit | Notes |
# #' |---|---|---|
# #' | `fill`, `col`, `bgcol` | color | name, hex, or palette string |
# #' | `fill_alpha`, `col_alpha`, `bgcol_alpha` | proportion 0-1 | 0 = transparent, 1 = opaque |
# #' | `size` (symbols, bubbles, squares, dots) | typographic lines | 1 line approx. 1/6 inch; scaled by `values.scale` |
# #' | `size` (circles) | meters | plain numeric or a `units` object |
# #' | `size` (text, labels) | multiplier | 1 = 12 pt (plot) / 12 px (view) |
# #' | `lwd` | lwd | base R units; 1 lwd approx. 0.75 pt at 96 dpi |
# #' | `lty` | -- | integer 1-6 or name ("solid", "dashed", ...) |
# #' | `shape` | -- | integer `pch` 1-25 or single character |
# #' | `angle` | degrees | 0-360, clockwise from north |
# #' | `fontface` | -- | "plain", "bold", "italic", "bold.italic" |
# #'
# #' ### Symbol size (`size` in `tm_symbols`, `tm_bubbles`, `tm_squares`, `tm_dots`)
# #' "Lines" is a typographic unit: one line is approximately 1/6 inch (the
# #' default base line-height in R graphics). The global multiplier
# #' `tmap_options(values.scale = list(size.bubbles = 1.5))` scales all symbol
# #' sizes without changing the data mapping.
# #'
# #' ### Circle size (`size` in `tm_circles`)
# #' The value is a geographic radius in meters. A plain numeric vector is
# #' interpreted as meters; a `units` object (from the **units** package) is
# #' automatically converted, so `units::as_units(1, "mi")` gives a 1-mile
# #' radius. Because the radius is geographic, circles scale with zoom in
# #' interactive (view) mode -- unlike bubble symbols which keep a fixed screen
# #' size.
# #'
# #' ### Text size (`size` in `tm_text`, `tm_labels`)
# #' The value is a multiplier of the base font size. `size = 1` renders at
# #' 12 pt in plot mode (R's default `par("ps")`) and at 12 px in view
# #' mode (`gp$cex * 12` px, see `tmapLeafletDataPlot.tm_data_text`); the two
# #' modes are consistent by design.

#' Arrange small multiples in grid layout
#'
#' Arrange small multiples in a grid layout. Normally, small multiples are created
#' by specifying multiple variables for one aesthetic or by specifying the by argument
#' (see [tm_facets()]). This function can be used to arrange custom small multiples in a grid layout.
#'
#' The global option `tmap.limits` controls the limit of the number of facets that are plotted.
#' By default, `tmap_options(tmap.limits = c(facets.view=4, facets.plot=64))`.
#' The maximum number of interactive facets is set to four since otherwise it may become very slow.
#'
#' @param ... [`tmap`] objects or one list of [`tmap`] objects.
#'   The number of multiples that can be plot is limited (see details).
#' @param ncol number of columns
#' @param nrow number of rows
#' @param widths vector of column widths. It should add up to 1 and the length
#'   should be equal to `ncol`.
#' @param heights vector of row heights. It should add up to 1 and the length
#'   should be equal to `nrow`.
#' @param sync logical. Should the navigation in view mode (zooming and panning)
#'   be synchronized? By default `FALSE`.
#' @param asp aspect ratio. The aspect ratio of each map. Normally, this is
#'   controlled by the `asp` argument from [tm_layout()] (also a tmap option).
#'   This argument will overwrite it, unless set to `NULL`.
#'   The default value for `asp` is 0, which means that the aspect ratio is
#'   adjusted to the size of the device divided by the number of columns and rows.
#'   When `asp` is set to `NA`, which is also the default value for `tm_layout()`,
#'   the aspect ratio will be adjusted to the used shapes.
#' @param outer.margins outer.margins, numeric vector four or a single value.
#'   If defines the outer margins for each multiple. If will overwrite the
#'   `outer.margins` argument from [tm_layout()], unless set to `NULL`.
#' @param height height of the arranged output. Only used in the interactive
#'   modes (`"view"`, `"maplibre"`, `"mapbox"`). A bare number is interpreted as
#'   pixels; a string is used as-is (e.g. `"80vh"`). The default `NULL` fills the
#'   available viewport height.
#' @param x a `tmap_arrange` object (returned from `tmap_arrange()`).
#' @param knit should [knitr::knit_print()] be enabled, or the normal [base::print()] function?
#' @param options options passed on to [knitr::knit_print()]
#' @example ./examples/tmap_arrange.R
#' @export
tmap_arrange = function(..., ncol = NA, nrow = NA, widths = NA, heights = NA, sync = FALSE, asp = 0, outer.margins = .02, height = NULL) {
	tms = list(...)
	if (!inherits(tms[[1]], "tmap")) {
		if (!is.list(tms[[1]])) stop("The first argument of tmap_arrange is neither a tmap object nor a list.")
		tms = tms[[1]]
	}
	istmap = vapply(tms, FUN = inherits, FUN.VALUE = logical(1), what = "tmap")
	if (!all(istmap)) stop("Not all arguments are tmap objects.")
	opts = list(ncol=ncol, nrow=nrow, widths=widths, heights=heights, sync=sync, asp=asp, outer.margins=outer.margins, height=height)
	attr(tms, "opts") = opts
	class(tms) = c("tmap_arrange", class(tms))
	tms
}

#' @rdname tmap_arrange
#' @exportS3Method knitr::knit_print
knit_print.tmap_arrange = function(x, ..., options = NULL) {
	print_tmap_arrange(x, knit=TRUE, ..., options = options)
}

#' @export
#' @rdname tmap_arrange
print.tmap_arrange = function(x, knit = FALSE, ..., options = NULL) {
	print_tmap_arrange(x, knit=knit, ..., options = options)
}




print_tmap_arrange = function(tms, knit = FALSE, show = TRUE, add.titles = TRUE, ..., options = options) {
	args = list(...)

	opts = attr(tms, "opts")

	tms_len = length(tms)

	nx = limit_nx(tms_len)
	if (nx != tms_len) tms = tms[1:nx]

	if (is.na(opts$ncol) && is.na(opts$nrow)) {
		devsize = dev.size()
		dasp = devsize[1] / devsize[2]

		## determine 'overall' aspect ratio by overlaying the maps
		#tmp = tempfile(fileext = ".png")
		#png( tmp, width=700, height=700, res = 100)

		curdev = grDevices::dev.cur()
		rmc = tmap_options(raster.max_cells = 36)

		on.exit(tmap_options(rmc))

		tasps = suppressMessages({
			vapply(tms, function(tm) {
				asp = get_asp_ratio(tm, width = 700, height = 700, res = 100)
				#asp = print_tmap(tm, return.asp = TRUE, mode = "plot")
				# dev.off()
				asp
			}, numeric(1))
		})

		grDevices::dev.set(curdev)
		tmap_options(rmc)

		#dev.off()
		#dev.off()
		hs = vapply(tasps, function(tasp) ifelse(tasp>1, 1, 1/tasp), numeric(1))
		ws = vapply(tasps, function(tasp) ifelse(tasp>1, tasp, 1), numeric(1))
		iasp = max(ws) / max(hs)

		asp_ratio = iasp / dasp

		nrowcol = process_get_arrangement(nx = nx, asp_ratio = asp_ratio)
		nrow = nrowcol[1]
		ncol = nrowcol[2]


	} else {
		ncol = if (is.na(opts$ncol)) ceiling(nx / opts$nrow) else opts$ncol
		nrow = if (is.na(opts$nrow)) ceiling(nx / opts$ncol) else opts$nrow
	}


	interactive = (getOption("tmap.mode")=="view")
	gs = tmap_graphics_name()

	fun = paste0("tmap", gs, "Arrange")

	do.call(fun, list(tms = tms, nx = nx, ncol = ncol, nrow = nrow, opts = opts, knit = knit, show = show, args = args, options = options))

}





process_get_arrangement = function(nx, asp_ratio) {
	#           asp ~ nrow
	#       |--------------
	#   1   |
	# ~ncol |       nx
	#       |
	ncol_init = sqrt(nx/asp_ratio)
	nrow_init = nx / ncol_init

	# rounding:
	nrow_ceiling = min(ceiling(nrow_init), nx)
	ncol_ceiling = min(ceiling(ncol_init), nx)

	# find minimal change
	nrow_xtra = abs(nrow_ceiling - nrow_init) * ncol_init
	ncol_xtra = abs(ncol_ceiling - ncol_init) * nrow_init

	# calculaet the other, and subtract 1 when possible
	if (nrow_xtra < ncol_xtra) {
		nrow = nrow_ceiling
		ncol = ceiling(nx / nrow)
		if ((nrow-1) * ncol >= nx) nrow = nrow - 1
	} else {
		ncol = ncol_ceiling
		nrow = ceiling(nx / ncol)
		if ((ncol-1) * nrow >= nx) ncol = ncol - 1
	}
	c(nrow=nrow, ncol=ncol)
}

#' Internal tmap function to create a tmap shape
#'
#' Internal tmap function to create a tmap shape
#'
#' @param shp Shape file
#' @param tmapID tmap Identifier
#' @param bbox bounding box
#' @export
#' @keywords internal
shapeTM = function(shp, tmapID = NULL, bbox = NULL, ...) {
	# filter empty geometries
	if (inherits(shp, "sfc")) {
		is_empty = sf::st_is_empty(shp)
		shp = shp[!is_empty]
		tmapID = tmapID[!is_empty]
	}
	x = structure(list(shp = shp, tmapID = tmapID, bbox = bbox, ...), class = c("shapeTM", "list"))
}


stm_bbox = function(shpTM, tmapID, crs) {
	bbox = shpTM$bbox

	if (!is.null(bbox$x)) {
		if (identical(bbox$x, "FULL")) {
			bbox$x = sf::st_bbox(c(xmin = -180, xmax = 180, ymax = 90, ymin = -90), crs = sf::st_crs(4326))
		}
		# to make sure the supplied bounding box will be converted in the correct crs
		bbox$projection = crs
	} else if (inherits(shpTM$shp, "character")) {
		bbox$x = sf::st_bbox(c(xmin = -180, xmax = 180, ymax = 90, ymin = -90), crs = sf::st_crs(4326))
	} else {
		shp = shpTM$shp
		shpID = shpTM$tmapID

		if (length(tmapID) == 0) return(sf::st_bbox(NA_real_))
		# filter the shape?
		do_filter = (length(tmapID) != length(shpID)) || (!all(tmapID == shpID))

		if (do_filter) {
			ids = match(tmapID, shpID)

			if (inherits(shp, "dimensions")) {
				m = matrix(NA, nrow = nrow(shp), ncol = ncol(shp))
				m[ids] = TRUE

				s = stars::st_as_stars(list(values = m), dimensions = shp)
				shp = trim_stars(s)
			} else if (inherits(shp, c("sf", "sfc"))) {
				shp = shp[ids]
			} else {
				stop("unknown shape class")
			}
		}
		bbox$x = sf::st_bbox(shp)
	}
	if (inherits(bbox$x, "SpatExtent")) bbox$x = sf::st_bbox(bbox$x)
	do.call(tmaptools::bb, bbox)
}


stm_merge_bbox = function(blist) {
	crs = sf::st_crs(blist[[1]])
	m = do.call(rbind, blist)
	sf::st_bbox(c(apply(m[, 1:2, drop = FALSE], 2, min), apply(m[, 3:4, drop = FALSE], 2, max)), crs = crs)
}


stm_bbox_all = function(shpTM) {
	sf::st_bbox(shpTM$shp)
}

bb_ext = function(bbx, ext = c(0, 0, 0, 0)) {
	dx = bbx[3] - bbx[1]
	dy = bbx[4] - bbx[2]

	bbx[2] = bbx[2] - ext[1] * dy
	bbx[1] = bbx[1] - ext[2] * dx
	bbx[4] = bbx[4] + ext[3] * dy
	bbx[3] = bbx[3] + ext[4] * dx
	bbx
}

bb_asp = function(bbx, asp) {
	sasp = get_asp_ratio(bbx)

	fact = sasp / asp

	if (fact > 1) {
		cy = mean(bbx[c(2,4)])
		dy = bbx[4] - bbx[2]
		dy2 = dy * fact
		bbx[2] = cy - (dy2 / 2)
		bbx[4] = cy + (dy2 / 2)
	} else if (fact < 1) {
		cx = mean(bbx[c(1,3)])
		dx = bbx[3] - bbx[1]
		dx2 = dx / fact
		bbx[1] = cx - (dx2 / 2)
		bbx[3] = cx + (dx2 / 2)
	}
	bbx
}

bb_ll_valid = function(bbx) {
	bbx[2] = max(bbx[2], -90)
	bbx[4] = min(bbx[4], 90)
	bbx
}

#' @export
#' @rdname tmap_internal
tmapUsrCls = function(x) {
	structure(x, class = "tmapUsrCls")
}



#' @export
#' @rdname tmap_internal
format_aes_results = function(values, ord = NULL, legend, chart) {
	legnr = vector(mode = "integer", length = length(values))
	legnr[1] = legend_save(legend)

	crtnr = vector(mode = "integer", length = length(values))
	crtnr[1] = chart_save(chart)

	if (is.null(ord)) {
		list(values = values,
			 legnr = legnr,
			 crtnr = crtnr)
	} else {
		list(values = values,
			 ord = ord,
			 legnr = legnr,
			 crtnr = crtnr)
	}
}





legends_init = function() {
	assign("legs", list(), envir = .TMAP)
}

charts_init = function() {
	assign("charts", list(), envir = .TMAP)
}


legend_save = function(legend) {
	if (!exists("legs", envir = .TMAP)) legends_init()
	legs = get("legs", envir = .TMAP)
	legs = c(legs, (list(legend)))
	assign("legs", legs, envir = .TMAP)
	length(legs)
}

#' @export
#' @rdname tmap_internal
tmapGetLegend = function(id) {
	if (!exists("legs", envir = .TMAP)) return(NULL)
	legs = get("legs", envir = .TMAP)
	if (id < length(legs)) {
		legs[[id]]
	} else {
		NULL
	}
}

#' @export
#' @rdname tmap_internal
chart_save = function(legend) {
	if (!exists("charts", envir = .TMAP)) charts_init()
	charts = get("charts", envir = .TMAP)
	charts = c(charts, (list(legend)))
	assign("charts", charts, envir = .TMAP)
	length(charts)
}

#' @export
#' @rdname tmap_internal
data_type = function(x) {
	if (all(is.na(x))) {
		"na"
	} else if (is.ordered(x)) {
		"order"
	} else if (is.logical(x) || is.character(x) || is.factor(x)) {
		"factor"
	} else if (is.numeric(x)) {
		if (any(x < 0 & !is.na(x)) && any(x > 0 & !is.na(x))) {
			"div"
		} else {
			"seq"
		}
	} else {
		"unknown"
	}
}

data_type_grp = function(x) {
	if (x %in% c("seq", "div")) {
		"num"
	} else {
		x
	}
}

#' @export
#' @rdname tmap_internal
data_class = function(x, check_for_color_class = FALSE, midpoint_enabled = FALSE) {
	# if (all(is.na(x))) {
	# 	"na"
	# } else
	cls = if (inherits(x, c("POSIXct", "POSIXlt"))) {
		c("datetime", "seq")
	} else if (inherits(x, "Date")) {
		c("date", "seq")
	} else if (is.numeric(x)) {
		y = without_units(x)
		subclass1 = if (is.integer(x)) "int" else "real"
		subclass2 = if ((any(y < 0 & !is.na(y)) && any(y > 0 & !is.na(y))) || midpoint_enabled) {
			"div"
		} else {
			"seq"
		}
		c("num", subclass1, subclass2)
	} else {
		if (check_for_color_class) {
			w = which(!is.na(x))
			if (length(w) && all(valid_colors(head(x[w], 100)))) {
				c("asis", "color")
			} else {
				subclass = if (is.ordered(x)) "ord" else "unord"
				c("fact", subclass)
			}
		} else {
			subclass = if (is.ordered(x)) "ord" else "unord"
			c("fact", subclass)
		}
	}

	attr(cls, "units") = if (inherits(x, "units")) {
		paste0(" [", units(x), "]")
	} else ""
	attr(cls, "unique") = (length(na.omit(unique(x)))==1)
	cls
}


#' @export
#' @rdname tmap_internal
tmapScale = function(aes, value, scale, legend, chart, free) {
	if (is.null(legend)) legend = tm_legend_hide()
	structure(list(aes = aes, value = tmapVV(value), scale = scale, legend = legend, chart = chart, free = free), class = c("tmapScale", "list"))
}

#' @export
#' @rdname tmap_internal
tmapScaleAuto = function(x1, scale, legend, chart, o, aes, layer, layer_args, sortRev, bypass_ord, submit_legend = TRUE, ...) {
	args = list(...)
	k = length(args) + 1L
	if (length(args)) {
		names(args) = paste0("x", 2L:(length(args)+1L))
	}

	cls = data_class(x1, check_for_color_class = aes %in% c("col", "fill"))

	#if (cls[1] == "na")
	sc_opt = getAesOption("scales.var", o, aes, layer, cls = cls)
#if (aes == "fill") browser()
	if (k == 2) {
		sc = "bivariate"
	} else if (k > 2) {
		stop("No default scale for multivariate variables", call. = FALSE)
		#sc = "composition"
	} else if (cls[1] == "asis") {
		sc = "asis"
	} else if (attr(cls, "unique") && !(sc_opt == "asis")) {
		if ("num" %in% cls) {
			sc = "ordinal"
			cli::cli_inform("The visual variable {.arg {aes}} of the layer {.str {layer}} contains a unique value. Therefore a discrete scale is applied (tm_scale_discrete).")
		} else {
			sc = "categorical"
			cli::cli_inform("The visual variable {.arg {aes}} of the layer {.str {layer}} contains a unique value. Therefore a categorical scale is applied (tm_scale_categorical).")
		}
	} else {

		sc_pref = scale$fun_pref

		if (!is.null(sc_pref)) {
			if (sc_pref %in% c("categorical", "continuous", "continuous_log", "rank")) {
				sc = sc_pref
			} else {
				sc = sc_opt
			}
		} else {
			sc = sc_opt
		}
	}


	tm_scalefun = paste0("tm_scale_", sc)

	scale = scale[names(scale) %in% names(formals(tm_scalefun))]

	scale_new = do.call(tm_scalefun, args = scale)

	FUN = scale_new$FUN
	scale_new$FUN = NULL

	if (sc == "bivariate") {
		do.call(FUN, list(x1 = x1, x2 = args[[1]], scale = scale_new, legend = legend, chart = chart, o = o, aes = aes, layer = layer, layer_args = layer_args, sortRev = sortRev, bypass_ord = bypass_ord, submit_legend = submit_legend))
	} else if (sc == "multi_continuous") {
		do.call(FUN, c(list(x1 = x1), args, list(scale = scale_new, legend = legend, chart = chart, o = o, aes = aes, layer = layer, layer_args = layer_args, sortRev = sortRev, bypass_ord = bypass_ord, submit_legend = submit_legend)))
	} else {
		do.call(FUN, list(x1 = x1, scale = scale_new, legend = legend, chart = chart, o = o, aes = aes, layer = layer, layer_args = layer_args, sortRev, bypass_ord, submit_legend))
	}

}

#' Map layer: polygons
#'
#' Map layer that draws polygons. Supported visual variables are: `fill` (the fill color),
#' `col` (the border color), `lwd` (line width), `lty` (line type),
#' `fill_alpha` (fill color alpha transparency) and `col_alpha` (border color alpha transparency).
#'
#' The visual variable arguments (e.g. `col`) can be specified with a data
#' variable name (e.g., a spatial vector attribute or a raster layer of the object
#' specified in [tm_shape()]), with a visual value (for `col`, a color is expected), or with a geometry-derived variable (see below).
#' See `r .doc_see_vig("basics_vv")`.
#'
#' Multiple values can be specified: in that case facets are created.
#' These facets can be combined with other faceting data variables, specified with [tm_facets()].
#' See `r .doc_see_vig("basics_facets")`.
#'
#' - The `*.scale` arguments determine the used scale to map the data values to
#' visual variable values. These can be specified with one of the available
#' `tm_scale_*()` functions. The default is specified by the tmap option ([tm_options()]) `scales.var`.
#' See `r .doc_see_vig("basics_scales")`.
#'
#' - The `*.legend` arguments determine the used legend, specified with [tm_legend()].
#' The default legend and its settings are determined by the tmap options ([tm_options()]) `legend.` .
#' See `r .doc_see_vig("basics_legends")`.
#'
#' - The `*.chart` arguments specify additional charts, specified with `tm_chart_`, e.g. [tm_chart_histogram()].
#' See `r .doc_see_vig("basics_charts")`.
#'
#' - The `*.free` arguments determine whether scales are applied freely across facets, or shared.
#' A logical value is required. They can also be specified with a vector of three
#' logical values; these determine whether scales are applied freely per facet dimension.
#' This is only useful when facets are applied (see [tm_facets()]).
#' There are maximally three facet dimensions: rows, columns, and pages. This only
#' applies for a facet grid ([tm_facets_grid()]). For instance, `col.free = c(TRUE, FALSE, FALSE)`
#' means that for the visual variable `col`, each row of facets will have its own
#' scale, and therefore its own legend. For facet wraps and stacks
#' ([tm_facets_wrap()] and [tm_facets_stack()]) there is only one facet dimension,
#' so the `*.free` argument requires only one logical value.
#'
#' Currently, three geometry-derived variables are implemented:
#'
#' - `"AREA"` (polygons only), which uses the feature area;
#'
#' - `"LENGTH"` (lines only), which uses the feature length; and
#'
#' - `"MAP_COLORS"`, which assigns values so that adjacent features receive
#' different values, making it particularly suitable for coloring
#' neighbouring polygons.
#'
#' Note that geometry-derived variables do not generate a legend automatically.
#' If a legend is required, compute the corresponding variable explicitly,
#' for example with [sf::st_area()], [sf::st_length()], or
#' [tmaptools::map_coloring()], and use the resulting values instead.
#'
#' ## Visual variable units
#'
#' Every visual variable maps data values to a specific output unit.
#' Knowing the unit matters when supplying constant values via [tm_const()],
#' or output ranges via `values.range` / `values.scale` in the scale
#' functions.
#'
#' | Variable | Output unit | Notes |
#' |---|---|---|
#' | `fill`, `col`, `bgcol` | color | name, hex, or palette string |
#' | `fill_alpha`, `col_alpha`, `bgcol_alpha` | proportion 0-1 | 0 = transparent, 1 = opaque |
#' | `size` (symbols, bubbles, squares, dots) | typographic lines | 1 line approx. 1/6 inch; scaled by `values.scale` |
#' | `size` (circles) | meters | plain numeric or a `units` object |
#' | `size` (text, labels) | multiplier | 1 = 12 pt (plot) / 12 px (view) |
#' | `lwd` | lwd | base R units; 1 lwd approx. 0.75 pt at 96 dpi |
#' | `lty` | -- | integer 1-6 or name ("solid", "dashed", ...) |
#' | `shape` | -- | integer `pch` 1-25 or single character |
#' | `angle` | degrees | 0-360, clockwise from north |
#' | `fontface` | -- | "plain", "bold", "italic", "bold.italic" |
#'
#' ### Symbol size (`size` in `tm_symbols`, `tm_bubbles`, `tm_squares`, `tm_dots`)
#' "Lines" is a typographic unit: one line is approximately 1/6 inch (the
#' default base line-height in R graphics). The global multiplier
#' `tmap_options(values.scale = list(size.bubbles = 1.5))` scales all symbol
#' sizes without changing the data mapping.
#'
#' ### Circle size (`size` in `tm_circles`)
#' The value is a geographic radius in meters. A plain numeric vector is
#' interpreted as meters; a `units` object (from the **units** package) is
#' automatically converted, so `units::as_units(1, "mi")` gives a 1-mile
#' radius. Because the radius is geographic, circles scale with zoom in
#' interactive (view) mode -- unlike bubble symbols which keep a fixed screen
#' size.
#'
#' ### Text size (`size` in `tm_text`, `tm_labels`)
#' The value is a multiplier of the base font size. `size = 1` renders at
#' 12 pt in plot mode (R's default `par("ps")`) and at 12 px in view
#' mode (`gp$cex * 12` px, see `tmapLeafletDataPlot.tm_data_text`); the two
#' modes are consistent by design.
#'
#' ## Layer blending (`blend`)
#'
#' Blend modes control how a layer's pixels are combined with the pixels
#' beneath it. For each pixel, let \eqn{S} be the source (top layer) RGB
#' value and \eqn{D} be the destination (bottom layer) RGB value, both
#' normalised to \eqn{[0, 1]}.
#'
#' | `blend` | Formula | Use case |
#' | --- | --- | --- |
#' | `"over"` | \eqn{S \cdot \alpha + D \cdot (1 - \alpha)} | Standard alpha compositing (default) |
#' | `"multiply"` | \eqn{S \times D} | Hillshading over colour raster; both layers darken each other |
#' | `"screen"` | \eqn{1 - (1 - S)(1 - D)} | Inverse of multiply; brightens |
#' | `"overlay"` | multiply if \eqn{D < 0.5}, screen if \eqn{D \geq 0.5} | Boosts contrast; preserves midtones |
#' | `"darken"` | \eqn{\min(S, D)} | Keeps the darker of the two layers per channel |
#' | `"lighten"` | \eqn{\max(S, D)} | Keeps the lighter of the two layers per channel |
#'
#' Requires R >= 4.2 and a compatible graphics device (e.g.
#' `png(type = "cairo")`, `svg()`). In view mode, blending is applied via
#' CSS `mix-blend-mode`. See [`grid::groupGrob()`] for the full list of
#' supported operators.
#'
#' ## zindex and pane names
#'
#' In view mode, each layer is rendered in a Leaflet pane named `"tmap{zindex}"`
#' (e.g., `"tmap401"`, `"tmap402"`), with base tile layers placed in the
#' standard `"tile"` pane.
#' @param fill,fill.scale,fill.legend,fill.chart,fill.free `r .doc_vv("fill")` `r .doc_unit("fill")`
#' @param col,col.scale,col.legend,col.chart,col.free `r .doc_vv("col")` `r .doc_unit("col")`
#' @param lwd,lwd.scale,lwd.legend,lwd.chart,lwd.free `r .doc_vv("lwd")` `r .doc_unit("lwd")`
#' @param lty,lty.scale,lty.legend,lty.chart,lty.free `r .doc_vv("lty")` `r .doc_unit("lty")`
#' @param fill_alpha,fill_alpha.scale,fill_alpha.chart,fill_alpha.legend,fill_alpha.free `r .doc_vv("fill_alpha")` `r .doc_unit("fill_alpha")`
#' @param col_alpha,col_alpha.scale,col_alpha.legend,col_alpha.chart,col_alpha.free `r .doc_vv("col_alpha")` `r .doc_unit("col_alpha")`
#' @param linejoin,lineend Line join and line end. See [gpar()][grid::gpar()] for details.
#' @param plot.order Specification in which order the spatial features are drawn.
#'   See [tm_plot_order()] for details.
#' @param zindex Controls the stacking order of map layers. Should be set to a
#'   value above 400. By default, layers are stacked in call order, starting
#'   at 401. See details.
#' @param group Name of the group to which this layer belongs. This is only
#'   relevant in view mode, where layer groups can be switched (see `group.control`)
#' @param group.control In view mode, the group control determines how layer groups
#'   can be switched on and off. Options: `"radio"` for radio buttons
#'   (meaning only one group can be shown), `"check"` for check boxes
#'   (so multiple groups can be shown), and `"none"` for no control
#'   (the group cannot be (de)selected).
#' @param options options passed on to the corresponding `opt_<layer_function>` function
#' @param popup popup specification for `"view"` mode, the output of [tm_popup()].
#'   It determines the data variables shown in the popup table, the popup title,
#'   and (in the future) the popup layout. This replaces the deprecated
#'   arguments `popup.vars` and `popup.format`.
#' @param popup.vars (Deprecated.) Use `popup` with
#'   [tm_popup()] instead (via its `vars` argument). Names of data variables that
#'   are shown in the popups in `"view"` mode. Set `popup.vars` to `TRUE` to show
#'   all variables in the shape object. Set `popup.vars` to `FALSE` to disable
#'   popups. Set `popup.vars` to a character vector of variable names to show
#'   those variables in the popups. The default (`NA`) depends on whether visual
#'   variables (e.g. `fill`) are used. If so, only those are shown. If not, all
#'   variables in the shape object are shown.
#' @param popup.format (Deprecated.) Use `popup` with
#'   [tm_popup()] instead (via its `format` argument). List of formatting options
#'   for the popup values. Output of [tm_label_format()]. Only applicable for
#'   numeric data variables. If one list of formatting options is provided,
#'   it is applied to all numeric variables of `popup.vars`. Also, a (named)
#'   list of lists can be provided. In that case, each list of formatting options
#'   is applied to the named variable.
#' @param hover name of the data variable that specifies the hover labels (view mode only). Set to `FALSE` to disable hover labels. By default `FALSE`, unless `id` is specified. In that case, it is set to `id`,
#' @param id name of the data variable that specifies the indices of the spatial
#'   features. Only used for `"view"` mode.
#' @param blend Compositing operator for layer blending. Default `"over"` applies
#'   no blending. See the "Layer blending" section for the supported values.
#' @param ... to catch deprecated arguments from version < 4.0
#' @example ./examples/tm_polygons.R
#' @seealso `r .doc_see_also_layers("polygons")`
#' @export
tm_polygons = function(fill = tm_const(),
					   fill.scale = tm_scale(),
					   fill.legend = tm_legend(),
					   fill.chart = tm_chart_none(),
					   fill.free = NA,
					   col = tm_const(),
					   col.scale = tm_scale(),
					   col.legend = tm_legend(),
					   col.chart = tm_chart_none(),
					   col.free = NA,
					   lwd = tm_const(),
					   lwd.scale = tm_scale(),
					   lwd.legend = tm_legend(),
					   lwd.chart = tm_chart_none(),
					   lwd.free = NA,
					   lty = tm_const(),
					   lty.scale = tm_scale(),
					   lty.legend = tm_legend(),
					   lty.chart = tm_chart_none(),
					   lty.free = NA,
					   fill_alpha = tm_const(),
					   fill_alpha.scale = tm_scale(),
					   fill_alpha.legend = tm_legend(),
					   fill_alpha.chart = tm_chart_none(),
					   fill_alpha.free = NA,
					   col_alpha = tm_const(),
					   col_alpha.scale = tm_scale(),
					   col_alpha.legend = tm_legend(),
					   col_alpha.chart = tm_chart_none(),
					   col_alpha.free = NA,
					   linejoin = "round",
					   lineend = "round",
					   plot.order = tm_plot_order("lwd", reverse = TRUE, na.order = "bottom"),
					   zindex = NA,
					   group = NA,
					   group.control = "check",
					   popup = tm_popup(),
					   popup.vars = NA,
					   popup.format = tm_label_format(),
					   hover = NA,
					   id = "",
					   blend = "over",
					   options = opt_tm_polygons(),
					   ...) {

	args_called = names(rlang::call_match()[-1])
	args = list(...)

	layer_fun = if ("called_from" %in% names(args)) {
		args$called_from
	} else {
		"tm_polygons"
	}

	if (any(v3_only("tm_polygons") %in% names(args))) {



		v3_start_message()


		if (!("style" %in% names(args))) {
			if (!"breaks" %in% names(args) || is.null(args$breaks)) {
				style = "pretty"
			} else {
				style = "fixed"
			}
		} else {
			style = args$style
		}


		v3_list_init()
		if (length(style) > 1) {
			style = style[1]
			.TMAP$v3_list$mult = TRUE
		}



		fill.scale.args = c(list(n = v3_impute(args, "n", 5),
								 style = style,
								 style.args = v3_impute(args, "style.args", list())),
							if (style %in% c("cont", "log10")) {
								c({
									if (!is.null(args$breaks)) {
										if (length(args$breaks) > 2) {
											list(ticks = v3_impute(args, "breaks", NULL, "ticks"))
										} else {
											list(ticks = v3_impute(args, "breaks", NULL, "limits"))
										}
									} else {
										list()
									}
								},
								list(outliers.trunc = c(TRUE, FALSE)))
							} else {
								list(breaks = v3_impute(args, "breaks", NULL),
									 interval.closure = v3_impute(args, "interval.closure", "left"),
									 drop.levels = v3_impute(args, "drop.levels", FALSE))
							},
							list(midpoint = v3_impute(args, "midpoint", NULL),
								 as.count = v3_impute(args, "as.count", NA),
								 values = v3_impute(args, "palette", NA, "values"),
								 values.repeat = !v3_impute(args, "stretch.palette", TRUE, "values.repeat"),
								 values.range = v3_impute(args, "contrast", NA, "values.range"),
								 values.scale = 1,
								 value.na = v3_impute(args, "colorNA", NA, "value.na"),
								 value.null = v3_impute(args, "colorNULL", NA, "value.null"),
								 value.neutral = NA,
								 labels = v3_impute(args, "labels", NULL),
								 label.na = v3_impute(args, "textNA", "Missing", "label.na"),
								 label.null = NA,
								 label.format = v3_impute(args, "legend.format", list(), "label.format")))


		fill.scale.args$fun_pref = if (style[1] == "cat") {
			"categorical"
		} else if (style %in% c("fixed", "sd", "equal", "pretty", "quantile", "kmeans", "hclust", "bclust", "fisher", "jenks", "dpih", "headtails", "log10_pretty")) {
			"intervals"
		} else if (style == "cont") {
			"continuous"
		} else if (style == "log10") {
			"continuous_log"
		} else if (style == "order") {
			"rank"
		} else {
			stop("unknown style")
		}

		if ("style" %in% names(args)) {
			v3_tm_scale_instead_of_style(style[1], scale_fun = fill.scale.args$fun_pref, vv = "fill", layer_fun = layer_fun, arg_list = v3_list_get())
		} else {
			v3_tm_scale(scale_fun = "", vv = "fill", layer_fun = layer_fun, arg_list = v3_list_get())
		}

		fill.scale = do.call("tm_scale", args = fill.scale.args)

		if ("convert2density" %in% names(args) && args$convert2density) {
			fill.scale$convert2density = TRUE
			v3_convert2density(layer_fun)
		}

		if ("col" %in% args_called && (!layer_fun %in% c("tm_fill", "qtm"))) {
			fill = col
			col = tm_const()
			v3_message_col_fill(layer_fun = layer_fun)
		}



		if ("border.col" %in% names(args)) {
			col = args$border.col
			if (!("col" %in% args_called)) v3_message_col_fill(layer_fun = layer_fun)
		}
		if (identical(args$called_from, "tm_borders")) {
			fill = NA
		}

		if ("alpha" %in% names(args)) {
			fill_alpha = args$alpha
			v3_message_fill_alpha(layer_fun = layer_fun)
		}

		if ("border.alpha" %in% names(args)) {
			col_alpha = args$border.alpha
			v3_message_col_alpha(layer_fun = layer_fun)
		}


		v3_list_init()
		if ("legend.show" %in% names(args) && !args$legend.show) {
			v3_tm_legend_hide(layer_fun, arg = "legend.show", vv = "fill")
			fill.legend = tm_legend_hide()
		} else {
			fill.legend.args = alist(title = v3_impute(args, "title", NA),
									na.show = v3_impute(args, "showNA", NA, "na.show"),
									format = v3_impute(args, "legend.format", list(), "format"),
									orientation = ifelse(v3_impute(args, "legend.is.portrait", TRUE, "orientation"), "portrait", "landscape"),
									reverse = v3_impute(args, "legend.reverse", FALSE, "reverse"))
			fill.legend = do.call("tm_legend", fill.legend.args)
			v3_tm_legend(fun = layer_fun, vv = "fill", arg_list = v3_list_get())
		}

		if ("legend.hist" %in% names(args) && args$legend.hist) {
			fill.chart = tm_chart_histogram()
			v3_tm_chart_hist(layer_fun = layer_fun, vv = "fill", arg = "legend.hist")

			# to do: histogram title
		}

	}

	# unused arguments: typos?
	unused = setdiff(names(args), c(v3_only("tm_polygons"), "called_from"))

	if (length(unused)) {
		message_layer_unused_args(layer_fun, unused)
	}

	# make sure required options are there
	options = complete_options(options, opt_tm_polygons())

	# blend has been migrated from opt_tm_polygons() to the layer function root;
	# it is still passed on internally via mapping.args (read as a$blend).
	options$mapping.args$blend = blend

	# resolve popup specification (popup.vars/popup.format are deprecated in
	# favour of popup = tm_popup(...))
	popup = process_popup(popup = popup,
						  popup.vars = popup.vars,
						  popup.format = popup.format,
						  popup.called = "popup" %in% args_called,
						  popup.vars.called = "popup.vars" %in% args_called,
						  popup.format.called = "popup.format" %in% args_called,
						  layer_fun = layer_fun)

	tm_element_list(tm_element(
		layer = "polygons",
		trans.fun = tmapTransPolygons,
		trans.args = options$trans.args,
		trans.aes = list(),
		trans.apply_to = "this",
		mapping.aes = list(fill = tmapScale(aes = "fill",
											value = fill,
											scale = fill.scale,
											legend = fill.legend,
											chart = fill.chart,
											free = fill.free),
						   col = tmapScale(aes = "col",
						   				value = col,
						   				scale = col.scale,
						   				legend = col.legend,
						   				chart = col.chart,
						   				free = col.free),
						   lwd = tmapScale(aes = "lwd",
						   				value = lwd,
						   				scale = lwd.scale,
						   				legend = lwd.legend,
						   				chart = lwd.chart,
						   				free = lwd.free),
						   lty = tmapScale(aes = "lty",
						   				value = lty,
						   				scale = lty.scale,
						   				legend = lty.legend,
						   				chart = lty.chart,
						   				free = lty.free),
						   fill_alpha = tmapScale(aes = "fill_alpha",
						   					   value = fill_alpha,
						   					   scale = fill_alpha.scale,
						   					   legend = fill_alpha.legend,
						   					   chart = fill_alpha.chart,
						   					   free = fill_alpha.free),
						   col_alpha = tmapScale(aes = "col_alpha",
						   					  value = col_alpha,
						   					  scale = col_alpha.scale,
						   					  legend = col_alpha.legend,
						   					  chart = col_alpha.chart,
						   					  free = col_alpha.free)),

		gpar = tmapGpar(fill = "__fill",
						col = "__col",
						shape = NA,
						size = NA,
						fill_alpha = "__fill_alpha",
						col_alpha = "__col_alpha",
						pattern = "fill",
						lty = "__lty",
						lwd = "__lwd",
						linejoin = linejoin,
						lineend = lineend),
		tpar = tmapTpar(area = "AREA"),
		plot.order = plot.order,
		mapping.args = options$mapping.args,
		zindex = zindex,
		group = group,
		group.control = group.control,
		popup.vars = popup$vars,
		popup.format = popup$format,
		popup.title = popup$title,
		popup.layout = popup$layout,
		hover = hover,
		id = id,
		subclass = c("tm_aes_layer", "tm_layer")))
}

#' @rdname tm_polygons
#' @export
tm_fill = function(...) {
	args = list(...)
	# tricks to make backward comp. work
	if (!("col" %in% names(args))) {
		args["col"] = list(NULL)
	}
	args$called_from = if (names(args)[1] == "") "tm_fill" else  "tm_polygons"
	tm = do.call(tm_polygons, args)
	tm[[1]]$layer = c("fill", "polygons")
	tm
}

#' @rdname tm_polygons
#' @export
tm_borders = function(col = tm_const(), ...) {
	args = list(...)
	if (!("fill" %in% names(args))) {
		args["fill"] = list(NULL)
	}
	args$called_from = "tm_borders"
	args$popup = FALSE
	args$hover = FALSE
	tm = do.call(tm_polygons, c(list(col = col), args))
	tm[[1]]$layer = c("borders", "polygons")
	tm
}

#' @param polygons.only should only polygon geometries of the shape object (defined in [tm_shape()]) be plotted? By default `"ifany"`, which means `TRUE` in case a geometry collection is specified.
#' @rdname tm_polygons
#' @export
opt_tm_polygons = function(polygons.only = "ifany") {
	list(trans.args = list(polygons.only = polygons.only),
		 mapping.args = list())
}

#' Quick thematic map plot
#'
#' Draw a thematic map quickly. This function is a convenient wrapper of the main
#' plotting method of stacking [`tmap-element`]s. Without arguments or with a
#' search term, this functions draws an interactive map.
#'
#' The first argument is a shape object (normally specified by [tm_shape()]).
#' The next arguments, from `fill` to `raster`, are the aesthetics from the main
#' layers. The remaining arguments are related to the map layout. Any argument
#' from any main layer function, such as [tm_polygons()], can be specified (see `...`).
#' It is also possible to stack [`tmap-element`]s on a `qtm` plot. See examples.
#'
#' By default, a scale bar is shown. This option can be set with [tmap_options()]
#' (argument `qtm.scalebar`). A minimap is shown by default when `qtm` is called
#' without arguments of with a search term. This option can be set with [tmap_options()]
#' (argument `qtm.minimap`).
#'
#' ## zindex and pane names
#'
#' In view mode, each layer is rendered in a Leaflet pane named `"tmap{zindex}"`
#' (e.g., `"tmap401"`, `"tmap402"`), with base tile layers placed in the
#' standard `"tile"` pane.
#' @param shp One of:
#'  * shape object, which is an object from a class defined by the [`sf`][`sf::sf`]
#'    or [`stars`][stars::st_as_stars()] package. Objects from the packages `sp`
#'    and `raster` are also supported, but discouraged.
#'  * Not specified, i.e. `qtm()` is executed. In this case a plain interactive
#'    map is shown.
#'  * An OpenStreetMap search string, e.g. `qtm("Amsterdam")`. In this case a
#'    plain interactive map is shown positioned according to the results of the
#'    search query (from OpenStreetMap nominatim)
#' @param fill,col,size,shape,lwd,lty,fill_alpha,col_alpha Visual variables.
#' @param text,text_col,text_size Visual variables.
#' @param by data variable name by which the data is split, or a vector of two
#'   variable names to split the data by two variables (where the first is used
#'   for the rows and the second for the columns). See also [tm_facets()].
#' @param scale numeric value that serves as the global scale parameter. All font
#'   sizes, symbol sizes, border widths, and line widths are controlled by this value.
#'   The parameters `symbols.size`, `text.size`, and `lines.lwd` can be scaled
#'   separately with respectively `symbols.scale`, `text.scale`, and
#'   `lines.scale`. See also `...`.
#' @param title main title. For legend titles, use `X.legend`, where X is the
#'   layer name (see `...`).
#' @param crs Either a [`crs`][sf::st_crs()] object or a character value
#'   (`PROJ.4` character string). By default, the projection is used that is
#'   defined in the `shp` object itself.
#' @param bbox bounding box. Argument passed on to [tm_shape()]
#' @param basemaps name(s) of the provider or an URL of a tiled basemap.
#'   It is a shortcut to [tm_basemap()]. Set to `NULL` to disable basemaps.
#'   By default, it is set to the tmap option `basemaps`.
#' @param overlays name(s) of the provider or an URL of a tiled overlay map.
#'   It is a shortcut to [tm_tiles()].
#' @param zindex Controls the stacking order of map layers. Should be set to a
#'   value above 400. By default, layers are stacked in call order, starting
#'   at 401. See details.
#' @param group group
#' @param group.control group.control
#' @param style Layout options (see [tm_layout()]) that define the style.
#'   See [tmap_style()] for details.
#' @param format Deprecated, see [tm_format()] for alternatives
#' @param ... arguments associated with the visual variables are passed on
#'   to the layer functions [tm_polygons()], [tm_lines()], [tm_symbols()],
#'   and [tm_text()].
#'   For instance, `fill.scale` is the scale specifications of the fill color
#'   of polygons (see [tm_polygons()]).
#' @return A [`tmap-element`]
#' @example ./examples/qtm.R
#' @references Tennekes, M., 2018, {tmap}: Thematic Maps in {R},
#'   Journal of Statistical Software, 84(6), 1-39, \doi{10.18637/jss.v084.i06}
#' @export
qtm = function(shp = NULL,
				fill = tmap::tm_const(),
				col = tmap::tm_const(),
				size = tmap::tm_const(),
				shape = tmap::tm_const(),
				lwd = tmap::tm_const(),
				lty = tmap::tm_const(),
				fill_alpha = tmap::tm_const(),
				col_alpha = tmap::tm_const(),
				text = tmap::tm_const(),
				text_col = tmap::tm_const(),
				text_size = tmap::tm_const(),
				by = NULL,
				scale = NULL,
				title = NULL,
				crs = NULL,
				bbox = NULL,
				basemaps = NA,
				overlays = NA,
				zindex = NA,
				group = NA,
				group.control = "check",
				style = NULL,
				format = NULL,
				...) {

	args = lapply(as.list(rlang::call_match(defaults = TRUE, dots_expand = TRUE)[-1]), eval, envir = parent.frame())
	shp_name = deparse(substitute(shp))[1]
	args_called = names(rlang::call_match(dots_expand = TRUE)[-1])

	if (any(v3_only("qtm") %in% args_called)) {
		v3_start_message()
		args_new = list(args = args, called = args_called) |>
			v3_instead("symbols.size", "size", "qtm", extra_called = "shape") |>
			v3_instead("symbols.col", "size", "qtm", extra_called = "shape") |>
			v3_instead("dots.col", "fill", "qtm") |>
			v3_instead("lines.lwd", "lwd", "qtm") |>
			v3_instead("lines.col", "col", "qtm") |>
			v3_instead("raster", "col", "qtm") |>
			v3_instead_value("borders", "col", "qtm", value_old = NULL, value_new = NA) |>
			v3_instead("text.size", "text_size", "qtm") |>
			v3_instead("text.col", "text_col", "qtm") |>
			v3_instead("projection", "crs", "qtm")
		args = args_new$args
		args_called = args_new$called
	}

	o = tmap_options_mode(mode.specific = FALSE)
	show.warnings = o$show.warnings

	if (missing(shp) || is.character(shp)) {
		viewargs = args[intersect(names(args), names(formals(tm_view)))]
		if (!missing(shp)) viewargs$bbox = shp

		g = c(tm_basemap(basemaps), tm_tiles(overlays), do.call("tm_view", viewargs))
		if (o$qtm.minimap) g = c(g, tm_minimap())
		if (o$qtm.scalebar) g = c(g, tm_scalebar())
		if (o$qtm.mouse_coordinates) g = c(g, tm_mouse_coordinates())

		attr(g, "qtm_shortcut") = TRUE
		class(g) = "tmap"
		return(g)
	}


	if (!missing(shp)) {
		nms_shp = intersect(args_called, names(formals(tm_shape)))
		g = do.call(tm_shape, c(args[nms_shp], list(name = shp_name)))

		is_rst = inherits(shp, c("stars", "SpatRaster"))

		prefix = list(tm_polygons = "fill",
					  tm_lines = "lines",
					  tm_symbols = c("symbols", "dots"),
					  tm_raster = "raster")

		if (is_rst) {
			nms_rst = intersect(names(args), funs_v4$tm_raster)
			args_rst = args[nms_rst]

			if (!any(c("col", "raster") %in% args_called)) {
				args_rst$col = tm_vars()
			}

			nms_rst_v3 = names(args)[substr(names(args), 1, 7) == "raster."]
			args_rst_v3 = args[nms_rst_v3]
			names(args_rst_v3) = substr(names(args_rst_v3), 8, nchar(names(args_rst_v3)))

			g = g + do.call(tm_raster, c(args_rst, args_rst_v3))
		} else {

			for (f in c("tm_polygons", "tm_lines", "tm_symbols")) {
				nms_f = intersect(names(args), funs_v4[[f]])
				args_f = args[nms_f]

				nms_other = intersect(setdiff(names(args), c(nms_f, nms_shp, "basemaps", "overlays", "style", "format")), args_called)
				v3_nms = intersect(nms_other, unlist(lapply(prefix[[f]], function(p) paste(p, setdiff(v3_only(f), "scale"), sep = "."))))
				args_v3 = args[v3_nms]
				names(args_v3) = sub("^[^.]+[.]", "", names(args_v3))
				if (f == "tm_symbols") {
					if (!"shape" %in% args_called) args_f$shape = NULL
					if (!"col" %in% args_called) args_f$col = NULL
				}
				if (f == "tm_lines") {
					if (!"col" %in% args_called) args_f$col = NULL
				}

				args_f$called_from = "qtm"
				options = opt_tm_sf()[[c(tm_polygons = "polygons", tm_lines = "lines", tm_symbols = "points")[f]]]
				g = g + do.call(f, c(args_f, args_v3, list(options = options)))
			}
			if ("text" %in% args_called) {
				args_txt = args[substr(names(args), 1, 5) == "text_" | names(args) == "text"]
				names(args_txt) = sub("text_", "", names(args_txt))

				v3_nms = substr(names(args), 1, 5) == "text."
				args_v3 = args[v3_nms]
				names(args_v3) = sub("^[^.]+[.]", "", names(args_v3))

				args_f$called_from = "qtm"
				g = g + do.call(tm_text, c(args_txt, args_v3))
			}
		}

		nms_fct = intersect(names(args), names(formals(tm_facets)))
		if (length(nms_fct)) {
			g = g + do.call(tm_facets, args[nms_fct])
		}
	} else {
		g = NULL
	}


	if ("basemaps" %in% args_called || o$basemap.show) {
		g = g + do.call(tm_basemap, list(server = args$basemaps))
	}
	if ("tiles" %in% args_called || o$tiles.show) {
		g = g + do.call(tm_tiles, list(server = args$overlays))
	}
	if ("style" %in% args_called) {
		g = g + tm_style(args$style)
	}
	if ("title" %in% args_called) {
		g = g + tm_title(args$title)
	}

	if (o$qtm.scalebar) g = g + tm_scalebar()
	if (o$qtm.minimap) g = g + tm_minimap()
	if (o$qtm.mouse_coordinates) g = g + tm_mouse_coordinates()

	if (is.null(g)) {
		message_qtm_empty()
		invisible(NULL)
	} else {
		assign("last_map_new", rlang::call_match(), envir = .TMAP)
		attr(g, "qtm_shortcut") = FALSE
		g
	}
}

tmapGridXlab = function(facet_page, o) {
	gts = get("gts", envir = .TMAP_GRID)
	g = get("g", envir = .TMAP_GRID)

	gt = gts[[facet_page]]

	row = g$xlab_row_id
	col = g$xlab_col_ids

	gpar_text = rescale_gp(grid::gpar(cex = o$xlab.size, col = o$xlab.color, fontfamily = o$xlab.fontfamily, fontface = o$xlab.fontface, alpha = o$xlab.alpha), o$scale_down)

	grb = grid::grobTree(
		grid::textGrob(label = o$xlab.text, rot = o$xlab.rotation, gp = gpar_text)
	)

	gt = add_to_gt(gt, grb, row = row, col = col)
	gts[[facet_page]] = gt

	assign("gts", gts, envir = .TMAP_GRID)

}


tmapGridYlab = function(facet_page, o) {
	gts = get("gts", envir = .TMAP_GRID)
	g = get("g", envir = .TMAP_GRID)

	gt = gts[[facet_page]]

	row = g$ylab_row_ids
	col = g$ylab_col_id

	gpar_text = rescale_gp(grid::gpar(cex = o$ylab.size, col = o$ylab.color, fontfamily = o$ylab.fontfamily, fontface = o$ylab.fontface), o$scale_down)

	grb = grid::grobTree(
		grid::textGrob(label = o$ylab.text, rot = o$ylab.rotation, gp = gpar_text)
	)

	gt = add_to_gt(gt, grb, row = row, col = col)
	gts[[facet_page]] = gt

	assign("gts", gts, envir = .TMAP_GRID)

}

#' Deprecated: format
#'
#' In tmap < 4.0 it was possible to set shape-specific options, such as margins and legend position. However, this has become superfluous because in tmap > 4.0 legends are by default placed outside the map area. If needed, a shape-specific set of options can be stored as a style with tmap_options_save.
#'
#' @param format Name of the format
#' @param ... not used
#' @param name Name of the format
#' @keywords internal
#' @rdname tmap-deprecated
#' @export
tmap_format = function(format) {
	if (format == "World") {
		x = list(inner.margins=c(0, 0.05, 0.025, 0.01), legend.position=tm_pos_in("left", "bottom"), component.position=c("right", "bottom"), scale=.8, title.size = 1.3)
	} else if (format == "World_wide") {
		x = list(inner.margins=c(0, 0.2, 0.025, 0.01), legend.position=tm_pos_in("left", "bottom"), component.position=c("right", "bottom"), scale=.8)
	} else if (format == "NLD") {
		x = list(frame = FALSE, inner.margins = c(.02, .2, .06, .02), legend.position = tm_pos_in("left", "top"), component.position=c("left", "bottom"))
	} else if (format == "NLD_wide") {
		x = list(frame = FALSE, inner.margins = c(.02, .3, .06, .02), legend.position = tm_pos_in("left", "top"), component.position=c("left", "bottom"))
	} else {
		x = NULL
	}
	v3_tmap_format(format)
	x
}

#' @rdname tmap-deprecated
#' @export
tmap_format_add = function(..., name) {
	v3_tmap_format_add(name)
	invisible(NULL)
}


#' @rdname tmap-deprecated
#' @export
tm_format = function(format, ...) {
	if (format == "World") {
		x = tm_layout(inner.margins=c(0, 0.05, 0.025, 0.01), legend.position=tm_pos_in("left", "bottom"), component.position=c("right", "bottom"), scale=.8, title.size = 1.3)
		code = "tm_layout(inner.margins=c(0, 0.05, 0.025, 0.01), legend.position=tm_pos_in(\"left\", \"bottom\"), component.position=c(\"right\", \"bottom\"), scale=.8, title.size = 1.3)" } else if (format == "World_wide") {
			x = tm_layout(inner.margins=c(0, 0.2, 0.025, 0.01), legend.position=tm_pos_in("left", "bottom"), component.position=c("right", "bottom"), scale=.8)
			code = "tm_ayout(inner.margins=c(0, 0.2, 0.025, 0.01), legend.position=tm_pos_in(\"left\", \"bottom\"), component.position=c(\"right\", \"bottom\"), scale=.8)"
		} else if (format == "NLD") {
			x = tm_layout(frame = FALSE, inner.margins = c(.02, .2, .06, .02), legend.position = tm_pos_in("left", "top"), component.position=c("left", "bottom"))
			code = "tm_layout(frame = FALSE, inner.margins = c(.02, .2, .06, .02), legend.position = tm_pos_in(\"left\", \"top\"), component.position=c(\"left\", \"bottom\"))"
		} else if (format == "NLD_wide") {
			x = tm_layout(frame = FALSE, inner.margins = c(.02, .3, .06, .02), legend.position = tm_pos_in("left", "top"), component.position=c("left", "bottom"))
			code = "tm_layout(frame = FALSE, inner.margins = c(.02, .3, .06, .02), legend.position = tm_pos_in(\"left\", \"top\"), component.position=c(\"left\", \"bottom\"))"
		} else {
			x = NULL
			code = ""
		}
	v3_tm_format(format, code)
	x
}

text_width_npc = function(txt, space = TRUE, to_height = FALSE) {
	brks = attr(txt, "brks")
	if (is.null(brks)) {
		if (space) txt = paste(txt, " ", sep = "")

		if (to_height) {
			convertHeight(convertWidth(stringWidth(txt), "inch"), "npc", TRUE)
		} else {
			convertWidth(stringWidth(txt), "npc", TRUE)
		}

	} else {
		txt_splits = split_legend_labels(txt, brks)
		spx = if (space) convertWidth(stringWidth(" "), "npc", TRUE) * 1.5 else 0
		res = lapply(txt_splits, function(tx) convertWidth(stringWidth(tx), "npc", TRUE) + spx)

		max1 = max(sapply(res, "[", 1))
		max2 = max(sapply(res, "[", 2))
		max3 = max(sapply(res, "[", 3))
		#r3 = sapply(res, "[", 3)
		widths = max1 + max2 + max3 #r3

		attr(widths, "cw") =  do.call(rbind, res)
		widths
	}
}

text_width_inch = function(txt, space = TRUE) {
	brks = attr(txt, "brks")
	if (is.null(brks)) {
		if (space) txt = paste(txt, " ", sep = "")

		convertWidth(stringWidth(txt), "inch", TRUE)

	} else {
		txt_splits = split_legend_labels(txt, brks)
		spx = if (space) convertWidth(stringWidth(" "), "inch", TRUE) * 1.5 else 0
		res = lapply(txt_splits, function(tx) convertWidth(stringWidth(tx), "inch", TRUE) + spx)

		max1 = max(sapply(res, "[", 1))
		max2 = max(sapply(res, "[", 2))
		max3 = max(sapply(res, "[", 3))
		#r3 = sapply(res, "[", 3)
		widths = max1 + max2 + max3 #r3

		attr(widths, "cw") =  do.call(rbind, res)
		widths
	}
}

text_height_npc = function(txt, to_width = FALSE) {
	if (to_width) {
		convertWidth(convertHeight(stringHeight(txt), "inch"), "npc", TRUE)
	} else {
		convertHeight(stringHeight(txt), "npc", TRUE)
	}
}

text_height_inch = function(txt, to_width = FALSE) {
	if (to_width) {
		grid::convertWidth(grid::convertHeight(stringHeight(txt), "inch"), "inch", TRUE)
	} else {
		grid::convertHeight(stringHeight(txt), "inch", TRUE)
	}
}

split_legend_labels = function(txt, brks) {
	lapply(1L:length(txt), function(i) {
		c(substr(txt[i], 1, brks[i,1]-2),
		  substr(txt[i], brks[i,1], brks[i,2]-2),
		  substr(txt[i], brks[i,2], nchar(txt[i])))
	})
}

is.ena = function(x) {
	if (is.expression(x)) {
		rep_len(FALSE, length(x))
	} else is.na(x)
}


nonempty_text = function(txt) {
	if (is.character(txt)) {
		nzchar(txt)
	} else rep_len(TRUE, length(txt))
}

number_text_lines = function(txt) {
	if (is.character(txt)) {
		length(strsplit(txt, "\n")[[1]])
	} else 1
}

expr_to_char = function(txt) {
	if (is.character(txt)) {
		txt
	} else {
		as.character(txt)
	}
}


encode_expr = function(txt) {
	if (is.character(txt)) {
		txt
	} else {
		paste0("__expr__", as.character(txt))
	}
}

decode_expr = function(txt) {
	if (!is.na(txt) && substr(txt, 1, 8) == "__expr__") {
		txt = substr(txt, 9, nchar(txt))
		parse(text = txt)[[1]]
	} else {
		txt
	}
}

# Is a label expression-encoded (i.e. produced by encode_expr)?
is_encoded_expr = function(txt) {
	is.character(txt) & !is.na(txt) & substr(txt, 1, 8) == "__expr__"
}

# Decode a character vector of (possibly expression-encoded) labels into a form
# accepted by grid::textGrob(). When none of the labels are expressions the
# input character vector is returned unchanged (so plain text keeps its exact
# behaviour, including multi-line "\n" labels). When at least one label is an
# expression, an expression vector (plotmath) is returned; plain strings in the
# mix become plotmath string constants, which render as literal text.
decode_expr_vec = function(txt) {
	if (!is.character(txt)) return(txt)
	if (!any(is_encoded_expr(txt))) return(txt)
	as.expression(lapply(txt, decode_expr))
}

# Decode expression-encoded labels to plain character (deparsed), for backends
# that cannot render plotmath (e.g. Leaflet/HTML). Consistent with how titles
# and credits fall back to expr_to_char() in view mode.
decode_expr_chr = function(txt) {
	if (!is.character(txt)) return(expr_to_char(txt))
	enc = is_encoded_expr(txt)
	txt[enc] = substr(txt[enc], 9, nchar(txt[enc]))
	txt
}


is_num_string = function(x) {
	suppressWarnings(!is.na(as.numeric(x)))
}

#' @param fill,col,shape,size,fill_alpha,col_alpha,lty,lwd,linejoin,lineend visual variables
#' @param ... args
#' @export
#' @rdname tmap_internal
tmapGpar = function(fill = NULL,
					col = NULL,
					shape = NULL,
					size = NULL,
					fill_alpha = NULL,
					col_alpha = NULL,
					#pattern = NULL,
					lty = NULL,
					lwd = NULL,
					linejoin = NULL,
					lineend = NULL,
					...) {
	args = c(as.list(environment()), list(...))
	structure(args, class = "tmapGpar")
}

#' @export
#' @rdname tmap_internal
tmapTpar = function(...) {
	args = c(as.list(environment()), list(...))
	structure(args, class = "tmapTpar")
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tm_xylab.R
\name{tm_xlab}
\alias{tm_xlab}
\alias{tm_ylab}
\title{Map: x and y labels}
\usage{
tm_xlab(text, size, color, rotation, space, fontface, fontfamily, alpha, side)

tm_ylab(text, size, color, rotation, space, fontface, fontfamily, alpha, side)
}
\arguments{
\item{text}{text of the title}

\item{size}{font size of the title}

\item{color}{color}

\item{rotation}{rotation in degrees}

\item{space}{space between label and map in number of line heights}

\item{fontface}{font face}

\item{fontfamily}{font family}

\item{alpha}{alpha transparency of the text}

\item{side}{side: \code{"top"} or \code{"bottom"} for \code{tm_xlab} and \code{"left"} or \code{"right"} for \code{tm_ylab}}
}
\description{
The x and y labels for maps
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/pkg.R
\docType{package}
\name{tmap-package}
\alias{tmap}
\alias{tmap-package}
\title{Thematic Map Visualization}
\description{
Thematic maps are geographical maps in which spatial data distributions are visualized.
This package offers a flexible, layer-based, and easy to use approach to create
thematic maps, such as choropleths and bubble maps. It is based on the grammar
of graphics, and resembles the syntax of ggplot2.
}
\references{
Tennekes, M., 2018, {tmap}: Thematic Maps in {R}, Journal of Statistical Software, 84(6), 1-39, \doi{10.18637/jss.v084.i06}
}
\seealso{
\href{https://r-tmap.github.io/tmap/index.html}{Main documentation}
}
\author{
Martijn Tennekes \email{mtennekes@gmail.com}
}
\concept{GIS}
\concept{bubble map}
\concept{choropleth}
\concept{statistical maps}
\concept{thematic maps}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/basemap_equivalents.R
\name{tmapSubmitBasemapEquivalents}
\alias{tmapSubmitBasemapEquivalents}
\title{Register cross-mode basemap equivalences}
\usage{
tmapSubmitBasemapEquivalents(equivalents, replace = FALSE)
}
\arguments{
\item{equivalents}{a named list of concept entries. Each entry is itself a
named list with mode names (e.g. \code{plot}, \code{view}, \code{maplibre}, \code{mapbox}) as
names and provider name(s) as values.}

\item{replace}{if \code{TRUE}, replace the whole registry; otherwise the entries
are merged by concept name (existing names are overwritten), so repeated
submissions are idempotent and multiple extensions compose.}
}
\description{
Used by tmap mode extensions to declare which basemap providers are visually
equivalent across modes. This lets a basemap be approximated when a map is
reproduced in a different mode, instead of falling back to the mode default.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tm_components.R
\name{tm_components}
\alias{tm_components}
\title{Group components}
\usage{
tm_components(
  group_id = "",
  position,
  stack,
  frame_combine,
  equalize,
  resize_as_group,
  stack_margin,
  offset,
  frame,
  frame.color,
  frame.alpha,
  frame.lwd,
  frame.r,
  bg,
  bg.color,
  bg.alpha
)
}
\arguments{
\item{group_id}{id of the component group. By default set to \code{""}, which will apply to all components. There are two other options. 1) To use the same (self-chosen) name that corresponds to the \code{group_id} argument of a component function, such as \code{\link[=tm_legend]{tm_legend()}} and \code{\link[=tm_title]{tm_title()}}. 2) To specify one (or more) component function names, e.g. \code{"tm_legend"} or \code{c("tm_scalebar", "tm_compass")}.}

\item{position}{The position specification of the components in this group: an object created with \code{tm_pos_in()} or \code{tm_pos_out()}. Or, as a shortcut, a vector of two values, specifying the x and y coordinates. The first is \code{"left"}, \code{"center"} or \code{"right"} (or upper case, meaning tighter to the map frame), the second \code{"top"}, \code{"center"} or \code{"bottom"}. Numeric values are also supported, where 0, 0 means left bottom and 1, 1 right top. See also \href{https://r-tmap.github.io/tmap/articles/adv_positions}{vignette about positioning}.}

\item{stack}{stacking \code{"horizontal"} or \code{"vertical"}}

\item{frame_combine}{put frame around all components that are drawn on the same location. Whether a frame is drawn is still decided by the \code{frame} argument of the  'main' (first) component.}

\item{equalize}{in case \code{frame_combine} is \code{FALSE}, should the separate frames be equalized, i.e. have the same width (when stacked vertically) or height (when stacked horizontally)?}

\item{resize_as_group}{in case a component if rescaled because of the limited space, rescale the other components proportionally?}

\item{stack_margin}{Margin between components}

\item{offset}{Offset margin between frame and the components block}

\item{frame}{Should a frame be drawn? By default \code{TRUE} for legends, charts and insets, and \code{FALSE} otherwise.}

\item{frame.color}{frame color}

\item{frame.alpha}{frame alpha transparancy}

\item{frame.lwd}{frame line width}

\item{frame.r}{Radius of the rounded frame corners. 0 means no rounding.}

\item{bg}{Background color the components block. Is usually set in each component function, but if specified here, it will overwrite them.}

\item{bg.color}{Background color the components block. Is usually set in each component function, but if specified here, it will overwrite them.}

\item{bg.alpha}{Background alpha transparency of the components block. Is usually set in each component function, but if specified here, it will overwrite them.}
}
\value{
A \code{\link{tmap-element}}
}
\description{
Group components
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tm_components.R
\name{tm_components}
\alias{tm_components}
\title{Group components}
\usage{
tm_components(
  group_id = "",
  position,
  stack,
  frame_combine,
  equalize,
  resize_as_group,
  stack_margin,
  offset,
  frame,
  frame.color,
  frame.alpha,
  frame.lwd,
  frame.r,
  bg,
  bg.color,
  bg.alpha
)
}
\arguments{
\item{group_id}{id of the component group. By default set to \code{""}, which will apply to all components. There are two other options. 1) To use the same (self-chosen) name that corresponds to the \code{group_id} argument of a component function, such as \code{\link[=tm_legend]{tm_legend()}} and \code{\link[=tm_title]{tm_title()}}. 2) To specify one (or more) component function names, e.g. \code{"tm_legend"} or \code{c("tm_scalebar", "tm_compass")}.}

\item{position}{The position specification of the components in this group: an object created with \code{tm_pos_in()} or \code{tm_pos_out()}. Or, as a shortcut, a vector of two values, specifying the x and y coordinates. The first is \code{"left"}, \code{"center"} or \code{"right"} (or upper case, meaning tighter to the map frame), the second \code{"top"}, \code{"center"} or \code{"bottom"}. Numeric values are also supported, where 0, 0 means left bottom and 1, 1 right top. See also vignette: \href{https://r-tmap.github.io/tmap/articles/adv_positions}{Positioning of components}.}

\item{stack}{stacking \code{"horizontal"} or \code{"vertical"}}

\item{frame_combine}{put frame around all components that are drawn on the same location. Whether a frame is drawn is still decided by the \code{frame} argument of the  'main' (first) component.}

\item{equalize}{in case \code{frame_combine} is \code{FALSE}, should the separate frames be equalized, i.e. have the same width (when stacked vertically) or height (when stacked horizontally)?}

\item{resize_as_group}{in case a component if rescaled because of the limited space, rescale the other components proportionally?}

\item{stack_margin}{Margin between components. One or two values. If two values are provided, the first is the horizontal stacking margin and the second the vertical. The unit is text line heights.}

\item{offset}{Offset margin between frame and the components block. One or two values. If two values are provided, the first is the horizontal offset and the second the vertical. The unit is text line heights.}

\item{frame}{Should a frame be drawn? By default \code{TRUE} for legends, charts and insets, and \code{FALSE} otherwise.}

\item{frame.color}{frame color}

\item{frame.alpha}{frame alpha transparancy}

\item{frame.lwd}{frame line width}

\item{frame.r}{Radius of the rounded frame corners. 0 means no rounding.}

\item{bg}{Background color the components block. Is usually set in each component function, but if specified here, it will overwrite them.}

\item{bg.color}{Background color the components block. Is usually set in each component function, but if specified here, it will overwrite them.}

\item{bg.alpha}{Background alpha transparency of the components block. Is usually set in each component function, but if specified here, it will overwrite them.}
}
\value{
A \code{\link{tmap-element}}
}
\description{
Group components
}
\seealso{
\href{https://r-tmap.github.io/tmap/articles/basics_components}{Components}, \href{https://r-tmap.github.io/tmap/articles/adv_positions}{Positioning of components}, \href{https://r-tmap.github.io/tmap/articles/adv_comp_group}{Grouping of components}, \href{https://r-tmap.github.io/tmap/articles/examples_choro_NLD}{Choropleth (Netherlands)}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tmapGridAux.R, R/tmapGridAux_grid.R,
%   R/tmapGridAux_tiles.R, R/tmapGridComp_leg_landscape.R,
%   R/tmapGridDataPlot.R, R/tmapGridDataPlot_lines.R,
%   R/tmapGridDataPlot_polygons.R, R/tmapGridDataPlot_raster.R,
%   R/tmapGridDataPlot_symbols.R, R/tmapGridDataPlot_text.R, R/tmapLeafletAux.R,
%   R/tmapLeafletComp_leg_portrait.R, R/tmapLeafletDataPlot_lines.R,
%   R/tmapLeafletDataPlot_polygons.R, R/tmapLeafletDataPlot_raster.R,
%   R/tmapLeafletDataPlot_symbols.R, R/tmapLeafletDataPlot_text.R,
%   R/tmapLeaflet_layers.R
\name{tmapGridAuxPrepare}
\alias{tmapGridAuxPrepare}
\alias{tmapGridAuxPrepare.default}
\alias{tmapGridAuxPlot}
\alias{tmapGridAuxPlot.default}
\alias{tmapGridAuxPrepare.tm_aux_graticules}
\alias{tmapGridAuxPrepare.tm_aux_grid}
\alias{tmapGridAuxPlot.tm_aux_basemap}
\alias{tmapGridAuxPlot.tm_aux_tiles}
\alias{tmapGridAuxPlot.tm_aux_graticules}
\alias{tmapGridAuxPlot.tm_aux_grid}
\alias{tmapGridAuxPrepare.tm_aux_basemap}
\alias{tmapGridAuxPrepare.tm_aux_tiles}
\alias{tmapGridCompPrepare}
\alias{tmapGridCompHeight}
\alias{tmapGridCompWidth}
\alias{tmapGridCompPlot}
\alias{tmapGridCompPrepare.default}
\alias{tmapGridDataPlot}
\alias{tmapGridDataPlot.default}
\alias{tmapGridDataPlot.tm_data_lines}
\alias{tmapGridDataPlot.tm_data_iso}
\alias{tmapGridDataPlot.tm_data_polygons}
\alias{tmapGridDataPlot.tm_data_fill}
\alias{tmapGridDataPlot.tm_data_borders}
\alias{tmapGridDataPlot.tm_data_raster}
\alias{tmapGridDataPlot.tm_data_symbols}
\alias{tmapGridDataPlot.tm_data_dots}
\alias{tmapGridDataPlot.tm_data_bubbles}
\alias{tmapGridDataPlot.tm_data_squares}
\alias{tmapGridDataPlot.tm_data_markers}
\alias{tmapGridDataPlot.tm_data_text}
\alias{tmapGridDataPlot.tm_data_labels}
\alias{tmapGridDataPlot.tm_data_labels_highlighted}
\alias{tmapLeafletAuxPrepare}
\alias{tmapLeafletAuxPlot}
\alias{tmapLeafletAuxPrepare.tm_aux_basemap}
\alias{tmapLeafletAuxPrepare.tm_aux_tiles}
\alias{tmapLeafletAuxPlot.tm_aux_basemap}
\alias{tmapLeafletAuxPlot.tm_aux_tiles}
\alias{tmapLeafletAuxPrepare.tm_aux_graticules}
\alias{tmapLeafletAuxPrepare.tm_aux_grid}
\alias{tmapLeafletAuxPlot.tm_aux_graticules}
\alias{tmapLeafletAuxPlot.tm_aux_grid}
\alias{tmapLeafletCompPrepare}
\alias{tmapLeafletCompPrepare.tm_chart_none}
\alias{tmapLeafletCompPrepare.default}
\alias{tmapLeafletDataPlot.tm_data_lines}
\alias{tmapLeafletDataPlot.tm_data_iso}
\alias{tmapLeafletDataPlot.tm_data_polygons}
\alias{tmapLeafletDataPlot.tm_data_fill}
\alias{tmapLeafletDataPlot.tm_data_borders}
\alias{tmapLeafletDataPlot.tm_data_raster}
\alias{tmapLeafletDataPlot.tm_data_symbols}
\alias{tmapLeafletDataPlot.tm_data_dots}
\alias{tmapLeafletDataPlot.tm_data_bubbles}
\alias{tmapLeafletDataPlot.tm_data_squares}
\alias{tmapLeafletDataPlot.tm_data_markers}
\alias{tmapLeafletDataPlot.tm_data_text}
\alias{tmapLeafletDataPlot.tm_data_labels}
\alias{tmapLeafletDataPlot.tm_data_labels_highlighted}
\alias{tmapLeafletDataPlot}
\alias{tmapLeafletDataPlot.default}
\title{Internal methods for map components in plot and view mode}
\usage{
tmapGridAuxPrepare(a, bs, id, o)

\method{tmapGridAuxPrepare}{default}(a, bs, id, o)

tmapGridAuxPlot(
  a,
  bi,
  bbx,
  facet_row,
  facet_col,
  facet_page,
  id,
  pane,
  group,
  o
)

\method{tmapGridAuxPlot}{default}(
  a,
  bi,
  bbx,
  facet_row,
  facet_col,
  facet_page,
  id,
  pane,
  group,
  o
)

\method{tmapGridAuxPrepare}{tm_aux_graticules}(a, bs, id, o)

\method{tmapGridAuxPrepare}{tm_aux_grid}(a, bs, id, o)

\method{tmapGridAuxPlot}{tm_aux_basemap}(
  a,
  bi,
  bbx,
  facet_row,
  facet_col,
  facet_page,
  id,
  pane,
  group,
  o
)

\method{tmapGridAuxPlot}{tm_aux_tiles}(
  a,
  bi,
  bbx,
  facet_row,
  facet_col,
  facet_page,
  id,
  pane,
  group,
  o
)

\method{tmapGridAuxPlot}{tm_aux_graticules}(
  a,
  bi,
  bbx,
  facet_row,
  facet_col,
  facet_page,
  id,
  pane,
  group,
  o
)

\method{tmapGridAuxPlot}{tm_aux_grid}(
  a,
  bi,
  bbx,
  facet_row,
  facet_col,
  facet_page,
  id,
  pane,
  group,
  o
)

\method{tmapGridAuxPrepare}{tm_aux_basemap}(a, bs, id, o)

\method{tmapGridAuxPrepare}{tm_aux_tiles}(a, bs, id, o)

tmapGridCompPrepare(comp, o)

tmapGridCompHeight(comp, o)

tmapGridCompWidth(comp, o)

tmapGridCompPlot(comp, o, fH, fW)

\method{tmapGridCompPrepare}{default}(comp, o)

tmapGridDataPlot(
  a,
  shpTM,
  dt,
  gp,
  bbx,
  facet_row,
  facet_col,
  facet_page,
  id,
  pane,
  group,
  glid,
  o,
  ...
)

\method{tmapGridDataPlot}{default}(
  a,
  shpTM,
  dt,
  gp,
  bbx,
  facet_row,
  facet_col,
  facet_page,
  id,
  pane,
  group,
  glid,
  o,
  ...
)

\method{tmapGridDataPlot}{tm_data_lines}(
  a,
  shpTM,
  dt,
  gp,
  bbx,
  facet_row,
  facet_col,
  facet_page,
  id,
  pane,
  group,
  glid,
  o,
  ...
)

\method{tmapGridDataPlot}{tm_data_iso}(
  a,
  shpTM,
  dt,
  gp,
  bbx,
  facet_row,
  facet_col,
  facet_page,
  id,
  pane,
  group,
  glid,
  o,
  ...
)

\method{tmapGridDataPlot}{tm_data_polygons}(
  a,
  shpTM,
  dt,
  gp,
  bbx,
  facet_row,
  facet_col,
  facet_page,
  id,
  pane,
  group,
  glid,
  o,
  ...
)

\method{tmapGridDataPlot}{tm_data_fill}(
  a,
  shpTM,
  dt,
  gp,
  bbx,
  facet_row,
  facet_col,
  facet_page,
  id,
  pane,
  group,
  glid,
  o,
  ...
)

\method{tmapGridDataPlot}{tm_data_borders}(
  a,
  shpTM,
  dt,
  gp,
  bbx,
  facet_row,
  facet_col,
  facet_page,
  id,
  pane,
  group,
  glid,
  o,
  ...
)

\method{tmapGridDataPlot}{tm_data_raster}(
  a,
  shpTM,
  dt,
  gp,
  bbx,
  facet_row,
  facet_col,
  facet_page,
  id,
  pane,
  group,
  glid,
  o,
  ...
)

\method{tmapGridDataPlot}{tm_data_symbols}(
  a,
  shpTM,
  dt,
  gp,
  bbx,
  facet_row,
  facet_col,
  facet_page,
  id,
  pane,
  group,
  glid,
  o,
  ...
)

\method{tmapGridDataPlot}{tm_data_dots}(
  a,
  shpTM,
  dt,
  gp,
  bbx,
  facet_row,
  facet_col,
  facet_page,
  id,
  pane,
  group,
  glid,
  o,
  ...
)

\method{tmapGridDataPlot}{tm_data_bubbles}(
  a,
  shpTM,
  dt,
  gp,
  bbx,
  facet_row,
  facet_col,
  facet_page,
  id,
  pane,
  group,
  glid,
  o,
  ...
)

\method{tmapGridDataPlot}{tm_data_squares}(
  a,
  shpTM,
  dt,
  gp,
  bbx,
  facet_row,
  facet_col,
  facet_page,
  id,
  pane,
  group,
  glid,
  o,
  ...
)

\method{tmapGridDataPlot}{tm_data_markers}(
  a,
  shpTM,
  dt,
  gp,
  bbx,
  facet_row,
  facet_col,
  facet_page,
  id,
  pane,
  group,
  glid,
  o,
  ...
)

\method{tmapGridDataPlot}{tm_data_text}(
  a,
  shpTM,
  dt,
  gp,
  bbx,
  facet_row,
  facet_col,
  facet_page,
  id,
  pane,
  group,
  glid,
  o,
  ...
)

\method{tmapGridDataPlot}{tm_data_labels}(
  a,
  shpTM,
  dt,
  gp,
  bbx,
  facet_row,
  facet_col,
  facet_page,
  id,
  pane,
  group,
  glid,
  o,
  ...
)

\method{tmapGridDataPlot}{tm_data_labels_highlighted}(
  a,
  shpTM,
  dt,
  gp,
  bbx,
  facet_row,
  facet_col,
  facet_page,
  id,
  pane,
  group,
  glid,
  o,
  ...
)

tmapLeafletAuxPrepare(a, bs, id, o)

tmapLeafletAuxPlot(
  a,
  bi,
  bbx,
  facet_row,
  facet_col,
  facet_page,
  id,
  pane,
  group,
  o
)

\method{tmapLeafletAuxPrepare}{tm_aux_basemap}(a, bs, id, o)

\method{tmapLeafletAuxPrepare}{tm_aux_tiles}(a, bs, id, o)

\method{tmapLeafletAuxPlot}{tm_aux_basemap}(
  a,
  bi,
  bbx,
  facet_row,
  facet_col,
  facet_page,
  id,
  pane,
  group,
  o
)

\method{tmapLeafletAuxPlot}{tm_aux_tiles}(
  a,
  bi,
  bbx,
  facet_row,
  facet_col,
  facet_page,
  id,
  pane,
  group,
  o
)

\method{tmapLeafletAuxPrepare}{tm_aux_graticules}(a, bs, id, o)

\method{tmapLeafletAuxPrepare}{tm_aux_grid}(a, bs, id, o)

\method{tmapLeafletAuxPlot}{tm_aux_graticules}(
  a,
  bi,
  bbx,
  facet_row,
  facet_col,
  facet_page,
  id,
  pane,
  group,
  o
)

\method{tmapLeafletAuxPlot}{tm_aux_grid}(
  a,
  bi,
  bbx,
  facet_row,
  facet_col,
  facet_page,
  id,
  pane,
  group,
  o
)

tmapLeafletCompPrepare(comp, o)

\method{tmapLeafletCompPrepare}{tm_chart_none}(comp, o)

\method{tmapLeafletCompPrepare}{default}(comp, o)

\method{tmapLeafletDataPlot}{tm_data_lines}(
  a,
  shpTM,
  dt,
  pdt,
  popup.format,
  hdt,
  idt,
  gp,
  bbx,
  facet_row,
  facet_col,
  facet_page,
  id,
  pane,
  group,
  glid,
  o,
  ...
)

\method{tmapLeafletDataPlot}{tm_data_iso}(
  a,
  shpTM,
  dt,
  pdt,
  popup.format,
  hdt,
  idt,
  gp,
  bbx,
  facet_row,
  facet_col,
  facet_page,
  id,
  pane,
  group,
  glid,
  o,
  ...
)

\method{tmapLeafletDataPlot}{tm_data_polygons}(
  a,
  shpTM,
  dt,
  pdt,
  popup.format,
  hdt,
  idt,
  gp,
  bbx,
  facet_row,
  facet_col,
  facet_page,
  id,
  pane,
  group,
  glid,
  o,
  ...
)

\method{tmapLeafletDataPlot}{tm_data_fill}(
  a,
  shpTM,
  dt,
  pdt,
  popup.format,
  hdt,
  idt,
  gp,
  bbx,
  facet_row,
  facet_col,
  facet_page,
  id,
  pane,
  group,
  glid,
  o,
  ...
)

\method{tmapLeafletDataPlot}{tm_data_borders}(
  a,
  shpTM,
  dt,
  pdt,
  popup.format,
  hdt,
  idt,
  gp,
  bbx,
  facet_row,
  facet_col,
  facet_page,
  id,
  pane,
  group,
  glid,
  o,
  ...
)

\method{tmapLeafletDataPlot}{tm_data_raster}(
  a,
  shpTM,
  dt,
  pdt,
  popup.format,
  hdt,
  idt,
  gp,
  bbx,
  facet_row,
  facet_col,
  facet_page,
  id,
  pane,
  group,
  glid,
  o,
  ...
)

\method{tmapLeafletDataPlot}{tm_data_symbols}(
  a,
  shpTM,
  dt,
  pdt,
  popup.format,
  hdt,
  idt,
  gp,
  bbx,
  facet_row,
  facet_col,
  facet_page,
  id,
  pane,
  group,
  glid,
  o,
  ...
)

\method{tmapLeafletDataPlot}{tm_data_dots}(
  a,
  shpTM,
  dt,
  pdt,
  popup.format,
  hdt,
  idt,
  gp,
  bbx,
  facet_row,
  facet_col,
  facet_page,
  id,
  pane,
  group,
  glid,
  o,
  ...
)

\method{tmapLeafletDataPlot}{tm_data_bubbles}(
  a,
  shpTM,
  dt,
  pdt,
  popup.format,
  hdt,
  idt,
  gp,
  bbx,
  facet_row,
  facet_col,
  facet_page,
  id,
  pane,
  group,
  glid,
  o,
  ...
)

\method{tmapLeafletDataPlot}{tm_data_squares}(
  a,
  shpTM,
  dt,
  pdt,
  popup.format,
  hdt,
  idt,
  gp,
  bbx,
  facet_row,
  facet_col,
  facet_page,
  id,
  pane,
  group,
  glid,
  o,
  ...
)

\method{tmapLeafletDataPlot}{tm_data_markers}(
  a,
  shpTM,
  dt,
  pdt,
  popup.format,
  hdt,
  idt,
  gp,
  bbx,
  facet_row,
  facet_col,
  facet_page,
  id,
  pane,
  group,
  glid,
  o,
  ...
)

\method{tmapLeafletDataPlot}{tm_data_text}(
  a,
  shpTM,
  dt,
  pdt,
  popup.format,
  hdt,
  idt,
  gp,
  bbx,
  facet_row,
  facet_col,
  facet_page,
  id,
  pane,
  group,
  glid,
  o,
  ...
)

\method{tmapLeafletDataPlot}{tm_data_labels}(
  a,
  shpTM,
  dt,
  pdt,
  popup.format,
  hdt,
  idt,
  gp,
  bbx,
  facet_row,
  facet_col,
  facet_page,
  id,
  pane,
  group,
  glid,
  o,
  ...
)

\method{tmapLeafletDataPlot}{tm_data_labels_highlighted}(
  a,
  shpTM,
  dt,
  pdt,
  popup.format,
  hdt,
  idt,
  gp,
  bbx,
  facet_row,
  facet_col,
  facet_page,
  id,
  pane,
  group,
  glid,
  o,
  ...
)

tmapLeafletDataPlot(
  a,
  shpTM,
  dt,
  pdt,
  popup.format,
  hdt,
  idt,
  gp,
  bbx,
  facet_row,
  facet_col,
  facet_page,
  id,
  pane,
  group,
  glid,
  o,
  ...
)

\method{tmapLeafletDataPlot}{default}(
  a,
  shpTM,
  dt,
  pdt,
  popup.format,
  hdt,
  idt,
  gp,
  bbx,
  facet_row,
  facet_col,
  facet_page,
  id,
  pane,
  group,
  glid,
  o,
  ...
)
}
\arguments{
\item{o}{options}

\item{comp}{component}
}
\description{
Internal methods for map components
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tmapGridAux.R, R/tmapGridAux_grid.R,
%   R/tmapGridAux_tiles.R, R/tmapGridComp_leg_landscape.R,
%   R/tmapGridDataPlot.R, R/tmapGridDataPlot_circles.R,
%   R/tmapGridDataPlot_lines.R, R/tmapGridDataPlot_polygons.R,
%   R/tmapGridDataPlot_raster.R, R/tmapGridDataPlot_symbols.R,
%   R/tmapGridDataPlot_text.R, R/tmapLeafletAux.R,
%   R/tmapLeafletComp_leg_portrait.R, R/tmapLeafletDataPlot_circles.R,
%   R/tmapLeafletDataPlot_lines.R, R/tmapLeafletDataPlot_polygons.R,
%   R/tmapLeafletDataPlot_raster.R, R/tmapLeafletDataPlot_symbols.R,
%   R/tmapLeafletDataPlot_text.R, R/tmapLeaflet_layers.R
\name{tmapGridAuxPrepare}
\alias{tmapGridAuxPrepare}
\alias{tmapGridAuxPrepare.default}
\alias{tmapGridAuxPlot}
\alias{tmapGridAuxPlot.default}
\alias{tmapGridAuxPrepare.tm_aux_graticules}
\alias{tmapGridAuxPrepare.tm_aux_grid}
\alias{tmapGridAuxPlot.tm_aux_basemap}
\alias{tmapGridAuxPlot.tm_aux_tiles}
\alias{tmapGridAuxPlot.tm_aux_graticules}
\alias{tmapGridAuxPlot.tm_aux_grid}
\alias{tmapGridAuxPrepare.tm_aux_basemap}
\alias{tmapGridAuxPrepare.tm_aux_tiles}
\alias{tmapGridCompPrepare}
\alias{tmapGridCompHeight}
\alias{tmapGridCompWidth}
\alias{tmapGridCompPlot}
\alias{tmapGridCompPrepare.default}
\alias{tmapGridDataPlot}
\alias{tmapGridDataPlot.default}
\alias{tmapGridDataPlot.tm_data_circles}
\alias{tmapGridDataPlot.tm_data_lines}
\alias{tmapGridDataPlot.tm_data_iso}
\alias{tmapGridDataPlot.tm_data_polygons}
\alias{tmapGridDataPlot.tm_data_fill}
\alias{tmapGridDataPlot.tm_data_borders}
\alias{tmapGridDataPlot.tm_data_raster}
\alias{tmapGridDataPlot.tm_data_symbols}
\alias{tmapGridDataPlot.tm_data_dots}
\alias{tmapGridDataPlot.tm_data_bubbles}
\alias{tmapGridDataPlot.tm_data_squares}
\alias{tmapGridDataPlot.tm_data_markers}
\alias{tmapGridDataPlot.tm_data_text}
\alias{tmapGridDataPlot.tm_data_labels}
\alias{tmapGridDataPlot.tm_data_labels_highlighted}
\alias{tmapLeafletAuxPrepare}
\alias{tmapLeafletAuxPlot}
\alias{tmapLeafletAuxPrepare.tm_aux_basemap}
\alias{tmapLeafletAuxPrepare.tm_aux_tiles}
\alias{tmapLeafletAuxPlot.tm_aux_basemap}
\alias{tmapLeafletAuxPlot.tm_aux_tiles}
\alias{tmapLeafletAuxPrepare.tm_aux_graticules}
\alias{tmapLeafletAuxPrepare.tm_aux_grid}
\alias{tmapLeafletAuxPlot.tm_aux_graticules}
\alias{tmapLeafletAuxPlot.tm_aux_grid}
\alias{tmapLeafletCompPrepare}
\alias{tmapLeafletCompPrepare.tm_chart_none}
\alias{tmapLeafletCompPrepare.default}
\alias{tmapLeafletDataPlot.tm_data_circles}
\alias{tmapLeafletDataPlot.tm_data_lines}
\alias{tmapLeafletDataPlot.tm_data_iso}
\alias{tmapLeafletDataPlot.tm_data_polygons}
\alias{tmapLeafletDataPlot.tm_data_fill}
\alias{tmapLeafletDataPlot.tm_data_borders}
\alias{tmapLeafletDataPlot.tm_data_raster}
\alias{tmapLeafletDataPlot.tm_data_symbols}
\alias{tmapLeafletDataPlot.tm_data_dots}
\alias{tmapLeafletDataPlot.tm_data_bubbles}
\alias{tmapLeafletDataPlot.tm_data_squares}
\alias{tmapLeafletDataPlot.tm_data_markers}
\alias{tmapLeafletDataPlot.tm_data_text}
\alias{tmapLeafletDataPlot.tm_data_labels}
\alias{tmapLeafletDataPlot.tm_data_labels_highlighted}
\alias{tmapLeafletDataPlot}
\alias{tmapLeafletDataPlot.default}
\title{Internal methods for map components in plot and view mode}
\usage{
tmapGridAuxPrepare(a, bs, id, o)

\method{tmapGridAuxPrepare}{default}(a, bs, id, o)

tmapGridAuxPlot(
  a,
  bi,
  bbx,
  facet_row,
  facet_col,
  facet_page,
  id,
  pane,
  group,
  o
)

\method{tmapGridAuxPlot}{default}(
  a,
  bi,
  bbx,
  facet_row,
  facet_col,
  facet_page,
  id,
  pane,
  group,
  o
)

\method{tmapGridAuxPrepare}{tm_aux_graticules}(a, bs, id, o)

\method{tmapGridAuxPrepare}{tm_aux_grid}(a, bs, id, o)

\method{tmapGridAuxPlot}{tm_aux_basemap}(
  a,
  bi,
  bbx,
  facet_row,
  facet_col,
  facet_page,
  id,
  pane,
  group,
  o
)

\method{tmapGridAuxPlot}{tm_aux_tiles}(
  a,
  bi,
  bbx,
  facet_row,
  facet_col,
  facet_page,
  id,
  pane,
  group,
  o
)

\method{tmapGridAuxPlot}{tm_aux_graticules}(
  a,
  bi,
  bbx,
  facet_row,
  facet_col,
  facet_page,
  id,
  pane,
  group,
  o
)

\method{tmapGridAuxPlot}{tm_aux_grid}(
  a,
  bi,
  bbx,
  facet_row,
  facet_col,
  facet_page,
  id,
  pane,
  group,
  o
)

\method{tmapGridAuxPrepare}{tm_aux_basemap}(a, bs, id, o)

\method{tmapGridAuxPrepare}{tm_aux_tiles}(a, bs, id, o)

tmapGridCompPrepare(comp, o)

tmapGridCompHeight(comp, o)

tmapGridCompWidth(comp, o)

tmapGridCompPlot(comp, o, fH, fW)

\method{tmapGridCompPrepare}{default}(comp, o)

tmapGridDataPlot(
  a,
  shpTM,
  dt,
  gp,
  bbx,
  facet_row,
  facet_col,
  facet_page,
  id,
  pane,
  group,
  glid,
  o,
  ...
)

\method{tmapGridDataPlot}{default}(
  a,
  shpTM,
  dt,
  gp,
  bbx,
  facet_row,
  facet_col,
  facet_page,
  id,
  pane,
  group,
  glid,
  o,
  ...
)

\method{tmapGridDataPlot}{tm_data_circles}(
  a,
  shpTM,
  dt,
  gp,
  bbx,
  facet_row,
  facet_col,
  facet_page,
  id,
  pane,
  group,
  glid,
  o,
  ...
)

\method{tmapGridDataPlot}{tm_data_lines}(
  a,
  shpTM,
  dt,
  gp,
  bbx,
  facet_row,
  facet_col,
  facet_page,
  id,
  pane,
  group,
  glid,
  o,
  ...
)

\method{tmapGridDataPlot}{tm_data_iso}(
  a,
  shpTM,
  dt,
  gp,
  bbx,
  facet_row,
  facet_col,
  facet_page,
  id,
  pane,
  group,
  glid,
  o,
  ...
)

\method{tmapGridDataPlot}{tm_data_polygons}(
  a,
  shpTM,
  dt,
  gp,
  bbx,
  facet_row,
  facet_col,
  facet_page,
  id,
  pane,
  group,
  glid,
  o,
  ...
)

\method{tmapGridDataPlot}{tm_data_fill}(
  a,
  shpTM,
  dt,
  gp,
  bbx,
  facet_row,
  facet_col,
  facet_page,
  id,
  pane,
  group,
  glid,
  o,
  ...
)

\method{tmapGridDataPlot}{tm_data_borders}(
  a,
  shpTM,
  dt,
  gp,
  bbx,
  facet_row,
  facet_col,
  facet_page,
  id,
  pane,
  group,
  glid,
  o,
  ...
)

\method{tmapGridDataPlot}{tm_data_raster}(
  a,
  shpTM,
  dt,
  gp,
  bbx,
  facet_row,
  facet_col,
  facet_page,
  id,
  pane,
  group,
  glid,
  o,
  ...
)

\method{tmapGridDataPlot}{tm_data_symbols}(
  a,
  shpTM,
  dt,
  gp,
  bbx,
  facet_row,
  facet_col,
  facet_page,
  id,
  pane,
  group,
  glid,
  o,
  ...
)

\method{tmapGridDataPlot}{tm_data_dots}(
  a,
  shpTM,
  dt,
  gp,
  bbx,
  facet_row,
  facet_col,
  facet_page,
  id,
  pane,
  group,
  glid,
  o,
  ...
)

\method{tmapGridDataPlot}{tm_data_bubbles}(
  a,
  shpTM,
  dt,
  gp,
  bbx,
  facet_row,
  facet_col,
  facet_page,
  id,
  pane,
  group,
  glid,
  o,
  ...
)

\method{tmapGridDataPlot}{tm_data_squares}(
  a,
  shpTM,
  dt,
  gp,
  bbx,
  facet_row,
  facet_col,
  facet_page,
  id,
  pane,
  group,
  glid,
  o,
  ...
)

\method{tmapGridDataPlot}{tm_data_markers}(
  a,
  shpTM,
  dt,
  gp,
  bbx,
  facet_row,
  facet_col,
  facet_page,
  id,
  pane,
  group,
  glid,
  o,
  ...
)

\method{tmapGridDataPlot}{tm_data_text}(
  a,
  shpTM,
  dt,
  gp,
  bbx,
  facet_row,
  facet_col,
  facet_page,
  id,
  pane,
  group,
  glid,
  o,
  ...
)

\method{tmapGridDataPlot}{tm_data_labels}(
  a,
  shpTM,
  dt,
  gp,
  bbx,
  facet_row,
  facet_col,
  facet_page,
  id,
  pane,
  group,
  glid,
  o,
  ...
)

\method{tmapGridDataPlot}{tm_data_labels_highlighted}(
  a,
  shpTM,
  dt,
  gp,
  bbx,
  facet_row,
  facet_col,
  facet_page,
  id,
  pane,
  group,
  glid,
  o,
  ...
)

tmapLeafletAuxPrepare(a, bs, id, o)

tmapLeafletAuxPlot(
  a,
  bi,
  bbx,
  facet_row,
  facet_col,
  facet_page,
  id,
  pane,
  group,
  o
)

\method{tmapLeafletAuxPrepare}{tm_aux_basemap}(a, bs, id, o)

\method{tmapLeafletAuxPrepare}{tm_aux_tiles}(a, bs, id, o)

\method{tmapLeafletAuxPlot}{tm_aux_basemap}(
  a,
  bi,
  bbx,
  facet_row,
  facet_col,
  facet_page,
  id,
  pane,
  group,
  o
)

\method{tmapLeafletAuxPlot}{tm_aux_tiles}(
  a,
  bi,
  bbx,
  facet_row,
  facet_col,
  facet_page,
  id,
  pane,
  group,
  o
)

\method{tmapLeafletAuxPrepare}{tm_aux_graticules}(a, bs, id, o)

\method{tmapLeafletAuxPrepare}{tm_aux_grid}(a, bs, id, o)

\method{tmapLeafletAuxPlot}{tm_aux_graticules}(
  a,
  bi,
  bbx,
  facet_row,
  facet_col,
  facet_page,
  id,
  pane,
  group,
  o
)

\method{tmapLeafletAuxPlot}{tm_aux_grid}(
  a,
  bi,
  bbx,
  facet_row,
  facet_col,
  facet_page,
  id,
  pane,
  group,
  o
)

tmapLeafletCompPrepare(comp, o)

\method{tmapLeafletCompPrepare}{tm_chart_none}(comp, o)

\method{tmapLeafletCompPrepare}{default}(comp, o)

\method{tmapLeafletDataPlot}{tm_data_circles}(
  a,
  shpTM,
  dt,
  pdt,
  popup.format,
  hdt,
  idt,
  gp,
  bbx,
  facet_row,
  facet_col,
  facet_page,
  id,
  pane,
  group,
  glid,
  o,
  ...
)

\method{tmapLeafletDataPlot}{tm_data_lines}(
  a,
  shpTM,
  dt,
  pdt,
  popup.format,
  hdt,
  idt,
  gp,
  bbx,
  facet_row,
  facet_col,
  facet_page,
  id,
  pane,
  group,
  glid,
  o,
  ...
)

\method{tmapLeafletDataPlot}{tm_data_iso}(
  a,
  shpTM,
  dt,
  pdt,
  popup.format,
  hdt,
  idt,
  gp,
  bbx,
  facet_row,
  facet_col,
  facet_page,
  id,
  pane,
  group,
  glid,
  o,
  ...
)

\method{tmapLeafletDataPlot}{tm_data_polygons}(
  a,
  shpTM,
  dt,
  pdt,
  popup.format,
  hdt,
  idt,
  gp,
  bbx,
  facet_row,
  facet_col,
  facet_page,
  id,
  pane,
  group,
  glid,
  o,
  ...
)

\method{tmapLeafletDataPlot}{tm_data_fill}(
  a,
  shpTM,
  dt,
  pdt,
  popup.format,
  hdt,
  idt,
  gp,
  bbx,
  facet_row,
  facet_col,
  facet_page,
  id,
  pane,
  group,
  glid,
  o,
  ...
)

\method{tmapLeafletDataPlot}{tm_data_borders}(
  a,
  shpTM,
  dt,
  pdt,
  popup.format,
  hdt,
  idt,
  gp,
  bbx,
  facet_row,
  facet_col,
  facet_page,
  id,
  pane,
  group,
  glid,
  o,
  ...
)

\method{tmapLeafletDataPlot}{tm_data_raster}(
  a,
  shpTM,
  dt,
  pdt,
  popup.format,
  hdt,
  idt,
  gp,
  bbx,
  facet_row,
  facet_col,
  facet_page,
  id,
  pane,
  group,
  glid,
  o,
  ...
)

\method{tmapLeafletDataPlot}{tm_data_symbols}(
  a,
  shpTM,
  dt,
  pdt,
  popup.format,
  hdt,
  idt,
  gp,
  bbx,
  facet_row,
  facet_col,
  facet_page,
  id,
  pane,
  group,
  glid,
  o,
  ...
)

\method{tmapLeafletDataPlot}{tm_data_dots}(
  a,
  shpTM,
  dt,
  pdt,
  popup.format,
  hdt,
  idt,
  gp,
  bbx,
  facet_row,
  facet_col,
  facet_page,
  id,
  pane,
  group,
  glid,
  o,
  ...
)

\method{tmapLeafletDataPlot}{tm_data_bubbles}(
  a,
  shpTM,
  dt,
  pdt,
  popup.format,
  hdt,
  idt,
  gp,
  bbx,
  facet_row,
  facet_col,
  facet_page,
  id,
  pane,
  group,
  glid,
  o,
  ...
)

\method{tmapLeafletDataPlot}{tm_data_squares}(
  a,
  shpTM,
  dt,
  pdt,
  popup.format,
  hdt,
  idt,
  gp,
  bbx,
  facet_row,
  facet_col,
  facet_page,
  id,
  pane,
  group,
  glid,
  o,
  ...
)

\method{tmapLeafletDataPlot}{tm_data_markers}(
  a,
  shpTM,
  dt,
  pdt,
  popup.format,
  hdt,
  idt,
  gp,
  bbx,
  facet_row,
  facet_col,
  facet_page,
  id,
  pane,
  group,
  glid,
  o,
  ...
)

\method{tmapLeafletDataPlot}{tm_data_text}(
  a,
  shpTM,
  dt,
  pdt,
  popup.format,
  hdt,
  idt,
  gp,
  bbx,
  facet_row,
  facet_col,
  facet_page,
  id,
  pane,
  group,
  glid,
  o,
  ...
)

\method{tmapLeafletDataPlot}{tm_data_labels}(
  a,
  shpTM,
  dt,
  pdt,
  popup.format,
  hdt,
  idt,
  gp,
  bbx,
  facet_row,
  facet_col,
  facet_page,
  id,
  pane,
  group,
  glid,
  o,
  ...
)

\method{tmapLeafletDataPlot}{tm_data_labels_highlighted}(
  a,
  shpTM,
  dt,
  pdt,
  popup.format,
  hdt,
  idt,
  gp,
  bbx,
  facet_row,
  facet_col,
  facet_page,
  id,
  pane,
  group,
  glid,
  o,
  ...
)

tmapLeafletDataPlot(
  a,
  shpTM,
  dt,
  pdt,
  popup.format,
  hdt,
  idt,
  gp,
  bbx,
  facet_row,
  facet_col,
  facet_page,
  id,
  pane,
  group,
  glid,
  o,
  ...
)

\method{tmapLeafletDataPlot}{default}(
  a,
  shpTM,
  dt,
  pdt,
  popup.format,
  hdt,
  idt,
  gp,
  bbx,
  facet_row,
  facet_col,
  facet_page,
  id,
  pane,
  group,
  glid,
  o,
  ...
)
}
\arguments{
\item{o}{options}

\item{comp}{component}
}
\description{
Internal methods for map components
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tm_layout.R
\name{tm_place_legends_right}
\alias{tm_place_legends_right}
\alias{tm_place_legends_left}
\alias{tm_place_legends_bottom}
\alias{tm_place_legends_top}
\alias{tm_place_legends_inside}
\alias{tm_extra_innner_margin}
\alias{tm_extra_inner_margin}
\title{tmap layout: helper functions}
\usage{
tm_place_legends_right(width = NA)

tm_place_legends_left(width = NA)

tm_place_legends_bottom(height = NA)

tm_place_legends_top(height = NA)

tm_place_legends_inside(pos.h = NULL, pos.v = NULL)

tm_extra_inner_margin(left = 0, right = 0, top = 0, bottom = 0)
}
\arguments{
\item{width}{width}

\item{height}{height}

\item{pos.h, pos.v}{position (horizontal and vertical)}

\item{left, right, top, bottom}{extra margins}
}
\description{
tmap layout: helper functions
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tm_component_functions.R
\name{tm_logo}
\alias{tm_logo}
\title{Map component: logo}
\usage{
tm_logo(
  file,
  height,
  margins,
  between_margin,
  stack,
  position,
  group_id,
  frame,
  frame.color,
  frame.alpha,
  frame.lwd,
  frame.r,
  z
)
}
\arguments{
\item{file}{either a filename or url of a png image. If multiple files/urls are provided with a character vector, the logos are placed near each other.
To specify logos for small multiples use a list of character values/vectors. In order to stack logos vertically, multiple \code{tm_logo} elements can be stacked.}

\item{height}{height of the logo in number of text line heights. The width is scaled based the height and the aspect ratio of the logo.
If multiple logos are specified by a vector or list, the heights can be specified accordingly.}

\item{margins}{margins}

\item{between_margin}{Margin between}

\item{stack}{stack with other map components, either \code{"vertical"} or \code{"horizontal"}.}

\item{position}{The position specification of the component: an object created with \code{tm_pos_in()} or \code{tm_pos_out()}. Or, as a shortcut, a vector of two values, specifying the x and y coordinates. The first is \code{"left"}, \code{"center"} or \code{"right"} (or upper case, meaning tighter to the map frame), the second \code{"top"}, \code{"center"} or \code{"bottom"}. Numeric values are also supported, where 0, 0 means left bottom and 1, 1 right top. See also \href{https://r-tmap.github.io/tmap/articles/adv_positions}{vignette about positioning}. In case multiple components should be combined (stacked), use \code{group_id} and specify \code{component} in \code{\link[=tm_components]{tm_components()}}.}

\item{group_id}{Component group id name. All components (e.g. legends, titles, etc) with the same \code{group_id} will be grouped. The specifications of how they are placed (e.g. stacking, margins etc.) are determined in \code{\link[=tm_components]{tm_components()}} where its argument \code{id} should correspond to \code{group_id}.}

\item{frame}{frame should a frame be drawn?}

\item{frame.color}{frame color}

\item{frame.alpha}{frame alpha transparancy}

\item{frame.lwd}{frame line width}

\item{frame.r}{Radius of the rounded frame corners. 0 means no rounding.}

\item{z}{z index, e.g. the place of the component relative to the other componets}
}
\description{
Map component that adds a logo.
}
\examples{
data(World)

tm_shape(World) +
	tm_polygons("HPI", fill.scale = tm_scale_intervals(values = "brewer.rd_yl_gn")) +
	tm_logo(c("https://www.r-project.org/logo/Rlogo.png",
			  system.file("help", "figures", "logo.png", package = "tmap"))) +
	tm_logo("https://happyplanetindex.org/wp-content/themes/hpi/public/images/hpi-logo.svg",
			height=5, position = c("left", "center"))
}
\seealso{
\href{https://r-tmap.github.io/tmap/articles/basics_components}{Vignette about components}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tmapShape_misc.R
\name{make_by_vars}
\alias{make_by_vars}
\title{Internal tmap function to create by variables (used for faceting)}
\usage{
make_by_vars(dt, tmf, smeta)
}
\arguments{
\item{dt}{data.table}

\item{tmf}{tmf object}

\item{smeta}{smeta object}
}
\description{
Internal tmap function to create by variables (used for faceting)
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tm_component_functions.R
\name{tm_logo}
\alias{tm_logo}
\title{Map component: logo}
\usage{
tm_logo(
  file,
  height,
  margins,
  between_margin,
  stack,
  position,
  group_id,
  frame,
  frame.color,
  frame.alpha,
  frame.lwd,
  frame.r,
  z
)
}
\arguments{
\item{file}{either a filename or url of a png image. If multiple files/urls are provided with a character vector, the logos are placed near each other.
To specify logos for small multiples use a list of character values/vectors. In order to stack logos vertically, multiple \code{tm_logo} elements can be stacked.}

\item{height}{height of the logo in number of text line heights. The width is scaled based the height and the aspect ratio of the logo.
If multiple logos are specified by a vector or list, the heights can be specified accordingly.}

\item{margins}{margins}

\item{between_margin}{Margin between}

\item{stack}{stack with other map components, either \code{"vertical"} or \code{"horizontal"}.}

\item{position}{The position specification of the component: an object created with \code{tm_pos_in()} or \code{tm_pos_out()}. Or, as a shortcut, a vector of two values, specifying the x and y coordinates. The first is \code{"left"}, \code{"center"} or \code{"right"} (or upper case, meaning tighter to the map frame), the second \code{"top"}, \code{"center"} or \code{"bottom"}. Numeric values are also supported, where 0, 0 means left bottom and 1, 1 right top. See also vignette: \href{https://r-tmap.github.io/tmap/articles/adv_positions}{Positioning of components}. In case multiple components should be combined (stacked), use \code{group_id} and specify \code{component} in \code{\link[=tm_components]{tm_components()}}.}

\item{group_id}{Component group id name. All components (e.g. legends, titles, etc) with the same \code{group_id} will be grouped. The specifications of how they are placed (e.g. stacking, margins etc.) are determined in \code{\link[=tm_components]{tm_components()}} where its argument \code{id} should correspond to \code{group_id}.}

\item{frame}{frame should a frame be drawn?}

\item{frame.color}{frame color}

\item{frame.alpha}{frame alpha transparancy}

\item{frame.lwd}{frame line width}

\item{frame.r}{Radius of the rounded frame corners. 0 means no rounding.}

\item{z}{z index, e.g. the place of the component relative to the other componets}
}
\description{
Map component that adds a logo.
}
\examples{
data(World)

tm_shape(World) +
	tm_polygons("HPI", fill.scale = tm_scale_intervals(values = "brewer.rd_yl_gn")) +
	tm_logo(c("https://www.r-project.org/logo/Rlogo.png",
			  system.file("help", "figures", "logo.png", package = "tmap"))) +
	tm_logo("https://happyplanetindex.org/wp-content/themes/hpi/public/images/hpi-logo.svg",
			height=5, position = c("left", "center"))
}
\seealso{
\href{https://r-tmap.github.io/tmap/articles/basics_components}{Components}, \href{https://r-tmap.github.io/tmap/articles/adv_positions}{Positioning of components}, \href{https://r-tmap.github.io/tmap/articles/adv_comp_group}{Grouping of components}, \href{https://r-tmap.github.io/tmap/articles/examples_choro_NLD}{Choropleth (Netherlands)}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tm_scale_.R
\name{tm_const}
\alias{tm_const}
\title{tmap function to define a constant visual value}
\usage{
tm_const()
}
\description{
tmap function to define a constant visual value
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tm_scale_.R
\name{tm_scale_rank}
\alias{tm_scale_rank}
\title{Scales: rank scale}
\usage{
tm_scale_rank(
  n = NULL,
  ticks = NULL,
  values = NA,
  values.repeat = FALSE,
  values.range = NA,
  values.scale = NA,
  value.na = NA,
  value.null = NA,
  value.neutral = NA,
  labels = NULL,
  label.na = NA,
  label.null = NA,
  label.format = tm_label_format(),
  unit = "rank"
)
}
\arguments{
\item{n}{Preferred number of tick labels. Only used if \code{ticks} is not specified}

\item{ticks}{Tick values. If not specified, it is determined automatically with \code{n}}

\item{values}{(generic scale argument) The visual values. For colors (e.g. \code{fill} or \code{col} for \code{\link[=tm_polygons]{tm_polygons()}}) this is a palette name from the \code{cols4all} package (see \code{\link[cols4all:c4a]{cols4all::c4a()}}) or vector of colors, for size (e.g. \code{size} for \code{\link[=tm_symbols]{tm_symbols()}}) these are a set of sizes (if two values are specified they are interpret as range), for symbol shapes (e.g. \code{shape} for \code{\link[=tm_symbols]{tm_symbols()}}) these are a set of symbols, etc. The tmap option \code{values.var} contains the default values per visual variable and in some cases also per data type.}

\item{values.repeat}{(generic scale argument) Should the values be repeated in case there are more categories?}

\item{values.range}{(generic scale argument) Range of the values, especially useful for color palettes. Vector of two numbers (both between 0 and 1) where the first determines the minimum and the second the maximum. Full range, which means that all values are used, is encoded as \code{c(0, 1)}. For instance, when a gray scale is used for color (from black to white), \code{c(0,1)} means that all colors are used, \verb{0.25, 0.75} means that only colors from dark gray to light gray are used (more precisely \code{"grey25"} to \code{"grey75"}), and \verb{0, 0.5} means that only colors are used from black to middle gray (\code{"grey50"}). When only one number is specified, this is interpreted as the second number (where the first is set to 0). Default values can be set via the tmap option \code{values.range}.}

\item{values.scale}{(generic scale argument) Scaling of the values. Only useful for size-related visual variables, such as \code{size} of \code{\link[=tm_symbols]{tm_symbols()}} and \code{lwd} of \code{\link[=tm_lines]{tm_lines()}}.}

\item{value.na}{(generic scale argument) Value used for missing values. See tmap option \code{"value.na"} for defaults per visual variable.}

\item{value.null}{(generic scale argument) Value used for NULL values. See tmap option \code{"value.null"} for defaults per visual variable. Null data values occur when out-of-scope features are shown (e.g. for a map of Europe showing a data variable per country, the null values are applied to countries outside Europe).}

\item{value.neutral}{(generic scale argument) Value that can be considered neutral. This is used for legends of other visual variables of the same map layer. E.g. when both \code{fill} and \code{size} are used for \code{\link[=tm_symbols]{tm_symbols()}} (using filled circles), the size legend items are filled with the \code{value.neutral} color from the \code{fill.scale} scale, and fill legend items are bubbles of size \code{value.neutral} from the \code{size.scale} scale.}

\item{labels}{(generic scale argument) Labels}

\item{label.na}{(generic scale argument) Label for missing values}

\item{label.null}{(generic scale argument) Label for null (out-of-scope) values}

\item{label.format}{(generic scale argument) Label formatting. Output of \code{\link[=tm_label_format]{tm_label_format()}}}

\item{unit}{The unit name of the values. By default \code{"rank"}.}
}
\description{
Scales in tmap are configured by the family of functions with prefix \code{tm_scale}.
Such function should be used for the input of the \code{.scale} arguments in the layer
functions (e.g. \code{fill.scale} in \code{\link[=tm_polygons]{tm_polygons()}}).
The function \code{\link[=tm_scale_rank]{tm_scale_rank()}} is used to rank numeric data.
}
\seealso{
\code{\link[=tm_scale]{tm_scale()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tm_scale_.R
\name{tm_scale_rank}
\alias{tm_scale_rank}
\title{Scales: rank scale}
\usage{
tm_scale_rank(
  n = NULL,
  ticks = NULL,
  values = NA,
  values.repeat = FALSE,
  values.range = NA,
  values.scale = NA,
  value.na = NA,
  value.null = NA,
  value.neutral = NA,
  labels = NULL,
  label.na = NA,
  label.null = NA,
  label.format = tm_label_format(),
  unit = "rank"
)
}
\arguments{
\item{n}{Preferred number of tick labels. Only used if \code{ticks} is not specified}

\item{ticks}{Tick values. If not specified, it is determined automatically with \code{n}}

\item{values}{(generic scale argument) The visual values. For colors (e.g. \code{fill} or \code{col} for \code{\link[=tm_polygons]{tm_polygons()}}) this is a palette name from the \code{cols4all} package (see \code{\link[cols4all:c4a]{cols4all::c4a()}}) or vector of colors, for size (e.g. \code{size} for \code{\link[=tm_symbols]{tm_symbols()}}) these are a set of sizes (if two values are specified they are interpret as range), for symbol shapes (e.g. \code{shape} for \code{\link[=tm_symbols]{tm_symbols()}}) these are a set of symbols, etc. The tmap option \code{values.var} contains the default values per visual variable and in some cases also per data type.}

\item{values.repeat}{(generic scale argument) Should the values be repeated in case there are more categories?}

\item{values.range}{(generic scale argument) Range of the values, especially useful for color palettes. Vector of two numbers (both between 0 and 1) where the first determines the minimum and the second the maximum. Full range, which means that all values are used, is encoded as \code{c(0, 1)}. For instance, when a gray scale is used for color (from black to white), \code{c(0,1)} means that all colors are used, \verb{0.25, 0.75} means that only colors from dark gray to light gray are used (more precisely \code{"grey25"} to \code{"grey75"}), and \verb{0, 0.5} means that only colors are used from black to middle gray (\code{"grey50"}). When only one number is specified, this is interpreted as the second number (where the first is set to 0). Default values can be set via the tmap option \code{values.range}.}

\item{values.scale}{(generic scale argument) Scaling of the values. Only useful for size-related visual variables, such as \code{size} of \code{\link[=tm_symbols]{tm_symbols()}} and \code{lwd} of \code{\link[=tm_lines]{tm_lines()}}.}

\item{value.na}{(generic scale argument) Value used for missing values. See tmap option \code{"value.na"} for defaults per visual variable.}

\item{value.null}{(generic scale argument) Value used for NULL values. See tmap option \code{"value.null"} for defaults per visual variable. Null data values occur when out-of-scope features are shown (e.g. for a map of Europe showing a data variable per country, the null values are applied to countries outside Europe).}

\item{value.neutral}{(generic scale argument) Value that can be considered neutral. This is used for legends of other visual variables of the same map layer. E.g. when both \code{fill} and \code{size} are used for \code{\link[=tm_symbols]{tm_symbols()}} (using filled circles), the size legend items are filled with the \code{value.neutral} color from the \code{fill.scale} scale, and fill legend items are bubbles of size \code{value.neutral} from the \code{size.scale} scale.}

\item{labels}{(generic scale argument) Labels}

\item{label.na}{(generic scale argument) Label for missing values}

\item{label.null}{(generic scale argument) Label for null (out-of-scope) values}

\item{label.format}{(generic scale argument) Label formatting. Output of \code{\link[=tm_label_format]{tm_label_format()}}}

\item{unit}{The unit name of the values. By default \code{"rank"}.}
}
\description{
Scales in tmap are configured by the family of functions with prefix \code{tm_scale}.
Such function should be used for the input of the \code{.scale} arguments in the layer
functions (e.g. \code{fill.scale} in \code{\link[=tm_polygons]{tm_polygons()}}).
The function \code{\link[=tm_scale_rank]{tm_scale_rank()}} is used to rank numeric data.
}
\seealso{
\href{https://r-tmap.github.io/tmap/articles/basics_scales}{Scales}, \href{https://r-tmap.github.io/tmap/articles/basics_vv}{Visual variables}, \href{https://r-tmap.github.io/tmap/articles/adv_multivariate}{Multivariate visual variables} \code{\link[=tm_scale]{tm_scale()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tm_layers_circles.R
\name{tm_circles}
\alias{tm_circles}
\alias{opt_tm_circles}
\title{Map layer: geographic circles}
\usage{
tm_circles(
  size = tm_const(),
  size.scale = tm_scale(),
  size.legend = tm_legend(),
  size.chart = tm_chart_none(),
  size.free = NA,
  fill = tm_const(),
  fill.scale = tm_scale(),
  fill.legend = tm_legend(),
  fill.chart = tm_chart_none(),
  fill.free = NA,
  col = tm_const(),
  col.scale = tm_scale(),
  col.legend = tm_legend(),
  col.chart = tm_chart_none(),
  col.free = NA,
  lwd = tm_const(),
  lwd.scale = tm_scale(),
  lwd.legend = tm_legend(),
  lwd.chart = tm_chart_none(),
  lwd.free = NA,
  lty = tm_const(),
  lty.scale = tm_scale(),
  lty.legend = tm_legend(),
  lty.chart = tm_chart_none(),
  lty.free = NA,
  fill_alpha = tm_const(),
  fill_alpha.scale = tm_scale(),
  fill_alpha.legend = tm_legend(),
  fill_alpha.chart = tm_chart_none(),
  fill_alpha.free = NA,
  col_alpha = tm_const(),
  col_alpha.scale = tm_scale(),
  col_alpha.legend = tm_legend(),
  col_alpha.chart = tm_chart_none(),
  col_alpha.free = NA,
  plot.order = tm_plot_order("size"),
  zindex = NA,
  group = NA,
  group.control = "check",
  popup = tm_popup(),
  popup.vars = NA,
  popup.format = tm_label_format(),
  hover = NA,
  id = "",
  blend = "over",
  options = opt_tm_circles(),
  ...
)

opt_tm_circles(
  points_only = "ifany",
  point_per = "feature",
  on_surface = FALSE
)
}
\arguments{
\item{size, size.scale, size.legend, size.chart, size.free}{Visual variable that determines the size. See details. \emph{Unit:} Meters. Accepts a plain numeric vector (values already in meters) or a \code{units} object from the \strong{units} package (any linear unit, e.g. \code{units::as_units(50, "km")}), which is converted to meters automatically.
\emph{Unit:} Meters. Accepts a plain numeric vector (values already in meters) or a \code{units} object from the \strong{units} package (any linear unit, e.g. \code{units::as_units(50, "km")}), which is converted to meters automatically.}

\item{fill, fill.scale, fill.legend, fill.chart, fill.free}{Visual variable that determines the fill color. See details. \emph{Unit:} Color -- a color name, hex string.}

\item{col, col.scale, col.legend, col.chart, col.free}{Visual variable that determines the color. See details. \emph{Unit:} Color -- a color name, hex string.}

\item{lwd, lwd.scale, lwd.legend, lwd.chart, lwd.free}{Visual variable that determines the line width. See details. \emph{Unit:} Base R line-width units; 1 lwd is approx. 0.75 pt at 96 dpi.}

\item{lty, lty.scale, lty.legend, lty.chart, lty.free}{Visual variable that determines the line type. See details. \emph{Unit:} Integer (1-6) or name: "solid", "dashed", "dotted", "dotdash", "longdash", "twodash".}

\item{fill_alpha, fill_alpha.scale, fill_alpha.legend, fill_alpha.chart, fill_alpha.free}{Visual variable that determines the fill color transparency. See details. \emph{Unit:} Proportion -- numeric 0-1 (0 = fully transparent, 1 = fully opaque).}

\item{col_alpha, col_alpha.scale, col_alpha.legend, col_alpha.chart, col_alpha.free}{Visual variable that determines the color transparency. See details. \emph{Unit:} Proportion -- numeric 0-1 (0 = fully transparent, 1 = fully opaque).}

\item{plot.order}{Specification in which order the spatial features are
drawn. See \code{\link[=tm_plot_order]{tm_plot_order()}} for details.}

\item{zindex}{Controls the stacking order of map layers. Should be set to a
value above 400. By default, layers are stacked in call order, starting
at 401. See details.}

\item{group}{Name of the group to which this layer belongs. This is only
relevant in view mode, where layer groups can be switched (see \code{group.control})}

\item{group.control}{In view mode, the group control determines how layer groups
can be switched on and off. Options: \code{"radio"} for radio buttons
(meaning only one group can be shown), \code{"check"} for check boxes
(so multiple groups can be shown), and \code{"none"} for no control
(the group cannot be (de)selected).}

\item{popup}{popup specification for \code{"view"} mode, the output of \code{\link[=tm_popup]{tm_popup()}}.
It determines the data variables shown in the popup table, the popup title,
and (in the future) the popup layout. This replaces the deprecated
arguments \code{popup.vars} and \code{popup.format}.}

\item{popup.vars}{(Deprecated.) Use \code{popup} with
\code{\link[=tm_popup]{tm_popup()}} instead (via its \code{vars} argument). Names of data variables that
are shown in the popups in \code{"view"} mode. Set \code{popup.vars} to \code{TRUE} to show
all variables in the shape object. Set \code{popup.vars} to \code{FALSE} to disable
popups. Set \code{popup.vars} to a character vector of variable names to show
those variables in the popups. The default (\code{NA}) depends on whether visual
variables (e.g. \code{fill}) are used. If so, only those are shown. If not, all
variables in the shape object are shown.}

\item{popup.format}{(Deprecated.) Use \code{popup} with
\code{\link[=tm_popup]{tm_popup()}} instead (via its \code{format} argument). List of formatting options
for the popup values. Output of \code{\link[=tm_label_format]{tm_label_format()}}. Only applicable for
numeric data variables. If one list of formatting options is provided,
it is applied to all numeric variables of \code{popup.vars}. Also, a (named)
list of lists can be provided. In that case, each list of formatting options
is applied to the named variable.}

\item{hover}{name of the data variable that specifies the hover labels (view mode only). Set to \code{FALSE} to disable hover labels. By default \code{FALSE}, unless \code{id} is specified. In that case, it is set to \code{id},}

\item{id}{name of the data variable that specifies the indices of the spatial
features. Only used for \code{"view"} mode.}

\item{blend}{Compositing operator for layer blending. Default \code{"over"} applies
no blending. See the "Layer blending" section for the supported values.}

\item{options}{Options passed on to \code{\link[=opt_tm_circles]{opt_tm_circles()}}.}

\item{...}{To catch deprecated arguments from version < 4.0.}

\item{points_only}{Should only point geometries of the shape object be
plotted? Default \code{"ifany"} means \code{TRUE} whenever the geometry collection
contains points.}

\item{point_per}{How spatial points are generated for non-point geometries.
One of \code{"feature"} (one point per feature, default), \code{"segment"} (one per
sub-feature), or \code{"largest"} (only the largest sub-feature).}

\item{on_surface}{For polygon inputs, should the centroid be guaranteed to
lie on the surface? If \code{TRUE} (slower), centroids outside the polygon are
replaced via \code{\link[sf:st_point_on_surface]{sf::st_point_on_surface()}}.}
}
\description{
Map layer that draws circles with geographically fixed radii — i.e. the
radius is expressed in meters and the circles scale with zoom in interactive
(view) mode. This is in contrast to \code{\link[=tm_bubbles]{tm_bubbles()}}, where the symbol size
is a fixed number of screen pixels.
}
\details{
Supported visual variables: \code{fill} (fill colour), \code{col} (border colour),
\code{size} (radius in meters), \code{lwd} (line width), \code{lty} (line type),
\code{fill_alpha} (fill transparency), \code{col_alpha} (border transparency).
}
\examples{
## Three concentric layers of geographic circles at different administrative
## levels, each with a fixed radius that corresponds to a real-world distance.
## Because the radius is in meters the circles scale with zoom in view mode.
tm_shape(NLD_prov) +
	tm_circles(size = 5000, fill = "#0033ff", col = NULL) +
tm_shape(NLD_muni) +
	tm_circles(size = 2000, fill = "#99dd99", col = NULL) +
tm_shape(NLD_dist) +
	tm_circles(size = 1000, fill = "#ff8833", col = NULL)

## Use a units object — any linear unit is accepted and converted to meters
NLD_prov$one_mile <- units::as_units(1:12, "mi")
tm_shape(NLD_prov) +
	tm_circles(size = "one_mile", size.scale = tm_scale_asis())
}
\seealso{
\itemize{
\item \code{\link[=tm_bubbles]{tm_bubbles()}} for screen-fixed proportional circles (pixel radius).
\item \code{\link[=tm_symbols]{tm_symbols()}} for the general symbol layer with configurable shapes.
\item \code{\link[=tm_scale_asis]{tm_scale_asis()}} to pass data values directly as metre radii.
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tm_add_legend.R, R/tm_element.R, R/tmapChart.R,
%   R/tmapGpar.R, R/tmapGridComp.R, R/tmapGridComp_charts.R, R/tmapOnLoad.R,
%   R/tmapScaleAsIs.R, R/tmapScaleBivariate.R, R/tmapScaleCategorical.R,
%   R/tmapScaleIntervals.R, R/tmapScaleRank.R, R/tmapScale_.R,
%   R/tmapScale_defaults.R, R/tmapTrans.R
\docType{data}
\name{toTitleCase}
\alias{toTitleCase}
\alias{tmap_internal}
\alias{tm_element}
\alias{tm_element_list}
\alias{tmapChartBinned}
\alias{tmapChartBinned2d}
\alias{tmapChartBinned2d_numcat}
\alias{tmapChartBinned2d_numnum}
\alias{tmapChartBinned2d_catcat}
\alias{tmapChartRaw}
\alias{tmapChartRaw_nna}
\alias{tmapChartNone}
\alias{tmapChartPass}
\alias{tmapChartBinned_categorical}
\alias{tmapChartBinned_numeric}
\alias{tmapGpar}
\alias{tmapTpar}
\alias{tmapGetCompGroupArgs}
\alias{lwd_to_mm}
\alias{.TMAP}
\alias{.TMAP_LEAFLET}
\alias{.TMAP_GRID}
\alias{tmapScaleAsIs}
\alias{tmapScaleBivariate}
\alias{tmapScaleCategorical}
\alias{tmapScaleIntervals}
\alias{tmapScaleRank}
\alias{tmapUsrCls}
\alias{format_aes_results}
\alias{chart_save}
\alias{data_type}
\alias{data_class}
\alias{tmapScale}
\alias{tmapScaleAuto}
\alias{tmapValuesCheck_col}
\alias{tmapValuesCheck_fill}
\alias{tmapValuesCheck_bgcol}
\alias{tmapValuesCheck_shape}
\alias{tmapValuesCheck_size}
\alias{tmapValuesCheck_area}
\alias{tmapValuesCheck_lwd}
\alias{tmapValuesCheck_lty}
\alias{tmapValuesCheck_xmod}
\alias{tmapValuesCheck_angle}
\alias{tmapValuesCheck_col_alpha}
\alias{tmapValuesCheck_fill_alpha}
\alias{tmapValuesCheck_bgcol_alpha}
\alias{tmapValuesCheck_text}
\alias{tmapValuesCheck_fontface}
\alias{tmapValuesIsDiv_fill}
\alias{tmapValuesIsDiv_col}
\alias{tmapValuesIsDiv_bgcol}
\alias{tmapValuesIsDiv_size}
\alias{tmapValuesIsDiv_area}
\alias{tmapValuesIsDiv_lwd}
\alias{tmapValuesIsDiv_lty}
\alias{tmapValuesIsDiv_col_alpha}
\alias{tmapValuesIsDiv_fill_alpha}
\alias{tmapValuesIsDiv_bgcol_alpha}
\alias{tmapValuesIsDiv_xmod}
\alias{tmapValuesIsDiv_angle}
\alias{tmapValuesIsDiv_shape}
\alias{tmapValuesIsDiv_text}
\alias{tmapValuesIsDiv_fontface}
\alias{tmapValuesRange_fill}
\alias{tmapValuesRange_col}
\alias{tmapValuesRange_bgcol}
\alias{tmapValuesRange_shape}
\alias{tmapValuesRange_lty}
\alias{tmapValuesRange_size}
\alias{tmapValuesRange_area}
\alias{tmapValuesRange_lwd}
\alias{tmapValuesRange_col_alpha}
\alias{tmapValuesRange_fill_alpha}
\alias{tmapValuesRange_bgcol_alpha}
\alias{tmapValuesRange_xmod}
\alias{tmapValuesRange_angle}
\alias{tmapValuesRange_text}
\alias{tmapValuesRange_fontface}
\alias{tmapValuesVV_fill}
\alias{tmapValuesVV_col}
\alias{tmapValuesVV_bgcol}
\alias{tmapValuesVV_shape}
\alias{tmapValuesVV_lty}
\alias{tmapValuesVV_size}
\alias{tmapValuesVV_area}
\alias{tmapValuesVV_lwd}
\alias{tmapValuesVV_col_alpha}
\alias{tmapValuesVV_fill_alpha}
\alias{tmapValuesVV_bgcol_alpha}
\alias{tmapValuesVV_xmod}
\alias{tmapValuesVV_angle}
\alias{tmapValuesVV_text}
\alias{tmapValuesVV_fontface}
\alias{tmapValuesSubmit_col}
\alias{tmapValuesSubmit_fill}
\alias{tmapValuesSubmit_bgcol}
\alias{tmapValuesSubmit_size}
\alias{tmapValuesSubmit_area}
\alias{tmapValuesSubmit_xmod}
\alias{tmapValuesSubmit_ymod}
\alias{tmapValuesSubmit_angle}
\alias{tmapValuesSubmit_lwd}
\alias{tmapValuesSubmit_lty}
\alias{tmapValuesSubmit_shape}
\alias{tmapValuesSubmit_col_alpha}
\alias{tmapValuesSubmit_fill_alpha}
\alias{tmapValuesSubmit_bgcol_alpha}
\alias{tmapValuesSubmit_text}
\alias{tmapValuesSubmit_fontface}
\alias{tmapValuesScale_col}
\alias{tmapValuesScale_fill}
\alias{tmapValuesScale_bgcol}
\alias{tmapValuesScale_size}
\alias{tmapValuesScale_area}
\alias{tmapValuesScale_lwd}
\alias{tmapValuesScale_lty}
\alias{tmapValuesScale_shape}
\alias{tmapValuesScale_col_alpha}
\alias{tmapValuesScale_fill_alpha}
\alias{tmapValuesScale_bgcol_alpha}
\alias{tmapValuesScale_text}
\alias{tmapValuesScale_fontface}
\alias{tmapValuesScale_xmod}
\alias{tmapValuesScale_angle}
\alias{tmapValuesColorize_col}
\alias{tmapValuesColorize_fill}
\alias{tmapValuesColorize_bgcol}
\alias{tmapValuesColorize_size}
\alias{tmapValuesColorize_area}
\alias{tmapValuesColorize_lwd}
\alias{tmapValuesColorize_lty}
\alias{tmapValuesColorize_shape}
\alias{tmapValuesColorize_col_alpha}
\alias{tmapValuesColorize_fill_alpha}
\alias{tmapValuesColorize_bgcol_alpha}
\alias{tmapValuesColorize_text}
\alias{tmapValuesColorize_fontface}
\alias{tmapValuesColorize_xmod}
\alias{tmapValuesColorize_angle}
\alias{tmapSeq}
\alias{transform_values}
\alias{tmapValuesCVV_fill}
\alias{tmapValuesCVV_col}
\alias{tmapValuesCVV_bgcol}
\alias{tmapValuesCVV_size}
\alias{tmapValuesCVV_area}
\alias{tmapValuesCVV_lwd}
\alias{tmapValuesCVV_col_alpha}
\alias{tmapValuesCVV_fill_alpha}
\alias{tmapValuesCVV_bgcol_alpha}
\alias{tmapValuesCVV_xmod}
\alias{tmapValuesCVV_ymod}
\alias{tmapValuesCVV_angle}
\alias{tmapValuesCVV_shape}
\alias{tmapValuesCVV_lty}
\alias{tmapValuesCVV_text}
\alias{tmapValuesCVV_fontface}
\alias{tmapValuesBVV_fill}
\alias{tmapValuesBVV_col}
\alias{tmapValuesBVV_bgcol}
\alias{tmapValuesCheck_num}
\alias{tmapValuesIsDiv_num}
\alias{tmapValuesRange_num}
\alias{tmapValuesVV_num}
\alias{tmapValuesSubmit_num}
\alias{tmapValuesScale_num}
\alias{tmapValuesColorize_num}
\alias{tmapValuesCVV_num}
\alias{tmapValuesCheck_skip}
\alias{tmapValuesIsDiv_skip}
\alias{tmapValuesRange_skip}
\alias{tmapValuesVV_skip}
\alias{tmapValuesSubmit_skip}
\alias{tmapValuesScale_skip}
\alias{tmapValuesColorize_skip}
\alias{tmapValuesCVV_skip}
\alias{tmapTransCentroid}
\alias{tmapTransRaster}
\alias{tmapTransPolygons}
\alias{tmapTransLines}
\title{Internal methods for tmap extensions}
\format{
An object of class \code{environment} of length 10.

An object of class \code{environment} of length 0.

An object of class \code{environment} of length 1.
}
\usage{
toTitleCase(x)

tm_element(..., subclass = NULL)

tm_element_list(...)

tmapChartBinned(chart, bin_colors, breaks_def, na.show, x1)

tmapChartBinned2d(chart, chart1, chart2)

tmapChartBinned2d_numcat(chart, chart1, chart2)

tmapChartBinned2d_numnum(chart, chart1, chart2)

tmapChartBinned2d_catcat(chart, chart1, chart2)

tmapChartRaw(chart, na.show, x1, ...)

tmapChartRaw_nna(chart, na.show, x1, ...)

tmapChartNone(chart, na.show, x1, ...)

tmapChartPass(chart, na.show, x1, ...)

tmapChartBinned_categorical(chart, bin_colors, breaks_def, na.show, x1)

tmapChartBinned_numeric(chart, bin_colors, breaks_def, na.show, x1)

tmapGpar(
  fill = NULL,
  col = NULL,
  shape = NULL,
  size = NULL,
  fill_alpha = NULL,
  col_alpha = NULL,
  lty = NULL,
  lwd = NULL,
  linejoin = NULL,
  lineend = NULL,
  ...
)

tmapTpar(...)

tmapGetCompGroupArgs(comp)

lwd_to_mm(value, unit = "bigpts")

.TMAP

.TMAP_LEAFLET

.TMAP_GRID

tmapScaleAsIs(
  x1,
  scale,
  legend,
  chart,
  o,
  aes,
  layer,
  layer_args,
  sortRev,
  bypass_ord,
  submit_legend = TRUE
)

tmapScaleBivariate(
  x1,
  x2,
  scale,
  legend,
  chart,
  o,
  aes,
  layer,
  layer_args,
  sortRev,
  bypass_ord,
  submit_legend = TRUE
)

tmapScaleCategorical(
  x1,
  scale,
  legend,
  chart,
  o,
  aes,
  layer,
  layer_args,
  sortRev,
  bypass_ord,
  submit_legend = TRUE
)

tmapScaleIntervals(
  x1,
  scale,
  legend,
  chart,
  o,
  aes,
  layer,
  layer_args,
  sortRev,
  bypass_ord,
  submit_legend = TRUE
)

tmapScaleRank(
  x1,
  scale,
  legend,
  chart,
  o,
  aes,
  layer,
  layer_args,
  sortRev,
  bypass_ord,
  submit_legend = TRUE
)

tmapUsrCls(x)

format_aes_results(values, ord = NULL, legend, chart)

chart_save(legend)

data_type(x)

data_class(x, check_for_color_class = FALSE, midpoint_enabled = FALSE)

tmapScale(aes, value, scale, legend, chart, free)

tmapScaleAuto(
  x1,
  scale,
  legend,
  chart,
  o,
  aes,
  layer,
  layer_args,
  sortRev,
  bypass_ord,
  submit_legend = TRUE,
  ...
)

tmapValuesCheck_col(x, is_var = TRUE)

tmapValuesCheck_fill(x, is_var = TRUE)

tmapValuesCheck_bgcol(x, is_var = TRUE)

tmapValuesCheck_shape(x, is_var = TRUE)

tmapValuesCheck_size(x, is_var = TRUE)

tmapValuesCheck_area(x, is_var = TRUE)

tmapValuesCheck_lwd(x, is_var = TRUE)

tmapValuesCheck_lty(x, is_var = TRUE)

tmapValuesCheck_xmod(x, is_var = TRUE)

tmapValuesCheck_angle(x, is_var = TRUE)

tmapValuesCheck_col_alpha(x, is_var = TRUE)

tmapValuesCheck_fill_alpha(x, is_var = TRUE)

tmapValuesCheck_bgcol_alpha(x, is_var = TRUE)

tmapValuesCheck_area(x, is_var = TRUE)

tmapValuesCheck_text(x, is_var = TRUE)

tmapValuesCheck_fontface(x, is_var = TRUE)

tmapValuesIsDiv_fill(x)

tmapValuesIsDiv_col(x)

tmapValuesIsDiv_bgcol(x)

tmapValuesIsDiv_size(x)

tmapValuesIsDiv_area(x)

tmapValuesIsDiv_lwd(x)

tmapValuesIsDiv_lty(x)

tmapValuesIsDiv_col_alpha(x)

tmapValuesIsDiv_fill_alpha(x)

tmapValuesIsDiv_bgcol_alpha(x)

tmapValuesIsDiv_area(x)

tmapValuesIsDiv_xmod(x)

tmapValuesIsDiv_angle(x)

tmapValuesIsDiv_shape(x)

tmapValuesIsDiv_text(x)

tmapValuesIsDiv_fontface(x)

tmapValuesRange_fill(x, n, isdiv)

tmapValuesRange_col(x, n, isdiv)

tmapValuesRange_bgcol(x, n, isdiv)

tmapValuesRange_shape(x, n, isdiv)

tmapValuesRange_lty(x, n, isdiv)

tmapValuesRange_size(x, n, isdiv)

tmapValuesRange_area(x, n, isdiv)

tmapValuesRange_lwd(x, n, isdiv)

tmapValuesRange_col_alpha(x, n, isdiv)

tmapValuesRange_fill_alpha(x, n, isdiv)

tmapValuesRange_bgcol_alpha(x, n, isdiv)

tmapValuesRange_area(x, n, isdiv)

tmapValuesRange_xmod(x, n, isdiv)

tmapValuesRange_angle(x, n, isdiv)

tmapValuesRange_text(x, n, isdiv)

tmapValuesRange_fontface(x, n, isdiv)

tmapValuesVV_fill(
  x,
  value.na,
  isdiv,
  n,
  dvalues,
  are_breaks,
  midpoint,
  range,
  scale,
  rep,
  o,
  aes = "fill"
)

tmapValuesVV_col(...)

tmapValuesVV_bgcol(...)

tmapValuesVV_shape(
  x,
  value.na,
  isdiv,
  n,
  dvalues,
  are_breaks,
  midpoint,
  range,
  scale,
  rep,
  o
)

tmapValuesVV_lty(
  x,
  value.na,
  isdiv,
  n,
  dvalues,
  are_breaks,
  midpoint,
  range,
  scale,
  rep,
  o
)

tmapValuesVV_size(
  x,
  value.na,
  isdiv,
  n,
  dvalues,
  are_breaks,
  midpoint,
  range,
  scale,
  rep,
  o
)

tmapValuesVV_area(...)

tmapValuesVV_lwd(...)

tmapValuesVV_col_alpha(...)

tmapValuesVV_fill_alpha(...)

tmapValuesVV_bgcol_alpha(...)

tmapValuesVV_area(...)

tmapValuesVV_xmod(...)

tmapValuesVV_angle(...)

tmapValuesVV_text(
  x,
  value.na,
  isdiv,
  n,
  dvalues,
  are_breaks,
  midpoint,
  range,
  scale,
  rep,
  o
)

tmapValuesVV_fontface(
  x,
  value.na,
  isdiv,
  n,
  dvalues,
  are_breaks,
  midpoint,
  range,
  scale,
  rep,
  o
)

tmapValuesSubmit_col(x, args)

tmapValuesSubmit_fill(x, args)

tmapValuesSubmit_bgcol(x, args)

tmapValuesSubmit_size(x, args)

tmapValuesSubmit_area(x, args)

tmapValuesSubmit_xmod(x, args)

tmapValuesSubmit_ymod(x, args)

tmapValuesSubmit_angle(x, args)

tmapValuesSubmit_lwd(x, args)

tmapValuesSubmit_lty(x, args)

tmapValuesSubmit_shape(x, args)

tmapValuesSubmit_col_alpha(x, args)

tmapValuesSubmit_fill_alpha(x, args)

tmapValuesSubmit_bgcol_alpha(x, args)

tmapValuesSubmit_text(x, args)

tmapValuesSubmit_fontface(x, args)

tmapValuesScale_col(x, scale)

tmapValuesScale_fill(x, scale)

tmapValuesScale_bgcol(x, scale)

tmapValuesScale_size(x, scale)

tmapValuesScale_area(x, scale)

tmapValuesScale_lwd(x, scale)

tmapValuesScale_lty(x, scale)

tmapValuesScale_shape(x, scale)

tmapValuesScale_col_alpha(x, scale)

tmapValuesScale_fill_alpha(x, scale)

tmapValuesScale_bgcol_alpha(x, scale)

tmapValuesScale_text(x, scale)

tmapValuesScale_fontface(x, scale)

tmapValuesScale_xmod(x, scale)

tmapValuesScale_angle(x, scale)

tmapValuesColorize_col(x, pc)

tmapValuesColorize_fill(x, pc)

tmapValuesColorize_bgcol(x, pc)

tmapValuesColorize_size(x, pc)

tmapValuesColorize_area(x, pc)

tmapValuesColorize_lwd(x, pc)

tmapValuesColorize_lty(x, pc)

tmapValuesColorize_shape(x, pc)

tmapValuesColorize_col_alpha(x, pc)

tmapValuesColorize_fill_alpha(x, pc)

tmapValuesColorize_bgcol_alpha(x, pc)

tmapValuesColorize_text(x, pc)

tmapValuesColorize_fontface(x, pc)

tmapValuesColorize_xmod(x, pc)

tmapValuesColorize_angle(x, pc)

tmapSeq(s, n = NULL)

transform_values(x, lim, rng, power, scale, include.neutral = TRUE)

tmapValuesCVV_fill(x, value.na, n, range, scale, rep, o)

tmapValuesCVV_col(x, value.na, n, range, scale, rep, o)

tmapValuesCVV_bgcol(x, value.na, n, range, scale, rep, o)

tmapValuesCVV_size(x, value.na, n, range, scale, rep, o)

tmapValuesCVV_area(x, value.na, n, range, scale, rep, o)

tmapValuesCVV_lwd(x, value.na, n, range, scale, rep, o)

tmapValuesCVV_col_alpha(x, value.na, n, range, scale, rep, o)

tmapValuesCVV_fill_alpha(x, value.na, n, range, scale, rep, o)

tmapValuesCVV_bgcol_alpha(x, value.na, n, range, scale, rep, o)

tmapValuesCVV_area(x, value.na, n, range, scale, rep, o)

tmapValuesCVV_xmod(x, value.na, n, range, scale, rep, o)

tmapValuesCVV_ymod(x, value.na, n, range, scale, rep, o)

tmapValuesCVV_angle(x, value.na, n, range, scale, rep, o)

tmapValuesCVV_shape(x, value.na, n, range, scale, rep, o)

tmapValuesCVV_lty(x, value.na, n, range, scale, rep, o)

tmapValuesCVV_text(x, value.na, n, range, scale, rep, o)

tmapValuesCVV_fontface(x, value.na, n, range, scale, rep, o)

tmapValuesBVV_fill(x, value.na, m, n, scale, rep, o)

tmapValuesBVV_col(x, value.na, m, n, scale, rep, o)

tmapValuesBVV_bgcol(x, value.na, m, n, scale, rep, o)

tmapValuesCheck_num(x, is_var = TRUE)

tmapValuesIsDiv_num(x)

tmapValuesRange_num(x, n, isdiv)

tmapValuesVV_num(
  x,
  value.na,
  isdiv,
  n,
  dvalues,
  are_breaks,
  midpoint,
  range,
  scale,
  rep,
  o
)

tmapValuesSubmit_num(x, args)

tmapValuesScale_num(x, scale)

tmapValuesColorize_num(x, pc)

tmapValuesCVV_num(x, value.na, n, range, scale, rep, o)

tmapValuesCheck_skip(x, is_var)

tmapValuesIsDiv_skip(x)

tmapValuesRange_skip(x, n, isdiv)

tmapValuesVV_skip(
  x,
  value.na,
  isdiv,
  n,
  dvalues,
  are_breaks,
  midpoint,
  range,
  scale,
  rep,
  o
)

tmapValuesSubmit_skip(x, args)

tmapValuesScale_skip(x, scale)

tmapValuesColorize_skip(x, pc)

tmapValuesCVV_skip(x, value.na, n, range, scale, rep, o)

tmapTransCentroid(
  shpTM,
  xmod = NULL,
  ymod = NULL,
  ord__,
  plot.order,
  args,
  scale
)

tmapTransRaster(shpTM, ord__, plot.order, args)

tmapTransPolygons(shpTM, ord__, plot.order, args, scale)

tmapTransLines(shpTM, ord__, plot.order, args, scale)
}
\arguments{
\item{...}{args}

\item{subclass}{subclass}

\item{fill, col, shape, size, fill_alpha, col_alpha, lty, lwd, linejoin, lineend}{visual variables}

\item{scale}{scale}

\item{args}{args}

\item{shpTM}{shpTM}

\item{xmod, ymod}{xmod and ymod}

\item{ord__}{ord}

\item{plot.order}{plot.order}
}
\description{
Internal methods for tmap extensions
}
\keyword{datasets}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tmap_mode.R
\name{tmap_devel_mode}
\alias{tmap_devel_mode}
\title{Set the development mode}
\usage{
tmap_devel_mode(devel.mode)
}
\arguments{
\item{devel.mode}{logical value that determines the development mode.
If omitted then the development mode is toggled.}
}
\description{
When the so-called "development mode" is enabled, helpful messages and timings
are printed in the console
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tm_layers_iso.R
\name{tm_iso}
\alias{tm_iso}
\title{Map layer: iso (contour)}
\usage{
tm_iso(
  col = tm_const(),
  text = tm_vars(x = 1),
  ...,
  options_lines = opt_tm_lines(),
  options_labels = opt_tm_labels()
)
}
\arguments{
\item{col}{Visual variable that determines the color. See details.}

\item{text}{Visual variable that determines the text. See details.}

\item{...}{passed on to \code{\link[=tm_lines]{tm_lines()}} and \code{\link[=tm_labels_highlighted]{tm_labels_highlighted()}}. For the text color and alpha transparency of the text labels, please use \code{text_col} and \code{text_alpha} instead of \code{col} and \code{col_alpha}.}

\item{options_lines}{The options for \code{\link[=tm_lines]{tm_lines()}}}

\item{options_labels}{The options for \code{\link[=tm_labels_highlighted]{tm_labels_highlighted()}}}
}
\description{
Map layer that draws iso (contour) lines. Stack of \code{\link[=tm_lines]{tm_lines()}} and \link{tm_labels_highlighted}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tm_add_legend.R, R/tm_element.R,
%   R/tmapChart.R, R/tmapGpar.R, R/tmapGridComp.R, R/tmapGridComp_charts.R,
%   R/tmapOnLoad.R, R/tmapScaleAsIs.R, R/tmapScaleBivariate.R,
%   R/tmapScaleCategorical.R, R/tmapScaleIntervals.R, R/tmapScaleRank.R,
%   R/tmapScale_.R, R/tmapScale_defaults.R, R/tmapTrans.R
\name{toTitleCase}
\alias{toTitleCase}
\alias{tmap_internal}
\alias{tm_element}
\alias{tm_element_list}
\alias{tmapChartBinned}
\alias{tmapChartBinned2d}
\alias{tmapChartBinned2d_numcat}
\alias{tmapChartBinned2d_numnum}
\alias{tmapChartBinned2d_catcat}
\alias{tmapChartRaw}
\alias{tmapChartRaw_nna}
\alias{tmapChartNone}
\alias{tmapChartPass}
\alias{tmapChartBinned_categorical}
\alias{tmapChartBinned_numeric}
\alias{tmapGpar}
\alias{tmapTpar}
\alias{tmapGetCompGroupArgs}
\alias{lwd_to_mm}
\alias{.TMAP}
\alias{.TMAP_LEAFLET}
\alias{.TMAP_GRID}
\alias{tmapScaleAsIs}
\alias{tmapScaleBivariate}
\alias{tmapScaleCategorical}
\alias{tmapScaleIntervals}
\alias{tmapScaleRank}
\alias{tmapUsrCls}
\alias{format_aes_results}
\alias{tmapGetLegend}
\alias{chart_save}
\alias{data_type}
\alias{data_class}
\alias{tmapScale}
\alias{tmapScaleAuto}
\alias{tmapValuesCheck_col}
\alias{tmapValuesCheck_fill}
\alias{tmapValuesCheck_bgcol}
\alias{tmapValuesCheck_shape}
\alias{tmapValuesCheck_size}
\alias{tmapValuesCheck_area}
\alias{tmapValuesCheck_lwd}
\alias{tmapValuesCheck_lty}
\alias{tmapValuesCheck_xmod}
\alias{tmapValuesCheck_angle}
\alias{tmapValuesCheck_col_alpha}
\alias{tmapValuesCheck_fill_alpha}
\alias{tmapValuesCheck_bgcol_alpha}
\alias{tmapValuesCheck_text}
\alias{tmapValuesCheck_fontface}
\alias{tmapValuesIsDiv_fill}
\alias{tmapValuesIsDiv_col}
\alias{tmapValuesIsDiv_bgcol}
\alias{tmapValuesIsDiv_size}
\alias{tmapValuesIsDiv_area}
\alias{tmapValuesIsDiv_lwd}
\alias{tmapValuesIsDiv_lty}
\alias{tmapValuesIsDiv_col_alpha}
\alias{tmapValuesIsDiv_fill_alpha}
\alias{tmapValuesIsDiv_bgcol_alpha}
\alias{tmapValuesIsDiv_xmod}
\alias{tmapValuesIsDiv_angle}
\alias{tmapValuesIsDiv_shape}
\alias{tmapValuesIsDiv_text}
\alias{tmapValuesIsDiv_fontface}
\alias{tmapValuesRange_fill}
\alias{tmapValuesRange_col}
\alias{tmapValuesRange_bgcol}
\alias{tmapValuesRange_shape}
\alias{tmapValuesRange_lty}
\alias{tmapValuesRange_size}
\alias{tmapValuesRange_area}
\alias{tmapValuesRange_lwd}
\alias{tmapValuesRange_col_alpha}
\alias{tmapValuesRange_fill_alpha}
\alias{tmapValuesRange_bgcol_alpha}
\alias{tmapValuesRange_xmod}
\alias{tmapValuesRange_angle}
\alias{tmapValuesRange_text}
\alias{tmapValuesRange_fontface}
\alias{tmapValuesVV_fill}
\alias{tmapValuesVV_col}
\alias{tmapValuesVV_bgcol}
\alias{tmapValuesVV_shape}
\alias{tmapValuesVV_lty}
\alias{tmapValuesVV_size}
\alias{tmapValuesVV_area}
\alias{tmapValuesVV_lwd}
\alias{tmapValuesVV_col_alpha}
\alias{tmapValuesVV_fill_alpha}
\alias{tmapValuesVV_bgcol_alpha}
\alias{tmapValuesVV_xmod}
\alias{tmapValuesVV_angle}
\alias{tmapValuesVV_text}
\alias{tmapValuesVV_fontface}
\alias{tmapValuesSubmit_col}
\alias{tmapValuesSubmit_fill}
\alias{tmapValuesSubmit_bgcol}
\alias{tmapValuesSubmit_size}
\alias{tmapValuesSubmit_area}
\alias{tmapValuesSubmit_xmod}
\alias{tmapValuesSubmit_ymod}
\alias{tmapValuesSubmit_angle}
\alias{tmapValuesSubmit_lwd}
\alias{tmapValuesSubmit_lty}
\alias{tmapValuesSubmit_shape}
\alias{tmapValuesSubmit_col_alpha}
\alias{tmapValuesSubmit_fill_alpha}
\alias{tmapValuesSubmit_bgcol_alpha}
\alias{tmapValuesSubmit_text}
\alias{tmapValuesSubmit_fontface}
\alias{tmapValuesScale_col}
\alias{tmapValuesScale_fill}
\alias{tmapValuesScale_bgcol}
\alias{tmapValuesScale_size}
\alias{tmapValuesScale_area}
\alias{tmapValuesScale_lwd}
\alias{tmapValuesScale_lty}
\alias{tmapValuesScale_shape}
\alias{tmapValuesScale_col_alpha}
\alias{tmapValuesScale_fill_alpha}
\alias{tmapValuesScale_bgcol_alpha}
\alias{tmapValuesScale_text}
\alias{tmapValuesScale_fontface}
\alias{tmapValuesScale_xmod}
\alias{tmapValuesScale_angle}
\alias{tmapValuesColorize_col}
\alias{tmapValuesColorize_fill}
\alias{tmapValuesColorize_bgcol}
\alias{tmapValuesColorize_size}
\alias{tmapValuesColorize_area}
\alias{tmapValuesColorize_lwd}
\alias{tmapValuesColorize_lty}
\alias{tmapValuesColorize_shape}
\alias{tmapValuesColorize_col_alpha}
\alias{tmapValuesColorize_fill_alpha}
\alias{tmapValuesColorize_bgcol_alpha}
\alias{tmapValuesColorize_text}
\alias{tmapValuesColorize_fontface}
\alias{tmapValuesColorize_xmod}
\alias{tmapValuesColorize_angle}
\alias{tmapSeq}
\alias{transform_values}
\alias{tmapValuesCVV_fill}
\alias{tmapValuesCVV_col}
\alias{tmapValuesCVV_bgcol}
\alias{tmapValuesCVV_size}
\alias{tmapValuesCVV_area}
\alias{tmapValuesCVV_lwd}
\alias{tmapValuesCVV_col_alpha}
\alias{tmapValuesCVV_fill_alpha}
\alias{tmapValuesCVV_bgcol_alpha}
\alias{tmapValuesCVV_xmod}
\alias{tmapValuesCVV_ymod}
\alias{tmapValuesCVV_angle}
\alias{tmapValuesCVV_shape}
\alias{tmapValuesCVV_lty}
\alias{tmapValuesCVV_text}
\alias{tmapValuesCVV_fontface}
\alias{tmapValuesBVV_fill}
\alias{tmapValuesBVV_col}
\alias{tmapValuesBVV_bgcol}
\alias{tmapValuesCheck_num}
\alias{tmapValuesIsDiv_num}
\alias{tmapValuesRange_num}
\alias{tmapValuesVV_num}
\alias{tmapValuesSubmit_num}
\alias{tmapValuesScale_num}
\alias{tmapValuesColorize_num}
\alias{tmapValuesCVV_num}
\alias{tmapValuesCheck_skip}
\alias{tmapValuesIsDiv_skip}
\alias{tmapValuesRange_skip}
\alias{tmapValuesVV_skip}
\alias{tmapValuesSubmit_skip}
\alias{tmapValuesScale_skip}
\alias{tmapValuesColorize_skip}
\alias{tmapValuesCVV_skip}
\alias{tmapTransCentroid}
\alias{tmapTransRaster}
\alias{tmapTransPolygons}
\alias{tmapTransLines}
\title{Internal methods for tmap extensions}
\usage{
toTitleCase(x)

tm_element(..., subclass = NULL)

tm_element_list(...)

tmapChartBinned(chart, bin_colors, breaks_def, na.show, x1)

tmapChartBinned2d(chart, chart1, chart2)

tmapChartBinned2d_numcat(chart, chart1, chart2)

tmapChartBinned2d_numnum(chart, chart1, chart2)

tmapChartBinned2d_catcat(chart, chart1, chart2)

tmapChartRaw(chart, na.show, x1, ...)

tmapChartRaw_nna(chart, na.show, x1, ...)

tmapChartNone(chart, na.show, x1, ...)

tmapChartPass(chart, na.show, x1, ...)

tmapChartBinned_categorical(chart, bin_colors, breaks_def, na.show, x1)

tmapChartBinned_numeric(chart, bin_colors, breaks_def, na.show, x1)

tmapGpar(
  fill = NULL,
  col = NULL,
  shape = NULL,
  size = NULL,
  fill_alpha = NULL,
  col_alpha = NULL,
  lty = NULL,
  lwd = NULL,
  linejoin = NULL,
  lineend = NULL,
  ...
)

tmapTpar(...)

tmapGetCompGroupArgs(comp)

lwd_to_mm(value, unit = "bigpts")

.TMAP

.TMAP_LEAFLET

.TMAP_GRID

tmapScaleAsIs(
  x1,
  scale,
  legend,
  chart,
  o,
  aes,
  layer,
  layer_args,
  sortRev,
  bypass_ord,
  submit_legend = TRUE
)

tmapScaleBivariate(
  x1,
  x2,
  scale,
  legend,
  chart,
  o,
  aes,
  layer,
  layer_args,
  sortRev,
  bypass_ord,
  submit_legend = TRUE
)

tmapScaleCategorical(
  x1,
  scale,
  legend,
  chart,
  o,
  aes,
  layer,
  layer_args,
  sortRev,
  bypass_ord,
  submit_legend = TRUE
)

tmapScaleIntervals(
  x1,
  scale,
  legend,
  chart,
  o,
  aes,
  layer,
  layer_args,
  sortRev,
  bypass_ord,
  submit_legend = TRUE
)

tmapScaleRank(
  x1,
  scale,
  legend,
  chart,
  o,
  aes,
  layer,
  layer_args,
  sortRev,
  bypass_ord,
  submit_legend = TRUE
)

tmapUsrCls(x)

format_aes_results(values, ord = NULL, legend, chart)

tmapGetLegend(id)

chart_save(legend)

data_type(x)

data_class(x, check_for_color_class = FALSE, midpoint_enabled = FALSE)

tmapScale(aes, value, scale, legend, chart, free)

tmapScaleAuto(
  x1,
  scale,
  legend,
  chart,
  o,
  aes,
  layer,
  layer_args,
  sortRev,
  bypass_ord,
  submit_legend = TRUE,
  ...
)

tmapValuesCheck_col(x, is_var = TRUE)

tmapValuesCheck_fill(x, is_var = TRUE)

tmapValuesCheck_bgcol(x, is_var = TRUE)

tmapValuesCheck_shape(x, is_var = TRUE)

tmapValuesCheck_size(x, is_var = TRUE)

tmapValuesCheck_area(x, is_var = TRUE)

tmapValuesCheck_lwd(x, is_var = TRUE)

tmapValuesCheck_lty(x, is_var = TRUE)

tmapValuesCheck_xmod(x, is_var = TRUE)

tmapValuesCheck_angle(x, is_var = TRUE)

tmapValuesCheck_col_alpha(x, is_var = TRUE)

tmapValuesCheck_fill_alpha(x, is_var = TRUE)

tmapValuesCheck_bgcol_alpha(x, is_var = TRUE)

tmapValuesCheck_area(x, is_var = TRUE)

tmapValuesCheck_text(x, is_var = TRUE)

tmapValuesCheck_fontface(x, is_var = TRUE)

tmapValuesIsDiv_fill(x)

tmapValuesIsDiv_col(x)

tmapValuesIsDiv_bgcol(x)

tmapValuesIsDiv_size(x)

tmapValuesIsDiv_area(x)

tmapValuesIsDiv_lwd(x)

tmapValuesIsDiv_lty(x)

tmapValuesIsDiv_col_alpha(x)

tmapValuesIsDiv_fill_alpha(x)

tmapValuesIsDiv_bgcol_alpha(x)

tmapValuesIsDiv_area(x)

tmapValuesIsDiv_xmod(x)

tmapValuesIsDiv_angle(x)

tmapValuesIsDiv_shape(x)

tmapValuesIsDiv_text(x)

tmapValuesIsDiv_fontface(x)

tmapValuesRange_fill(x, n, isdiv)

tmapValuesRange_col(x, n, isdiv)

tmapValuesRange_bgcol(x, n, isdiv)

tmapValuesRange_shape(x, n, isdiv)

tmapValuesRange_lty(x, n, isdiv)

tmapValuesRange_size(x, n, isdiv)

tmapValuesRange_area(x, n, isdiv)

tmapValuesRange_lwd(x, n, isdiv)

tmapValuesRange_col_alpha(x, n, isdiv)

tmapValuesRange_fill_alpha(x, n, isdiv)

tmapValuesRange_bgcol_alpha(x, n, isdiv)

tmapValuesRange_area(x, n, isdiv)

tmapValuesRange_xmod(x, n, isdiv)

tmapValuesRange_angle(x, n, isdiv)

tmapValuesRange_text(x, n, isdiv)

tmapValuesRange_fontface(x, n, isdiv)

tmapValuesVV_fill(
  x,
  value.na,
  isdiv,
  n,
  dvalues,
  are_breaks,
  midpoint,
  range,
  scale,
  rep,
  o,
  aes = "fill"
)

tmapValuesVV_col(...)

tmapValuesVV_bgcol(...)

tmapValuesVV_shape(
  x,
  value.na,
  isdiv,
  n,
  dvalues,
  are_breaks,
  midpoint,
  range,
  scale,
  rep,
  o
)

tmapValuesVV_lty(
  x,
  value.na,
  isdiv,
  n,
  dvalues,
  are_breaks,
  midpoint,
  range,
  scale,
  rep,
  o
)

tmapValuesVV_size(
  x,
  value.na,
  isdiv,
  n,
  dvalues,
  are_breaks,
  midpoint,
  range,
  scale,
  rep,
  o
)

tmapValuesVV_area(...)

tmapValuesVV_lwd(...)

tmapValuesVV_col_alpha(...)

tmapValuesVV_fill_alpha(...)

tmapValuesVV_bgcol_alpha(...)

tmapValuesVV_area(...)

tmapValuesVV_xmod(...)

tmapValuesVV_angle(...)

tmapValuesVV_text(
  x,
  value.na,
  isdiv,
  n,
  dvalues,
  are_breaks,
  midpoint,
  range,
  scale,
  rep,
  o
)

tmapValuesVV_fontface(
  x,
  value.na,
  isdiv,
  n,
  dvalues,
  are_breaks,
  midpoint,
  range,
  scale,
  rep,
  o
)

tmapValuesSubmit_col(x, args)

tmapValuesSubmit_fill(x, args)

tmapValuesSubmit_bgcol(x, args)

tmapValuesSubmit_size(x, args)

tmapValuesSubmit_area(x, args)

tmapValuesSubmit_xmod(x, args)

tmapValuesSubmit_ymod(x, args)

tmapValuesSubmit_angle(x, args)

tmapValuesSubmit_lwd(x, args)

tmapValuesSubmit_lty(x, args)

tmapValuesSubmit_shape(x, args)

tmapValuesSubmit_col_alpha(x, args)

tmapValuesSubmit_fill_alpha(x, args)

tmapValuesSubmit_bgcol_alpha(x, args)

tmapValuesSubmit_text(x, args)

tmapValuesSubmit_fontface(x, args)

tmapValuesScale_col(x, scale)

tmapValuesScale_fill(x, scale)

tmapValuesScale_bgcol(x, scale)

tmapValuesScale_size(x, scale)

tmapValuesScale_area(x, scale)

tmapValuesScale_lwd(x, scale)

tmapValuesScale_lty(x, scale)

tmapValuesScale_shape(x, scale)

tmapValuesScale_col_alpha(x, scale)

tmapValuesScale_fill_alpha(x, scale)

tmapValuesScale_bgcol_alpha(x, scale)

tmapValuesScale_text(x, scale)

tmapValuesScale_fontface(x, scale)

tmapValuesScale_xmod(x, scale)

tmapValuesScale_angle(x, scale)

tmapValuesColorize_col(x, pc)

tmapValuesColorize_fill(x, pc)

tmapValuesColorize_bgcol(x, pc)

tmapValuesColorize_size(x, pc)

tmapValuesColorize_area(x, pc)

tmapValuesColorize_lwd(x, pc)

tmapValuesColorize_lty(x, pc)

tmapValuesColorize_shape(x, pc)

tmapValuesColorize_col_alpha(x, pc)

tmapValuesColorize_fill_alpha(x, pc)

tmapValuesColorize_bgcol_alpha(x, pc)

tmapValuesColorize_text(x, pc)

tmapValuesColorize_fontface(x, pc)

tmapValuesColorize_xmod(x, pc)

tmapValuesColorize_angle(x, pc)

tmapSeq(s, n = NULL)

transform_values(x, lim, rng, power, scale, include.neutral = TRUE)

tmapValuesCVV_fill(x, value.na, n, range, scale, rep, o)

tmapValuesCVV_col(x, value.na, n, range, scale, rep, o)

tmapValuesCVV_bgcol(x, value.na, n, range, scale, rep, o)

tmapValuesCVV_size(x, value.na, n, range, scale, rep, o)

tmapValuesCVV_area(x, value.na, n, range, scale, rep, o)

tmapValuesCVV_lwd(x, value.na, n, range, scale, rep, o)

tmapValuesCVV_col_alpha(x, value.na, n, range, scale, rep, o)

tmapValuesCVV_fill_alpha(x, value.na, n, range, scale, rep, o)

tmapValuesCVV_bgcol_alpha(x, value.na, n, range, scale, rep, o)

tmapValuesCVV_area(x, value.na, n, range, scale, rep, o)

tmapValuesCVV_xmod(x, value.na, n, range, scale, rep, o)

tmapValuesCVV_ymod(x, value.na, n, range, scale, rep, o)

tmapValuesCVV_angle(x, value.na, n, range, scale, rep, o)

tmapValuesCVV_shape(x, value.na, n, range, scale, rep, o)

tmapValuesCVV_lty(x, value.na, n, range, scale, rep, o)

tmapValuesCVV_text(x, value.na, n, range, scale, rep, o)

tmapValuesCVV_fontface(x, value.na, n, range, scale, rep, o)

tmapValuesBVV_fill(x, value.na, m, n, scale, rep, o)

tmapValuesBVV_col(x, value.na, m, n, scale, rep, o)

tmapValuesBVV_bgcol(x, value.na, m, n, scale, rep, o)

tmapValuesCheck_num(x, is_var = TRUE)

tmapValuesIsDiv_num(x)

tmapValuesRange_num(x, n, isdiv)

tmapValuesVV_num(
  x,
  value.na,
  isdiv,
  n,
  dvalues,
  are_breaks,
  midpoint,
  range,
  scale,
  rep,
  o
)

tmapValuesSubmit_num(x, args)

tmapValuesScale_num(x, scale)

tmapValuesColorize_num(x, pc)

tmapValuesCVV_num(x, value.na, n, range, scale, rep, o)

tmapValuesCheck_skip(x, is_var)

tmapValuesIsDiv_skip(x)

tmapValuesRange_skip(x, n, isdiv)

tmapValuesVV_skip(
  x,
  value.na,
  isdiv,
  n,
  dvalues,
  are_breaks,
  midpoint,
  range,
  scale,
  rep,
  o
)

tmapValuesSubmit_skip(x, args)

tmapValuesScale_skip(x, scale)

tmapValuesColorize_skip(x, pc)

tmapValuesCVV_skip(x, value.na, n, range, scale, rep, o)

tmapTransCentroid(
  shpTM,
  xmod = NULL,
  ymod = NULL,
  ord__,
  plot.order,
  args,
  scale
)

tmapTransRaster(shpTM, ord__, plot.order, args)

tmapTransPolygons(shpTM, ord__, plot.order, args, scale)

tmapTransLines(shpTM, ord__, plot.order, args, scale)
}
\arguments{
\item{...}{args}

\item{subclass}{subclass}

\item{fill, col, shape, size, fill_alpha, col_alpha, lty, lwd, linejoin, lineend}{visual variables}

\item{scale}{scale}

\item{args}{args}

\item{shpTM}{shpTM}

\item{xmod, ymod}{xmod and ymod}

\item{ord__}{ord}

\item{plot.order}{plot.order}
}
\description{
Internal methods for tmap extensions
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tm_layers_iso.R
\name{tm_iso}
\alias{tm_iso}
\title{Map layer: iso (contour)}
\usage{
tm_iso(
  col = tm_const(),
  text = tm_vars(x = 1),
  ...,
  options_lines = opt_tm_lines(),
  options_labels = opt_tm_labels()
)
}
\arguments{
\item{col}{Visual variable that determines the color. See details. \emph{Unit:} Color -- a color name, hex string.}

\item{text}{Visual variable that determines the text. See details. \emph{Unit:} Character string.}

\item{...}{passed on to \code{\link[=tm_lines]{tm_lines()}} and \code{\link[=tm_labels_highlighted]{tm_labels_highlighted()}}. For the text color and alpha transparency of the text labels, please use \code{text_col} and \code{text_alpha} instead of \code{col} and \code{col_alpha}.}

\item{options_lines}{The options for \code{\link[=tm_lines]{tm_lines()}}}

\item{options_labels}{The options for \code{\link[=tm_labels_highlighted]{tm_labels_highlighted()}}}
}
\description{
Map layer that draws iso (contour) lines. Stack of \code{\link[=tm_lines]{tm_lines()}} and \link{tm_labels_highlighted}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tmap_options.R
\name{tmapAddLayerOptions}
\alias{tmapAddLayerOptions}
\title{Internal tmap function to add a default value for the layer functions. Deprecated}
\usage{
tmapAddLayerOptions(option, id, value)
}
\arguments{
\item{option, }{one of: \code{"value.const"}, \code{"value.na"}, \code{"value.blank"}, \code{"values.var"}, \code{'values.range'}, \code{"value.neutral"}, \code{"scales.var"}}

\item{id}{name of the visual variable with layer, in the format \code{"x.y"},
where \code{x} is the visual variable and \code{y} is the layer.
It is also possible to set \code{x} only; then it applies to all layer functions.}

\item{value}{value}
}
\description{
Internal tmap function to add a default value for the layer functions. Deprecated. Use \code{\link[=tmapSubmitOptions]{tmapSubmitOptions()}} instead.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tm_component_functions.R
\name{tm_title}
\alias{tm_title}
\alias{tm_title_in}
\alias{tm_title_out}
\title{Map component: title}
\usage{
tm_title(
  text,
  size,
  color,
  padding,
  fontface,
  fontfamily,
  alpha,
  stack,
  just,
  frame,
  frame.color,
  frame.alpha,
  frame.lwd,
  frame.r,
  bg,
  bg.color,
  bg.alpha,
  position,
  group_id,
  width,
  height,
  z
)

tm_title_in(text, ..., position = tm_pos_in("left", "top"))

tm_title_out(text, ..., position = tm_pos_out("center", "top"))
}
\arguments{
\item{text}{text}

\item{size}{font size}

\item{color}{font color}

\item{padding}{padding}

\item{fontface}{font face, bold, italic}

\item{fontfamily}{font family}

\item{alpha}{alpha transparency of the text}

\item{stack}{stack with other map components, either \code{"vertical"} or \code{"horizontal"}.}

\item{just}{just}

\item{frame}{frame should a frame be drawn?}

\item{frame.color}{frame color}

\item{frame.alpha}{frame alpha transparancy}

\item{frame.lwd}{frame line width}

\item{frame.r}{Radius of the rounded frame corners. 0 means no rounding.}

\item{bg}{Show background?}

\item{bg.color}{Background color}

\item{bg.alpha}{Background transparency}

\item{position}{The position specification of the component: an object created with \code{tm_pos_in()} or \code{tm_pos_out()}. Or, as a shortcut, a vector of two values, specifying the x and y coordinates. The first is \code{"left"}, \code{"center"} or \code{"right"} (or upper case, meaning tighter to the map frame), the second \code{"top"}, \code{"center"} or \code{"bottom"}. Numeric values are also supported, where 0, 0 means left bottom and 1, 1 right top. See also vignette: \href{https://r-tmap.github.io/tmap/articles/adv_positions}{Positioning of components}. In case multiple components should be combined (stacked), use \code{group_id} and specify \code{component} in \code{\link[=tm_components]{tm_components()}}.}

\item{group_id}{Component group id name. All components (e.g. legends, titles, etc) with the same \code{group_id} will be grouped. The specifications of how they are placed (e.g. stacking, margins etc.) are determined in \code{\link[=tm_components]{tm_components()}} where its argument \code{id} should correspond to \code{group_id}.}

\item{width, height}{width and height of the component.}

\item{z}{z index, e.g. the place of the component relative to the other componets}

\item{...}{passed on to \code{tm_title()}}
}
\description{
Map component that adds a title
}
\seealso{
\href{https://r-tmap.github.io/tmap/articles/basics_components}{Components}, \href{https://r-tmap.github.io/tmap/articles/adv_positions}{Positioning of components}, \href{https://r-tmap.github.io/tmap/articles/adv_comp_group}{Grouping of components}, \href{https://r-tmap.github.io/tmap/articles/examples_choro_NLD}{Choropleth (Netherlands)}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tm_legend.R
\name{tm_legend}
\alias{tm_legend}
\alias{tm_legend_hide}
\alias{tm_legend_combine}
\alias{tm_legend_bivariate}
\title{Legend}
\usage{
tm_legend(
  title,
  show,
  orientation,
  reverse,
  na.show,
  position,
  group_id,
  width,
  height,
  z,
  title.color,
  title.size,
  title.fontface,
  title.fontfamily,
  title.alpha,
  title.padding,
  title.align,
  text.color,
  text.size,
  text.fontface,
  text.fontfamily,
  text.alpha,
  format,
  frame,
  frame.lwd,
  frame.r,
  bg,
  bg.color,
  bg.alpha,
  absolute_fontsize,
  item.height,
  item.width,
  item.space,
  item.na.height,
  item.na.width,
  item.na.space,
  item.shape,
  ticks,
  ticks.disable.na,
  ticks.col,
  ticks.lwd,
  margins,
  item_text.margin,
  ...
)

tm_legend_hide()

tm_legend_combine(variable)

tm_legend_bivariate(
  xlab,
  ylab,
  xlab.color,
  xlab.size,
  xlab.rot,
  xlab.fontface,
  xlab.fontfamily,
  xlab.alpha,
  xlab.padding,
  xlab.align,
  ylab.color,
  ylab.size,
  ylab.rot,
  ylab.fontface,
  ylab.fontfamily,
  ylab.alpha,
  ylab.padding,
  ylab.align,
  ...
)
}
\arguments{
\item{title}{Legend title}

\item{show}{Show legend?}

\item{orientation}{Orientation of the legend: \code{"portrait"} or \code{"landscape"}}

\item{reverse}{Should the legend items be reversed?}

\item{na.show}{Show \code{NA} values in legend?}

\item{position}{The position specification of the component: an object created with \code{tm_pos_in()} or \code{tm_pos_out()}. Or, as a shortcut, a vector of two values, specifying the x and y coordinates. The first is \code{"left"}, \code{"center"} or \code{"right"} (or upper case, meaning tighter to the map frame), the second \code{"top"}, \code{"center"} or \code{"bottom"}. Numeric values are also supported, where 0, 0 means left bottom and 1, 1 right top. See also \href{https://r-tmap.github.io/tmap/articles/adv_positions}{vignette about positioning}. In case multiple components should be combined (stacked), use \code{group_id} and specify \code{component} in \code{\link[=tm_components]{tm_components()}}.}

\item{group_id}{Component group id name. All components (e.g. legends, titles, etc) with the same \code{group_id} will be grouped. The specifications of how they are placed (e.g. stacking, margins etc.) are determined in \code{\link[=tm_components]{tm_components()}} where its argument \code{id} should correspond to \code{group_id}.}

\item{width}{Width of the legend. Units are 'text line heights'. In case a negative number is specified, the units are (approximate) pixels. The relation between these two is configured via the option \code{absolute_fontsize}.}

\item{height}{Height of the legend. Units are 'text line heights'. In case a negative number is specified, the units are (approximate) pixels. The relation between these two is configured via the option \code{absolute_fontsize}.}

\item{z}{z index, e.g. the place of the component relative to the other componets}

\item{title.color}{The color of the title of the legend.}

\item{title.size}{The size of the title of the legend.}

\item{title.fontface}{The font face of the title of the legend. See \code{graphics::par}, option 'font'.}

\item{title.fontfamily}{The font family of the title of the legend. See \code{graphics::par}, option 'family'.}

\item{title.alpha}{The alpha transparency of the title of the legend.}

\item{title.padding}{The padding of the title of the legend. A vector of 4 values: bottom, left, top, right. The unit is the device height (for bottom and top) or width (for left and right).}

\item{title.align}{The align of the title of the legend.}

\item{text.color}{The color of the text of the legend.}

\item{text.size}{The size of the text of the legend.}

\item{text.fontface}{The font face of the text of the legend. See \code{graphics::par}, option 'font'.}

\item{text.fontfamily}{The font family of the text of the legend. See \code{graphics::par}, option 'family'.}

\item{text.alpha}{The alpha transparency of the text of the legend.}

\item{format}{Not used anymore: use the format argument of the \verb{tm_scale_*()} functions instead.}

\item{frame}{frame should a frame be drawn?}

\item{frame.lwd}{frame line width}

\item{frame.r}{Radius of the rounded frame corners. 0 means no rounding.}

\item{bg}{Show background?}

\item{bg.color}{The color of the bg of the legend.}

\item{bg.alpha}{The alpha transparency of the bg of the legend.}

\item{absolute_fontsize}{The absolute fontsize of the legend. So far, only used to calculate legend dimensions}

\item{item.height}{The height of the item of the legend.}

\item{item.width}{The width of the item of the legend.}

\item{item.space}{The space of the item of the legend. In terms of number of text line heights.}

\item{item.na.height}{The height of the na item of the legend.}

\item{item.na.width}{The width of the na item of the legend.}

\item{item.na.space}{The space of the na item of the legend. In terms of number of text line heights.}

\item{item.shape}{The shape of the item of the legend.}

\item{ticks}{List of vectors of size 2 that determines a tick mark line (for portrait legends). The values are the y-values of begin and endpoint of a tick mark. For a solid line, only one vector is required, for dashed lines one for each dash. See \href{https://r-tmap.github.io/tmap/articles/versus_ggplot2#mimicking-ggplot2-layout}{ggplot2 style example}.}

\item{ticks.disable.na}{Remove ticks for \code{NA} values}

\item{ticks.col}{The color of the ticks of the legend.}

\item{ticks.lwd}{The line width of the ticks of the legend. See \code{graphics::par}, option 'lwd'.}

\item{margins}{The margins of the legend. A vector of 4 values: bottom, left, top, right. The unit is the device height (for bottom and top) or width (for left and right).}

\item{item_text.margin}{The margin of the space between item and text of the legend.}

\item{...}{visual values, e.g. \code{col}, \code{fill}, \code{lwd}, can be specified. If so, they overrule the default visual values, which are determined by the drawn map objects (e.g. polygons)}

\item{variable}{visual (or transformation) variable to combine the legend with: e.g. \code{"fill"} or \code{"size"}}

\item{xlab}{label for the x dimension (rows)}

\item{ylab}{label for the y dimension (columns)}

\item{xlab.color}{The color of the xlab of the legend.}

\item{xlab.size}{The size of the xlab of the legend.}

\item{xlab.rot}{The rot of the xlab of the legend.}

\item{xlab.fontface}{The font face of the xlab of the legend. See \code{graphics::par}, option 'font'.}

\item{xlab.fontfamily}{The font family of the xlab of the legend. See \code{graphics::par}, option 'family'.}

\item{xlab.alpha}{The alpha transparency of the xlab of the legend.}

\item{xlab.padding}{The padding of the xlab of the legend. A vector of 4 values: bottom, left, top, right. The unit is the device height (for bottom and top) or width (for left and right).}

\item{xlab.align}{The align of the xlab of the legend.}

\item{ylab.color}{The color of the ylab of the legend.}

\item{ylab.size}{The size of the ylab of the legend.}

\item{ylab.rot}{The rot of the ylab of the legend.}

\item{ylab.fontface}{The font face of the ylab of the legend. See \code{graphics::par}, option 'font'.}

\item{ylab.fontfamily}{The font family of the ylab of the legend. See \code{graphics::par}, option 'family'.}

\item{ylab.alpha}{The alpha transparency of the ylab of the legend.}

\item{ylab.padding}{The padding of the ylab of the legend. A vector of 4 values: bottom, left, top, right. The unit is the device height (for bottom and top) or width (for left and right).}

\item{ylab.align}{The align of the ylab of the legend.}
}
\value{
A tm_legend component
}
\description{
Legend specification
}
\examples{
# Example using different settings from tm_legend()

tm_shape(World) +
  tm_polygons(
    fill = "HPI",
    fill.legend = tm_legend(
      title = "Home Price Index",
      title.color = "orange",
      bg.color = "purple",
      show = TRUE
	),
    id = "name",
    # Format the labels using dollar sign
    fill.scale = tm_scale_intervals(
      label.format = function(x) format(x, big.mark = " ")),
)
}
\seealso{
\href{https://r-tmap.github.io/tmap/articles/basics_legends}{Vignette about legends}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tm_vars.R
\name{tm_vars}
\alias{tm_vars}
\title{tmap function to specify variables}
\usage{
tm_vars(
  x = NA,
  dimvalues = NULL,
  n = NA,
  multivariate = FALSE,
  animate = FALSE
)
}
\arguments{
\item{x}{variable names, variable indices, or a dimension name}

\item{dimvalues}{dimension values}

\item{n}{if specified the first \code{n} variables are taken (or the first \code{n} dimension values)}

\item{multivariate}{in case multiple variables are specified, should they serve as facets (FALSE) or as a multivariate visual variable?}

\item{animate}{should the variable(s) be animated? (experimental)}
}
\description{
tmap function to specify all variables in the shape object
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tmapShape.R
\name{tmapReproject}
\alias{tmapReproject}
\title{Internal method that reprojects a shape object}
\usage{
tmapReproject(shp, tmapID, bbox, ..., crs)
}
\arguments{
\item{shp}{shape}

\item{tmapID}{tmap id values}

\item{bbox}{bounding box}

\item{...}{other args}

\item{crs}{crs}
}
\description{
Internal method that reprojects a shape object
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tm_legend.R
\name{tm_legend}
\alias{tm_legend}
\alias{tm_legend_hide}
\alias{tm_legend_combine}
\alias{tm_legend_bivariate}
\title{Legend}
\usage{
tm_legend(
  title,
  show,
  orientation,
  reverse,
  na.show,
  position,
  group_id,
  width,
  height,
  z,
  title.color,
  title.size,
  title.fontface,
  title.fontfamily,
  title.alpha,
  title.padding,
  title.align,
  text.color,
  text.size,
  text.fontface,
  text.fontfamily,
  text.alpha,
  format,
  frame,
  frame.color,
  frame.alpha,
  frame.lwd,
  frame.r,
  bg,
  bg.color,
  bg.alpha,
  absolute_fontsize,
  item.height,
  item.width,
  item.space,
  item.na.height,
  item.na.width,
  item.na.space,
  item.shape,
  ticks,
  ticks.disable.na,
  ticks.col,
  ticks.lwd,
  margins,
  item_text.margin,
  ...
)

tm_legend_hide()

tm_legend_combine(variable)

tm_legend_bivariate(
  xlab,
  ylab,
  xlab.color,
  xlab.size,
  xlab.rot,
  xlab.fontface,
  xlab.fontfamily,
  xlab.alpha,
  xlab.padding,
  xlab.align,
  ylab.color,
  ylab.size,
  ylab.rot,
  ylab.fontface,
  ylab.fontfamily,
  ylab.alpha,
  ylab.padding,
  ylab.align,
  ...
)
}
\arguments{
\item{title}{Legend title}

\item{show}{Show legend?}

\item{orientation}{Orientation of the legend: \code{"portrait"} or \code{"landscape"}}

\item{reverse}{Should the legend items be reversed?}

\item{na.show}{Show \code{NA} values in legend?}

\item{position}{The position specification of the component: an object created with \code{tm_pos_in()} or \code{tm_pos_out()}. Or, as a shortcut, a vector of two values, specifying the x and y coordinates. The first is \code{"left"}, \code{"center"} or \code{"right"} (or upper case, meaning tighter to the map frame), the second \code{"top"}, \code{"center"} or \code{"bottom"}. Numeric values are also supported, where 0, 0 means left bottom and 1, 1 right top. See also vignette: \href{https://r-tmap.github.io/tmap/articles/adv_positions}{Positioning of components}. In case multiple components should be combined (stacked), use \code{group_id} and specify \code{component} in \code{\link[=tm_components]{tm_components()}}.}

\item{group_id}{Component group id name. All components (e.g. legends, titles, etc) with the same \code{group_id} will be grouped. The specifications of how they are placed (e.g. stacking, margins etc.) are determined in \code{\link[=tm_components]{tm_components()}} where its argument \code{id} should correspond to \code{group_id}.}

\item{width}{Width of the legend. Units are 'text line heights'. In case a negative number is specified, the units are (approximate) pixels. The relation between these two is configured via the option \code{absolute_fontsize}.}

\item{height}{Height of the legend. Units are 'text line heights'. In case a negative number is specified, the units are (approximate) pixels. The relation between these two is configured via the option \code{absolute_fontsize}.}

\item{z}{z index, e.g. the place of the component relative to the other componets}

\item{title.color}{The color of the title of the legend.}

\item{title.size}{The size of the title of the legend.}

\item{title.fontface}{The font face of the title of the legend. See \code{graphics::par}, option 'font'.}

\item{title.fontfamily}{The font family of the title of the legend. See \code{graphics::par}, option 'family'.}

\item{title.alpha}{The alpha transparency of the title of the legend.}

\item{title.padding}{The padding of the title of the legend. A vector of 4 values: bottom, left, top, right. The unit is the device height (for bottom and top) or width (for left and right).}

\item{title.align}{The align of the title of the legend.}

\item{text.color}{The color of the text of the legend.}

\item{text.size}{The size of the text of the legend.}

\item{text.fontface}{The font face of the text of the legend. See \code{graphics::par}, option 'font'.}

\item{text.fontfamily}{The font family of the text of the legend. See \code{graphics::par}, option 'family'.}

\item{text.alpha}{The alpha transparency of the text of the legend.}

\item{format}{Not used anymore: use the format argument of the \verb{tm_scale_*()} functions instead.}

\item{frame}{frame should a frame be drawn?}

\item{frame.color}{frame color}

\item{frame.alpha}{frame alpha transparancy}

\item{frame.lwd}{frame line width}

\item{frame.r}{Radius of the rounded frame corners. 0 means no rounding.}

\item{bg}{Show background?}

\item{bg.color}{The color of the bg of the legend.}

\item{bg.alpha}{The alpha transparency of the bg of the legend.}

\item{absolute_fontsize}{The absolute fontsize of the legend. So far, only used to calculate legend dimensions}

\item{item.height}{The height of the item of the legend.}

\item{item.width}{The width of the item of the legend.}

\item{item.space}{The space of the item of the legend. In terms of number of text line heights.}

\item{item.na.height}{The height of the na item of the legend.}

\item{item.na.width}{The width of the na item of the legend.}

\item{item.na.space}{The space of the na item of the legend. In terms of number of text line heights.}

\item{item.shape}{The shape of the item of the legend.}

\item{ticks}{List of vectors of size 2 that determines a tick mark line (for portrait legends). The values are the y-values of begin and endpoint of a tick mark. For a solid line, only one vector is required, for dashed lines one for each dash. See vignette: \href{https://r-tmap.github.io/tmap/articles/versus_ggplot2}{tmap vs ggplot2}.}

\item{ticks.disable.na}{Remove ticks for \code{NA} values}

\item{ticks.col}{The color of the ticks of the legend.}

\item{ticks.lwd}{The line width of the ticks of the legend. See \code{graphics::par}, option 'lwd'.}

\item{margins}{The margins of the legend. A vector of 4 values: bottom, left, top, right. The unit is the device height (for bottom and top) or width (for left and right).}

\item{item_text.margin}{The margin of the space between item and text of the legend.}

\item{...}{visual values, e.g. \code{col}, \code{fill}, \code{lwd}, can be specified. If so, they overrule the default visual values, which are determined by the drawn map objects (e.g. polygons)}

\item{variable}{visual (or transformation) variable to combine the legend with: e.g. \code{"fill"} or \code{"size"}}

\item{xlab}{label for the x dimension (rows)}

\item{ylab}{label for the y dimension (columns)}

\item{xlab.color}{The color of the xlab of the legend.}

\item{xlab.size}{The size of the xlab of the legend.}

\item{xlab.rot}{The rot of the xlab of the legend.}

\item{xlab.fontface}{The font face of the xlab of the legend. See \code{graphics::par}, option 'font'.}

\item{xlab.fontfamily}{The font family of the xlab of the legend. See \code{graphics::par}, option 'family'.}

\item{xlab.alpha}{The alpha transparency of the xlab of the legend.}

\item{xlab.padding}{The padding of the xlab of the legend. A vector of 4 values: bottom, left, top, right. The unit is the device height (for bottom and top) or width (for left and right).}

\item{xlab.align}{The align of the xlab of the legend.}

\item{ylab.color}{The color of the ylab of the legend.}

\item{ylab.size}{The size of the ylab of the legend.}

\item{ylab.rot}{The rot of the ylab of the legend.}

\item{ylab.fontface}{The font face of the ylab of the legend. See \code{graphics::par}, option 'font'.}

\item{ylab.fontfamily}{The font family of the ylab of the legend. See \code{graphics::par}, option 'family'.}

\item{ylab.alpha}{The alpha transparency of the ylab of the legend.}

\item{ylab.padding}{The padding of the ylab of the legend. A vector of 4 values: bottom, left, top, right. The unit is the device height (for bottom and top) or width (for left and right).}

\item{ylab.align}{The align of the ylab of the legend.}
}
\value{
A tm_legend component
}
\description{
Legend specification
}
\examples{
# Example using different settings from tm_legend()

tm_shape(World) +
  tm_polygons(
    fill = "HPI",
    fill.legend = tm_legend(
      title = "Home Price Index",
      title.color = "orange",
      bg.color = "purple",
      show = TRUE
	),
    id = "name",
    # Format the labels using dollar sign
    fill.scale = tm_scale_intervals(
      label.format = function(x) format(x, big.mark = " ")),
)
}
\seealso{
\href{https://r-tmap.github.io/tmap/articles/adv_legends}{Legends (advanced)}, \href{https://r-tmap.github.io/tmap/articles/adv_positions}{Positioning of components}, \href{https://r-tmap.github.io/tmap/articles/examples_biv_choro}{Bivariate choropleth} \href{https://r-tmap.github.io/tmap/articles/basics_components}{Components}, \href{https://r-tmap.github.io/tmap/articles/adv_positions}{Positioning of components}, \href{https://r-tmap.github.io/tmap/articles/adv_comp_group}{Grouping of components}, \href{https://r-tmap.github.io/tmap/articles/examples_choro_NLD}{Choropleth (Netherlands)}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/shiny.R
\name{renderTmap}
\alias{renderTmap}
\alias{tmapOutput}
\alias{tmapProxy}
\alias{tm_remove_layer}
\title{Wrapper functions for using \strong{tmap} in \strong{shiny}}
\usage{
renderTmap(
  expr,
  env = parent.frame(),
  quoted = FALSE,
  execOnResize = TRUE,
  mode = NA
)

tmapOutput(outputId, width = "100\%", height = 400, mode = NA)

tmapProxy(mapId, session = shiny::getDefaultReactiveDomain(), x, mode = NA)

tm_remove_layer(zindex)
}
\arguments{
\item{expr}{A tmap object. A tmap object is created with \code{\link[=qtm]{qtm()}} or by stacking \code{\link{tmap-element}}s.}

\item{env}{The environment in which to evaluate expr}

\item{quoted}{Is \code{expr} a quoted expression (with \code{quote()})? This is useful if you want to save an expression in a variable}

\item{execOnResize}{If \code{TRUE} (default), when the plot is resized, the map is regenerated. When set to \code{FALSE} the map is rescaled: the aspect ratio is kept, but the layout will be less desirable.}

\item{mode}{tmap mode, see \code{\link[=tmap_mode]{tmap_mode()}} If not defined, the current mode is used}

\item{outputId}{Output variable to read from}

\item{width, height}{the width and height of the map}

\item{mapId}{single-element character vector indicating the output ID of the map to modify (if invoked from a Shiny module, the namespace will be added automatically)}

\item{session}{the Shiny session object to which the map belongs; usually the default value will suffice}

\item{x}{the tmap object that specifies the added and removed layers.}

\item{zindex}{the z index of the pane in which the layer is contained that is going to be removed. It is recommended to specify the \code{zindex} for this layer when creating the map (inside \code{renderTmap()}).}
}
\description{
\itemize{
\item \code{tmapOutput()} creates a UI element
\item \code{renderTmap()} renders a tmap map
\item \code{tmapProxy()} updates a tmap map in \code{view} mode
}

Adding layers is as usual via the map layer functions like \code{\link[=tm_polygons]{tm_polygons()}}.
Removing layers can be done, removing with \code{tm_remove_layer()}.
}
\details{
Two features from tmap are not (yet) supported in Shiny: small multiples (facets) and colored backgrounds (argument \code{bg.color} of \code{\link[=tm_layout]{tm_layout()}}).
Workarounds for small multiples: create multiple independent maps or specify \code{as.layers = TRUE} in \code{\link[=tm_facets]{tm_facets()}}.
}
\examples{
if (interactive() && require("shiny")) {

	data(World)
	world_vars <- setdiff(names(World), c("iso_a3", "name", "sovereignt", "geometry"))

	current.mode <- tmap_mode("plot")

	shinyApp(
		ui = fluidPage(
			tmapOutput("map", height = "600px"),
			selectInput("var", "Variable", world_vars)
		),
		server <- function(input, output, session) {
			output$map <- renderTmap({
				tm_shape(World) +
					tm_polygons(input$var, zindex = 401)
			})
		}
	)

	tmap_mode("view")

	shinyApp(
		ui = fluidPage(
			tmapOutput("map", height = "600px"),
			selectInput("var", "Variable", world_vars)
		),
		server <- function(input, output, session) {
			output$map <- renderTmap({
				tm_shape(World, id = "iso_a3") +
					tm_polygons(fill = world_vars[1], zindex = 401)
			})
			observe({
				var <- input$var
				tmapProxy("map", session, {
					tm_remove_layer(401) +
						tm_shape(World, id = "iso_a3") +
						tm_polygons(fill = var, zindex = 401)
				})
			})
		},options = list(launch.browser=TRUE)
	)

	tmap_mode(current.mode)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tm_vars.R
\name{tm_vars}
\alias{tm_vars}
\title{tmap function to specify variables}
\usage{
tm_vars(
  x = NA,
  dimvalues = NULL,
  n = NA,
  multivariate = FALSE,
  animate = FALSE
)
}
\arguments{
\item{x}{variable names, variable indices, or a dimension name}

\item{dimvalues}{dimension values}

\item{n}{if specified the first \code{n} variables are taken (or the first \code{n} dimension values)}

\item{multivariate}{in case multiple variables are specified, should they serve as facets (FALSE) or as a multivariate visual variable?}

\item{animate}{should the variable(s) be animated? (experimental)}
}
\description{
tmap function to specify all variables in the shape object
}
\seealso{
\href{https://r-tmap.github.io/tmap/articles/basics_scales}{Scales}, \href{https://r-tmap.github.io/tmap/articles/basics_vv}{Visual variables}, \href{https://r-tmap.github.io/tmap/articles/adv_multivariate}{Multivariate visual variables}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tm_layers_sf.R
\name{tm_sf}
\alias{tm_sf}
\alias{opt_tm_sf}
\title{Map layer: simple features}
\usage{
tm_sf(
  fill = tm_const(),
  fill.scale = tm_scale(),
  fill.legend = tm_legend(),
  fill.free = NA,
  col = tm_const(),
  col.scale = tm_scale(),
  col.legend = tm_legend(),
  col.free = NA,
  size = tm_const(),
  size.scale = tm_scale(),
  size.legend = tm_legend(),
  size.free = NA,
  shape = tm_const(),
  shape.scale = tm_scale(),
  shape.legend = tm_legend(),
  shape.free = NA,
  lwd = tm_const(),
  lwd.scale = tm_scale(),
  lwd.legend = tm_legend(),
  lwd.free = NA,
  lty = tm_const(),
  lty.scale = tm_scale(),
  lty.legend = tm_legend(),
  lty.free = NA,
  fill_alpha = tm_const(),
  fill_alpha.scale = tm_scale(),
  fill_alpha.legend = tm_legend(),
  fill_alpha.free = NA,
  col_alpha = tm_const(),
  col_alpha.scale = tm_scale(),
  col_alpha.legend = tm_legend(),
  col_alpha.free = NA,
  linejoin = "round",
  lineend = "round",
  plot.order.list = list(polygons = tm_plot_order("AREA"), lines =
    tm_plot_order("LENGTH"), points = tm_plot_order("size")),
  options = opt_tm_sf(),
  zindex = NA,
  group = NA,
  group.control = "check",
  ...
)

opt_tm_sf(
  polygons.only = "yes",
  lines.only = "yes",
  points_only = "yes",
  point_per = "feature",
  points.icon.scale = 3,
  points.just = NA,
  points.grob.dim = c(width = 48, height = 48, render.width = 256, render.height = 256)
)
}
\arguments{
\item{fill, fill.scale, fill.legend, fill.free}{Visual variable that determines the fill color. See details.}

\item{col, col.scale, col.legend, col.free}{Visual variable that determines the color. See details.}

\item{size, size.scale, size.legend, size.free}{Visual variable that determines the size. See details.}

\item{shape, shape.scale, shape.legend, shape.free}{Visual variable that determines the shape. See details.}

\item{lwd, lwd.scale, lwd.legend, lwd.free}{Visual variable that determines the line width. See details.}

\item{lty, lty.scale, lty.legend, lty.free}{Visual variable that determines the line type. See details.}

\item{fill_alpha, fill_alpha.scale, fill_alpha.legend, fill_alpha.free}{Visual variable that determines the fill color transparency. See details.}

\item{col_alpha, col_alpha.scale, col_alpha.legend, col_alpha.free}{Visual variable that determines the color transparency. See details.}

\item{linejoin, lineend}{line join and line end. See \link[grid:gpar]{gpar()} for details.}

\item{plot.order.list}{Specification in which order the spatial features are drawn.
This consists of a list of three elementary geometry types: for polygons, lines and, points.
For each of these types, which are drawn in that order, a \code{\link[=tm_plot_order]{tm_plot_order()}} is required.}

\item{options}{options passed on to the corresponding \verb{opt_<layer_function>} function}

\item{zindex}{Map layers are drawn on top of each other. The \code{zindex} numbers
(one for each map layer) determines the stacking order.
By default the map layers are drawn in the order they are called.}

\item{group}{Name of the group to which this layer belongs. This is only
relevant in view mode, where layer groups can be switched (see \code{group.control})}

\item{group.control}{In view mode, the group control determines how layer groups
can be switched on and off. Options: \code{"radio"} for radio buttons
(meaning only one group can be shown), \code{"check"} for check boxes
(so multiple groups can be shown), and \code{"none"} for no control
(the group cannot be (de)selected).}

\item{...}{passed on to \code{\link[=tm_polygons]{tm_polygons()}}, \code{\link[=tm_lines]{tm_lines()}}, and \code{\link[=tm_dots]{tm_dots()}}}

\item{polygons.only}{should only polygon geometries of the shape object (defined in \code{\link[=tm_shape]{tm_shape()}}) be plotted? By default \code{"ifany"}, which means \code{TRUE} in case a geometry collection is specified.}

\item{lines.only}{should only line geometries of the shape object (defined in \code{\link[=tm_shape]{tm_shape()}}) be plotted, or also other geometry types (like polygons)? By default \code{"ifany"}, which means \code{TRUE} in case a geometry collection is specified.}

\item{points_only}{should only point geometries of the shape object (defined in \code{\link[=tm_shape]{tm_shape()}}) be plotted? By default \code{"ifany"}, which means \code{TRUE} in case a geometry collection is specified.}

\item{point_per}{specification of how spatial points are mapped when the geometry is a multi line or a multi polygon. One of \code{"feature"}, \code{"segment"} or \code{"largest"}. The first generates a spatial point for every feature, the second for every segment (i.e. subfeature), the third only for the largest segment (subfeature). Note that the last two options can be significant slower.}

\item{points.icon.scale}{scaling number that determines how large the icons (or grobs) are in plot mode in comparison to proportional symbols (such as bubbles). For view mode, use the argument \code{grob.dim}}

\item{points.just}{justification of the points relative to the point coordinates. Either one of the following values: \code{"left"} , \code{"right"}, \code{"center"}, \code{"bottom"}, and \code{"top"}, or a vector of two values where first value specifies horizontal and the second value vertical justification. Besides the mentioned values, also numeric values between 0 and 1 can be used. 0 means left justification for the first value and bottom justification for the second value. Note that in view mode, only one value is used.}

\item{points.grob.dim}{vector of four values that determine how grob objects (see details) are shown in view mode. The first and second value are the width and height of the displayed icon. The third and fourth value are the width and height of the rendered png image that is used for the icon. Generally, the third and fourth value should be large enough to render a ggplot2 graphic successfully. Only needed for the view mode.}
}
\description{
Map layer that draws simple features as they are. Supported visual variables
are: \code{fill} (the fill color), \code{col} (the border color), \code{size} the point size,
\code{shape} the symbol shape, \code{lwd} (line width), \code{lty} (line type), \code{fill_alpha} (fill color alpha transparency)
and \code{col_alpha} (border color alpha transparency).
}
\details{
The visual variable arguments (e.g. \code{col}) can be specified with a data
variable name (e.g., a spatial vector attribute or a raster layer of the object
specified in \code{\link[=tm_shape]{tm_shape()}}), with a visual value (for \code{col}, a color is expected), or with a geometry-derived variable (see below).
See \href{https://r-tmap.github.io/tmap/articles/basics_vv}{vignette about visual variables}.

Multiple values can be specified: in that case facets are created.
These facets can be combined with other faceting data variables, specified with \code{\link[=tm_facets]{tm_facets()}}.
See \href{https://r-tmap.github.io/tmap/articles/basics_facets}{vignette about facets}.
\itemize{
\item The \verb{*.scale} arguments determine the used scale to map the data values to
visual variable values. These can be specified with one of the available
\verb{tm_scale_*()} functions. The default is specified by the tmap option (\code{\link[=tm_options]{tm_options()}}) \code{scales.var}.
See \href{https://r-tmap.github.io/tmap/articles/basics_scales}{vignette about scales}.
\item The \verb{*.legend} arguments determine the used legend, specified with \code{\link[=tm_legend]{tm_legend()}}.
The default legend and its settings are determined by the tmap options (\code{\link[=tm_options]{tm_options()}}) \code{legend.} .
See \href{https://r-tmap.github.io/tmap/articles/basics_legends}{vignette about legends}.
\item The \verb{*.chart} arguments specify additional charts, specified with \code{tm_chart_}, e.g. \code{\link[=tm_chart_histogram]{tm_chart_histogram()}}.
See \href{https://r-tmap.github.io/tmap/articles/basics_charts}{vignette about charts}.
\item The \verb{*.free} arguments determine whether scales are applied freely across facets, or shared.
A logical value is required. They can also be specified with a vector of three
logical values; these determine whether scales are applied freely per facet dimension.
This is only useful when facets are applied (see \code{\link[=tm_facets]{tm_facets()}}).
There are maximally three facet dimensions: rows, columns, and pages. This only
applies for a facet grid (\code{\link[=tm_facets_grid]{tm_facets_grid()}}). For instance, \code{col.free = c(TRUE, FALSE, FALSE)}
means that for the visual variable \code{col}, each row of facets will have its own
scale, and therefore its own legend. For facet wraps and stacks
(\code{\link[=tm_facets_wrap]{tm_facets_wrap()}} and \code{\link[=tm_facets_stack]{tm_facets_stack()}}) there is only one facet dimension,
so the \verb{*.free} argument requires only one logical value.
}

Currently, three geometry-derived variables are implemented:
\itemize{
\item \code{"AREA"} (polygons only), which uses the feature area;
\item \code{"LENGTH"} (lines only), which uses the feature length; and
\item \code{"MAP_COLORS"}, which assigns values so that adjacent features receive
different values, making it particularly suitable for coloring
neighbouring polygons.
}

Note that geometry-derived variables do not generate a legend automatically.
If a legend is required, compute the corresponding variable explicitly,
for example with \code{\link[sf:geos_measures]{sf::st_area()}}, \code{\link[sf:geos_measures]{sf::st_length()}}, or
\code{\link[tmaptools:map_coloring]{tmaptools::map_coloring()}}, and use the resulting values instead.
}
\examples{
data(World)

World$geometry[World$continent == "Africa"] <- 
  sf::st_centroid(World$geometry[World$continent == "Africa"])
World$geometry[World$continent == "South America"] <- 
  sf::st_cast(World$geometry[World$continent == "South America"], 
    "MULTILINESTRING", group_or_split = FALSE)

tm_shape(World, crs = "+proj=robin") + 
	tm_sf()
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/shiny.R, R/shiny_dispatch.R
\name{renderTmap}
\alias{renderTmap}
\alias{tmapOutput}
\alias{tmapProxy}
\alias{tm_remove_layer}
\alias{renderTmapGS}
\alias{tmapOutputGS}
\alias{tmapProxyGS}
\title{Wrapper functions for using \strong{tmap} in \strong{shiny}}
\usage{
renderTmap(
  expr,
  env = parent.frame(),
  quoted = FALSE,
  execOnResize = TRUE,
  mode = NA
)

tmapOutput(outputId, width = "100\%", height = 400, mode = NA)

tmapProxy(mapId, session = shiny::getDefaultReactiveDomain(), x, mode = NA)

tm_remove_layer(zindex)

renderTmapGS(x, ...)

tmapOutputGS(x, ...)

tmapProxyGS(x, ...)
}
\arguments{
\item{expr}{A tmap object. A tmap object is created with \code{\link[=qtm]{qtm()}} or by stacking \code{\link{tmap-element}}s.}

\item{env}{The environment in which to evaluate expr}

\item{quoted}{Is \code{expr} a quoted expression (with \code{quote()})? This is useful if you want to save an expression in a variable}

\item{execOnResize}{If \code{TRUE} (default), when the plot is resized, the map is regenerated. When set to \code{FALSE} the map is rescaled: the aspect ratio is kept, but the layout will be less desirable.}

\item{mode}{tmap mode, see \code{\link[=tmap_mode]{tmap_mode()}} If not defined, the current mode is used}

\item{outputId}{Output variable to read from}

\item{width, height}{the width and height of the map}

\item{mapId}{single-element character vector indicating the output ID of the map to modify (if invoked from a Shiny module, the namespace will be added automatically)}

\item{session}{the Shiny session object to which the map belongs; usually the default value will suffice}

\item{x}{the tmap object that specifies the added and removed layers.}

\item{zindex}{The stacking number of the layer to be removed. It is recommended to specify the \code{zindex} for this layer when creating the map (inside \code{renderTmap()}).}

\item{...}{passed on to the mode-specific methods}
}
\description{
\itemize{
\item \code{tmapOutput()} creates a UI element
\item \code{renderTmap()} renders a tmap map
\item \code{tmapProxy()} updates a tmap map in \code{view} mode
}

Adding layers is as usual via the map layer functions like \code{\link[=tm_polygons]{tm_polygons()}}.
Removing layers can be done, removing with \code{tm_remove_layer()}.
}
\details{
Two features from tmap are not (yet) supported in Shiny: small multiples (facets) and colored backgrounds (argument \code{bg.color} of \code{\link[=tm_layout]{tm_layout()}}).
Workarounds for small multiples: create multiple independent maps or specify \code{as.layers = TRUE} in \code{\link[=tm_facets]{tm_facets()}}.
}
\examples{
if (interactive() && require("shiny")) {

	data(World)
	world_vars <- setdiff(names(World), c("iso_a3", "name", "sovereignt", "geometry"))

	current.mode <- tmap_mode("plot")

	shinyApp(
		ui = fluidPage(
			tmapOutput("map", height = "600px"),
			selectInput("var", "Variable", world_vars)
		),
		server <- function(input, output, session) {
			output$map <- renderTmap({
				tm_shape(World) +
					tm_polygons(input$var, zindex = 401)
			})
		}
	)

	tmap_mode("view")

	shinyApp(
		ui = fluidPage(
			tmapOutput("map", height = "600px"),
			selectInput("var", "Variable", world_vars)
		),
		server <- function(input, output, session) {
			output$map <- renderTmap({
				tm_shape(World, id = "iso_a3") +
					tm_polygons(fill = world_vars[1], zindex = 401)
			})
			observe({
				var <- input$var
				tmapProxy("map", session, {
					tm_remove_layer(401) +
						tm_shape(World, id = "iso_a3") +
						tm_polygons(fill = var, zindex = 401)
				})
			})
		},options = list(launch.browser=TRUE)
	)

	tmap_mode(current.mode)
}
}
\seealso{
\href{https://r-tmap.github.io/tmap/articles/adv_shiny}{Shiny integration}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tmap_tip.R
\name{tmap_tip}
\alias{tmap_tip}
\title{Print a random tip to the console}
\usage{
tmap_tip()
}
\value{
A message
}
\description{
Print a random tip to the console
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tm_scale_.R
\name{tm_scale_asis}
\alias{tm_scale_asis}
\title{Scales: as is}
\usage{
tm_scale_asis(values.scale = NA, value.neutral = NA, value.na = NA, ...)
}
\arguments{
\item{values.scale}{(generic scale argument) Scaling of the values. Only useful for size-related visual variables, such as \code{size} of \code{\link[=tm_symbols]{tm_symbols()}} and \code{lwd} of \code{\link[=tm_lines]{tm_lines()}}.}

\item{value.neutral}{(generic scale argument) Value that can be considered neutral. This is used for legends of other visual variables of the same map layer. E.g. when both \code{fill} and \code{size} are used for \code{\link[=tm_symbols]{tm_symbols()}} (using filled circles), the size legend items are filled with the \code{value.neutral} color from the \code{fill.scale} scale, and fill legend items are bubbles of size \code{value.neutral} from the \code{size.scale} scale.}

\item{value.na}{(generic scale argument) Value used for missing values. See tmap option \code{"value.na"} for defaults per visual variable.}

\item{...}{Arguments caught (and not used) from the automatic function \code{\link[=tm_scale]{tm_scale()}}}
}
\description{
Scales in tmap are configured by the family of functions with prefix \code{tm_scale}.
Such function should be used for the input of the \code{.scale} arguments in the
layer functions (e.g. \code{fill.scale} in \code{\link[=tm_polygons]{tm_polygons()}}).
The function \code{tm_scale_asis()} is used to take data values as they are and use them as such for the visual variable.
}
\seealso{
\code{\link[=tm_scale]{tm_scale()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tmapShape_misc.R
\name{get_fact_levels_na}
\alias{get_fact_levels_na}
\title{Internal tmap function that gets factor levels with NA's}
\usage{
get_fact_levels_na(x, o)
}
\arguments{
\item{x}{vector}

\item{o}{options}
}
\description{
Internal tmap function that gets factor levels with NA's
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tmapOnLoad.R, R/tmap_providers.R
\docType{data}
\name{.tmap_providers}
\alias{.tmap_providers}
\alias{tmap_provider_credits}
\alias{tmap_providers}
\title{Get basemap tiles providers}
\format{
An object of class \code{environment} of length 44.
}
\usage{
.tmap_providers

tmap_provider_credits(provider)

tmap_providers(mode, credits = FALSE, as.list = credits)
}
\arguments{
\item{provider}{provider name}

\item{mode}{mode. If not specified the default mode is used}

\item{credits}{If \code{TRUE} the credit (attribution) text is returned. If \code{FALSE} (default) the provider name.}

\item{as.list}{Should the output be returned as list where names are provider names? By default \code{TRUE} when \code{credits} is also \code{TRUE}.}
}
\value{
list or vector (see \code{as.list}) with providers (or credits). \code{\link[=tmap_provider_credits]{tmap_provider_credits()}} returns the credits text for the provided provider.
}
\description{
Get basemap tiles providers and the credits (attribution text). \code{\link[=tmap_providers]{tmap_providers()}} returns a list (or vector) with provider nams (or credits). \code{\link[=tmap_provider_credits]{tmap_provider_credits()}}
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tm_scale_.R
\name{tm_scale_asis}
\alias{tm_scale_asis}
\title{Scales: as is}
\usage{
tm_scale_asis(values.scale = NA, value.neutral = NA, value.na = NA, ...)
}
\arguments{
\item{values.scale}{(generic scale argument) Scaling of the values. Only useful for size-related visual variables, such as \code{size} of \code{\link[=tm_symbols]{tm_symbols()}} and \code{lwd} of \code{\link[=tm_lines]{tm_lines()}}.}

\item{value.neutral}{(generic scale argument) Value that can be considered neutral. This is used for legends of other visual variables of the same map layer. E.g. when both \code{fill} and \code{size} are used for \code{\link[=tm_symbols]{tm_symbols()}} (using filled circles), the size legend items are filled with the \code{value.neutral} color from the \code{fill.scale} scale, and fill legend items are bubbles of size \code{value.neutral} from the \code{size.scale} scale.}

\item{value.na}{(generic scale argument) Value used for missing values. See tmap option \code{"value.na"} for defaults per visual variable.}

\item{...}{Arguments caught (and not used) from the automatic function \code{\link[=tm_scale]{tm_scale()}}}
}
\description{
Scales in tmap are configured by the family of functions with prefix \code{tm_scale}.
Such function should be used for the input of the \code{.scale} arguments in the
layer functions (e.g. \code{fill.scale} in \code{\link[=tm_polygons]{tm_polygons()}}).
The function \code{tm_scale_asis()} is used to take data values as they are and use them as such for the visual variable.
}
\seealso{
\href{https://r-tmap.github.io/tmap/articles/basics_scales}{Scales}, \href{https://r-tmap.github.io/tmap/articles/basics_vv}{Visual variables}, \href{https://r-tmap.github.io/tmap/articles/adv_multivariate}{Multivariate visual variables} \code{\link[=tm_scale]{tm_scale()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tm_layers_rgb.R
\name{tm_rgb}
\alias{tm_rgb}
\alias{tm_rgba}
\alias{opt_tm_rgb}
\title{Map layer: rgb images}
\usage{
tm_rgb(
  col = tm_vars(n = 3, multivariate = TRUE),
  col.scale = tm_scale_rgb(),
  col.legend = tm_legend(),
  col.chart = tm_chart_none(),
  col.free = NA,
  col_alpha = tm_const(),
  col_alpha.scale = tm_scale(),
  col_alpha.legend = tm_legend(),
  col_alpha.chart = tm_chart_none(),
  col_alpha.free = NA,
  options = opt_tm_rgb(),
  ...
)

tm_rgba(
  col = tm_vars(n = 4, multivariate = TRUE),
  col.scale = tm_scale_rgba(),
  col.legend = tm_legend(),
  col.chart = tm_chart_none(),
  col.free = NA,
  options = opt_tm_rgb()
)

opt_tm_rgb(interpolate = FALSE, saturation = 1)
}
\arguments{
\item{col, col.scale, col.legend, col.chart, col.free}{Visual variable that determines
the color. \code{col} is a multivariate variable, with 3 (\code{tm_rgb}) or 4 (\code{tm_rgba}) numeric data variables. These can be specified via \code{\link[=tm_vars]{tm_vars()}} with \code{multivariate = TRUE}}

\item{col_alpha, col_alpha.scale, col_alpha.legend, col_alpha.chart, col_alpha.free}{Visual variable that determines the color transparency. See details.}

\item{options}{options passed on to the corresponding \verb{opt_<layer_function>} function}

\item{...}{to catch deprecated arguments from version < 4.0}

\item{interpolate}{Should the raster image be interpolated? Currently only applicable in view mode (passed on to \code{\link[grid:grid.raster]{grid}})}

\item{saturation}{The saturation of the rgb.}
}
\description{
Map layer that an rgb image.. The used (multivariate) visual variable is \code{col},
which should be specified with 3 or 4 variables for \code{tm_rgb()} and \code{tm_rgba()} respectively.
The first three correspond to the red, green, and blue channels. The optional
fourth is the alpha transparency channel.
}
\examples{
\dontrun{
require(stars)
file = system.file("tif/L7_ETMs.tif", package = "stars")

L7 = stars::read_stars(file)

tm_shape(L7) +
	tm_rgb()

# the previous example was a shortcut of this call
tm_shape(L7) +
	tm_rgb(col = tm_vars("band", dimvalues = 1:3, multivariate = TRUE))

# alternative format: using a stars dimension instead of attributes
L7_alt = split(L7, "band")
tm_shape(L7_alt) +
	tm_rgb()

# with attribute names
tm_shape(L7_alt) +
	tm_rgb(col = tm_vars(c("X1", "X2", "X3"), multivariate = TRUE))

# with attribute indices
tm_shape(L7_alt) +
	tm_rgb(col = tm_vars(1:3, multivariate = TRUE))

if (requireNamespace("terra")) {
	L7_terra = terra::rast(file)

	tm_shape(L7_terra) +
		tm_rgb()

	# with layer names
	tm_shape(L7_terra) +
		tm_rgb(tm_vars(names(L7_terra)[1:3], multivariate = TRUE))

	# with layer indices
	tm_shape(L7_alt) +
		tm_rgb(col = tm_vars(1:3, multivariate = TRUE))

}
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tmapOnLoad.R, R/tmap_providers.R
\name{.tmap_providers}
\alias{.tmap_providers}
\alias{tmap_provider_credits}
\alias{tmap_providers}
\title{Get basemap tile providers}
\usage{
.tmap_providers

tmap_provider_credits(provider)

tmap_providers(mode, credits = FALSE, as.list = credits)
}
\arguments{
\item{provider}{provider name}

\item{mode}{mode. If not specified the current active mode is used.}

\item{credits}{If \code{TRUE} the credit (attribution) text is returned.
If \code{FALSE} (default) the provider name.}

\item{as.list}{Should the output be returned as a list where names are
provider names? By default \code{TRUE} when \code{credits} is also
\code{TRUE}.}
}
\value{
\code{tmap_providers()} returns a list or vector (see
\code{as.list}) of provider names or credits.
\code{tmap_provider_credits()} returns the attribution text for the
specified provider. \code{.tmap_providers} is an environment; see
Details.
}
\description{
Get basemap tile providers and their credits (attribution text).
\code{tmap_providers()} returns a list or vector of provider names or
credits. \code{tmap_provider_credits()} returns the attribution text for a
specific provider.
}
\details{
\code{.tmap_providers} is an environment populated with all available
provider names as named entries. Its primary purpose is to enable
autocomplete in IDEs such as RStudio: typing \code{.tmap_providers$} in the
console or a script triggers a dropdown list of all available providers,
making it easy to discover and select provider names without consulting the
documentation. It is not intended to be called as a function.
}
\examples{
# List all providers for the current mode
tmap_providers()

# Use IDE autocomplete to discover providers interactively:
# type .tmap_providers$ in the RStudio console
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tm_layers_rgb.R
\name{tm_rgb}
\alias{tm_rgb}
\alias{tm_rgba}
\alias{opt_tm_rgb}
\title{Map layer: rgb images}
\usage{
tm_rgb(
  col = tm_vars(n = 3, multivariate = TRUE),
  col.scale = tm_scale_rgb(),
  col.legend = tm_legend(),
  col.chart = tm_chart_none(),
  col.free = NA,
  col_alpha = tm_const(),
  col_alpha.scale = tm_scale(),
  col_alpha.legend = tm_legend(),
  col_alpha.chart = tm_chart_none(),
  col_alpha.free = NA,
  blend = "over",
  options = opt_tm_rgb(),
  ...
)

tm_rgba(
  col = tm_vars(n = 4, multivariate = TRUE),
  col.scale = tm_scale_rgba(),
  col.legend = tm_legend(),
  col.chart = tm_chart_none(),
  col.free = NA,
  blend = "over",
  options = opt_tm_rgb()
)

opt_tm_rgb(interpolate = FALSE, saturation = 1)
}
\arguments{
\item{col, col.scale, col.legend, col.chart, col.free}{Visual variable that determines
the color. \code{col} is a multivariate variable, with 3 (\code{tm_rgb}) or 4 (\code{tm_rgba}) numeric data variables. These can be specified via \code{\link[=tm_vars]{tm_vars()}} with \code{multivariate = TRUE}}

\item{col_alpha, col_alpha.scale, col_alpha.legend, col_alpha.chart, col_alpha.free}{Visual variable that determines the color transparency. See details. \emph{Unit:} Proportion -- numeric 0-1 (0 = fully transparent, 1 = fully opaque).}

\item{blend}{Compositing operator for layer blending. Default \code{"over"} applies no blending. See the "Layer blending" section of \code{\link[=tm_polygons]{tm_polygons()}} for supported values.}

\item{options}{options passed on to the corresponding \verb{opt_<layer_function>} function}

\item{...}{to catch deprecated arguments from version < 4.0}

\item{interpolate}{Should the raster image be interpolated? Currently only applicable in view mode (passed on to \code{\link[grid:rasterGrob]{grid}})}

\item{saturation}{The saturation of the rgb.}
}
\description{
Map layer that an rgb image.. The used (multivariate) visual variable is \code{col},
which should be specified with 3 or 4 variables for \code{tm_rgb()} and \code{tm_rgba()} respectively.
The first three correspond to the red, green, and blue channels. The optional
fourth is the alpha transparency channel.
}
\examples{
\dontrun{
require(stars)
file = system.file("tif/L7_ETMs.tif", package = "stars")

L7 = stars::read_stars(file)

tm_shape(L7) +
	tm_rgb()

# the previous example was a shortcut of this call
tm_shape(L7) +
	tm_rgb(col = tm_vars("band", dimvalues = 1:3, multivariate = TRUE))

# alternative format: using a stars dimension instead of attributes
L7_alt = split(L7, "band")
tm_shape(L7_alt) +
	tm_rgb()

# with attribute names
tm_shape(L7_alt) +
	tm_rgb(col = tm_vars(c("X1", "X2", "X3"), multivariate = TRUE))

# with attribute indices
tm_shape(L7_alt) +
	tm_rgb(col = tm_vars(1:3, multivariate = TRUE))

if (requireNamespace("terra")) {
	L7_terra = terra::rast(file)

	tm_shape(L7_terra) +
		tm_rgb()

	# with layer names
	tm_shape(L7_terra) +
		tm_rgb(tm_vars(names(L7_terra)[1:3], multivariate = TRUE))

	# with layer indices
	tm_shape(L7_alt) +
		tm_rgb(col = tm_vars(1:3, multivariate = TRUE))

}
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tmap_mode.R
\name{tmap_design_mode}
\alias{tmap_design_mode}
\title{Set the design mode}
\usage{
tmap_design_mode(design.mode)
}
\arguments{
\item{design.mode}{Logical value that determines the design mode.
If omitted then the design mode is toggled.}
}
\description{
When the so-called "design mode" is enabled,
the composition of the plot is shown explicitly in plot mode.
The used color codings is printed in the console as well as information about plot size and aspect ratio.
}
\details{
This function sets the global option \code{tmap.design.mode}.
It can be used as toggle function without arguments.
}
\seealso{
\code{\link[=tmap_options]{tmap_options()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tm_layout.R
\name{tm_view}
\alias{tm_view}
\title{View mode options}
\usage{
tm_view(
  use_browser,
  use_WebGL,
  control.position,
  control.bases,
  control.overlays,
  control.collapse,
  set_bounds,
  set_view,
  set_zoom_limits,
  use_circle_markers,
  leaflet.options,
  ...
)
}
\arguments{
\item{use_browser}{If \code{TRUE} it opens an external browser, and \code{FALSE} (default) it opens the internal IDEs (e.g. RStudio) browser.}

\item{use_WebGL}{use webGL for points, lines, and polygons. For large spatial objects, this is much faster than the standard leaflet layer functions. However, it can not always be used for two reasons. First, the number of visual variables are limited; only fill, size, and color (for lines) are supported. Second, projected CRS's are not supported. Furthermore, it has the drawback that polygon borders are not as sharp. By default only \code{TRUE} for large spatial objects (1000 or more features) when the mentioned criteria are met.
By default \code{TRUE} if no other visual variables are used.}

\item{control.position}{The position of the control. A tm_pos object, or a shortcut of two values: horizontal (left, center, right) and vertical (top, center, bottom). See tm_pos for details}

\item{control.bases}{base layers}

\item{control.overlays}{overlay layers}

\item{control.collapse}{Should the box be collapsed or expanded?}

\item{set_bounds}{logical that determines whether maximum bounds are set,
or a bounding box. Not applicable in plot mode.
In view mode, this is passed on to \link[leaflet:map-methods]{setMaxBounds()}}

\item{set_view}{numeric vector that determines the view.
Either a vector of three: \code{lng}, \code{lat}, and \code{zoom}, or a single value:
\code{zoom}. See \link[leaflet:map-methods]{setView()}.
Only applicable if \code{bbox} is not specified}

\item{set_zoom_limits}{numeric vector of two that set the minimum and maximum
zoom levels (see \link[leaflet:map-options]{tileOptions()}).}

\item{use_circle_markers}{If \code{TRUE} (default) circle shaped symbols (e.g. \code{tm_dots()} and \code{tm_symbols()}) will be rendered as \link[leaflet:map-layers]{addCircleMarkers()} instead of \link[leaflet:map-layers]{addMarkers()}. The former is faster, the latter can support any symbol since it is based on icons}

\item{leaflet.options}{options passed on to
\link[leaflet:leaflet]{leafletOptions()}}

\item{...}{to catch deprecated arguments}
}
\description{
View mode options. These options are specific to the view mode.
}
\seealso{
\code{\link[=tm_group]{tm_group()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/shapeTM.R
\name{shapeTM}
\alias{shapeTM}
\title{Internal tmap function to create a tmap shape}
\usage{
shapeTM(shp, tmapID = NULL, bbox = NULL, ...)
}
\arguments{
\item{shp}{Shape file}

\item{tmapID}{tmap Identifier}

\item{bbox}{bounding box}
}
\description{
Internal tmap function to create a tmap shape
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tm_layout.R
\name{tm_view}
\alias{tm_view}
\title{View mode options}
\usage{
tm_view(
  use_browser,
  use_WebGL,
  control.position,
  control.bases,
  control.overlays,
  control.collapse,
  set_bounds,
  set_view,
  set_zoom_limits,
  use_circle_markers,
  leaflet.options,
  ...
)
}
\arguments{
\item{use_browser}{If \code{TRUE} it opens an external browser, and \code{FALSE} (default) it opens the internal IDEs (e.g. RStudio) browser.}

\item{use_WebGL}{use webGL for points, lines, and polygons. For large spatial objects, this is much faster than the standard leaflet layer functions. However, it can not always be used for two reasons. First, the number of visual variables are limited; only fill, size, and color (for lines) are supported. Second, projected CRS's are not supported. Furthermore, it has the drawback that polygon borders are not as sharp. By default only \code{TRUE} for large spatial objects (1000 or more features) when the mentioned criteria are met.
By default \code{TRUE} if no other visual variables are used.}

\item{control.position}{The position of the control. A tm_pos object, or a shortcut of two values: horizontal (left, center, right) and vertical (top, center, bottom). See tm_pos for details}

\item{control.bases}{base layers}

\item{control.overlays}{overlay layers}

\item{control.collapse}{Should the box be collapsed or expanded?}

\item{set_bounds}{logical that determines whether maximum bounds are set,
or a bounding box. Not applicable in plot mode.
In view mode, this is passed on to \link[leaflet:setMaxBounds]{setMaxBounds()}}

\item{set_view}{numeric vector that determines the view.
Either a vector of three: \code{lng}, \code{lat}, and \code{zoom}, or a single value:
\code{zoom}. See \link[leaflet:setView]{setView()}.
Only applicable if \code{bbox} is not specified}

\item{set_zoom_limits}{numeric vector of two that set the minimum and maximum
zoom levels (see \link[leaflet:tileOptions]{tileOptions()}).}

\item{use_circle_markers}{If \code{TRUE} (default) circle shaped symbols (e.g. \code{tm_dots()} and \code{tm_symbols()}) will be rendered as \link[leaflet:addCircleMarkers]{addCircleMarkers()} instead of \link[leaflet:addMarkers]{addMarkers()}. The former is faster, the latter can support any symbol since it is based on icons}

\item{leaflet.options}{options passed on to
\link[leaflet:leafletOptions]{leafletOptions()}}

\item{...}{to catch deprecated arguments}
}
\description{
View mode options. These options are specific to the view mode.
}
\seealso{
\href{https://r-tmap.github.io/tmap/articles/basics_modes}{Modes} \href{https://r-tmap.github.io/tmap/articles/adv_groups}{Layer groups (view mode)}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tmap_format.R
\name{tmap_format}
\alias{tmap_format}
\alias{tmap_format_add}
\alias{tm_format}
\title{Deprecated: format}
\usage{
tmap_format(format)

tmap_format_add(..., name)

tm_format(format, ...)
}
\arguments{
\item{format}{Name of the format}

\item{...}{not used}

\item{name}{Name of the format}
}
\description{
In tmap < 4.0 it was possible to set shape-specific options, such as margins and legend position. However, this has become superfluous because in tmap > 4.0 legends are by default placed outside the map area. If needed, a shape-specific set of options can be stored as a style with tmap_options_save.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tmap_mode.R
\name{tmap_mode}
\alias{tmap_mode}
\alias{ttm}
\alias{rtm}
\alias{ttmp}
\alias{rtmp}
\title{Set tmap mode to static plotting or interactive viewing}
\usage{
tmap_mode(mode = NULL, silent = FALSE)

ttm()

rtm()

ttmp()

rtmp()
}
\arguments{
\item{mode}{One of \code{"plot"} or \code{"view"}. See Details for more info.}

\item{silent}{Should the mode be switched silently (by default \code{FALSE})}
}
\value{
\itemize{
\item \code{tmap_mode()} returns the current tmap mode invisibly (when called without argument).
Otherwise, returns the previous mode.
\item \code{ttm()} switches mode and returns previous tmap mode invisibly.
The previous tmap mode before switching.
}
}
\description{
\itemize{
\item \code{tmap_mode()} informs of the current mode (if called without argument).
\item \code{ttm()} toggles between the most recent two modes.
\item \code{ttmp()} same as \code{ttm()} and calls \code{\link[=tmap_last]{tmap_last()}} to display the last map in the new mode.
\item \code{rtm()} rotate between between all modes
\item \code{rtmp()} same as \code{rtm()} and calls \code{\link[=tmap_last]{tmap_last()}} to display the last map in the new mode.
}

Set tmap mode to static plotting or interactive viewing.
The global option \code{tmap.mode} determines the whether thematic maps are plot
in the graphics device, or shown as an interactive leaflet map (see also \code{\link[=tmap_options]{tmap_options()}}.
The function \code{tmap_mode()} is a wrapper to set this global option.
The convenient function \code{ttm()}, which stands for toggle thematic map,
is a toggle switch between the two modes. The function \code{ttmp()} stands for
toggle thematic map and print last map: it does the same as \code{ttm()} followed
by \code{\link[=tmap_last]{tmap_last()}}; in order words, it shows the last map in the other mode.
It is recommended to use \code{tmap_mode()} in scripts and \code{ttm()}/\code{ttmp()} in the console.
}
\section{\code{mode = "plot"}}{
Thematic maps are shown in the graphics device.
This is the default mode, and supports all tmap's features,
such as small multiples (see \code{\link[=tm_facets]{tm_facets()}}) and extensive layout settings (see \code{\link[=tm_layout]{tm_layout()}}).
It is recommended to use \code{\link[=tmap_save]{tmap_save()}} for saving static maps.
}

\section{\code{mode = "view"}}{
Thematic maps are viewed interactively in the web browser or RStudio's Viewer pane.
Maps are fully interactive with tiles from OpenStreetMap or other map providers
(see \code{\link[=tm_tiles]{tm_tiles()}}). See also \code{\link[=tm_view]{tm_view()}} for options related to the \code{"view"} mode.
This mode generates a \code{\link[leaflet:leaflet]{leaflet::leaflet()}} widget, which can also be directly
obtained with \code{\link[=tmap_leaflet]{tmap_leaflet()}}.
With R Markdown, it is possible to publish it to an HTML page.

However, there are a couple of constraints in comparison to \code{"plot"}:
}

\examples{
current.mode = tmap_mode()

tmap_mode("plot")

tm_shape(World) + tm_polygons("HPI")

tmap_mode("view")

tm_shape(World) + tm_polygons("HPI")

ttm()

tm_shape(World) + tm_polygons("HPI")

tmap_mode(current.mode)
}
\references{
Tennekes, M., 2018, {tmap}: Thematic Maps in {R},
Journal of Statistical Software, 84(6), 1-39, \doi{10.18637/jss.v084.i06}
}
\seealso{
\href{https://r-tmap.github.io/tmap/articles/basics_modes}{vignette about modes}
\itemize{
\item \code{\link[=tmap_last]{tmap_last()}} to show the last map
\item \code{\link[=tm_view]{tm_view()}} for viewing options
\item \code{\link[=tmap_leaflet]{tmap_leaflet()}} for obtaining a leaflet widget
\item \code{\link[=tmap_options]{tmap_options()}} for tmap options
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tmap_mode.R
\name{tmap_mode}
\alias{tmap_mode}
\alias{ttm}
\alias{rtm}
\alias{tmap_mode_pool}
\alias{ttmp}
\alias{rtmp}
\title{Set tmap mode}
\usage{
tmap_mode(mode = NULL, silent = FALSE)

ttm()

rtm()

tmap_mode_pool(modes = NULL, silent = FALSE)

ttmp()

rtmp()
}
\arguments{
\item{mode}{A string specifying the mode. See \code{\link[=tmap_options]{tmap_options()}} for available modes.}

\item{silent}{Should the mode be switched silently? Default \code{FALSE}.}

\item{modes}{Character vector of mode names (minimum 2), or \code{NULL} to reset.}
}
\value{
\itemize{
\item \code{tmap_mode()} returns the current mode invisibly when called without argument,
otherwise the previous mode.
\item \code{ttm()}, \code{rtm()} return the previous mode invisibly.
\item \code{tmap_mode_pool()} returns the previous pool invisibly.
}
}
\description{
\itemize{
\item \code{tmap_mode()} gets (no argument) or sets the current mode.
\item \code{ttm()} toggles between the two most recent modes.
\item \code{ttmp()} same as \code{ttm()}, then calls \code{\link[=tmap_last]{tmap_last()}}.
\item \code{rtm()} rotates through all modes in the pool.
\item \code{rtmp()} same as \code{rtm()}, then calls \code{\link[=tmap_last]{tmap_last()}}.
\item \code{tmap_mode_pool()} restricts which modes are cycled by \code{\link[=ttm]{ttm()}} and \code{\link[=rtm]{rtm()}}.
Call without arguments to inspect the current pool, or pass \code{NULL} to reset.
}

It is recommended to use \code{tmap_mode()} in scripts and \code{ttm()}/\code{ttmp()} in the console.
}
\details{
The default modes are \code{"plot"} (static, graphics device) and \code{"view"} (interactive,
browser or RStudio Viewer). Additional modes such as \code{"maplibre"} and \code{"mapbox"} become
available when \pkg{tmap.mapgl} is loaded.
}
\examples{
current.mode = tmap_mode()

tmap_mode("plot")

tm_shape(World) + tm_polygons("HPI")

tmap_mode("view")

tm_shape(World) + tm_polygons("HPI")

ttm()

tm_shape(World) + tm_polygons("HPI")

tmap_mode(current.mode)
}
\references{
Tennekes, M., 2018, {tmap}: Thematic Maps in {R},
Journal of Statistical Software, 84(6), 1-39, \doi{10.18637/jss.v084.i06}
}
\seealso{
\itemize{
\item .doc_see_also_modes()
\item \code{\link[=tmap_last]{tmap_last()}} to show the last map
\item \code{\link[=tm_view]{tm_view()}} for viewing options
\item \code{\link[=tmap_leaflet]{tmap_leaflet()}} for obtaining a leaflet widget
\item \code{\link[=tmap_options]{tmap_options()}} for tmap options
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tm_plot_order.R
\name{tm_plot_order}
\alias{tm_plot_order}
\title{Determine plotting order of features}
\usage{
tm_plot_order(
  aes,
  reverse = TRUE,
  na.order = c("mix", "bottom", "top"),
  null.order = c("bottom", "mix", "top"),
  null.below.na = TRUE
)
}
\arguments{
\item{aes}{Visual variable for which the values determine the plotting order.
Example: bubble map where the \code{"size"} aesthetic is used.
A data variable (say population) is mapped via a continuous scale (\code{\link[=tm_scale_continuous]{tm_scale_continuous()}})
to bubble sizes. The bubbles are plotted in order of size.
How is determined by the other arguments. Use \code{"DATA"} to keep the same
order as in the data. Another special value are \code{"AREA"} and \code{"LENGTH"}
which are preserved for polygons and lines respectively: rather than a data
variable the polygon area / line lengths determines the plotting order.}

\item{reverse}{Logical that determines whether the visual values are plotted
in reversed order. The visual values (specified with tmap option \code{"values.var"})
are by default reversed, so plotted starting from the last value.
In the bubble map example, this means that large bubbles are plotted first,
hence at the bottom.}

\item{na.order}{Where should features be plotted that have an \code{NA} value for
(at least) one other aesthetic variable? In the (order) \code{"mix"}, at the \code{"bottom"},
or on \code{"top"}? In the bubble map example: if fill color is missing for some bubble,
where should those bubbles be plotted?}

\item{null.order}{Where should non-selected (aka null) features be plotted?}

\item{null.below.na}{Should null features be plotted below NA features?}
}
\description{
Determine plotting order of features.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tm_layers_aux.R
\name{tm_graticules}
\alias{tm_graticules}
\alias{tm_grid}
\title{Coordinate grid / graticule lines}
\usage{
tm_graticules(
  x = NA,
  y = NA,
  n.x = NA,
  n.y = NA,
  crs = 4326,
  labels.format = list(suffix = intToUtf8(176)),
  labels.cardinal = TRUE,
  ...
)

tm_grid(
  x = NA,
  y = NA,
  n.x = NA,
  n.y = NA,
  crs = NA,
  col = NA,
  lwd = 1,
  alpha = NA,
  labels.show = TRUE,
  labels.pos = c("left", "bottom"),
  labels.size = 0.6,
  labels.col = NA,
  labels.rot = c(0, 0),
  labels.format = list(big.mark = ","),
  labels.cardinal = FALSE,
  labels.margin.x = 0,
  labels.margin.y = 0,
  labels.space.x = NA,
  labels.space.y = NA,
  labels.inside_frame = FALSE,
  ticks = labels.show & !labels.inside_frame,
  lines = TRUE,
  ndiscr = 100,
  zindex = NA,
  group = NA,
  group.control = "none",
  ...
)
}
\arguments{
\item{x}{X coordinates for vertical grid lines. If \code{NA}, it is specified
with a pretty scale and \code{n.x}.}

\item{y}{Y coordinates for horizontal grid lines. If \code{NA}, it is specified
with a pretty scale and \code{n.y}.}

\item{n.x}{Preferred number of grid lines for the x axis. For the labels,
a \code{\link[=pretty]{pretty()}} sequence is used, so the number of actual labels may be different than \code{n.x}.}

\item{n.y}{Preferred number of grid lines for the y axis. For the labels, a
\code{\link[=pretty]{pretty()}} sequence is used, so the number of actual labels may be different than \code{n.y}.}

\item{crs}{Projection character. If specified, the grid lines are projected
accordingly. Many world maps are projected, but still have latitude longitude
(EPSG 4326) grid lines.}

\item{labels.format}{List of formatting options for the grid labels. Parameters are:
\describe{
\item{fun}{Function to specify the labels. It should take a numeric vector,
and should return a character vector of the same size.
By default it is not specified. If specified, the list items \code{scientific},
\code{format}, and \code{digits} (see below) are not used.}
\item{scientific}{Should the labels be formatted scientifically?
If so, square brackets are used, and the \code{format} of the numbers is \code{"g"}.
Otherwise, \code{format="f"}, and \code{text.separator}, \code{text.less.than}, and
\code{text.or.more} are used. Also, the numbers are automatically rounded to
millions or billions if applicable.}
\item{format}{By default, \code{"f"}, i.e. the standard notation \code{xxx.xxx}, is used.
If \code{scientific=TRUE} then \code{"g"}, which means that numbers are formatted scientifically,
i.e. \code{n.dddE+nn} if needed to save space.}
\item{digits}{Number of digits after the decimal point if \code{format="f"}, and
the number of significant digits otherwise.}
\item{...}{Other arguments passed on to \code{\link[=formatC]{formatC()}}}
}}

\item{labels.cardinal}{Add the four cardinal directions (N, E, S, W) to the labels,
instead of using negative coordinates for west and south (so it assumes that
the coordinates are positive in the north-east direction).}

\item{...}{Used to catch deprecated arguments from tmap v3.}

\item{col}{Color of the grid lines.}

\item{lwd}{Line width of the grid lines}

\item{alpha}{Alpha transparency of the grid lines. Number between 0 and 1.
By default, the alpha transparency of \code{col} is taken.}

\item{labels.show}{Show tick labels. Either one value for both \code{x} and \code{y} axis,
or a vector two: the first for \code{x} and latter for \code{y}.}

\item{labels.pos}{position of the labels. Vector of two: the horizontal
("left" or "right") and the vertical ("top" or "bottom") position.}

\item{labels.size}{Font size of the tick labels}

\item{labels.col}{Font color of the tick labels}

\item{labels.rot}{Rotation angles of the labels. Vector of two values: the
first is the rotation angle (in degrees) of the tick labels on the x axis
and the second is the rotation angle of the tick labels on the y axis.
Only \code{0}, \code{90}, \code{180}, and \code{270} are valid values.}

\item{labels.margin.x}{Margin between tick labels of x axis and the frame.
Note that when \code{labels.inside_frame = FALSE} and \code{ticks = TRUE},
the ticks will be adjusted accordingly.}

\item{labels.margin.y}{Margin between tick labels of y axis and the frame.
Note that when \code{labels.inside_frame = FALSE} and \code{ticks = TRUE},
the ticks will be adjusted accordingly.}

\item{labels.space.x}{Space that is used for the labels and ticks for the x-axis
when \code{labels.inside_frame = FALSE}. By default, it is determined automatically
using the widths and heights of the tick labels. The unit of this parameter is text line height.}

\item{labels.space.y}{Space that is used for the labels and ticks for the y-axis
when \code{labels.inside_frame = FALSE}. By default, it is determined automatically
using the widths and heights of the tick labels. The unit of this parameter is text line height.}

\item{labels.inside_frame}{Show labels inside the frame? By default \code{FALSE}.}

\item{ticks}{If \code{labels.inside_frame = FALSE}, should ticks can be drawn between the labels and the frame?
Either one value for both \code{x} and \code{y} axis, or a vector two: the first for \code{x} and latter for \code{y}.}

\item{lines}{If \code{labels.inside_frame = FALSE}, should grid lines can be drawn?}

\item{ndiscr}{Number of points to discretize a parallel or meridian
(only applicable for curved grid lines)}

\item{zindex}{zindex of the pane in view mode. By default, it is set to the
layer number plus 400. By default, the tmap layers will therefore be placed
in the custom panes \code{"tmap401"}, \code{"tmap402"}, etc., except for the base tile
layers, which are placed in the standard \code{"tile"}. This parameter determines
both the name of the pane and the z-index, which determines the pane order
from bottom to top. For instance, if \code{zindex} is set to 500,
the pane will be named \code{"tmap500"}.}

\item{group}{Name of the group to which this layer belongs.
This is only relevant in view mode, where layer groups can be
switched (see \code{group.control})}

\item{group.control}{In view mode, the group control determines how layer groups
can be switched on and off. Options: \code{"radio"} for radio buttons (meaning only one group can be shown),
\code{"check"} for check boxes (so multiple groups can be shown), and \code{"none"}
for no control (the group cannot be (de)selected).}
}
\description{
Draws horizontal and vertical lines according to a coordinate reference system (CRS). \code{tm_grid()} uses the CRS of the (master) shape object, and \code{tm_graticules()} uses latitude and longitude coordinates (EPSG 4326). It creates a \code{\link{tmap-element}} that draws coordinate grid lines. It serves as a
layer that can be drawn anywhere between other layers.
}
\examples{
\dontrun{
current.mode <- tmap_mode("plot")

tm_shape(NLD_muni) +
	tm_polygons() +
	tm_grid()

tm_shape(NLD_muni) +
	tm_polygons() +
	tm_grid(crs = 4326)

tm_shape(NLD_muni) +
	tm_polygons() +
	tm_grid(crs = 3035, labels.inside.frame = TRUE)

tm_shape(World) +
	tm_polygons() +
	tm_facets(by = "continent") +
	tm_grid(labels.inside.frame = TRUE)

tm_shape(NLD_muni) +
	tm_polygons() +
	tm_graticules()

tm_shape(NLD_muni) +
	tm_polygons() +
	tm_graticules(labels.pos = c("right", "top"))


data(NLD_muni, World)

tmap_arrange(
	qtm(NLD_muni) + tm_grid(),
	qtm(NLD_muni) + tm_graticules()
)

qtm(World, shape.crs = "+proj=robin", style = "natural") +
	tm_graticules(ticks = FALSE) +
	tm_layout(frame=FALSE)

tmap_mode(current.mode)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tm_component_functions.R
\name{tm_minimap}
\alias{tm_minimap}
\title{Map component: minimap}
\usage{
tm_minimap(
  server,
  toggle,
  height,
  width,
  margins,
  between_margin,
  position,
  group_id,
  frame,
  frame.color,
  frame.alpha,
  frame.lwd,
  frame.r,
  bg,
  bg.color,
  bg.alpha,
  z,
  ...
)
}
\arguments{
\item{server}{name of the provider or an URL (see \code{\link{tm_tiles}}).
By default, it shows the same map as the basemap, and moreover, it will automatically change when the user switches basemaps.
Note the latter does not happen when \code{server} is specified.}

\item{toggle}{should the minimap have a button to minimise it? By default \code{TRUE}.}

\item{width, height}{width and height of the component.}

\item{margins}{margins}

\item{between_margin}{Margin between}

\item{position}{The position specification of the component: an object created with \code{tm_pos_in()} or \code{tm_pos_out()}. Or, as a shortcut, a vector of two values, specifying the x and y coordinates. The first is \code{"left"}, \code{"center"} or \code{"right"} (or upper case, meaning tighter to the map frame), the second \code{"top"}, \code{"center"} or \code{"bottom"}. Numeric values are also supported, where 0, 0 means left bottom and 1, 1 right top. See also \href{https://r-tmap.github.io/tmap/articles/adv_positions}{vignette about positioning}. In case multiple components should be combined (stacked), use \code{group_id} and specify \code{component} in \code{\link[=tm_components]{tm_components()}}.}

\item{group_id}{Component group id name. All components (e.g. legends, titles, etc) with the same \code{group_id} will be grouped. The specifications of how they are placed (e.g. stacking, margins etc.) are determined in \code{\link[=tm_components]{tm_components()}} where its argument \code{id} should correspond to \code{group_id}.}

\item{frame}{frame should a frame be drawn?}

\item{frame.color}{frame color}

\item{frame.alpha}{frame alpha transparancy}

\item{frame.lwd}{frame line width}

\item{frame.r}{Radius of the rounded frame corners. 0 means no rounding.}

\item{bg}{Show background?}

\item{bg.color}{Background color}

\item{bg.alpha}{Background transparency}

\item{z}{z index, e.g. the place of the component relative to the other componets}

\item{...}{
  Arguments passed on to \code{\link[leaflet:addMiniMap]{leaflet::addMiniMap}}
  \describe{
    \item{\code{map}}{a map widget object}
    \item{\code{collapsedWidth}}{The width of the toggle marker and the minimap
when collapsed, in pixels. Defaults to 19.}
    \item{\code{collapsedHeight}}{The height of the toggle marker and the minimap
when collapsed, in pixels. Defaults to 19.}
    \item{\code{zoomLevelOffset}}{The offset applied to the zoom in the minimap compared
to the zoom of the main map. Can be positive or negative, defaults to -5.}
    \item{\code{zoomLevelFixed}}{Overrides the offset to apply a fixed zoom level to
the minimap regardless of the main map zoom.
Set it to any valid zoom level, if unset zoomLevelOffset is used instead.}
    \item{\code{centerFixed}}{Applies a fixed position to the minimap regardless of
the main map's view / position. Prevents panning the minimap, but does
allow zooming (both in the minimap and the main map).
If the minimap is zoomed, it will always zoom around the centerFixed point.
You can pass in a LatLng-equivalent object. Defaults to false.}
    \item{\code{zoomAnimation}}{Sets whether the minimap should have an animated zoom.
(Will cause it to lag a bit after the movement of the main map.)
Defaults to false.}
    \item{\code{toggleDisplay}}{Sets whether the minimap should have a button to minimize it.
Defaults to false.}
    \item{\code{autoToggleDisplay}}{Sets whether the minimap should hide automatically,
if the parent map bounds does not fit within the minimap bounds.
Especially useful when 'zoomLevelFixed' is set.}
    \item{\code{minimized}}{Sets whether the minimap should start in a minimized position.}
    \item{\code{aimingRectOptions}}{Sets the style of the aiming rectangle by passing in
a Path.Options (\url{https://web.archive.org/web/20220702182250/https://leafletjs.com/reference-1.3.4.html#path-options}) object.
(Clickable will always be overridden and set to false.)}
    \item{\code{shadowRectOptions}}{Sets the style of the aiming shadow rectangle by passing in
a Path.Options (\url{https://web.archive.org/web/20220702182250/https://leafletjs.com/reference-1.3.4.html#path-option}) object.
(Clickable will always be overridden and set to false.)}
    \item{\code{strings}}{Overrides the default strings allowing for translation.}
    \item{\code{tiles}}{URL for tiles or one of the pre-defined providers.}
    \item{\code{mapOptions}}{Sets Leaflet options for the MiniMap map.
It does not override the MiniMap default map options but extends them.}
  }}
}
\description{
Map component that adds a \link[leaflet:addMiniMap]{minimap} in view mode.
}
\seealso{
\href{https://r-tmap.github.io/tmap/articles/basics_components}{Vignette about components}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tmap_icons.R
\name{tmap_icons}
\alias{tmap_icons}
\alias{marker_icon}
\title{Specify icons}
\usage{
tmap_icons(
  file,
  names = NULL,
  width = 48,
  height = 48,
  keep.asp = TRUE,
  just = c("center", "center"),
  merge = NA,
  as.local = TRUE,
  ...
)

marker_icon()
}
\arguments{
\item{file}{character value/vector containing the file path(s) or url(s).}

\item{names}{names to be given to the icons. Useful when icons are assigned to factor levels.}

\item{width}{width of the icon. If \code{keep.asp}, this is interpreted as the maximum width.}

\item{height}{height of the icon. If \code{keep.asp}, this is interpreted as the maximum height.}

\item{keep.asp}{keep the aspect ratio of the png image. If \code{TRUE} and the aspect
ratio differs from \code{width/height}, either \code{width} or \code{height} is adjusted accordingly.}

\item{just}{justification of the icons relative to the point coordinates.
The first value specifies horizontal and the second value vertical justification.
Possible values are: \code{"left"} , \code{"right"}, \code{"center"}, \code{"bottom"}, and \code{"top"}.
Numeric values of 0 specify left alignment and 1 right alignment.
The default value of \code{just} is \code{c("center", "center")}.}

\item{merge}{merge icons to one icon list (see return value)? If \code{FALSE}, a list is created per file. By default \code{TRUE}, unless \code{names} are specified.}

\item{as.local}{if the \code{file} is a url, should it be saved to local temporary file?}

\item{...}{arguments passed on to \code{\link[leaflet:icons]{leaflet::icons()}}.
When \code{iconWidth}, \code{iconHeight}, \code{iconAnchorX}, and \code{iconAnchorY} are specified,
they override \code{width} and \code{height}, and \code{just}.}
}
\value{
icon data (see \code{\link[leaflet:icons]{leaflet::icons()}})
}
\description{
Specifies icons from a png images, which can be used as markers in thematic maps.
The function \code{marker_icon()} is the specification of the default marker.
}
\seealso{
\code{\link[=tm_symbols]{tm_symbols()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tm_layers_aux.R
\name{tm_graticules}
\alias{tm_graticules}
\alias{tm_grid}
\title{Coordinate grid / graticule lines}
\usage{
tm_graticules(
  x = NA,
  y = NA,
  n.x = NA,
  n.y = NA,
  crs = 4326,
  labels.format = list(suffix = intToUtf8(176)),
  labels.cardinal = TRUE,
  ...
)

tm_grid(
  x = NA,
  y = NA,
  n.x = NA,
  n.y = NA,
  crs = NA,
  col = NA,
  lwd = 1,
  alpha = NA,
  labels.show = TRUE,
  labels.pos = c("left", "bottom"),
  labels.size = 0.6,
  labels.col = NA,
  labels.rot = c(0, 0),
  labels.format = list(big.mark = ","),
  labels.cardinal = FALSE,
  labels.margin.x = 0,
  labels.margin.y = 0,
  labels.space.x = NA,
  labels.space.y = NA,
  labels.inside_frame = FALSE,
  ticks = labels.show & !labels.inside_frame,
  lines = TRUE,
  ndiscr = 100,
  zindex = NA,
  group = NA,
  group.control = "none",
  ...
)
}
\arguments{
\item{x}{X coordinates for vertical grid lines. If \code{NA}, it is specified
with a pretty scale and \code{n.x}.}

\item{y}{Y coordinates for horizontal grid lines. If \code{NA}, it is specified
with a pretty scale and \code{n.y}.}

\item{n.x}{Preferred number of grid lines for the x axis. For the labels,
a \code{\link[=pretty]{pretty()}} sequence is used, so the number of actual labels may be different than \code{n.x}.}

\item{n.y}{Preferred number of grid lines for the y axis. For the labels, a
\code{\link[=pretty]{pretty()}} sequence is used, so the number of actual labels may be different than \code{n.y}.}

\item{crs}{Projection character. If specified, the grid lines are projected
accordingly. Many world maps are projected, but still have latitude longitude
(EPSG 4326) grid lines.}

\item{labels.format}{List of formatting options for the grid labels. Parameters are:
\describe{
\item{fun}{Function to specify the labels. It should take a numeric vector,
and should return a character vector of the same size.
By default it is not specified. If specified, the list items \code{scientific},
\code{format}, and \code{digits} (see below) are not used.}
\item{scientific}{Should the labels be formatted scientifically?
If so, square brackets are used, and the \code{format} of the numbers is \code{"g"}.
Otherwise, \code{format="f"}, and \code{text.separator}, \code{text.less.than}, and
\code{text.or.more} are used. Also, the numbers are automatically rounded to
millions or billions if applicable.}
\item{format}{By default, \code{"f"}, i.e. the standard notation \code{xxx.xxx}, is used.
If \code{scientific=TRUE} then \code{"g"}, which means that numbers are formatted scientifically,
i.e. \code{n.dddE+nn} if needed to save space.}
\item{digits}{Number of digits after the decimal point if \code{format="f"}, and
the number of significant digits otherwise.}
\item{...}{Other arguments passed on to \code{\link[=formatC]{formatC()}}}
}}

\item{labels.cardinal}{Add the four cardinal directions (N, E, S, W) to the labels,
instead of using negative coordinates for west and south (so it assumes that
the coordinates are positive in the north-east direction).}

\item{...}{Used to catch deprecated arguments from tmap v3.}

\item{col}{Color of the grid lines.}

\item{lwd}{Line width of the grid lines}

\item{alpha}{Alpha transparency of the grid lines. Number between 0 and 1.
By default, the alpha transparency of \code{col} is taken.}

\item{labels.show}{Show tick labels. Either one value for both \code{x} and \code{y} axis,
or a vector two: the first for \code{x} and latter for \code{y}.}

\item{labels.pos}{position of the labels. Vector of two: the horizontal
("left" or "right") and the vertical ("top" or "bottom") position.}

\item{labels.size}{Font size of the tick labels}

\item{labels.col}{Font color of the tick labels}

\item{labels.rot}{Rotation angles of the labels. Vector of two values: the
first is the rotation angle (in degrees) of the tick labels on the x axis
and the second is the rotation angle of the tick labels on the y axis.
Only \code{0}, \code{90}, \code{180}, and \code{270} are valid values.}

\item{labels.margin.x}{Margin between tick labels of x axis and the frame.
Note that when \code{labels.inside_frame = FALSE} and \code{ticks = TRUE},
the ticks will be adjusted accordingly.}

\item{labels.margin.y}{Margin between tick labels of y axis and the frame.
Note that when \code{labels.inside_frame = FALSE} and \code{ticks = TRUE},
the ticks will be adjusted accordingly.}

\item{labels.space.x}{Space that is used for the labels and ticks for the x-axis
when \code{labels.inside_frame = FALSE}. By default, it is determined automatically
using the widths and heights of the tick labels. The unit of this parameter is text line height.}

\item{labels.space.y}{Space that is used for the labels and ticks for the y-axis
when \code{labels.inside_frame = FALSE}. By default, it is determined automatically
using the widths and heights of the tick labels. The unit of this parameter is text line height.}

\item{labels.inside_frame}{Show labels inside the frame? By default \code{FALSE}.}

\item{ticks}{If \code{labels.inside_frame = FALSE}, should ticks can be drawn between the labels and the frame?
Either one value for both \code{x} and \code{y} axis, or a vector two: the first for \code{x} and latter for \code{y}.}

\item{lines}{If \code{labels.inside_frame = FALSE}, should grid lines can be drawn?}

\item{ndiscr}{Number of points to discretize a parallel or meridian
(only applicable for curved grid lines)}

\item{zindex}{Controls the stacking order of map layers. Should be set to a
value above 400. By default, layers are stacked in call order, starting
at 401. See details.}

\item{group}{Name of the group to which this layer belongs.
This is only relevant in view mode, where layer groups can be
switched (see \code{group.control})}

\item{group.control}{In view mode, the group control determines how layer groups
can be switched on and off. Options: \code{"radio"} for radio buttons (meaning only one group can be shown),
\code{"check"} for check boxes (so multiple groups can be shown), and \code{"none"}
for no control (the group cannot be (de)selected).}
}
\description{
Draws horizontal and vertical lines according to a coordinate reference system (CRS). \code{tm_grid()} uses the CRS of the (master) shape object, and \code{tm_graticules()} uses latitude and longitude coordinates (EPSG 4326). It creates a \code{\link{tmap-element}} that draws coordinate grid lines. It serves as a
layer that can be drawn anywhere between other layers.
}
\details{
#' In view mode, each layer is rendered in a Leaflet pane named \code{"tmap{zindex}"}
(e.g., \code{"tmap401"}, \code{"tmap402"}), with base tile layers placed in the
standard \code{"tile"} pane.
}
\examples{
\dontrun{
current.mode <- tmap_mode("plot")

tm_shape(NLD_muni) +
	tm_polygons() +
	tm_grid()

tm_shape(NLD_muni) +
	tm_polygons() +
	tm_grid(crs = 4326)

tm_shape(NLD_muni) +
	tm_polygons() +
	tm_grid(crs = 3035, labels.inside.frame = TRUE)

tm_shape(World) +
	tm_polygons() +
	tm_facets(by = "continent") +
	tm_grid(labels.inside.frame = TRUE)

tm_shape(NLD_muni) +
	tm_polygons() +
	tm_graticules()

tm_shape(NLD_muni) +
	tm_polygons() +
	tm_graticules(labels.pos = c("right", "top"))


data(NLD_muni, World)

tmap_arrange(
	qtm(NLD_muni) + tm_grid(),
	qtm(NLD_muni) + tm_graticules()
)

qtm(World, shape.crs = "+proj=robin", style = "natural") +
	tm_graticules(ticks = FALSE) +
	tm_layout(frame=FALSE)

tmap_mode(current.mode)
}
}
\seealso{
\href{https://r-tmap.github.io/tmap/articles/foundations_crs}{Map projections (CRS)}, \href{https://r-tmap.github.io/tmap/articles/examples_choro_NLD}{Choropleth (Netherlands)}, \href{https://r-tmap.github.io/tmap/articles/examples_choro_World}{Choropleth (World)}, \href{https://r-tmap.github.io/tmap/articles/examples_terrain}{Terrain map}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tm_component_functions.R
\name{tm_minimap}
\alias{tm_minimap}
\title{Map component: minimap}
\usage{
tm_minimap(
  server,
  toggle,
  height,
  width,
  margins,
  between_margin,
  position,
  group_id,
  frame,
  frame.color,
  frame.alpha,
  frame.lwd,
  frame.r,
  bg,
  bg.color,
  bg.alpha,
  z,
  ...
)
}
\arguments{
\item{server}{name of the provider or an URL (see \code{\link{tm_tiles}}).
By default, it shows the same map as the basemap, and moreover, it will automatically change when the user switches basemaps.
Note the latter does not happen when \code{server} is specified.}

\item{toggle}{should the minimap have a button to minimise it? By default \code{TRUE}.}

\item{width, height}{width and height of the component.}

\item{margins}{margins}

\item{between_margin}{Margin between}

\item{position}{The position specification of the component: an object created with \code{tm_pos_in()} or \code{tm_pos_out()}. Or, as a shortcut, a vector of two values, specifying the x and y coordinates. The first is \code{"left"}, \code{"center"} or \code{"right"} (or upper case, meaning tighter to the map frame), the second \code{"top"}, \code{"center"} or \code{"bottom"}. Numeric values are also supported, where 0, 0 means left bottom and 1, 1 right top. See also vignette: \href{https://r-tmap.github.io/tmap/articles/adv_positions}{Positioning of components}. In case multiple components should be combined (stacked), use \code{group_id} and specify \code{component} in \code{\link[=tm_components]{tm_components()}}.}

\item{group_id}{Component group id name. All components (e.g. legends, titles, etc) with the same \code{group_id} will be grouped. The specifications of how they are placed (e.g. stacking, margins etc.) are determined in \code{\link[=tm_components]{tm_components()}} where its argument \code{id} should correspond to \code{group_id}.}

\item{frame}{frame should a frame be drawn?}

\item{frame.color}{frame color}

\item{frame.alpha}{frame alpha transparancy}

\item{frame.lwd}{frame line width}

\item{frame.r}{Radius of the rounded frame corners. 0 means no rounding.}

\item{bg}{Show background?}

\item{bg.color}{Background color}

\item{bg.alpha}{Background transparency}

\item{z}{z index, e.g. the place of the component relative to the other componets}

\item{...}{
  Arguments passed on to \code{\link[leaflet:addMiniMap]{leaflet::addMiniMap}}
  \describe{
    \item{\code{map}}{a map widget object}
    \item{\code{collapsedWidth}}{The width of the toggle marker and the minimap
when collapsed, in pixels. Defaults to 19.}
    \item{\code{collapsedHeight}}{The height of the toggle marker and the minimap
when collapsed, in pixels. Defaults to 19.}
    \item{\code{zoomLevelOffset}}{The offset applied to the zoom in the minimap compared
to the zoom of the main map. Can be positive or negative, defaults to -5.}
    \item{\code{zoomLevelFixed}}{Overrides the offset to apply a fixed zoom level to
the minimap regardless of the main map zoom.
Set it to any valid zoom level, if unset zoomLevelOffset is used instead.}
    \item{\code{centerFixed}}{Applies a fixed position to the minimap regardless of
the main map's view / position. Prevents panning the minimap, but does
allow zooming (both in the minimap and the main map).
If the minimap is zoomed, it will always zoom around the centerFixed point.
You can pass in a LatLng-equivalent object. Defaults to false.}
    \item{\code{zoomAnimation}}{Sets whether the minimap should have an animated zoom.
(Will cause it to lag a bit after the movement of the main map.)
Defaults to false.}
    \item{\code{toggleDisplay}}{Sets whether the minimap should have a button to minimize it.
Defaults to false.}
    \item{\code{autoToggleDisplay}}{Sets whether the minimap should hide automatically,
if the parent map bounds does not fit within the minimap bounds.
Especially useful when 'zoomLevelFixed' is set.}
    \item{\code{minimized}}{Sets whether the minimap should start in a minimized position.}
    \item{\code{aimingRectOptions}}{Sets the style of the aiming rectangle by passing in
a Path.Options (\url{https://web.archive.org/web/20220702182250/https://leafletjs.com/reference-1.3.4.html#path-options}) object.
(Clickable will always be overridden and set to false.)}
    \item{\code{shadowRectOptions}}{Sets the style of the aiming shadow rectangle by passing in
a Path.Options (\url{https://web.archive.org/web/20220702182250/https://leafletjs.com/reference-1.3.4.html#path-option}) object.
(Clickable will always be overridden and set to false.)}
    \item{\code{strings}}{Overrides the default strings allowing for translation.}
    \item{\code{tiles}}{URL for tiles or one of the pre-defined providers.}
    \item{\code{mapOptions}}{Sets Leaflet options for the MiniMap map.
It does not override the MiniMap default map options but extends them.}
  }}
}
\description{
Map component that adds a \link[leaflet:addMiniMap]{minimap} in view mode.
}
\seealso{
\href{https://r-tmap.github.io/tmap/articles/basics_components}{Components}, \href{https://r-tmap.github.io/tmap/articles/adv_positions}{Positioning of components}, \href{https://r-tmap.github.io/tmap/articles/adv_comp_group}{Grouping of components}, \href{https://r-tmap.github.io/tmap/articles/examples_choro_NLD}{Choropleth (Netherlands)} \code{.doc_see_also_insets()}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tm_layers_symbols.R
\name{tm_symbols}
\alias{tm_symbols}
\alias{tm_dots}
\alias{tm_bubbles}
\alias{tm_squares}
\alias{tm_markers}
\alias{opt_tm_markers}
\alias{opt_tm_symbols}
\alias{opt_tm_dots}
\alias{opt_tm_bubbles}
\alias{opt_tm_squares}
\title{Map layer: symbols}
\usage{
tm_symbols(
  size = tm_const(),
  size.scale = tm_scale(),
  size.legend = tm_legend(),
  size.chart = tm_chart_none(),
  size.free = NA,
  fill = tm_const(),
  fill.scale = tm_scale(),
  fill.legend = tm_legend(),
  fill.chart = tm_chart_none(),
  fill.free = NA,
  col = tm_const(),
  col.scale = tm_scale(),
  col.legend = tm_legend(),
  col.chart = tm_chart_none(),
  col.free = NA,
  shape = tm_const(),
  shape.scale = tm_scale(),
  shape.legend = tm_legend(),
  shape.chart = tm_chart_none(),
  shape.free = NA,
  lwd = tm_const(),
  lwd.scale = tm_scale(),
  lwd.legend = tm_legend(),
  lwd.chart = tm_chart_none(),
  lwd.free = NA,
  lty = tm_const(),
  lty.scale = tm_scale(),
  lty.legend = tm_legend(),
  lty.chart = tm_chart_none(),
  lty.free = NA,
  fill_alpha = tm_const(),
  fill_alpha.scale = tm_scale(),
  fill_alpha.legend = tm_legend(),
  fill_alpha.chart = tm_chart_none(),
  fill_alpha.free = NA,
  col_alpha = tm_const(),
  col_alpha.scale = tm_scale(),
  col_alpha.legend = tm_legend(),
  col_alpha.chart = tm_chart_none(),
  col_alpha.free = NA,
  plot.order = tm_plot_order("size"),
  zindex = NA,
  group = NA,
  group.control = "check",
  popup.vars = NA,
  popup.format = tm_label_format(),
  hover = NA,
  id = "",
  options = opt_tm_symbols(),
  ...
)

tm_dots(
  fill = tm_const(),
  fill.scale = tm_scale(),
  fill.legend = tm_legend(),
  fill.free = NA,
  size = tm_const(),
  size.scale = tm_scale(),
  size.legend = tm_legend(),
  size.free = NA,
  lwd = tm_const(),
  lwd.scale = tm_scale(),
  lwd.legend = tm_legend(),
  lwd.free = NA,
  lty = tm_const(),
  lty.scale = tm_scale(),
  lty.legend = tm_legend(),
  lty.free = NA,
  fill_alpha = tm_const(),
  fill_alpha.scale = tm_scale(),
  fill_alpha.legend = tm_legend(),
  fill_alpha.free = NA,
  plot.order = tm_plot_order("DATA"),
  zindex = NA,
  group = NA,
  group.control = "check",
  options = opt_tm_dots(),
  ...
)

tm_bubbles(
  size = tm_const(),
  size.scale = tm_scale(),
  size.legend = tm_legend(),
  size.free = NA,
  fill = tm_const(),
  fill.scale = tm_scale(),
  fill.legend = tm_legend(),
  fill.free = NA,
  col = tm_const(),
  col.scale = tm_scale(),
  col.legend = tm_legend(),
  col.free = NA,
  lwd = tm_const(),
  lwd.scale = tm_scale(),
  lwd.legend = tm_legend(),
  lwd.free = NA,
  lty = tm_const(),
  lty.scale = tm_scale(),
  lty.legend = tm_legend(),
  lty.free = NA,
  fill_alpha = tm_const(),
  fill_alpha.scale = tm_scale(),
  fill_alpha.legend = tm_legend(),
  fill_alpha.free = NA,
  col_alpha = tm_const(),
  col_alpha.scale = tm_scale(),
  col_alpha.legend = tm_legend(),
  col_alpha.free = NA,
  plot.order = tm_plot_order("size"),
  zindex = NA,
  group = NA,
  group.control = "check",
  options = opt_tm_bubbles(),
  ...
)

tm_squares(
  size = tm_const(),
  size.scale = tm_scale(),
  size.legend = tm_legend(),
  size.free = NA,
  fill = tm_const(),
  fill.scale = tm_scale(),
  fill.legend = tm_legend(),
  fill.free = NA,
  col = tm_const(),
  col.scale = tm_scale(),
  col.legend = tm_legend(),
  col.free = NA,
  lwd = tm_const(),
  lwd.scale = tm_scale(),
  lwd.legend = tm_legend(),
  lwd.free = NA,
  lty = tm_const(),
  lty.scale = tm_scale(),
  lty.legend = tm_legend(),
  lty.free = NA,
  fill_alpha = tm_const(),
  fill_alpha.scale = tm_scale(),
  fill_alpha.legend = tm_legend(),
  fill_alpha.free = NA,
  col_alpha = tm_const(),
  col_alpha.scale = tm_scale(),
  col_alpha.legend = tm_legend(),
  col_alpha.free = NA,
  plot.order = tm_plot_order("size"),
  zindex = NA,
  group = NA,
  group.control = "check",
  options = opt_tm_squares(),
  ...
)

tm_markers(
  text = tm_const(),
  text.scale = tm_scale(),
  text.legend = tm_legend(),
  text.chart = tm_chart_none(),
  text.free = NA,
  size = tm_const(),
  size.scale = tm_scale(),
  size.legend = tm_legend(),
  size.chart = tm_chart_none(),
  size.free = NA,
  col = tm_const(),
  col.scale = tm_scale(),
  col.legend = tm_legend(),
  col.chart = tm_chart_none(),
  col.free = NA,
  col_alpha = tm_const(),
  col_alpha.scale = tm_scale(),
  col_alpha.legend = tm_legend(),
  col_alpha.chart = tm_chart_none(),
  col_alpha.free = NA,
  fontface = tm_const(),
  fontface.scale = tm_scale(),
  fontface.legend = tm_legend(),
  fontface.chart = tm_chart_none(),
  fontface.free = NA,
  fontfamily = "",
  bgcol = tm_const(),
  bgcol.scale = tm_scale(),
  bgcol.legend = tm_legend(),
  bgcol.chart = tm_chart_none(),
  bgcol.free = NA,
  bgcol_alpha = tm_const(),
  bgcol_alpha.scale = tm_scale(),
  bgcol_alpha.legend = tm_legend(),
  bgcol_alpha.chart = tm_chart_none(),
  bgcol_alpha.free = NA,
  xmod = 0,
  xmod.scale = tm_scale(),
  xmod.legend = tm_legend_hide(),
  xmod.chart = tm_chart_none(),
  xmod.free = NA,
  ymod = 0,
  ymod.scale = tm_scale(),
  ymod.legend = tm_legend_hide(),
  ymod.chart = tm_chart_none(),
  ymod.free = NA,
  angle = 0,
  angle.scale = tm_scale(),
  angle.legend = tm_legend_hide(),
  angle.chart = tm_chart_none(),
  angle.free = NA,
  plot.order = tm_plot_order("AREA", reverse = FALSE, na.order = "bottom"),
  zindex = NA,
  group = NA,
  group.control = "check",
  options = opt_tm_markers(),
  ...
)

opt_tm_markers(
  markers_on_top_of_text = FALSE,
  points_only = "ifany",
  point_per = "feature",
  on_surface = FALSE,
  shadow = FALSE,
  shadow.offset.x = 0.1,
  shadow.offset.y = 0.1,
  just = "center",
  along_lines = TRUE,
  bg.padding = 0.4,
  clustering = TRUE,
  point.label = TRUE,
  point.label.gap = 0.4,
  point.label.method = "SANN",
  remove_overlap = FALSE,
  dots.just = NA,
  dots.icon.scale = 3,
  dots.grob.dim = c(width = 48, height = 48, render.width = 256, render.height = 256)
)

opt_tm_symbols(
  points_only = "ifany",
  point_per = "feature",
  on_surface = FALSE,
  clustering = FALSE,
  icon.scale = 3,
  just = NA,
  grob.dim = c(width = 48, height = 48, render.width = 256, render.height = 256)
)

opt_tm_dots(
  points_only = "ifany",
  point_per = "feature",
  on_surface = FALSE,
  clustering = FALSE,
  icon.scale = 3,
  just = NA,
  grob.dim = c(width = 48, height = 48, render.width = 256, render.height = 256)
)

opt_tm_bubbles(
  points_only = "ifany",
  point_per = "feature",
  on_surface = FALSE,
  clustering = FALSE,
  icon.scale = 3,
  just = NA,
  grob.dim = c(width = 48, height = 48, render.width = 256, render.height = 256)
)

opt_tm_squares(
  points_only = "ifany",
  point_per = "feature",
  on_surface = FALSE,
  clustering = FALSE,
  icon.scale = 3,
  just = NA,
  grob.dim = c(width = 48, height = 48, render.width = 256, render.height = 256)
)
}
\arguments{
\item{size, size.scale, size.legend, size.chart, size.free}{Visual variable that determines the size. See details.}

\item{fill, fill.scale, fill.legend, fill.chart, fill.free}{Visual variable that determines the fill color. See details.}

\item{col, col.scale, col.legend, col.chart, col.free}{Visual variable that determines the color. See details.}

\item{shape, shape.scale, shape.legend, shape.chart, shape.free}{Visual variable that determines the shape. See details.}

\item{lwd, lwd.scale, lwd.legend, lwd.chart, lwd.free}{Visual variable that determines the line width. See details.}

\item{lty, lty.scale, lty.legend, lty.chart, lty.free}{Visual variable that determines the line type. See details.}

\item{fill_alpha, fill_alpha.scale, fill_alpha.legend, fill_alpha.chart, fill_alpha.free}{Visual variable that determines the fill color transparency. See details.
the fill color alpha transparency See details.}

\item{col_alpha, col_alpha.scale, col_alpha.legend, col_alpha.chart, col_alpha.free}{Visual variable that determines the color transparency. See details.}

\item{plot.order}{Specification in which order the spatial features are drawn.
See \code{\link[=tm_plot_order]{tm_plot_order()}} for details.}

\item{zindex}{Map layers are drawn on top of each other. The \code{zindex} numbers
(one for each map layer) determines the stacking order.
By default the map layers are drawn in the order they are called.}

\item{group}{Name of the group to which this layer belongs. This is only
relevant in view mode, where layer groups can be switched (see \code{group.control})}

\item{group.control}{In view mode, the group control determines how layer groups
can be switched on and off. Options: \code{"radio"} for radio buttons
(meaning only one group can be shown), \code{"check"} for check boxes
(so multiple groups can be shown), and \code{"none"} for no control
(the group cannot be (de)selected).}

\item{popup.vars}{names of data variables that are shown in the popups
in \code{"view"} mode. Set popup.vars to \code{TRUE} to show all variables in the
shape object. Set popup.vars to \code{FALSE} to disable popups. Set \code{popup.vars}
to a character vector of variable names to those those variables in the popups.
The default (\code{NA}) depends on whether visual variables (e.g.\code{fill}) are used.
If so, only those are shown. If not all variables in the shape object are shown.}

\item{popup.format}{list of formatting options for the popup values. Output of \code{\link[=tm_label_format]{tm_label_format()}}. Only applicable for
numeric data variables. If one list of formatting options is provided,
it is applied to all numeric variables of \code{popup.vars}. Also, a (named)
list of lists can be provided. In that case, each list of formatting options
is applied to the named variable.}

\item{hover}{name of the data variable that specifies the hover labels (view mode only). Set to \code{FALSE} to disable hover labels. By default \code{FALSE}, unless \code{id} is specified. In that case, it is set to \code{id},}

\item{id}{name of the data variable that specifies the indices of the spatial
features. Only used for \code{"view"} mode.}

\item{options}{options passed on to the corresponding \verb{opt_<layer_function>} function}

\item{...}{to catch deprecated arguments from version < 4.0}

\item{text, text.scale, text.legend, text.chart, text.free}{Visual variable that determines the text. See details.}

\item{fontface, fontface.scale, fontface.legend, fontface.chart, fontface.free}{Visual variable that determines the font face. See details.}

\item{fontfamily}{The font family. See \link[grid:gpar]{gpar()} for details.}

\item{bgcol, bgcol.scale, bgcol.legend, bgcol.chart, bgcol.free}{Visual variable that determines
the background color. See Details.}

\item{bgcol_alpha, bgcol_alpha.scale, bgcol_alpha.legend, bgcol_alpha.chart, bgcol_alpha.free}{Visual variable that determines
the background color transparency. See Details.}

\item{xmod, xmod.scale, xmod.legend, xmod.chart, xmod.free}{Transformation variable that determines the x offset. See details.}

\item{ymod, ymod.scale, ymod.legend, ymod.chart, ymod.free}{Transformation variable that determines the y offset. See details.
the text. See details.}

\item{angle, angle.scale, angle.legend, angle.chart, angle.free}{Rotation angle}

\item{markers_on_top_of_text}{should markers be plot on top of the text (by default \code{FALSE})}

\item{points_only}{should only point geometries of the shape object (defined in \code{\link[=tm_shape]{tm_shape()}}) be plotted? By default \code{"ifany"}, which means \code{TRUE} in case a geometry collection is specified.}

\item{point_per}{specification of how spatial points are mapped when the geometry is a multi line or a multi polygon. One of \code{"feature"}, \code{"segment"} or \code{"largest"}. The first generates a spatial point for every feature, the second for every segment (i.e. subfeature), the third only for the largest segment (subfeature). Note that the last two options can be significant slower.}

\item{on_surface}{In case of polygons, centroids are computed. Should the points be on the surface? If \code{TRUE}, which is slower than the default \code{FALSE}, centroids outside the surface are replaced with points computed with \code{\link[sf:geos_unary]{sf::st_point_on_surface()}}.}

\item{shadow}{Shadow behind the text. Logical.}

\item{shadow.offset.x, shadow.offset.y}{Shadow offset in line heights}

\item{just}{justification of the text relative to the point coordinates. Either one of the following values: \code{"left"} , \code{"right"}, \code{"center"}, \code{"bottom"}, and \code{"top"}, or a vector of two values where first value specifies horizontal and the second value vertical justification. Besides the mentioned values, also numeric values between 0 and 1 can be used. 0 means left justification for the first value and bottom justification for the second value. Note that in view mode, only one value is used.}

\item{along_lines}{logical that determines whether labels are rotated along the spatial lines. Only applicable if a spatial lines shape is used.}

\item{bg.padding}{The padding of the background in terms of line heights.}

\item{clustering}{in interactive modes (e.g. \code{"view"} mode), should clustering be applied at lower zoom levels? Either \code{FALSE} (default), \code{TRUE}, or a mode specific specification, e.g. for \code{"view"} mode \code{\link[leaflet:markerClusterOptions]{markerClusterOptions}}.}

\item{point.label}{logical that determines whether the labels are placed automatically. By default \code{FALSE} for \code{tm_text}, and \code{TRUE} for \code{tm_labels} if the number of labels is less than 500 (otherwise it will be too slow).}

\item{point.label.gap}{numeric that determines the gap between the point and label}

\item{point.label.method}{the optimization method, either \code{"SANN"} for simulated annealing (the default) or \code{"GA"} for a genetic algorithm.}

\item{remove_overlap}{logical that determines whether the overlapping labels are removed}

\item{dots.just}{justification of the text relative to the point coordinates. Either one of the following values: \code{"left"} , \code{"right"}, \code{"center"}, \code{"bottom"}, and \code{"top"}, or a vector of two values where first value specifies horizontal and the second value vertical justification. Besides the mentioned values, also numeric values between 0 and 1 can be used. 0 means left justification for the first value and bottom justification for the second value. Note that in view mode, only one value is used.}

\item{dots.icon.scale}{scaling number that determines how large the icons (or grobs) are in plot mode in comparison to proportional symbols (such as bubbles). In view mode, the size is determined by the icon specification (see \code{\link{tmap_icons}}) or, if grobs are specified by \code{grob.width} and \code{grob.height}}

\item{dots.grob.dim}{vector of four values that determine how grob objects (see details) are shown in view mode. The first and second value are the width and height of the displayed icon. The third and fourth value are the width and height of the rendered png image that is used for the icon. Generally, the third and fourth value should be large enough to render a ggplot2 graphic successfully. Only needed for the view mode.}

\item{icon.scale}{scaling number that determines how large the icons (or grobs) are in plot mode in comparison to proportional symbols (such as bubbles). For view mode, use the argument \code{grob.dim}}

\item{grob.dim}{vector of four values that determine how grob objects (see details) are shown in view mode. The first and second value are the width and height of the displayed icon. The third and fourth value are the width and height of the rendered png image that is used for the icon. Generally, the third and fourth value should be large enough to render a ggplot2 graphic successfully. Only needed for the view mode.}
}
\description{
Map layer that draws symbols Supported visual variables are:
\code{fill} (the fill color), \code{col} (the border color), \code{size} the symbol size,
\code{shape} the symbol shape, \code{lwd} (line width), \code{lty} (line type), \code{fill_alpha}
(fill color alpha transparency) and \code{col_alpha} (border color alpha transparency).
}
\details{
A symbol shape specification is one of the following three options.
\enumerate{
\item{A numeric value that specifies the plotting character of the symbol. See parameter \code{pch} of \code{\link[graphics:points]{points}} and the last example to create a plot with all options. Note that this is not supported for the \code{"view" mode.}}
\item{A \code{\link[grid:grid.grob]{grob}} object, which can be a ggplot2 plot object created with \code{\link[ggplot2:ggplotGrob]{ggplotGrob}}. To specify multiple shapes, a list of grob objects is required. Tip: for proportional symbols, such as donuts or pies, see the extension package \href{https://r-tmap.github.io/tmap/articles/ext_glyphs}{tmap.glyphs}}.
\item{An icon specification, which can be created with \code{\link{tmap_icons}}.}
}
To specify multiple shapes (needed for the \code{shapes} argument), a vector or list of these shape specification is required. The shape specification options can also be mixed. For the \code{shapes} argument, it is possible to use a named vector or list, where the names correspond to the value of the variable specified by the \code{shape} argument.
For small multiples, a list of these shape specification(s) should be provided.
}
\examples{
########################
## plot symbol shapes
########################

# create grid of 25 points in the Atlantic
atlantic_grid = cbind(expand.grid(x = -51:-47, y = 20:24), id = seq_len(25))
x = sf::st_as_sf(atlantic_grid, coords = c("x", "y"), crs = 4326)

tm_shape(x, bbox = tmaptools::bb(x, ext = 1.2)) +
	tm_symbols(shape = "id",
			   size = 2,
			   lwd = 2,
			   fill = "orange",
			   col = "black",
			   shape.scale = tm_scale_asis()) +
	tm_text("id", ymod = -2)
}
\seealso{
\href{https://r-tmap.github.io/tmap/articles/examples_bubble}{Bubble map example} and \href{https://r-tmap.github.io/tmap/articles/examples_terrain}{terrain map example}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tm_crs.R
\name{tm_crs}
\alias{tm_crs}
\title{Set the map projection (CRS)}
\usage{
tm_crs(crs = NA, property = NA, bbox = NULL)
}
\arguments{
\item{crs}{Map projection (CRS). Can be set to an \code{crs} object (see \code{\link[sf:st_crs]{sf::st_crs()}}), a proj4string, an EPSG number, the value \code{"auto"} (automatic crs recommendation), or one the the following generic projections: \code{c("laea", "aeqd", "utm", "pconic", "eqdc", "stere")}. See details.}

\item{property}{Which property should the projection have? One of: \code{"global"}, \code{"area"} (equal-area), \code{"distance"} (equidistant), \code{"shape"} (conformal). Only applicable if \code{crs = "auto"}. See details.}

\item{bbox}{bounding box. Three options: a \code{\link[sf:st_bbox]{sf::st_bbox()}} object, an Open Street Map query (passed on to \code{\link[tmaptools:geocode_OSM]{tmaptools::geocode_OSM()}}), or \code{"FULL"}, which means the whole earth , which means the whole earth (this also guarantees that transformations to another CRS keep the whole earth, unlike \code{\link[sf:st_bbox]{sf::st_bbox()}}).}
}
\description{
This function sets the map projection. It can also be set via \code{\link[=tm_shape]{tm_shape()}}, but \code{tm_crs} is generally recommended for most cases. It can also be determined automatically (see details); however, this is still work-in-progress.
}
\details{
The map projection (\code{crs}) determines in which coordinate system the spatial object is processed and plotted.
See \href{https://r-tmap.github.io/tmap/articles/foundations_crs}{vignette about CRS}. The \code{crs} can be specified in two places: 1) \code{tm_shape()} and \code{tm_crs()}.
In both cases, the map is plotted into the specified \code{crs}. The difference is that in the first option, the \code{crs} is also taken into account in spatial transformation functions, such as the calculation of centroids and cartograms. In the second option, the \code{crs} is only used in the plotting phase.

The automatic crs recommendation (which is still work-in-progress) is the following:\tabular{ll}{
   \strong{Property} \tab \strong{Recommendation} \cr
   \code{global} (for world maps) \tab A pseudocylindrical projection tmap option \code{crs_global}, by default \verb{"eqearth} (Equal Earth). See \url{https://r-tmap.github.io/tmap/articles/41_advanced_crs.html} for more options \cr
   \code{area} (equal area) \tab Lambert Azimuthal Equal Area (\code{laea}) \cr
   \code{distance}	(equidistant) \tab Azimuthal Equidistant (\code{aeqd}) \cr
   \code{shape}	(conformal) \tab Stereographic (\code{stere}) \cr
}


For further info about the available "generic" projects see:
for utm: \url{https://proj.org/en/9.4/operations/projections/utm.html}
for laea: \url{https://proj.org/en/9.4/operations/projections/laea.html}
for aeqd: \url{https://proj.org/en/9.4/operations/projections/aeqd.html}
for pconic: \url{https://proj.org/en/9.4/operations/projections/pconic.html}
for eqdc: \url{https://proj.org/en/9.4/operations/projections/eqdc.html}
}
\note{
Plans are to migrate the functionality regarding generic crs and automatic crs recommendation to a separate package.
}
\examples{
SA = World[World$continent == "South America", ]

# latlon coordinates (WGS84)
tm_shape(SA) +
	tm_polygons() +
	tm_graticules() +
	tm_crs(4326)

tm_list = lapply(c("global", "area", "distance", "shape"), FUN = function(property) {
	tm_shape(SA) +
		tm_polygons() +
		tm_graticules() +
		tm_crs(property = property)	+
	tm_title(property)
})

tmap_arrange(tm_list, nrow = 1)
}
\seealso{
\href{https://r-tmap.github.io/tmap/articles/foundations_crs}{vignette about CRS}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tm_scale_.R
\name{tm_scale_intervals}
\alias{tm_scale_intervals}
\title{Scales: interval scale}
\usage{
tm_scale_intervals(
  n = 5,
  style = ifelse(is.null(breaks), "pretty", "fixed"),
  style.args = list(),
  breaks = NULL,
  interval.closure = "left",
  label.style = "discrete",
  label.select = TRUE,
  midpoint = NULL,
  as.count = FALSE,
  values = NA,
  values.repeat = FALSE,
  values.range = NA,
  values.scale = NA,
  value.na = NA,
  value.null = NA,
  value.neutral = NA,
  labels = NULL,
  label.na = NA,
  label.null = NA,
  label.format = tm_label_format()
)
}
\arguments{
\item{n}{Number of intervals. For some styles (see argument \code{style} below) it is the preferred number rather than the exact number.}

\item{style}{Method to create intervals. Options are \code{"cat"}, \code{"fixed"}, \code{"sd"}, \code{"equal"}, \code{"pretty"}, \code{"quantile"}, \code{"kmeans"}, \code{"hclust"}, \code{"bclust"}, \code{"fisher"}, \code{"jenks"}, \code{"dpih"}, \code{"headtails"}, and \code{"log10_pretty"}. See the details in \code{\link[classInt:classIntervals]{classInt::classIntervals()}} (extra arguments can be passed on via \code{style.args}).}

\item{style.args}{List of extra arguments passed on to \code{\link[classInt:classIntervals]{classInt::classIntervals()}}.}

\item{breaks}{Interval breaks (only used and required when \code{style = "fixed"})}

\item{interval.closure}{value that determines whether where the intervals are closed: \code{"left"} or \code{"right"}. If \code{as.count = TRUE}, \code{inverval.closure} is always set to \code{"left"}.}

\item{label.style}{Either \code{"discrete"} or \code{"continuous"}. If \code{discrete} (default) intervals will be labeled. If \code{continuous}, the legend will be similar to a continuous scale legend (but with discrete colors), where the breaks are labeled.}

\item{label.select}{Which labels are shown? A logical vector is expected, which is repeated along the number of labels. By default \code{TRUE}, which means all.}

\item{midpoint}{The data value that is interpreted as the midpoint. By default it is set to 0 if negative and positive values are present. Useful when values are diverging colors. In that case, the two sides of the color palette are assigned to negative respectively positive values. If all values are positive or all values are negative, then the midpoint is set to \code{NA}, which means that the value that corresponds to the middle color class (see \code{style}) is mapped to the middle color. If it is specified for sequential color palettes (e.g. \code{"Blues"}), then this color palette will be treated as a diverging color palette.}

\item{as.count}{Should the data variable be processed as a count variable? For instance, if \code{style = "pretty"}, \code{n = 2}, and the value range of the variable is 0 to 10, then the column classes for \code{as.count = TRUE} are 0; 1 to 5; 6 to 10 (note that 0 is regarded as an own category) whereas for \code{as.count = FALSE} they are 0 to 5; 5 to 10. Only applicable if \code{style} is \code{"pretty"}, \code{"fixed"}, or \code{"log10_pretty"}. By default \code{FALSE}.}

\item{values}{(generic scale argument) The visual values. For colors (e.g. \code{fill} or \code{col} for \code{\link[=tm_polygons]{tm_polygons()}}) this is a palette name from the \code{cols4all} package (see \code{\link[cols4all:c4a]{cols4all::c4a()}}) or vector of colors, for size (e.g. \code{size} for \code{tm_symbols}) these are a set of sizes (if two values are specified they are interpret as range), for symbol shapes (e.g. \code{shape} for \code{tm_symbols}) these are a set of symbols, etc. The tmap option \code{values.var} contains the default values per visual variable and in some cases also per data type.}

\item{values.repeat}{(generic scale argument) Should the values be repeated in case there are more categories?}

\item{values.range}{(generic scale argument) Range of the values. Vector of two numbers (both between 0 and 1) where the first determines the minimum and the second the maximum. Full range, which means that all values are used, is encoded as \code{c(0, 1)}. For instance, when a gray scale is used for color (from black to white), \code{c(0,1)} means that all colors are used, \verb{0.25, 0.75} means that only colors from dark gray to light gray are used (more precisely \code{"gray25"} to \code{"gray75"}), and \verb{0, 0.5} means that only colors are used from black to middle grey (\code{"grey50"}). When only one number is specified, this is interpreted as the second number (where the first is set to 0). Default values can be set via the tmap option \code{values.range}.}

\item{values.scale}{(generic scale argument) Scaling of the values. Only useful for size-related visual variables, such as \code{size} of \code{\link[=tm_symbols]{tm_symbols()}} and \code{lwd} of \code{\link[=tm_lines]{tm_lines()}}.}

\item{value.na}{(generic scale argument) Value used for missing values. See tmap option \code{"value.na"} for defaults per visual variable.}

\item{value.null}{(generic scale argument) Value used for NULL values. See tmap option \code{"value.null"} for defaults per visual variable. Null data values occur when out-of-scope features are shown (e.g. for a map of Europe showing a data variable per country, the null values are applied to countries outside Europe).}

\item{value.neutral}{(generic scale argument) Value that can be considered neutral. This is used for legends of other visual variables of the same map layer. E.g. when both \code{fill} and \code{size} are used for \code{\link[=tm_symbols]{tm_symbols()}} (using filled circles), the size legend items are filled with the \code{value.neutral} color from the \code{fill.scale} scale, and fill legend items are bubbles of size \code{value.neutral} from the \code{size.scale} scale.}

\item{labels}{(generic scale argument) Labels}

\item{label.na}{(generic scale argument) Label for missing values}

\item{label.null}{(generic scale argument) Label for null (out-of-scope) values}

\item{label.format}{(generic scale argument) Label formatting. Output of \code{\link[=tm_label_format]{tm_label_format()}}}
}
\description{
Scales in tmap are configured by the family of functions with prefix \code{tm_scale}.
Such function should be used for the input of the \code{.scale} arguments in the
layer functions (e.g. \code{fill.scale} in \code{\link[=tm_polygons]{tm_polygons()}}).
The function \code{tm_scale_intervals()} is used for numerical data.
}
\seealso{
\href{https://r-tmap.github.io/tmap/articles/basics_scales}{Scales}, \href{https://r-tmap.github.io/tmap/articles/basics_vv}{Visual variables}, \href{https://r-tmap.github.io/tmap/articles/adv_multivariate}{Multivariate visual variables} \code{\link[=tm_scale]{tm_scale()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tm_layers_symbols.R
\name{tm_symbols}
\alias{tm_symbols}
\alias{tm_dots}
\alias{tm_bubbles}
\alias{tm_squares}
\alias{tm_markers}
\alias{opt_tm_markers}
\alias{opt_tm_symbols}
\alias{opt_tm_dots}
\alias{opt_tm_bubbles}
\alias{opt_tm_squares}
\title{Map layer: symbols}
\usage{
tm_symbols(
  size = tm_const(),
  size.scale = tm_scale(),
  size.legend = tm_legend(),
  size.chart = tm_chart_none(),
  size.free = NA,
  fill = tm_const(),
  fill.scale = tm_scale(),
  fill.legend = tm_legend(),
  fill.chart = tm_chart_none(),
  fill.free = NA,
  col = tm_const(),
  col.scale = tm_scale(),
  col.legend = tm_legend(),
  col.chart = tm_chart_none(),
  col.free = NA,
  shape = tm_const(),
  shape.scale = tm_scale(),
  shape.legend = tm_legend(),
  shape.chart = tm_chart_none(),
  shape.free = NA,
  lwd = tm_const(),
  lwd.scale = tm_scale(),
  lwd.legend = tm_legend(),
  lwd.chart = tm_chart_none(),
  lwd.free = NA,
  lty = tm_const(),
  lty.scale = tm_scale(),
  lty.legend = tm_legend(),
  lty.chart = tm_chart_none(),
  lty.free = NA,
  fill_alpha = tm_const(),
  fill_alpha.scale = tm_scale(),
  fill_alpha.legend = tm_legend(),
  fill_alpha.chart = tm_chart_none(),
  fill_alpha.free = NA,
  col_alpha = tm_const(),
  col_alpha.scale = tm_scale(),
  col_alpha.legend = tm_legend(),
  col_alpha.chart = tm_chart_none(),
  col_alpha.free = NA,
  plot.order = tm_plot_order("size"),
  zindex = NA,
  group = NA,
  group.control = "check",
  popup = tm_popup(),
  popup.vars = NA,
  popup.format = tm_label_format(),
  hover = NA,
  id = "",
  blend = "over",
  options = opt_tm_symbols(),
  ...
)

tm_dots(
  fill = tm_const(),
  fill.scale = tm_scale(),
  fill.legend = tm_legend(),
  fill.free = NA,
  size = tm_const(),
  size.scale = tm_scale(),
  size.legend = tm_legend(),
  size.free = NA,
  lwd = tm_const(),
  lwd.scale = tm_scale(),
  lwd.legend = tm_legend(),
  lwd.free = NA,
  lty = tm_const(),
  lty.scale = tm_scale(),
  lty.legend = tm_legend(),
  lty.free = NA,
  fill_alpha = tm_const(),
  fill_alpha.scale = tm_scale(),
  fill_alpha.legend = tm_legend(),
  fill_alpha.free = NA,
  plot.order = tm_plot_order("DATA"),
  zindex = NA,
  group = NA,
  group.control = "check",
  options = opt_tm_dots(),
  ...
)

tm_bubbles(
  size = tm_const(),
  size.scale = tm_scale(),
  size.legend = tm_legend(),
  size.free = NA,
  fill = tm_const(),
  fill.scale = tm_scale(),
  fill.legend = tm_legend(),
  fill.free = NA,
  col = tm_const(),
  col.scale = tm_scale(),
  col.legend = tm_legend(),
  col.free = NA,
  lwd = tm_const(),
  lwd.scale = tm_scale(),
  lwd.legend = tm_legend(),
  lwd.free = NA,
  lty = tm_const(),
  lty.scale = tm_scale(),
  lty.legend = tm_legend(),
  lty.free = NA,
  fill_alpha = tm_const(),
  fill_alpha.scale = tm_scale(),
  fill_alpha.legend = tm_legend(),
  fill_alpha.free = NA,
  col_alpha = tm_const(),
  col_alpha.scale = tm_scale(),
  col_alpha.legend = tm_legend(),
  col_alpha.free = NA,
  plot.order = tm_plot_order("size"),
  zindex = NA,
  group = NA,
  group.control = "check",
  options = opt_tm_bubbles(),
  ...
)

tm_squares(
  size = tm_const(),
  size.scale = tm_scale(),
  size.legend = tm_legend(),
  size.free = NA,
  fill = tm_const(),
  fill.scale = tm_scale(),
  fill.legend = tm_legend(),
  fill.free = NA,
  col = tm_const(),
  col.scale = tm_scale(),
  col.legend = tm_legend(),
  col.free = NA,
  lwd = tm_const(),
  lwd.scale = tm_scale(),
  lwd.legend = tm_legend(),
  lwd.free = NA,
  lty = tm_const(),
  lty.scale = tm_scale(),
  lty.legend = tm_legend(),
  lty.free = NA,
  fill_alpha = tm_const(),
  fill_alpha.scale = tm_scale(),
  fill_alpha.legend = tm_legend(),
  fill_alpha.free = NA,
  col_alpha = tm_const(),
  col_alpha.scale = tm_scale(),
  col_alpha.legend = tm_legend(),
  col_alpha.free = NA,
  plot.order = tm_plot_order("size"),
  zindex = NA,
  group = NA,
  group.control = "check",
  options = opt_tm_squares(),
  ...
)

tm_markers(
  text = tm_const(),
  text.scale = tm_scale(),
  text.legend = tm_legend(),
  text.chart = tm_chart_none(),
  text.free = NA,
  size = tm_const(),
  size.scale = tm_scale(),
  size.legend = tm_legend(),
  size.chart = tm_chart_none(),
  size.free = NA,
  col = tm_const(),
  col.scale = tm_scale(),
  col.legend = tm_legend(),
  col.chart = tm_chart_none(),
  col.free = NA,
  col_alpha = tm_const(),
  col_alpha.scale = tm_scale(),
  col_alpha.legend = tm_legend(),
  col_alpha.chart = tm_chart_none(),
  col_alpha.free = NA,
  fontface = tm_const(),
  fontface.scale = tm_scale(),
  fontface.legend = tm_legend(),
  fontface.chart = tm_chart_none(),
  fontface.free = NA,
  fontfamily = "",
  bgcol = tm_const(),
  bgcol.scale = tm_scale(),
  bgcol.legend = tm_legend(),
  bgcol.chart = tm_chart_none(),
  bgcol.free = NA,
  bgcol_alpha = tm_const(),
  bgcol_alpha.scale = tm_scale(),
  bgcol_alpha.legend = tm_legend(),
  bgcol_alpha.chart = tm_chart_none(),
  bgcol_alpha.free = NA,
  xmod = 0,
  xmod.scale = tm_scale(),
  xmod.legend = tm_legend_hide(),
  xmod.chart = tm_chart_none(),
  xmod.free = NA,
  ymod = 0,
  ymod.scale = tm_scale(),
  ymod.legend = tm_legend_hide(),
  ymod.chart = tm_chart_none(),
  ymod.free = NA,
  angle = 0,
  angle.scale = tm_scale(),
  angle.legend = tm_legend_hide(),
  angle.chart = tm_chart_none(),
  angle.free = NA,
  plot.order = tm_plot_order("AREA", reverse = FALSE, na.order = "bottom"),
  zindex = NA,
  group = NA,
  group.control = "check",
  options = opt_tm_markers(),
  ...
)

opt_tm_markers(
  markers_on_top_of_text = FALSE,
  points_only = "ifany",
  point_per = "feature",
  on_surface = FALSE,
  shadow = FALSE,
  shadow.offset.x = 0.1,
  shadow.offset.y = 0.1,
  just = "center",
  along_lines = TRUE,
  bg.padding = 0.4,
  clustering = TRUE,
  point.label = TRUE,
  point.label.gap = 0.4,
  point.label.method = "SANN",
  remove_overlap = FALSE,
  dots.just = NA,
  dots.icon.scale = 3,
  dots.grob.dim = c(width = 48, height = 48, render.width = 256, render.height = 256)
)

opt_tm_symbols(
  points_only = "ifany",
  point_per = "feature",
  on_surface = FALSE,
  clustering = FALSE,
  icon.scale = 3,
  just = NA,
  hitbox = "auto",
  grob.dim = c(width = 48, height = 48, render.width = 256, render.height = 256)
)

opt_tm_dots(
  points_only = "ifany",
  point_per = "feature",
  on_surface = FALSE,
  clustering = FALSE,
  icon.scale = 3,
  just = NA,
  hitbox = "auto",
  grob.dim = c(width = 48, height = 48, render.width = 256, render.height = 256)
)

opt_tm_bubbles(
  points_only = "ifany",
  point_per = "feature",
  on_surface = FALSE,
  clustering = FALSE,
  icon.scale = 3,
  just = NA,
  hitbox = "auto",
  grob.dim = c(width = 48, height = 48, render.width = 256, render.height = 256)
)

opt_tm_squares(
  points_only = "ifany",
  point_per = "feature",
  on_surface = FALSE,
  clustering = FALSE,
  icon.scale = 3,
  just = NA,
  hitbox = "auto",
  grob.dim = c(width = 48, height = 48, render.width = 256, render.height = 256)
)
}
\arguments{
\item{size, size.scale, size.legend, size.chart, size.free}{Visual variable that determines the size. See details. \emph{Unit:} Typographic lines ("lines"); 1 line is approx. 1/6 inch. Controlled by \code{values.scale} and \code{tmap_options(values.scale = ...)}.}

\item{fill, fill.scale, fill.legend, fill.chart, fill.free}{Visual variable that determines the fill color. See details. \emph{Unit:} Color -- a color name, hex string.}

\item{col, col.scale, col.legend, col.chart, col.free}{Visual variable that determines the color. See details. \emph{Unit:} Color -- a color name, hex string.}

\item{shape, shape.scale, shape.legend, shape.chart, shape.free}{Visual variable that determines the shape. See details. \emph{Unit:} Integer \code{pch} code (1-25) used as a plotting symbol. See example of \code{tm_symbols()}}

\item{lwd, lwd.scale, lwd.legend, lwd.chart, lwd.free}{Visual variable that determines the line width. See details. \emph{Unit:} Base R line-width units; 1 lwd is approx. 0.75 pt at 96 dpi.}

\item{lty, lty.scale, lty.legend, lty.chart, lty.free}{Visual variable that determines the line type. See details. \emph{Unit:} Integer (1-6) or name: "solid", "dashed", "dotted", "dotdash", "longdash", "twodash".}

\item{fill_alpha, fill_alpha.scale, fill_alpha.legend, fill_alpha.chart, fill_alpha.free}{Visual variable that determines the fill color transparency. See details. \emph{Unit:} Proportion -- numeric 0-1 (0 = fully transparent, 1 = fully opaque).
the fill color alpha transparency See details.}

\item{col_alpha, col_alpha.scale, col_alpha.legend, col_alpha.chart, col_alpha.free}{Visual variable that determines the color transparency. See details. \emph{Unit:} Proportion -- numeric 0-1 (0 = fully transparent, 1 = fully opaque).}

\item{plot.order}{Specification in which order the spatial features are drawn.
See \code{\link[=tm_plot_order]{tm_plot_order()}} for details.}

\item{zindex}{Controls the stacking order of map layers. Should be set to a
value above 400. By default, layers are stacked in call order, starting
at 401. See details.}

\item{group}{Name of the group to which this layer belongs. This is only
relevant in view mode, where layer groups can be switched (see \code{group.control})}

\item{group.control}{In view mode, the group control determines how layer groups
can be switched on and off. Options: \code{"radio"} for radio buttons
(meaning only one group can be shown), \code{"check"} for check boxes
(so multiple groups can be shown), and \code{"none"} for no control
(the group cannot be (de)selected).}

\item{popup}{popup specification for \code{"view"} mode, the output of \code{\link[=tm_popup]{tm_popup()}}.
It determines the data variables shown in the popup table, the popup title,
and (in the future) the popup layout. This replaces the deprecated
arguments \code{popup.vars} and \code{popup.format}.}

\item{popup.vars}{(Deprecated.) Use \code{popup} with
\code{\link[=tm_popup]{tm_popup()}} instead (via its \code{vars} argument). Names of data variables that
are shown in the popups in \code{"view"} mode. Set \code{popup.vars} to \code{TRUE} to show
all variables in the shape object. Set \code{popup.vars} to \code{FALSE} to disable
popups. Set \code{popup.vars} to a character vector of variable names to show
those variables in the popups. The default (\code{NA}) depends on whether visual
variables (e.g. \code{fill}) are used. If so, only those are shown. If not, all
variables in the shape object are shown.}

\item{popup.format}{(Deprecated.) Use \code{popup} with
\code{\link[=tm_popup]{tm_popup()}} instead (via its \code{format} argument). List of formatting options
for the popup values. Output of \code{\link[=tm_label_format]{tm_label_format()}}. Only applicable for
numeric data variables. If one list of formatting options is provided,
it is applied to all numeric variables of \code{popup.vars}. Also, a (named)
list of lists can be provided. In that case, each list of formatting options
is applied to the named variable.}

\item{hover}{name of the data variable that specifies the hover labels (view mode only). Set to \code{FALSE} to disable hover labels. By default \code{FALSE}, unless \code{id} is specified. In that case, it is set to \code{id},}

\item{id}{name of the data variable that specifies the indices of the spatial
features. Only used for \code{"view"} mode.}

\item{blend}{Compositing operator for layer blending. Default \code{"over"} applies
no blending. See the "Layer blending" section for the supported values.}

\item{options}{options passed on to the corresponding \verb{opt_<layer_function>} function}

\item{...}{to catch deprecated arguments from version < 4.0}

\item{text, text.scale, text.legend, text.chart, text.free}{Visual variable that determines the text. See details. \emph{Unit:} Character string.}

\item{fontface, fontface.scale, fontface.legend, fontface.chart, fontface.free}{Visual variable that determines the font face. See details. \emph{Unit:} "plain", "bold", "italic", or "bold.italic".}

\item{fontfamily}{The font family. See \link[grid:gpar]{gpar()} for details.}

\item{bgcol, bgcol.scale, bgcol.legend, bgcol.chart, bgcol.free}{Visual variable that determines
the background color. See Details. \emph{Unit:} Color -- a color name, hex string.}

\item{bgcol_alpha, bgcol_alpha.scale, bgcol_alpha.legend, bgcol_alpha.chart, bgcol_alpha.free}{Visual variable that determines
the background color transparency. See Details. \emph{Unit:} Proportion -- numeric 0-1 (0 = fully transparent, 1 = fully opaque).}

\item{xmod, xmod.scale, xmod.legend, xmod.chart, xmod.free}{Transformation variable that determines the x offset. See details. \emph{Unit:} Line heights, relative to the label anchor. Positive = right.}

\item{ymod, ymod.scale, ymod.legend, ymod.chart, ymod.free}{Transformation variable that determines the y offset. See details. \emph{Unit:} Line heights, relative to the label anchor. Positive = up.
the text. See details.}

\item{angle, angle.scale, angle.legend, angle.chart, angle.free}{Rotation angle \emph{Unit:} Degrees, clockwise from north (0-360).}

\item{markers_on_top_of_text}{should markers be plot on top of the text (by default \code{FALSE})}

\item{points_only}{should only point geometries of the shape object (defined in \code{\link[=tm_shape]{tm_shape()}}) be plotted? By default \code{"ifany"}, which means \code{TRUE} in case a geometry collection is specified.}

\item{point_per}{specification of how spatial points are mapped when the geometry is a multi line or a multi polygon. One of \code{"feature"}, \code{"segment"} or \code{"largest"}. The first generates a spatial point for every feature, the second for every segment (i.e. subfeature), the third only for the largest segment (subfeature). Note that the last two options can be significant slower.}

\item{on_surface}{In case of polygons, centroids are computed. Should the points be on the surface? If \code{TRUE}, which is slower than the default \code{FALSE}, centroids outside the surface are replaced with points computed with \code{\link[sf:st_point_on_surface]{sf::st_point_on_surface()}}.}

\item{shadow}{Shadow behind the text. Logical.}

\item{shadow.offset.x, shadow.offset.y}{Shadow offset in line heights}

\item{just}{justification of the text relative to the point coordinates. Either one of the following values: \code{"left"} , \code{"right"}, \code{"center"}, \code{"bottom"}, and \code{"top"}, or a vector of two values where first value specifies horizontal and the second value vertical justification. Besides the mentioned values, also numeric values between 0 and 1 can be used. 0 means left justification for the first value and bottom justification for the second value. Note that in view mode, only one value is used.}

\item{along_lines}{logical that determines whether labels are rotated along the spatial lines. Only applicable if a spatial lines shape is used.}

\item{bg.padding}{The padding of the background in terms of line heights.}

\item{clustering}{in interactive modes (e.g. \code{"view"} mode), should clustering be applied at lower zoom levels? Either \code{FALSE} (default), \code{TRUE}, or a mode specific specification, e.g. for \code{"view"} mode \code{\link[leaflet:markerClusterOptions]{markerClusterOptions}}.}

\item{point.label}{logical that determines whether the labels are placed automatically. By default \code{FALSE} for \code{tm_text}, and \code{TRUE} for \code{tm_labels} if the number of labels is less than 500 (otherwise it will be too slow).}

\item{point.label.gap}{numeric that determines the gap between the point and label}

\item{point.label.method}{the optimization method, either \code{"SANN"} for simulated annealing (the default) or \code{"GA"} for a genetic algorithm.}

\item{remove_overlap}{logical that determines whether the overlapping labels are removed}

\item{dots.just}{justification of the text relative to the point coordinates. Either one of the following values: \code{"left"} , \code{"right"}, \code{"center"}, \code{"bottom"}, and \code{"top"}, or a vector of two values where first value specifies horizontal and the second value vertical justification. Besides the mentioned values, also numeric values between 0 and 1 can be used. 0 means left justification for the first value and bottom justification for the second value. Note that in view mode, only one value is used.}

\item{dots.icon.scale}{scaling number that determines how large the icons (or grobs) are in plot mode in comparison to proportional symbols (such as bubbles). In view mode, the size is determined by the icon specification (see \code{\link{tmap_icons}}) or, if grobs are specified by \code{grob.width} and \code{grob.height}}

\item{dots.grob.dim}{vector of four values that determine how grob objects (see details) are shown in view mode. The first and second value are the width and height of the displayed icon. The third and fourth value are the width and height of the rendered png image that is used for the icon. Generally, the third and fourth value should be large enough to render a ggplot2 graphic successfully. Only needed for the view mode.}

\item{icon.scale}{scaling number that determines how large the icons (or grobs) are in plot mode in comparison to proportional symbols (such as bubbles). For view mode, use the argument \code{grob.dim}}

\item{hitbox}{Controls whether an invisible interaction layer with a
larger clickable area (\"hitbox\") is added on top of the symbols.

This can improve click and popup behaviour for small or densely packed
symbols by enlarging the effective mouse interaction area.

Possible values:
\describe{
\item{"none"}{No additional hitbox layer is added. Symbols are clickable
only at their visible size.}

\if{html}{\out{<div class="sourceCode">}}\preformatted{\\item\{"plusX"\}\{Adds \code{X} pixels to the visible symbol diameter to
compute the interaction size. For example, \code{"plus8"} widens the
clickable area by 4 pixels around each symbol edge.\}

\\item\{"pmaxX"\}\{Ensures a minimum interaction diameter of \code{X} pixels:
\code{pmax(symbol_diameter, X)}. Useful only for very small symbols, as
it adds no margin to symbols already larger than \code{X}.\}

\\item\{"auto"\}\{\code{"pmax12"} if and only if interactive features are
enabled (popup or hover), symbols are small (median visible diameter
< 12px), and there are fewer than 10000 features. Otherwise \code{"none"}.\}
}\if{html}{\out{</div>}}

}

\code{plus} and \code{pmax} can be combined, e.g. \code{"plus4pmax8"}.
The hitbox is not applied when clustering is enabled.}

\item{grob.dim}{vector of four values that determine how grob objects (see details) are shown in view mode. The first and second value are the width and height of the displayed icon. The third and fourth value are the width and height of the rendered png image that is used for the icon. Generally, the third and fourth value should be large enough to render a ggplot2 graphic successfully. Only needed for the view mode.}
}
\description{
Map layer that draws symbols Supported visual variables are:
\code{fill} (the fill color), \code{col} (the border color), \code{size} the symbol size,
\code{shape} the symbol shape, \code{lwd} (line width), \code{lty} (line type), \code{fill_alpha}
(fill color alpha transparency) and \code{col_alpha} (border color alpha transparency).
}
\details{
A symbol shape specification is one of the following three options.
\enumerate{
\item{A numeric value that specifies the plotting character of the symbol. See parameter \code{pch} of \code{\link[graphics:points]{points}} and the last example to create a plot with all options. Note that this is not supported for the \code{"view" mode.}}
\item{A \code{\link[grid:grid.grob]{grob}} object, which can be a ggplot2 plot object created with \code{\link[ggplot2:ggplotGrob]{ggplotGrob}}. To specify multiple shapes, a list of grob objects is required. Tip: for proportional symbols, such as donuts or pies, see the extension package vignette: \href{https://r-tmap.github.io/tmap/articles/ext_glyphs}{tmap.glyphs}.}
\item{An icon specification, which can be created with \code{\link{tmap_icons}}.}
}
To specify multiple shapes (needed for the \code{shapes} argument), a vector or list of these shape specification is required. The shape specification options can also be mixed. For the \code{shapes} argument, it is possible to use a named vector or list, where the names correspond to the value of the variable specified by the \code{shape} argument.
For small multiples, a list of these shape specification(s) should be provided.
}
\examples{
########################
## plot symbol shapes
########################

# create grid of 25 points in the Atlantic
atlantic_grid = cbind(expand.grid(x = -51:-47, y = 20:24), id = seq_len(25))
x = sf::st_as_sf(atlantic_grid, coords = c("x", "y"), crs = 4326)

tm_shape(x, bbox = tmaptools::bb(x, ext = 1.2)) +
	tm_symbols(shape = "id",
			   size = 2,
			   lwd = 2,
			   fill = "orange",
			   col = "black",
			   shape.scale = tm_scale_asis()) +
	tm_text("id", ymod = -2)
}
\seealso{
\href{https://r-tmap.github.io/tmap/articles/basics_vv}{Visual variables}, \href{https://r-tmap.github.io/tmap/articles/basics_scales}{Scales}, \href{https://r-tmap.github.io/tmap/articles/basics_legends}{Legends}, \href{https://r-tmap.github.io/tmap/articles/basics_facets}{Facets}, \href{https://r-tmap.github.io/tmap/articles/foundations_units}{Units}, \href{https://r-tmap.github.io/tmap/articles/adv_blend}{Layer blending}, \href{https://r-tmap.github.io/tmap/articles/examples_bubble}{Bubble map}, \href{https://r-tmap.github.io/tmap/articles/examples_terrain}{Terrain map}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tm_add_legend.R
\name{tm_add_legend}
\alias{tm_add_legend}
\title{Map component: manual legend}
\usage{
tm_add_legend(
  ...,
  labels = "",
  type = "symbols",
  title = "",
  orientation = NULL,
  position = NULL,
  group_id = NA_character_,
  group = NA,
  group.control = "check",
  z = NA_integer_
)
}
\arguments{
\item{...}{visual variables and arguments passed on to \code{tm_legend()}.
By default, the argument \code{type} is set to \code{"symbols"}, which means that the
supported visual variables are: \code{"fill"}, \code{"col"}, \code{"shape"}, \code{"size"},
\code{"fill_alpha"}, \code{"col_alpha"}, \code{"lty"}, \code{"lwd"}, \code{"linejoin"}, and \code{"lineend"}.
The number of legend items will be equal to the maximum number of specific values (and specified labels.)}

\item{labels}{labels by default \code{""} (so omitted)}

\item{type}{the layer type from which the visual variables (see \code{...}) are taken.
Options: \code{"symbols"} (default), \code{"lines"}, \code{"polygons"}, and \code{"text"}.}

\item{title}{The title of the legend.}

\item{orientation}{The orientation of the legend.}

\item{position}{The position of the legend. A tm_pos object, or a shortcut of two values: horizontal (left, center, right) and vertical (top, center, bottom). See tm_pos for details}

\item{group_id}{Component group id name. All components (e.g. legends, titles, etc) with the same \code{group_id} will be grouped. The specifications of how they are placed (e.g. stacking, margins etc.) are determined in \code{\link[=tm_components]{tm_components()}} where its argument \code{id} should correspond to \code{group_id}.}

\item{group}{Name of the group to which this layer belongs. This is only
relevant in view mode, where layer groups can be switched (see \code{group.control})}

\item{group.control}{In view mode, the group control determines how
layer groups can be switched on and off. Options: \code{"radio"} for radio
buttons (meaning only one group can be shown), \code{"check"} for check boxes
(so multiple groups can be shown), and \code{"none"} for no control
(the group cannot be (de)selected).}

\item{z}{z index, e.g. the place of the component relative to the other componets}
}
\description{
Map component that adds a manual legend.
}
\examples{
\dontrun{
tm_shape(NLD_muni) +
	tm_borders() +
	tm_basemap("OpenStreetMap") +
	tm_add_legend(labels = c("Motorway", "Primary road", "Secondary road", "Railway"),
				  col = c("#E892A1", "#FCD6A4", "#F8FABF", "#707070"),
				  lty = c("solid", "solid", "solid", "dotted"),
				  lwd = 3,
				  type = "lines",
				  bg.color = "grey92",
				  bg.alpha = 1)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tm_crs.R
\name{tm_crs}
\alias{tm_crs}
\title{Set the map projection (CRS)}
\usage{
tm_crs(crs = NA, property = NA, bbox = NULL, ...)
}
\arguments{
\item{crs}{Map projection (CRS). Can be set to an \code{crs} object (see \code{\link[sf:st_crs]{sf::st_crs()}}), a proj4string, an EPSG number, the value \code{"auto"} (automatic crs recommendation), or one the the following generic projections: \code{c("laea", "aeqd", "utm", "pconic", "eqdc", "stere")}. See details.}

\item{property}{Which property should the projection have? One of: \code{"global"}, \code{"area"} (equal-area), \code{"distance"} (equidistant), \code{"shape"} (conformal). Only applicable if \code{crs = "auto"}. See details.}

\item{bbox}{bounding box. Three options: a \code{\link[sf:st_bbox]{sf::st_bbox()}} object, an Open Street Map query (passed on to \code{\link[tmaptools:geocode_OSM]{tmaptools::geocode_OSM()}}), or \code{"FULL"}, which means the whole earth , which means the whole earth (this also guarantees that transformations to another CRS keep the whole earth, unlike \code{\link[sf:st_bbox]{sf::st_bbox()}}).}

\item{...}{passed on to \code{\link[tmaptools:bb]{bb}} (e.g. \code{ext} can be used to enlarge or shrink a bounding box)}
}
\description{
This function sets the map projection. It can also be set via \code{\link[=tm_shape]{tm_shape()}}, but \code{tm_crs} is generally recommended for most cases. It can also be determined automatically (see details); however, this is still work-in-progress.
}
\details{
The map projection (\code{crs}) determines in which coordinate system the spatial object is processed and plotted.
See vignette: \href{https://r-tmap.github.io/tmap/articles/foundations_crs}{Map projections (CRS)}. The \code{crs} can be specified in two places: 1) \code{tm_shape()} and \code{tm_crs()}.
In both cases, the map is plotted into the specified \code{crs}. The difference is that in the first option, the \code{crs} is also taken into account in spatial transformation functions, such as the calculation of centroids and cartograms. In the second option, the \code{crs} is only used in the plotting phase.

The automatic crs recommendation (which is still work-in-progress) is the following:\tabular{ll}{
   \strong{Property} \tab \strong{Recommendation} \cr
   \code{global} (for world maps) \tab A pseudocylindrical projection tmap option \code{crs_global}, by default \verb{"eqearth} (Equal Earth). See vignette: \href{https://r-tmap.github.io/tmap/articles/foundations_crs}{Map projections (CRS)} for more options \cr
   \code{area} (equal area) \tab Lambert Azimuthal Equal Area (\code{laea}) \cr
   \code{distance}	(equidistant) \tab Azimuthal Equidistant (\code{aeqd}) \cr
   \code{shape}	(conformal) \tab Stereographic (\code{stere}) \cr
}


For further info about the available "generic" projects see:
for utm: \url{https://proj.org/en/9.4/operations/projections/utm.html}
for laea: \url{https://proj.org/en/9.4/operations/projections/laea.html}
for aeqd: \url{https://proj.org/en/9.4/operations/projections/aeqd.html}
for pconic: \url{https://proj.org/en/9.4/operations/projections/pconic.html}
for eqdc: \url{https://proj.org/en/9.4/operations/projections/eqdc.html}
}
\note{
Plans are to migrate the functionality regarding generic crs and automatic crs recommendation to a separate package.
}
\examples{
SA = World[World$continent == "South America", ]

# latlon coordinates (WGS84)
tm_shape(SA) +
	tm_polygons() +
	tm_graticules() +
	tm_crs(4326)

tm_list = lapply(c("global", "area", "distance", "shape"), FUN = function(property) {
	tm_shape(SA) +
		tm_polygons() +
		tm_graticules() +
		tm_crs(property = property)	+
	tm_title(property)
})

tmap_arrange(tm_list, nrow = 1)
}
\seealso{
\href{https://r-tmap.github.io/tmap/articles/foundations_crs}{Map projections (CRS)}, \href{https://r-tmap.github.io/tmap/articles/examples_choro_NLD}{Choropleth (Netherlands)}, \href{https://r-tmap.github.io/tmap/articles/examples_choro_World}{Choropleth (World)}, \href{https://r-tmap.github.io/tmap/articles/examples_terrain}{Terrain map}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tm_layers_polygons.R
\name{tm_polygons}
\alias{tm_polygons}
\alias{tm_fill}
\alias{tm_borders}
\alias{opt_tm_polygons}
\title{Map layer: polygons}
\usage{
tm_polygons(
  fill = tm_const(),
  fill.scale = tm_scale(),
  fill.legend = tm_legend(),
  fill.chart = tm_chart_none(),
  fill.free = NA,
  col = tm_const(),
  col.scale = tm_scale(),
  col.legend = tm_legend(),
  col.chart = tm_chart_none(),
  col.free = NA,
  lwd = tm_const(),
  lwd.scale = tm_scale(),
  lwd.legend = tm_legend(),
  lwd.chart = tm_chart_none(),
  lwd.free = NA,
  lty = tm_const(),
  lty.scale = tm_scale(),
  lty.legend = tm_legend(),
  lty.chart = tm_chart_none(),
  lty.free = NA,
  fill_alpha = tm_const(),
  fill_alpha.scale = tm_scale(),
  fill_alpha.legend = tm_legend(),
  fill_alpha.chart = tm_chart_none(),
  fill_alpha.free = NA,
  col_alpha = tm_const(),
  col_alpha.scale = tm_scale(),
  col_alpha.legend = tm_legend(),
  col_alpha.chart = tm_chart_none(),
  col_alpha.free = NA,
  linejoin = "round",
  lineend = "round",
  plot.order = tm_plot_order("lwd", reverse = TRUE, na.order = "bottom"),
  zindex = NA,
  group = NA,
  group.control = "check",
  popup.vars = NA,
  popup.format = tm_label_format(),
  hover = NA,
  id = "",
  options = opt_tm_polygons(),
  ...
)

tm_fill(...)

tm_borders(col = tm_const(), ...)

opt_tm_polygons(polygons.only = "ifany")
}
\arguments{
\item{fill, fill.scale, fill.legend, fill.chart, fill.free}{Visual variable that determines the fill color. See details.}

\item{col, col.scale, col.legend, col.chart, col.free}{Visual variable that determines the color. See details.}

\item{lwd, lwd.scale, lwd.legend, lwd.chart, lwd.free}{Visual variable that determines the line width. See details.}

\item{lty, lty.scale, lty.legend, lty.chart, lty.free}{Visual variable that determines the line type. See details.}

\item{fill_alpha, fill_alpha.scale, fill_alpha.chart, fill_alpha.legend, fill_alpha.free}{Visual variable that determines the fill color transparency. See details.}

\item{col_alpha, col_alpha.scale, col_alpha.legend, col_alpha.chart, col_alpha.free}{Visual variable that determines the color transparency. See details.}

\item{linejoin, lineend}{Line join and line end. See \link[grid:gpar]{gpar()} for details.}

\item{plot.order}{Specification in which order the spatial features are drawn.
See \code{\link[=tm_plot_order]{tm_plot_order()}} for details.}

\item{zindex}{Map layers are drawn on top of each other. The \code{zindex} numbers
(one for each map layer) determines the stacking order.
By default the map layers are drawn in the order they are called.}

\item{group}{Name of the group to which this layer belongs. This is only
relevant in view mode, where layer groups can be switched (see \code{group.control})}

\item{group.control}{In view mode, the group control determines how layer groups
can be switched on and off. Options: \code{"radio"} for radio buttons
(meaning only one group can be shown), \code{"check"} for check boxes
(so multiple groups can be shown), and \code{"none"} for no control
(the group cannot be (de)selected).}

\item{popup.vars}{names of data variables that are shown in the popups
in \code{"view"} mode. Set popup.vars to \code{TRUE} to show all variables in the
shape object. Set popup.vars to \code{FALSE} to disable popups. Set \code{popup.vars}
to a character vector of variable names to those those variables in the popups.
The default (\code{NA}) depends on whether visual variables (e.g.\code{fill}) are used.
If so, only those are shown. If not all variables in the shape object are shown.}

\item{popup.format}{list of formatting options for the popup values. Output of \code{\link[=tm_label_format]{tm_label_format()}}. Only applicable for
numeric data variables. If one list of formatting options is provided,
it is applied to all numeric variables of \code{popup.vars}. Also, a (named)
list of lists can be provided. In that case, each list of formatting options
is applied to the named variable.}

\item{hover}{name of the data variable that specifies the hover labels (view mode only). Set to \code{FALSE} to disable hover labels. By default \code{FALSE}, unless \code{id} is specified. In that case, it is set to \code{id},}

\item{id}{name of the data variable that specifies the indices of the spatial
features. Only used for \code{"view"} mode.}

\item{options}{options passed on to the corresponding \verb{opt_<layer_function>} function}

\item{...}{to catch deprecated arguments from version < 4.0}

\item{polygons.only}{should only polygon geometries of the shape object (defined in \code{\link[=tm_shape]{tm_shape()}}) be plotted? By default \code{"ifany"}, which means \code{TRUE} in case a geometry collection is specified.}
}
\description{
Map layer that draws polygons. Supported visual variables are: \code{fill} (the fill color),
\code{col} (the border color), \code{lwd} (line width), \code{lty} (line type),
\code{fill_alpha} (fill color alpha transparency) and \code{col_alpha} (border color alpha transparency).
}
\details{
The visual variable arguments (e.g. \code{col}) can be specified with a data
variable name (e.g., a spatial vector attribute or a raster layer of the object
specified in \code{\link[=tm_shape]{tm_shape()}}), with a visual value (for \code{col}, a color is expected), or with a geometry-derived variable (see below).
See \href{https://r-tmap.github.io/tmap/articles/basics_vv}{vignette about visual variables}.

Multiple values can be specified: in that case facets are created.
These facets can be combined with other faceting data variables, specified with \code{\link[=tm_facets]{tm_facets()}}.
See \href{https://r-tmap.github.io/tmap/articles/basics_facets}{vignette about facets}.
\itemize{
\item The \verb{*.scale} arguments determine the used scale to map the data values to
visual variable values. These can be specified with one of the available
\verb{tm_scale_*()} functions. The default is specified by the tmap option (\code{\link[=tm_options]{tm_options()}}) \code{scales.var}.
See \href{https://r-tmap.github.io/tmap/articles/basics_scales}{vignette about scales}.
\item The \verb{*.legend} arguments determine the used legend, specified with \code{\link[=tm_legend]{tm_legend()}}.
The default legend and its settings are determined by the tmap options (\code{\link[=tm_options]{tm_options()}}) \code{legend.} .
See \href{https://r-tmap.github.io/tmap/articles/basics_legends}{vignette about legends}.
\item The \verb{*.chart} arguments specify additional charts, specified with \code{tm_chart_}, e.g. \code{\link[=tm_chart_histogram]{tm_chart_histogram()}}.
See \href{https://r-tmap.github.io/tmap/articles/basics_charts}{vignette about charts}.
\item The \verb{*.free} arguments determine whether scales are applied freely across facets, or shared.
A logical value is required. They can also be specified with a vector of three
logical values; these determine whether scales are applied freely per facet dimension.
This is only useful when facets are applied (see \code{\link[=tm_facets]{tm_facets()}}).
There are maximally three facet dimensions: rows, columns, and pages. This only
applies for a facet grid (\code{\link[=tm_facets_grid]{tm_facets_grid()}}). For instance, \code{col.free = c(TRUE, FALSE, FALSE)}
means that for the visual variable \code{col}, each row of facets will have its own
scale, and therefore its own legend. For facet wraps and stacks
(\code{\link[=tm_facets_wrap]{tm_facets_wrap()}} and \code{\link[=tm_facets_stack]{tm_facets_stack()}}) there is only one facet dimension,
so the \verb{*.free} argument requires only one logical value.
}

Currently, three geometry-derived variables are implemented:
\itemize{
\item \code{"AREA"} (polygons only), which uses the feature area;
\item \code{"LENGTH"} (lines only), which uses the feature length; and
\item \code{"MAP_COLORS"}, which assigns values so that adjacent features receive
different values, making it particularly suitable for coloring
neighbouring polygons.
}

Note that geometry-derived variables do not generate a legend automatically.
If a legend is required, compute the corresponding variable explicitly,
for example with \code{\link[sf:geos_measures]{sf::st_area()}}, \code{\link[sf:geos_measures]{sf::st_length()}}, or
\code{\link[tmaptools:map_coloring]{tmaptools::map_coloring()}}, and use the resulting values instead.
}
\examples{
# load Africa country data
data(World)
Africa = World[World$continent == "Africa", ]
Africa_border = sf::st_make_valid(sf::st_union(sf::st_buffer(Africa, 0.001))) # slow and ugly

# without specifications
tm_shape(Africa_border) + tm_polygons()
tm_shape(Africa_border) + tm_fill()
tm_shape(Africa_border) + tm_borders()

# specification with visual variable values
tm_shape(Africa) +
  tm_polygons(fill = "limegreen", col = "purple", lwd = 2, lty = "solid", col_alpha = 0.3) +
  tm_text("name", options = opt_tm_text(remove_overlap = TRUE)) +
tm_shape(Africa_border) +
  tm_borders("darkred", lwd = 3)

# specification with a data variable
tm_shape(Africa) +
  tm_polygons(fill = "income_grp", fill.scale = tm_scale_categorical(values = "-tol.muted"))


# continuous color scale with landscape legend
tm_shape(Africa) +
  tm_polygons(fill = "inequality",
    fill.scale = tm_scale_continuous(values = "-scico.roma"),
    fill.legend = tm_legend(
    	title = "", orientation = "landscape",
    	position = tm_pos_out("center", "bottom"), frame = FALSE
    	)) +
tm_shape(Africa_border) +
	tm_borders(lwd = 2) +
tm_title("Inequality index", position = tm_pos_in("right", "TOP"), frame = FALSE) +
tm_layout(frame = FALSE)

# bivariate scale
tm_shape(World) +
	tm_polygons(tm_vars(c("inequality", "well_being"), multivariate = TRUE))
}
\seealso{
\href{https://r-tmap.github.io/tmap/articles/examples_choro_World}{Choropleth example (1)} and \href{https://r-tmap.github.io/tmap/articles/examples_choro_NLD}{choropleth example (2)}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tm_add_legend.R
\name{tm_add_legend}
\alias{tm_add_legend}
\title{Map component: manual legend}
\usage{
tm_add_legend(
  ...,
  labels = "",
  type = "symbols",
  title = "",
  orientation = NULL,
  position = NULL,
  group_id = NA_character_,
  group = NA,
  group.control = "check",
  z = NA_integer_
)
}
\arguments{
\item{...}{visual variables and arguments passed on to \code{tm_legend()}.
By default, the argument \code{type} is set to \code{"symbols"}, which means that the
supported visual variables are: \code{"fill"}, \code{"col"}, \code{"shape"}, \code{"size"},
\code{"fill_alpha"}, \code{"col_alpha"}, \code{"lty"}, \code{"lwd"}, \code{"linejoin"}, and \code{"lineend"}.
The number of legend items will be equal to the maximum number of specific values (and specified labels.)}

\item{labels}{labels by default \code{""} (so omitted)}

\item{type}{the layer type from which the visual variables (see \code{...}) are taken.
Options: \code{"symbols"} (default), \code{"lines"}, \code{"polygons"}, and \code{"text"}.}

\item{title}{The title of the legend.}

\item{orientation}{The orientation of the legend.}

\item{position}{The position of the legend. A tm_pos object, or a shortcut of two values: horizontal (left, center, right) and vertical (top, center, bottom). See tm_pos for details}

\item{group_id}{Component group id name. All components (e.g. legends, titles, etc) with the same \code{group_id} will be grouped. The specifications of how they are placed (e.g. stacking, margins etc.) are determined in \code{\link[=tm_components]{tm_components()}} where its argument \code{id} should correspond to \code{group_id}.}

\item{group}{Name of the group to which this layer belongs. This is only
relevant in view mode, where layer groups can be switched (see \code{group.control})}

\item{group.control}{In view mode, the group control determines how
layer groups can be switched on and off. Options: \code{"radio"} for radio
buttons (meaning only one group can be shown), \code{"check"} for check boxes
(so multiple groups can be shown), and \code{"none"} for no control
(the group cannot be (de)selected).}

\item{z}{z index, e.g. the place of the component relative to the other componets}
}
\description{
Map component that adds a manual legend.
}
\examples{
\dontrun{
tm_shape(NLD_muni) +
	tm_borders() +
	tm_basemap("OpenStreetMap") +
	tm_add_legend(labels = c("Motorway", "Primary road", "Secondary road", "Railway"),
				  col = c("#E892A1", "#FCD6A4", "#F8FABF", "#707070"),
				  lty = c("solid", "solid", "solid", "dotted"),
				  lwd = 3,
				  type = "lines",
				  bg.color = "grey92",
				  bg.alpha = 1)
}
}
\seealso{
\href{https://r-tmap.github.io/tmap/articles/adv_legends}{Legends (advanced)}, \href{https://r-tmap.github.io/tmap/articles/adv_positions}{Positioning of components}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tm_layers_text.R
\name{tm_text}
\alias{tm_text}
\alias{tm_labels}
\alias{tm_labels_highlighted}
\alias{opt_tm_text}
\alias{opt_tm_labels}
\title{Map layer: text}
\usage{
tm_text(
  text = tm_const(),
  text.scale = tm_scale(),
  text.legend = tm_legend(),
  text.chart = tm_chart_none(),
  text.free = NA,
  size = tm_const(),
  size.scale = tm_scale(),
  size.legend = tm_legend(),
  size.chart = tm_chart_none(),
  size.free = NA,
  col = tm_const(),
  col.scale = tm_scale(),
  col.legend = tm_legend(),
  col.chart = tm_chart_none(),
  col.free = NA,
  col_alpha = tm_const(),
  col_alpha.scale = tm_scale(),
  col_alpha.legend = tm_legend(),
  col_alpha.chart = tm_chart_none(),
  col_alpha.free = NA,
  fontface = tm_const(),
  fontface.scale = tm_scale(),
  fontface.legend = tm_legend(),
  fontface.chart = tm_chart_none(),
  fontface.free = NA,
  fontfamily = NA,
  bgcol = tm_const(),
  bgcol.scale = tm_scale(),
  bgcol.legend = tm_legend(),
  bgcol.chart = tm_chart_none(),
  bgcol.free = NA,
  bgcol_alpha = tm_const(),
  bgcol_alpha.scale = tm_scale(),
  bgcol_alpha.legend = tm_legend(),
  bgcol_alpha.chart = tm_chart_none(),
  bgcol_alpha.free = NA,
  xmod = 0,
  xmod.scale = tm_scale(),
  xmod.legend = tm_legend_hide(),
  xmod.chart = tm_chart_none(),
  xmod.free = NA,
  ymod = 0,
  ymod.scale = tm_scale(),
  ymod.legend = tm_legend_hide(),
  ymod.chart = tm_chart_none(),
  ymod.free = NA,
  angle = 0,
  angle.scale = tm_scale(),
  angle.legend = tm_legend_hide(),
  angle.chart = tm_chart_none(),
  angle.free = NA,
  plot.order = tm_plot_order("size", reverse = FALSE),
  zindex = NA,
  group = NA,
  group.control = "check",
  options = opt_tm_text(),
  ...
)

tm_labels(
  text = tm_const(),
  text.scale = tm_scale(),
  text.legend = tm_legend(),
  text.chart = tm_chart_none(),
  text.free = NA,
  size = tm_const(),
  size.scale = tm_scale(),
  size.legend = tm_legend(),
  size.chart = tm_chart_none(),
  size.free = NA,
  col = tm_const(),
  col.scale = tm_scale(),
  col.legend = tm_legend(),
  col.chart = tm_chart_none(),
  col.free = NA,
  col_alpha = tm_const(),
  col_alpha.scale = tm_scale(),
  col_alpha.legend = tm_legend(),
  col_alpha.chart = tm_chart_none(),
  col_alpha.free = NA,
  fontface = tm_const(),
  fontface.scale = tm_scale(),
  fontface.legend = tm_legend(),
  fontface.chart = tm_chart_none(),
  fontface.free = NA,
  fontfamily = "",
  bgcol = tm_const(),
  bgcol.scale = tm_scale(),
  bgcol.legend = tm_legend(),
  bgcol.chart = tm_chart_none(),
  bgcol.free = NA,
  bgcol_alpha = tm_const(),
  bgcol_alpha.scale = tm_scale(),
  bgcol_alpha.legend = tm_legend(),
  bgcol_alpha.chart = tm_chart_none(),
  bgcol_alpha.free = NA,
  xmod = 0,
  xmod.scale = tm_scale(),
  xmod.legend = tm_legend_hide(),
  xmod.chart = tm_chart_none(),
  xmod.free = NA,
  ymod = 0,
  ymod.scale = tm_scale(),
  ymod.legend = tm_legend_hide(),
  ymod.chart = tm_chart_none(),
  ymod.free = NA,
  angle = 0,
  angle.scale = tm_scale(),
  angle.legend = tm_legend_hide(),
  angle.chart = tm_chart_none(),
  angle.free = NA,
  plot.order = tm_plot_order("AREA", reverse = FALSE, na.order = "bottom"),
  zindex = NA,
  group = NA,
  group.control = "check",
  options = opt_tm_labels(),
  ...
)

tm_labels_highlighted(
  text = tm_const(),
  text.scale = tm_scale(),
  text.legend = tm_legend(),
  text.chart = tm_chart_none(),
  text.free = NA,
  size = tm_const(),
  size.scale = tm_scale(),
  size.legend = tm_legend(),
  size.chart = tm_chart_none(),
  size.free = NA,
  col = tm_const(),
  col.scale = tm_scale(),
  col.legend = tm_legend(),
  col.chart = tm_chart_none(),
  col.free = NA,
  col_alpha = tm_const(),
  col_alpha.scale = tm_scale(),
  col_alpha.legend = tm_legend(),
  col_alpha.chart = tm_chart_none(),
  col_alpha.free = NA,
  fontface = tm_const(),
  fontface.scale = tm_scale(),
  fontface.legend = tm_legend(),
  fontface.chart = tm_chart_none(),
  fontface.free = NA,
  fontfamily = "",
  bgcol = tm_const(),
  bgcol.scale = tm_scale(),
  bgcol.legend = tm_legend(),
  bgcol.chart = tm_chart_none(),
  bgcol.free = NA,
  bgcol_alpha = tm_const(),
  bgcol_alpha.scale = tm_scale(),
  bgcol_alpha.legend = tm_legend(),
  bgcol_alpha.chart = tm_chart_none(),
  bgcol_alpha.free = NA,
  xmod = 0,
  xmod.scale = tm_scale(),
  xmod.legend = tm_legend_hide(),
  xmod.chart = tm_chart_none(),
  xmod.free = NA,
  ymod = 0,
  ymod.scale = tm_scale(),
  ymod.legend = tm_legend_hide(),
  ymod.chart = tm_chart_none(),
  ymod.free = NA,
  angle = 0,
  angle.scale = tm_scale(),
  angle.legend = tm_legend_hide(),
  angle.chart = tm_chart_none(),
  angle.free = NA,
  plot.order = tm_plot_order("AREA", reverse = FALSE, na.order = "bottom"),
  zindex = NA,
  group = NA,
  group.control = "check",
  options = opt_tm_labels(),
  ...
)

opt_tm_text(
  points_only = "ifany",
  point_per = "feature",
  on_surface = FALSE,
  shadow = FALSE,
  shadow.col = NA,
  shadow.offset.x = 0.05,
  shadow.offset.y = 0.05,
  halo = FALSE,
  halo.col = NA,
  halo.width = 0.05,
  halo.blur = 0,
  halo.alpha = 0.8,
  just = "center",
  along_lines = FALSE,
  bg.padding = 0.4,
  bg.border = FALSE,
  bg.border.col = "black",
  bg.border.lwd = 1,
  clustering = FALSE,
  point.label = FALSE,
  point.label.gap = 0,
  point.label.method = "SANN",
  remove_overlap = FALSE
)

opt_tm_labels(
  points_only = "ifany",
  point_per = "feature",
  on_surface = FALSE,
  shadow = FALSE,
  shadow.col = NA,
  shadow.offset.x = 0.05,
  shadow.offset.y = 0.05,
  halo = FALSE,
  halo.col = NA,
  halo.width = 0.05,
  halo.blur = 0,
  halo.alpha = 0.8,
  just = "center",
  along_lines = TRUE,
  bg.padding = 0.4,
  bg.border = FALSE,
  bg.border.col = "black",
  bg.border.lwd = 1,
  clustering = FALSE,
  point.label = NA,
  point.label.gap = 0.4,
  point.label.method = "SANN",
  remove_overlap = FALSE
)
}
\arguments{
\item{text, text.scale, text.legend, text.chart, text.free}{Visual variable that determines the text. See details.}

\item{size, size.scale, size.legend, size.chart, size.free}{Visual variable that determines the size. See details.}

\item{col, col.scale, col.legend, col.chart, col.free}{Visual variable that determines the color. See details.}

\item{col_alpha, col_alpha.scale, col_alpha.legend, col_alpha.chart, col_alpha.free}{Visual variable that determines the color transparency. See details.}

\item{fontface, fontface.scale, fontface.legend, fontface.chart, fontface.free}{Visual variable that determines the font face. See details.}

\item{fontfamily}{The font family. See \link[grid:gpar]{gpar()} for details.}

\item{bgcol, bgcol.scale, bgcol.legend, bgcol.chart, bgcol.free}{Visual variable that determines
the background color. See Details.}

\item{bgcol_alpha, bgcol_alpha.scale, bgcol_alpha.legend, bgcol_alpha.chart, bgcol_alpha.free}{Visual variable that determines
the background color transparency. See Details.}

\item{xmod, xmod.scale, xmod.legend, xmod.chart, xmod.free}{Transformation variable that determines the x offset. See details.}

\item{ymod, ymod.scale, ymod.legend, ymod.chart, ymod.free}{Transformation variable that determines the y offset. See details.
the text. See details.}

\item{angle, angle.scale, angle.legend, angle.chart, angle.free}{Rotation angle}

\item{plot.order}{Specification in which order the spatial features are drawn.
See \code{\link[=tm_plot_order]{tm_plot_order()}} for details.}

\item{zindex}{Map layers are drawn on top of each other. The \code{zindex} numbers
(one for each map layer) determines the stacking order.
By default the map layers are drawn in the order they are called.}

\item{group}{Name of the group to which this layer belongs. This is only
relevant in view mode, where layer groups can be switched (see \code{group.control})}

\item{group.control}{In view mode, the group control determines how layer groups
can be switched on and off. Options: \code{"radio"} for radio buttons
(meaning only one group can be shown), \code{"check"} for check boxes
(so multiple groups can be shown), and \code{"none"} for no control
(the group cannot be (de)selected).}

\item{options}{options passed on to the corresponding \verb{opt_<layer_function>} function}

\item{...}{to catch deprecated arguments from version < 4.0}

\item{points_only}{should only point geometries of the shape object (defined in \code{\link[=tm_shape]{tm_shape()}}) be plotted? By default \code{"ifany"}, which means \code{TRUE} in case a geometry collection is specified.}

\item{point_per}{specification of how spatial points are mapped when the geometry is a multi line or a multi polygon. One of \code{"feature"}, \code{"segment"} or \code{"largest"}. The first generates a spatial point for every feature, the second for every segment (i.e. subfeature), the third only for the largest segment (subfeature). Note that the last two options can be significant slower.}

\item{on_surface}{In case of polygons, centroids are computed. Should the points be on the surface? If \code{TRUE}, which is slower than the default \code{FALSE}, centroids outside the surface are replaced with points computed with \code{\link[sf:geos_unary]{sf::st_point_on_surface()}}.}

\item{shadow}{Shadow behind the text. Logical.}

\item{shadow.col}{Color of the shadow.}

\item{shadow.offset.x, shadow.offset.y}{Shadow offset in line heights}

\item{halo}{Halo behind the text. In plot mode, it is just an outline, in view mode also a subtle glow.}

\item{halo.col}{Color of the halo.}

\item{halo.width}{Width (thickness) of the halo outline. In line heights}

\item{halo.blur}{Blur radius of the halo glow (view mode only). In line heights. Should be sufficiently larger than \code{halo.width} in order to see the effect.}

\item{halo.alpha}{Alpha transparency of the halo}

\item{just}{justification of the text relative to the point coordinates. Either one of the following values: \code{"left"} , \code{"right"}, \code{"center"}, \code{"bottom"}, and \code{"top"}, or a vector of two values where first value specifies horizontal and the second value vertical justification. Besides the mentioned values, also numeric values between 0 and 1 can be used. 0 means left justification for the first value and bottom justification for the second value. Note that in view mode, only one value is used.}

\item{along_lines}{logical that determines whether labels are rotated along the spatial lines. Only applicable if a spatial lines shape is used.}

\item{bg.padding}{The padding of the background in terms of line heights.}

\item{bg.border}{Should the background have borders?}

\item{bg.border.col}{Color of the borders}

\item{bg.border.lwd}{Line width of the borders}

\item{clustering}{in interactive modes (e.g. \code{"view"} mode), should clustering be applied at lower zoom levels? Either \code{FALSE} (default), \code{TRUE}, or a mode specific specification, e.g. for \code{"view"} mode \code{\link[leaflet:markerClusterOptions]{markerClusterOptions}}.}

\item{point.label}{logical that determines whether the labels are placed automatically. By default \code{FALSE} for \code{tm_text}, and \code{TRUE} for \code{tm_labels} if the number of labels is less than 500 (otherwise it will be too slow).}

\item{point.label.gap}{numeric that determines the gap between the point and label}

\item{point.label.method}{the optimization method, either \code{"SANN"} for simulated annealing (the default) or \code{"GA"} for a genetic algorithm.}

\item{remove_overlap}{logical that determines whether the overlapping labels are removed}
}
\description{
Map layer that draws symbols Supported visual variables are: \code{text}
(the text itself) \code{col} (color), \code{size} (font size), and \code{fontface} (font face).
}
\details{
The visual variable arguments (e.g. \code{col}) can be specified with a data
variable name (e.g., a spatial vector attribute or a raster layer of the object
specified in \code{\link[=tm_shape]{tm_shape()}}), with a visual value (for \code{col}, a color is expected), or with a geometry-derived variable (see below).
See \href{https://r-tmap.github.io/tmap/articles/basics_vv}{vignette about visual variables}.

Multiple values can be specified: in that case facets are created.
These facets can be combined with other faceting data variables, specified with \code{\link[=tm_facets]{tm_facets()}}.
See \href{https://r-tmap.github.io/tmap/articles/basics_facets}{vignette about facets}.
\itemize{
\item The \verb{*.scale} arguments determine the used scale to map the data values to
visual variable values. These can be specified with one of the available
\verb{tm_scale_*()} functions. The default is specified by the tmap option (\code{\link[=tm_options]{tm_options()}}) \code{scales.var}.
See \href{https://r-tmap.github.io/tmap/articles/basics_scales}{vignette about scales}.
\item The \verb{*.legend} arguments determine the used legend, specified with \code{\link[=tm_legend]{tm_legend()}}.
The default legend and its settings are determined by the tmap options (\code{\link[=tm_options]{tm_options()}}) \code{legend.} .
See \href{https://r-tmap.github.io/tmap/articles/basics_legends}{vignette about legends}.
\item The \verb{*.chart} arguments specify additional charts, specified with \code{tm_chart_}, e.g. \code{\link[=tm_chart_histogram]{tm_chart_histogram()}}.
See \href{https://r-tmap.github.io/tmap/articles/basics_charts}{vignette about charts}.
\item The \verb{*.free} arguments determine whether scales are applied freely across facets, or shared.
A logical value is required. They can also be specified with a vector of three
logical values; these determine whether scales are applied freely per facet dimension.
This is only useful when facets are applied (see \code{\link[=tm_facets]{tm_facets()}}).
There are maximally three facet dimensions: rows, columns, and pages. This only
applies for a facet grid (\code{\link[=tm_facets_grid]{tm_facets_grid()}}). For instance, \code{col.free = c(TRUE, FALSE, FALSE)}
means that for the visual variable \code{col}, each row of facets will have its own
scale, and therefore its own legend. For facet wraps and stacks
(\code{\link[=tm_facets_wrap]{tm_facets_wrap()}} and \code{\link[=tm_facets_stack]{tm_facets_stack()}}) there is only one facet dimension,
so the \verb{*.free} argument requires only one logical value.
}

Currently, three geometry-derived variables are implemented:
\itemize{
\item \code{"AREA"} (polygons only), which uses the feature area;
\item \code{"LENGTH"} (lines only), which uses the feature length; and
\item \code{"MAP_COLORS"}, which assigns values so that adjacent features receive
different values, making it particularly suitable for coloring
neighbouring polygons.
}

Note that geometry-derived variables do not generate a legend automatically.
If a legend is required, compute the corresponding variable explicitly,
for example with \code{\link[sf:geos_measures]{sf::st_area()}}, \code{\link[sf:geos_measures]{sf::st_length()}}, or
\code{\link[tmaptools:map_coloring]{tmaptools::map_coloring()}}, and use the resulting values instead.
}
\examples{
tm_shape(World, bbox = World) +
	tm_text("name", size="pop_est", col="continent",
			col.scale = tm_scale_categorical(values = "seaborn.dark"),
			col.legend = tm_legend_hide(),
			size.scale = tm_scale_continuous(values.scale = 4),
			size.legend = tm_legend_hide())

metro$upside_down = ifelse(sf::st_coordinates(metro)[,2] < 0, 180, 0)
tm_shape(metro) +
	tm_text(text = "name", size = "pop2020",
			angle = "upside_down", size.legend = tm_legend_hide(),
			col = "upside_down",
			col.scale = tm_scale_categorical(values = c("#9900BB", "#228822")),
			col.legend = tm_legend_hide()) +
	tm_title_out("Which Hemisphere?", position = tm_pos_out("center", "top", pos.v = "bottom"))

}
\seealso{
\href{https://r-tmap.github.io/tmap/articles/examples_terrain}{Terrain map example}

\href{https://r-tmap.github.io/tmap/articles/examples_topo_Africa}{Topographic map}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tm_layers_polygons.R
\name{tm_polygons}
\alias{tm_polygons}
\alias{tm_fill}
\alias{tm_borders}
\alias{opt_tm_polygons}
\title{Map layer: polygons}
\usage{
tm_polygons(
  fill = tm_const(),
  fill.scale = tm_scale(),
  fill.legend = tm_legend(),
  fill.chart = tm_chart_none(),
  fill.free = NA,
  col = tm_const(),
  col.scale = tm_scale(),
  col.legend = tm_legend(),
  col.chart = tm_chart_none(),
  col.free = NA,
  lwd = tm_const(),
  lwd.scale = tm_scale(),
  lwd.legend = tm_legend(),
  lwd.chart = tm_chart_none(),
  lwd.free = NA,
  lty = tm_const(),
  lty.scale = tm_scale(),
  lty.legend = tm_legend(),
  lty.chart = tm_chart_none(),
  lty.free = NA,
  fill_alpha = tm_const(),
  fill_alpha.scale = tm_scale(),
  fill_alpha.legend = tm_legend(),
  fill_alpha.chart = tm_chart_none(),
  fill_alpha.free = NA,
  col_alpha = tm_const(),
  col_alpha.scale = tm_scale(),
  col_alpha.legend = tm_legend(),
  col_alpha.chart = tm_chart_none(),
  col_alpha.free = NA,
  linejoin = "round",
  lineend = "round",
  plot.order = tm_plot_order("lwd", reverse = TRUE, na.order = "bottom"),
  zindex = NA,
  group = NA,
  group.control = "check",
  popup = tm_popup(),
  popup.vars = NA,
  popup.format = tm_label_format(),
  hover = NA,
  id = "",
  blend = "over",
  options = opt_tm_polygons(),
  ...
)

tm_fill(...)

tm_borders(col = tm_const(), ...)

opt_tm_polygons(polygons.only = "ifany")
}
\arguments{
\item{fill, fill.scale, fill.legend, fill.chart, fill.free}{Visual variable that determines the fill color. See details. \emph{Unit:} Color -- a color name, hex string.}

\item{col, col.scale, col.legend, col.chart, col.free}{Visual variable that determines the color. See details. \emph{Unit:} Color -- a color name, hex string.}

\item{lwd, lwd.scale, lwd.legend, lwd.chart, lwd.free}{Visual variable that determines the line width. See details. \emph{Unit:} Base R line-width units; 1 lwd is approx. 0.75 pt at 96 dpi.}

\item{lty, lty.scale, lty.legend, lty.chart, lty.free}{Visual variable that determines the line type. See details. \emph{Unit:} Integer (1-6) or name: "solid", "dashed", "dotted", "dotdash", "longdash", "twodash".}

\item{fill_alpha, fill_alpha.scale, fill_alpha.chart, fill_alpha.legend, fill_alpha.free}{Visual variable that determines the fill color transparency. See details. \emph{Unit:} Proportion -- numeric 0-1 (0 = fully transparent, 1 = fully opaque).}

\item{col_alpha, col_alpha.scale, col_alpha.legend, col_alpha.chart, col_alpha.free}{Visual variable that determines the color transparency. See details. \emph{Unit:} Proportion -- numeric 0-1 (0 = fully transparent, 1 = fully opaque).}

\item{linejoin, lineend}{Line join and line end. See \link[grid:gpar]{gpar()} for details.}

\item{plot.order}{Specification in which order the spatial features are drawn.
See \code{\link[=tm_plot_order]{tm_plot_order()}} for details.}

\item{zindex}{Controls the stacking order of map layers. Should be set to a
value above 400. By default, layers are stacked in call order, starting
at 401. See details.}

\item{group}{Name of the group to which this layer belongs. This is only
relevant in view mode, where layer groups can be switched (see \code{group.control})}

\item{group.control}{In view mode, the group control determines how layer groups
can be switched on and off. Options: \code{"radio"} for radio buttons
(meaning only one group can be shown), \code{"check"} for check boxes
(so multiple groups can be shown), and \code{"none"} for no control
(the group cannot be (de)selected).}

\item{popup}{popup specification for \code{"view"} mode, the output of \code{\link[=tm_popup]{tm_popup()}}.
It determines the data variables shown in the popup table, the popup title,
and (in the future) the popup layout. This replaces the deprecated
arguments \code{popup.vars} and \code{popup.format}.}

\item{popup.vars}{(Deprecated.) Use \code{popup} with
\code{\link[=tm_popup]{tm_popup()}} instead (via its \code{vars} argument). Names of data variables that
are shown in the popups in \code{"view"} mode. Set \code{popup.vars} to \code{TRUE} to show
all variables in the shape object. Set \code{popup.vars} to \code{FALSE} to disable
popups. Set \code{popup.vars} to a character vector of variable names to show
those variables in the popups. The default (\code{NA}) depends on whether visual
variables (e.g. \code{fill}) are used. If so, only those are shown. If not, all
variables in the shape object are shown.}

\item{popup.format}{(Deprecated.) Use \code{popup} with
\code{\link[=tm_popup]{tm_popup()}} instead (via its \code{format} argument). List of formatting options
for the popup values. Output of \code{\link[=tm_label_format]{tm_label_format()}}. Only applicable for
numeric data variables. If one list of formatting options is provided,
it is applied to all numeric variables of \code{popup.vars}. Also, a (named)
list of lists can be provided. In that case, each list of formatting options
is applied to the named variable.}

\item{hover}{name of the data variable that specifies the hover labels (view mode only). Set to \code{FALSE} to disable hover labels. By default \code{FALSE}, unless \code{id} is specified. In that case, it is set to \code{id},}

\item{id}{name of the data variable that specifies the indices of the spatial
features. Only used for \code{"view"} mode.}

\item{blend}{Compositing operator for layer blending. Default \code{"over"} applies
no blending. See the "Layer blending" section for the supported values.}

\item{options}{options passed on to the corresponding \verb{opt_<layer_function>} function}

\item{...}{to catch deprecated arguments from version < 4.0}

\item{polygons.only}{should only polygon geometries of the shape object (defined in \code{\link[=tm_shape]{tm_shape()}}) be plotted? By default \code{"ifany"}, which means \code{TRUE} in case a geometry collection is specified.}
}
\description{
Map layer that draws polygons. Supported visual variables are: \code{fill} (the fill color),
\code{col} (the border color), \code{lwd} (line width), \code{lty} (line type),
\code{fill_alpha} (fill color alpha transparency) and \code{col_alpha} (border color alpha transparency).
}
\details{
The visual variable arguments (e.g. \code{col}) can be specified with a data
variable name (e.g., a spatial vector attribute or a raster layer of the object
specified in \code{\link[=tm_shape]{tm_shape()}}), with a visual value (for \code{col}, a color is expected), or with a geometry-derived variable (see below).
See vignette: \href{https://r-tmap.github.io/tmap/articles/basics_vv}{Visual variables}.

Multiple values can be specified: in that case facets are created.
These facets can be combined with other faceting data variables, specified with \code{\link[=tm_facets]{tm_facets()}}.
See vignette: \href{https://r-tmap.github.io/tmap/articles/basics_facets}{Facets}.
\itemize{
\item The \verb{*.scale} arguments determine the used scale to map the data values to
visual variable values. These can be specified with one of the available
\verb{tm_scale_*()} functions. The default is specified by the tmap option (\code{\link[=tm_options]{tm_options()}}) \code{scales.var}.
See `rvignette: \href{https://r-tmap.github.io/tmap/articles/basics_scales}{Scales}
\item The \verb{*.legend} arguments determine the used legend, specified with \code{\link[=tm_legend]{tm_legend()}}.
The default legend and its settings are determined by the tmap options (\code{\link[=tm_options]{tm_options()}}) \code{legend.} .
See `rvignette: \href{https://r-tmap.github.io/tmap/articles/basics_legends}{Legends}
\item The \verb{*.chart} arguments specify additional charts, specified with \code{tm_chart_}, e.g. \code{\link[=tm_chart_histogram]{tm_chart_histogram()}}.
See `rvignette: \href{https://r-tmap.github.io/tmap/articles/basics_charts}{Charts}
\item The \verb{*.free} arguments determine whether scales are applied freely across facets, or shared.
A logical value is required. They can also be specified with a vector of three
logical values; these determine whether scales are applied freely per facet dimension.
This is only useful when facets are applied (see \code{\link[=tm_facets]{tm_facets()}}).
There are maximally three facet dimensions: rows, columns, and pages. This only
applies for a facet grid (\code{\link[=tm_facets_grid]{tm_facets_grid()}}). For instance, \code{col.free = c(TRUE, FALSE, FALSE)}
means that for the visual variable \code{col}, each row of facets will have its own
scale, and therefore its own legend. For facet wraps and stacks
(\code{\link[=tm_facets_wrap]{tm_facets_wrap()}} and \code{\link[=tm_facets_stack]{tm_facets_stack()}}) there is only one facet dimension,
so the \verb{*.free} argument requires only one logical value.
}

Currently, three geometry-derived variables are implemented:
\itemize{
\item \code{"AREA"} (polygons only), which uses the feature area;
\item \code{"LENGTH"} (lines only), which uses the feature length; and
\item \code{"MAP_COLORS"}, which assigns values so that adjacent features receive
different values, making it particularly suitable for coloring
neighbouring polygons.
}

Note that geometry-derived variables do not generate a legend automatically.
If a legend is required, compute the corresponding variable explicitly,
for example with \code{\link[sf:st_area]{sf::st_area()}}, \code{\link[sf:st_length]{sf::st_length()}}, or
\code{\link[tmaptools:map_coloring]{tmaptools::map_coloring()}}, and use the resulting values instead.
\subsection{Visual variable units}{

Every visual variable maps data values to a specific output unit.
Knowing the unit matters when supplying constant values via \code{\link[=tm_const]{tm_const()}},
or output ranges via \code{values.range} / \code{values.scale} in the scale
functions.\tabular{lll}{
   Variable \tab Output unit \tab Notes \cr
   \code{fill}, \code{col}, \code{bgcol} \tab color \tab name, hex, or palette string \cr
   \code{fill_alpha}, \code{col_alpha}, \code{bgcol_alpha} \tab proportion 0-1 \tab 0 = transparent, 1 = opaque \cr
   \code{size} (symbols, bubbles, squares, dots) \tab typographic lines \tab 1 line approx. 1/6 inch; scaled by \code{values.scale} \cr
   \code{size} (circles) \tab meters \tab plain numeric or a \code{units} object \cr
   \code{size} (text, labels) \tab multiplier \tab 1 = 12 pt (plot) / 12 px (view) \cr
   \code{lwd} \tab lwd \tab base R units; 1 lwd approx. 0.75 pt at 96 dpi \cr
   \code{lty} \tab -- \tab integer 1-6 or name ("solid", "dashed", ...) \cr
   \code{shape} \tab -- \tab integer \code{pch} 1-25 or single character \cr
   \code{angle} \tab degrees \tab 0-360, clockwise from north \cr
   \code{fontface} \tab -- \tab "plain", "bold", "italic", "bold.italic" \cr
}

\subsection{Symbol size (\code{size} in \code{tm_symbols}, \code{tm_bubbles}, \code{tm_squares}, \code{tm_dots})}{

"Lines" is a typographic unit: one line is approximately 1/6 inch (the
default base line-height in R graphics). The global multiplier
\code{tmap_options(values.scale = list(size.bubbles = 1.5))} scales all symbol
sizes without changing the data mapping.
}

\subsection{Circle size (\code{size} in \code{tm_circles})}{

The value is a geographic radius in meters. A plain numeric vector is
interpreted as meters; a \code{units} object (from the \strong{units} package) is
automatically converted, so \code{units::as_units(1, "mi")} gives a 1-mile
radius. Because the radius is geographic, circles scale with zoom in
interactive (view) mode -- unlike bubble symbols which keep a fixed screen
size.
}

\subsection{Text size (\code{size} in \code{tm_text}, \code{tm_labels})}{

The value is a multiplier of the base font size. \code{size = 1} renders at
12 pt in plot mode (R's default \code{par("ps")}) and at 12 px in view
mode (\code{gp$cex * 12} px, see \code{tmapLeafletDataPlot.tm_data_text}); the two
modes are consistent by design.
}

}

\subsection{Layer blending (\code{blend})}{

Blend modes control how a layer's pixels are combined with the pixels
beneath it. For each pixel, let \eqn{S} be the source (top layer) RGB
value and \eqn{D} be the destination (bottom layer) RGB value, both
normalised to \eqn{[0, 1]}.\tabular{lll}{
   \code{blend} \tab Formula \tab Use case \cr
   \code{"over"} \tab \eqn{S \cdot \alpha + D \cdot (1 - \alpha)} \tab Standard alpha compositing (default) \cr
   \code{"multiply"} \tab \eqn{S \times D} \tab Hillshading over colour raster; both layers darken each other \cr
   \code{"screen"} \tab \eqn{1 - (1 - S)(1 - D)} \tab Inverse of multiply; brightens \cr
   \code{"overlay"} \tab multiply if \eqn{D < 0.5}, screen if \eqn{D \geq 0.5} \tab Boosts contrast; preserves midtones \cr
   \code{"darken"} \tab \eqn{\min(S, D)} \tab Keeps the darker of the two layers per channel \cr
   \code{"lighten"} \tab \eqn{\max(S, D)} \tab Keeps the lighter of the two layers per channel \cr
}


Requires R >= 4.2 and a compatible graphics device (e.g.
\code{png(type = "cairo")}, \code{svg()}). In view mode, blending is applied via
CSS \code{mix-blend-mode}. See \code{\link[grid:groupGrob]{grid::groupGrob()}} for the full list of
supported operators.
}

\subsection{zindex and pane names}{

In view mode, each layer is rendered in a Leaflet pane named \code{"tmap{zindex}"}
(e.g., \code{"tmap401"}, \code{"tmap402"}), with base tile layers placed in the
standard \code{"tile"} pane.
}
}
\examples{
# load Africa country data
data(World)
Africa = World[World$continent == "Africa", ]
Africa_border = sf::st_make_valid(sf::st_union(sf::st_buffer(Africa, 0.001))) # slow and ugly

# without specifications
tm_shape(Africa_border) + tm_polygons()
tm_shape(Africa_border) + tm_fill()
tm_shape(Africa_border) + tm_borders()

# specification with visual variable values
tm_shape(Africa) +
  tm_polygons(fill = "limegreen", col = "purple", lwd = 2, lty = "solid", col_alpha = 0.3) +
  tm_text("name", options = opt_tm_text(remove_overlap = TRUE)) +
tm_shape(Africa_border) +
  tm_borders("darkred", lwd = 3)

# specification with a data variable
tm_shape(Africa) +
  tm_polygons(fill = "income_grp", fill.scale = tm_scale_categorical(values = "-tol.muted"))


# continuous color scale with landscape legend
tm_shape(Africa) +
  tm_polygons(fill = "inequality",
    fill.scale = tm_scale_continuous(values = "-scico.roma"),
    fill.legend = tm_legend(
    	title = "", orientation = "landscape",
    	position = tm_pos_out("center", "bottom"), frame = FALSE
    	)) +
tm_shape(Africa_border) +
	tm_borders(lwd = 2) +
tm_title("Inequality index", position = tm_pos_in("right", "TOP"), frame = FALSE) +
tm_layout(frame = FALSE)

# bivariate scale
tm_shape(World) +
	tm_polygons(tm_vars(c("inequality", "well_being"), multivariate = TRUE))
}
\seealso{
\href{https://r-tmap.github.io/tmap/articles/basics_vv}{Visual variables}, \href{https://r-tmap.github.io/tmap/articles/basics_scales}{Scales}, \href{https://r-tmap.github.io/tmap/articles/basics_legends}{Legends}, \href{https://r-tmap.github.io/tmap/articles/basics_facets}{Facets}, \href{https://r-tmap.github.io/tmap/articles/foundations_units}{Units}, \href{https://r-tmap.github.io/tmap/articles/adv_blend}{Layer blending}, \href{https://r-tmap.github.io/tmap/articles/examples_choro_World}{Choropleth (World)}, \href{https://r-tmap.github.io/tmap/articles/examples_choro_NLD}{Choropleth (Netherlands)}, \href{https://r-tmap.github.io/tmap/articles/examples_biv_choro}{Bivariate choropleth}, \href{https://r-tmap.github.io/tmap/articles/examples_gridmaps}{Gridmap}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tm_scale_.R
\name{tm_scale}
\alias{tm_scale}
\title{Scales: automatic scale}
\usage{
tm_scale(...)
}
\arguments{
\item{...}{arguments passed on to the applied scale function \verb{tm_scale_*()}}
}
\description{
Scales in tmap are configured by the family of functions with prefix \code{tm_scale}.
Such function should be used for the input of the \code{.scale} arguments in the
layer functions (e.g. \code{fill.scale} in \code{\link[=tm_polygons]{tm_polygons()}}). The function \code{tm_scale()}
is a scale that is set automatically given by the data type (factor, numeric, and integer)
and the visual variable. The tmap option \code{scales.var} contains information
which scale is applied when.
}
\seealso{
\code{\link[=tm_scale_asis]{tm_scale_asis()}}, \code{\link[=tm_scale_ordinal]{tm_scale_ordinal()}}, \code{\link[=tm_scale_categorical]{tm_scale_categorical()}},
\code{\link[=tm_scale_intervals]{tm_scale_intervals()}}, \code{\link[=tm_scale_discrete]{tm_scale_discrete()}}, \code{\link[=tm_scale_continuous]{tm_scale_continuous()}},
\code{\link[=tm_scale_rank]{tm_scale_rank()}}, \code{\link[=tm_scale_continuous_log]{tm_scale_continuous_log()}}, \code{\link[=tm_scale_continuous_log2]{tm_scale_continuous_log2()}},
\code{\link[=tm_scale_continuous_log10]{tm_scale_continuous_log10()}}, \code{\link[=tm_scale_continuous_log1p]{tm_scale_continuous_log1p()}}, \code{\link[=tm_scale_continuous_sqrt]{tm_scale_continuous_sqrt()}},
\code{\link[=tm_scale_continuous_pseudo_log]{tm_scale_continuous_pseudo_log()}}, \code{\link[=tm_scale_rgb]{tm_scale_rgb()}}, \code{\link[=tm_scale_bivariate]{tm_scale_bivariate()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tm_animate.R
\name{tm_animate_fast}
\alias{tm_animate_fast}
\alias{tm_animate}
\title{Specify an animation}
\usage{
tm_animate_fast(
  frames = "VARS__",
  nframes = 60L,
  fps = 24L,
  play = c("loop", "pingpong", "once"),
  dpr = 2,
  ...
)

tm_animate(
  frames = "VARS__",
  nframes = 60L,
  fps = 2L,
  play = c("loop", "pingpong", "once"),
  dpr = 2,
  ...
)
}
\arguments{
\item{frames}{group by variable used to create the animation frames. This is similar to the \code{by} argument of \code{\link[=tm_facets_wrap]{tm_facets_wrap()}}. Instead of showing facets next to each other, they are shown as animation frames. However, under the hood \code{frames} will be used to specify \code{pages} of \code{\link[=tm_facets]{tm_facets()}}. This makes it possible to create an animation of regular facets.}

\item{nframes}{number of animation frames. So far, this only applied experimentally in transition map variables. See the extension package tmap.cartogram.}

\item{fps}{frames per second. Default: 30 for \code{tm_facets_animate} and 2 for \code{tm_facets_animate_slow}.}

\item{play}{how should the animation be played? One of \code{"loop"} (default), \code{"pingpong"}, and \code{"once"}, where \code{"loop"} means that the animation will loop indefinitely, \code{"pingpong"} means that it will play forward and then backward, and \code{"once"} means that it will play only once.}

\item{dpr}{device pixel ratio. The ratio between the physical pixel density of a device and its logical pixel density.}

\item{...}{passed on to \code{\link[=tm_facets]{tm_facets()}}. Note that for animated facets, \code{by} can be specified to create animated facet wraps, and \code{rows} and \code{cols} to created animated facet grids.}
}
\description{
Specify an animation from a tmap plot. This is similar to creating facets with \code{\link[=tm_facets]{tm_facets()}}. The animation subsequently can be exported to a gif or video file (e.g. mp4) with \code{\link[=tmap_animation]{tmap_animation()}}. If the tmap plot with \code{\link[=tm_animate]{tm_animate()}} is printed, the animation will be previewed. The default \code{\link[=tm_animate]{tm_animate()}} will show the individual frames slowly (frame per seconds (fps) set to 2) whereas \code{\link[=tm_animate_fast]{tm_animate_fast()}} will show them like a movie (with a fps set to 24).
}
\note{
In older versions (< 4.1) \code{\link[=tm_facets]{tm_facets()}} with \code{page} specification was used to create animations frames and \code{\link[=tmap_animation]{tmap_animation()}} to create the animation itself using inputs like the frame rate specification. As of version 4.2, the whole animation, including frame rate, is specified in \code{\link[=tm_animate]{tm_animate()}}. The animation can still be saved via \code{\link[=tmap_animation]{tmap_animation()}}.
}
\seealso{
\code{\link[=tm_facets]{tm_facets()}} which is the core function, and \code{\link[=tmap_animation]{tmap_animation()}} used to save the animation
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tm_scale_.R
\name{tm_scale}
\alias{tm_scale}
\title{Scales: automatic scale}
\usage{
tm_scale(...)
}
\arguments{
\item{...}{arguments passed on to the applied scale function \verb{tm_scale_*()}}
}
\description{
Scales in tmap are configured by the family of functions with prefix \code{tm_scale}.
Such function should be used for the input of the \code{.scale} arguments in the
layer functions (e.g. \code{fill.scale} in \code{\link[=tm_polygons]{tm_polygons()}}). The function \code{tm_scale()}
is a scale that is set automatically given by the data type (factor, numeric, and integer)
and the visual variable. The tmap option \code{scales.var} contains information
which scale is applied when.
}
\seealso{
\href{https://r-tmap.github.io/tmap/articles/basics_scales}{Scales}, \href{https://r-tmap.github.io/tmap/articles/basics_vv}{Visual variables}, \href{https://r-tmap.github.io/tmap/articles/adv_multivariate}{Multivariate visual variables} \code{\link[=tm_scale_asis]{tm_scale_asis()}}, \code{\link[=tm_scale_ordinal]{tm_scale_ordinal()}}, \code{\link[=tm_scale_categorical]{tm_scale_categorical()}},
\code{\link[=tm_scale_intervals]{tm_scale_intervals()}}, \code{\link[=tm_scale_discrete]{tm_scale_discrete()}}, \code{\link[=tm_scale_continuous]{tm_scale_continuous()}},
\code{\link[=tm_scale_rank]{tm_scale_rank()}}, \code{\link[=tm_scale_continuous_log]{tm_scale_continuous_log()}}, \code{\link[=tm_scale_continuous_log2]{tm_scale_continuous_log2()}},
\code{\link[=tm_scale_continuous_log10]{tm_scale_continuous_log10()}}, \code{\link[=tm_scale_continuous_log1p]{tm_scale_continuous_log1p()}}, \code{\link[=tm_scale_continuous_sqrt]{tm_scale_continuous_sqrt()}},
\code{\link[=tm_scale_continuous_pseudo_log]{tm_scale_continuous_pseudo_log()}}, \code{\link[=tm_scale_rgb]{tm_scale_rgb()}}, \code{\link[=tm_scale_bivariate]{tm_scale_bivariate()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tm_animate.R
\name{tm_animate_fast}
\alias{tm_animate_fast}
\alias{tm_animate}
\title{Specify an animation}
\usage{
tm_animate_fast(
  frames = "VARS__",
  nframes = 60L,
  fps = 24L,
  play = c("loop", "pingpong", "once"),
  dpr = 2,
  ...
)

tm_animate(
  frames = "VARS__",
  nframes = 60L,
  fps = 2L,
  play = c("loop", "pingpong", "once"),
  dpr = 2,
  ...
)
}
\arguments{
\item{frames}{group by variable used to create the animation frames. This is similar to the \code{by} argument of \code{\link[=tm_facets_wrap]{tm_facets_wrap()}}. Instead of showing facets next to each other, they are shown as animation frames. However, under the hood \code{frames} will be used to specify \code{pages} of \code{\link[=tm_facets]{tm_facets()}}. This makes it possible to create an animation of regular facets.}

\item{nframes}{number of animation frames. So far, this only applied experimentally in transition map variables. See the extension package tmap.cartogram.}

\item{fps}{frames per second. Default: 30 for \code{tm_facets_animate} and 2 for \code{tm_facets_animate_slow}.}

\item{play}{how should the animation be played? One of \code{"loop"} (default), \code{"pingpong"}, and \code{"once"}, where \code{"loop"} means that the animation will loop indefinitely, \code{"pingpong"} means that it will play forward and then backward, and \code{"once"} means that it will play only once.}

\item{dpr}{device pixel ratio. The ratio between the physical pixel density of a device and its logical pixel density.}

\item{...}{passed on to \code{\link[=tm_facets]{tm_facets()}}. Note that for animated facets, \code{by} can be specified to create animated facet wraps, and \code{rows} and \code{cols} to created animated facet grids.}
}
\description{
Specify an animation from a tmap plot. This is similar to creating facets with \code{\link[=tm_facets]{tm_facets()}}. The animation subsequently can be exported to a gif or video file (e.g. mp4) with \code{\link[=tmap_animation]{tmap_animation()}}. If the tmap plot with \code{\link[=tm_animate]{tm_animate()}} is printed, the animation will be previewed. The default \code{\link[=tm_animate]{tm_animate()}} will show the individual frames slowly (frame per seconds (fps) set to 2) whereas \code{\link[=tm_animate_fast]{tm_animate_fast()}} will show them like a movie (with a fps set to 24).
}
\note{
In older versions (< 4.1) \code{\link[=tm_facets]{tm_facets()}} with \code{page} specification was used to create animations frames and \code{\link[=tmap_animation]{tmap_animation()}} to create the animation itself using inputs like the frame rate specification. As of version 4.2, the whole animation, including frame rate, is specified in \code{\link[=tm_animate]{tm_animate()}}. The animation can still be saved via \code{\link[=tmap_animation]{tmap_animation()}}.
}
\examples{
if (interactive()) {
	tm_shape(NLD_prov) +
		tm_polygons("yellow") +
		tm_animate(frames = "name")

	tm_shape(metro) +
		tm_symbols(size = paste0("pop", seq(1950, 2030, by=10)),
				   size.free = FALSE,
				   size.legend = tm_legend("Population")) +
		tm_layout(panel.labels = seq(1970, 2030, by=10)) +
		tm_animate()
}

}
\seealso{
\href{https://r-tmap.github.io/tmap/articles/adv_animations}{Animations}, \href{https://r-tmap.github.io/tmap/articles/ext_cartogram}{tmap.cartogram}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/print.R
\name{print.tmap}
\alias{print.tmap}
\alias{knit_print.tmap}
\title{Draw thematic map}
\usage{
\method{print}{tmap}(
  x,
  return.asp = FALSE,
  show = TRUE,
  vp = NULL,
  knit = FALSE,
  options = NULL,
  in.shiny = FALSE,
  proxy = FALSE,
  ...
)

\method{knit_print}{tmap}(x, ..., options = NULL)
}
\arguments{
\item{x}{tmap object.}

\item{return.asp}{should the aspect ratio be returned?}

\item{show}{show the map}

\item{vp}{viewport (for \code{"plot"} mode)}

\item{knit}{A logical, should knit?}

\item{options}{A vector of options}

\item{in.shiny}{A logical, is the map drawn in \strong{shiny}?}

\item{proxy}{A logical, if \code{in.shiny}, is \code{\link[=tmapProxy]{tmapProxy()}} used?}

\item{...}{passed on internally (for developers: in \code{"view"} mode, the proxy leaflet object is passed to \code{tmapLeafletInit}).}
}
\description{
Draw thematic map
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tmapScale_defaults.R
\name{tm_seq}
\alias{tm_seq}
\title{Specify a numeric sequence}
\usage{
tm_seq(
  from = 0,
  to = 1,
  power = c("lin", "sqrt", "sqrt_perceptual", "quadratic")
)
}
\arguments{
\item{from, to}{The numeric range, default 0 and 1 respectively}

\item{power}{The power component, a number or or one of \code{"lin"}, \code{"sqrt"}, \code{"sqrt_perceptual"}, \code{"quadratic"}, which correspond to 1, 0.5, 0.5716, 2 respectively. See details.}
}
\description{
Specify a numeric sequence, for numeric scales like \code{\link[=tm_scale_continuous]{tm_scale_continuous()}}. This function is needed when there is a non-linear relationship between the numeric data values and the visual variables. E.g. to make relationship with the area of bubbles linear, the square root of input variables should be used to calculate the radius of the bubbles.
}
\details{
The perceived area of larger symbols is often underestimated. Flannery (1971) experimentally derived a method to compensate this for symbols. This compensation is obtained by using the power exponent of 0.5716 instead of 0.5, or by setting \code{power} to \code{"sqrt_perceptual"}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tm_options.R
\name{tm_options}
\alias{tm_options}
\title{tmap options}
\usage{
tm_options(
  crs,
  facet.max,
  free.scales,
  raster.max_cells,
  raster.warp,
  show.messages,
  show.warnings,
  output.format,
  output.size,
  output.dpi,
  animation.dpi,
  value.const,
  value.na,
  value.null,
  value.blank,
  values.var,
  values.range,
  value.neutral,
  values.scale,
  scales.var,
  scale.misc.args,
  continuous.nclass_per_legend_break,
  continuous.nclasses,
  label.format,
  label.na,
  scale,
  asp,
  bg,
  bg.color,
  outer.bg,
  outer.bg.color,
  frame,
  frame.color,
  frame.alpha,
  frame.lwd,
  frame.r,
  frame.double_line,
  outer.margins,
  inner.margins,
  inner.margins.extra,
  meta.margins,
  meta.auto_margins,
  between_margin,
  panel.margin,
  xlab.show,
  xlab.text,
  xlab.size,
  xlab.color,
  xlab.rotation,
  xlab.space,
  xlab.fontface,
  xlab.fontfamily,
  xlab.alpha,
  xlab.side,
  ylab.show,
  ylab.text,
  ylab.size,
  ylab.color,
  ylab.rotation,
  ylab.space,
  ylab.fontface,
  ylab.fontfamily,
  ylab.alpha,
  ylab.side,
  panel.type,
  panel.wrap.pos,
  panel.xtab.pos,
  unit,
  color.sepia_intensity,
  color.saturation,
  color_vision_deficiency_sim,
  text.fontface,
  text.fontfamily,
  r,
  component.position,
  component.offset,
  component.stack_margin,
  component.autoscale,
  component.resize_as_group,
  component.frame_combine,
  component.stack,
  legend.stack,
  chart.stack,
  component.equalize,
  component.frame,
  component.frame.color,
  component.frame.alpha,
  component.frame.lwd,
  component.frame.r,
  component.bg,
  component.bg.color,
  component.bg.alpha,
  legend.show,
  legend.orientation,
  legend.position,
  legend.width,
  legend.height,
  legend.reverse,
  legend.na.show,
  legend.title.color,
  legend.title.size,
  legend.title.fontface,
  legend.title.fontfamily,
  legend.title.alpha,
  legend.xlab.color,
  legend.xlab.size,
  legend.xlab.rot,
  legend.xlab.fontface,
  legend.xlab.fontfamily,
  legend.xlab.alpha,
  legend.ylab.color,
  legend.ylab.size,
  legend.ylab.rot,
  legend.ylab.fontface,
  legend.ylab.fontfamily,
  legend.ylab.alpha,
  legend.text.color,
  legend.text.size,
  legend.text.fontface,
  legend.text.fontfamily,
  legend.text.alpha,
  legend.frame,
  legend.frame.color,
  legend.frame.alpha,
  legend.frame.lwd,
  legend.frame.r,
  legend.bg,
  legend.bg.color,
  legend.bg.alpha,
  legend.only,
  legend.absolute_fontsize,
  legend.settings.portrait,
  legend.settings.landscape,
  add_legend.position,
  chart.show,
  chart.plot.axis.x,
  chart.plot.axis.y,
  chart.position,
  chart.width,
  chart.height,
  chart.reverse,
  chart.na.show,
  chart.title.color,
  chart.title.size,
  chart.title.fontface,
  chart.title.fontfamily,
  chart.title.alpha,
  chart.xlab.color,
  chart.xlab.size,
  chart.xlab.fontface,
  chart.xlab.fontfamily,
  chart.xlab.alpha,
  chart.ylab.color,
  chart.ylab.size,
  chart.ylab.fontface,
  chart.ylab.fontfamily,
  chart.ylab.alpha,
  chart.text.color,
  chart.text.size,
  chart.text.fontface,
  chart.text.fontfamily,
  chart.text.alpha,
  chart.frame,
  chart.frame.color,
  chart.frame.alpha,
  chart.frame.lwd,
  chart.frame.r,
  chart.bg,
  chart.bg.color,
  chart.bg.alpha,
  chart.object.color,
  title.size,
  title.color,
  title.fontface,
  title.fontfamily,
  title.alpha,
  title.padding,
  title.frame,
  title.frame.color,
  title.frame.alpha,
  title.frame.lwd,
  title.frame.r,
  title.position,
  title.width,
  credits.size,
  credits.color,
  credits.fontface,
  credits.fontfamily,
  credits.alpha,
  credits.padding,
  credits.position,
  credits.width,
  credits.height,
  compass.north,
  compass.type,
  compass.text.size,
  compass.size,
  compass.show.labels,
  compass.cardinal.directions,
  compass.text.color,
  compass.color.dark,
  compass.color.light,
  compass.lwd,
  compass.margins,
  compass.position,
  inset.position,
  logo.height,
  logo.margins,
  logo.between_margin,
  logo.position,
  inset_map.height,
  inset_map.width,
  inset_map.margins,
  inset_map.between_margin,
  inset_map.position,
  inset_map.frame,
  inset.height,
  inset.width,
  inset.margins,
  inset.between_margin,
  inset.frame,
  inset.bg,
  inset.bg.color,
  inset.bg.alpha,
  inset_grob.height,
  inset_grob.width,
  inset_gg.height,
  inset_gg.width,
  scalebar.breaks,
  scalebar.width,
  scalebar.allow_clipping,
  scalebar.text.size,
  scalebar.text.color,
  scalebar.text.fontface,
  scalebar.text.fontfamily,
  scalebar.color.dark,
  scalebar.color.light,
  scalebar.lwd,
  scalebar.size,
  scalebar.margins,
  scalebar.position,
  grid.show,
  grid.labels.pos,
  grid.x,
  grid.y,
  grid.n.x,
  grid.n.y,
  grid.crs,
  grid.col,
  grid.lwd,
  grid.alpha,
  grid.labels.show,
  grid.labels.size,
  grid.labels.col,
  grid.labels.fontface,
  grid.labels.fontfamily,
  grid.labels.rot,
  grid.labels.format,
  grid.labels.cardinal,
  grid.labels.margin.x,
  grid.labels.margin.y,
  grid.labels.space.x,
  grid.labels.space.y,
  grid.labels.inside_frame,
  grid.ticks,
  grid.lines,
  grid.ndiscr,
  mouse_coordinates.position,
  minimap.server,
  minimap.toggle,
  minimap.position,
  panel.show,
  panel.labels,
  panel.label.size,
  panel.label.color,
  panel.label.fontface,
  panel.label.fontfamily,
  panel.label.alpha,
  panel.label.bg,
  panel.label.bg.color,
  panel.label.bg.alpha,
  panel.label.frame,
  panel.label.frame.color,
  panel.label.frame.alpha,
  panel.label.frame.lwd,
  panel.label.frame.r,
  panel.label.height,
  panel.label.rot,
  qtm.scalebar,
  qtm.minimap,
  qtm.mouse_coordinates,
  earth_boundary,
  earth_boundary.color,
  earth_boundary.lwd,
  earth_datum,
  space,
  space.color,
  space_overlay,
  check_and_fix,
  basemap.show,
  basemap.server,
  basemap.alpha,
  basemap.zoom,
  tiles.show,
  tiles.server,
  tiles.alpha,
  tiles.zoom,
  attr.color,
  crs_extra,
  crs_global,
  crs_basemap,
  title = NULL,
  main.title = NULL,
  main.title.size = NULL,
  main.title.color = NULL,
  main.title.fontface = NULL,
  main.title.fontfamily = NULL,
  main.title.position = NULL,
  fontface = NULL,
  fontfamily = NULL,
  style,
  ...
)
}
\arguments{
\item{crs}{Map crs (see \code{\link[=tm_shape]{tm_shape()}}). \code{NA} means the crs is specified in \code{\link[=tm_shape]{tm_shape()}}. The crs that is used by the transformation functions is defined in \code{\link[=tm_shape]{tm_shape()}}.}

\item{facet.max}{Maximum number of facets}

\item{free.scales}{For backward compatibility: if this value is set, it will be used to impute the free arguments in the layer functions}

\item{raster.max_cells}{Maximum number of raster grid cells. Can be mode specific \code{c(plot = 3000, view = 1000, 1000)} (the last value is the fall back default)}

\item{raster.warp}{Should rasters be warped or transformed in case a different projection (crs) is used? Warping creates a new regular raster in the target crs, whereas transforming creates a (usually non-regular) raster in the target crs. The former is lossy, but much faster and is therefore the default.
When a different projection (crs) is used, a (usually) regular raster will be}

\item{show.messages}{Show messages?}

\item{show.warnings}{Show warnings?}

\item{output.format}{Output format}

\item{output.size}{Output size}

\item{output.dpi}{Output dpi}

\item{animation.dpi}{Output dpi for animations}

\item{value.const}{Default visual value constants e.g. the default fill color for \code{tm_shape(World) + tm_polygons()}. A list is required with per visual variable a value.}

\item{value.na}{Default visual values that are used to visualize NA data values. A list is required with per visual variable a value.}

\item{value.null}{Default visual values that are used to visualize null (out-of-scope) data values. A list is required with per visual variable a value.}

\item{value.blank}{Default visual values that correspond to blank. For color these are \code{"#00000000"} meaning transparent. A list is required with per visual variable a value.}

\item{values.var}{Default values when a data variable to mapped to a visual variable, e.g. a color palette. A list is required with per visual variable a value.}

\item{values.range}{Default range for values. See \code{values.range} of \code{\link[=tm_scale_categorical]{tm_scale_categorical()}}. A list is required with per visual variable a value.}

\item{value.neutral}{Default values for when a data variable to mapped to a visual variable, e.g. a color palette. A list is required with per visual variable a value.}

\item{values.scale}{Default scales (as in object sizes) for values. See \code{values.range} of \code{\link[=tm_scale_categorical]{tm_scale_categorical()}}. A list is required with per visual variable a value.}

\item{scales.var}{Default scale functions per visual variable and type of data variable. A list is required with per visual variable per data type.}

\item{scale.misc.args}{Default values of scale function-specific arguments. A list is required with per scale function and optional per visual variable.}

\item{continuous.nclass_per_legend_break}{The number of continuous legend breaks within one 'unit' (label).  The default value is 50.}

\item{continuous.nclasses}{the number of classes of a continuous scale. Should be odd.  The default value is 101.}

\item{label.format}{Format for the labels. These are the default values for \code{\link[=tm_label_format]{tm_label_format()}}}

\item{label.na}{Default label for missing values.}

\item{scale}{Overall scale of the map}

\item{asp}{Aspect ratio of each map. When \code{asp} is set to \code{NA} (default) the aspect ratio will be adjusted to the used shapes. When set to 0 the aspect ratio is
adjusted to the size of the device divided by the number of columns and rows.}

\item{bg}{Draw map background?}

\item{bg.color}{Background color of the map.}

\item{outer.bg}{Draw map background (outside the frame)?}

\item{outer.bg.color}{Background color of map outside the frame.}

\item{frame}{Draw map frame?}

\item{frame.color}{The color of the frame.}

\item{frame.alpha}{The alpha transparency of the frame.}

\item{frame.lwd}{The line width of the frame. See \code{graphics::par}, option 'lwd'.}

\item{frame.r}{The r (radius) of the frame.}

\item{frame.double_line}{The double line of the frame. \code{TRUE} or \code{FALSE}.}

\item{outer.margins}{The margins of the outer space (outside the frame. A vector of 4 values: bottom, left, top, right. The unit is the device height (for bottom and top) or width (for left and right).}

\item{inner.margins}{The margins of the inner space (inside the frame). A vector of 4 values: bottom, left, top, right. The unit is the device height (for bottom and top) or width (for left and right).}

\item{inner.margins.extra}{The extra arguments of the margins of the inner space (inside the frame). A list of arguments.}

\item{meta.margins}{The margins of the meta. A vector of 4 values: bottom, left, top, right. The unit is the device height (for bottom and top) or width (for left and right).}

\item{meta.auto_margins}{The auto_margins of the meta.}

\item{between_margin}{Margin between the map.}

\item{panel.margin}{The margin of the panel.}

\item{xlab.show}{The visibility of the xlab. \code{TRUE} or \code{FALSE}.}

\item{xlab.text}{The text of the xlab.}

\item{xlab.size}{The size of the xlab.}

\item{xlab.color}{The color of the xlab.}

\item{xlab.rotation}{The rotation of the xlab.}

\item{xlab.space}{The space of the xlab. In terms of number of text line heights.}

\item{xlab.fontface}{The font face of the xlab. See \code{graphics::par}, option 'font'.}

\item{xlab.fontfamily}{The font family of the xlab. See \code{graphics::par}, option 'family'.}

\item{xlab.alpha}{The alpha transparency of the xlab.}

\item{xlab.side}{The side of the xlab.}

\item{ylab.show}{The visibility of the ylab. \code{TRUE} or \code{FALSE}.}

\item{ylab.text}{The text of the ylab.}

\item{ylab.size}{The size of the ylab.}

\item{ylab.color}{The color of the ylab.}

\item{ylab.rotation}{The rotation of the ylab.}

\item{ylab.space}{The space of the ylab. In terms of number of text line heights.}

\item{ylab.fontface}{The font face of the ylab. See \code{graphics::par}, option 'font'.}

\item{ylab.fontfamily}{The font family of the ylab. See \code{graphics::par}, option 'family'.}

\item{ylab.alpha}{The alpha transparency of the ylab.}

\item{ylab.side}{The side of the ylab.}

\item{panel.type}{The type of the panel.}

\item{panel.wrap.pos}{The panel positions for wrapped facets created with \code{\link[=tm_facets_grid]{tm_facets_grid()}}. One of \code{"left"}, \code{"right"}, \code{"top"} (default) or \code{"bottom"}.}

\item{panel.xtab.pos}{The panel positions for grid facets created with \code{\link[=tm_facets_grid]{tm_facets_grid()}}. Vector of two, where the first determines the locations of row panels (\code{"left"} or \code{"right"}) and the second the location of column panels ( \code{"top"} or `"bottom")}

\item{unit}{Unit of the coordinate}

\item{color.sepia_intensity}{The sepia_intensity of the color.}

\item{color.saturation}{The saturation of the color.}

\item{color_vision_deficiency_sim}{Color vision deficiency simulation. Either \code{"protan"}, \code{"deutan"}, or \code{"tritan"}.}

\item{text.fontface}{The font face of the text. See \code{graphics::par}, option 'font'.}

\item{text.fontfamily}{The font family of the text. See \code{graphics::par}, option 'family'.}

\item{r}{The r (radius) (overall).}

\item{component.position}{The position of the component. A tm_pos object, or a shortcut of two values: horizontal (left, center, right) and vertical (top, center, bottom). See tm_pos for details}

\item{component.offset}{The offset of the component.}

\item{component.stack_margin}{The stack_margin of the component.}

\item{component.autoscale}{The autoscale of the component.}

\item{component.resize_as_group}{The resize_as_group of the component.}

\item{component.frame_combine}{The frame_combine of the component.}

\item{component.stack}{The stack of the component.}

\item{legend.stack}{The stack of the legend.}

\item{chart.stack}{The stack of the chart.}

\item{component.equalize}{The equalize of the component.}

\item{component.frame}{The frame of the component.}

\item{component.frame.color}{The color of the frame of the component.}

\item{component.frame.alpha}{The alpha transparency of the frame of the component.}

\item{component.frame.lwd}{The line width of the frame of the component. See \code{graphics::par}, option 'lwd'.}

\item{component.frame.r}{The r (radius) of the frame of the component.}

\item{component.bg}{The bg of the component.}

\item{component.bg.color}{The color of the bg of the component.}

\item{component.bg.alpha}{The alpha transparency of the bg of the component.}

\item{legend.show}{The visibility of the legend. \code{TRUE} or \code{FALSE}.}

\item{legend.orientation}{The orientation of the legend.}

\item{legend.position}{The position of the legend. A tm_pos object, or a shortcut of two values: horizontal (left, center, right) and vertical (top, center, bottom). See tm_pos for details}

\item{legend.width}{The width of the legend.}

\item{legend.height}{The height of the legend.}

\item{legend.reverse}{The reverse of the legend.}

\item{legend.na.show}{The visibility of the na of the legend. \code{TRUE} or \code{FALSE}.}

\item{legend.title.color}{The color of the title of the legend.}

\item{legend.title.size}{The size of the title of the legend.}

\item{legend.title.fontface}{The font face of the title of the legend. See \code{graphics::par}, option 'font'.}

\item{legend.title.fontfamily}{The font family of the title of the legend. See \code{graphics::par}, option 'family'.}

\item{legend.title.alpha}{The alpha transparency of the title of the legend.}

\item{legend.xlab.color}{The color of the xlab of the legend.}

\item{legend.xlab.size}{The size of the xlab of the legend.}

\item{legend.xlab.rot}{The rot of the xlab of the legend.}

\item{legend.xlab.fontface}{The font face of the xlab of the legend. See \code{graphics::par}, option 'font'.}

\item{legend.xlab.fontfamily}{The font family of the xlab of the legend. See \code{graphics::par}, option 'family'.}

\item{legend.xlab.alpha}{The alpha transparency of the xlab of the legend.}

\item{legend.ylab.color}{The color of the ylab of the legend.}

\item{legend.ylab.size}{The size of the ylab of the legend.}

\item{legend.ylab.rot}{The rot of the ylab of the legend.}

\item{legend.ylab.fontface}{The font face of the ylab of the legend. See \code{graphics::par}, option 'font'.}

\item{legend.ylab.fontfamily}{The font family of the ylab of the legend. See \code{graphics::par}, option 'family'.}

\item{legend.ylab.alpha}{The alpha transparency of the ylab of the legend.}

\item{legend.text.color}{The color of the text of the legend.}

\item{legend.text.size}{The size of the text of the legend.}

\item{legend.text.fontface}{The font face of the text of the legend. See \code{graphics::par}, option 'font'.}

\item{legend.text.fontfamily}{The font family of the text of the legend. See \code{graphics::par}, option 'family'.}

\item{legend.text.alpha}{The alpha transparency of the text of the legend.}

\item{legend.frame}{The frame of the legend.}

\item{legend.frame.color}{The color of the frame of the legend.}

\item{legend.frame.alpha}{The alpha transparency of the frame of the legend.}

\item{legend.frame.lwd}{The line width of the frame of the legend. See \code{graphics::par}, option 'lwd'.}

\item{legend.frame.r}{The r (radius) of the frame of the legend.}

\item{legend.bg}{The bg of the legend.}

\item{legend.bg.color}{The color of the bg of the legend.}

\item{legend.bg.alpha}{The alpha transparency of the bg of the legend.}

\item{legend.only}{Should only legends be printed (so without map)?}

\item{legend.absolute_fontsize}{The absolute fontsize of the legend. So far, only used to calculate legend dimensions}

\item{legend.settings.portrait}{The portrait of the settings of the legend.}

\item{legend.settings.landscape}{The landscape of the settings of the legend.}

\item{add_legend.position}{The position of the add_legend. A tm_pos object, or a shortcut of two values: horizontal (left, center, right) and vertical (top, center, bottom). See tm_pos for details}

\item{chart.show}{The visibility of the chart. \code{TRUE} or \code{FALSE}.}

\item{chart.plot.axis.x}{The x of the axis of the plot of the chart.}

\item{chart.plot.axis.y}{The y of the axis of the plot of the chart.}

\item{chart.position}{The position of the chart. A tm_pos object, or a shortcut of two values: horizontal (left, center, right) and vertical (top, center, bottom). See tm_pos for details}

\item{chart.width}{The width of the chart.}

\item{chart.height}{The height of the chart.}

\item{chart.reverse}{The reverse of the chart.}

\item{chart.na.show}{The visibility of the na of the chart. \code{TRUE} or \code{FALSE}.}

\item{chart.title.color}{The color of the title of the chart.}

\item{chart.title.size}{The size of the title of the chart.}

\item{chart.title.fontface}{The font face of the title of the chart. See \code{graphics::par}, option 'font'.}

\item{chart.title.fontfamily}{The font family of the title of the chart. See \code{graphics::par}, option 'family'.}

\item{chart.title.alpha}{The alpha transparency of the title of the chart.}

\item{chart.xlab.color}{The color of the xlab of the chart.}

\item{chart.xlab.size}{The size of the xlab of the chart.}

\item{chart.xlab.fontface}{The font face of the xlab of the chart. See \code{graphics::par}, option 'font'.}

\item{chart.xlab.fontfamily}{The font family of the xlab of the chart. See \code{graphics::par}, option 'family'.}

\item{chart.xlab.alpha}{The alpha transparency of the xlab of the chart.}

\item{chart.ylab.color}{The color of the ylab of the chart.}

\item{chart.ylab.size}{The size of the ylab of the chart.}

\item{chart.ylab.fontface}{The font face of the ylab of the chart. See \code{graphics::par}, option 'font'.}

\item{chart.ylab.fontfamily}{The font family of the ylab of the chart. See \code{graphics::par}, option 'family'.}

\item{chart.ylab.alpha}{The alpha transparency of the ylab of the chart.}

\item{chart.text.color}{The color of the text of the chart.}

\item{chart.text.size}{The size of the text of the chart.}

\item{chart.text.fontface}{The font face of the text of the chart. See \code{graphics::par}, option 'font'.}

\item{chart.text.fontfamily}{The font family of the text of the chart. See \code{graphics::par}, option 'family'.}

\item{chart.text.alpha}{The alpha transparency of the text of the chart.}

\item{chart.frame}{The frame of the chart.}

\item{chart.frame.color}{The color of the frame of the chart.}

\item{chart.frame.alpha}{The alpha transparency of the frame of the chart.}

\item{chart.frame.lwd}{The line width of the frame of the chart. See \code{graphics::par}, option 'lwd'.}

\item{chart.frame.r}{The r (radius) of the frame of the chart.}

\item{chart.bg}{The bg of the chart.}

\item{chart.bg.color}{The color of the bg of the chart.}

\item{chart.bg.alpha}{The alpha transparency of the bg of the chart.}

\item{chart.object.color}{The color of the object of the chart.}

\item{title.size}{The size of the title.}

\item{title.color}{The color of the title.}

\item{title.fontface}{The font face of the title. See \code{graphics::par}, option 'font'.}

\item{title.fontfamily}{The font family of the title. See \code{graphics::par}, option 'family'.}

\item{title.alpha}{The alpha transparency of the title.}

\item{title.padding}{The padding of the title. A vector of 4 values: bottom, left, top, right. The unit is the device height (for bottom and top) or width (for left and right).}

\item{title.frame}{The frame of the title.}

\item{title.frame.color}{The color of the frame of the title.}

\item{title.frame.alpha}{The alpha transparency of the frame of the title.}

\item{title.frame.lwd}{The line width of the frame of the title. See \code{graphics::par}, option 'lwd'.}

\item{title.frame.r}{The r (radius) of the frame of the title.}

\item{title.position}{The position of the title. A tm_pos object, or a shortcut of two values: horizontal (left, center, right) and vertical (top, center, bottom). See tm_pos for details}

\item{title.width}{The width of the title.}

\item{credits.size}{The size of the credits.}

\item{credits.color}{The color of the credits.}

\item{credits.fontface}{The font face of the credits. See \code{graphics::par}, option 'font'.}

\item{credits.fontfamily}{The font family of the credits. See \code{graphics::par}, option 'family'.}

\item{credits.alpha}{The alpha transparency of the credits.}

\item{credits.padding}{The padding of the credits. A vector of 4 values: bottom, left, top, right. The unit is the device height (for bottom and top) or width (for left and right).}

\item{credits.position}{The position of the credits. A tm_pos object, or a shortcut of two values: horizontal (left, center, right) and vertical (top, center, bottom). See tm_pos for details}

\item{credits.width}{The width of the credits.}

\item{credits.height}{The height of the credits.}

\item{compass.north}{The north of the compass.}

\item{compass.type}{The type of the compass.}

\item{compass.text.size}{The size of the text of the compass.}

\item{compass.size}{The size of the compass.}

\item{compass.show.labels}{The labels of the show of the compass.}

\item{compass.cardinal.directions}{The directions of the cardinal of the compass.}

\item{compass.text.color}{The color of the text of the compass.}

\item{compass.color.dark}{The dark of the color of the compass.}

\item{compass.color.light}{The light of the color of the compass.}

\item{compass.lwd}{The line width of the compass. See \code{graphics::par}, option 'lwd'.}

\item{compass.margins}{The margins of the compass. A vector of 4 values: bottom, left, top, right. The unit is the device height (for bottom and top) or width (for left and right).}

\item{compass.position}{The position of the compass. A tm_pos object, or a shortcut of two values: horizontal (left, center, right) and vertical (top, center, bottom). See tm_pos for details}

\item{inset.position}{The position of the inset. A tm_pos object, or a shortcut of two values: horizontal (left, center, right) and vertical (top, center, bottom). See tm_pos for details}

\item{logo.height}{The height of the logo.}

\item{logo.margins}{The margins of the logo. A vector of 4 values: bottom, left, top, right. The unit is the device height (for bottom and top) or width (for left and right).}

\item{logo.between_margin}{The between_margin of the logo.}

\item{logo.position}{The position of the logo. A tm_pos object, or a shortcut of two values: horizontal (left, center, right) and vertical (top, center, bottom). See tm_pos for details}

\item{inset_map.height}{The height of the inset_map.}

\item{inset_map.width}{The width of the inset_map.}

\item{inset_map.margins}{The margins of the inset_map. A vector of 4 values: bottom, left, top, right. The unit is the device height (for bottom and top) or width (for left and right).}

\item{inset_map.between_margin}{The between_margin of the inset_map.}

\item{inset_map.position}{The position of the inset_map. A tm_pos object, or a shortcut of two values: horizontal (left, center, right) and vertical (top, center, bottom). See tm_pos for details}

\item{inset_map.frame}{The frame of the inset_map.}

\item{inset.height}{The height of the inset.}

\item{inset.width}{The width of the inset.}

\item{inset.margins}{The margins of the inset. A vector of 4 values: bottom, left, top, right. The unit is the device height (for bottom and top) or width (for left and right).}

\item{inset.between_margin}{The between_margin of the inset.}

\item{inset.frame}{The frame of the inset.}

\item{inset.bg}{The bg of the inset.}

\item{inset.bg.color}{The color of the bg of the inset.}

\item{inset.bg.alpha}{The alpha transparency of the bg of the inset.}

\item{inset_grob.height}{The height of the inset_grob.}

\item{inset_grob.width}{The width of the inset_grob.}

\item{inset_gg.height}{The height of the inset_gg.}

\item{inset_gg.width}{The width of the inset_gg.}

\item{scalebar.breaks}{See \code{\link[=tm_scalebar]{tm_scalebar()}}}

\item{scalebar.width}{See \code{\link[=tm_scalebar]{tm_scalebar()}}}

\item{scalebar.allow_clipping}{See \code{\link[=tm_scalebar]{tm_scalebar()}}}

\item{scalebar.text.size}{The size of the text of the scalebar.}

\item{scalebar.text.color}{The color of the text of the scalebar.}

\item{scalebar.text.fontface}{The font face of the text of the scalebar. See \code{graphics::par}, option 'font'.}

\item{scalebar.text.fontfamily}{The font family of the text of the scalebar. See \code{graphics::par}, option 'family'.}

\item{scalebar.color.dark}{The dark of the color of the scalebar.}

\item{scalebar.color.light}{The light of the color of the scalebar.}

\item{scalebar.lwd}{The line width of the scalebar. See \code{graphics::par}, option 'lwd'.}

\item{scalebar.size}{The size of the scalebar.}

\item{scalebar.margins}{The margins of the scalebar. A vector of 4 values: bottom, left, top, right. The unit is the device height (for bottom and top) or width (for left and right).}

\item{scalebar.position}{The position of the scalebar. A tm_pos object, or a shortcut of two values: horizontal (left, center, right) and vertical (top, center, bottom). See tm_pos for details}

\item{grid.show}{The visibility of the grid. \code{TRUE} or \code{FALSE}.}

\item{grid.labels.pos}{The pos of the labels of the grid.}

\item{grid.x}{The x of the grid.}

\item{grid.y}{The y of the grid.}

\item{grid.n.x}{The x of the n of the grid.}

\item{grid.n.y}{The y of the n of the grid.}

\item{grid.crs}{The coordinate reference system (CRS) of the grid.}

\item{grid.col}{The color of the grid.}

\item{grid.lwd}{The line width of the grid. See \code{graphics::par}, option 'lwd'.}

\item{grid.alpha}{The alpha transparency of the grid.}

\item{grid.labels.show}{The visibility of the labels of the grid. \code{TRUE} or \code{FALSE}.}

\item{grid.labels.size}{The size of the labels of the grid.}

\item{grid.labels.col}{The color of the labels of the grid.}

\item{grid.labels.fontface}{The font face of the labels of the grid. See \code{graphics::par}, option 'font'.}

\item{grid.labels.fontfamily}{The font family of the labels of the grid. See \code{graphics::par}, option 'family'.}

\item{grid.labels.rot}{The rot of the labels of the grid.}

\item{grid.labels.format}{The format of the labels of the grid.}

\item{grid.labels.cardinal}{The cardinal of the labels of the grid.}

\item{grid.labels.margin.x}{The x of the margin of the labels of the grid.}

\item{grid.labels.margin.y}{The y of the margin of the labels of the grid.}

\item{grid.labels.space.x}{The x of the space of the labels of the grid.}

\item{grid.labels.space.y}{The y of the space of the labels of the grid.}

\item{grid.labels.inside_frame}{The inside_frame of the labels of the grid.}

\item{grid.ticks}{The ticks of the grid.}

\item{grid.lines}{The lines of the grid.}

\item{grid.ndiscr}{The ndiscr of the grid.}

\item{mouse_coordinates.position}{The position of the mouse_coordinates. A tm_pos object, or a shortcut of two values: horizontal (left, center, right) and vertical (top, center, bottom). See tm_pos for details}

\item{minimap.server}{The server of the minimap.}

\item{minimap.toggle}{The toggle of the minimap.}

\item{minimap.position}{The position of the minimap. A tm_pos object, or a shortcut of two values: horizontal (left, center, right) and vertical (top, center, bottom). See tm_pos for details}

\item{panel.show}{The visibility of the panel. \code{TRUE} or \code{FALSE}.}

\item{panel.labels}{The labels of the panel.}

\item{panel.label.size}{The size of the label of the panel.}

\item{panel.label.color}{The color of the label of the panel.}

\item{panel.label.fontface}{The font face of the label of the panel. See \code{graphics::par}, option 'font'.}

\item{panel.label.fontfamily}{The font family of the label of the panel. See \code{graphics::par}, option 'family'.}

\item{panel.label.alpha}{The alpha transparency of the label of the panel.}

\item{panel.label.bg}{The bg of the label of the panel.}

\item{panel.label.bg.color}{The color of the bg of the label of the panel.}

\item{panel.label.bg.alpha}{The alpha transparency of the bg of the label of the panel.}

\item{panel.label.frame}{The frame of the label of the panel.}

\item{panel.label.frame.color}{The color of the frame of the label of the panel.}

\item{panel.label.frame.alpha}{The alpha transparency of the frame of the label of the panel.}

\item{panel.label.frame.lwd}{The line width of the frame of the label of the panel. See \code{graphics::par}, option 'lwd'.}

\item{panel.label.frame.r}{The r (radius) of the frame of the label of the panel.}

\item{panel.label.height}{The height of the label of the panel.}

\item{panel.label.rot}{Rotation angles of the panel labels. Vector of four values that determine the panel label rotation when they are placed left, top, right, and bottom. The default angles are 90, 0, 270 and 0 respectively. Note that the second value is the most common, since labels are by default shown on top (see \code{panel.wrap.pos}). In cross-table facets created with \code{tm_facets_grid()}, the first two values are used by default (see \code{panel.xtab.pos}).}

\item{qtm.scalebar}{The scalebar of the qtm.}

\item{qtm.minimap}{The minimap of the qtm.}

\item{qtm.mouse_coordinates}{The mouse_coordinates of the qtm.}

\item{earth_boundary}{The earth boundary}

\item{earth_boundary.color}{The color of the earth_boundary.}

\item{earth_boundary.lwd}{The line width of the earth_boundary. See \code{graphics::par}, option 'lwd'.}

\item{earth_datum}{Earth datum}

\item{space}{Should the space be drawn? Only applicable is earth_boundary is enabled.}

\item{space.color}{The color of the space.}

\item{space_overlay}{Should the space be drawn as overlay (to make sure spatial features or rasters do not exceed the earth boundary), or as background? By default \code{TRUE} when a raster is warped.}

\item{check_and_fix}{Should attempt to fix an invalid shapefile}

\item{basemap.show}{The visibility of the basemap. \code{TRUE} or \code{FALSE}.}

\item{basemap.server}{The server of the basemap.}

\item{basemap.alpha}{The alpha transparency of the basemap.}

\item{basemap.zoom}{The zoom of the basemap.}

\item{tiles.show}{The visibility of the tiles. \code{TRUE} or \code{FALSE}.}

\item{tiles.server}{The server of the tiles.}

\item{tiles.alpha}{The alpha transparency of the tiles.}

\item{tiles.zoom}{The zoom of the tiles.}

\item{attr.color}{The color of the attr.}

\item{crs_extra}{Only used internally (work in progress)}

\item{crs_global}{The used crs for world maps}

\item{crs_basemap}{The crs_basemap (overall).}

\item{title}{deprecated See \code{\link[=tm_title]{tm_title()}}}

\item{main.title}{deprecated See \code{\link[=tm_title]{tm_title()}}}

\item{main.title.size, main.title.color, main.title.fontface, main.title.fontfamily, main.title.position}{deprecated. Use the \code{title.} options instead.}

\item{fontface, fontfamily}{renamed to \code{text.fontface} and \code{text.fontfamily}}

\item{style}{style see \code{tm_style()}}

\item{...}{List of tmap options to be set, or option names (characters) to be returned (see details)}
}
\description{
tmap options
}
\seealso{
\href{https://r-tmap.github.io/tmap/articles/basics_layout}{Vignette about layout}, \href{https://r-tmap.github.io/tmap/articles/adv_margins}{vignette about margins and aspect ratio} and \href{https://r-tmap.github.io/tmap/articles/adv_options}{vignette about options}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tmapScale_misc.R
\name{get_scale_defaults}
\alias{get_scale_defaults}
\title{Internal tmap function get scale values}
\usage{
get_scale_defaults(scale, o, aes, layer, cls, ct = NULL)
}
\arguments{
\item{scale}{scale}

\item{o}{o}

\item{aes}{aes}

\item{layer}{layer}

\item{cls}{cls}

\item{ct}{ct}
}
\description{
Internal tmap function get scale values
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tmapShape.R
\name{tmapGetShapeMeta2}
\alias{tmapGetShapeMeta2}
\title{Internal method that extracts more meta data from shape objects}
\usage{
tmapGetShapeMeta2(shp, smeta, o)
}
\arguments{
\item{shp}{the shape}

\item{smeta}{meta (from \code{tmapGetShapeMeta1()})}

\item{o}{the list of options}
}
\description{
Internal method that extracts meta data from shape objects
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tm_options.R
\name{tm_options}
\alias{tm_options}
\title{tmap options}
\usage{
tm_options(
  crs,
  facet.max,
  facet_levels.max,
  free.scales,
  raster.max_cells,
  raster.warp,
  show.messages,
  show.warnings,
  output.format,
  output.size,
  output.dpi,
  animation.dpi,
  value.const,
  value.na,
  value.null,
  value.blank,
  values.var,
  values.range,
  value.neutral,
  values.scale,
  scales.var,
  scale.misc.args,
  continuous.nclass_per_legend_break,
  continuous.nclasses,
  label.format,
  label.na,
  scale,
  asp,
  bg,
  bg.color,
  outer.bg,
  outer.bg.color,
  frame,
  frame.color,
  frame.alpha,
  frame.lwd,
  frame.r,
  frame.double_line,
  outer.margins,
  inner.margins,
  inner.margins.extra,
  meta.margins,
  meta.auto_margins,
  between_margin,
  panel.margin,
  xlab.show,
  xlab.text,
  xlab.size,
  xlab.color,
  xlab.rotation,
  xlab.space,
  xlab.fontface,
  xlab.fontfamily,
  xlab.alpha,
  xlab.side,
  ylab.show,
  ylab.text,
  ylab.size,
  ylab.color,
  ylab.rotation,
  ylab.space,
  ylab.fontface,
  ylab.fontfamily,
  ylab.alpha,
  ylab.side,
  panel.type,
  panel.wrap.pos,
  panel.xtab.pos,
  unit,
  color.sepia_intensity,
  color.saturation,
  color_vision_deficiency_sim,
  text.fontface,
  text.fontfamily,
  r,
  component.position,
  component.offset,
  component.stack_margin,
  component.autoscale,
  component.resize_as_group,
  component.frame_combine,
  component.stack,
  legend.stack,
  chart.stack,
  component.equalize,
  component.frame,
  component.frame.color,
  component.frame.alpha,
  component.frame.lwd,
  component.frame.r,
  component.bg,
  component.bg.color,
  component.bg.alpha,
  legend.show,
  legend.orientation,
  legend.position,
  legend.width,
  legend.height,
  legend.reverse,
  legend.na.show,
  legend.title.color,
  legend.title.size,
  legend.title.fontface,
  legend.title.fontfamily,
  legend.title.alpha,
  legend.xlab.color,
  legend.xlab.size,
  legend.xlab.rot,
  legend.xlab.fontface,
  legend.xlab.fontfamily,
  legend.xlab.alpha,
  legend.ylab.color,
  legend.ylab.size,
  legend.ylab.rot,
  legend.ylab.fontface,
  legend.ylab.fontfamily,
  legend.ylab.alpha,
  legend.text.color,
  legend.text.size,
  legend.text.fontface,
  legend.text.fontfamily,
  legend.text.alpha,
  legend.frame,
  legend.frame.color,
  legend.frame.alpha,
  legend.frame.lwd,
  legend.frame.r,
  legend.bg,
  legend.bg.color,
  legend.bg.alpha,
  legend.only,
  legend.absolute_fontsize,
  legend.settings.portrait,
  legend.settings.landscape,
  add_legend.position,
  chart.show,
  chart.plot.axis.x,
  chart.plot.axis.y,
  chart.position,
  chart.width,
  chart.height,
  chart.reverse,
  chart.na.show,
  chart.title.color,
  chart.title.size,
  chart.title.fontface,
  chart.title.fontfamily,
  chart.title.alpha,
  chart.xlab.color,
  chart.xlab.size,
  chart.xlab.fontface,
  chart.xlab.fontfamily,
  chart.xlab.alpha,
  chart.ylab.color,
  chart.ylab.size,
  chart.ylab.fontface,
  chart.ylab.fontfamily,
  chart.ylab.alpha,
  chart.text.color,
  chart.text.size,
  chart.text.fontface,
  chart.text.fontfamily,
  chart.text.alpha,
  chart.frame,
  chart.frame.color,
  chart.frame.alpha,
  chart.frame.lwd,
  chart.frame.r,
  chart.bg,
  chart.bg.color,
  chart.bg.alpha,
  chart.object.color,
  title.size,
  title.color,
  title.fontface,
  title.fontfamily,
  title.alpha,
  title.padding,
  title.frame,
  title.frame.color,
  title.frame.alpha,
  title.frame.lwd,
  title.frame.r,
  title.position,
  title.width,
  credits.size,
  credits.color,
  credits.fontface,
  credits.fontfamily,
  credits.alpha,
  credits.padding,
  credits.position,
  credits.width,
  credits.height,
  compass.north,
  compass.type,
  compass.text.size,
  compass.size,
  compass.show.labels,
  compass.cardinal.directions,
  compass.text.color,
  compass.color.dark,
  compass.color.light,
  compass.lwd,
  compass.margins,
  compass.position,
  inset.position,
  logo.height,
  logo.margins,
  logo.between_margin,
  logo.position,
  inset_map.height,
  inset_map.width,
  inset_map.margins,
  inset_map.between_margin,
  inset_map.position,
  inset_map.frame,
  inset.height,
  inset.width,
  inset.margins,
  inset.between_margin,
  inset.frame,
  inset.bg,
  inset.bg.color,
  inset.bg.alpha,
  inset_grob.height,
  inset_grob.width,
  inset_gg.height,
  inset_gg.width,
  scalebar.breaks,
  scalebar.width,
  scalebar.allow_clipping,
  scalebar.text.size,
  scalebar.text.color,
  scalebar.text.fontface,
  scalebar.text.fontfamily,
  scalebar.color.dark,
  scalebar.color.light,
  scalebar.lwd,
  scalebar.size,
  scalebar.margins,
  scalebar.position,
  grid.show,
  grid.labels.pos,
  grid.x,
  grid.y,
  grid.n.x,
  grid.n.y,
  grid.crs,
  grid.col,
  grid.lwd,
  grid.alpha,
  grid.labels.show,
  grid.labels.size,
  grid.labels.col,
  grid.labels.fontface,
  grid.labels.fontfamily,
  grid.labels.rot,
  grid.labels.format,
  grid.labels.cardinal,
  grid.labels.margin.x,
  grid.labels.margin.y,
  grid.labels.space.x,
  grid.labels.space.y,
  grid.labels.inside_frame,
  grid.ticks,
  grid.lines,
  grid.ndiscr,
  mouse_coordinates.position,
  minimap.server,
  minimap.toggle,
  minimap.position,
  panel.show,
  panel.labels,
  panel.label.size,
  panel.label.color,
  panel.label.fontface,
  panel.label.fontfamily,
  panel.label.alpha,
  panel.label.bg,
  panel.label.bg.color,
  panel.label.bg.alpha,
  panel.label.frame,
  panel.label.frame.color,
  panel.label.frame.alpha,
  panel.label.frame.lwd,
  panel.label.frame.r,
  panel.label.height,
  panel.label.rot,
  qtm.scalebar,
  qtm.minimap,
  qtm.mouse_coordinates,
  earth_boundary,
  earth_boundary.color,
  earth_boundary.lwd,
  earth_datum,
  space,
  space.color,
  space_overlay,
  check_and_fix,
  basemap.show,
  basemap.server,
  basemap.alpha,
  basemap.zoom,
  tiles.show,
  tiles.server,
  tiles.alpha,
  tiles.zoom,
  attr.color,
  crs_extra,
  crs_global,
  crs_basemap,
  title = NULL,
  main.title = NULL,
  main.title.size = NULL,
  main.title.color = NULL,
  main.title.fontface = NULL,
  main.title.fontfamily = NULL,
  main.title.position = NULL,
  fontface = NULL,
  fontfamily = NULL,
  style,
  ...
)
}
\arguments{
\item{crs}{Map crs (see \code{\link[=tm_shape]{tm_shape()}}). \code{NA} means the crs is specified in \code{\link[=tm_shape]{tm_shape()}}. The crs that is used by the transformation functions is defined in \code{\link[=tm_shape]{tm_shape()}}.}

\item{facet.max}{Maximum number of facets in a static plot (rows x columns); animation frames are not counted.}

\item{facet_levels.max}{Maximum number of facet levels processed across all faceting dimensions, animation frames included (rows x columns x frames). Default \code{1e5}.}

\item{free.scales}{For backward compatibility: if this value is set, it will be used to impute the free arguments in the layer functions}

\item{raster.max_cells}{Maximum number of raster grid cells. Can be mode specific \code{c(plot = 3000, view = 1000, 1000)} (the last value is the fall back default)}

\item{raster.warp}{Should rasters be warped or transformed in case a different projection (crs) is used? Warping creates a new regular raster in the target crs, whereas transforming creates a (usually non-regular) raster in the target crs. The former is lossy, but much faster and is therefore the default.
When a different projection (crs) is used, a (usually) regular raster will be}

\item{show.messages}{Show messages?}

\item{show.warnings}{Show warnings?}

\item{output.format}{Output format}

\item{output.size}{Output size}

\item{output.dpi}{Output dpi}

\item{animation.dpi}{Output dpi for animations}

\item{value.const}{Default visual value constants e.g. the default fill color for \code{tm_shape(World) + tm_polygons()}. A list is required with per visual variable a value.}

\item{value.na}{Default visual values that are used to visualize NA data values. A list is required with per visual variable a value.}

\item{value.null}{Default visual values that are used to visualize null (out-of-scope) data values. A list is required with per visual variable a value.}

\item{value.blank}{Default visual values that correspond to blank. For color these are \code{"#00000000"} meaning transparent. A list is required with per visual variable a value.}

\item{values.var}{Default values when a data variable to mapped to a visual variable, e.g. a color palette. A list is required with per visual variable a value.}

\item{values.range}{Default range for values. See \code{values.range} of \code{\link[=tm_scale_categorical]{tm_scale_categorical()}}. A list is required with per visual variable a value.}

\item{value.neutral}{Default values for when a data variable to mapped to a visual variable, e.g. a color palette. A list is required with per visual variable a value.}

\item{values.scale}{Default scales (as in object sizes) for values. See \code{values.range} of \code{\link[=tm_scale_categorical]{tm_scale_categorical()}}. A list is required with per visual variable a value.}

\item{scales.var}{Default scale functions per visual variable and type of data variable. A list is required with per visual variable per data type.}

\item{scale.misc.args}{Default values of scale function-specific arguments. A list is required with per scale function and optional per visual variable.}

\item{continuous.nclass_per_legend_break}{The number of continuous legend breaks within one 'unit' (label).  The default value is 50.}

\item{continuous.nclasses}{the number of classes of a continuous scale. Should be odd.  The default value is 101.}

\item{label.format}{Format for the labels. These are the default values for \code{\link[=tm_label_format]{tm_label_format()}}}

\item{label.na}{Default label for missing values.}

\item{scale}{Overall scale of the map}

\item{asp}{Aspect ratio of each map. When \code{asp} is set to \code{NA} (default) the aspect ratio will be adjusted to the used shapes. When set to 0 the aspect ratio is
adjusted to the size of the device divided by the number of columns and rows.}

\item{bg}{Draw map background?}

\item{bg.color}{Background color of the map.}

\item{outer.bg}{Draw map background (outside the frame)?}

\item{outer.bg.color}{Background color of map outside the frame.}

\item{frame}{Draw map frame?}

\item{frame.color}{The color of the frame.}

\item{frame.alpha}{The alpha transparency of the frame.}

\item{frame.lwd}{The line width of the frame. See \code{graphics::par}, option 'lwd'.}

\item{frame.r}{The r (radius) of the frame.}

\item{frame.double_line}{The double line of the frame. \code{TRUE} or \code{FALSE}.}

\item{outer.margins}{The margins of the outer space (outside the frame. A vector of 4 values: bottom, left, top, right. The unit is the device height (for bottom and top) or width (for left and right).}

\item{inner.margins}{The margins of the inner space (inside the frame). A vector of 4 values: bottom, left, top, right. The unit is the device height (for bottom and top) or width (for left and right).}

\item{inner.margins.extra}{The extra arguments of the margins of the inner space (inside the frame). A list of arguments.}

\item{meta.margins}{The margins of the 'meta' space (reserved for outside map components). A vector of 4 values: bottom, left, top, right. The unit is the device height (for bottom and top) or width (for left and right).}

\item{meta.auto_margins}{The auto_margins of the 'meta' space (reserved for outside map components).}

\item{between_margin}{Margin between the map.}

\item{panel.margin}{The margin of the panel.}

\item{xlab.show}{The visibility of the xlab. \code{TRUE} or \code{FALSE}.}

\item{xlab.text}{The text of the xlab.}

\item{xlab.size}{The size of the xlab.}

\item{xlab.color}{The color of the xlab.}

\item{xlab.rotation}{The rotation of the xlab.}

\item{xlab.space}{The space of the xlab. In terms of number of text line heights.}

\item{xlab.fontface}{The font face of the xlab. See \code{graphics::par}, option 'font'.}

\item{xlab.fontfamily}{The font family of the xlab. See \code{graphics::par}, option 'family'.}

\item{xlab.alpha}{The alpha transparency of the xlab.}

\item{xlab.side}{The side of the xlab.}

\item{ylab.show}{The visibility of the ylab. \code{TRUE} or \code{FALSE}.}

\item{ylab.text}{The text of the ylab.}

\item{ylab.size}{The size of the ylab.}

\item{ylab.color}{The color of the ylab.}

\item{ylab.rotation}{The rotation of the ylab.}

\item{ylab.space}{The space of the ylab. In terms of number of text line heights.}

\item{ylab.fontface}{The font face of the ylab. See \code{graphics::par}, option 'font'.}

\item{ylab.fontfamily}{The font family of the ylab. See \code{graphics::par}, option 'family'.}

\item{ylab.alpha}{The alpha transparency of the ylab.}

\item{ylab.side}{The side of the ylab.}

\item{panel.type}{The type of the panel.}

\item{panel.wrap.pos}{The panel positions for wrapped facets created with \code{\link[=tm_facets_grid]{tm_facets_grid()}}. One of \code{"left"}, \code{"right"}, \code{"top"} (default) or \code{"bottom"}.}

\item{panel.xtab.pos}{The panel positions for grid facets created with \code{\link[=tm_facets_grid]{tm_facets_grid()}}. Vector of two, where the first determines the locations of row panels (\code{"left"} or \code{"right"}) and the second the location of column panels ( \code{"top"} or `"bottom")}

\item{unit}{Unit of the coordinate}

\item{color.sepia_intensity}{The sepia_intensity of the color.}

\item{color.saturation}{The saturation of the color.}

\item{color_vision_deficiency_sim}{Color vision deficiency simulation. Either \code{"protan"}, \code{"deutan"}, or \code{"tritan"}.}

\item{text.fontface}{The font face of the text. See \code{graphics::par}, option 'font'.}

\item{text.fontfamily}{The font family of the text. See \code{graphics::par}, option 'family'.}

\item{r}{The r (radius) (overall).}

\item{component.position}{The position of the component. A tm_pos object, or a shortcut of two values: horizontal (left, center, right) and vertical (top, center, bottom). See tm_pos for details}

\item{component.offset}{The offset of the component.}

\item{component.stack_margin}{The stack_margin of the component.}

\item{component.autoscale}{The autoscale of the component.}

\item{component.resize_as_group}{The resize_as_group of the component.}

\item{component.frame_combine}{The frame_combine of the component.}

\item{component.stack}{The stack of the component.}

\item{legend.stack}{The stack of the legend.}

\item{chart.stack}{The stack of the chart.}

\item{component.equalize}{The equalize of the component.}

\item{component.frame}{The frame of the component.}

\item{component.frame.color}{The color of the frame of the component.}

\item{component.frame.alpha}{The alpha transparency of the frame of the component.}

\item{component.frame.lwd}{The line width of the frame of the component. See \code{graphics::par}, option 'lwd'.}

\item{component.frame.r}{The r (radius) of the frame of the component.}

\item{component.bg}{The bg of the component.}

\item{component.bg.color}{The color of the bg of the component.}

\item{component.bg.alpha}{The alpha transparency of the bg of the component.}

\item{legend.show}{The visibility of the legend. \code{TRUE} or \code{FALSE}.}

\item{legend.orientation}{The orientation of the legend.}

\item{legend.position}{The position of the legend. A tm_pos object, or a shortcut of two values: horizontal (left, center, right) and vertical (top, center, bottom). See tm_pos for details}

\item{legend.width}{The width of the legend.}

\item{legend.height}{The height of the legend.}

\item{legend.reverse}{The reverse of the legend.}

\item{legend.na.show}{The visibility of the na of the legend. \code{TRUE} or \code{FALSE}.}

\item{legend.title.color}{The color of the title of the legend.}

\item{legend.title.size}{The size of the title of the legend.}

\item{legend.title.fontface}{The font face of the title of the legend. See \code{graphics::par}, option 'font'.}

\item{legend.title.fontfamily}{The font family of the title of the legend. See \code{graphics::par}, option 'family'.}

\item{legend.title.alpha}{The alpha transparency of the title of the legend.}

\item{legend.xlab.color}{The color of the xlab of the legend.}

\item{legend.xlab.size}{The size of the xlab of the legend.}

\item{legend.xlab.rot}{The rot of the xlab of the legend.}

\item{legend.xlab.fontface}{The font face of the xlab of the legend. See \code{graphics::par}, option 'font'.}

\item{legend.xlab.fontfamily}{The font family of the xlab of the legend. See \code{graphics::par}, option 'family'.}

\item{legend.xlab.alpha}{The alpha transparency of the xlab of the legend.}

\item{legend.ylab.color}{The color of the ylab of the legend.}

\item{legend.ylab.size}{The size of the ylab of the legend.}

\item{legend.ylab.rot}{The rot of the ylab of the legend.}

\item{legend.ylab.fontface}{The font face of the ylab of the legend. See \code{graphics::par}, option 'font'.}

\item{legend.ylab.fontfamily}{The font family of the ylab of the legend. See \code{graphics::par}, option 'family'.}

\item{legend.ylab.alpha}{The alpha transparency of the ylab of the legend.}

\item{legend.text.color}{The color of the text of the legend.}

\item{legend.text.size}{The size of the text of the legend.}

\item{legend.text.fontface}{The font face of the text of the legend. See \code{graphics::par}, option 'font'.}

\item{legend.text.fontfamily}{The font family of the text of the legend. See \code{graphics::par}, option 'family'.}

\item{legend.text.alpha}{The alpha transparency of the text of the legend.}

\item{legend.frame}{The frame of the legend.}

\item{legend.frame.color}{The color of the frame of the legend.}

\item{legend.frame.alpha}{The alpha transparency of the frame of the legend.}

\item{legend.frame.lwd}{The line width of the frame of the legend. See \code{graphics::par}, option 'lwd'.}

\item{legend.frame.r}{The r (radius) of the frame of the legend.}

\item{legend.bg}{The bg of the legend.}

\item{legend.bg.color}{The color of the bg of the legend.}

\item{legend.bg.alpha}{The alpha transparency of the bg of the legend.}

\item{legend.only}{Should only legends be printed (so without map)?}

\item{legend.absolute_fontsize}{The absolute fontsize of the legend. So far, only used to calculate legend dimensions}

\item{legend.settings.portrait}{The portrait of the settings of the legend.}

\item{legend.settings.landscape}{The landscape of the settings of the legend.}

\item{add_legend.position}{The position of the add_legend. A tm_pos object, or a shortcut of two values: horizontal (left, center, right) and vertical (top, center, bottom). See tm_pos for details}

\item{chart.show}{The visibility of the chart. \code{TRUE} or \code{FALSE}.}

\item{chart.plot.axis.x}{The x of the axis of the plot of the chart.}

\item{chart.plot.axis.y}{The y of the axis of the plot of the chart.}

\item{chart.position}{The position of the chart. A tm_pos object, or a shortcut of two values: horizontal (left, center, right) and vertical (top, center, bottom). See tm_pos for details}

\item{chart.width}{The width of the chart.}

\item{chart.height}{The height of the chart.}

\item{chart.reverse}{The reverse of the chart.}

\item{chart.na.show}{The visibility of the na of the chart. \code{TRUE} or \code{FALSE}.}

\item{chart.title.color}{The color of the title of the chart.}

\item{chart.title.size}{The size of the title of the chart.}

\item{chart.title.fontface}{The font face of the title of the chart. See \code{graphics::par}, option 'font'.}

\item{chart.title.fontfamily}{The font family of the title of the chart. See \code{graphics::par}, option 'family'.}

\item{chart.title.alpha}{The alpha transparency of the title of the chart.}

\item{chart.xlab.color}{The color of the xlab of the chart.}

\item{chart.xlab.size}{The size of the xlab of the chart.}

\item{chart.xlab.fontface}{The font face of the xlab of the chart. See \code{graphics::par}, option 'font'.}

\item{chart.xlab.fontfamily}{The font family of the xlab of the chart. See \code{graphics::par}, option 'family'.}

\item{chart.xlab.alpha}{The alpha transparency of the xlab of the chart.}

\item{chart.ylab.color}{The color of the ylab of the chart.}

\item{chart.ylab.size}{The size of the ylab of the chart.}

\item{chart.ylab.fontface}{The font face of the ylab of the chart. See \code{graphics::par}, option 'font'.}

\item{chart.ylab.fontfamily}{The font family of the ylab of the chart. See \code{graphics::par}, option 'family'.}

\item{chart.ylab.alpha}{The alpha transparency of the ylab of the chart.}

\item{chart.text.color}{The color of the text of the chart.}

\item{chart.text.size}{The size of the text of the chart.}

\item{chart.text.fontface}{The font face of the text of the chart. See \code{graphics::par}, option 'font'.}

\item{chart.text.fontfamily}{The font family of the text of the chart. See \code{graphics::par}, option 'family'.}

\item{chart.text.alpha}{The alpha transparency of the text of the chart.}

\item{chart.frame}{The frame of the chart.}

\item{chart.frame.color}{The color of the frame of the chart.}

\item{chart.frame.alpha}{The alpha transparency of the frame of the chart.}

\item{chart.frame.lwd}{The line width of the frame of the chart. See \code{graphics::par}, option 'lwd'.}

\item{chart.frame.r}{The r (radius) of the frame of the chart.}

\item{chart.bg}{The bg of the chart.}

\item{chart.bg.color}{The color of the bg of the chart.}

\item{chart.bg.alpha}{The alpha transparency of the bg of the chart.}

\item{chart.object.color}{The color of the object of the chart.}

\item{title.size}{The size of the title.}

\item{title.color}{The color of the title.}

\item{title.fontface}{The font face of the title. See \code{graphics::par}, option 'font'.}

\item{title.fontfamily}{The font family of the title. See \code{graphics::par}, option 'family'.}

\item{title.alpha}{The alpha transparency of the title.}

\item{title.padding}{The padding of the title. A vector of 4 values: bottom, left, top, right. The unit is the device height (for bottom and top) or width (for left and right).}

\item{title.frame}{The frame of the title.}

\item{title.frame.color}{The color of the frame of the title.}

\item{title.frame.alpha}{The alpha transparency of the frame of the title.}

\item{title.frame.lwd}{The line width of the frame of the title. See \code{graphics::par}, option 'lwd'.}

\item{title.frame.r}{The r (radius) of the frame of the title.}

\item{title.position}{The position of the title. A tm_pos object, or a shortcut of two values: horizontal (left, center, right) and vertical (top, center, bottom). See tm_pos for details}

\item{title.width}{The width of the title.}

\item{credits.size}{The size of the credits.}

\item{credits.color}{The color of the credits.}

\item{credits.fontface}{The font face of the credits. See \code{graphics::par}, option 'font'.}

\item{credits.fontfamily}{The font family of the credits. See \code{graphics::par}, option 'family'.}

\item{credits.alpha}{The alpha transparency of the credits.}

\item{credits.padding}{The padding of the credits. A vector of 4 values: bottom, left, top, right. The unit is the device height (for bottom and top) or width (for left and right).}

\item{credits.position}{The position of the credits. A tm_pos object, or a shortcut of two values: horizontal (left, center, right) and vertical (top, center, bottom). See tm_pos for details}

\item{credits.width}{The width of the credits.}

\item{credits.height}{The height of the credits.}

\item{compass.north}{The north of the compass.}

\item{compass.type}{The type of the compass.}

\item{compass.text.size}{The size of the text of the compass.}

\item{compass.size}{The size of the compass.}

\item{compass.show.labels}{The labels of the show of the compass.}

\item{compass.cardinal.directions}{The directions of the cardinal of the compass.}

\item{compass.text.color}{The color of the text of the compass.}

\item{compass.color.dark}{The dark of the color of the compass.}

\item{compass.color.light}{The light of the color of the compass.}

\item{compass.lwd}{The line width of the compass. See \code{graphics::par}, option 'lwd'.}

\item{compass.margins}{The margins of the compass. A vector of 4 values: bottom, left, top, right. The unit is the device height (for bottom and top) or width (for left and right).}

\item{compass.position}{The position of the compass. A tm_pos object, or a shortcut of two values: horizontal (left, center, right) and vertical (top, center, bottom). See tm_pos for details}

\item{inset.position}{The position of the inset. A tm_pos object, or a shortcut of two values: horizontal (left, center, right) and vertical (top, center, bottom). See tm_pos for details}

\item{logo.height}{The height of the logo.}

\item{logo.margins}{The margins of the logo. A vector of 4 values: bottom, left, top, right. The unit is the device height (for bottom and top) or width (for left and right).}

\item{logo.between_margin}{The between_margin of the logo.}

\item{logo.position}{The position of the logo. A tm_pos object, or a shortcut of two values: horizontal (left, center, right) and vertical (top, center, bottom). See tm_pos for details}

\item{inset_map.height}{The height of the inset_map.}

\item{inset_map.width}{The width of the inset_map.}

\item{inset_map.margins}{The margins of the inset_map. A vector of 4 values: bottom, left, top, right. The unit is the device height (for bottom and top) or width (for left and right).}

\item{inset_map.between_margin}{The between_margin of the inset_map.}

\item{inset_map.position}{The position of the inset_map. A tm_pos object, or a shortcut of two values: horizontal (left, center, right) and vertical (top, center, bottom). See tm_pos for details}

\item{inset_map.frame}{The frame of the inset_map.}

\item{inset.height}{The height of the inset.}

\item{inset.width}{The width of the inset.}

\item{inset.margins}{The margins of the inset. A vector of 4 values: bottom, left, top, right. The unit is the device height (for bottom and top) or width (for left and right).}

\item{inset.between_margin}{The between_margin of the inset.}

\item{inset.frame}{The frame of the inset.}

\item{inset.bg}{The bg of the inset.}

\item{inset.bg.color}{The color of the bg of the inset.}

\item{inset.bg.alpha}{The alpha transparency of the bg of the inset.}

\item{inset_grob.height}{The height of the inset_grob.}

\item{inset_grob.width}{The width of the inset_grob.}

\item{inset_gg.height}{The height of the inset_gg.}

\item{inset_gg.width}{The width of the inset_gg.}

\item{scalebar.breaks}{See \code{\link[=tm_scalebar]{tm_scalebar()}}}

\item{scalebar.width}{See \code{\link[=tm_scalebar]{tm_scalebar()}}}

\item{scalebar.allow_clipping}{See \code{\link[=tm_scalebar]{tm_scalebar()}}}

\item{scalebar.text.size}{The size of the text of the scalebar.}

\item{scalebar.text.color}{The color of the text of the scalebar.}

\item{scalebar.text.fontface}{The font face of the text of the scalebar. See \code{graphics::par}, option 'font'.}

\item{scalebar.text.fontfamily}{The font family of the text of the scalebar. See \code{graphics::par}, option 'family'.}

\item{scalebar.color.dark}{The dark of the color of the scalebar.}

\item{scalebar.color.light}{The light of the color of the scalebar.}

\item{scalebar.lwd}{The line width of the scalebar. See \code{graphics::par}, option 'lwd'.}

\item{scalebar.size}{The size of the scalebar.}

\item{scalebar.margins}{The margins of the scalebar. A vector of 4 values: bottom, left, top, right. The unit is the device height (for bottom and top) or width (for left and right).}

\item{scalebar.position}{The position of the scalebar. A tm_pos object, or a shortcut of two values: horizontal (left, center, right) and vertical (top, center, bottom). See tm_pos for details}

\item{grid.show}{The visibility of the grid. \code{TRUE} or \code{FALSE}.}

\item{grid.labels.pos}{The pos of the labels of the grid.}

\item{grid.x}{The x of the grid.}

\item{grid.y}{The y of the grid.}

\item{grid.n.x}{The x of the n of the grid.}

\item{grid.n.y}{The y of the n of the grid.}

\item{grid.crs}{The coordinate reference system (CRS) of the grid.}

\item{grid.col}{The color of the grid.}

\item{grid.lwd}{The line width of the grid. See \code{graphics::par}, option 'lwd'.}

\item{grid.alpha}{The alpha transparency of the grid.}

\item{grid.labels.show}{The visibility of the labels of the grid. \code{TRUE} or \code{FALSE}.}

\item{grid.labels.size}{The size of the labels of the grid.}

\item{grid.labels.col}{The color of the labels of the grid.}

\item{grid.labels.fontface}{The font face of the labels of the grid. See \code{graphics::par}, option 'font'.}

\item{grid.labels.fontfamily}{The font family of the labels of the grid. See \code{graphics::par}, option 'family'.}

\item{grid.labels.rot}{The rot of the labels of the grid.}

\item{grid.labels.format}{The format of the labels of the grid.}

\item{grid.labels.cardinal}{The cardinal of the labels of the grid.}

\item{grid.labels.margin.x}{The x of the margin of the labels of the grid.}

\item{grid.labels.margin.y}{The y of the margin of the labels of the grid.}

\item{grid.labels.space.x}{The x of the space of the labels of the grid.}

\item{grid.labels.space.y}{The y of the space of the labels of the grid.}

\item{grid.labels.inside_frame}{The inside_frame of the labels of the grid.}

\item{grid.ticks}{The ticks of the grid.}

\item{grid.lines}{The lines of the grid.}

\item{grid.ndiscr}{The ndiscr of the grid.}

\item{mouse_coordinates.position}{The position of the mouse_coordinates. A tm_pos object, or a shortcut of two values: horizontal (left, center, right) and vertical (top, center, bottom). See tm_pos for details}

\item{minimap.server}{The server of the minimap.}

\item{minimap.toggle}{The toggle of the minimap.}

\item{minimap.position}{The position of the minimap. A tm_pos object, or a shortcut of two values: horizontal (left, center, right) and vertical (top, center, bottom). See tm_pos for details}

\item{panel.show}{The visibility of the panel. \code{TRUE} or \code{FALSE}.}

\item{panel.labels}{The labels of the panel.}

\item{panel.label.size}{The size of the label of the panel.}

\item{panel.label.color}{The color of the label of the panel.}

\item{panel.label.fontface}{The font face of the label of the panel. See \code{graphics::par}, option 'font'.}

\item{panel.label.fontfamily}{The font family of the label of the panel. See \code{graphics::par}, option 'family'.}

\item{panel.label.alpha}{The alpha transparency of the label of the panel.}

\item{panel.label.bg}{The bg of the label of the panel.}

\item{panel.label.bg.color}{The color of the bg of the label of the panel.}

\item{panel.label.bg.alpha}{The alpha transparency of the bg of the label of the panel.}

\item{panel.label.frame}{The frame of the label of the panel.}

\item{panel.label.frame.color}{The color of the frame of the label of the panel.}

\item{panel.label.frame.alpha}{The alpha transparency of the frame of the label of the panel.}

\item{panel.label.frame.lwd}{The line width of the frame of the label of the panel. See \code{graphics::par}, option 'lwd'.}

\item{panel.label.frame.r}{The r (radius) of the frame of the label of the panel.}

\item{panel.label.height}{The height of the label of the panel.}

\item{panel.label.rot}{Rotation angles of the panel labels. Vector of four values that determine the panel label rotation when they are placed left, top, right, and bottom. The default angles are 90, 0, 270 and 0 respectively. Note that the second value is the most common, since labels are by default shown on top (see \code{panel.wrap.pos}). In cross-table facets created with \code{tm_facets_grid()}, the first two values are used by default (see \code{panel.xtab.pos}).}

\item{qtm.scalebar}{The scalebar of the qtm.}

\item{qtm.minimap}{The minimap of the qtm.}

\item{qtm.mouse_coordinates}{The mouse_coordinates of the qtm.}

\item{earth_boundary}{The earth boundary}

\item{earth_boundary.color}{The color of the earth_boundary.}

\item{earth_boundary.lwd}{The line width of the earth_boundary. See \code{graphics::par}, option 'lwd'.}

\item{earth_datum}{Earth datum}

\item{space}{Should the space be drawn? Only applicable is earth_boundary is enabled.}

\item{space.color}{The color of the space.}

\item{space_overlay}{Should the space be drawn as overlay (to make sure spatial features or rasters do not exceed the earth boundary), or as background? By default \code{TRUE} when a raster is warped.}

\item{check_and_fix}{Should attempt to fix an invalid shapefile}

\item{basemap.show}{The visibility of the basemap. \code{TRUE} or \code{FALSE}.}

\item{basemap.server}{The server of the basemap.}

\item{basemap.alpha}{The alpha transparency of the basemap.}

\item{basemap.zoom}{The zoom of the basemap.}

\item{tiles.show}{The visibility of the tiles. \code{TRUE} or \code{FALSE}.}

\item{tiles.server}{The server of the tiles.}

\item{tiles.alpha}{The alpha transparency of the tiles.}

\item{tiles.zoom}{The zoom of the tiles.}

\item{attr.color}{The color of the attr.}

\item{crs_extra}{Only used internally (work in progress)}

\item{crs_global}{The used crs for world maps}

\item{crs_basemap}{The crs_basemap (overall).}

\item{title}{deprecated See \code{\link[=tm_title]{tm_title()}}}

\item{main.title}{deprecated See \code{\link[=tm_title]{tm_title()}}}

\item{main.title.size, main.title.color, main.title.fontface, main.title.fontfamily, main.title.position}{deprecated. Use the \code{title.} options instead.}

\item{fontface, fontfamily}{renamed to \code{text.fontface} and \code{text.fontfamily}}

\item{style}{style see \code{tm_style()}}

\item{...}{List of tmap options to be set, or option names (characters) to be returned (see details)}
}
\description{
tmap options
}
\seealso{
\href{https://r-tmap.github.io/tmap/articles/basics_layout}{Layout}, \href{https://r-tmap.github.io/tmap/articles/adv_margins}{Margins and aspect ratio}, \href{https://r-tmap.github.io/tmap/articles/adv_options}{Options and styles}, \href{https://r-tmap.github.io/tmap/articles/examples_choro_World}{Choropleth (World)}, \href{https://r-tmap.github.io/tmap/articles/examples_terrain}{Terrain map}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tm_chart.R
\name{tm_chart}
\alias{tm_chart}
\alias{tm_chart_histogram}
\alias{tm_chart_bar}
\alias{tm_chart_donut}
\alias{tm_chart_violin}
\alias{tm_chart_box}
\alias{tm_chart_none}
\alias{tm_chart_heatmap}
\title{Legend charts}
\usage{
tm_chart_histogram(
  breaks,
  plot.axis.x,
  plot.axis.y,
  extra.ggplot2,
  position,
  group_id,
  width,
  height,
  stack,
  z,
  ...
)

tm_chart_bar(
  plot.axis.x,
  plot.axis.y,
  extra.ggplot2,
  position,
  group_id,
  width,
  height,
  stack,
  z,
  ...
)

tm_chart_donut(position, group_id, width, height, stack, z, ...)

tm_chart_violin(position, group_id, width, height, stack, z, ...)

tm_chart_box(position, group_id, width, height, stack, z, ...)

tm_chart_none()

tm_chart_heatmap(position, group_id, width, height, stack, z, ...)
}
\arguments{
\item{breaks}{The breaks of the bins (for histograms)}

\item{plot.axis.x, plot.axis.y}{Should the x axis and y axis be plot?}

\item{extra.ggplot2}{Extra ggplot2 code}

\item{position}{The position specification of the component: an object created with \code{tm_pos_in()} or \code{tm_pos_out()}. Or, as a shortcut, a vector of two values, specifying the x and y coordinates. The first is \code{"left"}, \code{"center"} or \code{"right"} (or upper case, meaning tighter to the map frame), the second \code{"top"}, \code{"center"} or \code{"bottom"}. Numeric values are also supported, where 0, 0 means left bottom and 1, 1 right top. See also \href{https://r-tmap.github.io/tmap/articles/adv_positions}{vignette about positioning}. In case multiple components should be combined (stacked), use \code{group_id} and specify \code{component} in \code{\link[=tm_components]{tm_components()}}.}

\item{group_id}{Component group id name. All components (e.g. legends, titles, etc) with the same \code{group_id} will be grouped. The specifications of how they are placed (e.g. stacking, margins etc.) are determined in \code{\link[=tm_components]{tm_components()}} where its argument \code{id} should correspond to \code{group_id}.}

\item{width, height}{width and height of the component.}

\item{stack}{stack with other map components, either \code{"vertical"} or \code{"horizontal"}.}

\item{z}{z index, e.g. the place of the component relative to the other componets}

\item{...}{passed on to \code{tm_title()}}
}
\description{
Legend charts are small charts that are added to the map, usually in addition to legends.
}
\details{
Note that these charts are different from charts drawn inside the map. Those are called glyphs (to be implemented).
}
\examples{
tm_shape(World) +
  tm_polygons("HPI",
    fill.scale = tm_scale_intervals(),
    fill.chart = tm_chart_histogram())
}
\seealso{
\href{https://r-tmap.github.io/tmap/articles/basics_charts}{Vignette about charts}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tmapShape.R
\name{tmapSubsetShp}
\alias{tmapSubsetShp}
\title{Internal method that subsets data from shape objects}
\usage{
tmapSubsetShp(shp, vars)
}
\arguments{
\item{shp}{shape}

\item{vars}{a vector of variable names}
}
\description{
Internal method that subsets data from shape objects
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data.R
\docType{data}
\name{metro}
\alias{metro}
\title{Spatial data of metropolitan areas}
\format{
An object of class \code{sf} (inherits from \code{data.frame}) with 436 rows and 13 columns.
}
\source{
\url{https://population.un.org/wup/}
}
\usage{
metro
}
\description{
\code{metro} includes a population time series from 1950 to (forecasted) 2030. All metro areas with over 1 million inhabitants in 2010 are included.
}
\references{
United Nations, Department of Economic and Social Affairs, Population Division (2014). World Urbanization Prospects: The 2014 Revision, CD-ROM Edition.
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tm_label_format.R
\name{tm_label_format}
\alias{tm_label_format}
\title{tmap function to specify labels}
\usage{
tm_label_format(
  fun,
  scientific,
  format,
  digits,
  interval.disjoint,
  big.num.abbr,
  prefix,
  suffix,
  text.separator,
  text.less.than,
  text.less.than_as.prefix,
  text.or.more,
  text.or.more_as.prefix,
  text.align,
  text.to.columns,
  html.escape,
  ...
)
}
\arguments{
\item{fun}{Function to specify the labels. It should take a numeric vector, and should return a character vector of the same size. By default it is not specified. If specified, the list items \code{scientific}, \code{format}, and \code{digits} (see below) are not used.}

\item{scientific}{Should the labels be formatted scientifically? If so, square brackets are used, and the \code{format} of the numbers is \code{"g"}. Otherwise, \code{format="f"}, and \code{text.separator}, \code{text.less.than}, and \code{text.or.more} are used. Also, the numbers are automatically  rounded to millions or billions if applicable. By default, \code{FALSE}}

\item{format}{By default, \code{"f"}, i.e. the standard notation \code{xxx.xxx}, is used. If \code{scientific = TRUE} then \code{"g"}, which means that numbers are formatted scientifically, i.e. \code{n.dddE+nn} if needed to save space.}

\item{digits}{Number of digits after the decimal point if \code{format="f"}, and the number of significant digits otherwise. By default \code{NA}, meaning as many as needed to have distinct numbers}

\item{interval.disjoint}{In case of intervals (see \code{\link[=tm_scale_intervals]{tm_scale_intervals()}}), should the intervals appear disjoint, e.g. \code{0 to 999}, \code{1000 - 1999}, \code{2000 - 2999} (\code{TRUE}, default), or not: \code{0 - 1000}, \code{1000 - 2000}, \code{2000- 3000}.}

\item{big.num.abbr}{Vector that defines whether and which abbrevations are used for large numbers. It is a named numeric vector, where the name indicated the abbreviation, and the number the magnitude (in terms on numbers of zero). Numbers are only abbrevation when they are large enough. Set it to \code{NA} to disable abbrevations.  The default is \code{c("mln" = 6, "bln" = 9)}. For layers where \code{style} is set to \code{log10} or \code{log10_pretty}, the default is \code{NA}.}

\item{prefix}{Prefix of each number}

\item{suffix}{Suffix of each number}

\item{text.separator}{Character string to use to separate numbers in an interval legend (default: \code{"to"}).}

\item{text.less.than}{Character value(s) to use for 'less than'. Default \code{"Less than"}. When a character vector of length 2 is specified, one for each word, these words are aligned when \code{text.to.columns = TRUE}}

\item{text.less.than_as.prefix}{Should \code{text.less.than} be used as prefix?}

\item{text.or.more}{Character value(s) to use to  'or more'. Default is \code{"or more"}. When a character vector of length 2 is specified, one for each word, these words are aligned when \code{text.to.columns = TRUE}}

\item{text.or.more_as.prefix}{Should \code{text.or.more} be used as prefix?}

\item{text.align}{Not implemented in v4 (yet). Value that determines how the numbers are aligned, \code{"left"}, \code{"center"} or \code{"right"}. By default \code{"left"}.}

\item{text.to.columns}{Not implemented in v4 (yet). Logical that determines whether the text is aligned to three columns (from, text.separator, to). By default \code{FALSE}.}

\item{html.escape}{Logical that determins whther HTML code is escaped in the popups in view mode. By default \code{TRUE}. If set to \code{FALSE} HTML code can be added, e.g. to added white space via \code{&nbsp;}.}

\item{...}{arguments passed on to \code{\link[base:formatC]{formatC}}}
}
\value{
list with formatting options
}
\description{
tmap function to specify labels used in the scale functions, e.g. via the argument \code{label.format} in \code{\link[=tm_scale_intervals]{tm_scale_intervals()}}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tm_chart.R
\name{tm_chart}
\alias{tm_chart}
\alias{tm_chart_histogram}
\alias{tm_chart_bar}
\alias{tm_chart_donut}
\alias{tm_chart_violin}
\alias{tm_chart_box}
\alias{tm_chart_none}
\alias{tm_chart_heatmap}
\title{Legend charts}
\usage{
tm_chart_histogram(
  breaks,
  plot.axis.x,
  plot.axis.y,
  extra.ggplot2,
  position,
  group_id,
  width,
  height,
  stack,
  z,
  ...
)

tm_chart_bar(
  plot.axis.x,
  plot.axis.y,
  extra.ggplot2,
  position,
  group_id,
  width,
  height,
  stack,
  z,
  ...
)

tm_chart_donut(position, group_id, width, height, stack, z, ...)

tm_chart_violin(position, group_id, width, height, stack, z, ...)

tm_chart_box(position, group_id, width, height, stack, z, ...)

tm_chart_none()

tm_chart_heatmap(position, group_id, width, height, stack, z, ...)
}
\arguments{
\item{breaks}{The breaks of the bins (for histograms)}

\item{plot.axis.x, plot.axis.y}{Should the x axis and y axis be plot?}

\item{extra.ggplot2}{Extra ggplot2 code}

\item{position}{The position specification of the component: an object created with \code{tm_pos_in()} or \code{tm_pos_out()}. Or, as a shortcut, a vector of two values, specifying the x and y coordinates. The first is \code{"left"}, \code{"center"} or \code{"right"} (or upper case, meaning tighter to the map frame), the second \code{"top"}, \code{"center"} or \code{"bottom"}. Numeric values are also supported, where 0, 0 means left bottom and 1, 1 right top. See also vignette: \href{https://r-tmap.github.io/tmap/articles/adv_positions}{Positioning of components}. In case multiple components should be combined (stacked), use \code{group_id} and specify \code{component} in \code{\link[=tm_components]{tm_components()}}.}

\item{group_id}{Component group id name. All components (e.g. legends, titles, etc) with the same \code{group_id} will be grouped. The specifications of how they are placed (e.g. stacking, margins etc.) are determined in \code{\link[=tm_components]{tm_components()}} where its argument \code{id} should correspond to \code{group_id}.}

\item{width, height}{width and height of the component.}

\item{stack}{stack with other map components, either \code{"vertical"} or \code{"horizontal"}.}

\item{z}{z index, e.g. the place of the component relative to the other componets}

\item{...}{passed on to \code{tm_title()}}
}
\description{
Legend charts are small charts that are added to the map, usually in addition to legends.
}
\details{
Note that these charts are different from charts drawn inside the map. Those are called glyphs (to be implemented).
}
\examples{
tm_shape(World) +
  tm_polygons("HPI",
    fill.scale = tm_scale_intervals(),
    fill.chart = tm_chart_histogram())
}
\seealso{
\href{https://r-tmap.github.io/tmap/articles/basics_charts}{Charts}, \href{https://r-tmap.github.io/tmap/articles/basics_components}{Components}, \href{https://r-tmap.github.io/tmap/articles/adv_positions}{Positioning of components}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tm_label_format.R
\name{tm_label_format}
\alias{tm_label_format}
\title{tmap function to specify labels}
\usage{
tm_label_format(
  fun,
  scientific,
  format,
  digits,
  interval.disjoint,
  big.num.abbr,
  prefix,
  suffix,
  text.separator,
  text.less.than,
  text.less.than_as.prefix,
  text.or.more,
  text.or.more_as.prefix,
  text.align,
  text.to.columns,
  html.escape,
  ...
)
}
\arguments{
\item{fun}{Function to specify the labels. It should take a numeric vector, and should return a character vector of the same size. By default it is not specified. If specified, the list items \code{scientific}, \code{format}, and \code{digits} (see below) are not used.}

\item{scientific}{Should the labels be formatted scientifically? If so, square brackets are used, and the \code{format} of the numbers is \code{"g"}. Otherwise, \code{format="f"}, and \code{text.separator}, \code{text.less.than}, and \code{text.or.more} are used. Also, the numbers are automatically  rounded to millions or billions if applicable. By default, \code{FALSE}}

\item{format}{By default, \code{"f"}, i.e. the standard notation \code{xxx.xxx}, is used. If \code{scientific = TRUE} then \code{"g"}, which means that numbers are formatted scientifically, i.e. \code{n.dddE+nn} if needed to save space.}

\item{digits}{Number of digits after the decimal point if \code{format="f"}, and the number of significant digits otherwise. By default \code{NA}, meaning as many as needed to have distinct numbers}

\item{interval.disjoint}{In case of intervals (see \code{\link[=tm_scale_intervals]{tm_scale_intervals()}}), should the intervals appear disjoint, e.g. \code{0 to 999}, \code{1000 - 1999}, \code{2000 - 2999} (\code{TRUE}, default), or not: \code{0 - 1000}, \code{1000 - 2000}, \code{2000- 3000}.}

\item{big.num.abbr}{Vector that defines whether and which abbrevations are used for large numbers. It is a named numeric vector, where the name indicated the abbreviation, and the number the magnitude (in terms on numbers of zero). Numbers are only abbrevation when they are large enough. Set it to \code{NA} to disable abbrevations.  The default is \code{c("mln" = 6, "bln" = 9)}. For layers where \code{style} is set to \code{log10} or \code{log10_pretty}, the default is \code{NA}.}

\item{prefix}{Prefix of each number}

\item{suffix}{Suffix of each number}

\item{text.separator}{Character string to use to separate numbers in an interval legend (default: \code{"to"}).}

\item{text.less.than}{Character value(s) to use for 'less than'. Default \code{"Less than"}. When a character vector of length 2 is specified, one for each word, these words are aligned when \code{text.to.columns = TRUE}}

\item{text.less.than_as.prefix}{Should \code{text.less.than} be used as prefix?}

\item{text.or.more}{Character value(s) to use to  'or more'. Default is \code{"or more"}. When a character vector of length 2 is specified, one for each word, these words are aligned when \code{text.to.columns = TRUE}}

\item{text.or.more_as.prefix}{Should \code{text.or.more} be used as prefix?}

\item{text.align}{Not implemented in v4 (yet). Value that determines how the numbers are aligned, \code{"left"}, \code{"center"} or \code{"right"}. By default \code{"left"}.}

\item{text.to.columns}{Not implemented in v4 (yet). Logical that determines whether the text is aligned to three columns (from, text.separator, to). By default \code{FALSE}.}

\item{html.escape}{Logical that determins whther HTML code is escaped in the popups in view mode. By default \code{TRUE}. If set to \code{FALSE} HTML code can be added, e.g. to added white space via \code{&nbsp;}.}

\item{...}{arguments passed on to \code{\link[base:formatC]{formatC}}}
}
\value{
list with formatting options
}
\description{
tmap function to specify labels used in the scale functions, e.g. via the argument \code{label.format} in \code{\link[=tm_scale_intervals]{tm_scale_intervals()}}.
}
\seealso{
\href{https://r-tmap.github.io/tmap/articles/adv_legends}{Legends (advanced)}, \href{https://r-tmap.github.io/tmap/articles/adv_positions}{Positioning of components}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tmap_export.R
\name{tmap_leaflet}
\alias{tmap_leaflet}
\alias{tmap_grob}
\title{Export tmap to the format of the used graphics mode}
\usage{
tmap_leaflet(x, show = FALSE, ...)

tmap_grob(x, asp = NA, scale = 1, show = FALSE, ...)
}
\arguments{
\item{x}{a tmap object.}

\item{show}{show the map?}

\item{...}{
  Arguments passed on to \code{\link[=print.tmap]{print.tmap}}
  \describe{
    \item{\code{return.asp}}{should the aspect ratio be returned?}
    \item{\code{vp}}{viewport (for \code{"plot"} mode)}
    \item{\code{knit}}{A logical, should knit?}
    \item{\code{in.shiny}}{A logical, is the map drawn in \strong{shiny}?}
    \item{\code{proxy}}{A logical, if \code{in.shiny}, is \code{\link[=tmapProxy]{tmapProxy()}} used?}
    \item{\code{options}}{A vector of options}
  }}

\item{asp, scale}{the desired aspect ratio and scale of the map. Only applicable for \code{"plot"} mode.}
}
\value{
\itemize{
\item \code{tmap_grob()} returns a \code{\link[grid:grid.grob]{grob}} object (\code{"plot"} mode)
\item \code{tmap_leaflet()} a \code{\link[leaflet:leaflet]{leaflet}} object (\code{"view"} mode).
In case small multiples are shown, a list is returned.
}
}
\description{
\itemize{
\item \code{tmap_grob()} returns a \code{\link[grid:grid.grob]{grob}} object (\verb{"plot" mode})
\item \code{tmap_leaflet()} a \code{\link[leaflet:leaflet]{leaflet}} object (\code{"view"} mode).
}
}
\examples{
map = tm_shape(World) + tm_polygons()
tmap_leaflet(map, show = TRUE)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tm_layers_raster.R
\name{tm_raster}
\alias{tm_raster}
\alias{opt_tm_raster}
\title{Map layer: raster}
\usage{
tm_raster(
  col = tm_vars(),
  col.scale = tm_scale(),
  col.legend = tm_legend(),
  col.chart = tm_chart_none(),
  col.free = NA,
  col_alpha = tm_const(),
  col_alpha.scale = tm_scale(),
  col_alpha.legend = tm_legend(),
  col_alpha.chart = tm_chart_none(),
  col_alpha.free = NA,
  zindex = NA,
  group = NA,
  group.control = "check",
  options = opt_tm_raster(),
  ...
)

opt_tm_raster(interpolate = FALSE)
}
\arguments{
\item{col, col.scale, col.legend, col.chart, col.free}{Visual variable that determines the color. See details.}

\item{col_alpha, col_alpha.scale, col_alpha.legend, col_alpha.chart, col_alpha.free}{Visual variable that determines the color transparency. See details.}

\item{zindex}{Map layers are drawn on top of each other. The \code{zindex} numbers
(one for each map layer) determines the stacking order.
By default the map layers are drawn in the order they are called.}

\item{group}{Name of the group to which this layer belongs. This is only
relevant in view mode, where layer groups can be switched (see \code{group.control})}

\item{group.control}{In view mode, the group control determines how layer groups
can be switched on and off. Options: \code{"radio"} for radio buttons
(meaning only one group can be shown), \code{"check"} for check boxes
(so multiple groups can be shown), and \code{"none"} for no control
(the group cannot be (de)selected).}

\item{options}{options passed on to the corresponding \verb{opt_<layer_function>} function}

\item{...}{to catch deprecated arguments from version < 4.0}

\item{interpolate}{Should the raster image be interpolated? Currently only applicable in view mode (passed on to \code{\link[grid:grid.raster]{grid}})}
}
\description{
Map layer that draws rasters. Supported visual variable is: \code{col} (the  color).
}
\details{
The visual variable arguments (e.g. \code{col}) can be specified with a data
variable name (e.g., a spatial vector attribute or a raster layer of the object
specified in \code{\link[=tm_shape]{tm_shape()}}), with a visual value (for \code{col}, a color is expected), or with a geometry-derived variable (see below).
See \href{https://r-tmap.github.io/tmap/articles/basics_vv}{vignette about visual variables}.

Multiple values can be specified: in that case facets are created.
These facets can be combined with other faceting data variables, specified with \code{\link[=tm_facets]{tm_facets()}}.
See \href{https://r-tmap.github.io/tmap/articles/basics_facets}{vignette about facets}.
\itemize{
\item The \verb{*.scale} arguments determine the used scale to map the data values to
visual variable values. These can be specified with one of the available
\verb{tm_scale_*()} functions. The default is specified by the tmap option (\code{\link[=tm_options]{tm_options()}}) \code{scales.var}.
See \href{https://r-tmap.github.io/tmap/articles/basics_scales}{vignette about scales}.
\item The \verb{*.legend} arguments determine the used legend, specified with \code{\link[=tm_legend]{tm_legend()}}.
The default legend and its settings are determined by the tmap options (\code{\link[=tm_options]{tm_options()}}) \code{legend.} .
See \href{https://r-tmap.github.io/tmap/articles/basics_legends}{vignette about legends}.
\item The \verb{*.chart} arguments specify additional charts, specified with \code{tm_chart_}, e.g. \code{\link[=tm_chart_histogram]{tm_chart_histogram()}}.
See \href{https://r-tmap.github.io/tmap/articles/basics_charts}{vignette about charts}.
\item The \verb{*.free} arguments determine whether scales are applied freely across facets, or shared.
A logical value is required. They can also be specified with a vector of three
logical values; these determine whether scales are applied freely per facet dimension.
This is only useful when facets are applied (see \code{\link[=tm_facets]{tm_facets()}}).
There are maximally three facet dimensions: rows, columns, and pages. This only
applies for a facet grid (\code{\link[=tm_facets_grid]{tm_facets_grid()}}). For instance, \code{col.free = c(TRUE, FALSE, FALSE)}
means that for the visual variable \code{col}, each row of facets will have its own
scale, and therefore its own legend. For facet wraps and stacks
(\code{\link[=tm_facets_wrap]{tm_facets_wrap()}} and \code{\link[=tm_facets_stack]{tm_facets_stack()}}) there is only one facet dimension,
so the \verb{*.free} argument requires only one logical value.
}

Currently, three geometry-derived variables are implemented:
\itemize{
\item \code{"AREA"} (polygons only), which uses the feature area;
\item \code{"LENGTH"} (lines only), which uses the feature length; and
\item \code{"MAP_COLORS"}, which assigns values so that adjacent features receive
different values, making it particularly suitable for coloring
neighbouring polygons.
}

Note that geometry-derived variables do not generate a legend automatically.
If a legend is required, compute the corresponding variable explicitly,
for example with \code{\link[sf:geos_measures]{sf::st_area()}}, \code{\link[sf:geos_measures]{sf::st_length()}}, or
\code{\link[tmaptools:map_coloring]{tmaptools::map_coloring()}}, and use the resulting values instead.
}
\examples{
\dontrun{
# load land data
data(land, World)

tm_shape(land) +
	tm_raster("cover")

tm_shape(land) +
	tm_raster("elevation", col.scale = tm_scale_continuous(values = terrain.colors(9))) +
	tm_shape(World) +
	tm_borders()
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tmapShape.R
\name{tmapGetShapeMeta1}
\alias{tmapGetShapeMeta1}
\title{Internal method that extracts meta data from shape objects}
\usage{
tmapGetShapeMeta1(shp, layer, o)
}
\arguments{
\item{shp}{the shape object}

\item{layer}{(optional) the layer of the shape object}

\item{o}{the list of options}
}
\description{
Internal method that extracts meta data from shape objects
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tmap_export.R
\name{tmap_leaflet}
\alias{tmap_leaflet}
\alias{tmap_grob}
\title{Export tmap to the format of the used graphics mode}
\usage{
tmap_leaflet(x, show = FALSE, ...)

tmap_grob(x, asp = NA, scale = NA, show = FALSE, ...)
}
\arguments{
\item{x}{a tmap object.}

\item{show}{show the map?}

\item{...}{
  Arguments passed on to \code{\link{print.tmap}}
  \describe{
    \item{\code{return.asp}}{should the aspect ratio be returned?}
    \item{\code{vp}}{viewport (for \code{"plot"} mode)}
    \item{\code{knit}}{A logical, should knit?}
    \item{\code{in.shiny}}{A logical, is the map drawn in \strong{shiny}?}
    \item{\code{proxy}}{A logical, if \code{in.shiny}, is \code{\link[=tmapProxy]{tmapProxy()}} used?}
    \item{\code{options}}{A vector of options}
  }}

\item{asp, scale}{the desired aspect ratio and scale of the map. Only applicable for \code{"plot"} mode.}
}
\value{
\itemize{
\item \code{tmap_grob()} returns a \code{\link[grid:grob]{grob}} object (\code{"plot"} mode)
\item \code{tmap_leaflet()} a \code{\link[leaflet:leaflet]{leaflet}} object (\code{"view"} mode).
In case small multiples are shown, a list is returned.
}
}
\description{
\itemize{
\item \code{tmap_grob()} returns a \code{\link[grid:grob]{grob}} object (\verb{"plot" mode})
\item \code{tmap_leaflet()} a \code{\link[leaflet:leaflet]{leaflet}} object (\code{"view"} mode).
}
}
\examples{
map = tm_shape(World) + tm_polygons()
tmap_leaflet(map, show = TRUE)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tm_group.R
\name{tm_group}
\alias{tm_group}
\title{Layer group control}
\usage{
tm_group(name, control = NA, zoom_levels = NA)
}
\arguments{
\item{name}{group name that corresponds with the group name specified in the layer functions (e.g.  \code{\link[=tm_polygons]{tm_polygons()}})}

\item{control}{The group control determines how
layer groups can be switched on and off. Options: \code{"radio"} for radio
buttons (meaning only one group can be shown), \code{"check"} for check boxes
(so multiple groups can be shown), and \code{"none"} for no control
(the group cannot be (de)selected).}

\item{zoom_levels}{The zoom levels at which the group is displays at. When specified \code{control} will be set to \code{"none"}.}
}
\description{
Controls the layer groups in interactive maps (view mode): the layer control box (radio buttons or check boxes) and at which zoom levels the layers are displayed at.
}
\seealso{
\href{https://r-tmap.github.io/tmap/articles/adv_groups}{vignette about layer groups}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tm_layers_raster.R
\name{tm_raster}
\alias{tm_raster}
\alias{opt_tm_raster}
\title{Map layer: raster}
\usage{
tm_raster(
  col = tm_vars(),
  col.scale = tm_scale(),
  col.legend = tm_legend(),
  col.chart = tm_chart_none(),
  col.free = NA,
  col_alpha = tm_const(),
  col_alpha.scale = tm_scale(),
  col_alpha.legend = tm_legend(),
  col_alpha.chart = tm_chart_none(),
  col_alpha.free = NA,
  zindex = NA,
  group = NA,
  group.control = "check",
  blend = "over",
  options = opt_tm_raster(),
  ...
)

opt_tm_raster(interpolate = FALSE)
}
\arguments{
\item{col, col.scale, col.legend, col.chart, col.free}{Visual variable that determines the color. See details. \emph{Unit:} Color -- a color name, hex string.}

\item{col_alpha, col_alpha.scale, col_alpha.legend, col_alpha.chart, col_alpha.free}{Visual variable that determines the color transparency. See details. \emph{Unit:} Proportion -- numeric 0-1 (0 = fully transparent, 1 = fully opaque).}

\item{zindex}{Controls the stacking order of map layers. Should be set to a
value above 400. By default, layers are stacked in call order, starting
at 401. See details.}

\item{group}{Name of the group to which this layer belongs. This is only
relevant in view mode, where layer groups can be switched (see \code{group.control})}

\item{group.control}{In view mode, the group control determines how layer groups
can be switched on and off. Options: \code{"radio"} for radio buttons
(meaning only one group can be shown), \code{"check"} for check boxes
(so multiple groups can be shown), and \code{"none"} for no control
(the group cannot be (de)selected).}

\item{blend}{Compositing operator for layer blending. Default \code{"over"} applies
no blending. See the "Layer blending" section for the supported values.}

\item{options}{options passed on to the corresponding \verb{opt_<layer_function>} function}

\item{...}{to catch deprecated arguments from version < 4.0}

\item{interpolate}{Should the raster image be interpolated? Currently only applicable in view mode (passed on to \code{\link[grid:rasterGrob]{grid}})}
}
\description{
Map layer that draws rasters. Supported visual variable is: \code{col} (the  color).
}
\details{
The visual variable arguments (e.g. \code{col}) can be specified with a data
variable name (e.g., a spatial vector attribute or a raster layer of the object
specified in \code{\link[=tm_shape]{tm_shape()}}), with a visual value (for \code{col}, a color is expected), or with a geometry-derived variable (see below).
See vignette: \href{https://r-tmap.github.io/tmap/articles/basics_vv}{Visual variables}.

Multiple values can be specified: in that case facets are created.
These facets can be combined with other faceting data variables, specified with \code{\link[=tm_facets]{tm_facets()}}.
See vignette: \href{https://r-tmap.github.io/tmap/articles/basics_facets}{Facets}.
\itemize{
\item The \verb{*.scale} arguments determine the used scale to map the data values to
visual variable values. These can be specified with one of the available
\verb{tm_scale_*()} functions. The default is specified by the tmap option (\code{\link[=tm_options]{tm_options()}}) \code{scales.var}.
See `rvignette: \href{https://r-tmap.github.io/tmap/articles/basics_scales}{Scales}
\item The \verb{*.legend} arguments determine the used legend, specified with \code{\link[=tm_legend]{tm_legend()}}.
The default legend and its settings are determined by the tmap options (\code{\link[=tm_options]{tm_options()}}) \code{legend.} .
See `rvignette: \href{https://r-tmap.github.io/tmap/articles/basics_legends}{Legends}
\item The \verb{*.chart} arguments specify additional charts, specified with \code{tm_chart_}, e.g. \code{\link[=tm_chart_histogram]{tm_chart_histogram()}}.
See `rvignette: \href{https://r-tmap.github.io/tmap/articles/basics_charts}{Charts}
\item The \verb{*.free} arguments determine whether scales are applied freely across facets, or shared.
A logical value is required. They can also be specified with a vector of three
logical values; these determine whether scales are applied freely per facet dimension.
This is only useful when facets are applied (see \code{\link[=tm_facets]{tm_facets()}}).
There are maximally three facet dimensions: rows, columns, and pages. This only
applies for a facet grid (\code{\link[=tm_facets_grid]{tm_facets_grid()}}). For instance, \code{col.free = c(TRUE, FALSE, FALSE)}
means that for the visual variable \code{col}, each row of facets will have its own
scale, and therefore its own legend. For facet wraps and stacks
(\code{\link[=tm_facets_wrap]{tm_facets_wrap()}} and \code{\link[=tm_facets_stack]{tm_facets_stack()}}) there is only one facet dimension,
so the \verb{*.free} argument requires only one logical value.
}

Currently, three geometry-derived variables are implemented:
\itemize{
\item \code{"AREA"} (polygons only), which uses the feature area;
\item \code{"LENGTH"} (lines only), which uses the feature length; and
\item \code{"MAP_COLORS"}, which assigns values so that adjacent features receive
different values, making it particularly suitable for coloring
neighbouring polygons.
}

Note that geometry-derived variables do not generate a legend automatically.
If a legend is required, compute the corresponding variable explicitly,
for example with \code{\link[sf:st_area]{sf::st_area()}}, \code{\link[sf:st_length]{sf::st_length()}}, or
\code{\link[tmaptools:map_coloring]{tmaptools::map_coloring()}}, and use the resulting values instead.
\subsection{Visual variable units}{

Every visual variable maps data values to a specific output unit.
Knowing the unit matters when supplying constant values via \code{\link[=tm_const]{tm_const()}},
or output ranges via \code{values.range} / \code{values.scale} in the scale
functions.\tabular{lll}{
   Variable \tab Output unit \tab Notes \cr
   \code{fill}, \code{col}, \code{bgcol} \tab color \tab name, hex, or palette string \cr
   \code{fill_alpha}, \code{col_alpha}, \code{bgcol_alpha} \tab proportion 0-1 \tab 0 = transparent, 1 = opaque \cr
   \code{size} (symbols, bubbles, squares, dots) \tab typographic lines \tab 1 line approx. 1/6 inch; scaled by \code{values.scale} \cr
   \code{size} (circles) \tab meters \tab plain numeric or a \code{units} object \cr
   \code{size} (text, labels) \tab multiplier \tab 1 = 12 pt (plot) / 12 px (view) \cr
   \code{lwd} \tab lwd \tab base R units; 1 lwd approx. 0.75 pt at 96 dpi \cr
   \code{lty} \tab -- \tab integer 1-6 or name ("solid", "dashed", ...) \cr
   \code{shape} \tab -- \tab integer \code{pch} 1-25 or single character \cr
   \code{angle} \tab degrees \tab 0-360, clockwise from north \cr
   \code{fontface} \tab -- \tab "plain", "bold", "italic", "bold.italic" \cr
}

\subsection{Symbol size (\code{size} in \code{tm_symbols}, \code{tm_bubbles}, \code{tm_squares}, \code{tm_dots})}{

"Lines" is a typographic unit: one line is approximately 1/6 inch (the
default base line-height in R graphics). The global multiplier
\code{tmap_options(values.scale = list(size.bubbles = 1.5))} scales all symbol
sizes without changing the data mapping.
}

\subsection{Circle size (\code{size} in \code{tm_circles})}{

The value is a geographic radius in meters. A plain numeric vector is
interpreted as meters; a \code{units} object (from the \strong{units} package) is
automatically converted, so \code{units::as_units(1, "mi")} gives a 1-mile
radius. Because the radius is geographic, circles scale with zoom in
interactive (view) mode -- unlike bubble symbols which keep a fixed screen
size.
}

\subsection{Text size (\code{size} in \code{tm_text}, \code{tm_labels})}{

The value is a multiplier of the base font size. \code{size = 1} renders at
12 pt in plot mode (R's default \code{par("ps")}) and at 12 px in view
mode (\code{gp$cex * 12} px, see \code{tmapLeafletDataPlot.tm_data_text}); the two
modes are consistent by design.
}

}

\subsection{Layer blending (\code{blend})}{

Blend modes control how a layer's pixels are combined with the pixels
beneath it. For each pixel, let \eqn{S} be the source (top layer) RGB
value and \eqn{D} be the destination (bottom layer) RGB value, both
normalised to \eqn{[0, 1]}.\tabular{lll}{
   \code{blend} \tab Formula \tab Use case \cr
   \code{"over"} \tab \eqn{S \cdot \alpha + D \cdot (1 - \alpha)} \tab Standard alpha compositing (default) \cr
   \code{"multiply"} \tab \eqn{S \times D} \tab Hillshading over colour raster; both layers darken each other \cr
   \code{"screen"} \tab \eqn{1 - (1 - S)(1 - D)} \tab Inverse of multiply; brightens \cr
   \code{"overlay"} \tab multiply if \eqn{D < 0.5}, screen if \eqn{D \geq 0.5} \tab Boosts contrast; preserves midtones \cr
   \code{"darken"} \tab \eqn{\min(S, D)} \tab Keeps the darker of the two layers per channel \cr
   \code{"lighten"} \tab \eqn{\max(S, D)} \tab Keeps the lighter of the two layers per channel \cr
}


Requires R >= 4.2 and a compatible graphics device (e.g.
\code{png(type = "cairo")}, \code{svg()}). In view mode, blending is applied via
CSS \code{mix-blend-mode}. See \code{\link[grid:groupGrob]{grid::groupGrob()}} for the full list of
supported operators.
}

\subsection{zindex and pane names}{

In view mode, each layer is rendered in a Leaflet pane named \code{"tmap{zindex}"}
(e.g., \code{"tmap401"}, \code{"tmap402"}), with base tile layers placed in the
standard \code{"tile"} pane.
}
}
\examples{
\dontrun{
# load land data
data(land, World)

tm_shape(land) +
	tm_raster("cover")

tm_shape(land) +
	tm_raster("elevation", col.scale = tm_scale_continuous(values = terrain.colors(9))) +
	tm_shape(World) +
	tm_borders()
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tm_group.R
\name{tm_group}
\alias{tm_group}
\title{Layer group control}
\usage{
tm_group(name, control = NA, zoom_levels = NA)
}
\arguments{
\item{name}{group name that corresponds with the group name specified in the layer functions (e.g.  \code{\link[=tm_polygons]{tm_polygons()}})}

\item{control}{The group control determines how
layer groups can be switched on and off. Options: \code{"radio"} for radio
buttons (meaning only one group can be shown), \code{"check"} for check boxes
(so multiple groups can be shown), and \code{"none"} for no control
(the group cannot be (de)selected).}

\item{zoom_levels}{The zoom levels at which the group is displays at. When specified \code{control} will be set to \code{"none"}.}
}
\description{
Controls the layer groups in interactive maps (view mode): the layer control box (radio buttons or check boxes) and at which zoom levels the layers are displayed at.
}
\seealso{
\href{https://r-tmap.github.io/tmap/articles/adv_groups}{Layer groups (view mode)}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data.R
\docType{data}
\name{World}
\alias{World}
\title{World dataset}
\source{
NED: Natural Earth Data \url{https://www.naturalearthdata.com/}

HPI: Happy Planet Index \url{https://happyplanetindex.org/}

UNDP: Human Development Report (2024) \url{https://hdr.undp.org/content/human-development-report-2023-24}

WB: World Bank \url{https://data.worldbank.org}

OWiD: Our World in Data \url{https://ourworldindata.org}

RSF: Reporters Without Borders \url{https://rsf.org/en/index}
}
\usage{
World
}
\description{
World dataset, class \code{\link[sf:sf]{sf}}
}
\details{
\tabular{lll}{
   \strong{Variable} \tab \strong{Source} \tab \strong{Description} \cr
   \code{iso_a3} \tab NED \tab ISO 3166-1 alpha-3 three-letter country code (see below) \cr
   \code{name} \tab NED \tab Country name \cr
   \code{sovereignt} \tab NED \tab Sovereignt country name \cr
   \code{continent} \tab NED \tab Continent (primary; some countries are transcontinental) \cr
   \code{area} \tab NED \tab Area in km2 \cr
   \code{pop_est} \tab NED \tab Population estimation \cr
   \code{pop_est_dens} \tab NED \tab Population estimation per km2 \cr
   \code{economy} \tab NED \tab Economy class \cr
   \code{income_grp} \tab NED \tab Income group \cr
   \code{gdp_cap_est} \tab NED \tab GDP per capita (estimated) \cr
   \code{life_exp} \tab HPI \tab Life expectancy. The average number of years an infant born in that country is expected to live \cr
   \code{well_being} \tab HPI \tab Well being. Self-reported from 0 (worst) to 10 (best) \cr
   \code{footprint} \tab HPI \tab Carbon footprint. Per capita greendwelling gas emissions associated with consumption and economic activity \cr
   \code{HPI} \tab HPI \tab Happy Planet Indicator. An index of human well-being and environmental impact that was introduced by the New Economics Foundation in 2006. Approximate formula: \code{(life_exp * well_being) / footprint} \cr
   \code{inequality} \tab WB \tab Income inequality: Gini coefficient (World Bank variable SI.POV.GINI) A value of 0 represents perfect equality, while 100 implies perfect inequality \cr
   \code{gender} \tab UNDP/OWiD \tab Gender Inequality Index (GII) Composite metric using reproductive health, empowerment and the labour market. A low value indicates low inequality between women and men, and vice-versa \cr
   \code{press} \tab RSF \tab World Press Freedom Index. Degree of freedom that journalists, news organizations and netizens have \cr
}


See sources for more detailed information about the variables.

This dataset, created Noveber 2024, is an update from the old version, which has been created around 2016. All variables from the old version are included, but updated. Furthermore, gender ineuqlity and press freedom have been added.

ISO country-code: two countries have user-assigned codes, namely: XKX is used for Kosovo (conform European Union and World Bank) (was UNK in the old version); XNC is used for Northern Cyprus (was CYN in the old version).

For some variables data were available from multiple years, but availability was different across countries. In those cases, the most recent values were taken.
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tm_element.R
\name{tmap-element}
\alias{tmap-element}
\alias{+.tmap}
\title{Stacking of tmap elements}
\usage{
\method{+}{tmap}(e1, e2)
}
\arguments{
\item{e1}{first tmap element}

\item{e2}{second tmap element}
}
\description{
The plus operator allows you to stack tmap elements (functions with a prefix \code{tm_})
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tm_scale_.R
\name{tm_scale_continuous}
\alias{tm_scale_continuous}
\alias{tm_scale_continuous_log}
\alias{tm_scale_continuous_log2}
\alias{tm_scale_continuous_log10}
\alias{tm_scale_continuous_log1p}
\alias{tm_scale_continuous_sqrt}
\alias{tm_scale_continuous_pseudo_log}
\title{Scales: continuous scale}
\usage{
tm_scale_continuous(
  n = NULL,
  limits = NULL,
  outliers.trunc = NULL,
  ticks = NULL,
  trans = NULL,
  midpoint = NULL,
  values = NA,
  values.repeat = FALSE,
  values.range = NA,
  values.scale = NA,
  value.na = NA,
  value.null = NA,
  value.neutral = NA,
  labels = NULL,
  label.na = NA,
  label.null = NA,
  label.format = tm_label_format(),
  trans.args = list()
)

tm_scale_continuous_log(..., base = exp(1))

tm_scale_continuous_log2(...)

tm_scale_continuous_log10(...)

tm_scale_continuous_log1p(...)

tm_scale_continuous_sqrt(...)

tm_scale_continuous_pseudo_log(..., base = exp(1), sigma = 1)
}
\arguments{
\item{n}{Preferred number of tick labels. Only used if \code{ticks} is not specified}

\item{limits}{Limits of the data values that are mapped to the continuous scale. When \code{NA}, the range of data values is taken. When only one value is provided, the range of data values with this provided value is taken. The default depends on the visual variable: it is 0 for all visual variables other than color when \code{tm_scale_continuous} is used. For the transformation scale functions, it is \code{NA}.}

\item{outliers.trunc}{Should outliers be truncated? An outlier is a data value that is below or above the respectively lower and upper limit. A logical vector of two values is expected. The first and second value determines whether values lower than the lower limit respectively higher than the upper limit are truncated to the lower respectively upper limit. If \code{FALSE} (default), they are considered as missing values.}

\item{ticks}{Tick values. If not specified, it is determined automatically with \code{n}}

\item{trans}{Transformation function. One of \code{"identity"} (default), \code{"log"}, and \code{"log1p"}. Note: the base of the log scale is irrelevant, since the log transformed values are normalized before mapping to visual values.}

\item{midpoint}{The data value that is interpreted as the midpoint. By default it is set to 0 if negative and positive values are present. Useful when values are diverging colors. In that case, the two sides of the color palette are assigned to negative respectively positive values. If all values are positive or all values are negative, then the midpoint is set to \code{NA}, which means that the value that corresponds to the middle color class (see \code{style}) is mapped to the middle color. If it is specified for sequential color palettes (e.g. \code{"Blues"}), then this color palette will be treated as a diverging color palette.}

\item{values}{(generic scale argument) The visual values. For colors (e.g. \code{fill} or \code{col} for \code{\link[=tm_polygons]{tm_polygons()}}) this is a palette name from the \code{cols4all} package (see \code{\link[cols4all:c4a]{cols4all::c4a()}}) or vector of colors, for size (e.g. \code{size} for \code{\link[=tm_symbols]{tm_symbols()}}) these are a set of sizes (if two values are specified they are interpret as range), for symbol shapes (e.g. \code{shape} for \code{\link[=tm_symbols]{tm_symbols()}}) these are a set of symbols, etc. The tmap option \code{values.var} contains the default values per visual variable and in some cases also per data type.}

\item{values.repeat}{(generic scale argument) Should the values be repeated in case there are more categories?}

\item{values.range}{(generic scale argument) Range of the values, especially useful for color palettes. Vector of two numbers (both between 0 and 1) where the first determines the minimum and the second the maximum. Full range, which means that all values are used, is encoded as \code{c(0, 1)}. For instance, when a gray scale is used for color (from black to white), \code{c(0,1)} means that all colors are used, \verb{0.25, 0.75} means that only colors from dark gray to light gray are used (more precisely \code{"grey25"} to \code{"grey75"}), and \verb{0, 0.5} means that only colors are used from black to middle gray (\code{"grey50"}). When only one number is specified, this is interpreted as the second number (where the first is set to 0). Default values can be set via the tmap option \code{values.range}.}

\item{values.scale}{(generic scale argument) Scaling of the values. Only useful for size-related visual variables, such as \code{size} of \code{\link[=tm_symbols]{tm_symbols()}} and \code{lwd} of \code{\link[=tm_lines]{tm_lines()}}.}

\item{value.na}{(generic scale argument) Value used for missing values. See tmap option \code{"value.na"} for defaults per visual variable.}

\item{value.null}{(generic scale argument) Value used for NULL values. See tmap option \code{"value.null"} for defaults per visual variable. Null data values occur when out-of-scope features are shown (e.g. for a map of Europe showing a data variable per country, the null values are applied to countries outside Europe).}

\item{value.neutral}{(generic scale argument) Value that can be considered neutral. This is used for legends of other visual variables of the same map layer. E.g. when both \code{fill} and \code{size} are used for \code{\link[=tm_symbols]{tm_symbols()}} (using filled circles), the size legend items are filled with the \code{value.neutral} color from the \code{fill.scale} scale, and fill legend items are bubbles of size \code{value.neutral} from the \code{size.scale} scale.}

\item{labels}{(generic scale argument) Labels}

\item{label.na}{(generic scale argument) Label for missing values}

\item{label.null}{(generic scale argument) Label for null (out-of-scope) values}

\item{label.format}{(generic scale argument) Label formatting. Output of \code{\link[=tm_label_format]{tm_label_format()}}}

\item{trans.args}{list of additional argument for the transformation (generic transformation arguments)}

\item{...}{passed on to \code{\link[=tm_scale_continuous]{tm_scale_continuous()}}}

\item{base}{base of logarithm}

\item{sigma}{Scaling factor for the linear part of pseudo-log transformation.}
}
\description{
Scales in tmap are configured by the family of functions with prefix \code{tm_scale}.
Such function should be used for the input of the \code{.scale} arguments in the layer
functions (e.g. \code{fill.scale} in \code{\link[=tm_polygons]{tm_polygons()}}).
The function \code{tm_scale_continuous()} is used for continuous data.
The functions \verb{tm_scale_continuous_<x>()} use transformation functions x.
}
\examples{
tm_shape(World) +
  tm_polygons(
  	fill = "HPI",
    fill.scale = tm_scale_continuous(values = "scico.roma", midpoint = 30))

tm_shape(metro) +
  tm_bubbles(
  	size = "pop1950", 
    size.scale = tm_scale_continuous(
      values.scale = 1),
    size.legend = tm_legend("Population in 1950", frame = FALSE))

tm_shape(metro) +
  tm_bubbles(
    size = "pop1950", 
    size.scale = tm_scale_continuous(
      values.scale = 2, 
      limits = c(0, 12e6), 
      ticks = c(1e5, 3e5, 8e5, 4e6, 1e7), 
      labels = c("0 - 200,000", "200,000 - 500,000", "500,000 - 1,000,000", 
        "1,000,000 - 10,000,000", "10,000,000 or more"), 
      outliers.trunc = c(TRUE, TRUE)),
    size.legend = tm_legend("Population in 1950", frame = FALSE))
# Note that for this type of legend, we recommend tm_scale_intervals()
}
\seealso{
\href{https://r-tmap.github.io/tmap/articles/basics_scales}{Scales}, \href{https://r-tmap.github.io/tmap/articles/basics_vv}{Visual variables}, \href{https://r-tmap.github.io/tmap/articles/adv_multivariate}{Multivariate visual variables} \code{\link[=tm_scale]{tm_scale()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data.R
\docType{data}
\name{NLD_prov}
\alias{NLD_prov}
\alias{NLD_muni}
\alias{NLD_dist}
\title{Netherlands datasets}
\source{
https://www.cbs.nl/nl-nl/maatwerk/2024/11/kerncijfers-wijken-en-buurten-2022
}
\usage{
NLD_prov

NLD_muni

NLD_dist
}
\description{
Datasets of the Netherlands for 2022 at three levels: \code{NLD_prov} (12) provinces, \code{NLD_muni} (345) municipalities and \code{NLD_dist} (3340) districts , all class \code{\link[sf:sf]{sf}}
}
\details{
The data variables for \code{NLD_muni} and \code{NLD_dist} are identical:\tabular{ll}{
   \strong{Variable} \tab \strong{Description} \cr
   \code{code} \tab Code. Format is "GMaaaa" (municipality/'\strong{g}e\strong{m}eente') and "WKaaaabb" (district/\strong{w}ij\strong{k}). Here, "aaaa" represents the municipality id number, and "bb" the additional district number. \cr
   \code{name} \tab Name. \cr
   \code{province} \tab Province name. \cr
   \code{area} \tab Total area in km2. This area corresponds to the area of the polygons (including inland waters, excluding coastal waters), but is more precise because it is based on non-simplified geometries. \cr
   \code{urbanity} \tab Level of urbanity. Five classes, determined by the number of addresses per km2 (break values are 2500, 1500, 1000, and 500). \cr
   \code{population} \tab The total population count at 2022-01-01. \cr
   \code{pop_0_14} \tab Percentage (rounded) of people between 0 and 15. \cr
   \code{pop_15_24} \tab Percentage (rounded) of people between 15 and 25. \cr
   \code{pop_25_44} \tab Percentage (rounded) of people between 25 and 45. \cr
   \code{pop_45_64} \tab Percentage (rounded) of people between 45 and 65. \cr
   \code{pop_65plus} \tab Percentage (rounded) of people of 65 and older. \cr
   \code{dwelling_total} \tab Number of dwellings. \cr
   \code{dwelling_value} \tab Average dwelling value (Dutch: WOZ-value). \cr
   \code{dwelling_ownership} \tab Percentage of dwellings owned by the residents. \cr
   \code{employment_rate} \tab Share of the employed population within the total population from 15 to 75 years old. \cr
   \code{income_low} \tab Percentage of individuals in private households belonging to the lowest 40\% of personal income nationwide. \cr
   \code{income_high} \tab Percentage of individuals in private households belonging to the highest 20\% of personal income nationwide. \cr
   \code{edu_appl_sci} \tab Percentage of people aged 15 to 75 with a university of applied sciences (Dutch: HBO) or university (Dutch: WO) degree. \cr
}


See source for detailed information about the variables.

This dataset, created Noveber 2024, is an update from the datasets \code{NLD_muni} and \code{NLD_prov} used in tmap <= 3, which has been created around 2016. Note that the number of municipalities have been reduced (due to mergings). All old variables are included, except for variables related to ethnicity. Many new variable have been added, and moreover, district (Dutch: wijk) level data have added: \code{NLD_dist}.

The CRS (coordinate reference system) used is the Rijksdriekhoekstelsel New, EPSG 28992. Coordinates have been rounded to meters to reduce file size.
}
\references{
Statistics Netherlands (2024), The Hague/Heerlen, Netherlands, \url{https://www.cbs.nl/}.
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tm_scale_.R
\name{tm_scale_ordinal}
\alias{tm_scale_ordinal}
\alias{tm_scale_categorical}
\title{Scales: categorical and ordinal scale}
\usage{
tm_scale_ordinal(
  n.max = 30,
  values = NA,
  values.repeat = FALSE,
  values.range = 1,
  values.scale = NA,
  value.na = NA,
  value.null = NA,
  value.neutral = NA,
  levels = NULL,
  levels.drop = FALSE,
  labels = NULL,
  label.na = NA,
  label.null = NA,
  label.format = list()
)

tm_scale_categorical(
  n.max = 30,
  values = NA,
  values.repeat = TRUE,
  values.range = NA,
  values.scale = NA,
  value.na = NA,
  value.null = NA,
  value.neutral = NA,
  levels = NULL,
  levels.drop = FALSE,
  labels = NULL,
  label.na = NA,
  label.null = NA,
  label.format = list()
)
}
\arguments{
\item{n.max}{Maximum number of categories (factor levels). In case there are more, they are grouped into \code{n.max} groups.}

\item{values}{(generic scale argument) The visual values. For colors (e.g. \code{fill} or \code{col} for \code{tm_polygons()}) this is a palette name from the \code{cols4all} package (see \code{\link[cols4all:c4a]{cols4all::c4a()}}) or vector of colors, for size (e.g. \code{size} for \code{tm_symbols()}) these are a set of sizes (if two values are specified they are interpret as range), for symbol shapes (e.g. \code{shape} for \code{\link[=tm_symbols]{tm_symbols()}}) these are a set of symbols, etc. The tmap option \code{values.var} contains the default values per visual variable and in some cases also per data type.}

\item{values.repeat}{(generic scale argument) Should the values be repeated in case there are more categories?}

\item{values.range}{(generic scale argument) Range of the values. Vector of two numbers (both between 0 and 1) where the first determines the minimum and the second the maximum. Full range, which means that all values are used, is encoded as \code{c(0, 1)}. For instance, when a gray scale is used for color (from black to white), \code{c(0,1)} means that all colors are used, \verb{0.25, 0.75} means that only colors from dark gray to light gray are used (more precisely \code{"grey25"} to \code{"grey75"}), and \verb{0, 0.5} means that only colors are used from black to middle gray (\code{"gray50"}). When only one number is specified, this is interpreted as the second number (where the first is set to 0). Default values can be set via the tmap option \code{values.range}.}

\item{values.scale}{(generic scale argument) Scaling of the values. Only useful for size-related visual variables, such as \code{size} of \code{\link[=tm_symbols]{tm_symbols()}} and \code{lwd} of \code{\link[=tm_lines]{tm_lines()}}.}

\item{value.na}{(generic scale argument) Value used for missing values. See tmap option \code{"value.na"} for defaults per visual variable.}

\item{value.null}{(generic scale argument) Value used for NULL values. See tmap option \code{"value.null"} for defaults per visual variable. Null data values occur when out-of-scope features are shown (e.g. for a map of Europe showing a data variable per country, the null values are applied to countries outside Europe).}

\item{value.neutral}{(generic scale argument) Value that can be considered neutral. This is used for legends of other visual variables of the same map layer. E.g. when both \code{fill} and \code{size} are used for \code{\link[=tm_symbols]{tm_symbols()}} (using filled circles), the size legend items are filled with the \code{value.neutral} color from the \code{fill.scale} scale, and fill legend items are bubbles of size \code{value.neutral} from the \code{size.scale} scale.}

\item{levels}{Levels to show. Other values are treated as missing.}

\item{levels.drop}{Should unused levels be dropped (and therefore are not assigned to a visual value and shown in the legend)?}

\item{labels}{(generic scale argument) Labels}

\item{label.na}{(generic scale argument) Label for missing values}

\item{label.null}{(generic scale argument) Label for null (out-of-scope) values}

\item{label.format}{(generic scale argument) Label formatting. Output of \code{\link[=tm_label_format]{tm_label_format()}}}
}
\description{
Scales in tmap are configured by the family of functions with prefix \code{tm_scale}.
Such function should be used for the input of the \code{.scale} arguments in the
layer functions (e.g. \code{fill.scale} in \code{\link[=tm_polygons]{tm_polygons()}}).
The functions \code{tm_scale_categorical()} and \code{tm_scale_ordinal()} are used
for categorical data. The only difference between these functions is that the
former assumes unordered categories whereas the latter assumes ordered categories.
For colors (the visual variable \code{fill} or \code{col}), different default color
palettes are used (see the tmap option \code{values.var}).
}
\seealso{
\code{\link[=tm_scale]{tm_scale()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tm_layers_lines.R
\name{tm_lines}
\alias{tm_lines}
\alias{opt_tm_lines}
\title{Map layer: lines}
\usage{
tm_lines(
  col = tm_const(),
  col.scale = tm_scale(),
  col.legend = tm_legend(),
  col.chart = tm_chart_none(),
  col.free = NA,
  lwd = tm_const(),
  lwd.scale = tm_scale(),
  lwd.legend = tm_legend(),
  lwd.chart = tm_chart_none(),
  lwd.free = NA,
  lty = tm_const(),
  lty.scale = tm_scale(),
  lty.legend = tm_legend(),
  lty.chart = tm_chart_none(),
  lty.free = NA,
  col_alpha = tm_const(),
  col_alpha.scale = tm_scale(),
  col_alpha.legend = tm_legend(),
  col_alpha.chart = tm_chart_none(),
  col_alpha.free = NA,
  linejoin = "round",
  lineend = "round",
  plot.order = tm_plot_order("lwd", reverse = TRUE, na.order = "bottom"),
  zindex = NA,
  group = NA,
  group.control = "check",
  popup.vars = NA,
  popup.format = tm_label_format(),
  hover = NA,
  id = "",
  options = opt_tm_lines(),
  ...
)

opt_tm_lines(lines.only = "ifany", hitbox = "auto")
}
\arguments{
\item{col, col.scale, col.legend, col.chart, col.free}{Visual variable that determines the color. See details.}

\item{lwd, lwd.scale, lwd.legend, lwd.chart, lwd.free}{Visual variable that determines the line width. See details.}

\item{lty, lty.scale, lty.legend, lty.chart, lty.free}{Visual variable that determines the line type. See details.}

\item{col_alpha, col_alpha.scale, col_alpha.legend, col_alpha.chart, col_alpha.free}{Visual variable that determines the color transparency. See details.}

\item{linejoin, lineend}{line join and line end. See \link[grid:gpar]{gpar()} for details.}

\item{plot.order}{Specification in which order the spatial features are drawn. See \code{\link[=tm_plot_order]{tm_plot_order()}} for details.}

\item{zindex}{Map layers are drawn on top of each other. The \code{zindex} numbers
(one for each map layer) determines the stacking order.
By default the map layers are drawn in the order they are called.}

\item{group}{Name of the group to which this layer belongs. This is only
relevant in view mode, where layer groups can be switched (see \code{group.control})}

\item{group.control}{In view mode, the group control determines how layer groups
can be switched on and off. Options: \code{"radio"} for radio buttons
(meaning only one group can be shown), \code{"check"} for check boxes
(so multiple groups can be shown), and \code{"none"} for no control
(the group cannot be (de)selected).}

\item{popup.vars}{names of data variables that are shown in the popups
in \code{"view"} mode. Set popup.vars to \code{TRUE} to show all variables in the
shape object. Set popup.vars to \code{FALSE} to disable popups. Set \code{popup.vars}
to a character vector of variable names to those those variables in the popups.
The default (\code{NA}) depends on whether visual variables (e.g.\code{fill}) are used.
If so, only those are shown. If not all variables in the shape object are shown.}

\item{popup.format}{list of formatting options for the popup values. Output of \code{\link[=tm_label_format]{tm_label_format()}}. Only applicable for
numeric data variables. If one list of formatting options is provided,
it is applied to all numeric variables of \code{popup.vars}. Also, a (named)
list of lists can be provided. In that case, each list of formatting options
is applied to the named variable.}

\item{hover}{name of the data variable that specifies the hover labels (view mode only). Set to \code{FALSE} to disable hover labels. By default \code{FALSE}, unless \code{id} is specified. In that case, it is set to \code{id},}

\item{id}{name of the data variable that specifies the indices of the spatial
features. Only used for \code{"view"} mode.}

\item{options}{options passed on to the corresponding \verb{opt_<layer_function>} function}

\item{...}{to catch deprecated arguments from version < 4.0}

\item{lines.only}{should only line geometries of the shape object (defined in \code{\link[=tm_shape]{tm_shape()}}) be plotted, or also other geometry types (like polygons)? By default \code{"ifany"}, which means \code{TRUE} in case a geometry collection is specified.}

\item{hitbox}{Controls whether an invisible interaction layer with a
larger clickable area (\"hitbox\") is added on top of the lines.

This can improve click and popup behaviour for thin or densely packed
lines by increasing the effective mouse interaction width.

Possible values:
\describe{
\item{"none"}{No additional hitbox layer is added. Lines are clickable
only at their visible width.}

\if{html}{\out{<div class="sourceCode">}}\preformatted{\\item\{"pmaxX"\}\{Ensures a minimum interaction width of \code{X} pixels.
The hitbox width is computed as \code{pmax(line_width, X)}.
For example, \code{"pmax8"} guarantees at least 8 pixels.\}

\\item\{"timesX"\}\{Multiplies the visible line width by \code{X} to compute
the interaction width. For example, \code{"times2"} doubles the
clickable width.\}

\\item\{"auto"\}\{Automatically: \code{"pmax8"} by default if and only if interactive features are enabled (popup or hover) and lines are thin (median line width < 4) and there are less than 10000 features. Otherwise, \code{"none"} \}
}\if{html}{\out{</div>}}

}

Adding a hitbox improves usability for thin lines but may reduce
performance for very large datasets, as an additional invisible
layer is rendered.}
}
\description{
Map layer that draws lines. Supported visual variables are: \code{col} (the color), \code{lwd} (line width), \code{lty} (line type), and \code{col_alpha} (color alpha transparency).
}
\details{
The visual variable arguments (e.g. \code{col}) can be specified with a data
variable name (e.g., a spatial vector attribute or a raster layer of the object
specified in \code{\link[=tm_shape]{tm_shape()}}), with a visual value (for \code{col}, a color is expected), or with a geometry-derived variable (see below).
See \href{https://r-tmap.github.io/tmap/articles/basics_vv}{vignette about visual variables}.

Multiple values can be specified: in that case facets are created.
These facets can be combined with other faceting data variables, specified with \code{\link[=tm_facets]{tm_facets()}}.
See \href{https://r-tmap.github.io/tmap/articles/basics_facets}{vignette about facets}.
\itemize{
\item The \verb{*.scale} arguments determine the used scale to map the data values to
visual variable values. These can be specified with one of the available
\verb{tm_scale_*()} functions. The default is specified by the tmap option (\code{\link[=tm_options]{tm_options()}}) \code{scales.var}.
See \href{https://r-tmap.github.io/tmap/articles/basics_scales}{vignette about scales}.
\item The \verb{*.legend} arguments determine the used legend, specified with \code{\link[=tm_legend]{tm_legend()}}.
The default legend and its settings are determined by the tmap options (\code{\link[=tm_options]{tm_options()}}) \code{legend.} .
See \href{https://r-tmap.github.io/tmap/articles/basics_legends}{vignette about legends}.
\item The \verb{*.chart} arguments specify additional charts, specified with \code{tm_chart_}, e.g. \code{\link[=tm_chart_histogram]{tm_chart_histogram()}}.
See \href{https://r-tmap.github.io/tmap/articles/basics_charts}{vignette about charts}.
\item The \verb{*.free} arguments determine whether scales are applied freely across facets, or shared.
A logical value is required. They can also be specified with a vector of three
logical values; these determine whether scales are applied freely per facet dimension.
This is only useful when facets are applied (see \code{\link[=tm_facets]{tm_facets()}}).
There are maximally three facet dimensions: rows, columns, and pages. This only
applies for a facet grid (\code{\link[=tm_facets_grid]{tm_facets_grid()}}). For instance, \code{col.free = c(TRUE, FALSE, FALSE)}
means that for the visual variable \code{col}, each row of facets will have its own
scale, and therefore its own legend. For facet wraps and stacks
(\code{\link[=tm_facets_wrap]{tm_facets_wrap()}} and \code{\link[=tm_facets_stack]{tm_facets_stack()}}) there is only one facet dimension,
so the \verb{*.free} argument requires only one logical value.
}

Currently, three geometry-derived variables are implemented:
\itemize{
\item \code{"AREA"} (polygons only), which uses the feature area;
\item \code{"LENGTH"} (lines only), which uses the feature length; and
\item \code{"MAP_COLORS"}, which assigns values so that adjacent features receive
different values, making it particularly suitable for coloring
neighbouring polygons.
}

Note that geometry-derived variables do not generate a legend automatically.
If a legend is required, compute the corresponding variable explicitly,
for example with \code{\link[sf:geos_measures]{sf::st_area()}}, \code{\link[sf:geos_measures]{sf::st_length()}}, or
\code{\link[tmaptools:map_coloring]{tmaptools::map_coloring()}}, and use the resulting values instead.
}
\examples{
tm_shape(World_rivers) +
	tm_lines(lwd = "strokelwd",
			 lwd.scale = tm_scale_asis(values.scale = 0.2, value.neutral = 2),
			 col = "scalerank",
			 col.scale = tm_scale_categorical(values = "seaborn.dark"))

tm_shape(World) +
	tm_lines(col = "continent",
			 col.scale = tm_scale_categorical(values = "seaborn.dark"),
			 lty = "continent",
			 lwd = 1.5,
			 lty.legend = tm_legend_combine("col"))
}
\seealso{
\href{https://r-tmap.github.io/tmap/articles/examples_terrain}{Terrain map example}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tmap_options.R
\name{tmapMode}
\alias{tmapMode}
\title{Internal method for submitting a new mode (deprecated)}
\usage{
tmapMode(id, name, ..., styleOptions = NULL)
}
\arguments{
\item{id}{id of the mode: please use lowercase and one-word. This will be used with \code{\link[=tmap_mode]{tmap_mode()}}.}

\item{name}{name of the mode: please use title case. This will be used to recognize internal functions, e.g. \code{tmapLeafletInit}}

\item{...}{mode specific options}
}
\description{
Internal method for submitting a new mode. Deprecated. Use \code{\link[=tmapSubmitOptions]{tmapSubmitOptions()}} instead.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tm_scale_.R
\name{tm_scale_ordinal}
\alias{tm_scale_ordinal}
\alias{tm_scale_categorical}
\title{Scales: categorical and ordinal scale}
\usage{
tm_scale_ordinal(
  n.max = 30,
  values = NA,
  values.repeat = FALSE,
  values.range = 1,
  values.scale = NA,
  value.na = NA,
  value.null = NA,
  value.neutral = NA,
  levels = NULL,
  levels.drop = FALSE,
  labels = NULL,
  label.na = NA,
  label.null = NA,
  label.format = list()
)

tm_scale_categorical(
  n.max = 30,
  values = NA,
  values.repeat = TRUE,
  values.range = NA,
  values.scale = NA,
  value.na = NA,
  value.null = NA,
  value.neutral = NA,
  levels = NULL,
  levels.drop = FALSE,
  labels = NULL,
  label.na = NA,
  label.null = NA,
  label.format = list()
)
}
\arguments{
\item{n.max}{Maximum number of categories (factor levels). In case there are more, they are grouped into \code{n.max} groups.}

\item{values}{(generic scale argument) The visual values. For colors (e.g. \code{fill} or \code{col} for \code{tm_polygons()}) this is a palette name from the \code{cols4all} package (see \code{\link[cols4all:c4a]{cols4all::c4a()}}) or vector of colors, for size (e.g. \code{size} for \code{tm_symbols()}) these are a set of sizes (if two values are specified they are interpret as range), for symbol shapes (e.g. \code{shape} for \code{\link[=tm_symbols]{tm_symbols()}}) these are a set of symbols, etc. The tmap option \code{values.var} contains the default values per visual variable and in some cases also per data type.}

\item{values.repeat}{(generic scale argument) Should the values be repeated in case there are more categories?}

\item{values.range}{(generic scale argument) Range of the values. Vector of two numbers (both between 0 and 1) where the first determines the minimum and the second the maximum. Full range, which means that all values are used, is encoded as \code{c(0, 1)}. For instance, when a gray scale is used for color (from black to white), \code{c(0,1)} means that all colors are used, \verb{0.25, 0.75} means that only colors from dark gray to light gray are used (more precisely \code{"grey25"} to \code{"grey75"}), and \verb{0, 0.5} means that only colors are used from black to middle gray (\code{"gray50"}). When only one number is specified, this is interpreted as the second number (where the first is set to 0). Default values can be set via the tmap option \code{values.range}.}

\item{values.scale}{(generic scale argument) Scaling of the values. Only useful for size-related visual variables, such as \code{size} of \code{\link[=tm_symbols]{tm_symbols()}} and \code{lwd} of \code{\link[=tm_lines]{tm_lines()}}.}

\item{value.na}{(generic scale argument) Value used for missing values. See tmap option \code{"value.na"} for defaults per visual variable.}

\item{value.null}{(generic scale argument) Value used for NULL values. See tmap option \code{"value.null"} for defaults per visual variable. Null data values occur when out-of-scope features are shown (e.g. for a map of Europe showing a data variable per country, the null values are applied to countries outside Europe).}

\item{value.neutral}{(generic scale argument) Value that can be considered neutral. This is used for legends of other visual variables of the same map layer. E.g. when both \code{fill} and \code{size} are used for \code{\link[=tm_symbols]{tm_symbols()}} (using filled circles), the size legend items are filled with the \code{value.neutral} color from the \code{fill.scale} scale, and fill legend items are bubbles of size \code{value.neutral} from the \code{size.scale} scale.}

\item{levels}{Levels to show. Other values are treated as missing.}

\item{levels.drop}{Should unused levels be dropped (and therefore are not assigned to a visual value and shown in the legend)?}

\item{labels}{(generic scale argument) Labels}

\item{label.na}{(generic scale argument) Label for missing values}

\item{label.null}{(generic scale argument) Label for null (out-of-scope) values}

\item{label.format}{(generic scale argument) Label formatting. Output of \code{\link[=tm_label_format]{tm_label_format()}}}
}
\description{
Scales in tmap are configured by the family of functions with prefix \code{tm_scale}.
Such function should be used for the input of the \code{.scale} arguments in the
layer functions (e.g. \code{fill.scale} in \code{\link[=tm_polygons]{tm_polygons()}}).
The functions \code{tm_scale_categorical()} and \code{tm_scale_ordinal()} are used
for categorical data. The only difference between these functions is that the
former assumes unordered categories whereas the latter assumes ordered categories.
For colors (the visual variable \code{fill} or \code{col}), different default color
palettes are used (see the tmap option \code{values.var}).
}
\seealso{
\href{https://r-tmap.github.io/tmap/articles/basics_scales}{Scales}, \href{https://r-tmap.github.io/tmap/articles/basics_vv}{Visual variables}, \href{https://r-tmap.github.io/tmap/articles/adv_multivariate}{Multivariate visual variables} \code{\link[=tm_scale]{tm_scale()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tm_layers_lines.R
\name{tm_lines}
\alias{tm_lines}
\alias{opt_tm_lines}
\title{Map layer: lines}
\usage{
tm_lines(
  col = tm_const(),
  col.scale = tm_scale(),
  col.legend = tm_legend(),
  col.chart = tm_chart_none(),
  col.free = NA,
  lwd = tm_const(),
  lwd.scale = tm_scale(),
  lwd.legend = tm_legend(),
  lwd.chart = tm_chart_none(),
  lwd.free = NA,
  lty = tm_const(),
  lty.scale = tm_scale(),
  lty.legend = tm_legend(),
  lty.chart = tm_chart_none(),
  lty.free = NA,
  col_alpha = tm_const(),
  col_alpha.scale = tm_scale(),
  col_alpha.legend = tm_legend(),
  col_alpha.chart = tm_chart_none(),
  col_alpha.free = NA,
  linejoin = "round",
  lineend = "round",
  plot.order = tm_plot_order("lwd", reverse = TRUE, na.order = "bottom"),
  zindex = NA,
  group = NA,
  group.control = "check",
  popup = tm_popup(),
  popup.vars = NA,
  popup.format = tm_label_format(),
  hover = NA,
  id = "",
  blend = "over",
  options = opt_tm_lines(),
  ...
)

opt_tm_lines(lines.only = "ifany", hitbox = "auto")
}
\arguments{
\item{col, col.scale, col.legend, col.chart, col.free}{Visual variable that determines the color. See details. \emph{Unit:} Color -- a color name, hex string.}

\item{lwd, lwd.scale, lwd.legend, lwd.chart, lwd.free}{Visual variable that determines the line width. See details. \emph{Unit:} Base R line-width units; 1 lwd is approx. 0.75 pt at 96 dpi.}

\item{lty, lty.scale, lty.legend, lty.chart, lty.free}{Visual variable that determines the line type. See details. \emph{Unit:} Integer (1-6) or name: "solid", "dashed", "dotted", "dotdash", "longdash", "twodash".}

\item{col_alpha, col_alpha.scale, col_alpha.legend, col_alpha.chart, col_alpha.free}{Visual variable that determines the color transparency. See details. \emph{Unit:} Proportion -- numeric 0-1 (0 = fully transparent, 1 = fully opaque).}

\item{linejoin, lineend}{line join and line end. See \link[grid:gpar]{gpar()} for details.}

\item{plot.order}{Specification in which order the spatial features are drawn. See \code{\link[=tm_plot_order]{tm_plot_order()}} for details.}

\item{zindex}{Controls the stacking order of map layers. Should be set to a
value above 400. By default, layers are stacked in call order, starting
at 401. See details.}

\item{group}{Name of the group to which this layer belongs. This is only
relevant in view mode, where layer groups can be switched (see \code{group.control})}

\item{group.control}{In view mode, the group control determines how layer groups
can be switched on and off. Options: \code{"radio"} for radio buttons
(meaning only one group can be shown), \code{"check"} for check boxes
(so multiple groups can be shown), and \code{"none"} for no control
(the group cannot be (de)selected).}

\item{popup}{popup specification for \code{"view"} mode, the output of \code{\link[=tm_popup]{tm_popup()}}.
It determines the data variables shown in the popup table, the popup title,
and (in the future) the popup layout. This replaces the deprecated
arguments \code{popup.vars} and \code{popup.format}.}

\item{popup.vars}{(Deprecated.) Use \code{popup} with
\code{\link[=tm_popup]{tm_popup()}} instead (via its \code{vars} argument). Names of data variables that
are shown in the popups in \code{"view"} mode. Set \code{popup.vars} to \code{TRUE} to show
all variables in the shape object. Set \code{popup.vars} to \code{FALSE} to disable
popups. Set \code{popup.vars} to a character vector of variable names to show
those variables in the popups. The default (\code{NA}) depends on whether visual
variables (e.g. \code{fill}) are used. If so, only those are shown. If not, all
variables in the shape object are shown.}

\item{popup.format}{(Deprecated.) Use \code{popup} with
\code{\link[=tm_popup]{tm_popup()}} instead (via its \code{format} argument). List of formatting options
for the popup values. Output of \code{\link[=tm_label_format]{tm_label_format()}}. Only applicable for
numeric data variables. If one list of formatting options is provided,
it is applied to all numeric variables of \code{popup.vars}. Also, a (named)
list of lists can be provided. In that case, each list of formatting options
is applied to the named variable.}

\item{hover}{name of the data variable that specifies the hover labels (view mode only). Set to \code{FALSE} to disable hover labels. By default \code{FALSE}, unless \code{id} is specified. In that case, it is set to \code{id},}

\item{id}{name of the data variable that specifies the indices of the spatial
features. Only used for \code{"view"} mode.}

\item{blend}{Compositing operator for layer blending. Default \code{"over"} applies
no blending. See the "Layer blending" section for the supported values.}

\item{options}{options passed on to the corresponding \verb{opt_<layer_function>} function}

\item{...}{to catch deprecated arguments from version < 4.0}

\item{lines.only}{should only line geometries of the shape object (defined in \code{\link[=tm_shape]{tm_shape()}}) be plotted, or also other geometry types (like polygons)? By default \code{"ifany"}, which means \code{TRUE} in case a geometry collection is specified.}

\item{hitbox}{Controls whether an invisible interaction layer with a
larger clickable area (\"hitbox\") is added on top of the lines.

This can improve click and popup behaviour for thin or densely packed
lines by increasing the effective mouse interaction width.

Possible values:
\describe{
\item{"none"}{No additional hitbox layer is added. Lines are clickable
only at their visible width.}
\item{"plusX"}{Adds \code{X} pixels to the visible line width to compute
the interaction width: \code{line_width + X}. For example, \code{"plus8"}
widens the clickable area by 4 pixels on each side.}
\item{"pmaxX"}{Ensures a minimum interaction width of \code{X} pixels:
\code{pmax(line_width, X)}. For example, \code{"pmax8"} guarantees at
least 8 pixels. Useful for very thin lines.}
\item{"auto"}{\code{"pmax8"} if and only if interactive features are
enabled (popup or hover), lines are thin (median line width < 4), and
there are fewer than 10000 features. Otherwise \code{"none"}.}
}

\code{plus} and \code{pmax} can be combined, e.g. \code{"plus4pmax8"} means
\code{pmax(line_width + 4, 8)}.

Adding a hitbox improves usability for thin lines but may reduce
performance for very large datasets, as an additional invisible
layer is rendered.}
}
\description{
Map layer that draws lines. Supported visual variables are: \code{col} (the color), \code{lwd} (line width), \code{lty} (line type), and \code{col_alpha} (color alpha transparency).
}
\details{
The visual variable arguments (e.g. \code{col}) can be specified with a data
variable name (e.g., a spatial vector attribute or a raster layer of the object
specified in \code{\link[=tm_shape]{tm_shape()}}), with a visual value (for \code{col}, a color is expected), or with a geometry-derived variable (see below).
See vignette: \href{https://r-tmap.github.io/tmap/articles/basics_vv}{Visual variables}.

Multiple values can be specified: in that case facets are created.
These facets can be combined with other faceting data variables, specified with \code{\link[=tm_facets]{tm_facets()}}.
See vignette: \href{https://r-tmap.github.io/tmap/articles/basics_facets}{Facets}.
\itemize{
\item The \verb{*.scale} arguments determine the used scale to map the data values to
visual variable values. These can be specified with one of the available
\verb{tm_scale_*()} functions. The default is specified by the tmap option (\code{\link[=tm_options]{tm_options()}}) \code{scales.var}.
See `rvignette: \href{https://r-tmap.github.io/tmap/articles/basics_scales}{Scales}
\item The \verb{*.legend} arguments determine the used legend, specified with \code{\link[=tm_legend]{tm_legend()}}.
The default legend and its settings are determined by the tmap options (\code{\link[=tm_options]{tm_options()}}) \code{legend.} .
See `rvignette: \href{https://r-tmap.github.io/tmap/articles/basics_legends}{Legends}
\item The \verb{*.chart} arguments specify additional charts, specified with \code{tm_chart_}, e.g. \code{\link[=tm_chart_histogram]{tm_chart_histogram()}}.
See `rvignette: \href{https://r-tmap.github.io/tmap/articles/basics_charts}{Charts}
\item The \verb{*.free} arguments determine whether scales are applied freely across facets, or shared.
A logical value is required. They can also be specified with a vector of three
logical values; these determine whether scales are applied freely per facet dimension.
This is only useful when facets are applied (see \code{\link[=tm_facets]{tm_facets()}}).
There are maximally three facet dimensions: rows, columns, and pages. This only
applies for a facet grid (\code{\link[=tm_facets_grid]{tm_facets_grid()}}). For instance, \code{col.free = c(TRUE, FALSE, FALSE)}
means that for the visual variable \code{col}, each row of facets will have its own
scale, and therefore its own legend. For facet wraps and stacks
(\code{\link[=tm_facets_wrap]{tm_facets_wrap()}} and \code{\link[=tm_facets_stack]{tm_facets_stack()}}) there is only one facet dimension,
so the \verb{*.free} argument requires only one logical value.
}

Currently, three geometry-derived variables are implemented:
\itemize{
\item \code{"AREA"} (polygons only), which uses the feature area;
\item \code{"LENGTH"} (lines only), which uses the feature length; and
\item \code{"MAP_COLORS"}, which assigns values so that adjacent features receive
different values, making it particularly suitable for coloring
neighbouring polygons.
}

Note that geometry-derived variables do not generate a legend automatically.
If a legend is required, compute the corresponding variable explicitly,
for example with \code{\link[sf:st_area]{sf::st_area()}}, \code{\link[sf:st_length]{sf::st_length()}}, or
\code{\link[tmaptools:map_coloring]{tmaptools::map_coloring()}}, and use the resulting values instead.
\subsection{Visual variable units}{

Every visual variable maps data values to a specific output unit.
Knowing the unit matters when supplying constant values via \code{\link[=tm_const]{tm_const()}},
or output ranges via \code{values.range} / \code{values.scale} in the scale
functions.\tabular{lll}{
   Variable \tab Output unit \tab Notes \cr
   \code{fill}, \code{col}, \code{bgcol} \tab color \tab name, hex, or palette string \cr
   \code{fill_alpha}, \code{col_alpha}, \code{bgcol_alpha} \tab proportion 0-1 \tab 0 = transparent, 1 = opaque \cr
   \code{size} (symbols, bubbles, squares, dots) \tab typographic lines \tab 1 line approx. 1/6 inch; scaled by \code{values.scale} \cr
   \code{size} (circles) \tab meters \tab plain numeric or a \code{units} object \cr
   \code{size} (text, labels) \tab multiplier \tab 1 = 12 pt (plot) / 12 px (view) \cr
   \code{lwd} \tab lwd \tab base R units; 1 lwd approx. 0.75 pt at 96 dpi \cr
   \code{lty} \tab -- \tab integer 1-6 or name ("solid", "dashed", ...) \cr
   \code{shape} \tab -- \tab integer \code{pch} 1-25 or single character \cr
   \code{angle} \tab degrees \tab 0-360, clockwise from north \cr
   \code{fontface} \tab -- \tab "plain", "bold", "italic", "bold.italic" \cr
}

\subsection{Symbol size (\code{size} in \code{tm_symbols}, \code{tm_bubbles}, \code{tm_squares}, \code{tm_dots})}{

"Lines" is a typographic unit: one line is approximately 1/6 inch (the
default base line-height in R graphics). The global multiplier
\code{tmap_options(values.scale = list(size.bubbles = 1.5))} scales all symbol
sizes without changing the data mapping.
}

\subsection{Circle size (\code{size} in \code{tm_circles})}{

The value is a geographic radius in meters. A plain numeric vector is
interpreted as meters; a \code{units} object (from the \strong{units} package) is
automatically converted, so \code{units::as_units(1, "mi")} gives a 1-mile
radius. Because the radius is geographic, circles scale with zoom in
interactive (view) mode -- unlike bubble symbols which keep a fixed screen
size.
}

\subsection{Text size (\code{size} in \code{tm_text}, \code{tm_labels})}{

The value is a multiplier of the base font size. \code{size = 1} renders at
12 pt in plot mode (R's default \code{par("ps")}) and at 12 px in view
mode (\code{gp$cex * 12} px, see \code{tmapLeafletDataPlot.tm_data_text}); the two
modes are consistent by design.
}

}

\subsection{Layer blending (\code{blend})}{

Blend modes control how a layer's pixels are combined with the pixels
beneath it. For each pixel, let \eqn{S} be the source (top layer) RGB
value and \eqn{D} be the destination (bottom layer) RGB value, both
normalised to \eqn{[0, 1]}.\tabular{lll}{
   \code{blend} \tab Formula \tab Use case \cr
   \code{"over"} \tab \eqn{S \cdot \alpha + D \cdot (1 - \alpha)} \tab Standard alpha compositing (default) \cr
   \code{"multiply"} \tab \eqn{S \times D} \tab Hillshading over colour raster; both layers darken each other \cr
   \code{"screen"} \tab \eqn{1 - (1 - S)(1 - D)} \tab Inverse of multiply; brightens \cr
   \code{"overlay"} \tab multiply if \eqn{D < 0.5}, screen if \eqn{D \geq 0.5} \tab Boosts contrast; preserves midtones \cr
   \code{"darken"} \tab \eqn{\min(S, D)} \tab Keeps the darker of the two layers per channel \cr
   \code{"lighten"} \tab \eqn{\max(S, D)} \tab Keeps the lighter of the two layers per channel \cr
}


Requires R >= 4.2 and a compatible graphics device (e.g.
\code{png(type = "cairo")}, \code{svg()}). In view mode, blending is applied via
CSS \code{mix-blend-mode}. See \code{\link[grid:groupGrob]{grid::groupGrob()}} for the full list of
supported operators.
}

\subsection{zindex and pane names}{

In view mode, each layer is rendered in a Leaflet pane named \code{"tmap{zindex}"}
(e.g., \code{"tmap401"}, \code{"tmap402"}), with base tile layers placed in the
standard \code{"tile"} pane.
}
}
\examples{
tm_shape(World_rivers) +
	tm_lines(lwd = "strokelwd",
			 lwd.scale = tm_scale_asis(values.scale = 0.2, value.neutral = 2),
			 col = "scalerank",
			 col.scale = tm_scale_categorical(values = "seaborn.dark"))

tm_shape(World) +
	tm_lines(col = "continent",
			 col.scale = tm_scale_categorical(values = "seaborn.dark"),
			 lty = "continent",
			 lwd = 1.5,
			 lty.legend = tm_legend_combine("col"))
}
\seealso{
\href{https://r-tmap.github.io/tmap/articles/basics_vv}{Visual variables}, \href{https://r-tmap.github.io/tmap/articles/basics_scales}{Scales}, \href{https://r-tmap.github.io/tmap/articles/basics_legends}{Legends}, \href{https://r-tmap.github.io/tmap/articles/basics_facets}{Facets}, \href{https://r-tmap.github.io/tmap/articles/foundations_units}{Units}, \href{https://r-tmap.github.io/tmap/articles/adv_blend}{Layer blending}, \href{https://r-tmap.github.io/tmap/articles/examples_terrain}{Terrain map}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tmap_arrange.R
\name{tmap_arrange}
\alias{tmap_arrange}
\alias{knit_print.tmap_arrange}
\alias{print.tmap_arrange}
\title{Arrange small multiples in grid layout}
\usage{
tmap_arrange(
  ...,
  ncol = NA,
  nrow = NA,
  widths = NA,
  heights = NA,
  sync = FALSE,
  asp = 0,
  outer.margins = 0.02
)

\method{knit_print}{tmap_arrange}(x, ..., options = NULL)

\method{print}{tmap_arrange}(x, knit = FALSE, ..., options = NULL)
}
\arguments{
\item{...}{\code{\link{tmap}} objects or one list of \code{\link{tmap}} objects.
The number of multiples that can be plot is limited (see details).}

\item{ncol}{number of columns}

\item{nrow}{number of rows}

\item{widths}{vector of column widths. It should add up to 1 and the length
should be equal to \code{ncol}.}

\item{heights}{vector of row heights. It should add up to 1 and the length
should be equal to \code{nrow}.}

\item{sync}{logical. Should the navigation in view mode (zooming and panning)
be synchronized? By default \code{FALSE}.}

\item{asp}{aspect ratio. The aspect ratio of each map. Normally, this is
controlled by the \code{asp} argument from \code{\link[=tm_layout]{tm_layout()}} (also a tmap option).
This argument will overwrite it, unless set to \code{NULL}.
The default value for \code{asp} is 0, which means that the aspect ratio is
adjusted to the size of the device divided by the number of columns and rows.
When \code{asp} is set to \code{NA}, which is also the default value for \code{tm_layout()},
the aspect ratio will be adjusted to the used shapes.}

\item{outer.margins}{outer.margins, numeric vector four or a single value.
If defines the outer margins for each multiple. If will overwrite the
\code{outer.margins} argument from \code{\link[=tm_layout]{tm_layout()}}, unless set to \code{NULL}.}

\item{x}{a \code{tmap_arrange} object (returned from \code{tmap_arrange()}).}

\item{options}{options passed on to \code{\link[knitr:knit_print]{knitr::knit_print()}}}

\item{knit}{should \code{\link[knitr:knit_print]{knitr::knit_print()}} be enabled, or the normal \code{\link[base:print]{base::print()}} function?}
}
\description{
Arrange small multiples in a grid layout. Normally, small multiples are created
by specifying multiple variables for one aesthetic or by specifying the by argument
(see \code{\link[=tm_facets]{tm_facets()}}). This function can be used to arrange custom small multiples in a grid layout.
}
\details{
The global option \code{tmap.limits} controls the limit of the number of facets that are plotted.
By default, \code{tmap_options(tmap.limits = c(facets.view=4, facets.plot=64))}.
The maximum number of interactive facets is set to four since otherwise it may become very slow.
}
\examples{
tm1 = tm_shape(World) + tm_polygons("HPI")
tm2 = tm_shape(metro) + tm_bubbles(size = "pop2020")

tmap_arrange(tm1, tm2)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data.R
\docType{data}
\name{World_rivers}
\alias{World_rivers}
\title{Spatial data of rivers}
\format{
An object of class \code{sf} (inherits from \code{data.frame}) with 1632 rows and 5 columns.
}
\source{
\url{https://www.naturalearthdata.com}
}
\usage{
World_rivers
}
\description{
Spatial data of rivers
}
\note{
In tmap <= 3, this dataset was called \code{rivers}.
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tmapSplitShp.R
\name{tmapSplitShp}
\alias{tmapSplitShp}
\title{Internal method that split shape objects}
\usage{
tmapSplitShp(shp, split_stars_dim, smeta)
}
\arguments{
\item{shp}{shape}

\item{split_stars_dim}{name of the dimension to split (\code{""} to skip)}
}
\description{
Internal method that split shape objects. So far, only used to split stars object
(from dimension to attributes)
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tm_scale_bivariate.R
\name{tm_scale_bivariate}
\alias{tm_scale_bivariate}
\title{Scales: bivariate scale}
\usage{
tm_scale_bivariate(
  scale1 = tm_scale(),
  scale2 = tm_scale(),
  values = NA,
  values.repeat = FALSE,
  values.range = NA,
  values.scale = 1,
  value.na = NA,
  value.null = NA,
  value.neutral = NA,
  labels = NULL,
  label.na = NA,
  label.null = NA
)
}
\arguments{
\item{scale1, scale2}{two \code{tm_scale} objects. Currently, all \verb{tm_scale_*()} functions are supported except \code{tm_scale_continuous()}.}

\item{values}{(generic scale argument) The visual values. For colors
(e.g. \code{fill} or \code{col} for \code{tm_polygons()}) this is a palette name from the \code{cols4all} package (see \code{\link[cols4all:c4a]{cols4all::c4a()}})
or vector of colors, for size (e.g. \code{size} for \code{tm_symbols()}) these are a set of sizes (if two values are specified they are interpret as range),
for symbol shapes (e.g. \code{shape} for \code{\link[=tm_symbols]{tm_symbols()}}) these are a set of symbols, etc.
The tmap option \code{values.var} contains the default values per visual variable and in some cases also per data type.}

\item{values.repeat}{(generic scale argument) Should the values be repeated in case there are more categories?}

\item{values.range}{(generic scale argument) Range of the values. Vector of two numbers (both between 0 and 1) where the first determines the minimum and the second the maximum.
Full range, which means that all values are used, is encoded as \code{c(0, 1)}.
For instance, when a grey scale is used for color (from black to white), \code{c(0,1)}
means that all colors are used, \verb{0.25, 0.75} means that only colors from dark grey to light
grey are used (more precisely \code{"grey25"} to \code{"grey75"}), and \verb{0, 0.5} means
that only colors are used from black to middle grey (\code{"grey50"}).
When only one number is specified, this is interpreted as the second number
(where the first is set to 0). Default values can be set via the tmap option \code{values.range}.}

\item{values.scale}{(generic scale argument) Scaling of the values. Only useful for size-related visual variables, such as \code{size} of \code{\link[=tm_symbols]{tm_symbols()}} and \code{lwd} of \code{\link[=tm_lines]{tm_lines()}}.}

\item{value.na}{(generic scale argument) Value used for missing values. See tmap option \code{"value.na"} for defaults per visual variable.}

\item{value.null}{(generic scale argument) Value used for NULL values.
See tmap option \code{"value.null"} for defaults per visual variable. Null data
values occur when out-of-scope features are shown (e.g. for a map of Europe showing a data variable per country,
the null values are applied to countries outside Europe).}

\item{value.neutral}{(generic scale argument) Value that can be considered neutral.
This is used for legends of other visual variables of the same map layer.
E.g. when both \code{fill} and \code{size} are used for \code{\link[=tm_symbols]{tm_symbols()}} (using filled circles),
the size legend items are filled with the \code{value.neutral} color from the \code{fill.scale} scale,
and fill legend items are bubbles of size \code{value.neutral} from the \code{size.scale} scale.}

\item{labels}{(generic scale argument) Labels}

\item{label.na}{(generic scale argument) Label for missing values}

\item{label.null}{(generic scale argument) Label for null (out-of-scope) values}
}
\description{
Scales in tmap are configured by the family of functions with prefix \code{tm_scale}.
Such function should be used for the input of the \code{.scale} arguments in the
layer functions (e.g. \code{fill.scale} in \code{\link[=tm_polygons]{tm_polygons()}}).
The function \code{tm_scale_bivariate()} is used for \code{bivariate.scales}.
}
\seealso{
\code{\link[=tm_scale]{tm_scale()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/qtm.R
\name{qtm}
\alias{qtm}
\title{Quick thematic map plot}
\usage{
qtm(
  shp = NULL,
  fill = tmap::tm_const(),
  col = tmap::tm_const(),
  size = tmap::tm_const(),
  shape = tmap::tm_const(),
  lwd = tmap::tm_const(),
  lty = tmap::tm_const(),
  fill_alpha = tmap::tm_const(),
  col_alpha = tmap::tm_const(),
  text = tmap::tm_const(),
  text_col = tmap::tm_const(),
  text_size = tmap::tm_const(),
  by = NULL,
  scale = NULL,
  title = NULL,
  crs = NULL,
  bbox = NULL,
  basemaps = NA,
  overlays = NA,
  zindex = NA,
  group = NA,
  group.control = "check",
  style = NULL,
  format = NULL,
  ...
)
}
\arguments{
\item{shp}{One of:
\itemize{
\item shape object, which is an object from a class defined by the \code{\link[sf:sf]{sf}}
or \code{\link[stars:st_as_stars]{stars}} package. Objects from the packages \code{sp}
and \code{raster} are also supported, but discouraged.
\item Not specified, i.e. \code{qtm()} is executed. In this case a plain interactive
map is shown.
\item An OpenStreetMap search string, e.g. \code{qtm("Amsterdam")}. In this case a
plain interactive map is shown positioned according to the results of the
search query (from OpenStreetMap nominatim)
}}

\item{fill, col, size, shape, lwd, lty, fill_alpha, col_alpha}{Visual variables.}

\item{text, text_col, text_size}{Visual variables.}

\item{by}{data variable name by which the data is split, or a vector of two
variable names to split the data by two variables (where the first is used
for the rows and the second for the columns). See also \code{\link[=tm_facets]{tm_facets()}}.}

\item{scale}{numeric value that serves as the global scale parameter. All font
sizes, symbol sizes, border widths, and line widths are controlled by this value.
The parameters \code{symbols.size}, \code{text.size}, and \code{lines.lwd} can be scaled
separately with respectively \code{symbols.scale}, \code{text.scale}, and
\code{lines.scale}. See also \code{...}.}

\item{title}{main title. For legend titles, use \code{X.legend}, where X is the
layer name (see \code{...}).}

\item{crs}{Either a \code{\link[sf:st_crs]{crs}} object or a character value
(\code{PROJ.4} character string). By default, the projection is used that is
defined in the \code{shp} object itself.}

\item{bbox}{bounding box. Argument passed on to \code{\link[=tm_shape]{tm_shape()}}}

\item{basemaps}{name(s) of the provider or an URL of a tiled basemap.
It is a shortcut to \code{\link[=tm_basemap]{tm_basemap()}}. Set to \code{NULL} to disable basemaps.
By default, it is set to the tmap option \code{basemaps}.}

\item{overlays}{name(s) of the provider or an URL of a tiled overlay map.
It is a shortcut to \code{\link[=tm_tiles]{tm_tiles()}}.}

\item{zindex}{zindex}

\item{group}{group}

\item{group.control}{group.control}

\item{style}{Layout options (see \code{\link[=tm_layout]{tm_layout()}}) that define the style.
See \code{\link[=tmap_style]{tmap_style()}} for details.}

\item{format}{Deprecated, see \code{\link[=tm_format]{tm_format()}} for alternatives}

\item{...}{arguments associated with the visual variables are passed on
to the layer functions \code{\link[=tm_polygons]{tm_polygons()}}, \code{\link[=tm_lines]{tm_lines()}}, \code{\link[=tm_symbols]{tm_symbols()}},
and \code{\link[=tm_text]{tm_text()}}.
For instance, \code{fill.scale} is the scale specifications of the fill color
of polygons (see \code{\link[=tm_polygons]{tm_polygons()}}).}
}
\value{
A \code{\link{tmap-element}}
}
\description{
Draw a thematic map quickly. This function is a convenient wrapper of the main
plotting method of stacking \code{\link{tmap-element}}s. Without arguments or with a
search term, this functions draws an interactive map.
}
\details{
The first argument is a shape object (normally specified by \code{\link[=tm_shape]{tm_shape()}}).
The next arguments, from \code{fill} to \code{raster}, are the aesthetics from the main
layers. The remaining arguments are related to the map layout. Any argument
from any main layer function, such as \code{\link[=tm_polygons]{tm_polygons()}}, can be specified (see \code{...}).
It is also possible to stack \code{\link{tmap-element}}s on a \code{qtm} plot. See examples.

By default, a scale bar is shown. This option can be set with \code{\link[=tmap_options]{tmap_options()}}
(argument \code{qtm.scalebar}). A minimap is shown by default when \code{qtm} is called
without arguments of with a search term. This option can be set with \code{\link[=tmap_options]{tmap_options()}}
(argument \code{qtm.minimap}).
}
\examples{
data(World, World_rivers, metro)

# just the map
qtm(World)

# choropleth
qtm(World, fill = "economy", style = "cobalt", crs = "+proj=eck4")

qtm(World, col = NULL) +
qtm(metro, size = "pop2010",
	size.legend = tm_legend("Metropolitan Areas"))

# dot map
\dontrun{
current.mode <- tmap_mode("view")
qtm(metro, bbox = "China")
tmap_mode(current.mode) # restore mode
}

\dontrun{
# without arguments, a plain interactive map is shown (the mode is set to view)
qtm()

# search query for OpenStreetMap nominatim
qtm("Amsterdam")
}
}
\references{
Tennekes, M., 2018, {tmap}: Thematic Maps in {R},
Journal of Statistical Software, 84(6), 1-39, \doi{10.18637/jss.v084.i06}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tmap_arrange.R
\name{tmap_arrange}
\alias{tmap_arrange}
\alias{knit_print.tmap_arrange}
\alias{print.tmap_arrange}
\title{Arrange small multiples in grid layout}
\usage{
tmap_arrange(
  ...,
  ncol = NA,
  nrow = NA,
  widths = NA,
  heights = NA,
  sync = FALSE,
  asp = 0,
  outer.margins = 0.02,
  height = NULL
)

\method{knit_print}{tmap_arrange}(x, ..., options = NULL)

\method{print}{tmap_arrange}(x, knit = FALSE, ..., options = NULL)
}
\arguments{
\item{...}{\code{\link{tmap}} objects or one list of \code{\link{tmap}} objects.
The number of multiples that can be plot is limited (see details).}

\item{ncol}{number of columns}

\item{nrow}{number of rows}

\item{widths}{vector of column widths. It should add up to 1 and the length
should be equal to \code{ncol}.}

\item{heights}{vector of row heights. It should add up to 1 and the length
should be equal to \code{nrow}.}

\item{sync}{logical. Should the navigation in view mode (zooming and panning)
be synchronized? By default \code{FALSE}.}

\item{asp}{aspect ratio. The aspect ratio of each map. Normally, this is
controlled by the \code{asp} argument from \code{\link[=tm_layout]{tm_layout()}} (also a tmap option).
This argument will overwrite it, unless set to \code{NULL}.
The default value for \code{asp} is 0, which means that the aspect ratio is
adjusted to the size of the device divided by the number of columns and rows.
When \code{asp} is set to \code{NA}, which is also the default value for \code{tm_layout()},
the aspect ratio will be adjusted to the used shapes.}

\item{outer.margins}{outer.margins, numeric vector four or a single value.
If defines the outer margins for each multiple. If will overwrite the
\code{outer.margins} argument from \code{\link[=tm_layout]{tm_layout()}}, unless set to \code{NULL}.}

\item{height}{height of the arranged output. Only used in the interactive
modes (\code{"view"}, \code{"maplibre"}, \code{"mapbox"}). A bare number is interpreted as
pixels; a string is used as-is (e.g. \code{"80vh"}). The default \code{NULL} fills the
available viewport height.}

\item{x}{a \code{tmap_arrange} object (returned from \code{tmap_arrange()}).}

\item{options}{options passed on to \code{\link[knitr:knit_print]{knitr::knit_print()}}}

\item{knit}{should \code{\link[knitr:knit_print]{knitr::knit_print()}} be enabled, or the normal \code{\link[base:print]{base::print()}} function?}
}
\description{
Arrange small multiples in a grid layout. Normally, small multiples are created
by specifying multiple variables for one aesthetic or by specifying the by argument
(see \code{\link[=tm_facets]{tm_facets()}}). This function can be used to arrange custom small multiples in a grid layout.
}
\details{
The global option \code{tmap.limits} controls the limit of the number of facets that are plotted.
By default, \code{tmap_options(tmap.limits = c(facets.view=4, facets.plot=64))}.
The maximum number of interactive facets is set to four since otherwise it may become very slow.
}
\examples{
tm1 = tm_shape(World) + tm_polygons("HPI")
tm2 = tm_shape(metro) + tm_bubbles(size = "pop2020")

tmap_arrange(tm1, tm2)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tm_scale_bivariate.R
\name{tm_scale_bivariate}
\alias{tm_scale_bivariate}
\title{Scales: bivariate scale}
\usage{
tm_scale_bivariate(
  scale1 = tm_scale(),
  scale2 = tm_scale(),
  values = NA,
  values.repeat = FALSE,
  values.range = NA,
  values.scale = 1,
  value.na = NA,
  value.null = NA,
  value.neutral = NA,
  labels = NULL,
  label.na = NA,
  label.null = NA
)
}
\arguments{
\item{scale1, scale2}{two \code{tm_scale} objects. Currently, all \verb{tm_scale_*()} functions are supported except \code{tm_scale_continuous()}.}

\item{values}{(generic scale argument) The visual values. For colors
(e.g. \code{fill} or \code{col} for \code{tm_polygons()}) this is a palette name from the \code{cols4all} package (see \code{\link[cols4all:c4a]{cols4all::c4a()}})
or vector of colors, for size (e.g. \code{size} for \code{tm_symbols()}) these are a set of sizes (if two values are specified they are interpret as range),
for symbol shapes (e.g. \code{shape} for \code{\link[=tm_symbols]{tm_symbols()}}) these are a set of symbols, etc.
The tmap option \code{values.var} contains the default values per visual variable and in some cases also per data type.}

\item{values.repeat}{(generic scale argument) Should the values be repeated in case there are more categories?}

\item{values.range}{(generic scale argument) Range of the values. Vector of two numbers (both between 0 and 1) where the first determines the minimum and the second the maximum.
Full range, which means that all values are used, is encoded as \code{c(0, 1)}.
For instance, when a grey scale is used for color (from black to white), \code{c(0,1)}
means that all colors are used, \verb{0.25, 0.75} means that only colors from dark grey to light
grey are used (more precisely \code{"grey25"} to \code{"grey75"}), and \verb{0, 0.5} means
that only colors are used from black to middle grey (\code{"grey50"}).
When only one number is specified, this is interpreted as the second number
(where the first is set to 0). Default values can be set via the tmap option \code{values.range}.}

\item{values.scale}{(generic scale argument) Scaling of the values. Only useful for size-related visual variables, such as \code{size} of \code{\link[=tm_symbols]{tm_symbols()}} and \code{lwd} of \code{\link[=tm_lines]{tm_lines()}}.}

\item{value.na}{(generic scale argument) Value used for missing values. See tmap option \code{"value.na"} for defaults per visual variable.}

\item{value.null}{(generic scale argument) Value used for NULL values.
See tmap option \code{"value.null"} for defaults per visual variable. Null data
values occur when out-of-scope features are shown (e.g. for a map of Europe showing a data variable per country,
the null values are applied to countries outside Europe).}

\item{value.neutral}{(generic scale argument) Value that can be considered neutral.
This is used for legends of other visual variables of the same map layer.
E.g. when both \code{fill} and \code{size} are used for \code{\link[=tm_symbols]{tm_symbols()}} (using filled circles),
the size legend items are filled with the \code{value.neutral} color from the \code{fill.scale} scale,
and fill legend items are bubbles of size \code{value.neutral} from the \code{size.scale} scale.}

\item{labels}{(generic scale argument) Labels}

\item{label.na}{(generic scale argument) Label for missing values}

\item{label.null}{(generic scale argument) Label for null (out-of-scope) values}
}
\description{
Scales in tmap are configured by the family of functions with prefix \code{tm_scale}.
Such function should be used for the input of the \code{.scale} arguments in the
layer functions (e.g. \code{fill.scale} in \code{\link[=tm_polygons]{tm_polygons()}}).
The function \code{tm_scale_bivariate()} is used for \code{bivariate.scales}.
}
\seealso{
\href{https://r-tmap.github.io/tmap/articles/basics_scales}{Scales}, \href{https://r-tmap.github.io/tmap/articles/basics_vv}{Visual variables}, \href{https://r-tmap.github.io/tmap/articles/adv_multivariate}{Multivariate visual variables} \href{https://r-tmap.github.io/tmap/articles/basics_legends}{Legends}, \href{https://r-tmap.github.io/tmap/articles/adv_legends}{Legends (advanced)}, \href{https://r-tmap.github.io/tmap/articles/basics_components}{Components}, \href{https://r-tmap.github.io/tmap/articles/adv_positions}{Positioning of components}, \href{https://r-tmap.github.io/tmap/articles/examples_biv_choro}{Bivariate choropleth} \code{\link[=tm_scale]{tm_scale()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tm_layers_aux.R
\name{tm_basemap}
\alias{tm_basemap}
\alias{tm_tiles}
\title{Map layer: basemap / overlay tiles}
\usage{
tm_basemap(
  server = NA,
  alpha = NULL,
  zoom = NULL,
  api = NULL,
  max.native.zoom = 17,
  sub = "abc",
  zindex = 0,
  group = NA,
  group.control = "radio"
)

tm_tiles(
  server = NA,
  alpha = NULL,
  zoom = NULL,
  max.native.zoom = 17,
  sub = "abc",
  zindex = NA,
  group = NA,
  group.control = "check"
)
}
\arguments{
\item{server}{Name of the provider or an URL. Or a vector of multiple values. The list of available providers
can be obtained with \code{providers} (tip: in RStudio, type \verb{leaflet::providers$} to see
the options). See \url{https://leaflet-extras.github.io/leaflet-providers/preview/}
for a preview of those. When a URL is provided, it should be in template format,
e.g. \code{"https://{s}.tile.openstreetmap.org/{z}/{x}/{y}.png"}.
Use \code{NULL} in \code{tm_basemap()} to disable basemaps. It can be a named vector. In that case these names will be used a group names, as alternative to the argument \code{group}.}

\item{alpha}{Transparency level}

\item{zoom}{Zoom level (only used in plot mode)}

\item{api}{API key. Needed for \code{Stadia} and \code{Thunderforest} maps in plot mode. See details}

\item{max.native.zoom}{Maximum native zoom level (only used in view mode).
The minimum and maximum zoom levels are determined in \code{tm_view()}.}

\item{sub}{subdomain of the tile server. Only used when \code{server} is a url template. The
default is \code{"abc"} which works for most tile servers.}

\item{zindex}{zindex of the pane in view mode. By default, it is set to the
layer number plus 400. By default, the tmap layers will therefore be placed
in the custom panes \code{"tmap401"}, \code{"tmap402"}, etc., except for the base tile
layers, which are placed in the standard \code{"tile"}. This parameter determines
both the name of the pane and the z-index, which determines the pane order
from bottom to top. For instance, if \code{zindex} is set to 500, the pane will
be named \code{"tmap500"}.}

\item{group}{Name of the group to which this layer belongs. This is only
relevant in view mode, where layer groups can be switched (see \code{group.control})}

\item{group.control}{In view mode, the group control determines how layer
groups can be switched on and off. Options: \code{"radio"} for radio buttons
(meaning only one group can be shown), \code{"check"} for check boxes
(so multiple groups can be shown), and \code{"none"} for no control
(the group cannot be (de)selected).}
}
\description{
Map layer that draws tiles from a tile server. \code{tm_basemap()} draws the tile
layer as basemap, i.e. as bottom layer. In contrast, \code{tm_tiles()} draws the
tile layer as overlay layer, where the stacking order corresponds with the
order in which this layer is called, just like other map layers.
}
\details{
API keys. For Stadia and Thunderforest maps, an API key is required.
This can be set via the argument \code{api}. Alternatively they can be stored in environment variables \code{"STADIA_MAPS"} and \code{THUNDERFOREST_MAPS} with \code{Sys.setenv}
}
\examples{
\dontrun{
if (requireNamespace("maptiles")) {
	# view mode
	current_mode = tmap_mode("view")
	tm_basemap("Stadia.StamenWatercolor") +
		tm_shape(World) +
		tm_polygons(
		  "HPI",
		  fill.scale = tm_scale(values = "reds"),
		  fill_alpha.scale = 0.5)

	tm_shape(World, crs = "+proj=eqearth") +
		tm_polygons(
			"HPI",
			fill.scale = tm_scale(values = "reds"),
			fill_alpha.scale = 0.5) +
	tm_basemap(NULL)

	# plot mode:
	tmap_mode("plot")
	tm_basemap() +
		tm_shape(World) +
		tm_polygons("HPI")

	tm_basemap("OpenTopoMap") +
		tm_shape(World) +
		tm_polygons(fill = NA, col = "black")

	tm_basemap("CartoDB.PositronNoLabels") +
	tm_shape(NLD_prov, crs = 4236) +
		tm_borders() +
		tm_facets_wrap("name") +
		tm_tiles("CartoDB.PositronOnlyLabels")

	# restore mode
	tmap_mode(current_mode)
}
}
}
\seealso{
\href{https://r-tmap.github.io/tmap/articles/basics_basemaps}{Basemap examples}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/qtm.R
\name{qtm}
\alias{qtm}
\title{Quick thematic map plot}
\usage{
qtm(
  shp = NULL,
  fill = tmap::tm_const(),
  col = tmap::tm_const(),
  size = tmap::tm_const(),
  shape = tmap::tm_const(),
  lwd = tmap::tm_const(),
  lty = tmap::tm_const(),
  fill_alpha = tmap::tm_const(),
  col_alpha = tmap::tm_const(),
  text = tmap::tm_const(),
  text_col = tmap::tm_const(),
  text_size = tmap::tm_const(),
  by = NULL,
  scale = NULL,
  title = NULL,
  crs = NULL,
  bbox = NULL,
  basemaps = NA,
  overlays = NA,
  zindex = NA,
  group = NA,
  group.control = "check",
  style = NULL,
  format = NULL,
  ...
)
}
\arguments{
\item{shp}{One of:
\itemize{
\item shape object, which is an object from a class defined by the \code{\link[sf:sf]{sf}}
or \code{\link[stars:st_as_stars]{stars}} package. Objects from the packages \code{sp}
and \code{raster} are also supported, but discouraged.
\item Not specified, i.e. \code{qtm()} is executed. In this case a plain interactive
map is shown.
\item An OpenStreetMap search string, e.g. \code{qtm("Amsterdam")}. In this case a
plain interactive map is shown positioned according to the results of the
search query (from OpenStreetMap nominatim)
}}

\item{fill, col, size, shape, lwd, lty, fill_alpha, col_alpha}{Visual variables.}

\item{text, text_col, text_size}{Visual variables.}

\item{by}{data variable name by which the data is split, or a vector of two
variable names to split the data by two variables (where the first is used
for the rows and the second for the columns). See also \code{\link[=tm_facets]{tm_facets()}}.}

\item{scale}{numeric value that serves as the global scale parameter. All font
sizes, symbol sizes, border widths, and line widths are controlled by this value.
The parameters \code{symbols.size}, \code{text.size}, and \code{lines.lwd} can be scaled
separately with respectively \code{symbols.scale}, \code{text.scale}, and
\code{lines.scale}. See also \code{...}.}

\item{title}{main title. For legend titles, use \code{X.legend}, where X is the
layer name (see \code{...}).}

\item{crs}{Either a \code{\link[sf:st_crs]{crs}} object or a character value
(\code{PROJ.4} character string). By default, the projection is used that is
defined in the \code{shp} object itself.}

\item{bbox}{bounding box. Argument passed on to \code{\link[=tm_shape]{tm_shape()}}}

\item{basemaps}{name(s) of the provider or an URL of a tiled basemap.
It is a shortcut to \code{\link[=tm_basemap]{tm_basemap()}}. Set to \code{NULL} to disable basemaps.
By default, it is set to the tmap option \code{basemaps}.}

\item{overlays}{name(s) of the provider or an URL of a tiled overlay map.
It is a shortcut to \code{\link[=tm_tiles]{tm_tiles()}}.}

\item{zindex}{Controls the stacking order of map layers. Should be set to a
value above 400. By default, layers are stacked in call order, starting
at 401. See details.}

\item{group}{group}

\item{group.control}{group.control}

\item{style}{Layout options (see \code{\link[=tm_layout]{tm_layout()}}) that define the style.
See \code{\link[=tmap_style]{tmap_style()}} for details.}

\item{format}{Deprecated, see \code{\link[=tm_format]{tm_format()}} for alternatives}

\item{...}{arguments associated with the visual variables are passed on
to the layer functions \code{\link[=tm_polygons]{tm_polygons()}}, \code{\link[=tm_lines]{tm_lines()}}, \code{\link[=tm_symbols]{tm_symbols()}},
and \code{\link[=tm_text]{tm_text()}}.
For instance, \code{fill.scale} is the scale specifications of the fill color
of polygons (see \code{\link[=tm_polygons]{tm_polygons()}}).}
}
\value{
A \code{\link{tmap-element}}
}
\description{
Draw a thematic map quickly. This function is a convenient wrapper of the main
plotting method of stacking \code{\link{tmap-element}}s. Without arguments or with a
search term, this functions draws an interactive map.
}
\details{
The first argument is a shape object (normally specified by \code{\link[=tm_shape]{tm_shape()}}).
The next arguments, from \code{fill} to \code{raster}, are the aesthetics from the main
layers. The remaining arguments are related to the map layout. Any argument
from any main layer function, such as \code{\link[=tm_polygons]{tm_polygons()}}, can be specified (see \code{...}).
It is also possible to stack \code{\link{tmap-element}}s on a \code{qtm} plot. See examples.

By default, a scale bar is shown. This option can be set with \code{\link[=tmap_options]{tmap_options()}}
(argument \code{qtm.scalebar}). A minimap is shown by default when \code{qtm} is called
without arguments of with a search term. This option can be set with \code{\link[=tmap_options]{tmap_options()}}
(argument \code{qtm.minimap}).
\subsection{zindex and pane names}{

In view mode, each layer is rendered in a Leaflet pane named \code{"tmap{zindex}"}
(e.g., \code{"tmap401"}, \code{"tmap402"}), with base tile layers placed in the
standard \code{"tile"} pane.
}
}
\examples{
data(World, World_rivers, metro)

# just the map
qtm(World)

# choropleth
qtm(World, fill = "economy", style = "cobalt", crs = "+proj=eck4")

qtm(World, col = NULL) +
qtm(metro, size = "pop2010",
	size.legend = tm_legend("Metropolitan Areas"))

# dot map
\dontrun{
current.mode <- tmap_mode("view")
qtm(metro, bbox = "China")
tmap_mode(current.mode) # restore mode
}

\dontrun{
# without arguments, a plain interactive map is shown (the mode is set to view)
qtm()

# search query for OpenStreetMap nominatim
qtm("Amsterdam")
}
}
\references{
Tennekes, M., 2018, {tmap}: Thematic Maps in {R},
Journal of Statistical Software, 84(6), 1-39, \doi{10.18637/jss.v084.i06}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tm_layers_aux.R
\name{tm_basemap}
\alias{tm_basemap}
\alias{tm_tiles}
\title{Map layer: basemap / overlay tiles}
\usage{
tm_basemap(
  server = NA,
  alpha = NULL,
  zoom = NULL,
  api = NULL,
  max.native.zoom = 17,
  sub = "abc",
  zindex = 0,
  group = NA,
  group.control = "radio"
)

tm_tiles(
  server = NA,
  alpha = NULL,
  zoom = NULL,
  max.native.zoom = 17,
  sub = "abc",
  zindex = NA,
  group = NA,
  group.control = "check"
)
}
\arguments{
\item{server}{Name of the provider or an URL. Or a vector of multiple values. The list of available providers
can be obtained with \code{providers} (tip: in RStudio, type \verb{leaflet::providers$} to see
the options). See \url{https://leaflet-extras.github.io/leaflet-providers/preview/}
for a preview of those. When a URL is provided, it should be in template format,
e.g. \code{"https://{s}.tile.openstreetmap.org/{z}/{x}/{y}.png"}.
Use \code{NULL} in \code{tm_basemap()} to disable basemaps. It can be a named vector. In that case these names will be used a group names, as alternative to the argument \code{group}.}

\item{alpha}{Transparency level}

\item{zoom}{Zoom level (only used in plot mode)}

\item{api}{API key. Needed for \code{Stadia} and \code{Thunderforest} maps in plot mode, and for \code{Esri}/\code{MapTiler} basemaps in \code{"maplibre"}/\code{"mapbox"} mode (alternatively set the env var \code{ARCGIS_API_KEY} / \code{MAPTILER_API_KEY}). See details}

\item{max.native.zoom}{Maximum native zoom level (only used in view mode).
The minimum and maximum zoom levels are determined in \code{tm_view()}.}

\item{sub}{subdomain of the tile server. Only used when \code{server} is a url template. The
default is \code{"abc"} which works for most tile servers.}

\item{zindex}{Controls the stacking order of map layers. Should be set to a
value above 400. By default, layers are stacked in call order, starting
at 401. See details.}

\item{group}{Name of the group to which this layer belongs. This is only
relevant in view mode, where layer groups can be switched (see \code{group.control})}

\item{group.control}{In view mode, the group control determines how layer
groups can be switched on and off. Options: \code{"radio"} for radio buttons
(meaning only one group can be shown), \code{"check"} for check boxes
(so multiple groups can be shown), and \code{"none"} for no control
(the group cannot be (de)selected).}
}
\description{
Map layer that draws tiles from a tile server. \code{tm_basemap()} draws the tile
layer as basemap, i.e. as bottom layer. In contrast, \code{tm_tiles()} draws the
tile layer as overlay layer, where the stacking order corresponds with the
order in which this layer is called, just like other map layers.
}
\details{
API keys. For Stadia and Thunderforest maps, an API key is required.
This can be set via the argument \code{api}. Alternatively they can be stored in environment variables \code{"STADIA_MAPS"} and \code{THUNDERFOREST_MAPS} with \code{Sys.setenv}

In view mode, each layer is rendered in a Leaflet pane named \code{"tmap{zindex}"}
(e.g., \code{"tmap401"}, \code{"tmap402"}), with base tile layers placed in the
standard \code{"tile"} pane.
}
\examples{
\dontrun{
if (requireNamespace("maptiles")) {
	# view mode
	current_mode = tmap_mode("view")
	tm_basemap("Stadia.StamenWatercolor") +
		tm_shape(World) +
		tm_polygons(
		  "HPI",
		  fill.scale = tm_scale(values = "reds"),
		  fill_alpha.scale = 0.5)

	tm_shape(World, crs = "+proj=eqearth") +
		tm_polygons(
			"HPI",
			fill.scale = tm_scale(values = "reds"),
			fill_alpha.scale = 0.5) +
	tm_basemap(NULL)

	# plot mode:
	tmap_mode("plot")
	tm_basemap() +
		tm_shape(World) +
		tm_polygons("HPI")

	tm_basemap("OpenTopoMap") +
		tm_shape(World) +
		tm_polygons(fill = NA, col = "black")

	tm_basemap("CartoDB.PositronNoLabels") +
	tm_shape(NLD_prov, crs = 4236) +
		tm_borders() +
		tm_facets_wrap("name") +
		tm_tiles("CartoDB.PositronOnlyLabels")

	# restore mode
	tmap_mode(current_mode)
}
}
}
\seealso{
\href{https://r-tmap.github.io/tmap/articles/basics_basemaps}{Basemaps}, \href{https://r-tmap.github.io/tmap/articles/basics_modes}{Modes}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tmap_save.R
\name{tmap_save}
\alias{tmap_save}
\title{Save tmap}
\usage{
tmap_save(
  tm = NULL,
  filename = NA,
  device = NULL,
  width = NA,
  height = NA,
  units = NA,
  dpi = NA,
  outer.margins = NA,
  asp = NULL,
  scale = NA,
  insets_tm = NULL,
  insets_vp = NULL,
  add.titles = TRUE,
  in.iframe = FALSE,
  selfcontained = !in.iframe,
  verbose = NULL,
  ...
)
}
\arguments{
\item{tm}{tmap object}

\item{filename}{filename including extension, and optionally the path.
The extensions pdf, eps, svg, wmf (Windows only), png, jpg, bmp, tiff, and html are supported.
If the extension is missing, the file will be saved as a static plot in \code{"plot"}
mode and as an interactive map (html) in another mode.
The default format for static plots is png, but this can be changed using
the option \code{"output.format"} in \code{\link[=tmap_options]{tmap_options()}}. If \code{NA} (the default),
the file is saved as "tmap01" in the default format, and the number incremented
if the file already exists.}

\item{device}{graphic device to use. Either a device function
(e.g., \code{\link[grDevices:png]{png}} or \code{\link[grDevices:cairo]{cairo_pdf}})
or a text indicating selected graphic device: "pdf", "eps", "svg", "wmf" (Windows only), "png", "jpg", "bmp", "tiff".
If \code{NULL}, the graphic device is guessed based on the \code{filename} argument.}

\item{height, width}{The dimensions of the plot (not applicable for html files).
Units are set with the argument \code{units}. If one of them is not specified,
this is calculated using the formula asp = width / height, where asp is the
estimated aspect ratio of the map. If both are missing, they are set such
that \code{width * height} is equal to the option \code{"output.size"} in \code{\link[=tmap_options]{tmap_options()}}.
This is by default 49, meaning that is the map is a square (so aspect ratio of 1)
both width and height are set to 7.}

\item{units}{units for width and height (\code{"in"}, \code{"cm"}, or \code{"mm"}).
By default, pixels (\code{"px"}) are used if either width or height is set to a
value greater than 50. Else, the units are inches (\code{"in"}).}

\item{dpi}{dots per inch. Only applicable for raster graphics. By default it
is set to 300, but this can be changed using the option \code{"output.dpi"} in \code{\link[=tmap_options]{tmap_options()}}.}

\item{outer.margins}{overrides the outer.margins argument of \code{\link[=tm_options]{tm_options()}} (unless set to \code{NA})}

\item{asp}{if specified, it overrides the asp argument of \code{\link[=tm_options]{tm_options()}}.
\strong{Tip}: set to \code{0} if map frame should be placed on the edges of the image.}

\item{scale}{overrides the scale argument of \code{\link[=tm_options]{tm_options()}} (unless set to \code{NA})}

\item{insets_tm}{tmap object of an inset map, or a list of tmap objects of
multiple inset maps. The number of tmap objects should be equal to the number
of viewports specified with \code{insets_vp}.}

\item{insets_vp}{\code{\link[grid:viewport]{viewport}} of an inset map, or a list
of \code{\link[grid:viewport]{viewport}}s of multiple inset maps. The number of
viewports should be equal to the number of tmap objects specified with \code{insets_tm}.}

\item{add.titles}{add titles to leaflet object.}

\item{in.iframe}{should an interactive map be saved as an iframe?
If so, two HTML files will be saved; one small parent HTML file with the
iframe container, and one large child HTML file with the actual widget.
See \code{\link[widgetframe:saveWidgetframe]{widgetframe::saveWidgetframe()}} for details. By default \code{FALSE},
which means that one large HTML file is saved (see \link[htmlwidgets:saveWidget]{saveWidget()}).}

\item{selfcontained}{when an interactive map is saved, should the resources
(e.g. JavaScript libraries) be contained in the HTML file? If \code{FALSE}, they
are placed in an adjacent directory (see also \code{\link[htmlwidgets:saveWidget]{htmlwidgets::saveWidget()}}).
Note that the HTML file will often still be large when \code{selfcontained = FALSE},
since the map data (polygons and popups), which are also contained in the HTML file,
usually take more space then the map resources.}

\item{verbose}{Deprecated. It is now controlled by the tmap option \code{show.messages}
(see \code{\link[=tmap_options]{tmap_options()}})}

\item{...}{
  Arguments passed on to \code{\link[htmlwidgets:saveWidget]{htmlwidgets::saveWidget}}, \code{\link[widgetframe:saveWidgetframe]{widgetframe::saveWidgetframe}}
  \describe{
    \item{\code{widget}}{Widget to save}
    \item{\code{file}}{File to save HTML into}
    \item{\code{libdir}}{Directory to copy HTML dependencies into (defaults to
filename_files).}
    \item{\code{background}}{Text string giving the html background color of the widget.
Defaults to white.}
    \item{\code{title}}{Text to use as the title of the generated page.}
    \item{\code{knitrOptions}}{A list of \pkg{knitr} chunk options.}
  }}
}
\value{
the filename, invisibly, if export is successful.
}
\description{
Save tmap to a file. This can be either a static plot (e.g. png) or an interactive map (html).
}
\examples{
\dontrun{
	data(NLD_muni, NLD_prov)
	m <- tm_shape(NLD_muni) +
		     tm_fill(col="population", convert2density=TRUE,
                 style="kmeans",
                 title=expression("Population (per " * km^2 * ")")) +
		     tm_borders("black", alpha=.5) +
		 tm_shape(NLD_prov) +
		     tm_borders("grey25", lwd=2) +
		tm_style("classic") +
		tm_format("NLD", inner.margins = c(.02, .15, .06, .15)) +
	    tm_scalebar(position = c("left", "bottom")) +
	    tm_compass(position=c("right", "bottom"))

	tmap_save(m, "choropleth.png", height = 7) # height interpreted in inches
	tmap_save(m, "choropleth_icon.png", height = 100, scale = .1) # height interpreted in pixels

	data(World)
	m2 <- tm_shape(World) +
		tm_fill("well_being", id="name", title="Well-being") +
		tm_format("World")

	# save image
	tmap_save(m2, "World_map.png", width=1920, height=1080, asp=0)

	# cut left inner margin to make sure Antarctica is snapped to frame
	tmap_save(m2 + tm_layout(inner.margins = c(0, -.1, 0.05, 0.01)),
        "World_map2.png", width=1920, height=1080, asp=0)

	# save interactive plot
	tmap_save(m2, "World_map.html")
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tm_inset.R
\name{tm_inset}
\alias{tm_inset}
\title{Map component: inset maps and other objects}
\usage{
tm_inset(
  x = NULL,
  height,
  width,
  margins,
  between_margin,
  position,
  group_id,
  frame,
  frame.color,
  frame.alpha,
  frame.lwd,
  frame.r,
  bg,
  bg.color,
  bg.alpha,
  box_frame,
  box_frame.color,
  box_frame.alpha,
  box_frame.lwd,
  box_frame.lty,
  box_bg,
  box_bg.color,
  box_bg.alpha,
  main_frame,
  main_frame.r,
  main_frame.color,
  main_frame.alpha,
  main_frame.lwd,
  z
)
}
\arguments{
\item{x}{object to draw. Can be: bounding box, tmap object, ggplot2 object, grob object, image file name.}

\item{height}{height of the component in number of text line heights.}

\item{width}{width of the component in number of text line heights.}

\item{margins}{margins}

\item{between_margin}{Margin between}

\item{position}{The position specification of the component: an object created with \code{tm_pos_in()} or \code{tm_pos_out()}. Or, as a shortcut, a vector of two values, specifying the x and y coordinates. The first is \code{"left"}, \code{"center"} or \code{"right"} (or upper case, meaning tighter to the map frame), the second \code{"top"}, \code{"center"} or \code{"bottom"}. Numeric values are also supported, where 0, 0 means left bottom and 1, 1 right top. See also \href{https://r-tmap.github.io/tmap/articles/adv_positions}{vignette about positioning}. In case multiple components should be combined (stacked), use \code{group_id} and specify \code{component} in \code{\link[=tm_components]{tm_components()}}.}

\item{group_id}{Component group id name. All components (e.g. legends, titles, etc) with the same \code{group_id} will be grouped. The specifications of how they are placed (e.g. stacking, margins etc.) are determined in \code{\link[=tm_components]{tm_components()}} where its argument \code{id} should correspond to \code{group_id}.}

\item{frame}{frame should a frame be drawn?}

\item{frame.color}{frame color}

\item{frame.alpha}{frame alpha transparancy}

\item{frame.lwd}{frame line width}

\item{frame.r}{Radius of the rounded frame corners. 0 means no rounding.}

\item{bg}{Show background?}

\item{bg.color}{Background color}

\item{bg.alpha}{Background transparency}

\item{box_frame}{Should a box frame be drawn in the main map that shows where the inset is? \code{TRUE} by default}

\item{box_frame.color, box_frame.alpha, box_frame.lwd, box_frame.lty}{Properties of the box frame}

\item{box_bg}{Should the frame box have a background? \code{FALSE} by default}

\item{box_bg.color, box_bg.alpha}{Properties of the box background}

\item{main_frame}{Should a frame be drawn around the inset map? Note that this is different from the general map component frame (the argument \code{frame})}

\item{main_frame.r, main_frame.color, main_frame.alpha, main_frame.lwd}{Properties of the main frame}

\item{z}{z index, e.g. the place of the component relative to the other componets}
}
\description{
Map component that adds an inset object, e.g. a mini map
}
\examples{
## map
bb = tmaptools::bb(NLD_prov[NLD_prov$name == "Utrecht",], ext = 1.05)

bb_Randstad =
  sf::st_bbox(c(xmin = 120000, xmax = 150000, ymin = 460000, ymax = 500000), crs = 28992)

tm_shape(NLD_dist) +
	tm_polygons(
		fill = "dwelling_value",
		fill.scale = tm_scale_continuous_pseudo_log(values = "-cols4all.pu_gn_div"),
		col = NULL) +
tm_shape(NLD_muni) +
	tm_borders(col = "black", lwd = 0.5) +
tm_shape(NLD_prov) +
	tm_borders(col = "black", lwd = 1.5) +
tm_inset(bb_Randstad, height = 12, width = 12, position = c("left", "top")) +
	tm_compass(position = c("left", "top"), )

## ggplot2
if (requireNamespace("ggplot2")) {
	library(ggplot2)
	p = ggplot(World, aes(x = gender, y = press, colour = continent)) +
		geom_point() +
		theme_bw()

	tm_shape(World) +
		tm_polygons(
			fill = "gender",
			fill.scale = tm_scale(values = "-cols4all.pu_gn_div")) +
		tm_inset(p, height = 15, width = 20, position = tm_pos_in("left", "bottom"))
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tm_component_functions.R
\name{tm_scalebar}
\alias{tm_scalebar}
\title{Map component: scale bar}
\usage{
tm_scalebar(
  breaks,
  width,
  allow_clipping,
  text.size,
  text.color,
  color.dark,
  color.light,
  lwd,
  position,
  group_id,
  bg,
  bg.color,
  bg.alpha,
  size = "deprecated",
  stack,
  frame,
  frame.color,
  frame.alpha,
  frame.lwd,
  frame.r,
  margins,
  z
)
}
\arguments{
\item{breaks}{breaks. E.g. \code{c(0, 10, 50)} places scale bar breaks at 0, 10, and 50 units. These units are specified in \code{\link[=tm_shape]{tm_shape()}}.}

\item{width}{width of the scale bar. Units are number of text line heights, which is similar to the number of characters. In case \code{beaks} are specified, the \code{width} is only handy to finetune the approximated width, e.g. in case clipping of the labels occurs, or there is too much whitespace.}

\item{allow_clipping}{should clipping of the last label by allowed? If \code{TRUE} (default), the last break label including unit is printed even when it doesn't fit the frame. If \code{FALSE} it will not be printed. Instead the unit suffix is added to the second last label.}

\item{text.size}{text size}

\item{text.color}{text.color}

\item{color.dark}{color.dark}

\item{color.light}{color.light}

\item{lwd}{linewidth}

\item{position}{The position specification of the component: an object created with \code{tm_pos_in()} or \code{tm_pos_out()}. Or, as a shortcut, a vector of two values, specifying the x and y coordinates. The first is \code{"left"}, \code{"center"} or \code{"right"} (or upper case, meaning tighter to the map frame), the second \code{"top"}, \code{"center"} or \code{"bottom"}. Numeric values are also supported, where 0, 0 means left bottom and 1, 1 right top. See also \href{https://r-tmap.github.io/tmap/articles/adv_positions}{vignette about positioning}. In case multiple components should be combined (stacked), use \code{group_id} and specify \code{component} in \code{\link[=tm_components]{tm_components()}}.}

\item{group_id}{Component group id name. All components (e.g. legends, titles, etc) with the same \code{group_id} will be grouped. The specifications of how they are placed (e.g. stacking, margins etc.) are determined in \code{\link[=tm_components]{tm_components()}} where its argument \code{id} should correspond to \code{group_id}.}

\item{bg}{Show background?}

\item{bg.color}{Background color}

\item{bg.alpha}{Background transparency}

\item{size}{Deprecated (use \code{text.size} instead)}

\item{stack}{stack with other map components, either \code{"vertical"} or \code{"horizontal"}.}

\item{frame}{frame should a frame be drawn?}

\item{frame.color}{frame color}

\item{frame.alpha}{frame alpha transparancy}

\item{frame.lwd}{frame line width}

\item{frame.r}{Radius of the rounded frame corners. 0 means no rounding.}

\item{margins}{margins}

\item{z}{z index, e.g. the place of the component relative to the other componets}
}
\description{
Map component that adds a scale bar.
}
\seealso{
\href{https://r-tmap.github.io/tmap/articles/basics_components}{Vignette about components}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tmap_save.R
\name{tmap_save}
\alias{tmap_save}
\title{Save tmap}
\usage{
tmap_save(
  tm = NULL,
  filename = NA,
  device = NULL,
  width = NA,
  height = NA,
  units = NA,
  dpi = NA,
  outer.margins = NA,
  asp = NULL,
  scale = NA,
  insets_tm = NULL,
  insets_vp = NULL,
  add.titles = TRUE,
  in.iframe = FALSE,
  selfcontained = !in.iframe,
  verbose = NULL,
  ...
)
}
\arguments{
\item{tm}{tmap object}

\item{filename}{filename including extension, and optionally the path.
The extensions pdf, eps, svg, wmf (Windows only), png, jpg, bmp, tiff, and html are supported.
If the extension is missing, the file will be saved as a static plot in \code{"plot"}
mode and as an interactive map (html) in another mode.
The default format for static plots is png, but this can be changed using
the option \code{"output.format"} in \code{\link[=tmap_options]{tmap_options()}}. If \code{NA} (the default),
the file is saved as "tmap01" in the default format, and the number incremented
if the file already exists.}

\item{device}{graphic device to use. Either a device function
(e.g., \code{\link[grDevices:png]{png}} or \code{\link[grDevices:cairo_pdf]{cairo_pdf}})
or a text indicating selected graphic device: "pdf", "eps", "svg", "wmf" (Windows only), "png", "jpg", "bmp", "tiff".
If \code{NULL}, the graphic device is guessed based on the \code{filename} argument.}

\item{height, width}{The dimensions of the plot (not applicable for html files).
Units are set with the argument \code{units}. If one of them is not specified,
this is calculated using the formula asp = width / height, where asp is the
estimated aspect ratio of the map. If both are missing, they are set such
that \code{width * height} is equal to the option \code{"output.size"} in \code{\link[=tmap_options]{tmap_options()}}.
This is by default 49, meaning that is the map is a square (so aspect ratio of 1)
both width and height are set to 7.}

\item{units}{units for width and height (\code{"in"}, \code{"cm"}, or \code{"mm"}).
By default, pixels (\code{"px"}) are used if either width or height is set to a
value greater than 50. Else, the units are inches (\code{"in"}).}

\item{dpi}{dots per inch. Only applicable for raster graphics. By default it
is set to 300, but this can be changed using the option \code{"output.dpi"} in \code{\link[=tmap_options]{tmap_options()}}.}

\item{outer.margins}{overrides the outer.margins argument of \code{\link[=tm_options]{tm_options()}} (unless set to \code{NA})}

\item{asp}{if specified, it overrides the asp argument of \code{\link[=tm_options]{tm_options()}}.
\strong{Tip}: set to \code{0} if map frame should be placed on the edges of the image.}

\item{scale}{overrides the scale argument of \code{\link[=tm_options]{tm_options()}} (unless set to \code{NA})}

\item{insets_tm}{tmap object of an inset map, or a list of tmap objects of
multiple inset maps. The number of tmap objects should be equal to the number
of viewports specified with \code{insets_vp}.}

\item{insets_vp}{\code{\link[grid:viewport]{viewport}} of an inset map, or a list
of \code{\link[grid:viewport]{viewport}}s of multiple inset maps. The number of
viewports should be equal to the number of tmap objects specified with \code{insets_tm}.}

\item{add.titles}{add titles to leaflet object.}

\item{in.iframe}{should an interactive map be saved as an iframe?
If so, two HTML files will be saved; one small parent HTML file with the
iframe container, and one large child HTML file with the actual widget.
See \code{\link[widgetframe:saveWidgetframe]{widgetframe::saveWidgetframe()}} for details. By default \code{FALSE},
which means that one large HTML file is saved (see \link[htmlwidgets:saveWidget]{saveWidget()}).}

\item{selfcontained}{when an interactive map is saved, should the resources
(e.g. JavaScript libraries) be contained in the HTML file? If \code{FALSE}, they
are placed in an adjacent directory (see also \code{\link[htmlwidgets:saveWidget]{htmlwidgets::saveWidget()}}).
Note that the HTML file will often still be large when \code{selfcontained = FALSE},
since the map data (polygons and popups), which are also contained in the HTML file,
usually take more space then the map resources.}

\item{verbose}{Deprecated. It is now controlled by the tmap option \code{show.messages}
(see \code{\link[=tmap_options]{tmap_options()}})}

\item{...}{additional arguments passed on to the underlying output function
(see Details). For interactive (html) maps these are passed to
\code{\link[htmlwidgets:saveWidget]{htmlwidgets::saveWidget()}} or \code{\link[widgetframe:saveWidgetframe]{widgetframe::saveWidgetframe()}}. For static
maps they are passed to the graphic device; the most useful are:
\describe{
\item{\code{colormodel} (pdf, eps)}{Color model of the output. The default is
\code{"srgb"}; use \code{colormodel = "cmyk"} for print workflows that require
CMYK. See \code{\link[grDevices:pdf]{grDevices::pdf()}}.}
\item{\code{compression} (tiff)}{Compression method, e.g. \code{"lzw"}, \code{"zip"}, or
\code{"jpeg"}. The default \code{"none"} can produce very large files; the
lossless \code{"lzw"} or \code{"zip"} is usually a good choice for publication.}
\item{\code{quality} (jpg)}{JPEG quality, from \code{0} to \code{100} (default \code{75}).}
\item{\code{bg} (all)}{Background color. Use \code{bg = "transparent"} for a
transparent background (png, tiff, and svg support transparency; jpg and
bmp do not).}
\item{\code{pointsize} (all)}{Default point size of text, which scales the text
drawn by the device.}
\item{\code{family} (pdf, eps, svg), \code{fonts} (pdf, eps)}{Font family, and fonts
to embed in the file.}
\item{\code{antialias} (png, jpg, bmp, tiff)}{Type of anti-aliasing, e.g.
\code{"default"}, \code{"none"}, \code{"gray"}, or \code{"subpixel"}. This only applies to
cairo-type devices. \code{png} always uses cairo (\code{type = "cairo-png"}), but
\code{jpg}, \code{bmp}, and \code{tiff} use the platform default device type, so on a
build without cairo support \code{antialias} is ignored for those formats.}
}}
}
\value{
the filename, invisibly, if export is successful.
}
\description{
Save tmap to a file. This can be either a static plot (e.g. png) or an interactive map (html).
}
\details{
Extra arguments (\code{...}) are passed on to the function that writes the
output. For interactive (html) maps this is \code{\link[htmlwidgets:saveWidget]{htmlwidgets::saveWidget()}}, or
\code{\link[widgetframe:saveWidgetframe]{widgetframe::saveWidgetframe()}} when \code{in.iframe = TRUE}. For static maps it
is the graphic device: the function supplied to \code{device}, or, when
\code{device = NULL}, the default device for the file extension:
\code{\link[grDevices:pdf]{grDevices::pdf()}} (\code{"pdf"}), \code{\link[grDevices:postscript]{grDevices::postscript()}} (\code{"eps"}),
\code{\link[grDevices:svg]{grDevices::svg()}} (\code{"svg"}), \code{\link[grDevices:png]{grDevices::png()}} (\code{"png"}),
\code{\link[grDevices:jpeg]{grDevices::jpeg()}} (\code{"jpg"}/\code{"jpeg"}), \code{\link[grDevices:bmp]{grDevices::bmp()}} (\code{"bmp"}), and
\code{\link[grDevices:tiff]{grDevices::tiff()}} (\code{"tiff"}). The most useful device arguments are listed
under \code{...}.

Note that \code{filename}, \code{width}, and \code{height} are always set by \code{tmap_save},
as are \code{res} and \code{units} for the raster formats (png, jpg, bmp, tiff); these
cannot be overridden through \code{...}. Use the \code{dpi}, \code{width}, \code{height}, and
\code{units} arguments of \code{tmap_save} instead.
}
\examples{
\dontrun{
	data(NLD_muni, NLD_prov)
	m <- tm_shape(NLD_muni) +
		     tm_fill(col="population", convert2density=TRUE,
                 style="kmeans",
                 title=expression("Population (per " * km^2 * ")")) +
		     tm_borders("black", alpha=.5) +
		 tm_shape(NLD_prov) +
		     tm_borders("grey25", lwd=2) +
		tm_style("classic") +
		tm_format("NLD", inner.margins = c(.02, .15, .06, .15)) +
	    tm_scalebar(position = c("left", "bottom")) +
	    tm_compass(position=c("right", "bottom"))

	tmap_save(m, "choropleth.png", height = 7) # height interpreted in inches
	tmap_save(m, "choropleth_icon.png", height = 100, scale = .1) # height interpreted in pixels

	data(World)
	m2 <- tm_shape(World) +
		tm_fill("well_being", id="name", title="Well-being") +
		tm_format("World")

	# save image
	tmap_save(m2, "World_map.png", width=1920, height=1080, asp=0)

	# cut left inner margin to make sure Antarctica is snapped to frame
	tmap_save(m2 + tm_layout(inner.margins = c(0, -.1, 0.05, 0.01)),
        "World_map2.png", width=1920, height=1080, asp=0)

	# save interactive plot
	tmap_save(m2, "World_map.html")
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tm_inset.R
\name{tm_inset}
\alias{tm_inset}
\title{Map component: inset maps and other objects}
\usage{
tm_inset(
  x = NULL,
  height,
  width,
  margins,
  between_margin,
  position,
  group_id,
  frame,
  frame.color,
  frame.alpha,
  frame.lwd,
  frame.r,
  bg,
  bg.color,
  bg.alpha,
  box_frame,
  box_frame.color,
  box_frame.alpha,
  box_frame.lwd,
  box_frame.lty,
  box_bg,
  box_bg.color,
  box_bg.alpha,
  main_frame,
  main_frame.r,
  main_frame.color,
  main_frame.alpha,
  main_frame.lwd,
  z
)
}
\arguments{
\item{x}{object to draw. Can be: bounding box, tmap object, ggplot2 object, grob object, image file name.}

\item{height}{height of the component in number of text line heights.}

\item{width}{width of the component in number of text line heights.}

\item{margins}{margins}

\item{between_margin}{Margin between}

\item{position}{The position specification of the component: an object created with \code{tm_pos_in()} or \code{tm_pos_out()}. Or, as a shortcut, a vector of two values, specifying the x and y coordinates. The first is \code{"left"}, \code{"center"} or \code{"right"} (or upper case, meaning tighter to the map frame), the second \code{"top"}, \code{"center"} or \code{"bottom"}. Numeric values are also supported, where 0, 0 means left bottom and 1, 1 right top. See also vignette: \href{https://r-tmap.github.io/tmap/articles/adv_positions}{Positioning of components}. In case multiple components should be combined (stacked), use \code{group_id} and specify \code{component} in \code{\link[=tm_components]{tm_components()}}.}

\item{group_id}{Component group id name. All components (e.g. legends, titles, etc) with the same \code{group_id} will be grouped. The specifications of how they are placed (e.g. stacking, margins etc.) are determined in \code{\link[=tm_components]{tm_components()}} where its argument \code{id} should correspond to \code{group_id}.}

\item{frame}{frame should a frame be drawn?}

\item{frame.color}{frame color}

\item{frame.alpha}{frame alpha transparancy}

\item{frame.lwd}{frame line width}

\item{frame.r}{Radius of the rounded frame corners. 0 means no rounding.}

\item{bg}{Show background?}

\item{bg.color}{Background color}

\item{bg.alpha}{Background transparency}

\item{box_frame}{Should a box frame be drawn in the main map that shows where the inset is? \code{TRUE} by default}

\item{box_frame.color, box_frame.alpha, box_frame.lwd, box_frame.lty}{Properties of the box frame}

\item{box_bg}{Should the frame box have a background? \code{FALSE} by default}

\item{box_bg.color, box_bg.alpha}{Properties of the box background}

\item{main_frame}{Should a frame be drawn around the inset map? Note that this is different from the general map component frame (the argument \code{frame})}

\item{main_frame.r, main_frame.color, main_frame.alpha, main_frame.lwd}{Properties of the main frame}

\item{z}{z index, e.g. the place of the component relative to the other componets}
}
\description{
Map component that adds an inset object, e.g. a mini map
}
\examples{
## map
bb = tmaptools::bb(NLD_prov[NLD_prov$name == "Utrecht",], ext = 1.05)

bb_Randstad =
  sf::st_bbox(c(xmin = 120000, xmax = 150000, ymin = 460000, ymax = 500000), crs = 28992)

tm_shape(NLD_dist) +
	tm_polygons(
		fill = "dwelling_value",
		fill.scale = tm_scale_continuous_pseudo_log(values = "-cols4all.pu_gn_div"),
		col = NULL) +
tm_shape(NLD_muni) +
	tm_borders(col = "black", lwd = 0.5) +
tm_shape(NLD_prov) +
	tm_borders(col = "black", lwd = 1.5) +
tm_inset(bb_Randstad, height = 12, width = 12, position = c("left", "top")) +
	tm_compass(position = c("left", "top"), )

## ggplot2
if (requireNamespace("ggplot2")) {
	library(ggplot2)
	p = ggplot(World, aes(x = gender, y = press, colour = continent)) +
		geom_point() +
		theme_bw()

	tm_shape(World) +
		tm_polygons(
			fill = "gender",
			fill.scale = tm_scale(values = "-cols4all.pu_gn_div")) +
		tm_inset(p, height = 15, width = 20, position = tm_pos_in("left", "bottom"))
}
}
\seealso{
\href{https://r-tmap.github.io/tmap/articles/adv_inset_maps}{Inset maps}, \href{https://r-tmap.github.io/tmap/articles/adv_ggplot2}{ggplot2 integration} \href{https://r-tmap.github.io/tmap/articles/basics_components}{Components}, \href{https://r-tmap.github.io/tmap/articles/adv_positions}{Positioning of components}, \href{https://r-tmap.github.io/tmap/articles/adv_comp_group}{Grouping of components}, \href{https://r-tmap.github.io/tmap/articles/examples_choro_NLD}{Choropleth (Netherlands)}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tm_component_functions.R
\name{tm_scalebar}
\alias{tm_scalebar}
\title{Map component: scale bar}
\usage{
tm_scalebar(
  breaks,
  width,
  allow_clipping,
  text.size,
  text.color,
  color.dark,
  color.light,
  lwd,
  position,
  group_id,
  bg,
  bg.color,
  bg.alpha,
  size = "deprecated",
  stack,
  frame,
  frame.color,
  frame.alpha,
  frame.lwd,
  frame.r,
  margins,
  z
)
}
\arguments{
\item{breaks}{Scale bar break positions. E.g. \code{c(0, 10, 50)} places breaks
at 0, 10, and 50 units. The unit is controlled by the \code{unit} argument from \code{\link[=tm_shape]{tm_shape()}}.
When \code{NULL} (default), break positions are chosen automatically.}

\item{width}{Width of the scale bar, in number of text line heights (roughly
equivalent to character widths). When \code{breaks} are specified, \code{width} is
only useful for fine-tuning, e.g. to prevent label clipping or reduce
excess whitespace.}

\item{allow_clipping}{Should clipping of the last label be allowed? If
\code{TRUE} (default), the last break label including its unit suffix is printed
even when it extends beyond the frame. If \code{FALSE}, that label is suppressed
and the unit suffix is appended to the second-to-last label instead.}

\item{text.size}{text size}

\item{text.color}{text.color}

\item{color.dark}{color.dark}

\item{color.light}{color.light}

\item{lwd}{linewidth}

\item{position}{The position specification of the component: an object created with \code{tm_pos_in()} or \code{tm_pos_out()}. Or, as a shortcut, a vector of two values, specifying the x and y coordinates. The first is \code{"left"}, \code{"center"} or \code{"right"} (or upper case, meaning tighter to the map frame), the second \code{"top"}, \code{"center"} or \code{"bottom"}. Numeric values are also supported, where 0, 0 means left bottom and 1, 1 right top. See also vignette: \href{https://r-tmap.github.io/tmap/articles/adv_positions}{Positioning of components}. In case multiple components should be combined (stacked), use \code{group_id} and specify \code{component} in \code{\link[=tm_components]{tm_components()}}.}

\item{group_id}{Component group id name. All components (e.g. legends, titles, etc) with the same \code{group_id} will be grouped. The specifications of how they are placed (e.g. stacking, margins etc.) are determined in \code{\link[=tm_components]{tm_components()}} where its argument \code{id} should correspond to \code{group_id}.}

\item{bg}{Show background?}

\item{bg.color}{Background color}

\item{bg.alpha}{Background transparency}

\item{size}{Deprecated (use \code{text.size} instead)}

\item{stack}{stack with other map components, either \code{"vertical"} or \code{"horizontal"}.}

\item{frame}{frame should a frame be drawn?}

\item{frame.color}{frame color}

\item{frame.alpha}{frame alpha transparancy}

\item{frame.lwd}{frame line width}

\item{frame.r}{Radius of the rounded frame corners. 0 means no rounding.}

\item{margins}{margins}

\item{z}{z index, e.g. the place of the component relative to the other componets}
}
\description{
Map component that adds a scale bar.
}
\seealso{
\href{https://r-tmap.github.io/tmap/articles/basics_components}{Components}, \href{https://r-tmap.github.io/tmap/articles/adv_positions}{Positioning of components}, \href{https://r-tmap.github.io/tmap/articles/adv_comp_group}{Grouping of components}, \href{https://r-tmap.github.io/tmap/articles/examples_choro_NLD}{Choropleth (Netherlands)}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tm_component_functions.R
\name{tm_compass}
\alias{tm_compass}
\title{Map component: compass}
\usage{
tm_compass(
  north,
  type,
  text.size,
  size,
  show.labels,
  cardinal.directions,
  text.color,
  color.dark,
  color.light,
  lwd,
  position,
  group_id,
  bg,
  bg.color,
  bg.alpha,
  stack,
  just,
  frame,
  frame.color,
  frame.alpha,
  frame.lwd,
  frame.r,
  margins,
  z,
  ...
)
}
\arguments{
\item{north}{north}

\item{type}{compass type, one of: \code{"arrow"}, \code{"4star"}, \code{"8star"}, \code{"radar"}, \code{"rose"}. The default is controlled by \code{\link{tm_layout}} (which uses \code{"arrow"} for the default style)}

\item{text.size}{text.size}

\item{size}{size}

\item{show.labels}{show.labels}

\item{cardinal.directions}{cardinal.directions}

\item{text.color}{text.color}

\item{color.dark}{color.dark}

\item{color.light}{color.light}

\item{lwd}{lwd}

\item{position}{The position specification of the component: an object created with \code{tm_pos_in()} or \code{tm_pos_out()}. Or, as a shortcut, a vector of two values, specifying the x and y coordinates. The first is \code{"left"}, \code{"center"} or \code{"right"} (or upper case, meaning tighter to the map frame), the second \code{"top"}, \code{"center"} or \code{"bottom"}. Numeric values are also supported, where 0, 0 means left bottom and 1, 1 right top. See also \href{https://r-tmap.github.io/tmap/articles/adv_positions}{vignette about positioning}. In case multiple components should be combined (stacked), use \code{group_id} and specify \code{component} in \code{\link[=tm_components]{tm_components()}}.}

\item{group_id}{Component group id name. All components (e.g. legends, titles, etc) with the same \code{group_id} will be grouped. The specifications of how they are placed (e.g. stacking, margins etc.) are determined in \code{\link[=tm_components]{tm_components()}} where its argument \code{id} should correspond to \code{group_id}.}

\item{bg}{Show background?}

\item{bg.color}{Background color}

\item{bg.alpha}{Background transparency}

\item{stack}{stack with other map components, either \code{"vertical"} or \code{"horizontal"}.}

\item{just}{just}

\item{frame}{frame should a frame be drawn?}

\item{frame.color}{frame color}

\item{frame.alpha}{frame alpha transparancy}

\item{frame.lwd}{frame line width}

\item{frame.r}{Radius of the rounded frame corners. 0 means no rounding.}

\item{margins}{margins}

\item{z}{z index, e.g. the place of the component relative to the other componets}

\item{...}{to catch deprecated arguments (alpha)}
}
\description{
Map component that adds a compass
}
\seealso{
\href{https://r-tmap.github.io/tmap/articles/basics_components}{Vignette about components}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tm_popup.R
\name{tm_popup}
\alias{tm_popup}
\title{Popup specification for interactive maps}
\usage{
tm_popup(
  vars = NA,
  title = NA,
  format = tm_label_format(),
  width = "auto",
  max.height = "25em",
  title.align = "left",
  title.color = NULL,
  label.align = "left",
  label.color = "#888888",
  value.align = "right",
  value.color = NULL,
  css = NULL
)
}
\arguments{
\item{vars}{Names of the data variables that are shown in the popup table.
A (possibly named) character vector; when named, the names are used as
labels in the popup table. Besides a character vector, the following
special values are supported (identical to the former \code{popup.vars}
argument):
\describe{
\item{\code{TRUE}}{show all variables of the shape object;}
\item{\code{FALSE}}{disable popups;}
\item{\code{NA} (default)}{automatic: if visual variables (e.g. \code{fill}) are
used, only those are shown, otherwise all variables of the shape
object are shown.}
}}

\item{title}{Name of the data variable used as the popup title (the bold
header shown above the popup table). This overrules the layer argument
\code{id}, analogous to how \code{hover} overrules \code{id} for hover labels. The
default (\code{NA}) means that the popup title is derived from \code{id} (the former,
and still default, behaviour). A length-one character vector is expected;
a named value is allowed and reserved for future use.}

\item{format}{A list of formatting options for the popup values, the output
of \code{\link[=tm_label_format]{tm_label_format()}}. Only applicable to numeric data variables. If one
list of formatting options is provided, it is applied to all numeric
variables of \code{vars}. A (named) list of lists can also be provided; in that
case, each list of formatting options is applied to the named variable.}

\item{width}{Width of the popup content (view mode). A bare number is
interpreted as pixels (e.g. \code{300} means \code{"300px"}); a character string is
used as-is, so any CSS length is accepted (\code{"300px"}, \code{"20em"}, \code{"50\%"}).
The default \code{"auto"} lets the popup size to its content.}

\item{max.height}{Maximum height of the popup table before it becomes
vertically scrollable (view mode). A bare number is interpreted as \code{em}
(e.g. \code{5} means \code{"5em"}, roughly "show 5 lines"); a character string is
used as-is. Default \code{"25em"}. Use \code{max.height = "none"} (or \code{NA}/\code{Inf}) to
remove the cap, so the popup grows to fit its content and never scrolls.}

\item{title.align, label.align, value.align}{Horizontal alignment of the
popup title (bold header), the variable-name (label) column, and the value
column respectively. Each one of \code{"left"}, \code{"center"}, or \code{"right"}.
Defaults: title \code{"left"}, label \code{"left"}, value \code{"right"}.}

\item{title.color, label.color, value.color}{Text color of the popup title,
the label column, and the value column respectively. \code{NULL} (the default
for title and value) inherits the browser/popup default; \code{label.color}
defaults to \code{"#888888"} (grey).}

\item{css}{Optional free-form CSS, injected verbatim as a \verb{<style>} block in
each popup, for full restyling beyond the arguments above. Target the
semantic classes \code{.tmap-popup} (container), \code{.tmap-popup-table},
\code{.tmap-popup-title}, \code{.tmap-popup-label}, and \code{.tmap-popup-value}. To size
the popup box itself, target the backend's own popup element
(\code{.leaflet-popup-content} in view mode, \code{.maplibregl-popup-content} in
maplibre mode). Default \code{NULL}.}
}
\value{
A \code{tm_popup} object.
}
\description{
\code{tm_popup()} specifies the popups that are shown in interactive (\code{"view"})
mode when a feature is clicked. It is passed to the \code{popup} argument of the
layer functions (\code{\link[=tm_polygons]{tm_polygons()}}, \code{\link[=tm_symbols]{tm_symbols()}}, \code{\link[=tm_lines]{tm_lines()}}, etc.). It
replaces the (now deprecated) layer arguments \code{popup.vars} and \code{popup.format}.
}
\seealso{
\code{\link[=tm_polygons]{tm_polygons()}}, \code{\link[=tm_symbols]{tm_symbols()}}, \code{\link[=tm_lines]{tm_lines()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tm_shape.R
\name{tm_shape}
\alias{tm_shape}
\title{Shape (spatial object) specification}
\usage{
tm_shape(
  shp = NULL,
  bbox = NULL,
  crs = NULL,
  is.main = NA,
  layer = NULL,
  name = NULL,
  unit = NULL,
  filter = NULL,
  ...
)
}
\arguments{
\item{shp}{Spatial data object. Typically an object from \pkg{sf}, \pkg{terra}, or \pkg{stars}.
Additional spatial data types can be supported via extension packages, such as
\pkg{tmap.networks} and \pkg{tmap.sources} (experimental). These may include, for example,
remote or streaming data sources.}

\item{bbox}{Bounding box of the map. Only used when \code{shp} is the main shape (see \code{is.main}).
Three options are supported:
\itemize{
\item a \code{\link[sf:st_bbox]{sf::st_bbox()}} object,
\item a character string specifying a location, passed to \code{\link[tmaptools:geocode_OSM]{tmaptools::geocode_OSM()}},
\item \code{"FULL"}, which represents the whole earth. This option ensures that reprojection
retains the full global extent, unlike a regular bounding box.
}}

\item{crs}{Map projection (CRS). Can be set to an \code{crs} object (see \code{\link[sf:st_crs]{sf::st_crs()}}), a proj4string, an EPSG number, the value \code{"auto"} (automatic crs recommendation), or one the the following generic projections: \code{c("laea", "aeqd", "utm", "pconic", "eqdc", "stere")}. See details.}

\item{is.main}{Is \code{shp} the main shape, which determines the crs and
bounding box of the map? By default, \code{TRUE} if it is the first \code{tm_shape} call}

\item{layer}{Name of the layer to use. This is primarily relevant for multi-layer or remote
data sources (e.g. PMTiles or vector tiles), where multiple layers may be available.}

\item{name}{of the spatial object}

\item{unit}{Unit of the coordinates}

\item{filter}{Optional filter expression used to subset features. The exact syntax depends
on the data source. For in-memory objects (e.g. \pkg{sf}), this is typically evaluated in R,
whereas for remote sources it may be translated to a query and executed on the server side.}

\item{...}{passed on to \code{\link[tmaptools:bb]{bb}} (e.g. \code{ext} can be used to enlarge or shrink a bounding box)}
}
\description{
Specify a shape, which is a spatial object from one of these spatial object
class packages: \code{\link[sf:sf]{sf}}, \code{\link[stars:st_as_stars]{stars}}, or \code{terra}.
}
\details{
The map projection (\code{crs}) determines in which coordinate system the spatial object is processed and plotted.
See \href{https://r-tmap.github.io/tmap/articles/foundations_crs}{vignette about CRS}. The \code{crs} can be specified in two places: 1) \code{tm_shape()} and \code{tm_crs()}.
In both cases, the map is plotted into the specified \code{crs}. The difference is that in the first option, the \code{crs} is also taken into account in spatial transformation functions, such as the calculation of centroids and cartograms. In the second option, the \code{crs} is only used in the plotting phase.

The automatic crs recommendation (which is still work-in-progress) is the following:\tabular{ll}{
   \strong{Property} \tab \strong{Recommendation} \cr
   \code{global} (for world maps) \tab A pseudocylindrical projection tmap option \code{crs_global}, by default \verb{"eqearth} (Equal Earth). See \url{https://r-tmap.github.io/tmap/articles/41_advanced_crs.html} for more options \cr
   \code{area} (equal area) \tab Lambert Azimuthal Equal Area (\code{laea}) \cr
   \code{distance}	(equidistant) \tab Azimuthal Equidistant (\code{aeqd}) \cr
   \code{shape}	(conformal) \tab Stereographic (\code{stere}) \cr
}


For further info about the available "generic" projects see:
for utm: \url{https://proj.org/en/9.4/operations/projections/utm.html}
for laea: \url{https://proj.org/en/9.4/operations/projections/laea.html}
for aeqd: \url{https://proj.org/en/9.4/operations/projections/aeqd.html}
for pconic: \url{https://proj.org/en/9.4/operations/projections/pconic.html}
for eqdc: \url{https://proj.org/en/9.4/operations/projections/eqdc.html}
}
\note{
as of tmap 4.0, simplify has been removed. Please use \code{\link[tmaptools:simplify_shape]{tmaptools::simplify_shape()}} instead
}
\examples{
tm_shape(World, crs = "auto") +
	tm_polygons()

tm_shape(World, crs = 3035, bb = "Europe") +
	tm_polygons()

tm_shape(World, crs = "+proj=robin", filter = World$continent=="Africa") +
	tm_polygons()
}
\seealso{
\href{https://r-tmap.github.io/tmap/articles/foundations_crs}{vignette about CRS}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tmap_style_catalogue.R
\name{tmap_style_catalogue}
\alias{tmap_style_catalogue}
\alias{tmap_style_catalog}
\title{Create a style catalogue}
\usage{
tmap_style_catalogue(path = "./tmap_style_previews", styles = NA)

tmap_style_catalog(path = "./tmap_style_previews", styles = NA)
}
\arguments{
\item{path}{path where the png images are stored}

\item{styles}{vector of styles function names (see \code{\link[=tmap_style]{tmap_style()}}) for which a preview is generated. By default, a preview is generated for all loaded styles.}
}
\description{
Create a style catalogue for each predefined tmap style. The result is a set of png images, one for each style.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tmap_options.R
\name{tmapSubmitOptions}
\alias{tmapSubmitOptions}
\title{Internal method for submitting options}
\usage{
tmapSubmitOptions(options = NULL, styleOptions = NULL)
}
\arguments{
\item{options}{list of options}

\item{styleOptions}{list of lists of options, each with a style name}
}
\description{
Internal method for submitting options
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tmap_animation.R
\name{tmap_animation}
\alias{tmap_animation}
\title{Create animation}
\usage{
tmap_animation(
  tm,
  filename = NULL,
  width = NA,
  height = NA,
  dpi = NA,
  outer.margins = NA,
  asp = NULL,
  scale = NA,
  ...
)
}
\arguments{
\item{tm}{tmap or a list of tmap objects. If \code{tm} is a tmap object,
animation frames should be created using either \code{\link[=tm_animate]{tm_animate()}} or
\code{\link[=tm_animate_fast]{tm_animate_fast()}}.}

\item{filename}{filename. If omitted (default), the animation will be shown
in the viewer or browser. If specified, it should be a gif file or a video
file (i.e. mp4). The package \code{gifski} is required to create a gif animation.
The package \code{av} (which uses the \code{FFmpeg} library) is required for video formats.
The mp4 format is recommended but many other video formats are supported,
such as wmv, avi, and mkv.}

\item{width, height}{Dimensions of the animation file (in pixels).
Required when \code{tm} is a list, and recommended to specify in advance when
\code{tm} is a \code{tmap} object. If not specified in the latter case, it will be
determined by the aspect ratio of the map.}

\item{dpi}{dots per inch. By default 100, but this can be set with the option
\code{animation.dpi} in \code{\link[=tmap_options]{tmap_options()}}.}

\item{outer.margins}{(passed on to \code{\link[=tmap_save]{tmap_save()}}) overrides the outer.margins
argument of \code{tm_layout()} (unless set to \code{NA})}

\item{asp}{(passed on to \code{\link[=tmap_save]{tmap_save()}}) if specified, it overrides the \code{asp}
argument of \code{tm_layout()}. Tip: set to \code{0} if map frame should be placed on
the edges of the image.}

\item{scale}{(passed on to \code{\link[=tmap_save]{tmap_save()}}) overrides the scale argument of
\code{\link[=tm_layout]{tm_layout()}} (unless set to \code{NA})}

\item{...}{arguments passed on to \code{\link[av:encoding]{av::av_encode_video()}}}
}
\description{
Create a gif animation or video from an animated tmap plot. First use \code{\link[=tm_animate]{tm_animate()}} or \code{\link[=tm_animate_fast]{tm_animate_fast()}} to animate the plot, and then apply \code{\link[=tmap_animation]{tmap_animation()}} to save it as a gif or video file (e.g. mp4).
}
\note{
Not only tmap plots are supported, but any series of R plots.
}
\examples{
\dontrun{
data(NLD_prov)

m1 <- tm_shape(NLD_prov) + 
	      tm_polygons("yellow") +
	  tm_facets(along = "name")

tmap_animation(m1, delay=40)

data(World, metro)

m2 <- tm_shape(World, projection = "+proj=eck4", simplify = 0.5) +
          tm_fill() +
      tm_shape(metro) + 
          tm_bubbles(size = paste0("pop", seq(1970, 2030, by=10)),
          		   col = "purple",
          		   border.col = "black", border.alpha = .5,
          		   scale = 2) +
      tm_facets(free.scales.symbol.size = FALSE, nrow=1,ncol=1) + 
      tm_format("World")

tmap_animation(m2, delay=100, outer.margins = 0)

m3 <- lapply(seq(50, 85, by = 5), function(age) {
	World$at_most <- World$life_exp <= age
	World_sel <- World[which((World$life_exp <= age) & (World$life_exp > (age - 5))), ]
	tm_shape(World) +
		tm_polygons("at_most", palette = c("gray95", "gold"), legend.show = FALSE) +
		tm_shape(World_sel) +
		tm_text("name", size = "AREA", root = 5, remove_overlap = TRUE) +
		tm_layout(main.title = paste0("Life expectency at most ", age), frame = FALSE)
})

tmap_animation(m3, width = 1200, height = 600, delay = 100)

m4 <- tm_shape(World) +
	tm_polygons() +
tm_shape(metro) +
	tm_bubbles(col = "red") +
	tm_text("name", ymod = -1) +
tm_facets(by = "name", free.coords = FALSE, nrow = 1, ncol = 1) +
	tm_layout(panel.show = FALSE, frame = FALSE)

tmap_animation(m4, filename = "World_cities.mp4", 
    width=1200, height = 600, fps = 2, outer.margins = 0)
}
}
\concept{animation}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tm_scale_.R
\name{tm_scale_rgb}
\alias{tm_scale_rgb}
\alias{tm_scale_rgba}
\title{Scales: RGB}
\usage{
tm_scale_rgb(
  value.na = NA,
  stretch = FALSE,
  probs = c(0, 1),
  max_color_value = 255L
)

tm_scale_rgba(
  value.na = NA,
  stretch = FALSE,
  probs = c(0, 1),
  max_color_value = 255
)
}
\arguments{
\item{value.na}{value for missing values}

\item{stretch}{should each (r, g, b) band be stretched? Possible values: \code{"percent"} (same as \code{TRUE}), \code{"histogram"}, \code{FALSE}.
In the first case, the values are stretched to \verb{probs[1]...probs[2]}. In the second case, a histogram equalization is performed}

\item{probs}{probability (quantile) values when \code{stretch = "percent"}}

\item{max_color_value}{maximum value}
}
\description{
Scales in tmap are configured by the family of functions with prefix \code{tm_scale}.
Such function should be used for the input of the \code{.scale} arguments in the layer
functions (e.g. \code{fill.scale} in \code{\link[=tm_polygons]{tm_polygons()}}).
The function \code{\link[=tm_scale_rgb]{tm_scale_rgb()}} is used to transform r, g, b band variables to colors. This function is adopted from (and works similar as) \code{\link[stars:st_rgb]{stars::st_rgb()}}
}
\examples{
\dontrun{
require(stars)
file = system.file("tif/L7_ETMs.tif", package = "stars")

L7 = stars::read_stars(file)

tm_shape(L7) +
	tm_rgb(col.scale = tm_scale_rgb(probs = c(0, .99), stretch = TRUE))

tm_shape(L7) +
	tm_rgb(col.scale = tm_scale_rgb(stretch = "histogram"))
}
}
\seealso{
\code{\link[=tm_scale]{tm_scale()}} and \code{\link[stars:st_rgb]{stars::st_rgb()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tm_shape.R
\name{tm_shape}
\alias{tm_shape}
\title{Shape (spatial object) specification}
\usage{
tm_shape(
  shp = NULL,
  bbox = NULL,
  crs = NULL,
  is.main = NA,
  layer = NULL,
  name = NULL,
  unit = NULL,
  filter = NULL,
  ...
)
}
\arguments{
\item{shp}{Spatial data object. Typically an object from \pkg{sf}, \pkg{terra}, or \pkg{stars}.
Additional spatial data types can be supported via extension packages, such as
\pkg{tmap.networks} and \pkg{tmap.sources} (experimental). These may include, for example,
remote or streaming data sources.}

\item{bbox}{Bounding box of the map. Only used when \code{shp} is the main shape (see \code{is.main}).
Three options are supported:
\itemize{
\item a \code{\link[sf:st_bbox]{sf::st_bbox()}} object,
\item a character string specifying a location, passed to \code{\link[tmaptools:geocode_OSM]{tmaptools::geocode_OSM()}},
\item \code{"FULL"}, which represents the whole earth. This option ensures that reprojection
retains the full global extent, unlike a regular bounding box.
}}

\item{crs}{Map projection (CRS). Can be set to an \code{crs} object (see \code{\link[sf:st_crs]{sf::st_crs()}}), a proj4string, an EPSG number, the value \code{"auto"} (automatic crs recommendation), or one the the following generic projections: \code{c("laea", "aeqd", "utm", "pconic", "eqdc", "stere")}. See details.}

\item{is.main}{Is \code{shp} the main shape, which determines the crs and
bounding box of the map? By default, \code{TRUE} if it is the first \code{tm_shape} call}

\item{layer}{Name of the layer to use. This is primarily relevant for multi-layer or remote
data sources (e.g. PMTiles or vector tiles), where multiple layers may be available.}

\item{name}{of the spatial object}

\item{unit}{Unit of distance measurement, used by \code{\link[=tm_scalebar]{tm_scalebar()}}. Either a
specific unit string such as \code{"km"}, \code{"m"}, \code{"mi"}, \code{"yd"}, \code{"ft"}, or
\code{"in"} (see \code{\link[units:valid_udunits]{units::valid_udunits()}} for all options), or one of two
automatic options: \code{"metric"} (default) selects the most readable unit
from km, m, and mm; \code{"imperial"} selects from mi, yd, ft, and in.
In both cases the unit is chosen as the largest one for which the
map width expressed in that unit is at least 10.}

\item{filter}{Optional filter expression used to subset features. The exact syntax depends
on the data source. For in-memory objects (e.g. \pkg{sf}), this is typically evaluated in R,
whereas for remote sources it may be translated to a query and executed on the server side.}

\item{...}{passed on to \code{\link[tmaptools:bb]{bb}} (e.g. \code{ext} can be used to enlarge or shrink a bounding box)}
}
\description{
Specify a shape, which is a spatial object from one of these spatial object
class packages: \code{\link[sf:sf]{sf}}, \code{\link[stars:st_as_stars]{stars}}, or \code{terra}.
}
\details{
The map projection (\code{crs}) determines in which coordinate system the spatial object is processed and plotted.
See vignette: \href{https://r-tmap.github.io/tmap/articles/foundations_crs}{Map projections (CRS)}. The \code{crs} can be specified in two places: 1) \code{tm_shape()} and \code{tm_crs()}.
In both cases, the map is plotted into the specified \code{crs}. The difference is that in the first option, the \code{crs} is also taken into account in spatial transformation functions, such as the calculation of centroids and cartograms. In the second option, the \code{crs} is only used in the plotting phase.

The automatic crs recommendation (which is still work-in-progress) is the following:\tabular{ll}{
   \strong{Property} \tab \strong{Recommendation} \cr
   \code{global} (for world maps) \tab A pseudocylindrical projection tmap option \code{crs_global}, by default \verb{"eqearth} (Equal Earth). See vignette: \href{https://r-tmap.github.io/tmap/articles/foundations_crs}{Map projections (CRS)} for more options \cr
   \code{area} (equal area) \tab Lambert Azimuthal Equal Area (\code{laea}) \cr
   \code{distance}	(equidistant) \tab Azimuthal Equidistant (\code{aeqd}) \cr
   \code{shape}	(conformal) \tab Stereographic (\code{stere}) \cr
}


For further info about the available "generic" projects see:
for utm: \url{https://proj.org/en/9.4/operations/projections/utm.html}
for laea: \url{https://proj.org/en/9.4/operations/projections/laea.html}
for aeqd: \url{https://proj.org/en/9.4/operations/projections/aeqd.html}
for pconic: \url{https://proj.org/en/9.4/operations/projections/pconic.html}
for eqdc: \url{https://proj.org/en/9.4/operations/projections/eqdc.html}
}
\note{
as of tmap 4.0, simplify has been removed. Please use \code{\link[tmaptools:simplify_shape]{tmaptools::simplify_shape()}} instead
}
\examples{
tm_shape(World, crs = "auto") +
	tm_polygons()

tm_shape(World, crs = 3035, bb = "Europe") +
	tm_polygons()

tm_shape(World, crs = "+proj=robin", filter = World$continent=="Africa") +
	tm_polygons()
}
\seealso{
\href{https://r-tmap.github.io/tmap/articles/foundations_crs}{Map projections (CRS)}, \href{https://r-tmap.github.io/tmap/articles/examples_choro_NLD}{Choropleth (Netherlands)}, \href{https://r-tmap.github.io/tmap/articles/examples_choro_World}{Choropleth (World)}, \href{https://r-tmap.github.io/tmap/articles/examples_terrain}{Terrain map}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tm_scale_.R
\name{tm_scale_discrete}
\alias{tm_scale_discrete}
\title{Scales: discrete scale}
\usage{
tm_scale_discrete(
  ticks = NA,
  midpoint = NULL,
  values = NA,
  values.repeat = FALSE,
  values.range = NA,
  values.scale = NA,
  value.na = NA,
  value.null = NA,
  value.neutral = NA,
  labels = NULL,
  label.na = NA,
  label.null = NA,
  label.format = list()
)
}
\arguments{
\item{ticks}{Discrete values. If not specified, it is determined automatically: unique values are put on a discrete scale.}

\item{midpoint}{The data value that is interpreted as the midpoint. By default it is set to 0 if negative and positive values are present. Useful when values are diverging colors. In that case, the two sides of the color palette are assigned to negative respectively positive values. If all values are positive or all values are negative, then the midpoint is set to \code{NA}, which means that the value that corresponds to the middle color class (see \code{style}) is mapped to the middle color. If it is specified for sequential color palettes (e.g. \code{"Blues"}), then this color palette will be treated as a diverging color palette.}

\item{values}{(generic scale argument) The visual values. For colors (e.g. \code{fill} or \code{col} for \code{\link[=tm_polygons]{tm_polygons()}}) this is a palette name from the \code{cols4all} package (see \code{\link[cols4all:c4a]{cols4all::c4a()}}) or vector of colors, for size (e.g. \code{size} for \code{tm_symbols}) these are a set of sizes (if two values are specified they are interpret as range), for symbol shapes (e.g. \code{shape} for \code{\link[=tm_symbols]{tm_symbols()}}) these are a set of symbols, etc. The tmap option \code{values.var} contains the default values per visual variable and in some cases also per data type.}

\item{values.repeat}{(generic scale argument) Should the values be repeated in case there are more categories?}

\item{values.range}{(generic scale argument) Range of the values. Vector of two numbers (both between 0 and 1) where the first determines the minimum and the second the maximum. Full range, which means that all values are used, is encoded as \code{c(0, 1)}. For instance, when a gray scale is used for color (from black to white), \code{c(0,1)} means that all colors are used, \verb{0.25, 0.75} means that only colors from dark gray to light gray are used (more precisely \code{"grey25"} to \code{"grey75"}), and \verb{0, 0.5} means that only colors are used from black to middle grey (\code{"grey50"}). When only one number is specified, this is interpreted as the second number (where the first is set to 0). Default values can be set via the tmap option \code{values.range}.}

\item{values.scale}{(generic scale argument) Scaling of the values. Only useful for size-related visual variables, such as \code{size} of \code{\link[=tm_symbols]{tm_symbols()}} and \code{lwd} of \code{\link[=tm_lines]{tm_lines()}}.}

\item{value.na}{(generic scale argument) Value used for missing values. See tmap option \code{"value.na"} for defaults per visual variable.}

\item{value.null}{(generic scale argument) Value used for NULL values. See tmap option \code{"value.null"} for defaults per visual variable. Null data values occur when out-of-scope features are shown (e.g. for a map of Europe showing a data variable per country, the null values are applied to countries outside Europe).}

\item{value.neutral}{(generic scale argument) Value that can be considered neutral. This is used for legends of other visual variables of the same map layer. E.g. when both \code{fill} and \code{size} are used for \code{\link[=tm_symbols]{tm_symbols()}} (using filled circles), the size legend items are filled with the \code{value.neutral} color from the \code{fill.scale} scale, and fill legend items are bubbles of size \code{value.neutral} from the \code{size.scale} scale.}

\item{labels}{(generic scale argument) Labels}

\item{label.na}{(generic scale argument) Label for missing values}

\item{label.null}{(generic scale argument) Label for null (out-of-scope) values}

\item{label.format}{(generic scale argument) Label formatting. Output of \code{\link[=tm_label_format]{tm_label_format()}}}
}
\description{
Scales in tmap are configured by the family of functions with prefix \code{tm_scale}.
Such function should be used for the input of the \code{.scale} arguments in the
layer functions (e.g. \code{fill.scale} in \code{\link[=tm_polygons]{tm_polygons()}}).
The function \code{\link[=tm_scale_discrete]{tm_scale_discrete()}} is used for discrete numerical data, such as integers.
}
\seealso{
\code{\link[=tm_scale]{tm_scale()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tmap_animation.R
\name{tmap_animation}
\alias{tmap_animation}
\title{Create animation}
\usage{
tmap_animation(
  tm,
  filename = NULL,
  width = NA,
  height = NA,
  dpi = NA,
  outer.margins = NA,
  asp = NULL,
  scale = NA,
  ...
)
}
\arguments{
\item{tm}{tmap or a list of tmap objects. If \code{tm} is a tmap object,
animation frames should be created using either \code{\link[=tm_animate]{tm_animate()}} or
\code{\link[=tm_animate_fast]{tm_animate_fast()}}.}

\item{filename}{filename. If omitted (default), the animation will be shown
in the viewer or browser. If specified, it should be a gif file or a video
file (i.e. mp4). The package \code{gifski} is required to create a gif animation.
The package \code{av} (which uses the \code{FFmpeg} library) is required for video formats.
The mp4 format is recommended but many other video formats are supported,
such as wmv, avi, and mkv.}

\item{width, height}{Dimensions of the animation file (in pixels).
Required when \code{tm} is a list, and recommended to specify in advance when
\code{tm} is a \code{tmap} object. If not specified in the latter case, it will be
determined by the aspect ratio of the map.}

\item{dpi}{dots per inch. By default 100, but this can be set with the option
\code{animation.dpi} in \code{\link[=tmap_options]{tmap_options()}}.}

\item{outer.margins}{(passed on to \code{\link[=tmap_save]{tmap_save()}}) overrides the outer.margins
argument of \code{tm_layout()} (unless set to \code{NA})}

\item{asp}{(passed on to \code{\link[=tmap_save]{tmap_save()}}) if specified, it overrides the \code{asp}
argument of \code{tm_layout()}. Tip: set to \code{0} if map frame should be placed on
the edges of the image.}

\item{scale}{(passed on to \code{\link[=tmap_save]{tmap_save()}}) overrides the scale argument of
\code{\link[=tm_layout]{tm_layout()}} (unless set to \code{NA})}

\item{...}{arguments passed on to \code{\link[av:av_encode_video]{av::av_encode_video()}}}
}
\description{
Create a gif animation or video from an animated tmap plot. First use \code{\link[=tm_animate]{tm_animate()}} or \code{\link[=tm_animate_fast]{tm_animate_fast()}} to animate the plot, and then apply \code{\link[=tmap_animation]{tmap_animation()}} to save it as a gif or video file (e.g. mp4).
}
\note{
Not only tmap plots are supported, but any series of R plots.
}
\examples{
\donttest{
	if (interactive()) {
		m1 <- tm_shape(NLD_prov) +
			tm_polygons("yellow") +
			tm_animate(frames = "name")

		tmap_animation(m1, filename = "countries.gif")

		m2 <- tm_shape(metro) +
			tm_symbols(size = paste0("pop", seq(1950, 2030, by=10)),
					   size.free = FALSE,
					   size.legend = tm_legend("Population")) +
			tm_layout(panel.labels = seq(1970, 2030, by=10)) +
			tm_animate()

		tmap_animation(m2, filename = "cities.gif")
	}
}
}
\concept{animation}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tm_element.R
\name{tmap_last}
\alias{tmap_last}
\title{Retrieve the last map to be modified or created}
\usage{
tmap_last()
}
\value{
call
}
\description{
Retrieve the last map to be modified or created. Works in the same way
as \code{ggplot2::last_plot()}, although there is a difference:
\code{tmap_last()} returns the last call instead of the stacked \code{\link{tmap-element}}s.
}
\seealso{
\code{\link[=tmap_save]{tmap_save()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tm_scale_.R
\name{tm_scale_rgb}
\alias{tm_scale_rgb}
\alias{tm_scale_rgba}
\title{Scales: RGB}
\usage{
tm_scale_rgb(
  value.na = NA,
  stretch = FALSE,
  probs = c(0, 1),
  max_color_value = 255L
)

tm_scale_rgba(
  value.na = NA,
  stretch = FALSE,
  probs = c(0, 1),
  max_color_value = 255
)
}
\arguments{
\item{value.na}{value for missing values}

\item{stretch}{should each (r, g, b) band be stretched? Possible values: \code{"percent"} (same as \code{TRUE}), \code{"histogram"}, \code{FALSE}.
In the first case, the values are stretched to \verb{probs[1]...probs[2]}. In the second case, a histogram equalization is performed}

\item{probs}{probability (quantile) values when \code{stretch = "percent"}}

\item{max_color_value}{maximum value}
}
\description{
Scales in tmap are configured by the family of functions with prefix \code{tm_scale}.
Such function should be used for the input of the \code{.scale} arguments in the layer
functions (e.g. \code{fill.scale} in \code{\link[=tm_polygons]{tm_polygons()}}).
The function \code{\link[=tm_scale_rgb]{tm_scale_rgb()}} is used to transform r, g, b band variables to colors. This function is adopted from (and works similar as) \code{\link[stars:st_rgb]{stars::st_rgb()}}
}
\examples{
\dontrun{
require(stars)
file = system.file("tif/L7_ETMs.tif", package = "stars")

L7 = stars::read_stars(file)

tm_shape(L7) +
	tm_rgb(col.scale = tm_scale_rgb(probs = c(0, .99), stretch = TRUE))

tm_shape(L7) +
	tm_rgb(col.scale = tm_scale_rgb(stretch = "histogram"))
}
}
\seealso{
\href{https://r-tmap.github.io/tmap/articles/basics_scales}{Scales}, \href{https://r-tmap.github.io/tmap/articles/basics_vv}{Visual variables}, \href{https://r-tmap.github.io/tmap/articles/adv_multivariate}{Multivariate visual variables} \code{\link[=tm_scale]{tm_scale()}} \code{\link[stars:st_rgb]{stars::st_rgb()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tmap_style.R
\name{tmap_style}
\alias{tmap_style}
\title{Set or get the default tmap style}
\usage{
tmap_style(style)
}
\arguments{
\item{style}{Name of the style. When omitted, \code{tmap_style()} returns the current style
and also shows all available styles. When the style is specified,\code{tmap_style()}
sets the style accordingly. Note that in that case, all tmap options (see \code{\link[=tmap_options]{tmap_options()}})
will be reset according to the style definition.
See \code{\link[=tm_layout]{tm_layout()}} for predefined styles, and \code{tmap_style_catalogue()} for creating a catalogue.}
}
\value{
The style before changing
}
\description{
Set or get the default tmap style. Without arguments, the current style is returned.
Also the available styles are displayed. When a style is set, the corresponding tmap
options (see \code{\link[=tmap_options]{tmap_options()}}) will be set accordingly.
The default style (i.e. when loading the package) is \code{"white"}.
}
\details{
Note that \code{\link[=tm_style]{tm_style()}} is used within a plot call (so it only affects that plot),
whereas \code{tmap_style()} sets the style globally.

After loading a style, the options that defined this style
(i.e. the difference with the default \code{"white"} style) can be obtained by \code{\link[=tmap_options_diff]{tmap_options_diff()}}.

The documentation of \code{\link[=tmap_options]{tmap_options()}} (details and the examples) shows how to create a new style.
}
\examples{
tmap_style()

tm_shape(World) + tm_polygons("HPI")

tmap_style("cobalt")

tm_shape(World) + tm_polygons("HPI")

# for backwards compatibility, the styles of tmap versions 1-3 are also included:

tmap_style("v3")

tm_shape(World) + tm_polygons("HPI")

tmap_style("cobalt_v3")

tm_shape(World) + tm_polygons("HPI")
}
\seealso{
\itemize{
\item \code{\link[=tmap_options]{tmap_options()}} for tmap options
\item \code{\link[=tmap_style_catalogue]{tmap_style_catalogue()}}  to create a style catalogue of all available styles.
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tmapShape.R
\name{tmapShape}
\alias{tmapShape}
\title{Internal method that processed shape objects}
\usage{
tmapShape(
  shp,
  is.main,
  crs,
  bbox,
  unit,
  filter,
  layer,
  shp_name,
  smeta,
  o,
  tmf
)
}
\arguments{
\item{shp}{shp}

\item{is.main}{is.main}

\item{crs}{crs}

\item{bbox}{Bounding box}

\item{unit}{unit}

\item{filter}{filter}

\item{shp_name}{shp_name}

\item{smeta}{smeta}

\item{o}{o}

\item{tmf}{tmf}
}
\description{
Internal method that processed shape objects
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tm_layout.R
\name{tm_style}
\alias{tm_style}
\alias{tm_layout}
\title{Layout options}
\usage{
tm_style(style, ...)

tm_layout(
  scale,
  asp,
  legend.only,
  bg,
  bg.color,
  outer.bg,
  outer.bg.color,
  frame,
  frame.color,
  frame.alpha,
  frame.lwd,
  frame.r,
  frame.double_line,
  outer.margins,
  inner.margins,
  inner.margins.extra,
  meta.margins,
  meta.auto_margins,
  between_margin,
  text.fontfamily,
  text.fontface,
  r,
  attr.color,
  panel.margin,
  panel.type,
  panel.wrap.pos,
  panel.xtab.pos,
  color.sepia_intensity,
  color.saturation,
  color_vision_deficiency_sim,
  panel.show,
  panel.labels,
  panel.label.size,
  panel.label.color,
  panel.label.fontface,
  panel.label.fontfamily,
  panel.label.alpha,
  panel.label.bg,
  panel.label.bg.color,
  panel.label.bg.alpha,
  panel.label.frame,
  panel.label.frame.color,
  panel.label.frame.alpha,
  panel.label.frame.lwd,
  panel.label.frame.r,
  panel.label.height,
  panel.label.rot,
  earth_boundary,
  earth_boundary.color,
  earth_boundary.lwd,
  earth_datum,
  space,
  space.color,
  ...
)
}
\arguments{
\item{style}{name of the style}

\item{...}{List of tmap options to be set, or option names (characters) to be returned (see details)}

\item{scale}{Overall scale of the map}

\item{asp}{Aspect ratio of each map. When \code{asp} is set to \code{NA} (default) the aspect ratio will be adjusted to the used shapes. When set to 0 the aspect ratio is
adjusted to the size of the device divided by the number of columns and rows.}

\item{legend.only}{Should only legends be printed (so without map)?}

\item{bg}{Draw map background?}

\item{bg.color}{Background color of the map.}

\item{outer.bg}{Draw map background (outside the frame)?}

\item{outer.bg.color}{Background color of map outside the frame.}

\item{frame}{Draw map frame?}

\item{frame.color}{The color of the frame.}

\item{frame.alpha}{The alpha transparency of the frame.}

\item{frame.lwd}{The line width of the frame. See \code{graphics::par}, option 'lwd'.}

\item{frame.r}{The r (radius) of the frame.}

\item{frame.double_line}{The double line of the frame. \code{TRUE} or \code{FALSE}.}

\item{outer.margins}{The margins of the outer space (outside the frame. A vector of 4 values: bottom, left, top, right. The unit is the device height (for bottom and top) or width (for left and right).}

\item{inner.margins}{The margins of the inner space (inside the frame). A vector of 4 values: bottom, left, top, right. The unit is the device height (for bottom and top) or width (for left and right).}

\item{inner.margins.extra}{The extra arguments of the margins of the inner space (inside the frame). A list of arguments.}

\item{meta.margins}{The margins of the meta. A vector of 4 values: bottom, left, top, right. The unit is the device height (for bottom and top) or width (for left and right).}

\item{meta.auto_margins}{The auto_margins of the meta.}

\item{between_margin}{Margin between the map.}

\item{text.fontfamily}{The font family of the text. See \code{graphics::par}, option 'family'.}

\item{text.fontface}{The font face of the text. See \code{graphics::par}, option 'font'.}

\item{r}{The r (radius) (overall).}

\item{attr.color}{The color of the attr.}

\item{panel.margin}{The margin of the panel.}

\item{panel.type}{The type of the panel.}

\item{panel.wrap.pos}{The panel positions for wrapped facets created with \code{\link[=tm_facets_grid]{tm_facets_grid()}}. One of \code{"left"}, \code{"right"}, \code{"top"} (default) or \code{"bottom"}.}

\item{panel.xtab.pos}{The panel positions for grid facets created with \code{\link[=tm_facets_grid]{tm_facets_grid()}}. Vector of two, where the first determines the locations of row panels (\code{"left"} or \code{"right"}) and the second the location of column panels ( \code{"top"} or `"bottom")}

\item{color.sepia_intensity}{The sepia_intensity of the color.}

\item{color.saturation}{The saturation of the color.}

\item{color_vision_deficiency_sim}{Color vision deficiency simulation. Either \code{"protan"}, \code{"deutan"}, or \code{"tritan"}.}

\item{panel.show}{The visibility of the panel. \code{TRUE} or \code{FALSE}.}

\item{panel.labels}{The labels of the panel.}

\item{panel.label.size}{The size of the label of the panel.}

\item{panel.label.color}{The color of the label of the panel.}

\item{panel.label.fontface}{The font face of the label of the panel. See \code{graphics::par}, option 'font'.}

\item{panel.label.fontfamily}{The font family of the label of the panel. See \code{graphics::par}, option 'family'.}

\item{panel.label.alpha}{The alpha transparency of the label of the panel.}

\item{panel.label.bg}{The bg of the label of the panel.}

\item{panel.label.bg.color}{The color of the bg of the label of the panel.}

\item{panel.label.bg.alpha}{The alpha transparency of the bg of the label of the panel.}

\item{panel.label.frame}{The frame of the label of the panel.}

\item{panel.label.frame.color}{The color of the frame of the label of the panel.}

\item{panel.label.frame.alpha}{The alpha transparency of the frame of the label of the panel.}

\item{panel.label.frame.lwd}{The line width of the frame of the label of the panel. See \code{graphics::par}, option 'lwd'.}

\item{panel.label.frame.r}{The r (radius) of the frame of the label of the panel.}

\item{panel.label.height}{The height of the label of the panel.}

\item{panel.label.rot}{Rotation angles of the panel labels. Vector of four values that determine the panel label rotation when they are placed left, top, right, and bottom. The default angles are 90, 0, 270 and 0 respectively. Note that the second value is the most common, since labels are by default shown on top (see \code{panel.wrap.pos}). In cross-table facets created with \code{tm_facets_grid()}, the first two values are used by default (see \code{panel.xtab.pos}).}

\item{earth_boundary}{The earth boundary}

\item{earth_boundary.color}{The color of the earth_boundary.}

\item{earth_boundary.lwd}{The line width of the earth_boundary. See \code{graphics::par}, option 'lwd'.}

\item{earth_datum}{Earth datum}

\item{space}{Should the space be drawn? Only applicable is earth_boundary is enabled.}

\item{space.color}{The color of the space.}
}
\description{
Specify the layout of the maps. \code{\link[=tm_layout]{tm_layout()}} is identical as \code{\link[=tm_options]{tm_options()}} but only contain the tmap options that are directly related to the layout. \code{\link[=tm_style]{tm_style()}} sets the style for the map. A style is a specified set of options (that can be changed afterwards with \code{\link[=tm_layout]{tm_layout()}}). These functions are used within a plot a plot call (stacked with the \code{+} operator). Their counterparts \code{\link[=tmap_options]{tmap_options()}} and \code{\link[=tmap_style]{tmap_style()}} can be used to set the (layout) options globally.
}
\examples{
tm_shape(World) +
  tm_polygons() +
tm_layout(
  bg.color = "steelblue",
  outer.bg.color = "gold",
  frame.lwd = 3,
  inner.margins = 0)

tm_shape(World) +
  tm_polygons(fill = "HPI") +
tm_style("classic")

tm_shape(World) +
	tm_polygons(fill = "HPI") +
	tm_style("cobalt")
}
\seealso{
\href{https://r-tmap.github.io/tmap/articles/basics_layout}{Vignette about layout}, \href{https://r-tmap.github.io/tmap/articles/adv_margins}{vignette about margins and aspect ratio} and \href{https://r-tmap.github.io/tmap/articles/adv_options}{vignette about options}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tmap_overview.R
\name{tmap_overview}
\alias{tmap_overview}
\title{Overview of tmap layers}
\usage{
tmap_overview()
}
\value{
A list of three layer types are returned: data layers, aux layers, and components.
}
\description{
Overview of tmap layers, organized by layer type.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tm_element.R
\name{print.tm_element}
\alias{print.tm_element}
\alias{print.tm_shape}
\title{Print tm_element}
\usage{
\method{print}{tm_element}(x, ...)

\method{print}{tm_shape}(x, ...)
}
\arguments{
\item{x}{x}

\item{...}{passed on}
}
\description{
Print tm_element
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tm_pos.R
\name{tm_pos}
\alias{tm_pos}
\alias{tm_pos_in}
\alias{tm_pos_out}
\alias{tm_pos_on_top}
\alias{tm_pos_auto_out}
\alias{tm_pos_auto_in}
\title{Set the position of map components}
\usage{
tm_pos(cell.h, cell.v, pos.h, pos.v, align.h, align.v, just.h, just.v)

tm_pos_in(pos.h, pos.v, align.h, align.v, just.h, just.v)

tm_pos_out(cell.h, cell.v, pos.h, pos.v, align.h, align.v, just.h, just.v)

tm_pos_on_top(pos.h, pos.v, align.h, align.v, just.h, just.v)

tm_pos_auto_out(cell.h, cell.v, pos.h, pos.v, align.h, align.v, just.h, just.v)

tm_pos_auto_in(align.h, align.v, just.h, just.v)
}
\arguments{
\item{cell.h, cell.v}{The plotting area is overlaid with a 3x3 grid, of which
the middle grid cell is the map area. Components can be drawn into any cell.
\code{cell.h} specifies the horizontal position (column) can take values
\code{"left"}, \code{"center"}, and \code{"right"}. \code{cell.v} specifies the vertical position
(row) and can take values \code{"top"}, \code{"center"}, and \code{"bottom"}.
See details for a graphical explanation.}

\item{pos.h, pos.v}{The position of the component within the cell.
The main options for \code{pos.h} are \code{"left"}, \code{"center"}, and \code{"right"}.
For \code{cell.v} these are \code{"top"}, \code{"center"}, and \code{"bottom"}.
These options can also be provided in upper case; in that case there is no offset
(see the tmap option \code{component.offset}). Also numbers between 0 and 1 can be provided,
which determine the position of the component inside the cell (with (0,0) being left bottom).
The arguments \code{just.h} and \code{just.v} determine the justification point.}

\item{align.h, align.v}{The alignment of the component in case multiple components are stacked.
When they are stacked horizontally, \code{align.v} determines how components
that are smaller in height than the available height (determined by the outer.margins
if specified and otherwise by the highest component) are justified:
\code{"top"}, \code{"center"}, or \code{"bottom"}. Similarly, \code{align.h} determines how components
are justified horizontally when they are stacked vertically: \code{"left"}, \code{"center"}, or \code{"right"}.}

\item{just.h, just.v}{The justification of the components.
Only used in case \code{pos.h} and \code{pos.v} are numbers.}
}
\description{
Set the position of map components, such as legends, title, compass, scale bar, etc.
\code{tm_pos()} is the function to position these components:
\code{tm_pos_out()} places the components outside the map area, \code{tm_pos_in()} inside the map area, and \code{tm_pos_on_top()} on top of the map
Each \code{position} argument of a map layer or component should be specified with
one of these functions.
The functions \code{tm_pos_auto_out()} and \code{tm_pos_auto_in()}
are used to set the components automatically, and should be used
via \code{\link[=tmap_options]{tmap_options()}}. See Details how the positioning works.
}
\details{
\code{tm_pos_in()} sets the position of the component(s) inside the maps area,
which is equivalent to the center-center cell (in case there are facets,
these are all drawn in this center-center cell).

\code{tm_pos_out()} sets the position of the component(s) outside the map.

\code{tm_pos_on_top()} is the same as \code{tm_pos_out}, but with the cell set to the center cell
It may be therefore seem similar to \code{tm_pos_in()}, but with an essential difference:
\code{tm_pos_in()} takes the map frame into account whereas \code{tm_pos_on_top()} does not.
#'
The amount of space that the top and bottom rows, and left and right columns
occupy is determined by the \code{\link[=tm_layout]{tm_layout()}} arguments \code{meta.margins} and \code{meta.auto_margins}.
The former sets the relative space of the bottom, left, top, and right side.
In case these are set to \code{NA}, the space is set automatically based on 1)
the maximum relative space specified by \code{meta.auto_margins} and 2)
the presence and size of components in each cell.
For instance, if there is one landscape oriented legend in the center-bottom cell,
then the relative space of the bottom row is set to the height of that legend
(given that it is smaller than the corresponding value of \code{meta.auto_margins}),
while the other four sides are set to 0.

\code{tm_pos_auto_out()} is more complex: the \code{cell.h} and \code{cell.v} arguments should be set
to one of the four corners. It does not mean that the components are drawn in a corner.
The corner represents the sides of the map that the components are drawn.
By default, legends are drawn either at the bottom or on the right-side of
the map by default (see \code{tmap_options("legend.position")}).
Only when there are row- and column-wise legends and a general legend (using \code{\link[=tm_facets_grid]{tm_facets_grid()}}),
the general legend is drawn in the corner, but in practice this case will be rare.

The arguments \code{pos.h} and \code{pos.v} determine where the components are drawn within the cell.
Again, with \code{"left"}, \code{"center"}, and \code{"right"} for \code{pos.h}
and \code{"top"}, \code{"center"}, and \code{"bottom"} for \code{pos.v}.
The values can also be specified in upper-case, which influences the offset
with the cell borders, which is determined by tmap option \code{component.offset}.
By default, there is a small offset when components are drawn inside and no
offset when they are drawn outside or with upper-case.

\code{tm_pos_auto_in()} automatically determines \code{pos.h} and \code{pos.v} given the
available space inside the map. This is similar to the default positioning in tmap3.

In case multiple components are draw in the same cell and the same position
inside that cell, they are stacked (determined which the \code{stack} argument in
the legend or component function). The \code{align.h} and \code{align.v} arguments
determine how these components will be justified with each other.

Note that legends and components may be different for a facet row or column.
This is the case when \code{\link[=tm_facets_grid]{tm_facets_grid()}} or \code{\link[=tm_facets_stack]{tm_facets_stack()}} are applied
and when scales are set to free (with the \code{.free} argument of the map layer functions).
In case a legends or components are draw row- or column wise, and the position
of the legends (or components) is right next to the maps, these legends
(or components) will be aligned with the maps.
}
\seealso{
\href{https://r-tmap.github.io/tmap/articles/adv_positions}{Vignette about positioning}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tm_layout.R
\name{tm_style}
\alias{tm_style}
\alias{tm_layout}
\title{Layout options}
\usage{
tm_style(style, ...)

tm_layout(
  scale,
  asp,
  legend.only,
  bg,
  bg.color,
  outer.bg,
  outer.bg.color,
  frame,
  frame.color,
  frame.alpha,
  frame.lwd,
  frame.r,
  frame.double_line,
  outer.margins,
  inner.margins,
  inner.margins.extra,
  meta.margins,
  meta.auto_margins,
  between_margin,
  text.fontfamily,
  text.fontface,
  r,
  attr.color,
  panel.margin,
  panel.type,
  panel.wrap.pos,
  panel.xtab.pos,
  color.sepia_intensity,
  color.saturation,
  color_vision_deficiency_sim,
  panel.show,
  panel.labels,
  panel.label.size,
  panel.label.color,
  panel.label.fontface,
  panel.label.fontfamily,
  panel.label.alpha,
  panel.label.bg,
  panel.label.bg.color,
  panel.label.bg.alpha,
  panel.label.frame,
  panel.label.frame.color,
  panel.label.frame.alpha,
  panel.label.frame.lwd,
  panel.label.frame.r,
  panel.label.height,
  panel.label.rot,
  earth_boundary,
  earth_boundary.color,
  earth_boundary.lwd,
  earth_datum,
  space,
  space.color,
  ...
)
}
\arguments{
\item{style}{name of the style}

\item{...}{List of tmap options to be set, or option names (characters) to be returned (see details)}

\item{scale}{Overall scale of the map}

\item{asp}{Aspect ratio of each map. When \code{asp} is set to \code{NA} (default) the aspect ratio will be adjusted to the used shapes. When set to 0 the aspect ratio is
adjusted to the size of the device divided by the number of columns and rows.}

\item{legend.only}{Should only legends be printed (so without map)?}

\item{bg}{Draw map background?}

\item{bg.color}{Background color of the map.}

\item{outer.bg}{Draw map background (outside the frame)?}

\item{outer.bg.color}{Background color of map outside the frame.}

\item{frame}{Draw map frame?}

\item{frame.color}{The color of the frame.}

\item{frame.alpha}{The alpha transparency of the frame.}

\item{frame.lwd}{The line width of the frame. See \code{graphics::par}, option 'lwd'.}

\item{frame.r}{The r (radius) of the frame.}

\item{frame.double_line}{The double line of the frame. \code{TRUE} or \code{FALSE}.}

\item{outer.margins}{The margins of the outer space (outside the frame. A vector of 4 values: bottom, left, top, right. The unit is the device height (for bottom and top) or width (for left and right).}

\item{inner.margins}{The margins of the inner space (inside the frame). A vector of 4 values: bottom, left, top, right. The unit is the device height (for bottom and top) or width (for left and right).}

\item{inner.margins.extra}{The extra arguments of the margins of the inner space (inside the frame). A list of arguments.}

\item{meta.margins}{The margins of the 'meta' space (reserved for outside map components). A vector of 4 values: bottom, left, top, right. The unit is the device height (for bottom and top) or width (for left and right).}

\item{meta.auto_margins}{The auto_margins of the 'meta' space (reserved for outside map components).}

\item{between_margin}{Margin between the map.}

\item{text.fontfamily}{The font family of the text. See \code{graphics::par}, option 'family'.}

\item{text.fontface}{The font face of the text. See \code{graphics::par}, option 'font'.}

\item{r}{The r (radius) (overall).}

\item{attr.color}{The color of the attr.}

\item{panel.margin}{The margin of the panel.}

\item{panel.type}{The type of the panel.}

\item{panel.wrap.pos}{The panel positions for wrapped facets created with \code{\link[=tm_facets_grid]{tm_facets_grid()}}. One of \code{"left"}, \code{"right"}, \code{"top"} (default) or \code{"bottom"}.}

\item{panel.xtab.pos}{The panel positions for grid facets created with \code{\link[=tm_facets_grid]{tm_facets_grid()}}. Vector of two, where the first determines the locations of row panels (\code{"left"} or \code{"right"}) and the second the location of column panels ( \code{"top"} or `"bottom")}

\item{color.sepia_intensity}{The sepia_intensity of the color.}

\item{color.saturation}{The saturation of the color.}

\item{color_vision_deficiency_sim}{Color vision deficiency simulation. Either \code{"protan"}, \code{"deutan"}, or \code{"tritan"}.}

\item{panel.show}{The visibility of the panel. \code{TRUE} or \code{FALSE}.}

\item{panel.labels}{The labels of the panel.}

\item{panel.label.size}{The size of the label of the panel.}

\item{panel.label.color}{The color of the label of the panel.}

\item{panel.label.fontface}{The font face of the label of the panel. See \code{graphics::par}, option 'font'.}

\item{panel.label.fontfamily}{The font family of the label of the panel. See \code{graphics::par}, option 'family'.}

\item{panel.label.alpha}{The alpha transparency of the label of the panel.}

\item{panel.label.bg}{The bg of the label of the panel.}

\item{panel.label.bg.color}{The color of the bg of the label of the panel.}

\item{panel.label.bg.alpha}{The alpha transparency of the bg of the label of the panel.}

\item{panel.label.frame}{The frame of the label of the panel.}

\item{panel.label.frame.color}{The color of the frame of the label of the panel.}

\item{panel.label.frame.alpha}{The alpha transparency of the frame of the label of the panel.}

\item{panel.label.frame.lwd}{The line width of the frame of the label of the panel. See \code{graphics::par}, option 'lwd'.}

\item{panel.label.frame.r}{The r (radius) of the frame of the label of the panel.}

\item{panel.label.height}{The height of the label of the panel.}

\item{panel.label.rot}{Rotation angles of the panel labels. Vector of four values that determine the panel label rotation when they are placed left, top, right, and bottom. The default angles are 90, 0, 270 and 0 respectively. Note that the second value is the most common, since labels are by default shown on top (see \code{panel.wrap.pos}). In cross-table facets created with \code{tm_facets_grid()}, the first two values are used by default (see \code{panel.xtab.pos}).}

\item{earth_boundary}{The earth boundary}

\item{earth_boundary.color}{The color of the earth_boundary.}

\item{earth_boundary.lwd}{The line width of the earth_boundary. See \code{graphics::par}, option 'lwd'.}

\item{earth_datum}{Earth datum}

\item{space}{Should the space be drawn? Only applicable is earth_boundary is enabled.}

\item{space.color}{The color of the space.}
}
\description{
Specify the layout of the maps. \code{\link[=tm_layout]{tm_layout()}} is identical as \code{\link[=tm_options]{tm_options()}} but only contain the tmap options that are directly related to the layout. \code{\link[=tm_style]{tm_style()}} sets the style for the map. A style is a specified set of options (that can be changed afterwards with \code{\link[=tm_layout]{tm_layout()}}). These functions are used within a plot call (stacked with the \code{+} operator). Their counterparts \code{\link[=tmap_options]{tmap_options()}} and \code{\link[=tmap_style]{tmap_style()}} can be used to set the (layout) options globally.
}
\details{
When \code{\link[=tm_style]{tm_style()}} is called, all layout options are reset according to that
style. To override specific layout options for a chosen style, call
\code{\link[=tm_layout]{tm_layout()}} \emph{after} \code{\link[=tm_style]{tm_style()}}.
}
\note{
In earlier versions of tmap, the \code{"natural"} style enabled the earth
boundary automatically. This is no longer the case, because it did not work
well for many projections. The earth boundary can still be enabled
explicitly via the \code{earth_boundary} argument of \code{\link[=tm_layout]{tm_layout()}} (see the
example above).
}
\examples{
tm_shape(World) +
  tm_polygons() +
tm_layout(
  bg.color = "steelblue",
  outer.bg.color = "gold",
  frame.lwd = 3,
  inner.margins = 0)

tm_shape(World) +
  tm_polygons(fill = "HPI") +
tm_style("classic")

tm_shape(World) +
	tm_polygons(fill = "HPI") +
	tm_style("cobalt")
}
\seealso{
\href{https://r-tmap.github.io/tmap/articles/basics_layout}{Layout}, \href{https://r-tmap.github.io/tmap/articles/adv_margins}{Margins and aspect ratio}, \href{https://r-tmap.github.io/tmap/articles/adv_options}{Options and styles}, \href{https://r-tmap.github.io/tmap/articles/examples_choro_World}{Choropleth (World)}, \href{https://r-tmap.github.io/tmap/articles/examples_terrain}{Terrain map}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/theme_ps.R
\name{theme_ps}
\alias{theme_ps}
\title{ggplot2 theme for proportional symbols}
\usage{
theme_ps(
  base_size = 12,
  base_family = "",
  plot.axes = FALSE,
  plot.legend = FALSE
)
}
\arguments{
\item{base_size}{base size}

\item{base_family}{base family}

\item{plot.axes}{should the axes be shown?}

\item{plot.legend}{should the legend(s) be shown?}
}
\description{
ggplot2 theme for proportional symbols. By default, this theme only shows the
plotting area, so without titles, axes, and legend.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tm_layout.R
\name{tm_plot}
\alias{tm_plot}
\title{Plot mode options}
\usage{
tm_plot(use_gradient, limit_latitude_3857)
}
\arguments{
\item{use_gradient}{Use gradient fill using \link[grid:patterns]{linearGradient()}}

\item{limit_latitude_3857}{Vector of two limit latitude values for maps printed in Web Mercator projection (EPSG 3857). If \code{c(-90, 90)} the poles will be inflated too much. The Web Mercator is defines as \code{c(-85.06, 85.06)}, but the default setting in tmap is \code{c(-84, 84)}.}
}
\description{
Plot mode options. This option is specific to the plot mode.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tm_pos.R
\name{tm_pos}
\alias{tm_pos}
\alias{tm_pos_in}
\alias{tm_pos_out}
\alias{tm_pos_on_top}
\alias{tm_pos_auto_out}
\alias{tm_pos_auto_in}
\title{Set the position of map components}
\usage{
tm_pos(cell.h, cell.v, pos.h, pos.v, align.h, align.v, just.h, just.v)

tm_pos_in(pos.h, pos.v, align.h, align.v, just.h, just.v)

tm_pos_out(cell.h, cell.v, pos.h, pos.v, align.h, align.v, just.h, just.v)

tm_pos_on_top(pos.h, pos.v, align.h, align.v, just.h, just.v)

tm_pos_auto_out(cell.h, cell.v, pos.h, pos.v, align.h, align.v, just.h, just.v)

tm_pos_auto_in(align.h, align.v, just.h, just.v)
}
\arguments{
\item{cell.h, cell.v}{The plotting area is overlaid with a 3x3 grid, of which
the middle grid cell is the map area. Components can be drawn into any cell.
\code{cell.h} specifies the horizontal position (column) can take values
\code{"left"}, \code{"center"}, and \code{"right"}. \code{cell.v} specifies the vertical position
(row) and can take values \code{"top"}, \code{"center"}, and \code{"bottom"}.
See details for a graphical explanation.}

\item{pos.h, pos.v}{The position of the component within the cell.
The main options for \code{pos.h} are \code{"left"}, \code{"center"}, and \code{"right"}.
For \code{cell.v} these are \code{"top"}, \code{"center"}, and \code{"bottom"}.
These options can also be provided in upper case; in that case there is no offset
(see the tmap option \code{component.offset}). Also numbers between 0 and 1 can be provided,
which determine the position of the component inside the cell (with (0,0) being left bottom).
The arguments \code{just.h} and \code{just.v} determine the justification point.}

\item{align.h, align.v}{The alignment of the component in case multiple components are stacked.
When they are stacked horizontally, \code{align.v} determines how components
that are smaller in height than the available height (determined by the outer.margins
if specified and otherwise by the highest component) are justified:
\code{"top"}, \code{"center"}, or \code{"bottom"}. Similarly, \code{align.h} determines how components
are justified horizontally when they are stacked vertically: \code{"left"}, \code{"center"}, or \code{"right"}.}

\item{just.h, just.v}{The justification of the components.
Only used in case \code{pos.h} and \code{pos.v} are numbers.}
}
\description{
Set the position of map components, such as legends, title, compass, scale bar, etc.
\code{tm_pos()} is the function to position these components:
\code{tm_pos_out()} places the components outside the map area, \code{tm_pos_in()} inside the map area, and \code{tm_pos_on_top()} on top of the map
Each \code{position} argument of a map layer or component should be specified with
one of these functions.
The functions \code{tm_pos_auto_out()} and \code{tm_pos_auto_in()}
are used to set the components automatically, and should be used
via \code{\link[=tmap_options]{tmap_options()}}. See Details how the positioning works.
}
\details{
\code{tm_pos_in()} sets the position of the component(s) inside the maps area,
which is equivalent to the center-center cell (in case there are facets,
these are all drawn in this center-center cell).

\code{tm_pos_out()} sets the position of the component(s) outside the map.

\code{tm_pos_on_top()} is the same as \code{tm_pos_out}, but with the cell set to the center cell
It may be therefore seem similar to \code{tm_pos_in()}, but with an essential difference:
\code{tm_pos_in()} takes the map frame into account whereas \code{tm_pos_on_top()} does not.
#'
The amount of space that the top and bottom rows, and left and right columns
occupy is determined by the \code{\link[=tm_layout]{tm_layout()}} arguments \code{meta.margins} and \code{meta.auto_margins}.
The former sets the relative space of the bottom, left, top, and right side.
In case these are set to \code{NA}, the space is set automatically based on 1)
the maximum relative space specified by \code{meta.auto_margins} and 2)
the presence and size of components in each cell.
For instance, if there is one landscape oriented legend in the center-bottom cell,
then the relative space of the bottom row is set to the height of that legend
(given that it is smaller than the corresponding value of \code{meta.auto_margins}),
while the other four sides are set to 0.

\code{tm_pos_auto_out()} is more complex: the \code{cell.h} and \code{cell.v} arguments should be set
to one of the four corners. It does not mean that the components are drawn in a corner.
The corner represents the sides of the map that the components are drawn.
By default, legends are drawn either at the bottom or on the right-side of
the map by default (see \code{tmap_options("legend.position")}).
Only when there are row- and column-wise legends and a general legend (using \code{\link[=tm_facets_grid]{tm_facets_grid()}}),
the general legend is drawn in the corner, but in practice this case will be rare.

The arguments \code{pos.h} and \code{pos.v} determine where the components are drawn within the cell.
Again, with \code{"left"}, \code{"center"}, and \code{"right"} for \code{pos.h}
and \code{"top"}, \code{"center"}, and \code{"bottom"} for \code{pos.v}.
The values can also be specified in upper-case, which influences the offset
with the cell borders, which is determined by tmap option \code{component.offset}.
By default, there is a small offset when components are drawn inside and no
offset when they are drawn outside or with upper-case.

\code{tm_pos_auto_in()} automatically determines \code{pos.h} and \code{pos.v} given the
available space inside the map. This is similar to the default positioning in tmap3.

In case multiple components are draw in the same cell and the same position
inside that cell, they are stacked (determined which the \code{stack} argument in
the legend or component function). The \code{align.h} and \code{align.v} arguments
determine how these components will be justified with each other.

Note that legends and components may be different for a facet row or column.
This is the case when \code{\link[=tm_facets_grid]{tm_facets_grid()}} or \code{\link[=tm_facets_stack]{tm_facets_stack()}} are applied
and when scales are set to free (with the \code{.free} argument of the map layer functions).
In case a legends or components are draw row- or column wise, and the position
of the legends (or components) is right next to the maps, these legends
(or components) will be aligned with the maps.
}
\seealso{
\href{https://r-tmap.github.io/tmap/articles/basics_components}{Components}, \href{https://r-tmap.github.io/tmap/articles/adv_positions}{Positioning of components}, \href{https://r-tmap.github.io/tmap/articles/adv_comp_group}{Grouping of components}, \href{https://r-tmap.github.io/tmap/articles/examples_choro_NLD}{Choropleth (Netherlands)}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tm_facets.R
\name{tm_facets}
\alias{tm_facets}
\alias{tm_facets_grid}
\alias{tm_facets_wrap}
\alias{tm_facets_pagewise}
\alias{tm_facets_stack}
\alias{tm_facets_hstack}
\alias{tm_facets_vstack}
\alias{tm_facets_flip}
\title{Specify facets}
\usage{
tm_facets(
  by = NULL,
  rows = NULL,
  columns = NULL,
  pages = NULL,
  as.layers = FALSE,
  nrow = NA,
  ncol = NA,
  byrow = TRUE,
  orientation = NA,
  free.coords = NA,
  drop.units = TRUE,
  drop.empty.facets = TRUE,
  drop.NA.facets = FALSE,
  sync = TRUE,
  swipe = FALSE,
  na.text = NA,
  scale.factor = 2,
  type = NA,
  free.scales = NULL,
  ...
)

tm_facets_grid(rows = NULL, columns = NULL, pages = NULL, ...)

tm_facets_wrap(by = "VARS__", nrow = NA, ncol = NA, byrow = TRUE, ...)

tm_facets_pagewise(by = "VARS__", byrow = TRUE, ...)

tm_facets_stack(by = "VARS__", orientation = NA, ...)

tm_facets_hstack(by = "VARS__", ...)

tm_facets_vstack(by = "VARS__", ...)

tm_facets_flip(...)
}
\arguments{
\item{by}{Group by variable (only for a facet wrap or facet stack)}

\item{rows}{Variable that specifies the rows (only for a facet grid)}

\item{columns}{Variable that specifies the columns (only for a facet grid)}

\item{pages}{Variable that specifies the pages (only for a facet grid)}

\item{as.layers}{show facets as layers?}

\item{nrow}{Number of rows}

\item{ncol}{Number of columns}

\item{byrow}{Should facets be wrapped by row?}

\item{orientation}{For facet stack: horizontal or vertical?}

\item{free.coords}{Logical. If the \code{by} argument is specified, should each
map has its own coordinate ranges? By default \code{TRUE}, unless facets are shown
in as different layers (\code{as.layers = TRUE})}

\item{drop.units}{Logical. If the \code{by} argument is specified, should
non-selected spatial units be dropped? If \code{FALSE}, they are plotted where
mapped aesthetics are regarded as missing values. Not applicable for
raster shapes. By default \code{TRUE}.}

\item{drop.empty.facets}{Logical. If the \code{by} argument is specified, should
empty facets be dropped? Empty facets occur when the \code{by}-variable contains
unused levels. When \code{TRUE} and two \code{by}-variables are specified, empty rows
and columns are dropped.}

\item{drop.NA.facets}{Logical. If the \code{by} argument is specified, and all
data values for specific facets are missing, should these facets be dropped?
\code{FALSE} by default. In v3, it was called \code{showNA}.}

\item{sync}{Logical. Should the navigation in view mode (zooming and panning)
be synchronized? By default \code{TRUE} if the facets have the same bounding box.
This is generally the case when rasters are plotted, or when \code{free.coords} is
\code{FALSE}.}

\item{swipe}{Logical. If \code{TRUE}, the two maps are overlaid with a draggable divider to reveal either map. Requires exactly two maps, and only works in the \code{"mapbox"} and \code{"maplibre"} modes provided by the \pkg{tmap.mapgl} package. Default is \code{FALSE}.}

\item{na.text}{Text used for facets of missing values. In v3, it was \code{textNA}.}

\item{scale.factor}{Number that determines how the elements (e.g. font sizes,
symbol sizes, line widths) of the small multiples are scaled in relation to
the scaling factor of the shapes. The elements are scaled to the \code{scale.factor}th
root of the scaling factor of the shapes. So, for \code{scale.factor = 1},
they are scaled proportional to the scaling of the shapes. Since elements,
especially text, are often too small to read, a higher value is recommended.
By default, \code{scale.factor = 2}.}

\item{type}{\code{"grid"}, \code{"wrap"} or \code{"stack"}}

\item{free.scales}{deprecated. Please use the \code{.free} arguments in the layer functions, e.g. \code{fill.free} in \code{tm_polygons}.}

\item{...}{passed on to \code{tm_facets()}}
}
\description{
\itemize{
\item \code{tm_facets_wrap()} specify facets for one grouping variable (so one faceting dimension)
\item \verb{tm_facets_(hv)stack()} stacks the facets either horizontally or vertically (one grouping variable).
\item \code{tm_facets_grid()} specify facets for two grouping variables in a grid of rows and columns.
\item \code{tm_facets_pagewise()} same as wrap, but the facets are drawn on different plots (pages). Replaces the \code{along} argument from version 3.
\item \code{tm_facets()} is the core function, but it is recommended to use the other functions.
}
}
\note{
In older versions (< 4.1) \code{\link[=tm_facets]{tm_facets()}} with \code{page} specification was used to create animations frames and \code{\link[=tmap_animation]{tmap_animation()}} to create the animation itself using inputs like the frame rate specification. As of version 4.2, the whole animation, including frame rate, is specified in \code{\link[=tm_animate]{tm_animate()}}. The animation can still be saved via \code{\link[=tmap_animation]{tmap_animation()}}.
}
\examples{
\dontrun{
tm_shape(NLD_dist) +
  tm_polygons("edu_appl_sci",
    fill.scale = tm_scale_intervals(values = "pu_gn", style = "kmeans", n = 7)) +
  tm_facets(by = "province") +
tm_shape(NLD_muni) +
  tm_borders(lwd = 3) +
  tm_facets(by = "province") +
tm_title("Population with a univeristy degree (incl appl. sciences), percentages")

tm_shape(World) +
  tm_polygons(c("gender", "press"),
    fill.scale = list(tm_scale_intervals(values = "bu_br_div", midpoint = 0.5),
      tm_scale_intervals(values = "pu_gn_div", midpoint = 50)),
    fill.legend = tm_legend("")) +
tm_layout(panel.labels = c("Gender Inequality Index (GII)", "World Press Freedom Index"))
}
}
\seealso{
\code{\link[=tm_animate]{tm_animate()}}

\href{https://r-tmap.github.io/tmap/articles/basics_facets}{Vignette about facets}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data.R
\docType{data}
\name{land}
\alias{land}
\title{Spatial data of global land cover}
\format{
An object of class \code{stars} with 1080 rows and 540 columns.
}
\usage{
land
}
\description{
Spatial data of global land cover, percent tree cover, and elevation of class \code{\link[stars:st_as_stars]{stars}}.
Two attributes in this object relates to global land cover.
The cover layer classifies the status of land cover of the whole globe into 20 categories, while
the cover_cls layer uses 8 simplified categories.
Percent Tree Cover (trees) represents the density of trees on the ground, and the last attribute represents elevation.
}
\details{
\strong{Important:} publication of these maps is only allowed when cited to Tateishi et al. (2014), and when "Geospatial Information Authority of Japan, Chiba University and collaborating organizations." is shown.
}
\references{
Production of Global Land Cover Data - GLCNMO2008, Tateishi, R., Thanh Hoan, N., Kobayashi, T., Alsaaideh, B., Tana, G., Xuan Phong, D. (2014), Journal of Geography and Geology, 6 (3).
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tm_layout.R
\name{tm_plot}
\alias{tm_plot}
\title{Plot mode options}
\usage{
tm_plot(use_gradient, limit_latitude_3857)
}
\arguments{
\item{use_gradient}{Use gradient fill using \link[grid:linearGradient]{linearGradient()}}

\item{limit_latitude_3857}{Vector of two limit latitude values for maps printed in Web Mercator projection (EPSG 3857). If \code{c(-90, 90)} the poles will be inflated too much. The Web Mercator is defines as \code{c(-85.06, 85.06)}, but the default setting in tmap is \code{c(-84, 84)}.}
}
\description{
Plot mode options. This option is specific to the plot mode.
}
\seealso{
\href{https://r-tmap.github.io/tmap/articles/basics_modes}{Modes}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tm_facets.R
\name{tm_facets}
\alias{tm_facets}
\alias{tm_facets_grid}
\alias{tm_facets_wrap}
\alias{tm_facets_pagewise}
\alias{tm_facets_stack}
\alias{tm_facets_hstack}
\alias{tm_facets_vstack}
\alias{tm_facets_flip}
\title{Specify facets}
\usage{
tm_facets(
  by = NULL,
  rows = NULL,
  columns = NULL,
  pages = NULL,
  as.layers = FALSE,
  nrow = NA,
  ncol = NA,
  byrow = TRUE,
  orientation = NA,
  free.coords = NA,
  drop.units = TRUE,
  drop.empty.facets = TRUE,
  drop.NA.facets = FALSE,
  sync = TRUE,
  swipe = FALSE,
  na.text = NA,
  scale.factor = 2,
  type = NA,
  free.scales = NULL,
  ...
)

tm_facets_grid(rows = NULL, columns = NULL, pages = NULL, ...)

tm_facets_wrap(by = "VARS__", nrow = NA, ncol = NA, byrow = TRUE, ...)

tm_facets_pagewise(by = "VARS__", byrow = TRUE, ...)

tm_facets_stack(by = "VARS__", orientation = NA, ...)

tm_facets_hstack(by = "VARS__", ...)

tm_facets_vstack(by = "VARS__", ...)

tm_facets_flip(...)
}
\arguments{
\item{by}{Group by variable (only for a facet wrap or facet stack)}

\item{rows}{Variable that specifies the rows (only for a facet grid)}

\item{columns}{Variable that specifies the columns (only for a facet grid)}

\item{pages}{Variable that specifies the pages (only for a facet grid)}

\item{as.layers}{show facets as layers?}

\item{nrow}{Number of rows}

\item{ncol}{Number of columns}

\item{byrow}{Should facets be wrapped by row?}

\item{orientation}{For facet stack: horizontal or vertical?}

\item{free.coords}{Logical. If the \code{by} argument is specified, should each
map has its own coordinate ranges? By default \code{TRUE}, unless facets are shown
in as different layers (\code{as.layers = TRUE})}

\item{drop.units}{Logical. If the \code{by} argument is specified, should
non-selected spatial units be dropped? If \code{FALSE}, they are plotted where
mapped aesthetics are regarded as missing values. Not applicable for
raster shapes. By default \code{TRUE}.}

\item{drop.empty.facets}{Logical. If the \code{by} argument is specified, should
empty facets be dropped? Empty facets occur when the \code{by}-variable contains
unused levels. When \code{TRUE} and two \code{by}-variables are specified, empty rows
and columns are dropped.}

\item{drop.NA.facets}{Logical. If the \code{by} argument is specified, and all
data values for specific facets are missing, should these facets be dropped?
\code{FALSE} by default. In v3, it was called \code{showNA}.}

\item{sync}{Logical. Should the navigation in view mode (zooming and panning)
be synchronized? By default \code{TRUE} if the facets have the same bounding box.
This is generally the case when rasters are plotted, or when \code{free.coords} is
\code{FALSE}.}

\item{swipe}{Logical. If \code{TRUE}, the two maps are overlaid with a draggable divider to reveal either map. Requires exactly two maps, and only works in the \code{"mapbox"} and \code{"maplibre"} modes provided by the \pkg{tmap.mapgl} package. Default is \code{FALSE}.}

\item{na.text}{Text used for facets of missing values. In v3, it was \code{textNA}.}

\item{scale.factor}{Number that determines how the elements (e.g. font sizes,
symbol sizes, line widths) of the small multiples are scaled in relation to
the scaling factor of the shapes. The elements are scaled to the \code{scale.factor}th
root of the scaling factor of the shapes. So, for \code{scale.factor = 1},
they are scaled proportional to the scaling of the shapes. Since elements,
especially text, are often too small to read, a higher value is recommended.
By default, \code{scale.factor = 2}.}

\item{type}{\code{"grid"}, \code{"wrap"} or \code{"stack"}}

\item{free.scales}{deprecated. Please use the \code{.free} arguments in the layer functions, e.g. \code{fill.free} in \code{tm_polygons}.}

\item{...}{passed on to \code{tm_facets()}}
}
\description{
\itemize{
\item \code{tm_facets_wrap()} specify facets for one grouping variable (so one faceting dimension)
\item \verb{tm_facets_(hv)stack()} stacks the facets either horizontally or vertically (one grouping variable).
\item \code{tm_facets_grid()} specify facets for two grouping variables in a grid of rows and columns.
\item \code{tm_facets_pagewise()} same as wrap, but the facets are drawn on different plots (pages). Replaces the \code{along} argument from version 3.
\item \code{tm_facets()} is the core function, but it is recommended to use the other functions.
}
}
\note{
In older versions (< 4.1) \code{\link[=tm_facets]{tm_facets()}} with \code{page} specification was used to create animations frames and \code{\link[=tmap_animation]{tmap_animation()}} to create the animation itself using inputs like the frame rate specification. As of version 4.2, the whole animation, including frame rate, is specified in \code{\link[=tm_animate]{tm_animate()}}. The animation can still be saved via \code{\link[=tmap_animation]{tmap_animation()}}.
}
\examples{
\dontrun{
tm_shape(NLD_dist) +
  tm_polygons("edu_appl_sci",
    fill.scale = tm_scale_intervals(values = "pu_gn", style = "kmeans", n = 7)) +
  tm_facets(by = "province") +
tm_shape(NLD_muni) +
  tm_borders(lwd = 3) +
  tm_facets(by = "province") +
tm_title("Population with a univeristy degree (incl appl. sciences), percentages")

tm_shape(World) +
  tm_polygons(c("gender", "press"),
    fill.scale = list(tm_scale_intervals(values = "bu_br_div", midpoint = 0.5),
      tm_scale_intervals(values = "pu_gn_div", midpoint = 50)),
    fill.legend = tm_legend("")) +
tm_layout(panel.labels = c("Gender Inequality Index (GII)", "World Press Freedom Index"))
}
}
\seealso{
\code{\link[=tm_animate]{tm_animate()}}

\href{https://r-tmap.github.io/tmap/articles/basics_facets}{Facets}, \href{https://r-tmap.github.io/tmap/articles/adv_animations}{Animations}, \href{https://r-tmap.github.io/tmap/articles/examples_gridmaps}{Gridmap}
}
